/* Copyright 2021 Kevin Ryde

   This file is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3, or (at your option) any later
   version.

   This file is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along
   with this file.  See the file COPYING.  If not, see
   <http://www.gnu.org/licenses/>.
*/

/* Kevin Ryde, May 2021

   Usage: ./a.out [-v] n n lo-hi

   Generate terms of A337114 by iterating through all free trees of n
   vertices with gentreeg.c, finding their automorphism orbits with Nauty,
   and counting the distinct orbit sizes vectors (partitions of n) which
   occur.  This is a brute-force approach, suitable only for small n.
   Created for use with Nauty 2.7r1.

   The default is to try all n from 1 to 32.  Individual n or ranges can be
   given on the command line

          ./a.out 2 5-7 13

   Output is like

       n=1  a(n) = 1
       n=2  a(n) = 1
       n=3  a(n) = 1
       n=4  a(n) = 2
       n=5  a(n) = 3

   Command-line option -v prints some more verbose information.

   Compiling
   ---------

   Put this file in a Nauty build directory which has had its "./configure"
   and "make nauty1.a", then compile to an executable a.out,

       gcc -O2 \
         -DOUTPROC=my_gentreeg_outproc -DGENTREEG_MAIN=my_gentreeg_main \
         A337114-nauty.c gentreeg.c nauty1.a

   If Nauty is installed in system locations (such as Debian or similar),
   then #include "nauty.h" below must be #include <nauty/nauty.h>, and find
   where gentreeg.c is kept (perhaps /usr/share/nauty/), otherwise get a
   copy from the Nauty sources,

       gcc -O2 \
         -DOUTPROC=my_gentreeg_outproc -DGENTREEG_MAIN=my_gentreeg_main \
         A337114-nauty.c /usr/share/nauty/gentreeg.c -lnauty1

   The -D defines tell gentreeg.c to be its "calling from a program" style
   rather than the gentreeg program as such (see comments in gentreeg.c).

   The code here uses 1-word graphs so that the "nauty1.a" version of the
   library suffices, and is a little faster than the general "nauty.a".
   1-word graphs limit to n <= WORDSIZE, where WORDSIZE is either 32 or 64
   (bits).  More than 32 is expected to be too slow anyway.  (To attempt
   more, increase SEEN_SIZE defined below, and on a 32-bit machine compile
   with -DWORDSIZE=64 and link to nautyL1.a.)

   Notes
   -----

   Nauty's densenauty() gives automorphism orbits as an array orbits[v] = u
   which means vertex v is in the same orbit as vertex u, and u is the
   smallest vertex in that orbit.  Here this is turned into the size of each
   orbit (number of vertices in it), and those orbit sizes are sorted.
   Duplicate orbit sizes vectors are detected by building a seen[] tree.

   An orbit sizes vector is a partition of n (sum of sizes is n) so the
   number of vectors is at most the number of partitions of n (A000041).
   The number of partitions grows much slower than the number of free trees,
   which allows the seen[] storage to be a relatively modest size.
   Not every partition occurs, but that doesn't reduce the final seen[] size
   by very much, in the form constructed here.
*/

/* Change this to WANT_ASSERTS 1 to run the assert() statements for various
   consistency checks.  */
#define WANT_ASSERTS  1

/* Whether include \r progress messages in -v option. */
#define WANT_PROGRESS  1

/* Change this to WANT_DEBUG 1 for mucho debug prints. */
#define WANT_DEBUG  0


/* The code here is intended for 1-word graph storage and operations, so set
   MAXN accordingly.  Do this before nauty.h so as to get from it some
   1-word-only versions of macros.
*/
#define MAXN  WORDSIZE

#if ! WANT_ASSERTS
#define NDEBUG  /* no assert() from assert.h */
#endif

#include <assert.h>
#include <stdio.h>
#include <time.h>
#include "nauty.h"

#if WANT_DEBUG
#define DEBUG(expr)  expr
#else
#define DEBUG(expr)
#endif

int option_verbose = 0;

/* my_gentreeg_main() is in gentreeg.c, when that file is compiled with
   -DGENTREEG_MAIN=my_gentreeg_main */
int my_gentreeg_main(int argc, char**argv);

#define MIN(x,y)  ((x)<(y) ? (x) : (y))


/* ------------------------------------------------------------------------- */
/* Progress messages are printed roughly every PROGRESS_INTERVAL seconds.
   num_free_trees[n] is the total number of rooted trees of n
   vertices, which is used to show progress as a percentage of total work.

   The test for whether or not to print a message is based on a
   progress_remaining counter, so as to avoid the slowdown of say a time()
   system call for every forest.
*/

/* A000055, num_free_trees[n] = number of free trees of n vertices */
const double num_free_trees[] = {
  1.0, 1.0, 1.0, 1.0, 2.0, 3.0, 6.0, 11.0, 23.0, 47.0, 106.0, 235.0,
  551.0, 1301.0, 3159.0, 7741.0, 19320.0, 48629.0, 123867.0, 317955.0,
  823065.0, 2144505.0, 5623756.0, 14828074.0, 39299897.0, 104636890.0,
  279793450.0, 751065460.0, 2023443032.0, 5469566585.0, 14830871802.0,
  40330829030.0, 109972410221.0, 300628862480.0, 823779631721.0,
  2262366343746.0, 6226306037178.0
};
#define NUM_FREE_TREES_LEN \
  (sizeof(num_free_trees) / sizeof(num_free_trees[0]))
#define PROGRESS_INTERVAL  5  /* seconds between prints */
time_t progress_lasttime;
int progress_block = 100;  /* how many trees between progress prints */
int progress_remaining;    /* how many remaining to the next print */

#define PROGRESS(n,trees_count)                                         \
  do {                                                                  \
    if (WANT_PROGRESS && option_verbose && --progress_remaining < 0)    \
      progress_print(n,trees_count);                                    \
  } while (0)

void
progress_print (int n, double trees_count)
{
  if (n < NUM_FREE_TREES_LEN && num_free_trees[n])
    printf("n=%d, %.0lf trees, %.1lf%%\r",
           n, trees_count,
           100.0 * trees_count / num_free_trees[n]);
  else
    printf("n=%d, %.0lf trees\r", n, trees_count);

  {
    /* The previous progress_block many trees took elapsed_time to run.
       Recalculate progress_block to be about 5 seconds worth of trees.
       But try to beware of clock jumps or process stops
     */
    time_t t = time(NULL);
    double elapsed_time = abs(t - progress_lasttime);
    if (elapsed_time) {
      int per_second = progress_block / elapsed_time;
      progress_block = MIN(PROGRESS_INTERVAL*per_second+1,
                           2*progress_block+1);
    } else {
      /* took 0 seconds for the last block, try doubling */
      progress_block = 2*progress_block + 1;
    }
    if (progress_block < 0) progress_block = 100;
    progress_remaining = progress_block;
    progress_lasttime = t;
  }
}

/* ------------------------------------------------------------------------- */
/* seen[] is a search tree recording the partitions of n which have been
   seen as orbit sizes vectors.

   Each node in this search tree is a vector of pointers t[].  Orbit sizes
   are each size>=1 and each descends to sub-node t[size-1], successively.
   If t[size-1]==NULL then there's no such node yet, so this is a new orbit
   sizes vector, and new nodes are created for its descents.

   SEEN_SIZE must be big enough for all the nodes created.  For n vertices,
   an upper bound is SEEN_SIZE <= A026905(n) = the cumulative partition
   numbers.  Eg. for n=32 at most A026905(32) = 43819.

   If every partition of n occurred then seen[] space used would be exactly
   A026905(n).  Not every partition occurs, but this seems to reduce the
   space needed by only a small amount.
*/
#define SEEN_SIZE  65536
typedef void *POINTER;
POINTER seen[SEEN_SIZE];
POINTER *seen_end;  /* pointer into seen[], immediately after used space */

void
init_seen (int n)
{
  memset(seen, '\0', sizeof(seen));  /* all NULLs */
  seen_end = seen + n;   /* after first node of sizes 1..n */
}

/* orbitsizes[0 .. num_orbits-1] is a partition of n, so that sum
   orbitsizes[0] + ... + orbitsizes[num_orbits-1] == n, with terms in
   descending order so that orbitsizes[i] >= orbitsizes[i+1].
   If orbitsizes[] is not already in seen[], then add it and return 1.
   If orbitsizes[] is already in seen[], then do nothing and return 0.
*/
int
add_seen (int n, const int *orbitsizes, int num_orbits)
{
  int is_new = 0;
  POINTER *t = seen;
  int i;
  for (i = 0; i < num_orbits; i++) {
    int size = orbitsizes[i];
    assert (n > 0);
    assert (t != NULL);
    assert (1 <= size && size <= n);
    assert (i==0 || orbitsizes[i-1] >= orbitsizes[i]);  /* decreasing */

    n -= size;  /* new remaining total in orbitsizes[] */
    assert (n >= 0);

    if (t[size-1]) {
      t = t[size-1];   /* go to existing node */

    } else {
      /* New node_size is at most as big as "size" again since sizes are in
         descending order, and also at most the remaining "n", so whichever
         is smaller.
      */
      int node_size = MIN (n, size);
      DEBUG(printf("  seen insert i=%d size=%d at seen_end=%p\n", i, size, seen_end));

      if (seen_end >= seen + SEEN_SIZE - node_size) {
        fprintf(stderr, "oops, exceeded SEEN_SIZE\n");
        exit(1);
      }

      is_new = 1;
      t = t[size-1] = seen_end;
      seen_end += node_size;
    }
  }
  assert (n == 0);  /* after subtracting all orbit sizes */
  return (is_new);
}

/* ------------------------------------------------------------------------- */

/* vpar[1..n] is a rooted tree of vertices v=1..n where vpar[v] = parent of v
   or vpar[v]=0 if v is the root of the tree.
   Convert vpar[] to Nauty graph g,m,n, with vertices mapped from 1..n in
   vpar to instead 0..n-1 in g.
   The root in vpar is not distinguished in g, so vpar treated as a free tree.
   g must have space for n vertices.
   Any existing contents in g are replaced.
   Arcs are added both directions between v and its parent vpar[v], so an
   undirected graph g.
*/
void
vpar_to_graph (const int *vpar, graph *g, int m, int n)
{
  int v;
  assert (n >= 0);
  assert (m >= SETWORDSNEEDED(n));

  EMPTYGRAPH(g, m,n);
  for (v = 1; v <= n; v++) {
    int p = vpar[v];  /* parent of v */
    assert (0 <= p && p <= n);
    if (p) {
      int v0 = v-1;
      int p0 = p-1;
      ADDONEEDGE(g, v0,p0, m);
    }
  }
}

/* xp and yp are pointers to ints */
int
compare_int_descending (const void *xp, const void *yp)
{
  return (*(const int *)yp - *(const int *)xp);
}

DEFAULTOPTIONS_GRAPH(options);
double trees_count;  /* number of trees generated */
int num_seen;        /* number of distinct partitions seen */

/* callback from gentreeg */
void
my_gentreeg_outproc (FILE *outfile, int *vpar, int n)
{
  const int m = SETWORDSNEEDED(MAXN);
  setword g[MAXN*SETWORDSNEEDED(MAXN)];
  setword g_out[MAXN*SETWORDSNEEDED(MAXN)];
  int lab[MAXN];
  int ptn[MAXN];
  int orbits[MAXN];
  int orbitsizes[MAXN];
  int v;
  statsblk stats;

  DEBUG(printf("my_gentreeg_outproc() n=%d\n", n));
  trees_count++;
  PROGRESS(n,trees_count);
  assert(n <= MAXN);
  assert(m >= SETWORDSNEEDED(n));

  vpar_to_graph(vpar, g,m,n);
  densenauty ((graph*)g, lab, ptn, orbits, &options, &stats, m,n, g_out);

  DEBUG(printf("  orbits ");
        for(v=0; v<n; v++) printf(" %d", orbits[v]);
        printf("\n"));

  memset (orbitsizes, '\0', n*sizeof(orbitsizes[0]));
  for (v = 0; v < n; v++) {
    int u = orbits[v];  /* u is smallest vertex in the orbit of v */
    assert (0 <= u && u < n);
    orbitsizes[u]++;    /* how many in the orbit of u */
  }

  DEBUG(printf("  orbitsizes ");
        for(v=0; v<n; v++) printf(" %d", (int) orbitsizes[v]);
        printf("\n"));

  /* sort all n, so as to push 0s to the end */
  qsort (orbitsizes, n, sizeof(orbitsizes[0]), compare_int_descending);

  DEBUG(printf("  sorted     ");
        for(v=0; v<n; v++) printf(" %d", (int) orbitsizes[v]);
        printf("\n"));
#if WANT_ASSERTS
  /* stats.numorbits many non-0, so rest of orbitsizes[] all 0 */
  for(v=stats.numorbits; v<n; v++) assert(orbitsizes[v]==0);
#endif

  num_seen += add_seen(n, orbitsizes, stats.numorbits);
}

/* calculate and print A337114(n) */
void
calculate_n (int n)
{
  int gentreeg_argc;
  char *gentreeg_argv[10];
  char n_str[20];

  if (! (1 <= n && n <= MAXN)) {
    fprintf(stderr, "n out of range, must be 1 to %d, got %d\n", MAXN, n);
    exit(1);
  }
  {
    int m = SETWORDSNEEDED(n);
    nauty_check(WORDSIZE,m,n,NAUTYVERSIONID);
  }

  init_seen(n);
  num_seen = 0;
  trees_count = 0;

  if (option_verbose) {
    printf("gentreeg  n=%d\n", n);
  }
  sprintf(n_str, "%d", n);

  gentreeg_argc = 0;
  gentreeg_argv[gentreeg_argc++] = "gentreeg"; /* progname */
  gentreeg_argv[gentreeg_argc++] = "-q";       /* quiet */
  gentreeg_argv[gentreeg_argc++] = n_str;
  gentreeg_argv[gentreeg_argc] = NULL;
  my_gentreeg_main (gentreeg_argc, gentreeg_argv);

  if (option_verbose) {
    printf("n=%d end, tree count %.0lf, seen array %u used, a(%d) = %d\n",
           n, trees_count, seen_end - seen, n, num_seen);
    printf("\n");

  } else {
    printf("n=%d  a(n) = %d                     \n", n, num_seen);
  }
}

int
main (int argc, char **argv)
{
  int seen_n = 0;
  int i;
  setbuf(stdout,NULL);

  for (i = 1; i < argc; i++) {
    const char *arg = argv[i];
    int n, len;
    unsigned lo, hi;
    if (strcmp(arg,"-v") == 0) {
      option_verbose = 1;
      printf("NAUTYVERSION is %s\n", NAUTYVERSION);
      printf("SEEN_SIZE %u is %u bytes\n", SEEN_SIZE, sizeof(seen));
      printf("\n");
    } else if (sscanf(arg, "%d%n", &n, &len) == 1
               && len == strlen(arg)) {
      seen_n = 1;
      calculate_n(n);
    } else if (sscanf(arg, "%u-%u%n", &lo, &hi, &len) == 2
               && len == strlen(arg)) {
      seen_n = 1;
      for (n = lo; n <= hi; n++)
        calculate_n(n);
    } else {
      fprintf(stderr, "Unrecognised command line option: %s\n", arg);
      exit(1);
    }
  }

  if (!seen_n) {
    int n;
    for (n = 1; n <= MAXN; n++) {
      calculate_n(n);
    }
  }
  return 0;
}
