\\ Copyright 2019 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.

default(recover,0)
default(strictargs,1)
read("../vpar.gp");
read("../vpar-wip.gp");
read("../test-helpers.gp");


\\ Gang Li, "Generation of Rooted Trees and Free Trees", M.S. Thesis,
\\ Dept. of Computer Science, University of Victoria, Canada
\\ http://webhome.cs.uvic.ca/~ruskey/Theses/GangLiMScThesis.pdf
\\
\\ Gang Li and Frank Ruskey, "The Advantages of Forward Thinking in
\\ Generating Rooted and Free Trees", SODA 99.
\\ http://webhome.cs.uvic.ca/~ruskey/Publications/RootedFreeTree/RootedFreeTree.html
\\
\\ Li's thesis and Li and Ruskey's paper give recursive algorithms for
\\ generating unlabelled trees and free trees.
\\
\\ An attraction of their approach is that trees are formed by successively
\\ adding a new leaf below the existing tree.  If some property of the tree
\\ so far is not to your liking, then an immediate return can skip the
\\ generation of the rest, effectively pruning off that generating.  They
\\ give for example variants which limit the height of the trees made, or
\\ limit the number of children a vertex can have.


\\-----------------------------------------------------------------------------
\\ Unlabelled Trees

\\ Gen() is the algorithm of Li and Ruskey's paper, adapted from the Pascal
\\ there.  It generates trees (stored to a Gen_output list) in the same
\\ order and manner of Beyer and Hedetniemi (here vpar_unlabelled_next()),
\\ but recursively rather than iteratively.
\\
\\ p=next vertex to store, s=root of a subtree being copied, cL=size of s.
\\ Make trees of n vertices by
\\     vpar = vector(n)    \\ global variable (or local())
\\     Gen(1,0,0).
\\
Gen(p,s,cL) =
{
  if(p>#vpar, listput(Gen_output,vpar); return);

  if (cL==0,        vpar[p]=p-1,
      vpar[p-cL]<s, vpar[p] = vpar[s],
                    vpar[p] = cL + vpar[p-cL]);
  Gen(p+1,s,cL);
  while(vpar[p]>1,
        s=vpar[p];
        vpar[p]=vpar[s];
        Gen(p+1,s,p-s));
}

{
  \\ compare Gen() against vpar_unlabelled_next()
  for(n=1,8,
     Gen_output = List([]);
     local(vpar=vector(n));
     Gen(1,0,0);

     #Gen_output == vpar_count_unlabelled(n) || error();
     my(vpar=vpar_unlabelled_first(n),
        i=0);
     until((vpar=vpar_unlabelled_next(vpar))=='none,
           Gen_output[i++] == vpar || error()));
}



vector(10,n, Gen_output=List([]); local(vpar=vector(n)); Gen(1,0,0); #Gen_output)

n=4;
Gen_output = List([]);
vpar=vector(n); Gen(1,0,0);
print(Gen_output);
print(vpar_free_all(n));
check_equal(Vec(Gen_output), vpar_unlabelled_all(n));


\\-----------------------------------------------------------------------------

expand(p,h,n,N) =
{
  print("expand "p" "h" "n" "N);
  if(N-p >= h,
     Gen(p+1, 2, p-1, h, n, N, 1, 0));
  if((p-1)*2 >= N
     || (p-h-1==1 && par[p] > 2)
     || (p-h-1>=2 && (par[h+2]>2 || (par[h+3]>2))),
     Gen(p+1, 2, p-2, h, n, N, 1, 1));
}

Gen(p,s,cL,h,l,n,f,g) =
{
  print("Gen "p" "s" "cL" "h" "l" "n" "f" "g);
  if(p > n,
      if(f==0, expand(p-1,h,n,N),
         listput(Gen_output,vpar));
     return);

  if(cL==0 && p <= N/2 +2,  vpar[p] = p-1,
     vpar[p-cL] < s,        vpar[p] = vpar[p-cL],
     
     vpar[p] = cL + vpar[p-cL];
     if(g==1,
        if(   (l-1)*2 < n
           && (p-cL+1<l)
           && ((vpar[p-cL+1]==2
                && p-cL+2<=l
                && vpar[p-cL+2]==2 
           || (p-cL+1==l && vpar[p-cL+1]==2)
           || p-cL+1>l,
           s = vpar[p]; cL = p-s; vpar[p] = vpar[vpar[p]];
           ,
           vpar[p-cL]==2, vpar[p]=1)));

  Gen( p+1, s, cL, h, l, n, f, g );

  while(vpar[p] > 2 && (f==0 || p>l+h-g),
        if(s==0, h = p-2);
        s = vpar[p]; vpar[p] = vpar[s];
        if(f==0,
           Gen( p+1, s, p-s, h, 0, N-h+1, f, g ),
           Gen( p+1, s, p-s, h, l,    n,   f, g )));
  if(f==0, expand( p-1, h, p-1, N ));
}

N=4;
Gen_output = List([]);
vpar=vector(N); vpar[2]=1; Gen(3,0,0, N\2, 0, (N+3)\2, 0,0);
print(Gen_output);
print(vpar_free_all(N));
check_equal(Vec(Gen_output), vpar_unlabelled_all(N));
