\\ Copyright 2022 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file.  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.

\\-----------------------------------------------------------------------------

\\ Kevin Ryde, July 2022
\\
\\ Usage: gp <a111299.gp
\\
\\ This is some PARI/GP code to calculate terms of A111299, the
\\ Matula-Goebel numbers of rooted binary trees (vertices 0 or 2 children).
\\
\\ The strategy is to assemble trees as a root above any two previously
\\ calculated trees, up to a given maximum resulting tree number.

default(recover,0);
default(strictargs,1);

log_lower_bound(x) = logint(floor(x),2) * 11/16;

prime_lower_bound(n) =
{
  if(n<100000, prime(n),
     my(L = log_lower_bound(n));
     n*(L + log_lower_bound(L) - 3/2) + 1);
}

\\ Return a vector of all terms of A111299 with values <= limiit.
a_vector(limit) =
{
  my(ret=List([1]),
     pr=List([2]));  \\ prime(ret[i]) or <0 lower bound
  my(i=0);
  while(i++ <= #ret,
        for(j=1,i,
           \\ pr[i] and pr[j] only if needed
           foreach(vecsort([1,i,j],e->ret[e]),k,
              abs(pr[i])*abs(pr[j]) <= limit || next(2);
              if(pr[k]<0, pr[k]=prime(ret[k])));

           my(t = pr[i]*pr[j]);
           t<=limit || next;
           listput(ret,t);
           listput(pr, - prime_lower_bound(t))));
  vecsort(Vec(ret)[^1]);
}

{
  my(v=a_vector(5000));
  print1("A111299 = ");
  for(n=1,#v, print1(v[n],", "));
  print("...");
}

\\ If x and y are terms of the sequence, and with 1 reckoned also in the
\\ sequence, then
\\ 
\\     prime(x) * prime(y)
\\     
\\ is also a term.
\\
\\ Terms are generated in this way by running x and y through the previously
\\ generated terms, and in a "triangular" way so
\\
\\ The "limit" is enforced by
\\ 
\\     prime(x) * prime(y)  <=  limit
\\ 
\\ But circa PARI version 2.13, prime() becomes slow when x (or y) is beyond
\\ the prime table size.  So certain lower bounds are tried so as to avoid
\\ large prime(),
\\
\\     prime_lower_bound(x) * prime_lower_bound(y)  <=  limit
\\
\\ quickly excludes terms which are too big, and then a second check
\\
\\     prime(x) * prime_lower_bound(y)  <=  limit
\\
\\ where x is the smaaller of x,y.
\\ 
\\ In the code, primes are calcualted on an as-needed basis pr[n] = prime(n)
\\ if has been needed, otherwise pr[n] = - prime_lower_bound(y).
\\ Hence use of abs(pr[n]) for either the exact prime or the lower bound.
\\
\\ In a 32-bit bulid, again circa PARI 2.13, the prime table is limited to
\\ 4 billion and with that terms here can be found up to about 5 billion
\\ before unfeasibly slow.
\\
\\
\\ Prime Lower Bound
\\ -----------------
\\ 
\\ The prime lower bound used is per Rosser and Schoenfeld as in A000040.
\\
\\     n*(log n + log log n - 3/2) < prime(n),  for n >= 2
\\     
\\ For convenient exact calculation, decrease the logs to
\\
\\     log_lower_bound(x) = floor(log_2(x)) * 16/11
\\                        <= log(x)


\\-----------------------------------------------------------------------------
print("end");
