\\ Copyright 2021 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file.  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.

default(recover,0);
default(strictargs,1);


\\ Usage; gp <a193405.gp
\\
\\ Find terms of A193405 = Matula-Goebel numbers of trees with a perfect
\\ matching.


\\------------------------------------------------------------------------------
\\ Predicate
\\
\\ A tree with a perfect matching has the root matched to one of its
\\ children, and any number of perfect matched subtrees below that pair.
\\
\\      R           root R and child U pair
\\    / | \         each P = perfect matched subtree
\\   P  P  U             
\\  /   |  | \
\\         P  P
\\         |  |
\\
\\ The U subtree (U and everything below it) is near perfect except the root
\\ unmatched.  R has exactly one U child subtree.
\\
\\ A tree with a perfect matching can be identified by working up from the
\\ childless vertices and at a vertex v establish a "state" either P or U,
\\
\\     state   type 
\\     -----   ----
\\       0      P     have perfect matching, everything matched
\\       1      U     near-perfect matching in which the root is unmatched
\\
\\ When a given vertex v has a child subtree c added below it, the state
\\ becomes
\\
\\     new = v - c           accumulate c under v
\\
\\ If new = -1 then the tree has no perfect matching, because it means new v
\\ has two or more subtrees of type U.
\\
\\
\\ This type of accumulation of states is a generic approach.  It can be
\\ applied starting at leaves of a tree and working inwards to any desired
\\ final vertex -- which in the code here is the root.
\\
\\ The leaves as starting points are singleton vertices (nothing below) and
\\ are of the U type
\\ 
\\     state = 0        at singleton
\\
\\ In the code, f[i,2] is the number of copies of the child subtree c so
\\ subtract c*f[i,2] from v for those copies.  Any consequent v<0 means no
\\ perfect matching.


\\ n is a Matula-Goebel tree number.
\\ Return  0 if n has a perfect matching
\\         1 if n has a near-perfect matching where the root is unmatched
\\        -1 if n does not have a perfect matching, and can never have a
\\           perfect matching no matter what is added at the root
A193405_perfect_matching_state(n) =
{
  my(f=factor(n),
     v=1);  \\ state for a singleton
  for(i=1,matsize(f)[1],
     my(c=A193405_perfect_matching_state(primepi(f[i,1])));  \\ child
     if(c<0, return(-1),
        c,   v -= f[i,2];  \\ repetitions of child c==1
             if(v<0, return(-1))));
  v;
}

is_A193405(n) = A193405_perfect_matching_state(n)==0;

{
  print("A193405 perfect matching tree codes by is_A193405():");
  print1("  ");
  for(n=1,305,
     if(is_A193405(n),
        print1(n", ");
        if(n==93||n==202, print1("\n  "))));
  print("...\n");
}
{
  print("A193406 not-perfect-matching tree codes:");
  print1("  ");
  for(n=1,67,
     if(! is_A193405(n),
        print1(n", ");
        if(n==27||n==47, print1("\n  "))));
  print("...\n");
}


\\-----------------------------------------------------------------------------
\\ The above predicate tests a single treecode by factoring.  For treecodes
\\ in a given range, forfactored() is more efficient than individual
\\ factoring.  P and U trees are both rather rare, so a Map() suits for
\\ recording those of each type, rather than for instance a vector of states
\\ 0,1,-1 (which would be overwhelmingly -1).

\\ return a vector of terms of A193405 which are <= limit
A193405_vector_by_forfactored(limit) =
{
  my(P=Map(),U=Map());
  forfactored(nf=1,limit,
     my(n,f); [n,f]=nf;
     my(v=1);      \\ state 1 root as singleton
     for(i=1,matsize(f)[1],
        my(b=primepi(f[i,1]));
        if(mapisdefined(U,b),  if((v-=f[i,2])<0, next(2)),
           !mapisdefined(P,b), next(2)));
     
     if(v, mapput(U,n,1),
           mapput(P,n,1)));
  P=Vec(Mat(P));
  if(#P,P[1]~,P);
}

{
  \\ verify against is_A193405()
  my(limit=10000);
  my(by_forfactored=A193405_vector_by_forfactored(limit),
     by_predicate=select(is_A193405,[1..limit]));
  by_forfactored == by_predicate || error();
}

{
  \\ This is a time comparison.  Change to "if(1," to try it, and increase
  \\ or decrease "limit" for the size of interest, or to enough to see a
  \\ difference on a fast or slow computer.
  \\ Expect A193405_vector_by_forfactored() about 3* faster.
  if(1,
     my(limit=500000);
     \\ factor() and forfactored() both use the primes table do they?
     \\ Try running once as a warm-up (so the one or the other doesn't
     \\ suffer from growing the table).
     forfactored(nf=1,limit, );

     gettime();
     my(v=A193405_vector_by_forfactored(limit));
     print("A193405_vector_by_forfactored() time ",gettime());

     gettime();
     my(l=List([])); for(n=1,limit,if(is_A193405(n), listput(l,n)));
     l=Vec(l);
     print("is_A193405() time                    ",gettime());
     v == l || error();
     );
}

\\-----------------------------------------------------------------------------
print("end");
