/* Copyright 2021 Kevin Ryde

   This file is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3, or (at your option) any later
   version.

   This file is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along
   with this file.  See the file COPYING.  If not, see
   <http://www.gnu.org/licenses/>.
*/

/* Kevin Ryde, May 2021

   Usage: ./a.out [-v]

   Generate terms of A347414 by iterating through all rooted trees of n
   vertices and finding their automorphism orbits with Nauty.  This is a
   brute-force approach, suitable only for small n.

   The sizes of the automorphism orbits sum to the number of vertices n (so
   a partition of n).  The output is the number of distinct partitions of n
   occurring and is printed in b-file style "n a(n)",

       1 1
       2 1
       3 1
       4 2
       5 3

   Command-line option -v prints some extra information, including elapsed
   time.

   Made for use with Nauty 2.7.

   Compiling
   ---------

   Put this file in a Nauty build directory which has had its "./configure"
   and "make nauty.a", then compile to an executable a.out,

       gcc -O2 A347414-nauty.c nauty.a

   If Nauty is installed in system locations (such in Debian or similar),
   then #include "nauty.h" below must be #include <nauty/nauty.h> and

       gcc -O2 A347414-nauty.c -lnauty

   Nauty has some library MAXN and WORDSIZE variations.  MAXN=32 suits here,
   which is library nautyW1.a instead of nauty.a.  On a 64-bit machine,
   maybe WORDSIZE=64 suits memory access and manipulations, which is
   nautyL1.a instead.

   nautyL1.a for WORDSIZE=64 is also MAXN=64, but the code here is likely to
   be too slow to go beyond n=32 vertices.  (To try more, increase the
   #define MAXN 32 below, check SEEN_SIZE is enough, and perhaps start
   the loop in main() at a desired next n.)

   Notes
   -----

   Nauty's densenauty() gives automorphism orbits as an array orbits[v] = u
   which means vertex v is in the same orbit as vertex u, and u is the
   smallest vertex in that orbit.  Here this is turned into the size of each
   orbit (number of vertices), and those orbit sizes are sorted.

   Duplicate orbit sizes vectors are detected by building a seen[] search
   tree.  Each search node is a row[] vector of pointers.  Orbit size s
   descends to sub-node row[s].  If row[s]==NULL then there's no such node
   yet so this orbit sizes vector is new and new search nodes are created
   for its descents.

   The number of distinct orbit sizes vectors is at most the number of
   partitions of n (A000041).  This grows much slower than the number of
   rooted trees, and in particular allows seen[] to be a relatively modest
   size.  Not every partition of n occurs as orbit sizes, but that seems to
   reduce the search tree size only by a small amount.

   Speed
   -----

   Most of the time is spent in densenauty().  The roots of the forest are
   distinguished from other vertices by an initial partition lab,ptn.  An
   alternative would be to make a directed graph with each edge directed
   from child to parent, but Nauty circa its version 2.7 seems much faster
   with an initial partition.
*/

#define MAXN  32

/* Change this to WANT_ASSERTS 1 to run the assert() statements for various
   slow-ish consistency checks.  */
#define WANT_ASSERTS  0

/* Change this to WANT_DEBUG 1 for mucho debug prints. */
#define WANT_DEBUG  0


#if ! WANT_ASSERTS
#if ! defined(NDEBUG)  /* if NDEBUG not already set */
#define NDEBUG
#endif
#endif

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include "nauty.h"

#if WANT_DEBUG
#define DEBUG(expr)  expr
#else
#define DEBUG(expr)
#endif

/* Function my_gentreeg_main() is in gentreeg.c (when that file is compiled
   with -DGENTREEG_MAIN=my_gentreeg_main) */
int my_gentreeg_main(int argc, char**argv);

#define MIN(x,y)  ((x)<(y) ? (x) : (y))

void
vpar_rooted_forest_first (int *vpar, int n)
{
  int i;
  vpar[0]=0;
  for (i=1; i<=n; i++) vpar[i] = i-1;
}
int
vpar_rooted_forest_next (int *vpar, int n)
{
  DEBUG(printf("  vpar_rooted_forest_next()\n"));

  int p, q, len;
  for (p=n; ; p--) {
    if (p==0) return(0);
    if ((q=vpar[p]) != 0) break;
  }
  q = vpar[p];
  len = p-q;
  DEBUG(printf("    p=%d q=%d len=%d\n", p,q,len));
  for (;;) {
    int i;
    vpar[p] = vpar[q];
    DEBUG(printf("    common parent set p=%d to %d\n", p, vpar[p]));
    p++; if (p>n) return(1);
    for (i=1; i<len; i++) {
      vpar[p] = vpar[p-len] + len;
      DEBUG(printf("    replicate set p=%d to %d+%d = %d\n",
                   p, vpar[p-len], len, vpar[p]));
      p++; if (p>n) return(1);
    }
  }
}

/* Convert tree vpar[1..n] to Nauty graph g,m,n.
   g must have space for n vertices.
   Any existing contents of g are overwritten.
   Edges are both ways between child c and parent p, so an undirected graph.
*/
void
vpar_to_graph (const int *vpar, graph *g, int m, int n)
{
  int v;
  assert (n >= 0);
  assert (m >= SETWORDSNEEDED(n));

  EMPTYGRAPH(g, m,n);
  for (v = 1; v <= n; v++) {
    int p = vpar[v];
    assert (0 <= p && p <= n);
    if (p) {
      int v0 = v-1;
      int p0 = p-1;
      ADDONEEDGE(g, v0,p0, m);  /* up */
    }
  }
}

/* xp and yp are pointers to ints */
int
compare_int_descending (const void *xp, const void *yp)
{
  return ((int) * (const int *) yp - (int) * (const int *) xp);
}

/* SEEN_SIZE upper bound for n vertices is A026905(n) = cumulative partition
   numbers.  Eg. for n=32 this is A026905(32) = 43819.  This would be the
   exact size if every partition of n was orbit sizes, but some are not so
   actual size is a little smaller.
*/
#define SEEN_SIZE  65536
typedef void *POINTER;
POINTER seen[SEEN_SIZE];
POINTER *seen_end;
int num_seen;  /* number of distinct partitions seen */

static DEFAULTOPTIONS_GRAPH(options);
int m;       /* graph number of words for n, for usual g,m,n */
unsigned long long tree_count;


int
add_seen (int n, const int *orbitsizes, int num_orbits)
{
  int is_new = 0;
  POINTER *t = seen;
  int i;
  for (i = 0; i < num_orbits; i++) {
    int size = orbitsizes[i];
    assert (n > 0);
    assert (t != NULL);
    assert (1 <= size && size <= n);
    assert (i==0 || orbitsizes[i-1] >= orbitsizes[i]);  /* decreasing */

    n -= size;  /* new remaining total in orbitsizes[] */
    assert (n >= 0);

    if (t[size-1]) {
      t = t[size-1];   /* go to existing node */
      DEBUG(printf("    existing i=%d size=%d at t=%p\n", i, size, t));

    } else {
      /* new node_size is at most as big as "size" again (sizes decreasing),
         but is also at most the remaining "n", so whichever is smaller */
      int node_size = MIN (n, size);
      DEBUG(printf("    insert i=%d size=%d at seen_end=%p\n", i, size, seen_end));

      if (seen_end >= (POINTER*)seen + SEEN_SIZE - node_size) {
        fprintf(stderr, "oops, exceeded SEEN_SIZE\n");
        exit(1);
      }

      is_new = 1;
      t = t[size-1] = seen_end;
      memset(seen_end, '\0', node_size*sizeof(POINTER));
      seen_end += node_size;
    }
  }
  assert (n == 0);
  DEBUG(if (is_new) printf("    new partition\n"));
  return (is_new);
}

/* Fill lab[] and ptn[] for v to be a single distinguished vertex for Nauty.
   so here v then the rest: lab[0..n-1] = v, 0,1,...,v-1, v+1,...,n-1.
   ptn[] is filled with 1s except the last of each colour class in lab[] has
   ptn=0, so in this case ptn[0..n-1] = 0, 1,...,1, 0.
*/
void
lab_ptn_vpar_roots (const int *vpar, int n, int *lab, int *ptn)
{
  int v, num_roots = 0;
  for (v=1; v<=n; v++) num_roots += (vpar[v]==0);
  {
    /* lab[] is a list of the vertices 0..n-1 with those of each colour
       contiguous.  Here this is the roots first and the rest after.
    */
    int rpos = 0;
    int opos = num_roots;
    for (v=1; v<=n; v++) {
      if (vpar[v]) {
        lab[opos++] = v-1;
      } else {
        lab[rpos++] = v-1;
      }
    }
  }
  /* ptn[] is all 1s except a 0 marker at the last of each colour class in
     lab[], which in this case means at num_roots-1 and the end-most n-1.
   */
  for (v=0; v<n; v++) ptn[v] = 1;
  ptn[num_roots-1] = 0;
  ptn[n-1] = 0;
}


/* callback from gentreeg */
void
process_vpar (const int *vpar, int n)
{
  setword g[MAXN*SETWORDSNEEDED(MAXN)];
  setword g_out[MAXN*SETWORDSNEEDED(MAXN)];
  int lab[MAXN];
  int ptn[MAXN];
  int orbits[MAXN];
  int orbitsizes[MAXN];
  int v;
  statsblk stats;

  DEBUG(printf("process_vpar() n=%d\n", n));
  assert(n <= MAXN);                  /* array size here */
  assert(m <= SETWORDSNEEDED(MAXN));  /* array size here */
  assert(m >= SETWORDSNEEDED(n));
  tree_count++;
  DEBUG(printf("  vpar ");
        for(v=1; v<=n; v++) printf(" %d", vpar[v]);
        printf("\n"));

  vpar_to_graph(vpar, g,m,n);
  lab_ptn_vpar_roots (vpar,n, lab,ptn);
  densenauty ((graph*)g, lab, ptn, orbits, &options, &stats, m,n, g_out);

  memset (orbitsizes, '\0', n*sizeof(orbitsizes[0]));
  for (v = 0; v < n; v++) {
    int u = orbits[v];  /* smallest vertex in the orbit of v */
    assert (0 <= u && u < n);
    orbitsizes[u]++;    /* how many in the orbit of u */
  }

  DEBUG(printf("    orbitsizes ");
        for(v=0; v<n; v++) printf(" %d", (int) orbitsizes[v]);
        printf("\n"));

  qsort (orbitsizes, n, sizeof(orbitsizes[0]), compare_int_descending);

  DEBUG(printf("    sorted     ");
        for(v=0; v<n; v++) printf(" %d", (int) orbitsizes[v]);
        printf("\n"));
#if WANT_ASSERTS
  /* stats.numorbits many non-0, rest of orbitsizes[] all 0 */
  for(v=stats.numorbits; v<n; v++) assert(orbitsizes[v]==0);
#endif

  num_seen += add_seen(n, orbitsizes, stats.numorbits);
}

int
main (int argc, char **argv)
{
  int verbose = 0;
  int n;

  setbuf(stdout,NULL);
  options.defaultptn = 0;

  while (*++argv) {
    if (strcmp(*argv,"-v") == 0) {
      verbose = 1;
    } else {
      fprintf(stderr, "unrecognised command line option: %s\n", *argv);
      exit(1);
    }
  }

  if (verbose) {
    printf("SEEN_SIZE %u is %u bytes\n", SEEN_SIZE, sizeof(seen));
  }

  for (n = 1; n <= 20; n++) {
    time_t start_time = time(NULL);
    int vpar[MAXN+1];
    if (verbose) {
      printf("trees  n=%d\n", n);
    }

    m = SETWORDSNEEDED(n);
    num_seen = 0;
    memset(seen, '\0', n*sizeof(seen[0]));  /* top search tree node */
    seen_end = ((POINTER*) seen) + n;
    tree_count = 0;

    vpar_rooted_forest_first(vpar,n);
    do {
      process_vpar(vpar,n);
    } while (vpar_rooted_forest_next(vpar,n));

    if (verbose) {
      time_t end_time = time(NULL);
      time_t elapsed_time = end_time - start_time;
      int seconds = elapsed_time % 60;
      int minutes = (elapsed_time/60) % 60;
      int hours   = elapsed_time /60 / 60;

      printf("n=%d end, rooted tree count %Lu, seen array %u used, a(n) = %d\n",
             n, tree_count, seen_end - (POINTER*)seen, num_seen);
      printf("  elapsed time %ld seconds = %d:%02d:%02d\n",
             (long) elapsed_time, hours, minutes, seconds);
      printf("\n");

    } else {
      printf("%d %d\n", n, num_seen);
    }
  }

  return 0;
}
