\\ Copyright 2021 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file.  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.


\\-----------------------------------------------------------------------------

\\ Kevin Ryde, September 2021
\\
\\ Usage; gp <a347967.gp
\\
\\ Calculate terms of A347967(n) = number of maximum matchings in the tree
\\ of Matula-Goebel number n.

default(recover,0)
default(strictargs,1)


\\ A maximum matching is a matching which is the maximum possible size in
\\ the tree.  (That size is the matching number A206483(n), which is not
\\ calculated.)
\\
\\ Maximum matchings in a tree can be counted by calculating 3 quantities at
\\ each vertex v, representing matchings in the subtree comprising v and
\\ everything below v (all its descendants).
\\
\\     with = how many maximum matchings where v matched.
\\
\\     without = how many maximum matchings where v not matched.
\\
\\     without_is_same
\\       = 1 when the "with" matchings are the same size as the "without"
\\         matchings, or when with=0 so no "with" matchings at all
\\
\\       = 0 when one or more "with" matchings exist and the "without"
\\         matchings are 1 pair smaller than the "with" matchings
\\
\\ "without" matchings always exist so they always have a size to consider.
\\
\\ Then
\\
\\     total = / with + without     if without_is_same = 1
\\             \ with               if without_is_same = 0
\\
\\ When "without" matchings are 1 pair smaller than "with" matchings, those
\\ withouts are not part of the total.  But they're relevant within the
\\ calculation since pairing the root with a new unmatched vertex, such as a
\\ singleton child, makes more matchings of size equal to the "with" size.
\\
\\ Any "with" matchings smaller than the maximum "with" are ignored and
\\ likewise any "without" smaller than the maximum "without" are ignored.
\\ This is since if some matching were constructed from a size smaller than
\\ the respective maxima then the result is smaller than (hence unwanted)
\\ the same construction starting from the maxima.
\\
\\
\\ For a new child subtree c added underneath vertex v, the quantities from
\\ the c subtree are accumulated into the quantities at v.
\\
\\ New "without" is simply multiply
\\
\\     new_without = v_without * c_total
\\
\\ These are the combinations of v_without and c_total, which are the only
\\ ways to leave new v unmatched.  The matchings in the c part are all the
\\ same size, ie. the "with" size, hence c_total.
\\
\\ New "with" combinations can be formed in the following ways
\\
\\     W0 = v_with    * c_with
\\     W1 = v_with    * c_without
\\     W2 = v_without * c_without, by new pair v,c
\\
\\ The relative sizes of these ways, and consequently which are taken and
\\ which ignored, depend on without_is_same in v and c.
\\
\\     Neither v_without_is_same nor c_without_is_same:
\\       W0 is bigger than the others and is the sole new "with".
\\
\\     v_without_is_same but not c_without_is_same:
\\       Only W0,W2 are taken since W1 is 1 smaller.
\\
\\     Not v_without_is_same but have c_without_is_same:
\\       All of W0,W1,W2 are the same size and are taken.
\\
\\     Both v_without_is_same and c_without_is_same:
\\       W2 is bigger than the others and is the sole new "with".
\\
\\ New "without_is_same" follows from these considerations too,
\\
\\     Both v_without_is_same and c_without_is_same:
\\       new_without_is_same = 0 since W2 is new "with" bigger than new
\\       "without".
\\
\\     Otherwise:
\\       new_without_is_same = v_without_is_same, no change.
\\
\\ In the code, without_is_same cases are treated by a sum
\\
\\     s = v_without_is_same + c_without_is_same
\\
\\ s=0 and s=1 want either W0 or W0+W1 and use of c_total includes or
\\ excludes W1 appropriately.  s=1 and s=2 want W2 also.
\\
\\
\\ This type of accumulation of quantities is a generic approach.  It can be
\\ applied starting at leaves of a tree and working inwards to any desired
\\ final vertex -- which in the code here is the root.
\\
\\ The leaves which are the starting points are singletons (nothing below)
\\ and their quantities are
\\
\\     with            = 0
\\     without         = 1
\\     without_is_same = 1
\\
\\ The code uses recursion to get child subtree quantities.  It's expected
\\ that maximum depth in the tree, and hence maximum recursion level, will
\\ be modest for a tree in the range which factor() and primepi() are able
\\ to decode.


\\ Return a 3-element vector [with, without, without_is_same] of quantities
\\ for tree n.
A347967_maximum_matchings_quantities(n) =
{
  my(with            = 0,  \\ initial quantities for a singleton
     without         = 1,
     without_is_same = 1,
     f=factor(n));
  for(i=1,matsize(f)[1],   \\ each child subtree
     my(c = A347967_maximum_matchings_quantities(primepi(f[i,1])),
        c_total = c[1] + if(c[3],c[2]));

     for(j=1,f[i,2],       \\ copies of the child subtree
        my(s = without_is_same + c[3]);
        with = if(s!=2, with * c_total)   \\ W0 and possibly W1
             + if(s!=0, without * c[2]);  \\ W2
        without *= c_total;
        if(s==2, without_is_same = 0)));

  [with, without, without_is_same];
}

A347967(n) =
{
  my(v=A347967_maximum_matchings_quantities(n));
  v[1] + if(v[3],v[2]);
}

{
  print("A347967 number of maximum matchings in tree number n:");
  print1("  ");
  for(n=1,35,
     print1(A347967(n),","));
  print("...");
}

print("end");
