\\ Copyright 2022 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file.  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.

\\-----------------------------------------------------------------------------
\\ Kevin Ryde, September 2022
\\
\\ Usage: gp <a356917.gp
\\
\\ This is some PARI/GP code to calculate a row of A356917, which is
\\ Colijn and Plazzotta's L_n list of subtrees in rooted binary tree
\\ (every vertex 0 or 2 children) number n of their enumeration.
\\
\\ Requires PARI/GP 2.13 or higher for sqrtint() remainder.

default(recover,0);
default(strictargs,1);

{
  \\ children(n), for n>=2, returns a vector [x,y] of the child subtrees
  \\ numbers of n, being x = A002024(n-1) and y = A002260(n-1)
  my(children(n) =
    my(r, s=sqrtint((n-1)<<1,&r));
    if(r>s,  r-=s;s++, r+=s);
    [s,r>>1]);

  \\ return row n of A356917 but terms not sorted
  my(row_unsorted(n) =
     if(n==1, [1],
        concat(concat(apply(self(), children(n))), n)));

\\ A356917
row(n) = vecsort(row_unsorted(n));
}

{
  print("A356917 rows:");
  for(n=1,10,
     my(v=row(n));
     printf("  n=%2d: ",n);
     for(k=1,#v,
        print1(v[k],","));
     print());
  print("  ...");
}

\\ row_unsorted() is a recursive traversal of tree n, concatenating subtree
\\ lists and n itself.
\\
\\ Peak recursion depth is modest since the smallest n requiring depth d is
\\ A108225(d) which is very large very quickly.
\\
\\ row_unsorted() gives subtree numbers as a post-order traversal,
\\ which is not numerical order, hence the final vecsort().
\\
\\ Maybe some "vecmerge" would suit sorted lists throughout, merged instead
\\ of concat().  vecsort() might even have such a merge internally.
\\ Suspect any speedup would be small.

print("end");
