\\ Copyright 2018, 2019 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.


\\ Graphs in the form of adjacency matrices.
\\
\\ g is an adjacency matrix representing a labelled undirected graph.
\\ Since undirected, g is symmetric g==g~.


default(recover,0)
default(strictargs,1)
read("../test-helpers.gp");


gradj_Seidel_adjacency_matrix(g) =
{
  my(n=matsize(g)[1]);
  matrix(n,n,i,j, if(i!=j, 1-(g[i,j]<<1)));
}
gradj_Seidel_adjacency_poly(g,var='x) = \
  charpoly(gradj_Seidel_adjacency_matrix(g),var);

gradj_relabel_perm(g,perm) =
{
  my(n=matsize(g)[1]);
  matrix(n,n,i,j, g[perm[i],perm[j]]);
}

\\ perm is a permutation.  Return 1 if it is an automorphism of g.
gradj_is_automorphism(g,perm) =
{
  my(n=matsize(g)[1]);
  for(v=1,n,
     vecextract(g[,perm[v]],perm) == g[,v] || return(0));
  1;
}
check_equal(gradj_relabel_perm([0,1,0; 1,0,0; 0,0,0], [3,2,1]), \
            [0,0,0; 0,0,1; 0,1,0]);

gradj_make_paths(vec,n=vecmax(apply(vecmax,vec))) =
{
  my(g=matrix(n,n));
  for(i=1,#vec,
     for(j=1,#vec[i]-1,
        g[vec[i][j],vec[i][j+1]] = 
        g[vec[i][j+1],vec[i][j]] = 1));
  g;
}

gradj_make_adjtriangle(vec,n=sqrtint(2*#vec)+1) =
{
  my(g=matrix(n,n), upto=0);
  for(i=1,n,
     for(j=i+1,n,
        g[i,j] = g[j,i] = vec[upto++]));
  g;
}
check_equal(gradj_make_adjtriangle([1]), [0,1; 1,0]);
check_equal(gradj_make_adjtriangle([1,0,0]), [0,1,0; 1,0,0; 0,0,0]);

gradj_automorphisms(g) =
{
  my(n=matsize(g)[1],
     l=List([]));
  for(i=0,n!-1,
     my(perm=numtoperm(n,i));
     if(gradj_is_automorphism(g,perm),
        listput(l,Vecsmall(perm))));
  Vec(l);
}
{
  my(g=gradj_make_paths([ [1,2], [3,4] ]));
  check_equal(g, [0,1,0,0; 1,0,0,0; 0,0,0,1; 0,0,1,0], "g");

  check_equal(gradj_is_automorphism(g,[3,4, 1,2]), 1);
  check_equal(gradj_is_automorphism(g,[3,2,1,4]), 0);
  check_equal(gradj_relabel_perm(g,[2, 3,4, 1]) != g,  1, "g perm diff");
  check_equal(gradj_is_automorphism(g,[2, 3,4, 1]), 0, "g not aut");

  check_equal(gradj_automorphisms(g),
              [Vecsmall([1,2,3,4]),
               Vecsmall([1,2,4,3]),
               Vecsmall([2,1,3,4]),
               Vecsmall([2,1,4,3]),

               Vecsmall([3,4,1,2]),
               Vecsmall([3,4,2,1]),
               Vecsmall([4,3,1,2]),
               Vecsmall([4,3,2,1])
              ]);
}

gradj_isomorphism_perm(g1,g2) =
{
  my(n=matsize(g1)[1]);
  for(i=1,n!-1,
     if(gradj_relabel_perm(g1,numtoperm(n,i)) == g2,
        return(perm)));
  'none;
}
gradj_is_isomorphic(g1,g2) = gradj_isomorphism_perm(g1,g2)!='none;

\\ Return the automorphisms of g which have v fixed (perm[v]==v).
gradj_automorphisms_fixing(g,v) =
{
  my(n=matsize(g)[1],
     l=List([]));
  for(i=0,n!-1,
     my(perm=numtoperm(n,i));
     if(perm[v]==v && gradj_is_automorphism(g,perm),
        listput(l,Vecsmall(perm))));
  Vec(l);
}

gradj_delete_vertex(g,v) = g[^v,^v];

gradj_is_semi_stable(g,v) =
{
  my(perms=gradj_automorphisms_fixing(g,v));
  perms=test_perms_delete_n(perms,v);
  perms == gradj_automorphisms(gradj_delete_vertex(g,v));
}

gradj_is_stabilizing_seq(g,seq) =
{
  my(n=matsize(g)[1]);
  for(i=1,#seq,
     my(v=seq[i]);
     \\ print("v="v" in "g);
     gradj_is_semi_stable(g,v) || return(0);
     g = gradj_delete_vertex(g,seq[i]);
     for(j=i+1,#seq, if(seq[j]>seq[i], seq[j]--)));
  1;
}

gradj_max_stablizing_seq(g) =
{
  my(n=matsize(g)[1],
     m=Vecsmall([]));  \\ max seq found so far, including initial v
  for(v=1,n,
     if(gradj_is_semi_stable(g,v),
        my(seq=gradj_max_stablizing_seq(gradj_delete_vertex(g,v)));
        if(#seq >= #m,
           for(i=1,#seq, if(seq[i]>=v, seq[i]++));
           m=concat(Vecsmall([v]), seq);
           if(#m==n, break))));
  m;        
}

\\-----------------------------------------------------------------------------


{
  if(0,
     gradj_graphs=vector(6);

     \\ A000088 number of graphs
     \\ 1,1,2,4,11,34
     for(n=1,#gradj_graphs,
        my(l=List([]));
        forvec(vec=vector(n*(n-1)/2,i, [0,1]),
              my(g=gradj_make_adjtriangle(vec));
              for(i=1,#l,
                 if(gradj_is_isomorphic(g,l[i]),
                    next(2)));
              listput(l,g));
        gradj_graphs[n] = Vec(l);
        \\ print1(#l", ");
     );
     print("graphs");
     print(gradj_graphs);
    );
}

{ 
  gradj_graphs = [[Mat(0)], [[0, 0; 0, 0], [0, 1; 1, 0]], [[0, 0, 0; 0, 0,
0; 0, 0, 0], [0, 0, 0; 0, 0, 1; 0, 1, 0], [0, 0, 1; 0, 0, 1; 1, 1, 0], [0,
1, 1; 1, 0, 1; 1, 1, 0]], [[0, 0, 0, 0; 0, 0, 0, 0; 0, 0, 0, 0; 0, 0, 0, 0],
[0, 0, 0, 0; 0, 0, 0, 0; 0, 0, 0, 1; 0, 0, 1, 0], [0, 0, 0, 0; 0, 0, 0, 1;
0, 0, 0, 1; 0, 1, 1, 0], [0, 0, 0, 0; 0, 0, 1, 1; 0, 1, 0, 1; 0, 1, 1, 0],
[0, 0, 0, 1; 0, 0, 0, 1; 0, 0, 0, 1; 1, 1, 1, 0], [0, 0, 0, 1; 0, 0, 1, 0;
0, 1, 0, 0; 1, 0, 0, 0], [0, 0, 0, 1; 0, 0, 1, 0; 0, 1, 0, 1; 1, 0, 1, 0],
[0, 0, 0, 1; 0, 0, 1, 1; 0, 1, 0, 1; 1, 1, 1, 0], [0, 0, 1, 1; 0, 0, 1, 1;
1, 1, 0, 0; 1, 1, 0, 0], [0, 0, 1, 1; 0, 0, 1, 1; 1, 1, 0, 1; 1, 1, 1, 0],
[0, 1, 1, 1; 1, 0, 1, 1; 1, 1, 0, 1; 1, 1, 1, 0]], [[0, 0, 0, 0, 0; 0, 0, 0,
0, 0; 0, 0, 0, 0, 0; 0, 0, 0, 0, 0; 0, 0, 0, 0, 0], [0, 0, 0, 0, 0; 0, 0, 0,
0, 0; 0, 0, 0, 0, 0; 0, 0, 0, 0, 1; 0, 0, 0, 1, 0], [0, 0, 0, 0, 0; 0, 0, 0,
0, 0; 0, 0, 0, 0, 1; 0, 0, 0, 0, 1; 0, 0, 1, 1, 0], [0, 0, 0, 0, 0; 0, 0, 0,
0, 0; 0, 0, 0, 1, 1; 0, 0, 1, 0, 1; 0, 0, 1, 1, 0], [0, 0, 0, 0, 0; 0, 0, 0,
0, 1; 0, 0, 0, 0, 1; 0, 0, 0, 0, 1; 0, 1, 1, 1, 0], [0, 0, 0, 0, 0; 0, 0, 0,
0, 1; 0, 0, 0, 1, 0; 0, 0, 1, 0, 0; 0, 1, 0, 0, 0], [0, 0, 0, 0, 0; 0, 0, 0,
0, 1; 0, 0, 0, 1, 0; 0, 0, 1, 0, 1; 0, 1, 0, 1, 0], [0, 0, 0, 0, 0; 0, 0, 0,
0, 1; 0, 0, 0, 1, 1; 0, 0, 1, 0, 1; 0, 1, 1, 1, 0], [0, 0, 0, 0, 0; 0, 0, 0,
1, 1; 0, 0, 0, 1, 1; 0, 1, 1, 0, 0; 0, 1, 1, 0, 0], [0, 0, 0, 0, 0; 0, 0, 0,
1, 1; 0, 0, 0, 1, 1; 0, 1, 1, 0, 1; 0, 1, 1, 1, 0], [0, 0, 0, 0, 0; 0, 0, 1,
1, 1; 0, 1, 0, 1, 1; 0, 1, 1, 0, 1; 0, 1, 1, 1, 0], [0, 0, 0, 0, 1; 0, 0, 0,
0, 1; 0, 0, 0, 0, 1; 0, 0, 0, 0, 1; 1, 1, 1, 1, 0], [0, 0, 0, 0, 1; 0, 0, 0,
0, 1; 0, 0, 0, 1, 0; 0, 0, 1, 0, 0; 1, 1, 0, 0, 0], [0, 0, 0, 0, 1; 0, 0, 0,
0, 1; 0, 0, 0, 1, 0; 0, 0, 1, 0, 1; 1, 1, 0, 1, 0], [0, 0, 0, 0, 1; 0, 0, 0,
0, 1; 0, 0, 0, 1, 1; 0, 0, 1, 0, 1; 1, 1, 1, 1, 0], [0, 0, 0, 0, 1; 0, 0, 0,
1, 0; 0, 0, 0, 1, 1; 0, 1, 1, 0, 0; 1, 0, 1, 0, 0], [0, 0, 0, 0, 1; 0, 0, 0,
1, 0; 0, 0, 0, 1, 1; 0, 1, 1, 0, 1; 1, 0, 1, 1, 0], [0, 0, 0, 0, 1; 0, 0, 0,
1, 1; 0, 0, 0, 1, 1; 0, 1, 1, 0, 0; 1, 1, 1, 0, 0], [0, 0, 0, 0, 1; 0, 0, 0,
1, 1; 0, 0, 0, 1, 1; 0, 1, 1, 0, 1; 1, 1, 1, 1, 0], [0, 0, 0, 0, 1; 0, 0, 1,
1, 0; 0, 1, 0, 1, 0; 0, 1, 1, 0, 0; 1, 0, 0, 0, 0], [0, 0, 0, 0, 1; 0, 0, 1,
1, 0; 0, 1, 0, 1, 0; 0, 1, 1, 0, 1; 1, 0, 0, 1, 0], [0, 0, 0, 0, 1; 0, 0, 1,
1, 0; 0, 1, 0, 1, 1; 0, 1, 1, 0, 1; 1, 0, 1, 1, 0], [0, 0, 0, 0, 1; 0, 0, 1,
1, 1; 0, 1, 0, 1, 1; 0, 1, 1, 0, 1; 1, 1, 1, 1, 0], [0, 0, 0, 1, 1; 0, 0, 0,
1, 1; 0, 0, 0, 1, 1; 1, 1, 1, 0, 0; 1, 1, 1, 0, 0], [0, 0, 0, 1, 1; 0, 0, 0,
1, 1; 0, 0, 0, 1, 1; 1, 1, 1, 0, 1; 1, 1, 1, 1, 0], [0, 0, 0, 1, 1; 0, 0, 1,
0, 1; 0, 1, 0, 0, 1; 1, 0, 0, 0, 1; 1, 1, 1, 1, 0], [0, 0, 0, 1, 1; 0, 0, 1,
0, 1; 0, 1, 0, 1, 0; 1, 0, 1, 0, 0; 1, 1, 0, 0, 0], [0, 0, 0, 1, 1; 0, 0, 1,
0, 1; 0, 1, 0, 1, 0; 1, 0, 1, 0, 1; 1, 1, 0, 1, 0], [0, 0, 0, 1, 1; 0, 0, 1,
0, 1; 0, 1, 0, 1, 1; 1, 0, 1, 0, 1; 1, 1, 1, 1, 0], [0, 0, 0, 1, 1; 0, 0, 1,
1, 1; 0, 1, 0, 1, 1; 1, 1, 1, 0, 0; 1, 1, 1, 0, 0], [0, 0, 0, 1, 1; 0, 0, 1,
1, 1; 0, 1, 0, 1, 1; 1, 1, 1, 0, 1; 1, 1, 1, 1, 0], [0, 0, 1, 1, 1; 0, 0, 1,
1, 1; 1, 1, 0, 0, 1; 1, 1, 0, 0, 1; 1, 1, 1, 1, 0], [0, 0, 1, 1, 1; 0, 0, 1,
1, 1; 1, 1, 0, 1, 1; 1, 1, 1, 0, 1; 1, 1, 1, 1, 0], [0, 1, 1, 1, 1; 1, 0, 1,
1, 1; 1, 1, 0, 1, 1; 1, 1, 1, 0, 1; 1, 1, 1, 1, 0]]];
}
print("graphs range ",#gradj_graphs," of n");

poly_num_distinct_roots(poly) =
{
  my(f=factor(poly));
  vecsum(apply(poldegree,f[,1]));
}
check_equal(poly_num_distinct_roots(x-1), 1);
check_equal(poly_num_distinct_roots((x-1)^2), 1);
check_equal(poly_num_distinct_roots(x^2-1), 2);

{
  \\ A002854 Seidel matrices order n
  for(n=1,#gradj_graphs,
     my(l=List([]));
     for(i=1,#gradj_graphs[n],
        my(g=gradj_graphs[n][i],
           poly=gradj_Seidel_adjacency_poly(g));
        if(poly_num_distinct_roots(poly)==3,
           listput(l,poly)));
     l=Set(l);
     print(#l));
  quit;
}
{
  \\ A002854 Seidel matrices order n
  for(n=1,#gradj_graphs,
     my(polys=apply(gradj_Seidel_adjacency_poly,gradj_graphs[n]));
     polys=Set(polys);
     print1(#polys,","));
  quit;
}

{
  \\ path-4 not semistable
  my(n=4,
     g=gradj_make_paths([ [1..n]  ]));
  print(vector(n,v, gradj_is_semi_stable(g,v)));
}
{
  my(n=9,
     g=gradj_make_paths([[7,6,1,2,3,9,8,4,5,1],
                         [3,8], [4,9] ]),
     seq=[7,1,6,3,2,4,5,9,8]);
  print(g);
  print(gradj_is_stabilizing_seq(g,seq));
  print(vector(n,v, gradj_is_semi_stable(g,v)));

  print();
  g=gradj_make_paths([[8,7,9],
                      [7,6,1,2,3,4,5,1],
                      [3,1,4] ]);
  seq=[8,9,7,6,1];
  print(g);
  print(gradj_is_stabilizing_seq(g,seq));
  print(vector(n,v, gradj_is_semi_stable(g,v)));
  print("max stabilizing ",gradj_max_stablizing_seq(g));
  quit;
}

\\ per Holton, path-4 not semi-stable 1--4--3--2
{
  for(n=1,#gradj_graphs,
     my(gs=gradj_graphs[n],
        count=0);
     for(i=1,#gs,
        my(g=gs[i],
           semi=0);
        for(v=1,n,
           if(gradj_is_semi_stable(g,v),
              semi=1;
              \\ print(g" semi-stable at v="v);
             ));
        if(!semi, print(g" not semi-stable"));
        count+=semi));
}

