/* Copyright 2018 Kevin Ryde

   This file is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3, or (at your option) any later
   version.

   This file is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along
   with this file.  See the file COPYING.  If not, see
   <http://www.gnu.org/licenses/>.
*/

/* Usage: vpar-to-graph6 [-v] [-zN] <integers >output.g6

   Read vpar from stdin in the form of integers and write graph6, sparse6 or
   digraph6 to stdout.

   A bit rough, but enough for testing (test-slow.gp).
*/


#define NAUTY_CPU_DEFINED 1  /* not CPUDEFS */
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <nauty/nauty.h>
#include <nauty/gutils.h>
#include <nauty/gtools.h>

const char *
output_format_name (int format)
{
  switch (format) {
  case GRAPH6:   return "GRAPH6";
  case SPARSE6:  return "SPARSE6";
  case DIGRAPH6: return "DIGRAPH6";
  default:       return "UNKNOWN";
  }
}

int
main (int argc, char *argv[])
{
  int output_format = GRAPH6;
  int output_header  = 0;
  int option_verbose = 0;
  int option_up      = 1;   /* both ways for undirected */
  int option_down    = 1;

  while (*++argv) {
    if (strcmp(*argv,"-g")==0) {
      output_format = GRAPH6;
    } else if (strcmp(*argv,"-s")==0) {
      output_format = SPARSE6;
    } else if (strcmp(*argv,"-z0")==0) {
      output_format = DIGRAPH6;
      option_up   = 1;
      option_down = 0;
    } else if (strcmp(*argv,"-z1")==0) {
      output_format = DIGRAPH6;
      option_up   = 0;
      option_down = 1;
    } else if (strcmp(*argv,"-z2")==0) {
      output_format = DIGRAPH6;
    } else if (strcmp(*argv,"-h")==0) {
      output_header = 1;
    } else if (strcmp(*argv,"-v")==0) {
      option_verbose = 1;
    } else {
      fprintf(stderr, "unrecognised option: %s\n", *argv);
      exit(1);
    }
  }
  if (option_verbose)
    fprintf(stderr, "output format %s up=%d down=%d\n",
            output_format_name(output_format), option_up, option_down);

  DYNALLSTAT(int,vpar,vpar_sz);
  int n;
  {
    int v = 1;
    int p;
    while (scanf(" %d",&p)==1) {
      DYNREALLOC(int,vpar,vpar_sz, v+1, "malloc");
      vpar[v++] = p;
    }
    n = v-1;
  }
  if (option_verbose)
    fprintf(stderr, "n=%d vertices\n", n);

  int m = SETWORDSNEEDED(n);
  DYNALLSTAT(graph,g, g_sz);
  if (n) {
    DYNALLOC2(graph,g, g_sz, m,n, "malloc");
    EMPTYGRAPH(g,m,n);
    int v;
    for (v = 1; v<=n; v++) {
      int p = vpar[v];
      if (p < 0 || p > n) {
        fprintf(stderr, "bad parent p=%d\n", p);
        exit(1);
      }
      if (p) {
        int from = v - 1;
        int to   = vpar[v] - 1;
        if (option_verbose)
          fprintf(stderr, " edge v=%d p=%d\n", v,vpar[v]);

        if (option_up) {     /* up    v->p */
          if (option_verbose)
            fprintf(stderr, "   up %d -> %d\n", from,to);
          ADDONEARC(g, from,to, m);
        }
        if (option_down) {   /* down  p->v */
          if (option_verbose)
            fprintf(stderr, "   down %d -> %d\n", to,from);
          ADDONEARC(g, to,from, m);
        }
      }
    }
  }

  if (option_verbose) {
    unsigned long edges;
    int  mindeg, mincount, maxdeg, maxcount, eulerian;
    degstats(g,m,n, &edges, &mindeg, &mincount, &maxdeg, &maxcount, &eulerian);
    fprintf(stderr, "degstats edges %lu\n", edges);
  }

  if (option_verbose) {
    int i;
    for (i=0; i<n; i++)
      fprintf(stderr, "g 0x%4X\n", g[i]);
  }

  FILE *fp = stdout;
  int i;
  for (i=0; i<2; i++) {
    if (i==1) {
      if (! option_verbose) break;
      fp = stderr;
    }
    switch (output_format) {
    case GRAPH6:
      if (output_header) fputs(GRAPH6_HEADER,fp);
      writeg6(fp,g,m,n);
      break;
    case SPARSE6:
      if (output_header) fputs(GRAPH6_HEADER,fp);
      writes6(fp,g,m,n);
      break;
    case DIGRAPH6:
      if (output_header) fputs(GRAPH6_HEADER,fp);
      writed6(fp,g,m,n);
      break;
    default:
      abort();
    }
  }
  if (ferror(stdout)) {
    fprintf(stderr, "error writing");
    exit(1);
  }

  return 0;
}
