\\ Copyright 2018, 2019, 2020 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.

default(recover,0);
default(strictargs,1)
read("../vpar.gp");


\\ Usage: gp <disjoint-domnum-most-sets.gp
\\
\\ This is a brute-force search for most pairs of sets of size the disjoint
\\ domination number in a tree.
\\
\\ As described in vpar_disjoint_domnum(), disjoint dominating sets are two
\\ dominating sets with no vertex in common.  The disjoint domination number
\\ is the smallest combined number of vertices in those two sets.
\\
\\ There can be various different pairs of disjoint dominating sets in a
\\ tree, including various different pairs of size the disjoint domnum.
\\ The aim here is to look for trees with the most such pairs of size the
\\ disjoint domnum.
\\
\\ For example, path-4 has 2 ways to make disjoint dominating sets of total
\\ size 4.
\\
\\     1----2----3----4         1----2----3----4
\\     A    B    A    B         A    B    B    A
\\
vpar_disjoint_domnum_and_count(vpar_make_path(4)) == [4,2]  || error();
\\
\\ The count of pairs of sets is just one way around, so not also the flips
\\ BABA, BAAB.
\\
\\ For n even <= 10, it can be noticed the number of set pairs is 2^(n/2-1),
\\ and the trees (there's only a few) are any tree n/2 with a new leaf hung
\\ off each existing vertex (corona of singletons).  The disjoint dominating
\\ sets are by taking one each either new leaf or existing vertex, giving
\\ 2^(n/2) possible assignments of vertex to set, and so 2^(n/2-1) set pairs
\\ taken just one way around A,B.
\\
\\ n=12 is the first even n with more than 2^(n/2-1) set pairs.  The sole
\\ most tree there has 33 sets of disjoint domnum size, as opposed to the
\\ coronas all having just 32.
\\
\\       *---*---*---*---*---*---*
\\               |       |             n=12                        
\\               *       *             33 disjoint domnum set pairs  
\\              / \      |              
\\             *   *     *
\\

limit = 14;
{
  for(n=2,limit,
     my(vpar=vpar_free_first(n),
        max_count=-1,
        max_vpars=List([]));
     until((vpar=vpar_free_next(vpar))=='none,
           my(count=vpar_disjoint_domnum_and_count(vpar)[2]);
           if(count > max_count, max_count=count; max_vpars=List([vpar]),
              count == max_count, listput(max_vpars,vpar)));
     print("n="n" max count "max_count" set pairs, occurs in "#max_vpars" tree(s)");
     for(i=1,#max_vpars,
        print("  "max_vpars[i]));
     print();
  );
}

\\ vpar_count_free(3) == 1
\\ vpar_count_free(4) == 2
\\ vpar_count_free(5) == 3

\\-----------------------------------------------------------------------------
