\\ Copyright 2017, 2018, 2019, 2020 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.

default(recover,0);
default(strictargs,1)
read("../vpar.gp");


\\ Usage: gp <maximal-asymmetric-subtrees.gp
\\
\\ Requires GP 2.9.x or higher.
\\
\\     Jaroslav Nesetril, "A Congruence Theorem For Asymmetric Trees",
\\     Pacific Journal of Mathematics, volume 37, number 3, March 1971,
\\     pages 771-778.  https://projecteuclid.org/euclid.pjm/1102970478
\\
\\ Nesetril shows that a free asymmetric tree is uniquely determined by the
\\ list of its maximal asymmetric subtrees with repetition.  And without
\\ repetition it is uniquely determined except for two n=9 trees which have
\\ the same set.
\\
\\ Here subtree of a free tree means any set of adjacent vertices, so that
\\ their induced subgraph is connected (a tree).  (This is not like the
\\ various "subtree" functions in vpar.gp where subtree mean everything "at
\\ and below" a vertex in a rooted tree.)
\\
\\ The code below finds asymmetric subtrees of a given vpar by pruning off
\\ leaf (degree=1) vertices and checking them vpar_is_free_asymmetric().
\\ If not asymmetric then further prunings of that subtree are considered.
\\
\\ The implementation is a little naive.  It could be worth watching for
\\ subtrees which are the same as already descended into, and just repeating
\\ the counts of what was found.  Eg. if a given vertex has multiple leaves
\\ hanging off it then deleting any one of them gives the same.  In general,
\\ among leaves take just one from each free automorphism orbit, since same
\\ orbit means same rest of the tree.  Also, vpar_is_free_asymmetric()
\\ testing asymmetric does almost the same work as vpar_to_free() which is
\\ desired immediately after, so ideally would make some combination.
\\


\\-----------------------------------------------------------------------------

\\ maximal_asymmetric_subtrees(vpar) returns a vector of trees
\\ [vpar1,vpar2,...] which are asymmetric free subtrees of the given vpar.
\\
\\ Each is an inclusion-wise maximal subtree, meaning that if vpar1 was
\\ vertex numbers v1,v2,... in vpar then no subset of those vertices is
\\ considered in the rest of the returned vpar2 etc.
\\
\\ Each vpar1 etc is vpar with some vertices deleted and then made free
\\ canonical per vpar_to_free().  The return vector is vecsort()ed so as to
\\ make a canonical list.
\\
\\ There can be duplicates among the vpar1 etc returned.  That occurs when
\\ the same asymmetric subtree occurs in different parts of the original
\\ vpar.  Those parts could be partially overlapping.
\\ Singletons and the empty tree are excluded from the return.
\\
maximal_asymmetric_subtrees(vpar) =
{
  my(ret=List([]),
     pending=List([ vpar ]));  \\ trees to consider reducing
  while(#pending,
        my(vpar=pending[#pending]); listpop(pending);

        \\ subtrees tested are to be >=7 vertices, since smallest asymmetric
        \\ free tree is 7 vertices (apart from singleton or empty)
        #vpar > 7 || next;

        my(degrees=vpar_degrees_vector(vpar));
        for(v=1,#vpar,
           degrees[v]==1 || next;  \\ delete leaves only
           my(sub=vpar_delete_vertex(vpar,v));
           if(vpar_is_free_asymmetric(sub),
              listput(ret,vpar_to_free(sub)), \\ asymmetric, return it
              listput(pending,sub))));      \\ symmetric, try subtrees

  vecsort(Vec(ret));
}

print("maximal_asymmetric_subtrees() is unique");
{
  my(m=Map());
  for(n=3,12,
     print("n="n);
     my(vpar=vpar_free_first(n));
     until((vpar=vpar_free_next(vpar))=='none,
           vpar_is_free_asymmetric(vpar) || next;
           my(subtrees = maximal_asymmetric_subtrees(vpar));
           if(mapisdefined(m,subtrees),
              error("oops, maximal asymmetric subtrees should be unique"));
           mapput(m,subtrees,1)));
}

maximal_asymmetric_set(vpar) = Set(maximal_asymmetric_subtrees(vpar));

\\----------
\\ The pair of n=9 with same set are
\\ Path 7 with Leaves 3,4
\\    https://hog.grinvin.org/ViewGraphInfo.action?id=31095
\\ Star Arms 1,3,4
\\    https://hog.grinvin.org/ViewGraphInfo.action?id=31097

print();
print("maximal_asymmetric_set() is unique except a pair of n=9");
{
  my(m=Map());
  for(n=3,12,
     print("n="n);
     my(vpar=vpar_free_first(n));
     until((vpar=vpar_free_next(vpar))=='none,
           vpar_is_free_asymmetric(vpar) || next;
           my(subtrees = maximal_asymmetric_set(vpar),
              prev);
           if(mapisdefined(m,subtrees,&prev),
              print(" prev="prev);
              print(" vpar="vpar);
              print(" both have set "subtrees));
           mapput(m,subtrees,vpar)));
}

\\ The set of subtrees which are common to the n=9 trees comprises two
\\ trees.  They are the sole n=7 and n=8 free asymmetric trees,
\\
\\ n=7 asymmetric tree, the smallest (excluding singleton or empty)
\\    https://hog.grinvin.org/ViewGraphInfo.action?id=934
\\ n=8 asymmetric tree, the second smallest
\\    https://hog.grinvin.org/ViewGraphInfo.action?id=31099
\\
vpar_count_free_asymmetric(7) == 1  || error();
vpar_count_free_asymmetric(8) == 1  || error();

\\-----------------------------------------------------------------------------
