#!/usr/bin/perl -w

# Copyright 2018 Kevin Ryde.
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
# 
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
# 
# You should have received a copy of the GNU General Public License along
# with this file; see the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.


use strict;
use FindBin;
use File::Slurp;
use File::stat;
use IPC::Run;

my $program_filename = File::Spec->catfile($FindBin::Bin,
                                           $FindBin::Script);
chdir '/tmp' or die;

# File::Slurp::write_file('favicon.tex',<<'HERE');
# \documentclass[a4paper]{article}
# \pagestyle{empty}
# \usepackage{tikz}
# \begin{document}
# \begin{tikzpicture}[scale=.8,inner sep=.2em,thick]
#   \node at (0,0) [] (2) {$2$};
#   \node at (-120:1) [] (1) {$1$};
#   \node at (-60:1) [] (3) {$3$};
#   \draw[->] (1)--(2);
#   \draw[->] (3)--(2);
# \end{tikzpicture}
# \end{document}
# HERE
# 
# File::Slurp::write_file('favicon.tex',<<'HERE');
# \documentclass[a4paper]{article}
# \pagestyle{empty}
# \usepackage{tikz}
# \usetikzlibrary{calc}
# \begin{document}
# \definecolor{mydarkblue}{rgb}{0,0,.55}
# \begin{tikzpicture}
#     [scale=.8,
#      color=mydarkblue,
#      my dot/.style={fill,circle,inner sep=.1em,outer sep=.2em,node contents=},
#      my edge/.style={line width=.3em},
#     ]
#   \node (root) at (0,0) [my dot];
#   \node (1) at (-120:1) [my dot];
#   \node (2) at (-60:1) [my dot];
#   \node (3) at (-120:2) [my dot];
#   \node (4) at ($(1.center) + (-60:1)$) [my dot];
#   \node (5) at (-60:2) [my dot];
#   \draw[my edge] (1)--(root);
#   \draw[my edge] (2)--(root);
#   \draw[my edge] (3)--(1);
#   \draw[my edge] (4)--(1);
#   \draw[my edge] (5)--(2);
# \end{tikzpicture}
# \end{document}
# HERE

File::Slurp::write_file('favicon.tex',<<'HERE');
\documentclass[a4paper]{article}
\pagestyle{empty}
\usepackage{tikz}
\usetikzlibrary{calc}
\begin{document}
\definecolor{mydarkblue}{rgb}{0,0,.6}
\begin{tikzpicture}
    [scale=.8,
     color=mydarkblue,
     my dot/.style={fill,circle,inner sep=.1em,outer sep=.2em,node contents=},
     my edge/.style={line width=.42em},
     line cap=round,
     line join=round,
    ]
  \coordinate (root) at (0,0) [my dot];
  \coordinate (1) at (-120:1) [my dot];
  \coordinate (2) at (-60:1) [my dot];
  \coordinate (3) at (-120:2) [my dot];
  \coordinate (4) at ($(1.center) + (-60:1)$) [my dot];
  \coordinate (5) at (-60:2) [my dot];
  \draw[my edge] (3)--(root)--(5);
  \draw[my edge] (4)--(1);
\end{tikzpicture}
\end{document}
HERE

IPC::Run::run(['latexmk','-gg','-pdf','favicon.tex'], '<',File::Spec->devnull)
  or die;

IPC::Run::run(['pdfcrop',
               '--verbose',
               '--margins','5',
               'favicon.pdf',
               'favicon-cropped.pdf'],
              '<', File::Spec->devnull)
  or die;
rename 'favicon-cropped.pdf', 'favicon.pdf' or die;

# IPC::Run::run(['xpdf','favicon.pdf']);

IPC::Run::run(['gs',
               '-q', # quiet, cf -dQUIET
               '-dNOPAUSE',
               '-g16x16',
               '-sDEVICE=pnggray',
               '-dPDFFitPage',
               # '-dEPSFitPage',
               '-dORIENT=0',
               "-dTextAlphaBits=4",     # grey scale edge blurring
               "-dGraphicsAlphaBits=4",
               "-sOutputFile=favicon.png",
               'favicon.pdf'],
              '<', File::Spec->devnull)
  or die;
print "\n";

IPC::Run::run(['convert','-strip','favicon.png','favicon-stripped.png']);
rename 'favicon-stripped.png', 'favicon.png' or die;

print "uncompressed ",-s 'favicon.png'," bytes\n";
# no saving ...
IPC::Run::run(['optipng','-quiet','-o5','favicon.png']);
print "compressed   ",-s 'favicon.png'," bytes\n";

# PNG spec 11.3.4.2 suggests RFC822 (or rather RFC1123) for CreationTime
use constant STRFTIME_FORMAT_RFC822 => '%a, %d %b %Y %H:%M:%S %z';
my $stat = File::stat::stat($program_filename)
  or die "Oops, cannot stat $program_filename: $!";
my $creation_time = POSIX::strftime(STRFTIME_FORMAT_RFC822,
                                    localtime($stat->mtime));

my $copyright = File::Slurp::read_file($program_filename);
$copyright =~ /^# (Copyright.*)/m or die;
$copyright = $1;
$copyright =~ s/\.$//;

pngtextadd('Author',    'Kevin Ryde');
pngtextadd('Generator', "tools/$FindBin::Script");
pngtextadd('Title',     'vpar.gp - icon');
pngtextadd('Creation Time', $creation_time);
pngtextadd('Copyright', $copyright);
pngtextadd('Homepage',  'http://user42.tuxfamily.org/pari-vpar/index.html');

print "with text    ",-s 'favicon.png'," bytes\n";

# IPC::Run::run(['xzgv','favicon.png']);

exit 0;

# add text to the png image in 'favicon.png'
sub pngtextadd {
  my ($keyword, $value) = @_;
  system('pngtextadd', "--keyword=$keyword", "--text=$value", 'favicon.png') == 0
    or die "system(pngtextadd)";
}
