\\ Copyright 2020, 2021, 2022 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file.  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.


\\ Tests related to Matula-Goebel primecode.

default(recover,0);
default(strictargs,1);

option_verbose=0;
option_progress=1;

read("memoize.gp");
read("../vpar.gp");
read("../vpar-wip.gp");
read("../test-helpers.gp");
read("OEIS-data.gp");
read("OEIS-data-wip.gp");

\\ not:
\\ A228599 graph obtained by applying Mycielski's construction etc
\\

\\-----------------------------------------------------------------------------
\\ Helpers

\\ vector of the prime factors of n,
\\ with multiplicity and in ascending order
prime_factors(n) =
{
  n>=1 || error();
  if(n==1, [],
     my(f=factor(n));
     concat(vector(matsize(f)[1],i, vector(f[i,2],j,f[i,1]))));
}
{
  check_equal(vector(500,n, vecprod(prime_factors(n))),
              vector(500,n, n));
  \\ A027746 has an extra row n=1 with 1
  OEIS_check_func("A027746",
                  n->if(n==1,[1],prime_factors(n)),
                  ['offset,1]);

  \\ prime factor indices
  OEIS_check_func("A112798",
    n->apply(primepi,prime_factors(n)));
}

\\ Return a vector of the prime factors of n, with multiplicity, in
\\ ascending order (A027746).
test_factor_to_primes(n) =
{
  if(n==1,[],
     my(f=factor(n));
     concat([vector(f[i,2],j,f[i,1]) |i<-[1..matsize(f)[1]]]));
}
{
  check_equal(test_factor_to_primes(12*7^4), [2,2,3,7,7,7,7]);
  check_equal(test_factor_to_primes(12), [2,2,3]);
  my(want=OEIS_data("A027746"),
     got=[]);
  for(n=1,oo, if(#got>=#want,break);
     got=concat(got, if(n==1,[1], test_factor_to_primes(n))));
  got=got[1..#want];
  check_equal(got,want, "A027746 test_factor_to_primes()");
}

distinct_prime_factors(n) =
{
  n>=1 || error();
  Vec(factor(n)[,1]);
}
A027748_distinct_prime_factors(n) =
{
  n>=1 || error();
  if(n==1,[1], distinct_prime_factors(n));
}
OEIS_check_rowfunc("A027748_distinct_prime_factors",,['offset,1]);


\\=============================================================================

\\-----------------------------------------------------------------------------
progress("A111299 Binary Trees (0 or 2 children)");
\\
\\ A111299 primecodes which are binary trees
\\ all vertices 0 or 2 children
\\ but not the singleton tree

\\ return 1 if vpar has all vertices with 0 or 2 children
test_is_binary_tree(vpar) = \
  test_set_is_subset(Set(vpar_num_children_vector(vpar)), [0,2]);

my(s=[0,2]); \
is_A111299_binary_tree_primecode(n) =
{
  if(n==1,1,
     my(f=factor(n));
     setsearch(s,vecsum(f[,2]))
     && self()(primepi(f[1,1]))
     && (matsize(f)[1]<2 || self()(primepi(f[2,1]))));
}

{
  my(is_by_vpar(n) = test_is_binary_tree(vpar_from_primecode(n)));
  check_equal(vpar_to_primecode([0]), 1);
  my(want=OEIS_data("A111299"));
  while(want[#want]>2000, want=want[^#want]);  \\ explicit search a bit slow
  check_equal(select(primecode->is_by_vpar(primecode),[2..want[#want]]),
              want,
              "A111299 primecodes which are binary trees");

  check_equal(vector(1000,n,is_A111299_binary_tree_primecode(n)),
              vector(1000,n,is_by_vpar(n)),
              "binary tree is_A111299_binary_tree_primecode()");
}

print("=========");
print("read a111299-Matula-binary-trees.gp");
read("a111299-Matula-binary-trees.gp");
print("=========");
A111299_vector = a_vector;
kill(a_vector);

{
  \\ prime_lower_bound()
  \\ A000040 for n >= 2, Rosser and Schoenfeld
  \\ n*(log n + log log n - 3/2) < prime(n)

  \\ drop to log3
  check_equal(vector(1000,n, logint(n,3) <= log(n)),
              vector(1000,n, 1));

  if(0,
     for(k=1,30,
        my(n=2^k,
           p=prime(n),
           lo = prime_lower_bound(n));
        print(lo" "p" "p/lo*1.0));
     quit);

  if(0,
     \\ log2(n) = log(n)/log(2)
     \\ log(n) = log2(n)*log(2)
     my(x=log(2), prev=0); 
     for(n=1,10000, 
        my(f=bestappr(x,n));
        if(f!=prev, printf("%20s   %.8f\n",f,f);prev=f));
     quit);

  if(0,
     \\ 186  to <=    100 000 000
     \\ 305  to <=    500 000 000
     \\ 369  to <=  1 000 000 000
     \\ 559  to <=  4 000 000 000
     \\ 605  to <=  5 000 000 000
     gettime();
     my(v=A111299_vector(5 000 000 000));
     print("time "gettime()" len "#v);
     quit);

  \\ log log <= 2 for n>=100000
  \\ vector(100,n,n++; log(log(n)))
  \\ my(n=200000); log(log(n))
  check_equal(vector(100,n,n+=99999; 2 <= log(log(n))),
              vector(100,n,n+=99999; 1));

  check_equal(vector(1000,n,
    n*(logint(n,3) - 3/2) + 1 <= prime(n)),
              vector(1000,n, 1));
}

{
  my(want=OEIS_data("A111299"),
     got=A111299_vector(want[#want]));
  print("A111299 DATA end ",want[#want]);
  check_equal(got, want,
              "A111299_vector() primecodes binary trees");

  if(1,
     my(want=OEIS_data("A111299",'bfile),
        got=A111299_vector(want[#want]));
     print("A111299 b-file end ",want[#want]);
     check_equal(got, want,
                 "A111299_vector() primecodes binary trees"));
}
{
  \\ 33033 bytes
  \\  count 1833 max prime 1469906485609 log 12
  if(0,
     my_prime_max=1;
     my(my_prime(n) = if(n<100000, prime(n),
        my(ret=floor(n*log(n)));
        my_prime_max = max(my_prime_max,n)));
     my(binary_trees_estimate(limit) =
        my(ret=List([1]),
           pr=List([2]));  \\ prime(ret[i]) or <0 lower bound
        my(i=0);
        while(i++ <= #ret,
              for(j=1,i,
                 foreach(vecsort([1,i,j],e->ret[e]),k,
                    abs(pr[i])*abs(pr[j]) <= limit || next(2);
                    if(pr[k]<0, pr[k]=my_prime(ret[k])));
                 
                 my(t = pr[i]*pr[j]);
                 t<=limit || next;
                 listput(ret,t);
                 listput(pr, - prime_lower_bound(t))));
        vecsort(Vec(ret)[^1]);
        );

     for(i=1,20,
        my(limit=10^i,
           got=binary_trees_estimate(limit));
        system("mkdir -p /tmp");
        system("rm -f /tmp/b999999.txt");
        for(n=1,#got,
           write("/tmp/b999999.txt", Str(n," ",got[n])));
        system("ls -l /tmp/b999999.txt");
        print("  count ",#got," max prime ",
              my_prime_max,
              " log ",
              logint(my_prime_max,10));
     );
     quit);
}

\\-----------------------------------------------------------------------------
progress("A348067 New Vertex Under Each Existing");

prime_estimate(n) = if(n<10 000 000,prime(n),ceil(n*log(n)));
A348067_estimate(n) = my(f=factor(n)); 2*factorback([prime_estimate(self()(primepi(p))) | p<-f[,1]], f[,2]);

A348067_new_child_each(n) =
{
  my(f=factor(n)); 2*factorback([prime(self()(primepi(p))) | p<-f[,1]], f[,2]);
}
\\ 26, 18, 202, 78, 122, 54, 338, 606, 2462, 234, 794, 366, 2626, 162
\\ vector(14,n,n+=2; A348067_new_child_each(n))

\\ return vpar with a childless vertex added underneath each existing vertex
test_add_childless_below_each(vpar) = concat(vpar,[1..#vpar]);

my(v=OEIS_data("A076146")); A076146(n) = v[n];  \\ OFFSET=1
{
  my(want=OEIS_data("A076146"));  \\ OFFSET=1
  check_equal(vector(#want,n, A076146(n)),
              want,
              "A076146 samples");

  \\ A076146 top-down binomial tree by copy as new child
  check_equal(vector(5,n, A076146(n)),
              vector(5,n, if(n==1,1, A076146(n-1)*prime(A076146(n-1)))));
}

\\ A033286 a(n) = n*prime(n)
\\ is Matula-Goebel copy tree as new child of root
A033286(n) = n*prime(n);
\\ 2, 6, 15, 28, 55, 78, 119, 152, 207, 290, 341, 444
{
  OEIS_check_func("A033286",,['offset,1]);

  check_equal(vector(100,n, A033286(n)),
              vector(100,n, my(vpar=vpar_from_primecode(n));
                     vpar_to_primecode(vpar_join(vpar,1,vpar))),
              "A033286() vs vpar_join()");
}

{
  my(a=A348067_new_child_each);
  check_equal(vector(500,n, a(n)),
              vector(500,n, my(p=test_factor_to_primes(n));
                     2*prod(i=1,#p, prime(a(primepi(p[i]))))),
              "A348067_new_child_each() by product");

  check_equal(vector(100,n, A348067_new_child_each(n)),
              vector(100,n, my(vpar=vpar_from_primecode(n));
                     vpar=test_add_childless_below_each(vpar);
                     vpar_validate(vpar);
                     vpar_to_primecode(vpar)),
              "A348067_new_child_each() vs vpar");

  check_equal(apply(p->a(p),                   primes(50)),
              apply(p->2*prime(a(primepi(p))), primes(50)),
              "A348067_new_child_each() on primes");

  \\ check_equal(matrix(5,5,m,n,m++;n++; a(m*n)),
  \\             matrix(5,5,m,n,m++;n++; a(m)*a(n)),
  \\             "A348067_new_child_each() multiplicative m,n >= 2");
  \\ not multiplicative due to the doubling for child at the root
  check_equal(a(2)*a(3), 156);
  check_equal(a(6),      78);
  check_equal(2*78,      156);

  my(a_nested(k) = my(ret=1); for(i=1,k, ret=a(ret)); ret);
  check_equal(a_nested(0), 1);
  check_equal(a_nested(1), a(1));
  for(k=0,20,
     my(got=a_nested(k));
     check_equal(got, A076146(k+1),
                 "A348067_new_child_each() k nesting gives A076146(k+1)");
     if(got>10^8,break));
  check_equal(A076146(1), 1);
  check_equal(A076146(2), a(1));
  check_equal(A076146(3), a(a(1)));
  check_equal(A076146(4), a(a(a(1))), "nest 3");


  if(0,
     \\ childess below each vertex, in numerical order
     \\ not in OEIS: 2, 6, 18, 26, 54, 78, 122, 162, 202, 234, 338, 366, 486, 502, 606
     my(vec=vector(1000,n,
                   my(vpar=test_add_childless_below_each(vpar_from_primecode(n)));
                   vpar_validate(vpar);
                   vpar_to_primecode(vpar)));
     print(Set(vec)[1..20]);
     quit);

  if(0,
     \\ child below each vertex or at root above so new root, in numerical order
     \\ apparently gives A202854 palindromic matchings:
     \\ 2, 5, 6, 18, 23, 26, 41, 54, 78, 103, 122, 162, 167, 202, 234, 283, 338, 366, 419, 486
     my(vec=vector(200,n,
                   my(vpar=test_add_childless_below_each(vpar_from_primecode(n)));
                   vpar_validate(vpar);
                   vpar_to_primecode(vpar)));
     \\ also reroot at leaf child of root
     vec=concat(vec, apply(n->n%2==0||error(); prime(n/2), vec));
     print(Set(vec)[1..20]);
     quit);

  \\ DATA section
  if(0,
     for(n=1,46,print1(A348067_new_child_each(n),",");if(n==19||n==33,print()));
     print();
     quit);

  \\ ~/OEIS/b348067.txt
  \\ n=1..5000 estimate max 97563682858
  \\ big    4451 13198442398
  \\ max    4805 97563682858
  if(0,
     print("primelimit is ",default(primelimit));
     gettime();
     system("rm -f /tmp/b348067.txt");
     my(m=0);
     for(n=1,5000,
        \\ my(t=A348067_estimate(n));
        my(t=a(n));
        if(t>m,print("new maximum ",m=t));
        write("/tmp/b348067.txt", Str(n," ",t)));
     system("ls -l /tmp/b348067.txt");
     print("maximum ",m);
     print("time taken ",gettime());
     quit);

  if(0,
     my(want=OEIS_data("A348067",'bfile)); \\ OFFSET=1
     if(#want>4000, want=Vec(want,4000));
     print("b348067.txt length ",#want);
     check_equal(vector(#want,n, A348067_new_child_each(n)),
                 want,
                 "A348067_new_child_each() b-file samples"));
}


\\---------------
\\ A348067/2 would be multiplicative a(p) = a(prime(2*primepi(p)))

\\ Evaluate func(n) for a completely multiplicative function of n.
\\ func() is called func(p) for primes p.
\\ n = p^e * q^f * ... is return func(p)^e * func(q)^f * ...
test_completely_multiplicative(n,func) =
{
  n>=0 || error();
  if(n==0,0,
     my(f=factor(n));
     factorback(apply(func,f[,1]), f[,2]));
}

half_A348067(n) = \
  test_completely_multiplicative(n, (p)->prime(2*half_A348067(primepi(p))));
\\ not in OEIS: 13, 9, 101, 39, 61, 27, 169, 303, 1231, 117, 397, 183, 1313, 81
\\ A348067 / 2
\\ vector(14,n,n+=2; half_A348067(n))
{
  \\ compact
  \\ my(a(n) = my(f=factor(n));
  \\    for(i=1,matsize(f)[1], f[i,1]=prime(2*self()(primepi(f[i,1]))));
  \\    factorback(f));
  my(a(n) = my(f=factor(n)); factorback([prime(2*self()(primepi(p))) | p<-f[,1]], f[,2]));

  check_equal(vector(500,n, a(n)),
              vector(500,n, half_A348067(n)));
  check_equal(vector(500,n, 2*half_A348067(n)),
              vector(500,n, A348067_new_child_each(n)));
}
{
  \\ DATA section
  if(0,
     for(n=1,48,print1(half_A348067(n),",");if(n==20||n==34,print()));
     print();
     quit);

  my(want=[
 1,3,13,9,101,39,61,27,169,303,1231,117,397,183,1313,81,673,507,251,909,
 793,3693,2273,351,10201,1191,2197,549,4457,3939,21977,243,16003,2019,
 6161,1521,1481,753,5161,2727,6089,2379,2473,11079,17069,6819,23603,1053
           ]);
  check_equal(vector(#want,n, half_A348067(n)), want,
              "half_A348067 samples");
}

{
  my(a=half_A348067);
  check_equal(vector(5,n,n--; a(n)),
              [0, 1, 3, 13, 9]);

  check_equal(matrix(100,100,s,t, a(s*t)),
              matrix(100,100,s,t, a(s)*a(t)),
              "half_A348067() completely multiplicative");

  my(want=OEIS_data("A076146"),
     t=1,
     got=List([]));
  while(1,
        listput(got,t);
        if(t>10^8,break);
        t = 2*a(t));
  check_equal(Vec(got), vector(#got,n, A076146(n)),
              "half_A348067 2*nesting gives A076146");
  check_equal(A076146(1), 1);
  check_equal(A076146(2), 2*a(1));
  check_equal(A076146(3), 2*a( 2*a(1) ));
  check_equal(A076146(4), 2*a( 2*a( 2*a(1) )), "nest 3");

  \\ apply(prime,OEIS_data("A076146")[1..5])
  \\ not in OEIS: 3, 13, 397, 362801
  my(t=2,
     got=List([]));
  while(t<10^5, listput(got,t); t=a(t));
  check_equal(Vec(got),
              vector(#got,n, prime(A076146(n))),
              "half_A348067 nesting gives prime(A076146)");
  check_equal(prime(A076146(1)), 2, "nest 0");
  check_equal(prime(A076146(2)), a(2), "nest 1");
  check_equal(prime(A076146(3)), a(a(2)));
  check_equal(prime(A076146(4)), a(a(a(2))));
}

test_non_roots(vpar) = \
  setminus([1..#vpar],vpar_roots(vpar));
test_add_child_below_except_root(vpar) = \
  concat(vpar,test_non_roots(vpar));
{
  check_equal(vector(500,n, half_A348067(n)),
              vector(500,n, vpar_to_primecode
                              (test_add_child_below_except_root
                                 (vpar_from_primecode(n)))));
}
is_half_A348067(n) =
{
  my(f=factor(n));
  foreach(f[,1],p, p=primepi(p); if(p%4!=2 || !self()(p/2), return(0))); 1;
}
check_equal(is_half_A348067(1),1, "1");
check_equal(is_half_A348067(2),0, "1");
check_equal(is_half_A348067(3),1, "3");
check_equal(primepi(3), 2);
check_equal(isprime(61), 1);
check_equal(primepi(61), 18);
{
  \\ sorted
  \\ not in OEIS: 1, 3, 9, 13, 27, 39, 61, 81, 101, 117, 169, 183, 243
  \\ print(select(is_half_A348067,[1..250]));
}
{
  my(want=Set(vector(1000,n, half_A348067(n)))[1..50]);
  check_equal(select(is_half_A348067, [1..want[#want]]),
              want);
}

\\-----------------------------------------------------------------------------
progress("Binomial Tree");

\\----------
\\ order n
\\ A076146 = primecode of binomial tree order k
\\         a(1)=1; a(n)=a(n-1)*prime(a(n-1))
A076146(n) =
{
  n>=1 || error();
  my(ret=1); for(i=2,n, ret=ret*prime(ret)); ret;
}
{
  \\ trim to range able to prime(n)
  my(want=OEIS_data("A076146"));
  while(want[#want]>10^9, want=want[^#want]);
  check_equal(vector(#want,n, A076146(n)),
              want,
              "A076146 samples");
  check_equal(vector(#want,n,n--; A076146(n+1)),
              vector(#want,n,n--; vpar_to_primecode(vpar_make_binomial_tree(2^n))),
              "A076146() vs vpar");
}

\\-------
\\ A666666  n vertices

print("=========");
print("read a666666-Matula-binomial-tree-n.gp");
read("a666666-Matula-binomial-tree-n.gp");
print("=========");
A666666_binomial_tree_n = a;
A666666_binomial_tree_vector = a_vector;
kill(a);
kill(a_vector);

\\  0
\\  | \  \
\\  1  2  4
\\     |  | \
\\     3  5  6
\\           |
\\           7

{
  \\ big numbers quickly
  my(a=A666666_binomial_tree_n,
     a_vector=A666666_binomial_tree_vector);
  check_equal(vector(32,n, a(n)),
              vector(32,n, vpar_to_primecode(vpar_make_binomial_tree(n))),
              "A666666_binomial_tree_n() vs vpar");

  \\ a_vector()
  check_equal(vector(10,n,#A666666_binomial_tree_vector(n)),
              vector(10,n, n));
  check_equal(a_vector(32),
              vector(32,n,A666666_binomial_tree_n(n)));

  \\ a(2^k) = A076146(k+1)
  \\ a(2^k+r) = A076146(k) * prime(a(r)) for 1 <= r < 2^k
  check_equal(vector(5,h,h--; vector(2^h,k,k--; A666666_binomial_tree_n(2^h+k))),
              vector(5,h,h--; vector(2^h,k,k--;
                                A076146(h+1)
                                * if(k==0,1, prime(A666666_binomial_tree_n(k))))),
              "A666666_binomial_tree_n() high bit");

  \\ binomial tree n vertices
  \\ print(vector(10,n,n++; A666666_binomial_tree_n(n))); quit;
  \\ not in OEIS: 2, 4, 6, 12, 18, 42, 78, 156, 234, 546

  \\ DATA
  if(0,
     for(n=1,34,print1(a(n),",");if(n==19||n==28,print()));
     quit);

  \\ a(2n) new child under each vertex per A348067
  check_equal(vector(20,n, a(2*n)),
              vector(20,n, A348067_new_child_each(a(n))));

  \\ formula a(2^k + r)
  check_equal(vector(5,k,k--; vector(2^k,r, a(2^k+r))),
              vector(5,k,k--; vector(2^k,r, a(2^k) * prime(a(r)))));

  \\ example
  check_equal(vpar_to_primecode_vector(vpar_make_binomial_tree(7)),
              [42, 1, 2, 1, 4, 1, 1]);
  check_equal(a(4), 6);
  check_equal(a(3), 4);
  check_equal(prime(a(3)), 7);

  \\ A348067
  \\ vector(20,n, my(vpar=vpar_from_primecode(n)); vpar_to_primecode(concat(vpar,[1..#vpar])))

  my(by_recursive(n) =
     n>=1 || error();
     if(n<=1,1,
        my(k=logint(n-1,2), ret=A076146(k+1));
        n-=1<<k;
        print(n);
        if(n,ret*=prime(A666666_binomial_tree_n(n)));
        ret));
  check_equal(vector(32,n, a(n)),
              vector(32,n, by_recursive(n)),
              "A666666_binomial_tree_n() vs by_recursive");


  \\ code formula a(m+1)
  check_equal(vector(5,k,k--; vector(2^k,s,s--;
                my(m=2^k+s); a(m+1))),
              vector(5,k,k--; vector(2^k,s,s--;
                my(m=2^k+s); a(2^k) * prime(a(s+1)))));

  my(a_with_checks(n) =
     my(ret=1,
        seen=Map());
     if(n--,
        my(i=0,
           T=1,   \\ a(2^i)
           max_i = if(n,logint(n,2)));
        while(ret == a((n % 2^i) + 1)  || error();
              T   == a(2^i)            || error();

              if(bittest(n,i),
                 if(ret>10 && mapisdefined(seen,ret),
                    print("duplicate prime() ret=",ret));
                 mapput(seen,ret,1);

                 ret = T*prime(ret));

              i++ <= max_i
                    ,
              if(T>10 && mapisdefined(seen,T),
                 print("duplicate prime() T=",T));
              mapput(seen,T,1);

              T *= prime(T)));
     print("distinct prime() taken ",#seen);
     ret);
  check_equal(vector(32,n, a_with_checks(n)),
              vector(32,n, a(n)),
              "A666666 a_with_checks");

  if(0,
     \\ n=60 59918820260044270902 time 1080 785

     for(n=1,256,
        gettime();
        my(got=a(n),
           t=gettime());
        print("n="n" "got" time "t));
     quit);
}

\\-----------------------------------------------------------------------------
progress("A356082 Complete Binary Tree Primecodes");

{
  my(want=[1, 4, 49, 51529, 400034745289]);
  my(a(n) =  n>=1 || error();
     my(ret=1); for(i=2,n, ret=prime(ret)^2); ret;
     );
  check_equal(vector(#want,n, a(n)), want);
  check_equal(vector(4,n, a(n)),
              vector(4,n, vpar_to_primecode(vpar_make_complete_tree(2,n-1))));

  check_equal(vector(4,n,n++; a(n)),
              vector(4,n,n++; prime(a(n-1))^2));

  check_equal(vector(4,n,n++; 
                     vpar_is_isomorphic(vpar_from_primecode(a(n)),
                                        vpar_make_complete_tree(2,n-1))),
              vector(4,n,n++; 1));

  \\ example
  check_equal(vpar_to_primecode_vector(vpar_make_complete_tree(2,2)),
              [49, 4,4, 1,1,1,1]);

  \\ height cf A007097 path length h is h+1 vertices
  check_equal(vector(6,h,h--; vpar_to_primecode(vpar_make_path(h+1))),
              vector(6,h,h--; test_A007097_primeth(h)));
}


\\-----------------------------------------------------------------------------
progress("A356083 Complete Ternary Tree Primecodes");

{
  my(want=[1, 8, 6859, 328951971956791]);
  my(a(n) = n>=1 || error();
     my(ret=1); for(i=2,n, ret=prime(ret)^3); ret;
     );
  check_equal(vector(#want,n, a(n)), want);
  check_equal(vector(4,n, a(n)),
              vector(4,n, vpar_to_primecode(vpar_make_complete_tree(3,n-1))));

  check_equal(vector(3,n,n++; a(n)),
              vector(3,n,n++; prime(a(n-1))^3));

  check_equal(vector(4,n,
                     vpar_is_isomorphic(vpar_from_primecode(a(n)),
                                        vpar_make_complete_tree(3,n-1))),
              vector(4,n, 1));

  \\ example
  check_equal(vpar_to_primecode_vector(vpar_make_complete_tree(3,2)),
              [6859, 8,8,8, 1,1,1,1,1,1,1,1,1]);
}


\\-----------------------------------------------------------------------------
progress("A356084 Complete Quaternary Tree Primecodes");

{
  my(want=[1, 16, 7890481, 379723559137648771815613388050801]);
  my(a(n) = n>=1 || error();
     my(ret=1); for(i=2,n, ret=prime(ret)^4); ret;
     );
  check_equal(vector(#want,n, a(n)), want);
  check_equal(vector(4,n, a(n)),
              vector(4,n, vpar_to_primecode(vpar_make_complete_tree(4,n-1))));

  check_equal(vector(3,n,n++; a(n)),
              vector(3,n,n++; prime(a(n-1))^4));

  check_equal(vector(4,n,
                     vpar_is_isomorphic(vpar_from_primecode(a(n)),
                                        vpar_make_complete_tree(4,n-1))),
              vector(4,n, 1));

  \\ example
  check_equal(vpar_to_primecode_vector(vpar_make_complete_tree(4,2)),
              [7890481, 16,16,16,16, 1,1,1,1, 1,1,1,1, 1,1,1,1, 1,1,1,1]);
}


\\-----------------------------------------------------------------------------
progress("Binary Trees (0 or 2 children) by Leaves");
\\ A245824

A001190 = OEIS_data_func("A001190");

A245824_rows(len) =
{
  my(ret=Vec([ [1] ],len));
  for(n=1,len-1,
     my(l=List([]));
     ret[n+1] = vector(sum(i=1,(n+1)>>1, #ret[i]*#ret[n+1-i]));
     \\ #ret[n+1] == A001190(n) || error(n);
     my(upto=0);
     for(x=1,(n+1)>>1,
        my(y=n+1-x);
        for(i=1,#ret[x],
           for(j=i,#ret[y],
              listput(l, prime(ret[x][i]) * prime(ret[y][j])))));
     ret[n+1] = Vec(l);
     \\ upto==#ret[n+1] || error(n);
  );
  apply(vecsort,ret);
}
{
  my(got=A245824_rows(8),
     want=OEIS_data("A245824"));
  check_equal(concat(got)[1..#want],want);
}

\\-----------------------------------------------------------------------------
progress("Num Orbits");

\\ A214567 how many orbits in the automorphism group of primecode n,
\\ so how many distinct rooted trees by adding a new child under a vertex

\\ A000107 = total num orbits in all rooted n vertices

A214567_Matula_num_orbits(n) =
{
  1 + vecsum([self()(primepi(p)) |p<-factor(n)[,1]]);
}

{
  my(a=A214567_Matula_num_orbits,
     by_vpar(n) = vpar_num_orbits
                    (vpar_from_primecode(n)));
  OEIS_check_func("A214567",by_vpar);
  OEIS_check_func("A214567_Matula_num_orbits");

  \\ formula in A214567
  check_equal(a(1), 1);
  check_equal(vector(500,t,n=prime(t); a(n)),
              vector(500,t,n=prime(t); 1 + a(t))); 
  check_equal(vector(500,n, a(n)),
              vector(500,n, 
                     my(v=distinct_prime_factors(n));
                     1 + sum(i=1,#v, a(v[i])-1)));

  \\ example in A214567
  check_equal(a(987654321), 18);

  check_equal(vector(500,primecode,
    my(vpar=vpar_from_primecode(primecode),
       new_vpars=vector(#vpar,v, vpar_join(vpar,v,[0])),
       new_primecodes=apply(vpar_to_primecode,new_vpars));
    #Set(new_primecodes)),
              vector(500,primecode, by_vpar(primecode)),
              "num automorphism orbits by adding a child");

  \\ probably wrong
  \\ check_equal(vector(500,n,
  \\                    my(seen=Map(), pending=List([n]));
  \\                    while(#pending,
  \\                          my(n=pending[1]); pending=pending[^1];
  \\                          if(mapisdefined(seen,n),next);
  \\                          print(n);
  \\                          pending=concat(pending,
  \\                                         Vec(apply(primepi,factor(n)[,1]))));
  \\                    #seen),
  \\             vector(500,primecode,
  \\                    vpar_num_orbits(vpar_from_primecode(primecode))),
  \\             "num automorphism orbits by adding a child");
}

\\-----------------------------------------------------------------------------
\\ Num Pairs of Same Siblings

test_num_sibling_pairs(vpar) =
{
  my(ids=vpar_automorphism_ids_vector(vpar),
     siblings=vpar_siblings_sets(vpar));
  sum(i=1,#siblings,
      my(s=siblings[i]);
      sum(i=1,#s-1,
          sum(j=i+1,#s,
              ids[s[i]]==ids[s[j]])));
}
{
  my(a(n)=test_num_sibling_pairs(vpar_from_primecode(n)));
  my(want=OEIS_data("A276625"));
  check_equal(select(n->a(n)==0, [1..want[#want]]),
              want,
              "num_sibling_pairs = 0 iff asymmetric");

  if(0,
     \\ max num vertices at same depth level
     print(vector(30,n,n++; a(n)));
     \\ not in OEIS: 0, 0, 1, 0, 0, 1, 3, 1, 0, 0, 1, 0, 1, 0, 6, 1, 1, 3, 1, 1, 0, 1, 3, 1, 0, 3, 2, 0, 0, 0
     quit);
}


\\-----------------------------------------------------------------------------
\\ Breath = Max Num Vertices at Depths

\\ width of the tree, max vertices at any given depth
\\ 
test_max_breadth(vpar) =
{
  vecmax(vpar_INTERNAL_histogram(vpar_depths_vector(vpar),0));
}
{
  if(0,
     \\ max num vertices at same depth level
     \\ max or row of A196056
     print(vector(20,n,n++; vecmax(vpar_numdetphs_vector(vpar_from_primecode(n)))));
     \\ not in OEIS: 1, 1, 2, 1, 2, 2, 3, 2, 2, 1, 3, 2, 2, 2, 4, 2, 3, 3, 3, 3

     \\ min num vertices at same depth level,
     \\ other than the root
     print(vector(20,n,n++; vecmin(vpar_numdetphs_vector(vpar_from_primecode(n))[^1])));
     \\ not in OEIS: 1, 1, 2, 1, 1, 1, 3, 2, 1, 1, 1, 1, 2, 1, 4, 1, 2, 1, 1, 2
     quit);
}


\\-----------------------------------------------------------------------------
\\ A355661 Max Num Children

print("=========");
print("read a354322-Matula-subtree-rows.gp");
read("a354322-Matula-subtree-rows.gp");
print("=========");
A354322_row=row;
kill(row);
kill(rows_vector);
OEIS_check_rowfunc("A354322_row",,['offset,1]);

\\ max number of children under a vertex in tree n
A355661_max_num_children(n) =
{
  my(f=factor(n)); vecmax(concat(vecsum(f[,2]), [self()(primepi(p)) |p<-f[,1]]));
}
{
  my(a=A355661_max_num_children);
  check_equal(vector(1000,n, a(n)),
              vector(1000,n, vecmax(vpar_num_children_vector(vpar_from_primecode(n)))),
              "A355661_max_num_children() vs vpar");

  \\ record highs
  my(m=-oo,got=List([]));
  for(n=1,1024,
     my(t=a(n));
     if(t>m, listput(got,[n,t]); m=t));
  check_equal(Vec(got),
              vector(11,k,k--; [2^k,k]));

  \\ new root no change
  check_equal(vector(1000,n,n++; a(prime(n))),
              vector(1000,n,n++; a(n)));

  \\ <=1 at prime'th recurrence
  my(want=select(n->a(n)<=1,[1..10000]));
  check_equal(#want, 9);
  check_equal(vector(#want,n,n--; test_A007097_primeth(n)), want);

  \\ formula max bigomega()
  my(prime_factors(n)=Vec(factor(n)[,1]));
  check_equal(vector(1000,n, a(n)),
              vector(1000,n, 
    vecmax(concat(bigomega(n), [a(primepi(p)) | p<-prime_factors(n)]))));

  \\ formula A354322 row
  check_equal(vector(1000,n, a(n)),
              vector(1000,n, 
    vecmax(apply(bigomega,A354322_row(n)))));
  

  \\ example
  \\       629              1
  \\      /   \           /   \    
  \\     7     12        2      6  
  \\     |     /|\       |     /|\ 
  \\     4    1 1 2      3    7 8 9
  \\    / \       |     / \       |
  \\   1   1      1    4   5     10
  \\       1 2 3 4 5  6 7 8 9 10
  my(vpar=[0,1,2,3,3, 1,6,6,6,9]);
  vpar_validate(vpar);
  my(n=vpar_to_primecode(vpar));
  check_equal(n, 629);
  check_equal(a(629), 3);
  check_equal(vpar_from_primecode(n), vpar);
  check_equal(vpar_to_primecode_vector(vpar),
              [629, 7,4,1,1, 12,1,1,2,1]);

  \\ DATA
  if(0,
     for(n=1,87,print1(a(n),",");if(n==29||n==58,print()));
     quit);
  OEIS_check_func("A355661_max_num_children");

  \\ ~/OEIS/b355661.txt
  if(0,
     system("mkdir -p /tmp/new");
     system("rm -f /tmp/new/b355661.txt");
     for(n=1,10000,
        write("/tmp/new/b355661.txt", Str(n," ",a(n))));
     system("ls -l /tmp/new/b355661.txt");
     quit);
  if(0,
     OEIS_check_func("A355661_max_num_children",,'bfile));

  \\ [|] apply syntax always vec even if input col
  my(f=factor(30),
     v=[5*p |p<-f[,1]]);
  check_equal(type(f[,1]),"t_COL");
  check_equal(type(v),"t_VEC");
  check_equal(#v, 3, "vec length");
}


\\-----------------------------------------------------------------------------
\\ A355662 Min Num Children

A355662_min_num_children(n) =
{
  \\ my line of code
  n>=1 || error();
  my(f=factor(n)); vecmin(concat(vecsum(f[,2]), [self()(primepi(p)) |p<-f[,1], p!=2]));
}
{
  my(a=A355662_min_num_children);

  \\ singleton
  check_equal(a(1), 0);
  check_equal(factor(1), matrix(0,2));
  check_equal(factor(1)[,1], []~);
  check_equal(factor(1)[,2], []~);
  check_equal(vecsum(factor(1)[,2]), 0,
              "vecsum() of empty is 0");

  check_equal(vector(1000,n, a(n)),
              vector(1000,n, if(n==1,0, vecmin(select(c->c!=0, vpar_num_children_vector(vpar_from_primecode(n)))))),
              "A355662_min_num_children() vs vpar");
  
  \\ A001222 num primes with multiplicity, takes n=1 as 0 primes
  check_equal(bigomega(1), 0);
  OEIS_check_func("A001222","bigomega");

  \\ record highs
  my(m=-oo,got=List([]));
  for(n=1,1024,
     my(t=a(n));
     if(t>m, listput(got,[n,t]); m=t));
  check_equal(Vec(got),
              vector(11,k,k--; [2^k,k]));

  \\ new root no change
  check_equal(vector(1000,n, a(prime(n))),
              vector(1000,n, 1));

  \\ formula min bigomega()
  my(prime_factors(n)=Vec(factor(n)[,1]));
  my(odd_prime_factors(n)=select(p->p%2,
                                 prime_factors(n)));
  check_equal(vector(1000,n, a(n)),
              vector(1000,n, 
    vecmin(concat(bigomega(n),
                  apply(p->a(primepi(p)),
                       odd_prime_factors(n))))));

  \\ formula A354322 row
  check_equal(vector(1000,n, a(n)),
              vector(1000,n, 
    my(v=apply(bigomega,
              select(s-> s>=2,  \\ not singletons
                     A354322_row(n))));
    if(#v,vecmin(v),0)));

  \\ example
  \\     31972               1
  \\    / |  \             / |  \         
  \\   1  1  1007         2  3    4
  \\        /    \             /    \     
  \\       8      16          5      9   
  \\      /|\    // \\       /|\    // \\ 
  \\     1 1 1  1 1 1 1     6 7 8  10 11 12 13
  \\       1 2 3 4  5 6 7 8  9 10    13
  my(vpar=[0,1,1,1, 4,5,5,5, 4,9,9,9,9]);
  vpar_validate(vpar);
  my(n=vpar_to_primecode(vpar));
  check_equal(n, 31972);
  check_equal(a(31972), 2);
  check_equal(vpar_from_primecode(n), vpar);
  check_equal(vpar_to_primecode_vector(vpar),
              [31972, 1,1, 1007, 8,1,1,1, 16,1,1,1,1]);
  check_equal(A355661_max_num_children(31972), 4);

  \\ DATA
  if(0,
     for(n=1,87,print1(a(n),",");if(n==29||n==58,print()));
     quit);
  my(want=[
0,1,1,2,1,1,1,3,1,1,1,1,1,2,1,4,1,1,1,1,1,1,1,1,1,1,1,2,1,
 1,1,5,1,1,1,1,1,2,1,1,1,1,1,1,1,1,1,1,2,1,1,1,1,1,1,2,1,1,
 1,1,1,1,1,6,1,1,1,1,1,1,1,1,1,1,1,3,1,1,1,1,1,1,1,1,1,2,1
           ]);
  check_equal(vector(#want,n,a(n)), want);
}

{
  if(0,
     \\ min num children
     print(vector(20,n,n++; a(n)));
     \\ not in OEIS: 1, 1, 2, 1, 1, 1, 3, 1, 1, 1, 1, 1, 2, 1, 4, 1, 1, 1, 1, 1

     print(select(n->a(n)<=1,[1..20]));
     \\ not in OEIS: 1, 2, 3, 5, 6, 7, 9, 10, 11, 12, 13, 15, 17, 18, 19, 20
     quit);
}
\\ lone-child avoiding, nowhere only child
is_A291636(n) = n>=1 && A355662_min_num_children(n)!=1;
{
  OEIS_check_predfunc("is_A291636")
}

\\-----------------------------------------------------------------------------
progress("Centre");

{
  \\ \\ num centre vertices
  \\ my(want=OEIS_data("A198330"));
  \\ check_equal(vector(#want,primecode,
  \\                    #vpar_centres(vpar_from_primecode(primecode)) - 1),
  \\             want);
  \\ not in OEIS: 1, 2, 1, 1, 2, 2, 1, 1, 1, 1, 1, 2, 2, 2, 2, 1, 2, 1, 1, 1
  \\ not in OEIS: 0, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 1, 0, 0, 0
}
{
  \\ A198330 primecodes which are unicentral
  my(is(n) = n>=1 &&
     #vpar_centres(vpar_from_primecode(n)) == 1);
  OEIS_check_predfunc("A198330",is);
}
{
  \\ A198331 primecodes which are bicentral
  my(is(n) = #vpar_centres(vpar_from_primecode(n)) == 2);
  OEIS_check_predfunc("A198331",is);
}
{
  \\ primecodes which are unicentroidal
  \\ maybe rooted at centroid too
  \\ maybe root is centroid of either unicentroid or bicentroid
  \\ could choose bicentroid root by free-primary rule

  if(0,
     \\ root is any of the centroid
     \\ not in OEIS: 1, 2, 4, 6, 8, 9, 12, 15, 16, 18, 20, 21, 24, 25, 27, 28, 30, 32, 35, 36, 40, 42
     print(select(n->my(vpar=vpar_from_primecode(n));
             setsearch(Set(vpar_centroids(vpar)),1),
                  [1..100]));
     quit);

  if(0,
     \\ root is unicentroid
     \\ not in OEIS: 1, 4, 8, 9, 12, 16, 18, 24, 25, 27, 30, 32, 35, 36
     print(select(n->my(vpar=vpar_from_primecode(n));
                     vpar_centroids(vpar)==[1],
                  [1..100]));
     quit);

  if(0,
     \\ unicentroidal
     \\ not in OEIS: 1, 3, 4, 7, 8, 9, 10, 11, 12, 13, 14, 16, 17, 18, 19, 23, 24
     print(select(n->my(vpar=vpar_from_primecode(n));
                     #vpar_centroids(vpar) == 1,
                  [1..100]));
     quit);

  if(0,
     \\ bicentroidal
     \\ not in OEIS: 2, 5, 6, 15, 20, 21, 22, 28, 29, 31, 34, 43, 55, 59, 65
     print(select(n->my(vpar=vpar_from_primecode(n));
                     #vpar_centroids(vpar) == 2,
                  [1..100]));
     quit);
}


\\-----------------------------------------------------------------------------
progress("Vertex Under Smallest Child of Root");

{
  \\ not in OEIS: 3, 5, 9, 11, 9, 17, 27, 25, 15, 31, 27, 41
  my(a(n) = my(f=factor(n));
     f[1,1] = prime(f[1,1]);
     factorback(f));
  if(0,
     print(vector(12,n,n++; a(n)));
     quit);
}


\\-----------------------------------------------------------------------------
progress("Delete Largest Child Vertex of Root");

\\ all primes drop to primepi
A003963_primes_to_primepi(n) =
{
  n>=1 || error();
  my(f=factor(n));
  factorback(apply(p->primepi(p), f[,1]), f[,2]);
}
{
  OEIS_check_func("A003963_primes_to_primepi",,['offset,1]);
  my(by_vpar(n) = my(vpar=vpar_from_primecode(n));
     vpar=vpar_delete_vertices(vpar,
            vpar_children_of_vertex(vpar,1));
     for(v=2,#vpar, if(vpar[v]==0, vpar[v]=1));
     vpar_to_primecode(vpar));
  check_equal(vector(500,n, A003963_primes_to_primepi(n)),
              vector(500,n, by_vpar(n)));
}                     

\\ smallest prime drop to primepi
smallest_prime_to_primepi(n) =
{
  n>=1 || error();
  if(n==1,1,
     my(f=factor(n));
     n * primepi(f[1,1]) / f[1,1]);
}
{
  if(0,
     print(vector(12,n,n+=2; smallest_prime_to_primepi(n)));
     quit);
  \\ not in OEIS: 2, 2, 3, 3, 4, 4, 6, 5, 5, 6, 6, 7
}

smallest_prime_to_primepi_multiplicity(n) =
{
  n>=1 || error();
  if(n==1,1,
     my(f=factor(n));
     f[1,1] = primepi(f[1,1]);
     factorback(f));
}
{
  if(0,
     print(vector(12,n,n+=2; smallest_prime_to_primepi_multiplicity(n)));
     quit);
  \\ not in OEIS: 2, 1, 3, 3, 4, 1, 4, 5, 5, 3, 6, 7
}

\\ largest prime drop to primepi
largest_prime_to_primepi(n) =
{
  n>=1 || error();
  if(n==1,1,
     my(f=factor(n), p=f[matsize(f)[1],1]);
     n * primepi(p) / p);
}
{
  if(0,
     print(vector(12,n,n+=2; largest_prime_to_primepi(n)));
     quit);
  \\ not in OEIS: 2, 2, 3, 4, 4, 4, 6, 6, 5, 8, 6, 8
}

largest_prime_to_primepi_multiplicity(n) =
{
  n>=1 || error();
  if(n==1,1,
     my(f=factor(n),
        r=matsize(f)[1]);
     f[r,1] = primepi(f[r,1]);
     factorback(f));
}
{
  if(0,
     print(vector(12,n,n+=2; largest_prime_to_primepi_multiplicity(n)));
     quit);
  \\ not in OEIS: 2, 1, 3, 4, 4, 1, 4, 6, 5, 8, 6, 8
}



\\-----------------------------------------------------------------------------
\\ vpar_depths_vector() rows

{
  \\ when ascending tree numbers is lex min
  if(0,
     my(is(n) = my(vpar=vpar_from_primecode(n));
        vpar == vpar_relabel_preorder(vpar_relabel_reverse(vpar_to_unlabelled(vpar))));
     print(select(is,[1..20]));
     print(select(n->!is(n),[1..500]));
     quit);  
  \\ not in OEIS: 1, 2, 3, 4, 5, 7, 8, 9, 11, 16, 17, 19, 23, 25, 27, 31, 32, 35, 49, 53, 59, 64, 67, 81, 83, 95, 97
  \\ not in OEIS: 6, 10, 12, 13, 14, 15, 18, 20, 21, 22, 24, 26, 28, 29, 30, 33, 34, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 50

  \\   1         35
  \\   | \        | \    
  \\   2  5       3  4   
  \\   |  | \     |  | \ 
  \\   3  6  7    2  1  1
  \\   |          |      
  \\   4          1      
  check_equal(vpar_from_primecode(35),
              [0, 1, 2, 3, 1, 5, 5]);
  \\           1  2  3  4  5  6  7
  check_equal(35, 5 * 7);
  check_equal(vpar_from_primecode(5),
              [0, 1, 2, 3]);

  \\      1
  \\     / \
  \\    1   2
  \\        |
  \\        3
  check_equal(vpar_from_primecode(6),
              [0, 1, 1, 3]);
  \\           1  2  3  4 
}
{
  \\ depths levels
  if(0,
     print(concat(vector(10,n,n+=4; vpar_depths_vector(vpar_from_primecode(n)))));
     print(concat(vector(10,n,n+=4; apply(t->t+1, vpar_depths_vector(vpar_from_primecode(n))))));
     print(vector(10,n,n+=4; vecsum(vpar_depths_vector(vpar_from_primecode(n)))));
     quit);
  \\ not in OEIS: 0, 1, 2, 3, 0, 1, 1, 2, 0, 1, 2, 2, 0, 1, 1, 1
  \\ not in OEIS: 1, 2, 3, 4, 1, 2, 2, 3, 1, 2, 3, 3, 1, 2, 2, 2
}
{
  \\ vpars
  if(0,
     print(concat(vector(10,n,n+=4; vpar_from_primecode(n))));
     quit);
  \\ not in OEIS: 0, 1, 2, 3, 0, 1, 1, 3, 0, 1, 2, 2, 0, 1, 1, 1
}


\\-----------------------------------------------------------------------------
progress("Series-Reduced");

\\ not in OEIS: 1, 2, 2, 4, 2, 4, 7, 8, 4, 4, 2, 8, 7
\\              1  2  3  4  5           9    

\\ Delete from vpar any vertices with exactly 1 child.
\\ 
test_series_reduce(vpar) =
{
  \\ print("vpar ",vpar);
  \\ vpar_validate(vpar);
  my(num_children=vpar_num_children_vector(vpar),
     del_flags=vectorsmall(#vpar,v,
                 vpar[v]
                 && num_children[v]==1),
     new=vector(#vpar,v,
                my(p=vpar[v]);
                while(p && del_flags[p], p=vpar[p]);
                p));
  \\ print("del_flags "del_flags);
  \\ print(num_children);
  vpar_delete_flags(new,del_flags);

  \\ my(num_children=vpar_num_children_vector(vpar),
  \\    del=select(v->num_children[v]==1,[1..#vpar]),
  \\    new=vector(#vpar,v,
  \\               my(p=vpar[v]);
  \\               while(p && vpar[p] && num_children[p]==1,
  \\                     p=vpar[p]);
  \\               p));
  \\ vpar_delete_vertices(new,del);
}
{
  check_equal(test_series_reduce([0,1,2, 1,4,5]),
              [0,1,1]);
  check_equal(test_series_reduce([0,1,2,3, 1,5,6,7]),
              [0,1,1]);

  my(A999999_series_reduce(n) =
     vpar_to_primecode
     (test_series_reduce(vpar_from_primecode(n))));
  \\ OEIS_check_func("A000001",A999999_series_reduce,
  \\                 ['offset,1]);

check_equal(vpar_from_primecode(5), [0, 1, 2, 3]);
check_equal(vpar_from_primecode(3), [0, 1, 2]);
  check_equal(A999999_series_reduce(3), 2);
  check_equal(A999999_series_reduce(3*3), 2*2, "3*3");
  check_equal(A999999_series_reduce(5*5), 2*2);
}

{
  \\ A331489 primecodes of series reduced free trees
  my(is_A331489_series_reduced(n) =
     my(vpar=vpar_from_primecode(n),
        degrees=vpar_degrees_vector(vpar));
     sum(i=1,#degrees, degrees[i]==2)==0);
  OEIS_check_predfunc("A331489",is_A331489_series_reduced);
  \\ OEIS_data("A331489")

  \\ primecodes of not series reduced free trees (somewhere a degree 2)
  \\ not in OEIS: 3, 4, 5, 6, 9, 10, 11, 12, 13, 14, 15, 17, 18, 20, 21
}
{
  \\ A257538 subdivide each edge to path-2
  my(want=OEIS_data("A257538"),
     by_vpar(n) = vpar_to_primecode(vpar_subdivide(vpar_from_primecode(n),1)));
  check_equal(vector(#want,primecode,by_vpar(primecode)),
              want,
              "A257538 subdivide each edge to path-2");

  \\ like Antti Karttunen in A257538
  my(a(n) = my(f=factor(n));
     for(i=1,#f~, f[i,1] = prime(prime(self()(primepi(f[i,1])))));
     factorback(f);
    );
  check_equal(vector(200,n,a(n)),
              vector(200,n,by_vpar(n)));
}


\\-----------------------------------------------------------------------------
\\ Binary Treees (0 or 2 children)
\\ How many when childless depths range <= 2.
\\ Maximally balanced.

\\ A299037 number of rooted binary trees 
\\ with minimal Sackin index and n leaves.

{
  if(0,
     my(l=List([]));
     for(n=1,20,
        my(full_h = logint(n,2),
           full_vpar=vpar_make_complete_tree(2,full_h),
           full_c = vpar_num_childless(full_vpar),
           extras = n - full_c,
           childless=vpar_childless_vertices(full_vpar),
           m=Map());
        forsubset([full_c,extras],sub,
           my(app=apply(elem->vector(2,i,childless[elem]),
                       Vec(sub)));
           if(#app,app=concat(app));
           my(vpar=concat(full_vpar,app));
           vpar_validate(vpar);
           mapput(m,vpar_to_unlabelled(vpar),1));
        print("n="n" "full_c" extras "extras
             " ways "#m);
        listput(l,#m)
        );
     print(Vec(l));
     quit);
}

\\-----------------------------------------------------------------------------
\\ Binary Treees (0 or 2 children)
\\ Minimum height for n childless, how many ways.

\\            n=1  2  3  4  5  6  7  8  9 10
\\ not in OEIS: 1, 1, 1, 1, 2, 2, 1, 1, 8, 9

\\ n=9 h=4  8  [452411, 576157, 682711, 713573, 728443, 882899, 1264966, 1421587]
\\ n=10 h=4  9  [3041573, 3604079, 3972049, 4383377, 4427381, 5935757, 6395537, 7033511, 10297681]
\\ vpar_from_primecode(452411)
\\ vpar_from_primecode(10297681)

\\ one path down to the min height h, takes 2 childless
\\ rest as a binary forest of height <= h-1


logint_ceil(n,b) = if(n==1,0, logint(n-1,b)+1);
test_is_childless_depths_range_le_2(vpar) =
{
  my(c=Set(vecextract(vpar_depths_vector(vpar),
                      vpar_childless_vertices(vpar))));
  #c==1 || (#c==2 && c[1]+1==c[2]);
}  
{
  if(0,
     for(n=1,10,
        my(vpar=vpar_unlabelled_first(2*n-1),
           m=Map(),
           h=logint_ceil(n,2));
        until((vpar=vpar_unlabelled_next(vpar)) == 'none,  \\ trees
              setminus(Set(vpar_num_children_vector(vpar)),
                       [0,2]) == []  || next;
              vpar_height(vpar) == h || next;
              test_is_childless_depths_range_le_2(vpar) || next;
              my(primecode=vpar_to_primecode(vpar));
              mapput(m,primecode,1));
        print("n="n" h="h"  "#m"  ",Mat(m)[,1]~);
     );
     quit);
}


\\-----------------------------------------------------------------------------
\\ Binary Treees (0 or 2 children)

\\ u and v are vectors.
\\ Return a Set() of all u[i]*v[j] combinations.
test_all_cross_products(u,v) = \
  Set(concat(vector(#u,i, u[i]*v)))

\\ Return a vector of vectors which are rows of A245824,
\\ binary trees (0 or 2 children).
test_binary_tree_primecodes_vector(len) =
{
  my(ret=Vec([ [1] ], len),
     pp =Vec([ [2] ], len));
  for(n=2,len,
     my(s=[]);
     for(x=1,n\2,
        my(y=n-x);
        s=setunion(s,test_all_cross_products(pp[x],pp[y])));
     ret[n] = s;
     if(n<len, pp[n] = apply(prime,ret[n])));
  ret;
}
{
  \\ A245824 binary trees (0 or 2 children)
  \\ by rows of n childless
  my(want=OEIS_data("A245824"),
     got=concat(test_binary_tree_primecodes_vector(8)));
  check_equal(got[1..#want], want,
              "A245824 test_binary_tree_primecodes_vector()");

  \\ A001190 number of binary trees (0 or 2 children)
  \\ of n childless.
  \\ Wedderburn-Etherington numbers.
  \\ Row widths of A245824.
  \\   0, 1, 1, 1, 2, 3, 6, 11, 23, 46, 98,
  \\ n=0  1  2  3  4  5  6   7   8
  \\
  my(want=OEIS_data("A001190"),
     got=apply(length,test_binary_tree_primecodes_vector(8)));
  want=want[^1];  \\ skip its n=0
  check_equal(got,want[1..#got],
              "A001190 count binary 0,2 trees");
}

\\ A245824
\\ 1;
\\ 4;
\\ 14;
\\ 49, 86;
\\ 301, 454, 886;
\\ 1589, 1849, 3101, 3986, 6418, 13766;
\\
\\ min and max by rows
\\ not in OEIS: 1,4,14, 49, 301, 1589
\\ not in OEIS: 1,4,14, 86, 886, 13766

\\ Return the number of binary trees (0 or 2 children) for n childless vertices.
test_binary_tree_primecodes_by_search(n) =
{
  my(vpar=vpar_unlabelled_first(2*n-1),
     m=Map());
  until((vpar=vpar_unlabelled_next(vpar)) == 'none,  \\ trees
        setminus(Set(vpar_num_children_vector(vpar)),
                 [0,2]) == []  || next;
        my(primecode=vpar_to_primecode(vpar));
        mapput(m,primecode,1));
  Mat(m)[,1]~;
}
{
  my(len=6);
  check_equal(test_binary_tree_primecodes_vector(len),
              vector(len,n, test_binary_tree_primecodes_by_search(n)));
}
\\ A001190
\\   Wedderburn-Etherington numbers:
\\   unlabelled binary rooted trees
\\   (every node has out-degree 0 or 2)
\\   with n endpoints (and 2n-1 nodes in all).

\\ apply(t->2*t- 1,OEIS_data("A001190"))

\\ print(vector(12,n,test_binary_tree_count_by_search(n)));quit
\\ print(vector(12,n,(test_binary_tree_count(n)+1)/2));quit

\\-----------------------------------------------------------------------------
progress("Binary Tree (0 or 2 children) -- Smallest");
\\ my A352456

\\ n, code, x                 y
\\ 1  [1, 1]                  
\\ 2  [4, [1]]                 1
\\ 3  [14, [1]]                2
\\ 4  [49, [2]]                2
\\ 5  [301, [2]]               3
\\ 6  [1589, [2]]              4
\\ 7  [9761, [3]]              4
\\ 8  [51529, [4]]             4
\\ 9  [452411, [4]]            5
\\ 10  [3041573, [4]]          6
\\ 11  [23140153, [4]]         7
\\ 12  [143573641, [4]]        8
\\ 13  [1260538619, [5]]       8
\\
{
  \\ vpar_from_primecode(1260538619)
  check_equal(vpar_from_primecode(1260538619),
     [0,1,2,3,3, 2,6,6,8,8,
   \\ 1 2 3 4 5  6 7 8 9 10
   1,11,12,13,13, 12,16,16, 11,19,20,20, 19,23,23
\\11 12 13 14 15  16 17 18  19 20 21,22  23 24 25
     ]);
  \\
  \\                    1
  \\               /           \
  \\         2                     11
  \\       /   \                /       \
  \\     3       6           12            19
  \\    / \     / \         /  \         /    \
  \\   4   5   7   8      13     16     20     23
  \\              / \    / \    /  \    / \    / \
  \\             9  10  14  15 17  18  21 22 24  25
}

print("---------------------------------");
print("a352456-Matula-binary-smallest.gp");
read("a352456-Matula-binary-smallest.gp");
test_A352456_binary_tree_min_vector = a_vector;
kill(a_vector);
print("---------------------------------");

{
  my(a_vector=test_A352456_binary_tree_min_vector);
  check_equal(vector(6,len,len--; #a_vector(len)),
              vector(6,len,len--; len));
}

test_binary_tree_min(n) = \
  test_binary_tree_min_and_xy(n)[1];
test_binary_tree_min_and_xy(n) =
{
  if(n==1,[1,1],
     my(x_list=[1..n\2],
        p_list=apply(x->
                    my(y=n-x);
                    x+y == n || error();
                    x <= y || error();
                    prime(test_binary_tree_min(x))
                    * prime(test_binary_tree_min(y)),
                 x_list),
        m=vecmin(p_list),
        x_list_m = select(x->p_list[x]==m, x_list));
     my(y_list_m = apply(x->n-x, x_list_m));
     [m,x_list_m,y_list_m]);
}
test_binary_tree_min_and_xy=memoize(test_binary_tree_min_and_xy);

{
  my(want=test_A352456_binary_tree_min_vector(11));
  check_equal(vector(#want,n,test_binary_tree_min(n)),
              want);
  check_equal(vector(#want,n,
       if(n==1,1,
       vecmin(vector(n-1,x,my(y=n-x);
                prime(want[x])*prime(want[y]))))),
              want,
              "vs vecmax");

  if(0,
    for(n=1,15,
       my(v=test_binary_tree_min_and_xy(n),
          primecode=v[1]);
       print(n"  "v);
       \\ print(test_binary_tree_min_and_xy(n)[2]);
       if(test_is_pow2(n),
          my(c=vpar_make_complete_tree(2,logint(n,2)),
             cc=vpar_to_primecode(c));
          print("   ",cc);
          if(cc!=v[1], print("*******")););
       ));
}

{
  \\ DATA 
  my(A352456_samps=[1,4,14,49,301,1589,9761,51529,452411,3041573,23140153,143573641,1260538619]);
  check_equal(test_A352456_binary_tree_min_vector(#A352456_samps),
              A352456_samps);
  check_equal(#A352456_samps, 13);

  \\ so far, all minimum are childless depths range 2
  \\ so fill complete tree then possible row of extras
  for(n=1,#A352456_samps,
     my(vpar=vpar_from_primecode(A352456_samps[n]),
        depths=vpar_depths_vector(vpar),
        childless=vpar_childless_vertices(vpar),
        cdepths=vecextract(depths,childless),
        cmin=vecmin(cdepths),
       cmax=vecmax(cdepths));
     \\ print(cdepths"  "cmin" "cmax);
     check_equal(cmin==cmax || cmin+1==cmax,
                 1);
     );
}
\\print(vector(8,n,test_binary_tree_min(n)));quit
\\ not in OEIS: 1, 4, 14, 49, 301, 1589, 9761, 51529

\\print(vector(12,n,test_binary_tree_min_and_xy(n)[2]));quit

test_binary_tree_min_by_search(n) =
{
  my(vpar=vpar_unlabelled_first(2*n-1),
     m=oo);
  until((vpar=vpar_unlabelled_next(vpar)) == 'none,  \\ trees
        setminus(Set(vpar_num_children_vector(vpar)),
                 [0,2]) == []  || next;
        m=min(m,vpar_to_primecode(vpar)));
  m;
}
{
  my(limit=7);
  check_equal(vector(limit,n,test_binary_tree_min(n)),
              vector(limit,n,test_binary_tree_min_by_search(n)),
              "test_binary_tree_min() vs search");
}

\\ vpar_from_primecode(301)
\\ [0, 1, 2, 2, 1, 5, 5, 7, 7]
\\
\\          1
\\        /  \
\\       2    5
\\      / \   / \
\\     3  4  6   7
\\              / \
\\             8   9
\\
\\ vpar_from_primecode(1589)
\\ [0, 1, 2, 2, 1, 5, 6, 6, 5, 9, 9]
\\           root
\\         /     \
\\       a         b
\\      / \      /   \
\\     c  d     e     f
\\             / \   / \
\\            g  h  i   j

{
  check_equal(vpar_to_primecode([0,1,2,2,1,5,4,4,5,9,9]),
              1849);
}
\\-----------------------------------------------------------------------------
progress("Total Cophenetic");
\\ A352288 total cophenetic of Matula   -- prospective
\\
\\ phylogenetic tree = shared history of extrant species
\\ leaves (childless) = extant species
\\ internal vertices = common ancestors of some
\\ root = most recent common ancestor of all


print("-------------------");
print("Read a352288-total-cophenetic.gp ...");
read("a352288-total-cophenetic.gp");
print("-------------------");
A352288_total_cophenetic=a;
kill(a);

{
  check_equal(A352288_total_cophenetic(1), 0);
  if(0,
     \\ DATA
     for(n=1,87,print1(A352288_total_cophenetic(n),",");if(n==29||n==58,print()));
     quit);

  my(want=[
0,0,0,0,0,0,1,0,0,0,0,0,1,1,0,0,2,0,3,0,1,0,1,0,0,1,0,1,1,
 0,0,0,0,2,1,0,3,3,1,0,2,1,4,0,0,1,1,0,2,0,2,1,6,0,0,1,3,1,
 3,0,3,0,1,0,1,0,6,2,1,1,3,0,4,3,0,3,1,1,1,0,0,2,2,1,2,4,1
]);
  check_equal(vector(#want,n, A352288_total_cophenetic(n)),
              want,
              "A352288_total_cophenetic() vs data");
  \\ not in OEIS: 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 2, 0, 3, 0, 1, 0, 1, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 2, 1, 0, 3, 3, 1

  \\ ~/OEIS/b352288.txt
  if(0,
     system("mkdir -p /tmp/new");
     system("rm -f /tmp/new/b352288.txt");
     for(n=1,10000,
        write("/tmp/new/b352288.txt", Str(n," ",A352288_total_cophenetic(n))));
     system("ls -l /tmp/new/b352288.txt");
     quit);
  if(0,
     my(want=OEIS_data("A352288",'bfile));
     print("~/OEIS/b352288.txt length ",#want);
     check_equal(vector(#want,n, A352288_total_cophenetic(n)),
                 want,
                 "A352288_total_cophenetic() b-file samples"));

  my(by_vpar(n) = vpar_total_cophenetic(vpar_from_primecode(n)));
  check_equal(vector(1000,n, A352288_total_cophenetic(n)),
              vector(1000,n, by_vpar(n)),
              "A352288_total_cophenetic() vs vpar");

  my(A109129_num_childless(n)=if(n==1,0,
       vpar_num_childless(vpar_from_primecode(n))));
  my(want=OEIS_data("A109129"));    \\ OFFSET=1
  check_equal(vector(#want,n, A109129_num_childless(n)), want,
              "A109129() num non-root childless in primecode");

  \\ formula
  my(a=A352288_total_cophenetic,
     C=A109129_num_childless);
  check_equal(vector(1000,n, a(n)),
              vector(1000,n,
                     if(n==1,0,
                        my(p=prime_factors(n), k=#p);
                        sum(i=1,k, a(primepi(p[i])) + binomial(C(p[i]),2)))),
              "A352288_total_cophenetic() formula");
}

{
  \\   root  R         pair  ancestor depth
  \\        / \         G,D     A       1
  \\       A   B        G,E     A       1
  \\      /|\   \       D,E     A       1
  \\     C D E   F     any,F    R       0
  \\     |                             ---
  \\     G                 total a(n) = 3
  my(vpar=[0,1,1, 2,2,2,3, 4]);
  check_equal(vpar_total_cophenetic(vpar), 3);
  check_equal(vpar_to_primecode(vpar), 111);
}
{
  \\     1
  \\    / \
  \\   2   3
  \\        \
  \\         4
  check_equal(vpar_from_primecode(6),  [0,1,1,3]);
  check_equal(vpar_Wiener_index([0,1,1,3],1),3);
}
{
  \\     1
  \\    /|\
  \\   2 2 3
  my(vpar=[0,1,1,1],
     childless=vpar_childless_vertices(vpar),
     C=#childless,
     depths=vpar_depths_vector(vpar),
     childless_depths=vecextract(depths,childless),
     CD=vecsum(childless_depths););
  check_equal(vpar_from_primecode(8),  vpar);
  check_equal(childless, [2,3,4]);
  check_equal(depths, [0,1,1,1]);
  check_equal(childless_depths, [1,1,1]);
  check_equal(vpar_Wiener_index(vpar,1), 6);
  check_equal(vpar_total_cophenetic(vpar), 0);
  check_equal(CD, 3);
  check_equal((C-1)*CD, 2*3);

  \\ identity with TW on vpar
  check_equal(vector(50,n,
                     my(vpar=vpar_from_primecode(n));
                     2*vpar_total_cophenetic(vpar)
                     + vpar_Wiener_index(vpar,1)),
              vector(50,n,
                     my(vpar=vpar_from_primecode(n),
                        childless=vpar_childless_vertices(vpar),
                        C=#childless,
                        depths=vpar_depths_vector(vpar),
                        total_depths=vecsum(vecextract(depths,childless)));
                     total_depths * (C-!isprime(n))),
              "2*total cophenetic + terminal Wiener");

  \\ identity 2*a(n) + TW(n) = Ext(n)*(C(n) - 1)
  my(a = A352288_total_cophenetic);

  my(TW(n) = vpar_Wiener_index(vpar_from_primecode(n),2));
  my(v=OEIS_data("A348959"));
  check_equal(vector(#v,n,TW(n)), v,
              "A348959 childless TW");

  my(Ext(n) = my(vpar=vpar_from_primecode(n));
       vecsum(vecextract(vpar_depths_vector(vpar),
                vpar_childless_vertices(vpar))));
  my(v=OEIS_data("A196048"));
  check_equal(vector(#v,n,Ext(n)), v, "A196048 Ext");
  check_equal(Ext(1), 0);  \\ so C(1) doesn't matter

  my(C(n) = if(n==1,0,
       vpar_num_childless(vpar_from_primecode(n))));
  my(v=OEIS_data("A109129"));
  check_equal(vector(#v,n,C(n)), v, "A109129 C");

  check_equal(vector(50,n, 2*a(n) + TW(n)),
              vector(50,n, Ext(n)*(C(n) - 1)),
              "2*a(n) + TW identity");
}

\\ phylogenetic means series-reduced, nowhere 1 child
test_is_phylogenetic(vpar) =
{
  foreach(vpar_num_children_vector(vpar),c,
     if(c==1,return(0))); 
  1;
}
\\
\\ select(n->test_is_phylogenetic(vpar_from_primecode(n)), [1..200])
\\ A291636 lone-child-avoiding rooted trees
\\
{
  \\ Mir, Rossello, Rotger,
  \\ total cophenetic is maximum at caterpillar spine 1 childless each
  \\ binary phylogenetic trees = series reduced leaf-labelled n leaves
  if(0,
     my(limit=6,
        max_total=vector(limit),
        max_vpar=vector(limit));
     for(n=1,2*limit + 2,
        my(vpar=vpar_unlabelled_first(n));
        until((vpar=vpar_unlabelled_next(vpar)) == 'none,  \\ trees
              test_is_phylogenetic(vpar) || next;
              my(c=vpar_num_childless(vpar));
              if(c>limit,next);
              my(total=vpar_total_cophenetic(vpar));
              if(total>max_total[c],
                 max_total[c]=total;
                 max_vpar[c]=vpar)));
     print(max_total);
     print(vector(limit,n,binomial(n,3)),"  binomials");
     print(max_vpar);
     quit);

  check_equal(vector(20,n, my(vpar=vpar_make_caterpillar(vector(n,i,1)));
                           vpar_total_cophenetic(vpar)),
              vector(20,n, binomial(n,3)),
              "total cophenetic maximum");
  \\ vector(20,n, binomial(n,3))
  \\ A000292 tetrahedral
}

A200067(n) =
{
  \\ (2*n^3 - 3*n^2 - if(n%3==2,3*n) + (n%3))/27;
  \\ if(n%3==0, 2*n^3 - 3*n^2,
  \\    n%3==1, 2*n^3 - 3*n^2       + 1,           \\ (n-1)^2    *(2*n+1)
  \\    n%3==2, 2*n^3 - 3*n^2 - 3*n + 2) / 27;     \\ (n-2)*(n+2)*(2*n-1)

  ((2*n^2 - 3*(n+(n%3==2)))*n + 2)\27;
}
OEIS_check_func("A200067");
{
  \\ maximum over n vertices
  \\ n = 1 2 3 ...
  \\ max  0,0,0,1,3,6,12,20,30,45,63,84,112,144,180,225,
  \\ rep 1,1,2,1,1,2, 1, 1, 2, 1, 1, 2,  1,  1,  2,  1,
  \\ A200067 maximum sum of all products of absolute differences and distances between element pairs among the integer partitions of n.
  \\   0,0,0,1,3,6,12,20
  \\
  \\ n=7 reps 1
  \\   [0, 1, 2, 3, 3, 3, 3]
  \\ n=8 reps 1
  \\   [0, 1, 2, 3, 3, 3, 3, 3]
  \\ n=9 reps 2
  \\   [0, 1, 2, 3, 4, 4, 4, 4, 4]
  \\   [0, 1, 2, 3, 3, 3, 3, 3, 3]

  if(0,
     for(n=1,16,
        my(vpar=vpar_unlabelled_first(n),
           max_total=-1,
           max_reps=0,
           max_vpars=[]);
        until((vpar=vpar_unlabelled_next(vpar)) == 'none,  \\ trees
              my(total=vpar_total_cophenetic(vpar));
              if(total==max_total,
                 max_reps++; max_vpars=concat(max_vpars,[vpar]);
                 ,
                 total>max_total,
                 max_total=total; max_reps=1; max_vpars=[vpar]));
        \\ print1(max_total,",");
        \\ print1(max_reps,",");
        print("n="n" reps "max_reps);
        foreach(max_vpars,vpar, print("  "vpar" primecode ",vpar_to_primecode(vpar)));
     );
     print();
     \\ print(vector(limit,n,binomial(n,3)),"  binomials");
     \\ print(max_vpar);
     );
}

\\ A325663 rooted stars, any arm lengths
\\ product of prime'th recurrences
\\ 1,2,3,4,5,6,8,9,10
is_A325663(n) =
{
  foreach(factor(n)[,1],p,
     while((p=primepi(p))>1, isprime(p)||return(0)));
  1;
}
{
  my(want=OEIS_data("A325663"));
  check_equal(select(is_A325663,[1..want[#want]]),
              want,
              "is_A325663() vs samples");
}

\\ rooted stars exactly 2 paths down
\\ print(select(n->bigomega(n)==2 && is_A325663(n),[1..100])); quit;
\\ 4, 6, 9, 10, 15, 22, 25, 33, 55, 62, 93
\\
\\ or exactly 2 paths down starting from a child
\\ apply(prime,[4, 6, 9, 10, 15, 22, 25, 33, 55, 62, 93])
\\ not in OEIS: 7, 13, 23, 29, 47, 79, 97, 137, 257, 293, 487

\\ print(select(n->A352288_total_cophenetic(n)==1,[1..100])); quit;
\\ not in OEIS: 7, 13, 14, 21, 23, 26, 28, 29, 35, 39, 42, 46, 47, 52, 56

\\ has a branch node somewhere lower than the root
\\ print(select(n->A352288_total_cophenetic(n)!=0,[1..100])); quit;
\\ not in OEIS: 7, 13, 14, 17, 19, 21, 23, 26, 28, 29, 34, 35, 37, 38

\\ select(n->A352288_total_cophenetic(n)==1,[1..200])
\\ select(n->A352288_total_cophenetic(n)==2,[1..200])
\\ not in OEIS: 7, 13, 14, 21, 23, 26, 28, 29, 35, 39, 42
\\ not in OEIS: 17, 34, 41, 49, 51, 68, 82, 83, 85, 91, 98, 102, 109


\\-----------------------------------------------------------------------------
\\ A352289 n childless on path down so binary
\\    root  86       a(4) = 2*prime(14) = 86
\\         /  \
\\        14   1     tree numbers of subtrees shown,
\\       /  \        with "1" being childless
\\      4    1       (n=4 of those childless)
\\     / \
\\    1   1

A352289_caterpillar_binary(n) =
{
  n>=1 || error();
  my(ret=1); for(i=2,n, ret=2*prime(ret)); ret;

  \\ if(n==1,1, my(ret=4); for(i=3,n, ret=2*prime(ret)); ret);
}
{
  my(a=A352289_caterpillar_binary);
  check_equal(A352289_caterpillar_binary(2), 4);

  if(0,
     \\ DATA
     for(n=1,9,print1(A352289_caterpillar_binary(n),","));
     quit);
  my(want=[ 1,4,14,86,886,13766,298154,8455786,300427382
          ]);
  check_equal(vector(#want,n, A352289_caterpillar_binary(n)),
              want,
              "A352289_caterpillar_binary() vs DATA");

  \\ recurrence formula
  my(a=A352289_caterpillar_binary);
  check_equal(vector(8,n,n++; a(n)),    \\ n>=2
              vector(8,n,n++; 2*prime(a(n-1))),
              "A352289_caterpillar_binary() recurrence");

  check_equal(vector(8,n, vpar_num_childless(vpar_from_primecode(A352289_caterpillar_binary(n)))),
              vector(8,n, n),
              "A352289_caterpillar_binary() num childless = n");

  check_equal(2*prime(4), 14);
  check_equal(2*prime(4), 14);

  my(a=A352289_caterpillar_binary);
  check_equal(a(1), 1);

  \\ bigger than any combination of other childless which make n,
  \\ so numerically greatest
  for(n=2,8,
     my(sublens=[1..n-1]);
     for(i=1,#sublens,
        my(x=sublens[i],
           y=Vecrev(sublens)[i]);
        x + y == n || error();
        a(n) >= prime(a(x)) & prime(a(y))
         || error("n=",n," xy=",x," ",y);
    ));
}

\\ A000040
\\ n*(log n + log log n - 3/2)      for n >= 20,
\\  < prime(n)
\\  < n*(log n + log log n - 1/2).

\\ a(x+1) = 2*prime(a(x))
\\ prine(n) = n*log(n)
\\ prime(a(x+1))
\\ = prime(2*prime(a(x)))
\\ >= prime(a(x)) * 2*log(2*prime(a(x)))
\\
\\ prime(a(x)) * prime(a(y))
\\ prime(a(x+1)) * prime(a(y))  /  prime(a(x)) * prime(a(y+1))
\\ >= prime(a(x)) * 2*log(2*prime(a(x))) * prime(a(y)
\\    )  /  prime(a(x)) * prime(a(y+1))
\\ = 2*log(2*prime(a(x))) * prime(a(y) / prime(a(y+1))
\\ >= 2*log(2*prime(a(x))) * prime(a(y) /   prime(a(y) * 24log(2*prime(a(y))
\\ = og(2*prime(a(x)) - 3/2)  /( log(2*prime(a(y))) + log(log((2*prime(a(y))))) )
{
  if(0,
     my(a=A352289_caterpillar_binary);

     for(x=2,10, my(y=x-1);
        print(
              log(2*prime(a(x)) - 3/2)  /( log(2*prime(a(y))) + log(log((2*prime(a(y))))) )
              ));

     for(n=5,11,
        my(prev=0);
        for(x=n\/2,n-1,
           my(y=n-x, t=prime(a(x))*prime(a(y)));
           print1(" "t);
           if(prev, t>prev || error());
           prev=t);
        print());
     quit);;
}

{
  \\ A352289 as binary tree maximum
  \\ with DATA from Daniel Suteu
  my(data=[1, 4, 14, 86, 886, 13766, 298154,
              8455786, 300427382, 12942000398,
              659492202274, 38995629272042,
              2634767648759954, 200877694833442486,
              17101872791349773894]);
  my(a(n) = data[n]);
  my(p(n) = data[n+1]/2);
  for(n=2,#data,
     my(v=vector(n-1,x, my(y=n-x); p(x)*p(y)));
     check_equal(vecmax(v), a(n),
                 "A352289 as binary tree maximum"));
}




\\-----------------------------------------------------------------------------
progress("Binary Trees (0 or 2 children)");

{
  \\ A245824 binary trees by num leaves 
  \\         or equivalently num vertices
  \\
  \\    *       r = prime(x) * prime(y)
  \\   / \      n = #x + #y + 1
  \\  x   y     x <= y  for no duplication
  \\
  \\ 1 [0]
  \\ 4 [0, 1, 1]
  \\ 14 [0, 1, 1, 3, 3]
  \\ 86 [0, 1, 1, 3, 3, 5, 5]
  \\ 886 [0, 1, 1, 3, 3, 5, 5, 7, 7]
  \\ 13766 [0, 1, 1, 3, 3, 5, 5, 7, 7, 9, 9]
  \\ 298154 [0, 1, 1, 3, 3, 5, 5, 7, 7, 9, 9, 11, 11]
  \\ 8455786 [0, 1, 1, 3, 3, 5, 5, 7, 7, 9, 9, 11, 11, 13, 13]
  \\ 300427382 [0, 1, 1, 3, 3, 5, 5, 7, 7, 9, 9, 11, 11, 13, 13, 15, 15]
  \\
  \\ 886/86.
  \\ 8455786/298154.
  check_equal(14*86, 1204);
  check_equal(7+9+1, 17);


  my(limit=6,
     max_vpar=vector(limit));
  for(n=1,2*limit,
     my(vpar=vpar_unlabelled_first(n),
        codes=List([]));
     until((vpar=vpar_unlabelled_next(vpar)) == 'none,  \\ trees
           test_is_binary_tree(vpar) || next;
           listput(codes,vpar_to_primecode(vpar)));
     codes=vecsort(Vec(codes));
     if(#codes,
        my(t=codes[#codes],
           vpar=vpar_from_primecode(t));
        print(t" "vpar)
        );
     );
}

\\-------------
\\ A303024 not-binary, nowhere 2 chlidren
{
  my(is(n) = my(vpar=vpar_from_primecode(n),
                num_children=vpar_num_children_vector(vpar));
     for(v=1,#vpar, if(num_children[v]==2,return(0))); 1);
  my(want=OEIS_data("A303024"));
  check_equal(select(is,[1..want[#want]]),
              want,
              "A303024 Matula codes of nowhere 2 children");
}

\\ A303026 not-binary, nowhere 1 or 2 chlidren
{
  my(is(n) = my(vpar=vpar_from_primecode(n),
                num_children=vpar_num_children_vector(vpar));
     for(v=1,#vpar, if(num_children[v]==1||num_children[v]==2, return(0)));
     1);
  my(want=OEIS_data("A303026"));
  check_equal(select(is,[1..want[#want]]),
              want,
              "A303026 Matula codes of nowhere 1 or 2 children");
}


\\-----------------------------------------------------------------------------
\\ A331937   path down, one childless each

\\ 26, 202, 2462, 43954, 1063462, 33076174
 \\ a(1) = 1; a(2) = 2; a(n + 1) = 2 * prime(a(n)).
A331937_caterpillar_1s(n) = my(ret=min(n,2)); for(i=3,n, ret=2*prime(ret)); ret;
{
  \\     1
  \\    / \          and exception n=1 is singleton
  \\   2   3
  \\      / \
  \\     4   5
  \\        /
  \\       6
  check_equal(vpar_from_primecode(A331937_caterpillar_1s(4)),
              [0, 1,1, 3,3, 5]);
  \\           1  2 3  4 5  6
  check_equal(A331937_caterpillar_1s(2), 2);
  check_equal(A331937_caterpillar_1s(3), 2*prime(2));

  my(want=OEIS_data("A331937"));  \\ OFFSET=1
  while(want[#want]>10^7,want=want[^#want]);
  check_equal(vector(#want,n, A331937_caterpillar_1s(n)), want,
              "A331937_caterpillar_1s() samples");

  check_equal(vector(5,n, A331937_caterpillar_1s(n)),
              vector(5,n, if(n==1,1,
                             vpar_to_primecode(vpar_make_caterpillar(vector(n-1,i,1))))),
              "A331937_caterpillar_1s() vs vpar_make_caterpillar() spine n-1");
}

\\-----------------------------------------------------------------------------
progress("Branches");

\\ A191515 num vertices which have num children >= 2
{
  my(want=OEIS_data("A191515"));
  my(by_vpar(n)=my(vpar=vpar_from_primecode(n),
                   num_children=vpar_num_children_vector(vpar));
     sum(v=1,#vpar, num_children[v]>=2));
  check_equal(vector(#want,primecode, by_vpar(primecode)),
              want,
              "A191515 num vertices which have num children >= 2");

  my(a(n) = my(f=factor(n)); (vecsum(f[,2])>=2) + [self()(primepi(p)) |p<-f[,1]]*f[,2];
    );
  check_equal(vector(500,n, a(n)),
              vector(500,n, by_vpar(n)),
              "A191515 compat");

  \\ g(n)=G(n,x)
  my(rs(n) = my(r=factor(n)[1,1]); [r,n/r]);
  my(g(n,x='x) = if(n==1,0,
                    isprime(n), my(t=primepi(n)); x*self()(t,x),
                    my(r,s); [r,s]=rs(n); 1 + self()(r,x) + self()(s,x)
                    - self()(r,0) - self()(s,0)));
  check_equal(vector(500,n, g(n,1)),
              vector(500,n, a(n)),
              "A191515 by g()");

  \\ g(n,x) is polynomial count*x^depth how many at each depth
  check_equal(vector(500,n, my(vpar=vpar_from_primecode(n),
                               num_children=vpar_num_children_vector(vpar),
                               depths=vpar_depths_vector(vpar));
                     sum(v=1,#vpar, if(num_children[v]>=2, x^depths[v])) + x*0),
              vector(500,n, g(n,x)  + x*0),
              "A191515 by g()");

  \\-------------
  my(want=OEIS_data("A007097")[1..9]);
  check_equal(select(n->a(n)==0, [1..want[#want]]), want,
              "A191515 num branches=0 is path primeth recurrence");

  \\ print(concat(vector(50,n, Vecrev(g(n))))); quit;

  \\ num branches excluding root
  \\ print(vector(50,n,n++; a(n) - !isprime(n))); quit
  \\ not in OEIS: 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 1, 0, 1, 0, 1
}

\\-----------------------------------------------------------------------------
progress("Total Distances");

{
  \\ A196051 Wiener index of tree primecode n
  \\ sometimes called "area"
  my(want=OEIS_data("A196051"));
  check_equal(vector(#want,primecode,
                     vpar_Wiener_index(vpar_from_primecode(primecode))),
              want,
              "A196051 Wiener index of primecode");
}
{
  \\ return 2-element vector [W,t]
  my(W_and_num_vertices(n) =
       my(f=factor(n),wt);
       [ (wt=self()(primepi(p))) + [wt[2]*('total-wt[2]), 0]
        | p<-f[,1] ]   \\ each prime
       * f[,2]         \\ scalar product for repetitions
       + [0,1]);       \\ new vertex, this n subtree root

  A196051(n) = my(wt=W_and_num_vertices(n)); subst(wt[1],'total,wt[2]);
}
{
  check_equal(vector(1000,n, A196051(n)),
              vector(1000,n, vpar_Wiener_index(vpar_from_primecode(n))),
              "A196051() vs vpar_Wiener_index()");
}

\\----------------
{
  \\ A196055 terminal Wiener index of tree primecode n
  my(want=OEIS_data("A196055"));
  check_equal(vector(#want,primecode,
                     vpar_Wiener_index(vpar_from_primecode(primecode),1)),
              want,
              "A196055 terminal Wiener index of primecode");
}

{
  \\ return 2-element vector [W,t]
  my(W_and_leaves(n,root=0) =
       my(f=factor(n),wt);
       [ (wt=self()(primepi(p))) + [wt[2]*('total-wt[2]), 0]
        | p<-f[,1] ]   \\ each prime
       * f[,2]         \\ scalar product for repetitions
       + [0, n==1 || (root && f[,2]==[1]~)]);

  A196055(n) = my(wt=W_and_leaves(n,1)); subst(wt[1],'total,wt[2]);
}
{
  check_equal(vector(1000,n, A196055(n)),
              vector(1000,n, vpar_Wiener_index(vpar_from_primecode(n),1)),
              "A196055() vs vpar_Wiener_index() terminal Wiener");
}

\\---------------
{
  \\ A196066 reverse Wiener index of tree primecode n
  \\ measuring how much short of every pair at diameter apart
  my(want=OEIS_data("A196066"));
  check_equal(vector(#want,primecode,
                     my(vpar=vpar_from_primecode(primecode),
                        n=#vpar);
                     (1/2)*n*(n-1)*vpar_diameter(vpar)
                     - vpar_Wiener_index(vpar)),
              want,
              "A196066 reverse Wiener index of tree primecode n");
}

\\--------------
\\ vertices of degree >= 3
test_branching_vertices(vpar) = \
  Vec(select(d->d>=3,vpar_degrees_vector(vpar),1));

\\ total distance between pairs of degree >= 3
test_branching_Wiener_index(vpar) =
{
  my(m=vpar_path_lengths_matrix(vpar),
     branches=test_branching_vertices(vpar));
  sum(from=1,#branches,
      sum(to=from+1,#branches,
          m[branches[from],branches[to]]));
}
{
  \\ A206499 sum distances between branch nodes
  my(primecode=28,
     vpar=vpar_from_primecode(primecode));
  check_equal(vpar,[0,1,1,1, 4,4]);
  check_equal(vpar_degrees_vector(vpar),[3,1,1,3, 1,1]);
  check_equal(test_branching_Wiener_index(vpar), 1);
}
{
  my(want=OEIS_data("A206499"));
  check_equal(vector(#want,primecode,
                test_branching_Wiener_index(vpar_from_primecode(primecode))),
              want,
              "A206499 sum distances between branch nodes");
}

{
  \\ A196061 multiplicative Wiener index
  \\ product distances between pairs
  my(want=OEIS_data("A196061")); \\ OFFSET=2
  check_equal(vector(#want,primecode,primecode++;
                     my(vpar=vpar_from_primecode(primecode),
                        m=vpar_path_lengths_matrix(vpar));
                     prod(u=1,#vpar,
                          prod(v=u+1,#vpar,
                               m[u,v]))),
              want,
              "A196061 multiplicative Wiener index");
}

\\ If vpar is "planted" meaning its root has a single child, then delete the
\\ root and the path down to the first vertex which is not a single child.
\\ vpar can be a forest, in which case each component is unplanted.
test_unplanted(vpar) =
{
  while(1,
        my(roots=vpar_roots(vpar));
        roots=select(v->vpar_num_children_of_vertex(vpar,v)==1, roots);
        #roots || return(vpar);
        vpar=vpar_delete_vertices(vpar,roots));
}
{
  check_equal(test_unplanted([0,1,2,3]), [0]);

  my(want=OEIS_data("A288469"));  \\ OFFSET=1
  check_equal(vector(#want,primecode,
                vpar_to_primecode(test_unplanted(vpar_from_primecode(primecode)))),
              want,
              "A288469 test_unplanted() strip path down to non-single-child");

  \\ square apply primeth recurrence
  \\ my(limit=200,l=List([])); \
  \\ for(r=2,sqrtint(limit), my(s=r^2); \
  \\   until(s>limit, listput(l,s); s=prime(s))); Set(l)
}
{
  \\ total distances between childless vertices
  \\ not in OEIS: 0, 0, 0, 2, 0, 3, 2, 6, 4, 4, 0, 8, 3, 8, 5
  \\              1  2  3  4
  \\     4
  \\    / \   distance 2 between childless
  \\   1   1
  my(on_vpar(vpar) =
     my(m=vpar_path_lengths_matrix(vpar),
        childless=vpar_childless_vertices(vpar));
     sum(i=1,#childless, sum(j=1,i-1, m[childless[i],childless[j]])));
  my(by_vpar(primecode) = on_vpar(vpar_from_primecode(primecode)));
  check_equal(vector(10,primecode, by_vpar(primecode)),
              vector(10,primecode,
                     my(vpar=vpar_from_primecode(primecode));
                     vpar=test_unplanted(vpar);
                     vpar_Wiener_index(vpar,1)),
              "total between childless vs vpar_Wiener_index of unplanted");

  \\ print(vector(20,primecode, by_vpar(primecode))); quit;
}

\\-----------------------------------------------------------------------------
progress("Distinct Subtrees");

\\ Among each set of sibling subtrees, keep only one copy of each distinct
\\ subtree up to rooted isomorphism (vpar_is_isomorphic()).
\\ vpar can be a forest, in which case roots are considered siblings and
\\ only one copy of each distinct component tree is kept.
distinct_subtrees_of_vpar(vpar) =
{
  my(components=vpar_component_trees(vpar));
  components=apply(vpar->vpar_add_root_above_forest(     \\ recurse
                         distinct_subtrees_of_vpar(
                         vpar_delete_vertex(vpar,vpar_root(vpar)))),
                   components);
  components=Set(apply(vpar_to_unlabelled,components));  \\ canonicals
  vpar_concat_forests(components);
}

\\ (mine)
A348066_distinct_subtrees(n) =
{
  n>=1 || error();
  vecprod(Set([prime(self()(primepi(p))) | p<-factor(n)[,1]]));
}
{
  my(a=A348066_distinct_subtrees);
  OEIS_check_func("A348066_distinct_subtrees");

  check_equal(vector(500,n, a(n)),
              vector(500,n,
    vpar_to_primecode(distinct_subtrees_of_vpar(vpar_from_primecode(n)))),
    "A348066_distinct_subtrees() vs vpar");

  \\ unchanged iff asymmetric
  check_equal(vector(500,n, a(n) == n),
              vector(500,n, vpar_is_asymmetric(vpar_from_primecode(n))),
              "A348066_distinct_subtrees() unchanged iff asymmetric");

  \\ divisibility sequence: a(n) divides a(m*n) for any m and n
  check_equal(matrix(400,400,m,n, A348066_distinct_subtrees(m*n) % A348066_distinct_subtrees(n)),
              matrix(400,400,m,n, 0),
              "A348066_distinct_subtrees() divisibility");
  for(m=1,200,
     my(mset=prime_factors(A348066_distinct_subtrees(m)));
     mset==Set(mset) || error();
     for(n=1,200,
        my(nset=prime_factors(A348066_distinct_subtrees(n)));
        nset==Set(nset) || error();
        check_equal(prime_factors(A348066_distinct_subtrees(m*n)),
                    setunion(mset,nset),
                    "A348066_distinct_subtrees() set union of prime factors")));

  \\ DATA section
  if(0,
     for(n=1,75,print1(A348066_distinct_subtrees(n),",");if(n==28||n==51,print()));
     print();
     quit);

  \\ ~/OEIS/b348066.txt
  if(0,
     system("mkdir -p /tmp/new");
     system("rm -f /tmp/new/b348066.txt");
     for(n=1,10000,
        write("/tmp/new/b348066.txt", Str(n," ",a(n))));
     system("ls -l /tmp/new/b348066.txt");
     quit);
  if(0,
     OEIS_check_func("A348066",'bfile));

  \\ number of vertices after reduction
  \\       1
  \\      / \      primecode = 21
  \\     2   4
  \\    /   / \
  \\   3   5   6
  \\ not the same as number of orbits,
  \\ after reduction 4 same as 2 so merge
  my(primecode=21,
     vpar=vpar_from_primecode(primecode));
  check_equal(vpar, [0, 1, 2, 1, 4, 4]);
  check_equal(#vpar, 6);
  check_equal(vpar_num_orbits(vpar), 5);
  if(0,
     for(n=1,100,
        if(#vpar_from_primecode(a(n))
           != vpar_num_orbits(vpar_from_primecode(n)),
           print(n)));
     quit);
  if(0,
     print(vector(50,n, #vpar_from_primecode(a(n))));
     \\ not in OEIS: 1, 2, 3, 2, 4, 4, 3, 2, 3, 5, 5, 4, 5, 4, 6, 2, 4, 4, 3, 5, 3, 6, 4, 4, 4, 6, 3, 4, 6, 7, 6, 2, 7, 5, 6, 4, 5, 4, 7, 5, 6, 4, 5, 6, 6, 5, 7, 4, 3, 5

     quit);
}
{
  \\ not multiplicative at m=3, n=7
  \\ have divisible, but not multiplicative
  my(a=A348066_distinct_subtrees);
  my(m=3,n=7);
  check_equal(vpar_from_primecode(m), [0,1,2]);
  check_equal(vpar_from_primecode(n), [0,1,2,2]);
  check_equal(a(m), 3);
  check_equal(a(n), 3);
  check_equal(a(m*n), 3);
}
\\ print(vector(14,n,n+=2; A348066_distinct_subtrees(n)));quit;
\\ 3, 2, 5, 6, 3, 2, 3, 10, 11, 6, 13, 6, 15, 2
\\ print(Set(vector(1000,n, A348066_distinct_subtrees(n)))[1..14]);quit;
\\ not in OEIS: 1, 2, 3, 5, 6, 10, 11, 13, 15, 22, 26, 29, 30, 31

is_distinct_subtrees(n) =
{
  if(n>1, my(f=factor(n));
     vecmax(f[,2])==1 || return(0);
     for(i=1,matsize(f)[1], self()(primepi(f[i,1])) || return(0)));
  1;
}
{
  my(limit=1000,
     by_set=Set(vector(limit,n, A348066_distinct_subtrees(n))));
  check_equal(select(is_distinct_subtrees,[1..limit]),
              by_set);

  \\ A276625 asymmetrics
  check_equal(vector(1000,n, is_distinct_subtrees(n)),
              vector(1000,n, vpar_is_asymmetric(vpar_from_primecode(n))));
}

\\----------
\\ core(n) = d where n = d*f^2 with d square-free
check_equal(core(9), 1);
check_equal(core(81,1), [1,9]);

\\ done only at the root is one copy each prime factor
distinct_root_subtrees(n) = vecprod(factor(n)[,1]);
A007947(n) = distinct_root_subtrees(n);
{
  my(want=OEIS_data("A007947"));  \\ OFFSET=1
  check_equal(vector(#want,n, A007947(n)),
              want,
              "A007947 distinct_root_subtrees()");
  check_equal(matrix(200,200,m,n, if(gcd(m,n)!=1,'none, A007947(m)*A007947(n))),
              matrix(200,200,m,n, if(gcd(m,n)!=1,'none, A007947(m*n))),
              "A007947() multiplicative");

  check_equal(vector(500,n, my(vpar=vpar_from_primecode(n));
                     vpar=vpar_delete_vertex(vpar,vpar_root(vpar));
                     my(vpars=vpar_component_trees(vpar));
                     vpars=Set(apply(vpar_to_unlabelled,vpars));
                     vpar=vpar_add_root_above_forest
                          (vpar_concat_forests(vpars));
                     vpar_to_primecode(vpar)),
              vector(500,n, distinct_root_subtrees(n)),
              "distinct_root_subtrees() vs vpar");
}


\\-----------------------------------------------------------------------------
\\ A297002 new vertex under each child of the root

\\ apply(prime,OEIS_data("A297002")[1..10])
\\ not in OEIS: 2, 5, 17, 23, 41, 73, 67, 103, 227, 167
\\ Set(OEIS_data("A297002"))[1..12]
\\ not in OEIS: 1, 3, 7, 9, 13, 19, 21, 27, 29, 37, 39, 43

A297002(n) = test_completely_multiplicative(n, (p)->prime(2*primepi(p)));
{
  \\ A297002 multiplicative a(prime(k)) = prime(2 * k)
  \\ 1, 3, 7, 9, 13, 21, 19, 27, 49, 39, 29, 63, 37, 57, 91
  my(want=OEIS_data("A297002"));
  check_equal(vector(#want,n, A297002(n)),
              want,
              "A297002 multiplicative samples");

  \\ code by Remy Sigrist in A297002
  check_equal(vector(500,n, A297002(n)),
              vector(500,n, my(f=factor(n)); prod(i=1,#f~, prime(2*primepi(f[i,1]))^f[i,2])));
  \\ or using factorback
  check_equal(vector(500,n, A297002(n)),
              vector(500,n, my(f=factor(n)); factorback([prime(2*primepi(p)) | p<-f[,1]], f[,2])));

  check_equal(vector(#want,n, A297002(n)),
              want,
              "A297002 multiplicative prime(2*primepi(p))");

  check_equal(vector(500,n, A297002(n)),
              vector(500,n,
                  my(vpar=vpar_from_primecode(n),
                     children=vpar_children_of_vertex(vpar,1));
                  vpar=concat(vpar,children);  \\ childless under each child of root
                  vpar_validate(vpar);
                  vpar_to_primecode(vpar)),
              "A297002 multiplicative by vpar add childless each child of root");
}

\\-----------------------------------------------------------------------------
\\ A066207 All primes that divide n are of the form prime(2k)
is_A066207(n) = my(f=factor(n)); foreach(f[,1],p, primepi(p)%2==0 || return(0)); 1;
{
  my(want=OEIS_data("A066207"));
  check_equal(select(is_A066207, [1..want[#want]]),
              want,
              "is_A066207() vs samples");
}


\\-----------------------------------------------------------------------------
progress("Matchings");

\\ maximal/maximum matchings count
\\ vector(20,n,n+=5; vpar_maximal_matchings_count(vpar_from_primecode(n)))
\\ vector(20,n,n+=5; vpar_matchnum_and_count(vpar_from_primecode(n))[2])
\\ A347966 = 2, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 3, 3, 4, 5, 5, 4, 3, 4, 5
\\ A347967 = 1, 3, 3, 3, 3, 3, 2, 2, 2, 1, 4, 2, 1, 4, 5, 5, 1, 1, 3, 4

{
  \\ A193404 matchings count of tree primecode n
  my(want=OEIS_data("A193404"));
  check_equal(vector(#want,primecode,
                     vpar_matchings_count(vpar_from_primecode(primecode))),
              want,
              "A193404 matchings count of primecode");
}
{
  \\ A206483 matchnum of tree primecode n
  my(want=OEIS_data("A206483"),
     by_vpar(n) = vpar_matchnum(vpar_from_primecode(n)));
  check_equal(vector(#want,primecode, by_vpar(primecode)),
              want,
              "A206483 matchnum of primecode");

  check_equal(vpar_from_primecode(9), [0,1,2, 1,4]);

  my(matchnum_and_flag_root_unmatched(n) =
     my(f=factor(n),m=0,u=1,cm,cu);
     for(i=1,matsize(f)[1],
        [cm,cu]=self()(primepi(f[i,1]));
        m+=cm*f[i,2]; if(u&&cu, m++;u=0));
     [m,u]);
  my(a(n) = matchnum_and_flag_root_unmatched(n)[1];
    );
  check_equal(by_vpar(9),2, "by_vpar(9)");
  check_equal(a(9),2);
  check_equal(vector(2000,n, a(n)),
              vector(2000,n, by_vpar(n)),
              "A206483 matchnum a(n) vs by_vpar()");
}

{
  \\ Maple code in A193405 making matchpoly M(n) = [withpoly,withoutpoly]
  my(r(n) = factor(n)[1,1]);     \\ smallest prime factor of n
  my(s(n) = n/r(n));             \\ rest of n
  my(M(n) =
     my(M=self());
     if(n==1, [0, Pol(1)],
        bigomega(n)==1, [x*M(primepi(n))[2],
                         M(primepi(n))[1] + M(primepi(n))[2]],
        [M(r(n))[1]*M(s(n))[2] + M(r(n))[2]*M(s(n))[1],
         M(r(n))[2]*M(s(n))[2]]));

  for(n=1,150,
     my(maple=M(n),
        vpar=vpar_from_primecode(n),
        without=vpar_matchpoly(vpar_delete_vertex(vpar,1)),
        all=vpar_matchpoly(vpar),
        with=all-without);
     maple == [with,without]  || error(n));

  check_equal(vector(150,n, vecsum(M(n))),
              vector(150,n, vpar_matchpoly(vpar_from_primecode(n))),
              "A193405 M() Maple matchpoly of primecode");
}

{
  \\ A193403 rows adjacency poly, including 0s
  my(want=OEIS_data("A193403"));  \\ OFFSET=1
  my(got=[], n=1);
  while(#got < #want,
        got=concat(got,Vecrev(vpar_adjacency_poly(vpar_from_primecode(n))));
        n++);
  check_equal(got[1..#want], want,
              "A193403 rows adjacency poly");

  check_equal(vector(15,n, vpar_adjacency_poly(vpar_from_primecode(n))),
              vector(15,n, my(vpar=vpar_from_primecode(n),
                              m=vpar_matchpoly(vpar));
                     subst(m,x,-1/x^2)*x^#vpar),
              "adjacency poly = matchpoly recip x^2");

  \\ adjacency poly row sum
  \\ print(vector(50,n, subst(vpar_adjacency_poly(vpar_from_primecode(n)),x,1)));
  \\ not in OEIS: 1, 0, -1, -1, -1, -1, -2, -2, 0, 0, 0, -1, -1, -1, 1, -3, -1, 0, -3
  \\ for(n=1,25, print(Vecrev(vpar_adjacency_poly(vpar_from_primecode(n)))))
}
{
  \\ A202853 matchpoly rows
  my(want=OEIS_data("A202853"));  \\ OFFSET=1
  my(got=[], n=1);
  while(#got < #want,
        got=concat(got,Vecrev(vpar_matchpoly(vpar_from_primecode(n))));
        n++);
  check_equal(got[1..#want], want,
              "A202853 matchpoly rows, lowest term first");

  \\ matchpoly rows high to low instead
  \\ concat(vector(10,n,n+=2; Vec(vpar_matchpoly(vpar_from_primecode(n)))))
  \\ not in OEIS: 2, 1, 2, 1, 1, 3, 1, 1, 3, 1, 3, 1, 3, 1, 3, 4, 1, 3, 4, 1, 3, 4, 1, 2, 4, 1
  \\ for(n=1,10, print(Vecrev(vpar_matchpoly(vpar_from_primecode(n)))))
}

\\------------------
\\ A202854 treecodes which have palindromic matchpoly

test_poly_is_palindromic(p) = p==polrecip(p);
test_is_A202854(n) = \
  test_poly_is_palindromic(vpar_matchpoly(vpar_from_primecode(n)));

{
  my(want=OEIS_data("A202854"));  \\ OFFSET=1
  check_equal(select(test_is_A202854, [1..want[#want]]),
              want,
              "A202854 primecodes which have palindromic matchpoly");

  \\ 2*prime(n) for composite n
  foreach(OEIS_data("A202854"),n,
     n>=2 || next;
     if(isprime(n),
        check_equal(test_is_A202854(2*primepi(n)), 1,
                    "A202854 2*primei(n) of prime n");
        ,
        check_equal(test_is_A202854(2*prime(n)), 1,
                    "A202854 2*prime(n) of composite n"));

     \\ FIXME: Not quite right.
     \\ foreach(OEIS_data("A202854"),m,
     \\    m>=2 || next;
     \\    check_equal(test_is_A202854(n*m/2), 1,
     \\                "A202854 n*m/2"));
  );

  \\ Trees with palindromic matchpoly always have a perfect matching,
  \\ since a unique maximum matching means must be perfect.
  \\ Anywhere an unmatched u is beside a matched v-w can have that pairing
  \\ switch from v-w to instead v-u.
  for(rep=1,1000,
     my(vpar=vpar_make_random(random(50)),
        poly=vpar_matchpoly(vpar),
        ispal=test_poly_is_palindromic(poly));
     check_equal((vpar==[0] || vpar_has_perfect_matching(vpar)) >= ispal, 1,
                 vpar));

  \\ apply(n->vpar_has_perfect_matching(vpar_from_primecode(n)),OEIS_data("A202854"))

  \\ free trees 2n are free trees n with leaf attached at each, apparently
  \\ 1, 1, 0, 1, 0, 1, 0, 2, 0, 3, 0, 6, 0, 11, 0, 23
  \\ 2,3,6,11,23,47,106
  if(0,
     print(vector(12,n,
                  my(vpar=vpar_free_first(n),
                     count=0);
                  until((vpar=vpar_free_next(vpar))=='none,  \\ trees
                        count+=test_poly_is_palindromic(vpar_matchpoly(vpar)));
                  count));
     forstep(n=8,20,2,
        my(vpar=vpar_free_first(n),
           count=0);
        until((vpar=vpar_free_next(vpar))=='none,  \\ trees
              count+=test_poly_is_palindromic(vpar_matchpoly(vpar)));
        print1(count,","));
     print();
     quit);
}


\\-----------------------------------------------------------------------------
progress("Asymmetric Trees");

check_equal(issquarefree(7), 1);
check_equal(issquarefree(7^2), 0);
check_equal(issquarefree(7^3), 0);
check_equal(issquarefree(7^4), 0);
check_equal(issquarefree(7^3*13), 0);
check_equal(issquarefree(7*13), 1);

\\ primecodes of asymmetric trees
is_A276625(n) =
{
  \\ if(n>=4,
  \\    my(q=[],f=factor(n));
  \\    until(q[#q]<=3,
  \\          if(vecmax(f[,2])>1,return(0));
  \\          q=setunion(q[^#q],f[,1]~); f=factor(primepi(q[#q]))));
  \\ 1;

  \\ \\ looking at small factors first, remembering seen
  \\ issquarefree(n) || return(0);
  \\ my(q=factor(n)[,1]~,seen=[]);
  \\ while(#q,
  \\       n=primepi(q[1]);
  \\       \\ print(q" n="n"  seen "seen);
  \\       issquarefree(n)||return(0);
  \\       seen=setunion(seen,q[1]);
  \\       my(f=factor(n));
  \\       q=setunion(q[^1],setminus(f[,1]~,seen)));

  \\ looking at small factors first, remembering seen
  if(n>1,
     my(f=factor(n),seen=[2],q=[]);
     while(1,
           if(vecmax(f[,2])>1,return(0));
           q=setunion(q,setminus(f[,1]~,seen));
           #q||break;
           seen=setunion(seen,[q[1]]);
           f=factor(primepi(q[1]));
           q=q[^1]));
  1;
}

{
  \\ A276625 asymmetric primecodes

  check_equal(vector(5000,n, is_A276625(n)),
              vector(5000,n, vpar_is_asymmetric(vpar_from_primecode(n))),
              "is_A276625() vs vpar_is_asymmetric()");

  my(want=OEIS_data("A276625"));
  check_equal(select(is_A276625,[1..want[#want]]),
              want,
              "A276625 asymmetric primecodes");
  check_equal(vector(2000,n, is_A276625(n)),
              vector(2000,n, vpar_num_automorphisms(vpar_from_primecode(n))==1),
              "is_A276625() vs num automorphisms == 1");

  check_equal(matrix(400,400,x,y, is_A276625(x*y)),
              matrix(400,400,x,y, gcd(x,y)==1 && is_A276625(x) && is_A276625(y)),
              "is_A276625() two terms");

  check_equal(vector(1000,n, is_A276625(prime(n))),
              vector(1000,n, is_A276625(n)),
              "is_A276625() with prime");

  \\ primecodes of symmetrics (not asymmetric)
  \\ not in OEIS: 4, 7, 8, 9, 12, 14, 16, 17, 18, 19, 20, 21, 23, 24, 25
}

asymmetrics_row_by_vpar(n) =
{
  my(vpar=vpar_unlabelled_first(n),
     l=List([]));
  until((vpar=vpar_unlabelled_next(vpar)) == 'none,  \\ trees
        if(vpar_is_asymmetric(vpar),
           listput(l,vpar_to_primecode(vpar))));
  vecsort(Vec(l));
}

{
  \\ asymmetrics sorted by num vertices
  \\ not in OEIS: 2, 3, 5, 6, 10, 11, 13, 15, 22, 26, 29, 31, 41
  \\ print(concat(vector(5,n,n++; asymmetrics_row_by_vpar(n)))); quit;
}


\\-----------------------------------------------------------------------------
progress("Duplicate Subtrees");

duplicate_subtrees_of_vpar_forest(vpar) =
{
  my(components=vpar_component_trees(vpar));
  components=apply(duplicate_subtrees_of_vpar_tree,components);
  vpar_concat_forests(concat(components,components));
}
duplicate_subtrees_of_vpar_tree(vpar) =
{
  vpar_is_tree(vpar) || error();
  vpar_add_root_above_forest
  (duplicate_subtrees_of_vpar_forest
   (vpar_delete_vertex(vpar,vpar_root(vpar))));
}
{
  check_equal(duplicate_subtrees_of_vpar_forest([0]), [0,0]);
  my(got=duplicate_subtrees_of_vpar_forest([0,1]));
  check_equal(vpar_is_isomorphic(got, [0,1,1,0,4,4]), 1);
}
{ my(vpar=[0,1,1]);
  check_equal(duplicate_subtrees_of_vpar_tree(vpar), [5,5,5,5,0]);
}

duplicate_subtrees(n) = \
  my(f=factor(n)); sqr(factorback([prime(self()(primepi(p))) | p<-f[,1]], f[,2]));
{
  check_equal(vector(10,n, duplicate_subtrees(n)),
              vector(10,n, my(vpar=vpar_from_primecode(n));
                     vpar_to_primecode(duplicate_subtrees_of_vpar_tree(vpar))),
              "duplicate_subtrees() vs vpar");
}

\\ print(vector(10,n,n+=2; duplicate_subtrees(n)));quit;
\\ not in OEIS: 49, 16, 51529, 196, 2809, 64, 2401, 206116, 400034745289, 784, 1423249
\\
\\ print(vector(10,n,n+=2; sqrtint(duplicate_subtrees(n))));quit;
\\ not in OEIS: 7, 4, 227, 14, 53, 8, 49, 454, 632483, 28


\\-----------------------------------------------------------------------------
progress("Num Automorphisms");

{
  \\ A206497 num automorphisms
  \\ multiplicative
  my(want=OEIS_data("A206497"),
    a(n)=vpar_num_automorphisms(vpar_from_primecode(n)));
  check_equal(vector(#want,n, a(n)), want,
              "A206497 num automorphisms");

  \\ is multiplicative
  for(p=1,50,
     for(q=1,50,
        gcd(p,q)==1 || next;
        a(p*q) == a(p)*a(q)  || error()));
}


\\-----------------------------------------------------------------------------
progress("Primecodes by Num Vertices");


A061773_row(n) =
{
  my(vpar=vpar_unlabelled_first(n),
     l=List([]));
  until((vpar=vpar_unlabelled_next(vpar)) == 'none,  \\ trees
        listput(l,vpar_to_primecode(vpar)));
  vecsort(Vec(l));
}

\\ terms of A061773 for 1..n_limit vertices
A061773_to_num_vertices(n_limit) =
{
  my(ret=[]);
  for(n=1,n_limit,
     my(vpar=vpar_unlabelled_first(n),
        l=List([]));
     until((vpar=vpar_unlabelled_next(vpar)) == 'none,
           listput(l,vpar_to_primecode(vpar)));
     ret=concat(ret, vecsort(Vec(l))));
  ret;
}

{
  \\ A061773 triangle rows of primecodes of rooted trees of n vertices,
  \\ so some bigger primecodes wait until later, first such is 15
  \\ 1,      \\ n=1
  \\ 2,      \\ n=2
  \\ 3, 4,   \\ n=3
  \\ 5, 6, 7, 8,
  \\ 9,10,11,12,13,14,16,17,19;
  my(want=OEIS_data("A061773"));
  my(got=[], n=1);
  while(#got<#want,
        got=concat(got,A061773_row(n));
        n++);
  check_equal(got[1..#want],
              want,
              "A061773 rows of primecodes of rooted trees of n vertices");
  if(#got!=#want,
     "A061773 data length ",#want," versus whole row ",#got);

  for(i=1,#want,
     if(want[i]!=i,
        check_equal(i,15, "A061773 first out-of-sequence primecode at 15");
        break));
}



\\ {
\\   my(want=OEIS_data("A061775"),
\\     by_vpar(n)=#vpar_from_primecode(n));
\\   check_equal(vector(#want,primecode, primecode_to_num_vertices(primecode)),
\\               want,
\\               "A061775 num vertices in primecode");
\\ }

{
  \\ A005517 minimum primecode of n vertices
  \\   1, 2, 3, 5, 9, 15, 25, 45, 75, 125, 225, 375, 625, 1125, 1875, 3125,
  \\ A005518 maximum primecode of n vertices
  \\   1, 2, 4, 8, 19, 67, 331, 2221, 19577, 219613, 3042161, 50728129,
  \\ A091233 difference + 1
  my(want_max=OEIS_data("A005518"),
     limit=min(12,#want_max));
  want_max=want_max[1..limit];
  my(want_min =OEIS_data("A005517")[1..limit],
     want_diff=OEIS_data("A091233")[1..limit]);
  for(n=1,limit, \\ A005518_samples OFFSET=1
     my(vpar=vpar_unlabelled_first(n),
        c_min,c_max);
     c_min = c_max = vpar_to_primecode(vpar);
     until((vpar=vpar_unlabelled_next(vpar))=='none,
           my(c=vpar_to_primecode(vpar));
           c_min=min(c_min, c);
           c_max=max(c_max, c));
     check_equal(want_min[n], c_min,
                 "A005517 minimum primecode for n vertices");
     check_equal(want_max[n], c_max,
                 "A005518 maximum primecode for n vertices");
     check_equal(want_diff[n], c_max - c_min + 1,
                 "A091233 max-min primecode for n vertices"));

  \\ Gobel section "Range", table a_n minimum and omega_n maximum samples
  check_equal(want_min[1..9],
              [1,2,3,5,9,15,25,45,75]);
  check_equal(want_max[1..9],
              [1,2,4,8,19,67,331,2221,19577]);
}

\\--------------
\\ A347620 (mine) position of primecode when primecodes are sorted by num
\\ vertices, inverse of permutation A061773
\\ See a347620-Matula-inverse.gp
\\     a347620-Matula-inverse--test.gp

A347620_vector_by_rows(len) =
{
  my(ret=vector(len),
     got=[],
     n=1);
  while(hammingweight(ret) < #ret,
        got=concat(got,A061773_row(n)); n++;
        for(i=1,#got,
           if(got[i]<=#ret, ret[got[i]]=i)));
  ret;
}
{
  my(want=OEIS_data("A347620"));
  check_equal(A347620_vector_by_rows(#want), want,
              "A347620 position of primecode in sorted by num vertices");
}


\\-----------------------------------------------------------------------------
progress("Ordered Trees Codes");

\\ Balanced binary A014486 taken as encodings etc.
\\ A014486 = 0, 2, 10, 12, 42, 44, 50, 52, 56

balanced_binary_to_primecode(b) = \
  vpar_to_primecode(vpar_add_root_above_forest(vpar_from_balanced_binary(b)));
{
  my(want=OEIS_data("A127301"),
     b=0,
     got=vector(#want,i,
                my(primecode=balanced_binary_to_primecode(b));
                b=vpar_balanced_binary_next(b);
                primecode));
  check_equal(got, want, "A127301 balanced binary preorder to primecode");
}

preorder_tilt(vpar) =
{
  my(sibling=vpar_next_sibling_vector(vpar,1));
  vector(#vpar,v, if(sibling[v],sibling[v],vpar[v]));
}
check_equal(preorder_tilt([0,0,0,0]), [0,1,2,3]);

pad_to_binary_tree(vpar) =
{
  my(num_children=vpar_num_children_vector(vpar),
     ret=Vec(vpar,2*#vpar+1),
     upto=#vpar);
  for(v=1,#vpar,
     for(i=num_children[v],1,
        ret[upto++]=v));
  ret;
}
check_equal(pad_to_binary_tree([0]), [0,1,1]);
check_equal(pad_to_binary_tree([0,1]), [0,1, 1,2,2]);

{
  \\ empty, 10, 1010, 1100
  check_equal(vpar_from_primecode(1),  [0]);
  check_equal(vpar_from_primecode(4),  [0, 1, 1]);
  check_equal(vpar_from_primecode(14), [0, 1, 1, 3, 3]);

  \\ A014486(4..8) = 42, 44, 50, 52, 56
  \\                 86, 86, 49, 86, 86
  \\ ........\/.....\/.................\/.....\/...
  \\
  \\ .......\/.......\/.....\/.\/.....\/.......\/..
  \\
  \\ ......\/.......\/.......\_/.......\/.......\/.
  check_equal(vpar_from_balanced_binary(42), [0,0,0]);
  check_equal(vpar_from_balanced_binary(44), [0,0,2]);
  check_equal(vpar_from_balanced_binary(50), [0,1,0]);
  check_equal(vpar_from_balanced_binary(52), [0,1,1]);
  check_equal(vpar_from_balanced_binary(56), [0,1,2]);
  check_equal(preorder_tilt([0,0,0]), [0,1,2]);
  check_equal(preorder_tilt([0,0,2]), [0,1,2]);
  check_equal(preorder_tilt([0,1,1]), [0,1,2]);
  check_equal(preorder_tilt([0,1,2]), [0,1,2]);

  my(want=OEIS_data("A127302"),
     b=0,
     got=vector(#want,i,
                my(vpar=vpar_from_balanced_binary(b));
                vpar=preorder_tilt(vpar);
                vpar=pad_to_binary_tree(vpar);
                my(primecode=vpar_to_primecode(vpar));
                b=vpar_balanced_binary_next(b);
                primecode));
  check_equal(got, want, "A127302 balanced binary binary tree to primecode");
}



\\-----------------------------------------------------------------------------
{
  \\ A184159 childless deepest - shallowest difference
  my(want=OEIS_data("A184159")); \\ OFFSET=1
  check_equal(vector(#want,primecode,
                     my(vpar=vpar_from_primecode(primecode),
                        depths=vpar_depths_vector(vpar),
                        childless=vpar_childless_vertices(vpar));
                     depths=vecextract(depths,childless);
                     vecmax(depths)-vecmin(depths)),
              want,
              "A184159 childless deepest - shallowest difference");
}

\\ A184155 all childless are at same depth
{
  my(is(n) = my(vpar=vpar_from_primecode(n),
                depths=vpar_depths_vector(vpar));
     #Set(vecextract(depths,vpar_childless_vertices(vpar)))==1);
  my(want=OEIS_data("A184155"));
  check_equal(select(is,[1..want[#want]]),
              want,
              "A184155 Matula codes of all childless at same depth");
}

\\-----------------------------------------------------------------------------
progress("Num Childless Vertices");

\\ A184154 num childless by depth of tree primecode n, excluding depth 0
A184154_num_childless_at_depth_row(n) =
{
  my(vpar=vpar_from_primecode(n),
     depths=vpar_depths_vector(vpar)); \\ without root
  depths=vecextract(depths,vpar_childless_vertices(vpar));
  Vec(vpar_INTERNAL_histogram(depths,0))[^1];
}
A184154_at_depth(n,k) = A184154_num_childless_at_depth_row(n)[k];
{
  OEIS_check_func("A184154_num_childless_at_depth_row",,['offset,1]);

  check_equal(vector(1000,n, #A184154_num_childless_at_depth_row(n)),
              vector(1000,n, vpar_height(vpar_from_primecode(n))),
              "A184154_num_childless_at_depth_row() width = A109082_height()");
}


\\ this code by Francois Marques in A342507
A342507_formula(n) =  \\ childful
{
  my(f=factor(n));
  1 + sum(k=1, matsize(f)[1],
          A342507(primepi(f[k,1]))*f[k,2]);
}
A342507(n) = if(n==1, 0, A342507_formula(n));
{
  \\ A342507_formula(1) has the 1+ which gives 1 instead of desired 0
  check_equal(A342507(1),         0);
  check_equal(A342507_formula(1), 1);

  \\ A342507 num childful vertices in tree of primecode n
  \\ my(want=OEIS_data("A342507"));
  my(want=OEIS_data("A342507"));
  check_equal(vector(#want,primecode,
                     vpar_num_childful(vpar_from_primecode(primecode))),
              want,
              "A342507 num childful vertices in primecode");

  check_equal(vector(1024,n, A342507(n)),
              vector(1024,n, vpar_num_childful(vpar_from_primecode(n))),
              "A342507() num childful");
}

\\ Francois Marques in A109129,  childless
ML(n) =
{
  if(n==1,1,
     my(f=factor(n)); sum(k=1, matsize(f)[1], ML(primepi(f[k,1]))*f[k,2]));
}
A109129(n) = if(n==1, 0, ML(n));
{
  \\ A109129 num non-root childless vertices of tree primecode n
  \\ Non-root meaning singleton is 0 childless unlike vpar_num_childless()
  \\ gives 1.
  \\ 0, 1, 1, 2, 1, 2, 2, 3, 2, 2, 1, 3, 2, 3, 2, 4, 2, 3, 3, 3, 3, 2, 2, 4,
  my(want=OEIS_data("A109129"));    \\ OFFSET=1
  check_equal(vector(#want,n, A109129(n)),
              want,
              "A109129() num non-root childless in primecode");

  check_equal(vector(1000,n, A109129(n)),
              vector(1000,n,
                     \\ A109129 reckons singleton no non-root leaf
                     if(n==1,0,
                     vpar_num_childless(vpar_from_primecode(n)))),
              "A109129() num childless vs vpar");

  my(a=A109129);
  check_equal(matrix(10,10,r,s, a(r*s)),
              matrix(10,10,r,s, a(r) + a(s)),
              "A109129 num childless product formula");
}

{
  \\ A325543 num non-root childless vertices of tree primecode n!
  my(want=OEIS_data("A325543"));
  check_equal(vector(#want,n,n--; vpar_num_childless(vpar_from_primecode(n!))),
              want,
              "A325543 num childless n!");
}
{
  \\ A325612 num non-root childless vertices of tree primecode 2^n-1
  my(want=OEIS_data("A325612"));
  want=want[1..min(40,#want)];      \\ needs some big factoring
  check_equal(vector(#want,n,
                     vpar_num_childless(vpar_from_primecode(2^n-1))),
              want,
              "A325612 num childless 2^n-1");
}


\\-----------------------------------------------------------------------------
progress("Caterpillars");

is_caterpillar_at_end(vpar) =
{
  vpar_INTERNAL_is_state_table(vpar,vpar_INTERNAL_is_caterpillar_table,5,14);

  \\ all states accepting:
  \\ GP-Test  bitneg(0,3)*2 == 14
  \\ GP-Test  binary(14) == [         1, 1, 1, 0]
  \\                            5  4  3  2  1   state
}
{
  \\     3
  \\     |
  \\     4
  \\    / \
  \\   1   2
  \\   |   |
  \\   5   6
  my(vpar=[4, 4, 0, 3, 2, 1]);
  check_equal(vpar_is_caterpillar(vpar), 1);
  check_equal(vpar_root(vpar), 3);
  check_equal(vpar_eccentricity_of_vertex(vpar,3), 3);
  check_equal(vpar_diameter(vpar), 4);
}
test_is_caterpillar_at_end_by_diameter_endpoints(vpar) =
{
  if(#vpar==0, return(1));
  vpar_is_caterpillar(vpar) || return(0);
  my(ecc=vpar_eccentricities_vector(vpar),
     diameter=vecmax(ecc),
     root=vpar_root(vpar));
  if(ecc[root]==diameter, return(1));
  my(children=vpar_children_of_vertex(vpar,root));
  for(i=1,#children,
     if(ecc[children[i]]==diameter, return(1)));
  0;
}
{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i]);
     check_equal(is_caterpillar_at_end(vpar),
                 test_is_caterpillar_at_end_by_diameter_endpoints(vpar),
                 "is_caterpillar_at_end() vs test_is_caterpillar_at_end_by_diameter_endpoints()");
     );

  \\ 2*k and prime(k)
  my(limit=1000,
     set=[1],prevlen=0);
  while(#set!=prevlen,
        prevlen=#set;
        set=Set(concat([set, apply(k->2*k,set), apply(k->prime(k),set)]));
        set=select(k-> k<=limit, set));
  my(is_A209638(n) = is_caterpillar_at_end(vpar_from_primecode(n)));
  check_equal(set,
              select(is_A209638,[1..limit]),
              "2*k and primepi(k) vs select()");

  \\ A209638 Matula codes of caterpillars rooted at end
  \\ ( = A209636 in ascending order)
  my(want=OEIS_data("A209638"));
  check_equal(select(is_A209638,[1..want[#want]]),
              want,
              "A209638 Matula codes of caterpillars at end");

  \\ print(select(n->!is_A209638(n),[1..100]));
  \\ not in OEIS: 9, 15, 18, 21, 23, 25, 27, 30, 33, 35, 36, 39, 42, 45, 46, 47, 49, 50, 51, 54, 55, 57, 60, 61, 63, 65, 66, 69, 70, 72, 73, 75, 77, 78, 81, 83, 84, 85, 87, 90, 91, 92, 93, 94, 95, 97, 98, 99, 100

  \\ insisting that the root has at least one leaf child is 2*A209638
  my(want=OEIS_data("A209638"));
  check_equal(select(n->my(vpar=vpar_from_primecode(n),
                           children=vpar_children_of_vertex(vpar,vpar_root(vpar)),
                           leaf_children=select(v->vpar_is_leaf(vpar,v),children));
                     is_caterpillar_at_end(vpar) && #leaf_children>=1,
                     [1..2*want[#want]]),
              2*want,
              "A209638*2 Matula codes of caterpillars at end with >=1 leaf");
}


\\-----------------------------------------------------------------------------
progress("Num Vertices");

{
  \\ A196050 num edges in tree of primecode n
  my(want=OEIS_data("A196050"),
     by_vpar(n) = vpar_num_edges(vpar_from_primecode(n)));
  check_equal(vector(#want,primecode, by_vpar(primecode)),
              want,
              "A196050 num edges in primecode");

  \\ compact
  my(a(n) = my(f=factor(n)); [self()(primepi(p))+1 |p<-f[,1]]*f[,2];
    );
  check_equal(vector(2000,n,a(n)),
              vector(2000,n,by_vpar(n)),
              "num edges compact code");
}


{
  \\ A061775 num vertices in tree of primecode n
  my(want=OEIS_data("A061775"),
    by_vpar(n)=#vpar_from_primecode(n));
  check_equal(vector(#want,primecode, by_vpar(primecode)),
              want,
              "A061775 num vertices in primecode");

  \\ Antti Karttunen in A061775, mangled to recurse by self()
  my(AK(n) =
     if(n==1, 1,
        if(isprime(n),
           1+self()(primepi(n));
           ,
           my(pfs,t,i); pfs=factor(n);
           pfs[,1] = [self()(t) | t<- pfs[,1]]~;
           1-bigomega(n) + sum(i=1,omega(n), pfs[i,1]*pfs[i,2])));
     );
  check_equal(vector(500,n, AK(n)),
              vector(500,n, by_vpar(n)),
              "A061775 num vertices AK() vs by_vpar()");

  my(ma(n) = my(m=Map(factor(n)),ret=1);
     while(#m,
           my(p=vecextract(Mat(m)[,1],"-1")[1], \\ biggest prime in m
              c=mapget(m,p));                   \\ count
           mapdelete(m,p);
           \\ print("p="p" c="c"  "m);
           ret+=c;                              \\ this vertex
           my(f=factor(primepi(p)));
           for(i=1,matsize(f)[1],
              my(e=0); mapisdefined(m,f[i,1],&e);
              mapput(m,f[i,1], e + c*f[i,2]);
              \\ print("  store ",f[i,1]," count "mapget(m,f[i,1]));
           ));
     ret;
    );
  check_equal(vpar_from_primecode(9), [0, 1,2, 1,4]);
  check_equal(ma(9), 5);
  check_equal(ma(3), 3);
  check_equal(vector(2000,n, ma(n)),
              vector(2000,n, by_vpar(n)),
              "A061775 num vertices ma() map vs by_vpar()");

  \\ using lists
  my(L(n) = my(f=factor(n), q=List(f[,1]),qc=List(f[,2]), ret=1, p,c,pos);
     while(#q,
           my(p=q[#q],c=qc[#q]); listpop(q);listpop(qc); ret+=c;
           f=factor(primepi(p));
           for(i=1,matsize(f)[1],
              my(pos=setsearch(q,f[i,1]));
              if(pos, qc[pos]+=c*f[i,2],
                 pos=setsearch(q,f[i,1],1);
                 listinsert(q,f[i,1],pos);
                 listinsert(qc,c*f[i,2],pos))));
     ret;
    );
  check_equal(vector(2000,n, L(n)),
              vector(2000,n, ma(n)),
              "A061775 num vertices L() lists vs by_vpar()");

  \\ compact using recursion
  my(r(n) = my(f=factor(n)); [self()(primepi(p)) | p<-f[,1]]*f[,2] + 1;
    );
  check_equal(vector(2000,n, r(n)),
              vector(2000,n, ma(n)),
              "A061775 num vertices r() recursive vs a()");

  check_equal(normlp(Mat([20,3]),1), 20+3);
  check_equal(normlp([2,30; 20000,3000000],1), 3020032);    \\ sum terms

  \\ \\ Map and List about the same after small values,
  \\ \\ recurse faster by about 1.7x
  \\ \\ 14757.0/8377
  \\ my(nmax=200000);
  \\ gettime(); for(n=1,nmax,L(n)); print("L  ",gettime());
  \\ gettime(); for(n=1,nmax,ma(n)); print("ma ",gettime());
  \\ gettime(); for(n=1,nmax,r(n)); print("r  ",gettime());
  \\ quit;
}
{
  \\ A325544 num vertices in tree of primecode 2^n-1
  \\ needs some big factoring
  my(want=OEIS_data("A325544"));
  check_equal(vector(#want,n,n--; #vpar_from_primecode(n!)),
              want,
              "A325544 num vertices in primecode n!");
}

{
  \\ A325611 num vertices in tree of primecode 2^n-1
  my(want=OEIS_data("A325611"));
  want=want[1..min(40,#want)];      \\ needs some big factoring
  check_equal(vector(#want,n, #vpar_from_primecode(2^n-1)),
              want,
              "A325611 num vertices in primecode 2^n-1");
}

{
  \\ A214572 primecodes having 8 vertices
  my(want=OEIS_data("A214572"));
  check_equal(select(n->#vpar_from_primecode(n)==8, [1..want[#want]]),
              want,
              "A214572 having 8 vertices");

  my(vpar=vpar_unlabelled_first(8),
     m=Map());
  until((vpar=vpar_unlabelled_next(vpar)) == 'none,
        mapput(m,vpar_to_primecode(vpar),1));
  check_equal(Mat(m)[,1]~, want,
              "A214572 primecodes of 8 vertex trees");
}



\\-----------------------------------------------------------------------------
progress("A109082 Height");

\\ my line of code
\\ work downwards with set of all the primes at given depth, until no more
A109082_height(n) =
{
  my(v=factor(n)[,1],d=0);
  while(#v,
        d++; v=fold(setunion,apply(p->factor(primepi(p))[,1]~, v)));
  d;
}
{
  OEIS_check_func("A109082_height",,['offset,1]);

  my(by_vpar(n) = vpar_height(vpar_from_primecode(n)));
  check_equal(vector(1024,n, A109082_height(n)),
              vector(1024,n, by_vpar(n)),
              "A109082_height() vs by_vpar()");
}

\\-----------------------------------------------------------------------------
progress("Depths");

\\ A196056 triangle num vertices by depth of tree primecode n,
\\ excluding root
A196056_vertices_at_depth_row(n) =
{
  my(vpar=vpar_from_primecode(n),
     depths=vpar_depths_vector(vpar)[^1]); \\ without root
  Vec(vpar_INTERNAL_histogram(depths));
}
A196056_at_depth(n,k) = A196056_vertices_at_depth_row(n)[k];
{
  OEIS_check_func("A196056_vertices_at_depth_row");

  check_equal(vector(1000,n, #A196056_vertices_at_depth_row(n)),
              vector(1000,n, A109082_height(n)),
              "A196056_vertices_at_depth_row() width = A109082_height()");

  \\ this code by Francois Marques in A196056
  my(P(n) = if(n==1, 0,
               my(f=factor(n)~,
                  v=vector(#f,k, self()(primepi(f[1,k]))) );
               'x*sum(k=1, #f, (v[k]+1)*f[2,k]) );
    );
  check_equal(vector(500,n, Vecrev(P(n))[^1]),
              vector(500,n, A196056_vertices_at_depth_row(n)),
              "P(n) vs A196056_vertices_at_depth_row()");
}

{
  \\ If root included, which is an extra 1 at the start of each row:
  if(0,
     print(concat(vector(10,n,n+=2;
      Vec(vpar_INTERNAL_histogram
       (vpar_depths_vector(vpar_from_primecode(n)), 0)))));
     \\ not in OEIS: 1,1,1,1,2,1,1,1,1,1,2,1,1,1,2,1,3,1,2,2,1,2,1,1,1,1,1,1,1,1,3,1
     quit);
}

{
  \\ A184162 num chains, chain being non-empty set of comparable vertices,
  \\ so a list of ancestors, not necessarily consecutive
  \\ so sum 2^depth for each vertex

  my(by_vpar(n) = vecsum([2^d | d<-vpar_depths_vector(vpar_from_primecode(n))]));
  OEIS_check_func("A184162",by_vpar);

  my(a(n) = my(f=factor(n)); [self()(primepi(p)) | p<-f[,1] ]*f[,2]*2 + 1;
    );
  check_equal(vector(2000,n,a(n)),
              vector(2000,n,by_vpar(n)),
              "num chains compact code");

  check_equal(vector(2000,n, 1 + sum(k=1,A109082_height(n),
                                     A196056_at_depth(n,k)*2^k)),
              vector(2000,n, by_vpar(n)),
              "A184162 num chains by sum A196056 at depth");

  check_equal(vector(2000,n, my(v=A196056_vertices_at_depth_row(n));
                             1 + sum(k=1,#v, v[k]*2^k)),
              vector(2000,n, by_vpar(n)),
              "A184162 num chains by sum A196056 at depth");

  \\ this would be chains not involving the root ??
  \\ print(vector(100,n, sum(k=1,A109082_height(n), A196056_at_depth(n,k)*2^(k-1)))); quit;
  \\ not in OEIS: 0, 1, 3, 2, 7, 4, 5, 3, 6, 8, 15, 5, 9, 6, 10
}

\\ depths list
\\ concat(vector(4,n,n+=2; vecsort(vpar_depths_vector(vpar_from_primecode(n)))))
\\ apply(n->n+1,concat(vector(4,n,n+=2; vecsort(vpar_depths_vector(vpar_from_primecode(n))))))
\\ not in OEIS: 0, 1, 2, 0, 1, 1, 0, 1, 2, 3, 0, 1, 1, 2
\\ not in OEIS: 1, 2, 3, 1, 2, 2, 1, 2, 3, 4, 1, 2, 2, 3
\\ concat(vector(4,n,n+=2; vpar_depths_vector(vpar_from_primecode(n))))
\\ apply(n->n+1,concat(vector(4,n,n+=2; vpar_depths_vector(vpar_from_primecode(n)))))
\\ not in OEIS: 0, 1, 2, 0, 1, 1, 0, 1, 2, 3, 0, 1, 1, 2
\\ not in OEIS: 1, 2, 3, 1, 2, 2, 1, 2, 3, 4, 1, 2, 2, 3

\\-----------------------------------------------------------------------------
\\ A196047 "path length" = sum depths of tree primecode n

{
  my(by_vpar(n) = vecsum(vpar_depths_vector(vpar_from_primecode(n))));
  OEIS_check_func("A196047",by_vpar);

  my(a(n, d=0) = my(f=factor(n)); d + [self()(primepi(p),d+1) | p<-f[,1]] * f[,2];
    );
  check_equal(vector(2000,n,a(n)),
              vector(2000,n,by_vpar(n)));

  \\ code by Francois Marques in A196047
  my(NPl(n) = if(n==1, return([1, 0]),
                 my(f=factor(n)~, v=Mat(vector(#f,k, self()(primepi(f[1,k]))~)));
                 return([1 + sum(k=1,#f, v[1,k]*f[2,k]),
                         sum(k=1,#f, (v[1,k]+v[2,k])*f[2,k]) ] ));
    );
  check_equal(vector(2000,n,NPl(n)[2]),
              vector(2000,n,a(n)),
              "A196047 path length NPl");

  check_equal(vector(2000,n,  sum(k=1,A109082_height(n),
                                  k*A196056_at_depth(n,k))),
              vector(2000,n, a(n)),
              "A196047 path length by sum A196056 at depth");
}
{
  \\ A196068 visitation length = sum depth+1 of vartices
  my(want=OEIS_data("A196068"), \\ OFFSET=1
     by_vpar(n) = my(vpar=vpar_from_primecode(n));
                     #vpar + vecsum(vpar_depths_vector(vpar)));
  check_equal(vector(#want,primecode, by_vpar(primecode)),
              want,
              "A196068 visitation length of primecode");

  my(a(n, d=1) = my(f=factor(n)); d + [self()(primepi(p),d+1) | p<-f[,1]] * f[,2];
    );
  check_equal(vector(2000,n,a(n)),
              vector(2000,n,by_vpar(n)));

}
{
  \\ A196048 "external path length" = sum depths of childless tree primecode n
  \\ sometimes "Sackin index"
  my(want=OEIS_data("A196048"),   \\ OFFSET=1
     by_vpar(n) = my(vpar=vpar_from_primecode(n),
                     depths=vpar_depths_vector(vpar),
                     childless=vpar_childless_vertices(vpar));
     vecsum(vecextract(depths,childless)));
  check_equal(vector(#want,n, by_vpar(n)),
              want,
              Str("A196048 sum depths of childless vertices, len=",#want));

  my(b=(n,depth=0)->my(f=factor(n));
     if(n==1,depth,
        sum(i=1,matsize(f)[1], f[i,2]*self()(primepi(f[i,1]),depth+1))));
  check_equal(vector(500,n, b(n)),
              vector(500,n, by_vpar(n)));

  \\ Mir, Rossello, Rotger example figure 3 T
  \\           1
  \\          /
  \\         2   \
  \\        / \    \
  \\       3    6    9
  \\      / \  / \  / \
  \\      4 5  7 8  10 11
  \\          /
  my(vpar=[0,1,2,3,3, 2,6,6, 1,9,9],
     depths=vpar_depths_vector(vpar),
     childless=vpar_childless_vertices(vpar));
  vpar_validate(vpar);
  check_equal(vecsum(vecextract(depths,childless)), 16);
}
{
  \\ A343006 "internal path length" = sum depths of childful in primecode n
  my(want=OEIS_data("A343006"));
  check_equal(vector(#want,primecode,
                     my(vpar=vpar_from_primecode(primecode));
                     vecsum(vecextract(vpar_depths_vector(vpar),
                                       vpar_childful_vertices(vpar)))),
              want,
              "A343006 sum depths of childfuls");
}



\\-----------------------------------------------------------------------------
progress("Number of Subtrees (subcons)");

{
  \\ A184160 number of subtrees (subcons) where the root must be included
  my(want=OEIS_data("A184160"));
  my(by_vpar(n) = my(vpar=vpar_from_primecode(n));
     vpar_num_subcons(vpar)
     - vpar_num_subcons(vpar_delete_vertex(vpar,1));
    );
  check_equal(by_vpar(1), 1, "singleton");
  check_equal(by_vpar(2), 2, "path-2");
  check_equal(vector(#want,primecode,by_vpar(primecode)),
              want,
              "A184160 number of subtrees (subcons) which include the root");

  \\ similar to Andrew Howroyd in A184160
  my(a(n) = my(f=factor(n)); prod(i=1,#f~, (self()(primepi(f[i,1])) + 1)^f[i,2]);
    );
  check_equal(vector(2000,n,a(n)),
              vector(2000,n,by_vpar(n)),
              "A184160 a() vs by_vpar()");
}

{
  \\ A184161 number of subtrees (subcons)
  my(want=OEIS_data("A184161"),
     by_vpar(n) = vpar_num_subcons(vpar_from_primecode(n)));
  check_equal(vector(#want,primecode,by_vpar(primecode)),
              want,
              "A184161 number of subtrees (subcons)");

  my(quants(n) = my(with=1,without=0,f=factor(n));
     for(i=1,#f~, my(c=self()(primepi(f[i,1])));
        with *= (c[1]+1)^f[i,2];
        without += vecsum(c)*f[i,2]);
     [with,without]); \\ ways with or without root in subtree
  my(a(n) = vecsum(quants(n));
    );
  check_equal(vector(2000,n,a(n)),
              vector(2000,n,by_vpar(n)),
              "A184161 a() vs by_vpar()");
}

\\ if empty subtree were included: A184161 + 1
\\ vector(15,n, vpar_num_subcons(vpar_from_primecode(n))+1)
\\ not in OEIS: 2, 4, 7, 7, 11, 11, 12, 12, 16, 16, 16, 18, 18, 18, 22


\\-----------------------------------------------------------------------------
progress("Independent Sets");

{
  \\ A212625 indnum of tree primecode n
  my(want=OEIS_data("A212625"));
  check_equal(vector(#want,primecode,
                     vpar_indnum(vpar_from_primecode(primecode))),
              want,
              "A212625 indnum of primecode");
}
{
  \\ A212623 indpoly rows, of sets size k>=1
  \\ 1;
  \\ 2;
  \\ 1,2;
  \\ 1,1;
  \\ 0,3;
  my(want=OEIS_data("A212623"));  \\ OFFSET=1
  my(got=[], n=1);
  while(#got < #want,
        got=concat(got,Vecrev(vpar_indpoly(vpar_from_primecode(n))));
        n++);
  check_equal(got[1..#want], want,
              "A212623 indpoly rows");
}
{
  \\ A184165 indsets count of tree primecode n
  my(want=OEIS_data("A184165"));
  check_equal(vector(#want,primecode,
                     vpar_indsets_count(vpar_from_primecode(primecode))),
              want,
              "A184165 indsets count of primecode");
}
{
  \\ A228731 count indsets which include root of tree primecode n
  my(want=OEIS_data("A228731"));
  check_equal(vector(#want,primecode,
                     my(vpar=vpar_from_primecode(primecode),
                        root=vpar_root(vpar));
                     \\ root and its children deleted
                     vpar=vpar_delete_vertices(vpar,
                       concat(root, vpar_children_of_vertex(vpar,root)));
                     vpar_indsets_count(vpar)),
              want,
              "A228731 count indsets which include the root");
}
{
  \\ A228732 count indsets which exclude root of tree primecode n
  my(want=OEIS_data("A228732"));
  check_equal(vector(#want,primecode,
                     my(vpar=vpar_from_primecode(primecode));
                     vpar=vpar_delete_vertex(vpar, vpar_root(vpar));
                     vpar_indsets_count(vpar)),
              want,
              "A228732 count indsets which exclude the root");
}
{
  \\ A212624 total num vertices in all indsets of tree primecode n
  my(want=OEIS_data("A212624"));
  check_equal(vector(#want,primecode,
                     vecsum(apply(length,
                        vpar_indsets(vpar_from_primecode(primecode))))),
              want,
              "A212624 total num vertices in all indsets of primecode");
}


{
  \\ A212626 num maximum indsets of tree primecode n
  my(want=OEIS_data("A212626"));
  check_equal(vector(#want,primecode,
                vpar_indnum_and_count(vpar_from_primecode(primecode))[2]),
              want,
              "A212626 num maximum indsets of primecode");
}

is_maximal_indset(vpar,set) =
{
  if(!vpar_is_indset(vpar,set), return(0));
  for(v=1,#vpar,
     if(setsearch(set,v),next());
     my(new_set=setunion(set,[v]));
     if(vpar_is_indset(vpar,new_set), return(0)));
  1;
}
maximal_indsets(vpar) = \
  select(set->is_maximal_indset(vpar,set), vpar_indsets(vpar));
maximal_indsets_count(vpar) = #maximal_indsets(vpar);
maximal_indpoly(vpar) = \
  Polrev(vpar_INTERNAL_histogram(apply(length,maximal_indsets(vpar)),0));


{
  \\ A212627 maximal indpoly rows,
  \\ starting set size k=1, so not the 0 constant coeff
  my(want=OEIS_data("A212627"));  \\ OFFSET=1
  my(got=[], n=1);
  while(#got < #want,
        got=concat(got,Vecrev(maximal_indpoly(vpar_from_primecode(n)) / x));
        n++);
  check_equal(got[1..#want], want,
              "A212627 maximal indpoly rows");
}
{
  \\ A212628 num maximal indsets of tree primecode n
  my(want=OEIS_data("A212628"));
  check_equal(vector(#want,primecode,
                     maximal_indsets_count(vpar_from_primecode(primecode))),
              want,
              "A212628 maximal indsets of tree primecode n");
}
{
  \\ A212629 num vertices in all maximal indsets of tree primecode n
  my(want=OEIS_data("A212629"));
  check_equal(vector(#want,primecode,
    vecsum(apply(length,maximal_indsets(vpar_from_primecode(primecode))))),
              want,
              "A212620 num vertices in all maximal indsets");
}


\\-----------------------------------------------------------------------------
progress("Paths");

{
  \\ A198326 sum lengths directed paths
  my(want=OEIS_data("A198326"),
     by_vpar(n) = vecsum([d*(d+1)/2 | d<-vpar_depths_vector(vpar_from_primecode(n))]));
  check_equal(vector(#want,primecode, by_vpar(primecode)),
              want,
              "A198326 sum lengths directed paths");

  my(a(n, d=0) = my(f=factor(n)); d*(d+1)/2 + [self()(primepi(p),d+1) | p<-f[,1]] * f[,2];
    );
  check_equal(vector(2000,n,a(n)),
              vector(2000,n,by_vpar(n)));
}
{
  \\ A198325 triangle num directed paths length k >= 1
  my(want=OEIS_data("A198325"));
  my(got=[], n=1);
  while(#got < #want,
        my(depths=vpar_depths_vector(vpar_from_primecode(n)),
           hist=vpar_INTERNAL_histogram(depths,0));
        forstep(i=#hist-1,1,-1, hist[i]+=hist[i+1]);
        got=concat(got,Vec(hist[^1]));
        n++);
  check_equal(got[1..#want], want,
              "A198325 triangle num directed paths length k");
}

\\-----------------------------------------------------------------------------
progress("Initial Subtrees");

\\ Return vpar with all vertices below v deleted, but v itself retained.
delete_child_subtrees(vpar,v) =
{
  vpar_delete_vertices(vpar,
    select(c-> c!=v && vpar_is_ancestor(vpar,c,v), [1..#vpar]));
}

\\ must take all or none of siblings
num_initial_subtrees(vpar,v=vpar_root(vpar)) =
{
  my(seq=vpar_INTERNAL_seq_upwards(vpar),
     num=vectorsmall(#vpar+1));
  for(i=1,#seq,
     my(v=seq[i], p=vpar[v]);
     num[p+1] = max(1,num[p+1]) * (1+num[v+1]));
  num[1];
}
check_equal(num_initial_subtrees([0,1]), 2);
{
  \\     1
  \\    / \
  \\   2   3
  my(primecode=4,
     vpar=vpar_from_primecode(primecode));
  check_equal(vpar, [0, 1, 1]);
  check_equal(OEIS_data("A318046")[primecode], 2);
  check_equal(num_initial_subtrees(vpar), 2);
}
{
  \\        1        (o((o))(oo))
  \\      / | \
  \\     2  3  6     (o((o))(oo)), (o(o)(oo)), (oo(oo)),
  \\        |  | \   (o((o))o), (o(o)o), (ooo), o,
  \\        4  7  8
  \\        |
  \\        5
  my(primecode=70,
     vpar=vpar_from_primecode(primecode));
  check_equal(primecode, 2*5*7);
  check_equal(vpar, [0, 1, 1, 3, 4, 1, 6, 6]);
  check_equal(num_initial_subtrees(vpar), 7);
}
{
  \\    1       1      1      1     1
  \\   / \     / \    / \    / \
  \\  2   4   2   4  2   4  2   4
  \\  |   |       |  |
  \\  3   5       5  3
  my(vpar=[0,1,2,1,4],
     primecode=vpar_to_primecode(vpar));
  check_equal(primecode, 9);
  check_equal(num_initial_subtrees(vpar), 5);
  check_equal(OEIS_data("A318046")[primecode], 5);
}
{
  \\ A318046 distinct subtrees at the root
  \\ 1, 2, 3, 2, 4, 3, 3, 2, 5, 4, 5, 3, 4, 3, 7, 2

  my(want=OEIS_data("A318046")); \\ OFFSET=1
  \\ for(n=1,10,
  \\    my(vpar=vpar_from_primecode(n));
  \\    print("n="n"  ",want[n]" ",num_initial_subtrees(vpar)," ",vpar));
  check_equal(vector(#want,primecode,
                     num_initial_subtrees(vpar_from_primecode(primecode))),
              want,
              "A318046 distinct subtrees at the root");
}

\\----------------------
\\ initial_subtrees(vpar,v=vpar_root(vpar)) =
\\ {
\\   my(children=vpar_children_of_vertex(vpar,v));
\\   if(!#children, return([ [0] ]));
\\   my(child_subtrees=apply(c->initial_subtrees(vpar,c), children),
\\      ret=List([ [0] ]));
\\   print("children       ",children);
\\   print("child_subtrees ",child_subtrees);
\\   forvec(sel=apply(vec->[1,length(vec)], child_subtrees),
\\         print("sel     ",sel);
\\         my(ext=vector(#sel,i, child_subtrees[sel[i]]));
\\         print("extract ",ext);
\\         listput(ret,
\\                 vpar_to_unlabelled
\\                 (vpar_add_root_above_forest
\\                  (vpar_concat_forests(ext)))));
\\   print("return ",Set(ret));
\\   Set(ret);
\\ }
\\ check_equal(initial_subtrees([0]), [ [0] ]);
\\ check_equal(initial_subtrees([0,1]), [ [0], [0,1] ]);
\\ check_equal(initial_subtrees([0,1,2]), [ [0], [0,1], [0,1,2] ]);

\\ span(vpar) =
\\ {
\\   my(seq=vpar_INTERNAL_seq_upwards(vpar),
\\      num=vectorsmall(#vpar+1));
\\   for(i=1,#seq,
\\      my(v=seq[i], p=vpar[v]);
\\      num[p+1] = max(1,num[p+1]) * (1+num[v+1]));
\\   num[1];
\\ }
\\ {
\\   \\ A318048 span
\\   \\   terminal subtrees of initial subtrees
\\   \\ = initial subtrees of terminal subtrees
\\   \\
\\   \\          1        example 42   (o(o)(oo))
\\   \\        / | \
\\   \\       2  3  5     o, (o), (oo), (ooo), (oo(oo)), (o(o)o), (o(o)(oo))
\\   \\          |  |\
\\   \\          4  6 7
\\   my(primecode=42,
\\      vpar=[0, 1, 1, 3, 1, 5, 5]);
\\   check_equal(vpar_from_primecode(primecode), vpar);
\\   check_equal(num_initial_subtrees(vpar), 5);
\\   check_equal(vector(#vpar,v, num_initial_subtrees(vpar_subtree(vpar,v))),
\\               [5, 1, 2, 1, 2, 1, 1]);
\\   check_equal(initial_subtrees(vpar),
\\               [5, 1, 2, 1, 2, 1, 1]);
\\
\\   \\ my(want=OEIS_data("A318048")); \\ OFFSET=1
\\   \\ check_equal(vector(#want,primecode,
\\   \\                    num_initial_subtrees(vpar_from_primecode(primecode))),
\\   \\             want,
\\   \\             "A318046 distinct subtrees at the root");
\\ }

\\-----------------

\\ up to isomorphism
all_subtrees(vpar) =
{
  my(seen=Map(),
     pending=List([vpar]));
  while(#pending,
        vpar=pending[#pending]; listpop(pending);
        vpar=vpar_to_unlabelled(vpar);
        if(mapisdefined(seen,vpar),next);
        mapput(seen,vpar,1);
        for(v=1,#vpar,
           vpar[v] || next;
           my(cut=vpar); cut[v]=0;
           my(components=vpar_component_trees(cut));
           for(j=1,#components,
              listput(pending,components[j]))));
  Mat(seen)[,1]~;
}
{
  check_equal(all_subtrees([0,1,1]), [ [0], [0,1], [0,1,1] ]);

  \\ vector(15,primecode,
  \\        #all_subtrees(vpar_from_primecode(primecode)))
  \\ not in OEIS: 1, 2, 3, 3, 4, 5, 5, 4, 6, 7, 5, 7, 8, 7, 9, 5
}


\\-----------------------------------------------------------------------------
progress("Dominating Sets");

\\ How many minimum domsets:
\\ vector(20,n,n+=5; vpar_domnum_and_count(vpar_from_primecode(n))[2])
\\ not in OEIS: 4, 1, 1, 3, 3, 3, 2, 2, 2, 1, 1, 2, 8, 1, 2, 2, 1, 8, 2, 8
\\
\\ How many minimal domsets:
\\ vector(20,n,n+=5; vpar_minimal_domsets_count(vpar_from_primecode(n)))
\\ not in OEIS: 4, 2, 2, 4, 4, 4, 4, 4, 4, 7, 2, 4, 8, 2, 4, 4, 7, 8, 4, 9

{
  \\ A212630 dompoly rows
  \\ starting domset size k=1, so not the 0 constant coeff
  my(want=OEIS_data("A212630"));  \\ OFFSET=1
  my(got=[], n=1);
  while(#got < #want,
        got=concat(got,Vecrev(vpar_dompoly(vpar_from_primecode(n)) / x));
        n++);
  check_equal(got[1..#want], want,
              "A212630 dompoly rows");
}
{
  \\ A212631 num domsets of tree primecode n
  my(want=OEIS_data("A212631"));
  check_equal(vector(#want,primecode,
                     vpar_domsets_count(vpar_from_primecode(primecode))),
              want,
              "A212631 num domsets");
}
{
  \\ A212632 domnum of tree primecode n
  my(want=OEIS_data("A212632"));
  check_equal(vector(#want,primecode,
                     vpar_domnum(vpar_from_primecode(primecode))),
              want,
              "A212632 domnum of primecode");
}

\\-----------------------------------------------------------------------------
progress("Laplacian Eigenvalues");

{
  \\ A193401 Laplacian poly rows
  my(want=OEIS_data("A193401"));  \\ OFFSET=1
  my(got=[], n=1);
  while(#got < #want,
        got=concat(got,Vecrev(vpar_Laplacian_poly(vpar_from_primecode(n))));
        n++);
  check_equal(got[1..#want], want,
              "A193401 Laplacian poly rows");

  \\ Laplacian poly reversed instead
  \\ concat(vector(10,n,n+=2; Vec(vpar_Laplacian_poly(vpar_from_primecode(n)))))
  \\ not in OEIS: 1, -4, 3, 0, 1, -4, 3, 0, 1, -6, 10, -4, 0, 1, -6, 10
  \\ for(n=1,10, print(Vecrev(vpar_Laplacian_poly(vpar_from_primecode(n)))))
}
{
  \\ A193402 treecodes with Laplacian eigenvalue 2
  my(want=OEIS_data("A193402"));  \\ OFFSET=1
  check_equal(select(primecode->
    vpar_Laplacian_poly(vpar_from_primecode(primecode)) % (x-2) == 0,
                     [1..want[#want]]),
              want,
              "A193402 treecodes with Laplacian eigenvalue 2");

  \\ differs from A193405 trees with perfect matching at 343 has eigenvalue
  \\ 2 but no perfect matching
  \\         1
  \\      /  |  \
  \\    2    5    8
  \\   / \  / \  / \
  \\   3 4  6 7  9 10
  my(primecode=343,
     vpar=[0, 1, 2, 2, 1, 5, 5, 1, 8, 8]);
  check_equal(vpar_from_primecode(primecode),vpar);
  check_equal(vpar_Laplacian_poly(vpar),
              x * (x-5) * (x-2) * (x-1)^3 * (x^2 - 4*x + 1)^2 );
}
{
  \\ A202852 treecodes with Laplacian eigenvalue 2 without perfect matching
  my(want=OEIS_data("A202852"));  \\ OFFSET=1
  check_equal(select(primecode->
                     my(vpar=vpar_from_primecode(primecode));
                     ! vpar_has_perfect_matching(vpar)
                     && vpar_Laplacian_poly(vpar) % (x-2) == 0,
                     [1..want[#want]]),
              want,
              "A202852 Laplacian eigenvalue 2 without perfect matching");
}


\\-----------------------------------------------------------------------------
progress("Perfect Matching");

\\ primecodes with/without near perfect matching (odd num vertices)
\\ select(n->vpar_has_perfect_matching(vpar_from_primecode(n),1), [1..35])
\\ select(n->!vpar_has_perfect_matching(vpar_from_primecode(n),1), [1..35])
\\ not in OEIS: 1, 3, 4, 9, 10, 11, 12, 13, 14, 17, 25, 27, 30, 33, 35
\\ not in OEIS: 2, 5, 6, 7, 8, 15, 16, 18, 19, 20, 21, 22, 23, 24

\\ select(n->vpar_has_perfect_matching(vpar_delete_vertex(vpar_from_primecode(n),1)), [1..100])
\\ not in OEIS: 3, 9, 11, 13, 27, 33, 39, 47, 61, 79, 81, 83, 99
\\ vpar_from_primecode(3) == [0,1,2]

{
  check_equal(vpar_from_primecode(2), [0,1]);
  check_equal(2*prime(2), 6);
  check_equal(vpar_from_primecode(6), [0,1,1,3]);
  check_equal(prime(2), 3);
  check_equal(prime(6), 13);
  check_equal(3*3*13, 117);
  check_equal(vpar_from_primecode(117), [0,1,2, 1,4, 1,6,6,8]);
}

{
  \\ A193405 primecodes with perfect matching (so even num vertices too)
  my(want=OEIS_data("A193405"));
  check_equal(select(primecode->
                     vpar_has_perfect_matching(vpar_from_primecode(primecode)),
                     [1..want[#want]]),
              want,
              "A193405 primecode with perfect matching");
}
{
  \\ A193406 primecodes without perfect matching
  \\ any odd num vertices, or even but not perfect
  my(want=OEIS_data("A193406"));
  check_equal(select(primecode->
                     !vpar_has_perfect_matching(vpar_from_primecode(primecode)),
                     [1..want[#want]]),
              want,
              "A193405 primecode without perfect matching");
}

print("read a193405-perfect.gp");
read("a193405-perfect.gp");
{
  check_equal(vector(10000,n, is_A193405(n)),
              vector(10000,n, vpar_has_perfect_matching(vpar_from_primecode(n))),
              "a193405-perfect.gp is_A193405(n) vs vpar_has_perfect_matching()");
}

{
  my(imperfection(n) = my(f=factor(n),v=1,c); for(i=1,matsize(f)[1],
       if((c=self()(primepi(f[i,1])))<0 || (v-=c*f[i,2])<0, return(-1))); v);
  my(is(n) = !imperfection(n);
    );
  check_equal(vector(10000,n, is(n)),
              vector(10000,n, vpar_has_perfect_matching(vpar_from_primecode(n))),
              "perfect matching compact vs vpar_has_perfect_matching()");
}


\\-----------------------------------------------------------------------------
progress("Equimatchable");

\\ select(n->vpar_is_equimatchable(vpar_from_primecode(n)), [1..35])
\\ not in OEIS: 1, 2, 3, 4, 7, 8, 9, 10, 11, 16, 19, 20, 21, 27, 29, 32, 34

print("read a750002-equimatchable.gp");
read("a750002-equimatchable.gp");
{
  check_equal(vector(10000,n, is(n)),
              vector(10000,n, vpar_is_equimatchable(vpar_from_primecode(n))),
              "a750002-equimatchable.gp is(n) vs vpar_maximal_matchings_count()");
}
kill(is)
kill(equimatchable_state);


\\-----------------------------------------------------------------------------
progress("Generalized Bethe Trees");
\\ all vertices at same depth have same degree

test_is_generalized_Bethe(vpar) =
{
  my(depths=vpar_depths_vector(vpar),
     num_children=vpar_num_children_vector(vpar),
     want=vectorsmall(vecmax(depths)+1,i,-1));
  for(v=1,#vpar,
     my(d=depths[v]+1);
     if(want[d] >= 0,
        want[d] == num_children[v] || return(0),
        want[d] = num_children[v]));
  1;
}
{
  my(want=OEIS_data("A214577"));
  check_equal(select(primecode->
                     test_is_generalized_Bethe(vpar_from_primecode(primecode)),
                     [2..want[#want]]),
              want);

  \\ primecodes of not generalized Bethe trees
  \\ not in OEIS: 6, 10, 12, 13, 14, 15, 18, 20, 21, 22, 24, 26, 28, 29, 30, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47
}
{
  \\ row of num of vertices at depth (excluding root), or single 0 if not generalized Bethe
  my(want=OEIS_data("A214578"),
     got=[]);
  for(n=1,#want,
     if(#got>=#want,break);
     my(vpar=vpar_from_primecode(n));
     if(n>1 && test_is_generalized_Bethe(vpar),
        got=concat(got,Vecrev(vpar_INTERNAL_histogram(vpar_depths_vector(vpar)[^1]))),
        got=concat(got,0)));
  got=Vec(got,#want);
  check_equal(got, want, "A214578");
}

\\ Bethe_counts=vector(50);
\\ Bethe_find(n,prev) =
\\ {
\\   Bethe_counts[n]++;
\\   forstep(t=prev,#counts-n,prev,
\\      self()(n+t,t));
\\ }
\\ for(prev=1,#counts, Bethe_find(prev,prev));
\\ Bethe_counts
\\ partition_is_Bethe(p) = for(i=2,#p, if(p[i]%p[i-1],return(0))); 1;
\\ vector(20,n, my(count=0); forpart(p=n, count+=partition_is_Bethe(p)); count)
\\ A003238
\\ 1, 2, 3, 5, 6, 10, 11, 16, 19, 26, 27, 40, 41, 53, 61, 77, 78, 104, 105, 134


\\-----------------------------------------------------------------------------
progress("Uniform Trees");
\\ At each vertex, isomorphic subtrees occur with same multiplicity.

\\ Return a Map "elem -> count" where count is the number of times elem
\\ occurs in vector vec[].
test_histogram_map(vec) =
{
  my(m=Map());
  for(i=1,#vec,
     my(c=0);mapisdefined(m,vec[i],&c); mapput(m,vec[i],c+1));
  m;
}
test_is_uniform(vpar) =
{
  my(ids=vpar_automorphism_ids_vector(vpar),
     siblings=vpar_siblings_sets(vpar));
  for(i=1,#siblings,
     my(m=test_histogram_map(vecextract(ids,siblings[i])));
     #Set(Mat(m)[,2])==1 || return(0));
  1;
}
Matula_is_uniform(n) =
{
  if(n>1, my(f=factor(n)); #Set(f[,2])==1 || return(0);
     foreach(f[,1],p, self()(primepi(p)) || return(0))); 1;
}
{
  \\     1
  \\   / | \
  \\  2  3  4
  \\        |
  \\        5
  check_equal(vpar_from_primecode(12), [0, 1, 1, 1, 4]);
  check_equal(vpar_automorphism_ids_vector([0, 1, 1, 1, 4]), [4, 1, 1, 2, 1]);
  check_equal(factor(13), Mat([13,1]));
  check_equal(primepi(13), 6);
  check_equal(factor(6), [2,1; 3,1]);

  \\ A317710 primecodes of uniform trees
  \\ factor n, exponents of primes must be all the same,
  \\ which means power of a square-free
  my(want=OEIS_data("A317710"));
  check_equal(select(primecode->
                     test_is_uniform(vpar_from_primecode(primecode)),
                     [1..want[#want]]),
              want,
              "A317710 primecode of uniform tree");
}
{
  \\ A317711 primecodes of not uniform trees
  my(want=OEIS_data("A317711"));
  check_equal(select(primecode->
                     !test_is_uniform(vpar_from_primecode(primecode)),
                     [1..want[#want]]),
              want,
              "A317711 primecode of not uniform tree");
}
{
  check_equal(vector(1000,n, Matula_is_uniform(n)),
              vector(1000,n, test_is_uniform(vpar_from_primecode(n))),
              "Matula_is_uniform()");
}

{
  \\ A317712 number of uniform trees of n vertices
  my(want=OEIS_data("A317712"));
  want=want[1..min(#want,11)];  \\ bit slow, limit
  check_equal(vector(#want,n,
                     my(vpar=vpar_unlabelled_first(n),
                        count=0);
                     until((vpar=vpar_unlabelled_next(vpar))=='none,  \\ trees
                           count+=test_is_uniform(vpar));
                     count),
              want,
              "A317712 num uniform rooted trees of n vertices");
}

test_is_free_uniform(vpar) =
{
  my(ids=vpar_automorphism_ids_vector
         (vpar_INTERNAL_for_free_automorphisms(vpar))[1..#vpar],
     siblings=vpar_siblings_sets(vpar));
  for(i=1,#siblings,
     my(m=test_histogram_map(vecextract(ids,siblings[i])));
     #Set(Mat(m)[,2])==1 || return(0));
  1;
}
{
  \\ not in OEIS: 1, 2, 2, 3, 9, 12, 28, 56, 123, 249
  \\ vector(12,n,
  \\        my(vpar=vpar_free_first(n),
  \\           count=0);
  \\        until((vpar=vpar_free_next(vpar))=='none,  \\ trees
  \\              count+=test_is_free_uniform(vpar));
  \\        count)
}


\\-----------------------------------------------------------------------------
progress("Aperiodic Trees");
\\ At each vertex, gcd of isomorphic subtree multiplicities is 1.

test_is_aperiodic(vpar) =
{
  my(ids=vpar_automorphism_ids_vector(vpar),
     siblings=vpar_siblings_sets(vpar));
  for(i=1,#siblings,
     if(#Set(siblings[i]) < 2, next);
     my(m=test_histogram_map(vecextract(ids,siblings[i])));
     gcd(Mat(m)[,2])==1 || return(0));
  1;
}
Matula_is_aperiodic(n) =
{
  if(n>1, my(f=factor(n)); gcd(f[,2])==1 || return(0);
     foreach(f[,1],p, self()(primepi(p)) || return(0))); 1;
}
{
  \\ A303431 primecodes of aperiodic trees
  my(want=OEIS_data("A303431"));
  check_equal(select(primecode->
                     test_is_aperiodic(vpar_from_primecode(primecode)),
                     [1..want[#want]]),
              want,
              "A317710 primecode of aperiodic tree");

  \\ primecodes of not aperiodic trees
  \\ not in OEIS: 4, 7, 8, 9, 14, 16, 17, 19, 21, 23, 25, 27, 28, 32, 34, 35

  check_equal(vector(1000,n, Matula_is_aperiodic(n)),
              vector(1000,n, test_is_aperiodic(vpar_from_primecode(n))),
              "Matula_is_aperiodic()");
}

\\-----------------------------------------------------------------------------
\\ Total Domination

\\ vector(20,n,n+=5; vpar_totdomnum(vpar_from_primecode(n)))
\\ not in OEIS: 2, 2, 2, 3, 3, 3, 2, 2, 2, 4, 2, 2, 3, 2, 3, 3, 4, 3, 2, 4
\\
\\ vector(20,n,n+=5; vpar_totdomnum_and_count(vpar_from_primecode(n))[2])
\\ not in OEIS: 1, 3, 3, 1, 1, 1, 1, 1, 1, 4, 4, 1, 1, 4, 1, 1, 4, 1, 1, 3


\\-----------------------------------------------------------------------------
progress("Degrees");

{
  \\ A196046 max degree
  my(want=OEIS_data("A196046"));
  check_equal(vector(#want,primecode,
          vecmax(vpar_degrees_vector(vpar_from_primecode(primecode)))),
              want,
              "A196046 max degree");
}

\\ edge degree = num edges adjacent, = vertex degree - 1 each end
test_edge_degrees_vector(vpar) =
{
  my(degrees=vpar_degrees_vector(vpar));
  vector(#vpar,v, if(vpar[v], degrees[v] + degrees[vpar[v]] - 2, 'none));
}
{
  my(want=OEIS_data("A191517")); \\ OFFSET=2
  check_equal(vector(#want,primecode,primecode++;
    vecmax(test_edge_degrees_vector(vpar_from_primecode(primecode))[^1])),
              want,
              "A191517 max edge degree");
}
A191516_row(n) =
{
  Vec(vpar_INTERNAL_histogram
      (test_edge_degrees_vector(vpar_from_primecode(n))[^1]));
}
{
  my(want=OEIS_data("A191516"));  \\ OFFSET=3
  my(got=[], n=3);
  while(#got < #want, got=concat(got,A191516_row(n)); n++);
  check_equal(got[1..#want], want,
              "A191516 triangle edge degrees");
}

{
  my(want=OEIS_data("A198332")); \\ OFFSET=1
  check_equal(vector(#want,primecode,
    vecsum(test_edge_degrees_vector(vpar_from_primecode(primecode))[^1])),
              want,
              "A198332 Platt index, total edge degree");
}

\\------------------
{
  \\ A196046 max vertex degree of tree primecode n
  my(want=OEIS_data("A196046"),
     by_vpar(n) = vecmax(vpar_degrees_vector(vpar_from_primecode(n))));
  check_equal(vector(#want,primecode, by_vpar(primecode)),
              want,
              "A196046 max vertex degree in primecode");

  \\ at root no 1+ as no parent
  my(a = (n)->
     my(m=factor(n),v=m[,1]~,ret=vecsum(m[,2]));
     while(#v,
           m=factor(primepi(v[#v])); v=v[^#v];   \\ peel from end
           ret=max(ret,1+vecsum(m[,2]));
           v=setunion(v,m[,1]~));
     ret;
    );
  check_equal(vector(1024,n, a(n)),
              vector(1024,n, by_vpar(n)),
              "A196046 max vertex degree in primecode, by factoring");
}


\\-----------------------------------------------------------------------------
\\ A078442 distance down to first branch, how many prime of prime of etc

test_distance_to_first_branch(vpar) =
{
  my(v=vpar_root(vpar),ret=0,c);
  while(#(c=vpar_children_of_vertex(vpar,v))==1,
        v=c[1];
        ret++);
  ret;
}
{
  my(want=OEIS_data("A078442")); \\ OFFSET=1
  check_equal(vector(#want,primecode,
                     test_distance_to_first_branch(vpar_from_primecode(primecode))),
              want,
              "A078442 distance down to first branch");
}

\\-------------------
{
  \\ A196067 num degree=1 vertices in of tree primecode n
  my(want=OEIS_data("A196067"));
  check_equal(vector(#want,primecode,
                     my(vpar=vpar_from_primecode(primecode),
                        degrees=vpar_degrees_vector(vpar));
                     sum(v=1,#vpar, degrees[v]==1)),
              want,
              "A196067 num degreee=1 of primecode");
}
{
  \\ A198323 degrees 1 and K only
  my(want=OEIS_data("A198323"));
  check_equal(select(primecode->
              #Set(vpar_degrees_vector(vpar_from_primecode(primecode))) <= 2,
                     [1..want[#want]]),
              want,
              "A198323 primecodes with degrees 1 and K only");
}

\\ how many distinct degrees
\\ vector(30,primecode, \
\\   #Set(vpar_degrees_vector(vpar_from_primecode(primecode))))
\\ not in OEIS: 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 2, 2, 3, 3, 2, 3, 3


\\-----------------------------------------------------------------------------
progress("Total Walk Count");

{
  \\ A206486 total walk count
  \\ all walks lengths 1..n-1, including with backtracking
  my(want=OEIS_data("A206486"));
  check_equal(vector(#want,primecode,
                     my(vpar=vpar_from_primecode(primecode),
                        m=vpar_adjacency_matrix(vpar));
                     vecsum(Vec(concat(Vec(sum(k=1,#vpar-1,m^k)))))),
              want,
              "A206486 total walk count");
}

\\-----------------------------------------------------------------------------
\\ A320230  everywhere childful siblings are same (isomorphic) subtrees
\\          childless vertices freely anywhere
\\ so everywhere prime powers, ignoring all factors of 2
\\
\\ OEIS_data("A320230")
\\ 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, ...

is_A320230_by_ids(n) =
{
  my(vpar=vpar_from_primecode(n),
     ids=vpar_automorphism_ids_vector(vpar),
     want=vector(#vpar));
  forstep(v=#vpar,2,-1,
     if(ids[v]==1, next); \\ childless
     if(want[vpar[v]],
        want[vpar[v]]==ids[v] || return(0),
        want[vpar[v]]=ids[v]));
  1;
}
is_A320230(n) =
{
  \\ my line of code in A320230
  while((n>>=valuation(n,2)) > 1,
        isprimepower(n,&n) || return(0); n=primepi(n)); 1;
}
{
  check_equal(vector(2000,n, is_A320230(n)),
              vector(2000,n, is_A320230_by_ids(n)),
              "is_A320230()");
}
{
  my(want=OEIS_data("A320230"));
  check_equal(select(is_A320230,[1..want[#want]]), want,
              "A320230 primecodes where non-leaf branches same");
  \\ want=OEIS_data("A320230",'bfile);
  \\ check_equal(select(is_A320230,[1..want[#want]]), want,
  \\         Str("A320230 primecodes where non-leaf branches same, b-file length",
  \\             #want));
}

\\ select(n->!is_A320230_by_ids(n),[1..200])
\\ not in OEIS: 15, 21, 30, 33, 35, 39, 42, 45, 47, 51, 55, 57, 60, 63

\\ cf
\\ A000961  prime powers, including 1 (which isprimepower() does not include)
\\ select(isprimepower,[1..200])
{
  my(want=OEIS_data("A000961"));
  check_equal(select(n->n==1||isprimepower(n), [1..want[#want]]),
              want,
              "A000961 prime powers");

  check_equal(isprimepower(1), 0, "isprimepower() of 1 = false");
  check_equal(isprimepower(3), 1);
  check_equal(isprimepower(9), 2);
  check_equal(isprimepower(27), 3);
}

\\ A336101  exactly 1 odd prime factor
is_A336101(n) = isprimepower(n>>valuation(n,2));
{
  my(want=OEIS_data("A336101"));
  check_equal(select(is_A336101, [1..want[#want]]),
              want,
              "A336101 one odd prime factor");

  check_equal(vector(20,k,k--; is_A336101(2^k)),
              vector(20,k,k--; 0),
              "is_A336101() not powers-of-2");
}

\\ vector of all m such that prime(m) divides n = A304038
test_prime_indices(n) = apply(primepi,factor(n)[,1]~);
A304038_row = test_prime_indices;
{
  my(want=OEIS_data("A304038"));
  my(got=[], n=0);
  while(#got < #want, got=concat(got,A304038_row(n)); n++);
  check_equal(got[1..#want], want,
              "A304038 rows of prime indices");
}

\\ select(n-> n>>=valuation(n,2); n==1 || isprimepower(n), [5..200])
\\ not A324109 it has 2^i*p^j but also more
\\ A070776 Numbers n such that number of terms in the n-th cyclotomic polynomial equals largest prime factor of n.
\\  2^i*p^j, i>=0 and j>=0,  and i>0 or j>0 to exclude 1
is_A070776(n) = if(n<2,0, n>>=valuation(n,2); n==1 || isprimepower(n));
{
  my(want=OEIS_data("A070776"));
  check_equal(select(is_A070776, [0..want[#want]]),
              want,
              "A070776 2^i*p^j");
}
{
  check_equal(vector(2000,n, is_A320230(n)),
              vector(2000,n,
                     if(n==1,1,
                        my(v=test_prime_indices(n));
                        is_A070776(n) && sum(i=1,#v,is_A320230(v[i])==1)==#v)),
              "is_A320230() vs is_A070776 and test_prime_indices");
}

\\ select(n-> n>>=valuation(n,2); !(n==1 || isprimepower(n)), [1..200])
\\ A070537 cyclotomic polynomial has more terms than the largest prime factor of n


\\ select(n-> n>>=valuation(n,2); !isprimepower(n), [1..200])
\\ not in OEIS: 4, 8, 15, 16, 21, 30, 32, 33, 35, 39, 42, 45, 51
\\ not A098904 2* giving even numbers whose num distinct prime factors is odd


\\-----------------------------------------------------------------------------
progress("Re-Rooting");

{
  \\ A257539 smallest primecode re-rooting
  my(want=OEIS_data("A257539"), m=Map());
  check_equal(vector(#want,primecode,
                     my(vpar=vpar_from_primecode(primecode),
                        prev);
                     vpar=vpar_to_free(vpar);
                     if(!mapisdefined(m,vpar, &prev),
                        mapput(m,vpar, prev=primecode));
                     prev),
              want,
              "A257539 smallest primecode re-rooting, by successive");

  check_equal(vector(#want,primecode,
                     my(vpar=vpar_from_primecode(primecode));
                     vecmin(vector(#vpar,r, vpar_to_primecode(vpar_reroot(vpar,r))))),
              want,
              "A257539 smallest primecode re-rooting, by vpar_reroot()");
}
{
  \\ \\ A?????? biggest primecode re-rooting
  \\ my(want=OEIS_data("A257539"));
  \\ check_equal(vector(#want,primecode,
  \\                    my(vpar=vpar_from_primecode(primecode));
  \\                    vecmax(vector(#vpar,r, vpar_to_primecode(vpar_reroot(vpar,r))))),
  \\             want,
  \\             "A?????? biggest primecode re-rooting");
  \\ not in OEIS: 1, 2, 4, 4, 6, 6, 8, 8, 11, 11, 11, 17, 17, 17, 31, 19, 17
}
{
  \\ A235121 rows of all rootings of tree primecode n (including self),
  \\ so list of isomorphic to n as free trees
  my(want=OEIS_data("A235121"));
  my(got=[], primecode=1);
  while(#got < #want,
        my(vpar=vpar_from_primecode(primecode));
        got=concat(got,
                   Set(vector(#vpar,r, vpar_to_primecode(vpar_reroot(vpar,r)))));
        primecode++);
  check_equal(got[1..#want],
              want,
              "A235121 triangle of rerootings of primecode");
}
{
  \\ A235122 num distinct rootings of tree primecode n,
  \\ which is num orbits as a free tree
  my(want=OEIS_data("A235122"));
  check_equal(vector(#want,primecode,
                     vpar_num_free_orbits(vpar_from_primecode(primecode))),
              want,
              "A235122 num distinct rootings of primecode, by orbits");

  check_equal(vector(256,primecode,
                vpar_num_free_orbits(vpar_from_primecode(primecode))),
              vector(256,primecode,
                my(vpar=vpar_from_primecode(primecode));
                #Set(vector(#vpar,r,
                            vpar_to_primecode(vpar_reroot(vpar,r))))),
              "A235122 num distinct rootings of primecode, by Set");
}


\\-----------------------------------------------------------------------------
progress("Protection Number");
\\ "escape distance"

{
  \\ A184166 protection number, depth of shallowest leaf
  my(want=OEIS_data("A184166"));
  check_equal(vector(#want,primecode,
                     vpar_protnum(vpar_from_primecode(primecode))),
              want,
              "A184169 maximum subtree height in tree primecode n");
}
{
  \\ A184169 maximum vertex protection number
  my(want=OEIS_data("A184169"));
  check_equal(vector(#want,primecode,
                     vecmax(vpar_protnums_vector
                            (vpar_from_primecode(primecode)))),
              want,
              "A184169 maximum vertex protection number in tree primecode n");
}
{
  \\ A184170 how many vertices with the maximum vertex protection number
  my(want=OEIS_data("A184170"));
  check_equal(vector(#want,primecode,
                     my(protnums=vpar_protnums_vector
                                 (vpar_from_primecode(primecode)),
                        m=vecmax(protnums));
                     sum(v=1,#protnums, protnums[v]==m)),
              want,
    "A184170 how many vertices with the maximum vertex protection number");
}
{
  \\ A184168 total of vertex protection numbers
  my(want=OEIS_data("A184168"));
  check_equal(vector(#want,primecode,
                vecsum(vpar_protnums_vector(vpar_from_primecode(primecode)))),
              want,
              "A184168 total of vertex protection numbers");
}


\\-----------------------------------------------------------------------------
\\ "Star Vertices" all children are childless

{
  \\ A214566 num star vertices
  my(want=OEIS_data("A214566")); \\ OFFSET=1
  check_equal(vector(#want,primecode,
                     my(vpar=vpar_from_primecode(primecode),
                        heights=vpar_subtree_heights_vector(vpar));
                     sum(v=1,#heights, heights[v]==1)),
              want,
              "A214566 num star vertices");
}


\\-----------------------------------------------------------------------------
progress("Even Num Children Everywhere");

is_A298126_by_num_children(n) = \
  vpar_num_children_vector(vpar_from_primecode(n)) % 2 == 0;

\\ even num children everywhere
is_A298126_compact(n) =
{
  my(q=[], f=factor(n));
  while(#f[,1],
        if(vecsum(f[,2])%2, return(0));
        q=setunion(q[^#q], f[,1]~);
        f=factor(primepi(q[#q])));
  1;
}
{
  \\ A298126 primecodes with vertices even num children everywhere
  my(want=OEIS_data("A298126"), \\ OFFSET=1
     got=List([]));
  for(primecode=1,oo, if(#got>=#want,break);
     if(is_A298126_by_num_children(primecode),
        listput(got,primecode)));
  check_equal(Vec(got), want,
              "A298126 those with even num children");

  check_equal(vector(2000,n, is_A298126_compact(n)),
              vector(2000,n, is_A298126_by_num_children(n)),
              "is_A298126_compact()");
}


\\-----------------------------------------------------------------------------
progress("Odd Num Children Everywhere");

test_vector_is_all_odd_or_zero(v) = \
  for(i=1,#v, v[i]==0 || v[i]%2!=0 || return(0)); 1;
is_A298120_by_num_children(n) =
{
  test_vector_is_all_odd_or_zero
  (vpar_num_children_vector(vpar_from_primecode(n)));
}

\\ odd num children everywhere
is_A298120(n) =
{
  \\ if(n>1,
  \\    my(f=factor(n)); vecsum(f[,2])%2 || return(0);
  \\    foreach(f[,1],p, self()(primepi(p)) || return(0)));
  \\ 1;

  \\ my(q=[n]);
  \\ while(q[#q]>1, my(f=factor(q[#q]));
  \\       vecsum(f[,2])%2 || return(0);
  \\       q=setunion(q[^#q], apply(primepi,f[,1]~)));
  \\ 1;

  \\ \\ with setunion before primepi
  \\ if(n>1,
  \\    my(f=factor(n)); vecsum(f[,2])%2 || return(0);
  \\    my(q=f[,1]~);
  \\    while(q[#q]>2, f=factor(primepi(q[#q]));
  \\          vecsum(f[,2])%2 || return(0);
  \\          q=setunion(q[^#q], f[,1]~)));
  \\ 1;

  my(q=[], f=factor(n));
  while(#f[,1],
        vecsum(f[,2])%2 || return(0);
        q=setunion(q[^#q], f[,1]~);
        f=factor(primepi(q[#q])));
  1;
}
{
  \\ A298120 primecodes with vertices odd num children or no children
  my(want=OEIS_data("A298120"), \\ OFFSET=1
     got=List([]));
  for(primecode=1,oo, if(#got>=#want,break);
     if(is_A298120_by_num_children(primecode),
        listput(got,primecode)));
  check_equal(Vec(got), want,
              "A298120 those with vertices odd degree");

  check_equal(is_A298120_by_num_children(4), 0, "is_A298120_by_num_children() 4");
  check_equal(is_A298120(4),                 0, "is_A298120() 4");
  check_equal(vector(2000,n, is_A298120(n)),
              vector(2000,n, is_A298120_by_num_children(n)));
}


\\-----------------------------------------------------------------------------

\\ not in OEIS: 4, 5, 6, 9, 10, 11, 15, 16, 19, 22, 24, 25, 31, 33
\\ test_vector_is_all_even_or_1(v) = \
\\   for(i=1,#v, v[i]==1 || v[i]%2==0 || return(0)); 1;
\\ {
\\   \\ primecodes with vertices even degree
\\   my(want=OEIS_data("A190175"), \\ OFFSET=1
\\      got=List([]));
\\   for(primecode=2,oo, if(#got>=#want,break);
\\      if(test_vector_is_all_even_or_1
\\           (vpar_degrees_vector(vpar_from_primecode(primecode))),
\\         listput(got,primecode)));
\\   check_equal(Vec(got), want,
\\               "A190175 those with all vertices even degree");
\\ }

test_vector_is_all_odd(v) = \
  for(i=1,#v, v[i]%2!=0 || return(0)); 1;
{
  \\ A190175 primecodes with vertices odd degree
  my(want=OEIS_data("A190175"), \\ OFFSET=1
     got=List([]));
  for(primecode=1,oo, if(#got>=#want,break);
     if(test_vector_is_all_odd
          (vpar_degrees_vector(vpar_from_primecode(primecode))),
        listput(got,primecode)));
  check_equal(Vec(got), want,
              "A190175 those with all vertices even degree");
}



\\-----------------------------------------------------------------------------
progress("vpar_from_primecode() invariants");


{
  \\ A198328 strip childless vertices, but leave a singleton alone
  my(want=OEIS_data("A198328")); \\ OFFSET=1
  check_equal(vector(#want,primecode,
                     if(primecode==1,1,
                        my(vpar=vpar_from_primecode(primecode));
                        vpar_to_primecode
                        (vpar_delete_vertices(vpar,
                                              vpar_childless_vertices(vpar))))),
              want,
              "A198328 strip childless vertices");
}


{
  \\ A317713 num distinct subtrees
  \\ Take a vertex and all below it, how many distinct rooted trees.
  \\ This how many distinct prime factors seen when factoring all the way down.
  my(want=OEIS_data("A317713"),
     by_ids(n) = #Set(vpar_automorphism_ids_vector(vpar_from_primecode(n)));
     );
  check_equal(vector(#want,primecode, by_ids(primecode)),
              want,
              "A317713 num distinct rooted subtrees in primecode");

  check_equal(vector(500,n,
                     my(vpar=vpar_from_primecode(n));
                     #Set(vector(#vpar,v, vpar_to_unlabelled(vpar_subtree(vpar,v))))),
              vector(500,n, by_ids(n)),
              "A317713 num distinct rooted subtrees by vpar_to_unlabelled() canonical");
}

{
  \\ A182907 rows of degree sequence of tree primecode n
  \\ degrees in increasing order
  my(want=OEIS_data("A182907"));
  my(got=[], primecode=1);
  while(#got < #want,
        got=concat(got,
              vecsort(vpar_degrees_vector(vpar_from_primecode(primecode))));
        primecode++);
  check_equal(got[1..#want],
              want,
              "A182907 triangle of rerootings of primecode");
}

{
  \\ A190174 num vertices even degree of tree primecode n
  my(want=OEIS_data("A190174"));
  check_equal(vector(#want,primecode,
                     my(vpar=vpar_from_primecode(primecode),
                        degrees=vpar_degrees_vector(vpar));
                     #vpar - vecsum(degrees%2)),
              want,
              "A190174 num vertices even degree in primecode");
}

{
  \\ A198337 radius of tree primecode n
  my(want=OEIS_data("A198337"));
  check_equal(vector(#want,primecode,
                     vpar_radius(vpar_from_primecode(primecode))),
              want,
              "A198337 radius of primecode");
}


{
  \\ A196058 diameter of tree primecode n
  my(want=OEIS_data("A196058"));
  check_equal(vector(#want,primecode,
                     vpar_diameter(vpar_from_primecode(primecode))),
              want,
              "A196058 diameter of primecode");

  want=OEIS_data("A196058",'bfile);
  check_equal(vector(#want,primecode,
                     vpar_diameter(vpar_from_primecode(primecode))),
              want,
              Str("b196058.txt diameter of primecode, size ",#want));
}
{
  \\ A196049 num "branching vertices" is degree>=3s, of tree primecode n
  \\ goes as if free tree so the root must be 3 children to be branching
  my(want=OEIS_data("A196049"));
  check_equal(vector(#want,primecode,
                     my(vpar=vpar_from_primecode(primecode),
                        degrees=vpar_degrees_vector(vpar));
                     sum(v=1,#vpar, degrees[v]>=3)),
              want,
              "A196049 num degree>=3 of primecode");
}



{
  \\ A001222 bigomega num children of the root of tree primecode n
  my(want=OEIS_data("A001222"));
  check_equal(vector(#want,n, bigomega(n)),
              want,
              "A001222 bigomega()");

  check_equal(vector(1024,primecode,
                     my(vpar=vpar_from_primecode(primecode));
                     vpar_num_children_of_vertex(vpar,vpar_root(vpar))),
              vector(1024,primecode, bigomega(primecode)),
              "bigomega() num children of root of primecode");
}
{
  \\ A196052 sum degrees children of root of tree primecode n
  my(want=OEIS_data("A196052"));
  my(a_by_vpar = (primecode)->
     my(vpar=vpar_from_primecode(primecode),
        degrees=vpar_degrees_vector(vpar),
        children=vpar_children_of_vertex(vpar,vpar_root(vpar)));
     vecsum(vecextract(degrees,children)));
  check_equal(vector(#want,n, a_by_vpar(n)),
              want,
              "A196052 sum degrees children of root of primecode");

  \\ my(A196052=(n)->
  \\ scalar product for multiplicities
  \\ my line of code in A196052
  check_equal(vector(2^12,n,
        my(m=factor(n)); [bigomega(primepi(p))+1 | p<-m[,1]] * m[,2]
                    ),
              vector(2^12,n, a_by_vpar(n)),
              "A196052 by scalar product");

  \\ rows of factor matrix
  check_equal(vector(2^12,n,
        \\ vecsum(apply(v->(1+bigomega(primepi(v[1])))*v[2], Col(factor(n))))
        vecsum([(1+bigomega(primepi(v[1])))*v[2] | v<-Col(factor(n))])
                    ),
              vector(2^12,n, a_by_vpar(n)),
              "A196052 by matrix rows");

  \\ with sum() over matrix
  check_equal(vector(2^12,n,
 my(m=factor(n)); sum(i=1,matsize(m)[1], (1+bigomega(primepi(m[i,1])))*m[i,2])
                    ),
              vector(2^12,n, a_by_vpar(n)),
              "A196052 by sum()");
}
{
  \\ A196057 num sibling pairs of tree primecode n
  my(want=OEIS_data("A196057"));
  check_equal(vector(#want,primecode,
                     my(vpar=vpar_from_primecode(primecode),
                        sets=vpar_siblings_sets(vpar));
                     sum(i=1,#sets, binomial(#sets[i],2))),
              want,
              "A196057 num sibling pairs of primecode");
}
{
  \\ A196053 Zagreb index of tree primecode n
  \\ sum degree[v]^2
  my(want=OEIS_data("A196053"));
  check_equal(vector(#want,primecode,
                     my(vpar=vpar_from_primecode(primecode));
                     vecsum(apply(sqr,vpar_degrees_vector(vpar)))),
              want,
              "A196053 Zagreb index of primecode");
}
{
  \\ A196054 second Zagreb index of tree primecode n
  \\ sum degree[u]*degree[v] across edges
  my(want=OEIS_data("A196054"));
  check_equal(vector(#want,primecode,
                     my(vpar=vpar_from_primecode(primecode),
                        degrees=vpar_degrees_vector(vpar));
                     sum(v=1,#vpar, if(vpar[v], degrees[v]*degrees[vpar[v]]))),
              want,
              "A196054 second Zagreb index of primecode");
}
{
  \\ A196063 Narumi-Katayama index of tree primecode n
  \\ prod degrees
  my(want=OEIS_data("A196063"));
  check_equal(vector(#want,primecode,
                     my(vpar=vpar_from_primecode(primecode),
                        degrees=vpar_degrees_vector(vpar));
                     prod(v=1,#vpar, degrees[v])),
              want,
              "A196063 Narumi-Katayama index of tree primecode");
}
{
  \\ A196065 first multiplicative Zagrep index of tree primecode n
  \\ prod degrees[v]^degree[v]
  my(want=OEIS_data("A196065"));
  check_equal(vector(#want,primecode,
                     my(vpar=vpar_from_primecode(primecode),
                        degrees=vpar_degrees_vector(vpar));
                     prod(v=1,#vpar, degrees[v]^2)),
              want,
              "A196065 first multiplicative Zagrep index of tree primecode");
}
{
  \\ A196064 second multiplicative Zagrep index of tree primecode n
  \\ prod degrees^2
  my(want=OEIS_data("A196064"));
  check_equal(vector(#want,primecode,
                     my(vpar=vpar_from_primecode(primecode),
                        degrees=vpar_degrees_vector(vpar));
                     if(#vpar==1,0,
                        prod(v=1,#vpar, degrees[v]^degrees[v]))),
              want,
              "A196064 second multiplicative Zagrep index of tree primecode");
}




\\-----------------------------------------------------------------------------
check_equal(a,'a);
check_equal(a_vector,'a_vector);

progress("end");
