\\ Copyright 2017, 2018, 2019, 2020, 2021, 2022 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file.  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.


\\ Usage gp -f test.gp
\\
\\ Run some tests on vpar.gp.  Might take a long time.
\\
\\ "option_verbose" and/or "option_progress" below can be 1 to print test
\\ names as they run.  This can help identify roughly where a failing (or
\\ slow) test is if it doesn't have its own good identifying name.
\\
\\ option_progress is a heading for each test section.  option_verbose
\\ output is many megabytes due to test repetitions, so put that through
\\ "tail" or into a file.
\\
\\ The order of tests is jumbled due to variously bringing things up to the
\\ start during development.
\\
\\ Some test trees and forests are generated randomly in test-helpers.gp
\\ with a seed from the system time.  test-helpers.gp prints the setrand()
\\ seed it used.  Modify test-helpers.gp if you want to force setrand() so
\\ as to repeat a particular failing run.
\\
\\ There's some test_limit_... parameters in test-helpers.gp to limit the
\\ size of explicit set searches or permutation searches.  A particularly
\\ fast or slow machine might take higher or lower limits, but that's hard
\\ to automate.
\\

default(recover,0);
default(strictargs,1);

option_verbose = 0;
option_randomize = 1;
option_progress = 1;

read("vpar.gp");  \\ check multiple reads ok to various install() or alias()
read("vpar.gp");
read("vpar.gp");
read("test-helpers.gp");


\\-----------------------------------------------------------------------------
\\ vpar_is_edgecover_set()
\\ vpar_is_edgecover_flags()

{
  check_equal(vpar_is_edgecover_set([], []), 1);
  check_equal(vpar_is_edgecover_flags([], []), 1);

  check_equal(vpar_is_edgecover_set([0,1,2], []), 0);
  check_equal(vpar_is_edgecover_set([0,1,2], [1]), 0);
  check_equal(vpar_is_edgecover_set([0,1,2], [2]), 0);
  check_equal(vpar_is_edgecover_set([0,1,2], [2,3]),
              1,
              "vpar_is_edgecover_set() path-3 both");

  check_equal(vpar_is_edgecover_set([0,1,2,3], [2,4]),
              1,
              "vpar_is_edgecover_set() path-4 ends");

  \\ self-loop
  check_equal(vpar_is_edgecover_set([0], []), 0,
             "vpar_is_edgecover_set() self-loop");
  check_equal(vpar_is_edgecover_set([1], [1]), 1);

  \\ 4-cycle
  check_equal(vpar_is_edgecover_set([4,1,2,3], [1,3]), 1);
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     for(rep=1,#vpar+1,
        my(flags=vpar_random_flags(vpar),
           set=test_flags_to_Set(flags),
           by_flags=vpar_is_edgecover_flags(vpar,flags),
           by_set=vpar_is_edgecover_set(vpar,set));
        check_equal(by_set,by_flags,
                    "vpar_is_edgecover_set() vs vpar_is_edgecover_flags()");
     ));
}

\\-----------------------------------------------------------------------------
progress("vpar_edgecovers_poly()");

{
  check_equal(vpar_edgecovers_poly([]), Pol(1));
  check_equal(vpar_edgecovers_poly([0]), Pol(0));
  check_equal(vpar_edgecovers_poly([0,1]), x);
  check_equal(vpar_edgecovers_poly([0,1],y), y);
}

test_edgecovers_poly_by_flags(vpar) =
{
  my(ret=Pol(0));
  forvec(flags=vector(#vpar,i,[0,1]),
     if(vpar_is_edgecover_flags(vpar,flags),
        ret += x^hammingweight(flags)));
  ret;
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     \\ print(vpar);
     my(got=vpar_edgecovers_poly(vpar));

     if(#vpar <= test_limit_flags_search,
        check_equal(got, test_edgecovers_poly_by_flags(vpar)));

     my(component_vpar=vpar_component_trees(vpar));
     check_equal(got,
                 prod(i=1,#component_vpar,
          vpar_edgecovers_poly(component_vpar[i]),
                      Pol(1)),
                 "vpar_edgecovers_poly() product of components, per addhelp");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_edgecovers_count()");

\\ cf A352885  never a tree edge covers count
\\ OEIS_data("A352885")[1..15]

{
  check_equal(vpar_edgecovers_count([]), 1,
              "vpar_edgecovers_count() empty");
  check_equal(vpar_edgecovers_count([0]), 0);
  check_equal(vpar_edgecovers_count([0,1]),1);
}

test_edgecovers_count_by_flags(vpar) =
{
  my(count=0);
  forvec(flags=vector(#vpar,i,[0,1]),
        count += vpar_is_edgecover_flags(vpar,flags));
  count;
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     \\ print(vpar);
     my(got=vpar_edgecovers_count(vpar));

     if(#vpar <= test_limit_flags_search,
        check_equal(got, test_edgecovers_count_by_flags(vpar)));

     my(component_vpar=vpar_component_trees(vpar));
     check_equal(got,
                 prod(i=1,#component_vpar, vpar_edgecovers_count(component_vpar[i])),
                 "vpar_edgecovers_count() product of components, per addhelp");
  );
}

{
  \\ star only 1 edgecover, all edges
  \\ each leaf edge must be in
  check_equal(vector(10,n,n--;
    vpar_edgecovers_count(vpar_make_star(n))),
              vector(10,n,n--; if(n==1,0, 1)),
              "vpar_edgecovers_count() of star");
}
{
  \\ Matula number of edge covers
  \\ not in OEIS: 0, 1, 1, 1, 2, 2, 1, 1, 3, 3, 3, 2, 2, 2, 5, 1, 2, 4, 1, 3
  \\ print(vector(20,n, vpar_edgecovers_count(vpar_from_primecode(n)))); quit
}


{
  \\ vpar_edgecovers_count() of complete binary
  \\
  \\        R
  \\      //  \
  \\     X     Y
  \\    / \   / \
  \\   S   T
  my(a(n) = n>=1 || error();
     vpar_edgecovers_count(vpar_make_complete_tree(2,n-1)));
  my(optroot(n) = n>=1 || error();
     my(vpar=vpar_make_complete_tree(2,n-1));
     vpar=vpar_join([0],1,vpar);
     vpar_edgecovers_count(vpar));

  check_equal(vector(10,n,n+=2; a(n)),
              vector(10,n,n+=2;
       optroot(n-1) * (2*a(n-1) + optroot(n-1))));
  check_equal(vector(10,n,n+=2; a(n)),
              vector(10,n,n+=2;
         (  a(n-1) + a(n-2)^2)
       * (3*a(n-1) + a(n-2)^2)));

  check_equal(vector(10,n,n+=2; optroot(n)),
              vector(10,n,n+=2; a(n) + a(n-1)^2),
              "optroot() permits none,none to root");
  check_equal(vector(10,n,n+=2; optroot(n)),
              vector(10,n,n+=2;
                     (a(n-1) + optroot(n-1))^2));

  if(0,
     print(vector(6,h, vpar_edgecovers_count(vpar_make_complete_tree(2,h))));
     \\ not in OEIS: 1, 3, 40, 6321, 162879523, 107217054417536040
     quit);
}
  

\\-----------------------------------------------------------------------------
progress("vpar_count_min_descendant_increasing()");

{
  my(limit = 10,
     L = O('x^limit),
     gT = test_func_to_gf(vpar_count_min_descendant_increasing, limit),
     gF = test_func_to_gf(n->vpar_count_min_descendant_increasing(n,1), limit),
     gB = test_func_to_gf(n->vpar_count_min_descendant_increasing(n,-1), limit));
  check_equal(gB,  1 + gF*'x + L);

  my(gFactorial = test_func_to_gf(n->n!, limit),
     gFactorial_n_sub_1 = test_func_to_gf(n->if(n>=1,(n-1)!,0), limit+1));
  check_equal(gFactorial_n_sub_1,
              'x*gFactorial,
              "gFactorial_n_sub_1");

  \\ A167894 gf 1/suminf(k!*x^k)
  check_equal(1/gFactorial,
              Polrev(A167894_samples[1..limit]) + L,
              "1/gFactorial");

  my(gF_times_n_plus_1 = test_func_to_gf(n->
                         (n+1)*vpar_count_min_descendant_increasing(n,1), limit));
  check_equal(gF_times_n_plus_1,  (gF*'x)',
              "gF_times_n_plus_1");

  check_equal(gF, 1 + gF_times_n_plus_1 * gFactorial_n_sub_1 + L);
  check_equal(gF, 1 + (gF*'x)' * 'x * gFactorial + L);
  check_equal(gF, 1 + ('x*gF' + gF) * 'x * gFactorial + L);
  check_equal(gF' - (1 - 'x*gFactorial)/('x^2*gFactorial) * gF, -1/('x^2*gFactorial) + L);
  check_equal(gF' + (-1 + 'x*gFactorial)/('x^2*gFactorial) * gF, -1/('x^2*gFactorial) + L);
  check_equal(gF' + 1/'x^2*('x - 1/gFactorial)*gF, -1/('x^2*gFactorial));

  \\ y' + P(x)*y = Q(x)
  \\ y = 1/u(x) integral Q(x)*u(x)  where u(x) = exp(integral P(x))
  \\ P = 1/x^2*(x - 1/gFactorial)
  \\ u = exp(intformal(1/x^2*(x - 1/gFactorial)))
  \\   = exp(intformal(1/x) - intformal(1/gFactorial)))
  \\   = exp(log(x)) / exp(intformal(1/gFactorial))
  \\   = 1 / exp(intformal(1/gFactorial))
  \\ 1/u = exp(intformal(1/gFactorial))
  \\ Q*u = -1/(x^2*gFactorial) * 1 / exp(intformal(1/gFactorial))
  \\     = -1/(x^2*gFactorial) / exp(intformal(1/gFactorial))
  \\     = -1/( x^2 * gFactorial * exp(intformal(1/gFactorial)) )
  \\
  \\ gF = exp(intformal(1/gFactorial))
  \\      * intformal( -1/( x^2 * gFactorial * exp(intformal(1/gFactorial)) ) )

  \\ print(((1/gFactorial)));
  \\ print(test_gf_terms(intformal(1/gFactorial), 10));

  \\ this u wrong
  my(P = 1/x^2*(x - 1/gFactorial),
     u = 1 / exp(intformal(1/gFactorial)),
     Q = -1/(x^2*gFactorial));
  check_equal(x^2*(P - 1/x),  log(u)');
  check_equal(P,  log(u)'/x^2 + 1/x, "P");
  check_equal(x^2*(P - 1/x),  -1/gFactorial);
  \\ log(u)'/x^2 + log(x)'
  \\ z' = log(u)'/x^2
  \\ z = log(u)/x^2 - intformal(log(u)*-2/x^3)
  \\ check_equal(P,  (log(u)/x^2 - intformal(log(u)*-2/x^3) + log(x))');
  \\ print(u);
  \\ print(log(u));
  \\ print(log(u)');
  \\ print(test_gf_terms(log(u)',limit));
  \\ u = 1 / exp(intformal(1/gFactorial)) is A003319

  \\ gH' = x^2*gF'
  \\ gH = intformal(x^2*gF')           int(f*g') = f*g - int(f'*g)
  \\    = x^2*gF - 2*int(x*gF)
  \\    = x^2*gF - 2*(x*int(gF) - int(gF)
  my(gH=intformal(x^2*gF'));
  check_equal(gH', x^2*gF');
  check_equal(gH, x^2*gF - intformal((x^2)'*gF));
  check_equal(gH, x^2*gF - 2*intformal(x*gF));
  check_equal(gH, x^2*gF - 2*(x*intformal(gF) - intformal(intformal(gF))));

  check_equal(gF', gH'/x^2);
  check_equal(gF, 1 + intformal(gH'/x^2), "gF");
  \\ termwise *(n-2)/n ...
  check_equal(gF,
              1 + sum(n=3,limit+1, polcoeff(gH,n)*n/(n-2) * x^(n-2)) + L);
  \\ print(gH);
  \\ print(gF);
}


{
  \\ A104981 column 1 of triangle, but n=0 case here is 1 instead
  check_equal(vpar_count_min_descendant_increasing(0), 1);
  check_equal(A104981_samples[1],                      0);
  check_equal(vector(#A104981_samples,n,n--;
                     if(n==0,0,
                     vpar_count_min_descendant_increasing(n))),
              A104981_samples,
              "vpar_count_min_descendant_increasing() trees vs A104981_samples");
}
A104981(n) = if(n==0,0, (n-1)! + sum(p=1,n-1, p * (n-1-p)! * A104981(p)));
{
  \\ A104981 column 1 of triangle
  check_equal(vector(#A104981_samples,n,n--; A104981(n)),
              A104981_samples,    \\ OFFSET=0
              "A104981() samples");
}
A003319(n) =
{
  \\ formula per Bowen in A003319
  \\ not efficient
  n>=0 || error();
  n! - sum(k=1,n-1, k! * A003319(n-k));
}
{
  check_equal(vector(#A003319_samples,n,n--; A003319(n)),
              A003319_samples,  \\ OFFSET=0
              "A003319() samples");

  my(limit=#A003319_samples,
     L = O(x^limit),
     gA003319 = test_func_to_gf(A003319,limit),
     gFactorial = test_func_to_gf(n->n!, limit));
  check_equal(gA003319,
     2 + log(1 / exp(intformal(1/gFactorial)))',
             "A003319 from gFactorial");
  check_equal(gA003319, 2 - log(exp(intformal(1/gFactorial)))',
             "A003319 from gFactorial two");
  check_equal(gA003319, 2 - intformal(1/gFactorial)',
             "A003319 from gFactorial three");
  check_equal(gA003319, 2 - 1/gFactorial,
             "A003319 from gFactorial four");

  \\ Paul D. Hanna in A003319, ogf satisfies
  \\   A(x) = x - x*A(x) + A(x)^2 + x^2*A'(x)
  \\ broken by incompatible change to a(0)
  my(A=gA003319-1);
  check_equal(A,  x - x*A + A^2 + x^2*A' + L,
              "A003319 ogf satisfies");
}

{
  \\ A003319 indecomposable permutations, those where initial [1..j] is not
  \\ a permutation (so sends something above j), for any 1<j<n
  check_equal(vector(#A003319_samples,n,n-=2;
                     if(n==-1,1, vpar_count_min_descendant_increasing(n,1))),
              A003319_samples,
              "vpar_count_min_descendant_increasing() forests vs samples");
}

\\ per addhelp
test_count_min_descendant_increasing_C_forests(n) = \
  if(n==0,1, (n-1)! + sum(p=1,n-1, test_count_min_descendant_increasing_C_forests(p) * (p+1) * (n-p-1)!));
{
  check_equal(vector(15,n,n--; vpar_count_min_descendant_increasing(n,1)),
              vector(15,n,n--; test_count_min_descendant_increasing_C_forests(n)),
              "vpar_count_min_descendant_increasing() forests vs C");
}
\\ per addhelp
test_count_min_descendant_increasing_C_trees(n) = \
  if(n==0,1, (n-1)! + sum(p=1,n-1, test_count_min_descendant_increasing_C_trees(p)*p*(n-p-1)!));
{
  check_equal(vector(15,n,n--; vpar_count_min_descendant_increasing(n)),
              vector(15,n,n--; test_count_min_descendant_increasing_C_trees(n)),
              "vpar_count_min_descendant_increasing() forests vs C");
}

{
  \\ indecomposable permutations recurrence
  \\   C(n) = n! - sum(k=1,n-1, C(n)*(n-k)!)
  check_equal(vector(#want,n,n--; vpar_count_min_descendant_increasing(n,1)),
              vector(#want,n,n--;
    n! - sum(k=1,n-1, (n-k)! * vpar_count_min_descendant_increasing(k,1))));
}

{
  for(forest=-1,1,
     for(n=0, test_limit_trees_search - forest,
        my(count_min_increasing=0,
           count_min_decreasing=0,
           count_max_increasing=0,
           count_max_decreasing=0,
           count_both_increasing=0,
           count_both_decreasing=0);
        forvec(vec=vpar_forvec_from(n,forest),
              my(vpar=vpar_from_upascend(vec),
                 min_descendants=vpar_minmax_descendants_vector(vpar,min),
                 max_descendants=vpar_minmax_descendants_vector(vpar,max),
                 is_min_increasing=test_vector_is_increasing(min_descendants),
                 is_min_decreasing=test_vector_is_decreasing(min_descendants),
                 is_max_increasing=test_vector_is_increasing(max_descendants),
                 is_max_decreasing=test_vector_is_decreasing(max_descendants));

              count_min_increasing += is_min_increasing;
              count_min_decreasing += is_min_decreasing;
              count_max_increasing += is_max_increasing;
              count_max_decreasing += is_max_decreasing;
              count_both_increasing += (is_min_increasing && is_max_increasing);
              count_both_decreasing += (is_min_decreasing && is_max_decreasing);

              \\ if((is_min_increasing && is_max_increasing) && n==6, print(min_descendants" "max_descendants));
        );
        my(got=vpar_count_min_descendant_increasing(n,forest));
        check_equal(got, count_min_increasing,
                    "vpar_count_min_descendant_increasing() vs count_min_increasing");

        \\ max increasing root=1 - have max_descendant[1]=n so
        \\ vec=[n,n,...,n] which is v=n below everything else so a path down
        \\ to it and (n-2)! permute in between
        \\
        check_equal(if(forest==-1, if(n<2,1, (n-2)!),
                       got),
                    count_max_increasing,
          Str("vpar_count_min_descendant_increasing() vs count_max_increasing"
              ", n="n" forest="forest));

        \\ min decreasing - have min_descendant[1]=1 so vec=[1,1,1,...,1]
        \\ which is v=1 below everything else so a path down to it and
        \\ (n-1)! permute
        \\ for root=1 this is possibly only for n<=1, otherwise no such
        check_equal(if(n<=1,1, forest==-1,0, (n-1)!),
                    count_min_decreasing,
             Str("descendant count_min_decreasing n="n" forest="forest));

        \\ max decreasing - have max_descendant[n]=n so vec=[n,n,...,n]
        \\ which is v=n below everything else so a path down to it and
        \\ (n-1)! permute, or root=1 fixes it at top and (n-2)! below
        check_equal(if(n<2,1, forest==-1,(n-2)!, (n-1)!),
                    count_max_decreasing,
             Str("descendant count_max_decreasing n="n" forest="forest));

        \\ trees, and root=1 is 1 later
        \\              1,1,2,5,14,45,168
        \\ not in OEIS:     2,5,14,45,168
        \\ forests
        \\              1,1,3,10,35,129,504
        \\ not in OEIS:     3,10,35,129,504
        \\ print(count_both_increasing);
        \\
        \\ check_equal(if(n==0,1,(n-1+forest)!), count_both_increasing,
        \\             "min and max descendant increasing");

        \\ both min descendant and max descendant decreasing not possible as
        \\ v=1 always min descendant = 1 and v=n always max descendant = n.
        check_equal(if(n<=1,1,0), count_both_decreasing,
                    "descendant count_both_decreasing");
  ));
}


\\-----------------------------------------------------------------------------
progress("vpar_domsets_count()");

check_equal(vpar_domsets_count([]), 1, "vpar_domsets_count() empty");
check_equal(vpar_domsets_count(vpar_make_path(2)), 1+1+1);
check_equal(vpar_domsets_count(vpar_make_path(3)), 1+3+1);
check_equal(vpar_domsets_count(vpar_make_path(4)), 4+4+1);

test_domsets_count_by_flags(vpar) =
{
  my(count=0);
  forvec(flags=vector(#vpar,i,[0,1]),
        count += vpar_is_domset_flags(vpar,flags));
  count;
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got=vpar_domsets_count(vpar));
     check_equal(got>=1, 1);

     if(#vpar <= test_limit_flags_search,
        check_equal(got, test_domsets_count_by_flags(vpar)));

     my(component_vpar=vpar_component_trees(vpar));
     check_equal(got,
                 prod(i=1,#component_vpar, vpar_domsets_count(component_vpar[i])),
                 "vpar_domsets_count() product of components, per addhelp");
  );
}

{
  \\ vpar_domsets_count() in complete binary tree
  my(domsets(h) = vpar_domsets_count(vpar_make_complete_tree(2,h)));
  check_equal(vector(6,h,h--; domsets(h)),
    [1, 5, 49, 4325, 39394801, 3141021747711077],
             "complete binary domsets()");
  \\ not in OEIS: 5, 49, 4325, 39394801, 3141021747711077

  \\ without root, dom or undom
  my(without(h) =
     if(h==0,1,domsets(h-1)^2));
  check_equal(vector(6,h,h--; without(h)),
              vector(6,h,h--; 
    my(vpar=vpar_make_complete_tree(2,h));
    vpar=vpar_delete_vertex(vpar,1);  \\ sans root
    vpar_domsets_count(vpar)),
              "complete binary without()");

  my(without_dom(h) =
     if(h==0,0, without(h) - self()(h-1)^2));
  check_equal(vector(4,h,h--; without_dom(h)),
              vector(4,h,h--; 
   my(count=0, vpar=vpar_make_complete_tree(2,h));
   forvec(flags=vector(#vpar,i,if(i==1,[0,0],[0,1])),
     count += vpar_is_domset_flags(vpar,flags));
   count),
              "complete binary without_dom()");

  my(without_undom(h) = without(h) - without_dom(h));

  my(with(h) = if(h==0,1,
                  (self()(h-1) + without(h-1))^2));
  check_equal(vector(4,h,h--; with(h)),
              vector(4,h,h--; 
   my(count=0, vpar=vpar_make_complete_tree(2,h));
   forvec(flags=vector(#vpar,i,if(i==1,[1,1],[0,1])),
     count += vpar_is_domset_flags(vpar,flags));
   count),
              "complete binary with()");

  my(all(h) = with(h) + without(h));
  check_equal(vector(4,h,h--; with(h)),
              vector(4,h,h--; if(h==0,1, 
                                 all(h-1)^2)));
  check_equal(vector(6,h,h--; all(h)),
    [2, 5, 50, 4901, 42725426, 3377412372711077]);
}
\\
\\ domsets(h)     = with(h) + without_dom(h)
\\ with(h)        = all(h-1)^2
\\                = (with(h-1) + domesets(h-2)^2)^2
\\ all(h)         = with(h) + without(h)
\\ without(h)     = domesets(h-1)^2
\\ without_dom(h) = without(h) - without_dom(h-1)^2
\\          = domesets(h-1)^2 - without_dom(h-1)^2
\\ without_undom(h) = without(h) - without_dom(h)
\\ vector(6,h,h--; with(h))
\\ not in OEIS: 1, 4, 25, 2500, 24019801, 1825462026881476
\\ vector(6,h,h--; without(h))
\\              1, 1, 25, 2401, 18705625, 1551950345829601
\\ not in OEIS:       25, 2401, 18705625, 1551950345829601
\\ vector(6,h,h--; without_dom(h))
\\        0, 1, 24, 1825, 15375000, 1315559720829601
\\ not in OEIS: 24, 1825, 15375000, 1315559720829601


\\-----------------------------------------------------------------------------
progress("vpar_to_primecode_vector()");

{
  check_equal(vpar_to_primecode_vector([]), [],
              "vpar_to_primecode_vector() empty");
  check_equal(vpar_to_primecode_vector([0]), [1],
              "vpar_to_primecode_vector() singleton");

  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i]);
     check_equal(vpar_to_primecode_vector(vpar),
                 vector(#vpar,v, vpar_to_primecode(vpar_subtree(vpar,v)))))
}

\\-----------------------------------------------------------------------------
progress("vpar_to_primecode()");

{
  check_equal(vpar_to_primecode([]), 0, "vpar_to_primecode() empty");
  check_equal(vpar_to_primecode([0]), 1, "vpar_to_primecode() singleton");
  check_equal(vpar_to_primecode([0,1]), 2, "vpar_to_primecode() path-2");
  check_equal(vpar_to_primecode([2,0]), 2, "vpar_to_primecode() path-2");
}

{
  \\ Gobel's example figure 1
  \\
  \\      1   T        1  T1       1  T2
  \\     / \           |          / \
  \\    2   4          2         2   3
  \\   /   / \                       |
  \\  3   5   6                      4
  \\          |
  \\          7
  check_equal(vpar_to_primecode([0,1]),     2,
              "vpar_to_primecode() Goebel T1");

  check_equal(vpar_to_primecode([0,1,1,3]), 6,
              "vpar_to_primecode() Goebel T2");
  check_equal(vpar_to_primecode([0,1,2,1]), 6,
              "vpar_to_primecode() Goebel T2 other way");

  check_equal(vpar_to_primecode([0,1,2,1,4,4,6]), 39,
              "vpar_to_primecode() Goebel T");
}

{
  \\ Peter R Cappello, "A Note on a Bijection between Natural Numbers and
  \\ Rooted Trees", 15 July 1988.

  \\       1              figure 1 example
  \\    /  |   \          primecode 399
  \\  2    4      7
  \\  |   / \    /|\
  \\  3  5   6  8 9 10
  \\
  my(vpar=[0,1,2, 1,4,4, 1,7,7,7]);
  vpar_validate(vpar);
  check_equal(vpar_to_primecode(vpar), 399,
              "vpar_to_primecode() Cappello figure 1");

  \\ Cappello's example figure 2 trees primecode = 1 to 15
  \\ (cf Goebel to 45)
  check_equal(vpar_to_primecode([0]), 1);
  check_equal(vpar_to_primecode([0,1]), 2);
  check_equal(vpar_to_primecode([0,1,2]), 3);
  check_equal(vpar_to_primecode([0,1,1]), 4);
  check_equal(vpar_to_primecode([0,1,2,3]), 5);
  check_equal(vpar_to_primecode([0,1,1,3]), 6);
  check_equal(vpar_to_primecode([0,1,2,2]), 7);
  check_equal(vpar_to_primecode([0,1,1,1]), 8);
  check_equal(vpar_to_primecode([0,1,2,1,4]), 9);
  check_equal(vpar_to_primecode([0,1,1,3,4]), 10, "Cappello primecode=10");
  check_equal(vpar_to_primecode([0,1,2,3,4]), 11);
  check_equal(vpar_to_primecode([0,1,1,1,4]), 12);
  check_equal(vpar_to_primecode([0,1,2,2,4]), 13);
  check_equal(vpar_to_primecode([0,1,1,3,3]), 14);
  check_equal(vpar_to_primecode([0,1,2,1,4,5]), 15);
}

{
  \\ All relabellings same primecode.
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i]);
     if(#vpar>6, next());
     my(vpar_relabelled=vpar_relabel_random(vpar));
     check_equal(vpar_to_primecode(vpar_relabelled),
                 vpar_to_primecode(vpar),
                 "vpar_to_primecode() relabelled"));
}
{
  \\ is A007097 prime'th recurrence
  \\ prime() becomes a bit slow after about 14 iterations, and overflows
  \\ 32-bit table of primes before long
  \\ A007097(n) = primecode of path of n edges
  \\
  my(want=A007097_samples[1..14]);
  check_equal(vector(#want,n,n--;
                     my(vpar=vpar_make_path(n+1));
                     vpar_num_edges(vpar) == n || error();
                     vpar_to_primecode(vpar)),
              want,  \\ A007097_samples OFFSET=0
              "A007097 primeth recurrence, primecode of path n");

  \\ Peter R Cappello, "A Note on a Bijection between Natural Numbers and
  \\ Rooted Trees", 15 July 1988.
  \\ Section 3 paths samples.
  \\ Primecode of path rooted at an end.
  check_equal(A007097_samples[1..12],
              [1,2,3,5,11,31,127,709,5381,52711,648391,9737333]);
}


\\-----------------------------------------------------------------------------
\\ test_A007097_primeth()
\\ prime'th recurrence prime(prime(...prime(1)))

{
  \\ prime() becomes slow above the prime table,
  \\ so limit the test data
  my(want=A007097_samples[1..10]);  \\ OFFSET=0
  check_equal(vector(#want,n,n--; test_A007097_primeth(n)),
              want,
              "test_A007097_primeth() vs DATA")
}

\\-----------------------------------------------------------------------------
progress("vpar_total_cophenetic()");

{
  \\   1          1     
  \\   |  \       |  \ 
  \\   2    6     2    7
  \\   |          |
  \\   3          3
  \\  / \         |
  \\  4  5        4
  \\              /\
  \\             5  6
  check_equal(vpar_total_cophenetic([0,1,2,3,3,1]),
              2);
  check_equal(vpar_total_cophenetic([0,1,2,3,4,4, 1]),
              3);
}

{
  \\ Mir, Rossello, Rotger example figure 4
  check_equal(vpar_total_cophenetic([0,1,1,1,1]), 0, "star");
  check_equal(vpar_total_cophenetic([0,1,2,3,4, 1,2,3,4]), 10);
  \\          1
  \\        /   \            3,4
  \\       2     6           3,5
  \\      /|\   / \          4,5
  \\     3 4 5  7  8         7,8
  check_equal(vpar_total_cophenetic([0,1,2,2,2, 1,6,6]), 4);

  \\ star all cophenetic 0
  check_equal(vector(20,n,n--; vpar_total_cophenetic(vpar_make_star(n))),
              vector(20,n,n--; 0),
              "vpar_total_cophenetic() of star");
}
{
  \\ Mir, Rossello, Rotger examples figure 4
  \\ all trees with 5 childless vertices and phylogenic so series-reduced
  \\ (FIXME: what order is it?  Not by num internal vertices ...)
  my(is_phylogenic(vpar) = my(num_children=vpar_num_children_vector(vpar));
     for(v=1,#vpar, if(num_children[v]==1,return(0))); 1);
  my(vpars=concat([test_unlabelled_all(n) |n<-[6..9]]));
  vpars=select(vpar->vpar_num_childless(vpar)==5 && is_phylogenic(vpar), vpars);
  check_equal(#vpars,12);
  check_equal(vecsort(apply(vpar_total_cophenetic,vpars)),
              [0,1,2,3, 4,4,5,6, 7,8,9,10]);
}

\\ cophenetic = depth of deepest common ancestor of u,v
test_cophenetic_value_of_pair(vpar,u,v) =
{
  my(c=vpar_common_ancestor(vpar,u,v));
  if(c, vpar_depth_of_vertex(vpar,c), 0);
}
\\ sum cophenetic of all pairs of leaves
test_total_cophenetic_by_pairs(vpar) =
{
  my(childless=vpar_childless_vertices(vpar));
  sum(i=1,#childless-1,
      sum(j=i+1,#childless,
          test_cophenetic_value_of_pair(vpar,
                                        childless[i], childless[j])));
}

{
  check_equal(vpar_total_cophenetic([0,1,2,2]), 1);
  check_equal(vpar_total_cophenetic([0,1,1]), 0);
  check_equal(vpar_total_cophenetic([0,1,2,2]), 1);
  check_equal(vpar_total_cophenetic([0,1,2,2,4,5]), 1);
  check_equal(vpar_total_cophenetic([0,1,2,3,4,5,6,6]), 5);

  \\      2
  \\     / \
  \\    3   7
  \\        | \
  \\        4  6
  \\        |  |\
  \\        8  5 1
  my(vpar=[6, 0, 2, 7, 6, 7, 2, 4]);
  \\       1  2  3  4  5  6  7  8
  vpar_validate(vpar);
  check_equal(test_total_cophenetic_by_pairs(vpar), 2*1 + 2);
  check_equal(vpar_total_cophenetic(vpar),          2 + 2*1);
}
{
  my(vpar=[0,1,2, 3,3,3,3], want=2*binomial(4,2));
  \\       1 2 3  4 5 6 7
  vpar_validate(vpar);
  check_equal(want, 12);
  check_equal(test_total_cophenetic_by_pairs(vpar), want);
  check_equal(vpar_total_cophenetic(vpar),          want);
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     check_equal(vpar_total_cophenetic(vpar),
                 test_total_cophenetic_by_pairs(vpar),
                 "total_cophenetic_by_vector()");
     );
}

\\-----
\\ total cophenetic of binomial tree

\\ A100575 Half the number of permutations of 0..n with exactly two maxima.
\\   0,0,1,8,44,208,912,3840,15808
\\   4^(n-1) - (n+1)*2^(n-2)
test_A100575(n) = n>=0 || error(); 1<<(2*n-2) - (n+1)<<(n-2);
{
  check_equal(vector(#A100575_samples,n,n--; test_A100575(n)),
              A100575_samples,  \\ OFFSET=0
              "A000431() samples");

  my(a=test_A100575);
  check_equal(vector(100,n,n--; a(n+1)),
              vector(100,n,n--; 2*a(n) + binomial(2^n,2)),
              "test_A100575() recurrence");

  my(b(n) = if(n==0,0,a(n-1)));
  check_equal(vector(10,n,n++; b(n)),
              vector(10,n,n++; 2*b(n-1) + binomial(2^(n-2),2)),
              "b() recurrence");
  check_equal(vector(10,n, vpar_num_childless(vpar_make_binomial_tree(2^n))),
              vector(10,n, 2^(n-1)),
              "binomial tree num childless");

  check_equal(vector(12,n,n--; vpar_total_cophenetic(vpar_make_binomial_tree(2^n))),
              vector(12,n,n--; b(n)),
              "vpar_total_cophenetic() of binomial tree 2^n");
  check_equal(vector(12,n, vpar_total_cophenetic(vpar_make_binomial_tree(2^n))),
              vector(12,n, test_A100575(n-1)),
              "vpar_total_cophenetic() of binomial tree 2^n = A100575");
}

\\---------
\\ total cophenetic of complete binary tree

\\ (4^n - n*2^(n+1))/8
\\ 0,0,0,2,16,88,416
A000431(n) = 1<<(2*n-3) - n<<(n-2);
{
  check_equal(vector(#A000431_samples,n,n--; A000431(n)),
              A000431_samples,  \\ OFFSET=0
              "A000431() samples");
  check_equal(vector(100,n,n--; A000431(n)),
              test_gf_terms(2*x^3/((1-2*x)^2*(1-4*x)),100),
              "A000431() gf");

  check_equal(vector(10,n,n--; A000431(n+1)),
              vector(10,n,n--; sum(k=1,n, 2^k*binomial(2^(n-k),2))));
  \\ or A100575 with k=0 included
}
total_cophenetic_of_complete_binary(h) = A000431(h+1);
{
  \\ complete binary
  \\    1             1
  \\   / \           / \
  \\  2   3         2   5
  \\               / \ / \
  \\              3  4 6  7
  check_equal(vector(6,h,h--; vpar_total_cophenetic(vpar_make_complete_tree(2,h))),
              vector(6,h,h--; total_cophenetic_of_complete_binary(h)),
              "total cophenetic");
  check_equal(vpar_total_cophenetic([0,1,1]), 0);
  check_equal(vpar_total_cophenetic([0,1,2,2, 1,5,5]), 2, "complete binary h=2");
  check_equal(A000431(3), 2);
  check_equal(vector(6,h, vpar_num_childless(vpar_make_complete_tree(2,h))),
              vector(6,h, 2^h),
              "complete binary num childless");

  \\ increment is +1 deeper ancestor for each pair in the two halves
  check_equal(vector(6,h, vpar_total_cophenetic(vpar_make_complete_tree(2,h))
                       - 2*vpar_total_cophenetic(vpar_make_complete_tree(2,h-1))),
              vector(6,h, 2*binomial(2^(h-1),2)),
              "total cophenetic complete binary tree increment");

  check_equal(vector(50,n,n--; sum(k=1,n, 2^k*binomial(2^(n-k),2))),
              vector(50,n,n--; (2^n - (n+1))*2^(n-1)),
              "total cophenetic complete binary formula");
}


\\-----------------------------------------------------------------------------
\\ vpar_is_caterpillar()

check_equal(vpar_is_caterpillar([]), 1, "vpar_is_caterpillar() empty");
{
  \\     5
  \\     |\
  \\     2 4
  \\    /|
  \\   1 6
  \\     |
  \\     3
  my(vpar=[2, 5, 6, 5, 0, 2]);
  check_equal(vpar_is_caterpillar(vpar), 1);
}
{
  \\     1
  \\     |
  \\     2
  \\     |\
  \\     3 5
  \\     |  \
  \\     4   6
  my(vpar=[0,1,2,3, 2, 5]);
  check_equal(vpar_is_caterpillar(vpar), 1);
}

{
  check_equal(vpar_is_caterpillar([0,1,0,3]), 1,
              "forest 2 paths");

  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        by_forest = vpar_is_caterpillar(vpar),
        components=vpar_component_trees(vpar),
        each=apply(vpar_is_caterpillar,components),
        by_components = hammingweight(each)==#each);
     check_equal(by_forest, by_components,
                 "vpar_is_caterpillar() forest vs components"));
}

\\ Return a vertex which is a diameter endpoint, ie. a vertex of maximum
\\ eccentricity, in the "periphery".
\\ There may be several vertices of maximum eccentricity and it's
\\ unspecified which is returned.
test_diameter_endpoint(vpar) = \
  my(v=0); if(#vpar, vecmax(vpar_eccentricities_vector(vpar),&v)); v;
check_equal(test_diameter_endpoint([0,1,2,3]), 1);
check_equal(test_diameter_endpoint([0,1,2,3,4, 2,2,2,2]), 1);

test_is_caterpillar_by_reroot(vpar) =
{
  \\ Re-root to a diameter endpoint, then the spine is a height vertex up to
  \\ the root and every vertex must have a spine vertex as its parent
  \\ (true also of the spine itself).
  my(components=vpar_component_trees(vpar));
  for(i=1,#components,
     my(vpar=components[i]);
     vpar=vpar_reroot(vpar,test_diameter_endpoint(vpar));
     my(low=vpar_height_vertices(vpar)[1],
        spine=Set(vpar_ancestors(vpar,low)));
     for(v=1,#vpar,
        if(vpar[v] && !setsearch(spine,vpar[v]), return(0))));
  1;
}
{
  \\  1      2
  \\       / | \
  \\      3  5  7
  \\      |  |  |
  \\      4  6  8
  my(vpar=[0, 0,2,3,2,5,2,7]);
  vpar_validate(vpar);
  check_equal(#vpar,8);
  check_equal(vpar_is_caterpillar(vpar), 0);
  check_equal(test_is_caterpillar_by_reroot(vpar), 0);

  for(rep=1,100,
     my(n=random(20),
        vpar=vpar_make_random(n,1));
     check_equal(vpar_is_caterpillar(vpar),
                 test_is_caterpillar_by_reroot(vpar),
                 "vpar_is_caterpillar() vs test_is_caterpillar_by_reroot() random"));
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     check_equal(vpar_is_caterpillar(vpar),
                 test_is_caterpillar_by_reroot(vpar),
                 "vpar_is_caterpillar() vs test_is_caterpillar_by_reroot()");
     );
}

\\ A005418 1, 2, 3, 6, 10, 20, 36, 72, 136, 272, 528
\\ num caterpillars 1,1,2,3,6,10,20,36,72,136,272,
test_A005418_formula(n) = 2^(n-2) + 2^(floor(n/2) - 1);
test_A005418(n) = n>=1 || error(); test_A005418_formula(n);
test_count_free_caterpillars(n) = if(n<=2,1, test_A005418(n-2));
{
  check_equal(test_A005418_formula(0), 3/4);
  check_equal(vector(#A005418_samples,n, test_A005418(n)),
              A005418_samples,  \\ OFFSET=1
              "A005418 samples");
  check_equal(vector(10,n,n--; test_count_free_caterpillars(n)),
              vector(10,n,n--;
                     my(vpar=vpar_free_first(n),
                        count=0);
                     until((vpar=vpar_free_next(vpar))=='none,
                           count+=vpar_is_caterpillar(vpar));
                     count),
              "vpar_is_caterpillar() count A005418");
}

\\-----------------------------------------------------------------------------
progress("vpar_is_degrees_vector()");

\\ gp 2.11.2
\\ Vecsmall([]) == 0     \\ is not 0, unlike full Vec
\\ Vecsmall([0]) == 0
\\ !Vecsmall([])
\\ !Vecsmall([0])
\\ vpar_INTERNAL_gnot(Vecsmall([]))
\\ vpar_INTERNAL_gnot(Vecsmall([0]))
\\ vecsum(Vecsmall([1,2,3]))    \\ no vecsum of Vecsmall
\\ [] == 0
\\ [0] == 0

{
  \\ empty and singleton
  for(forest=-1,1,
     check_equal(vpar_is_degrees_vector([],forest), 1,
                 "vpar_is_degrees_vector() empty");
     check_equal(vpar_is_degrees_vector(Vecsmall([]),forest), 1,
                 "vpar_is_degrees_vector() empty Vecsmall");

     check_equal(vpar_is_degrees_vector([0],forest), 1,
                 "vpar_is_degrees_vector() singleton");
     check_equal(vpar_is_degrees_vector(Vecsmall([0]),forest), 1,
                 "vpar_is_degrees_vector() singleton Vecsmall");

     check_equal(vpar_is_degrees_vector([1],forest), 0,
                 "vpar_is_degrees_vector() self-loop");
     check_equal(vpar_is_degrees_vector(Vecsmall([1]),forest), 0,
                 "vpar_is_degrees_vector() self-loop Vecsmall");
     );
}

{
  my(vpar=[2,0,2]);
  check_equal(vpar_degrees_vector(vpar), [1,2,1]);
  check_equal(vpar_is_degrees_vector([1,2,1]), 1);
  check_equal(vpar_is_degrees_vector([1,2,1],1), 1);
  check_equal(vpar_is_degrees_vector([1,2,1],-1), 1);
}
{
  check_equal(vpar_is_degrees_vector([0,0],1), 1);
  check_equal(iferr(vpar_is_degrees_vector([0,0],'forest);0, e, 1),
              1,
  "vpar_is_degrees_vector() error on forest=symbol (whereas it should be 0,1)");
}

{
  \\ any partition of the integer 2*n-2 into n parts is a degrees vector
  for(n=1,10,
     forpart(p=2*n-2,  \\ partitions of 2*n-2
        \\ convert to full Vec() since Vecsmall not accepted by vecsum()
        \\ circa gp 2.11.x, used in vpar_is_degrees_vector()
        \\ print(n" "p);
        check_equal(vpar_is_degrees_vector(Vec(p)), 1,
                    "vpar_is_degrees_vector() of partition");
        ,, [n,n]));     \\ partitions into n parts
}
{
  \\ partitions of 2n-2 of n parts are equivalent to all partitions of n-2,
  \\ by part of n-2 add to vector of n many 1s
  \\ numbpart(20)
  \\ numbpart(30)
  \\
  check_equal(vector(30,n,n+=2; #partitions(2*n-2,,[n,n])),
              vector(30,n,numbpart(n)));
  for(n=2,20,
     my(degs=partitions(2*n-2,,[n,n]));
     \\ print(#degs" "degs);
     my(parts=partitions(n-2));
     \\ print(#parts" "parts);
     my(mods=apply(p->vectorsmall(n,i, my(x=#p-(n-i)); 1 + if(x>=1,p[x],0)),
                  parts));
     \\ print(#mods" "mods);
     check_equal(mods, degs);
  );
}

{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i],
        degrees=vpar_degrees_vector(vpar));
     \\ print(vpar"  "degrees);
     check_equal(vecsum(degrees), max(0, 2*#vpar - 2),
                "vecsum(degrees) == 2*n-2");
     check_equal(vpar_is_degrees_vector(degrees),1,
                "vpar_is_degrees_vector() of tree");
     check_equal(vpar_is_degrees_vector(degrees,-1), 1,
     "vpar_is_degrees_vector() of root=1, degrees don't depend on root"));
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        degrees=vpar_degrees_vector(vpar));
     check_equal(vpar_is_degrees_vector(degrees,1),1,
                 "vpar_is_degrees_vector() of forest"));
}

check_equal(vpar_degrees_vector([0,1,2]), [1, 2, 1]);

check_equal(vpar_degrees_vector([0,1]), [1,1], "vpar_degrees_vector() path-2");
check_equal(vpar_is_degrees_vector([1,1]), 1, "vpar_is_degrees_vector() path-2");

check_equal(vpar_is_degrees_vector([1,2,1]), 1, "vpar_is_degrees_vector() path-3");
check_equal(vpar_is_degrees_vector([3,0,1]), 0);
\\check_equal(vpar_is_degrees_vector([3,0,1],1), 0);

{
  \\ l = list of all degrees vectors found by iterating.
  \\ All must pass vpar_is_degrees_vector().

  for(forest=-1,1,
     for(n=0, test_limit_trees_search - forest - 1,
        my(l=List([]));
        forvec(vec=vpar_forvec_from(n,forest), \\ trees or forests
              my(vpar=vpar_from_upascend(vec),
                 degrees=vpar_degrees_vector(vpar),
                 pos=setsearch(l,degrees,1));
              \\ print(vpar" degs "degrees" pos "pos);
              if(pos,listinsert(l,degrees,pos));
              check_equal(vpar_is_degrees_vector(degrees,forest), 1,
                          "vpar_is_degrees_vector() of vpar_forvec_from()"));
        \\ print("forest="forest"  "l);

        forvec(degrees=vector(n,i,[-1,n+1]),
              \\ if(degrees==[3,0,1], print(setsearch(l,degrees)));
              check_equal(vpar_is_degrees_vector(degrees,forest),
                          setsearch(l,degrees)!=0,
                          "vpar_is_degrees_vector() vs setsearch"));
     ));
}


\\-----------------------------------------------------------------------------
\\ vpar_make_caterpillar()

check_equal(vpar_make_caterpillar([]), []);
check_equal(vpar_make_caterpillar([0]), [0]);
check_equal(vpar_make_caterpillar([2]), [0,1,1]);

\\-----------------------------------------------------------------------------
progress("vpar_Wiener_index()");

{
  for(n=0,#v,
     my(want=n*(n-1)/2,
        vpar=vpar_make_path(n));
     check_equal(vpar_Wiener_index(vpar),want,
                 Str("vpar_Wiener_index() path "n)));
}

check_equal(vpar_Wiener_index([0,1,1]),4);  \\ star-3 = path-3
check_equal(vpar_Wiener_index([0,3,1]),4);  \\ path-3 relabelled

test_Wiener_index_by_subtree_sizes(vpar) =
{
  \\ edge v--p crossings is num vertices below * num vertices above
  my(sizes=vpar_subtree_sizes_vector(vpar));
  sum(v=1,#vpar, sizes[v] * (#vpar - sizes[v]));
}
test_Wiener_index_by_pairwise_path_lengths(vpar) =
{
  my(W=0);
  for(u=1,#vpar-1,
     for(v=u+1,#vpar,
        W += max(0, vpar_path_length(vpar,u,v))));
  W;
}

\\ this one an alternative to going via vpar_subtree_sizes_vector()
test_Wiener_index_by_num_below(vpar) =
{
  my(seq=vpar_INTERNAL_seq_upwards(vpar),
     num_below=vectorsmall(#vpar,v,1), \\ num vertices below, including self
     W=0);
  for(i=1,#seq,
     my(v=seq[i], p=vpar[v]);
     if(p,
        \\ edge v -- p crossings according to vertices on each side
        W += num_below[v] * (#vpar - num_below[v]);
        num_below[p] += num_below[v]));
  W;
}

{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i],
        got=vpar_Wiener_index(Vecsmall(vpar)));
     check_equal(got,
                 test_Wiener_index_by_pairwise_path_lengths(vpar),
                 Str("vpar_Wiener_index() vs pairwise of ",vpar));
     check_equal(got,
                 test_Wiener_index_by_num_below(vpar),
                 Str("vpar_Wiener_index() vs num_below of ",vpar));
     check_equal(got,
                 test_Wiener_index_by_subtree_sizes(vpar),
                 Str("vpar_Wiener_index() vs subtree_sizes of ",vpar));
  );
}

{
  \\ n=7 mean distance between pairs exactly 2/3
  my(vpar=vpar_make_bistar(3,4));
  check_equal(vpar_Wiener_index(vpar) / binomial(#vpar,2) / vpar_diameter(vpar),
              2/3);
}
{
  \\ n=12 mean distance between pairs exactly 2/3
  \\         *
  \\         |
  \\     3   *   *
  \\     |   |   |
  \\ 1---2---5---10
  \\     |   |   |
  \\     4   *   *
  \\         |
  \\         *
  \\                5 6 7   9
  my(vpar=[0,1,2,2, 2,5,5,6,7, 5,10,10]);
  check_equal(vpar_Wiener_index(vpar) / binomial(#vpar,2) / vpar_diameter(vpar),
              2/3);
}
{
  \\ n=4 claw mean distance between pairs exactly 3/4
  my(vpar=vpar_make_star(4));
  check_equal(vpar_Wiener_index(vpar) / binomial(#vpar,2) / vpar_diameter(vpar),
              3/4);
}
{
  \\ n=16 mean distance between pairs exactly 3/4
  my(vpar=vpar_make_bistar(10,6));
  check_equal(vpar_Wiener_index(vpar) / binomial(#vpar,2) / vpar_diameter(vpar),
              3/4);
}


\\----------------------
\\ terminal Wiener Index

check_equal(vpar_Wiener_index([0,3,1],1),2);  \\ path-3 relabelled
check_equal(vpar_Wiener_index([0,1,1],1),2);  \\ star-3 = path-3

{
  for(n=0,#v,
     my(want=n*(n-1)/2,
        vpar=vpar_make_path(n));
     check_equal(vpar_Wiener_index(vpar,1),want,
                 Str("vpar_Wiener_index() terminal, path "n)));
}

test_terminal_Wiener_index_by_pairwise_path_lengths(vpar) =
{
  my(degrees=vpar_degrees_vector(vpar));
  my(W=0);
  for(u=1,#vpar-1,
     if(degrees[u]>1,next());
     for(v=u+1,#vpar,
        if(degrees[v]>1,next());
        W += vpar_path_length(vpar,u,v)));
  W;
}
{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i]);
     check_equal(vpar_Wiener_index(Vecsmall(vpar),1), \\ terminal
                 test_terminal_Wiener_index_by_pairwise_path_lengths(vpar),
                 Str("vpar_Wiener_index() terminal of ",vpar));
  );
}

{
  \\ n=10 equal maximum terminal Wiener index
  \\
  \\ Ivan Gutman, Boris Furtula and Miroslav Petrovic, "Terminal Wiener
  \\ Index", Journal of Mathematical Chemistry, volume 46, issue 2, August
  \\ 2009, pages 522--531.  DOI 10.1007/s10910-008-9476-2

  my(vpar=vpar_make_bistar(5,5));
  check_equal(#vpar,10);
  check_equal(vpar_Wiener_index(vpar,1), 72);
}
{
  \\ n=10 equal maximum terminal Wiener index
  my(vpar=vpar_make_star(10));
  check_equal(#vpar,10);
  check_equal(vpar_Wiener_index(vpar,1), 72);
}
{
  \\ n=11 maximum terminal Wiener index
  \\
  \\ Ivan Gutman, Boris Furtula, Miroslav Petrovic, "Terminal Wiener
  \\ Index", Journal of Mathematical Chemistry, volume 46, issue 2, August
  \\ 2009, pages 522--531.  DOI 10.1007/s10910-008-9476-2

  my(vpar=vpar_make_bistar(5,6));
  check_equal(#vpar,11);
  check_equal(vpar_Wiener_index(vpar,1), 92);
}
{
  \\ n=12 maximum terminal Wiener index
  \\
  \\ Ivan Gutman, Boris Furtula, Miroslav Petrovic, "Terminal Wiener
  \\ Index", Journal of Mathematical Chemistry, volume 46, issue 2, August
  \\ 2009, pages 522--531.  DOI 10.1007/s10910-008-9476-2

  my(vpar=vpar_make_bistar(6,6));
  check_equal(#vpar,12);
  check_equal(vpar_Wiener_index(vpar,1), 115);
}
{
  \\ Gutman, Furtula, Petrovic, figure 1 example
  \\     7   8   10
  \\     |   |   |
  \\ 1---2---3---4---5---6
  \\         |
  \\         9
  my(vpar=[0,1,2,3,4,5, 2,3,3,4]);
  vpar_validate(vpar);
  check_equal(#vpar,10);
  check_equal(vpar_Wiener_index(vpar,1), 51);
}


\\-----------------
\\ childless terminal Wiener index

test_childless_terminal_Wiener_index_by_unplant(vpar) =
{
  while(1,
        my(r=vpar_root(vpar));
        if(vpar_num_children_of_vertex(vpar,r)==1,
           vpar=vpar_delete_vertex(vpar,r),
           break));
  vpar_Wiener_index(vpar,1);
}
test_childless_terminal_Wiener_index_by_pairwise_path_lengths(vpar) =
{
  my(childless=vpar_childless_vertices(vpar),
     path_lengths=vpar_path_lengths_matrix(vpar));;
  sum(i=2,#childless,
      sum(j=1,i-1,
          path_lengths[childless[i],childless[j]]));
}
{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i],
        got=vpar_Wiener_index(vpar,2));
     check_equal(got,
                 test_childless_terminal_Wiener_index_by_pairwise_path_lengths(vpar),
                "vpar_Wiener_index() vs test_childless_terminal_Wiener_index_by_pairwise_path_lengths");
     check_equal(got,
                 test_childless_terminal_Wiener_index_by_unplant(vpar),
                "vpar_Wiener_index() vs test_childless_terminal_Wiener_index_by_unplant");

     my(root=vpar_root(vpar),
        depths=vpar_depths_vector(vpar),
        childless=vpar_childless_vertices(vpar));
     check_equal(got,
                 vpar_Wiener_index(vpar,1)
                 - if(root && vpar_degree_of_vertex(vpar,root)==1,
                      vecsum(vecextract(depths,childless))),
                 "vpar_Wiener_index() vs sans root pathsy");
     );
}


\\-----------------------------------------------------------------------------
progress("vpar_indsets_count()");

check_equal(vpar_indsets_count([]), 1);
check_equal(vpar_indsets_count([0]), 2);
check_equal(vpar_indsets_count([0,1,0,3]),3*3);

test_indsets_count_by_flags(vpar) =
{
  my(count=0);
  forvec(flags=vector(#vpar,i,[0,1]),
        count += vpar_is_indset_flags(vpar,flags));
  count;
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got=vpar_indsets_count(vpar));

     if(#vpar <= test_limit_flags_search,
        check_equal(got, test_indsets_count_by_flags(vpar)));

     my(component_vpar=vpar_component_trees(vpar));
     check_equal(got,
                 prod(i=1,#component_vpar, vpar_indsets_count(component_vpar[i])),
                 "vpar_indsets_count() product of components, per addhelp");
  );
}

\\------------
{
  \\ total indsets count over ordered trees
  \\ A007226 a(n+2) = 2*C(3n, n) - C(3n, n+1)
  \\                = 1/(n-2)*binomial(3*(n-2), n-1)
  for(n=0,10,
     my(vpar=vpar_preorder_first(n),
       count=0);
     until((vpar=vpar_preorder_next(vpar))=='none,
           count += vpar_indsets_count(vpar));

     check_equal(count, if(n==0,1, A007226_samples[n]),
                 "A007226 vpar_indsets_count() total over ordered trees");

     check_equal(count, if(n==0,1,
                           2*binomial(3*(n-1), n-1) - binomial(3*(n-1), n)),
     "vpar_indsets_count() total over ordered trees, two binomial formula");

     check_equal(count, if(n==0,1, n==1,2,
                           1/(n-1)*binomial(3*(n-1), n)),
     "vpar_indsets_count() total over ordered trees, one binomial formula");
  );

  check_equal(vector(#A007226_samples,n,n++;
                     if(n==2, 2,  1/(n-2)*binomial(3*(n-2), n-1))),
              A007226_samples,  \\ OFFSET=2
              "A007226 one binomial");
}
{
  \\ total indsets count over ordered forests
  \\ A006013 binomial(3*n+1,n)/(n+1)
  for(n=0,9,
     my(vpar=vpar_preorder_first(n),
       count=0);
     until((vpar=vpar_preorder_next(vpar,1))=='none, \\ forests
           count += vpar_indsets_count(vpar));

     check_equal(count,
                 A006013_samples[n+1],
                 "A006013 vpar_indsets_count() total over ordered forests");

     check_equal(count, 1/(n+1) * binomial(3*n+1,n),
     "vpar_indsets_count() total over ordered forests, binomial formula");
  );

  check_equal(vector(#A006013_samples,n,n--;
                     1/(n+1)*binomial(3*n+1,n)),
              A006013_samples,  \\ OFFSET=0
              "A006013 binomial");
}

\\------------
\\ A055779 "fat" trees
test_A055779(n) =
{
  \\ given with OFFSET=1, dunno what if anything would be right for n=0
  if(n<1, error("test_A055779() is for n>=1"));

  sum(k=1,n, binomial(n,k)*k^(n-k)*n^(k-2));
}
test_indsets_count_total_forests_egf(limit) = \
  test_func_to_egf(n->test_A055779(n+1), limit);
{
  \\ total indsets count over labelled forests
  \\ A055779 num "fat" trees
  \\ 1, 1, 2, 10, 89, 1156, 19897

  check_equal(vector(#A055779_samples,n, test_A055779(n)),
              A055779_samples,   \\ OFFSET=1
              "A055779 by test_A055779()");
  my(forest=1);
  for(n=0, test_limit_trees_search - forest,
     my(count=0);
     forvec(vec=vpar_forvec_from(n,forest), \\ forests
           my(vpar=vpar_from_upascend(vec));
           count += vpar_indsets_count(vpar));
     check_equal(count, test_A055779(n+1),
                 "A055779 total indsets count over labelled forests");
     check_equal(count,
                 sum(k=1,n+1, binomial(n+1,k)*k^(n-k+1)*(n+1)^(k-2)),
                 "total indsets count over labelled forests");
  );
}
\\ A055779 total over all forests, so log as total over trees
test_indsets_count_total_trees(n) = \
  test_egf_term(1+log(test_indsets_count_total_forests_egf(n+1)), n);
{
  \\ total indsets count over labelled trees
  \\ not in OEIS: 1, 2, 6, 45, 528, 8525, 175428, 4391135
  \\ print(vector(8,n,n--; test_indsets_count_total_trees(n)));

  my(forest=0);
  for(n=0, test_limit_trees_search - forest,
     my(count=0);
     forvec(vec=vpar_forvec_from(n,forest), \\ trees
           my(vpar=vpar_from_upascend(vec));
           count += vpar_indsets_count(vpar));
     check_equal(count, test_indsets_count_total_trees(n),
                 "A055779 total indsets count over labelled trees");
  );
}

\\------------
test_A295623(n) = sum(k=0,n, binomial(n,k)*k^(n-k)*n^k);
test_A295623_egf(limit) = test_func_to_egf(test_A295623,limit);
\\
\\ A295623 total over all rootless, so log as total over one-component
\\ rootless, so the exp() of that gives
test_indsets_count_total_rootless_onecomp(n) = \
  test_egf_term(1+log(test_A295623_egf(n+1)), n);
\\ not in OEIS: 1, 7, 68, 962, 17844, 410172, 11255614, 359124768
\\ print(vector(8,n, test_indsets_count_total_rootless_onecomp(n)));
\\
{
  \\ total indsets count over rootless functional digraphs
  \\ choose k vertices to be not in the set
  \\ those k can have any parent 1..n, so n^k
  \\ other n-k in set have one of the k non-set as parent, so k^(n-k)
  \\ can have k=0; that will be 0^(n-k) = 0 ways n!=0, or 0^0=1 when n=0

  check_equal(vector(#A295623_samples,n,n--;
                    test_A295623(n)),
              A295623_samples,  \\ OFFSET=0
              "test_A295623() vs A295623_samples");

  check_equal(vector(20,n,n--; test_A295623(n)),
              vector(20,n,n--; if(n==0,1, n^2 * test_A055779(n))),
              "A295623 = n^2 * A055779");

  \\ A295623 a(n) = n! * [x^n] exp(n*x*exp(x)).
  check_equal(vector(#A295623_samples,n,n--;
                my(L=O(x^(n+1)));
                n!*polcoeff(exp(n*sum(k=1,n,x^k/(k-1)!) + L), n)),
              A295623_samples,
              "A295623 with sum");
  check_equal(vector(#A295623_samples,n,n--;
                my(L=O(x^(n+1)));
                n!*polcoeff(exp(n*x*exp(x + L)), n)),
              A295623_samples,
              "A295623 with exp");
  check_equal(vector(#A295623_samples,n,n--;
                my(L=O(x^(n+1)));
                n!*test_gf_terms(exp(n*x*exp(x + L)), n+1)[n+1]),
              A295623_samples,
              "A295623 with test_gf_terms()");
  check_equal(vector(#A295623_samples,n,n--;
                my(L=O(x^(n+1)));
                test_egf_terms(exp(n*x*exp(x + L)), n+1)[n+1]),
              A295623_samples,
              "A295623 with test_egf_terms()");

  for(n=0,5,
     my(count=0,
        count_onecomp=0);
     forvec(vpar=vector(n,i,[1,n]), \\ rootless
           \\ PENDING: vpar_indsets_count() allowing cyclics
           my(is_onecomp = (vpar_num_components(vpar)<=1));
           forvec(flags=vector(n,i,[0,1]),
                 if(vpar_is_indset_flags(vpar,flags),
                    count++;
                    count_onecomp += is_onecomp)));

     check_equal(count, test_A295623(n),
     "A295623 total indsets count over rootless functional digraphs");
     check_equal(count_onecomp, test_indsets_count_total_rootless_onecomp(n),
     "total indsets count over rootless functional digraphs of one component");
  );
}

{
  \\ total indsets count over functional digraphs
  \\ choose k vertices to be not in the set
  \\ those k can have any parent 0..n, so (n+1)^k
  \\ other n-k in set have one of the k or 0 as parent, so (k+1)^(n-k)

  for(n=0,5,
     my(count=0);
     forvec(vpar=vector(n,i,[0,n]),
           \\ PENDING: vpar_indsets_count() allowing cyclics
           forvec(flags=vector(n,i,[0,1]),
                 count += vpar_is_indset_flags(vpar,flags)));
     check_equal(count,
                 sum(k=0,n, binomial(n,k)*(k+1)^(n-k)*(n+1)^k),
                 "total indsets count over functional digraphs"));
}

{
  \\ A000312 n^n
  \\ 1, 1, 4, 27, 256, 3125, 46656
  check_equal(vector(#A000312_samples,n,n--;
                     if(n==0,1,
                        test_egf_terms(exp( n*x*exp(x * O(x^n))),n+1)[n+1])),
              A000312_samples,  \\ OFFSET=0
              "A000312")
}


\\-----------------------------------------------------------------------------
progress("vpar_count_free()");

{
  check_equal(vpar_count_free(5,1), 10);
  check_equal(iferr(vpar_count_free(5,'forest);0, e, 1),
              1,
    "vpar_count_free() error on forest=symbol (whereas it should be 0,1)");
}

{
  \\ unlabelled free trees, so vpar up to free isomorphism
  \\ A000055
  \\    n= 0  1  2  3  4  5  6
  my(want=[1, 1, 1, 1, 2, 3, 6, 11, 23, 47, 106, 235, 551, 1301, 3159, 7741,
           19320, 48629, 123867, 317955, 823065, 2144505, 5623756, 14828074]);
  check_equal(vector(#want,n,n--; vpar_count_free(n)), want,
              "vpar_count_free() trees");
  check_equal(vector(#A000055_samples,n,n--; vpar_count_free(n)),
              A000055_samples,  \\ OFFSET=0
              "vpar_count_free() = A000055");

  \\ Gang Li thesis equation 4.2 formula
  my(A000081(n)=if(denominator(n)>1,'none, vpar_count_unlabelled(n)));
  check_equal(vector(100,n, vpar_count_free(n)),
              vector(100,n,
                     simplify(
                     A000081(n)
                     - sum(i=1,n-1,my(j=n-i);
                           i+j==n || error();
                           if(i<=j, A000081(i) * A000081(j)))
                     + (1-(-1)^(n-1)) * binomial(A000081(n/2)+1,2) / 2
                             )));
}
{
  \\ unlabelled free forests, so vpar forests up to free isomorphism
  \\ A005195
  \\    n= 0  1  2  3  4  5  6
  my(want=[1,1,2,3,6,10,20,37,76,153,329,710,1601,3658,8599,20514,49905,122963,
           307199,775529,1977878,5086638,13184156,34402932,90328674,238474986]);
  check_equal(vector(#want,n,n--; vpar_count_free(n,1)), want,
              "vpar_count_free() forests");
  check_equal(vector(#A005195_samples,n,n--; vpar_count_free(n,1)),
              A005195_samples,  \\ OFFSET=0
              "vpar_count_free() forests = A005195");
}

test_count_unlabelled_free_trees_by_poly(n) =
{
  \\ code by Michael Somos in A000055
  my(A, A1, an, t);
  if(n<=1, 1,
     an=Vec(A = A1 = 1+O(x^n));
     for(m=2,n,
        my(i=m\2);
        an[m] = sum(k=1,i, an[k] * an[m-k])
                + (t = polcoeff(if(m%2, A *= (A1 - x^i)^-an[i], A),
                                m-1)));
     t + if(n%2==0, binomial(-polcoeff(A, n\2-1), 2)));
}
check_equal(vector(15,n,n--; vpar_count_free(n)), \
            vector(15,n,n--; test_count_unlabelled_free_trees_by_poly(n)));

\\ vpar_count_unlabelled(n,1) forests, calculated by summing over
\\ vpar_count_unlabelled() trees.
test_count_isomorphic_free_forests_by_trees_partition(n) =
{
  if(n<=1,return(1));
  my(a=vector(n,i,vpar_count_free(i)));
  my(total=0);
  forpart(p=n,
          my(r=test_partition_to_repetitions(p));
          my(f=prod(i=1,#r,
          \\ print("binomial "r[i]"  ", (a[r[i][1]] + r[i][2]-1)," "r[i][2]);
                    binomial(a[r[i][1]] + r[i][2]-1, r[i][2])));
          total += f;
          \\ print(Vec(p)" = "f"   a="apply(i->a[i],Vec(p))"   "r);
         );
  total;
}
{
  check_equal(vector(12,n,n--; vpar_count_free(n,1)),
              vector(12,n,n--; test_count_isomorphic_free_forests_by_trees_partition(n)),
              "vpar_count_free() free forests vs by partition of trees");
}
{
  check_equal(vpar_INTERNAL_Euler_transform
                (vector(12,n, vpar_count_free(n))),
              vector(12,n, test_count_isomorphic_free_forests_by_trees_partition(n)),
              "vpar_count_free() free forests vs Euler transform of trees");
}
{
  \\ In generating functions (ogfs), free trees F(x) from rooted trees R(x) is
  \\
  \\     R(x)                unicentroial, root is the centroid
  \\     - 1/2*(R(x)-1)^2    the unbalanced in R(x)
  \\     + 1/2*R(x^2)        bicentroidal isomorphic halves
  \\
  \\ This is the usual root-at-centroid per the explicit sums in
  \\ vpar_INTERNAL_rooted_vec_to_free(), but slightly obscured.
  \\
  \\ R(x) includes trees where the root is not a centroid.  These are where
  \\ the root has a child subtree size > than root and the rest.  Eg. term
  \\ n=5 want to discard combinations 1-4, 2-3.
  \\
  \\ (R(x)-1)^2 is size combinations 1-4, 2-3, 3-2, 4-1.  So 1/2 of it
  \\ excludes the desired non-centroid sizes.
  \\
  \\ For n even such as n=6, want to discard 1-5, 2-4.  They are in R(x)^2,
  \\ but R(x)^2 also has 3-3.  The 3-3 cases in R(x) are pairs both ways AA,
  \\ AB, BA, BB, one under the other.  In R(x)^2 likewise.  Want to keep AA
  \\ and BB two halves same.  Subtracting 1/2*R(x)^2 takes away 1/2 them,
  \\ and adding back 1/2*R(x^2) restores.
  \\
  \\ If isomorphic halves are not desired, then subtracting -1/2*R(x^2)
  \\ eliminates them instead of +1/2 padding them out.

  my(limit = 50,
     L = O('x^limit),
     gF = test_func_to_gf(vpar_count_free, limit),
     gR = test_func_to_gf(vpar_count_unlabelled, limit));

  check_equal(gF,
              gR - 1/2*(gR-1)^2 + 1/2*subst(gR-1,'x,'x^2),
              "vpar_count_free() by gf");
  check_equal(gF,
              1/2*(3 - (gR-2)^2 + subst(gR-1,'x,'x^2)),
              "vpar_count_free() by gf");
}
{
  \\ number of free unicentroidal A027416
  \\                bicentroidal  A102911
  my(limit=13,
     gF = test_func_to_gf(vpar_count_free, limit),
     gR = test_func_to_gf(vpar_count_unlabelled, limit),
     gU = O(x^limit),
     gB = O(x^limit));
  for(n=0,limit-1,
     my(vpar=vpar_free_first(n),
        count_uni=0,
        count_bi=0);
     until((vpar=vpar_free_next(vpar))=='none,
           if(#vpar_centroids(vpar) <= 1,
              count_uni++,
              count_bi++));

     gU += count_uni * x^n;
     gB += count_bi  * x^n;

     check_equal(count_uni, A027416_samples[n +1],
                 "A027416 number of unicentroidal free trees");

     check_equal(count_bi, if(n%2==1,0, A102911_samples[n/2 +1]),
                 "A102911 number of bicentroidal free trees");
     if(n%2==0 && n>1,
        \\ bicentroidals choose an n/2 for each half, duplicates allowed
        my(h=vpar_count_unlabelled(n/2));
        check_equal(count_bi, binomial(h+1,2),
                    "num bicentroidal free trees by choose halves")));

  \\ n=6  star-6    *--*--c--*--*        c--*--*
  \\                      |            / | \
  \\                      *           *  *  *
  \\ not  path-6    *--c--c--*     *--c--c--*--*
  \\                   |  |           |
  \\                   *  *           *
}


\\-----------------------------------------------------------------------------
progress("vpar_from_balanced_binary()");

check_equal(vpar_from_balanced_binary(0), [], \
            "vpar_from_balanced_binary() empty");
check_equal(vpar_from_balanced_binary(2), [0], \
            "vpar_from_balanced_binary() singleton");
check_equal(vpar_from_balanced_binary(10), [0,0]);
check_equal(vpar_from_balanced_binary(12), [0,1]);

{
  \\ addhelp() describing b increases with increasing preorder lex()
  for(b=0,2^12,
     vpar_is_balanced_binary(b) || next;
     check_equal(if(b,logint(b,2),-1), #binary(b)-1);  \\ high 1-bit position

     my(vpar=vpar_from_balanced_binary(b));
     vpar_validate(vpar);
     check_equal(b, vpar_to_balanced_binary(vpar),
                 "vpar_from_balanced_binary() and vpar_to_balanced_binary"));
}


\\-----------------------------------------------------------------------------
progress("vpar_preorder_next()");
\\ vpar_preorder_prev()

{
  check_equal(vpar_count_ordered(2,1), 2);
  check_equal(vpar_preorder_first(2), [0,1]);
  check_equal(vpar_preorder_next([0,1],1), [0,0]);
  for(n=2,5,
     my(vpar=vpar_preorder_first(n));
     check_equal(iferr(vpar_preorder_next(vpar,'forest);0, e, 1),
                 1,
                 Str("vpar_preorder_next() error on forest=symbol (whereas it should be 0,1), n="n)));
}
{
  check_equal(vpar_count_ordered(2,1), 2);
  check_equal(vpar_preorder_last(2,1), [0,0]);
  check_equal(vpar_preorder_prev([0,0],1), [0,1]);
  for(n=3,5,
     my(vpar=vpar_preorder_last(n));
     check_equal(iferr(vpar_preorder_prev(vpar,'forest);0, e, 1),
                 1,
                 Str("vpar_preorder_prev() error on forest=symbol (whereas it should be 0,1), n="n)));
}

{
  for(forest=0,1,
     for(n=0,10-forest,
        my(vpar=vpar_preorder_first(n),
           prev_depths='none,
           prev_vpar='none,
           seen=List([]),
           count=0,
           want_tree=(forest==0));
        until((vpar=vpar_preorder_next(vpar,forest)) == 'none,
              vpar_validate(vpar);
              if(!forest,
                 check_equal(vpar_is_tree(vpar), 1));
              check_equal(vpar_is_preorder(vpar), 1,
                         "vpar_preorder_next() always vpar_is_preorder()");
              check_equal(vpar_is_downwards(vpar), 1,
                         "vpar_preorder_next() always vpar_is_downwards()");

              \\ print(vpar);
              check_equal(vpar_preorder_prev(vpar,forest), prev_vpar,
                          "vpar_preorder_prev()");

              \\ lex order shown in vpar_preorder_next() addhelp
              my(depths = vpar_depths_vector(vpar));
              if(prev_depths!='none,
                 check_equal(lex(prev_depths,depths), 1));

              if(prev_vpar!='none,
                 check_equal(lex(prev_vpar,vpar), 1));

              my(pos=setsearch(seen,vpar,1));
              check_equal(pos>0, 1, "vpar_preorder_next() not seen before");
              listinsert(seen,vpar,pos);
              count++;

              prev_vpar = vpar;
              prev_depths = depths);

        check_equal(#seen, count);
        check_equal(count, vpar_count_ordered(n,forest),
                    Str("vpar_count_ordered() n="n" forest="forest))));
}


\\-----------------------------------------------------------------------------
progress("vpar_unlabelled_next()");

{
  my(vpar=vpar_unlabelled_first(5));
  check_equal(iferr(vpar_unlabelled_next(vpar,'forest);0, e, 1),
              1,
  "vpar_unlabelled_next() error on forest=symbol (whereas it should be 0,1)");
}
{
  for(forest=0,1,
     for(n=0,10-forest,
        my(vpar=vpar_unlabelled_first(n),
           prev_depths='none,
           prev_vpar='none,
           seen=List([]),
           count=0);
        while(1,
              vpar_validate(vpar);
              if(!forest,
                 check_equal(vpar_is_tree(vpar), 1));
              check_equal(vpar_is_preorder(vpar), 1);
              check_equal(vpar_is_unlabelled(vpar), 1);

              my(depths = vpar_depths_vector(vpar));
              if(prev_depths!='none,
                 check_equal(lex(prev_depths, depths), 1,
                             "vpar_unlabelled_next() lex decreasing depths");
                 check_equal(lex(prev_vpar, vpar), 1,
                             "vpar_unlabelled_next() lex decreasing vpar");
                );

              my(pos=setsearch(seen,vpar,1));
              check_equal(pos>0, 1, "vpar_unlabelled_next() not seen before");
              listinsert(seen,vpar,pos);
              count++;

              prev_vpar = vpar;
              prev_depths = depths;

              my(new_vpar=vpar_unlabelled_next(vpar,forest));
              check_equal(vpar,prev_vpar, "vpar_unlabelled_next() no change original");
              if(new_vpar=='none, break);
              vpar=new_vpar);

        check_equal(prev_vpar, vpar_unlabelled_last(n,forest));
        check_equal(#seen, count);
        check_equal(count, vpar_count_unlabelled(n,forest));
     ));
}
{
  \\ expecting whole number of rows in sample data
  my(got=[]);
  for(n=1,#A346913_samples,
     my(vpar=vpar_unlabelled_first(n));
     until((vpar=vpar_unlabelled_next(vpar))=='none,  \\ trees
           if(#got>=#A346913_samples, break(2));
           my(levels = apply(d->d+1, vpar_depths_vector(vpar)));
           got=concat(got,levels)));
  check_equal(got, A346913_samples,
             "vpar_unlabelled_next() vs A346913 tree levels");
}
{
  \\ expecting whole number of rows in sample data
  my(got=[]);
  for(n=1,#A346914_samples,  \\ number of vertices
     my(vpar=vpar_unlabelled_first(n));
     until((vpar=vpar_unlabelled_next(vpar,1))=='none,  \\ forests
           if(#got>=#A346914_samples, break(2));
           got=concat(got,vpar)));
  check_equal(got, A346914_samples,
             "vpar_unlabelled_next() vs A346914 forests");
}


\\ check_equal(vpar_count_unlabelled(10),    719);
\\ check_equal(vpar_INTERNAL_Catalan_number(10-1), 4862);


\\-----------------------------------------------------------------------------
progress("vpar_to_free()");

{
  \\           1  2  3  4  5
  \\ split to [2, 3, 4, 9, 9, 5, 6, 7, 0]
  \\ perm [8, 7, 6, 1, 2, 3, 4, 5, 9]
  \\ seq  [4, 5, 6, 7, 8, 3, 2, 1, 9]
  \\
  check_equal(vpar_to_free([0,1,2,3,4,5,6,7]),
              [0,1,2,3,4, 1,6,7]);
}
{
  \\             1
  \\     L1     / \   \ \ \ \     L2
  \\           2   4  6 7 8 9
  \\          /    |             second more
  \\         3     5              vertices
  my(vpar=[0,1,2,1,4, 1,1,1,1]);
  check_equal(vpar_to_free(vpar), vpar);
}
{
  \\             1
  \\     L1     / \               L2
  \\           2    6
  \\          /|\   |            first more
  \\         3 4 5  7             vertices
  my(vpar=[0,1,2,2,2, 1,6]);
  check_equal(vpar_to_free(vpar), vpar);
}
{
  \\          1            3
  \\         /            / \
  \\        2            2   4
  \\       /            /     \
  \\      3            1       5
  \\     /
  \\    4
  \\   /
  \\  5
  my(vpar=[0,1,2,3,4]);
  vpar_validate(vpar);

  my(vpar_for=vpar_INTERNAL_for_free_automorphisms(vpar));
  check_equal(vpar_for, [2, 3, 0, 3, 4],
              "vpar_INTERNAL_for_free_automorphisms() path-5, rerooted");
  check_equal(vpar_INTERNAL_premax(vpar_for,1),
              Vecsmall([3,2,1,4,5]),
              "vpar_INTERNAL_premax() path-5 seq");

  my(got=vpar_to_free(vpar));
  vpar_validate(got);
  check_equal(got, [0,1,2,1,4], "vpar_to_free() path-5");
}
{
  for(n=9,13,
     my(all=Set(test_free_all(n)));
     # print("n="n" "#all" trees");

     \\ bicentral versus finding in list
     for(h=n\2-2,n\2,
        for(rep=1,10,
           my(L=vpar_make_random(h),
              R=vpar_make_random(n-h));
           for(i=0,1,
              my(vpar=vpar_join(L,random(#L)+1, R),
                 got=vpar_to_free(vpar));
              check_equal(setsearch(all,got)!=0, 1,
                          "vpar_to_free() bicentral vs list"));
           [L,R]=[R,L])));
}
{
  \\  free primary
  \\       1                    bicentral
  \\      /   \  \  \
  \\     2     6  8  9          2       1
  \\     |     |                |       | \ \
  \\     3     7                3       6  8 9
  \\     | \                    | \     |
  \\     4  5                   4  5    7
  \\                           L1        L2
  \\                         height 2 = height 2
  \\                         size 4   < size 5
  \\
  my(vpar=[0,1,2,3,3,1,6,1,1]);
  vpar_validate(vpar);

  my(got=vpar_to_free(vpar));
  vpar_validate(got);
  check_equal(got, vpar, "vpar_to_free() 9 primary already");
}

\\ {
\\   \\ forest
\\   \\        2    7   1           12      11       10
\\   \\        |    |              / \      / \      |
\\   \\        8    6             3   8    6   7     1
\\   \\       / \                /|   |\
\\   \\      3   5              4 9   2 5
\\   \\     / \                   for free
\\   \\    4   9
\\   \\
\\   \\       1     7    9
\\   \\      /|\    |
\\   \\     2 5 6   8
\\   \\    /|
\\   \\   3 4
\\   \\
\\   my(vpar=[0, 0, 8, 3, 8, 7, 0, 2, 3]);
\\   vpar_validate(vpar);
\\
\\   my(vpar_for=vpar_INTERNAL_for_free_automorphisms(vpar));
\\   check_equal(vpar_for, [10, 8, 12, 3, 8, 11,11, 12, 3, 0,0,0]);
\\   my(seq=vpar_INTERNAL_premax(vpar_for,5));
\\   check_equal(test_is_perm(seq), 1,
\\               "vpar_INTERNAL_premax() free seq 9 is perm");
\\   check_equal(seq, Vecsmall([3, 8,2,5, 4,9, 6,7, 1, 10,11,12]),
\\               "vpar_INTERNAL_premax() free seq 9");
\\
\\   my(got=vpar_to_free(vpar));
\\   vpar_validate(got);
\\   check_equal(got, [0,1,2,2,1,1, 0,7, 0],
\\               "vpar_to_free() 12");
\\ }

test_to_free_primary_bicentral_by_premax(vpar) =
{
  my(centres=vpar_centres(vpar));
  #centres==2 || error();
  \\ print("centres "centres);

  my(rerooted=vpar_reroot(vpar,centres[1]));
  rerooted[centres[2]] == centres[1] || error();
  rerooted[centres[2]] = 0;
  \\ print("rerooted "rerooted);
  rerooted[centres[1]] == 0 || error();
  rerooted[centres[2]] == 0 || error();

  my(subtree_sizes=vpar_subtree_sizes_vector(rerooted),
     s=lex(subtree_sizes[centres[1]], subtree_sizes[centres[2]]));
  if(s==-1,
     centres=Vecrev(centres),
     s==0,
     my(seq=vpar_seq_premax(rerooted));
     \\ print("seq "Vec(seq));
     for(i=1,#vpar,
        if(seq[i]==centres[1],
           \\ print("centres[1]="centres[1]" lex bigger");
           break());
        if(seq[i]==centres[2],
           \\ print("centres[2]="centres[2]" lex bigger, swap to [1]");
           centres=Vecrev(centres);
           break())));
  \\ print("centres sorted "centres", first bigger");
  vpar_to_unlabelled(vpar_reroot(vpar,centres[1]));
}

{
  \\        5            6               1
  \\        |           / \             /|\
  \\        3          2   3           2 4 5
  \\       / \         |   |\         /
  \\      2   1        4   1 5       3
  \\      |
  \\      4
  \\
  my(vpar=[3, 3, 5, 2, 0]);
  vpar_validate(vpar);

  my(vpar_for=vpar_INTERNAL_for_free_automorphisms(vpar));
  check_equal(vpar_for, [3, 6, 6, 2, 3, 0],
              "vpar_INTERNAL_for_free_automorphisms() 5");
  my(seq=vpar_INTERNAL_premax(vpar_for,5));
  check_equal(test_is_perm(seq), 1,
              "vpar_INTERNAL_premax() free seq 5 is perm");
  check_equal(seq, Vecsmall([3, 2,4, 1,5, 6]),
              "vpar_INTERNAL_premax() free seq 5");

  my(got=vpar_to_free(vpar));
  vpar_validate(got);
  my(want=[0,1,2,1,1]);
  vpar_validate(want);
  check_equal(got, want, "vpar_to_free() 5");

  check_equal(test_to_free_primary_bicentral_by_premax(vpar), want,
              "test_to_free_primary_bicentral_by_premax() 5");
}


{
  \\        1  L2          8                  1
  \\       / \ \          / \                / \
  \\   L1 2   5  7       2   1              2   6
  \\     /     \        /     \ \          /|    \
  \\    3       6      3       5  7       3 5     7
  \\   /              /         \        /         \
  \\  4              4           6      4           8

  my(vpar=[0,1,2,3, 1,5, 1]);
  vpar_validate(vpar);
  check_equal(vpar_to_unlabelled(vpar), vpar);

  my(free_list=Set(test_free_all(#vpar)));
  check_equal(setsearch(free_list,vpar), 9,
              "vpar_to_free() 7 is free");

  my(vpar_for=vpar_INTERNAL_for_free_automorphisms(vpar));
  check_equal(vpar_for, [8,8,2,3, 1,5, 1, 0]);
  check_equal(vpar_to_unlabelled(vpar_for),
              [0,1,2,3,2, 1,6,7]);

  my(got=vpar_to_free(vpar));
  vpar_validate(got);
  check_equal(got, vpar);
}



check_equal(vpar_to_free([]), [], \
            "vpar_to_free() empty");

check_equal(vpar_to_free([0]), [0], \
            "vpar_to_free() singleton");
check_equal(vpar_to_free([0,1]), [0,1], \
            "vpar_to_free() path-2");
check_equal(vpar_free_first(3), [0,1,1], "vpar_free_first() 3");
check_equal(vpar_to_free([0,1,2]), [0,1,1], \
            "vpar_to_free() path-3");

check_equal(vpar_to_free([0,1,2,3]), [0,1,2,1], \
            "vpar_to_free() path-4");

check_equal(vpar_to_free([0,0,2]), [0,1,0], \
            "vpar_to_free() forest path-2 and singleton");

\\ {
\\   \\ Forest Path-5 and Path-6
\\   \\
\\   \\       12         13
\\   \\        |        /  \
\\   \\        3       8    9           1          7
\\   \\       / \      |    |          / \        / \
\\   \\      2   4     7    10        2   5      8   10
\\   \\      |   |     |    |         |   |      |   |
\\   \\      1   5     6    11        3   6      9   11
\\   \\                               |
\\   \\                               4
\\   \\
\\   my(vpar=[0,1,2,3,4, 0,6,7,8,9,10]);
\\   vpar_validate(vpar);
\\
\\   my(vpar_above=vpar_add_roots_above_centres(vpar));
\\   check_equal(vpar_above,
\\               [2, 3, 12, 3, 4, 7, 8, 13, 13, 9, 10, 0, 0],
\\               "vpar_add_roots_above_centres() forest path-5 and path-6");
\\
\\   my(vpar_un=vpar_to_unlabelled(vpar_above));
\\   check_equal(vpar_un,
\\               [0, 1, 2, 3, 2, 5, 0, 7, 8, 9, 7, 11, 12],
\\               "vpar_to_unlabelled() of forest path-5 and path-6");
\\
\\   \\ FIXME: This is wrong.
\\   my(got=vpar_to_free(vpar));
\\   check_equal(got, [0, 1, 2, 1, 4, 0, 6, 7, 8, 6, 10],
\\               "vpar_to_free() forest path-5 and path-6");
\\
\\   \\ PENDING vpar_preorder_depths_is_free_primary()
\\   \\ check_equal(test_unlabelled_is_free_primary(got), 1,
\\   \\             "test_unlabelled_is_free_primary() forest path-5 and path-6");
\\
\\   \\ check_equal(vpar_to_unlabelled(got), got,
\\   \\             "vpar_to_unlabelled() forest path-5 and path-6 got");
\\ }


\\ {
\\   \\          1   6            1       6
\\   \\         /    |           / \      |
\\   \\        2     7          2   4     7
\\   \\       /                /     \
\\   \\      3                3       5
\\   \\     /
\\   \\    4
\\   \\   /
\\   \\  5
\\   my(vpar=[0,1,2,3,4, 0,6]);
\\   vpar_validate(vpar);
\\   my(got=vpar_to_free(vpar));
\\   vpar_validate(got);
\\   check_equal(got, [0,1,2,1,4, 0,6],
\\               "vpar_to_free() path-5 and path-2");
\\ }

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got_vpar=vpar_to_free(vpar));
     vpar_validate(got_vpar);
     check_equal(vpar_is_preorder(got_vpar), 1,
                 "vpar_to_free() is preorder");
     );
}
{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i],
        got_vpar=vpar_to_free(vpar));
     vpar_validate(got_vpar);
     check_equal(vpar_is_preorder(got_vpar), 1,
                 "vpar_to_free() is preorder");
     check_equal(got_vpar,
                 vpar_to_unlabelled(got_vpar),
                 "vpar_to_free() is rooted premax");

     my(centres=vpar_centres(vpar));
     if(#centres==1,
        my(by_premax=vpar_to_unlabelled
                             (vpar_reroot(vpar,centres[1])));
        check_equal(got_vpar,by_premax);
        ,
        #centres==2,
        my(by_premax =
           test_to_free_primary_bicentral_by_premax(vpar));
        check_equal(got_vpar,by_premax,
                    "vpar_to_free() bicentral by premax");
       );
  );
}

     \\ if(#vpar <= test_limit_perms_search-1,
     \\    my(want_vpar=test_relabel_free_premax_by_search(vpar));
     \\    check_equal(vpar_is_preorder(want_vpar), 1,
     \\                "test_relabel_free_premax_by_search() is preorder");
     \\
     \\    my(want_depths=vpar_depths_vector(want_vpar),
     \\       got_depths=vpar_depths_vector(got_vpar));
     \\    \\ print("vpar  "vpar);
     \\    \\ print("got_depths  "got_depths);
     \\    \\ print("want_depths "want_depths);
     \\
     \\    check_equal(got_depths,want_depths,
     \\                "vpar_to_free() vs perm search, depths");
     \\    check_equal(got_vpar,want_vpar,
     \\                "vpar_to_free() vs perm search, vpars"));

{
  my(forest=0);
  for(n=0, test_limit_trees_search - forest,
     if(option_verbose, print("  n="n));
     my(free_list=Set(test_free_all(n)));
     forvec(vec=vpar_forvec_from(n,forest), \\ trees
           my(vpar=vpar_from_upascend(vec),
              vpar_relabelled=vpar_to_free(vpar));
           \\ print(vpar" relabelled "vpar_relabelled);
           check_equal(setsearch(free_list,vpar_relabelled)!=0, 1,
                       "vpar_to_free() vs test_free_all()");
           check_equal(vpar_is_free_isomorphic(vpar,vpar_relabelled), 1,
                       "vpar_is_free_isomorphic() vs relabelled");
     );
  );
}

\\ WRONG
\\ test_relabel_free_premax_by_search(vpar) =
\\ {
\\   print("test_relabel_premax_by_search()");
\\   my(max_vpar=vpar_relabel_preorder(vpar),
\\      max_depths=vpar_depths_vector(max_vpar));
\\   print(" initial depths "max_depths);
\\   for(i=1,(#vpar)!,
\\      my(perm=numtoperm(#vpar,i),
\\         this_vpar = vpar_relabel_perm(vpar,perm));
\\      for(j=1,vpar_num_rootings(vpar),
\\         my(this_vpar=vpar_reroot_ith(this_vpar,j));
\\         vpar_is_preorder(this_vpar) || next();
\\         my(this_depths = vpar_depths_vector(this_vpar));
\\         print("  try i="i" j="j"  "this_vpar" depths "this_depths);
\\         if(lex(this_depths,max_depths) > 0,
\\            print("  new high vpar="this_vpar" depths "this_depths" vs "max_depths);
\\            max_vpar = this_vpar;
\\            max_depths = this_depths)));
\\
\\   print(" final return vpar="max_vpar" depths "max_depths);
\\   max_vpar;
\\ }



{
  \\        1  L2          8                  1
  \\       / \ \          / \                / \
  \\   L1 2   5  7       2   1              2   6
  \\     /     \        /     \ \          /|    \
  \\    3       6      3       5  7       3 5     7
  \\   /              /         \        /         \
  \\  4              4           6      4           8

  my(vpar=[0,1,2,3, 1,5, 1]);
  vpar_validate(vpar);
  check_equal(vpar_to_unlabelled(vpar), vpar);

  my(free_list=Set(test_free_all(#vpar)));
  check_equal(setsearch(free_list,vpar), 9,
              "vpar_to_free() 7 is free");

  my(vpar_for=vpar_INTERNAL_for_free_automorphisms(vpar));
  check_equal(vpar_for, [8,8,2,3, 1,5, 1, 0]);
  check_equal(vpar_to_unlabelled(vpar_for),
              [0,1,2,3,2, 1,6,7]);

  my(got=vpar_to_free(vpar));
  vpar_validate(got);
  check_equal(got, vpar);
}


\\-----------------------------------------------------------------------------
progress("vpar_matchings_count() of complete binary tree");

check_equal(vpar_make_complete_tree(2, 0), [0]);

\\ of n levels, so height n-1
test_complete_binary_tree_matchings_count(n) =
{
  \\ my(x=1,y=1); for(i=1,n, [x,y] = [x^2 + 2*x*y, x^2]); x;
  my(x=1,y=1); for(i=2,n, [x,y] = [x^2 + 2*x*y^2, x]); x;
}

\\ ret[n] = with n levels
test_complete_binary_tree_matchings_count_vector(len) =
{
  my(ret=Vec([1,3],len));
  \\ for(i=3,len, ret[i] = ret[i-1]^2 + 2*ret[i-1]*ret[i-2]^2);
  for(i=3,len, ret[i] = ret[i-1]*(ret[i-1] + 2*ret[i-2]^2));
  ret;
}

{
  check_equal(vector(#A338293_samples,n,n-=2;
                     vpar_matchings_count(vpar_make_complete_tree(2,n))),
              A338293_samples,
              "A338293 number of matchings in complete binary tree (mine)");

  check_equal(vector(10,n,n--;
                     test_complete_binary_tree_matchings_count(n)),
              vector(10,n,n--;
                     my(vpar=vpar_make_complete_tree(2,n-1));
                     #vpar_row_vertices_vector(vpar) == n || error();
                     vpar_matchings_count(vpar)),
              "test_complete_binary_tree_matchings_count() vs tree");

  check_equal(test_complete_binary_tree_matchings_count_vector(10),
              vector(10,n,
                     my(vpar=vpar_make_complete_tree(2,n-1));
                     #vpar_row_vertices_vector(vpar) == n || error();
                     vpar_matchings_count(vpar)),
              "test_complete_binary_tree_matchings_count_vector() vs tree");

  check_equal(vector(10,n, n),
              vector(10,n, #vpar_row_vertices_vector(vpar_make_complete_tree(2,n-1))),
              "complete binary tree n rows");
  check_equal(vector(10,n, 2^n-1),
              vector(10,n, #vpar_make_complete_tree(2,n-1)),
              "complete binary tree n rows num vertices");
}


\\-----------------------------------------------------------------------------
progress("vpar_count_free_vector()");

{
  check_equal(vpar_count_free_vector(5,1), [1,2,3,6,10]);
  check_equal(iferr(vpar_count_free_vector(5,'forest);0, e, 1),
              1,
  "vpar_count_free_vector() error on forest=symbol (whereas it should be 0,1)");
}
{
  for(forest=0,1,
     for(n=0,16,
        my(got=vpar_count_free_vector(n,forest));
        check_equal(#got,n,
                    "vpar_count_free_vector() length");
        check_equal(got, vector(n,i,vpar_count_free(i,forest)),
                    "vpar_count_free_vector() vs vpar_count_free()")));
}


\\-----------------------------------------------------------------------------
progress("vpar_from_primecode()");
\\
\\ More in devel/test-Matula.gp

check_equal(vpar_from_primecode(0),[], "vpar_from_primecode() empty");
check_equal(vpar_from_primecode(1),[0], "vpar_from_primecode() singleton");
check_equal(vpar_from_primecode(2),[0,1], "vpar_from_primecode() path-2");

{
  \\     1
  \\    / \
  \\   2   3
  \\       |
  \\       4
  check_equal(vpar_from_primecode(6),[0,1,1,3], "vpar_from_primecode() 6");
}

test_from_primecode_by_recurse(primecode) =
{
  if(primecode==0,return([]));
  if(primecode==1,return([0]));
  my(m=factor(primecode),
     vpar=[0]);
  for(i=1,matsize(m)[1],
     my(subtree=test_from_primecode_by_recurse(primepi(m[i,1])));
     for(j=1,m[i,2],
        vpar=vpar_join(vpar,1,subtree)));
  vpar;
}
{
  for(rep=1,20,
     my(primecode=random(100000));
     my(vpar=vpar_from_primecode(primecode));
     vpar_validate(vpar);
     check_equal(vpar, test_from_primecode_by_recurse(primecode));

     \\ current implementation makes pre-order
     check_equal(vpar_is_preorder(vpar), 1,
                 "vpar_from_primecode() currently is pre-order");

     my(primecode_again=vpar_to_primecode(vpar));
     check_equal(primecode_again,primecode));
}

{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i],
        primecode=vpar_to_primecode(vpar));
     vpar=vpar_from_primecode(primecode);
     check_equal(vpar_is_preorder(vpar), 1,
                 "vpar_from_primecode() currently is pre-order");

     my(subcodes=vector(#vpar,v, vpar_to_primecode(vpar_subtree(vpar,v))),
        sets=vpar_siblings_sets(vpar));
     for(i=1,#sets,
        my(siblings=sets[i]);
        for(j=2,#siblings,
           check_equal(subcodes[siblings[j-1]] <= subcodes[siblings[j]], 1,
                       "vpar_from_primecode() currently in ascending subcode order"))));
}


\\ WRONG, probably
\\ test_seq_premin(vpar) =
\\ {
\\   Vecrev(vpar_seq_postorder(vpar_to_unlabelled(vpar)));
\\ }
\\ test_to_premin(vpar) =
\\ {
\\   vpar_relabel_seq(vpar,test_seq_premin(vpar));
\\ }


\\------------------------------------------------------------------------------
progress("vpar_num_rootings()");

check_equal(vpar_num_rootings([]), 1, \
            "vpar_num_rootings() empty tree 1 rooting");
{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i],
        got=vpar_num_rootings(vpar));
     check_equal(1 <= got && got <= vpar_num_rootings_max(#vpar), 1,
                 "vpar_num_rootings() range vs vpar_num_rootings_max()");
     check_equal(got,
                 if(#vpar==0,1,#vpar),
                 "vpar_num_rootings() of tree"));
}


\\------------------------------------------------------------------------------
progress("vpar_num_free_orbits()");

check_equal(vpar_num_free_orbits([]), 0, \
           "vpar_num_free_orbits() empty");
check_equal(vpar_num_free_orbits([0,1]), 1, \
           "vpar_num_free_orbits() path-2");

\\ this is a bit slow by the time repeated canonicals
test_num_free_orbits_by_rerooting(vpar) =
{
  vpar_is_tree(vpar)
    || error("test_num_free_orbits_by_rerooting() is for trees only");
  if(#vpar==0, return(0));  \\ no orbits at all in empty tree
  my(seen=[]);
  for(i=1,vpar_num_rootings(vpar),
     my(reroot_vpar=vpar_reroot_ith(vpar,i));
     reroot_vpar=vpar_to_unlabelled(reroot_vpar);  \\ canonical
     seen=setunion(seen,[reroot_vpar]));
  #seen;
}

{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i]);
     my(num=vpar_num_free_orbits(vpar));
     check_equal(num, test_num_free_orbits_by_rerooting(vpar),
                 "vpar_num_free_orbits() vs test_num_free_orbits_by_rerooting()"));
}

{
  for(n=1,14,
     my(total=0,
        vpar=vpar_free_first(n));
     until((vpar=vpar_free_next(vpar))=='none,
           my(num=vpar_num_free_orbits(vpar));
           total += num);
     check_equal(total, vpar_count_unlabelled(n),
     "vpar_num_free_orbits() total = vpar_count_unlabelled()"));
}


\\-----------------------------------------------------------------------------
progress("vpar_INTERNAL_attach_transform_result()");

{
  \\ transform of 1,1,1,1 forest=0 is factorials
  check_equal(vector(20,n,
                     vpar_INTERNAL_attach_transform_result(vector(n,i,1), 0)),
              vector(20,n, n!));
}
{
  \\ transform of 1,1,1,1 forest=1 is even permutations (even num of swaps)
  my(samples=A001710_samples[3..#A001710_samples]);
  check_equal(vector(#samples,n,
                    vpar_INTERNAL_attach_transform_result(vector(n,i,1), 1)),
              samples,
  "vpar_INTERNAL_attach_transform_result() of 1,1,1 forest=1 is A001710");
}

test_odd_double_factorial(n) = prod(i=1,n, 2*i-1);

{
  \\ transform of 2,2,2,2 forest=0 is A097801
  check_equal(vector(20,n,
                     vpar_INTERNAL_attach_transform_result(vector(n,i,2), 0)),
              vector(20,n, (2*n)!/(n!*2^(n-1)) ));
}
{
  \\ transform of 2,2,2,2 forest=1 is A293962 chords = 2/3 * double factorial
  check_equal(vector(#A293962_samples,n,n-=2;
                     if(n<=0,1,
                        vpar_INTERNAL_attach_transform_result(vector(n,i,2), 1))),
              A293962_samples,
              "vpar_INTERNAL_attach_transform_result() of 2,2,2,2 forest=1");
  check_equal(vector(20,n, vpar_INTERNAL_attach_transform_result(vector(n,i,2),1)),
              vector(20,n,n++; 2/3*test_odd_double_factorial(n)));
}

{
  \\ transform of 1,2,3,4,... forest=0
  check_equal(vector(#A058799_samples,n,
                     vpar_INTERNAL_attach_transform_result([1..n], 0)),
              A058799_samples);
}

{
  \\ transform of 1,2,3,4 forest=1
  \\ A075374 -1, -2, -1, 0, 1, 4, 19, 110, ... drop initial 4 terms
  my(samples=A075374_samples[5..#A075374_samples]);
  check_equal(vector(#samples,n,
                     vpar_INTERNAL_attach_transform_result([1..n], 1)),
              - samples);
}
test_A075374_vector(len) =
{
  my(v=vector(len));
  if(len>=1,
     v[1]=1;
     if(len>=2,
        v[2]=2;
        for(n=1,len-2, v[n+2] = n*v[n+1] - v[n])));
  v;
}
{
  check_equal(test_A075374_vector(1), [1]);
  check_equal(vector(10,len,len--; #test_A075374_vector(len)),
              vector(10,len,len--; len),
              "test_A075374_vector() length");
  check_equal(test_A075374_vector(#A075374_samples),
              A075374_samples);
}
test_A075374(n) =
{
  n>=1 || error();
  test_A075374_vector(n)[n];
}
{
  check_equal(vector(10,n,n+=2; test_A075374(n)),
              vector(10,n,n+=2; (n-2)*test_A075374(n-1) - test_A075374(n-2)),
              "test_A075374() recurrence");
}

test_gA075374(len) = x*Polrev(test_A075374_vector(len));
{
  check_equal(poldegree(test_gA075374(1)), 1);  \\ high term x^1

  my(len=10,
     L = O(x^(len+1)),
     g=test_gA075374(len));
  check_equal(x^3*(1/x*g)' + L,
              sum(n=3,len, (n-2)*test_A075374(n-1)*x^n + L));
  check_equal(x^2*(g' - g/x) + L,
              sum(n=3,len, (n-2)*test_A075374(n-1)*x^n + L),
              "g'");
  check_equal(x^2*g + L,
              sum(n=3,len, test_A075374(n-2)*x^n + L));

  check_equal(x^2*(g' - g/x) - x^2*g  + x + 2*x^2 + L,
              g + L,
              "PDE");
  my(h=g/x);

  check_equal(x^3*h' + (-x-x^3)*h  + L,
              -x - 2*x^2  + L,
              "PDE in h");

  \\ \\ y' + P(x)*y = Q(x)
  \\ \\ y = 1/u(x) int Q(x)*u(x)  where u(x) = exp(int P(x))
  \\ my(P= (-x-x^3),
  \\    Q= -x - 2*x^2,
  \\    int_P = intformal(P));
  \\ check_equal(h' + P*h + L,  Q + L);
  \\ check_equal(int_P, -1/4*x^4 - 1/2*x^2);
  \\ my(u=exp(int_P + L),
  \\    int_Qu = intformal(Q*u));
  \\
  \\ check_equal(1/u * int_Qu  + L, h  + L,
  \\             "PDE solution h");
  \\ print(int_Qu);
  \\ quit;
}
{
  \\ EGF for A075374
  my(limit=10,
     L = O(x^limit),
     g=test_func_to_egf(n->if(n==0,0,test_A075374(n)),limit));
  check_equal(test_egf_terms(g,limit-1), \\ starting n=0
              concat(0,test_A075374_vector(limit-2)),
              "A075374 egf terms");

  \\ f(n-1) = intformal(g)
  check_equal(intformal(intformal(g))  + L,
              test_func_to_egf(n->if(n<3,0, test_A075374(n-2)),
                               limit),
              "g''");

  \\ n*f(n-1) = x*g
  check_equal(x*g - 2*intformal(g)  + L,
              test_func_to_egf(n->if(n<2,0, (n-2)*test_A075374(n-1)),
                               limit),
              "(n-2)*");

  check_equal(x*g - 2*intformal(g)  - intformal(intformal(g)) + L,
              g - x - x^2,
              "EGF recurrence");

  check_equal((x-1)*g - 2*intformal(g)  - intformal(intformal(g)) + L,
              - x - x^2  + L);
  check_equal(((x-1)*g)' - 2*g  - intformal(g) + L,
              -1 - 2*x  + L/x);
  check_equal((x-1)*g' - g - intformal(g) + L,
              -1 - 2*x  + L/x);

  check_equal(((x-1)*g')' - g' - g + L,
              -2  + L/x^2);
  check_equal((x-1)*g'' - g  + L  == -2, 1);
}

test_partA075374(n) =
{
  n>=1 || error();
  - test_A075374(n+4);
}
{
  check_equal(vector(5,n, test_partA075374(n)),
              [1, 4, 19, 110, 751]);
  check_equal(vector(15,n, vpar_INTERNAL_attach_transform_result([1..n],1)),
              vector(15,n, test_partA075374(n)));

  check_equal(vector(15,n,n+=2; (n+2)*test_partA075374(n-1)
                                    - test_partA075374(n-2)),
              vector(15,n,n+=2; test_partA075374(n)),
              "test_partA075374() recurrence");

  \\ Gary W. Adamson, partA075374 is "eigensequence" of A003991 as a triangle
  \\ http://qntmpkt.blogspot.com/2012/01/eigensequence-of-triangle-eigentriangle.html
  \\
  check_equal(vector(15,n, sum(i=1,n, (n+1-i)*i
                               * if(i==1,1,test_partA075374(i-1)))),
              vector(15,n, test_partA075374(n)));
  \\ or with exception taken out
  check_equal(vector(15,n, n + sum(i=2,n, (n+1-i)*i * test_partA075374(i-1))),
              vector(15,n, test_partA075374(n)));

  check_equal(vector(15,n,n+=2; n-2 + sum(i=2,n-2,
                                          (n+1-i-2)*i * test_partA075374(i-1))),
              vector(15,n,n+=2; test_partA075374(n-2)),
              "at n-2");

  check_equal(vector(15,n,n++; n-1 + sum(i=2,n-1,
                                     (n+1-i-1)*i * test_partA075374(i-1))),
              vector(15,n,n++; test_partA075374(n-1)),
              "at n-1");
  check_equal(vector(15,n,n++; (n+2)
                     *(n-1 + sum(i=2,n-1, (n+1-i-1)*i * test_partA075374(i-1)))),
              vector(15,n,n++; (n+2)*test_partA075374(n-1)),
              "(n+2)* at n-1");

  check_equal(vector(15,n,n+=2;
             (n+2)* (n-1 + sum(i=2,n-1, (n+1-i-1)*i * test_partA075374(i-1)))
                  - (n-2 + sum(i=2,n-2, (n+1-i-2)*i * test_partA075374(i-1))) ),
              vector(15,n,n+=2; test_partA075374(n)),
              "recurrence in sums");
}

\\-----------------------------------------------------------------------------
progress("vpar_adjacency_poly()");

{
  my(vpar=[0]);
  check_equal(vpar_matchpoly(vpar), Pol(1),
              "vpar_adjacency_poly() singleton, is a Pol");
  \\ subst() remains a Pol in GP 2.13 but lowered itself to t_INT in 2.11
  \\ and earlier, so compare with "=="
  check_equal(subst(vpar_matchpoly(vpar),'x,-'x^2) == 1,
              1);
  check_equal(Vecrev(1*'x/'x), [1]);
  check_equal(vpar_adjacency_poly(vpar), 'x);
}
check_equal(Polrev('x+6), 'x+6, "Polrev doesn't reverse terms");
check_equal(polrecip('x+6), 6*'x+1, "polrecip reverses terms");

{
  \\ clean to global x
  local(x='abcd, y='efgh);
  check_equal(vpar_adjacency_poly([0,1,2]),    'x^3 - 2*'x);
  check_equal(vpar_adjacency_poly([0,1,2],'y), 'y^3 - 2*'y);
}

{
  \\ A226594 number of distinct characteristic polynomials of trees
  my(limit=13);
  check_equal(vector(limit,n,n--;
                     test_count_free_distinct_func(n, vpar_adjacency_poly)),
              A226594_samples[1..limit]);  \\ OFFSET=0

  \\ n=4 vertices
  my(l=List([]),
     vpar=vpar_free_first(4));
  until((vpar=vpar_free_next(vpar))=='none,  \\ trees
        listput(l,vpar_adjacency_poly(vpar)));
  l=Set(l);
  check_equal(l, [x^4 - 3*x^2,
                  x^4 - 3*x^2 + 1]);
  check_equal(#l, 2);
  check_equal(test_count_free_distinct_func(4,vpar_adjacency_poly),
              2);
}

\\ random non-root
test_random_nonroot(vpar) =
{
  my(m=vpar_num_edges(vpar));
  if(m,
     m=random(m);
     for(v=1,#vpar,
        if(vpar[v] && m--<0, return(v))));
}

\\ Product of characteristic polynomial across an edge.
\\
test_adjacency_poly_by_withwithout(vpar,var='x) =
{
  my(seq=vpar_INTERNAL_seq_upwards(vpar),
     with     = vector(#vpar,v,var),    \\ with[v] = 'x
     without  = vector(#vpar,v,1),      \\ without[v] = 1
     ret      = var/var);           \\ 1 as poly
  for(i=1,#seq,
     my(v=seq[i],
        p=vpar[v]);
     if(p,
        with[p] = with[p]*with[v] - without[p]*without[v];
        without[p] *= with[v];
        ,
        ret *= with[v]);  \\ at root
     with[v]    =     \\ kill intermediate polys
     without[v] = 0);
  ret;
}
check_equal(test_adjacency_poly_by_withwithout([]), Pol(1));
check_equal(test_adjacency_poly_by_withwithout([0]), 'x);

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        poly=vpar_adjacency_poly(vpar));
     check_equal(type(poly), "t_POL");

     \\ print(vpar);
     \\ print("want "charpoly(vpar_adjacency_matrix(vpar)));
     \\ print("got  "vpar_adjacency_poly(vpar));
     \\ print("perf "vpar_has_perfect_matching(vpar));
     \\ print();

     check_equal(poly, charpoly(vpar_adjacency_matrix(vpar)),
                 "vpar_adjacency_poly() vs matrix charpoly");

     check_equal(poly, 'x^#vpar * subst(vpar_matchpoly(vpar), 'x, -1/'x^2),
                 "vpar_adjacency_poly() vs matchpoly shown in addhelp");
     check_equal(polcoeff(poly,#vpar), 1,
                 "vpar_adjacency_poly() high term 1");

     check_equal(test_poly_num_real_roots(poly), poldegree(poly),
                 "vpar_adjacency_poly() all roots real");

     \\ Collatz and Sinogowitz, odd terms = 0 for connected bipartite
     forstep(i=1,#vpar,2,
        check_equal(polcoeff(poly,#vpar-i), 0,
                 "vpar_adjacency_poly() odd term == 0"));

     \\ Collatz and Sinogowitz
     check_equal(polcoeff(poly,#vpar-2), -vpar_num_edges(vpar),
                 "vpar_adjacency_poly() third highest term");

     \\ Collatz and Sinogowitz n-3 term = -2*num triangles, which for a tree
     \\ is 0
     check_equal(polcoeff(poly,#vpar-3), 0,
                 "vpar_adjacency_poly() term n-3 = -2 * num triangles = 0");

     \\ terms high +1 then alternating downwards
     check_equal(abs(polcoeff(poly,0)), vpar_has_perfect_matching(vpar),
       "vpar_adjacency_poly() abs constant term vpar_has_perfect_matching()");
     check_equal(polcoeff(poly,0), (-1)^(#vpar/2) * vpar_has_perfect_matching(vpar),
       "vpar_adjacency_poly() constant term signed vpar_has_perfect_matching()");

     my(component_vpar=vpar_component_trees(vpar));
     check_equal(poly,
                 Pol(prod(i=1,#component_vpar, vpar_adjacency_poly(component_vpar[i]))),
                 "vpar_adjacency_poly() product of components");

     check_equal(poly, test_adjacency_poly_by_withwithout(vpar),
                 "vpar_adjacency_poly() vs test_adjacency_poly_by_withwithout()");

     \\ roots absolute value upper bound 2*sqrt(maxdegree-1)
     \\ need 2.11.x for polsturm on not square-free and this [] arg form
     if(lex(version(),[2,11]) >= 0,
       my(roots=polrootsreal(poly),
          degrees=vpar_degrees_vector(vpar),
          maxdegree=if(#degrees,vecmax(degrees),0),
          roots_bound=if(maxdegree==0, 0,
                         maxdegree==1, 1,
                         2*sqrt(maxdegree-1) + 1e-10));
       for(i=1,#roots,
          my(a=abs(roots[i]));
          \\ print(vpar" "a" "roots_bound);
          check_equal(a <= roots_bound, 1,
                      "vpar_adjacency_poly() roots upper bound"));
       check_equal(polsturm(poly,[-roots_bound, roots_bound]),
                   polsturm(poly),
                   "vpar_adjacency_poly() roots upper bound, polsturm same");
        );

     \\ per for example Brouwers "Small Integral Trees"
     \\ https://www.win.tue.nl/~aeb/graphs/integral_trees.html
     \\ product of characteristic polynomial across an edge c <-> p
     \\
     my(c=test_random_nonroot(vpar));
     if(c,
        my(p=vpar[c],
           c_vpar=vpar_subtree(vpar,c),
           p_vpar=vpar_delete_subtree(vpar,c),
           c_set=Set(vpar_subtree_vertices(vpar,c)),
           c_vpar_sans_c=vpar_keep_vertices(vpar,setminus(c_set,[c])),
           p_vpar_sans_p=vpar_delete_vertices(vpar,concat(c_set,[p])));
        check_equal(poly,
                    vpar_adjacency_poly(c_vpar)
                    * vpar_adjacency_poly(p_vpar)
                    -
                    vpar_adjacency_poly(c_vpar_sans_c)
                    * vpar_adjacency_poly(p_vpar_sans_p)));
  );
}

test_make_bistar_split_middle(a,b) =
{
  my(vpar=vpar_make_bistar(a,b));
  vpar=Vec(vpar,#vpar+1);
  vpar[1] = vpar[a+1] = #vpar;
  vpar;
}
check_equal(test_make_bistar_split_middle(1,1), [3,3,0]);
check_equal(test_make_bistar_split_middle(3,1), [5,1,1, 5,0]);

{
  \\ Abbe Mowshowitz, "The Characteristic Polynomial of a Graph", Journal of
  \\ Combinatorial Theory, series B, volume 12, 1972, pages 177-193.
  \\ https://www.sciencedirect.com/science/article/pii/0095895672900238
  \\
  \\ Cospectral tree pairs.
  \\
  \\     T1                T2
  \\  u      v         x         y    u,v x,y
  \\  *      *         *         *    many leaf each side
  \\   \    /           \       /
  \\  *-*--*-*         *-*--*--*-*
  \\   /    \           /       \
  \\  *      *         *         *
  \\
  \\ u+v=n-1  x+y=n-2  for total n+1 each
  \\ h2(T1) = u*v                 matchings with 2 pairs
  \\ h2(T2) = x*(y+1) + y
  \\ for equal h2 want
  \\ y^2 - v^2 + (n-1)*v - (n-2)*y - (n-2) = 0
  \\ take y=v+1 for
  \\ v = (2*n-5)/3
  \\ n= 3k+7, k>=0
  \\ vector(5,k,k--; my(n=3*k+7); (2*n-5)/3) == [3,5,7,9,11]
  \\ v odd >= 3
  \\ y=v+1 even >= 4

  my(poly = u*v - (x*(y+1)+y));
  poly = subst(poly, u, n-1-v);
  poly = subst(poly, x, n-2-y);
  check_equal(poly, y^2 - v^2 + (n-1)*v - (n-2)*y - (n-2));

  my(T1=(u,v)->vpar_make_bistar(u+1,v+1),
     T2=(x,y)->test_make_bistar_split_middle(x+1,y+1));
  for(u=2,5,
     for(v=2,5,
        my(vpar=T1(u,v));
        check_equal(polcoeff(vpar_matchpoly(vpar),2), u*v)));
  for(x=2,5,
     for(y=2,5,
        my(vpar=T2(x,y));
        check_equal(polcoeff(vpar_matchpoly(vpar),2), x*(y+1)+y)));

  my(k_to_v=(k)-> my(n=3*k+7); (2*n-5)/3,
     k_to_y=(k)-> k_to_v(k) + 1,
     k_to_u=(k)-> my(n=3*k+7); n-1 - k_to_v(k),
     k_to_x=(k)-> my(n=3*k+7); n-2 - k_to_y(k));
  my(T1_k=(k)->T1(k_to_u(k), k_to_v(k)),
     T2_k=(k)->T2(k_to_x(k), k_to_y(k)));

  check_equal(k_to_v(0), 3);
  check_equal(k_to_u(0), 3);
  check_equal(k_to_x(0), 1);
  check_equal(k_to_y(0), 4);

  \\  u      v         x         y   many leaf each
  \\  *      *                   *
  \\   \    /                   /_-*
  \\  *-*--*-*         *-*--*--* _
  \\   /    \                   \ -*
  \\  *      *                   *
  \\
  check_equal(T1_k(0), [0, 1, 1, 1, 1, 5, 5, 5]);
  check_equal(T2_k(0), [8, 1, 8, 3, 3, 3, 3, 0]);

  check_equal(vpar_is_free_isomorphic(T1_k(0), [0, 1, 2, 2, 2, 1, 1, 1]), 1);
  check_equal(vpar_is_free_isomorphic(T2_k(0), [0, 1, 2, 2, 2, 2, 1, 7]), 1);

  for(k=0,10,
     my(vpar1=T1_k(k),
        vpar2=T2_k(k));
     check_equal(vpar_is_free_isomorphic(vpar1,vpar2), 0);
     check_equal(vpar_adjacency_poly(vpar1),
                 vpar_adjacency_poly(vpar2));

     check_equal(vpar_matchpoly(vpar1),
                 vpar_matchpoly(vpar2));
     check_equal(vpar_matchnum(vpar1), 2);
     check_equal(vpar_matchnum(vpar2), 2);

     my(u = k_to_u(k),
        v = k_to_v(k),
        x = k_to_x(k),
        y = k_to_y(k),
        n = 3*k+7);
     check_equal(vpar_matchpoly(vpar1), u*v         * 'x^2 + n*'x + 1);
     check_equal(vpar_matchpoly(vpar2), (x*(y+1)+y) * 'x^2 + n*'x + 1);
     check_equal(u*v, (x*(y+1)+y));
  );
}

{
  \\ adjacency poly of path-n
  \\ 1
  \\ x
  \\ factor(x^2 - 1)
  \\ factor(x^3 - 2*x)
  \\ factor(x^4 - 3*x^2 + 1)
  my(limit=20,
     d=vector(limit));
  d[1] = 'x;
  d[2] = 'x*d[1] - 1;
  for(n=1,limit,
     my(vpar=vpar_make_path(n),
        poly=vpar_adjacency_poly(vpar),
        f=factor(poly),
        num_distinct_roots=sum(i=1,matsize(f)[1], poldegree(f[i,1])),
        num_ramifications=sum(i=1,matsize(f)[1], f[i,2]>=2));
     \\ print(f);
     check_equal(num_distinct_roots,n);
     check_equal(num_ramifications,0);

     \\ per Harary, King, Moshowitz
     if(n>2, d[n] = 'x*d[n-1] - d[n-2]);
     check_equal(poly,d[n]);

     \\ Gerd H. Frickey, Stephen T. Hedetniemi, David P. Jacobsz and Vilmar
     \\ Trevisan, "Reducing the Adjacency Matrix of a Tree", Electronic
     \\ Journal of Linear Algebra, volume 1, October 1996, pages 34-43.
     \\ http://gauss.technion.ac.il/iic/ela
     \\ For n even, adjacency poly of paths n and n+2 have no common root.
     if(n%2==0,
        my(vpar2=vpar_make_path(n+2),
           poly2=vpar_adjacency_poly(vpar2),
           f2=factor(poly2));
        check_equal(setintersect(Set(f[,1]),Set(f2[,1])), []));
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_adjacency_matrix()");

{
  my(vpar=[2,1],  \\ 2-cycle
     m=vpar_adjacency_matrix(vpar));
  check_equal(m, [0, 1;
                  1, 0]);
  check_equal(m^2, [1, 0;
                    0, 1]);
  check_equal(trace(m), 0);
}

{
  \\ Brouwer, "Spectra of Graphs" page 2 example path-3
  my(vpar=vpar_make_path(3),
     m=vpar_adjacency_matrix(vpar));
  check_equal(m, [0, 1, 0;
                  1, 0, 1;
                  0, 1, 0]);
  check_equal(charpoly(m), 'x^3 - 2*'x);
  \\ mateigen(m)
  \\ qfjacobi(m)
}
{
  \\ squared is paths length=2 ways u->v, with backtracking etc
  my(vpar=[0,1,1,1],
     m=vpar_adjacency_matrix(vpar));
  check_equal(m^2, [3, 0, 0, 0;
                    0, 1, 1, 1;
                    0, 1, 1, 1;
                    0, 1, 1, 1],
              "vpar_adjacency_matrix() star squared");
}

{
  \\ cubed   1 <- 2 <- 3 <- 4
  my(vpar=[0,1,2,3],
     m=vpar_adjacency_matrix(vpar));
  check_equal(m^3, [0, 2, 0, 1;
                    2, 0, 3, 0;
                    0, 3, 0, 2;
                    1, 0, 2, 0],
              "vpar_adjacency_matrix() path-4 cubed");
}

test_adjacency_matrix_by_loop(vpar) =
{
  my(m=matrix(#vpar,#vpar), p);
  for(v=1,#vpar, if((p=vpar[v]), m[v,p] = m[p,v] = 1));
  m;
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        m=vpar_adjacency_matrix(vpar));
     check_equal(m, test_adjacency_matrix_by_loop(vpar),
                 "vpar_adjacency_matrix() vs test_adjacency_matrix_by_loop()");
     check_equal(m, matrix(#vpar,#vpar,u,v, vpar_is_neighbour(vpar,u,v)),
                 "vpar_adjacency_matrix() vs individual vpar_is_neighbour()");

     \\ Collatz and Sinogowitz
     check_equal(trace(m), 0,
                 "vpar_adjacency_matrix() trace==0, per addhelp");

     \\ Bevis, Domke, Miller
     check_equal(matrank(m), 2*vpar_matchnum(vpar));

     \\ Gerd H. Frickey, Stephen T. Hedetniemi, David P. Jacobsz and Vilmar
     \\ Trevisan, "Reducing the Adjacency Matrix of a Tree", Electronic
     \\ Journal of Linear Algebra, volume 1, October 1996, pages 34-43.
     \\ http://gauss.technion.ac.il/iic/ela
     \\
     my(has_perfect_matching=vpar_has_perfect_matching(vpar));
     check_equal(matdet(m), if(#vpar%4==0 && has_perfect_matching, 1,
                               #vpar%4==2 && has_perfect_matching, -1,
                               0));
     check_equal(matdet(m), has_perfect_matching * (-1)^(#vpar\2));

     check_equal(trace(m^2), 2*vpar_num_edges(vpar));
     check_equal(vector(#vpar,u, (m^2)[u,u]), vpar_degrees_vector(vpar));
  );
}

test_num_triangles(vpar) =
{
  my(cycle_lengths = vpar_component_cycle_lengths(vpar));
  sum(i=1,#cycle_lengths, cycle_lengths[i]==3);
}
{
  \\ adjacency matrix trace(m^3) = 6*num_triangles
  check_equal(test_num_triangles([2,3,1]), 1);
  check_equal(vpar_skew_adjacency_matrix([2,3,1]), [0, 1, -1;
                                                    -1, 0, 1;
                                                    1, -1, 0]);
  check_equal(vpar_skew_adjacency_matrix([2,3,1])^2, [-2, 1, 1;
                                                      1, -2, 1;
                                                      1, 1, -2]);
  check_equal(vpar_adjacency_matrix([2,3,1])^3, [2, 3, 3;
                                                 3, 2, 3;
                                                 3, 3, 2]);
  check_equal(vpar_skew_adjacency_matrix([2,3,1])^3, [0, -3, 3;
                                                      3, 0, -3;
                                                      -3, 3, 0]);
}
{
  \\ self-loop allows its neighbours to have a 3-step path back to
  \\ themselves, 6--3--3--6
  \\ and the self-loop can step down to a neighbour
  \\                                      3--6--3--3
  \\                                      3--3--6--3
  \\                                      3--3--3--3  itself
  \\    v--+    v--+
  \\    3 -+    5 -+
  \\    |       |
  \\    6       4
  \\    |       |
  \\    2       1
  my(vpar=[4, 6, 3, 5, 5, 3]);
  check_equal(test_num_triangles(vpar), 0);
  check_equal(vpar_adjacency_matrix(vpar), [0, 0, 0, 1, 0, 0;
                                            0, 0, 0, 0, 0, 1;
                                            0, 0, 1, 0, 0, 1;
                                            1, 0, 0, 0, 1, 0;
                                            0, 0, 0, 1, 1, 0;
                                            0, 1, 1, 0, 0, 0]);
  check_equal(vpar_adjacency_matrix(vpar)^2, [1, 0, 0, 0, 1, 0;
                                              0, 1, 1, 0, 0, 0;
                                              0, 1, 2, 0, 0, 1;
                                              0, 0, 0, 2, 1, 0;
                                              1, 0, 0, 1, 2, 0;
                                              0, 0, 1, 0, 0, 2]);
  check_equal(vpar_adjacency_matrix(vpar)^3, [0, 0, 0, 2, 1, 0;
                                              0, 0, 1, 0, 0, 2;
                                              0, 1, 3, 0, 0, 3;
                                              2, 0, 0, 1, 3, 0;
                                              1, 0, 0, 3, 3, 0;
                                              0, 2, 3, 0, 0, 1]);
  check_equal(vpar_adjacency_matrix([1])^3, Mat([1]));
  check_equal(vpar_adjacency_matrix([1,2])^3, [1, 0;
                                               0, 1]);
}
{
  \\ adjacency matrix trace(m^3) = 6*num_triangles
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        m = vpar_adjacency_matrix(vpar),
        num_triangles = test_num_triangles(vpar),
        num_selfloops = sum(v=1,#vpar, vpar[v]==v));
     if(num_selfloops==0,
        check_equal(trace(m^3),
                    6*num_triangles,
                    "vpar_adjacency_matrix() ^3 trace = num triangles, no self loops")));
}

\\ Return the number of self-loops in vpar.
test_num_selfloop(vpar) = sum(v=1,#vpar, vpar[v]==v);

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        m=vpar_adjacency_matrix(vpar));

     check_equal(trace(m), test_num_selfloop(vpar),
                 "vpar_adjacency_matrix() trace==num self-loops, per addhelp");

     );
}
{
  \\ count of how many vpar_adjacency_matrix() occur, per the addhelp
  for(forest=0,1,
     for(n=0, test_limit_trees_search - forest,
        my(l=List([])); \\ distinct matrices
        forvec(vec=vpar_forvec_from(n,forest), \\ forests
              my(vpar=vpar_from_upascend(vec),
                 m=vpar_adjacency_matrix(vpar));
              my(pos=setsearch(l,m,1));
              if(pos, listinsert(l,m,pos)));

        check_equal(#l, vpar_count_unrooted(n,forest),
                    "vpar_adjacency_matrix() count distinct")));
}


\\-----------------------------------------------------------------------------
progress("vpar_skew_adjacency_matrix()");

check_equal(vpar_skew_adjacency_matrix([0]), Mat([0]), \
            "vpar_skew_adjacency_matrix() singleton");

\\ Not yet specified on self-loops.
\\ check_equal(vpar_skew_adjacency_matrix([1]), Mat([1]), \
\\             "vpar_skew_adjacency_matrix() self-loop");
\\ check_equal(vpar_skew_adjacency_matrix([2,1]), [0,1; -1,0], \
\\             "vpar_skew_adjacency_matrix() 2-cycle");

check_equal(vpar_skew_adjacency_matrix([3,1,2]), [0,-1,1; 1,0,-1; -1,1,0], \
            "vpar_skew_adjacency_matrix() 3-cycle");

test_skew_adjacency_matrix_by_loop(vpar) =
{
  my(m=matrix(#vpar,#vpar), p);
  for(v=1,#vpar, if((p=vpar[v]), m[v,p]=1; m[p,v]=-1));
  m;
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        m = vpar_skew_adjacency_matrix(vpar));
     check_equal(m, test_skew_adjacency_matrix_by_loop(vpar));

     \\ similar to vpar_adjacency_matrix()

     \\ Collatz and Sinogowitz
     check_equal(trace(m), 0, "vpar_skew_adjacency_matrix() trace==0");

     \\ Bevis, Domke, Miller
     check_equal(matrank(m), 2*vpar_matchnum(vpar));

     check_equal(matdet(m), vpar_has_perfect_matching(vpar));
     check_equal(trace(m^2), -2*vpar_num_edges(vpar));
     check_equal(vector(#vpar,u, -(m^2)[u,u]), vpar_degrees_vector(vpar));

     \\ usual substitutions
     check_equal(vpar_matchings_count(vpar),
                 matdet(matid(#vpar) + m));
     check_equal(charpoly(m),
                 Pol(matdet(matid(#vpar)*'x + m)));
  );
}

{
  \\ all vpar_skew_adjacency_matrix() are distinct
  my(forest=1);
  for(n=0, test_limit_trees_search - forest,
     my(l=List([])); \\ distinct matrices
     forvec(vec=vpar_forvec_from(n,forest), \\ forests
           my(vpar=vpar_from_upascend(vec),
              m=vpar_skew_adjacency_matrix(vpar));
           my(pos=setsearch(l,m,1));
           check_equal(pos!=0, 1);
           if(pos, listinsert(l,m,pos)));

        check_equal(#l, vpar_count(n,forest),
                    "vpar_skew_adjacency_matrix() count distinct"));
}


\\-----------------------------------------------------------------------------
progress("vpar_from_skew_adjacency_matrix()");

test_girth(vpar) =
{
  my(lens=vpar_component_cycle_lengths(vpar));
  lens=select(l->l!=0, lens);
  if(#lens, vecmin(lens), 0);
}
check_equal(test_girth([]), 0);
check_equal(test_girth([0]), 0);
check_equal(test_girth([1]), 1);
check_equal(test_girth([2,1]), 2);
check_equal(test_girth([3,1,2, 1,1,1]), 3);
check_equal(test_girth([2,1, 0,3]), 2);

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        girth=test_girth(vpar));
     if(girth==1||girth==2, next);

     my(m=vpar_skew_adjacency_matrix(vpar),
        got=vpar_from_skew_adjacency_matrix(m));
     check_equal(got, vpar,
                 "vpar_from_skew_adjacency_matrix() restore"));
}


\\-----------------------------------------------------------------------------
progress("vpar_num_childless()");

check_equal(vpar_num_childless([]), 0, "vpar_num_childless() empty");
check_equal(vpar_num_childless([0]), 1, "vpar_num_childless() singleton");
check_equal(vpar_num_childless([1]), 0, "vpar_num_childless() self-loop");

test_num_childless_by_hamming(vpar) =
{
  my(seen=vectorsmall(#vpar+1));
  for(v=1,#vpar, seen[vpar[v]+1]=1);
  seen[1] = 0;  \\ don't count vpar[v]==0 entries
  #vpar - hammingweight(seen);
}
{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        got=vpar_num_childless(vpar));
     check_equal(got, test_num_childless_by_hamming(vpar),
                 "vpar_num_childless() vs test_num_childless_by_hamming()");
  );
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got=vpar_num_childless(vpar));
     #vpar >= 1 || next;

     \\ Masanobu Ishikawa, Katsuhisa Yamanaka, Yota Otachi, Shin-Ichi
     \\ Nakano, "Enumerating All Rooted Trees Including k Leaves", IEICE
     \\ Trans Inf & Syst, volume E95-D, number 3, March 2012, pages 763-768.
     \\
     \\ Remark at end of section 3 concerning number of childless =
     \\ O(number of pairs of adjacent siblings).
     \\
     \\ Have total pairs = num childless - 1, for a non-empty tree, because
     \\ adding a new vertex under a childless is no change to num childless
     \\ and no change to num pairs.  Or adding a new vertex under a childful
     \\ is +1 childless and +1 sibling pair.
     \\
     my(siblings_sets=vpar_siblings_sets(vpar));
     check_equal(sum(i=1,#siblings_sets, #siblings_sets[i] - 1),
                 got - 1,
                 "vpar_num_childless() vs count adjacent sibling pairs");
  );
}

{
  \\ Ordered forests total num childless = binomial(2*n-1,n)
  \\ A001700
  check_equal(vector(#A001700_samples,n, binomial(2*n-1,n)),
              A001700_samples);  \\ up to offset
  for(n=0,10,
     my(vpar=vpar_preorder_first(n),
        total=0);
     until((vpar=vpar_preorder_next(vpar,1))=='none,  \\ forests
           total += vpar_num_childless(vpar));
     check_equal(total,
                 if(n==0,0, binomial(2*n-1,n)),
                 "vpar_num_childless() total over ordered forests"));
  my(n=0);
  check_equal(binomial(2*n-1,n), 1);  \\ whereas want 1 for total num childless
}
{
  \\ Ordered forests total num childless except last = binomial(2*n-1,n-2)
  for(n=0,10,
     my(vpar=vpar_preorder_first(n),
        total=0);
     until((vpar=vpar_preorder_next(vpar,1))=='none,  \\ forests
           my(num_childless=vpar_num_childless(vpar));
           if(num_childless,
              total += num_childless-1));
     check_equal(total,
                 if(n==0,0, binomial(2*n-1,n-2)),
     "vpar_num_childless() total except last childless over ordered forests"));
}
{
  \\ Ordered forests postorder num childless and not first child
  for(n=0,10,
     my(vpar=vpar_preorder_first(n),
        total=0);
     until((vpar=vpar_preorder_next(vpar,1))=='none,  \\ forests
           my(prev_sibling=vpar_next_sibling_vector(vpar,1),
              num_children=vpar_num_children_vector(vpar));
           for(v=1,#vpar,
              if(prev_sibling[v]!=0 && num_children[v]==0,
                 total++)));
     check_equal(total,
                 if(n==0,0, binomial(2*n-2,n-2)),
                 "postorder childless and have preceding sibling"));
}


\\-----------------------------------------------------------------------------
progress("vpar_childful_vertices()");

check_equal(vpar_childful_vertices([]),[]);
check_equal(vpar_childful_vertices([0]),[], "vpar_childful_vertices() singleton");
check_equal(vpar_childful_vertices([0,1]),[1], "vpar_childful_vertices() path-2");
check_equal(vpar_childful_vertices([2,0]),[2]);

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        got=vpar_childful_vertices(vpar));
     check_equal(got, test_select_vertices(vpar,vpar_is_childful),
                 "vpar_childful_vertices() vs vpar_is_childful()");
     check_equal(#got, vpar_num_childful(vpar),
                 "vpar_childful_vertices() count vs vpar_num_childful()");

     my(childless=vpar_childless_vertices(vpar));
     check_equal(setintersect(got,childless), [],
       "vpar_childful_vertices() nothing common to vpar_childless_vertices()");
     check_equal(got, vpar_set_complement(vpar,childless),
       "vpar_childful_vertices() complement of vpar_childless_vertices()");
  );
}


\\-----------------------------------------------------------------------------
progress("A084107");

\\ A084107 codes complete binary trees by steps, which for
\\ vpar_to_balanced_binary() means root-less binomial trees ... ?
{
  \\ A084107 balanced binary
  check_equal(vector(#A084107_samples,n,n--;
                     my(vpar=vpar_make_binomial_tree(2^n));
                     vpar=vpar_delete_vertex(vpar,1);  \\ forest
                     vpar=vpar_relabel_preorder(vpar,1);
                     vpar_to_balanced_binary(vpar)),
             A084107_samples);

  \\ A084108 index within balanced binaries
  \\
  \\ WRONG:
  \\ check_equal(vector(#A084108_samples,n,n--;
  \\                    my(vpar=vpar_make_binomial_tree(2^n));
  \\                    vpar=vpar_delete_vertex(vpar,1);  \\ forest
  \\                    vpar=vpar_relabel_preorder(vpar,1);
  \\                    vpar_is_preorder(vpar) || error();
  \\                    vpar_preorder_to_i(vpar)),
  \\            A084108_samples);
}


\\-----------------------------------------------------------------------------
progress("vpar_horizpos_vector()");

check_equal(vpar_horizpos_vector([]), [], "vpar_horizpos_vector() empty");
check_equal(vpar_horizpos_vector([0,0]), [0,1], "vpar_horizpos_vector() 2 singletons");
check_equal(vpar_horizpos_vector([0,0],1), [1,0], "vpar_horizpos_vector() 2 singletons, down=1");
check_equal(vpar_horizpos_vector([0,0,2,3,4]), [0,1,1,1,1]);
check_equal(vpar_horizpos_vector([0,3,4,5,0]), [0,1,1,1,1]);
check_equal(vpar_horizpos_vector([0,0,2]), [0,1,1]);
check_equal(vpar_horizpos_vector([0,1,2]), [0,0,0], \
            "vpar_horizpos_vector() path-3 down");

{
  \\      +---------+     1 is 0+1+0+2 = 3
  \\      |         |     2 is 1+1+0+2 = 4
  \\      6         |     3 is   2+1+0 = 3
  \\      |\        |     4 is 0+0+2+1 = 3
  \\      4 5       |
  \\        |\ \ \  |
  \\        1 2 3 7 |     7 is 3+1+0+2 = 6
  \\            |   |
  \\            +---+
  my(vpar=[5,5,5, 6,6, 3,5]);
  check_equal(vpar_num_cycles(vpar), 1);
  check_equal(vpar_num_cyclic_vertices(vpar), 3);
  check_equal(vpar_depths_vector(vpar),     [3,3,2, 3,2, 2, 3], "depths");
  check_equal(vpar_siblingpos_vector(vpar), [0,1,2, 0,1, 0, 3]);
  check_equal(vpar_horizpos_vector(vpar),   [3,4,3, 3,3, 3, 6]);
}

test_horizpos_vector_by_downwards(vpar) =
{
  my(seq=vpar_seq_downqueue(vpar),
     siblingpos=vpar_siblingpos_vector(vpar),
     ret=vector(#vpar));
  for(i=1,#seq,
     my(v=seq[i]);
     ret[v] = siblingpos[v] + if(vpar[v],ret[vpar[v]]));
  ret;
}
test_horizpos_vector_by_chains(vpar) =
{
  my(siblingpos=vpar_siblingpos_vector(vpar),
     ret=vector(#vpar));
  for(i=1,#vpar,
     my(v=i);
     while(v, ret[i] += siblingpos[v]; v=vpar[v]));
  ret;
}
check_equal(test_horizpos_vector_by_downwards([0,1]), [0,0]);
check_equal(test_horizpos_vector_by_downwards([2,0]), [0,0]);
check_equal(vpar_siblingpos_vector([0,0]), [0,1]);

{
  \\ horizpos of preorder not distinct
  my(vpar=[0,1,1]);
  vpar_validate(vpar);
  check_equal(vpar_is_preorder(vpar), 1);
  check_equal(vpar_horizpos_vector(vpar), [0,0,1]);

  my(vpar=[0,1,0]);
  vpar_validate(vpar);
  check_equal(vpar_is_preorder(vpar), 1);
  check_equal(vpar_horizpos_vector(vpar), [0,0,1]);
}


\\   7  8  9
\\         | \  \
\\         4  5  6
\\               | \  \
\\               1  2  3
check_equal(vpar_horizpos_vector([6,6,6, 9,9,9, 0,0,0]), \
            [4,5,6, 2,3,4, 0,1,2]);

\\ 1  3                 1  3
\\    |                 |
\\    2                 2
\\ vpar    =[0, 3, 0]
\\ horizpos=[0, 1, 1]   horizpos=[0, 0, 1]
\\ h down  =[1, 0, 0]   h down  =[1, 1, 0]

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got=vpar_horizpos_vector(vpar));
     check_equal(got,
                 test_horizpos_vector_by_downwards(vpar),
                 "vpar_horizpos_vector() vs test_horizpos_vector_by_downwards()");
     check_equal(got,
                 test_horizpos_vector_by_chains(vpar),
                 "vpar_horizpos_vector() vs test_horizpos_vector_by_chains()");

     check_equal(got,
                 Vecrev(vpar_horizpos_vector(vpar_relabel_reverse(vpar),1)),
                 "vpar_horizpos_vector() vs reverse down=1");

     my(got_down=vpar_horizpos_vector(vpar,1));
     check_equal(got_down,
                 Vecrev(vpar_horizpos_vector(vpar_relabel_reverse(vpar))),
                 "vpar_horizpos_vector(),down=1 vs reverse");
  );
}

\\ A003422 a(0)=0 no terms in sum
test_left_factorial(n) = sum(i=0,n-1, i!);
check_equal(vector(#A003422_samples,n,n--; test_left_factorial(n)), \
            A003422_samples);    \\ OFFSET=0

\\ num partitions of n which are 0 or 1 away from triangular
A074909(n) =
{
  my(t = test_untriangular_floor(n),
     r = n - test_triangular(t));
  binomial(t+1,r);
}
{
  check_equal(vector(#A074909_samples,n,n--; A074909(n)),
              A074909_samples,    \\ OFFSET=0
              "A074909 almost triangular partitions");
}

\\ num partitions of n which are exactly 1 diagonal away from triangular
A325191(n) = A074909(n) - test_is_triangular(n);
{
  check_equal(vector(#A325191_samples,n,n--; A325191(n)),
              A325191_samples,  \\ OFFSET=0
              "A318859");
}

\\ PENDING vpar_perm_count_indecomposable() maybe
test_perm_count_indecomposable(n) = \
  if(n, vpar_count_min_descendant_increasing(n-1,1), 1);

{
  for(forest=-1,1,
     for(n=0, test_limit_trees_search - forest,
        my(count_increasing=0,
           count_decreasing=0,
           count_increasing_down=0,
           count_decreasing_down=0);
        forvec(vec=vpar_forvec_from(n,forest),
              my(vpar=vpar_from_upascend(vec),
                 horizpos      = vpar_horizpos_vector(vpar),
                 horizpos_down = vpar_horizpos_vector(vpar,1),
                 is_increasing = test_vector_is_increasing(horizpos),
                 is_decreasing = test_vector_is_decreasing(horizpos),
                 is_increasing_down = test_vector_is_increasing(horizpos_down),
                 is_decreasing_down = test_vector_is_decreasing(horizpos_down));

              check_equal(is_decreasing, vecsum(horizpos)==0,
                         "vpar_horizpos_vector() decreasing only by all 0s");
              check_equal(is_increasing_down, vecsum(horizpos_down)==0,
                         "vpar_horizpos_vector(),down=1 increasing only by all 0s");

              count_increasing += is_increasing;
              count_decreasing += is_decreasing;
              count_increasing_down += is_increasing_down;
              count_decreasing_down += is_decreasing_down);

        \\----
        \\ Horizpos decreasing only when all 0s.
        \\ The only way some horizpos>0 can occur is when some v next
        \\ sibling of u, and that v > u is an increase in horizpos.
        \\ So merely a path-down all horizpos=0, with n! permutations of
        \\ that path, or (n-1)! when root=1 demanded.
        \\
        \\ Similarly for down=1 increasing.

        check_equal(count_decreasing, max(0, n-(forest==-1)) !,
                    "vpar_horizpos_vector() decreasing = n!");

        check_equal(count_increasing_down, max(0, n-(forest==-1)) !,
                    "vpar_horizpos_vector(),down=1 increasing = n!");

        \\----
        \\ Root=1 is horizpos[1]==0 so decreasing from there is all 0s
        \\ for n! paths as above.
        if(forest==-1,
           check_equal(count_decreasing_down, max(0,n-1)!,
                       "vpar_horizpos_vector(),down=1 decreasing = n!"));

        if(forest==1,
           check_equal(count_decreasing_down, test_perm_count_indecomposable(n+1),
                       "vpar_horizpos_vector(),down=1 decreasing = indecomposable"));

        \\----
        \\ labelled trees with horizpos increasing
        \\ labelled trees with horizpos,down=1 decreasing
        \\ See try.gp "vpar_horizpos_vector() increasing" for maybe A104982
        if(forest==0,
           check_equal(count_increasing, count_decreasing_down));

        if(forest==-1,
           check_equal(count_increasing,
                       test_perm_count_indecomposable(n),
                       "vpar_horizpos_vector() increasing = indecomposable"));
        if(forest==1,
           check_equal(count_increasing,
                       vpar_count_min_descendant_increasing(n,forest),
                       "vpar_horizpos_vector() increasing vs vpar_count_min_descendant_increasing()"));

        \\ vector(10,n,vpar_count_min_descendant_increasing(n,0))
        \\ vector(10,n,vpar_count_min_descendant_increasing(n,1))
        \\ vector(10,n,vpar_perm_count_indecomposable(n))


        \\ print1(count_increasing_down", ");
     ));
}


\\-----------------------------------------------------------------------------
progress("vpar_Wiener_index() of complete binary tree");

\\ R. Balakrishnan, K. Viswanathan Iyer, K. T. Raghavendra,
\\ "Wiener Index of Two Special Trees",
\\ MATCH Communications in Mathematical and in Computer Chemistry,
\\ volume 57, number 2, 2007, pages 385-392.
\\ http://match.pmf.kg.ac.rs/
\\ http://match.pmf.kg.ac.rs/volume57.htm
\\ http://match.pmf.kg.ac.rs/content57n2.htm
\\ http://match.pmf.kg.ac.rs/electronic_versions/Match57/n2/match57n2_385-392.pdf
\\ match57n2_385-392.pdf
\\
\\ of height n, so n+1 levels
test_complete_binary_tree_Wiener_index(n) = (n-2)*2^(2*n+2) + (n+4)*2^(n+1);
{
  check_equal(vector(#A158681_samples,n,
                     test_complete_binary_tree_Wiener_index(n)),
              A158681_samples,  \\ OFFSET=1
              "A158681 test_complete_binary_tree_Wiener_index()");

  \\ height n
  for(n=-1,10,
     my(vpar=vpar_make_complete_tree(2,n));
     check_equal(vpar_height(vpar), n);
     check_equal(test_complete_binary_tree_Wiener_index(n),
                 vpar_Wiener_index(vpar),
                 "vpar_Wiener_index() vs test_complete_binary_tree_Wiener_index()");

     check_equal((n-2)*4^(n+1) + (n+4)*2^(n+1),
                 test_complete_binary_tree_Wiener_index(n));
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_indsets_count() of complete binary tree");

A076725(n) =
{
  n>=0 || error("A076725() is for n>=0");
  my(x=1,y=1); for(i=2,n, [x,y] = [x^2 + y^4, x]); x;
}
{
  \\ A076725 indsets count in complete binary tree,
  \\ being a(n) = a(n-1)^2 + a(n-2)^4
  \\ initial 1, 1 in A076725
  \\ start from h=0 singleton tree vpar=[0] which is 2 indsets

  check_equal(vector(#A076725_samples,n,n--; A076725(n)),
              A076725_samples,  \\ OFFSET=0
              "A076725() vs A076725_samples");

  for(h=0,10,
     my(vpar=vpar_make_complete_tree(2,h),
        got=vpar_indsets_count(vpar));
     check_equal(vpar_height(vpar), h);
     check_equal(got, A076725(h+2));
     my(num_rows = h+1);
     check_equal(got, A076725(num_rows+1)));
}

\\ Return 1 if vector v has all entries == target.
test_vector_is_all_equal(v,target) = \
  for(i=1,#v, if(v[i]!=target, return(0))); 1;
check_equal(test_vector_is_all_equal([3,3], 3), 1);

\\ A242117 is count of indsets in complete binary tree but only those
\\ indsets made from among the degree-3 vertices, and not the empty set.
A242117(n) =
{
  n>=1 || error("A242117() is for n>=1");
  \\ A076725(n-1)^2 - 1;
  \\ my(x=1,y=1); for(i=3,n, [x,y] = [x^2 + y^4, x]); x^2-1;
  my(x=1,y=1); for(i=3,n, [x,y] = [(x + y^2)^2, x]); x-1;
}
{
  \\ h=2
  \\         1
  \\      /     \
  \\     2       3     degree=3
  \\    / \     / \
  \\   4   5   6   7
  \\              n = 1  2  3       levels
  \\              h = 0  1  2   3   height
  \\                  1  3  7  15   vertices, next 31 vertices is too many sets

  check_equal(vector(#A242117_samples,n, A242117(n)),
              A242117_samples,  \\ OFFSET=1
              "A242117() vs A242117_samples");

  \\ looking at the sets and the degree of their vertices
  for(n=1,4,
     my(vpar=vpar_make_complete_tree(2,n-1));  \\ n rows
     check_equal(#vpar_row_vertices_vector(vpar), n, "tree of n rows");

     my(sets=vpar_indsets(vpar),
        degrees=vpar_degrees_vector(vpar));
     \\ print("vpar    "vpar);
     \\ print("degrees "degrees);
     sets=select(set->set!=[], sets);  \\ not the empty set
     my(set_is_all_degree3 = (set) ->
        for(i=1,#set,
           if(degrees[set[i]]!=3, return(0)));
        1);
     sets=select(set_is_all_degree3,sets);
     check_equal(A242117(n), #sets,
                 "A242117() vs sets selected by degrees"));

  \\ The degree-3 vertices of complete binary tree are two complete binary
  \\ trees, so A242117 is A076725_samples squared, less the empty tree not
  \\ counted in A242117.
  for(n=1,8,
     my(h=n-1,
        vpar=vpar_make_complete_tree(2,h));  \\ n rows
     check_equal(#vpar_row_vertices_vector(vpar), n, "tree of n rows");

     my(degrees=vpar_degrees_vector(vpar),
        keep=select(v->degrees[v]==3,[1..#vpar]));
     vpar=vpar_keep_vertices(vpar,keep);

     check_equal(A242117(n),
                 vpar_indsets_count(vpar) - 1,  \\ A242117 excludes empty set
                 "A242117() vs degree=3 induced subgraph");

     check_equal(A242117(n),
                 A076725(n-1)^2 - 1,
                 "A242117 in terms of A076725");

     my(vpar_nsub2=if(h-2<-1,[],vpar_make_complete_tree(2,h-2)));  \\ n-2 rows
     check_equal(#vpar_row_vertices_vector(vpar), max(0,n-2), "tree n-2 rows");

     check_equal(A076725(n-2+1),
                 vpar_indsets_count(vpar_nsub2),
                 "A242117 indsets in each forest half");

     my(vpar_forest=vpar_concat_forest(vpar_nsub2,vpar_nsub2));
     check_equal(vpar_is_isomorphic(vpar,vpar_forest), 1,
                 "degree=3 induced subgraph is forest of two n-2");
     );
}

complete_binary_indsets_count(h) = A076725(h+2);

\\ not in OEIS: 1, 5, 98, 11161, 81317325
complete_binary_indsets_size_total(h) =
{
  if(h<0,0, h==0,1,
     \\ print(h" "A076725(1+ h-1)" "A076725(1+ h-2));
       2 * complete_binary_indsets_count(h-1) * complete_binary_indsets_size_total(h-1)
     + 4 * complete_binary_indsets_count(h-2)^3 * complete_binary_indsets_size_total(h-2)
     + complete_binary_indsets_count(h-2)^4);
}
{
  check_equal(vector(5,h,h-=2;
                     complete_binary_indsets_size_total(h)),
              vector(5,h,h-=2;
                     vecsum(apply(length,vpar_indsets(vpar_make_complete_tree(2,h))))),
              "complete_binary_indsets_size_total()");
}

\\ unique indset of alternate rows from the biggest
\\ vector(5,h,h-=2; vpar_indnum_and_count(vpar_make_complete_tree(2,h)))


\\-----------------------------------------------------------------------------
progress("vpar_free_orbit_sizes()");

{
  \\ A337114 how many distinct sizes vectors (unordered, so partition of n)
  my(samples=A337114_samples);  \\ OFFSET=1
  samples=samples[1..min(#samples,14)];

  my(a = (n)->
     my(vpar=vpar_free_first(n),
        seen=List([]));
     until((vpar=vpar_free_next(vpar))=='none,
           my(vec=vecsort(vpar_free_orbit_sizes(vpar)));
           my(pos=setsearch(seen,vec,1));
           if(pos, listinsert(seen,vec,pos)));
     #seen);
  check_equal(vector(#samples,n, a(n)),
              samples,
              "A337114 how many vpar_free_orbit_sizes() as partitions of n");
}

{
  \\ count orbit size = 1, fixed points

  my(samples=A005201_samples[1..min(12,#A005201_samples)]);
  check_equal(vector(#samples,n,
                     my(total=0,
                        vpar=vpar_free_first(n));
                     until((vpar=vpar_free_next(vpar))=='none,
                           my(orbit_sizes=vpar_free_orbit_sizes(vpar));
                           total += sum(i=1,#orbit_sizes, orbit_sizes[i]==1));
                     total),
              samples,
     "A005201 total number of automorphism fixed vertices in free trees");


  \\ gf form shown in A005201
  my(limit=vecmin([8,#A005201_samples,#A005200_samples]),
     L=O(x^(limit+1)),
     gA       = Polrev(A005201_samples)*x + L,
     gA005200 = Polrev(A005200_samples)*x + L, \\ total fixed in rooted trees
     gT = Polrev(vpar_count_unlabelled_vector(limit))*x + L);
  check_equal(gA,
              gT * (1 - subst(gA005200,x,x^2)) - subst(gA005200,x,x^2),
              "A005201 gf satisfying");
}



\\-----------------------------------------------------------------------------
progress("vpar_orbit_sizes()");

{
  \\ count orbit size = 1, fixed points

  my(samples=A005200_samples[1..min(10,#A005200_samples)]);
  check_equal(vector(#samples,n,
                     my(total=0,
                        vpar=vpar_unlabelled_first(n));
                     until((vpar=vpar_unlabelled_next(vpar))=='none,
                           my(orbit_sizes=vpar_orbit_sizes(vpar));
                           total += sum(i=1,#orbit_sizes, orbit_sizes[i]==1));
                     total),
              samples,
    "A005200 total number of automorphism fixed vertices in unlabelled trees");

  \\ gf form shown in A005200
  my(limit=min(20,#A005200_samples),
     L=O(x^(limit+1)),
     gA = 0 + Polrev(A005200_samples)*x + L,
     gT = Polrev(vpar_count_unlabelled_vector(limit))*x + L);
  check_equal(gA,
              gT * (1 + gA - subst(gA,x,x^2)),
              "A005200 gf vs vpar_count_unlabelled()");
}


\\-----------------------------------------------------------------------------
progress("vpar_matchpoly() and vpar_matchings_count() of path n");

{
  \\ Match poly of path of n edges, per for instance
  \\     Tomislav Doslic, Ivana Zubac, "Counting Maximal Matchings In Linear
  \\     Polymers", Ars Mathematica Contemporanea, volume 11, 2016, pages
  \\     255-276, http://amc-journal.eu
  for(n=0,12,
     my(vpar=vpar_make_path(n+1));
     check_equal(vpar_matchpoly(vpar),
                 sum(k=0,ceil(n/2), binomial(n+1-k,k)*'x^k)));
}
{
  \\ Match poly coeffs of path of n edges, high to low.
  my(v=[], n=0, len=max(#A098925_samples,#A102426_samples));
  while(#v<len,
        my(vpar=vpar_make_path(n),
           p=vpar_matchpoly(vpar));
        v=concat(v, Vec(p));
        n++);

  check_equal(v[1..#A098925_samples],
              A098925_samples,
              "A098925 matchpoly coeffs of path");

  \\ A102426 with initial 0 for the n=0 polynomial=0
  check_equal(concat(0, v[1..#A102426_samples-1]),
              A102426_samples,
              "A102426 matchpoly coeffs of path");
}
{
  \\ Match poly coeffs of path of n edges, low to high.
  my(v=[], n=0);
  while(#v < #A011973_samples,
        my(vpar=vpar_make_path(n),
           p=vpar_matchpoly(vpar));
        v=concat(v, Vecrev(p));
        n++);

  check_equal(v[1..#A011973_samples],
              A011973_samples,
              "A011973 matchpoly of path coeffs low to high");
}

{
  \\ A226594 number of distinct characteristic polynomials of trees
  \\   match polys equivalent to adjacency polys
  my(limit=13);
  check_equal(vector(limit,n,n--;
                     test_count_free_distinct_func(n, vpar_matchpoly)),
              A226594_samples[1..limit],  \\ OFFSET=0
              "A226594 num distinct match polys");
}

\\ Abbe Mowshowitz, "The Characteristic Polynomial of a Graph", Journal of
\\ Combinatorial Theory, series B, volume 12, 1972, pages 177-193.
\\ https://www.sciencedirect.com/science/article/pii/0095895672900238
\\
test_matchings_count_of_path(n,k) =
{
  n>=0 || error();
  k>=0 || error();
  if(k==0, 1,
     2*k>n, 0,
     n==0, 1,
     n==1, binomial(n-1,1),
     test_matchings_count_of_path(n-1,k) + test_matchings_count_of_path(n-2,k-1));
}
check_equal(vector(10,n, my(k=1); test_matchings_count_of_path(n,k)), \
            vector(10,n, vpar_num_edges(vpar_make_path(n))));
check_equal(vector(10,n, my(k=n-1); test_matchings_count_of_path(n,k)), \
            vector(10,n, my(k=n-1); binomial(n-k,k)));


\\-----------------------------------------------------------------------------
progress("vpar_count_unlabelled()");
\\ vpar_count_unlabelled_vector()

{
  for(forest=0,1,
     for(len=0,5,
        check_equal(vpar_count_unlabelled_vector(len,forest),
                    vector(len,n,  vpar_count_unlabelled(n,forest)),
     "vpar_count_unlabelled_vector() vs individual vpar_count_unlabelled()")));
}
{
  check_equal(vpar_INTERNAL_Euler_transform
                (vector(12,n, vpar_count_unlabelled(n))),
              vector(12,n, vpar_count_unlabelled(n,1)),
              "vpar_count_unlabelled() forests vs Euler transform of trees");
}
{
  for(forest=0,1,
     for(n=-5,-1,
        check_equal(iferr(vpar_count_unlabelled(n,forest);0, e, 1),
                    1,
                    "vpar_count_unlabelled() error on negative n")));
}

{
  \\ unlabelled trees, so vpar up to isomorphism
  \\ A000081 (though it reckons 0 trees of 0 vertices)
  \\    n= 1  2  3  4  5
  my(want=[1, 1, 2, 4, 9, 20, 48, 115, 286, 719, 1842, 4766, 12486, 32973,
           87811, 235381, 634847, 1721159, 4688676, 12826228, 35221832]);
  check_equal(vpar_count_unlabelled_vector(#want), want,
              "vpar_count_unlabelled_vector() samples trees");

  check_equal(A000081_samples[1], 0);  \\ n=0
  want=A000081_samples[^1];
  check_equal(vpar_count_unlabelled_vector(#want), want,
              "vpar_count_unlabelled_vector() = A000081");
}
{
  \\ unlabelled forests, is 1 later
  \\ A000081
  \\    n= 1  2  3  4  5
  my(want=[1, 2, 4, 9, 20, 48, 115, 286, 719, 1842, 4766, 12486, 32973,
           87811, 235381, 634847, 1721159, 4688676, 12826228, 35221832]);
  check_equal(vpar_count_unlabelled_vector(#want,1), want,
              "vpar_count_unlabelled_vector() samples forests");

  want=A000081_samples[3..#A000081_samples];
  check_equal(vpar_count_unlabelled_vector(#want,1), want,
              "vpar_count_unlabelled_vector() forests = A000081");
}

{
  \\ A087803 rooted unlabelled trees <= n
  \\ cumulative, not counting empty tree n=0
  check_equal(vector(#A087803_samples,n, sum(i=1,n,vpar_count_unlabelled(i))),
              A087803_samples);  \\ OFFSET=1
  check_equal(vector(#A087803_samples,n,
                     vecsum(vpar_count_unlabelled_vector(n))),
              A087803_samples);  \\ OFFSET=1
}

\\ 11 12 13 22 23 33      3 choose 2 with repetition
check_equal(binomial(3+1,2), 6);

\\ 12 13 14 23 24 34
check_equal(binomial(4,2), 6);
\\ 12 13 14 23 24 34  11 22 33 44
check_equal(binomial(4+1,2), 10);
\\ 123 124 134 234
check_equal(binomial(4,3), 4);
\\ 111 112 113 114  122 123 124  133 134  144  = 10
\\ 222 223 224   233 234   244  =  6
\\ 333 334 344    444           = 4
check_equal(binomial(4+2,3), 20);

check_equal(binomial(3+1,2), 6);

\\ vpar_count_unlabelled(n,1) forests, calculated by summing over partitions
\\ and tree choices for each element t of the partition.  Uses
\\ test_partition_to_repetitions() to get t and rep since there is no
\\ order among trees of the same size, so binomial to choose rep many from
\\ among vpar_count_unlabelled(t) many trees t, with duplication allowed.
test_count_unlabelled_forests_by_trees_partition(n) =
{
  if(n<=1,return(1));
  my(a=vector(n,i,vpar_count_unlabelled(i)));
  my(total=0);
  forpart(p=n,
          my(r=test_partition_to_repetitions(p));
          my(f=prod(i=1,#r,
          \\ print("binomial "r[i]"  ", (a[r[i][1]] + r[i][2]-1)," "r[i][2]);
                    binomial(a[r[i][1]] + r[i][2]-1, r[i][2])));
          total += f;
          \\ print(Vec(p)" = "f"   a="apply(i->a[i],Vec(p))"   "r);
         );
  total;
}
{
  check_equal(vector(12,n,n--; vpar_count_unlabelled(n,1)),
              vector(12,n,n--; test_count_unlabelled_forests_by_trees_partition(n)),
              "vpar_count_unlabelled() vs by trees partition");
}

\\ vpar_count_unlabelled(n,1) forests, calculated by summing over
\\ vpar_count_unlabelled() trees.
test_count_unlabelled_forests_by_trees_sum(n) =
{
  if(n<=1,return(1));
  sum(k=1,n,
        1/n * sumdiv(n-k+1,d, d*vpar_count_unlabelled(d))
            * vpar_count_unlabelled(k));
}
{
  check_equal(vector(12,n,n--; vpar_count_unlabelled(n,1)),
              vector(12,n,n--; test_count_unlabelled_forests_by_trees_sum(n)),
              "vpar_count_unlabelled() vs by trees sum");
}


\\-----------------------------------------------------------------------------
progress("A073059 opposite of A098725");

test_A073059(n) =
{
  n>=1 || error("test_A073059() is for n>=1");  \\ OFFSET=1
  valuation(3*n+1,2)%2;
}

\\ Ralf Stephan recurrences
check_equal(vector(1024,n, test_A073059(2*n)), \
            vector(1024,n, 0), \
            "A073059 even");
check_equal(vector(1024,n,n--; test_A073059(4*n+3)), \
            vector(1024,n,n--; 1), \
            "A073059  3 mod 4");
check_equal(vector(1024,n, test_A073059(4*n+1)), \
            vector(1024,n, test_A073059(n)), \
            "A073059  1 mod 4");

{
  check_equal(vector(2048,n, test_A073059(n)),
              vector(2048,n, 1 - test_A098725(n+1)),
              "A073059 by its opposite A098725");

  check_equal(vector(4096,n, test_A073059(n)),
              vector(4096,n, my(i=0,b); while((b=bittest(n,i))&&!(b=bittest(n,i++)),i++); b),
              "A073059 by bit b");
  check_equal(vector(4096,n, test_A073059(n)),
              vector(4096,n, my(i=0); while(bittest(n,i)&&!bittest(n,i++),i++); i%2),
              "A073059 by bit i");
  check_equal(vector(4096,n, test_A073059(n)),
              vector(4096,n, my(i=0); while(bittest(n,i)!=i%2,i++); i%2),
              "A073059 by bit i");
  my(f=(n)->for(i=0,logint(n+1,2)+10, if(bittest(n,i)==i%2,return(i%2))));
  check_equal(vector(4096,n, test_A073059(n)),
              vector(4096,n, f(n)),
              "A073059 by for loop");
}

\\ formula in A073059, looks wrong
\\ check_equal(vector(30,n,n++; test_A073059(n)), \
\\             vector(30,n,n++; 1 - test_A035263(n-1)));  \\ for n>1

{
  \\ trisections
  check_equal(vector(2048,n, test_A073059(n)),
              vector(2048,n, 1 - test_A035263(3*n+1)),
              "A073059 by A035263(3*n+1)");

  check_equal(vector(2048,n, test_A073059(n)),
              vector(2048,n, test_A096268(3*n)),
              "A073059 by A096268(3*n)");
}

my(k=3); check_equal(digits(3*4^k + (4^k-1)/3, 4), [3, 1,1,1]);

progress("A073059 gf");
{
  \\ ... 11010101 with k many low 01 bit pairs
  \\ terms up to k inclusive are good to < 01 010101 with k+1 low 1s
  \\ where next would begin
  for(k=0,5,
     my(limit = 3*4^k + (4^(k+1)-1)/3,
        L=O(x^limit),
        g = sum(k=0,k,
                x^((10*4^k-1)/3) / (1 - x^(4^(k+1)))
                + L),
        f=sum(n=1,limit-1, test_A073059(n)*x^n) + L);
     check_equal(g, f,
                 Str("gA098725 by bits  k="k" limit="limit));
     \\ print(g);
     \\ print(f);
     \\ print(g==f);
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_siblings_sets()");

check_equal(vpar_siblings_sets([0,1,1,3]), [ [1], [2,3], [4] ], \
            "vpar_siblings_sets()");

test_siblings_sets_by_children(vpar) =
{
  vecsort(select(length,
                 concat([vpar_roots(vpar)], vpar_children_vector(vpar))));
}
test_siblings_sets_by_next(vpar) =
{
  my(next_sibling=vpar_next_sibling_vector(vpar),
     seen=vectorsmall(#vpar),
     ret=List([]));
  for(i=1,#vpar,
     if(!seen[i],
        my(v=i, l=List([]));
        while(v, listput(l,v); seen[v]=1; v=next_sibling[v]);
        listput(ret,Vec(l))));
  Vec(ret);
}
test_siblings_sets_by_seq(vpar) =
{
  \\ vpar_seq_siblings() is sibling vertex numbers consecutive.
  \\ Cut seq up into separate vectors of the siblings.
  if(#vpar==0, return([]));
  my(seq=vpar_seq_siblings(vpar),
     beg=1,
     ret=List([]));
  for(i=1,#vpar,
     if(i==#vpar || vpar[seq[i+1]] != vpar[seq[i]],
        listput(ret, Vec(seq[beg..i]));
        beg=i+1));
  Vec(ret);
}
{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        got=vpar_siblings_sets(vpar));
     check_equal(vecsum(apply(length,got)), #vpar);
     check_equal(sum(i=1,#got, #got[i]), #vpar);
     for(i=1,#got, check_equal(setisset(got[i]), 1));

     check_equal(got,
                 vecsort(vector(#vpar,v,vpar_siblings_of_vertex(vpar,v)),,8),
                 "vpar_siblings_sets() vs vpar_siblings_of_vertex()");

     check_equal(#got, vpar_num_siblings_sets(vpar));
     check_equal(got, test_siblings_sets_by_children(vpar));
     check_equal(got, test_siblings_sets_by_next(vpar),
                 "vpar_siblings_sets() vs test_siblings_sets_by_next()");

     check_equal(got, test_siblings_sets_by_seq(vpar),
                 "vpar_siblings_sets() vs test_siblings_sets_by_seq()");
  );
}

{
  \\ num distinct vpar_siblings_sets() == vpar_next_sibling_vectors_count()
  for(forest=-1,1,
     for(n=0, test_limit_trees_search - forest,
        my(seen=List([]));  \\ Set() of siblings_sets vectors seen
        forvec(vec=vpar_forvec_from(n,forest),
              my(vpar=vpar_from_upascend(vec),
                 siblings_sets=vpar_siblings_sets(vpar));
              my(pos=setsearch(seen,siblings_sets,1));
              if(pos, listinsert(seen,siblings_sets,pos)));
        check_equal(#seen, vpar_next_sibling_vectors_count(n,forest),
                    "vpar_siblings_sets() num distinct is vpar_next_sibling_vectors_count()");
     ));
}

\\ vec is a vector of things.
\\ Return a vector of those elements of vec where flags[i]==want.
\\ Default want=1 selects those where flags[i]==1.
test_vector_select_flags(vec,flags,want=1) = \
  vecextract(vec, select(f->f==want, flags, 1));
check_equal(test_vector_select_flags([11,12,13,14], [0,1,1,0]),   [12,13]);
check_equal(test_vector_select_flags([11,12,13,14], [0,1,1,0],0), [11,14]);

\\ ret = test_all_set_partitions(vec)
\\ vec is a set of things.
\\ Return a vector of all partitions of them into sets.
\\ The return is a vector of vectors of vectors,
\\     [ [ [a,b], [c], [d,e,f] ],
\\       ...
\\     ]
\\ Each element of the return is a partition.
\\ A partition is a vector of sets of elements from vec.
test_all_set_partitions(vec) =
{
  my(ret=[]);
  if(#vec,
     forvec(flags=vector(#vec,i,if(i==1,[1,1],[0,1])),
           my(s=vecsum(flags));
           if(s==#vec, ret=concat(ret, [ [vec] ]); next);
           my(first=test_vector_select_flags(vec,flags),  \\ some elements
              nonvec=test_vector_select_flags(vec,flags,0),
              rest=test_all_set_partitions(nonvec));   \\ recurse choose more
           for(i=1,#rest,
              rest[i]=concat([first],rest[i]));  \\ combinations first+rest
           ret=concat(ret,rest)));
  ret;
}
{
  \\ check_equal(test_all_set_partitions([]), [ [] ]);
  check_equal(test_all_set_partitions([11]), [ [[11]] ],
              "test_all_set_partitions()");
  check_equal(test_all_set_partitions([11,12]),
              [ [[11],[12]], [[11,12]] ],
              "test_all_set_partitions()");
}
{
  my(vec=test_all_set_partitions([11,12,13]));
  check_equal(#vec,5);
  check_equal(vec,
              [[ [11], [12], [13] ],
               [ [11], [12,13] ],
               [ [11,13], [12] ],
               [ [11,12], [13] ],
               [ [11,12,13] ]
              ]);
}
{
  \\ A000110 Bell numbers, number of partitions of 1..n
  \\ 1,2,5,15,52,203,877,
  for(n=1,7,
     check_equal(#test_all_set_partitions([11..10+n]),
                 vpar_INTERNAL_Bell_number(n),
              "test_all_set_partitions() length vpar_INTERNAL_Bell_number()"));
}

\\ vec1 and vec2 are sets of numbers.
\\ Return 1 if they are non-crossing.
\\ Non-crossing means if their ranges overlap then the entire range low to
\\ high of one set falls within a gap in the elements of the other set.
\\ So if vec2 ranges lo to hi then there are no elements of vec1 between lo
\\ and hi.  vec1 and vec2 can be either way around.
test_sets_is_noncrossing(vec1,vec2) =
{
  if(vec1[1] > vec2[1],     \\ make vec1 start smallest
     my(t=vec1); vec1=vec2; vec2=t);
  my(m=vecmin(vec2));
  vec1=select(elem->elem>=m, vec1);
  m=vecmax(vec2);
  vec1=select(elem->elem<=m, vec1);
  #vec1==0;
}
check_equal(test_sets_is_noncrossing([11],[12]), 1);
check_equal(test_sets_is_noncrossing([11,12],[13]), 1);
check_equal(test_sets_is_noncrossing([11,13],[12]), 1);
check_equal(test_sets_is_noncrossing([12],[11,13]), 1);
check_equal(test_sets_is_noncrossing([11,13,15],[12,14]), 0);
\\
\\ vec is a partition of some integers, meaning a vector of sets
\\ like [ [1,2], [3], [4,5,6] ].
\\ Return 1 if this is a non-crossing partition, meaning all sets pairwise
\\ are non-crossing per test_sets_is_noncrossing().
test_set_partition_is_noncrossing(vec) =
{
  for(i=1,#vec,
     for(j=1,#vec,
        if(i!=j,
           test_sets_is_noncrossing(vec[i],vec[j]) || return(0))));
  1;
}
{
  my(vec=test_all_set_partitions([11,12,13]));
  vec=select(test_set_partition_is_noncrossing,vec);
  check_equal(#vec,5);
  check_equal(vec,
              [[ [11], [12], [13] ],
               [ [11], [12,13] ],
               [ [11,13], [12] ],
               [ [11,12], [13] ],
               [ [11,12,13] ]
              ]);
}
{
  for(n=1,7,
     my(vec=test_all_set_partitions([1..n]));
     vec=select(test_set_partition_is_noncrossing,vec);
     check_equal(#vec, vpar_INTERNAL_Catalan_number(n),
                "test_set_partition_is_noncrossing() length vpar_INTERNAL_Catalan_number()");

     my(l=List([]));    \\ list of sibings sets
     my(vpar=vpar_preorder_first(n));
     until((vpar=vpar_preorder_next(vpar,1))=='none,
           listput(l, vpar_siblings_sets(vpar)));
     l=Vec(l);
     check_equal(vecsort(vec), vecsort(l),
                "test_set_partition_is_noncrossing() vs vpar_siblings_sets()"));
}

\\ vec is a partition of some integers, meaning a vector of sets
\\ like [ [1,2], [3], [4,5,6] ].

\\ Return 1 if this is a non-overlapping partition, meaning all sets
\\ pairwise have their range low to high separate or nested, not partly
\\ overlapping.
test_set_partition_is_nonoverlapping(vec) =
{
  my(lo=apply(vecmin,vec),
     hi=apply(vecmax,vec));
  for(i=1,#vec,
     for(j=1,#vec,
        if(! ((lo[i] <= lo[j] && hi[j] <= hi[i])
           || (lo[j] <= lo[i] && hi[i] <= hi[j])
           || hi[i] <= lo[j]
           || hi[j] <= lo[i]),
           return(0))));
  1;
}
{
  \\ A006789 Bessel, nonoverlapping partition 1..n
  \\ 1,2,5,14,43,143,509,
  for(n=1,7,
     my(vec=test_all_set_partitions([11..10+n]));
     vec=select(test_set_partition_is_nonoverlapping,vec);
     check_equal(#vec, A006789_samples[n+1],
                "A006789 test_set_partition_is_nonoverlapping()"));
}


\\-----------------------------------------------------------------------------
progress("vpar_num_siblings_sets()");

check_equal(vpar_num_siblings_sets([]), 0, "vpar_num_siblings_sets() empty");
check_equal(vpar_num_siblings_sets([0]),1,"vpar_num_siblings_sets() singleton");
check_equal(vpar_num_siblings_sets([1]),1,"vpar_num_siblings_sets() self-loop");

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        num_siblings=vpar_num_siblings_sets(vpar));
     check_equal(num_siblings, #vecsort(vpar,,8),
                 "vpar_num_siblings_sets() vs #vecsort distinct");
     check_equal(num_siblings, #Set(vpar),
                 "vpar_num_siblings_sets() vs #Set()");

     my(num_childless=vpar_num_childless(vpar));
     check_equal(num_siblings + num_childless,  #vpar + (vpar_num_roots(vpar)!=0),
                 "vpar_num_siblings_sets() vs vpar_num_childless() and vpar_num_roots()");
     check_equal(num_siblings + num_childless,  #vpar + !vpar_is_childless(vpar,0),
                 "vpar_num_siblings_sets() vs vpar_num_childless() and vpar_is_childless of 0");

     my(num_childful=vpar_num_childful(vpar));
     check_equal(num_siblings,  num_childful + (vpar_num_roots(vpar)!=0),
                 "vpar_num_siblings_sets() vs vpar_num_childful() and vpar_num_roots()");
     check_equal(num_siblings,  num_childful + !vpar_is_childless(vpar,0),
                 "vpar_num_siblings_sets() vs vpar_num_childful() and vpar_is_childless of 0");
  );
}

\\ n items make ordered choice of k
test_binomialtimes(n,k) = binomial(n,k)*k!;
{
  for(n=1,10,
     for(k=1,n,
        check_equal(test_binomialtimes(n,k), prod(i=n-k+1,n, i));
        check_equal(test_binomialtimes(n,k), n!/(n-k)!)));
}

\\ setsizes is a vector of set sizes for a partition of the integers 1..n,
\\ where n = vecsum(sizesize).
\\ Return the number of non-crossing partitions of 1..n with these set sizes.
test_setsizes_num_noncrossing(setsizes) =
{
  \\ Kreweras theorem 4.
  \\ Notation m_h = binomial(m,h)*h!, ordered choice.
  my(n=sum(i=1,#setsizes, setsizes[i]),
     hist=vpar_INTERNAL_histogram(setsizes));
  test_binomialtimes(n,#setsizes-1) / prod(i=1,#hist, hist[i]!);
}

\\ WRONG
\\ {
\\   for(n=1,10,
\\      print();
\\      for(k=1,n,
\\         my(total=0);
\\         forpart(setsizes=n,
\\                 my(hist=vpar_INTERNAL_histogram(setsizes));
\\                 total += k! / prod(i=1,#hist, hist[i]!),
\\                 k); \\ k many parts
\\         print(total," "binomial(n-1,k-1))));
\\ }

\\ range lo to hi anywhere within size, and not the entire
\\ lo=1 to n, hi = lo to n
\\ if lo=1 then hi!=n
\\ Per Knuth fasc 4a 7.2.1.6 exercise 26(e) answer.
\\
test_setsize_num_noncrossing_splits(n) = test_triangular(n-1);

\\ vector(6,n,test_setsize_num_noncrossing_splits(n))

check_equal(test_setsize_num_noncrossing_splits(1), 0, \
           "test_setsize_num_noncrossing_splits() 1");
\\ 2:  11
check_equal(test_setsize_num_noncrossing_splits(2), 1, \
           "test_setsize_num_noncrossing_splits() 2");
\\ 3:  1,2 1_1_1 2,1
check_equal(test_setsize_num_noncrossing_splits(3), 3);
\\ 4:  1,3 1_1_2 1_2_1
\\     2,2 2_1_1
\\     3,1
check_equal(test_setsize_num_noncrossing_splits(4), 6);
\\ 5:  1,4 1_1_3 1_2_2 1_3_1
check_equal(test_setsize_num_noncrossing_splits(5), 10);

{
  \\ ordered forests count by k many siblings sets = Narayana(n,k)
  \\ ordered trees count by k many siblings sets = Narayana(n-1,k-1)
  \\
  \\ ordered forests count by k many childless = Narayana(n,k)
  \\ ordered trees count by k many childless = Narayana(n-1,k)

  for(forest=0,1,
     my(tree=!forest);
     for(n=1,10-forest,
        my(counts_by_num_siblings=vector(n),
           counts_by_num_childless=vector(n),
           setsizes_list=List([]),
           counts_by_setsizes=List([]),
           vpar=vpar_preorder_first(n));
        until((vpar=vpar_preorder_next(vpar,forest))=='none,
              my(siblings_sets=vpar_siblings_sets(vpar));
              counts_by_num_siblings[#siblings_sets]++;

              my(setsizes=vecsort(apply(length,siblings_sets)));
              my(pos=setsearch(setsizes_list,setsizes));
              if(!pos,
                 pos=setsearch(setsizes_list,setsizes,1);
                 listinsert(setsizes_list,setsizes,pos);
                 listinsert(counts_by_setsizes,0,  pos));
              counts_by_setsizes[pos]++;

              counts_by_num_childless[vpar_num_childless(vpar)]++;
             );

        check_equal(counts_by_num_siblings,
                    vector(n,k, test_Narayana(n-tree, k-tree)),
        "vpar_num_siblings_sets() ordered trees/forests = Narayana");

        if(forest,
           \\ G. Kreweras, "Sur les Partitions Non-Croisees d'Un Cycle",
           \\ Discrete Mathematics, volume 1, number 4, 1971, pages 333-350.
           \\ Corollary 4.1 num non-crossing partitions of 1..n with k sets.
           check_equal(counts_by_num_siblings,
                       vector(n,k, (n-1)!*n! /(k-1)! /k! / (n-k)!/(n-k+1)!),
           "vpar_num_siblings_sets() ordered trees/forests, Kreweras num sets");

           setsizes_list     =Vec(setsizes_list);
           counts_by_setsizes=Vec(counts_by_setsizes);
           check_equal(counts_by_setsizes,
                       apply(test_setsizes_num_noncrossing,setsizes_list),
           "vpar_num_siblings_sets() ordered trees/forests, Kreweras setsizes");

           \\ s[i] = num sets of size i
           \\ s1+s2+...+sk = h  total num sets
           \\ s1+2s2+3s3 + ... + ksk = m
          );

        check_equal(counts_by_num_childless,
                    vector(n,k, test_Narayana(n-tree, k)),
        "vpar_num_childless() in ordered trees/forests = Narayana")));
}
{
  \\ Ordered forests total siblings sets except the set of roots.
  \\ Binary tree total left child.
  \\ By symmetry total right child same.
  \\ n-1 lefts and rights each tree.
  \\ (n-1)/2*Catalan(n) total lefts.
  \\ = Tamari lattice edges, since rotate at each right child.

  for(n=1,9,
     my(total=0,
        total_sans_roots=0,
        vpar=vpar_preorder_first(n));
     until((vpar=vpar_preorder_next(vpar,1))=='none,  \\ forests
           my(num_siblings_sets=vpar_num_siblings_sets(vpar));
           total += num_siblings_sets;
           total_sans_roots += num_siblings_sets - 1);
     check_equal(total_sans_roots, (n-1)/2 * vpar_count_ordered(n,1),
                 "vpar_num_siblings_sets() excl roots total, ordered forests");
     check_equal(total, (n+1)/2 * vpar_count_ordered(n,1),
                 "vpar_num_siblings_sets() total, ordered forests");
     check_equal(total, 1/2*binomial(2*n,n),
                 "vpar_num_siblings_sets() total, ordered forests, binomial"));

  \\ vector(10,n, (n-1)/2 * vpar_count_ordered(n,1))
  \\ vector(10,n, (n+1)/2 * vpar_count_ordered(n,1))
}


\\-----------------------------------------------------------------------------
progress("vpar_seq_postorder()");

check_equal(vpar_seq_postorder([]), Vecsmall([]));
check_equal(vpar_seq_postorder([0,0]), Vecsmall([1,2]));
check_equal(vpar_seq_postorder([0,0],1), Vecsmall([2,1]));

\\  1    4
\\  |\   |
\\  2 3  5
check_equal(vpar_seq_postorder([0,1,1,0,4]),   Vecsmall([2,3,1,5,4]));
check_equal(vpar_seq_postorder([0,1,1,0,4],1), Vecsmall([5,4,3,2,1]));

{
  for(n=0,5,
     my(vpar=vpar_make_path(n));
     vpar_seq_postorder(vpar,0);
     vpar_seq_postorder(vpar,1);
     check_equal(iferr(vpar_seq_postorder(vpar,'down);0, e, 1),
                 1,
                 "vpar_seq_postorder() error on down=symbol (whereas it should be 0,1)"));

  \\ vpar_seq_postorder([0,1,2,3],'down)
}

test_seq_postorder_by_sort(vpar,rev=0) =
{
  \\ keys are paths negated [-root, ..., -v]
  \\ Or for rev=1, keys are positives [root, ..., v].
  \\ Sorting by that is pre-order with subtrees descending.
  \\ Sort reverse is post-order with subtrees ascending.
  \\
  my(keys=vector(#vpar,v, Vecrev(vpar_ancestors(vpar,v))));
  if(!rev, keys=-keys);
  vecsort(keys,,1+4);
}
test_seq_postorder_by_sort2(vpar,rev=0) =
{
  \\ keys are paths [root, ..., v, BIGNUM]
  \\ BIGNUM causes a longer path to sort as smaller than a shorter path,
  \\ hence descending into subtrees first.
  \\ For rev=1, keys are positives [-root, ..., -v, +BIGNUM] so bigger
  \\ vertex first.
  my(keys=vector(#vpar,v, concat([Vecrev(vpar_ancestors(vpar,v))*if(rev,-1,1),
                                  [#vpar+1]])));
  vecsort(keys,,1);
}

\\ Return a seq of the vertices at and below v in post-order.
test_seq_postorder_subtree_by_recurse(vpar,v,rev) =
{
  my(seq=Vecsmall([]),
     children=vpar_children_of_vertex(vpar,v));
  if(rev, children=Vecrev(children));
  for(i=1,#children,
     seq=concat(seq, test_seq_postorder_subtree_by_recurse(vpar,children[i],rev)));
  if(v, concat(seq,Vecsmall([v])), seq);
}
\\ Return a seq of vpar vertices in post-order.
test_seq_postorder_by_recurse(vpar,rev) =
  test_seq_postorder_subtree_by_recurse(vpar,0,rev);

\\ like binary tree natural correspondence traversed in-order
test_seq_postorder_by_inorder(vpar,rev=0,v='root) =
{
  if(!#vpar || !v, return(Vecsmall([])));
  if(v=='root,
     my(func=if(rev,vecmax,vecmin));
     v=func(vpar_roots(vpar)));
  my(child  =vpar_minmax_child_of_vertex(vpar,v,if(rev,max,min)),
     sibling=vpar_next_sibling_of_vertex(vpar,v,rev));
  concat([test_seq_postorder_by_inorder(vpar,rev,child),
          Vecsmall([v]),
          test_seq_postorder_by_inorder(vpar,rev,sibling)]);
}
check_equal(test_seq_postorder_by_inorder([0,0]),   Vecsmall([1,2]));
check_equal(test_seq_postorder_by_inorder([0,0],1), Vecsmall([2,1]));

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     for(rev=0,1,
        my(got=vpar_seq_postorder(Vecsmall(vpar), rev));
        check_equal(got, test_seq_postorder_by_recurse(vpar,rev),
                    Str("vpar_seq_postorder() vs test_seq_postorder_by_recurse() rev=",rev));
        check_equal(got, test_seq_postorder_by_sort(vpar,rev),
                    Str("vpar_seq_postorder() vs test_seq_postorder_by_sort() rev=",rev));
        check_equal(got, test_seq_postorder_by_sort2(vpar,rev),
                    Str("vpar_seq_postorder() vs test_seq_postorder_by_sort2() rev=",rev));
        check_equal(got, test_seq_postorder_by_inorder(vpar,rev),
                    Str("vpar_seq_postorder() vs test_seq_postorder_by_inorder() rev=",rev));
        check_equal(got, Vecsmall(Vecrev(vpar_seq_preorder(vpar,!rev))),
                    Str("vpar_seq_postorder() vs opposite vpar_seq_preorder() rev=",rev));
        check_equal(got,
                    Vecsmall(apply(v->#vpar+1-v,
                             Vec(vpar_seq_postorder(vpar_relabel_reverse(vpar), !rev)))),
                    "vpar_seq_postorder() vs reverse vpar_seq_preorder()");
        check_equal(vpar_seq_is_upwards(vpar,got), 1,
                    "vpar_seq_postorder() is vpar_seq_is_upwards() per addhelp");
        if(!rev,
           check_equal(vpar_seq_is_ordered(vpar,got), 1,
                       "vpar_seq_postorder() is vpar_seq_is_ordered() per docs"));

        \\ Not quite reversal, order among children is still ascending.
        \\ if(vpar_is_tree(vpar),
        \\ check_equal(pre_seq, Vecsmall(Vecrev(post_seq)),
        \\             "vpar_seq_preorder() reversal of vpar_seq_postorder() per docs"));
  ));
}

{
  \\ Antonio Guedes De Oliveira, Michel Las Vergnas, "Parking Functions
  \\ and Labelled Trees", Seminaire Lotharingien de Combinatoire, volume
  \\ 65, 2011, article B65e.
  \\ Figure 1 example (a).
  \\
  \\        10
  \\      /    \
  \\    1        7
  \\    |      / | \
  \\    9     2  3  5
  \\    |       / \
  \\    4      6   8
  my(vpar=[10,7,7,9,7, 3,10,3,1,0]);
  vpar_validate(vpar);
  check_equal(vpar_seq_postorder(vpar,1),
              Vecsmall([5,8,6,3,2,7,4,9,1,10]));

  \\ Figure 1 example (b).
  \\
  \\            10
  \\         /      \
  \\       8          9
  \\     / | \        |
  \\    2  5  7       4
  \\         / \      |
  \\        3   6     1
  my(vpar=[4,8,7,9, 8,7,8,10,10,0]);
  vpar_validate(vpar);
  check_equal(vpar_seq_postorder(vpar,1),
              Vecsmall([1,4,9,6,3,7,5,2,8,10]));

  \\ the depth-first numbering red annotations in the picture
  check_equal(vpar_seq_preorder(vpar),
              Vecsmall([10,8,2,5,7,3,6,9,4,1]));
}


\\------------------------------------------------------------------------------
progress("test_vector_insert_after()");

check_equal(test_vector_insert_after([],0,123), [123]);
check_equal(test_vector_insert_after([5,4],0,123), [123,5,4]);
check_equal(test_vector_insert_after([5,4],1,123), [5,123,4]);
check_equal(test_vector_insert_after([5,4],2,123), [5,4,123]);


\\-----------------------------------------------------------------------------
progress("vpar_INTERNAL_Euler_transform()");

{
  \\ a[n]=1 transformed
  \\ A000041 number of partitions, starting n=0
  check_equal(vector(#A000041_samples,n,n--; numbpart(n)),
              A000041_samples);  \\ OFFSET=0
  check_equal(vector(12,n,n--; numbpart(n)),
              vector(12,n,n--; #partitions(n)));

  check_equal(vector(20,n, numbpart(n)),
              vector(20,n, 1/n * sum(k=1,n, sumdiv(k,d,d)*numbpart(n-k))));

  my(samples=A000041_samples[^1]);
  check_equal(vpar_INTERNAL_Euler_transform(vector(#samples,n,1)),
              samples,
              "vpar_INTERNAL_Euler_transform() of 1 = A000219");

  \\ partitions per Euler (or is it Kronecker?) gf product
  my(limit=#A000041_samples,
     L=O('x^limit),
     gP=test_func_to_gf(n->A000041_samples[n+1],limit));
  check_equal(gP, prod(k=1,limit, 1/(1-'x^k + L)));
}
{
  \\ vector of 2s constant transformed
  \\ A000712 partitions into 2 kinds
  my(samples=A000712_samples[^1]);
  check_equal(vpar_INTERNAL_Euler_transform(vector(#samples,n,2)),
              samples,
              "vpar_INTERNAL_Euler_transform() of 2 = A000712");

  my(limit=#A000712_samples,
     L=O('x^limit),
     gT=test_func_to_gf(n->A000712_samples[n+1],limit));
  check_equal(gT, prod(k=1,limit, 1/(1-'x^k + L)^2));
}
{
  \\ a[n]=n transformed
  \\ A000219 something partitions of n, starting n=1
  my(samples=A000219_samples[^1]);
  check_equal(vpar_INTERNAL_Euler_transform(vector(#samples,n, n)),
              samples,
              "vpar_INTERNAL_Euler_transform() of n = A000219");
}

{
  \\ A000009 partitions of n into sum of distinct numbers
  my(limit=#A000009_samples,
     L=O('x^limit),
     gD=test_func_to_gf(n->A000009_samples[n+1],limit));
  check_equal(gD, prod(k=1,limit, 1+'x^k + L));
}

{
  \\ EulerTransform(A) = -1 + exp(gA(x)/1 + gA(x^2)/2 + ...)
  my(A=vector(10,i, random(10)),
     B=vpar_INTERNAL_Euler_transform(A),
     limit=#A+1,
     L=O('x^limit),
     gA = test_func_to_gf(n->if(n,A[n]), limit),
     gB = test_func_to_gf(n->if(n,B[n]), limit));
  check_equal(1+gB,
              exp(sum(k=1,limit, subst(gA,'x,'x^k)/k)));
}

{
  my(a=[5,7,13],
     c=vector(#a,k, sumdiv(k,d, d*a[d])),
     b=vpar_INTERNAL_Euler_transform(a));
  check_equal(c,[5,19,44]);
  check_equal(c,[5,
                 1*5 + 2*7,
                 1*5 + 3*13]);
  check_equal(b,[5,22,83]);
  check_equal(b,[   1 * 1*5     \\ k=1 n-k=0 d=1
                 ,
                 (  5 * 1*5     \\ k=1 n-k=1 d=1
                  + 1 * 1*5     \\ k=2 n-k=0 d=1
                  + 1 * 2*7     \\ k=2 n-k=0 d=2   <- 2s
                 ) / 2
                 ,
                 ( 22 * 1*5     \\ k=1 n-k=2 d=1
                  + 5 * 1*5     \\ k=2 n-k=1 d=1
                  + 5 * 2*7     \\ k=3 n-k=1 d=2
                  + 1 * 1*5     \\ k=3 n-k=0 d=1
                  + 1 * 3*13    \\ k=3 n-k=0 d=3   <- 3s
                 ) / 3
                ]);
  check_equal((22*5+5*5)/3, 45);
  check_equal((5*2*7+5)/3, 25);
  check_equal((3*13)/3, 13);
  check_equal(45+25+13, 83);
  check_equal(1 /(1-x)^5 /(1-x^2)^7 /(1-x^3)^13 + O(x^4),
              1 + 5*x + 22*x^2 + 83*x^3         + O(x^4));
  check_equal((1+x+x^2+x^3+x^4)^5 * (1+x^2+x^4)^7 * (1+x^3+x^6)^13  + O(x^4),
              1 + 5*x + 22*x^2 + 83*x^3         + O(x^4));
  check_equal((1+x+x^2+x^3+x^4)^5 * (1+x^2+x^4)^7   + O(x^4),
              1 + 5*x + 22*x^2 + 70*x^3         + O(x^4));
  check_equal((1+x+x^2+x^3+x^4)^5               + O(x^4),
              1 + 5*x + 15*x^2 + 35*x^3         + O(x^4));
  check_equal(( 22 * 1*5     \\ k=1 n-k=2 d=1
               + 5 * 1*5     \\ k=2 n-k=1 d=1
               + 1 * 1*5     \\ k=3 n-k=0 d=1
               ),
              140);

  my(L=O(x^4),
     gApoly= 5*x + 7*x^2 + 13*x^3 + L,
     gA=(x)->subst(gApoly,'x,x));
  check_equal(1 + 5*x + 22*x^2 + 83*x^3  + L,
              exp(sum(k=1,3, gA(x^k)/k)),
              "gf");

  \\ log(1-x)
  check_equal(log(1-x+L), -x - 1/2*x^2 - 1/3*x^3  +L);
  check_equal(log(1 + 5*x + 22*x^2 + 83*x^3  + O(x^4)),
              -5*log(1-x+L) - 7*log(1-x^2+L) - 13*log(1-x^3+L));
}
{
  check_equal(partitions(5),
              Set([Vecsmall([         5]),
                   Vecsmall([1,       4]),
                   Vecsmall([2,       3]),
                   Vecsmall([1,1,     3]),
                   Vecsmall([1,2,     2]),
                   Vecsmall([1,1,1,   2]),
                   Vecsmall([1,1,1,1, 1])]));
  check_equal(1*5 + 1*4 + 2*3 + 2*2 + 1*1,  20);
  check_equal(numbpart(1), 1);
  check_equal(numbpart(2), 2);
  check_equal(numbpart(3), 3);
  check_equal(numbpart(4), 5);
  check_equal(numbpart(5), 7);
}

\\-----------------------------------------------------------------------------
progress("test_A084091()");
\\ A084091  gf = sum(k>=0, x^2^k/(1 + x^2^k + x^2^(k+1)))

{
  check_equal(test_A084091(14), 0, "test_A084091(14)");
  check_equal(vector(#A084091_samples,n,n--; test_A084091(n)),  \\ OFFSET=0
              A084091_samples,
              "test_A084091() vs A084091_samples");
}

my(k=3); check_equal(x^2^k/(1+x^2^k+x^2^(k+1)),  x^8/(1 + x^8 + x^16));
my(k=3); check_equal(x^2^k/(1+x^2^k+x^2^(k+1)) == x^8*(1-x^8)/(1 - x^(3*8)),  1);
my(k=3); check_equal(x^2^k/(1+x^2^k+x^2^(k+1)) == (x^8 - x^(2*8))/(1 - x^(3*8)),  1);
my(k=0); check_equal( x /(1 - x^3) + O(x^16), \
                     x + x^4 + x^7 + x^10 + x^13 + O(x^16));  \\ 3n+1 periodic
my(k=1); check_equal( x^(2^k) /(1 - x^(3*(2^k))) + O(x^32), \
                     x^2 + x^8 + x^14 + x^20 + x^26 + O(x^32));
my(k=0); check_equal( x^2 /(1 - x^3) + O(x^16), \
                     x^2 + x^5 + x^8 + x^11 + x^14 + O(x^16));  \\ 3n+2 periodic

{
  \\ 1 - x^(3*(2^k))   is periodic 3*m * 2^k  for any m
  \\ x^(2^k)     puts +1 at 3*m *2^k + 2^k    = (3*m+1)*2^k
  \\ -x^(2*2^k)  puts -1 at 3*m *2^k + 2*2^k  = (3*m+2)*2^k
  for(k=0,8,
     my(limit = floor(3*2^(k-1)),
        L=O(x^limit),
        g = sum(k=0,k,
                ( x^(2^k) - x^(2*2^k) )/(1 - x^(3*(2^k)))
                + L),
        f=sum(n=0,limit-1, test_A084091(n) * x^n) + L);
     check_equal(g, f,
                 Str("gA084091  k="k" limit="limit));
  );
}

\\ A057078 periodic 1,0,-1
test_A057078(n) = if(n%3==0,1, n%3==1,0, -1);
check_equal(vector(#A057078_samples,n,n--; test_A057078(n)), /* OFFSET=0 */ \
            A057078_samples, \
            "test_A057078() vs A057078_samples");

{
  check_equal(vector(4096,n,n--; test_A084091(2*n)),
              vector(4096,n,n--; test_A084091(n) + 1 - ((n+1)%3)),
              "A084091 2n recurrence");  \\ Ralf Stephan
  check_equal(vector(4096,n,n--; test_A084091(2*n+1)),
              vector(4096,n,n--; 1 - (n%3)),
              "A084091 2n+1");  \\ Ralf Stephan

  check_equal(vector(4096,n,n--; test_A084091(2*n+1)),
              vector(4096,n,n--; test_A057078(n)),
              "A084091 2n+1 periodic");  \\ Michael Somos

  check_equal(vector(4096,n,n--; test_A084091(3*n)),
              vector(4096,n,n--; 0),
              "A084091 3n");  \\ Michael Somos

  check_equal(vector(4096,n,n--; test_A084091(3*n+1)),
              vector(4096,n,n--; test_A098725(n+1)),
              "A084091 3n+1");  \\ Michael Somos

  \\ n%3==1 and odd  \ unchanged by k%2==0 even mult 4*
  \\ n%3==2 and odd  /
  \\ 4n
  \\     1 - ((n+1)%3))  + 1 - ((2n+1)%3))
  \\   = 1 - ((n+1)%3))  + 1 - ((2(n+1)-1)%3))
  \\ vector(3,n,n--; [(n+1)%3, (2*n+1)%3])
  check_equal(vector(30,n,n--;    1 - ((n+1)%3)    ),
              vector(30,n,n--; - (1 - ((2*n+1)%3)) ));
  check_equal(vector(4096,n,n--; test_A084091(4*n)),
              vector(4096,n,n--; test_A084091(n)),
              "A084091 4n");

  \\ What is 3n+2 ?
  \\ check_equal(vector(40,n,n--; -A084091(3*n+2)),
  \\             vector(40,n,n--; test_A098725(n)),
  \\             "A084091 3n+2");

  my(by_loop=(n)->
     if(n==0,return(0));
     my(ret=0);
     while(n%2==0,
           \\ print("at even 2n = "test_to_binary(n)"  ret="ret);
           n/=2;
           \\ print("  to n = "test_to_binary(n)"  add "1 - ((n+1)%3));
           ret += 1 - ((n+1)%3));   \\ +/-1
     \\ print(n); \\ ret==0 || ret==-1 || error(ret);
     n \= 2;
     ret + 1-(n%3));
  check_equal(vector(4096,n,n--; test_A084091(n)),
              vector(4096,n,n--; by_loop(n)),
              "A084091 vs by_loop");
}
{
  my(by_bits=(n)->
     \\ return(if(n%3==0,0, -(-1)^(n%3)));

     my(k=0); while(n&&n%2==0,k++;n\=2);
     if(n%3==0, 0,
        n%3==1,   k%2==0,
        n%3==2, -(k%2==0));
     if(n%3==0, 0, -(-1)^(n%3)*(k%2==0));
     \\ if(n%3==0, 0, (k%2)-(n%3));
    );
  check_equal(vector(4096,n,n--; test_A084091(n)),
              vector(4096,n,n--; by_bits(n)),
              "A084091 vs by_bits");
}
{
  \\                                                     0 1 2  3 4  5
  check_equal(vector(4096,n,n--; test_A084091(n)),
              vector(4096,n,n--; while(n&&n%4==0,n\=4); [0,1,0, 0,0,-1][n%6+1]),
              "A084091 by base 4");

}
\\ for(k=0,3, for(m=1,32, my(n=3*m+2); n%2==1||next; n*=2^k; \
\\   printf("k=%d  %6s  %d\n", k,test_to_binary(n),test_A084091(n))));
\\ vector(20,k,k--; 4^k %6) == [1, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4]
{
  \\                                        0 1 2  3 4  5
  check_equal(vector(4096,n,n--; test_A084091(n)),
              vector(4096,n,n--; if(n&&valuation(n,2)%2,0, (n+1)%3-1)),
              "A084091 vs remainders");
}


\\-----------------------------------------------------------------------------
\\ Polmod Powering

{
  my(p='b*'x + 'c*'x^2 + 'd*'x^3,
     m = -10 + -20*'x + -30*'x^2 + 'x^3);
  check_equal(p % m == 10*d + (b + 20*d)*x + (c + 30*d)*x^2,
              1);
}
{
  my(m=-1-'x+'x^2);
  check_equal(polrecip(m), 1-'x-'x^2);
  check_equal(vector(10,n,n--; my(p=lift(Mod('x,m)^n));
                     [polcoeff(p,0),polcoeff(p,1)]),
              vector(10,n,n--; [fibonacci(n-1),fibonacci(n)]));
  \\ recurrence_guess(vector(20,n,n--; fibonacci(n)))
}

recurrence_coeffs_to_matrix(coeffs) = {
  matrix(#coeffs,#coeffs,r,c,
         if(c==r+1,1, r==#coeffs,coeffs[c], 0));
}
recurrence_eval(coeffs,initial,k) = {
  my(skip=#initial-#coeffs);
  if(skip && k<skip, return(initial[k+1]));
  initial=initial[skip+1 .. #initial];
  k-=skip;
  (recurrence_coeffs_to_matrix(coeffs)^k * Col(initial))[1];
}
dJA(k) = recurrence_eval([2,0,1],[0,0,1],k);
{
  check_equal(vector(16,n,n--; dJA(n)),
              [0, 0, 1, 1, 1, 3, 5, 7, 13, 23, 37, 63, 109, 183, 309, 527]);
  my(m=x^3 - x^2 - 2);
  check_equal(vector(10,n,n--; my(p=lift(Mod('x,m)^n));
                     polcoeff(p,2)),
              vector(10,n,n--; dJA(n)),
              "polcoeff 2");

  check_equal(vector(10,n,n--; my(p=lift(Mod('x,m)^n));
                     polcoeff(p,1)),
              vector(10,n,n--; dJA(n+1) - dJA(n) ),
              "polcoeff 1");

  check_equal(vector(10,n,n--; my(p=lift(Mod('x,m)^n));
                     polcoeff(p,0)),
              vector(10,n,n--; dJA(n+2) - dJA(n+1) ),
              "polcoeff 0");

  \\ recurrence_guess(vector(20,k,k--; dJA(k)))
}

{
  my(f=(k)->recurrence_eval([7,11,13,17],[0,0,0,1],k));
  check_equal(vector(10,n,n--; f(n)),
              [0, 0, 0, 1, 17, 302, 5366, 95342, 1694013, 30098807]);
  check_equal(vector(10,n,n--; f(n)),
              vector(10,n,n--; 17*f(n-1) + 13*f(n-2) + 11*f(n-3) + 7*f(n-4)));
  check_equal(vector(10,n,n--; f(n+4) - 17*f(n+3) - 13*f(n+2) - 11*f(n+1) ),
              vector(10,n,n--; 7*f(n)));
  \\ recurrence_guess(vector(20,k,k--; f(k)));

  my(m=x^4 - 17*x^3 - 13*x^2 - 11*x - 7);
  check_equal(vector(10,n,n--; my(p=lift(Mod('x,m)^n));
                     polcoeff(p,3)),
              vector(10,n,n--; f(n)),
              "polcoeff 3");

  check_equal(vector(10,n,n--; my(p=lift(Mod('x,m)^n));
                     polcoeff(p,2)),
              vector(10,n,n--; f(n+1) - 17*f(n) ),
              "polcoeff 2");

  check_equal(vector(10,n,n--; my(p=lift(Mod('x,m)^n));
                     polcoeff(p,1)),
              vector(10,n,n--; f(n+2) - 17*f(n+1) - 13*f(n) ),
              "polcoeff 1");

  check_equal(vector(10,n,n--; my(p=lift(Mod('x,m)^n));
                     polcoeff(p,0)),
              vector(10,n,n--; f(n+3) - 17*f(n+2) - 13*f(n+1) - 11*f(n) ),
              "polcoeff 0");

  check_equal((       'a*x +        'b*x^2       +  'c*x^3       + 'd*x^4) % m
              ==
              7*'d + ('a + 11*'d)*x + ('b + 13*'d)*x^2 + ('c + 17*'d)*x^3,
              1);

  \\ d = f(n)
  \\ c = f(n+1) - 17*f(n)
  \\ c + 17*d = f(n+1)    = new d
  \\
  \\ a = f(n+3) - 17*f(n+2) - 13*f(n+1) - 11*f(n)
  \\
}

\\-----------------------------------------------------------------------------
\\ test_antidiagonals_of_func()

{
  \\  11 12 13
  \\  21 22
  \\  31
  check_equal(test_antidiagonals_of_func(10,(n,m)->[n,m]),
              [[1,1],
               [2,1], [1,2],
               [3,1], [2,2], [1,3],
               [4,1], [3,2], [2,3], [1,4]],
              "test_antidiagonals_of_func()");

  check_equal(test_antidiagonals_of_func(6,(n,m)->[n,m], 0,0),
              [[0,0],
               [1,0], [0,1],
               [2,0], [1,1], [0,2]],
              "test_antidiagonals_of_func() 0,0");

  check_equal(test_antidiagonals_of_func(6,(n,m)->[n,m], 1,0),
              [[1,0],
               [2,0], [1,1],
               [3,0], [2,1], [1,2]],
              "test_antidiagonals_of_func() 1,0");

  \\  07 08 09     columns starting from 7
  \\  17 18        and downwards through
  \\  27
  check_equal(test_antidiagonals_of_func(6,(n,m)->[n,m], 0,7, 1),
              [[0,7],
               [0,8], [1,7],
               [0,9], [1,8], [2,7]],
              "test_antidiagonals_of_func() 0,7,1");
}


\\-----------------------------------------------------------------------------
progress("Binomial Tree Balanced Binary");

\\ A092124 a(0) = 2, a(n) = (2^(2^n)+2)*a(n-1)
\\ binomial tree 2^k in balanced binary integer
A092124(n) = 2*prod(i=1,n, 2^(2^i) + 2);
{
  check_equal(vector(#A092124_samples,n,n--; A092124(n)),
              A092124_samples,  \\ OFFSET=0
              "A092124() vs A092124_samples");
}

A333447(n) =
{
  \\ formula in A333447
  \\ extra bit above high copy, drop top bit of low copy
  my(a=2);
  for(n=1,n, a = 2^(2^(n+1)-1) - 2^(2^n-1)  + a*(2^(2^n-1) + 1));
  a;

  \\     z100 z100
  \\    z1100  100
  my(a=0);
  for(n=1,n,
     a = 2^(2^(n+1)-2) + a*(2^(2^n-1) + 1);
  );
  a + 2^(2^(n+1)-1);

  my(ret=0);
  for(i=1,n, ret += 1<<(2^(i+1)-2) + ret<<((1<<i)-1) );
  ret + 1<<((2<<n)-1);

  my(ret=0);
  for(i=1,n, ret += ret<<((1<<i)-1) + 1<<((2<<i)-2));
  ret + 1<<((2<<n)-1);

  my(b=1); for(i=1,n, b += b<<((1<<i)-1) + 1<<((2<<i)-1)); b+1;
}
{
  check_equal(vector(#A333447_samples,n,n--; test_to_binary(A333447(n))),
              apply(test_to_binary,A333447_samples),  \\ OFFSET=0
              "A333447() vs A333447_samples");
}

\\ A210995 blocks from A079559
\\ except a(0)=1,
\\ binomial tree 2^k in balanced binary integer, written in binary
A210995(n) = if(n==0,1, 10*prod(i=1,n-1, 10^(2^i) + 10));
{
  check_equal(vector(#A210995_samples,n,n--; A210995(n)),
              A210995_samples,  \\ OFFSET=0
              "A210995() vs A210995_samples");
}

{
  \\ binomial tree 2^k as balanced binary

  \\ A210995 = rows of A079559 = binary
  \\ 1, 10, 1100, 11011000, 1101100110110000, 11011001101100011011001101100000
  \\ balanced binary binomial tree starts 10,1100,...
  for(k=0,6,
     my(vpar=vpar_make_binomial_tree(2^k),
        b=vpar_to_balanced_binary(vpar));
     check_equal(b, A092124(k),
                 "A092124 binomial tree, as balanced binary");
     check_equal(test_to_binary(b), A210995(k+1),
                 "A210995 binomial tree, as balanced binary");

     \\ A333447 reversed balanced binary
     vpar=vpar_relabel_reverse(vpar);
     b=vpar_to_balanced_binary(vpar);
     check_equal(b, A333447(k),
                 "A333447 binomial tree reversed, as balanced binary");
  );

  \\ sans low 0-bit
  check_equal(vector(5,k,k--; test_to_binary(vpar_to_balanced_binary
                                             (vpar_make_binomial_tree(2^k))/2)),
       [1, 110, 1101100, 110110011011000, 1101100110110001101100110110000]);
  \\ not in OEIS: 1, 110, 1101100, 110110011011000, 1101100110110001101100110110000
}

test_binomial_tree_balanced_binary(n) =
{
  if(n==0,return(0));
  my(v=[1,0]);
  while(#v < 2*n, v=concat(apply(bit->if(bit,[1,1,0],[0]), v)));
  while(hammingweight(v)>n,
        forstep(i=#v,1,-1,
           if(v[i],
              v=concat(v[1..i-1],v[i+2..#v]);
              break)));
  vpar_INTERNAL_fromdigits(v,2);
}
{
  \\ binomial tree n as balanced binary

  for(n=0,32,
     my(vpar=vpar_make_binomial_tree(n));
     check_equal(vpar_to_balanced_binary(vpar),
                 test_binomial_tree_balanced_binary(n),
                 "binomial tree balanced binary vs test_binomial_tree_balanced_binary()"));

  check_equal(vector(10,n, vpar_to_balanced_binary(vpar_make_binomial_tree(n))),
              [2, 12, 52, 216, 868, 3480, 13928, 55728, 222916, 891672]);
  \\ not in OEIS: 2, 12, 52, 216, 868, 3480, 13928, 55728, 222916, 891672

  check_equal(vector(10,n, test_to_binary(vpar_to_balanced_binary
                                          (vpar_make_binomial_tree(n)))),
              [10, 1100, 110100, 11011000, 1101100100, 110110011000, 11011001101000, 1101100110110000, 110110011011000100, 11011001101100011000]);
  \\ not in OEIS: 10, 1100, 110100, 11011000, 1101100100, 110110011000
  \\ not in OEIS: 1, 110, 11010, 1101100, 110110010, 11011001100
}

\\ Return the number of low 1-bits on n.
\\ Going n+1 changes them to low 0-bits which valuation() counts.
test_count_low_ones(n) = valuation(n+1,2);
check_equal(test_count_low_ones(0), 0);
check_equal(test_count_low_ones(1), 1);
check_equal(test_count_low_ones(2), 0);
check_equal(test_count_low_ones(3), 2);

test_binomial_tree_balanced_binary_vector(n,trim=0) =
{
  my(ret=vector(2*n),
     upto=0);
  for(v=0,n-1,
     ret[upto++] = 1;                   \\ 1 for vertex v
     for(j=1,test_count_low_ones(v),    \\ then step up by low ones to next v
        ret[upto++]=0));
  if(trim, Vec(ret,upto),
     ret);
}
check_equal(test_binomial_tree_balanced_binary_vector(1),   [1,0]);
check_equal(test_binomial_tree_balanced_binary_vector(2),   [1,1,0,0]);
check_equal(test_binomial_tree_balanced_binary_vector(3),   [1,1,0,1,0,0]);
check_equal(test_binomial_tree_balanced_binary_vector(3,1), [1,1,0,1]);
check_equal(test_binomial_tree_balanced_binary_vector(5),   [1,1,0,1,1,0,0,1,0,0]);
check_equal(test_binomial_tree_balanced_binary_vector(5,1), [1,1,0,1,1,0,0,1]);
\\ test_binomial_tree_balanced_binary_vector(10,1)
{
  for(n=1,50,
     check_equal(test_binomial_tree_balanced_binary_vector(n),
                 binary(vpar_to_balanced_binary(vpar_make_binomial_tree(n))),
                 "binomial tree balanced binary vs test_binomial_tree_balanced_binary_vector()"));
}


\\-----------------------------------------------------------------------------
\\ test_bit_complement_and_reverse()
\\ A036044

{
  \\ A036044 reckons n=0 as one 0-bit rather than empty
  check_equal(vector(#A036044_samples,n,n--;
                     if(n==0,1, test_bit_complement_and_reverse(n))),
              A036044_samples,  \\ OFFSET=0
              "A036044 = test_bit_complement_and_reverse()");

  \\ low 1-bits become high 0-bits so are lost to the BCR
  check_equal(vector(#A036044_samples,n,n=2*n-1;
                     if(n==1,1, test_bit_complement_and_reverse(n))),
              A036044_samples,  \\ OFFSET=0
              "A036044 = test_bit_complement_and_reverse() of odds");
}
{
  \\ A059894 BCR of evens, meaning high bit unchanged and BCR the rest
  check_equal(vector(#A059894_samples,n,n=2*n;
                     test_bit_complement_and_reverse(n)/2),
              A059894_samples,  \\ OFFSET=1
              "A059894 = test_bit_complement_and_reverse() of evens, sans low 0");
}


\\-----------------------------------------------------------------------------
\\ test_bitlength()

check_equal(test_bitlength(5),3);

\\ A029837 bit length
check_equal(vector(#A029837_samples,n,n--; test_bitlength(n)), \
            A029837_samples);   \\ bit it OFFSET=1 actually
\\ A004258 duplicate
check_equal(vector(#A004258_samples,n, test_bitlength(n-1)), \
            A004258_samples);   \\ OFFSET=1


\\-----------------------------------------------------------------------------
\\ test_highbit()

check_equal(test_highbit(5),4);

\\ A053644 MSB
check_equal(vector(#A053644_samples,n,n--; test_highbit(n)), \
            A053644_samples);   \\ OFFSET=0


\\------------------------------------------------------------------------------
progress("Balanced Binary Stuff");

\\ A037800 count 01 bit pairs in n
test_count_01(n) = my(v=binary(n)); sum(i=1,#v-1, v[i]==0 && v[i+1]==1);
{
  check_equal(vector(#A037800_samples,n,n--; test_count_01(n)),
              A037800_samples,
              "test_count_01 vs A037800_samples");
}
{
  \\ A127284 count 01 in balanced binary
  \\ (Tamari Successors)
  my(l=List([]),b=0);
  while(#l < #A127284_samples,
        listput(l,test_count_01(b));
        b=vpar_balanced_binary_next(b));
  check_equal(Vec(l),
              A127284_samples,
              "A127284 balanced binary valleys");
}

\\ A033264 count 10 bit pairs in n
test_count_10(n) = my(v=binary(n)); sum(i=1,#v-1, v[i]==1 && v[i+1]==0);
{
  check_equal(vector(#A033264_samples,n, test_count_10(n)),
              A033264_samples,  \\ OFFSET=1
              "test_count_10 vs A033264_samples");
}
{
  \\ A057514 count 10 in balanced binary
  my(l=List([]),b=0);
  while(#l < #A057514_samples,
        listput(l,test_count_10(b));
        b=vpar_balanced_binary_next(b));
  check_equal(Vec(l), A057514_samples,
              "A057514 balanced binary peaks");
}

\\ A014081 count 11 bit pairs in n
test_count_11(n) = my(v=binary(n)); sum(i=1,#v-1, v[i]==1 && v[i+1]==1);
{
  check_equal(vector(#A014081_samples,n,n--; test_count_11(n)),
              A014081_samples,  \\ OFFSET=0
              "test_count_11 vs A014081_samples");
}
{
  \\ A126306 count 11 in balanced binary
  my(l=List([]),b=0);
  while(#l < #A126306_samples,
        listput(l,test_count_11(b));
        b=vpar_balanced_binary_next(b));
  check_equal(Vec(l), A126306_samples,
              "A126306 balanced binary 11");
}

{
  \\ A057515
  \\ balanced binary forest num component trees
  my(l=List([]),b=0);
  while(#l < #A057515_samples,
        my(vpar=vpar_from_balanced_binary(b));
        listput(l,vpar_num_components(vpar));
        b=vpar_balanced_binary_next(b));
  check_equal(Vec(l), A057515_samples,
              "A057515 balanced binary num mountains");
}
{
  \\ balanced binary first mountain size (returning to sea-level)
  \\ not in OEIS: 1, 1, 2, 3, 3, 1, 1, 1, 1, 1, 2, 2, 3, 4, 4, 3, 4, 4, 4
  my(l=List([]),b=2);
  while(#l < 20,
        my(vpar=vpar_from_balanced_binary(b),
           sizes=vpar_component_sizes(vpar));
        listput(l,sizes[1]);
        b=vpar_balanced_binary_next(b));
  \\ print(Vec(l));

  \\ n=4 first mountain
  \\ not in OEIS: 4,4,4,3,4,4,3,2,2,1,1,1,1,1
  \\ not in OEIS: 1,1,1,1,1,2,2,3,4,4,3,4,4,4
  \\ n=4 last mountain
  \\ not in OEIS: 4,4,4,1,4,4,1,2,1,3,3,1,2,1
  \\ not in OEIS: 1,2,1,3,3,1,2,1,4,4,1,4,4,4
  \\ my(n=4,
  \\    forest=1,
  \\    vpar=vpar_preorder_first(n)); \\ ,forest
  \\ until((vpar=vpar_preorder_next(vpar,forest))=='none,
  \\       my(sizes=vpar_component_sizes(vpar));
  \\       print1(sizes[#sizes],","));
  \\ print();
}

{
  \\ A072643 balanced binary lengths
  \\ inverse of A014138 Catalan cumulative
  my(l=List([]),b=0);
  while(#l < #A072643_samples,
        listput(l,test_bitlength(b)/2);
        b=vpar_balanced_binary_next(b));
  check_equal(Vec(l), A072643_samples,
              "A072643 balanced binary lengths");
}

{
  \\ balanced binary increments length    (lengths of A085192)
  \\ not in OEIS: 2, 4, 2, 5, 2, 3, 2, 3, 7, 2, 3
  \\ not in OEIS:    5, 3, 6, 3, 4, 3, 4, 8, 3, 4, 3, 4
  \\ not in OEIS:    3, 1, 4, 1, 2, 1, 2, 6, 1, 2, 1, 2
  \\ balanced binary increments num 1s at and below change
  \\ not in OEIS: 1, 1, 4, 1, 2, 1, 1, 4, 1, 2, 1, 1, 2, 1, 2, 1, 1
  \\
  \\ my(l=List([]),prev=0,b=0,n=0);
  \\ while(#l < #A072643_samples,
  \\       b=vpar_balanced_binary_next(b);
  \\       listput(l, test_bitlength(b-prev)-1);
  \\      prev=b);
  \\ check_equal(Vec(l), A072643_samples,
  \\             "A072643 balanced binary lengths");
}



\\-----------------------------------------------------------------------------
progress("vpar_INTERNAL_Catalan_number()");

check_equal(vpar_INTERNAL_Catalan_number(10-1), 4862);
check_equal(vector(6,n,vpar_INTERNAL_Catalan_number(n)), [1,2,5,14,42,132]);

\\ ratio  C(n) = (4*n-2)/(n+1) * C(n-1)
check_equal(vector(30,n, vpar_INTERNAL_Catalan_number(n) / vpar_INTERNAL_Catalan_number(n-1)), \
            vector(30,n, (4*n-2)/(n+1)));

{
  \\ Euler terms steps of 4
  check_equal(vector(30,n,n--; vpar_INTERNAL_Catalan_number(n)),
              vector(30,n,n--; prod(i=1,n, 4*i-2) / (n+1)!));
}
{
  \\ Senger, forward times reverse
  my(vec=vector(30,n, 1));
  for(n=3,#vec,
     vec[n] = sum(i=1,n-1, vec[i]*vec[n-i]));
  check_equal(vector(#vec,n,n--; vpar_INTERNAL_Catalan_number(n)),
              vec);

  \\ counting binary trees, size of left and size of right
  check_equal(vector(30,n, vpar_INTERNAL_Catalan_number(n)),
              vector(30,n, sum(l=0,n-1, vpar_INTERNAL_Catalan_number(l)
                               * vpar_INTERNAL_Catalan_number(n-1-l))));
}
{
  \\ Pascal's triangle centre column, subtract its adjacent
  \\ per Vern Hoggatt
  check_equal(vector(30,n,n--; vpar_INTERNAL_Catalan_number(n)),
              vector(30,n,n--; binomial(2*n,n) - binomial(2*n,n-1)));
}
{
  \\ A000108 Catalan numbers
  check_equal(vector(#A000108_samples,n,n--; vpar_INTERNAL_Catalan_number(n)),
              A000108_samples);    \\ OFFSET=0
}
{
  \\ Catalan gf
  my(limit = 10,
     L = O(x^limit),
     gC = test_func_to_gf(vpar_INTERNAL_Catalan_number, limit));  \\ ordinary gf
  check_equal(gC,       1 + x*gC^2 + L,          "gC Catalan");
  check_equal(x*gC^2 - gC + 1,     0+L,          "gC Catalan 2");
  check_equal(gC + L/x, (1-sqrt(1-4*x+L))/(2*x), "gC Catalan 3");
}

{
  \\ A120588 preorder and downwards trees, starting n=0
  \\ same as Catalan A000108 but an extra initial 1.
  \\ vpar_count_ordered() of n-1
  check_equal(vector(#A120588_samples,n,n--; if(n==0,1, vpar_count_ordered(n))),
              A120588_samples,    \\ OFFSET=0
              "A120588 vs vpar_count_ordered() forests");
}
{
  \\ A120588 gf
  my(limit=#A120588_samples,
     L = O(x^limit),
     gA120588 = test_func_to_gf(n->A120588_samples[n+1], limit));
  check_equal(3*gA120588,  2 + x + gA120588^2,
              "A120588 gf equality");
  check_equal(gA120588^2 - 3*gA120588 + x+2,  0+L,
              "A120588 gf quadratic");
  check_equal(gA120588, (3-sqrt(9-4*(x+2)+L))/2, "gA120588 gf quadratic formula");
  check_equal(gA120588, (3-sqrt(1-4*x+L))/2, "gA120588 gf quadratic formula");
}

{
  \\ A014137 Catalan numbers cumulative
  check_equal(vector(#A014137_samples,n,n--;
                     sum(i=0,n,vpar_INTERNAL_Catalan_number(i))),
              A014137_samples,    \\ OFFSET=0
              "A014137 Catalan numbers cumulative");
}
{
  \\ A014138 Catalan numbers cumulative, starting from 1
  check_equal(vector(#A014138_samples,n,n--;
                     sum(i=1,n,vpar_INTERNAL_Catalan_number(i))),
              A014138_samples,    \\ OFFSET=0
              "A014138 Catalan numbers cumulative");
}
{
  \\ A071724 Catalan(n+1) - Catalan(n), but starting 1 at n=0
  check_equal(vector(#A071724_samples,n,n--;
                     if(n==0, 1,
                        vpar_INTERNAL_Catalan_number(n+1)
                        - vpar_INTERNAL_Catalan_number(n))),
              A071724_samples,    \\ OFFSET=0
              "A071724 Catalan(n+1) - Catalan(n)");
}


\\-----------------------------------------------------------------------------
progress("vpar_make_binomial_tree()");

{
  my(vpar=vpar_make_binomial_tree(2^4));
  check_equal(vpar_Wiener_index(vpar),392);
  check_equal(vpar_indnum(vpar),8);
  check_equal(vpar_domnum(vpar),8);
  check_equal(vpar_matchnum(vpar),8);
}

\\ A090996 number of high 1-bits
test_A090996(n) =
{
  my(h=if(n,logint(n,2)));
  forstep(i=h,-1,-1,
     if(!bittest(n,i),
        return(h-i)));
  0;

  \\ if(n,
  \\ \\ my(h=logint(n,2));
  \\ \\ my(m=bitxor((n-(1<<h))<<1,n));
  \\ \\ print(m);
  \\
  \\    my(h=logint(n,2));
  \\    h - logint(bitxor((n-(1<<h))<<1,n), 2) + 1;
  \\   );
}
my(n=13,h=logint(n,2)); check_equal((n-(1<<h))<<1, 10);
check_equal(test_A090996(6), 2, "A090996(6)");
check_equal(test_A090996(2), 1, "A090996(2)");
check_equal(test_A090996(13), 2, "A090996(13)");
check_equal(test_A090996(26), 2, "A090996(26)");
check_equal(test_A090996(15), 4, "A090996(15)");
check_equal(test_A090996(7), 3, "A090996(7)");
check_equal(test_A090996(3), 2, "A090996(3)");
{
  for(i=1,#A090996_samples,  \\ OFFSET=0
     my(n=i-1);
     check_equal(test_A090996(n), A090996_samples[i],
                 "test_A090996() vs A090996_samples[]"));
}

\\ A126236 max Huffman codeword length, n symbols freq 1..n
\\ conjectured floor(log2(n)) + floor(log2(2n/3))
test_A126236(n) = logint(n,2) + logint((2*n)\3,2);
{
  check_equal(vector(#A126236_samples,n,n++; test_A126236(n)),
              A126236_samples);  \\ OFFSET=2
}

test_second_highest_bit(n) = if(n==0,0, bittest(n, logint(n,2)-1));

\\ A079944 second highest bit
test_A079944(n) = n+=2; bittest(n, logint(n,2)-1);
{
  check_equal(vector(#A079944_samples,n,n--; test_A079944(n)),
              A079944_samples,  \\ OFFSET=0
              "test_A079944() second highest bit");
}

\\ A086694 complement second highest bit
test_A086694(n) = n++; !bittest(n, logint(n,2)-1);
{
  check_equal(vector(#A086694_samples,n, test_A086694(n)),
              A086694_samples,  \\ OFFSET=1
              "test_A086694() complement second highest bit");
}
{
  \\ by n
  for(n=0,100,
     my(vpar=vpar_make_binomial_tree(n));
     vpar_validate(vpar);
     check_equal(#vpar,n);
     check_equal(vpar_is_downwards(vpar), 1,
                 "vpar_make_binomial_tree() n, always vpar_is_downwards()");
     check_equal(vpar_is_preorder(vpar), 1,
                 "vpar_make_binomial_tree() n, always vpar_is_preorder()");
     check_equal(vpar_is_equimatchable(vpar), n<=3,
                 "vpar_make_binomial_tree() n, equimatchable only n<=3 path");
     check_equal(vpar_has_perfect_matching(vpar), n%2==0,
                 "vpar_make_binomial_tree() n, has perfect matching");

     \\ depth = number of 1-bits
     my(depths=vpar_depths_vector(vpar));
     check_equal(depths, vector(#vpar,v, hammingweight(v-1)), \\ A000120
                 "vpar_make_binomial_tree() n, depths vector");

     \\ height
     \\ n=1xxxx  length(xxxx) = h = floor(log2(n))
     \\ can go 0xxxx up h many steps to root 0
     \\
     my(height=vpar_height(vpar));
     check_equal(height,
                 if(n==0,-1, logint(n,2)),  \\ A000523
                 "vpar_make_binomial_tree() n, height");

     \\ num vertices at height
     check_equal(vpar_height_num_vertices(vpar),
                 test_A090996(n),
                 "A090996 vpar_make_binomial_tree() n, height num vertices");

     \\ diameter, increases at v=101111 and again at v=111111
     \\ A126236
     \\ 1,2,3,3,4,4,5,5,5,5,6,6,6,6,7,7,7,7,7,7,7,7,8,8,8,8,8,8,8,
     if(n>=2,
        check_equal(vpar_diameter(vpar),
                    test_A126236(n),
                    "binomial tree n diameter"));

     \\ centres
     my(centres=vpar_centres(vpar));
     check_equal(#centres,
                 if(n<=1,n, 2-bittest(n,logint(n,2)-1)),
                 "binomial tree number of centres, second highest bit");
     if(#centres,
        check_equal(vecmax(centres),
                    if(test_second_highest_bit(n)==1, 1,
                       test_highbit(n)>>1 + 1),
                    "binomial tree bigger centre"));
  );
}

{
  \\ by order n = 2^k
  for(k=0,8,
     my(n=2^k,
        vpar=vpar_make_binomial_tree(n));
     vpar_validate(vpar);
     my(centres=vpar_centres(vpar));
     check_equal(#centres, if(k==0,1,2),
                 "vpar_make_binomial_tree() order k bicentral for k>0");
     check_equal(vpar_root(vpar), 1);
     check_equal(vecmin(centres), 1,
                 "vpar_make_binomial_tree() root is centre");

     check_equal(vpar_diameter(vpar), if(k==0,0, 2*k-1),
                 "vpar_make_binomial_tree() diameter");
  );
}

\\ {
\\   \\ binomial tree n indnum count, maximum independent sets
\\   for(n=2,#A126236_samples+1,
\\      my(vpar=vpar_make_binomial_tree(n),
\\         want=A126236_samples[n-1]);
\\      print1(vpar_indnum_and_count(vpar)[2],","));
\\ \\want, "binomial tree n indnum"));
\\ }

{
  \\ A192021 binomial tree 2^k Wiener index
  for(i=1,#A192021_samples,
     my(k=i-1);    \\ A192021 OFFSET=0
     if(k>14, break());  \\ limit to 2^14 vertices
     my(want=A192021_samples[i],
        vpar=vpar_make_binomial_tree(2^k));
     vpar_validate(vpar);
     check_equal(#vpar,2^k, "num vertices");
     check_equal(vpar_Wiener_index(vpar),want,
                 Str("vpar_make_binomial_tree() k="k" Wiener")));

  check_equal(vector(15,n,n++; vpar_Wiener_index(vpar_make_binomial_tree(n))),
            [1, 4, 10, 18, 31, 46, 68, 88, 117, 148, 190, 226, 275, 326, 392]);
  \\ not in OEIS: 1, 4, 10, 18, 31, 46, 68, 88, 117, 148, 190, 226, 275, 326, 392

  \\ total path lengths to root
  \\ A000788 total 1s in 0 .. n
  \\ 1, 2, 4, 5, 7, 9, 12, 13, 15, 17, 20, 22, 25, 28, 32
  for(n=1,50,
     check_equal(vecsum(vpar_path_lengths_vector(vpar_make_binomial_tree(n),1)),
                 sum(i=0,n-1, hammingweight(i))));

  \\ total path lengths to last vertex
  \\
  check_equal(vector(19,n,n++; vecsum(vpar_path_lengths_vector(vpar_make_binomial_tree(n),n))),
              [1,3,6,8,13,15,22,20,29,31,42,36,49,51,66,48,65,67,86]);
  check_equal(vector(19,n,n++; vecsum(vpar_path_lengths_vector(vpar_make_binomial_tree(n),n))
                     - vecsum(vpar_path_lengths_vector(vpar_make_binomial_tree(n-1),n-1))),
              [1, 2, 3, 2, 5, 2, 7, -2, 9, 2, 11, -6, 13, 2, 15, -18, 17, 2, 19]);

  \\ even to even total
  \\ odd to even  total + 1*paths
  \\ odd to odd   total + 2*paths
}
{
  \\ A192020 num pairs of vertices in the binomial tree at distance k
  \\ coeffs of distance polynomial
  \\
  \\     0      0          0
  \\     |      | \        | \  \
  \\     1      1  2       1  2   4
  \\               |          |   | \
  \\               3          3   5  6
  \\  1 at 1   3 at 1                |
  \\           2 at 2                7
  \\           1 at 3
  \\                        1 at 5
  my(l=[]);
  for(n=1,#A192020_samples,
     if(#l>=#A192020_samples, break);
     my(vpar=vpar_make_binomial_tree(2^n),
        m=vpar_path_lengths_matrix(vpar),
        histogram=vector(2*n-1));
     for(u=1,#vpar,
        for(v=1,u-1,
           histogram[m[u,v]]++));
     l=concat(l,histogram));
  l=Vec(l,#A192020_samples);
  check_equal(l, A192020_samples);

  \\ my a(n) formula
  my(a=(n)-> my(s=sqrtint(n),r=n-s^2); sum(i=0,s, 2^(s-i)*binomial(2*i,r)));
  check_equal(vector(#A192020_samples,n,n--; a(n)),
              A192020_samples,    \\ OFFSET=0
              "A192020 a(n) flattened");

  for(n=0,6,
     my(vpar=vpar_make_binomial_tree(2^n),
        m=vpar_path_lengths_matrix(vpar),
        histogram=vector(2*n));
     for(u=1,#vpar,
        for(v=1,u-1,
           histogram[m[u,v]]++));

     check_equal(vpar_num_edges(vpar), 2^n-1,
                 "binomial tree vpar_num_edges()");
     check_equal(vecsum(histogram), binomial(#vpar,2),
                 "binomial tree histogram[] vecsum = vertex pairs");
     check_equal(if(#histogram<1,0, histogram[1]),
                 2^n-1,
                 "binomial tree Wiener histogram[1] = num edges");
     check_equal(if(n<2,0, histogram[2]),
                 2^(n+1) - 2*n - 2,
                 "binomial tree Wiener histogram[2]");

     \\ Suppose two vertices u,v have their highest bit difference at bit
     \\ position j, counting least significant bit as j=1.  One of them has
     \\ a 1-bit and the other has a 0-bit.  Take u as the 0-bit and v as the
     \\ 1-bit.  The bits above are the same in both and those, with 0s
     \\ below, is the closest common ancestor c of u,v.
     \\
     \\ To be distance k apart, u,v must be total depth k down from c, which
     \\ means k many 1-bits at and below j.  v already has one 1-bit in v,
     \\ so there are k-1 more 1-bits to arrange in the 2*(j-1) bit positions
     \\ below j.  There are n-j bits above j and they can be arbitrary so
     \\ 2^(n-j).
     \\
     \\ If u gets no 1-bits below j then have u=c, so u an ancestor k levels
     \\ up from v.
     \\
     \\ If v gets no 1-bits below j, only a 1-bit at j, then it is a child
     \\ of c,
     \\          c
     \\         / \          v has 1-bit at j, child of c
     \\        *   v
     \\       /              u has other 1-bits, is further down
     \\      *
     \\     /
     \\    u
     \\
     for(k=1,#histogram,
        check_equal(histogram[k],
                    sum(j=1,n, 2^(n-j)*binomial(2*j-2,k-1))));

     check_equal(t,'t, "t unbound");

     \\ generating function for number of vertices at depth
     my(r=(n,t)->(t+1)^n);
     check_equal(r(n,t),
                 sum(v=1,#vpar, t^vpar_depth_of_vertex(vpar,v)),
                 "r(n,t) vs vpar_depths_vector()");
     check_equal(r(n,t),
                 sum(i=0,n, binomial(n,i)*t^i),
                 "r(n,t) binomials");

     \\ generating function for number A192020 row
     my(w=(n,t)-> t * ((t+1)^(2*n) - 2^n)/((t+1)^2 - 2)
                 + 0*t);  \\ force t_POL always
     if(n>=1,
        check_equal(w(n,t),
                    2*w(n-1,t) + t*(r(n-1,t))^2,  \\ Emeric Deutsch
                    "w(n,t) recurrence"));
     check_equal(w(n,t),
                 sum(i=0,n-1,  2^(n-1-i)*t*(t+1)^(2*i))     + 0*t,
                 "w(n,t) sum");
     check_equal(w(n,t),
                 t*Polrev(histogram,t),
                 "w(n,t) vs histogram[]");

     check_equal(vpar_Wiener_index(vpar),
                 (n-1)*2^(2*n-1)+2^(n-1),
                 "binomial tree vpar_Wiener_index() vs formula");
     check_equal(vpar_Wiener_index(vpar),
                 2^n * ((n-1)*2^n + 1) / 2,
                 "binomial tree vpar_Wiener_index() vs formula 2");
     check_equal(((1+1)^2 - 2), 2);

     check_equal(w(n,t)',
                 sum(i=0,n-1, 2^(n-1-i)*( (t+1)^(2*i) + (2*i)*t*(t+1)^(2*i-1) )) + 0*t,
                 "w(n,t)' derivative sum, product rule");

     check_equal(simplify( subst((w(n,t))',t,1) ),
                 vpar_Wiener_index(vpar),
                 "w(n,t)' at t=1 vs vpar_Wiener_index()");

     my(wdash=(n,t)->( (t * ((t+1)^(2*n) - 2^n))' * ((t+1)^2 - 2)
                      - (t * ((t+1)^(2*n) - 2^n)) * ((t+1)^2 - 2)' )
                     / ((t+1)^2 - 2)^2 );
     check_equal(wdash(n,t)          + 0*t,
                 w(n,t)',
                 "wdash 1");

     my(wdash=(n,t)->( ( t * ((t+1)^(2*n) - 2^n)'
                        + ((t+1)^(2*n) - 2^n)
                        ) * ((t+1)^2 - 2)
                      - (t * ((t+1)^(2*n) - 2^n)) * (2*t + 2) )
                     / ((t+1)^2 - 2)^2 );
     check_equal(wdash(n,t)          + 0*t,
                 w(n,t)',
                 "wdash 2");

     my(wdash=(n,t)->( (  2*n*t*(t+1)^(2*n-1)  + (t+1)*(t+1)^(2*n-1) - 2^n
                        ) * ((t+1)^2 - 2)
                      - (t * ((t+1)^(2*n) - 2^n)) * (2*t + 2) )
                     / ((t+1)^2 - 2)^2 );
     check_equal(wdash(n,t)          + 0*t,
                 w(n,t)',
                 "wdash 3");

     my(wdash=(n,t)-> ( ((2*n+1)*t +1) *(t+1)^(2*n-1) - 2^n
                        ) * ((t+1)^2 - 2)
                      / ((t+1)^2 - 2)^2
                      - (t * ((t+1)^(2*n) - 2^n)) * (2*t + 2)
                      / ((t+1)^2 - 2)^2 );
     check_equal(wdash(n,t)          + 0*t,
                 w(n,t)',
                 "wdash 4");

     my(wdash=(n,t)->  ( (2*n*t + t +1)*(t+1)^(2*n-1) - 2^n )
                       / ((t+1)^2 - 2)
                     - (t * ((t+1)^(2*n) - 2^n)) * (2*t + 2)
                       / ((t+1)^2 - 2)^2 );
     \\ print(w(n,t)"  "w(n,t)'"  "wdash(n,t));
     check_equal(wdash(n,t)          + 0*t,
                 w(n,t)',
                 "wdash 5");

     \\ (2*n*t+T)*(T^2 - 2) + 2*t*T^2
     my(wdash=(n,t)-> ( 2*n*t       *(t+1)^(2*n-1) * ((t+1)^2 - 2)
                      + (t+1)       *(t+1)^(2*n-1) * ((t+1)^2 - 2)
                      -              (t+1)^(2*n-1) * ((t+1)^2    )*2*t
                      + 2^n*(t^2+1)
                      )
                      / ((t+1)^2 - 2)^2 );
     check_equal(wdash(n,t)          + 0*t,
                 w(n,t)',
                 "wdash 6");

     \\ 2*n*t + (t+1) - 2*t
     \\  ((t+1)^2 - 2)*((2*n-1)*t+1) - (          2)*2*t
     my(wdash=(n,t)-> ( (t+1)^(2*n+1) *   ( 2*n*t - t + 1)
                      - (t+1)^(2*n-1) * 2*( 2*n*t + t + 1)
                      + 2^n*(t^2+1)
                      )
                      / ((t+1)^2 - 2)^2 );
     check_equal(wdash(n,t)          + 0*t,
                 w(n,t)',
                 "wdash 7");

     check_equal(vpar_Wiener_index(vpar),
                 (n-1)*2^(2*n-1) + 2^(n-1),
                 "binomial tree vpar_Wiener_index() vs formula");

     \\ This form looks closest to Iyer and Reddy (n-1)*2^(2*n-1) - 2^(n-1),
     \\ except there's a denominator which at t=1 is 4 and corresponding
     \\ factor in the numerator terms.  Not particularly attractive though.
     my(wdash=(n,t)-> ( (2*n*t*((t+1)^2-2) - (t+1)*(t^2+1)) * (t+1)^(2*n-1)
                      + (t^2+1) * 2^n
                      )
                      / ((t+1)^2-2)^2 );
     check_equal(wdash(n,t)          + 0*t,
                 w(n,t)',
                 "wdash 8");
     );
}

\\ A003059 n appears 2*n-1 times, is ceil(sqrt(n))
\\ for(n=2,6, for(j=1,2*n-1, print1(n",")))
test_A003059(n) = sqrtint(n-1)+1;
{
  check_equal(vector(#A003059_samples,n, test_A003059(n)),
              A003059_samples,   \\ OFFSET=1
              "test_A003059()");
}

\\ A135034 n appears 2*n-1 times, starting 0
test_A135034(n) = if(n==0,0, sqrtint(n-1)+1);
{
  check_equal(vector(#A135034_samples,n,n--; test_A135034(n)),
              A135034_samples,   \\ OFFSET=0
              "test_A135034()");
}

\\ A071797 runs 1 .. 2*n-1
\\ for(n=2,6, for(j=1,2*n-1, print1(j",")))
test_A071797(n) = n - (test_A003059(n)-1)^2;
{
  check_equal(vector(#A071797_samples,n, test_A071797(n)),
              A071797_samples,   \\ OFFSET=1
              "test_A071797()");
}

\\ A192020
test_A192020(n) =
{
  my(m=test_A003059(n+1),   \\ order m >= 1
     k=test_A071797(n+1));  \\ offset k >= 1
  sum(j=1,m, 2^(m-j)*binomial(2*j-2,k-1));
}
{
  check_equal(vector(#A192020_samples,n,n--; test_A192020(n)),
              A192020_samples,   \\ OFFSET=0
              "test_A192020() vs A192020_samples");
}
{
  check_equal(vector(500,n,n--;
                     my(s=sqrtint(n), k=n-s^2);
                     sum(i=0,s, 2^(s-i)*binomial(2*i,k));
                    ),
              vector(500,n,n--; test_A192020(n)),
              "A192020 by sqrt");
}

{
  \\ binomial tree total number of re-orderings of siblings

  \\ A052129  w[k] = k*w[k-1]^2  ways to permute siblings
  \\ also num indsets with root in the set
  check_equal(vector(#A052129_samples,n,n--;
                     my(vpar=vpar_make_binomial_tree(2^n),
                        sets=vpar_siblings_sets(vpar));
                     prod(i=1,#sets, (#sets[i])!)),
              A052129_samples,
              "A052129  binomial tree siblings permutations");

  \\ formula by Jonathan Sondow in A052129
  check_equal(vector(#A052129_samples,n,n--; prod(k=1,n, k^(2^(n-k)))),
              A052129_samples,
              "A052129 product");
}


\\-----------------------------------------------------------------------------
progress("vpar_unlabelled_first()");

{
  check_equal(vpar_unlabelled_first(0), []);
  check_equal(vpar_unlabelled_first(2), [0,1]);
  check_equal(vpar_unlabelled_first(5), [0,1,2,3,4]);
}
{
  for(n=0,10,
     my(vpar=vpar_unlabelled_first(n));
     vpar_validate(vpar);
     if(!forest,
        check_equal(vpar_is_tree(vpar), 1));
     check_equal(vpar_is_preorder(vpar), 1);
     check_equal(vpar_is_unlabelled(vpar), 1);
     check_equal(vpar_unlabelled_prev(vpar), 'none);
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_unlabelled_last()");

{
  check_equal(vpar_unlabelled_last(5,1), [0,0,0,0,0]);
  check_equal(iferr(vpar_unlabelled_last(5,'forest);0, e, 1),
              1,
  "vpar_unlabelled_last() error on forest=symbol (whereas it should be 0,1)");
}
{
  for(forest=0,1,
     for(n=0,10-forest,
        my(vpar=vpar_unlabelled_last(n,forest));
        vpar_validate(vpar);
        if(!forest,
           check_equal(vpar_is_tree(vpar), 1));
        check_equal(vpar_is_preorder(vpar), 1);
        check_equal(vpar_is_unlabelled(vpar), 1);
        check_equal(vpar_unlabelled_next(vpar,forest), 'none);
     ));
}


\\-----------------------------------------------------------------------------
progress("vpar_is_unlabelled()");

check_equal(vpar_is_unlabelled([]), 1);
check_equal(vpar_is_unlabelled([0]), 1);
check_equal(vpar_is_unlabelled([0,0]), 1);
check_equal(vpar_is_unlabelled([0,0,2]), 0);
check_equal(vpar_is_unlabelled([0,1,0]), 1);
check_equal(vpar_is_unlabelled([0,0,0]), 1, "vpar_is_unlabelled() 3 singletons");

{
  \\  1
  \\  | \
  \\  2  3    not lex descending
  \\     |
  \\     4
  my(vpar=[0,1,1,3]);
  check_equal(vpar_is_unlabelled(vpar), 0);
}
{
  \\  1
  \\  |\
  \\  2 4    not preorder
  \\  |\
  \\  3 5
  my(vpar=[0, 1, 2, 1, 2]);
  check_equal(vpar_is_unlabelled(vpar), 0);
}

test_is_unlabelled_by_relabel(vpar) = \
  ! vpar_is_cyclic(vpar) && vpar_to_unlabelled(vpar)==vpar;

test_is_unlabelled_by_seq(vpar) =
{
  ! vpar_is_cyclic(vpar)
  && vpar_is_preorder(vpar)
  && vpar_seq_premax(vpar) == vpar_INTERNAL_identity_perm(#vpar);
}

test_is_unlabelled_by_depths(vpar) =
{
  !vpar_is_cyclic(vpar) || return(0);
  vpar_is_preorder(vpar) || return(0);
  my(depths=vpar_depths_vector(vpar),
     subtree_sizes=vpar_subtree_sizes_vector(vpar),
     siblings_sets=vpar_siblings_sets(vpar));
  for(i=1,#siblings_sets,
     for(j=2,#siblings_sets[i],
        my(u=siblings_sets[i][j-1],
           v=siblings_sets[i][j]);
        lex(depths[u..u+subtree_sizes[u]-1],
            depths[v..v+subtree_sizes[v]-1]) >= 0 || return(0)));
  1;
}

{
  my(samples=concat([[[0, 1, 1, 2],
                      [0,1,1,3,0,5,6],
                      [0, 1, 2, 1, 4]
                     ],
                     test_sample_forests,
                     test_sample_preorder_forests,
                     test_sample_cyclics
                    ]));
  for(i=1,#samples,
     my(vpar=samples[i],
        got=vpar_is_unlabelled(vpar));
     \\ print(vpar" seq "Vec(vpar_seq_premax(vpar)));
     check_equal(got,
                 test_is_unlabelled_by_relabel(vpar),
                 "vpar_is_unlabelled() vs test_is_unlabelled_by_relabel() of sample forest");
     check_equal(got,
                 test_is_unlabelled_by_seq(vpar),
                 "vpar_is_unlabelled() vs test_is_unlabelled_by_seq() of sample forest");
     check_equal(got,
                 test_is_unlabelled_by_depths(vpar),
                 "vpar_is_unlabelled() vs test_is_unlabelled_by_depths() of sample forest"));
}

{
  my(forest=0);
  for(n=0, test_limit_trees_search - forest,
     my(count_unlabelled_trees=0);
     forvec(vec=vpar_forvec_from(n,forest), \\ forest
           my(vpar=vpar_from_upascend(vec),
              got=vpar_is_unlabelled(vpar));
           \\ print(vpar);
           check_equal(got, test_is_unlabelled_by_seq(vpar),
                       "vpar_is_unlabelled() vs test_is_unlabelled_by_seq()");
           count_unlabelled_trees += got;
     );

     check_equal(count_unlabelled_trees, vpar_count_unlabelled(n),
                 "vpar_is_unlabelled() count vs vpar_count_unlabelled()");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_seq_preorder()");

check_equal(vpar_seq_preorder([]), Vecsmall([]), \
            "vpar_seq_preorder() empty");
check_equal(vpar_seq_preorder([0,0]), Vecsmall([1,2]), \
            "vpar_seq_preorder() 2 singletons");
check_equal(vpar_seq_preorder([0,0],1), Vecsmall([2,1]));
check_equal(vpar_seq_preorder([0,1,1]), Vecsmall([1,2,3]));
check_equal(vpar_seq_preorder([0,1,1],1), Vecsmall([1,3,2]));

\\  1    4
\\  |\   |
\\  2 3  5
check_equal(vpar_seq_preorder([0,1,1,0,4]),   Vecsmall([1,2,3,4,5]));
check_equal(vpar_seq_preorder([0,1,1,0,4],1), Vecsmall([4,5,1,3,2]));

{
  for(n=0,5,
     my(vpar=vpar_make_path(n));
     vpar_seq_preorder(vpar,0);
     vpar_seq_preorder(vpar,1);
     check_equal(iferr(vpar_seq_preorder(vpar,'down);0, e, 1),
                 1,
                 "vpar_seq_preorder() error on down=symbol (whereas it should be 0,1)"));
}

\\ Return a seq of the vertices at and below v in pre-order.
test_seq_preorder_subtree_by_recurse(vpar,v,rev) =
{
  my(seq=if(v,Vecsmall([v]),Vecsmall([])),
     children=vpar_children_of_vertex(vpar,v));
  if(rev, children=Vecrev(children));
  for(i=1,#children,
     seq=concat(seq, test_seq_preorder_subtree_by_recurse(vpar,children[i],rev)));
  seq;
}
\\ Return a seq of vpar vertices in pre-order.
test_seq_preorder_by_recurse(vpar,rev) = \
  test_seq_preorder_subtree_by_recurse(vpar,0,rev);

test_seq_preorder_by_sort(vpar,rev=0) =
{
  \\ keys are paths [root, ..., v] (which is noted in the
  \\ vpar_seq_downqueue() addhelp too).
  \\ Or for rev=1, keys are negatives [-root, ..., -v] to go biggest first.
  my(keys=vector(#vpar,v, Vecrev(vpar_ancestors(vpar,v))));
  if(rev, keys=-keys);
  vecsort(keys,,1);
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     for(rev=0,1,
        my(got=vpar_seq_preorder(Vecsmall(vpar), rev));
        check_equal(got, test_seq_preorder_by_recurse(vpar,rev),
                    Str("vpar_seq_preorder() vs test_seq_preorder_by_recurse() rev=",rev));
        check_equal(got, test_seq_preorder_by_sort(vpar,rev),
                    Str("vpar_seq_preorder() vs test_seq_preorder_by_sort() rev=",rev));
        check_equal(got, Vecsmall(Vecrev(vpar_seq_postorder(vpar,!rev))),
                    Str("vpar_seq_preorder() vs opposite vpar_seq_postorder() rev=",rev));
        check_equal(got,
                    Vecsmall(apply(v->#vpar+1-v,
                             Vec(vpar_seq_preorder(vpar_relabel_reverse(vpar), !rev)))),
                    "vpar_seq_preorder() vs reverse vpar_seq_postorder()");
        check_equal(vpar_seq_is_downwards(vpar,got), 1,
                    "vpar_seq_preorder() is vpar_seq_is_downwards() per addhelp");
        if(!rev,
           check_equal(vpar_seq_is_ordered(vpar,got), 1,
                       "vpar_seq_preorder() is vpar_seq_is_ordered() per addhelp"));
  ));
}


\\-----------------------------------------------------------------------------
progress("vpar_seq_premax()");

check_equal(vpar_seq_premax([]), Vecsmall([]), \
            "vpar_seq_premax() empty");
check_equal(vpar_seq_premax([0]), Vecsmall([1]), \
            "vpar_seq_premax() singleton");
check_equal(vpar_seq_premax([0,1]), Vecsmall([1,2]), \
            "vpar_seq_premax() path-2");
check_equal(vpar_seq_premax([0,1,2,3]), Vecsmall([1,2,3,4]), \
            "vpar_seq_premax() path-4");
check_equal(vpar_seq_premax([0,1,1,1]), Vecsmall([1, 2,3,4]), \
            "vpar_seq_premax() star-4");
check_equal(vpar_seq_premax([0,1,2,1,4,1,6]), \
            Vecsmall([1, 2,3, 4,5, 6,7]));
{
  \\ same vecextracts
  \\     1           1
  \\    / \         / \
  \\   2   3       2   4
  \\   |           |
  \\   4           3
  my(vpar1 = [0, 1, 1, 2],
     vpar2 = [0, 1, 2, 1],
     seq1 = vpar_seq_premax(vpar1),
     seq2 = vpar_seq_premax(vpar2));
  check_equal(seq1, Vecsmall([1,2,4,3]), "vpar_seq_premax() same seq1");
  check_equal(seq2, Vecsmall([1,2,3,4]), "vpar_seq_premax() same seq2");

  my(vec1 = vecextract(vpar1,seq1),
     vec2 = vecextract(vpar2,seq2));
  check_equal(vec1, vec2);
  check_equal(vec1, [0, 1, 2, 1]);
}
{
  \\  1 3
  \\    |
  \\    2
  my(vpar=[0, 3, 0]);
  vpar_validate(vpar);
  my(got_seq = vpar_seq_premax(vpar));
  check_equal(got_seq, Vecsmall([3,2,1]), "vpar_seq_premax() 3");
}
{
  \\         1
  \\       / | \
  \\      2  4  6
  \\      |  |  |
  \\      3  5  7
  \\            |
  \\            8
  my(vpar=[0,1,2,1,4,1,6,7]);
  vpar_validate(vpar);
  my(got_seq = vpar_seq_premax(vpar));
  check_equal(got_seq, Vecsmall([1, 6,7,8, 2,3, 4,5]));
}
{
  \\         1
  \\       /| |\
  \\      2 3 4 5
  \\      | | | |
  \\      6 7 8 9
  \\            |
  \\            10
  \\
  my(vpar=[0,1,1,1,1,2,3,4,5, 9]);
  vpar_validate(vpar);
  my(got_seq = vpar_seq_premax(vpar));
  check_equal(got_seq, Vecsmall([1,5,9,10, 2,6, 3,7, 4,8]));
}

\\ WRONG
test_seq_premax_by_sort(vpar) =
{
  my(children=vpar_children_vector(vpar),
     perm=vpar_INTERNAL_identity_perm(#vpar),
     id=vpar_premax_ids_vector(vpar));
  print("test_seq_premax_by_sort() "vpar);
  print(" ids "id);
  children=concat(children,[vpar_roots(vpar)]);
  for(i=1,#children,
     print(" children "children[i]);
     my(children_ids=-vecextract(id,children[i]),
        order=vecsort(children_ids,,1));
     print("  children_ids "children_ids);
     print("  order "order);
     for(j=1,#children[i],
        perm[children[i][j]] = children[i][order[j]]));
  print(" perm "Vec(perm));
  test_is_perm(perm) || error();

  vpar=vpar_relabel_perm(vpar,perm);
  print(" vpar relabelled "vpar);
  my(seq=vpar_seq_preorder(vpar));
  print(" preorder "Vec(seq));

  print("  times perm "perm*seq);
  print("  times perm "perm^-1*seq);
  \\ print("  times perm "perm^-1*seq*perm);
  \\ print("  times perm "perm*seq*perm^-1);
  print("  times perm "seq*perm);
  print("  times perm "seq*perm^-1);
  \\ print("  times perm "perm*seq^-1);
  \\ print("  times perm "perm^-1*seq^-1);
  \\ print("  times perm "perm^-1*seq^-1*perm);
  \\ print("  times perm "perm*seq^-1*perm^-1);
  \\ print("  times perm "seq^-1*perm);
  \\ print("  times perm "seq^-1*perm^-1);
  perm*seq;
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        seq=vpar_seq_premax(vpar));
     check_equal(test_is_perm(seq), 1, "vpar_seq_premax() is perm");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_to_unlabelled()");

test_relabel_premax_by_search(vpar) =
{
  \\ print("test_relabel_premax_by_search()");
  my(max_vpar=vpar_relabel_preorder(vpar),
     max_depths=vpar_depths_vector(max_vpar));
  \\ print(" initial depths "max_depths);
  for(i=1,(#vpar)!,
     my(perm=numtoperm(#vpar,i),
        this_vpar = vpar_relabel_perm(vpar,perm));
     vpar_is_preorder(this_vpar) || next();
     my(this_depths = vpar_depths_vector(this_vpar));
     if(lex(this_depths,max_depths) > 0,
        \\ print(" new high vpar="this_vpar" depths "this_depths" vs "max_depths);
        max_vpar = this_vpar;
        max_depths = this_depths));

  \\ print(" final return vpar="max_vpar" depths "max_depths);
  max_vpar;
}

{
  \\  2     3     6
  \\  |    / \
  \\  4   7   8
  \\  |       |
  \\  1       5
  my(vpar=[4, 0, 0, 2, 8, 0, 3, 3]);
  vpar_validate(vpar);

  my(want_vpar=test_relabel_premax_by_search(vpar),
     want_depths=vpar_depths_vector(want_vpar));
  check_equal(vpar_is_preorder(want_vpar), 1,
              "test_relabel_premax_by_search() 8 preorder");
  check_equal(want_vpar, [0,1,2,1, 0,5,6, 0],
              "test_relabel_premax_by_search() 8");
  check_equal(want_depths, [0,1,2,1, 0,1,2, 0]);

  my(got_vpar=vpar_to_unlabelled(vpar));
  check_equal(vpar_is_preorder(got_vpar), 1);
  check_equal(got_vpar, want_vpar,
             "vpar_to_unlabelled() 8");
}

{
  \\  1   2   4
  \\      |
  \\      3
  my(vpar=[0, 0, 2, 0]);
  vpar_validate(vpar);

  my(want_vpar=test_relabel_premax_by_search(vpar),
     want_depths=vpar_depths_vector(want_vpar));
  check_equal(vpar_is_preorder(want_vpar), 1,
              "test_relabel_premax_by_search() 4 preorder");
  check_equal(want_vpar, [0,1,0,0],
              "test_relabel_premax_by_search() 4");
  check_equal(want_depths, [0,1,0,0]);  \\ depth=1 soonest

  my(got_vpar=vpar_to_unlabelled(vpar));
  check_equal(vpar_is_preorder(got_vpar), 1);
  check_equal(got_vpar, [0,1,0,0],
             "vpar_to_unlabelled() 4");
}

{
  \\       2               1
  \\   /  / \  \       /  / \  \
  \\  3  7  10  13   14 18  21  23
  \\  |  |   |        |  |   |
  \\  4  8  11       15 19  22
  \\  |  |   |        |  |
  \\  5  9  12       16 20
  \\  |               |
  \\  6              17
  \\
  \\                  6      9       12  13
  my(vpar=[0,0, 2,3,4,5, 2,7,8, 2,10,11,  2,
           1,14,15,16, 1,18,19, 1,21,  1     ]);
  \\               17       20    22  23
  vpar_validate(vpar);

  my(got_seq=vpar_seq_premax(vpar));
  check_equal(got_seq,
              Vecsmall([2,3,4,5,6, 7,8,9, 10,11,12, 13,
                        1, 14,15,16,17, 18,19,20, 21,22, 23]),
              "vpar_seq_premax() 23");
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got_vpar=vpar_to_unlabelled(vpar));
     vpar_validate(got_vpar);
     check_equal(vpar_is_preorder(got_vpar), 1,
                 "vpar_to_unlabelled() is preorder");

     my(siblings=vpar_siblings_sets(got_vpar),
        subtree_heights=vpar_subtree_heights_vector(got_vpar));
     for(i=1,#siblings,
        check_equal(test_vector_is_decreasing(vecextract(subtree_heights,
                                                         siblings[i])),
                    1,
          "vpar_to_unlabelled() sibling subtree heights are decreasing");
        for(a=1,#siblings[i],
           for(b=a+1,#siblings[i],
              my(u=siblings[i][a], v=siblings[i][b]);
              check_equal(vpar_INTERNAL_preorder_subtree_cmp(got_vpar, u,v)>=0,
                          1,
                "vpar_to_unlabelled() sibling subtrees are lex decreasing"))));

     if(#vpar <= test_limit_perms_search,
        my(want_vpar=test_relabel_premax_by_search(vpar));
        check_equal(vpar_is_preorder(want_vpar), 1,
                    "test_relabel_premax_by_search() is preorder");

        my(want_depths=vpar_depths_vector(want_vpar),
           got_depths=vpar_depths_vector(got_vpar));
        \\ print("vpar  "vpar);
        \\ print("got_depths  "got_depths);
        \\ print("want_depths "want_depths);

        check_equal(got_depths,want_depths,
                    "vpar_to_unlabelled() vs perm search, depths");
        check_equal(got_vpar,want_vpar,
                    "vpar_to_unlabelled() vs perm search, vpars"));
  );
}

{
  \\ Gang Li and Frank Ruskey, "The Advantages of Forward Thinking in
  \\ Generating Rooted and Free Trees", SODA 99.
  \\ http://webhome.cs.uvic.ca/~ruskey/Publications/RootedFreeTree/RootedFreeTree.html
  \\ Figure 1 example rooted trees vpars.
  \\
  \\         1                       1
  \\      /  |   \                /  |  \
  \\    2    4     5            2    8   10
  \\    |         / \          / \   |
  \\    3        6   10       3   7  9
  \\            / \          / \
  \\           7   8        4   6
  \\               |        |
  \\               9        5
  my(vpar=[0,1,2,1,1,5,6,6,8,5]);
  vpar_validate(vpar);
  my(got=vpar_to_unlabelled(vpar));
  check_equal(got, [0,1,2,3,4,3,2,1,8,1]);
}

{
  for(forest=0,1,
     for(n=0, test_limit_trees_search - forest,
        my(unlabelled_list=Set(test_unlabelled_all(n,1)));
        forvec(vec=vpar_forvec_from(n,forest),
              my(vpar=vpar_from_upascend(vec),
                 vpar_relabelled=vpar_to_unlabelled(vpar));
              \\ print(vpar"  "vpar_relabelled);
              check_equal(setsearch(unlabelled_list,vpar_relabelled)!=0, 1,
                          "vpar_to_unlabelled() vs test_unlabelled_all()"));
     ));
}


\\-----------------------------------------------------------------------------
progress("vpar_concat_forests()");

check_equal(vpar_concat_forests([]), []);
check_equal(vpar_concat_forests([ [] ]), []);
check_equal(vpar_concat_forests([ [],[] ]), []);
check_equal(vpar_concat_forests([ [0],[0,1] ]), [0,0,2], \
           "vpar_concat_forests() two");

check_equal(vpar_concat_forests([ Vecsmall([]) ]), Vecsmall([]));
check_equal(vpar_concat_forests([ Vecsmall([0,1,2]) ]), Vecsmall([0,1,2]));
check_equal(vpar_concat_forests([ Vecsmall([0,1]), Vecsmall([0,1]) ]), \
            Vecsmall([0,1,0,3]));

test_make_random_cyclic(n) = vector(n,j, random(n+1));

test_concat_forests_by_pairs(vpars) =
{
  my(vpar=[]);
  for(i=1,#vpars,
     vpar=vpar_concat_forest(vpar,vpars[i]));
  vpar;
}
{
  for(c=0,5,
     for(rep=1,5,
        my(vpars=vector(c,i, test_make_random_cyclic(random(5))));
        check_equal(vpar_concat_forests(vpars),
                    test_concat_forests_by_pairs(vpars),
                    "vpar_concat_forests() vs test_concat_forests_by_pairs()")));
}


\\-----------------------------------------------------------------------------
progress("vpar_is_preorder()");
\\ vpar_is_postorder()

check_equal(vpar_is_preorder([]), 1, "vpar_is_preorder() empty");
check_equal(vpar_is_postorder([]), 1);
check_equal(vpar_is_preorder([0]), 1);
check_equal(vpar_is_postorder([0]), 1);
check_equal(vpar_is_preorder([0,0,0]), 1, "vpar_is_preorder() forest");
check_equal(vpar_is_postorder([0,0,0]), 1);

\\     1 2 3
\\     |
\\     4
check_equal(vpar_is_preorder([0,0,0,1]), 0);
check_equal(vpar_is_postorder([0,0,0,1]), 0);

check_equal(vpar_is_preorder([0,0]), 1);
check_equal(vpar_is_preorder([0,1]), 1);
check_equal(vpar_is_preorder([2,0]), 0);
check_equal(vpar_is_postorder([0,0]), 1);
check_equal(vpar_is_postorder([0,1]), 0);
check_equal(vpar_is_postorder([2,0]), 1);

test_is_preorder_by_seq(vpar) = \
  vpar_seq_preorder(vpar) == vpar_INTERNAL_identity_perm(#vpar);
test_is_postorder_by_seq(vpar) = \
  vpar_seq_postorder(vpar) == vpar_INTERNAL_identity_perm(#vpar);

\\ per addhelp string of vpar_is_preorder(), each v must have parent v-1
\\ or ancestor of v-1
test_is_preorder_by_ancestor(vpar) =
{
  if(vpar,
     vpar[1]==0 || return(0);
     for(v=2,#vpar,
        vpar[v]==0 || vpar_is_ancestor(vpar,v-1,vpar[v]) || return(0)));
  1;
}
\\ 1 --> 2
check_equal(vpar_is_ancestor([0,1,1],1,1), 1);
check_equal(test_is_preorder_by_ancestor([2,0]), 0, \
            "test_is_preorder_by_ancestor() pre-order");
check_equal(test_is_preorder_by_ancestor([0,1,1]), 1, \
            "test_is_preorder_by_ancestor() pre-order, star-3");

\\ per addhelp string of vpar_is_postorder(), each v must have parent v+1
\\ or ancestor of v+1
test_is_postorder_by_ancestor(vpar) =
{
  if(vpar,
     vpar[#vpar]==0 || return(0);
     for(v=1,#vpar-1,
        vpar[v]==0 || vpar_is_ancestor(vpar,v+1,vpar[v]) || return(0)));
  1;
}
\\ 1 --> 2
check_equal(test_is_postorder_by_ancestor([2,0]), 1, \
            "test_is_postorder_by_ancestor() post-order");

\\ For post=0 return 1 if vpar is labelled in pre-order.
\\ For post=1 return 1 if varp is labelled in post-order.
test_is_porder_by_old_code(vpar,post) =
{
  \\ depths[v] is the depth of vertex v.
  \\ Depths d are an index into ancestors[] vector so start root depths[r]=1.
  \\ depths[v]=1 if not yet determined too.
  \\ ancestors[d] is the vertex number which is most recently seen at depth d.
  \\
  \\ prev_depth is the depth of vertex v-1.  The parent p of v must have
  \\ depth <= prev_depth.
  \\
  \\ For pre-order, each v must have its parent p as the most recently seen
  \\ vertex at the parent depth d=depths[p], ie. must ancestors[d]==p.
  \\ If p has not yet been seen, ie. p>v, then it is not pre-order.  This is
  \\ detected by depths[p]=1 for as yet unseen so fail ancestors[d]==p.
  \\ When p is ok set depths[v]=d+1 and recent at d+1 so ancestors[d+1]=v.
  \\
  \\ For post-order, the same but working end to start in vpar, so that an
  \\ unseen p is p<v so not post-order.

  my(depths    = vectorsmall(#vpar,v,1),
     ancestors = vectorsmall(#vpar),
     prev_depth=1, p);
  forstep(v=if(post,#vpar,1), if(post,1,#vpar), if(post,-1,1),
     my(d=1);
     if((p=vpar[v]),
        d=depths[p];
        if(d>prev_depth || ancestors[d]!=p,
           \\ print("  p="p" depth "depths[p]" "depths);
           return(0));
        d++);
     prev_depth=depths[v]=d;
     ancestors[d]=v);
  1;
}

test_is_porder_by_old_code2(vpar,post) =
{
  \\ Preorder is tested working down #vpar to 1 and asking that parents are
  \\ in certain ranges.  Must have vpar[v] < v always.  That is a downwards
  \\ tree (vpar_is_downwards()), of which pre-orders are a subset.
  \\ Then the pre-order condition can be illustrated as follows,
  \\
  \\      1    2
  \\           | \  \
  \\           3   7   8
  \\           |\
  \\           4 6
  \\           |
  \\           5
  \\
  \\ v=8 has parent vpar[8]=2.  This means vertices 3..7 in between must
  \\ have parents >=2.  In the code, m[] is a stack and the bound 2 is
  \\ pushed onto it.
  \\
  \\ v=7 has parent vpar[7]==2.  This is >=2 bound.  It is not pushed onto
  \\ m[] because the top m[d] is already equal to 2.
  \\
  \\ v=6 has parent vpar[6]=3.  This is pushed onto m[] likewise as a bound
  \\ for the 4..5 in between.
  \\
  \\ v=5 pushes bound 4 similarly.  At v=4 it is noticed v==m[d], so the
  \\ top-most bound has been reached, so pop it.  That leaves m[d]==3 to
  \\ compare against vpar[4]=3 which is ok.
  \\
  \\ m[1]=0 initial bound accepts all vpar[r]=0 roots, including v=1 which
  \\ is always a root vpar[1]=0.
  \\
  \\ Each push onto m[] narrows the range by at least 2, being the v and the
  \\ new bigger m.  So at most floor(#vpar/2) pushes.  m[] is sized for that
  \\ plus initial m[1]=0.  Actual space used in m[] will often be less.
  \\
  \\ For post-order, the loop goes up 1 to #vpar instead and comparisons are
  \\ reversed to v < vpar[v] <= m[d].
  \\ Roots parent vpar[r]=0 are changed to p=#vpar+1 so bigger than v.
  \\ m[1] is set to that so as to accept those roots.


  \\ print("test_is_porder_by_old_code2() post="post" "vpar);
  my(m=vectorsmall((#vpar >> 1) + 1),
     d=1,                   \\ stack top
     step=if(post,1,-1));   \\ preorder=-1, postorder=1
  m[1]=if(post,#vpar+1);    \\ sentinel postorder=0, preorder=#vpar+1

  \\ loop preorder #vpar .. 1, postorder 1 .. #vpar
  forstep(v=if(post,1,#vpar), if(post,#vpar,1), step,
     \\ print(" at v="v" p="vpar[v]"  d="d" m="Vec(m)[1..d]);

     if(v == m[d],
        \\ print("  pop m==v");
        d--);
     my(p=vpar[v]);
     p || (p=m[1]);

     \\ preorder p>=v or p<m[d] are bad
     if(lex(p,v) != step || lex(p,m[d]) == step,
        \\ print("  out of range");
        return(0));

     if(m[d] != p,
        \\ print("  push m "vpar[v]);
        m[d++]=vpar[v]));
  1;
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got_is_preorder=vpar_is_preorder(vpar),
        got_is_postorder=vpar_is_postorder(vpar));

     check_equal(got_is_preorder,
                 test_is_preorder_by_seq(vpar),
                 "vpar_is_preorder() vs test_is_preorder_by_seq() on samples");
     check_equal(got_is_preorder,
                 test_is_preorder_by_ancestor(vpar),
                 "vpar_is_preorder() vs test_is_preorder_by_ancestor() on samples");
     check_equal(got_is_preorder,
                 test_is_porder_by_old_code(vpar,0),
                 "vpar_is_preorder() vs test_is_porder_by_old_code() on samples");
     check_equal(got_is_preorder,
                 test_is_porder_by_old_code2(vpar,0),
                 "vpar_is_preorder() vs test_is_porder_by_old_code2() on samples");

     check_equal(got_is_postorder,
                 test_is_postorder_by_seq(vpar),
                 "vpar_is_postorder() vs test_is_postorder_by_seq() on samples");
     check_equal(got_is_postorder,
                 test_is_postorder_by_ancestor(vpar),
                 "vpar_is_postorder() vs test_is_postorder_by_ancestor() on samples");
     check_equal(got_is_postorder,
                 test_is_porder_by_old_code(vpar,1),
                 "vpar_is_postorder() vs test_is_porder_by_old_code() on samples");
     check_equal(got_is_postorder,
                 test_is_porder_by_old_code2(vpar,1),
                 "vpar_is_postorder() vs test_is_porder_by_old_code2() on samples");
  );
}
{
  \\ result on cyclics not guaranteed, but not an infinite loop
  for(i=1,#test_sample_cyclics,
     my(vpar=test_sample_cyclics[i]);
     vpar_is_preorder(vpar);
     vpar_is_postorder(vpar);
  );
}

{
  \\ per addhelp of vpar_is_preorder() and vpar_is_postorder(), number of
  \\ pre/post trees/forests is Catalan numbers
  \\
  my(forest=1);
  for(n=0, test_limit_trees_search - forest,
     my(count_pre_trees=0,
        count_pre_forests=0,
        count_post_trees=0,
        count_post_forests=0);
     forvec(vec=vpar_forvec_from(n,forest), \\ forest
           my(vpar=vpar_from_upascend(vec));

           my(pre=vpar_is_preorder(vpar));
           check_equal(pre, test_is_preorder_by_seq(vpar),
                       "vpar_is_preorder() vs test_is_preorder_by_seq()");
           check_equal(pre, test_is_preorder_by_ancestor(vpar),
                       "vpar_is_preorder() vs test_is_preorder_by_ancestor()");
           count_pre_forests += pre;
           count_pre_trees   += (pre && vpar_is_tree(vpar));

           my(post=vpar_is_postorder(vpar));
           check_equal(post, test_is_postorder_by_seq(vpar),
                       "vpar_is_postorder() vs test_is_postorder_by_seq()");
           check_equal(post, test_is_postorder_by_ancestor(vpar),
                       "vpar_is_postorder() vs test_is_postorder_by_ancestor()");
           count_post_forests += post;
           count_post_trees   += (post && vpar_is_tree(vpar));
     );

     my(want_count_forests = vpar_INTERNAL_Catalan_number(n),
        want_count_trees   = if(n==0,1,vpar_INTERNAL_Catalan_number(n-1)));
     check_equal(count_pre_forests,  want_count_forests);
     check_equal(count_post_forests, want_count_forests);
     check_equal(count_pre_trees,   want_count_trees);
     check_equal(count_post_trees,  want_count_trees);
     check_equal(want_count_trees,   vpar_count_ordered(n));
     check_equal(want_count_forests, vpar_count_ordered(n,1));
  );
}


\\-----------------------------------------------------------------------------
\\ A172447 and A075528 (relevant to A084089)

A002066(n) = 10*4^n;
A072197(n) = (10*4^n - 1)/3;
A172447(n) = (-1 + 10*4^n - 3*n)/9;
{
  check_equal(vector(#A002066_samples,n,n--; A002066(n)),
              A002066_samples, \\ OFFSET=0
              "A002066() vs A002066_samples");
  check_equal(vector(#A172447_samples,n,n--; A172447(n)),
              A172447_samples, \\ OFFSET=0
              "A172447() vs A172447_samples");
  check_equal(vector(#A072197_samples,n,n--; A072197(n)),
              A072197_samples, \\ OFFSET=0
              "A072197() vs A072197_samples");

  check_equal(vector(100,n,n--; A002066(n)),
              Vec(Ser( 10/(1-4*x) ,,100)),
              "A002066 gf");
  check_equal(vector(100,n,n--; A072197(n)),
              Vec(Ser( (3-2*x) / ((1-x)*(1-4*x)) ,,100)),
              "A072197 gf");
  check_equal(vector(100,n,n--; A172447(n)),
              Vec(Ser( (1 - 2*x + 2*x^2)/((1-4*x) * (x-1)^2 ) ,,100)),
              "A172447 gf");

  check_equal(1/x * (3-2*x)           % (1-4*x), Pol(10));
  check_equal(1/x * (1 - 2*x + 2*x^2) % ((1-4*x)*(x-1)), 3-2*x);

  \\ A002066 = diffs of A072197
  check_equal(vector(10,n,n--; A002066(n)),
              vector(10,n,n--; A072197(n+1) - A072197(n)),
              "A002066 = diffs of A072197");

  \\ A072197 = diffs of A172447
  check_equal(vector(10,n,n--; A072197(n)),
              vector(10,n,n--; A172447(n+1) - A172447(n)),
              "A072197 = diffs of A172447");
}


\\-----------------------------------------------------------------------------
progress("perfect matched unlabelled and free");

\\ Rooted unlabelled perfected matched trees of 2*n vertices.
\\ forest=1 is for a forest of 2*n vertices.
count_unlabelled_perfect_vector(len,forest=0) =
{
  my(ret=vector(len,n,1), s=vector(len));
  for(n=1,len-1,
     s[n] = 2*sumdiv(n,d, d*ret[d]);
     ret[n+1] = sum(i=1,n, ret[i]*s[n+1-i]) /n);   \\ next term EulerT(2*prevs)
  if(forest,
     vpar_INTERNAL_Euler_transform(ret),
     ret);

  \\ i=1..n-1  n-i=n-1..1
  \\ for(n=2,len,
  \\    ret[n] = sum(i=1,n-1, ret[i]*sumdiv(n-i,d, d*ret[d])) *2/(n-1));
}
{
  check_equal(count_unlabelled_perfect_vector(#A000151_samples),
              A000151_samples,
              "count_unlabelled_perfect_vector() vs A000151 samples");

  \\ Euler transform of 2* itself
  my(R=count_unlabelled_perfect_vector(50));
  check_equal(R,
              concat(1,vpar_INTERNAL_Euler_transform(2*R[^#R])),
              "count_unlabelled_perfect_vector() vs Euler 2*");
}

\\ rooted planted unlabelled perfected matched of 2*n vertices
\\ (root has a single child)
count_unlabelled_planted_perfect_vector(len,forest=0) =
{
  my(ret=concat([1],count_unlabelled_perfect_vector(len-1,1)));
  if(forest,
     vpar_INTERNAL_Euler_transform(ret),
     ret);
}
{
  check_equal(count_unlabelled_planted_perfect_vector(#A005750_samples),
              A005750_samples,
              "count_unlabelled_planted_perfect_vector() vs A005750 samples");

  \\ Euler transform of self squared
  \\ squared gives r, the transform to forest for p
  my(p=count_unlabelled_planted_perfect_vector(50),
     t=vector(#p-1,i, sum(j=1,i, p[j]*p[i-j+1])));
  check_equal(vpar_INTERNAL_Euler_transform(t),
              p[^1]);

  \\ forstep(n=1,#p,2
  \\    print(p[(n+1)/2],
  \\             x + x*gcount_unlabelled_perfect(5,1),
  \\             "gcount_unlabelled_planted_perfect() as forest");
}

gcount_unlabelled_perfect(limit,forest=0) =
{
  x*Ser(count_unlabelled_perfect_vector(limit-1,forest));
}
{
  check_equal(count_unlabelled_perfect_vector(5),
              [1, 2, 7, 26, 107]);
  check_equal(gcount_unlabelled_perfect(6),
              1*x + 2*x^2 + 7*x^3 + 26*x^4 + 107*x^5 + O(x^6),
              "gcount_unlabelled_perfect()");
}

gcount_unlabelled_planted_perfect(limit,forest=0) = \
  x*Ser(count_unlabelled_planted_perfect_vector(limit-1,forest));
{
  check_equal(count_unlabelled_planted_perfect_vector(5),
              [1, 1, 3, 10, 39]);
  check_equal(gcount_unlabelled_planted_perfect(6),
              1*x + 1*x^2 + 3*x^3 + 10*x^4 + 39*x^5 + O(x^6),
              "gcount_unlabelled_planted_perfect()");

  check_equal(gcount_unlabelled_planted_perfect(6),
              x + x*gcount_unlabelled_perfect(5,1),
              "gcount_unlabelled_planted_perfect() as forest");
}
{
  \\ Rodica Simion theorem 1.1 (a)
  my(limit=20,
     L=O(x^limit),
     G=gcount_unlabelled_perfect(limit),
     g=(x)->subst(G,'x,x));
  check_equal(x*exp(2*sum(i=1,limit, 1/i * g(x^i + L))) + L,
              g(x),
              "grooted = exp ...");
}
{
  \\ Rodica Simion theorem 1.1 (b)
  \\ P(x)^2 = x*R(x)
  check_equal(gcount_unlabelled_planted_perfect(20)^2,
              x*gcount_unlabelled_perfect(20),
              "gplanted^2 = x*grooted");

  \\ inverse as noted Andrew Howroyd in A005750
  \\ P(x) = sqrt(x*R(x))
  check_equal(gcount_unlabelled_planted_perfect(20),
              sqrt(x*gcount_unlabelled_perfect(20)),
              "gplanted = sqrt(x*grooted)");
}


count_free_perfect_vector(len,forest=0) =
{
  my(r=count_unlabelled_perfect_vector(len),
     p=concat(1,vpar_INTERNAL_Euler_transform(r)),
     ret=vector(len,n, (r[n]
                        - sum(m=1,n-1, r[m]*r[n-m])
                        + if(n%2==0, r[n/2], p[(n+1)/2])) / 2));
  if(forest,
     vpar_INTERNAL_Euler_transform(ret),
     ret);
}
{
  check_equal(count_free_perfect_vector(#A005751_samples),
              A005751_samples,
              "count_free_perfect_vector() vs A005751 samples");
}

gcount_free_perfect(limit,forest=0) = \
  x*Ser(count_free_perfect_vector(limit-1,forest));
{
  \\ Rodica Simion theorem 1.1 (c)
  my(limit=20,
     L=O(x^limit),
     R=gcount_unlabelled_perfect(limit),
     RF=gcount_unlabelled_perfect(limit,1),
     P=gcount_unlabelled_planted_perfect(limit),
     M=gcount_free_perfect(limit),
     r=(x)->subst(R,'x,x),
     rf=(x)->subst(RF,'x,x),
     p=(x)->subst(P,'x,x));
  check_equal(1/2*( r(x) - r(x)^2 + r(x^2) + 1/x*p(x^2)),
              M,
              "gfree = r and p");

  check_equal(P, x*(1+RF) + L);
  check_equal(p(x), x*(1+rf(x)) + L);
  check_equal(p(x^2) + L, x^2*(1+rf(x^2)) + L);
  check_equal(1/2*( r(x) - r(x)^2 + r(x^2) + 1/x*x^2*(1+rf(x^2))),
              M,
              "gfree = r and p");

  \\ print(Vec(rf(x)^2));
  \\ check_equal(r(x), x*rf(x)^2);

  check_equal(1/2*( r(x)*(1 - r(x))
                   + r(x^2)             \\ evens
                   + x + x*rf(x^2)),    \\ odds
              M,
              "gfree = r and p");
}

count_free_perfect_vector_compact(len) =
{
  my(r=vector(len),s=r); r[1]=1;
  for(n=1,len-1, s[n] = sumdiv(n,d, d*r[d]);
                 r[n+1] = sum(k=1,n, r[k]*s[n+1-k])*2/n);
  my(p=vector((len+1)\2)); p[1]=1;
  for(n=1,#p-1, p[n+1] = sum(k=1,n, s[k]*p[n+1-k])/n);
  vector(len,n,
         (r[n] + if(n%2, p[(n+1)/2], r[n/2] - r[n/2]^2)) /2
         - sum(k=1,(n-1)\2, r[k]*r[n-k])
        );

  \\ A000238 oriented trees n vertices
  \\ vector(len,n, r[n] - sum(k=1,n-1, r[k]*r[n-k]))

  \\ A198760 non-planted
  \\ vector(#p,n, r[n] - p[n])

  \\ my(p=vpar_INTERNAL_Euler_transform(r));
  \\ print("p="p);
  \\ print("b="b);
  \\ print("c="c);
  \\ print("s="s);
}

\\ count_unlabelled_perfect_vector(20)
\\ count_unlabelled_planted_perfect_vector(20)
\\ count_free_perfect_vector_compact(12)
\\ count_free_perfect_vector(12)

{
  for(len=3,20,
     check_equal(count_free_perfect_vector_compact(len),
                 count_free_perfect_vector(len)));
}


\\-----------------------------------------------------------------------------
progress("vpar_domsets()");

check_equal(vpar_domsets([]), [ [] ]);

{
  check_equal(vecsort(vpar_domsets([2,0,2])),
              [[1, 2], [1, 2, 3], [1, 3], [2], [2, 3]],
              "[2,0,2]");
}
{
  check_equal(vecsort(vpar_domsets([0,1,2])),
              [[1, 2], [1, 2, 3], [1, 3], [2], [2, 3]],
              "[0,1,2]");
}
check_equal(vpar_domsets([0,1]), [[1, 2], [1], [2]]);

test_domsets_by_flags(vpar) =
{
  my(l=List([]));
  forvec(flags=vector(#vpar,i,[0,1]),
        if(vpar_is_domset_flags(vpar,flags),
           listput(l,test_flags_to_Set(flags))));
  Vec(l);
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        want_count=vpar_domsets_count(vpar));
     if(want_count > test_limit_sets, next());

     my(got=vpar_domsets(vpar));
     check_equal(#got, want_count, "vpar_domsets() vs vpar_domsets_count()");
     for(i=1,#got,
        check_equal(setisset(got[i]), 1);
        check_equal(vpar_is_domset(vpar,got[i]), 1));

     check_equal(vecsort(got),
                 vecsort(test_domsets_by_flags(vpar)),
                 "vpar_domsets() vs test_domsets_by_flags()");

     check_equal(Pol(0,'y) + sum(i=1,#got, 'y^#got[i]),
                 vpar_dompoly(vpar,'y),
                 "vpar_domsets() size counts vs vpar_dompoly()");
  );
}


\\-----------------------------------------------------------------------------
progress("A079559 0,1 balanced binary of infinite binomial tree, and childless postorder");
\\
\\ 1, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 1, 0, 0, 0, 1, 1, 0, 1, 1, 0, 0, 1, 1
\\ OFFSET=0

\\ return a vector of first len many terms of A079559
A079559_vector(len) =
{
  my(v=[1]);
  while(#v<len, v=concat([v,v,0]));
  Vec(v,len);
}

\\ return gf ... + O(x^limit)
gA079559(limit) = Ser(A079559_vector(limit));

{
  check_equal(A079559_vector(#A079559_samples),
              A079559_samples,
              "A079559_vector() vs A079559_samples");

  \\ A079559 = balanced binary bits of big enough binomial tree
  my(len=1024,
     vpar=vpar_make_binomial_tree(len),
     b=vpar_to_balanced_binary(vpar),
     v=binary(b));
  check_equal(A079559_vector(len),
              v[1..len],
              "A079559_vector() vs balanced binary of binomial tree");

  my(is_childless=vpar_is_childless_vector(vpar));
  check_equal(is_childless,
              vector(#is_childless,n,n--; n%2),
              "binomial tree odd vertices are childless");

  my(seq=vpar_seq_postorder(vpar));
  is_childless=vecextract(is_childless,seq);  \\ in seq order
  seq=vector(#seq,i,seq[i]-1);    \\ 0-based vertex numbering, not 1-based

  check_equal(seq[#seq],                   0, "final root vertex");
  check_equal(is_childless[#is_childless], 0, "final root not childless");
  seq=seq[^len];                              \\ discard root
  is_childless=is_childless[^#is_childless];  \\ discard root

  check_equal(is_childless,
              A079559_vector(#is_childless),
              "binomial tree postorder childless vs A079559");
}

\\ A079559() isolated term by some bitwise.
\\
\\ Let W(k) = terms 0 to 2^k-2 inclusive, so length 2^k-1.
\\ Top-down replication is  W(k+1) = W(k),W(k),0 starting W(0)=1
\\ Last of W(k) is 0 for k>=1.
\\ So can extend back the second 2^k-1 block to its preceding 0
\\ to make a 2^k length block
\\       A079559(2^k-2 + r)      for 0 <= r < 2^k
\\     = A079559(   -2 + r+1)    and k>=2
\\ So +1 if bit k of r is 1.
\\
\\ If r < 2^k-1 then the new
\\    -2 + r+1 < 2^k-2 = 2^(k-1)-2 + 2^(k-1)
\\ ready to repeat a the reduction with k-1.
\\
\\ If r = 2^k-1 then that would reduce to -2+r+1 = 2^k-2 which would want a
\\ second A079559(2^k-2 + r) reduction, though this time with r=0.  The
\\ result wanted for r = 2^k-1 is
\\
\\     A079559(2^k-2 + 2^k-1) = A079559(2^(k+1)-3) = 0        k>=1
\\
\\ since this n = 2^(k+1)-3 is the second last in W(k+1) (it runs n=0 to
\\ n=2^(k+1)-2 inclusive) and that is the last of the second W(k) which is 0
\\ for k>=1.
\\
\\ It suffices to ignore the 2^k bit even if r=2^k-1 since a second
\\ reduction if done would drop it to r=0, and that is already what the bits
\\ without 2^k are.
\\
\\ After k=2, have 0 <= r <= 3 and its values are to be
\\
\\             r  =  0  1  2  3
\\             n  = -1  0  1  2
\\     A079559(n) =  0  1  1  0       understanding A079559(-1)=0
\\
\\ This A079559(-1)=0 understood is where the r=0 finishes.
\\ In any case second lowest bit of r is A079559(n).
\\
\\ Cf David Wasserman's formula (author) in A100661 (which is a variation of
\\ partial sums of A079559) recurrence n -> n-2^k+1.
\\   a(2^k-1) = 1.  For 2^k <= n <= 2^(k+1)-2,  a(n) = a(n-2^k+1)+1.
\\ ~/OEIS/a100661.txt
\\
\\ Cf Reinhard Zumkeller's formula in A079559 subtracting, sort of ?
\\   a(n) = p(n,1) with  p(n,k) = p(n-k,2*k+1) + p(n,2*k+1) if k <= n, otherwise 0^n

A079559(n) =
{
  n>=0 || error("A079559() is for n>=0");
  n+=2; forstep(k=logint(n,2),2,-1, n+=bittest(n,k)); bittest(n,1);
}

{
  check_equal(A079559(0), 1, "A079559(0)");
  check_equal(A079559(1), 1, "A079559(1)");
  check_equal(A079559(2), 0, "A079559(2)");
  check_equal(vector(2^14,n,n--; A079559(n)),
              A079559_vector(2^14),
              "A079559() vs A079559_vector()");

  check_equal(vector(#A079559_samples,n,n--; A079559(n)),
              A079559_samples,  \\ OFFSET=0
              "A079559() vs A079559_samples");

  \\ A079559 = binomial tree postorder vertices mod 2
  check_equal(vector(1024,n, test_negidx2lexrev(n) % 2),
              vector(1024,n, A079559(n-1)),
              "A079559(n-1) = test_negidx2lexrev() A108918 mod 2");

  check_equal(A079559_vector(1024),
              vector(1024,n, test_negidx2lexrev(n) % 2),
              "A079559_vector() vs test_negidx2lexrev() A108918 mod 2");

  check_equal(gA079559(256),
              sum(n=0,255, A079559(n)*x^n) + O(x^256),
              "gA079559() vs A079559()");
}


\\ Jackson and Ruskey, "Meta-Fibonacci Sequences, Binary Trees And
\\ Extremal Compact Codes", Electronic Journal of Combinatorics, volume
\\ 13, 2006, R26.  In proof of Corollary 2.3.
\\
\\ they start d(1) first term, and define preceding d(0)=0
A079559_d0(n) =
{
  n>=0 || error("A079559_d0() is for n>=0");
  if(n==0,0,
     A079559(n-1));
}
\\ allowing also n==-1
A079559_with_neg1(n) = if(n==-1,0, A079559(n));
{
  \\ 1 <= k <= 2^h-1
  check_equal(vector(7,h,h--; vector(2^h-1,k, A079559_d0(2^h-1 + k))),
              vector(7,h,h--; vector(2^h-1,k, A079559_d0(        k))),
              "A079559 Jackson and Ruskey d0 and 2^h-1");

  \\ also holds at k=0 since d(0)=0, for h>=2 at least
  check_equal(A079559_d0(0), 0);
  check_equal(vector(8,h,h++; my(k=0); A079559_d0(2^h-1 + k)),
              vector(8,h,h++; my(k=0); A079559_d0(        k)),
              "A079559 Jackson and Ruskey d0 and 2^h-1 at k=0");

  \\ in offset of A079559, -1 <= r < 2^k-1
  check_equal(vector(8,h,h++; vector(2^h,r,r-=2; A079559_with_neg1(2^h-1 + r))),
              vector(8,h,h++; vector(2^h,r,r-=2; A079559_with_neg1(        r))),
              "A079559 like Jackson and Ruskey");
}

{
  \\ isolated term is by string
  \\ W(k+1) = W(k),W(k),0
  \\ Each W is length h = 2^k-1.  The logint finds h where
  \\    W(k),W(k),0     h <= n <= 2*h
  \\         ^--h
  \\ So n is in the second W(k).  Subtract h to bring it down to the first.
  \\ If n==2*h then it is the 0 after and can return immediately.

  \\ subtracting h = 2^k-1 found by logint for where high bit of n+1
  my(a = (n)->
     while(n, my(h=2^logint(n+1,2)-1); n-=h; if(n==h,return(0))); 1);
  check_equal(vector(2^12,n,n--; a(n)),
              vector(2^12,n,n--; A079559(n)),
              "h = 2^k-1 by logint vs A079559()");
}
{
  \\ subtracting h = 2^k-1 where possible shifting down initial 111...111
  my(a = (n)->
     my(h=2^logint(n+1,2)-1);
     while(h, if(n>=h, n-=h; if(n==h,return(0))); h>>=1); 1);
  check_equal(vector(2^12,n,n--; a(n)),
              vector(2^12,n,n--; A079559(n)),
              "h = 2^k-1 by shifting vs A079559()");
}
{
  \\ subtracting h = 2^k-1 where possible shifting down initial 111...111,
  \\ with checks against the conditions on m=n+1 and p=h+1=2^k
  my(a = (n)->
     my(k=logint(n+1,2), h=2^k-1);
     while(h,
           my(m=n+1, p=h+1);
           h == 2^k-1 || error();
           p == 2^k   || error();
           (n>=h)== (m>=p) || error();
           (n>=h) == bittest(m,k) || error();
           if(n>=h,
              n-=h;
              m = m+1 - p;
              m==n+1 || error();
              (n==h) == (m==p) || error();
              (n==h) == (valuation(m,2)==k) || error();
              if(n==h,
                 return(0)));
           h>>=1;
           k--);
     1);
  check_equal(vector(2^12,n,n--; a(n)),
              vector(2^12,n,n--; A079559(n)),
              "h = 2^k-1 by shifting with m,p vs A079559()");
}


\\ 0   1;
\\ 1   1,0;
\\ 3   1,1,0,0;
\\ 7   1,1,0,1,1,0,0,0;
\\ 15  1,1,0,1,1,0,0,1,1,0,1,1,0,0,0,0;
\\ 31  1,1,0,1,1,0,0,1,1,0,1,1,0,0,0,1,1,0,1,1,0,0,1,1,0,1,1,0,0,0,0,0;
\\ m=n+2
\\ m=2^k is n=2^k-2 is last 0 of each row
my(k=3); check_equal(2^k-2, 6);
{
  \\ a(2^k-2 + r) = a(-1 + r)
  check_equal(vector(2,k,k++; apply(n->if(n<0,0, A079559(n)),
                                                  [   -1 .. 2^k-1     -1])),
              vector(2,k,k++; apply(A079559, [2^k-2 .. 2^(k+1)-2 -1])),
              "A079559 row");

  my(k=1);
  check_equal(apply(n->if(n<0,0, A079559(n)),
                                  [   -1 .. 2^k-1     -1]),
              [0, 1]);
  check_equal(apply(A079559, [2^k-2 .. 2^(k+1)-2 -1]),
              [1, 1]);
}
my(k=1); check_equal(2^k-2,   0, "k=1 low n");
my(k=1); check_equal(2*2^k-3, 1, "k=1 high n");

my(k=2); check_equal(2^k-2,   2, "k=2 low n");
my(k=2); check_equal(2*2^k-3, 5, "k=2 high n");

\\ initial W(k) = 2^k-1 terms a(0) to a(2^k-2)
\\ is binomial tree order k without one final 0
\\ replication W(k+1) = W(k),W(k),0

{
  \\ A079559 gf, per Vladeta Jovovic original submission there
  my(g = prod(n=1,12, (1+x^(2^n-1))));
  check_equal(g + O(x^1024),
              Ser(A079559_vector(1024)),
              "A079559 gf");
}


\\ whole of final n of the bit increments
A079559_whole_final_n(n) = \
  n+=2; forstep(k=logint(n,2),2,-1, n+=bittest(n,k)); n;

\\ vector(20,n, A079559_whole_final_n(n))
\\ not in OEIS: 3, 5, 6, 7, 8, 9, 10, 11, 13, 14, 15, 16, 16, 17, 18, 19, 21, 22, 23, 24
\\
\\ vector(20,n, A079559_whole_final_n(n) % 4)
\\ not in OEIS: 3, 1, 2, 3, 0, 1, 2, 3, 1, 2, 3, 0, 0, 1, 2, 3, 1, 2, 3, 0

\\ low bit
\\ vector(24,n, A079559_whole_final_n(n) % 2)
\\ not in OEIS: 1, 1, 0, 1, 0, 1, 0, 1, 1, 0, 1, 0, 0, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1, 1
\\

{
  \\ final n     = 1 2 3 0
  \\ xor two bits  1 1 0 0
  \\ cf A079559 =  0 1 1 0
  \\ is a shift because never more than two 1s consecutive,
  \\ which is everywhere first child of a vertex is childless
  check_equal(vector(2^12,n,n--; my(w=A079559_whole_final_n(n));
                     bitxor(bittest(w,0), bittest(w,1))),
              vector(2^12,n,n--; A079559(n + 1)))
}

\\ count low 0s
test_A007814(n) = valuation(n,2);
{
  \\ run lengths of 0s is A007814, per Franklin T. Adams-Watters
  \\
  my(v=vector(2^10), upto=1, zeros=0, n=1);
  while(upto<=#v,
        if(A079559(n),
           v[upto]=zeros;
           if(upto==#v,break);
           zeros=0;
           upto++;
           ,
           zeros++);
        n++);

  check_equal(v, vector(#v,n,test_A007814(n)),
             "A079559 run lengths of 0s is A007814");
}


\\-----------------------------------------------------------------------------
progress("A046699 partial sums of A079559 binomial 0,1");
\\ 1, 1, 2, 2, 3, 4, 4, 4, 5, 6, 6, 7, 8, 8, 8, 8, 9, 10, 10, 11, 12, 12,
\\   1, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 1, 0, 0, 0, 1, 1,  0,  1,  1,  0,

\\ defining recurrence in A046699
A046699_vector(len) =
{
  my(v=vector(len));
  if(len>=1,
     v[1]=1;
     if(len>=2,
        v[2]=1;
        for(n=3,len,
           v[n] = v[n - v[n-1]] + v[n-1 - v[n-2]])));
  v;
}
A046699(n) =
{
  n>=1 || error("A046699 is for n>=1");
  A046699_vector(n)[n];
}

{
  check_equal(A046699_vector(#A046699_samples),
              A046699_samples,
              "A046699_vector() vs A079559_samples");

  check_equal(A046699_vector(2^8),   \\ OFFSET=1
              vector(2^8,n,
                     if(n==1,1,
                        sum(i=0,n-2, A079559(i)))),
              "A046699 partial sums of A079559");

  my(v=A079559_vector(2^12));
  for(i=2,#v, v[i]+=v[i-1]);
  v=concat(1,v);
  check_equal(v,
              A046699_vector(#v),
              "A046699 partial sums of A079559")
}


\\ Jackson and Ruskey, "Meta-Fibonacci Sequences, Binary Trees And
\\ Extremal Compact Codes", Electronic Journal of Combinatorics, volume
\\ 13, 2006, R26.  Corollary 2.3 from:
\\
\\ B.W. Conolly, Meta-Fibonacci Sequences, Chapter XII in S. Vajda,
\\ Fibonacci & Lucas Numbers, and the Golden Section, Ellis Horwood
\\ Limited, 1989.  Equation 6 page 132.
\\
A046699_a0(n) =  \\ sum d0 from 0 to n inclusive, per proof of Corollary 2.3
{
  if(n==0,0, A046699(n+1));
}
{
  check_equal(vector(2^8,n,n--; A046699_a0(n)),
              vector(2^8,n,n--; sum(i=0,n, A079559_d0(i))),
              "A046699 Jackson and Ruskey a0 sums");

  check_equal(vector(7,h, A046699_a0(2^h-1)),
              vector(7,h, 2^(h-1)),
              "A046699 Jackson and Ruskey proof of corollary 2.3");

  \\                                            n = 0  1 2 3  4 5 6  7
  check_equal(vector(8,n,n--; A079559_d0(n)), [0, 1,1,0, 1,1,0, 0]);
  check_equal(vector(8,n,n--; A046699_a0(n)), [0, 1,2,2, 3,4,4, 4]);

  \\ 0 <= k < 2^h
  check_equal(vector(8,h, vector(2^h,k,k--; A046699_a0(   2^h-1 + k ))),
              vector(8,h, vector(2^h,k,k--; 2^(h-1) + A046699_a0( k ))),
              "A046699 Jackson and Ruskey proof of corollary 2.3 a0 formula by Conolly");
}

{
  \\ A046699 gf
  \\ Jorg Arndt fxtbook section 1.26.5 with offset 0
  my(g = 1 + x/(1-x) * prod(n=1,12, (1+x^(2^n-1))));
  check_equal(g + O(x^1024),
              Ser(A046699_vector(1024)),
              "A046699 gf offset 0");
}
{
  \\ A046699 gf
  \\ Frank Ruskey and Chris Deugau in A046699  OFFSET=1
  my(g = x + x^2/(1-x) * prod(n=1,12, (1+x^(2^n-1))),
     v=A046699_vector(1023));
  check_equal(g + O(x^1024),
              sum(n=1,#v, v[n]*x^n) + O(x^1024),
              "A046699 gf");
}


\\-----------------------------------------------------------------------------
progress("binomial tree postorder vertex numbers = A108918");
\\
\\ (and cf mod 2 = A079559 = balanced binary coding of the tree ...)

\\     0---
\\     | \  \
\\     1  2  4
\\        |  | \
\\        3  5  6
\\              |
\\              7

{
  check_equal(vector(#A108918_samples,n, test_A108918(n)),
              A108918_samples,
              "A108918() vs A108918_samples");

  my(len=1024,
     vpar=vpar_make_binomial_tree(len),
     seq=vpar_seq_postorder(vpar));  \\ vertex numbers
  seq=vector(#seq,i,seq[i]-1);  \\ 0-based numbering, not 1-based
  check_equal(seq[#seq], 0, "final root vertex");
  seq=seq[^len];                \\ discard root

  check_equal(seq,
              vector(#seq,n, test_A108918(n)),
              "binomial tree postorder vertices = A108918");

  \\ Knuth
  check_equal(vector(1024,n,
                     my(k=logint(n,2));
                     (2^k <= n && n < 2^(k+1)) || error();
                     if(n==2^(k+1)-1,
                        2^k,
                        2^k + test_A108918(n - 2^k + 1));
                    ),
              vector(1024,n, test_A108918(n)),
              "A108918 Knuth T recurrence");

  \\ bittest and separate ret
  check_equal(vector(1024,n,
                     my(ret=0);
                     forstep(k=logint(n,2),0,-1,
                        if(bittest(n,k),
                           n++;
                           ret += 1<<k;
                           if(!bittest(n,k), break)));
                     ret;
                    ),
              vector(1024,n, test_A108918(n)),
              "A108918 by bittest");

  \\ in-place and final s
  check_equal(vector(1024,n,
    my(s); forstep(k=logint(n,2),0,-1, if(bittest(n,k), n++;s=k)); n-(1<<s);
                    ),
              vector(1024,n, test_A108918(n)),
              "A108918 remember final s");
}
{
  \\ in-place and exit detected
  my(f=(n)->
     forstep(k=logint(n,2),0,-1,
        if(bittest(n,k), n++; if(!bittest(n,k), return(n-(1<<k)))));
     error());
  check_equal(vector(1024,n, f(n)),
              vector(1024,n, test_A108918(n)),
              "A108918 in-place");
}
{
  \\ bit h kept, less efficient than bittest if h a bignum
  my(f=(n)->
     my(h=1<<logint(n,2));
     while(h, if(bitand(h,n), n++; if(bitnegimply(h,n), return(n-h))); h>>=1);
     error());
  check_equal(vector(1024,n, f(n)),
              vector(1024,n, test_A108918(n)),
              "A108918 bit h");
}

test_A135136(n) =
{
  n>=0 || error("test_A135136 is for n>=0");
  floor(hammingweight(n)/2) % 2;
}
{
  check_equal(vector(#A135136_samples,n,n--; test_A135136(n)),
              A135136_samples,  \\ OFFSET=0
              "test_A135136 vs A135136_samples");

  \\ is not quite binomial postorder bigger than n
  \\ check_equal(vector(32,n,n++; test_A108918(n) >= n),
  \\             vector(32,n,n++; test_A135136(n+1)),
  \\             "A135136() binomial postorder >= n");
\\ vector(32,n,n++; test_A108918(n) >= n) - vector(32,n,n++; test_A135136(n+1))
}


\\-----------------------------------------------------------------------------
progress("binomial tree postorder subtree heights = A215020");

\\ A215020 defined as bitlength(A182105) where A182105 bottom-up merge sort
\\   = binomial tree postorder subtree heights
\\
A215020(n) = valuation(test_negidx2lexrev(n), 2);
{
  \\ A215020 = postorder count low 0-bits, which is final s
  check_equal(vector(#A215020_samples,n, A215020(n)),
              A215020_samples,  \\ OFFSET=1
              "A215020 binomial tree postor count low 0s");

  \\ A215020  last position of 1-bit
  check_equal(vector(1024,n,
    my(ret); forstep(k=logint(n,2),0,-1, if(bittest(n,k), n++;ret=k)); ret;
                    ),
              vector(1024,n, A215020(n)),
              "A215020 by remember last s");

  \\ binomial tree postorder subtree heights
  my(len=32,
     vpar=vpar_make_binomial_tree(len),
     subtree_heights=vpar_subtree_heights_vector(vpar),
     seq=vpar_seq_postorder(vpar));
  subtree_heights=vecextract(subtree_heights,seq);  \\ in seq order
  seq=vector(#seq,i,seq[i]-1);    \\ 0-based vertex numbering, not 1-based

  check_equal(seq[#seq], 0,   "final root vertex");
  check_equal(subtree_heights[#subtree_heights],
              logint(len,2),  "final root subtree height");
  seq=seq[^len];                                       \\ discard root
  subtree_heights=subtree_heights[^#subtree_heights];  \\ discard root

  check_equal(subtree_heights,
              vector(#subtree_heights,n, A215020(n)),
              "binomial tree postorder subtree height vs A215020()");
}

\\-------
\\ A082850 same but +1

\\ 1,1,2, 1,1,2, 3, 1,1,2, 1,1,2, 3, 4, 1,1,2, 1,1,2, 3, 1,1,2
A082850(n) =
{
  n>=1 || error("A082850() is for n>=1");
  valuation(test_negidx2lexrev(n), 2) + 1;
}
{
  \\ A082850 = runs up 1 to CountLowZeros(k) for successive k
  check_equal(vector(#A082850_samples,n, A082850(n)),
              A082850_samples,  \\ OFFSET=1
              "A082850 binomial tree postor count low 0s");

  check_equal(vector(100,m, A082850(2^m-1)),
              vector(100,m, m),
              "A082850 at 2^m-1");

  \\ A082850 recurrence by Franklin T. Adams-Watters
  check_equal(vector(1024,n,
                     my(m=logint(n+1,2), k = n - (2^m-1));
                     n == 2^m-1 + k || error("n=",n);
                     (0 <= k && k < 2^m) || error();
                     if(k==0, m,
                        A082850(k))),
              vector(1024,n, A082850(n)),
              "A082850 recurrence");

  check_equal(vector(1024,n,
    my(ret); forstep(k=logint(n,2),0,-1, if(bittest(n,k), n++;ret=k)); ret+1;
                     ),
              vector(1024,n, A082850(n)),
              "A082850 by final k");

  \\ A[k+1] = A[k]*(1+x^(2^k-1)) + (k+1)*x^(2^(k+1)-1)
  check_equal( (x+x^2+2*x^3)*(1+x^(2^2-1)) + (2+1)*x^(2^(2+1)-1),
              sum(n=1,7, A082850(n)*x^n));
  check_equal( (x^1*(1+x^(2^1-1)) + 2*x^3)*(1+x^(2^2-1)) + (2+1)*x^(2^(2+1)-1),
              sum(n=1,7, A082850(n)*x^n));

  \\ A082850 gf
  my(K=10,
     limit=2^K,
     L=O(x^limit),
     g=sum(n=1,limit-1, A082850(n)*x^n) + L);
  check_equal(sum(k=1,K, k*x^(2^k-1) * prod(j=k,K, 1+x^(2^j-1))) + L,
              g);
}


{
  \\ variant of A082850 runs 1 .. CountLowZeros
  \\ here instead runs of CountLowZeros many of CountLowZeros value

  \\ not in OEIS: 1, 2, 2, 1, 3, 3, 3, 1, 2, 2, 1, 4, 4, 4, 4, 1, 2, 2, 1, 3
  \\ not in OEIS: 0, 1, 1, 0, 2, 2, 2, 0, 1, 1, 0, 3, 3, 3, 3, 0, 1, 1, 0, 2
  my(f=(n)->
     forstep(k=logint(n,2),0,-1, n+=bittest(n,k)); valuation(n,2);
    );

  \\ repeat runs
  my(len=1024, v=vector(len,i,'none), pos=0, t);
  for(t=1,2*len+2,
     my(c=valuation(t,2));
     for(j=1,c, v[pos++]=c; if(pos>=len, break(2))));
  check_equal(v, vector(len,n, f(n)));
}

A001511(n) =
{
  n>=1 || error("A001511() is for n>=1");
  valuation(n,2)+1;
}
{
  check_equal(vector(#A001511_samples,n, A001511(n)),
              A001511_samples,  \\ OFFSET=1
              "A001511 count low 0-bits + 1");
}

\\ A252488 run lengths of 0s and 1s given by CountLowZeros+1
\\
\\ 1,0,0,1,0,0,0,1,0,0,1,0,0,0,0,1,0,0,1,0,0,0,1,0,0,1,0,0,0,0,0,1,0,0,1,0,0,0,1
\\ 1   2 1     3 1 2   1       4 1   2 1     3 1   2 1         5 1   2 1     3
\\       4     7              15                              31
A252488(n) =
{
  n>=1 || error("A252488() is for n>=1, got n=",n);
  forstep(k=logint(n,2),0,-1, n+=bittest(n,k)); bittest(n,1);
}
A252488_allowing_n0(n) = if(n==0, 0, A252488(n));

gA252488(limit) =
{
  my(L=O(x^limit),
     K=logint(limit+1,2));
  \\ K big enough that next multiply x * x^(2^(K+1)-1) bigger than needed
  1 + 2^(K+1)-1 >= limit  || error();

  x * prod(k=2,K, 1+x^(2^k-1) + L) + L;
}
check_equal(2^2-1, 3); \\ first term 1+x^3

{
  check_equal(vector(#A252488_samples,n, A252488(n)),
              A252488_samples,  \\ OFFSET=1
              "A252488 runs 0s and 1s");

  check_equal(vector(1024,n,
                     if(n<=2, A252488(n),
                        my(k=logint(n+1,2));
                        (2^k-1 <= n && n < 2^(k+1)-1) || error();
                        A252488_allowing_n0(n-(2^k-1)))
                     ),
              vector(1024,n, A252488(n)),
              "A252488 recurrence");

  \\ A252488 looping down to 1 (instead of 0) same but n++ for offset
  check_equal(vector(1024,n,
    n++; forstep(k=logint(n,2),1,-1, n+=bittest(n,k)); bittest(n,1);
                     ),
              vector(1024,n, A252488(n)),
              "A252488 down to bit 1 only");
}{
  my(len=64, v=vector(len,i,'none), pos=0, t);
  for(t=1,2*len+2,
     my(c=valuation(t,2)+1);
     for(j=1,c, v[pos++]=t%2; if(pos>=len, break(2))));
  check_equal(v, vector(len,n, A252488(n)),
              "A252488 by run lengths");
}{
  check_equal(gA252488(100),
              sum(n=1,100-1, A252488(n)*x^n) + O(x^100),
              "gA252488() vs A252488()");
  my(g_by_sum=0);
  for(n=1,128,
     g_by_sum += A252488(n)*x^n;
     my(got=gA252488(n+1));
     \\ length(t_SER) counts from lowest non-zero
     check_equal(#got,n, "gA252488() length");
     check_equal(got,
                 g_by_sum + O(x^(n+1)),
                 "gA252488() vs sum A252488()"));

  check_equal(gA252488(256) == x/(1+x) * gA079559(256), 1,
              "gA252488 vs gA079559");

  check_equal(vector(1024,n, A252488(n) + A252488(n+1)),
              vector(1024,n, A079559(n)),
              "two of A252488 is A079559");

  check_equal(vector(24,n, A252488(n)),
              vector(24,n, sum(i=0,n-1, (-1)^(n-1-i) * A079559(i))),
              "A252488 by alternating sum A079559");

  check_equal(vector(6,k, sum(i=1,2^(k-1), A001511(i))),
              vector(6,k, 2^k-1),
              "A001511 block sum");
}

\\ definition of A005187, and cf code by Michael Somos there
A005187(n) =
{
  n>=0 || error("A005187() is for n>=0");
  2*n - hammingweight(n);
}
{
  check_equal(vector(#A005187_samples,n,n--; A005187(n)),
              A005187_samples,  \\ OFFSET=0
              "A005187 2n-CountOnes(n)");

  \\ Max Hasler in A005187
  check_equal(vector(256,n,n--; A005187(2*n+1)),
              vector(256,n,n--; A005187(2*n)+1));
}

\\ definition of A120738
A120738(n) =
{
  n>=0 || error("A120738() is for n>=0");
  4*n - hammingweight(n);
}
{
  check_equal(vector(#A120738_samples,n,n--; A120738(n)),
              A120738_samples,  \\ OFFSET=0
              "A120738 4n-CountOnes(n)");
}

{
  \\ A252488 locations of 1s, Neil Sloane and Max Hasler
  my(v=select(n->A252488(n)==1, [1..1024]));
  check_equal(vector(#v,n,n--; A005187(2*n+1)),
              v,
              "A252488 locations of 1s vs A005187");
  check_equal(vector(#v,n,n--; A005187(2*n)+1),
              v,
              "A252488 locations of 1s vs A005187");
  check_equal(vector(#v,n,n--; A120738(n)+1),
              v,
              "A252488 locations of 1s vs A120738");
}


\\-----------------------------------------------------------------------------
progress("binomial tree postorder vertices inverse permutation = A118319");

\\ Jorg Arndt fxtbook 1.26.2 T(x), inverse of negidx2lexrev
test_binomial_tree_T_recursive(n) =
{
  n>=0 || error("test_binomial_tree_T_recursive() is for n>=0");
  if(n==0,return(0));
  my(h=test_highbit(n));
  h-1 + if(n-h!=0, test_binomial_tree_T_recursive(n-h),
                   h);
}
A118319(n) = test_binomial_tree_T_recursive(n);
{
  check_equal(vector(#A118319_samples,n, A118319(n)),
              A118319_samples,  \\ OFFSET=1
              "A118319() vs A118319_samples");

  check_equal(vector(1024,n, A118319(test_A108918(n))),
              vector(1024,n, n),
              "A118319 inverse of A108918");

  my(len=1024,
     vpar=vpar_make_binomial_tree(len),
     seq=vpar_seq_postorder(vpar),  \\ vertex numbers
     perm=seq^-1);
  check_equal(perm[1],len); \\ root goes to last place
  perm=perm[^1];            \\ discard root

  check_equal(Vec(perm),
              vector(#perm,n, A118319(n)),
              "binomial tree postorder locations of vertices = A118319");
}


\\-----------------------------------------------------------------------------
progress("binomial tree postorder depths = A100661");
\\
\\ A100661 how many maximal h=2^k-1 subtracted, binomial postorder depths
\\
\\ Fxtbook 1.26.3 Minimal Decompositions into Terms 2^k-1

\\ ~/OEIS/a100661.txt
\\     Leroy Quet's transform


\\ A100661 = postorder depth
A100661(n) =
{
  n>=1 || error("A100661() is for n>=1");
  n++; my(ret=0); while(n>1, n-=(1<<logint(n,2))-1; ret++); ret;
}

\\ A080468 = postorder depth-1
A080468(n) =
{
  n>=2 || error("A080468() is for n>=2");
  my(ret=-1); while(n>1, n-=(1<<logint(n,2))-1; ret++); ret;
}

\\ something not-already-in-sequence
A080578(n) =
{
  n>=1 || error("A080578() is for n>=1");
  my(ret=1);
  while(n>=2,
        my(h=1<<logint(n,2));
        ret += 2*h-1;   \\ h + h-1
        n -= h-1);
  n==1 || error();
  ret;
}

{
  \\ A100661 gf
  \\ Jorg Arndt in A100661
  \\ my(g = x*(2*(1-x)*prod(n=1,12, (1+x^(2^n-1))) - 1)
  \\          / (1-x)^2);
  my(k=11, len=2^(k-1));
  my(g = -x/(1-x)^2  + 2*x/(1-x)*prod(n=1,k, 1+x^(2^n-1)));
  check_equal(g + O(x^len),
              sum(n=1,len-1, A100661(n)*x^n  + O(x^len)),  \\ OFFSET=1
              "A100661 gf");
}
{
  \\ A100661 gf with offset 0
  \\ Jorg Arndt Fxtbook 1.26.3
  my(k=11, len=2^(k-1));
  my(g = (-1 + 2*(1-x) * prod(n=1,k, (1 + x^(2^n-1))))
         / (1-x)^2);
  check_equal(g + O(x^len),
              sum(n=1,len, A100661(n)*x^(n-1)  + O(x^len)),  \\ OFFSET=1
              "A100661 gf");
}

{
  check_equal(vector(#A100661_samples,n, A100661(n)),
              A100661_samples,
              "A100661() vs A100661_samples");

  \\ A100661 = depth in postorder
  \\         = hammingweight
  check_equal(vector(#A100661_samples,n, A100661(n)),
              A100661_samples,
              "A100661() vs A100661_samples");

  check_equal(vector(1024,n, A080578(n+1) - 2*n - 1),
              vector(1024,n, A100661(n)),
              "A100661() related not-already A080578()");

  check_equal(vector(1024,n, A100661(n)),
              vector(1024,n, hammingweight(test_A108918(n))),
              "A100661() count 1-bits in A108918 postorder");

  check_equal(vector(1024,n, hammingweight(test_negidx2lexrev(n))),
              vector(1024,n, A100661(n)),
              "test_negidx2lexrev() count 1-bits vs A100661()");

  check_equal(vector(1024,n,
                     my(ret=0); while(n, my(h=2^logint(n+1,2)-1); n-=h; ret++); ret;
                    ),
              vector(1024,n, A100661(n)),
              "A100661 number of maximal 2^k-1 subtractions");

  my(len=1024,
     vpar=vpar_make_binomial_tree(len),
     depths=vpar_depths_vector(vpar),
     seq=vpar_seq_postorder(vpar));
  depths=vecextract(depths,seq);  \\ in seq order
  seq=vector(#seq,i,seq[i]-1);    \\ 0-based vertex numbering, not 1-based

  check_equal(seq[#seq],       0, "final root vertex");
  check_equal(depths[#depths], 0, "final root depth");
  seq=seq[^#seq];                 \\ discard root
  depths=depths[^#depths];        \\ discard root

  check_equal(depths,
              vector(#seq,i, hammingweight(seq[i])),
              "binomial tree postorder depths vs hammingweight()");
  check_equal(depths,
              vector(#depths,n, A100661(n)),
              "binomial tree postorder depth vs A100661()");

  \\--------
  \\ A080468
  check_equal(apply(n->n-1, depths),
              vector(#depths,n,n++; A080468(n)),
              "binomial tree postorder depth-1 vs A080468()");

  \\ A080468 = hammingweight - 1 in postorder
  \\         = depth - 1
  check_equal(vector(#A080468_samples,n,n++; A080468(n)),
              A080468_samples,
              "A080468() vs A080468_samples");

  check_equal(vector(1024,n,n++; A080468(n)),
              vector(1024,n,n++; A100661(n-1) - 1),
              "A080468() related A100661()");
  check_equal(vector(1024,n, A100661(n)),
              vector(1024,n, A080468(n+1) + 1),
              "A100661() related A080468()");

  check_equal(vector(1024,n,n++; A080578(n) - 2*n),
              vector(1024,n,n++; A080468(n)),
              "A080468() related not-already A080578()");

  check_equal(vector(1024,n,n++; hammingweight(test_negidx2lexrev(n-1)) - 1),
              vector(1024,n,n++; A080468(n)),
              "test_binomial_tree_postorder() count 1-bits vs A080468()");
}

{
  \\ from the gfs of A046699 and A100661
  check_equal(vector(512,n, 2*A046699(n)),
              vector(512,n, if(n==1,2, n-1 + A100661(n-1))),
              "A046699 and A100661");

  check_equal(vector(512,n, A046699(n)),
              vector(512,n, if(n==1,1, (n-1 + A100661(n-1))/2)),
              "A046699 from A100661");

  check_equal(vector(512,n, A100661(n)),
              vector(512,n, 2*A046699(n+1) - n),
              "A100661 from A046699");
}

{
  \\ A080578
  check_equal(vector(#A080578_samples,n, A080578(n)),
              A080578_samples,
              "A080578() vs A080578_samples");

  check_equal(vector(1024,n,
                     my(ret=2*n-1); while(n>1, n-=(1<<logint(n,2))-1; ret++); ret;
                    ),
              vector(1024,n, A080578(n)),
              "A080578() by decreasing 2*n-1");
}

A007843(n) =
{
  n>=0 || error();
  if(n==0,1, my(ret=n);n++; while(n>1, n-=(1<<logint(n,2))-1; ret++); ret);
}
{
  \\ A007843 = smallest k for which k! has n low 0-bits
  \\         = A080578(n) - n up to offset
  check_equal(vector(#A007843_samples,n,n--; A007843(n)),
              A007843_samples,  \\ OFFSET=0
              "A007843() vs A007843_samples");

  check_equal(vector(1024,n,n--; A007843(n)),
              vector(1024,n,n--; if(n==0,1, A080578(n+1)-(n+1))),
              "A007843 = A080578 - n");

  my(n=0);
  check_equal(A080578(n+1)-(n+1), 0); \\ whereas A007843(0)=1
  check_equal(n!, 1);   \\ could have had A007843(0)=0 with 0! = 1
  check_equal(A007843_samples[1], 1);
}


\\-----------------------------------------------------------------------------
progress("vpar_perfect_domsets_count()");

check_equal(vpar_perfect_domsets_count([]), 1, \
            "vpar_perfect_domsets_count() empty");
check_equal(vpar_perfect_domsets_count([0,1]), 3, \
            "vpar_perfect_domsets_count() path-2 = one or both");

test_perfect_domsets_count_by_flags(vpar) =
{
  my(count=0);
  forvec(flags=vector(#vpar,i,[0,1]),
        count += vpar_is_perfect_domset_flags(vpar,flags));
  count;
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got = vpar_perfect_domsets_count(vpar));

     if(#vpar <= test_limit_flags_search,
        check_equal(got, test_perfect_domsets_count_by_flags(vpar),
                    "vpar_perfect_domsets_count() vs test_perfect_domsets_count_by_flags()");
       ));
}

A179070(n) =
{
  n>=1 || error("A179070() is for n>=1");
  \\ if(n<=3,1, my(x=3,y=1,z=1); for(i=5,n, [x,y,z] = [x+z,x,y]); x);
  if(n==1,1, polcoeff(lift((2*x-1)*Mod(x,x^3-x^2-1)^(n+1)), 0));
}
\\ recurrence_guess(vector(50,n,n--; if(n,if(n<=3,1, my(x=3,y=1,z=1); for(i=5,n, [x,y,z] = [x+z,x,y]); x))));
\\ recurrence_guess(vector(50,n,n--; if(n,A179070(n))))
{
  check_equal(vector(#A179070_samples,n, A179070(n)),
              A179070_samples,
              "A179070()");

  \\ perfect domsets count of path
  \\ A179070 a(1)=a(2)=a(3)=1, a(4)=3; thereafter a(n) = a(n-1) + a(n-3).
  \\
  \\ end 1       1 or 0dom          start 1      -> 1 or 0undom     n+1
  \\ end 0undom  1                        0dom   -> 1               n
  \\ end 0dom    0undom                   0undom -> 0dom            n-1
  \\ forbidden 101        allowed 1001
  \\          100 0       every 0 exactly one 1 adjacent
  \\          100 end     each 0 as 10 or 01 not both
  check_equal(vector(50,n,n--; A179070(n+2)),
              vector(50,n,n--; vpar_perfect_domsets_count(vpar_make_path(n))),
              "vpar_perfect_domsets_count() of path n");
}
\\ for(b=8,64, \
\\   my(flags=binary(b)[^1], \
\\      vpar=vpar_make_path(#flags)); \
\\   if(vpar_is_perfect_domset_flags(vpar,flags), \
\\      print1(b,",")));
\\      print(flags)));

{
  \\ perfect domsets count of star
  \\ must middle, as otherwise it is multi-dominated by the surrounds
  check_equal(vector(20,n,n--; if(n==0,1, n==2,3, 2^(n-1))),
              vector(20,n,n--; vpar_perfect_domsets_count(vpar_make_star(n))),
              "vpar_perfect_domsets_count() of star n");
}
{
  \\ A131831 perfect domsets count of subdivided star
  \\ A100314 sans initial n=0 term
  check_equal(vector(10,n,n--; vpar_perfect_domsets_count
                                (vpar_subdivide(vpar_make_star(n)))),
              vector(10,n,n--; if(n==0,1, 2^(n-1) + 2*n - 2)));

  check_equal(vector(#A131831_samples,n,
                     vpar_perfect_domsets_count
                      (vpar_subdivide(vpar_make_star(n)))),
              A131831_samples,
              "A131831 perfect domsets count of subdivided star");
}
{
  \\ perfect domsets count of complete binary tree
  check_equal(vector(6,h,h--;
                   vpar_perfect_domsets_count(vpar_make_complete_tree(2,h))),
              vector(6,h,h--; if(h==0,1, 2^(2^h))));
  \\ not in OEIS: 2, 4, 41, 1438, 4505183
  \\ factor(4505183)
  \\ same as minimal domsets count of complete binary tree
}
{
  \\ perfect domsets count of binomial tree n
  \\ is A228693 except initials, most indomsets of any tree (Wilf)
  check_equal(vector(20,n,n--; vpar_perfect_domsets_count
                                (vpar_make_binomial_tree(n))),
              vector(20,n,n--; if(n<=1,1, 2^floor((n+1)/2) + ((n+1)%2))));

  my(samples=A228693_samples[5..#A228693_samples]);
  check_equal(vector(#samples,n,n++; vpar_perfect_domsets_count
                                      (vpar_make_binomial_tree(n))),
              samples,
              "A228693 perfect domsets count of binomial tree n");
}
{
  \\ A000215 perfect domsets count of binomial tree 2^k
  \\ 2^(2^(k-1)) + 1
  check_equal(vector(7,k,k--;
                   vpar_perfect_domsets_count(vpar_make_binomial_tree(2^k))),
              vector(7,k,k--; if(k==0,1, 2^(2^(k-1)) + 1)));

  check_equal(vector(#A000215_samples,n,n--;
                   vpar_perfect_domsets_count(vpar_make_binomial_tree(2^(n+1)))),
              A000215_samples,  \\ OFFSET=0
              "A000215 perfect domsets count of binomial tree order n+1");
}


\\-----------------------------------------------------------------------------
progress("vpar_totdomnum_and_count()");

check_equal(vpar_totdomnum_and_count([0,1,2]), [2,2]);
check_equal(vpar_totdomnum_and_count([0,1,1]), [2,2]);

test_totdomnum_and_count_by_flags(vpar) =
{
  if(#vpar==0,return([0,1]));  \\ empty set only
  my(totdomnum=#vpar+1,count=0);
  forvec(flags=vector(#vpar,i,[0,1]),
        if(vpar_is_totdomset_flags(vpar,flags),
           my(size=vecsum(flags));
           if(size<totdomnum,  totdomnum=size;count=1,  \\ new low
              size==totdomnum, count++)));           \\ equal low
  if(count==0,totdomnum=-1);
  [totdomnum,count];
}

check_equal(test_totdomnum_and_count_by_flags(vpar_make_path(0)), [0,1], \
            "test_totdomnum_and_count_by_flags of path 0");
check_equal(test_totdomnum_and_count_by_flags(vpar_make_path(1)), [-1,0], \
            "test_totdomnum_and_count_by_flags of path 1");
check_equal(test_totdomnum_and_count_by_flags(vpar_make_path(2)), [2,1]);
check_equal(test_totdomnum_and_count_by_flags(vpar_make_path(3)), [2,2]);
check_equal(test_totdomnum_and_count_by_flags(vpar_make_path(4)), [2,1]);

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     my(by_flags=test_totdomnum_and_count_by_flags(vpar));

     check_equal(vpar_totdomnum(vpar),by_flags[1],
                 "vpar_totdomnum() vs test_totdomnum_and_count_by_flags()");

     check_equal(vpar_totdomnum_and_count(vpar), by_flags,
       "vpar_totdomnum_and_count() vs test_totdomnum_and_count_by_flags()");
  );
}

test_totdomnum_count_of_path(n) =
{
  my(m=n\4);
  if(n%4 == 0, 1,
     n%4 == 1, m,
     n%4 == 2, (m+1)^2,
     n%4 == 3, m+2,
     error());
}
{
  check_equal(vector(200,n,n--;
                     vpar_totdomnum_and_count(vpar_make_path(n))[2]),
              vector(200,n,n--; test_totdomnum_count_of_path(n)),
              "vpar_totdomnum_and_count() count of path");

  my(limit=test_limit_flags_search);
  check_equal(vector(limit,n,n--;
                 test_totdomnum_and_count_by_flags(vpar_make_path(n))[2]),
              vector(limit,n,n--; test_totdomnum_count_of_path(n)),
              "vpar_totdomnum_and_count() count of path");
}
{
  \\ A302654 number of minimum totdomnum sets in path-n
  check_equal(vector(#A302654_samples,n, test_totdomnum_count_of_path(n)),
              A302654_samples,  \\ OFFSET=1
             "A302654 number of totdomnum sets in path-n");
}

test_is_minimal_totdomset_flags(vpar,flags) =
{
  vpar_is_totdomset_flags(vpar,flags) || return(0);
  for(v=1,#vpar,
     if(flags[v],
        my(new_flags=flags);
        new_flags[v]=0;
        if(vpar_is_totdomset_flags(vpar,new_flags), return(0))));
  1;
}
test_minimal_totdomsets_count_by_flags(vpar) =
{
  my(count=0);
  forvec(flags=vector(#vpar,i,[0,1]),
        count += test_is_minimal_totdomset_flags(vpar,flags));
  count;
}
{
  \\ A302655 num minimal totdomsets in path n
  \\ grows as A060006
  \\
  my(limit=min(test_limit_flags_search,#A302655_samples),
     want=Vec(A302655_samples,limit));
  check_equal(vector(#want,n,  \\ A302655_samples OFFSET=1
                test_minimal_totdomsets_count_by_flags(vpar_make_path(n))),
              want,
              "A302655 vs test_minimal_totdomsets_count_by_flags()");
}


\\-----------------------------------------------------------------------------
progress("vpar_matchsets()");

check_equal(vpar_matchsets([]), [[]], "vpar_matchsets() empty");
check_equal(vpar_matchsets([0]), [[]], "vpar_matchsets() singleton");
check_equal(vpar_matchsets([0,1]), [[], [2]], "vpar_matchsets() path-2");
check_equal(vpar_matchsets([2,0]), [[], [1]], "vpar_matchsets() path-2");

test_matchsets_by_flags(vpar) =
{
  my(l=List([]));
  forvec(flags=vector(#vpar,i,[0,1]),
        if(vpar_is_matching_flags(vpar,flags),
           listput(l,test_flags_to_Set(flags))));
  Vec(l);
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        matchings_count=vpar_matchings_count(vpar));
     if(matchings_count > test_limit_sets, next());

     my(got=vpar_matchsets(vpar));
     check_equal(#got, matchings_count, "vpar_matchsets() count");
     for(i=1,#got,
        check_equal(setisset(got[i]), 1);
        check_equal(vpar_is_matching_set(vpar,got[i]), 1));

     check_equal(vecsort(got),
                 vecsort(test_matchsets_by_flags(vpar)));

     check_equal(Pol(0,'y) + sum(i=1,#got, 'y^#got[i]),
                 vpar_matchpoly(vpar,'y),
                 "vpar_matchsets() count by size vs vpar_matchpoly()");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_matchpoly()");

check_equal(vpar_matchpoly([]),    Pol(1),    "vpar_matchpoly() empty");
check_equal(type(vpar_matchpoly([])), "t_POL",  "vpar_matchpoly() empty type()");
check_equal(vpar_matchpoly([0]),   Pol(1),    "vpar_matchpoly() singleton");
check_equal(vpar_matchpoly([0,1]), 'x+1, "vpar_matchpoly() path-2");
check_equal(vpar_matchpoly([0,1,1,1,1]), 4*'x + 1, "vpar_matchpoly() star-5");
check_equal(vpar_matchpoly([0,1,2,3,4]), 3*'x^2 + 4*'x + 1, "path-5");

test_matchpoly_by_flags(vpar) =
{
  my(poly=Pol(0));
  forvec(flags=vector(#vpar,i,[0,1]),
        if(vpar_is_matching_flags(vpar,flags),
           poly += 'x^vecsum(flags)));
  poly;
}

\\ Past code:
test_matchpoly_by_paired(vpar,var='x) =
{
  \\ paired[v] is how many matchings have v already paired.
  \\ unpaired[v] is how many matchings have v not paired.
  \\ The code works upwards from leaf (childless) nodes to apply counts at v
  \\ to parent p.  v_either is all matchings in the subtree at and below v,
  \\ either v paired or unpaired.
  \\
  \\   unpaired[p] = prod v_either of its children
  \\   paired[p] = sum unpaired[v] * others either, ie. it pairs with one
  \\   child v which must be unpaired, and the other children can be anything.
  \\
  \\ Initially all paired[v]=0.  At the leaf nodes there are no paired
  \\ matchings below.  Nodes above the leaves start as no matchings and sum
  \\ the children.  Initially all unpaired[v]=1 which is the empty matching
  \\ as the sole possibility in the tree comprising vertex v alone.
  \\
  \\ This approach roughly the same as
  \\
  \\     Jingyuan Zhang, Xufeng Chen, Weigang Sun, "A Linear-Time Algorithm
  \\     for the Hosoya Index of an Arbitrary Tree", MATCH Communications in
  \\     Mathematical and in Computer Chemistry, volume 75, 2016, pages
  \\     703-714, ISSN 0340-6253.
  \\
  \\ They express the algorithm by a fraction alpha(v) at each vertex, which
  \\ is summed at the end.  The paired and unpaired counts are in their beta
  \\ as m(v) and m-(v).
  \\
  my(seq=vpar_INTERNAL_seq_upwards(vpar),
     paired   = vector(#vpar),      \\ 0 sets
     unpaired = vector(#vpar,v,1),  \\ 1 set size 0
     ret      = var/var);           \\ 1 as poly
  for(i=1,#seq,
     my(v=seq[i],
        p=vpar[v],
        v_either=paired[v]+unpaired[v]);
     if(p,
        paired[p] = paired[p]*v_either + unpaired[p]*unpaired[v]*var;
        unpaired[p] *= v_either;
        ,
        ret *= v_either);  \\ at root
     paired[v]   =     \\ kill intermediate polys
     unpaired[v] = 0);
  ret;
}


{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        poly=vpar_matchpoly(vpar));
     check_equal(type(poly), "t_POL");

     check_equal(poly,
                 test_matchpoly_by_paired(vpar),
                 "vpar_matchpoly() vs test_matchpoly_by_paired()");

     \\ print("vpar="vpar);
     \\ print("poly="poly);

     my(matchnum_and_count=vpar_matchnum_and_count(vpar));
     check_equal(matchnum_and_count[1], poldegree(poly));
     check_equal(matchnum_and_count[2], pollead(poly));

     check_equal(polcoeff(poly,0), 1,
                 "vpar_matchpoly() constant term 1 per addhelp");
     check_equal(polcoeff(poly,1), vpar_num_edges(vpar),
                 "vpar_matchpoly() x term num edges per addhelp");
     check_equal(poldegree(poly), vpar_matchnum(vpar),
                 "vpar_matchpoly() degree vpar_matchnum() per addhelp");
     check_equal(pollead(poly), matchnum_and_count[2],
                 "vpar_matchpoly() pollead() matchnum count per addhelp");
     check_equal(if(#vpar%2==0,
                    polcoeff(poly,#vpar\2),     \\ 0 or 1 when even
                    polcoeff(poly,#vpar\2)>0),
                 vpar_has_perfect_matching(vpar,#vpar%2),
                 "vpar_matchpoly() coeff for vpar_has_perfect_matching() per addhelp");
     check_equal(test_poly_num_real_roots(poly), poldegree(poly),
                 "vpar_matchpoly() all roots real");

     my(matchings_count=vpar_matchings_count(vpar));
     check_equal(matchings_count, vecsum(Vec(poly)));
     check_equal(matchings_count, simplify(subst(poly,'x,1)));
     if(#vpar <= test_limit_flags_search,
        check_equal(poly, test_matchpoly_by_flags(vpar),
                    "vpar_matchpoly() vs test_matchpoly_by_flags()"));

     \\ There are matchings of all sizes from 0 up to matchnum.
     \\ The sizes of maximal matchings are continuous as noted by
     \\     Tomislav Doslic, Ivana Zubac, "Counting Maximal Matchings In
     \\     Linear Polymers", Ars Mathematica Contemporanea, volume 11,
     \\     2016, pages 255-276, http://amc-journal.eu
     \\ and anything non-maximal can have a pair added for size+1.
     \\
     for(n=valuation(poly,'x), poldegree(poly),
        check_equal(polcoeff(poly,n)>0, 1));

     my(component_vpar=vpar_component_trees(vpar));
     check_equal(poly,
                 Pol(prod(i=1,#component_vpar, vpar_matchpoly(component_vpar[i]))),
                 "vpar_matchpoly() product of components");

     \\ FIXME: Is this the right reference?
     \\ Ole J Heilmann and Elliott H Lieb, "Theory of Monomer-Dimer
     \\ Systems", Communications in Mathematical Physics, volume 25, number
     \\ 3, 1972, pages 190-232.
     \\ Not to be confused with adjacency poly upper bound 2*sqrt(maxdegree-).
     \\
     \\ roots absolute value lower bound 1/(4*(maxdegree-1))
     \\ need 2.11.x for polsturm on not square-free and this [] arg form
     if(lex(version(),[2,11]) >= 0,
       my(roots=polrootsreal(poly),
          degrees=vpar_degrees_vector(vpar),
          maxdegree=if(#degrees,vecmax(degrees),0),
          roots_bound=if(maxdegree==0, 0,
                         maxdegree==1, 1 - 1e-10,
                         1/(4*(maxdegree-1)) - 1e-10));
       for(i=1,#roots,
          my(a=abs(roots[i]));
          \\ print(vpar" "a" "roots_bound);
          check_equal(a > roots_bound, 1,
                      "vpar_matchpoly() roots lower bound"));
       check_equal(polsturm(poly,[-roots_bound, roots_bound]), 0,
                   "vpar_matchpoly() roots lower bound, no polsturm in middle");
       );
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_matchings_count()");

{
  \\ Example in Zhang, Chen, Sun, "A Linear-Time Algorithm for the Hosoya
  \\ Index of an Arbitrary Tree" (ref with vpar_matchpoly()), figure 2 and
  \\ section 4.
  \\
  \\         3
  \\      /  |  \
  \\    2    0    4
  \\   / \       / \
  \\  1   6     7   5
  \\
  \\                   v=0      3
  my(vpar=apply(p->p+1, [3,2,3,-1,3,4,2,4]));
  vpar_validate(vpar);
  check_equal(vpar_matchings_count(vpar), 24);

  vpar=vpar_reroot(vpar,7+1);
  check_equal(vpar_matchings_count(vpar), 24);
}

test_matchings_count_by_flags(vpar) =
{
  my(matchnum=0,count=0);
  forvec(flags=vector(#vpar,i,[0,1]),
        count+=vpar_is_matching_flags(vpar,flags));         \\ equal high
  count;
}

{
  \\ path-n matchings count = Fibonacci numbers
  for(n=0,8,
     my(vpar=vpar_make_path(n),
        got=vpar_matchings_count(vpar));
     check_equal(got, test_matchings_count_by_flags(vpar),
                 Str("vpar_matchings_count() path "n));
     check_equal(got, fibonacci(n+1));
  );
}
{
  \\ star-n matchings count = n
  for(n=0,8,
     my(vpar=vpar_make_star(n),
        got=vpar_matchings_count(vpar));
     check_equal(got, test_matchings_count_by_flags(vpar),
                 Str("vpar_matchings_count() star "n));
     check_equal(got, max(1,n));
  );
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got=vpar_matchings_count(vpar));

     if(#vpar <= test_limit_flags_search,
        check_equal(vpar_matchings_count(vpar),
                    test_matchings_count_by_flags(vpar),
               "vpar_matchings_count() vs test_matchings_count_by_flags()"));
  );
}

{
  \\ binomial tree n vertices  vpar_matchings_count()
  \\ not in OEIS: 2, 3, 5, 7, 12, 17, 29, 39, 68, 97, 165
  my(want=[1, 2, 3, 5, 7, 12, 17, 29, 39, 68, 97, 165]);
  check_equal(vector(#want,n,
                     vpar_matchings_count(vpar_make_binomial_tree(n))),
              want);
}

\\-----------------------------------------------------------------------------
progress("vpar_matchings_count() of binomial tree order k");

\\ A073833 numerators of b(n) where b(1) = 1, b(i) = b(i-1) + 1/b(i-1).
\\ A073834 denominators of b(n) where b(1) = 1, b(i) = b(i-1) + 1/b(i-1).
\\
\\ num/den -> num/den + den/num
\\          = (num^2 + den^2) / (num*den)
\\ no common factor since gcd(num^2+den^2, num) = gcd(num,den)
\\ no common factor since gcd(num^2+den^2, den) = gcd(num,den)
\\ and they start num=den=1 which is gcd(num,den)=1

A073833(n) =
{
  n>=1 || error("A073833() is for n>=1");
  my(x=1,y=1); for(i=2,n, [x,y] = [sqr(x)+sqr(y), x*y]); x;

  \\ my(x=1,y=1); for(i=2,n, y*=x; x=sqr(x)+sqr(y)); x;
}
A073834(n) =
{
  n>=1 || error("A073834() is for n>=1");
  my(x=1,y=1); for(i=2,n, [x,y] = [sqr(x)+sqr(y),x*y]); y;
}

\\ this code by Michael Somos in A073833
A073833_recursively(n) =
{
 if(n<1, 0,
    n<3, n,
    my(x = A073833_recursively(n-2)^2, y = A073833_recursively(n-1));
    y^2 + x*y - x^2);
}

{
  check_equal(vector(#A073833_samples,n, A073833(n)),
              A073833_samples,  \\ OFFSET=1
              "A073833() vs A073833_samples");

  check_equal(vector(12,n, A073833(n)),
              vector(12,n, A073833_recursively(n)),
              "A073833() vs A073833_recursively()");

  check_equal(vector(12,n,n+=2; A073833(n)),
              vector(12,n,n+=2;
                     my(x = A073833(n-2)^2,
                        y = A073833(n-1));
                     y^2 + x*y - x^2),
              "A073833() recurrence in Michael Somos' code");

  check_equal(vector(12,n,n+=2; A073833(n)),
              vector(12,n,n+=2;
                     my(x = A073833(n-2),
                        y = A073833(n-1));
                     y^2 + y*x^2 - x^4),
              "A073833() recurrence variant");


  \\ A073833(n) is binomial tree 2^(n-1) vpar_matchings_count()
  check_equal(vector(12,n, A073833(n)),
              vector(12,n, my(vpar=vpar_make_binomial_tree(2^(n-1)));
                           vpar_matchings_count(vpar)),
              "A073833(n) binomial tree order n+1 matchings count");

  \\-------------
  check_equal(vector(#A073834_samples,n, A073834(n)),
              A073834_samples,  \\ OFFSET=1
              "A073834() vs A073834_samples");

  \\ A073834 is binomial tree 2^(n-1) sans its root
  check_equal(vector(12,n,A073834(n)),
              vector(12,n,
                     my(vpar=vpar_make_binomial_tree(2^(n-1)));
                     vpar=vpar_delete_vertex(vpar,1);
                     vpar_matchings_count(vpar)),
              "A073834(n) binomial tree order n-1 sans root matchings count");

  check_equal(vector(12,n, A073834(n)),
              vector(12,n, my(vpar=[]);
                     for(i=0,n-2,
                        my(part=vpar_make_binomial_tree(2^i));
                        vpar=vpar_concat_forest(vpar,part));
                     vpar_matchings_count(vpar)),
              "A073834() forest orders 0 to n-1");

  check_equal(vector(12,n, A073834(n)),
              vector(12,n, prod(i=1,n-1, A073833(i))),
              "A073834() as product of A073834()");

  check_equal(vector(12,n,n++; A073833(n-1)),
              vector(12,n,n++; A073834(n) / A073834(n-1)));

  check_equal(vector(10,n,n+=2; A073833(n-1)),
              vector(10,n,n+=2; A073833(n-2)^2 + A073834(n-2)^2));
  check_equal(vector(10,n,n+=2; A073833(n-1)),
              vector(10,n,n+=2; (A073834(n-1) / A073834(n-2))^2 + A073834(n-2)^2));
  \\ division not particularly efficient way to extract A073833(n-1) from A073834 n-1 and n-2
  check_equal(vector(10,n,n+=2; A073834(n)),
              vector(10,n,n+=2; my(x=A073834(n-1),
                                   y=A073834(n-2));  x * ((x/y)^2 + y^2)));

}

\\ num, den -> num^2 + den^2,             num*den
\\ num-den -> num^2 - num*den + den^2,
\\          = (num-den)^2 + num*den

A900035(n) =
{
  n>=1 || error("A900035() is for n>=1");
  if(n==1,return(0));
  my(x=1,y=1); for(i=3,n, y*=x+y; x=x^2+y); x;
}
\\ vector(10,n,n++; A900035(n))
\\ not in OEIS: 1, 3, 19, 651, 696691, 749967308571, 830891350723888949050531
\\ nothing from superseeker
{
  check_equal(vector(6,n, A900035(n)),
              vector(6,n, A073833(n) - A073834(n)),
              "A900035 = any - without");
  check_equal(vector(12,n,n++; A900035(n)),
              vector(12,n,n++; A900035(n-1)^2 + A073834(n)),
              "A900035 and A073834 recurrence 1");

  check_equal(vector(12,n,n++; A073834(n)),
              vector(12,n,n++; (A900035(n-1)+A073834(n-1)) * A073834(n-1)),
              "A900035 and A073834 recurrence 2");
  check_equal(vector(12,n,n++; A073834(n)),
              vector(12,n,n++; A073834(n-1)^2 + A900035(n-1)*A073834(n-1)),
              "A900035 and A073834 recurrence 2b");
}

\\ vector(8,n, digits(A073833(n),3))
\\ vector(8,n, digits(A073834(n),3))
\\ vector(8,n, digits(A900035(n),3))
\\ #A073833_samples
\\ #A073834_samples
\\ for(n=1,10,print1(A900035(n),",");if(n==8||n==9,print()))
\\ quit


\\-----------------------------------------------------------------------------
progress("vpar_minimal_domsets_count()");

check_equal(vpar_minimal_domsets_count([]), 1, \
            "vpar_minimal_domsets_count() empty");
check_equal(vpar_minimal_domsets_count(vpar_make_path(2)), 2);
check_equal(vpar_minimal_domsets_count(vpar_make_path(3)), 2);
check_equal(vpar_minimal_domsets_count(vpar_make_path(4)), 4);

\\ Return 1 if vpar contains any singletons (vertex with no neighbours).
test_any_non_singleton(vpar) =
{
  my(sizes=vpar_component_sizes(vpar));
  for(i=1,#sizes, if(sizes[i]>=2, return(1)));
  0;
}
check_equal(test_any_singletons([]), 0);
check_equal(test_any_singletons([0]), 1);
check_equal(test_any_singletons([0,1]), 0);
check_equal(test_any_singletons([2,0]), 0);
check_equal(test_any_singletons([2,0,0]), 1);
check_equal(test_any_singletons([2,0,0,0]), 1);


test_minimal_domsets_count_by_flags(vpar) =
{
  my(count=0);
  forvec(flags=vector(#vpar,i,[0,1]),
        count += vpar_is_minimal_domset_flags(vpar,flags));
  count;
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got = vpar_minimal_domsets_count(vpar));

     \\ No, count odd or even not like this.
     \\ check_equal(got%2==0, test_any_non_singleton(vpar),
     \\             "vpar_minimal_domsets_count() even");

     if(#vpar <= test_limit_flags_search,
        check_equal(got, test_minimal_domsets_count_by_flags(vpar),
                    Str("vpar_minimal_domsets_count() "
                        "vs test_minimal_domsets_count_by_flags()"));
       );

     my(component_vpar=vpar_component_trees(vpar));
     check_equal(got,
                 prod(i=1,#component_vpar, vpar_minimal_domsets_count(component_vpar[i])),
                 "vpar_minimal_domsets_count() product of components, per addhelp");
  );
}

{
  \\ complete binary tree domsets count
  check_equal(vector(6,h,h--;
                   vpar_minimal_domsets_count(vpar_make_complete_tree(2,h))),
              [1, 2, 4, 41, 1438, 4505183]);
  \\ not in OEIS: 2, 4, 41, 1438, 4505183
  \\ factor(4505183)
  \\ same as perfect domsets count of complete binary tree
}

{
  \\ subdivision of star
  check_equal(vector(8,n,n--; vpar_minimal_domsets_count
                                (vpar_subdivide(vpar_make_star(n)))),
              vector(8,n,n--; 2^max(0,n-1)));
}
{
  \\ binomial tree n
  check_equal(vector(8,n,n--; vpar_minimal_domsets_count
                                (vpar_make_binomial_tree(n))),
              vector(8,n,n--; 2^floor(n/2)));
}
{
  \\ binomial tree level k
  check_equal(vector(7,k,k--;
                   vpar_minimal_domsets_count(vpar_make_binomial_tree(2^k))),
              vector(7,k,k--; if(k==0,1, 2^(2^(k-1)))));
}

{
  \\ n=19 most minimal dominating sets, more than 2^(n/2)
  my(vpar=vpar_subdivide(vpar_make_bistar(5,5)));
  check_equal(#vpar, 19);
  my(got=vpar_minimal_domsets_count(vpar));
  check_equal(got, 737);
  check_equal(got > 2^(#vpar/2), 1);
}
{
  \\ n=19 second most minimal dominating sets, more than 2^(n/2)
  my(vpar=vpar_subdivide(vpar_make_bistar(6,4)));
  check_equal(#vpar, 19);
  my(got=vpar_minimal_domsets_count(vpar));
  check_equal(got, 729);
  check_equal(got > 2^(#vpar/2), 1);
}
{
  \\ n=27 Krzywkowski tree T1 many minimal dominating sets
  \\
  \\ Marcin Krzywkowski, "Trees Having Many Minimal Dominating Sets",
  \\ Information Processing Letters, volume 113, issue 8, April 2013, pages
  \\ 276-279.  http://www.krzywkowski.pl/ipl.pdf

  my(vpar=vpar_subdivide(vpar_make_bistar(7,7)));
  check_equal(#vpar, 27);
  check_equal(vpar_minimal_domsets_count(vpar), 12161);
}
{
  \\ n=32 most minimal dominating sets, n even with more than 2^(n/2)
  my(vpar=vpar_join(vpar_subdivide(vpar_make_bistar(4,5)),1,
                    vpar_subdivide(vpar_make_bistar(4,4))));
  check_equal(#vpar, 32);
  my(got=vpar_minimal_domsets_count(vpar));
  check_equal(got, 65960);
  check_equal(got > 2^(#vpar/2), 1);
}

{
  \\ n=34 most minimal dominating sets, n even with more than 2^(n/2)
  my(vpar=vpar_join(vpar_subdivide(vpar_make_bistar(4,5)),1,
                    vpar_subdivide(vpar_make_bistar(4,5))));
  check_equal(#vpar, 34);
  my(got=vpar_minimal_domsets_count(vpar));
  check_equal(got, 134432);
  check_equal(got > 2^(#vpar/2), 1);
}
{
  \\ n=34 most minimal dominating sets, n even with more than 2^(n/2)
  my(vpar=vpar_join(vpar_subdivide(vpar_make_bistar(5,4)),1,
                    vpar_subdivide(vpar_make_bistar(5,4))));
  check_equal(#vpar, 34);
  my(got=vpar_minimal_domsets_count(vpar));
  check_equal(got, 132400);
  check_equal(got > 2^(#vpar/2), 1);
}


\\-----------------------------------------------------------------------------
progress("vpar_indomsets_count()");

check_equal(vpar_indomsets_count([]), 1, "vpar_indomsets_count() empty");
check_equal(vpar_indomsets_count([0]), 1, "vpar_indomsets_count() singleton");
check_equal(vpar_indomsets_count(vpar_make_path(2)), 2);
check_equal(vpar_indomsets_count(vpar_make_path(3)), 2);
check_equal(vpar_indomsets_count(vpar_make_path(4)), 3);

test_indomsets_count_by_flags(vpar) =
{
  my(count=0);
  forvec(flags=vector(#vpar,i,[0,1]),
        count += vpar_is_indomset_flags(vpar,flags));
  count;
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got=vpar_indomsets_count(vpar));

     if(#vpar <= test_limit_flags_search,
        check_equal(vpar_indomsets_count(vpar),
                    test_indomsets_count_by_flags(vpar),
               "vpar_indomsets_count() vs test_indomsets_count_by_flags()"));
  );
}

{
  \\ A058925 = number of independent dominating sets in labelled trees
  \\ of n nodes
  \\
  \\ Changwoo Lee and Sanghee Cho, "The Number of Maximal Independent Sets
  \\ of Unlabelled Trees", J. Appl. Math. & Computing, volume 18, number 1-2,
  \\ 2005, pages 639-647
  \\ http://www.mathnet.or.kr/mathnet/kms_tex/982463.pdf
  \\
  my(forest=-1); \\ root=1
  for(n=1, test_limit_trees_search - forest,
     my(total=0);
     forvec(vec=vpar_forvec_from(n,forest),
           my(vpar=vpar_from_upascend(vec));
           total += vpar_indomsets_count(vpar));
     check_equal(total,
                 A058925_samples[n],  \\ OFFSET=1
                 "A058925 indomsets in labelled trees"));
}



\\-----------------------------------------------------------------------------
progress("test_A084089() values (6m+1)*4^k");

\\ A084089 == 1 mod 3 and even low 0-bits
\\   (6m-5)*4^k  = (6m+1)*4^k
\\ 1, 4, 7, 13, 16, 19, 25, 28, 31, 37, 43, 49, 52, 55, 61, 64, 67, 73,

\\                                                           64
\\                        16
\\     *          4                28               52
\\ * *   * *  * 1   7  13    19 25    31 37   43 49    55 61    67 73

\\ test whether a occurs as some a(n) = a in A084089
A084089_pred(a) = a>=1 && a%3==1 && valuation(a,2)%2==0;
{
  check_equal(select(A084089_pred, [0..A084089_samples[#A084089_samples]]),
              A084089_samples,
              "A084089_pred() vs A084089_samples");
}

\\ (6p-5)*4^(k-i) < 1*4^k
\\ (6p-5) < 4^i
\\ 6p < 4^i+5
\\ p < (4^i+5)/6
A084089_first_i(i) =
{
  i>=1 || error("A084089_first_i() is for i>=1");
  (4^i+2)/6;
}
\\ vector(10,i, A084089_first_i(i))
{
  for(k=1,50,
     for(i=1,k,
        my(p=A084089_first_i(i));
        check_equal((6*p-5)*4^(k-i) < 4^k, 1);
        p++;
        check_equal((6*p-5)*4^(k-i) > 4^k, 1);

        \\ (6p+1)*4^(k-i) < 4^k for p < (4^i+2)/6
        my(p = (4^i+2)/6 - 1);
        check_equal( (6*p+1)*4^(k-i) < 4^k,  1);
        p++;
        check_equal( (6*p+1)*4^(k-i) < 4^k,  0)));

  \\ A007583 (2^(2*n + 1) + 1)/3.
  check_equal(vector(#A007583_samples,i,i--; A084089_first_i(i+1)),
              A007583_samples,  \\ OFFSET=0
              "A084089_first_i() vs A007583_samples");
}
A007583(n) = (2^(2*n+1) + 1)/3;
{
  check_equal(vector(#A007583_samples,n,n--; A007583(n)),
              A007583_samples,  \\ OFFSET=0
              "A007583() vs A007583_samples");
}

\\ vector(10,k, sum(i=1,k, A084089_first_i(i)))

\\ 0-based: 0, 1, 4, 15, 58, 229
\\ 1-based: 1, 2, 5, 16, 59, 230
\\ not in OEIS: 1, 2, 5, 16, 59, 230
\\ not A151746 which is half, for k>=1
\\
A084089_first(k) = (2*4^k + 3*k + 7)/9;
{
  my(n=1,    \\ A084089_samples OFFSET=1
     k=0);
  for(a=1,2000,
     A084089_pred(a)||next;
     if(valuation(a,4)==k,
        check_equal(A084089_first(k), n,
                    "A084089_first() vs A084089_pred()");
        k++);
     n++);

  check_equal(vector(100,k,k--; A084089_first(k)),
              vector(100,k,k--; 1 + sum(i=1,k, A084089_first_i(i))),
              "A084089_first() vs sum A084089_first_i()");

  \\ A151746
  check_equal(vector(#A151746_samples-1,k,k--; A084089_first(k+1)),
              A151746_samples[^1] / 2, \\ OFFSET=0
              "A084089_first() vs A151746_samples / 2");
}

A160156(n) = (3*n + 1 + 2^(2*n+3))/9;
{
  check_equal(vector(#A160156_samples,n,n--; A160156(n)),
              A160156_samples, \\ OFFSET=0
              "A160156() vs A160156_samples");

  check_equal(vector(100,k,k--; A084089_first(k)),
              vector(100,k,k--; A160156(k-1) + 1),
              "A084089_first() vs A160156 + 1");
}

\\ (2*4^k + 3*k - 2)/9 < n < (2*4^(k+1) + 3*(k+1) - 2)/9
\\ 2/9*4^k + 3/9*k - 2/9 < n < 2/9*4^(k+1) + 3/9*(k+1) - 2/9
\\ n < 2/9*4^(k+1) + 3/9*(k+1) - 2/9
\\ 9/2*n < 4^(k+1) + 2/27*k + 2/81
\\
check_equal(1/4 - 2/9,  1/36);
\\ 1/36*4^k >= 1/3*k
\\ 4^k >= 12*k
my(k=1); check_equal([4^k,12*k],  [4, 12]);
my(k=2); check_equal([4^k,12*k],  [16, 24]);
my(k=3); check_equal([4^k,12*k],  [64, 36]);
\\
A084089_k(n) = logint(8*n,4);
{
  my(bad=0);
  for(n=1,2^10,
     my(k=A084089_k(n));
     if(! (A084089_first(k-1)-1 < n),
        print("A084089_k too big at n=",n,"  k="k" first="A084089_first(k));
        if(bad++ > 50, break));
     if(! (n < A084089_first(k+1)-1),
        print("A084089_k too small at n=",n,"  k="k" first="A084089_first(k));
        if(bad++ > 50, break));
  );
  check_equal(bad,0, "A084089_k() vs A084089_first()");
}
\\ vector(200,n,n--; A084089_k(n))

\\ A084089_step_i(k)
\\ 4^k < (6p-5)*4^(k-i) < 7*4^k
\\ 4^i < (6p-5) < 7*4^i
\\ 4^i+5 < 6p < 7*4^i+5
\\ (4^i+5)/6 < p < (7*4^i+5)/6
\\ (4^i+5)/6 < p < (4^i+5)/6 + 4^i
my(k=1); check_equal([4^k, 7*4^k], [4, 28]);
\\ in between  4,   7, 13,  16,  19, 25,   28
\\                          ^^
\\                        is k=2
check_equal(  7, 6*2-5);
check_equal( 13, 6*3-5);
check_equal( 19, 6*4-5);
check_equal( 25, 6*5-5);

A084089_step(k) = (4^(k+1) - 1)/3;
{
  \\ distance between first and second k, which are 4^k and 7*4^k
  check_equal(vector(10,k,k--; (7*4^k)%6),
              [1, 4, 4, 4, 4, 4, 4, 4, 4, 4]);
  for(k=0,8,
     my(n=-1,  \\ OFFSET=0, after pre-increment
        seen=0,
        first);
     for(a=1,2000000,
        A084089_pred(a) || next;
        valuation(a,4) <= k || next;
        n++;
        if(valuation(a,4)==k,
           seen++;
           if(seen==1, first=n; next);
           check_equal(first, A084089_first(k)-1,
                       "A084089_first() in step search");
           check_equal(A084089_step(k), n-first,
                       "A084089_step() vs A084089_pred count");
           next(2))));

  check_equal(vector(10,k,k--; A084089_step(k)),
              vector(10,k,k--; 1 + sum(i=1,k, 4^i)),
              "A084089_first() vs sum A084089_first_i()");

  \\ A002450 (4^n - 1)/3
  check_equal(vector(#A002450_samples,k,k--; A084089_step(k-1)),
              A002450_samples, \\ OFFSET=0
              "A084089_step() vs A002450_samples");
}

A084089_extra(k) = (10*4^k - 3*k - 10)/9;
{
  check_equal(vector(100,k,k--; A084089_extra(k)),
              vector(100,k,k--; A084089_step(k) - A084089_first(k)),
              "A084089_extra() vs step and first");

  \\ vector(8,k, A084089_extra(k))
  \\ not in OEIS: 3, 16, 69, 282, 1135, 4548, 18201, 72814
  \\ A172447 is extra + 1
  check_equal(vector(#A172447_samples,k,k--; A084089_extra(k) + 1),
              A172447_samples, \\ OFFSET=0
              "A084089_extra() vs A172447_samples");
}
\\ vector(20,k, A084089_extra(k))

A084089_by_parts(n) =
{
  \\ print("A084089 n="n);
  n>=1 || error("A084089_by_parts() is for n>=1");
  my(q,r);
  forstep(k=A084089_k(n), 1, -1,
     \\ print("  k="k"  n="n"  first "A084089_first(k)" step "A084089_step(k)" divrem "Vec(divrem(n - A084089_first(k) + A084089_step(k), A084089_step(k))));
     [q,r]=divrem(n + A084089_extra(k), A084089_step(k));
     q>=0 || error("A084089_by_parts() oops, negative quotient");
     if(r==0,
        \\ print("  exact q="qr[1]);
        return((6*q-5)*4^k));
     n -= q);
  \\ print("  final k=0 n="n);
  6*n-5;
}
{
  check_equal(vector(10,k, A084089_by_parts(A084089_first(k))),
              vector(10,k, 4^k),
              "A084089_by_parts() at A084089_first()");
  check_equal(vector(#A084089_samples,n, A084089_by_parts(n)),
              A084089_samples,  \\ OFFSET=1
              "A084089_by_parts() vs A084089_samples");
}

\\ based on watching for when (6j-5)*4^k for given k, then dividing out by
\\ how many of them below n
A084089(n) =
{
  n>=1 || error("A084089() is for n>=1");
  my(q,r);
  forstep(k=logint(8*n,4),1,-1,
     [q,r] = divrem(n + A084089_extra(k), A084089_step(k));
     q>=0 || error("A084089() oops, negative quotient");
     if(r==0, return((6*q-5)*4^k));
     n-=q);
  6*n-5;
}

{
  check_equal(vector(20,n, A084089(n)),
              vector(20,n, A084089_by_parts(n)),
              "A084089() vs A084089_by_parts()");
  check_equal(vector(10,k, A084089(A084089_first(k))),
              vector(10,k, 4^k),
              "A084089() at A084089_first()");
  check_equal(vector(#A084089_samples,n, A084089(n)),
              A084089_samples,  \\ OFFSET=1
              "A084089() vs A084089_samples");
}
{
  my(n=1,  \\ A084089_samples OFFSET=1
     target=A084089(n),
     bad=0);
  for(a=1,2^16, \\ prospective a(n)
     \\ if(bitand(a,32767)==0, print1("A084089 a="a,Strchr(13)));
     my(by_pred=A084089_pred(a),
        by_target=(a==target));
     if(by_pred != by_target,
        print("n="n" a="a"  by_pred="by_pred" vs by_target "by_target" expecting "target);
        if(bad++ > 10, break));
     if(by_target,
        n++;
        target=A084089(n)));
  check_equal(bad,0);
}


\\-----------------------------------------------------------------------------
progress("vpar_is_perfect_domset()");
progress("vpar_is_perfect_domset_flags()");

check_equal(vpar_is_perfect_domset([],[]), 1, \
            "vpar_is_perfect_domset() empty");
check_equal(vpar_is_perfect_domset([2,1],[1]), 0, \
            "vpar_is_perfect_domset() 2-cycle, one");
check_equal(vpar_is_perfect_domset([2,1],[1,2]), 1, \
            "vpar_is_perfect_domset() 2-cycle, two");

check_equal(vpar_is_perfect_domset_flags([],[]), 1);

test_is_perfect_domset_flags(vpar,flags) =
{
  my(dominated=Vecsmall(flags,#vpar));  \\ all of flags itself dominated
  for(v=1,#flags,
     my(p=vpar[v]);
     if(p, dominated[p] += flags[v];    \\ p dominated by v
           dominated[v] += flags[p]));  \\ v dominated by p
  for(v=1,#flags,
     if(!dominated[v], return(0));
     if(!flags[v] && dominated[v]>=2, return(0)));
  1;
}
test_is_perfect_domset_flags_by_neighbours(vpar,flags) =
{
  for(v=1,#vpar,
     if(!flags[v],
        my(neighbours=vpar_neighbours_of_vertex(vpar,v));
        if(hammingweight(vecextract(flags,neighbours)) != 1,
           return(0))));
  1;
}
{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     for(rep=1,max(1,#vpar),
        my(flags=vpar_random_flags(vpar),
           set=test_flags_to_Set(flags),
           got=vpar_is_perfect_domset_flags(vpar,flags),
           by_test=test_is_perfect_domset_flags(vpar,flags),
           by_neighbours=test_is_perfect_domset_flags_by_neighbours(vpar,flags));
        \\ print(vpar" flags "flags);
        check_equal(by_test, by_neighbours,
                    "test_is_perfect_domset_flags() vs test_is_perfect_domset_flags_by_neighbours()");
        check_equal(got, by_test,
                    "vpar_is_perfect_domset_flags() vs test_is_perfect_domset_flags");
        check_equal(vpar_is_perfect_domset(vpar,set),
                    by_test,
                    "vpar_is_perfect_domset() vs test_is_perfect_domset_flags");
     ));
}


\\-----------------------------------------------------------------------------
progress("vpar_totdomnum()");

check_equal(vpar_totdomnum([]), 0, \
            "vpar_totdomnum() empty");
check_equal(vpar_totdomnum([0]), -1, \
            "vpar_totdomnum() singleton, none");
check_equal(vpar_totdomnum([0,1]), 2, \
            "vpar_totdomnum() path-2");
check_equal(vpar_totdomnum([0,1,1]), 2, \
            "vpar_totdomnum() star-3");
check_equal(vpar_totdomnum([2,3,0]), 2, \
            "vpar_totdomnum() path-3 reverse");

check_equal(vpar_totdomnum([0,1,0]), -1, \
            "vpar_totdomnum() path-2 and singleton, none");

{
  \\ star arms 1,3,3 equal maximum totdomnum of n=8 connected graphs
  my(vpar=[0,1,1,3,4,1,6,7]);
  check_equal(#vpar, 8);
  check_equal(vpar_totdomnum(vpar), 5);
}
{
  \\ star arms 2,2,3 equal maximum totdomnum of n=8 connected graphs
  my(vpar=[0,1,2,1,4,1,6,7]);
  check_equal(#vpar, 8);
  check_equal(vpar_totdomnum(vpar), 5);
}

\\ This form by separate flags tracking whether a vertex is present etc.
\\
test_totdomnum_by_separate(vpar) =
{
  my(seq=vpar_INTERNAL_seq_upwards(vpar),
     present   =vectorsmall(#vpar),    \\ v in the set
     dominated =vectorsmall(#vpar),    \\ v dominated by below
     have_child=vectorsmall(#vpar),    \\ v has a child
     ret=0);
  for(i=1,#seq,
     my(v=seq[i], p=vpar[v]);
     if(p,
        if(!dominated[v],
           ret += !present[p];  \\ take parent to dominate leaf
           present[p]=1);
        have_child[p]=1;
        dominated[p] += present[v];  \\ child dominates parent
        ,
        !dominated[v],    \\ no parent and not already dominated
        if(have_child[v],
           ret++,         \\ take child to dominate parent
           return(-1)))); \\ no child, isolated vertex, cannot totdom
  ret;
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     check_equal(test_totdomnum_by_separate(vpar),
                 vpar_totdomnum(vpar));
  );
}

\\ A004524
A004524(n) = floor(n/2) - (n%4==2);
{
  check_equal(vector(#A004524_samples,n,n--; A004524(n)),
              A004524_samples,  \\ OFFSET=0
              "A004524() vs A004524_samples");
}

\\ cf Eric Weisstein in A004524 same totdomnum for complete graph or cycle
\\ graph
test_totdomnum_of_path(n) = if(n==1,-1, floor((n+2)/2) - (n%4==0));
{
  check_equal(vector(50,n,n--; vpar_totdomnum(vpar_make_path(n))),
              vector(50,n,n--; test_totdomnum_of_path(n)),
              "vpar_totdomnum() of path");

  check_equal(vector(50,n,n--; vpar_totdomnum(vpar_make_path(n))),
              vector(50,n,n--; if(n==1,-1, A004524(n+2))),
              "vpar_totdomnum() of path");

  \\ for n>=4
  check_equal(vector(50,n,n+=3; A004524(n)),
              vector(50,n,n+=3; vpar_totdomnum(vpar_make_path(n-2))),
              "A004524() as vpar_totdomnum() of path");
}

{
  \\ about the same as plain domnum
  check_equal(vector(50,n,n--; vpar_totdomnum(vpar_make_binomial_tree(n))),
              vector(50,n,n--; if(n==0,0, n==1,-1, n<=3,2, floor(n/2))),
              "vpar_totdomnum() of binomial tree n vertices");
}

\\-----------------------------------------------------------------------------
progress("vpar_semitotdomnum()");

check_equal(vpar_semitotdomnum([]), 0, \
            "vpar_semitotdomnum() empty");
check_equal(vpar_semitotdomnum([0]), -1, \
            "vpar_semitotdomnum() singleton, none");
check_equal(vpar_semitotdomnum([0,1]), 2, \
            "vpar_semitotdomnum() path-2");
check_equal(vpar_semitotdomnum([0,1,1]), 2, \
            "vpar_semitotdomnum() star-3");
check_equal(vpar_semitotdomnum([2,3,0]), 2, \
            "vpar_semitotdomnum() path-3 reverse");

check_equal(vpar_semitotdomnum([0,1,0]), -1, \
            "vpar_semitotdomnum() path-2 and singleton, none");

\\      1
\\      |
\\      2*     v=2 with 2 106 without 2 106 210
\\     / \
\\    3   4*
\\        |
\\        5
\\      4
\\      |
\\      3*     v=3 with 3 105 without 2 105 208
\\     / \
\\    2   5*   v=2 with 105 1 without 105 105 0
\\        |
\\        1
check_equal(vpar_semitotdomnum([0,1,2,2,4]), 2);
check_equal(vpar_semitotdomnum([5, 3, 4, 0, 3]), 2);

{
  \\ Path semitotdomnum per
  \\ W. Goddard, M. A. Henning, C. A. McPillan, "Semitotal Domination In
  \\ Graphs". Utilitas Mathematica, volume 94, 2014, pages 67-81.
  \\
  check_equal(vector(20,n,n--; vpar_semitotdomnum(vpar_make_path(n))),
              vector(20,n,n--; if(n==1,-1, n==2,2, ceil(2*n/5))),
              "vpar_semitotdomnum() of path");

  \\ vector(20,n,n++; test_semitotdomnum_of_path(n))
  \\ A057354 floor(2*n/5).

  check_equal(vector(100,n,n--; n+=2; floor(2*n/5)),
              vector(100,n,n--; ceil(2*n/5)),
              "floor 2*n/5 of n+2 is same as ceil 2*n/5");
}

test_semitotdomnum_of_complete_binary_tree(h) = \
  if(h==0,-1, 1/7*(2^(h+2) + test_vector_modulo([3,6,-2],h)));
{
  check_equal(vector(14,h,h--; vpar_semitotdomnum(vpar_make_complete_tree(2,h))),
              vector(14,h,h--; test_semitotdomnum_of_complete_binary_tree(h)),
              "vpar_semitotdomnum() of complete binary tree");
  \\              2, 2, 5, 10, 18, 37, 74, 146, 293, 586, 1170, 2341
  \\ not in OEIS:              18, 37, 74, 146, 293, 586, 1170, 2341

  \\ -> 2/7 of vertices
  \\ my(h=16); test_semitotdomnum_of_complete_binary_tree(h)/#vpar_make_complete_tree(2,h)*1.0

  check_equal(vector(10,h,h++; test_semitotdomnum_of_complete_binary_tree(h)),
              vector(10,h,h++; 2*test_semitotdomnum_of_complete_binary_tree(h-1)
                               + test_vector_modulo([1,0,-2],h)),
              "test_semitotdomnum_of_complete_binary_tree() recurrence");
}
\\ [-1, 0]    0
\\ [2, 3]     1
\\ [2, 1]     2
\\ [5, 3]     3
\\ for(h=0,7, print(test_semitotdomnum_and_count_by_flags(vpar_make_complete_tree(2,h))));

{
  \\ about the same as plain domnum
  check_equal(vector(50,n,n--; vpar_semitotdomnum(vpar_make_binomial_tree(n))),
              vector(50,n,n--; if(n==0,0, n==1,-1, n<=3,2, floor(n/2))),
              "vpar_semitotdomnum() of binomial tree n vertices");
}


\\-----------------------------------------------------------------------------
progress("vpar_is_semitotdomset()");
\\        vpar_is_semitotdomset_flags()

check_equal(vpar_is_semitotdomset([0,1,2],[1,2,3]), 1, \
            "vpar_is_semitotdomset() path-3 all");
check_equal(vpar_is_semitotdomset([0,1,2],[1,3]), 1, \
            "vpar_is_semitotdomset() path-3 ends");
check_equal(vpar_is_semitotdomset([0,1,1],[2,3]), 1, \
            "vpar_is_semitotdomset() star-3 children");

check_equal(vpar_is_semitotdomset_flags([],[]), 1, \
            "vpar_is_semitotdomset_flags() empty");
check_equal(vpar_is_semitotdomset_flags([0],[0]), 0, \
            "vpar_is_semitotdomset_flags() singleton");
check_equal(vpar_is_semitotdomset_flags([0],[1]), 0, \
            "vpar_is_semitotdomset_flags() singleton");

check_equal(vpar_is_semitotdomset_flags([0,1],[1,0]), 0, \
            "vpar_is_semitotdomset_flags() path-2");
check_equal(vpar_is_semitotdomset_flags([0,1],[1,1]), 1, \
            "vpar_is_semitotdomset_flags() path-2");

{
  \\    4
  \\    |
  \\    3
  \\   / \
  \\  1   2
  \\      |
  \\      5
  check_equal(vpar_is_semitotdomset([3, 3, 4, 0, 2], [1, 2, 4]), 1,
              "vpar_is_semitotdomset() star-3 children");
}

test_is_semitotdomset_by_path_length(vpar,set) =
{
  vpar_is_domset(vpar,set) || return(0);

  \\ each in the set must have another at distance 1 or 2
  \\ this doesn't allow for self-loops
  if(vpar_is_cyclic(vpar),
     error("test_is_semitotdomset_by_path_length() is not for cyclic"));
  for(i=1,#set,
     for(j=1,#set,
        my(l=vpar_path_length(vpar,set[i],set[j]));
        if(l==1 || l==2,
           next(2)));
     return(0));
  1;
}

\\ Return 1 if vertices u,v are distance 2 neighbours, meaning there is a
\\ path of 2 distinct edges going from u to v.
\\ A vertex v is its own neighbour2 when v is part of a 2-cycle.
\\
\\ If u,v are neighbours, and one of them is a self-loop, then the 2 edges
\\ are the edge between followed by the self-loop.
\\
\\    v    u      *
\\    |    |     / \    loop   loop
\\    *    *    u   v     v      u
\\    |    |              |      |
\\    u    v              u      v
\\
test_is_neighbour2(vpar,u,v) =
{
  my(p=vpar[u]);
  if(p,
     if(p!=u && vpar[p]==v, return(1));  \\ u up twice
     if(u!=v && vpar[v]==p, return(1));  \\ common parent
    );
  p=vpar[v];
  if(p && p!=v && vpar[p]==u, return(1)); \\ v up twice
  0;
}
{
  check_equal(test_is_neighbour2([0,1,2],1,3), 1);

  check_equal(test_is_neighbour2([0,1],1,2), 0);
  check_equal(test_is_neighbour2([0,1],2,1), 0);

  check_equal(test_is_neighbour2([2,0],1,2), 0);
  check_equal(test_is_neighbour2([2,0],2,1), 0);
  check_equal(test_is_neighbour2([2,0],1,1), 0);
  check_equal(test_is_neighbour2([2,0],2,2), 0);

  \\ self-loop
  check_equal(test_is_neighbour2([1],1,1), 0);

  \\ self-loop above
  check_equal(test_is_neighbour2([1,1],1,2), 1);
  check_equal(test_is_neighbour2([1,1],2,1), 1);
  check_equal(test_is_neighbour2([1,1,2],1,3), 1);
  check_equal(test_is_neighbour2([1,1,2,3],1,4), 0);

  \\ common parent
  check_equal(test_is_neighbour2([0,1,1],2,3), 1);
  check_equal(test_is_neighbour2([0,1,1],2,3), 1);
}

\\ Return a vector [ [child,parent], [child,parent], ... ]
test_edges(vpar) =
{
  my(edges=vector(vpar_num_edges(vpar)), upto=0);
  for(v=1,#vpar,
     if(vpar[v],
        edges[upto++] = [v,vpar[v]]));
  edges;
}

\\ If there is a 2-edge path leading from u to v then return those edges
\\ [ [u,t], [v,t] ].  It's unspecified which way around each edge is, but
\\ the first contains u, the second contains v, and the other is the same
\\ in each.
\\ It's possible for t==u or t==v when a self-loop is used in the path.
\\ If no such 2-edge path, then return symbol 'none.
\\
test_neighbour2_edges(vpar,u,v) =
{
  my(edges=test_edges(vpar));
  for(i=1,#edges,
     if(test_vector_search(edges[i],u),
        my(t=edges[i][1]);
        if(t==u,t=edges[i][2]);
        for(j=1,#edges,
           if(i==j,next);
           if(edges[j]==[t,v] || edges[j]==[v,t],
              return([edges[i],edges[j]])))));
  'none;
}

\\ Return 1 if there is a path of length 2 from u to v, or return 0 if not.
\\ This means two different edges traversed to go from u to v.
test_is_neighbour2_by_edges(vpar,u,v) = \
  test_neighbour2_edges(vpar,u,v) != 'none;

{
  \\   loop
  \\     1
  \\     |
  \\     4
  \\    / \
  \\   2   5
  \\       |
  \\       3
  my(vpar=[1, 4, 5, 1, 4],
     u=5,
     v=4);
  check_equal(vpar_is_neighbour(vpar,u,v), 1);
  check_equal(test_is_neighbour2(vpar,u,v), 0);
  check_equal(test_neighbour2_edges(vpar,u,v), 'none,
              "test_neighbour2_edges() 4 5");
  check_equal(test_is_neighbour2_by_edges(vpar,u,v), 0,
              "test_is_neighbour2_by_edges() 4 5");
}
{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     if(#vpar==0,next);
     for(rep=1,max(1,#vpar),
        my(u=vpar_random_vertex(vpar),
           v=vpar_random_vertex(vpar));
        check_equal(test_is_neighbour2(vpar,u,v),
                    test_is_neighbour2_by_edges(vpar,u,v),
        "test_is_neighbour2() vs test_is_neighbour2_by_edges()")));
}

test_is_neighbour2_in_set(vpar,v,set) =
{
  my(neighbours=vpar_neighbours_of_vertex(vpar,v));
  for(i=1,#set,
     if(test_is_neighbour2(vpar,set[i],v),
        return(1)));
  0;
}


test_is_semitotdomset_by_neighbours(vpar,set) =
{
  \\ print("test_is_semitotdomset_by_neighbours() "vpar" "set);
  for(v=1,#vpar,
     if(test_is_neighbour_in_set(vpar,v,set),
        \\ print("  v="v" neighbour");
        next);
     if(setsearch(set,v) && test_is_neighbour2_in_set(vpar,v,set),
        \\ print("  v="v" in set, neighbour2");
        next);
     \\ print("  v="v" not");
     return(0));
  1;
}

{
  \\    3
  \\    |
  \\    2
  \\    |
  \\    1
  my(vpar=[2, 3, 0],
     set=[2]);
  check_equal(vpar_is_semitotdomset(vpar,set), 0,
              "vpar_is_semitotdomset() path-2 middle");
  check_equal(test_is_semitotdomset_by_path_length(vpar,set), 0,
              "test_is_semitotdomset_by_path_length() path-2 middle");
}

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     for(rep=1,max(1,#vpar),
        my(flags=vpar_random_flags(vpar),
           set=test_flags_to_Set(flags),
           got=vpar_is_semitotdomset(vpar,set));
        \\ print(vpar" "set);

        if(!vpar_is_cyclic(vpar),
           check_equal(got,
                       test_is_semitotdomset_by_path_length(vpar,set),
        "vpar_is_semitotdomset() vs test_is_semitotdomset_by_path_length()"));

        check_equal(got,
                    test_is_semitotdomset_by_neighbours(vpar,set),
        "vpar_is_semitotdomset() vs test_is_semitotdomset_by_neighbours()");

        check_equal(got,
                    vpar_is_semitotdomset_flags(vpar,flags),
        "vpar_is_semitotdomset() vs vpar_is_semitotdomset_flags()");
  ));
}


\\-----------------------------------------------------------------------------
progress("vpar_semitotdomnum_and_count()");

\\ cf A001638 totdomsets count of n-cycle

check_equal(vpar_semitotdomnum_and_count([]), [0,1], \
            "vpar_semitotdomnum_and_count() empty");
check_equal(vpar_semitotdomnum_and_count([0]), [-1,0]);
check_equal(vpar_semitotdomnum_and_count([0,1]), [2,1], \
            "vpar_semitotdomnum_and_count() path-2");

check_equal(test_semitotdomnum_and_count_by_flags([0,1,2]), [2,3], \
            "test_semitotdomnum_and_count_by_flags() path-3");
check_equal(vpar_semitotdomnum_and_count([0,1,2]), [2,3], \
            "vpar_semitotdomnum_and_count() path-3");
check_equal(vpar_semitotdomnum_and_count([0,1,1]), [2,3]);

check_equal(vpar_semitotdomnum_and_count([0,1,0,3]), [4,1], \
            "vpar_semitotdomnum_and_count() two path-2");

test_semitotdomnum_and_count_by_flags(vpar) =
{
  if(#vpar==0,return([0,1]));  \\ empty set only
  my(semitotdomnum=#vpar+1,count=0);
  forvec(flags=vector(#vpar,i,[0,1]),
        if(vpar_is_semitotdomset_flags(vpar,flags),
           my(size=vecsum(flags));
           if(size<semitotdomnum,  semitotdomnum=size;count=1,  \\ new low
              size==semitotdomnum, count++)));           \\ equal low
  if(count==0,semitotdomnum=-1);
  [semitotdomnum,count];
}
check_equal(test_semitotdomnum_and_count_by_flags([]), [0,1], \
            "test_semitotdomnum_and_count_by_flags() empty");
check_equal(test_semitotdomnum_and_count_by_flags([0]), [-1,0], \
            "test_semitotdomnum_and_count_by_flags() singleton");
check_equal(test_semitotdomnum_and_count_by_flags(vpar_make_path(2)), [2,1]);
check_equal(test_semitotdomnum_and_count_by_flags(vpar_make_path(3)), [2,3]);
check_equal(test_semitotdomnum_and_count_by_flags(vpar_make_path(4)), [2,3]);

{
  \\  4 <- 1 <- 2 <- 5 <- 3
  \\       *         *
  my(vpar=[4, 1, 5, 0, 2]);
  check_equal(test_semitotdomnum_and_count_by_flags(vpar_make_path(5)), [2,1]);
  check_equal(test_semitotdomnum_and_count_by_flags(vpar), [2,1]);
  check_equal(vpar_semitotdomnum(vpar_make_path(5)), 2,
              "vpar_semitotdomnum() path-5");
  check_equal(vpar_semitotdomnum(vpar), 2);
}

{
  for(n=0,10,
     my(vpar=vpar_make_path(n));
     check_equal(vpar_semitotdomnum(vpar),
                 test_semitotdomnum_and_count_by_flags(vpar)[1],
                 Str("vpar_semitotdomnum() vs test_semitotdomnum_and_count_by_flags() path "n));
  );
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     my(by_flags=test_semitotdomnum_and_count_by_flags(vpar));

     check_equal(vpar_semitotdomnum(vpar),by_flags[1],
                 "vpar_semitotdomnum() vs test_semitotdomnum_and_count_by_flags()");

     check_equal(vpar_semitotdomnum_and_count(vpar), by_flags,
       "vpar_semitotdomnum_and_count() vs test_semitotdomnum_and_count_by_flags()");
  );
}

check_equal(poldisc(n^2 + 6*n + 17), -32);
check_equal(poldisc(n^2 + 3*n + 6),  -15);
check_equal(subst(1/6 * m*(m^2 + 6*m + 17), m, (n-1)/5), \
            1/750 * (n-1)*(n^2 + 28*n + 396));

test_semitotdomnum_count_of_path(n) =
{
  my(m=n\5);
  if(n==2, 1,
     n%5 == 0, 1,
     n%5 == 1, 1/6 * m*(m^2 + 6*m + 17),
     n%5 == 2, m,
     n%5 == 3, 1/24 * (m^4 + 14*m^3 + 71*m^2 + 106*m + 72),
     n%5 == 4, 1/2 * (m^2 + 3*m + 6),
     error());
}
{
  check_equal(vector(200,n,n--;
                     vpar_semitotdomnum_and_count(vpar_make_path(n))[2]),
              vector(200,n,n--; test_semitotdomnum_count_of_path(n)),
              "vpar_semitotdomnum_and_count() count of path");
}
{
  my(limit=test_limit_flags_search);
  check_equal(vector(limit,n,n--;
                 test_semitotdomnum_and_count_by_flags(vpar_make_path(n))[2]),
              vector(limit,n,n--; test_semitotdomnum_count_of_path(n)),
              "vpar_semitotdomnum_and_count() count of path");
}
\\ vector(18,n,n++; vpar_semitotdomnum_and_count(vpar_make_path(n))[2])
\\ not in OEIS: 1, 3, 3, 1, 4, 1, 11, 5, 1, 11, 2, 29, 8, 1, 22, 3, 62, 12


\\ A300738
\\ totdom count cycle


\\-----------------------------------------------------------------------------
progress("vpar_is_totdomset()");

check_equal(vpar_is_totdomset([],[]), 1, \
            "vpar_is_totdomset() empty is dominated");

check_equal(vpar_is_totdomset([0],[]), 0, \
            "vpar_is_totdomset() singleton not dominated");
check_equal(vpar_is_totdomset([0],[1]), 0, \
            "vpar_is_totdomset() singleton not dominated");

check_equal(vpar_is_neighbour([1],1,1), 1, \
            "vpar_is_neighbour() self-loop is own neighbour");
check_equal(vpar_is_totdomset([1],[1]), 1, \
            "vpar_is_totdomset() self-loop dominated");

check_equal(vpar_is_totdomset([2,1],[1]), 0, \
            "vpar_is_totdomset() 2-cycle");
check_equal(vpar_is_totdomset([2,1],[1,2]), 1, \
            "vpar_is_totdomset() 2-cycle");

test_is_totdomset_by_neighbours(vpar,set) =
{
  for(v=1,#vpar,
     test_is_neighbour_in_set(vpar,v,set) || return(0));
  1;
}

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     for(rep=1,max(1,#vpar),
        my(flags=vpar_random_flags(vpar),
           set=test_flags_to_Set(flags),
           got=vpar_is_totdomset(vpar,set));
        check_equal(got,
                    test_is_totdomset_by_neighbours(vpar,set),
                    "vpar_is_totdomset() vs test_is_totdomset_by_neighbours()");
        check_equal(got,
                    vpar_is_totdomset_flags(vpar,flags),
                    "vpar_is_totdomset() vs vpar_is_totdomset_flags()");

        \\ per addhelp() totdomsets subset of domsets
        check_equal(got <= vpar_is_domset(vpar,set),
                    1,
                    "vpar_is_totdomset() <= vpar_is_domset()")));
}


\\-----------------------------------------------------------------------------
progress("vpar_is_totdomset_flags()");

check_equal(vpar_is_totdomset_flags([],[]), 1, \
            "vpar_is_totdomset_flags() empty is dominated");

check_equal(vpar_is_totdomset_flags([0],[0]), 0, \
            "vpar_is_totdomset_flags() singleton not dominated");
check_equal(vpar_is_totdomset_flags([0],[1]), 0, \
            "vpar_is_totdomset_flags() singleton not dominated");



\\-----------------------------------------------------------------------------
progress("vpar_is_indomset_flags()");

check_equal(vpar_is_indomset_flags([],[]), 1, "vpar_is_indomset_flags() empty");
check_equal(vpar_is_indomset_flags(vpar_make_path(2), [1,0]), 1);

check_equal(vpar_is_indomset_flags([2,1], [1,0]), 1, \
            "vpar_is_indomset_flags() 2-cycle");
check_equal(vpar_is_indomset_flags([2,1], [1,1]), 0, \
            "vpar_is_indomset_flags() 2-cycle");
check_equal(vpar_is_indomset_flags([2,1], [0,0]), 0, \
            "vpar_is_indomset_flags() 2-cycle");

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     for(rep=1,max(1,#vpar),
        my(flags=vpar_random_flags(vpar),
           set=test_flags_to_Set(flags));
        check_equal(vpar_is_indomset(vpar,set),
                    vpar_is_indomset_flags(vpar,flags),
                    "vpar_is_indomset() vs vpar_is_indomset_flags()")));
}


\\-----------------------------------------------------------------------------
progress("vpar_incidence_matrix()");

{
  my(vpar=[0,1,2],
     m=vpar_incidence_matrix(vpar));
  \\      edge   1-2  2-3
  check_equal(m, [-1, 0;
                  1,  -1;
                  0,  1]);
  check_equal(matrank(m), #vpar-1);
}

test_matrix_random_row_number(m) = \
  random(matsize(m)[1]) + 1;

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        m   =vpar_incidence_matrix(vpar));
     check_equal(matrank(m), vpar_num_edges(vpar));
     if(#vpar>=1,
        \\ matrix tree theorem, r=reduced incidence matrix, giving number of
        \\ spanning trees, 1 for tree, 0 for forest
        if(matsize(m)[1],
           my(row=test_matrix_random_row_number(m),
              r=m[^row,]);
           check_equal(matdet(r*r~), vpar_is_tree(vpar));
          )));
}


\\-----------------------------------------------------------------------------
progress("vpar_height()");

check_equal(vpar_height([]), -1, "height of empty");
check_equal(vpar_height([0]), 0, "height of single");
check_equal(vpar_height([0,1]), 1, "height of path-2");

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        height=vpar_height(vpar));

     my(component_vpar=vpar_component_trees(vpar),
        component_heights=apply(vpar_height, component_vpar));
     check_equal(height,
                 if(#component_heights==0,-1, vecmax(component_heights)),
                 "vpar_height() maximum of components"));
}


\\-----------------------------------------------------------------------------
progress("vpar_count_equaldepths()");

{
  check_equal(vpar_count_equaldepths(5,-1), 14);
  check_equal(vpar_count_equaldepths(5,0), 14);
  check_equal(vpar_count_equaldepths(5,1), 43);
  check_equal(iferr(vpar_count_equaldepths(5,'forest);0, e, 1),
              1,
  "vpar_count_equaldepths() error on forest=symbol (instead of -1,0,1)");
}

{
  for(forest=-1,1,
     for(n=0, test_limit_trees_search - forest,
        my(count=0);
        forvec(vec=vpar_forvec_from(n,forest),
              my(vpar=vpar_from_upqueue(vec));
              count += vpar_is_equaldepths(vpar));
        check_equal(count,
                    vpar_count_equaldepths(n,forest),
                 Str("vpar_count_equaldepths() vs count vpar_is_equaldepths()"
                     " n="n" forest="forest));
     ));
}
{
  \\ A047970(n) = trees(n+2)          trees(n) = A047970(n-2)
  \\ A047970(n) = forests(n+1)        forests(n) = A047970(n-1)
  check_equal(vector(#A047970_samples,n,n--; vpar_count_equaldepths(n+2)),
              A047970_samples,  \\ OFFSET=0
              "A047970 = vpar_count_equaldepths()");
  check_equal(vector(#A047970_samples,n,n--; vpar_count_equaldepths(n+1,1)),
              A047970_samples,  \\ OFFSET=0
              "A047970 = vpar_count_equaldepths()");
}
test_A047970(n) =
{
  n>=0 || error("test_A047970() is for n>=0");
  1 + sum(k=1,n, k*(k+1)^(n-k));
}
{
  check_equal(vector(#A047970_samples,n,n--; test_A047970(n)),
              A047970_samples,  \\ OFFSET=0
              "test_A047970() vs samples");
  check_equal(vector(30,n,n--; test_A047970(n)),
              vector(30,n,n--; 1 + sum(k=1,n+1, (k-1)*k^(n+1-k))),
              "A047970 formula 2");

  my(len=50);
  check_equal(vector(len,n,n--; test_A047970(n)),
                             \\ Peter Bala in A047970
              test_gf_terms( (1-x) * sum(n=0,len, x^n/(1 - (n+2)*x)),
                            len),
              "A047970 gf");
}
{
  \\ A112532 diffs of A047970
  check_equal(vector(#A112532_samples,n,n--; test_A047970(n) - if(n==0,0,test_A047970(n-1))),
              A112532_samples,  \\ OFFSET=0
              "test_A112532() vs samples");

  \\ A026898 cumulative A047970, with initial 1
  check_equal(vector(#A026898_samples,n,n--; sum(i=-1,n-1, if(i==-1,1,test_A047970(i)))),
              A026898_samples,  \\ OFFSET=0
              "A026898 samples");

  \\ A003101 cumulative A047970, first n terms
  check_equal(vector(#A003101_samples,n,n--; sum(i=0,n-1, test_A047970(i))),
              A003101_samples,  \\ OFFSET=0
              "A003101 samples");
}

{
  \\ vpar_count_equaldepths() addhelp description of equaldepths preorder
  \\ trees and forests
  for(forest=0,1,
     for(n=0,7-forest,
        my(count=0,
           vpar=vpar_preorder_first(n));
        until((vpar=vpar_preorder_next(vpar,forest))=='none,
              count+=vpar_is_equaldepths(vpar));

        check_equal(count, if(forest,
                              if(n<1,1, 2^(n-1)),   \\ forests
                              if(n<2,1, 2^(n-2))),  \\ trees
                    "vpar_is_equaldepths() count of pre-order, per vpar_count_equaldepths() help string")));
}


\\-----------------------------------------------------------------------------
progress("vpar_is_minimal_domset()");
\\        vpar_is_minimal_domset_flags()

check_equal(vpar_is_minimal_domset([],[]), 1);

check_equal(vpar_is_minimal_domset([0],[]), 0, \
            "vpar_is_minimal_domset() singleton not");
check_equal(vpar_is_minimal_domset([0],[1]), 1, \
            "vpar_is_minimal_domset() singleton dom");

check_equal(vpar_is_minimal_domset([1],[]), 0, \
            "vpar_is_minimal_domset() self-loop not");
check_equal(vpar_is_minimal_domset([1],[1]), 1, \
            "vpar_is_minimal_domset() self-loop dom");

check_equal(vpar_is_minimal_domset([2,1],[1]), 1, \
            "vpar_is_minimal_domset() 2-cycle");
check_equal(vpar_is_minimal_domset([2,1,1],[1]), 1, \
            "vpar_is_minimal_domset() 2-cycle and tail");
check_equal(vpar_is_minimal_domset([2,1,1],[1,3]), 0, \
            "vpar_is_minimal_domset() 2-cycle and tail");

\\---
check_equal(vpar_is_minimal_domset_flags([],[]), 1, \
            "vpar_is_minimal_domset_flags() empty");

check_equal(vpar_is_minimal_domset_flags([0],[0]), 0, \
            "vpar_is_minimal_domset_flags() singleton not");
check_equal(vpar_is_minimal_domset_flags([0],[1]), 1, \
            "vpar_is_minimal_domset_flags() singleton dom");

check_equal(vpar_is_minimal_domset_flags([1],[0]), 0, \
            "vpar_is_minimal_domset_flags() self-loop not");
check_equal(vpar_is_minimal_domset_flags([1],[1]), 1, \
            "vpar_is_minimal_domset_flags() self-loop dom");

check_equal(vpar_is_minimal_domset_flags([2,1],[1,0]), 1, \
            "vpar_is_minimal_domset_flags() 2-cycle");


\\---

\\ by attempting to remove one from the set and still be a domset
test_is_minimal_domset_by_remove(vpar,set) =
{
  vpar_is_domset(vpar,set) || return(0);
  for(i=1,#set,
     if(vpar_is_domset(vpar,set[^i]), return(0)));
  1;
}

\\ by attempting to remove one of the flags and still be a domset
test_is_minimal_domset_flags_by_remove(vpar,flags) =
{
  vpar_is_domset_flags(vpar,flags) || return(0);
  for(v=1,#flags,
     if(flags[v],
        my(new_flags=flags);
        new_flags[v]=0;
        if(vpar_is_domset_flags(vpar,new_flags), return(0))));
  1;
}

{
  \\ 1 <-- 4*       9         3*
  \\   v  /        | |
  \\    6           7 *
  \\    |           |
  \\    5*          8 *
  \\                |
  \\                2
  \\
  my(vpar=[6, 8, 0, 1, 6, 4, 9, 7, 7],
     set=[3, 4, 5, 7, 8]);
  check_equal(#vpar, 9);
  check_equal(vpar_is_domset(vpar,set), 1,
              "vpar_is_domset() 6,8");
  check_equal(test_is_minimal_domset_by_remove(vpar,set), 1,
              "test_is_minimal_domset_by_remove() 6,8");
  check_equal(vpar_is_minimal_domset(vpar,set), 1,
              "vpar_is_minimal_domset() 6,8");
}

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     for(rep=1,3*#vpar,
        my(flags=vpar_random_flags(vpar),
           set=test_flags_to_Set(flags),
           want=test_is_minimal_domset_by_remove(vpar,set));
        \\ print(vpar" "set);
        check_equal(want,
                    test_is_minimal_domset_flags_by_remove(vpar,flags),
                    "test_is_minimal_domset_by_remove() vs test_is_minimal_domset_flags_by_remove()");

        check_equal(want,
                    vpar_is_minimal_domset(vpar,set),
                    "vpar_is_minimal_domset_flags() vs test_is_minimal_domset_by_remove()");
        check_equal(want,
                    vpar_is_minimal_domset_flags(vpar,flags),
                    "vpar_is_minimal_domset_flags() vs test_is_minimal_domset_by_remove()");
        ));
}


\\-----------------------------------------------------------------------------
progress("vpar_is_maximal_matching_set()");
\\        vpar_is_maximal_matching_flags()

\\   1 --> 2 --> 3
check_equal(vpar_is_matching_flags([2,3,0], [1,0,0]), 1);
check_equal(vpar_is_maximal_matching_flags([2,3,0], [1,0,0]), 1);

check_equal(vpar_is_maximal_matching_set([],[]), 1, \
            "vpar_is_maximal_matching_set() empty");
check_equal(vpar_is_maximal_matching_set([0],[]), 1, \
            "vpar_is_maximal_matching_set() singleton unmatched");
check_equal(vpar_is_maximal_matching_set([0],[1]), 0, \
            "vpar_is_maximal_matching_set() singleton matched");

check_equal(vpar_is_maximal_matching_set([1],[]), 1, \
            "vpar_is_maximal_matching_set() self-loop unmatched");
check_equal(vpar_is_maximal_matching_set([1],[1]), 0, \
            "vpar_is_maximal_matching_set() self-loop matched");

check_equal(vpar_is_maximal_matching_set([2,1],[]), 0, \
            "vpar_is_maximal_matching_set() 2-cycle unmatched");
check_equal(vpar_is_maximal_matching_set([2,1],[1]), 1, \
            "vpar_is_maximal_matching_set() 2-cycle matched 1");
check_equal(vpar_is_maximal_matching_set([2,1],[2]), 1, \
            "vpar_is_maximal_matching_set() 2-cycle matched 2");

\\ see if flags is a maximal matching by attempting to change some flags[]=0
\\ to 1 instead
test_is_maximal_matching_flags_by_extend(vpar,flags) =
{
  vpar_is_matching_flags(vpar,flags) || return(0);
  for(v=1,#vpar,
     if(!flags[v],
        my(new_flags=flags);
        new_flags[v]=1;
        if(vpar_is_matching_flags(vpar,new_flags),
           return(0))));  \\ can add a flag, not maximal
  1;
}

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     for(rep=1,#vpar+1,
        my(flags=vpar_random_flags(vpar),
           set=test_flags_to_Set(flags),
           want=test_is_maximal_matching_flags_by_extend(vpar,flags));

        check_equal(vpar_is_maximal_matching_set(vpar,set),
                    want,
"vpar_is_maximal_matching_set() vs test_is_maximal_matching_flags_by_extend()");

        check_equal(vpar_is_maximal_matching_flags(vpar,flags),
                    want,
"vpar_is_maximal_matching_flags() vs test_is_maximal_matching_flags_by_extend()")));
}


\\-----------------------------------------------------------------------------
progress("vpar_is_matching_set()");

check_equal(vpar_is_matching_set([],[]), 1, "vpar_is_matching_set() empty");
check_equal(vpar_is_matching_set([0],[1]), 0);
check_equal(vpar_is_matching_set([0,1],[1]), 0);
check_equal(vpar_is_matching_set([0,1],[2]), 1);

\\ singletons
\\ check_equal(vpar_is_matching_set([1,2],[]), 1);
\\ check_equal(vpar_is_matching_set([1,2],[1]), 1, \
\\             "vpar_is_matching_set() singleton");
\\ check_equal(vpar_is_matching_set([1,2],[2]), 1);

\\ Matching as independent edge set would allow singleton.
\\ Matching as pairs of vertices might allow v-v if adjacent like
\\ vpar_is_neighbour() says.
\\
\\ matchnum of n-cycle
\\ *      1
\\ *-*    1
\\ *-*-*  1
\\ *-*-*-*  2
\\ *-*-*-*-*  2
\\ *-*-*-*-*-*  3

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        flags=vpar_random_flags(vpar),
        set=test_flags_to_Set(flags));
     check_equal(vpar_is_matching_set(vpar,set),
                 vpar_is_matching_flags(vpar,flags)));
}


\\-----------------------------------------------------------------------------
progress("vpar_is_matching_flags()");

check_equal(vpar_is_matching_flags([],[]), 1, "vpar_is_matching_flags() empty");
check_equal(vpar_is_matching_flags([0],[0]), 1, \
            "vpar_is_matching_flags() singleton unmatched");
check_equal(vpar_is_matching_flags([0],[1]), 0, \
            "vpar_is_matching_flags() singleton matched, is not");
check_equal(vpar_is_matching_flags([0,0],[0,0]), 1, \
            "vpar_is_matching_flags() 2 singletons unmatched");

check_equal(vpar_is_matching_flags([0,1,1,1],[1,0,0,0]), 0);
check_equal(vpar_is_matching_flags([0,1,1,1],[0,1,0,0]), 1, \
            "the root");
check_equal(vpar_is_matching_flags([0,1,1,1],[0,1,0,1]), 0, \
            "two children of same parent");

check_equal(vpar_is_matching_flags([1],[0]), 1, \
            "vpar_is_matching_flags() self-loop unmatched");
check_equal(vpar_is_matching_flags([1],[1]), 0, \
            "vpar_is_matching_flags() self-loop matched, is not");

check_equal(vpar_is_matching_flags([1,2],[0,0]), 1, \
            "vpar_is_matching_flags() 2 self-loops unmatched");
check_equal(vpar_is_matching_flags([1,2],[1,0]), 0, \
            "vpar_is_matching_flags() 2 self-loops, one matched");
check_equal(vpar_is_matching_flags([1,2],[1,1]), 0, \
            "vpar_is_matching_flags() 2 self-loops matched, is not");


\\------------------------------------------------------------------------------
progress("vpar_is_bipartite()");

check_equal(vpar_is_bipartite([]), 1, "vpar_is_bipartite() empty");
check_equal(vpar_is_bipartite([0]), 1, "vpar_is_bipartite() singleton");
check_equal(vpar_is_bipartite([1]), 0, "vpar_is_bipartite() self-loop");

test_is_bipartite_by_cycle_lengths(vpar) =
{
  my(cycles=vpar_component_cycle_lengths(vpar));
  for(i=1,#cycles, if(cycles[i]%2, return(0)));
  1;
}
{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     check_equal(vpar_is_bipartite(vpar),
                 test_is_bipartite_by_cycle_lengths(vpar),
                 "vpar_is_bipartite() vs test_is_bipartite_by_cycle_lengths()");
  );
}


\\-----------------------------------------------------------------------------
progress("test_Fibonacci_polynomial()");

test_Fibonacci_polynomial(n,x) =
{
  n>=0 || error("test_Fibonacci_polynomial() is for n>=0");
  my(p=1 + 0*x, q=0*x);
  for(i=1,n,
     my(new_p = x*p + q);
     q=p; p=new_p);
  q;
}
{
  check_equal(test_Fibonacci_polynomial(0,x), Pol(0),
              "test_Fibonacci_polynomial() n=0");
  check_equal(test_Fibonacci_polynomial(1,x), Pol(1),
              "test_Fibonacci_polynomial() n=1");
  check_equal(test_Fibonacci_polynomial(2,x), 'x,
              "test_Fibonacci_polynomial() n=2");
  check_equal(test_Fibonacci_polynomial(3,x), 'x^2 + 1,
              "test_Fibonacci_polynomial() n=3");
  check_equal(test_Fibonacci_polynomial(4,x), 'x^3 + 2*'x,
              "test_Fibonacci_polynomial() n=4");
  check_equal(test_Fibonacci_polynomial(5,x), 'x^4 + 3*'x^2 + 1,
              "test_Fibonacci_polynomial() n=5");
  for(n=2,20,
     check_equal(test_Fibonacci_polynomial(n,x),
                 x*test_Fibonacci_polynomial(n-1,x)
                 + test_Fibonacci_polynomial(n-2,x),
                "test_Fibonacci_polynomial() recurrence");
     check_equal(polresultant(test_Fibonacci_polynomial(n,x),
                              test_Fibonacci_polynomial(n+1,x)),
                1,
                "test_Fibonacci_polynomial() successive resultant");
     check_equal(test_Fibonacci_polynomial(n,x)',
                 (2*n*test_Fibonacci_polynomial(n-1,x)
                  + (n-1)*x*test_Fibonacci_polynomial(n,x)) / (x^2+4),
                 "test_Fibonacci_polynomial() derivative");
  );
  for(n=1,30,
     check_equal(polisirreducible(test_Fibonacci_polynomial(n,x)),
                 isprime(n),
                 "test_Fibonacci_polynomial() irreducible iff n prime"));

  \\ coefficients
  my(v=[], n=0);
  while(#v < #A168561_samples,
        v=concat(v, Vecrev(test_Fibonacci_polynomial(n,x)));
        n++);
  check_equal(v[1..#A168561_samples],
              A168561_samples,
              "A168561 coeffs of Fibonacci polynomials low to high");
}


\\-----------------------------------------------------------------------------
progress("vpar_subtree_heights_vectors_count()");

check_equal(vector(6,n,n--; vpar_subtree_heights_vectors_count(n,1)), \
            [1, 1, 3, 10, 47, 246], \
            "vpar_subtree_heights_vectors_count() forests addhelp samples");
check_equal(vector(7,n,n--; vpar_subtree_heights_vectors_count(n,-1)), \
            [1, 1, 1, 3, 10, 47, 246], \
            "vpar_subtree_heights_vectors_count() root=1 addhelp samples");
check_equal(vector(6,n,n--; vpar_subtree_heights_vectors_count(n)), \
            [1,1,2,9,40,235], \
            "vpar_subtree_heights_vectors_count() trees addhelp samples");

check_equal(vpar_subtree_heights_vectors_count(0,1), 1, \
            "vpar_subtree_heights_vectors_count() n=0 forests");
check_equal(vpar_subtree_heights_vectors_count(1,1), 1, \
            "vpar_subtree_heights_vectors_count() n=1 forests");
check_equal(vpar_subtree_heights_vectors_count(2,1), 3, \
            "vpar_subtree_heights_vectors_count() n=2 forests");

check_equal(vpar_subtree_heights_vectors_count(0), 1, \
            "vpar_subtree_heights_vectors_count() n=0 trees");
check_equal(vpar_subtree_heights_vectors_count(1), 1, \
            "vpar_subtree_heights_vectors_count() n=1 trees");
\\ check_equal(vpar_subtree_heights_vectors_count(2),1, \
\\             "vpar_subtree_heights_vectors_count() n=2");   \\ path-2 only

test_A327827(n) = n * test_A005651(n-1);
{
  check_equal(vector(#A327827_samples,n,n--; test_A327827(n)),
              A327827_samples,   \\ OFFSET=0
              "test_A327827() vs A327827_samples");

  check_equal(vector(100,n, vpar_subtree_heights_vectors_count(n)),
              vector(100,n, test_A327827(n)),
              "A327827 = vpar_subtree_heights_vectors_count(), tree");

  check_equal(vector(#A005651_samples,n,n--; vpar_subtree_heights_vectors_count(n,1)),
              A005651_samples,    \\ OFFSET=0
              "A005651 = vpar_subtree_heights_vectors_count(), forest");

  check_equal(vpar_subtree_heights_vectors_count(0,-1), 1);
  check_equal(vpar_subtree_heights_vectors_count(1,-1), 1);
  check_equal(vector(20,n, vpar_subtree_heights_vectors_count(n,-1)),
              vector(20,n, vpar_subtree_heights_vectors_count(n-1,1)),
              "vpar_subtree_heights_vectors_count(), root=1 same one later");
}

{
  \\ Form shown in vpar_subtree_heights_vectors_count() addhelp.
  \\ Works for n=0 since partitions() is an empty partition and prod() of no
  \\ terms is 1, which is desired.
  my(by_partitions = (n)-> vecsum(apply(p->n!/prod(i=1,#p,p[i]!), partitions(n))));
  check_equal(vector(12,n,n--; vpar_subtree_heights_vectors_count(n,1)),
              vector(12,n,n--; by_partitions(n)),
              "vpar_subtree_heights_vectors_count() forest vs by_partitions()");
}

{
  \\ number of different vpar_subtree_heights_vector()
  for(forest=-1,1,
     for(n=1, test_limit_trees_search - forest,
        my(l=List([]), count=0);
        forvec(vec=vpar_forvec_from(n,forest),
              my(vpar=vpar_from_upascend(vec),
                 vec=vpar_subtree_heights_vector(vpar),
                 pos=setsearch(l,vec,1));
              if(pos, listinsert(l,vec,pos));
              count++);
        check_equal(count, vpar_count(n,forest));  \\ all trees or forests

        check_equal(vpar_subtree_heights_vectors_count(n,forest),
                    #l,
                    Str("vpar_subtree_heights_vectors_count() vs vpars n="n" forest="forest));

        \\ vpar_is_subtree_heights_vector() vs the set found
        forvec(vec=vector(n,i,[-1,n+1]),
              check_equal(vpar_is_subtree_heights_vector(vec,forest),
                          setsearch(l,vec)!=0,
                          "vpar_is_subtree_heights_vector() vs set from vpars"));
  ));
}


\\-----------------------------------------------------------------------------
progress("vpar_matchnum()");

check_equal(vpar_matchnum([]), 0, "vpar_matchnum() empty tree");
check_equal(vpar_matchnum([0]), 0, "vpar_matchnum() singleton tree");
check_equal(vpar_matchnum([0,1]), 1, "vpar_matchnum() path-2");

check_equal(vpar_matchnum([1]), 0, "vpar_matchnum() self-loop");
check_equal(vpar_matchnum([2,1]), 1, "vpar_matchnum() 2-cycle");

test_matchnum_by_seq(vpar) =
{
  \\ Previous code using a seq, before combined with perfect matching.
  if(vpar_is_cyclic(vpar), error("for acyclics only"));

  my(seq=vpar_INTERNAL_seq_upwards(vpar),
     matched  = vectorsmall(#vpar),
     matchnum = 0);
  for(i=1,#seq,
     my(v=seq[i], p);
     if(!matched[v] && (p=vpar[v]) && !matched[p],
        matchnum++;
        matched[p]=1));
  matchnum;
}

\\ like vpar_matchnum(), but by searching sets of flags
test_matchnum_by_flags(vpar) =
{
  my(ret=0);
  forvec(flags=vector(#vpar,i,[0,1]),
        if(vpar_is_matching_flags(vpar,flags),
           ret=max(ret,hammingweight(flags))));
  ret;
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got=vpar_matchnum(vpar));
     check_equal(got, test_matchnum_by_seq(vpar),
                 "vpar_matchnum() vs test_matchnum_by_seq()");

     if(#vpar <= test_limit_flags_search,
        check_equal(got, test_matchnum_by_flags(vpar),
                 "vpar_matchnum() vs test_matchnum_by_flags()"));

     my(component_vpar=vpar_component_trees(vpar));
     check_equal(got, sum(i=1,#component_vpar, vpar_matchnum(component_vpar[i])),
                 "vpar_matchnum() sum of components");

     check_equal(got, vpar_INTERNAL_matchnum(vpar,#vpar+2),
                 "vpar_matchnum() vs vpar_INTERNAL_matchnum()");
  );
}

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        got=vpar_matchnum(vpar));

     if(#vpar <= test_limit_flags_search,
        check_equal(got, test_matchnum_by_flags(vpar),
                 "vpar_matchnum() vs test_matchnum_by_flags()"));
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_transpose()");

check_equal(vpar_transpose([]), []);
check_equal(vpar_transpose([0]), [0]);
check_equal(vpar_transpose([0,1]), [0,0], \
            "vpar_transpose() path-2 -> 2 singletons");
check_equal(vpar_transpose([0,0]), [0,1], \
            "vpar_transpose() 2 singletons -> path-2");

\\  1        1..2     or opposite    1    3..1     1..3
\\  |\          |                  / |    |     =     |
\\  2..3        3                 2..3    2           2
check_equal(vpar_next_sibling_vector([0,1,1],1), [0,0,2]);
check_equal(vpar_transpose([0,1,1]),   [0,0,2], "vpar_transpose() 123");
check_equal(vpar_transpose([0,1,1],1), [0,3,0], "vpar_transpose() 123 down");
check_equal(vpar_transpose([0,0,2]), [0,1,1], \
            "vpar_transpose() of 0,0,2");

\\   3     1  2  3
\\   |
\\   2
\\   |
\\   1
check_equal(vpar_next_sibling_vector([2,3,0],1), [0,0,0]);
check_equal(vpar_transpose([2,3,0]), [0,0,0], \
            "vpar_transpose() upwards 2,3,0");

\\   3        3  1
\\  / \         /|\
\\ 1   2       2 4 5
\\     |
\\     5
\\     |
\\     4
check_equal(vpar_transpose([3,3, 0, 5,2]), [0,1,0,1,1], \
            "vpar_transpose() 3,3,0,5,2");

{
  \\ 1..2    1  3    or opposite    4..2  = 2..4
  \\ |  |    |\                       /|    |\
  \\ 3  4    2..4                   3..1    1..3
  \\
  my(vpar=[0,0,1,2]);
  vpar_validate(vpar);
  check_equal(vpar_transpose(vpar),   [0,1,0,1]);
  check_equal(vpar_transpose(vpar,1), [2,0,2,0]);
}

{
  \\ 6               6 .. 2 .. 5           2 .. 5 .. 6        2 .. 1
  \\ |         =>         |            is  |             =>   |    |
  \\ 2 .. 4               4 .. 3 .. 1      1 .. 3 .. 4        5    3
  \\ |    |                                                   |    |
  \\ 5    3                                                   6    4
  \\      |
  \\      1
  my(vpar=[3,6,4,6,2,0]);
  check_equal(vpar_transpose(vpar),   [2,0,2,2,0,0]);
}

{
  \\ Knuth fasc4a.ps exercise 12 example.
  \\   1 .. 3                           1 .. 2
  \\   |    |                           |
  \\   2    4 .. 6 ........12           3 .. 4 .. 5
  \\        |    |          |                |
  \\        5    7 .. 11   13 .. 14          6 .. 7 .. 8
  \\             |               |           |    |    |
  \\             8 .. 9          15          |   11    9 .. 10
  \\                  |                      |
  \\                 10                     12 .. 13
  \\                                               |
  \\                                              14 .. 15
  my(vpar=[0,1, 0,3,4, 3,6,7,7,9,6, 3,12,12,14]);
  my(want=[0,0, 1,1,1, 4,4,4, 8,8, 7, 6,6, 13,13]);
  vpar_validate(vpar);
  vpar_validate(want);
  check_equal(vpar_transpose(vpar), want,
              "vpar_transpose() Knuth fasc4a exercise 12");
}

{
  for(i=1,#test_sample_downwards_forests,
     my(vpar=test_sample_downwards_forests[i],
        transposed_vpar=vpar_transpose(vpar),
        vpar_again=vpar_transpose(transposed_vpar));
     check_equal(vpar_again, vpar,
                 "vpar_transpose() twice"));
}
{
  for(n=1,6,
     for(i=1,vpar_count_updown(n),
        my(vpar=vpar_downwards_ith(n,i,1), \\ forest
           transposed_vpar=vpar_transpose(vpar),
           vpar_again=vpar_transpose(transposed_vpar));
     check_equal(vpar_again, vpar,
                 "vpar_transpose() twice, all downwards")));
}

test_downwards_flip_by_seq(vpar) =
{
  \\ Vertex which has a prev_sibling changes to become a child of that
  \\ prev_sibling.
  \\ Vertex which does not have a prev_sibling changes to become sibling of
  \\ its parent.  Processing vertices downwards means the parent's parent
  \\ has been determined in ret[] already.

  my(seq=vpar_seq_rows(vpar),  \\ downwards
     prev_sibling=vpar_next_sibling_vector(vpar,1),
     ret=vector(#vpar));
  for(i=1,#seq,
     my(v=seq[i]);
     ret[v] = if(prev_sibling[v], prev_sibling[v],
                 vpar[v], ret[vpar[v]]));
  ret;
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        first_child=vpar_minmax_child_vector(vpar,min),
        next_sibling=vpar_next_sibling_vector(vpar),
        transposed_vpar=vpar_transpose(vpar));
     \\ print(vpar"  first_child "first_child" next_sibling "next_sibling"  transposed "transposed_vpar);

     check_equal(transposed_vpar, test_downwards_flip_by_seq(vpar),
                "vpar_transpose() vs test_downwards_flip_by_seq()");

     for(v=1,#vpar,
        \\ print("  v="v);
        if(first_child[v],
           check_equal(transposed_vpar[first_child[v]], transposed_vpar[v],
                       "vpar_transpose() expected sibling"));
        if(next_sibling[v],
           check_equal(transposed_vpar[next_sibling[v]], v)));

     my(vpar_again=vpar_transpose(transposed_vpar));
     check_equal(vpar==vpar_again,  vpar_is_downwards(vpar),
                 "vpar_transpose() back again iff vpar_is_downwards()");

     my(depths=vpar_depths_vector(vpar),
        horizpos=vpar_horizpos_vector(vpar),
        transposed_depths=vpar_depths_vector(transposed_vpar),
        transposed_horizpos=vpar_horizpos_vector(transposed_vpar));
     check_equal(transposed_depths,horizpos,
                 "vpar_transpose() original horizpos = tranpsose depths");
     check_equal(transposed_horizpos == depths,
                 vpar_is_downwards(vpar),
                 "vpar_transpose() of downwards: original depths = transpose horizpos");

     \\---------
     \\ opp=1

     check_equal(transposed_vpar,
            vpar_relabel_reverse(vpar_transpose(vpar_relabel_reverse(vpar), 1)),
            "vpar_transpose() opp=0 vs reverse with opp=1");

     my(opp_transposed_vpar=vpar_transpose(vpar,1));
     check_equal(opp_transposed_vpar,
            vpar_relabel_reverse(vpar_transpose(vpar_relabel_reverse(vpar))),
            "vpar_transpose() opp=1 vs reverse with opp=0");

     check_equal(vpar==vpar_transpose(opp_transposed_vpar,1),
                 vpar_is_upwards(vpar),
                 "vpar_transpose() opp=1 back again iff vpar_is_downwards()");

     my(opp_transposed_vpar=vpar_transpose(vpar,1));

     my(horizpos_down=vpar_horizpos_vector(vpar,1),
        opp_transposed_depths=vpar_depths_vector(opp_transposed_vpar),
        opp_transposed_horizpos=vpar_horizpos_vector(opp_transposed_vpar));
     check_equal(opp_transposed_depths,
                 horizpos_down,
                 "vpar_transpose() opp=1 transposed depths");
     if(vpar_is_downwards(vpar),
        check_equal(opp_transposed_horizpos, depths,
                    "vpar_transpose() opp=1 transposed horizpos"));
  );
}

{
  \\ how many labelled rooted are vpar_is_downwards after vpar_transpose()
  for(forest=-1,1,
     for(n=0, test_limit_trees_search - forest,
        my(count=0);
        forvec(vec=vpar_forvec_from(n,forest),
              my(vpar=vpar_from_upascend(vec));
              count += vpar_is_downwards(vpar_transpose(vpar)));

        if(forest==1,
           check_equal(count, test_odd_double_factorial(n)));
        if(forest==-1,
           check_equal(count, if(n==0,1, test_odd_double_factorial(n-1))));
        if(forest==0,
           check_equal(count, if(n==0,1, n*test_odd_double_factorial(n-1))))));
}


\\-----------------------------------------------------------------------------
progress("vpar_seq_rows()");

check_equal(vpar_seq_rows([]), Vecsmall([]));
check_equal(vpar_seq_rows([0,1,1,1]), Vecsmall([1,2,3,4]));
check_equal(vpar_seq_rows([0,1,1,1],1), Vecsmall([2,3,4,1]));

check_equal(vpar_seq_rows([0,1,1,1]), Vecsmall([1,2,3,4]));
check_equal(vpar_seq_rows([0,1,1,1],1), Vecsmall([2,3,4,1]));

test_seq_rows_by_sort(vpar,up=0) =
{
  my(depths=vpar_depths_vector(vpar),
     s=if(up,-1,1));
  Vecsmall(vecsort([1..#vpar],
                   t->[s*depths[t],t])); \\ key func

}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     for(up=0,1,
        my(got_seq = vpar_seq_rows(vpar,up));
        check_equal(got_seq, test_seq_rows_by_sort(vpar,up),
                    "vpar_seq_rows() vs test_seq_rows_by_sort()");
        check_equal(vpar_seq_is_ordered(vpar,got_seq), 1,
                    "vpar_seq_rows() is vpar_seq_is_ordered() per addhelp");

        check_equal(if(up, vpar_seq_is_upwards(vpar,got_seq),
                       vpar_seq_is_downwards(vpar,got_seq)),
                    1,
                    "vpar_seq_rows() is vpar_seq_is_downwards() or vpar_seq_is_upwards() per addhelp");

        my(row_vertices=vpar_row_vertices_vector(vpar));
        if(up, row_vertices=Vecrev(row_vertices));
        check_equal(Vec(got_seq),
                    if(#row_vertices, concat(row_vertices), []),
                    "vpar_seq_rows() vs vpar_row_vertices_vector() per addhelp");
     ));
}


\\-----------------------------------------------------------------------------
progress("test_is_gnil()");

{
  if(!have_test_is_gnil,
     print("have_test_is_gnil = "have_test_is_gnil", skip its tests");
     ,

     \\ absent else-case of an "if" is gnil
     check_equal(test_is_gnil(if(1,1)),   0);
     check_equal(test_is_gnil(if(1,1,0)), 0);
     check_equal(test_is_gnil(if(0,1)),   1);
     check_equal(test_is_gnil(if(0,1,0)), 0);

     \\ example in test_is_gnil.c
     my(mul_x_and_y = (x,y)->if(x, x*y));
     check_equal(test_is_gnil(mul_x_and_y(0,0)),  1);
     check_equal(test_is_gnil(mul_x_and_y(0,1)),  1);
     check_equal(test_is_gnil(mul_x_and_y(1,0)),  0);
     check_equal(test_is_gnil(mul_x_and_y(2,2)),  0);

     my(n=0);
     check_equal(test_is_gnil(if(n,valuation(n,2)) + valuation(n+1,2)),  0);
     check_equal(test_is_gnil(if(0,0)), 1);
     check_equal(test_is_gnil(if(0,0)+0), 0);
     check_equal(test_is_gnil(0+if(0,0)), 0);
    );
}


\\-----------------------------------------------------------------------------
progress("test_A084088()");
\\ A084088 == 2 mod 3 and even low 0-bits
\\   (6m-1)*4^k
\\   arising in Huffman tree weights 1 to n
\\ 5, 11, 17, 20, 23, 29, 35, 41, 44, 47, 53, 59, 65, 68, 71, 77, 80, 83,

\\ test whether a occurs as some a(n) in A084088
A084088_pred(a) = a>=5 && a%3==2 && valuation(a,2)%2==0;
{
  check_equal(select(A084088_pred, [0..A084088_samples[#A084088_samples]]),
              A084088_samples,
              "A084088_pred() vs A084088_samples");
}

\\ (6p-1)*4^(k-i) < 5*4^k
\\ (6p-1) < 5*4^i
\\ 6p < 5*4^i+1
\\ p < (5*4^i+1)/6
A084088_first_i(i) =
{
  i>=1 || error("A084088_first_i() is for i>=1");
  (5*4^i-2)/6;
}
\\ vector(10,i,i--; A084088_first_i(i))
{
  for(k=1,50,
     for(i=1,k,
        my(p=A084088_first_i(i));
        check_equal((6*p-1)*4^(k-i) < 5*4^k, 1);
        p++;
        check_equal((6*p-1)*4^(k-i) > 5*4^k, 1)));

  \\ A072197
  check_equal(vector(#A072197_samples,i,i--; A084088_first_i(i+1)),
              A072197_samples,  \\ OFFSET=0
              "A084088_first_i() vs A072197_samples");
}

A084088_first(k) = (10*4^k - 3*k - 1)/9;
{
  \\ where first of k, which is 5*4^k
  \\ 1,4,17,70,283,1136,4549,18202,
  \\ A172447(n) = (-1 + 5*2^(2*n+1) - 3*n)/9
  \\ vector(20,n, (-1 + 5*2^(2*n+1) - 3*n)%9)
  \\ vector(20,n, A172447(n) - A172447(n-1))
  \\
  \\ 5*4^(k-1) increase to 5*4^k and new 5 mod 6 below it
  \\ vector(10,k, (5*4^k)\6)
  \\ vector(10,k,k--; (5*4^k)%6) == [5, 2, 2, 2, 2, 2, 2, 2, 2, 2]
  \\ A072197 a(n) = 4*a(n-1) + 1 with a(0) = 3.
  \\ base4  3111..111
  \\
  my(n=1,     \\ OFFSET=1
     k=0);
  for(a=1,2000,
     A084088_pred(a)||next;
     if(valuation(a,4)==k,
        check_equal(n, A084088_first(k),
                    "A084088_first() vs A084088_pred()");
        k++);
     n++);

  check_equal(vector(100,k,k--; A084088_first(k)),
              vector(100,k,k--; 1 + sum(i=1,k, A084088_first_i(i))),
              "A084088_first() vs sum A084088_first_i()");

  \\ A172447
  check_equal(vector(#A172447_samples,k,k--; A084088_first(k)),
              A172447_samples, \\ OFFSET=0
              "A084089_extra() vs A172447_samples");
}

\\ n < (10*4^(k+1) - 3*(k+1) - 1)/9;
\\ n < 10/9*4^(k+1) - 3/9*(k+1) - 1
\\ n < 10/9*4^(k+1) - 1/3*k - 4/3
\\ n < 4^(k+1)
A084088_k(n) = logint(n,4);
{
  my(bad=0);
  for(n=1,2^10,
     my(k=A084088_k(n));
     if(! (A084088_first(k-1) < n),
        print("A084088_k too big at n=",n,"  k="k" first="A084088_first(k));
        if(bad++ > 10, break));
     if(! (n < A084088_first(k+1)),
        print("A084088_k too small at n=",n);
        if(bad++ > 10, break));
  );
  check_equal(bad,0, "A084088_k() vs A084088_first()");
}

\\ A084088_step_i(k)
\\ 5*4^k < (6p-1)*4^(k-i) < 11*4^k
\\ 5*4^i < (6p-1) < 11*4^i
\\ 5*4^i+1 < 6p < 11*4^i+1
\\ (5*4^i+1)/6 < p < (11*4^i+1)/6
\\ (5*4^i+1)/6 < p < (5*4^i+1)/6 + 4^i

A084088_step(k) = (4^(k+1) - 1)/3;
{
  \\ distance between first and second k, which are 5*4^k and 11*4^k
  check_equal(vector(10,k,k--; (11*4^k)%6),
              [5, 2, 2, 2, 2, 2, 2, 2, 2, 2]);
  for(k=0,8,
     my(n=0,  \\ OFFSET=1, after pre-increment
        seen=0,
        first);
     for(a=1,2000000, A084088_pred(a)||next;
        n++;
        if(valuation(a,4)==k,
           seen++;
           if(seen==1, first=n; next);
           check_equal(first, A084088_first(k),
                       "A084088_first() in step search");
           check_equal(A084088_step(k), n-first,
                       "A084088_step() vs A084088_pred count");
           next(2))));

  \\ A002450 (4^n - 1)/3
  check_equal(vector(#A002450_samples,k,k--; A084088_step(k-1)),
              A002450_samples, \\ OFFSET=0
              "A084088_step() vs A002450_samples");
}

A084088_extra(k) = (2*4^k + 3*k - 2)/9;
{
  check_equal(vector(100,k,k--; A084088_extra(k)),
              vector(100,k,k--; A084088_step(k) - A084088_first(k)),
              "A084088_extra() vs step and first");

  \\ A160156
  check_equal(vector(#A160156_samples,k,k--; A084088_extra(k+1)),
              A160156_samples, \\ OFFSET=0
              "A084088_extra() vs A160156_samples");
}
\\ vector(20,k, A084088_extra(k))


A084088_by_parts(n) =
{
  \\ print("A084088 n="n);
  forstep(k=A084088_k(n)+5, 1, -1,
     \\ print("  k="k"  n="n"  first "A084088_first(k)" step "A084088_step(k)" divrem "Vec(divrem(n + A084088_extra(k), A084088_step(k))));
     my(qr=divrem(n + A084088_extra(k), A084088_step(k)));
     qr[1]>=0 || error("A084088_by_parts() oops, negative quotient");
     if(qr[2]==0,
        \\ print("  exact q="qr[1]);
        return((6*qr[1]-1)*4^k));
     n -= qr[1]);
  \\ print("  final k=0 n="n);
  6*n-1;
}
{
  check_equal(vector(#A084088_samples,n, A084088_by_parts(n)), \\ OFFSET=1
              A084088_samples);
}

A084088(n) = {
  forstep(k=logint(n,4),1,-1,
     my(qr=divrem(n + (2*4^k+3*k-2)/9, (4^(k+1)-1)/3));
     if(qr[2]==0, return((6*qr[1]-1)*4^k));
     n -= qr[1]);
  6*n-1; }
{
  check_equal(vector(2000,n, A084088(n)),
              vector(2000,n, A084088_by_parts(n)));
  check_equal(vector(100,k, A084088(A084088_first(k))),
              vector(100,k, 5*4^k));
}
{
  my(n=1,  \\ OFFSET=0
     target=A084088(n),
     bad=0);
  for(a=1,2^16, \\ prospective a(n)
     \\ if(bitand(a,32767)==0, print1("A084088 a="a,Strchr(13)));
     my(by_pred=A084088_pred(a),
        by_target=(a==target));
     if(by_pred != by_target,
        print("n="n" a="a"  by_pred="by_pred" vs by_target "by_target" expecting "target);
        if(bad++ > 10, break));
     if(by_target,
        n++;
        target=A084088(n)));
  check_equal(bad,0);
}


\\-----------------------------------------------------------------------------
progress("test_A065883()");
\\ A065883 remove factors of 4

{
  check_equal(vector(#A065883_samples,n, test_A065883(n)), \\ OFFSET=1
              A065883_samples,
              "test_A065883() vs A065883_samples");

  check_equal(vector(4096,n,n--; test_A065883(n)),
              vector(4096,n,n--; while(n&&n%4==0,n\=4); n),
              "test_A134667() vs loop");
}


\\-----------------------------------------------------------------------------
progress("test_A134667()");
\\ A134667 periodic

{
  check_equal(vector(#A134667_samples,n,n--; test_A134667(n)), \\ OFFSET=0
              A134667_samples,
              "test_A134667() vs A134667_samples");
  check_equal(vector(4096,n,n--; test_A134667(n)),
              vector(4096,n,n--; kronecker(-12,n)),
              "test_A134667() vs kronecker(-12,n)");
}


\\-----------------------------------------------------------------------------
progress("test_A096268()");
\\ A096268 period doubling, parity count low 0-bits

{
  check_equal(vector(#A096268_samples,n,n--; test_A096268(n)), \\ OFFSET=0
              A096268_samples, "test_A096268()");
}

\\ A096268 other trisections
\\ 3n = A073059
{
  check_equal(vector(4096,n, test_A096268(3*n+2)),
              vector(4096,n, test_A096268(n)),
              "A096268 own trisection, since 3*n+2 no change to low 1-bits");
}
check_equal(binary(3*3+2), [1,0, 1,1])
check_equal(binary(3*5+2), [1,0,0, 0,1])
\\ 11110
\\  1111
\\   +1
\\ -----
\\ 01111

\\ for(n=1,50, print1(test_A096268(3*n+1)","));
\\ not in OEIS: 0,1,0,1,0,0,0,1,0,1,0,1,0,0,0,1,0,1,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,1,0,1,0,0,0,1,0,1,0,1,0,0,0,1,0,1
\\ not A096268 itself, only middle match
\\ not A217586 middle match
\\ not A108357 periodic
\\ not A284932 middle match


\\-----------------------------------------------------------------------------
progress("test_A035263() period doubling parity CountLowZeros as 1,0");
\\
\\ A035263 OFFSET=1 period doubling, parity CountLowZeros as 1,0
\\    1, 0, 1, 1, 1, 0, 1, 0, 1, 0, 1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 0, 1, 0
\\ n= 1  2  3  4  5  6  7  8
\\
{
  check_equal(vector(#A035263_samples,n, test_A035263(n)), \\ OFFSET=1
              A035263_samples,
              "test_A035263() vs A035263_samples");
}

\\ A035263 other trisections
\\ 3n+1 = A073059 by A035263(3*n+1)
{
  \\ A035263
  check_equal(vector(512,n, test_A035263(3*n)),
              vector(512,n, test_A035263(n)),
              "A035263 own trisection, since 3*n no change to low 0s");
}
\\ for(n=1,50, print1(test_A035263(3*n+2)","));
\\ not in OEIS: 1,0,1,0,1,1,1,0,1,0,1,0,1,1,1,0,1,0,1,0,1,1,1,0,1,1,1,0,1,1,1,0,1,0,1,0,1,1,1,0,1,0,1,0,1,1,1,0,1,0
\\ not A035263 itself, only middle match
\\ not A039982 middle match
\\ not A104106 middle match
\\ not A259044 periodic
\\ not A306559 partial middle


\\-----------------------------------------------------------------------------
progress("test_A098725()");

{
  \\ A098725
  \\ base4 low 2 is +1 to rest            0,x,0,1     0,1,2,3
  \\ must 1 above to be base4 low 2 again
  \\ x11112
  \\ 3*(n-1) = 3x 33333
  \\

  check_equal(vector(#A098725_samples,n,n--; test_A098725(n)), \\ OFFSET=0
              A098725_samples, "test_A098725() vs A098725_samples");

  check_equal(vector(4096,n,n--; test_A098725(n)),
              vector(4096,n,n--;
                     \\ Ralf Stephan in A098725
                     if(n%2==1,1, n%4==0,0, test_A098725((n-2)/4+1))),
             "test_A098725() recurrence");

  check_equal(vector(4096,n,n++; test_A098725(n)),
              vector(4096,n,n++; 1 - test_A073059(n-1)),
              "A098725 by its opposite A073059, for n>=2");


  my(by_loop=(n)-> while(n%4==2, n=(n\4)+1); n%2);
  check_equal(vector(4096,n,n--; test_A098725(n)),
              vector(4096,n,n--; by_loop(n)),
              "test_A098725() vs by_loop()");

  my(by_low_non1=(n)-> n--; while(n%4==1, n\=4); !(n%2));
  check_equal(vector(4096,n,n--; test_A098725(n)),
              vector(4096,n,n--; by_low_non1(n)),
              "test_A098725() vs by_low_non1()");

  my(by_alt_bits=(n)-> n--;
     while(n,
           if(n%2==0, return(1));
           n\=2;
           if(n%2==1, return(0));
           n\=2);
     1);
  check_equal(vector(4096,n,n--; test_A098725(n)),
              vector(4096,n,n--; by_alt_bits(n)),
              "test_A098725() vs by_alt_bits()");

  \\ print(select(n->test_A098725(n), [1..100]));
  \\ not in OEIS: 1, 2, 3, 5, 6, 7, 9, 10, 11, 13, 15, 17, 18, 19, 21, 22, 23, 25, 26, 27, 29
  \\
  \\ print(select(n->!test_A098725(n), [1..100]));
  \\ not in OEIS: 4, 8, 12, 14, 16, 20, 24, 28, 30, 32, 36, 40, 44, 46, 48, 52, 54, 56, 60, 62
}

\\ for(n=0,64,print(digits(n-1,4)" "test_A098725(n)));
\\ for(k=0,4,print(digits((4^k-1)/3,4)));
\\ for(k=0,4,print(digits(2*4^k + (4^k-1)/3,4)));

my(k=3); check_equal(digits((4^k-1)/3, 4), [1,1,1]);
my(k=3); check_equal(digits(2*4^k + (4^k-1)/3, 4), [2,1,1,1]);

{
  \\ ... [0,2] 11..11 base 4  with k many low 1s
  \\ terms up to k inclusive are good to < 11..11 with k+1 low 1s
  \\ where next would begin
  for(k=0,6,
     my(limit = (4^(k+1)-1)/3,
        L=O(x^limit),
        g = sum(k=0,k,
                x^((4^k+2)/3)*( 1 + x^(2*4^k) ) / (1 - x^(4*4^k))
                + L),
        f=sum(n=0,limit-1, test_A098725(n) * x^n) + L);
     check_equal(g, f,
                 Str("gA098725 by base-4  k="k" limit="limit));
     \\ print(g);
     \\ print(f);
     \\ print(g==f);
  );
}
{
  \\ ... 0 010101 with k many low 01 bit pairs
  \\ terms up to k inclusive are good to < 01 010101 with k+1 low 1s
  \\ where next would begin
  for(k=0,6,
     my(limit = (4^(k+1)-1)/3,
        L=O(x^limit),
        g = sum(k=0,k,
                x^((4^k+2)/3) / (1 - x^(2*4^k))
                + L),
        f=sum(n=0,limit-1, test_A098725(n) * x^n) + L);
     check_equal(g, f,
                 Str("gA098725 by bits  k="k" limit="limit));
     \\ print(g);
     \\ print(f);
     \\ print(g==f);
  );
}

\\ A035263 period doubling inverse
{
  check_equal(vector(4096,n, test_A098725(n)),
              vector(4096,n, test_A035263(3*n-2)),
              "test_A098725() using A035263");
  check_equal(test_A098725(0), 0);
  check_equal(valuation(-2,2), 1);
  my(n=0); check_equal(3*n-2, -2);
}

\\ A096268 period doubling
{
  check_equal(vector(40,n, test_A098725(n)),
              vector(40,n, 1 - test_A096268(3*n-3)),
              "test_A098725() using A096268");
  check_equal(test_A098725(0), 0);
  check_equal(valuation(-2,2), 1);
  my(n=0); check_equal(3*n-3, -3);
}


\\-----------------------------------------------------------------------------
progress("vpar_from_adjacency_matrix()");

check_equal(vpar_from_adjacency_matrix(Mat([0])), [0], \
            "vpar_from_adjacency_matrix() singleton");
check_equal(vpar_from_adjacency_matrix(Mat([1])), [1], \
            "vpar_from_adjacency_matrix() self-loop");

\\ PENDING
\\ this not documented yet
{
  \\  +- 1 <-- 2 <-- 3 <-+
  \\  |                  |
  \\  +------------------+
  my(vpar=[3,1,2],
     m=vpar_adjacency_matrix(vpar));
  check_equal(vpar_from_adjacency_matrix(m), vpar);
}
{
  \\  +-> 1 --> 2 --> 3 -+
  \\  |                  |
  \\  +------------------+
  my(vpar=[2,3,1],
     m=vpar_adjacency_matrix(vpar),
     got=vpar_from_adjacency_matrix(m));
  check_equal(got, [3,1,2]);
  check_equal(vpar_adjacency_matrix(got), m);
}

\\ PENDING possible vpar_add_edge
test_vpar_add_edge(vpar,v,p) =
{
  until(!v, my(new_v=vpar[v]); vpar[v]=p; p=v; v=new_v);
  vpar;
}
test_from_adjacency_matrix_by_add_edge(m) =
{
  my(vpar=vector(matsize(m)[1]));
  forstep(p=#vpar,1,-1,
     forstep(v=#vpar,p+1,-1,
        if(m[v,p],
           vpar = test_vpar_add_edge(vpar,v,p))));
  vpar;
}

\\      3  smallest of its component
\\      |\
\\      7 8
\\        |
\\        6
\\        |
\\        5
\\
\\ 3 smallest, all neighbours are its children > 3.
\\ 5 would get child 6.
\\ 6 neighbour 8 find 8 already under another, so that root of 6 must flip.


test_from_adjacency_matrix_by_rows(m) =
{
  \\ This is start at v=1 as first root and find its children by examining
  \\ the row m[v,c] for all c.  They can be put in a queue or stack to be
  \\ processed likewise find its children.  This avoids re-writing when a v
  \\ is found with a parent already.
  \\
  \\ But it looks at the rows of all vertices to find children, so n^2
  \\ instead of triangle 1/2*n^2 in the current vpar_from_adjacency_matrix().
  \\
  \\ Some measuring suggests this n^2 is about 2* slower on a random tree.
  \\ The average re-writing in vpar_from_adjacency_matrix() seems to grow
  \\ faster than linear, but not by very much.  The worst case for rewriting
  \\ is a star with root last [n,n,n,...,n, 0] which is triangular number of
  \\ steps.  But that case still measures a bit slower here.

  my(n=matsize(m)[1],
     vpar=vector(n,i,-1),
     pending=vectorsmall(#vpar),
     upto);
  for(i=1,n,
     if(vpar[i]<0,
        vpar[i]=0;
        pending[upto=1]=i;
        while(upto,
              my(v=pending[upto]); upto--;
              for(c=1,n,
                 if(vpar[c]<0 && m[c,v],     \\ c = child of v
                    vpar[c]=v; pending[upto++]=c)))));
  vpar;
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        m=vpar_adjacency_matrix(vpar),
        got=vpar_from_adjacency_matrix(m));
     \\ print(vpar" to "got);
     vpar_validate(got);
     check_equal(vpar_adjacency_matrix(got), m,
                 "vpar_from_adjacency_matrix() matrix again");
     check_equal(got, vpar_reroot_minmax(vpar,min),
                 "vpar_from_adjacency_matrix() is rooted at mins");
     check_equal(test_equal_after_rerooting(vpar,got), 1);
     check_equal(got, test_from_adjacency_matrix_by_add_edge(m));
     check_equal(got, test_from_adjacency_matrix_by_rows(m));
  );
}
{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        m=vpar_adjacency_matrix(vpar),
        got=vpar_from_adjacency_matrix(m));
     \\ print("cyclic "vpar" to "got);
     check_equal(vpar_adjacency_matrix(got), m,
                 "vpar_from_adjacency_matrix() cyclic matrix again");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_protnum_of_vertex()");

{
  check_equal(iferr(vpar_protnum_of_vertex([],1);0, e, 1),
              1,
              "vpar_protnum_of_vertex() error on empty");
  check_equal(iferr(vpar_protnum_of_vertex([0,0,0],4);0, e, 1),
              1,
              "vpar_protnum_of_vertex() error on v out of range");
}

check_equal(vpar_protnum_of_vertex([0],1), 0, \
            "vpar_protnum_of_vertex() singleton");

{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i],
        root=vpar_root(vpar));
     if(#vpar==0,next);
     check_equal(vpar_protnum_of_vertex(vpar,root),
                 vpar_protnum(vpar),
                 "vpar_protnum_of_vertex() root is vpar_protnum()"));
}


\\-----------------------------------------------------------------------------
progress("vpar_protnum()");

check_equal(vpar_protnum([]), 0, "vpar_protnum() empty");

test_protnum_by_upwards(vpar) =
{
  my(vertices=vpar_childless_vertices(vpar));
  for(d=0,#vpar,
        for(i=1,#vertices,
           (vertices[i]=vpar[vertices[i]]) || return(d)));
  0;
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     check_equal(vpar_protnum(vpar),
                 test_protnum_by_upwards(vpar),
       "vpar_protnum() vs test_protnum_by_upwards()"));
}
{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i]);
     check_equal(vpar_protnum(vpar),
                 if(#vpar, vpar_protnum_of_vertex(vpar,vpar_root(vpar)), 0),
       "vpar_protnum() of tree vs vpar_protnum_of_vertex() root"));
}

\\ Clemens Heuberger and Helmut Prodinger, "Protection Number In Plane
\\ Trees", Applicable Analysis and Discrete Mathematics, volume 11, 2017,
\\ pages 314-326, http://doi.org/10.2298/AADM1702314H
\\
\\ Return the number of rooted ordered trees with vpar_protnum() >= k.
\\ Formula per Heuberger and Prodinger.
test_protnum_count_rnk(n,k) =
{
  sum(j=1,n^2,
      (-1)^(j-1) * (  binomial(2*n-3 - (2*k-1)*j,  n   - (k+1)*j)
                    - binomial(2*n-3 - (2*k-1)*j,  n-3 - (k+1)*j)));
}
{
  \\ ordered count trees which are at least k-protected

  \\ for(n=1,8, print(vector(n,k, test_protnum_count_rnk(n,k))));
  for(n=2,8,
     my(counts=vector(n),
        counts_ge=vector(n),
        vpar=vpar_preorder_first(n));
     until((vpar=vpar_preorder_next(vpar))=='none,
           my(p=vpar_protnum(vpar));
           counts[p]++;
           for(i=1,p, counts_ge[i]++));
     check_equal(counts_ge,
                 vector(n,k, test_protnum_count_rnk(n,k)));
     );
}

{
  \\ labelled count forests of given protection number
  \\
  \\ [1]
  \\ [1, 2]
  \\ [7, 3, 6]
  \\ [49, 40, 12, 24]
  \\ [491, 425, 200, 60, 120]
  \\ [6181, 5436, 2910, 1200, 360, 720]

  my(forest=1,
     limit= test_limit_trees_search - forest,
     counts=matrix(limit,limit));
  for(n=1,limit,
     forvec(vec=vpar_forvec_from(n,forest),
           my(vpar=vpar_from_upascend(vec));
           counts[n,vpar_protnum(vpar)+1]++));

  \\ forests num=0, is >=1 singleton
  \\ not in OEIS: 1, 7, 49, 491, 6181, 94305,
  \\ vector(8,n, vpar_count(n,1) - vpar_count_no_singletons(n));
  check_equal(counts[,1],
              vectorv(limit,n, vpar_count(n,1) - vpar_count_no_singletons(n)),
              "vpar_protnum()==0 is how many forests of no singletons");

  check_equal(vector(limit,n, counts[n,n]),
              vector(limit,n, n!),
              "vpar_protnum()==n is n! forests all singletons");
}


\\-----------------------------------------------------------------------------
progress("vpar_protnums_vectors_count()");
\\
\\ and see vpar_is_protnums_vector() for count versus search

check_equal(vector(8,n,n--; vpar_protnums_vectors_count(n)), \
            [1, 1, 2, 9, 46, 245, 1601, 11480], \
            "vpar_protnums_vectors_count() trees");

test_A320566(n) =
{
  n>=0 || error();
  sum(k=0,n,binomial(n,k)*test_A005651(k));
}
{
  check_equal(vector(#A320566_samples,n,n--; test_A320566(n)),
              A320566_samples,     \\ OFFSET=0
              "A320566 sum binomials of A005651");
}

\\ with a[1] as the 0th term
test_binomial_transform(a) = \
  vector(#a,n,n--; sum(k=0,n,binomial(n,k)*a[k+1]));
{
  check_equal(test_binomial_transform(vector(#A320566_samples,n,n--;
                                             vpar_subtree_heights_vectors_count(n,1))),
              A320566_samples,
              "A320566 binomial transform of vpar_subtree_heights_vectors_count() forests");
}
{
  my(limit=#A320566_samples,
     L = O(x^limit),
     g = L + exp(x+L) / prod(k=1,limit, 1 - x^k/k! + L));
  check_equal(test_egf_terms(g,limit), A320566_samples[1..limit],
              "A320566 egf");
}
{
  my(samples=[1, 1, 2, 6, 23, 110, 617, 4035, 29927, 249926, 2316317, 23674841, 264329177]);
  check_equal(vector(#samples,n,n-=2; if(n==-1,1,test_A320566(n))),
              samples);
}
{
  check_equal(vector(12,n, vpar_protnums_vectors_count(n,-1)),
              vector(12,n, test_A320566(n-1) - test_A005651(n-1) + (n==1)),
              "vpar_protnums_vectors_count() = A320566 - A005651");

  \\ Per addhelp of vpar_protnums_vectors_count():
  \\ k is size of the root=1 term.
  \\ binomial(n-1,k-1) choose the other vertices in that term.
  \\ Rest n-k any forest.
  \\ Then subtract cases where root=1 is the last term, which is an n-1 forest
  \\ to which root=1 put in the last.
  \\
  check_equal(vector(12,n, vpar_protnums_vectors_count(n,-1)),
              vector(12,n,
                     sum(k=1,n,binomial(n-1,k-1)*vpar_protnums_vectors_count(n-k,1))
                     - vpar_protnums_vectors_count(n-1,1)
                     + (n==1)),
              "vpar_protnums_vectors_count() root=1 from binomial forests");

  \\ print((vector(16,n,n--; vpar_subtree_heights_vectors_count(n-1,1))));
  \\ print(test_binomial_transform(vector(16,n,n--; vpar_subtree_heights_vectors_count(n,1))));

  \\ starting at 1
  check_equal(vector(16,n, vpar_protnums_vectors_count(n,-1)),
              test_binomial_transform(vector(16,n,n--; vpar_subtree_heights_vectors_count(n,1)))
              - vector(16,n,n--; if(n, vpar_subtree_heights_vectors_count(n,1))),
              "vpar_protnums_vectors_count() root=1 from binomial transform forests");

  \\ starting at 0
  \\ want binomial transform of F(n) at term n-1, so concat in extra initial
  check_equal(vector(16,n,n--; vpar_protnums_vectors_count(n,-1)),
              concat(1,test_binomial_transform(vector(16-1,n,n--; vpar_subtree_heights_vectors_count(n,1))))
              - vector(16,n,n--; if(n-1>0, vpar_subtree_heights_vectors_count(n-1,1))),
              "vpar_protnums_vectors_count() root=1 from binomial transform forests");
}
{
  my(limit=16,
     L = O(x^limit),
     \\ g constant term corresponding to n=1
     g = 1 + (exp(x+L) - 1) / prod(k=1,limit, 1 - x^k/k! + L));
  check_equal(test_egf_terms(g,limit),
              vector(limit,n, vpar_protnums_vectors_count(n,-1)),
              "vpar_protnums_vectors_count() root=1 egf");

  check_equal(test_egf_terms(intformal(g)+1,limit),
              vector(limit,n,n--; vpar_protnums_vectors_count(n,-1)),
              "vpar_protnums_vectors_count() root=1 egf offset 0");
}

\\ previous code
test_protnums_vectors_count_by_arrays(n,forest=0) =
{
  \\ For root=1, at the start of the k loop,
  \\ a[i+1] = number of labelled partitions of i with all terms <k.
  \\   This is unrestricted, same as vpar_subtree_heights_vectors_count().
  \\ w[i+1] = number of labelled partitions with all terms <k and where
  \\   root=1 vertex is in the first of some set of partition terms.
  \\   Eg. 1,1,2,2,2,5,5 with root=1 in the first 1 or 2 or 5.
  \\ Each vector starts with a 0 term, so a[1]=ways for partitioning the
  \\ empty set, being 1 way.
  \\
  \\ w[n+1] is the desired n vertices root=1 forms, except not those where
  \\ root=1 is in the very last term.  Such last term are a[] of n-1.  This
  \\ is by taking a[] partition of n-1 and adding root=1 in its last term.
  \\ Doing so makes its last term bigger than the preceding, like w.
  \\
  \\ For n=1, there is just a single protnum vector [0] partition 1.  It is
  \\ in w[] and is good even though root=1 is in the last partition term.
  \\ So don't want to subtract a[].  That case is left to go through to the
  \\ forests and all trees code, both of which are count 1.
  \\
  \\ a[] is extended by binomial(t,k) choose vertices to be in new
  \\ partition term k.
  \\
  \\ w[] is extended in two ways.  Firstly extending from a[] by root=1 in
  \\ the first k term.  Other vertices in that term are binomial(t-1,k-1).
  \\ Secondly by extending from w[], leaving root=1 as already set in w[],
  \\ so new vertices binomial(t-1,k).  In both cases further k terms are
  \\ binomial(t-1,k) to leave root=1 as already set.
  \\
  \\ r runs from high to low so that the a[r+1],w[r+1] starting points for
  \\ extension are not yet changed for k.

  if(forest<0 && n>=2,
     my(a=vector(n+1)); a[1]=1;
     my(w=a);
     for(k=1,n,                 \\ new term k
        forstep(r=n,0,-1,       \\ remainder, so t=r+extra
           my(b = a[r+1]);
           forstep(e=k,n-r,k,   \\ e multiple of k terms
              my(t=r+e);        \\ total new size
              a[t+1] += (b *= binomial(t,k)));

           my(t=r+k);
           if(t<=n,
              w[t+1] += (b = (a[r+1] * binomial(t-1,k-1)
                            + w[r+1] * binomial(t-1,k)));
              forstep(e=2*k,n-r,k,
                 t=r+e;
                 w[t+1] += (b *= binomial(t-1,k))));
           ));

     \\ print("a="a);
     \\ print("w="w);
     \\ a == vector(n+1,n,n--; vpar_subtree_heights_vectors_count(n,1))
     \\     || error();

     w[n+1] - a[n];
     ,
     vpar_subtree_heights_vectors_count(n,1) - (n>=2 && !forest));
}
{
  for(forest=-1,1,
     check_equal(vector(12,n,n--;vpar_protnums_vectors_count(n,forest)),
                 vector(12,n,n--;test_protnums_vectors_count_by_arrays(n,forest)),
                 Str("vpar_protnums_vectors_count() vs test_protnums_vectors_count_by_arrays() forest="forest)));
}


\\-----------------------------------------------------------------------------
progress("vpar_is_protnums_vector()");

check_equal(vpar_is_protnums_vector([]), 1, \
            "vpar_is_protnums_vector() empty");
check_equal(vpar_is_protnums_vector([0]), 1, \
            "vpar_is_protnums_vector() singleton");
check_equal(vpar_is_protnums_vector([0,0]), 0);
check_equal(vpar_is_protnums_vector([0,0],1), 1);

check_equal(vpar_is_protnums_vector([0,-1]), 0);

{
  for(forest=-1,1,
     for(n=0, test_limit_trees_search - forest,
        my(seen=List([]));  \\ Set() of protnums vectors seen
        forvec(vec=vpar_forvec_from(n,forest),
              my(vpar=vpar_from_upascend(vec),
                 protnums=vpar_protnums_vector(vpar),
                 pos=setsearch(seen,protnums,1));
              if(pos, listinsert(seen,protnums,pos)));

        check_equal(#seen, vpar_protnums_vectors_count(n,forest),
                    "vpar_protnums_vectors_count() vs all set");

        forvec(vec=vector(n,i,[-1,n+1]),
              \\ print(" vec "vec"  forest="forest);
              check_equal(vpar_is_protnums_vector(vec,forest),
                          setsearch(seen,vec)!=0,
                          "vpar_is_protnums_vector() vs setsearch"));
     ));
}


\\-----------------------------------------------------------------------------
progress("vpar_protnums_vector()");

{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i],
        protnums=vpar_protnums_vector(vpar));
     check_equal(protnums,
                 vector(#vpar,v, vpar_protnum_of_vertex(vpar,v)),
                 "vpar_protnums_vector() vs individual vpar_protnum_of_vertex()");

     \\ protnum[v] in terms of protection numbers of its children,
     \\ in the manner of
     \\   Clemens Heuberger and Helmut Prodinger, "Protection Number In
     \\   Plane Trees", Applicable Analysis and Discrete Mathematics, volume
     \\   11, 2017, pages 314-326.  http://doi.org/10.2298/AADM1702314H
     my(children=vpar_children_vector(vpar));
     for(v=1,#vpar,
        check_equal(protnums[v],
                    if(#children[v]==0, 0,
                       1+vecmin(vecextract(protnums,children[v]))))));
}

{
  my(got=[]);
  for(n=2,10,
     if(#got > #A143362_samples, break);
     my(counts=vector(n-1),
        vpar=vpar_preorder_first(n));
     until((vpar=vpar_preorder_next(vpar))=='none,
           my(vec=vpar_protnums_vector(vpar));
           counts[sum(i=1,#vec,vec[i]>=2)+1]++);
     got=concat(got,counts));
  my(len=min(#got,#A143362_samples));
  check_equal(got[1..len], A143362_samples[1..len],
              "A143362 count ordered trees n vertices of which k many are 2-protected");
}
{
  my(got=vector(10));
  for(n=1,#got,
     my(vpar=vpar_preorder_first(n));
     until((vpar=vpar_preorder_next(vpar))=='none,
           my(vec=vpar_protnums_vector(vpar));
           got[n] += (sum(i=1,#vec,vec[i]>=2)==0)));  \\ none >=2
  check_equal(got, A143363_samples[1..#got],
              "A143363 count ordered trees with no 2-protected vertices");
}
{
  my(got=[]);
  for(n=2,10,
     if(#got > #A143364_samples, break);
     my(counts=vector(n-1),
        vpar=vpar_preorder_first(n));
     until((vpar=vpar_preorder_next(vpar))=='none,
           if(vecmax(vpar_num_children_vector(vpar)) > 2, next);
           my(vec=vpar_protnums_vector(vpar));
           counts[sum(i=1,#vec,vec[i]>=2)+1]++);
     got=concat(got,counts));
  my(len=min(#got,#A143364_samples));
  check_equal(got[1..len], A143364_samples[1..len],
              "A143364 count ordered trees <=2 children, n vertices of which k are 2-protected");
}


\\-----------------------------------------------------------------------------
progress("vpar_is_protected_vertex()");

check_equal(vpar_protnum_of_vertex([0],1), 0);
check_equal(vpar_is_protected_vertex([0],1, 2),  0);
check_equal(vpar_is_protected_vertex([0],1, 1),  0);
check_equal(vpar_is_protected_vertex([0],1, 0),  1, \
            "vpar_is_protected_vertex() singleton k=0");

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     for(v=1,#vpar,
        my(protnum=vpar_protnum_of_vertex(vpar,v));
        for(k=0,#vpar,
           \\ print(vpar"  v="v" k="k);
           check_equal(vpar_is_protected_vertex(vpar,v,k),
                       protnum >= k,
                       Str("vpar_is_protected_vertex() vs vpar_protnum_of_vertex()"))));

     check_equal(test_select_vertices(vpar, (vpar,v) ->
                                      vpar_is_protected_vertex(vpar,v,0)),
                 [1..#vpar],
       "vpar_is_protected_vertex() 0 = all vertices");

     check_equal(test_select_vertices(vpar, (vpar,v) ->
                                      vpar_is_protected_vertex(vpar,v,1)),
                 vpar_set_complement(vpar,vpar_childless_vertices(vpar)),
       "vpar_is_protected_vertex() 1 vs non vpar_childless_vertices()");
  );
}


\\------------------------------------------------------------------------------
progress("vpar_automorphism_generators()");

check_equal(vpar_automorphism_generators([]), [], \
            "vpar_automorphism_generators() empty");
check_equal(vpar_automorphism_generators([0]), [], \
            "vpar_automorphism_generators() singleton");
check_equal(vpar_automorphism_generators([0,0]), [ Vecsmall([2,1]) ], \
            "vpar_automorphism_generators() 2 singletons");

check_equal(vpar_automorphism_generators([0,1,1]), \
            [Vecsmall([1, 3,2])], \
            "vpar_automorphism_generators() swap star-3");
{
  \\  1  2  3    three singletons
  my(vpar=[0,0,0],
     got_generators=vpar_automorphism_generators(vpar));
  check_equal(got_generators,
              [ Vecsmall([1, 3,2]), Vecsmall([3,1,2]) ],
              "vpar_automorphism_generators() 3 singletons");
  check_equal(#got_generators, 2/3*#vpar,
              "vpar_automorphism_generators() 3 singletons, 2/3*n many");
}
{
  \\  1  3  4
  \\     |
  \\     2
  check_equal(vpar_automorphism_generators([0,3,0,0]),
              [Vecsmall([4, 2,3, 1])],
              "vpar_automorphism_generators() 4, swap singletons");
}
{
  \\  3  4  5
  \\     |  |
  \\     1  2
  check_equal(vpar_automorphism_generators([4, 5, 0, 0, 0]),
              [Vecsmall([2,1, 3, 5,4])],
              "vpar_automorphism_generators() swap 2 paths");
}
check_equal(vpar_automorphism_generators([0,0]), \
            [ Vecsmall([2,1]) ], \
            "vpar_automorphism_generators() 2 singletons");
check_equal(vpar_automorphism_generators([0,0,0,0]), \
            [ Vecsmall([1,2, 4,3]), Vecsmall([4,1,2,3]) ], \
            "vpar_automorphism_generators() 4 singletons");
{
  \\       1
  \\     /   \       complete binary tree h=2
  \\   2      3
  \\  / \    / \
  \\ 4   5  6   7
  my(vpar=[0,1,1,2,2,3,3]);
  vpar_validate(vpar);
  my(got=vpar_automorphism_generators(vpar));
  check_equal(got,
              [ Vecsmall([1, 3,2, 6,7, 4,5]),   \\ subtrees 2,3 swap
                Vecsmall([1,2,3, 5,4, 6,7]) ],   \\ leaf 4,5 swap
              "vpar_automorphism_generators() complete binary h=2");
  check_equal(test_perm_generators_is_minimal(got), 1);
  \\ PENDING
  \\ check_equal(#got, vpar_num_automorphism_generators(vpar),
  \\             Str("vpar_automorphism_generators() complete binary h=2,"
  \\                 " vs vpar_num_automorphism_generators()"));
}
{
  \\       1     n=6
  \\       |
  \\       6
  \\     /| |\
  \\    2 3 4 5
  my(vpar=[0, 6, 6, 6, 6, 1]);
  vpar_validate(vpar);
  check_equal(vpar_automorphism_generators(vpar),
              [ Vecsmall([1,2,3, 5,4, 6]),
                Vecsmall([1, 5,2,3,4, 6]) ],
              "vpar_automorphism_generators() tree 6");
  \\ PENDING
  \\ check_equal(vpar_num_automorphism_generators(vpar), 2);
}
{
  \\       1      n=8
  \\     /   \
  \\   2      3
  \\  / \    /|\
  \\ 4   5  6 7 8
  \\
  my(vpar=[0,1,1,2,2,3,3,3]);
  vpar_validate(vpar);
  check_equal(vpar_automorphism_generators(vpar),
            [ Vecsmall([1,2,3,4,5,6, 8,7]),
              Vecsmall([1,2,3,4,5, 8,6,7]),
              Vecsmall([1,2,3, 5,4, 6,7,8]) ],
            "vpar_automorphism_generators() tree 8");
}

{
  \\        3
  \\      / | \
  \\     5  6  7
  \\     |  |  |
  \\     1  2  4
  my(vpar=[5, 6, 0, 7, 3, 3, 3]);
  vpar_validate(vpar);
  check_equal(vpar_automorphism_generators(vpar),
            [ Vecsmall([1,4,3,2,5, 7,6]),
              Vecsmall([4,1,3,2, 7,5,6]) ],
            "vpar_automorphism_generators() tree 7");
}


test_automorphisms_by_search(vpar) =
{
  my(l=List([]));
  for(i=0,(#vpar)!-1,
     my(perm=numtoperm(#vpar,i));
     if(vpar_is_automorphism(vpar,perm),
        listput(l,Vecsmall(perm))));
  Vec(l);
}

\\ WRONG
test_perm_generators_is_minimal_by_powers(vec) =
{
  if(#vec==0, return(1));
  my(id=vpar_INTERNAL_identity_perm(#vec[1]));
  for(i=1,#vec, if(vec[i]==id, return(0)));

  my(pows=apply(test_perm_powers,vec),
     set=Set(vec));
  forvec(v=vector(#pows,i,[1,#pows[i]]),
        if(vecsum(v)<=#pows+1,next());
        my(p=test_vector_product(vector(#v,i, pows[i][v[i]])));
        \\ print(v"  "p);
        if(setsearch(set,p),return(0)));
  1;
}

{
  my(vec=[ Vecsmall([2,3,1]), Vecsmall([3,1,2]) ]);
  check_equal(test_perm_generators_is_minimal(vec), 0);
  check_equal(test_perm_generators_is_minimal_by_powers(vec), 0);
}
{
  my(vec=[Vecsmall([1, 5, 6, 4, 2, 3]),    \\ (2,5) (3,6)
          Vecsmall([1, 6, 3, 5, 4, 2]),    \\ (2,6) (4,5)
          Vecsmall([2, 3, 1, 5, 6, 4]),    \\ (1,2,3) (4,5,6)
          Vecsmall([6, 4, 2, 1, 5, 3])]);  \\ (1,2,3,4,6)
  check_equal(test_perm_powers(vec[1]),
              [Vecsmall([1, 2, 3, 4, 5, 6]), Vecsmall([1, 5, 6, 4, 2, 3])]);
  check_equal(test_perm_powers(vec[2]),
              [Vecsmall([1, 2, 3, 4, 5, 6]), Vecsmall([1, 6, 3, 5, 4, 2])]);
  check_equal(test_perm_powers(vec[3]),
              [Vecsmall([1, 2, 3, 4, 5, 6]), Vecsmall([2, 3, 1, 5, 6, 4]),
               Vecsmall([3, 1, 2, 6, 4, 5])]);
  check_equal(test_perm_powers(vec[4]),
              [Vecsmall([1, 2, 3, 4, 5, 6]), Vecsmall([6, 4, 2, 1, 5, 3]),
               Vecsmall([3, 1, 4, 6, 5, 2]), Vecsmall([2, 6, 1, 3, 5, 4]),
               Vecsmall([4, 3, 6, 2, 5, 1])]);

  check_equal(#test_perms_generators_to_group(vec), 6!/2);
  check_equal(vector(#vec,i, #test_perms_generators_to_group(vec[^i])),
              [360, 360, 360, 60]);
  check_equal(test_perm_generators_is_minimal(vec), 0,
              "test_perm_generators_is_minimal() 6");
  \\ check_equal(test_perm_generators_is_minimal_by_powers(vec), 0,
  \\             "test_perm_generators_is_minimal_by_powers() 6");
}

\\ {
\\   for(n=3,6,
\\      for(rep=1,n,
\\         my(vec=vector(4,i, vpar_INTERNAL_random_perm(n)));
\\         print(vec);
\\         check_equal(test_perm_generators_is_minimal(vec),
\\                     test_perm_generators_is_minimal_by_powers(vec),
\\                     Str("test_perm_generators_is_minimal()"
\\                         " vs test_perm_generators_is_minimal_by_powers()"))));
\\ }

{
  \\     1      2
  \\          / | \
  \\         3  4  5
  my(vpar=[0, 0, 2, 2, 2]);
  check_equal(vpar_relabel_perm(vpar,[1,2,3, 5, 4]), vpar);
  check_equal(vpar_relabel_perm(vpar,[1,2, 5, 3, 4]), vpar);
  check_equal(3!, 6);
}

\\ vec1 and vec2 are vectors of integers in the range 1..n.
\\ The vectors must be the same length and all their entries distinct.
\\ Return a permutation which is a swap of each vec1[i] <-> vec2[i].
\\
test_perm_swap_lists(n,vec1,vec2) =
{
  my(perm=test_identity_perm(n));
  for(i=1,#vec1,
     perm[vec1[i]] = vec2[i];
     perm[vec2[i]] = vec1[i]);
  perm;
}

test_automorphism_generators_by_ids(vpar) =
{
  \\ print("vpar_automorphism_generators() "vpar);

  \\ subtree_vertices[v] is built as a vector of the vertices under v,
  \\ sorted by descending lexicographic order of the subtree at and below
  \\ each vertex.  v is always first in the vector.

  my(ids=vpar_automorphism_ids_vector(vpar),
     subtree_vertices=vector(#vpar,v,[v]),
     keys=vector(#vpar,v, [ids[v],vpar[v],v]),
     seq=vecsort(keys,,1),
     ret=List([]));
  for(i=1,#vpar-1,
     my(u=seq[i], v=seq[i+1], p=vpar[u]);
     if(p==vpar[v] && ids[u]==ids[v],
        listput(ret, test_perm_swap_lists(#vpar,
                                          subtree_vertices[u],subtree_vertices[v])));

     if(p,
        subtree_vertices[p]=concat(subtree_vertices[p],subtree_vertices[u]));

     subtree_vertices[u]=0; \\ kill intermediate vector
  );
  Vec(ret);
}
check_equal(test_automorphism_generators_by_ids([]), [], \
           "test_automorphism_generators_by_ids() empty");

\\ same as vpar_num_automorphism_generators()
test_num_automorphism_generators(vpar) = \
  vpar_INTERNAL_automorphisms(vpar,44);

test_perms_is_normal_in_S(n,vec) =
{
  my(l=vector(2,i, List([])));
  for(i=1,n!,
     my(perm=Vecsmall(numtoperm(n,i)));
     for(j=1,#vec,
        my(p=perm*vec[j]);
        for(k=1,2,
           my(pos=setsearch(l[k],p,1));
           if(pos, listinsert(l[k],p,pos));
           p=vec[j]*perm)));
  l[1]==l[2];
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     \\ print("vpar "vpar);

     my(got_generators = vpar_automorphism_generators(vpar));
     check_equal(#got_generators, test_num_automorphism_generators(vpar),
                 "vpar_automorphism_generators() count vs test_num_automorphism_generators()");
     for(i=1,#got_generators,
        check_equal(#got_generators[i], #vpar);
        check_equal(test_is_perm(got_generators[i]), 1));
     check_equal(test_perm_generators_is_minimal(got_generators), 1);

     if(#vpar <= test_limit_perms_search,
        my(got_perms = test_perms_generators_to_group(got_generators,#vpar));
        got_perms  = vecsort(got_perms);
        check_equal(#got_perms, vpar_num_automorphisms(vpar),
          "vpar_automorphism_generators() group size vs vpar_num_automorphisms()");

        my(generators_by_ids=test_automorphism_generators_by_ids(vpar));
        my(perms_by_ids=test_perms_generators_to_group(generators_by_ids,#vpar));
        perms_by_ids = vecsort(perms_by_ids);
        check_equal(got_perms, perms_by_ids,
                    "vpar_automorphism_generators() vs by ids");

        my(perms_by_search = test_automorphisms_by_search(vpar));
        if(got_perms!=perms_by_search, print(vpar));
        check_equal(got_perms, perms_by_search,
          "vpar_automorphism_generators() vs test_automorphisms_by_search()");

        check_equal(test_perms_is_normal_in_S(#vpar, got_perms), 1);
       );
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_seq_is_upwards()");

check_equal(vpar_seq_is_upwards([],[]), 1);
check_equal(vpar_seq_is_upwards([0,1],[2,1]), 1, "vpar_seq_is_upwards() 2,1");
check_equal(vpar_seq_is_upwards([0,1],[1,2]), 0, "vpar_seq_is_upwards() 1,2");

check_equal(vpar_seq_is_upwards([2,1],[2,1]), 0, \
            "vpar_seq_is_upwards() cycle 2,1");
check_equal(vpar_seq_is_upwards([2,1],[1,2]), 0, \
            "vpar_seq_is_upwards() cycle 1,2");

\\ PENDING
\\ check_equal(vpar_seq_is_upwards([1,1,2,3],[1,2,3,4]), 0, \
\\             "vpar_seq_is_upwards() self-loop down");
\\ check_equal(vpar_seq_is_upwards([1,1,2,3],[4,3,2,1]), 1, \
\\             "vpar_seq_is_upwards() self-loop up");


\\-----------------------------------------------------------------------------
progress("vpar_seq_is_downwards()");

check_equal(vpar_seq_is_downwards([],[]), 1);
check_equal(vpar_seq_is_downwards([0,1],[2,1]), 0, \
            "vpar_seq_is_downwards() 2,1");
check_equal(vpar_seq_is_downwards([0,1],[1,2]), 1, \
            "vpar_seq_is_downwards() 1,2");

check_equal(vpar_seq_is_downwards([2,1],[2,1]), 0, \
            "vpar_seq_is_downwards() cycle 2,1");
check_equal(vpar_seq_is_downwards([2,1],[1,2]), 0, \
            "vpar_seq_is_downwards() cycle 1,2");

\\ PENDING
\\ check_equal(vpar_seq_is_downwards([1,1,2,3],[1,2,3,4]), 1, \
\\             "vpar_seq_is_downwards() self-loop down");
\\ check_equal(vpar_seq_is_downwards([1,1,2,3],[4,3,2,1]), 0, \
\\             "vpar_seq_is_downwards() self-loop up");


\\-----------------------------------------------------------------------------
progress("vpar_free_first()");

my(n=8); check_equal(floor(n/2)+1, 5);
\\                               1 2 3 4 5  6
check_equal(vpar_free_first(8), [0,1,2,3,4, 1,6,7]);
check_equal(vpar_free_first(10), [0,1,2,3,4,5, 1,7,8,9]);

\\ levels [1,2,...,k, 2,3,...,n-k+1]     n>=4
\\ k=floor(n/2)+1     in 1-based levels
\\
test_free_first_depths_by_k(n) =
{
  my(k=floor(n/2)+1,
     depths=List([]));
  for(i=1,k, listput(depths,i));
  for(i=2,n-k+1, listput(depths,i));
  depths=Vec(depths);
  apply(d->d-1, depths);
}
test_free_first_by_k(n) = \
  vpar_make_preorder_depths(test_free_first_depths_by_k(n));

check_equal(vpar_free_first(0), []);
check_equal(vpar_free_first(1), [0], "vpar_free_first() singleton");
check_equal(vpar_free_first(2), [0,1]);
check_equal(vpar_free_first(3), [0,1,1]);
check_equal(vpar_free_first(4), [0,1,2,1]);
check_equal(vpar_free_first(5), [0,1,2,1,4]);
check_equal(vpar_free_first(8), [0,1,2,3,4, 1,6,7]);

{
  for(n=1,16,
     my(want_depths=test_free_first_depths_by_k(n));
     check_equal(vpar_is_depths_vector(want_depths), 1,
                 "test_free_first_depths_by_k() is depths");

     my(vpar=vpar_free_first(n));
     check_equal(vpar, test_free_first_by_k(n));

     my(childless=vpar_childless_vertices(vpar),
        L1_height=want_depths[childless[1]],
        L2_height=if(#childless<2,0,want_depths[childless[2]]));
     if(n%2==0,
        check_equal(L1_height, L2_height + 1),
        check_equal(L1_height, L2_height));

     \\    m=n);
     \\ for(i=1,n,
     \\    if(i==n || want_depths[i]>=want_depths[i+1],
     \\       L1_height=want_depths[i]; m=i; break()));
     \\ my(L2_height=if(m==n,0,want_depths[n]));

     \\ print("n="n"  m="m"  "L1_height" "L2_height"  "vpar);
  );
}

check_equal(vpar_free_first(8), [0,1,2,3,4, 1,6,7]);


\\-----------------------------------------------------------------------------
progress("vpar_is_preorder_horizpos_vector()");

check_equal(vpar_is_preorder_horizpos_vector([],1), 1);
check_equal(vpar_is_preorder_horizpos_vector([0],1), 1);
check_equal(vpar_is_preorder_horizpos_vector([0,1],1), 1);

\\ [0,0,1,0]    [0,1,1,0]   [0,0,1,2,1]
\\    1            1..2       1..5
\\    |               |       |
\\    2..3            3       2..3..4
\\
check_equal(vpar_horizpos_vector([0,1,1,1,0]), [0,0,1,2,1]);
check_equal(vpar_is_preorder_horizpos_vector([0,0,1,2,1],1), 1);
check_equal(vpar_is_preorder_horizpos_vector(vector(100)), 1);
check_equal(vpar_is_preorder_horizpos_vector(vector(100),1), 1);

{
  for(forest=1,1,
     for(n=0, test_limit_trees_search - forest,
        my(l=List([]),
           vpar=vpar_preorder_first(n));
        until((vpar=vpar_preorder_next(vpar,forest))=='none,
              my(horizpos=vpar_horizpos_vector(vpar),
                 pos=setsearch(l,horizpos,1));
              if(pos, listinsert(l,horizpos,pos)));
        \\ print(l);

        forvec(vec=vector(n,i,[-1,n+1]),
              \\ print("n="n" forest="forest"  vec="vec);
              check_equal(vpar_is_preorder_horizpos_vector(vec,forest),
                          setsearch(l,vec)!=0,
                          "vpar_is_horizpos_vector() vs vpar_horizpos_vector() occurring"));

        \\ check_equal(#l,
        \\             if(forest==-1, if(n==0,1, (n-1)!),
        \\                forest==0,  if(n<=1,1, n!/2),
        \\                            n!),
        \\             "vpar_is_horizpos_vector() count");
     ));
}


\\-----------------------------------------------------------------------------
progress("vpar_seq_diagqueue()");

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);

     for(down=0,1,
        my(seq=vpar_seq_diagqueue(vpar));
        check_equal(test_is_perm(seq), 1,
                    "vpar_seq_diagqueue() is perm");
        check_equal(vpar_seq_is_ordered(vpar,seq), 1,
                    "vpar_seq_diagqueue() is vpar_seq_is_ordered()");
        check_equal(vpar_seq_is_downwards(vpar,seq), 1,
                    "vpar_seq_diagqueue() is vpar_seq_is_downwards()"));
  );
}

{
  \\ count vpar_seq_diagqueue() up == vpar_seq_diagqueue() down
  for(forest=0,1,
     for(n=0, test_limit_trees_search - forest,
        my(count=0);
        forvec(vec=vpar_forvec_from(n,forest), \\ forests
              my(vpar=vpar_from_upascend(vec),
                 seq_up  =vpar_seq_diagqueue(vpar),
                 seq_down=vpar_seq_diagqueue(vpar,1),
                 same=(seq_up==seq_down));
              \\ print(vpar" "same" "seq_up" "seq_down);
              check_equal(same,
                          vecsort(vpar_diagdepths_vector(vpar))==[0..n-1],
                    "vpar_seq_diagqueue() seq up == seq down by diagdepths");
              count += same);

        check_equal(count, vpar_depths_vectors_count(n,forest),
                    "vpar_seq_diagqueue() seq up == seq down")));
}


\\-----------------------------------------------------------------------------
progress("vpar_is_horizpos_vector()");

check_equal(vpar_is_horizpos_vector([]),   1, \
            "vpar_is_horizpos_vector() empty");
check_equal(vpar_is_horizpos_vector([],1), 1);
check_equal(vpar_is_horizpos_vector([],-1),1);

check_equal(vpar_is_horizpos_vector([0],1), 1);

\\ 0,1 only by forest 2 singletons
check_equal(vpar_is_horizpos_vector([0,1],1),  1, \
            "vpar_is_horizpos_vector() 0,1");
check_equal(vpar_is_horizpos_vector([0,1],0),  0);
check_equal(vpar_is_horizpos_vector([0,1],-1), 0);

\\ 2  3    1,0,1
\\    |
\\    1
check_equal(vpar_is_horizpos_vector([1,0,1],1),  1);
check_equal(vpar_is_horizpos_vector([1,0,1],0),  0);
check_equal(vpar_is_horizpos_vector([1,0,1],-1), 0);

check_equal(vpar_is_horizpos_vector([6,6,6,6,6],1), 0, \
            "vpar_is_horizpos_vector() 66666");
check_equal(vpar_is_horizpos_vector([6,6,6,6,6],0), 0);

\\ 1  2       3       0,1,0
\\ |     or   |\
\\ 3          1 2
check_equal(vpar_is_horizpos_vector([0,1,0],1),  1, \
            "vpar_is_horizpos_vector() 010");
check_equal(vpar_is_horizpos_vector([0,1,0]),    1);
check_equal(vpar_is_horizpos_vector([0,1,0],-1), 0);

{
  for(forest=-1,1,
     for(n=0, test_limit_trees_search - forest,
        my(l=List([]));
        forvec(vec=vpar_forvec_from(n,forest),
              my(vpar=vpar_from_upascend(vec),
                 horizpos=vpar_horizpos_vector(vpar),
                 pos=setsearch(l,horizpos,1));
              if(pos, listinsert(l,horizpos,pos)));
        \\ print(l);
        forvec(vec=vector(n,i,[-1,n+1]),
              \\ print("n="n" forest="forest"  vec="vec);
              check_equal(vpar_is_horizpos_vector(vec,forest),
                          setsearch(l,vec)!=0,
                          "vpar_is_horizpos_vector() vs vpar_horizpos_vector() occurring"));

        check_equal(#l,
                    if(forest==-1, if(n==0,1, (n-1)!),
                       forest==0,  if(n<=1,1, n!/2),
                                   n!),
                    "vpar_is_horizpos_vector() count");
     ));
}


\\-----------------------------------------------------------------------------
progress("vpar_diagdepths_vector()");

check_equal(vpar_diagdepths_vector([0,1,2,3,4]), [0,1,2,3,4]);
check_equal(vpar_diagdepths_vector([0,1,0]), [0,1,1]);
check_equal(vpar_diagdepths_vector([]), []);
check_equal(vpar_diagdepths_vector([0]), [0]);

{
  \\ diagdepths of preorder not distinct
  my(vpar=[0,1]);
  vpar_validate(vpar);
  check_equal(vpar_is_preorder(vpar), 1);
  check_equal(vpar_diagdepths_vector(vpar), [0,1]);

  my(vpar=[0,0]);
  vpar_validate(vpar);
  check_equal(vpar_is_preorder(vpar), 1);
  check_equal(vpar_diagdepths_vector(vpar), [0,1]);
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got=vpar_horizpos_vector(vpar));
     check_equal(vpar_diagdepths_vector(vpar),
                 vpar_horizpos_vector(vpar) + vpar_depths_vector(vpar),
                 "vpar_diagdepths_vector() vs horizpos + depths"));
}


\\-----------------------------------------------------------------------------
progress("test_INTERNAL_siblingpos_cumulative()");

\\ Old code:
\\ down=0,1 same as vpar_siblingpos_of_vertex()
\\ d=1 for vpar_horizpos_vector(), d=0 for vpar_diagdepths_vector()
\\
test_INTERNAL_siblingpos_cumulative(vpar,down,d) =
{
  if(vpar_is_cyclic(vpar), error("for acyclics only"));

  \\ ret[] is first built as -1-siblingpos[v], so -1 for smallest sibling,
  \\ -2 for next bigger, etc.  This is like vpar_siblingpos_vector() but
  \\ -1-pos each.
  \\
  \\ Then go up chain v total siblingpos = horizpos at v, fill that in,
  \\ and also the chain upwards by subtracting siblingpos at each step.
  \\ When filling, must get the old siblingpos out before storing the
  \\ horizpos.  The "upto" variable is re-used for that.
  \\ ret[v]>=0 is a horizpos already determined.
  \\ ret[v]<0 is a siblingpos.
  \\ Checks for vpar[v]!=0 are to avoid going up past a root.
  \\ The re-traversal up could be done from a count of the first traversal,
  \\ but it's about as easy to test vpar[v] and ret[v]<0 conditions again.
  \\
  \\ d=0 is for vpar_diagdepths_vector().  The -1 in the siblingpos
  \\ negatives becomes +1 each depth of the chain up from v.
  \\ d=1 is for vpar_horizpos_vector().  It adjusts out the -1 siblingpos.
  \\

  \\ print("test_INTERNAL_siblingpos_cumulative() d="d"  "vpar);
  my(ret=apply(p->-1-p, vpar_siblingpos_vector(vpar,down)));
  \\ print(" siblingpos negs "ret);

  for(i=1,#vpar,
     if(ret[i]<0,
        \\ print(" i="i);
        my(v=i,
           t=d);  \\ total this upwards
        until(ret[v]>=0 || !(v=vpar[v]),
              t += abs(ret[v]) - d);
        t -= !v;
        \\ print("  t="t" at v="v" ret="if(v,ret[v]));
        v=i;
        my(prev=ret[v]);
        until(!(v=vpar[v]) || (prev=ret[v])>=0,
              \\ print("  set v="v" t="t); ret[v]<0 || error();
              ret[v] = t;
              t += prev + d);
        ));
  ret;
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     for(down=0,1,
        check_equal(vpar_horizpos_vector(vpar,down),
                    test_INTERNAL_siblingpos_cumulative(vpar,down,1),
                    "vpar_horizpos_vector() vs test_INTERNAL_siblingpos_cumulative()");

        check_equal(vpar_diagdepths_vector(vpar,down),
                    test_INTERNAL_siblingpos_cumulative(vpar,down,0),
                    "vpar_diagdepths_vector() vs test_INTERNAL_siblingpos_cumulative()");
        ));
}


\\-----------------------------------------------------------------------------
progress("vpar_is_automorphisms_le() Solvable Count");

{
  \\ Solvable, <=4 Siblings
  my(limit=12,
     got=vector(limit));
  for(n=1,limit,
     my(vpar=vpar_unlabelled_first(n));
     until((vpar=vpar_unlabelled_next(vpar))=='none,
           got[n] += vpar_is_automorphisms_le(vpar,4)));

  check_equal(got, A318851_samples[2..#got+1],  \\ starts OFFSET=0
              "A318851 vpar_is_automorphisms_le() solvable count");

  \\ cf A318798 count rooted unlabelled with sibling subtrees <=4 same size.
  \\ OFFSET=0 so n=0..15 same, then first differing at n=16.
  check_equal(A318851_samples[1..16],
              A318798_samples[1..16]);
  check_equal(A318851_samples[17] != A318798_samples[17],  1);

  my(L=O(x^(limit+1)),
     g = x + L);
  for(n=1,limit,     \\ root above 0 to 4 of each solvable size n=1..limit
     g *= (1 + x^n + x^(2*n) + x^(3*n) + x^(4*n) + L)^polcoeff(g,n));
  check_equal(got,Vec(g), "vpar_is_automorphisms_le() solvable count vs gf product");
}


\\-----------------------------------------------------------------------------
progress("vpar_is_automorphisms_le() Abelian Count");

{
  my(limit=11,
     got = vector(limit));
  for(n=1,limit,
     my(vpar=vpar_unlabelled_first(n));
     until((vpar=vpar_unlabelled_next(vpar))=='none,
           got[n] += vpar_is_automorphisms_le(vpar,2,1)));

  my(a_vec = vpar_count_unlabelled_asymmetric_vector(limit),
     got_symmetric = got - a_vec);  \\ Abelians with at least one symmetry

  \\ print(got);
  \\ 1, 1, 2, 3, 7, 15, 33, 77, 184, 446, 1098
  \\ not in OEIS: 2, 3, 7, 15, 33, 77, 184, 446, 1098

  \\ print(got_symmetric);
  \\ 0, 0, 1, 1, 4, 9, 21, 52, 132, 333, 851
  \\ not in OEIS: 4, 9, 21, 52, 132, 333, 851

  my(L=O(x^(limit+1)),
     g=x+L);      \\ c*x^n = how many n vertex trees
  for(n=1,limit-1,
     my(a = a_vec[n],         \\ num asymmetrics
        A = polcoeff(g,n),    \\ num Abelians
        A_symmetric = A - a); \\ num Abelian symmetrics
     g *= (1 + x^n +L          )^A_symmetric  \\ 1 Abelian symmetric
        * (1 + x^n + x^(2*n) +L)^a);          \\ 1 or 2 asymmetric
  check_equal(got, Vec(g),
              "vpar_is_automorphisms_le() Abelian count vs gf product");

  \\ WRONG
  \\ my(got_func          =(n)->if(n==0,1,got[n]));
  \\ my(got_symmetric_func=(n)->if(n==0,1,got_symmetric[n]));
  \\ check_equal(vector(limit,n,
  \\                    sum(k=0,n-1, got_func(n-k) * got_symmetric_func(k))
  \\                    +
  \\                    sum(k=1,n\2, got_func(n-2*k) * vpar_count_unlabelled_asymmetric(k))),
  \\            got);

  \\ WRONG
  \\ my(g=x+L);      \\ c*x^n = how many n vertex trees
  \\ for(n=1,limit-1,
  \\    my(a = vpar_count_unlabelled_asymmetric(n));
  \\    g *= (1 + x^n     +L)^polcoeff(g,n)    \\ 1 Abelian
  \\       * (1 + x^(2*n) +L)^a);              \\ 2 asymmetric
  \\ check_equal(got, Vec(g),
  \\             "vpar_is_automorphisms_le() Abelian count vs gf product");

  \\ WRONG
  \\ my(g = x^3 + L);    \\ c*x^n = how many n vertex trees
  \\ for(n=1,limit-1,
  \\    my(a = vpar_count_unlabelled_asymmetric(n));
  \\    print(a"  "Vecrev(Pol(g)));
  \\    g *= (1 + x^n     +L)^polcoeff(g,n+1)  \\ 1 Abelian symmetric
  \\       * (1 + x^(2*n+1) +L)^a);            \\ 2 asymmetric
  \\ check_equal(got_symmetric, Vecrev(Pol(g)),
  \\             "vpar_is_automorphisms_le() Abelian symmetrics count vs gf product");
}


\\-----------------------------------------------------------------------------
progress("vpar_is_automorphisms_le()");

{
  \\    1       2      3
  \\   / \     / \    / \
  \\  4   5   6   7  8   9
  my(vpar = [0,0,0, 1,1, 2,2, 3,3]);
  check_equal(vpar_is_automorphisms_le(vpar,3,2), 0,
             "vpar_is_automorphisms_le() 3 of 2");
  check_equal(vpar_is_automorphisms_le(vpar,3,3), 1);

  check_equal(vpar_is_automorphisms_le([0,0,0],3,2), 1,
             "vpar_is_automorphisms_le() 3 singletons");
}
{
  \\ Supersolvable
  \\        1
  \\      /   \
  \\    2       3
  \\   /|\     /|\
  \\  4 5 6   7 8 9
  \\
  my(vpar=[0,1,1, 2,2,2, 3,3,3]);
  vpar_validate(vpar);

  my(perms=vpar_automorphisms(vpar));
  check_equal(test_perms_is_2group(perms), 0, "tree complete 2,3 not 2-group");
  check_equal(test_perms_is_supersolvable(perms), 0);
  check_equal(test_perms_is_solvable(perms), 1);

  check_equal(vpar_is_automorphisms_le(vpar,2), 0);
  check_equal(vpar_is_automorphisms_le(vpar,3), 1);
  check_equal(vpar_is_automorphisms_le(vpar,3,2), 0);
}

{
  \\               1
  \\           /       \
  \\        2               3
  \\   / /  | \ \      /  / | \ \
  \\  4  5  6  7 8    9 10 11 12 13
  \\  |  |  |         |  |  |
  \\ 14 15 16        17 18 19
  \\
  my(vpar=[0,1,1, 2,2,2,2,2, 3,3,3,3,3, 4,5,6, 9,10,11]);
  vpar_validate(vpar);
  check_equal(vpar_num_automorphisms(vpar), (3!*2)^2 * 2);
  check_equal(vpar_num_automorphisms(vpar), 288);

  check_equal(vpar_is_automorphisms_le(vpar,2), 0);
  check_equal(vpar_is_automorphisms_le(vpar,3), 1);
  check_equal(vpar_is_automorphisms_le(vpar,3,2), 0, "2,3,2");
}
{
  \\               1
  \\           /       \
  \\        2               3
  \\   / /  | \ \      /  / | \ \
  \\  4  5  6  7 8    9 10 11 12 13
  \\  |  |            |  |
  \\ 14 15           16 17
  \\
  my(vpar=[0,1,1, 2,2,2,2,2, 3,3,3,3,3, 4,5, 9,10]);
  vpar_validate(vpar);
  check_equal(#vpar,17);
  check_equal(vpar_num_automorphisms(vpar), (3!*2)^2 * 2);
  check_equal(vpar_num_automorphisms(vpar), 288);

  check_equal(vpar_is_automorphisms_le(vpar,2), 0, "2,2,3 not le 2");
  check_equal(vpar_is_automorphisms_le(vpar,3), 1);
  check_equal(vpar_is_automorphisms_le(vpar,3,2), 0, "2,2,3");
}
{
  \\ Supersolvable
  \\        1
  \\      /   \
  \\    2       3
  \\    |       |
  \\    4       5
  \\    |       |
  \\    6       7
  \\   /|\     /|\
  \\  8 9 10 11 12 13
  \\
  my(vpar=[0,1,1, 2,3, 4,5, 6,6,6, 7,7,7]);
  vpar_validate(vpar);

  my(perms=vpar_automorphisms(vpar));
  check_equal(test_perms_is_2group(perms), 0, "tree stretched complete 2,3 not 2-group");
  check_equal(test_perms_is_supersolvable(perms), 0);
  check_equal(test_perms_is_solvable(perms), 1);

  check_equal(vpar_is_automorphisms_le(vpar,2), 0);
  check_equal(vpar_is_automorphisms_le(vpar,3), 1);
  check_equal(vpar_is_automorphisms_le(vpar,3,2), 0,
              "vpar_is_automorphisms_le() 3,2 of tree stretched complete 2,3");
}
{
  \\ Abelian
  my(vpar=vpar_make_complete_tree(2,2),   \\ S2 wr S2
     perms=vpar_automorphisms(vpar));
  check_equal(#perms, 8);
  check_equal(test_perms_is_Abelian(perms), 0,
             "test_perms_is_Abelian() complete tree 2,2, no");

  check_equal(vpar_is_automorphisms_le(vpar,2,1), 0,
             "vpar_is_automorphisms_le() Abelian, complete tree 2,2, no");

  \\ relabelled as [0, 1, 2, 2, 1, 5, 5]
  \\        1
  \\      /   \
  \\    2       5
  \\   / \     / \
  \\  3   4   6   7
  vpar = vpar_to_unlabelled(vpar);
  check_equal(vpar_is_automorphisms_le(vpar,2,1), 0,
             "vpar_is_automorphisms_le() Abelian complete tree 2,2, no");
}

{
  for(h=0,4,
     my(vpar=vpar_make_complete_tree(2,h));
     check_equal(vpar_is_automorphisms_le(vpar,3,2), 1,
                 "vpar_is_automorphisms_le() supersolvable, complete binary tree"));
  for(h=0,2,
     my(vpar=vpar_make_complete_tree(3,h));
     check_equal(vpar_is_automorphisms_le(vpar,3,2), h<=1,
                 "vpar_is_automorphisms_le() supersolvable, complete ternary tree"));
}

\\ Looking for anywhere >2 siblings same, or =2 siblings with some symmetry below.
test_is_automorphisms_Abelian_by_ids(vpar) =
{
  my(ids=vpar_automorphism_ids_vector(vpar),
     seq=vecsort(ids,,1),
     prev_id=vectorsmall(#vpar+1),
     count =vectorsmall(#vpar+1));
  for(i=1,#seq,
     my(v=seq[i],
        p=vpar[v]);
     if(prev_id[p+1] == ids[v],
        count[p+1]++;
        count[p+1]<=2 || return(0);
        vpar_is_asymmetric(vpar_subtree(vpar,v)) || return(0);
        ,
        prev_id[p+1] = ids[v];
        count[p+1]=1);
     );
  1;
}
check_equal(test_is_automorphisms_Abelian_by_ids(vpar_make_complete_tree(3,2)), 0);
check_equal(test_is_automorphisms_Abelian_by_ids(vpar_make_complete_tree(2,2)), 0);
check_equal(test_is_automorphisms_Abelian_by_ids(vpar_make_path(4)), 1);

\\ Disallowing anywhere >2 siblings same
test_is_automorphisms_2group_by_ids(vpar) =
{
  my(ids=vpar_automorphism_ids_vector(vpar),
     seq=vecsort(ids,,1),
     prev_id=vectorsmall(#vpar+1),
     count =vectorsmall(#vpar+1));
  for(i=1,#seq,
     my(v=seq[i],
        p=vpar[v]);
     if(prev_id[p+1] == ids[v],
        count[p+1]++;
        count[p+1]<=2 || return(0);
        ,
        prev_id[p+1] = ids[v];
        count[p+1]=1);
     );
  1;
}
check_equal(test_is_automorphisms_2group_by_ids(vpar_make_complete_tree(3,2)), 0);
check_equal(test_is_automorphisms_2group_by_ids(vpar_make_complete_tree(2,2)), 1);
check_equal(test_is_automorphisms_2group_by_ids(vpar_make_path(4)), 1);

{
  my(vpar=vpar_make_complete_tree(3,2),   \\ S3 wr S3
     perms=vpar_automorphisms(vpar));
  check_equal(#perms, 1296);
  check_equal(test_perms_is_solvable(perms), 1,
             "test_perms_is_solvable() complete tree 3,2, yes");
  check_equal(test_perms_is_supersolvable(perms), 0,
             "test_perms_is_solvable() complete tree 3,2, is not");
}

{
  \\ Charles Wells, "Some Applications of the Wreath Product Construction",
  \\ American Mathematical Monthly, volume 83, number 5, May 1976, pages
  \\ 317-338.  https://www.jstor.org/stable/2318641
  \\ Figure 15.3 example wreath product tree automorphism group.
  \\ (S2 wr S3) x (S2 wr S3 wr S2).
  \\
  \\                        1
  \\         /    /    /          \                  \
  \\        /    /    /             \                  \
  \\        |    |    |              11                 12
  \\        |    |    |           /   |   \          /   |    \
  \\        2    3    4         13   14    15      16   17    18
  \\       / \  / \  / \       / \   / \   / \    / \   / \   / \
  \\      5  6  7 8  9 10     19 20 21 22 23 24  25 26 27 28 29 30
  my(vpar=[0, 1,1,1, 2,2,3,3,4,4,
           1,1, 11,11,11, 12,12,12, 13,13,14,14,15,15, 16,16,17,17,18,18]);
  vpar_validate(vpar);

  check_equal(vpar_num_automorphisms(vpar), 2^3*3! * (2^3 * 3!)^2 * 2!);
  check_equal(221184,                       2^3*3! * (2^3 * 3!)^2 * 2!);

}

{
  my(samples=concat(test_sample_forests,
                    [
                     [0,1,1,2,2,2,3,3,3],  \\ tree 2,3 = S3 wr S2
                     [0,1,1,1,1, 0,0 ],    \\ S4 x S2
                     [0],
                     [0,1,1],
                     [0,1,1,1],
                     [0,1,1,1,1],
                     [0,1,1,1,1,1],
                     [0,1,1,1,1,1,1],
                     vpar_make_complete_tree(2,2),     \\ 8
                     vpar_make_complete_tree(2,3),     \\ 128
                     \\ vpar_make_complete_tree(2,4)   \\ 32768 automorphisms
                     vpar_make_complete_tree(3,1),     \\ 6
                     vpar_make_complete_tree(3,2)      \\ 1296
                    ]));
  for(i=1,#samples,
     my(vpar=samples[i],
        got_asymmetric    = vpar_is_automorphisms_le(vpar,0,0),
        got_Abelian       = vpar_is_automorphisms_le(vpar,2,1),
        got_2group        = vpar_is_automorphisms_le(vpar,2),
        got_supersolvable = vpar_is_automorphisms_le(vpar,2,3),
        got_solvable      = vpar_is_automorphisms_le(vpar,4),
        num_automorphisms = vpar_num_automorphisms(vpar));

     my(subdivided=vpar_subdivide(vpar,3));
     check_equal(matrix(6,6,m,t,m--;t--; vpar_is_automorphisms_le(vpar,m,t)),
                 matrix(6,6,m,t,m--;t--; vpar_is_automorphisms_le(subdivided,m,t)));

     \\ finitely generated Abelian groups are subset of the 2-groups
     \\
     check_equal(got_asymmetric    <= got_Abelian,       1);
     check_equal(got_Abelian       <= got_2group,        1);
     check_equal(got_2group        <= got_supersolvable, 1);
     check_equal(got_supersolvable <= got_solvable,      1);

     check_equal(got_asymmetric, vpar_is_asymmetric(vpar));

     \\ Feit-Thompson: finite groups of odd order are solvable, though for
     \\ tree automorphism groups only the trivial group is odd
     if(num_automorphisms % 2,
        check_equal(got_solvable, 1,
          "vpar_is_automorphisms_le() solvable, vs odd num_automorphisms"));

     check_equal(got_2group, test_is_automorphisms_2group_by_ids(vpar),
       "vpar_is_automorphisms_le() 2-group vs test_is_automorphisms_2group_by_ids()");

     check_equal(got_Abelian, test_is_automorphisms_Abelian_by_ids(vpar),
       "vpar_is_automorphisms_le() Abelian vs test_is_automorphisms_Abelian_by_ids()");

     if(num_automorphisms <= test_limit_sets,
        \\ print(vpar" has "vpar_num_automorphisms(vpar));
        my(perms = vpar_automorphisms(vpar));

        \\ -------
        \\ Abelian

        check_equal(got_Abelian, test_perms_is_Abelian(perms),
          "vpar_is_automorphisms_le() Abelian vs test_perms_is_Abelian()");
        check_equal(got_Abelian, test_perms_is_involutions(perms),
          "vpar_is_automorphisms_le() Abelian vs test_perms_is_involutions()");

        \\----------
        \\ 2-group

        check_equal(got_2group, test_is_pow2(num_automorphisms),
          "vpar_is_automorphisms_le() 2-group vs test_is_pow2(vpar_num_automorphisms())");
        check_equal(got_2group, test_perms_is_2group(perms),
          "vpar_is_automorphisms_le() 2-group vs test_perms_is_2group()");

        \\----------
        \\ nilpotent == 2-group for tree automorphism groups

        check_equal(got_2group, test_perms_is_nilpotent(perms),
         "vpar_is_automorphisms_le() 2-group vs test_perms_is_nilpotent()");
        check_equal(got_2group, test_perms_nilpotency_class(perms)!='none,
         "vpar_is_automorphisms_le() 2-group vs test_perms_nilpotency_class()");

        \\--------------
        \\ supersolvable

        \\ FIXME: test_perms_is_supersolvable not right
        \\ check_equal(got_solvable, test_perms_is_solvable(perms),
        \\   "vpar_is_automorphisms_le() supersolvable vs test_perms_is_supersolvable()");

        \\---------
        \\ solvable

        check_equal(got_solvable, test_perms_is_solvable(perms),
          "vpar_is_automorphisms_solvable() vs test_perms_is_solvable()");
        );

     \\ print("vpar="vpar);
     \\ print("perms="perms);
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_is_automorphisms_le() Unlabelled 2-Group Count");
\\ A248869 unlabelled
\\ A309352 free (mine)

check_equal(test_count_unlabelled_2group_gf(6), \
            1 + x + x^2 + 2*x^3 + 3*x^4 + 7*x^5 + O(x^6));  \\ 0 up, 6 terms
check_equal(test_count_unlabelled_2group_vector(6), \
            [1,1,2,3,7,15]);  \\ 1 to 6
check_equal(#test_count_unlabelled_2group_vector(6), 6);

check_equal(test_count_unlabelled_2group(0), 1);
check_equal(test_count_unlabelled_2group(3), 2);
check_equal(vpar_count_unlabelled       (3), 2, "vpar_count_unlabelled(3)");
check_equal(test_count_unlabelled_2group(4), 3);
check_equal(vpar_count_unlabelled       (4), 4, "vpar_count_unlabelled(4)");

{
  for(n=0,12,
     my(vpar=vpar_unlabelled_first(n),
        count=0);
     until((vpar=vpar_unlabelled_next(vpar))=='none,
           count += vpar_is_automorphisms_le(vpar,2));
     check_equal(count, test_count_unlabelled_2group(n),
                 "vpar_is_automorphisms_le() count vs test_count_unlabelled_2group()"));

  check_equal(concat(0,
                     test_count_unlabelled_2group_vector(#A248869_samples-1)),
              A248869_samples,
              "A248869 number of rooted unlabelled 2-group");

  \\ same as test_count_unlabelled_2group_vector() code, for now
  my(limit=20,
     L=O(x^limit),
     g = x + L);
  for(n=1,limit-1,   \\ root above 0 to 2 of each solvable size n=1..limit
     g *= (1 + x^n + x^(2*n) + L)^polcoeff(g,n));
  check_equal(test_count_unlabelled_2group_gf(limit) - 1,
              g,
              "vpar_is_automorphisms_le() 2-group count vs gf product");

  \\ 2-group with symmetry, meaning 2-group excluding the trivial group
  check_equal((test_count_unlabelled_2group_vector(#A318859_samples+1)
               - vpar_count_unlabelled_asymmetric_vector(#A318859_samples+1))
              [^1],
              A318859_samples,  \\ OFFSET=2
              "A318859 count max isomorphic siblings = 2, so 2-groups without asymmetrics");
}
A318859_by_samples(n) = if(n<2,0, A318859_samples[n-1]);
{
  check_equal(vector(#A318859_samples,n,n++; A318859_by_samples(n)),
              A318859_samples,   \\ OFFSET=2
              "A318859");
}

A248869(n) = if(n==0,0, test_count_unlabelled_2group(n));
{
  check_equal(vector(#A248869_samples,n,n--; A248869(n)),
              A248869_samples,   \\ OFFSET=0
              "A248869");
}

\\ A004111 reckons 0 of n=0
A004111(n) = if(n==0,0, vpar_count_unlabelled_asymmetric(n));
{
  check_equal(vector(#A004111_samples,n,n--; A004111(n)),
              A004111_samples,    \\ OFFSET=0
              "A004111");

  check_equal(vector(30,n,n--; A248869(n)),
              vector(30,n,n--; A318859_by_samples(n) + A004111(n)),
              "A248869 = A004111 + A318859");
}


\\-----------------------------------------------------------------------------
progress("vpar_is_automorphisms_le() Supersolvable Count");

{
  my(limit=12,
     got_supersolvable=vector(limit));
  for(n=1,limit,
     my(vpar=vpar_unlabelled_first(n));
     until((vpar=vpar_unlabelled_next(vpar))=='none,
           got_supersolvable[n] += vpar_is_automorphisms_le(vpar,3,2)));

  \\ print(got_supersolvable);
  \\ not in OEIS: 1, 2, 4, 8, 18, 43, 102, 249, 620, 1571, 4022

  my(count_2group=test_count_unlabelled_2group_vector(limit),
     L=O(x^(limit+1)),
     g = x + L);
  for(n=1,limit,
     \\ 0 to 1  supersolvable with 3 siblings
     \\ 0 to 2  supersolvable without 3 siblings but with symmetry,
     \\           which is symmetric 2-group
     \\ 0 to 3  asymmetrics
     my(a = vpar_count_unlabelled_asymmetric(n),
        s           = polcoeff(g,n),
        t           = count_2group[n],
        s_with_3    = s - t,
        t_symmetric = t - a);
     g *= (1 + x^n                     + L)^s_with_3
        * (1 + x^n + x^(2*n)           + L)^t_symmetric
        * (1 + x^n + x^(2*n) + x^(3*n) + L)^a;

     if(n>=2,
        check_equal(t_symmetric, A318859_samples[n-1],
                    "A318859 count max isomorphic siblings = 2, so 2-groups with symmetry")));

  \\ print(g);
  \\ print(got_supersolvable);
  check_equal(got_supersolvable, Vec(g),
              "vpar_is_automorphisms_le() supersolvable count vs gf product");

  \\ asym          1, 1, 1, 2, 3
  \\ 2-groups      1, 1, 2, 3, 7
  \\ 2-groups sym  0, 0, 1, 1, 4
  \\ supersolvable 1, 1, 2, 4, 8
  \\ n=5 supersolvables
  \\ [0, 1, 2, 3, 4]   path-5
  \\ [0, 1, 2, 3, 3]  1--2--3--4
  \\                         \-5
  \\ [0, 1, 2, 3, 2]  1--2--3--4
  \\                      \-5
  \\ [0, 1, 2, 3, 1]  1--2--3--4
  \\                   \-5
  \\ [0, 1, 2, 2, 2]    1
  \\                    |
  \\                    2
  \\                   /|\
  \\                  3 4 5
  \\ [0, 1, 2, 2, 1]      1
  \\                     / \
  \\                    2   5
  \\                   / \
  \\                  3   4
  \\ [0, 1, 2, 1, 4]    1
  \\                   / \
  \\                  2   4
  \\                  |   |
  \\                  3   5
  \\ [0, 1, 2, 1, 1]    1
  \\                   /|\
  \\                  2 4 5
  \\                  |
  \\                  3
}


\\-----------------------------------------------------------------------------
progress("vpar_is_automorphisms_le() <=3 Count");

{
  my(limit=12,
     got=vector(limit));
  for(n=1,limit,
     my(vpar=vpar_unlabelled_first(n));
     until((vpar=vpar_unlabelled_next(vpar))=='none,
           got[n] += vpar_is_automorphisms_le(vpar,3)));

  check_equal(got,
              A318797_samples[2..limit+1],  \\ OFFSET=0
  "A318797 number of rooted unlabelled trees with isomorphic sibling sets <=3");

  my(L=O(x^(limit+1)),
     g = x + L);
  for(n=1,limit,     \\ root above 0 to 3 of each solvable size n=1..limit
     g *= (1 + x^n + x^(2*n) + x^(3*n) + L)^polcoeff(g,n));
  check_equal(got,Vec(g), "vpar_is_automorphisms_le() <=3 count vs gf product");
}


\\-----------------------------------------------------------------------------
progress("vpar_automorphism_ids_vector()");

check_equal(vpar_automorphism_ids_vector([0,0]),[1,1], \
            "vpar_automorphism_ids_vector() two isolated");
check_equal(vpar_automorphism_ids_vector([0,1]),[2,1]);
check_equal(vpar_automorphism_ids_vector([0,1,1,1]),[4,1,1,1]);
check_equal(vpar_automorphism_ids_vector([0,1,2,3]),[4,3,2,1], "path-4");
check_equal(vpar_automorphism_ids_vector([0,1,2,2]),[4,3,1,1]);

{
  my(vpar=[0,1,1, 2,2,2,2, 3,3,3,3]);
  check_equal(vpar_automorphism_ids_vector(vpar), [7, 5,5, 1,1,1,1, 1,1,1,1]);
  check_equal(#vpar, 11);
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        ids=vpar_automorphism_ids_vector(vpar));
     check_equal(#ids, #vpar);
     if(#vpar,
        check_equal(vecmin(ids), 1,
                   "vpar_automorphism_ids_vector() minimum id");
        check_equal(vecmax(ids) <= #vpar, 1,
                   "vpar_automorphism_ids_vector() maximum id");

        for(rep=1,#vpar+3,
           my(u=vpar_random_vertex(vpar),
              v=vpar_random_vertex(vpar));
           check_equal(ids[u]==ids[v],
                       vpar_is_isomorphic(vpar_subtree(vpar,u),
                                          vpar_subtree(vpar,v)),
                       "vpar_automorphism_ids_vector() same ids are vpar_is_isomorphic() subtrees"));
        );
     \\ print(vpar" ids "ids);

     \\ FIXME:
     \\ my(distinct=Set(ids));
     \\ check_equal(#distinct==#vpar, test_is_path(vpar),
     \\             "vpar_automorphism_ids_vector() full set ids is path");
     \\ check_equal(#distinct, vpar_num_automorphism_ids(vpar),
     \\   "vpar_automorphism_ids_vector() distinct = vpar_num_automorphism_ids()");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_num_automorphism_ids()");

check_equal(vpar_num_automorphism_ids([]), 0, \
            "vpar_num_automorphism_ids() empty");
check_equal(vpar_height([]), -1);

check_equal(vpar_num_automorphism_ids([0]), 1, \
            "vpar_num_automorphism_ids() singleton");
check_equal(vpar_height([0]), 0);

check_equal(vpar_num_automorphism_ids([0,1]), 2, \
            "vpar_num_automorphism_ids() path-2");
check_equal(vpar_num_automorphism_ids([0,0]), 1, \
            "vpar_num_automorphism_ids() 2 singletons");

{
  for(n=1,10,
     my(vpar=vpar_make_path(n));
     check_equal(vpar_num_automorphism_ids(vpar), n,
                 Str("vpar_num_automorphism_ids() path n="n));
     check_equal(vpar_height(vpar), n-1));
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got=vpar_num_automorphism_ids(vpar));

     check_equal(vpar_height(vpar)+1 <= got, 1,
                 "vpar_num_automorphism_ids() range low");
     check_equal(got <= #vpar, 1,
                 "vpar_num_automorphism_ids() range low");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_automorphism_group()");

check_equal(vpar_automorphism_group([]), [[],Vecsmall([])], \
            "vpar_automorphism_group() empty");
{
  \\ S2
  my(group=vpar_automorphism_group([0,0]));
  test_smallgroup_validate(group);
  check_equal(group,
              [[Vecsmall([2,1])], Vecsmall([2])],
              "vpar_automorphism_group() 2 singletons");

}
{
  \\ S3
  my(want_gens=apply(s->for(i=1,#s,s[i]++);s,
                    vpar_INTERNAL_S3_generators),
     group=vpar_automorphism_group([0,0,0]));
  test_smallgroup_validate(group);
  check_equal(group,
              [ want_gens, Vecsmall([3,2]) ],
              "vpar_automorphism_group() 3 singletons");
}
{
  \\ S4
  my(want_gens=apply(s->for(i=1,#s,s[i]++);s,
                    vpar_INTERNAL_S4_generators),
     group=vpar_automorphism_group([0,0,0,0]));
  test_smallgroup_validate(group);
  check_equal(group,
              [ want_gens, Vecsmall([2,2,3,2]) ],
              "vpar_automorphism_group() 4 singletons");
}
{
  \\ S2 x S2
  my(group=vpar_automorphism_group([0,1,1, 0,4,4, 5,6]));
  test_smallgroup_validate(group);
  check_equal(group,
              [ [Vecsmall([2,1, 4,3]),
                 Vecsmall([3,4, 1,2]) ],
               Vecsmall([2,2])],
              "vpar_automorphism_group() S2xS2");
  my(S2 = test_perms_Sn(2),
     S2xS2 = test_perms_direct_product(S2,S2),
     gens=group[1],
     perms=test_perms_generators_to_group(gens));
  check_equal(test_perms_is_group(perms), 1);
  check_equal(test_perms_is_isomorphic(perms,S2xS2), 1,
              "vpar_automorphism_group() S2xS2 test_perms_is_isomorphic()");
}
{
  \\ S3 x S4
  \\          1
  \\       /     \
  \\     2        3
  \\   /| |\    /| |
  \\  4 5 6 7  8 9 10
  \\
  my(group=vpar_automorphism_group([0,1,1, 2,2,2,2, 3,3,3]));
  test_smallgroup_validate(group);
  check_equal(group[2], Vecsmall([3,2, 2,2,3,2]));

  my(S3 = test_perms_Sn(3),
     S4 = test_perms_Sn(4),
     S3xS4 = test_perms_direct_product(S3,S4),
     gens=group[1],
     perms=test_perms_generators_to_group(gens));
  check_equal(#perms, 144);
  check_equal(test_perms_is_group(perms), 1);

  check_equal(vecsort(apply(test_permorder,S3xS4)),
              vecsort(apply(test_permorder,perms)));
  \\ bit slow
  \\ check_equal(test_perms_is_isomorphic(perms,S3xS4), 1,
  \\             "vpar_automorphism_group() S3xS4 test_perms_is_isomorphic()");
}
{
  \\ S2 wr S2
  my(vpar=vpar_make_complete_tree(2,2),
     group=vpar_automorphism_group(vpar));
  test_smallgroup_validate(group);
  my(gens=group[1],
     perms=test_perms_generators_to_group(gens),
     S2 = test_perms_Sn(2),
     S2wrS2 = test_perms_wreath_product(S2,S2));
  check_equal(test_perms_is_isomorphic(perms,S2wrS2), 1);
}
{
  \\ S2 wr S2 wr S2
  my(vpar=vpar_make_complete_tree(2,3),
     group=vpar_automorphism_group(vpar));
  test_smallgroup_validate(group);
  my(gens=group[1],
     orders=group[2]);
  test_perms_validate(gens);
  check_equal(#gens, 4+2+1);
  check_equal(orders, Vecsmall([2,2,2,2, 2,2, 2]));

  my(perms=test_perms_generators_to_group(gens));
  test_perms_validate(perms);
  check_equal(#perms, 128);
  check_equal(test_perms_is_group(perms), 1);

  my(S2 = test_perms_Sn(2),
     S2wrS2wrS2 = test_perms_wreath_power(S2,3));
  check_equal(#S2wrS2wrS2, 128, "S2wrS2wrS2 num perms");
  check_equal(vecsort(apply(test_permorder,S2wrS2wrS2)),
              vecsort(apply(test_permorder,perms)));
  check_equal(test_perms_is_group(S2wrS2wrS2), 1);
  check_equal(test_perms_is_isomorphic(vecsort(perms),vecsort(S2wrS2wrS2)), 1,
             "vpar_automorphism_group() vs test_perms_is_isomorphic() S2wrS2wrS2 sorted");
  check_equal(test_perms_is_isomorphic(perms,S2wrS2wrS2), 1,
             "vpar_automorphism_group() vs test_perms_is_isomorphic() S2wrS2wrS2");
}

{
  \\ (S2 x S2) wr S2
  \\          1
  \\       /     \
  \\     2        3
  \\   /| |\    /| | \
  \\  4 5 6 7  8 9 10 11
  \\  | |          |  |
  \\ 12 13        14 15
  \\
  my(vpar=[0,1,1, 2,2,2,2, 3,3,3,3, 4,5, 10,11]);
  vpar_validate(vpar);
  my(group=vpar_automorphism_group(vpar));
  test_smallgroup_validate(group);
  my(S2 = test_perms_Sn(2),
     S2xS2 = test_perms_direct_product(S2,S2),
     S2xS2wrS2 = test_perms_wreath_product(S2xS2,S2));
  check_equal(#S2xS2wrS2, 32, "S2xS2wrS2 size");
  my(gens=group[1],
     perms=test_perms_generators_to_group(gens));
  check_equal(#perms, 4^2*2);
  check_equal(#perms, 32);
  check_equal(test_perms_is_group(perms), 1,
              "vpar_automorphism_group() S2xS2wrS2 vs test_perms_is_group()");
  check_equal(vecsort(apply(test_permorder,S2xS2wrS2)),
              vecsort(apply(test_permorder,perms)));

  \\ bit slow
  \\ check_equal(test_perms_is_isomorphic(perms,S2xS2wrS2), 1,
  \\             "vpar_automorphism_group() S2xS2wrS2 test_perms_is_isomorphic()");
}

{
  for(n=1,8,
     my(vpar=vector(n),  \\ singletons
        group=vpar_automorphism_group(vpar));
     check_equal(group=='none, n>=5));
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        num_automorphisms=vpar_num_automorphisms(vpar));
     num_automorphisms <= 1024 || next;
     \\ print(vpar);

     my(group=vpar_automorphism_group(vpar));
     if(vpar_is_automorphisms_le(vpar,3,2),
        check_equal(group=='none, 0,
              "vpar_automorphism_group() not 'none on supersolvable"));
     if(group=='none, next);

     check_equal(#group, 2, "vpar_automorphism_group() two elems");
     for(i=1,#group[1],
        check_equal(type(group[1][i]), "t_VECSMALL");
        check_equal(#group[1][i], num_automorphisms);
        check_equal(test_is_perm(group[1][i]), 1));
     check_equal(type(group[2]), "t_VECSMALL");
     check_equal(Vecsmall(apply(test_permorder,group[1])), group[2]);

     num_automorphisms <= 256 || next;
     my(whole=test_perms_generators_to_group(group[1],
                                             if(group[1]==0,1,-1)));
     check_equal(#whole, num_automorphisms,
              "vpar_automorphism_group() whole vs vpar_num_automorphisms()");
     my(perms=vpar_automorphisms(vpar));
     check_equal(#perms, num_automorphisms);
     check_equal(test_perms_is_isomorphic(whole,perms), 1,
              "vpar_automorphism_group() isomorphic to vpar_automorphisms()");
     );
}

\\ {
\\   \\ S4 x S4
\\   my(vpar=[0,1,1,1,1, 1,6, 1,8, 1,10, 1,12]);
\\   vpar_validate(vpar);
\\   check_equal(vpar_num_automorphisms(vpar), 4!^2);
\\   check_equal(vpar_automorphism_group(vpar), 'none);
\\ }


\\-----------------------------------------------------------------------------
progress("vpar_is_childful_vector()");

{
  \\ preorder
  \\ childful = at each balanced binary 1, the next bit
  \\            is 1 for child or 0 for not
  for(n=0,8,
     my(vpar=vpar_preorder_first(n));
     until((vpar=vpar_preorder_next(vpar,1))=='none,
           my(b=binary(vpar_to_balanced_binary(vpar)),
              l=List([]));
           for(i=1,#b, if(b[i], listput(l,b[i+1])));
           check_equal(vpar_is_childful_vector(vpar), Vec(l),
                       "vpar_is_childful_vector() vs balanced binary next bit")));
}


\\-----------------------------------------------------------------------------
progress("vpar_is_childless_vector()");

{
  \\ preorder
  \\ childful = at each balanced binary 1, the next bit
  \\            is 1 for child or 0 for not
  for(n=0,8,
     my(vpar=vpar_preorder_first(n));
     until((vpar=vpar_preorder_next(vpar,1))=='none,
           my(b=binary(vpar_to_balanced_binary(vpar)),
              l=List([]));
           for(i=1,#b, if(b[i], listput(l,!b[i+1])));
           check_equal(vpar_is_childless_vector(vpar), Vec(l))));
}


\\-----------------------------------------------------------------------------
progress("vpar_join()");

check_equal(vpar_join([0,1,2],2, [0,1,0,3]), [0,1,2, 2,4,2,6]);
check_equal(vpar_join(Vecsmall([0,1,2]),2, Vecsmall([0,1,0,3])), \
            Vecsmall([0,1,2, 2,4,2,6]));

\\ FIXME: Maybe should validate the vertex number in the top half, to guard
\\ against making rubbish.
check_equal(vpar_join([],1,[0]), [1]);
check_equal(vpar_join([],99,[0]), [99]);

{
  for(n=1,5,
     for(m=1,5,
        check_equal(vpar_join(vpar_make_path(n),n,vpar_make_path(m)),
                    vpar_make_path(n+m),
                    "vpar_join() paths")));
}


\\-----------------------------------------------------------------------------
progress("test_is_triangular()");

{
  check_equal(vector(#A010054_samples,n,n--; test_is_triangular(n)),
              A010054_samples,   \\ OFFSET=0
              "A010054 test_is_triangular()");

  check_equal(vector(#A023532_samples,n, ! test_is_triangular(n)),
              A023532_samples,
              "A023532 not test_is_triangular(), apart from offset");
}

\\-----------------------------------------------------------------------------
progress("vpar_Laplacian_poly()");

check_equal(vpar_Laplacian_poly([]), Pol(1));
check_equal(vpar_Wiener_index([]),   0);

check_equal(vpar_Laplacian_poly([0]),    'x);
check_equal(vpar_Laplacian_poly([0],'y), 'y);
check_equal(vpar_Laplacian_poly([0,0]),  'x^2);
check_equal(vpar_Laplacian_poly([0,1,0,3]),  (x-2)^2 * x^2);

test_nearly_equal_epsilon = 1e-10;
test_nearly_equal(x,y, epsilon=test_nearly_equal_epsilon) =
{
  my(d=abs(x*1.0 - y*1.0));
  if(type(d)=="t_MAT", d=vecsum(Vec(d)));
  if(type(d)=="t_VEC" || type(d)=="t_COL", d=vecsum(d));
  abs(d) <= epsilon;
}
\\ x<=y means x-y negative, and "nearly" allow small positive
test_nearly_le(x,y, epsilon=test_nearly_equal_epsilon) = \
  x*1.0 - y*1.0 <= epsilon;

\\ if x majorized by y, meaning y bigger
test_vector_is_nearly_majorized(x,y) =
{
  #x==#y || error();
  for(k=1,#x-1,
     test_nearly_le(vecsum(x[1..k]), vecsum(y[1..k])) || return(0));
  test_nearly_equal(vecsum(x), vecsum(y));
}

test_vector_unique_nearly(vec, epsilon=test_nearly_equal_epsilon) = \
{
  my(ret=List([]));
  for(i=1,#vec,
     for(j=1,#ret,
        if(test_nearly_equal(vec[i],ret[j]), next(2)));
     listput(ret,vec[i]));
  Vec(ret);
}
check_equal(test_vector_unique_nearly([1, 2, 1+1e-20]), [1,2]);

test_vector_most_repeats_nearly(vec, epsilon=test_nearly_equal_epsilon) = \
{
  my(value=List([]),
     count=List([]));
  for(i=1,#vec,
     for(j=1,#value,
        if(test_nearly_equal(vec[i],value[j]),
           count[j]++; next(2)));
     listput(value,vec[i]);
     listput(count,1));
  if(#count, vecmax(Vec(count)), 0);
}
check_equal(test_vector_most_repeats_nearly([1, 2, 1, 2,2]), 3);

test_perms_any_order_ge3(perms) =
{
  for(i=1,#perms,
     my(id=test_identity_perm(#perms[i]));
     if(perms[i] != id        \\ order 1
        && perms[i]^2 != id,  \\ order 2
        return(1)));
  0;
}
check_equal(test_perms_any_order_ge3([ Vecsmall([1,2]), Vecsmall([2,1]) ]), 0);
check_equal(test_perms_any_order_ge3([ Vecsmall([2,3,1]) ]), 1);

\\ term x^n
\\ choose a set of n vertices as roots,
\\ choose an out-edge from each non-root,
\\ out-edges must be distinct
test_unsigned_Laplacian_poly_by_sets(vpar) =
{
  \\ flags[v] = 1 for v as root
  \\          = 0 for v as non-root (so want to choose an out-edge)
  \\
  my(neighbours=vpar_neighbours_vector(vpar),
     ret=0*'x);
  forvec(flags=vector(#vpar,v,[0,1]),
        my(p = x^hammingweight(flags),
           outs=select(f->!f, flags, 1),
           outs_neighbours=vecextract(neighbours,outs));
        forvec(toindices=vector(#outs,i,[1,#outs_neighbours[i]]),
              my(tovertices=vector(#toindices,i,
                                   outs_neighbours[i][toindices[i]]),
                 edges=vector(#tovertices,i, vecsort([outs[i],tovertices[i]])),
                 edges_distinct = (#edges==#Set(edges)));
              ret += edges_distinct * p));
  ret;
}

test_unsigned_Laplacian_poly_by_edge_deletes(vpar) =
{
  \\ flags[v] = 1 to delete edge v to parent
  \\          = 0 not delete (or no parent)
  my(ret=0*'x);
  forvec(flags=vector(#vpar,v, [0, vpar[v]!=0]),
        my(reduced_vpar=vpar);
        for(v=1,#reduced_vpar, if(flags[v], reduced_vpar[v]=0));
        ret += vpar_num_rootings(reduced_vpar)
               * 'x^vpar_num_roots(reduced_vpar));
  ret;
}

test_Laplacian_poly_by_ru(vpar,var='x) =
{
  \\ path-2  p   p   p
  \\         ^   |
  \\         |   v
  \\         v   v   v
  my(seq=vpar_INTERNAL_seq_upwards(vpar),
     root = vector(#vpar,v,var),  \\ root[v] = x
     non  = vector(#vpar),      \\ non[v] = 0
     ret = var/var);            \\ 1 as poly

  for(i=1,#seq,
     my(v=seq[i],
        p=vpar[v]);
     \\ print("v="v"  root "root[v]" non "non[v]);
     if(p,
        \\ print(" to p="p"  root "root[p]" non "non[p]);
        \\ print(" edge up   "root[p]*root[v] / var );
        \\ print(" no edge   "root[p]*(root[v]+non[v]) );
        \\ print(" edge down "root[p]*root[v] / var );
        \\ print(" edge down "root[p]*non[v] / var );
        \\ print(" edge up   "non[p]*root[v] / var );
        \\ print(" no edge   "non[p]*(root[v]+non[v]) );

        my(rootp = root[p]*root[v] / var       \\ edge up
                + root[p]*(root[v]+non[v]));   \\ no edge
        non[p]  = root[p]*root[v] / var       \\ edge down
                + root[p]*non[v] / var        \\ edge down
                + non[p]*root[v] / var        \\ edge up
                + non[p]*(root[v]+non[v]);    \\ no edge
        root[p]=rootp;
        \\ print(" set p  root "root[p]" non "non[p]);
        ,
        \\ print(" top "root[v]+non[v]);
        ret *= (root[v] + non[v]));  \\ at root

     root[v] =     \\ kill intermediate polys
     non[v]  = 0);

  ret;
}
check_equal(test_Laplacian_poly_by_ru([0]), x, "test_Laplacian_poly_by_ru()");
check_equal(test_Laplacian_poly_by_ru([0,0]), x^2);
check_equal(test_Laplacian_poly_by_ru([0,1]), x^2 + 2*x);
check_equal(vpar_Laplacian_poly([0,1,2]), x^3 - 4*x^2 + 3*x);
check_equal(test_Laplacian_poly_by_ru([0,1,2]), x^3 + 4*x^2 + 3*x);
check_equal(vpar_Laplacian_poly([0,1,1]), x^3 - 4*x^2 + 3*x);
check_equal(test_Laplacian_poly_by_ru([0,1,1]), x^3 + 4*x^2 + 3*x);
check_equal(vpar_Laplacian_poly([0,1,1,1]), x^4 - 6*x^3 + 9*x^2 - 4*x);
check_equal(test_Laplacian_poly_by_ru([0,1,1,1]), x^4 + 6*x^3 + 9*x^2 + 4*x, "star");

test_Laplacian_poly_by_ru_sub(vpar,var='x) =
{
  \\ root[], non[] are kept with 1 factor of x too small

  my(seq=vpar_INTERNAL_seq_upwards(vpar),
     root = vector(#vpar,v,1),  \\ root[v] = 1
     non  = vector(#vpar),      \\ non[v] = 0
     ret = var/var);

  for(i=1,#seq,
     my(v=seq[i],
        p=vpar[v]);
     \\ print("v="v"  root "root[v]" non "non[v]);
     if(p,
        my(rootp = root[p]*root[v]      \\ edge up
                + root[p]*(root[v]+non[v]) *var);   \\ no edge
        non[p]  = root[p]*root[v]       \\ edge down
                + root[p]*non[v]        \\ edge down
                + non[p]*root[v]        \\ edge up
                + non[p]*(root[v]+non[v]) *var;    \\ no edge
        root[p]=rootp;

        \\ print(" set p  root "root[p]" non "non[p]);
        ,
        \\ print(" top "root[v]+non[v]);
        ret *= (root[v]+non[v]));  \\ at root

     root[v] =     \\ kill intermediate polys
     non[v]  = 0);

  var^vpar_num_roots(vpar) * ret;
}

test_Laplacian_poly_by_ru_sub2(vpar,var='x) =
{
  \\ root[], non[] are kept with 1 factor of x too small

  my(seq=vpar_INTERNAL_seq_upwards(vpar),
     root = vector(#vpar,v,1),  \\ root[v] = 1
     non  = vector(#vpar),      \\ non[v] = 0
     ret = var/var);

  for(i=1,#seq,
     my(v=seq[i],
        p=vpar[v],
        a = root[v]+non[v]);
     \\ print("v="v"  root "root[v]" non "non[v]);
     if(p,
        my(r = root[p],
           c = root[v] + a*var);
        root[p] *= c;
        non[p]  = non[p]*c + r*a;

        \\ print(" set p  root "root[p]" non "non[p]);
        ,
        \\ print(" top "root[v]+non[v]);
        ret *= a);  \\ at root

     root[v] =     \\ kill intermediate polys
     non[v]  = 0);

  var^vpar_num_roots(vpar) * ret;
}

test_abspoly(poly) = Polrev(abs(Vecrev(poly)),variable(poly));
check_equal(test_abspoly(x^3), x^3);
check_equal(test_abspoly(-x^3), x^3);
check_equal(test_abspoly(x^3-1), x^3+1);
check_equal(test_abspoly(-x^3-1), x^3+1);

{
  for(i=1,#test_sample_trees_and_forests,
     my(vpar=test_sample_trees_and_forests[i],
        num_edges  = vpar_num_edges(vpar),
        m          = vpar_Laplacian_matrix(vpar),
        is_tree    = vpar_is_tree(vpar));

     my(poly = vpar_Laplacian_poly(vpar));
     check_equal(type(poly), "t_POL",
                 "vpar_Laplacian_poly() type");
     check_equal(poly, charpoly(m),
                 "vpar_Laplacian_poly() vs charpoly(vpar_Laplacian_matrix())");

     check_equal(test_abspoly(poly),
                 test_Laplacian_poly_by_ru(vpar),
                 "vpar_Laplacian_poly() vs test_Laplacian_poly_by_ru()");
     check_equal(test_abspoly(poly),
                 test_Laplacian_poly_by_ru_sub(vpar),
                 "vpar_Laplacian_poly() vs test_Laplacian_poly_by_ru_sub()");
     check_equal(test_abspoly(poly),
                 test_Laplacian_poly_by_ru_sub2(vpar),
                 "vpar_Laplacian_poly() vs test_Laplacian_poly_by_ru_sub2()");

     if(#vpar <= test_limit_flags_search,
        check_equal(test_abspoly(poly),
                    test_unsigned_Laplacian_poly_by_sets(vpar),
        "vpar_Laplacian_poly() vs test_unsigned_Laplacian_poly_by_sets()"));

     if(num_edges <= test_limit_flags_search,
        check_equal(test_abspoly(poly),
                    test_unsigned_Laplacian_poly_by_edge_deletes(vpar),
    "vpar_Laplacian_poly() vs test_unsigned_Laplacian_poly_by_edge_deletes()"));

     \\ Gutman and Zhou, "A Connection Between Ordinary and Laplacian
     \\ Spectra of Bipartite Graphs", Linear and Multilinear Algebra.
     \\ Matchpoly coeffs are the Laplacian coeffs without signs.
     \\
     \\ (GP 2.7.x subst can give back t_INT which polrecip() doesn't like,
     \\ so Pol() to force it.)
     my(matchpoly_of_subdivide=vpar_matchpoly(vpar_subdivide(vpar)));
     check_equal(poly,
                 'x^vpar_num_roots(vpar)
                 * polrecip(Pol(subst(matchpoly_of_subdivide,'x,-'x))),
       "vpar_Laplacian_poly() vs vpar_matchpoly() of vpar_subdivide()");

     check_equal(polcoeff(poly,#vpar), 1,
                 "vpar_Laplacian_poly() coeff #vpar = 1");

     check_equal(polcoeff(poly,#vpar-1), -2*vpar_num_edges(vpar),
                 "vpar_Laplacian_poly() coeff #vpar-1 = -2*num edges");

     check_equal(polcoeff(poly,0), if(#vpar,0,1),
                 "vpar_Laplacian_poly() coeff 0 = 0");

     my(low_nonzero_pos = valuation(poly,'x));
     check_equal(low_nonzero_pos, vpar_num_roots(vpar),
                 "vpar_Laplacian_poly() valuation = vpar_num_roots()");

     for(i=low_nonzero_pos,poldegree(poly),
        check_equal(sign(polcoeff(poly,i)),
                    if((poldegree(poly)-i)%2==0, 1, -1),
                    "vpar_Laplacian_poly() coeffs alternating signs"));

     my(component_sizes=vpar_component_sizes(vpar));
     check_equal(polcoeff(poly,low_nonzero_pos),
                 prod(i=1,#component_sizes, -component_sizes[i]) * (-1)^#vpar,
                 "vpar_Laplacian_poly() low term = product");
     check_equal(polcoeff(poly,valuation(poly,'x)),
                 prod(i=1,#component_sizes, component_sizes[i]) * (-1)^(#vpar-#component_sizes),
                 "vpar_Laplacian_poly() low term = product");
     check_equal(abs(polcoeff(poly,valuation(poly,'x))),
                 vpar_num_rootings(vpar),
                 "vpar_Laplacian_poly() coeff lowest!=0 = num rootings");

     my(component_vpar=vpar_component_trees(vpar));
     check_equal(poly,
                 Pol(prod(i=1,#component_vpar, vpar_Laplacian_poly(component_vpar[i]))),
                 "vpar_Laplacian_poly() product of components");

     check_equal(test_poly_num_real_roots(poly), poldegree(poly),
                 "vpar_Laplacian_poly() all roots real");

     my(roots=polroots(poly));
     for(i=1,#roots,
        check_equal(abs(imag(roots[i])) < 1e-10, 1,
                    "vpar_Laplacian_poly() all roots real");
        roots[i]=real(roots[i]));
     roots=vecsort(roots,,4);  \\ biggest first
     if(#vpar >= 1,
        \\ always valuation >=1 forest components
        check_equal(roots[#vpar], 0.0,
                    "vpar_Laplacian_poly() smallest root == 0"));
     if(#vpar >= 2,
        \\ second smallest root == 0 means valuation >= 2 forest components
        check_equal(roots[#vpar-1] != 0.0, is_tree,
           "vpar_Laplacian_poly() second smallest root vs vpar_is_tree()"));

     \\------
     if(is_tree,
        check_equal(polcoeff(poly,2),
                    (-1)^#vpar * vpar_Wiener_index(vpar),
          "vpar_Laplacian_poly() signed second term is vpar_Wiener_index()");
        check_equal(abs(polcoeff(poly,2)),
                    vpar_Wiener_index(vpar),
          "vpar_Laplacian_poly() abs second term is vpar_Wiener_index()"));

     \\------
     \\ Properties summarized in
     \\ R.B. Bapat, "The Laplacian Matrix of a Graph", The Mathematics
     \\ Student, voume 65, 1996, pages 214-223.

     my(degrees=vpar_degrees_vector(vpar));
     degrees=vecsort(degrees,,4);  \\ biggest first
     if(#vpar >= 1,
        \\ variational principle for largest eigenvalue
        check_equal(roots[1] >= degrees[1], 1,
                  "vpar_Laplacian_poly() biggest root >= biggest degree");

        \\ Gershgorin Discs Theorem
        check_equal(roots[1] <= 2*degrees[1], 1,
                  "vpar_Laplacian_poly() biggest root <= 2* biggest degree"));

     if(#vpar >= 2 && is_tree,
        \\ Merris, given in Bapat.  For connected graph >= 2, maybe ...
        check_equal(roots[1] >= degrees[1]+1, 1,
                  "vpar_Laplacian_poly() biggest root >= biggest degree"));

     \\ print(vpar"  "degrees" "roots);

     \\ Schur, real symmetric matrix diagonal elements majorized by eigenvalues
     check_equal(test_vector_is_nearly_majorized(degrees,roots), 1,
                 "vpar_Laplacian_poly() degrees majorized by roots");

     \\ R. Grone, "Eigenvalues and the Degree Sequences of Graphs", Linear
     \\ and Multilinear Algebra, volume 39, 1995, pages 133-136.
     \\ For a connected graph.
     if(#vpar>=1 && is_tree,
        my(vec=degrees);
        vec[1]++;      \\ biggest +1
        vec[#vec]--;   \\ smallest -1
        check_equal(test_vector_is_nearly_majorized(vec,roots), 1,
                 "vpar_Laplacian_poly() degrees +/- majorized by roots"));

     if(is_tree,
        my(distinct_roots=test_vector_unique_nearly(roots),
           gens=vpar_free_automorphism_generators(vpar),
           any_3cycle = test_perms_any_order_ge3(gens));
        \\ print(#distinct_roots" "#roots" "any_3cycle"  "vpar" gen orders "apply(test_permorder,gens));
        if(any_3cycle,
           check_equal(#distinct_roots<#roots, 1,
             "vpar_Laplacian_poly() automorphism cycle>=3 means repeat root"));
        if(#distinct_roots == #roots,
           check_equal(test_perms_is_involutions(gens), 1,
             "vpar_Laplacian_poly() all distinct roots means automorphisms all order 2"));
        );

     if(#vpar >= 2 && is_tree,
        \\ print(test_vector_most_repeats_nearly(roots)" "vpar_num_leaves(vpar)"  "vpar);
        check_equal(test_vector_most_repeats_nearly(roots)
                    <= vpar_num_leaves(vpar)-1,  1,
          "vpar_Laplacian_poly() most repeat roots vs vpar_num_leaves()"));

     \\ tree of n vertices
     if(is_tree,
        for(i=1,#roots,
           my(int=round(roots[i]));
           if(int!=0 && test_nearly_equal(int,roots[i]),
              \\ print(int" "#vpar"  "vpar);
              check_equal(#vpar % int, 0,
                "vpar_Laplacian_poly() integer eigenvalue divides #vpar"))));

     \\ Fielder algebraic connectivity of a graph = second smallest eigenvalue.
     \\ Range for a tree 2(1-cos pi/n) <= a <= 1.
     \\ M. Fielder, "Eigenvectors of Acyclic Matrices", Czechoslovak Math J,
     \\ volume 23, 1973, pages 298-305.
     if(#vpar>=3 && is_tree,
        my(algebraic_connectivity=roots[#roots-1]);
        \\ print(algebraic_connectivity" "2*(1-cos(Pi/#vpar))"  "vpar"  roots="roots);
        check_equal(test_nearly_le(2*(1-cos(Pi/#vpar)), algebraic_connectivity),
                    1,
                    "vpar_Laplacian_poly() algebraic connectivity min");
        check_equal(algebraic_connectivity <= 1,  1,
                    "vpar_Laplacian_poly() algebraic connectivity <= 1"));

     \\-------------------

     \\ Ivan Gutman and Ljiljana Pavlovic, "On the Coefficients of the
     \\ Laplacian Characteristic Polynomial of Trees", Bulletin Academie
     \\ Serbe Des Sciences Et Des Arts. Classe Des Sciences Mathematiques Et
     \\ Naturelles. Sciences Mathematiques, number 28, 2003, pages 31-40.
     \\ https://www.jstor.org/stable/44095650

     if(#vpar>=1 && is_tree,
        my(n=#vpar);

        \\ Gutman and Pavlovic equations 6,7, referring to C. S. Oliveira,
        \\ N. M. M. De Abreu, S. Jurkewicz, "The Characteristic Polynomial of
        \\ the Laplacian of Graphs in (a,b)-Linear Classes", Lin. Algebra
        \\ Appl, volume 356, 2002, pages 113-121.
        \\
        check_equal(polcoeff(poly,n-2),
                    2*n^2 - 5*n + 3 - 1/2*sum(i=1,n, degrees[i]^2),
                    "vpar_Laplacian_poly() term n-2");

        check_equal(polcoeff(poly,n-3),
                    -1/3*(4*n^3 - 18*n^2 + 24*n - 10
                          + sum(i=1,n, degrees[i]^3)
                          - 3*(n-2)*sum(i=1,n, degrees[i]^2)),
                    "vpar_Laplacian_poly() term n-3");

        \\ Gutman and Pavlovic, terms 2, 3, n-2 and n-3 terms are min star
        \\ max path, and those extremes attained only by star and path.
        \\
        \\ Term 2 is Wiener index, so that star and path are unique extremes
        \\ of Wiener index.
        \\
        if(n>=5 && !test_is_free_star(vpar) && !test_is_free_path(vpar),
           my(star_poly=vpar_Laplacian_poly(vpar_make_star(n)));
           my(path_poly=vpar_Laplacian_poly(vpar_make_path(n)));
           \\ print("star "star_poly);
           \\ print("poly "poly);
           \\ print("path "path_poly);

           check_equal(abs(polcoeff(star_poly, 2)) < abs(polcoeff(poly,2)), 1);
           check_equal(abs(polcoeff(poly, 2)) < abs(polcoeff(path_poly,2)), 1);
           check_equal(abs(polcoeff(star_poly,2)), (n-1)^2,
                       "star Wiener index");
           check_equal(abs(polcoeff(path_poly,2)), binomial(n+1,3),
                       "path Wiener index");

           check_equal(abs(polcoeff(star_poly, 3)) < abs(polcoeff(poly,3)), 1);
           check_equal(abs(polcoeff(poly, 3)) < abs(polcoeff(path_poly,3)), 1);

           check_equal(polcoeff(star_poly, n-2) < polcoeff(poly,n-2), 1);
           check_equal(polcoeff(poly, n-2) < polcoeff(path_poly,n-2), 1);

           check_equal(abs(polcoeff(star_poly, n-3)) < abs(polcoeff(poly,n-3)),
                       1,
                       "vpar_Laplacian_poly() term n-3 vs star");
           check_equal(abs(polcoeff(poly, n-3)) < abs(polcoeff(path_poly,n-3)),
                       1,
                       "vpar_Laplacian_poly() term n-3 vs path");
           );
     );
  );
}

\\-----------------------------------------------------------------------------
progress("vpar_Laplacian_poly() of Path");

\\ Algebraic connectivity of a graph is the second smallest eigenvalue of
\\ its Laplacian matrix.
\\ All Laplacian eigenvalues are real since the matrix is symmetric, but
\\ polroots() gives complex with float 0.0 imaginary parts.
test_algebraic_connectivity_inexact(vpar) = \
  vecsort(real(polroots(vpar_Laplacian_poly(vpar))))[2];

{
  \\ path-n
  for(n=3,10,
     my(vpar=vpar_make_path(n),
       poly=vpar_Laplacian_poly(vpar));

     check_equal(polcoeff(poly,0), 0);
     check_equal(polcoeff(poly,1), -(-1)^n * n);

     \\ algebraic connectivity min over trees = 2*(1-cos(Pi/n))
     check_equal(test_nearly_equal(test_algebraic_connectivity_inexact(vpar),
                                   2*(1-cos(Pi/n))),
                 1));
}
\\ for(n=3,5, print(vpar_Laplacian_poly(vpar_make_path(n))))

{
  \\ A053122 coeffs of Laplacian poly of path-n, x^1 up to x^n
  \\ low to high coeffs, omit low 0 constant coeff
  \\ 1,
  \\ -2, 1,
  \\ 3, -4, 1,
  \\ -4, 10, -6, 1,
  \\ 5, -20, 21, -8, 1,

  check_equal(vpar_Laplacian_poly(vpar_make_path(1)), x);
  check_equal(vpar_Laplacian_poly(vpar_make_path(2)), x^2 -2*x);
  check_equal(vpar_Laplacian_poly(vpar_make_path(3)), x^3 -4*x^2 +3*x);

  my(l=List([]));
  for(n=1,#A053122_samples,
     if(#l>#A053122_samples, break());
     my(poly=vpar_Laplacian_poly(vpar_make_path(n)));
     for(i=1,n, listput(l,polcoeff(poly,i))));
  l=Vec(l,#A053122_samples);
  check_equal(l, A053122_samples,
             "A053122 Laplacian poly coeffs of path-n");
}
{
  \\ A053123 Laplacian poly path coeffs, x^n down to x^1
  my(got=[]);
  for(n=1,#A053123_samples,
     my(vec=Vec(vpar_Laplacian_poly(vpar_make_path(n))));
     got=concat(got,vec[^#vec]);
     if(#got >= #A053123_samples, break));
  got=Vec(got,#A053123_samples);
  check_equal(got,A053123_samples,
              "A053123 coefficients of Laplacian poly of path");
}
{
  \\ A285072 coeffs of Laplacian poly of path-n, x^1 up to x^n
  \\ low to high coeffs, low coeff negative, high coeff alternating signs
  \\ 0,
  \\ -1,
  \\ 0, -2, 1,
  \\ 0, -3, 4, -1,
  \\ 0, -4, 10, -6, 1,
  \\ 0, -5, 20, -21, 8, -1,

  check_equal(vpar_Laplacian_poly(vpar_make_path(1)), x);
  check_equal(vpar_Laplacian_poly(vpar_make_path(2)), -2*x +x^2);
  check_equal(vpar_Laplacian_poly(vpar_make_path(3)), 3*x -4*x^2 + x^3);

  my(l=List([]));
  for(n=1,#A285072_samples,
     if(#l>#A285072_samples, break());
     my(poly=vpar_Laplacian_poly(vpar_make_path(n))
             * (-1)^n);
     for(i=0,poldegree(poly), listput(l,polcoeff(poly,i))));
  l=Vec(l,#A285072_samples);
  check_equal(l, A285072_samples,
             "A285072 Laplacian poly coeffs of path-n");
}

{
  \\ path-n evaluates
  check_equal(vector(30,n,n--; subst(vpar_Laplacian_poly(vpar_make_path(n)),
                                     'x,1)),
              vector(30,n,n--; if(n==0,1, ((n+1)%3)-1)),  \\ -1,0,1 repeating
              "vpar_Laplacian_poly() path-n evaluate x=1");
  check_equal(vector(30,n,n--; subst(vpar_Laplacian_poly(vpar_make_path(n)),
                                     'x,-1)),
              vector(30,n,n--; if(n==0,1, (-1)^n * fibonacci(2*n))),
              "vpar_Laplacian_poly() path-n evaluate x=-1");
}


\\-----------------------------------------------------------------------------
progress("vpar_Laplacian_poly() of Star");

{
  \\ star-n
  for(n=3,10,
     my(vpar=vpar_make_star(n),
        poly=vpar_Laplacian_poly(vpar));
     check_equal(polcoeff(poly,0), 0);
     check_equal(polcoeff(poly,1), -(-1)^n * n);

     \\ algebraic connectivity max over trees = 1
     check_equal(test_nearly_equal(test_algebraic_connectivity_inexact(vpar),
                                   1),
                 1));
}
{
  \\ star-n evaluates
  check_equal(vector(30,n,n--; subst(vpar_Laplacian_poly(vpar_make_star(n)),'x,1)),
              vector(30,n,n--; if(n<=1,1, n==2,-1, 0)),
              "vpar_Laplacian_poly() star-n subst 1");
  check_equal(vector(30,n,n--; subst(vpar_Laplacian_poly(vpar_make_star(n)),'x,-1)),
              vector(30,n,n--; if(n==0,1, (n+1)*(-2)^(n-2))),
              "vpar_Laplacian_poly() star-n subst -1");
}

{
  \\ A103406 Laplacian poly star coeffs, absolute values x^n down to x^1
  my(got=[]);
  for(n=1,#A103406_samples,
     my(vec=Vec(vpar_Laplacian_poly(vpar_make_star(n))));
     got=concat(got, abs(vec[^#vec]));
     if(#got >= #A103406_samples, break));
  got=Vec(got,#A103406_samples);
  check_equal(got, A103406_samples,
              "A103406 coefficients of Laplacian poly of star");
}
{
  \\ A093375 Laplacian poly star coeffs, absolute values x^1 up to x^n
  my(got=[]);
  for(n=1,#A093375_samples,
     my(vec=Vecrev(vpar_Laplacian_poly(vpar_make_star(n))));
     got=concat(got, abs(vec[^1]));
     if(#got >= #A093375_samples, break));
  got=Vec(got,#A093375_samples);
  check_equal(got, A093375_samples,
              "A093375 coefficients of Laplacian poly of star");
}


\\-----------------------------------------------------------------------------
progress("test_A005651() - sum multinomial over partitions");
\\ A005651
\\ n=0  1  2   3
\\   1, 1, 3, 10, 47, 246, 1602, 11481

{
  check_equal(vector(#A005651_samples,n,n--; test_A005651(n)),
              A005651_samples,  \\ OFFSET=0
              "test_A005651() vs A005651_samples");
}

\\ S recurrence per Vladimir Kruchinin in A005651.
\\ Sge(n,m) = sum over those partitions with all terms >= m.
\\ Would cache or vector of rows since n,m are variously re-used, but without
\\ is enough for testing.
test_A005651_Sge(n,m) = \
  if(n==m,1, n<m,0, 1 + sum(k=m,n\2, binomial(n,k)*test_A005651_Sge(n-k,k)));
{
  check_equal(vector(10,n,n--; if(n==0,1, test_A005651_Sge(n,1))),
              vector(10,n,n--; test_A005651(n)));
}
{
  for(n=1,8,
     for(m=1,n,
        my(total=0);
        forpart(p=n, total += test_multinomial(p), [m,n]); \\ terms >=m
        check_equal(total, test_A005651_Sge(n,m),
                    "test_A005651_Sge() vs forpart() multinomials")));
}
\\ for(n=1,5, for(m=1,n, printf(" %8d",test_A005651_Sge(n,m)));print());
\\ vpar_subtree_heights_vectors_count(5,1)
\\ my(n=5,m=1); for(k=m,n\2, print("k="k"  "test_A005651_Sge(n-k,k)" * "binomial(n,k)));


\\ Sle(n,m) = sum over those partitions with all terms <= m
\\ One partition of single term n, provided m>=n.
\\ Otherwise k is the biggest term, up to n-1 or m.
\\ Binomial choose its elements, and partition terms <=k in the rest.
\\ Would cache or vector of rows since n,m are variously re-used, but without
\\ is enough for testing.
test_A005651_Sle(n,m) = \
{
  n>=0 || error();
  m>=1 || (m==0 && n==0) || error();
  (m>=n) + sum(k=1,min(n-1,m), binomial(n,k)*test_A005651_Sle(n-k,k));
}
{
  for(n=1,8,
     for(m=1,n,
        my(total=0);
        forpart(p=n, total += test_multinomial(p), m);  \\ terms <=m
        check_equal(total, test_A005651_Sle(n,m))));
}
check_equal(vector(12,n,n--; test_A005651_Sle(n,n)), \
            vector(12,n,n--; test_A005651(n)));


\\ test_A005651_vector0_by_steps(n) returns a vector of A005651 terms 0 to n
\\ inclusive.
\\ The loop starts a[n+1] = sum multinomials of partitions with all terms <k.
\\ The loop adds terms k to such partitions, making bigger ones.
\\ r is the starting partition size.
\\ e is a multiple of k which are added terms.
\\ The binomial advances the existing r size to t=r+e.
\\ Or equivalently it is choice of which vertices, among the t, are going to
\\ be in the new partition.
\\ a_new[] is with the k terms added, which replaces a[].
\\ Both a[] and a_new[] begin with the n=0 term, hence indexes a[n+1] etc.
\\ max_k default is to take k up to all partitions.  Smaller can stop and
\\ count just those of limited partition terms.
\\ For efficiency might like to cache binomial(t,k) in the k loop, since t
\\ arises more than one way.  Or rearrange the loop maybe.
\\ k=1 is a(n) = n! which could be set explicitly rather than running the loop.
test_A005651_vector0_by_steps(n,max_k=n) =
{
  \\ m limit
  \\   t = r+m*k <= n
  \\   m*k <= n-r
  \\   m <= floor (n-r)/k
  \\
  \\ r+k   binomial(r+k,k)
  \\ r+2k  binomial(r+2k,k) * binomial(r+k,k)
  \\ r+3k  binomial(r+3k,k) * binomial(r+2k,k) * binomial(r+k,k)

  my(a=vector(n+1,i,0));
  a[1]=1;
  for(k=1,max_k,             \\ new term k
     my(new_a=a);
     for(r=0,#a,             \\ remainder, so t=r+extra
        my(b=1);
        forstep(e=k,n-r,k,   \\ e multiple of k terms
           my(t=r+e);        \\ total new size
           new_a[t+1] += a[r+1] * (b *= binomial(t,k))));
     a=new_a);
  a;
}
check_equal(test_A005651_vector0_by_steps(12), \
            vector(13,n,n--; test_A005651(n)));


\\ A005651 mathematica code by Jean-Francois Alcover
\\ term j >= i, and exactly t terms altogether
\\ builds sum as fractions with denominators as partitions, then final n!
\\ numerator for a(n)
\\ Would cache results since n,i,t are variously re-used, but without is
\\ enough for testing.
\\
test_A005651_b(n,i,t) = if(t==1, 1/n!, sum(j=i,n\t, test_A005651_b(n-j, j, t-1) /j!));
test_A005651_by_b(n) = if(n==0, 1, n!*test_A005651_b(n, 0, n));
check_equal(vector(10,n,n--; test_A005651_by_b(n)), \
            vector(10,n,n--; test_A005651(n)));
{
  for(n=1,7,
     for(i=1,n,
        for(t=1,n,
           my(total=0);
           \\ each term >=i, number of terms ==t
           forpart(p=n, total += test_multinomial(p), [i,n], [t,t]);
           check_equal(total, n!*test_A005651_b(n,i,t)))));
}

\\ A005651 formula by Vladeta Jovovic
\\ Like exponential transform of A182926, but done with b here as fractions.
\\ Multiplying b by (k-1)! gives the integers A182926.
\\
\\ test_A005651_sd_b(k) = sumdiv(k,d, d / d!^(k/d));
\\ test_A005651_sd_b(k) = sumdiv(k,d, k /(k/d) / d!^(k/d));
\\ test_A005651_sd_b(k) = k * sumdiv(k,d, 1 / (k/d * d!^(k/d)));
test_A005651_sd_b(k) = sumdiv(k,d, d * d!^(-k/d));
test_A005651_by_sd(n) =
{
  if(n==0, 1,
     sum(k=1,n, (n-1)!/(n-k)!*test_A005651_sd_b(k)*test_A005651_by_sd(n-k)));
}
check_equal(vector(16,n,n--; test_A005651_by_sd(n)), \
            vector(16,n,n--; test_A005651(n)));
check_equal(vector(#A182926_samples,k,test_A005651_sd_b(k)*(k-1)!), \
            A182926_samples);
\\
\\ vector(10,k,test_A005651_sd_b(k)*k!)   \\ extra factor k
\\ not in OEIS: 1, 4, 9, 40, 125, 966, 5047, 45648, 367929, 3856870
\\
\\ vector(10,k,numerator(test_A005651_sd_b(k)))
\\ vector(10,k,denominator(test_A005651_sd_b(k)))
\\ not in OEIS: 1, 2, 3, 5, 25, 161, 721, 317, 13627, 385687
\\ not in OEIS: 1, 1, 2, 3, 24, 120, 720, 280, 13440, 362880

test_A182926(n) = my(f=n!); sumdiv(n,d, f/(d*((n/d)!)^d));
{
  \\ A182926 1, 2, 3, 10, 25, 161, 721, 5706, 40881, 385687
  check_equal(vector(#A182926_samples,n, test_A182926(n)),
              A182926_samples,  \\ OFFSET=1
              "test_A182926()");
  check_equal(vector(#A182926_samples,n,
                     sumdiv(n,d, n! / d!^(n/d) / (n/d))),
              A182926_samples,  \\ OFFSET=1
              "A182926");
  \\ 1..n into ordered list of equal size sets, with fixed first set.
  \\ When 1 set, the fixing has no effect.
  \\ When >1 set ...

  \\ n=1 setsize=1  1 way              1
  \\ n=2 setsize=1  2 ways  *1/2 = 1 \ 2
  \\     setsize=2  1 way   *2/2 = 1 /
  \\ n=3 setsize=1  6 ways  *1/3 = 2 \ 3
  \\     setsize=3  1 way   *3/3 = 1 /
  \\ n=4 setsize=1 24 ways  *1/4 = 6 \        fixed first set
  \\     setsize=2  6 ways  *2/4 = 3 | 10
  \\     setsize=4  1 way   *4/4 = 1 /
}
{
  \\ A061095 partition 1..n into ordered list of equal size sets
  check_equal(vector(#A061095_samples,n,
                     sumdiv(n,d, n! / d!^(n/d))),
              A061095_samples,  \\ OFFSET=1
              "A061095");

  my(limit=min(20,#A061095_samples),
     L = O(x^limit),
     g = L + sum(k=1,limit, x^k/k! / (1 - x^k/k! + L)));
  check_equal(test_egf_terms(g,limit),
              concat(0,A061095_samples[1..limit-1]),   \\ starts at n=1
              "A061095 egf");
  check_equal(test_egf_terms(g',limit),
              A061095_samples[1..limit-1],   \\ starts at n=1
              "A061095 egf'");

  \\ n=1 setsize=1  1 way    1
  \\ n=2 setsize=1  2 ways \ 3
  \\     setsize=2  1 way  /
  \\ n=3 setsize=1  6 ways \ 7
  \\     setsize=3  1 way  /
  \\ vpar_INTERNAL_exponential_transform([1, 3, 7, 31, 121, 831, 5041, 42911, 364561])
  \\ not in OEIS: 1, 4, 17, 105, 722, 6163, 58011, 627490, 7433997
  \\ exp n=2  [1],[1]  1 way      because exp is with duplication
  \\          [1,1]    1 way
  \\          [2]      2 ways
  \\ exp n=3  [1],[1],[1]  1           \ cf only 6 in A005651           1
  \\          [1],[1,1]    3*2 = 6     |      [1,1] discounted to 1     3
  \\          [1,1,1]      6           /      [1,1,1] discounted to 2   2
  \\          [1],[2]      3
  \\          [3]          1
  \\ GP-Test           1+6+6+3+1 == 17
  \\ c,b,a                a,b,c
  \\ bc,a b,ac c,ab       a,cb b,ca ba,c
  \\ abc, acb             cab, cba
  \\ in A182926 take first set as the smallest (smallest element)
  \\ in exp take blocks in descending order of their first element
  \\ or other way around

  \\ v=vpar_INTERNAL_exponential_transform([1, 3, 7, 31, 121, 831, 5041, 42911, 364561])
}

\\ exp(A182926) = A005651 per Peter Luschny in A182926 (written in binomials)
check_equal(vpar_INTERNAL_exponential_transform(vector(20,n,test_A182926(n))),\
            vector(20,n, test_A005651(n)), \
            "A005651 = exponential transform of A182926");

{
  \\ A005651 egf from Simon Plouffe (is it?) there.
  \\ 1/(1-x^k/k!) = 1 + x^k/k! + x^(2k)/k!^2 + x^(3k)/k!^3 + ...
  \\ Product over k is term x^n all sums various k, with repetitions
  \\ allowed, which equal n (so partitions of n), and with corresponding k!
  \\ denominators, thus giving the denominators of the multinomials in
  \\ A005651.
  \\
  my(limit=min(20,#A005651_samples),
     L = O(x^limit),
     g = L + 1 / prod(k=1,limit, 1 - x^k/k! + L));
  check_equal(test_egf_terms(g,limit),
              A005651_samples[1..limit],   \\ starting at n=0
              "A005651 egf");

  \\ A182926 starts n=1, so initial 0 for exponential untransform
  check_equal(test_egf_terms(log(g),limit),
              concat(0, A182926_samples[1..limit-1]),
              "A182926 egf log of A005651");
}
{
  my(limit=20,
     L = O(x^limit));
  check_equal(test_egf_terms(-log(1 - x^1/1! + L), limit),
              vector(limit,n,n--; if(n==0,0, (n-1)!)));  \\ factorials

  check_equal(test_egf_terms(-log(1 - x^2/2! + L), limit),
              vector(limit,n,n--; if(n>=2 && n%2==0, (n-1)!/2^((n-2)/2),  0)));
  \\ A007019
  \\ vector(10,n,n--; (2*n+1)!/2^n)

  check_equal(test_egf_terms(-log(1 - x^3/3! + L), limit),
              vector(limit,n,n--; if(n>=3 && n%3==0,  n! / (n/3) / 3!^(n/3),  0)));
  \\ vector(12,n,n--; if(n%3==0,(3*n/3)!/((n/3)!*(3!)^(n/3))))
  \\ v=[1, 10, 560, 92400, 33633600, 22870848000]
  \\ for(i=1,#v,v[i]/=(i-1)!); v

  \\ A025035 number partitions of 1..3n into sets of size 3 each
  check_equal(vector(#A025035_samples,n,n--; (3*n)!/(n!*(3!)^n)),
              A025035_samples);  \\ OFFSET=0
}
{
  for(k=1,5,
     for(q=1,8\k,
        my(n=q*k);
        if(n! / k!^q > 10000, next);
        \\ print("n="n" k="k" q="q);
        my(s=vector(n,i, (i+k-1)\k),
           l=List([]));
        \\ print(s)
        for(i=1,n!,
           my(perm=numtoperm(n,i),
              elem_to_setnum=vecextract(s,perm),
              sets=vector(q,i, Set(select(j->elem_to_setnum[j]==i,[1..n]))),
              pos=setsearch(l,sets,1));
           if(pos, listinsert(l,sets,pos)));
        check_equal(#l, n! / k!^q);        \\ ordered list of sets
        l = Set(apply(v->Set(v),l));
        check_equal(#l, n! / q! / k!^q);   \\ set of sets
        ));
}


\\-----------------------------------------------------------------------------
progress("vpar_relabel_reverse()");

check_equal(vpar_relabel_reverse([]),[], \
            "vpar_relabel_reverse() empty");
check_equal(vpar_relabel_reverse([0,0,0]),[0,0,0], \
            "vpar_relabel_reverse() singletons");
check_equal(vpar_relabel_reverse(vpar_make_star(5)),[5,5,5,5,0]);

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     check_equal(vpar_relabel_reverse(vpar),
                 vpar_relabel_perm(vpar, test_perm_reversal(#vpar))));
}

{
  \\ reverse of preorder is postorder
  for(n=0,6,
     my(vpar=vpar_preorder_first(n),
        total=0);
     until((vpar=vpar_preorder_next(vpar,1))=='none,  \\ forests
           my(vparrev=vpar_relabel_reverse(vpar));
           check_equal(vpar_is_postorder(vparrev), 1)));
}


\\-----------------------------------------------------------------------------
progress("test_partition_to_repetitions()");

check_equal(test_partition_to_repetitions([1,3,3,3,4,4,4,4,4,4]), \
            [[1,1],[3,3],[4,6]]);


\\-----------------------------------------------------------------------------
progress("vpar_relabel_seq()");

check_equal(vpar_relabel_seq([],[]), [], \
            "vpar_relabel_seq() empty")
check_equal(vpar_relabel_seq([0,0,0],[3,2,1]), [0,0,0], \
            "vpar_relabel_seq() singletons")

{
  my(vpar=[0,1,2,3,4, 5,6,7,8],
     seq=[1,2,3,4,5, 9,8,7,6]);
  check_equal(vpar_relabel_seq(vpar,seq),
              [0,1,2,3,4, 7,8,9,5])
}

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        seq=vpar_INTERNAL_random_perm(#vpar),
        got_vpar=vpar_relabel_seq(vpar,seq));
     check_equal(got_vpar,
                 vpar_relabel_perm(vpar,Vecsmall(seq)^-1),
                 "vpar_relabel_seq() vs inverse vpar_relabel_perm()"));
}


\\-----------------------------------------------------------------------------
progress("test_setintersect_sets()");

check_equal(test_setintersect_sets([ [1,2,3], [2,3,4], [3,4,5] ]), [3]);
check_equal(test_setintersect_sets([ [1,2], [2,3,4], [3,4,5] ]), []);


\\-----------------------------------------------------------------------------
progress("test_poly_low_coefficient()");

check_equal(test_poly_low_coefficient(0), 0);
check_equal(test_poly_low_coefficient(3*x^5 + 2*x), 2);


\\-----------------------------------------------------------------------------
progress("test_compnum_to_smallest()");

check_equal(test_compnum_to_smallest([]), [], \
            "test_compnum_to_smallest() empty");
check_equal(test_compnum_to_smallest([1,2,2]), [1,2]);
check_equal(test_compnum_to_smallest([2,1,2]), [2,1]);
check_equal(test_compnum_to_smallest([2,2,2,1,3]), [4,1,5]);


\\-----------------------------------------------------------------------------
progress("test_flags_is_one_first()");

check_equal(test_flags_is_one_first([]), 1, \
            "test_flags_is_one_first() empty");
check_equal(test_flags_is_one_first([0]), 1);
check_equal(test_flags_is_one_first([0,0,0]), 1);
check_equal(test_flags_is_one_first([0,0,0,1]), 1);
check_equal(test_flags_is_one_first([0,0,0,2]), 0);
check_equal(test_flags_is_one_first([1]), 1);
check_equal(test_flags_is_one_first([2,1,1,1]), 0);
check_equal(test_flags_is_one_first([1,2,1,1]), 1);
check_equal(test_flags_is_one_first([0,2,1,1,1]), 0);
check_equal(test_flags_is_one_first([0,1,2,1,1]), 1);


\\-----------------------------------------------------------------------------
progress("vpar_has_perfect_matching()");

check_equal(vpar_has_perfect_matching([]), 1, \
            "vpar_has_perfect_matching() empty tree");
check_equal(vpar_has_perfect_matching([],1), 0, \
            "vpar_has_perfect_matching() empty tree, near");

check_equal(vpar_has_perfect_matching([0]), 0, \
            "vpar_has_perfect_matching() single vertex");
check_equal(vpar_has_perfect_matching([0],1), 1);

check_equal(vpar_has_perfect_matching([0,0]), 0, \
            "vpar_has_perfect_matching() two single vertices");
check_equal(vpar_has_perfect_matching([0,0],1), 0);

check_equal(vpar_has_perfect_matching([0,1]), 1, \
            "vpar_has_perfect_matching() path-2");
check_equal(vpar_has_perfect_matching([0,1],1), 0);

check_equal(vpar_has_perfect_matching([0,1,2]), 0, \
            "vpar_has_perfect_matching() path-3");
check_equal(vpar_has_perfect_matching([0,1,2],1), 1);

\\  1    2     4
\\  |   / \
\\  3  5   6
\\
check_equal(vpar_has_perfect_matching([0,0,1,0,2,2]), 0);
check_equal(vpar_has_perfect_matching([0,0,1,0,2,2],1), 0);

check_equal(vpar_has_perfect_matching([2,3,4,0]), 1);

\\        1 --> 2 --> 3 --> 8
\\  4 --> 5 --> 6 --> 7 --/
check_equal(vpar_has_perfect_matching([2,3,8, 5,6,7,8, 0]), 1);

check_equal(vpar_has_perfect_matching([0,1,2,3]), 1);

check_equal(vpar_has_perfect_matching([0,1,1,1]), 0);

\\  1 <- 2 <- 4
\\    \- 3
check_equal(vpar_has_perfect_matching([0,1,1,2]), 1);

\\    2
\\   /|\
\\  1 3 6
\\  | |
\\  4 5
check_equal(vpar_has_perfect_matching([2,0,2,1,3,2]), 1);

\\     5
\\     |
\\     3
\\     |
\\     1
\\    / \
\\   2   6
\\   |
\\   4
check_equal(vpar_has_perfect_matching([3,1,5,2,0,1]), 1);

\\     4
\\     | \  \
\\     3  5  6
\\     |
\\     2
\\     |
\\     1
check_equal(vpar_has_perfect_matching([2,3,4,0,4,4]), 0);
check_equal(vpar_has_perfect_matching([2,3,4,0,4,4],1), 0);


{
  \\       5
  \\      / \
  \\     1   6
  \\    / \
  \\   2   7
  \\       |
  \\       4
  \\       |
  \\       3
  my(vpar=[5, 1, 4, 7, 0, 5, 1]);
  vpar_validate(vpar);
  check_equal(#vpar, 7);
  my(matchnum=vpar_matchnum(vpar));
  check_equal(matchnum, 3);
  check_equal(vpar_has_perfect_matching(vpar), 0);
  check_equal(vpar_has_perfect_matching(vpar,1), 1);
}

\\ Original code from perfect matching without the near perfect option.
\\
test_has_perfect_matching_only(vpar) =
{
  \\ A perfect matching is made by taking a leaf and its attachment,
  \\ reckoning them removed from the tree, and repeating.
  \\
  \\ seen[v] >= 0 is number of children at v.
  \\ seen[v] < 0 means already matched.
  \\
  \\ When a pair v,p is taken, reckoning them removed from the tree reduces
  \\ the children of vpar[p].  If that becomes childless then can continue
  \\ upwards.  If vpar[p] > i then it would be reached later in the
  \\ outermost loop, but those < i must be continued up.
  \\
  \\ A childless unseen v without a parent, or with a parent which is
  \\ already matched, cannot be paired so no perfect matching.  Any tree
  \\ with no perfect matching eventually reaches such a v, so the final
  \\ return is 1 with no checking of the seen[] array (it will be all <0 ).
  \\
  \\ Using a seq to work upwards is also possible, and slightly less code,
  \\ but combining the num_children and the flags for when matched means
  \\ just one workspace vector.

  \\ print("vpar_has_perfect_matching "vpar);
  if(bitand(#vpar,1), 0, \\ odd size
     my(seen=vpar_INTERNAL_num_children_vecsmall(vpar));
     \\ print(" num children "Vec(seen));
     for(i=1,#vpar,
        my(v=i, p);
        \\ print(" v="i);
        while(!seen[v],
              ((p=vpar[v]) && seen[p]>0) || return(0);
              \\ print("  pair v="v" p="p);
              seen[v] = -1;
              seen[p] = -1;
              (v=vpar[p]) || break();
              seen[v]--));
     1);
}

\\ Second code with perfect and near perfect together.
\\
test_has_perfect_matching_with_both(vpar,near=0) =
{
  \\ print("test_has_perfect_matching_with_both() near="near" "vpar);

  \\ At a leaf, can pair child+parent if both unmatched.
  \\ Continue doing so at grandparent and above too.
  \\ If a child has no parent or the parent is already matched then that is
  \\ the child left unmatched.
  \\ For perfect matching this is immediately fatal.
  \\ For near perfect 2 or more unmatched is fatal.
  \\ Continuing up to grandparent etc is constrained v<=i so that an
  \\ unmatched v > i is not counted a second time when the i loop reaches
  \\ that v.
  \\ seen[v] = num children, decremented as children are processed
  \\      or = -1 when in a matched pair

  if(bitand(#vpar,1) != near, return(0));

  my(seen=vpar_INTERNAL_num_children_vecsmall(vpar),
     count=0);
  \\ print(" num children "Vec(seen));
  for(i=1,#vpar,
     \\ print(" i="i);
     my(v=i);
     while(v<=i && !seen[v],
           \\ print("  v="v);
           my(p=vpar[v]);
           if(p && seen[p]>=0,
              \\ print("  pair v="v" p="p);
              seen[v] = -1;
              seen[p] = -1;
              (v=vpar[p]) || break();  \\ v = grandparent
              seen[v]--;  \\ p processed, decrease children of grandparent
              ,
              \\ print("  v no match (at root or p matched), count="count);
              if(count++>near, return(0));
              break())));
  \\ print(" final count="count);
  count==near;
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);

     check_equal(vpar_has_perfect_matching(vpar),
                 test_has_perfect_matching_only(vpar));

     my(matchnum=vpar_matchnum(vpar));
     for(near=0,1,
        check_equal(vpar_has_perfect_matching(vpar,near),
                    matchnum == (#vpar-near)/2,
                    "vpar_has_perfect_matching() vs matchnum");
        check_equal(vpar_has_perfect_matching(vpar,near),
                    test_has_perfect_matching_with_both(vpar,near),
                    "vpar_has_perfect_matching() vs test_has_perfect_matching_with_both()");
     );
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_count_perfect_matching()");

check_equal(vpar_count_perfect_matching(0), 1);
check_equal(vpar_count_perfect_matching(0,,1), 1);
check_equal(vpar_count_perfect_matching(0,,-1), 1);

{
  \\ nautyextra misc.c perfect trees
  \\      n = 1  2      4        6          8
  my(samples=[0, 2, 0, 48, 0, 4320, 0, 860160, 0, 302400000, 0]);
  check_equal(vector(#samples,n, vpar_count_perfect_matching(n)), samples);
}
{
  \\ nautyextra misc.c near perfect trees
  \\      n = 1     3       5          7            9
  my(samples=[1, 0, 9, 0, 600, 0, 102900, 0, 33067440, 0]);
  check_equal(vector(#samples,n, vpar_count_perfect_matching(n,1)), samples);
}
{
  \\ nautyextra misc.c perfect forests
  my(samples=[0, 2, 0, 60, 0, 5880, 0, 1224720, 0, 442743840, 0]);
  check_equal(vector(#samples,n, vpar_count_perfect_matching(n,,1)), samples);
}
{
  \\ nautyextra misc.c near perfect forests
  \\      n = 1      3        5          7            9
  my(samples=[1, 0, 15, 0, 1080, 0, 188160, 0, 60480000, 0]);
  check_equal(vector(#samples,n, vpar_count_perfect_matching(n,1,1)), samples);
}

{
  \\ perfect
  \\ component tree containing v=1, size t, choose t-1 others,
  \\ perfect trees t and perfect forests n-t remaining
  check_equal(vector(20,n, vpar_count_perfect_matching(n,,1)),
              vector(20,n, sum(t=1,n,
                               binomial(n-1,t-1)
                               * vpar_count_perfect_matching(t)
                               * vpar_count_perfect_matching(n-t,,1))),
              "vpar_count_perfect_matching() forest from trees recurrence");
}
{
  \\ A036363 line labelled 2-trees = 1/2 * perfect matching forests
  check_equal(vector(#A036363_samples,n,n*=2; 1/2*vpar_count_perfect_matching(n,,1)),
              A036363_samples,  \\ OFFSET=3 in fact
              "A036363 = 1/2 * vpar_count_perfect_matching() perfect forests even size");
}
{
  \\ near perfect
  \\ component tree containing v=1, size t, choose t-1 others,
  \\ trees t as perfect if t even or near if t odd
  \\ forests n-t remaining, opposite
  check_equal(vector(20,n, vpar_count_perfect_matching(n,1,1)),
              vector(20,n, sum(t=1,n,
                               binomial(n-1,t-1)
                               * vpar_count_perfect_matching(t,t%2)
                               * vpar_count_perfect_matching(n-t,1-(t%2),1))),
              "vpar_count_perfect_matching() near forest from trees recurrence");
}

{
  for(near=0,1,
     for(forest=-1,1,
        for(n=0, test_limit_trees_search - forest,
           my(count=0);
           forvec(vec=vpar_forvec_from(n,forest), \\ forests
                 my(vpar=vpar_from_upascend(vec));
                 count += vpar_has_perfect_matching(vpar,near));

           check_equal(count,
                       vpar_count_perfect_matching(n,near,forest),
                       Str("vpar_count_perfect_matching()"
                           " n="n" near="near" forest="forest));
  )));
}


\\-----------------------------------------------------------------------------
progress("vpar_is_free_automorphism()");

{
  \\  1  2  4                 5     6     7
  \\     |       mangled      |    / \    |
  \\     3                    1   2   3   4
  my(vpar=[0, 0, 2, 0],
     perm=Vecsmall([4, 2,3, 1]));
  check_equal(vpar_relabel_perm(vpar,perm), vpar);
  check_equal(vpar_is_automorphism(vpar,perm), 1);
  check_equal(vpar_is_free_automorphism(vpar,perm), 1,
              "vpar_is_free_automorphism() 3-component");
  check_equal(vpar_INTERNAL_for_free_automorphisms(vpar), [5, 6, 6, 7, 0, 0, 0]);
}
{
  \\   1        2
  \\   ^   =>   ^
  \\   |        |
  \\   2        1

  my(vpar=[0,1],
     perm=Vecsmall([2,1]));
  vpar_validate(vpar);
  check_equal(test_is_perm(perm), 1);
  check_equal(vpar_relabel_perm(vpar,perm), [2,0]);
  check_equal(vpar_is_automorphism(vpar,perm), 0,
              "vpar_is_automorphism() path-2 reverse");

  my(v=1, p=vpar[v]);
  check_equal(p, 0);
  check_equal(if(p,perm[p],0),       0);
  check_equal(if(p,vpar[perm[p]],0), 0);
  check_equal(perm[v], 2);

  check_equal(vpar_is_free_automorphism(vpar,perm), 1,
              "vpar_is_free_automorphism() path-2 flip");
}
{
  \\    2      4
  \\    |  =>  |
  \\    1      3
  \\
  my(vpar=[2,0, 4,0],
     perm=Vecsmall([3,4, 1,2]));
  vpar_validate(vpar);
  check_equal(test_is_perm(perm), 1);
  check_equal(vpar_relabel_perm(vpar,perm), [2,0, 4,0]);
  check_equal(vpar_is_automorphism(vpar,perm), 1);
  check_equal(vpar_is_free_automorphism(vpar,perm), 1,
              "vpar_is_free_automorphism() path-2s");

  perm=Vecsmall([4,3, 2,1]);
  check_equal(test_is_perm(perm), 1);
  check_equal(vpar_relabel_perm(vpar,perm), [0,1, 0,3]);
  check_equal(vpar_is_automorphism(vpar,perm), 0);
  check_equal(vpar_is_free_automorphism(vpar,perm), 1,
              "vpar_is_free_automorphism() path-2s invert");
}
{
  \\    1          3
  \\    |    =>    |
  \\    2          2
  \\   / \        / \
  \\  3   4      4   1
  \\
  my(vpar=[0,1,2,2],
     perm=Vecsmall([3, 2, 4,1]));

  check_equal(vpar_relabel_perm(vpar,perm), [2, 3, 0, 2]);
  check_equal(vpar_reroot([2, 3, 0, 2],1), vpar);
  check_equal(vpar_roots(vpar), [1]);
  check_equal(vecextract(perm,vpar_roots(vpar)), Vecsmall([3]));

  check_equal(vpar_is_free_automorphism([0,1,2,2], [3, 2, 4,1]), 1,
              "vpar_is_free_automorphism() claw");
}

\\ like vpar_neighbours_vector(), but Set()s of neighbours
test_neighbours_sets(vpar) = \
  apply(vec->Set(vec), vpar_neighbours_vector(vpar));
check_equal(test_neighbours_sets([2,3,0]), [ [2], [1,3], [2] ]);
check_equal(vpar_neighbours_vector([2,3,0]), [ [2], [3,1], [2] ]);

\\ equal after relabelling, up to rerooting
\\
\\ vpar_relabel_perm(vpar,perm)
\\ == vpar_reroot_forest(vpar, vecextract(perm,vpar_roots(vpar)));
\\
test_is_free_automorphism_by_rerooting(vpar,perm) = \
  test_equal_after_rerooting(vpar, vpar_relabel_perm(vpar,perm));

\\ neighbouring vertices all unchanged
\\
test_is_free_automorphism_by_neighbours(vpar,perm) =
{
  test_neighbours_sets(vpar)
  == test_neighbours_sets(vpar_relabel_perm(vpar,perm));
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        perm=vpar_INTERNAL_random_perm(#vpar),
        got=vpar_is_free_automorphism(vpar,perm));
     \\ print(vpar"  perm="perm"  got="got);

     \\ cf devel/test-slow.gp for adapting the rooted vpar_is_automorphism()
     \\ to test free automorphism
     check_equal(got, test_is_free_automorphism_by_rerooting(vpar,perm),
       "vpar_is_free_automorphism() vs test_is_free_automorphism_by_rerooting()");
     check_equal(got, test_is_free_automorphism_by_neighbours(vpar,perm),
       "vpar_is_free_automorphism() vs test_is_free_automorphism_by_neighbours()");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_num_claws_total()");

check_equal(vpar_num_claws_total(0), 0, "vpar_num_claws_total() n=0 trees");
check_equal(vpar_num_claws_total(0,1), 0, \
            "vpar_num_claws_total() n=0 forests");
check_equal(vpar_num_claws_total(0,-1), 0, \
            "vpar_num_claws_total() n=0 root=1");

\\ n=4 root=1 is   1        1
\\                /|\       |
\\               2 3 4      2      with 3 choices for the middle so total 4
\\                         / \
\\                        3   4
check_equal(vpar_num_claws_total(4,-1), 4, \
            "vpar_num_claws_total() n=4 root1");

check_equal(vpar_num_claws_total(4,1), 16, \
            "vpar_num_claws_total() n=4 forest");

{
  for(forest=-1,1,
     for(n=1, test_limit_trees_search - forest,
        my(total=0);
        forvec(vec=vpar_forvec_from(n,forest), \\ trees or forests
              my(vpar=vpar_from_upascend(vec));
              total += vpar_num_claws(vpar));
        check_equal(total, vpar_num_claws_total(n,forest),
                    Str("vpar_num_claws_total() n="n" forest="forest))));
}

{
  \\ A053508 = binomial(n-1,3)*n^(n-4)), for n>=1
  check_equal(vector(#A053508_samples,n, vpar_num_claws_total(n,-1)/4),
              A053508_samples);  \\ OFFSET=1
}

{
  \\ A053509 = binomial(n,4)*n^(n-5)), for n>=1
  \\ here vpar_num_claws_total forests total starting n=0
  \\ egf -1/5!*LambertW(-x)^5
  check_equal(vector(#A053509_samples,n,n--; vpar_num_claws_total(n,1)/16),
              A053509_samples);  \\ OFFSET=1 instead
}

{
  my(limit = 11,
     L = O(x^limit),
     gB = test_func_to_egf(n->vpar_num_claws_total(n,-1), limit),
     gT = test_func_to_egf(vpar_num_claws_total, limit),
     gF = test_func_to_egf(n->if(n,[0, 0, 0, 16, 480, 11760, 286720, 7348320, 201600000][n]), 10));

  check_equal(test_egf_terms(gT,limit),
              vector(limit,n,n--; vpar_num_claws_total(n)));
  check_equal(gT, x*gB' + L);  \\ factor n
  check_equal(gT', (x*gB')' + L, "vpar_num_claws_total() gT'");
  check_equal(gT', x'*gB' + x*gB'' + L);
  check_equal(gT', gB' + x*gB'' + L);
  check_equal(gB, intformal(gT/x));  \\ divide n

  check_equal(gB', gT/x, "vpar_num_claws_total() gB'");

  my(gNumT = test_func_to_egf(vpar_count, limit),
     gNumF = test_func_to_egf(n->vpar_count(n,1), limit),
     gNumB = test_func_to_egf(n->vpar_count(n,-1), limit));
  check_equal(4*(gNumT-1)' + O(x^((limit+1)\2)),
              gF/gB,
              "vpar_num_claws_total() gF/gB");

  check_equal(gF, 4*gB*gNumT' + O(x^10));

  my(F = n->if(n,sum(t=1,n, 4* binomial(n,t)
                             * vpar_num_claws_total(t,-1) * (n-t+1)^(n-t))));
  check_equal(gF, test_func_to_egf(F,10), "vpar_num_claws_total() gF");

  my(F = n->if(n, 4*vpar_num_claws_total(n,-1)
                + sum(t=1,n-1, 4*binomial(n-1,t-1)*n/t^2*vpar_num_claws_total(t) * (n-t+1)^(n-t))));
  check_equal(gF, test_func_to_egf(F,10));


  check_equal(vector(limit,n,n--; vpar_num_claws_total(n,1)),
              vector(limit,n,n--;
                vpar_num_claws_total(n)
                + sum(t=1,n-1, binomial(n,t) * t/n * (
                          vpar_num_claws_total(t) * vpar_count(n-t,1)
                        + vpar_num_claws_total(n-t,1) * vpar_count(t)))));

  \\   claws in the component containing v=1, times num forests of n-t rest
  \\ + claws in the n-t rest, times num trees t the first component
  check_equal(vector(limit,n,n--; vpar_num_claws_total(n,1)),
              vector(limit,n,n--;
                vpar_num_claws_total(n)
                + sum(t=1,n-1, binomial(n-1,t-1) * (
                          vpar_num_claws_total(t) * vpar_count(n-t,1)
                        + vpar_num_claws_total(n-t,1) * vpar_count(t)))));

  check_equal(matrix(6,6,n,t, binomial(n-1,t-1)),
              matrix(6,6,n,t, binomial(n,t) * t/n));

  \\ n!/(t-1)!/(n-t+1)!
  \\ (n-1)!/(t-1)!/(n-t)!  * n / (n-t+1)
  check_equal(vector(6,n,vector(n,t, binomial(n-1,t-1))),
              vector(6,n,vector(n,t, binomial(n,t-1) * (n-t+1) / n)));
  check_equal(matrix(6,6,n,t, binomial(n-1,t-1)),
              matrix(6,6,n,t, binomial(n,t-1) * (n-t+1) / n));

  check_equal(vector(limit,n,n--; vpar_num_claws_total(n,1)),
              vector(limit,n,n--;
                my(m=n-1);
                vpar_num_claws_total(n)
                + sum(k=0,m-1,
                      my(t=k+1);
                               binomial(m,k) * (
                          vpar_num_claws_total(k+1) * vpar_count(m-k,1)
                        + vpar_num_claws_total(m-k,1) * vpar_count(k+1)))));

  check_equal(vector(limit,n,n--; vpar_num_claws_total(n,1)),
              vector(limit,n,n--;
                my(m=n-1);
                - binomial(m,m) * (
                          vpar_num_claws_total(m+1) * vpar_count(m-m,1)
                        + vpar_num_claws_total(m-m,1) * vpar_count(m+1))
                +
                vpar_num_claws_total(n)
                + sum(k=0,m,
                               binomial(m,k) * (
                          vpar_num_claws_total(k+1) * vpar_count(m-k,1)
                        + vpar_num_claws_total(m-k,1) * vpar_count(k+1)))));

  check_equal(vector(limit,n,n--; vpar_num_claws_total(n,1)),
              vector(limit,n,n--;
                my(m=n-1);
                sum(k=0,m,
                               binomial(m,k) * (
                          vpar_num_claws_total(k+1) * vpar_count(m-k,1)
                        + vpar_num_claws_total(m-k,1) * vpar_count(k+1)))));

  check_equal(gF, intformal( gT' * gNumF
                           + gF  * gNumT'  ) + L/x,
              "vpar_num_claws_total() gF intformal" );

  check_equal(gF' - gF*gNumT',      gT'*gNumF        + L/x^2);
  check_equal(gF' - gF*(x*gNumF)',  gT'*gNumF        + L/x^2);
  check_equal(gF' - gF*gNumT',      gT'*(gNumT-1)/x  + L/x^2);

  check_equal(gF' - gF*gNumT',        (x*gB')'*(gNumT-1)/x       + L/x^2,  "gF' - gF*gNumT'");
  check_equal(gF' - gF*gNumT',        (gB' + x*gB'')*(gNumT-1)/x + L/x^2);
  check_equal(gF' - gF*gNumT',        (gB'/x + gB'')*(gNumT-1)   + L/x^2);

  check_equal(gF,  4*gB*gNumT' + L/x,
              "vpar_num_claws_total() gF = 4*gB*gNumT'");  \\ target
  check_equal((4*gB*gNumT')'               + L,
              4*gB'*gNumT' + 4*gB*gNumT''  + L,
              "vpar_num_claws_total() expr");


  \\ first order ODE
  \\ y' + P(x)*y = Q(x)
  \\ y = 1/u(x) integral Q(x)*u(x)  where u(x) = exp(integral P(x))

  check_equal(gF' - gF*gNumT',   gT'*gNumF + L/x^2);
  my(P = - gNumT',
     Q = gT'*gNumF,
     u = exp(intformal(P)));
  check_equal(gF' + P*gF, Q + L/x^2, "PQ ODE");
  check_equal(gF, 1/u*intformal(Q*u) + L/x, "PQ ODE gF");

  check_equal(u, exp(intformal(- gNumT')));
  check_equal(u, exp(-gNumT+1));
  check_equal(u, 1/gNumF);
  check_equal(gF, gNumF * intformal( Q/gNumF ) + L/x,
              "gF intformal");
  check_equal(gF, gNumF * gT + L/x);
}


\\-----------------------------------------------------------------------------
progress("test_make_vecsearch_func()");

{
  my(vec=[c,a,x],
     func=test_make_vecsearch_func(vec));
  check_equal(func(c), 1, "test_make_vecsearch_func() 1");
  check_equal(func(a), 2);
  check_equal(func(x), 3);
  check_equal(func(z), 0, "test_make_vecsearch_func() not-found");
  check_equal(func(123), 0);
}


\\-----------------------------------------------------------------------------
progress("test_untriangular_ceil()");

\\ 0 1 2 3 4 5 6 7 8 9 10
\\ ^ ^   ^     ^        ^
\\ check_equal(test_untriangular_ceil(0), 0);
check_equal(test_triangular(0), 0, "test_triangular() n=0");
check_equal(test_triangular(1), 1);
check_equal(test_triangular(2), 3);
check_equal(test_triangular(3), 6);
check_equal(test_untriangular_ceil(0), 0, "test_untriangular_ceil() n=0");
check_equal(test_untriangular_ceil(1), 1);
check_equal(test_untriangular_ceil(2), 2);
check_equal(test_untriangular_ceil(3), 2);
check_equal(test_untriangular_ceil(4), 3);
check_equal(test_untriangular_ceil(6), 3);
check_equal(test_untriangular_ceil(7), 4);
{
  for(n=1,100,
     my(t=test_untriangular_ceil(n));
     check_equal(test_triangular(t-1) < n, 1);
     check_equal(test_triangular(t) >= n, 1);
  );
  check_equal(vector(#A123578_samples,n,n--; test_untriangular_ceil(n)),
              A123578_samples,
              "A123578 n appears n times, and 0 = test_untriangular_ceil() up to offset");
}


\\-----------------------------------------------------------------------------
progress("vpar_num_roots()");
\\ vpar_num_edges()

check_equal(vpar_num_roots([]),0, "vpar_num_roots() empty");
check_equal(vpar_num_edges([]),0, "vpar_num_edges() empty");
check_equal(vpar_num_roots([0,0,2,3,4,5]),2);
check_equal(vpar_num_edges([0,0,2,3,4,5]),4);

{
  \\ addhelp() of vpar_num_roots(), total roots among all forests n vertices
  my(T = (n)-> 2*n*(n+1)^(n-2),
     forest=1);
  for(n=0, test_limit_trees_search - forest,
     my(total=0);
     forvec(vec=vpar_forvec_from(n,forest), \\ forests
           my(vpar=vpar_from_upascend(vec));
           total += vpar_num_roots(vpar));
     check_equal(total, T(n)));
}

\\ total number of roots in ordered forests where the last vertex is depth d
test_num_roots_total_Fd(n,d) =
{
  if(n>=0 && d>=0 && d<=n-1,
     vpar_INTERNAL_Catalan_triangle(n,n-1-d),
     0);
}

check_equal(test_num_roots_total_Fd(0,0), 0);
check_equal(test_num_roots_total_Fd(2,-1), 0, "Fd(2,-1)");
check_equal(test_num_roots_total_Fd(2,0), 2, "Fd(2,0)");
check_equal(test_num_roots_total_Fd(2,1), 1, "Fd(2,1)");
check_equal(test_num_roots_total_Fd(2,2), 0, "Fd(2,2)");
\\
check_equal(test_num_roots_total_Fd(3,-1), 0, "Fd(3,-1)");
check_equal(test_num_roots_total_Fd(3,0), 3+2, "Fd(3,0)");
check_equal(test_num_roots_total_Fd(3,1), 2+1, "Fd(3,1)");
check_equal(test_num_roots_total_Fd(3,2), 1, "Fd(3,2)");
check_equal(test_num_roots_total_Fd(3,3), 0, "Fd(3,3)");
{
  \\ Recurrence:
  \\ When d=0 count each of the Catalan number forests of n-1.
  \\ When d!=0, last depth of n-1 can be any d-1 up to n-1.
  for(n=1,8,
     for(d=0,n,
        check_equal(test_num_roots_total_Fd(n,d),
                    if(d==0, vpar_INTERNAL_Catalan_number(n),
                       sum(i=max(0,d-1),n-1, test_num_roots_total_Fd(n-1,i))),
                    "Fd by recurrence");
     ));
}
\\ for(n=1,8,for(k=1,n,print1(vpar_INTERNAL_Catalan_triangle(n,k),",")); print());

{
  \\ ordered forests total roots
  my(F = (n)-> if(n==0,0, 3*(2*n)! / ((n+2)!*(n-1)!)));
  check_equal(vector(#A000245_samples,n,n--; F(n)),
              A000245_samples,  \\ OFFSET=0
              "A000245 samples");
  \\ Corresponds to the Dyck path downsteps going to zero level, as noted by
  \\ Cheyne Homberger in A000245.

  check_equal(vector(20,n,n++; F(n)),
              vector(20,n,n++; binomial(2*n,n) *3*n/(n+2)/(n+1)));
  check_equal(vector(20,n,n++; F(n)),
              vector(20,n,n++; binomial(2*n,n+1) *3/(n+2)));

  \\ Wolfdieter Lang in A000245, from the binomial
  check_equal(vector(20,n,n++; F(n)),
             vector(20,n,n++; vpar_INTERNAL_Catalan_number(n) *3*n/(n+2)));

  \\ adapting David Callan in A002057
  check_equal(vector(20,n,n++; F(n)),
              vector(20,n,n++; vpar_INTERNAL_Catalan_number(n+1) - vpar_INTERNAL_Catalan_number(n)));

  check_equal(vector(15,n,n++; F(n)),
              vector(15,n,n++; sum(d=0,n-1, test_num_roots_total_Fd(n,d))),
              "vpar_num_roots() total over ordered forests, F vs sum Fd");

  for(n=0,10,
     my(total=0,
        by_d=vector(n),
        vpar=vpar_preorder_first(n));
     until((vpar=vpar_preorder_next(vpar,1))=='none, \\ forests
           my(num_roots=vpar_num_roots(vpar));
           total += num_roots;
           if(n, by_d[vpar_depth_of_vertex(vpar,n)+1] += num_roots));
     check_equal(total, F(n),
                 "vpar_num_roots() total over ordered forests");
     check_equal(by_d, vector(n,d,d--; test_num_roots_total_Fd(n,d)),
                 "vpar_num_roots() total over ordered forests, by depth of last vertex");
     \\ print(by_d"   "vector(n,d,d--; test_num_roots_total_Fd(n,d)));
  );

  \\ gf in terms of Catalan numbers per A000245
  my(limit=20,
     L = O(x^limit),
     g = test_func_to_gf(F,limit),
     gCatalan = test_func_to_gf(vpar_INTERNAL_Catalan_number,limit));
  check_equal(g + L/x, (-1 + (1-x)*gCatalan)/x);
  check_equal(g, x*(1 + x*gCatalan^2)*gCatalan^2 + L);

  \\-----------
  \\ total right arm edges in binary trees of n vertices,
  \\ being forests total num roots less the first root in each forest
  my(E = (n)->F(n)-vpar_INTERNAL_Catalan_number(n));
  check_equal(vector(#A002057_samples,n,n++; E(n)),
              A002057_samples,
              "A002057 samples");
  check_equal(vector(#A002057_samples,n,n--; 4*binomial(2*n+3,n)/(n+4)),
              A002057_samples,
              "A002057 samples");

  check_equal(vector(20,n, E(n)),
              vector(20,n, binomial(2*n-1,n-2) * 4/(n+2)));
  check_equal(vector(20,n,n++; E(n)),
              vector(20,n,n++; vpar_INTERNAL_Catalan_number(n) * 2*(n-1)/(n+2)));
  check_equal(2*(n-1)/(n+2) + 1,  3*n/(n+2));

  \\ David Callan in A002057
  check_equal(vector(20,n,n++; E(n)),
              vector(20,n,n++; vpar_INTERNAL_Catalan_number(n+1) - 2*vpar_INTERNAL_Catalan_number(n)));

  \\ k vertices right of root
  \\ E(k) total right arm there
  \\ root right edge is to each of Cat(k) many right subtrees
  \\ then multiply Cat(n-1-k) different left subtrees
  \\
  check_equal(vector(20,n, E(n)),
              vector(20,n, sum(k=1,n-1,
                               (E(k) + vpar_INTERNAL_Catalan_number(k))   \\ right arm edges
                               * vpar_INTERNAL_Catalan_number(n-1-k))));  \\ left subtrees
}

{
  \\ Ordered forests count by num roots is row of Catalan triangle.
  for(n=1,8,
     my(count_by_num_roots=vector(n),
        vpar=vpar_preorder_first(n));
     until((vpar=vpar_preorder_next(vpar,1))=='none, \\ forests
           count_by_num_roots[vpar_num_roots(vpar)]++);
     check_equal(count_by_num_roots,
                 vector(n,k, vpar_INTERNAL_Catalan_triangle(n-1,n-k)));
     \\ print();
     \\ print(count_by_num_roots);
     \\ print(vector(n,k, vpar_INTERNAL_Catalan_triangle(n-1,n-k)));
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_siblingpos_vectors_count()");

check_equal(vpar_siblingpos_vectors_count(0), 1, \
            "vpar_siblingpos_vectors_count() n=0 trees");
check_equal(vpar_siblingpos_vectors_count(0,1), 1, \
            "vpar_siblingpos_vectors_count() n=0 forests");

{
  check_equal(vpar_siblingpos_vectors_count(5,1), 26);
  check_equal(iferr(vpar_siblingpos_vectors_count(5,'forest);0, e, 1),
              1,
  "vpar_siblingpos_vectors_count() error on forest=symbol (whereas it should be 0,1)");
}
{
  \\ trees
  my(want=[1,  1, 1, 3, 7, 20, 56, 182, 589]);
  check_equal(vector(#want,n,n--; vpar_siblingpos_vectors_count(n)),
              want,
              "vpar_siblingpos_vectors_count() trees samples");
}
{
  \\ forests
  my(want=[1,  1, 2, 4, 10, 26, 76, 232, 764]);
  check_equal(vector(#want,n,n--; vpar_siblingpos_vectors_count(n,1)),
              want,
              "vpar_siblingpos_vectors_count() forests samples");
}
{
  \\ trees root=1
  my(want=[1,  1, 1, 2, 4, 10, 26, 76, 232, 764]);
  check_equal(vector(#want,n,n--; vpar_siblingpos_vectors_count(n,-1)),
              want,
              "vpar_siblingpos_vectors_count() root=1 samples");
}

\\ root=1 trees are forests 1 smaller
check_equal(vector(20,n, vpar_siblingpos_vectors_count(n,-1)), \
            vector(20,n, vpar_siblingpos_vectors_count(n-1,1)));

{
  \\ forests
  \\ 1, 1, 2, 4, 10, 26, 76, 232, 764, 2620
  my(limit = 16,
     L = O('x^limit),
     gT = test_func_to_egf(vpar_siblingpos_vectors_count, limit),
     gF = test_func_to_egf(n->vpar_siblingpos_vectors_count(n,1), limit),
     gB = test_func_to_egf(n->vpar_siblingpos_vectors_count(n,-1), limit));

  check_equal(gF, exp('x + 'x^2/2 + L),
              "vpar_siblingpos_vectors_count() vs exp");
  check_equal(gF + L/x,  gB',
              "vpar_siblingpos_vectors_count() trees vs root=1");
}

{
  \\ forests, and root=1 trees
  \\ A000085 involutions = ballot sequences
  check_equal(vector(#A000085_samples,n,n--; vpar_siblingpos_vectors_count(n,1)),
              A000085_samples);  \\ OFFSET=0
  check_equal(vector(#A000085_samples,n, vpar_siblingpos_vectors_count(n,-1)),
              A000085_samples);
}

\\ forests=involutions
\\ per Muir, "A Treatise on the Theory of Determinants" as shown at
\\ mathworld "Permutation Involution"
\\ http://mathworld.wolfram.com/PermutationInvolution.html
\\
test_involutions_count_by_binomials(n) = \
  1 + sum(k=0,floor((n-1)/2), 1/((k+1)!) * prod(i=0,k, binomial(n-2*i,2)));
\\
check_equal(vector(20,n, test_involutions_count_by_binomials(n)), \
            vector(20,n, vpar_siblingpos_vectors_count(n,1)));

{
  \\ trees
  \\ A238124 ballots with last candidate 1 vote = ballots with first outright winner
  \\ has 0 at n=0, whereas here 1 empty siblingpos vector
  check_equal(vector(#A238124_samples,n,n--;
                     if(n==0,0, vpar_siblingpos_vectors_count(n))),
              A238124_samples,  \\ OFFSET=0
              "A238124 vpar_siblingpos_vectors_count() trees");
}

\\ vpar_INTERNAL_exponential_transform(vector(10,n, if(n<=2,1,0)))
\\ vector(10,n, vpar_siblingpos_vectors_count(n))
\\ test_egf_terms(exp(x+x^2/2+L),limit)
\\ vector(10,n, vpar_siblingpos_vectors_count(n,1) - vpar_siblingpos_vectors_count(n))


\\-----------------------------------------------------------------------------
progress("vpar_random_vertex()");

check_equal(vpar_random_vertex([]), 0)
check_equal(vpar_random_vertex([0]), 1)


\\-----------------------------------------------------------------------------
progress("test_perm_commute()");

check_equal(test_perm_commute(Vecsmall([1,2,3]), Vecsmall([2,3,1])), 1);
check_equal(test_perm_commute(Vecsmall([4,1,2,3]), Vecsmall([2,3,4,1])), 1);


\\-----------------------------------------------------------------------------
progress("test_perm_elem_order()");

check_equal(test_perm_elem_order(Vecsmall([2,1]),1), 2);
check_equal(test_perm_elem_order(Vecsmall([2,3,1]),2), 3);


\\-----------------------------------------------------------------------------
progress("test_perm_elem_orders_vector()");

check_equal(test_perm_elem_orders_vector(Vecsmall([2,1,4,5,3])), [2,2,3,3,3]);


\\-----------------------------------------------------------------------------
progress("Ordered Even Trees");

\\ Return 1 if all vertices have an even number of children.
test_is_num_children_even(vpar) =
{
  my(num_children=vpar_num_children_vector(vpar));
  for(v=1,#vpar, if(num_children[v]%2, return(0)));
  1;
}

\\ i=2, n=5 vertices
\\    1            1          1
\\  // \\         / \        / \
\\ 2 3 4 5       2   5      2   3
\\              / \            / \
\\             3   4          4   5

test_A001764(n) = binomial(3*n,n)/(2*n+1);
check_equal(vector(#A001764_samples,n,n--; test_A001764(n)), \
            A001764_samples);  \\ OFFSET=0

{
  \\ A001764 = number of ordered even trees
  for(i=0,5,
     my(n=2*i+1,
        count=0,
        total_depths=0,
        vpar=vpar_preorder_first(n));
     until((vpar=vpar_preorder_next(vpar))=='none,
           if(test_is_num_children_even(vpar),
              \\ if(i==2,vpar_view(vpar));
              \\ 1,1,2,3,7,
              \\ if(i<7 && vpar_is_ordered_isomorphic(vpar,vpar_relabel_reverse(vpar)),
              \\    print("i="i" "vpar));

              count++;
              total_depths += vecsum(vpar_depths_vector(vpar))));

     \\ print1(total_depths",");
     if(i>=1,
        \\ A062236 total depths in non-crossing trees
        check_equal(total_depths, 2*A062236_samples[i])); \\ apparently

     check_equal(count,test_A001764(i)));
}


\\-----------------------------------------------------------------------------
progress("vpar_depths_vector()");

check_equal(vpar_depths_vector([]), [], "vpar_depths_vector() empty");
check_equal(vpar_depths_vector([0]), [0], "vpar_depths_vector() singleton");
check_equal(vpar_depths_vector([0,0]), [0,0], "vpar_depths_vector() 2 singletons");
check_equal(vpar_depths_vector([2,0]), [1,0], "vpar_depths_vector() path-2 reverse");
check_equal(vpar_depths_vector([0,1]), [0,1], "vpar_depths_vector() path-2");

\\   3 self-loop      2
\\   |
\\   1
check_equal(vpar_depths_vector([3,0,3]), [1,0,0]);

\\     1 -> 2
\\    /  <-
\\    3
\\    |
\\    4
check_equal(vpar_depths_vector([2,1,1,3]), [1,1,2,3]);

\\ previous code, cycles not allowed
test_depths_vector_by_acyclic(vpar) = \
{
  \\ At each v=i search upwards for a depths[v] already known, or the root.
  \\ depths[v] = -1 when not yet known, or >=0 when known.
  \\ A re-traversal upwards fills in the new depths.

  my(depths=vector(#vpar,v,-1));
  for(i=1,#vpar,
     if(depths[i]<0,
        my(v=i, d=1);
        while(v=vpar[v], d++; if(depths[v]>=0, d+=depths[v]; break));
        v=i;
        until(!(v=vpar[v]) || depths[v]>=0, depths[v] = d--));
     );
  depths;
}

test_depths_vector_by_seq(vpar) =      \\ acyclic
{
  my(seq=vpar_INTERNAL_seq_upwards(vpar),
     depths=vector(#vpar));
  forstep(i=#seq,1,-1,    \\ downwards
     my(v=seq[i], p=vpar[v]);
     if(p, depths[v]=depths[p]+1));
  depths;
}
test_depths_vector_by_updown(vpar) =      \\ acyclic
{
  \\ At each i search upwards for a depths[v] already known, or the root.
  \\ depths[v]=-1 initially for not known, then >=0 when known.
  \\ Keep a linked list downwards through the v traversed, stored in depths[].
  \\ Fill them in back down again starting from the found depth or from -1
  \\ above the root.  Pre-increment means the latter sets depths[root] = 0.

  my(depths=vector(#vpar,v,-1));
  for(i=1,#vpar,
     my(v=i,c=0,d);      \\ c = child of v, or 0
     while((d=depths[v]) < 0,
           depths[v]=c; c=v;
           (v=vpar[v]) || break());
     while(c, v=c; c=depths[c]; depths[v]=d++)); \\ pre-increment
  depths;
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got=vpar_depths_vector(vpar));
     check_equal(got, vector(#vpar,v,vpar_depth_of_vertex(vpar,v)),
       "vpar_depths_vector() vs individual vpar_depth_of_vertex()");
     check_equal(got, test_depths_vector_by_seq(vpar),
       "vpar_depths_vector() vs test_depths_vector_by_seq()");
     check_equal(got, test_depths_vector_by_updown(vpar),
       "vpar_depths_vector() vs test_depths_vector_by_updown()");
     check_equal(got, test_depths_vector_by_acyclic(vpar),
       "vpar_depths_vector() vs test_depths_vector_by_acyclic()");
  );
}

test_depths_vector_by_cyclics(vpar) =
{
  my(depths=vector(#vpar,v,-1),
     compnum=vpar_compnum_vector(vpar),
     cyclelens=vpar_component_cycle_lengths(vpar),
     cyclics=vpar_cyclic_vertices(vpar));
  for(i=1,#cyclics, depths[cyclics[i]] = cyclelens[compnum[cyclics[i]]] - 1);

  my(stop);
  until(stop,
        stop=1;
        for(v=1,#vpar,
           if(depths[v]>=0, next);
           if(vpar[v],
              if(depths[vpar[v]]>=0,
                 depths[v] = depths[vpar[v]] + 1;
                 stop=0);
              ,
              depths[v]=0; stop=0)));
  depths;
}

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        got=vpar_depths_vector(vpar));
     check_equal(got,
                 test_depths_vector_by_cyclics(vpar),
                 "vpar_depths_vector() vs test_depths_vector_by_cyclics()");
     check_equal(got,
                 vector(#vpar,v, vpar_depth_of_vertex(vpar,v)),
                 "vpar_depths_vector() vs individual vpar_depth_of_vertex()");
  );
}

{
  for(c=1,8,
     forstep(t=0,6,3,
        my(vpar=test_make_cycle(c,6),
           want=vector(#vpar,i, if(i<=c,c-1,i-1)),
           perm=vpar_INTERNAL_random_perm(#vpar));
        vpar=vpar_relabel_seq(vpar,perm);
        want=vecextract(want,perm);

        check_equal(vpar_component_cycle_lengths(vpar), [c]);
        my(got=vpar_depths_vector(vpar));
        check_equal(got, want,
                    "vpar_depths_vector() on cycle and tail");
        check_equal(got,
                    test_depths_vector_by_cyclics(vpar),
                    "vpar_depths_vector() vs test_depths_vector_by_cyclics() on cycle and tail");
        check_equal(got,
                    vector(#vpar,v, vpar_depth_of_vertex(vpar,v)),
                    "vpar_depths_vector() vs individual vpar_depth_of_vertex() on cycle and tail");
     ));
}

\\-----------
\\ lex ordering

{
  for(n=1,10,
     for(rep=1,min(n,5),
        my(forest=random(2),
           vpar1=vpar_make_random_preorder(n,forest),
           vpar2=vpar_make_random_preorder(n,forest),
           depths1=vpar_depths_vector(vpar1),
           depths2=vpar_depths_vector(vpar2));
        check_equal(lex(vpar1,vpar2),
                    lex(depths1,depths2),
                    "vpar_depths_vector() lex same as vpar lex");
     ));
}

\\-----------
\\ total depths - labelled
{
  \\ A000435 total depths, labelled rooted root=1
  my(forest=-1);
  for(n=1, test_limit_trees_search - forest,
     my(total=0);
     forvec(vec=vpar_forvec_from(n,forest), \\ root=1
           my(vpar=vpar_from_upascend(vec));
           total += vecsum(vpar_depths_vector(vpar)));
     check_equal(total, A000435_samples[n],
                 "A000435 depths total, root=1"));

  check_equal(vector(#A000435_samples,n, (n-1)!*sum(k=0,n-2, n^k/k!)),
              A000435_samples,  \\ OFFSET=1
              "A000435 formula");
}
{
  \\ A001864 total depths, labelled rooted trees
  \\   n*A000435(n)
  my(forest=0);
  for(n=1, test_limit_trees_search - forest,
     my(total=0);
     forvec(vec=vpar_forvec_from(n,forest), \\ trees
           my(vpar=vpar_from_upascend(vec));
           total += vecsum(vpar_depths_vector(vpar)));
     check_equal(total, A001864_samples[n],
                "A001864 subtree sizes total, trees"));
}
{
  \\ A057500 1/2 * forests total depths
  \\         num connected graphs
  \\         num unicyclic connected graphs
  \\         num trees with 1 inversion (i>j and j descendant of i)
  my(forest=1);
  for(n=0, test_limit_trees_search - forest,
     my(total=0);
     forvec(vec=vpar_forvec_from(n,forest), \\ forests
           my(vpar=vpar_from_upascend(vec));
           total += vecsum(vpar_depths_vector(vpar)));
     check_equal(1/2*total, A057500_samples[n+1],
                "A057500 depths total, forests"));
}

\\-----------
\\ total depths - ordered
{
  \\ ordered total depths
  \\ trees   A000346 a(n) = 2^(2*n+1) - binomial(2*n+1, n+1)
  \\ forests A006419 a(n) = 2^(2*n+1) - C(2*n+3,n+1) + C(2*n+1,n).
  for(forest=0,1,
     for(n=0,7-forest,
        my(total=0,
           vpar=vpar_preorder_first(n));
        until((vpar=vpar_preorder_next(vpar,forest))=='none,
              total += vecsum(vpar_depths_vector(vpar)));
        \\ print1(total", ");
        check_equal(total,
                    if(n<=1, 0,
                       forest==0,
                       2^(2*n-3) - binomial(2*n-3,n-1),
                       2^(2*n-1) - binomial(2*n+1,n) + binomial(2*n-1,n-1)),
                    "vpar_depths_vector() total over ordered")));
}

\\-----------
\\ total depths - downwards
{
  \\ down labelled total depths
  \\ trees   A001705 generalized Sterling n! * Sum_{k=0..n-1} (k+1)/(n-k)
  \\   0, 0, 1, 5, 26, 154, 1044, 8028, 69264,
  \\   vector(10,n, abs(stirling(n+1,2)) - n!)
  \\   a(n) = (n+1)*a(n-1) + n!
  \\
  \\ forests A002538 Second-order Eulerian numbers <<n+1,n-1>>.
  \\   0, 0, 1, 8, 58, 444, 3708, 33984,
  \\   a(n) = (n+2)*a(n-1) + n*n!
  \\
  \\ For trees, let total depths be D(n).  The last vertex is underneath any
  \\ of the n-1 preceding.  So (n-1)*D(n-1) copies of those n-1, then last
  \\ under each preceding means a further D(n-1), plus 1 deeper of each
  \\ which is (n-1) * vpar_count_updown(n-1) = (n-1)!.
  \\
  \\ For forests similarly, total depths E(n).  Last vertex underneath any
  \\ of the n-1 preceding and also a new root.  So n*E(n-1) copies of those
  \\ E(n-1), then last vertex depths are total further E(n-1), plus 1 deeper
  \\ each (n-1) * vpar_count_updown(n-1,1) = (n-1)*(n-1)!.

  my(D = (n)-> if(n==0,0, abs(stirling(n+1,2)) - n!),      \\ trees
     E = (n)-> sum(k=1,n-1, k*abs(stirling(n+1,k+2))));    \\ forests

  check_equal(vector(10,n,n++; n*D(n-1) + (n-1)!),
              vector(10,n,n++; D(n)));
  check_equal(vector(10,n, vpar_count_updown(n)),
              vector(10,n, (n-1)!));
  check_equal(vector(10,n,n++; (n-1)*vpar_count_updown(n-1)),
              vector(10,n,n++; (n-1)!));

  check_equal(vector(10,n,n++; (n+1)*E(n-1) + (n-1)*(n-1)!),
              vector(10,n,n++; E(n)));
  check_equal(vector(10,n, vpar_count_updown(n,1)),
              vector(10,n, n!));
  check_equal(vector(10,n,n++; (n-1)*vpar_count_updown(n-1,1)),
              vector(10,n,n++; (n-1)*(n-1)!));

  for(forest=0,1,
     for(n=0,8-forest,
        my(total=0);
        for(i=1,vpar_count_updown(n,forest),
           my(vpar=vpar_downwards_ith(n,i,forest));
           total += vecsum(vpar_depths_vector(vpar)));
        \\ print1(total", ");
        check_equal(total,
                    if(n==0,0,
                       forest==0,     abs(stirling(n+1,2)) - n!,
                       forest && n==2, 1,
                       sum(k=1,n-1, k*abs(stirling(n+1,k+2)))),
                    "vpar_depths_vector() total over downwards");
     ));

  my(limit=20,
     L = O(x^limit));
  check_equal(test_func_to_egf(D,limit),
              - intformal(log(1-x +L) / (1-x)^2) + L,
              "vpar_depths_vector() total trees downwards, egf");

  check_equal(test_func_to_egf(E,limit),
              - intformal( (x + 2*log(1-x + L)) / (1-x)^3 ) + L,
              "vpar_depths_vector() total forests downwards, egf");
  check_equal(intformal(x/(x-1)^3), (1/2-x)/(x-1)^2);
  check_equal(-1/2 + (1/2-x)/(x-1)^2, -1/2*x^2/(x-1)^2);
  check_equal(test_func_to_egf(E,limit),
              -1/2*x^2/(x-1)^2 - 2*intformal( log(1-x + L)/(1-x)^3 ) + L,
              "vpar_depths_vector() total forests downwards, egf");
}
{
  my(limit=min(20,#A001705_samples),
     L = O(x^limit));
  \\ per A001705 by Neil Sloane
  check_equal(test_vector_to_egf(A001705_samples,0,limit),  \\ OFFSET=0
              -log(1-x +L) / (1-x)^2,
              "A001705 egf");
}
{
  my(limit=min(20,#A002538_samples),
     L = O(x^limit));
  \\ per Vladeta Jovovic in A002538
  check_equal(test_vector_to_egf(A002538_samples,1,limit),  \\ OFFSET=1
              (x + 2*log(1-x + L)) / (x-1)^3,
              "A002538 egf");
}
{
  check_equal(vector(#A000254_samples,n,n--; abs(stirling(n+1,2))),
              A000254_samples, \\ OFFSET=0
              "A000254 abs Stirling numbers first kind (n+1,2)");

  check_equal(vector(#A000254_samples,n,n--;
                     n!*sum(k=1,n, (-1)^(k+1)*binomial(n,k)/k)),
              A000254_samples,  \\ OFFSET=0
              "A000254 abs Stirling numbers first kind (n+1,2)");

  check_equal(vector(#A000254_samples,n,n--;
                     - sum(k=1,n, n!/k * (-1)^k*binomial(n,k))),
              A000254_samples,  \\ OFFSET=0
              "A000254 abs Stirling numbers first kind (n+1,2)");
}

\\-------
\\ vpar_depths_vector() Pre-Order Increase One Entry

\\ A002054 David Callan
\\ Dyck path UUDUUDDD
\\           |------|    subpaths, 5 of
\\            ||
\\            |----|
\\              |--|
\\               ||
\\ A002054 Emeric Deutsch
\\ Number of valleys in all Dyck paths of semilength n+1.
\\ a(2)=5 because UD*UD*UD, UD*UUDD, UUDD*UD, UUD*UDD, UUUDDD
\\ valleys marked by *, is each DU

\\ Return the number of 01 bit pairs in integer n written in binary.
test_count_01_bit_pairs(n) =
{
  my(vec=binary(n));
  sum(i=1,#vec-1, vec[i]==0 &&  vec[i+1]==1);
}

{
  my(T = (n)-> binomial(2*n-1,n-2));
  for(n=1,8,
     my(count=0,
        total_01=0,
        vpar=vpar_preorder_first(n));
     until((vpar=vpar_preorder_next(vpar,1))=='none, \\ forests
           my(depths=vpar_depths_vector(vpar));
           for(i=2,#depths,
              if(depths[i]<=depths[i-1],
                 count++));
           total_01 += test_count_01_bit_pairs(vpar_to_balanced_binary(vpar)));
     \\ print(count" "binomial(2*n-1,n-2));
     check_equal(count, T(n),
                 "vpar_depths_vector() ordered count increase one entry");
     check_equal(total_01, T(n),
                 "vpar_to_balanced_binary() total 01 pairs"));

  check_equal(vector(#A002054_samples,n, T(n+1)),
              A002054_samples);  \\ OFFSET=1
  check_equal(vector(#A002054_samples,n, binomial(2*n+1,n-1)),
              A002054_samples);  \\ OFFSET=1
  check_equal(vector(#A002054_samples,n,n--; binomial(2*n+3,n)),
              A002054_samples);
}
{
  \\ can decrease depths[i] when not already 0, and where depths[i+1] <=
  \\ depths[i] so that i+1 isn't depending on i
  for(n=1,5,
     my(vpar=vpar_preorder_first(n));
     until((vpar=vpar_preorder_next(vpar,1))=='none, \\ forests
           my(depths=vpar_depths_vector(vpar));
           for(i=1,#depths,
              my(new_depths=depths);
              new_depths[i]--;
              \\ print("i="i"  "depths" -> "new_depths);
              my(good=vpar_is_preorder_depths_vector(new_depths,1));
              check_equal(good,
                          depths[i]>0 && (i==n || depths[i+1]<=depths[i]));
              if(good,
                 my(new_vpar=vpar_make_preorder_depths(new_depths),
                    bb=vpar_to_balanced_binary(vpar),
                    new_bb=vpar_to_balanced_binary(new_vpar),
                    x=bitxor(bb,new_bb));  \\ binary 1100..00
                 check_equal(x, 3<<valuation(x,2),
                             "depths change by 1 is vpar_to_balanced_binary() flip 01<->10")))));
}


\\-----------------------------------------------------------------------------
progress("test_make_cycle()");

check_equal(test_make_cycle(0), []);
check_equal(test_make_cycle(1), [1]);
check_equal(test_make_cycle(2), [2,1]);
check_equal(test_make_cycle(5), [5,1,2,3,4]);
check_equal(test_make_cycle(5,3), [5,1,2,3,4,5,6,7]);
check_equal(test_make_cycle(0,3), [0,1,2]);

for(t=0,5, check_equal(test_make_cycle(0,t), vpar_make_path(t)));


\\-----------------------------------------------------------------------------
progress("test_is_isomorphic_halves()");

check_equal(test_is_isomorphic_halves([0,1]), 1);
check_equal(test_is_isomorphic_halves([0,1,2]), 0);
{
  for(same=0,1,
     for(k=1,8,
        check_equal(test_is_isomorphic_halves(vpar_make_binomial_tree(2^k),
                                              same), 1,
                    "test_is_isomorphic_halves() binomial tree");
        check_equal(test_is_isomorphic_halves(vpar_make_path(2^k),
                                              same), 1,
                    "test_is_isomorphic_halves() path")));
}


\\-----------------------------------------------------------------------------
progress("test_perms_wreath_product()");

{
  my(S2=test_perms_Sn(2),
     S2wrS2=test_perms_wreath_product(S2,S2),
     D4=test_perms_Dn(4));

  check_equal(S2wrS2,
              [Vecsmall([1, 2, 3, 4]), Vecsmall([1, 2, 4, 3]),
               Vecsmall([2, 1, 3, 4]), Vecsmall([2, 1, 4, 3]),
               Vecsmall([3, 4, 1, 2]), Vecsmall([3, 4, 2, 1]),
               Vecsmall([4, 3, 1, 2]), Vecsmall([4, 3, 2, 1])],
              "test_perms_wreath_product() S2 x S2");

  \\ print(Set(S2wrS2));
  \\ print(Set(D4));
  check_equal(test_perms_is_isomorphic(S2wrS2, D4), 1,
              "test_perms_wreath_product() S2wrS2 = D4");
  check_equal(test_perms_is_isomorphic(S2wrS2, test_quaternion_perms), 0,
              "test_perms_wreath_product() S2wrS2 != quaternions");
  check_equal(apply(test_permorder,D4), [1, 4, 2, 4, 2, 2, 2, 2],
              "D4 permorder()");

  my(S2_pow6=test_perms_direct_power(S2,6),
     S2wrS2_x_S2wrS2=test_perms_direct_product(S2wrS2,S2wrS2));
  check_equal(#S2_pow6, 64);
  check_equal(#S2wrS2_x_S2wrS2, 64);
  check_equal(test_perms_is_isomorphic(S2_pow6, S2wrS2_x_S2wrS2),
              0,
              "S2^6 != (S2wrS2)^2");

  my(S2xS2xS2       = test_perms_direct_power(S2,3),
     S2xS2xS2_wr_S2 = test_perms_wreath_product(S2xS2xS2,S2),
     S2wrS2_x_S2wrS2_x_S2 = test_perms_direct_product(S2wrS2_x_S2wrS2, S2));
  check_equal(#S2xS2xS2, 8);
  check_equal(#S2xS2xS2_wr_S2,       128, "S2xS2xS2_wr_S2 size");
  check_equal(#S2wrS2_x_S2wrS2_x_S2, 128, "S2wrS2_x_S2wrS2_x_S2 size");

  check_equal(vecsort(apply(test_permorder,S2xS2xS2_wr_S2)),
              vecsort(apply(test_permorder,S2wrS2_x_S2wrS2_x_S2)),
              "permorders S2xS2xS2_wr_S2 = permorders S2wrS2_x_S2wrS2_x_S2");
  \\ print(vecsort(apply(test_permorder,S2xS2xS2_wr_S2)));
  \\ print(vpar_INTERNAL_histogram(apply(test_permorder,S2xS2xS2_wr_S2)));

  \\ too big
  \\ check_equal(test_perms_is_isomorphic(S2xS2xS2_wr_S2, S2wrS2_x_S2wrS2_x_S2),
  \\             1,
  \\             "S2xS2xS2_wr_S2 = S2wrS2_x_S2wrS2_x_S2");
}


\\-----------------------------------------------------------------------------
progress("vpar_relabel_preorder()");

check_equal(vpar_relabel_preorder([]), [], "vpar_relabel_preorder() empty");
check_equal(vpar_relabel_preorder([0]), [0]);
check_equal(vpar_relabel_preorder([0,1]), [0,1]);
check_equal(vpar_relabel_preorder([2,0]), [0,1]);

\\ down=1
check_equal(vpar_relabel_preorder([0,1],1), [0,1]);
check_equal(vpar_relabel_preorder([2,0],1), [0,1]);
\\  1  2       1  3
\\     |  =>   |
\\     3       2
{
  my(vpar=[0,0,2],
     want_down=[0,1,0],
     rev=vpar_relabel_reverse(vpar));
  check_equal(rev, [2,0,0]);

  check_equal(vpar_is_preorder(vpar), 1);
  check_equal(vpar_relabel_preorder(vpar),   vpar);
  check_equal(vpar_relabel_preorder(vpar,1), want_down);

  \\ rev then plain   -> down=1
  check_equal(vpar_relabel_preorder(rev),    want_down);

  \\ rev then down=1   -> plain
  check_equal(vpar_relabel_preorder(rev,1),  vpar);
}

{
  \\ rev then down=1  -> plain
  \\
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got=vpar_relabel_preorder(vpar),
        got_down=vpar_relabel_preorder(vpar,1));
     vpar_validate(got);
     vpar_validate(got_down);
     check_equal(vpar_is_preorder(got), 1,
                 "vpar_relabel_preorder() is preorder");
     check_equal(vpar_is_preorder(got_down), 1,
                 "vpar_relabel_preorder() down=1 is preorder");

     \\ print(vpar" down="down"  rev="vpar_relabel_reverse(vpar));
     \\ print(" got "got);
     \\ print(" opp "vpar_relabel_preorder(vpar_relabel_reverse(vpar), 1));

     check_equal(got_down,
                 vpar_relabel_preorder(vpar_relabel_reverse(vpar)),
                 "vpar_relabel_preorder() of reverse is down=1");
     check_equal(got,
                 vpar_relabel_preorder(vpar_relabel_reverse(vpar), 1),
                 "vpar_relabel_preorder() down=1 of reverse is plain");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_relabel_postorder()");

check_equal(vpar_relabel_postorder([]), [], "vpar_relabel_postorder() empty");
check_equal(vpar_relabel_postorder([0]), [0]);
check_equal(vpar_relabel_postorder([0,1]), [2,0]);
check_equal(vpar_relabel_postorder([2,0]), [2,0]);
check_equal(vpar_relabel_postorder([0,1,1,3,0,5]), [4,3,4,0,6,0]);

\\ down=1
check_equal(vpar_relabel_postorder([0,1],1), [2,0]);
check_equal(vpar_relabel_postorder([2,0],1), [2,0]);
\\  1  3       2  3
\\     |  =>   |
\\     2       1
check_equal(vpar_relabel_postorder([0,3,0]),   [0,3,0]);
check_equal(vpar_relabel_postorder([0,3,0],1), [2,0,0]);
check_equal(vpar_relabel_postorder([2,0,0]),   [2,0,0]);
check_equal(vpar_relabel_postorder([2,0,0],1), [0,3,0]);

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got=vpar_relabel_postorder(vpar),
        got_down=vpar_relabel_postorder(vpar,1));
     vpar_validate(got);
     vpar_validate(got_down);
     check_equal(vpar_is_postorder(got), 1,
                 "vpar_relabel_postorder() is postorder");
     check_equal(vpar_is_postorder(got_down), 1,
                 "vpar_relabel_postorder() down=1 is postorder");

     check_equal(got_down,
                 vpar_relabel_postorder(vpar_relabel_reverse(vpar)),
                 "vpar_relabel_postorder() of reverse is down=1");
     check_equal(got,
                 vpar_relabel_postorder(vpar_relabel_reverse(vpar), 1),
                 "vpar_relabel_postorder() down=1 of reverse is plain");
  );
}


\\------------------------------------------------------------------------------
progress("vpar_INTERNAL_identity_perm()");
\\        test_identity_perm()

check_equal(vpar_INTERNAL_identity_perm(0), Vecsmall([]));
check_equal(vpar_INTERNAL_identity_perm(1), Vecsmall([1]));
check_equal(vpar_INTERNAL_identity_perm(3), Vecsmall([1,2,3]));

check_equal(test_identity_perm(0), Vecsmall([]));
check_equal(test_identity_perm(1), Vecsmall([1]));
check_equal(test_identity_perm(3), Vecsmall([1,2,3]));


\\-----------------------------------------------------------------------------
progress("test_odd_double_factorial()");

{
  check_equal(vector(#A001147_samples,n,n--; test_odd_double_factorial(n)),
              A001147_samples,    \\ OFFSET=0
              "A001147 test_odd_double_factorial()");

  \\ partitions of 1..2n into pairs
  \\ (2*n)! permute all,
  \\ /n! unordered among pairs
  \\ /2^n unordered within each pairs
  check_equal(vector(100,n,n--; test_odd_double_factorial(n)),
              vector(100,n,n--; (2*n)! / n! /2!^n));

  \\ egf 1/sqrt(1-2x)
  my(limit=100);
  check_equal(vector(limit,n,n--; test_odd_double_factorial(n)),
              test_egf_terms( 1 / sqrt(1-2*x + O('x^limit)), limit),
              "test_odd_double_factorial() egf");
}


\\-----------------------------------------------------------------------------
progress("vpar_is_upwards()");
\\ vpar_is_downwards()

\\   1
\\   |
\\   2
check_equal(vpar_is_upwards([0,1]), 0);
check_equal(vpar_is_downwards([0,1]), 1);

\\   2
\\   |
\\   1
check_equal(vpar_is_upwards([2,0]), 1);
check_equal(vpar_is_downwards([2,0]), 0);

{
  \\   2
  \\   |
  \\   1
  \\   |
  \\   3
  my(vpar=[2, 0, 1]);
  check_equal(vpar_is_upwards(vpar), 0);
  check_equal(vpar_is_downwards(vpar), 0);
}

\\ cyclics
check_equal(vpar_is_upwards([2,1]), 0, "vpar_is_upwards() 2-cycle");
check_equal(vpar_is_downwards([2,1]), 0, "vpar_is_downwards() 2-cycle");
check_equal(vpar_is_upwards([1]), 0, "vpar_is_upwards() self-loop");
check_equal(vpar_is_downwards([1]), 0, "vpar_is_downwards() self-loop");

{
  \\ Mitchell, Cockayne, Hedetniemi, "Linear Algorithms on Recursive
  \\ Representations of Trees", Journal of Computer and System Sciences,
  \\ volume 18, 1979, pages 76-85.
  \\ https://www.sciencedirect.com/science/article/pii/0022000079900539
  \\ Figure 1 (a),(b) tree T.
  \\
  \\                   9
  \\                   |
  \\  10---8---4---1---2---3---5
  \\           |       |
  \\           7       6
  my(vpar=[0, 1, 2, 1, 3, 2, 4, 4, 2, 8]);   \\ their "C"
  check_equal(#vpar, 10);
  check_equal(vpar_is_downwards(vpar), 1);
}

\\ Monotonic tree has all vpar[v]<v so each parent smaller number (or bigger
\\ number).  Vertices 1..n can be added under any existing, so (n-1)!.
{
  \\ counts (n-1)! shown in the addhelp
  for(forest=0,1,
     for(n=0, test_limit_trees_search - forest,
        my(forest_of_singletons=vector(n));  \\ [0,0,...,0]
        my(count_upwards=0,
           count_downwards=0);
        forvec(vec=vpar_forvec_from(n,forest),
              my(vpar=vpar_from_upascend(vec),
                 is_upwards  =vpar_is_upwards(vpar),
                 is_downwards=vpar_is_downwards(vpar));
              count_upwards   += is_upwards;
              count_downwards += is_downwards;

              \\ the only both upwards and downwards is the [0,0,..] forest
              \\ of singletons
              check_equal(is_upwards && is_downwards,
                          vpar==forest_of_singletons));
        my(want = if(forest, n!,  if(n==0,1,(n-1)!)));
        check_equal(count_upwards,   want);
        check_equal(count_downwards, want);
     ));
}
\\ vector(10,n,n--; if(n==0,1, (n-1)!))

{
  for(i=1,#test_sample_cyclics,
     my(vpar=test_sample_cyclics[i]);
     check_equal(vpar_is_upwards(vpar), 0,
                 "vpar_is_upwards() false on cyclic");
     check_equal(vpar_is_downwards(vpar), 0,
                 "vpar_is_downwards() false on cyclic"));
}


\\------------------
\\ Bruce Sagan, "Probabilistic Proofs of Hook Length Formulas Involving Trees"
\\ Abstract 1052-05-58.pdf
\\ Increasing labelling child < parent.
\\ fT = number of increasing labellings.
\\ Hook length = subtree size at and below v.
\\ Hook length formula
\\    fT = n! / prod hooklength(v)

\\ "hook length" formula
test_num_upwards_relabellings_by_formula(vpar) = \
  (#vpar)! / test_vector_product(vpar_subtree_sizes_vector(vpar));
test_num_upwards_relabellings_by_search(vpar) =
{
  sum(n=1,(#vpar)!,
      vpar_is_upwards(vpar_relabel_perm(vpar,numtoperm(#vpar,n))));
}
{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i]);
     #vpar <= test_limit_perms_search || next();
     check_equal(test_num_upwards_relabellings_by_formula(vpar),
                 test_num_upwards_relabellings_by_search(vpar)));
}


\\-----------------------------------------------------------------------------
progress("vpar_count_ordered()");

{
  for(n=1,5,
     vpar_count_ordered(n,0);
     vpar_count_ordered(n,1);
     check_equal(iferr(vpar_count_ordered(n,'forest);0, e, 1),
                 1,
                 "vpar_count_ordered() error on forest=symbol (whereas it should be 0,1)"));
}

{
  my(C = (n)->vpar_count_ordered(n,1));
  for(n=0,10,
     \\ Catalan C(n+1) = 2*(2*n+1)/(n+2) * C(n) in examples/random-decorated.gp
     check_equal(C(n+1), 2*(2*n+1)/(n+2) * C(n)));
}
{
  for(forest=0,1,
     for(n=0,12-forest,
        check_equal(vpar_count_ordered(n,forest) <= vpar_count_updown(n,forest),
                    1)));
}

{
  check_equal(vector(#A001813_samples,n,n--; (n+1)! * vpar_count_ordered(n,1)),
              A001813_samples,    \\ OFFSET=0
              "A001813 decorated binary trees");
  check_equal(vector(50,n,n--; (n+1)! * vpar_count_ordered(n,1)),
              vector(50,n,n--; (2*n)! / n!));
}
{
  my(got=vector(10));
  for(n=1,#got,
     my(vpar=vpar_preorder_first(n));
     until((vpar=vpar_preorder_next(vpar))=='none,
           got[n] += (vecmax(vpar_num_children_vector(vpar)) <= 2)));
  check_equal(got, A001006_samples[1..#got],
              "A001006 Motzkin = count ordered trees with <=2 children");
}


\\-----------------------------------------------------------------------------
progress("vpar_make_random_preorder()");

{
  for(n=2,5,
     vpar_make_random_preorder(n,1);
     check_equal(iferr(vpar_make_random_preorder(n,'forest);0, e, 1),
                 1,
     "vpar_make_random_preorder() error on forest=symbol (whereas it should be 0,1)"));
}
{
  for(forest=0,1,
     for(n=1,20,
        for(rep=1,min(n,5),
           my(vpar=vpar_make_random_preorder(n,forest));
           vpar_validate(vpar);
           check_equal(vpar_is_preorder(vpar), 1,
                       "vpar_make_random_preorder() is preorder");
           if(!forest,
              check_equal(vpar_is_tree(vpar), 1,
                          "vpar_make_random_preorder() is tree")))));
}

\\ sum i=0,k of Cat(n,i)
test_INTERNAL_Catalan_triangle_cumulative(n,k) = \
  binomial(n+k+1, n+1)*(n-k+2)/(n+2);
{
  for(n=0,20,
     for(k=0,n,
        check_equal(sum(i=0,k, vpar_INTERNAL_Catalan_triangle(n,i)),
                    test_INTERNAL_Catalan_triangle_cumulative(n,k))));
}
{
  for(n=1,10,
     for(k=0,n-1,
        check_equal(vpar_INTERNAL_Catalan_triangle(n,k)
                    / vpar_INTERNAL_Catalan_triangle(n-1,k),
                    (n-k+1)*(n+k) / ( (n+1)*(n-k) ))));
}
{
  for(n=1,10,
     for(k=0,n,
        check_equal(vpar_INTERNAL_Catalan_triangle(n,k),
                    test_INTERNAL_Catalan_triangle_cumulative(n-1,k))));
}
{
  for(n=0,10,
     for(k=1,n,
        check_equal(vpar_INTERNAL_Catalan_triangle(n,k)
                    / test_INTERNAL_Catalan_triangle_cumulative(n,k-1),
                    (n+2)*(n-k+1) / ( (n-k+3)*k ))));
}

\\ t = 1/probability taken, which should come out as == num ordered trees or
\\     forests
\\ list = flags of whether rand compare is to be true or false,
\\     except that if comp >= limit then always false
test_make_random_preorder_with_probabilities__t = 0;
test_make_random_preorder_with_probabilities__list = List([]);
test_make_random_preorder_with_probabilities__rand(limit,comp) =
{
  \\ print(limit" "comp);
  my(ret=test_make_random_preorder_with_probabilities__list[1]);
  listpop(test_make_random_preorder_with_probabilities__list,1);
  if(comp>=limit, return(0));

  \\ comparing rand(limit) >= comp
  \\ so false = comp       chance
  \\    true  = limit-comp chance
  test_make_random_preorder_with_probabilities__t
    *= if(ret, limit-comp, comp) / limit;
  ret;
}
test_make_random_preorder_with_probabilities(n,forest=0) =
{
  \\ same code as vpar_make_random_preorder(), but
  \\ test_make_random_preorder_with_probabilities__rand() for the random() and
  \\ compare, and total probability accumulated

  my(vpar=vector(n));
  if(n>=2,
     my(p=2-forest,     \\ parent vertex of v if v is set to depth d
        d=1);
     vpar[2]=1;
     for(v=3-forest,n,
        my(k=n-v+1);  \\ how many remaining, including v
        while(test_make_random_preorder_with_probabilities__rand
              ((d+(k<<1))*(d+1), (d+2)*k),      \\ compare >=
              p=vpar[p];
              d--);
        vpar[v]=p;
        p=v;
        d++;
     ));
  vpar;
}
{
  for(forest=0,1,
     for(n=1,6,
        my(want_t = 1 / vpar_count_ordered(n,forest));
        forvec(vec=vector(2*n,i,[0,1]),
              \\ print("flags "vec);
              test_make_random_preorder_with_probabilities__list=List(vec);
              test_make_random_preorder_with_probabilities__t = 1;
              my(vpar=test_make_random_preorder_with_probabilities(n,forest));
              check_equal(test_make_random_preorder_with_probabilities__t,
                          want_t,
                          "vpar_make_random_preorder() total probability"))));
}


\\-----------------------------------------------------------------------------
progress("test_string_replace()");

check_equal(test_string_replace("aaaa", "a", "XY"), "XYXYXYXY", \
            "test_string_replace() aa");
check_equal(test_string_replace("", "a", "X"), "");
check_equal(test_string_replace("abb", "a", "XY"), "XYbb");
check_equal(test_string_replace("bba", "a", "XY"), "bbXY");
check_equal(test_string_replace("bbadda", "a", "XY"), "bbXYddXY");
check_equal(test_string_replace("abcdefab", "ab", "XY"), "XYcdefXY");
check_equal(test_string_replace("abcdef", "cd", "XY"), "abXYef");
check_equal(test_string_replace("abcdef", "ef", "XY"), "abcdXY");


\\-----------------------------------------------------------------------------
progress("vpar_next_sibling_vectors_count()");

{
  for(n=1,4,
     vpar_next_sibling_vectors_count(n);
     check_equal(iferr(vpar_next_sibling_vectors_count(n,'forest);0, e, 1),
                 1,
                 "vpar_next_sibling_vectors_count() error on forest=symbol (instead of -1,0,1)"));
}
{
  \\ forests
  check_equal(vector(20,n,n--; vpar_next_sibling_vectors_count(n,1)),
              vector(20,n,n--; vpar_INTERNAL_Bell_number(n)));
  check_equal(vector(#A000110_samples,n,n--; vpar_next_sibling_vectors_count(n,1)),
              A000110_samples,    \\ OFFSET=0
              "vpar_next_sibling_vectors_count() forests = A000110");
}
{
  \\ root=1 trees
  check_equal(vector(20,n,n--; vpar_next_sibling_vectors_count(n,-1)),
              vector(20,n,n--; if(n==0,1,vpar_INTERNAL_Bell_number(n-1))));
  check_equal(vector(#A000110_samples,n, vpar_next_sibling_vectors_count(n,-1)),
              A000110_samples,
              "vpar_next_sibling_vectors_count() root=1 trees = A000110");
}

\\ Return a Bell numbers variant counting how many partitions of 1..n into
\\ sets where at least one set is a singleton.
test_Bell_ge1_single(n) =
{
  if(n<=1,1,
     vpar_INTERNAL_Bell_number(n-1) + sum(t=2,n-1, binomial(n-1,t-1)*test_Bell_ge1_single(n-t)));
}
{
  \\ trees
  check_equal(vector(20,n,n--; vpar_next_sibling_vectors_count(n)),
              vector(20,n,n--; test_Bell_ge1_single(n)));

  my(samples=A000296_samples);
  check_equal(samples[2], 0);  \\ starts   1,0,1,1,4
  samples=samples[^2];         \\ but want   1,1,1,4
  check_equal(vector(#samples,n,n--; vpar_next_sibling_vectors_count(n)),
              samples,
              "vpar_next_sibling_vectors_count() trees = A000296");
  check_equal(vector(20,n, vpar_next_sibling_vectors_count(n)),
              vector(20,n, sum(i=1,n, (-1)^(n-i)*vpar_INTERNAL_Bell_number(i))));

  \\ H. D. Becker, "Solution to problem E461", American Mathematical
  \\ Monthly, volume 48, 1941, pages 701-702, as presented in A000296
  check_equal(vector(20,n,n++; vpar_INTERNAL_Bell_number(n)),
              vector(20,n,n++; vpar_next_sibling_vectors_count(n)
                             + vpar_next_sibling_vectors_count(n-1)));

  my(gA000296 = test_func_to_egf(n->A000296_samples[n+1], #A000296_samples));
  check_equal(gA000296, exp(-1 - 'x + exp('x + O('x^#A000296_samples))),
              "A000296 egf");

  my(limit = 20,
     L = O('x^limit),
     gT = test_func_to_egf(vpar_next_sibling_vectors_count, limit-1));
  check_equal(gT,
              1 + exp(-1 - 'x + exp('x+L))',
              "vpar_next_sibling_vectors_count() egf like A000296");
}

\\ Return a vector of vector of vectors which are partitions of the elements
\\ of vector v.
\\ The return is [ partition, partition, ... ].
\\ Each partition = [ subset_of_v, subset_of_v, ... ].
\\ Each subset_of_v is a non-empty vector of some of the elements of v.
\\ The subsets are disjoint and together are all of v.
\\ The subsets in each partition are in order of their smallest element.
\\
test_partition_vec(v) =
{
  if(#v==0, [ []  ],   \\ list of empty partition only
     #v==1, [ [v] ],   \\ list of partition of v alone
     my(b=test_partition_vec(v[^1]),
        l=List([]));
     for(i=1,#b,
        my(p=b[i]);  \\ partition of rest
        listput(l, concat([[v[1]]], p));  \\ v[1] as own set
        for(j=1,#p,
           my(s=p); s[j]=concat([v[1]],s[j]);
           listput(l,vecsort(s))));   \\ v[1] together in another set
     Vec(l));
}
check_equal(test_partition_vec([]),    [ [] ]);
check_equal(test_partition_vec([1]),   [ [[1]] ]);
check_equal(test_partition_vec([1,2]), [ [[1],[2]], [[1,2]] ]);
{
  check_equal(test_partition_vec([1,2,3]), [[[1], [2], [3]],
                                            [[1,2], [3]],
                                            [[1,3], [2]],
                                            [[1], [2,3]],
                                            [[1,2,3]]
                                           ]);
}
check_equal(vector(7,n,n--; #test_partition_vec(vector(n,i))), \
            vector(7,n,n--; vpar_INTERNAL_Bell_number(n)));

\\ p is a single partition as from test_partition_vec(v) above.
\\ Return the number of single-element subsets in it.
test_partition_num_singles(p) = sum(i=1,#p, #p[i]==1);

{
  check_equal(vector(7,n, vpar_next_sibling_vectors_count(n)),
              vector(7,n, my(b=test_partition_vec(vector(n,i)));
                     sum(i=1,#b, test_partition_num_singles(b[i])>=1)),
    "vpar_next_sibling_vectors_count() vs count partitions with >=1 single");

  check_equal(vector(7,n,n++; vpar_next_sibling_vectors_count(n-1)),
              vector(7,n,n++; my(b=test_partition_vec(vector(n,i)));
                     sum(i=1,#b, test_partition_num_singles(b[i])==0)),
    "vpar_next_sibling_vectors_count() vs count partitions with no singles");
}


\\-----------------------------------------------------------------------------
progress("vpar_is_next_sibling_vector()");

check_equal(vpar_is_next_sibling_vector([],0,-1), 1, \
            "vpar_is_next_sibling_vector() empty");
check_equal(vpar_is_next_sibling_vector([],0,0), 1);
check_equal(vpar_is_next_sibling_vector([],0,1), 1);

check_equal(vpar_is_next_sibling_vector([0],0,-1), 1, \
            "vpar_is_next_sibling_vector() singleton root=1");
check_equal(vpar_is_next_sibling_vector([0],0,0), 1);
check_equal(vpar_is_next_sibling_vector([0],0,1), 1);

check_equal(vpar_is_next_sibling_vector([2,0,0],0), 1, \
            "vpar_is_next_sibling_vector() star-3 root at 3");
check_equal(vpar_is_next_sibling_vector([-99,0,0],0), 0);
check_equal(vpar_is_next_sibling_vector([1,0,0],0), 0);
check_equal(vpar_is_next_sibling_vector([99,0,0],0), 0);

{
  for(n=1,4,
     my(vec=vector(n));
     check_equal(vpar_is_next_sibling_vector(vec), 1);
     check_equal(vpar_is_next_sibling_vector(vec,0,0), 1);
     check_equal(iferr(vpar_is_next_sibling_vector(vec,'down);0, e, 1),
                 1,
                 "vpar_is_next_sibling_vector() error on down=symbol (whereas it should be 0,1)");
     check_equal(iferr(vpar_is_next_sibling_vector(vec,0,'forest);0, e, 1),
                 1,
                 "vpar_is_next_sibling_vector() error on forest=symbol (instead of -1,0,1)"));
}
{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i]);
     for(down=0,1,
        my(next_sibling=vpar_next_sibling_vector(vpar,down));
        \\ print(vpar"  "next_sibling);
        check_equal(vpar_is_next_sibling_vector(next_sibling,down),1,
                    Str("vpar_is_next_sibling_vector() of tree, down="down));
        check_equal(vpar_is_next_sibling_vector(next_sibling,down,1),1,
                    Str("vpar_is_next_sibling_vector() of tree as forest, down="down)));

     if(#vpar, vpar=vpar_reroot(vpar,1));
     for(down=0,1,
        my(next_sibling=vpar_next_sibling_vector(vpar,down));
        \\ print(vpar"  "next_sibling);
        check_equal(vpar_is_next_sibling_vector(next_sibling,down,-1), 1,
                    Str("vpar_is_next_sibling_vector() of tree root=1, down="down));
        check_equal(vpar_is_next_sibling_vector(next_sibling,down), 1,
                    Str("vpar_is_next_sibling_vector() of tree root=1 as tree, down="down));
        check_equal(vpar_is_next_sibling_vector(next_sibling,down,1), 1,
                    Str("vpar_is_next_sibling_vector() of tree root=1 as forest, down="down))));
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     for(down=0,1,
        my(next_sibling=vpar_next_sibling_vector(vpar,down));
        check_equal(vpar_is_next_sibling_vector(next_sibling,down,1),1,
                    "vpar_is_next_sibling_vector() of forest")));
}

{
  \\ l = list of all next_sibling vectors which occur from iterating.
  \\ All must pass vpar_is_next_sibling_vector().

  for(down=0,1,
     for(forest=0,1,
        for(n=0, test_limit_trees_search - forest,
           my(l=List([]));
           forvec(vec=vpar_forvec_from(n,forest), \\ trees or forests
                 my(vpar=vpar_from_upascend(vec),
                    next_sibling=vpar_next_sibling_vector(vpar,down),
                    pos=setsearch(l,next_sibling,1));
                 \\ print(vpar" degs "next_sibling" pos "pos);
                 if(pos,listinsert(l,next_sibling,pos));
                 check_equal(vpar_is_next_sibling_vector(next_sibling,down,forest), 1,
                 "vpar_is_next_sibling_vector() of vpar_forvec_from()"));


           check_equal(vpar_next_sibling_vectors_count(n,forest),
                       #l,
                       Str("vpar_next_sibling_vectors_count() forest="forest));

           forvec(next_sibling=vector(n,i,[-1,n+1]),
                 \\ print(next_sibling"  "setsearch(l,next_sibling));
                 check_equal(vpar_is_next_sibling_vector(next_sibling,down,forest),
                             setsearch(l,next_sibling)!=0,
                             Str("vpar_is_next_sibling_vector() vs setsearch"
                                 " forest="forest
                                 " down="down)));
        )));
}


\\-----------------------------------------------------------------------------
progress("vpar_next_sibling_vector()");

check_equal(vpar_next_sibling_vector([3,3, 0, 3]), [2,4, 0, 0], \
           "vpar_next_sibling_vector()");

{
  for(n=0,4,
     my(vpar=vpar_make_path(n));
     check_equal(iferr(vpar_next_sibling_vector(vpar,'down);0, e, 1),
                 1,
                 "vpar_next_sibling_vector() error on down=symbol (whereas it should be 0,1)"));
}

test_next_sibling_vector_by_prev(vpar,down=0) =
{
  \\ prev_child[p+1] is the previously seen child of parent p, including
  \\ p==0 for previously seen root
  \\
  my(prev_child=vectorsmall(#vpar+1),
     vec=vector(#vpar));
  forstep(v=if(down,1,#vpar), if(down,#vpar,1), if(down,1,-1),
         my(p=vpar[v]+1);
         vec[v]=prev_child[p];
         prev_child[p]=v);
  vec;
}

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     for(down=0,1,
        my(got=vpar_next_sibling_vector(vpar,down));
        check_equal(got,
                    vector(#vpar,v, vpar_next_sibling_of_vertex(vpar,v,down)),
                    "vpar_next_sibling_vector() vs vpar_next_sibling_of_vertex() individual");
        check_equal(got,
                    test_next_sibling_vector_by_prev(vpar,down),
                    "vpar_next_sibling_vector() vs test_next_sibling_vector_by_prev()")));
}


\\-----------------------------------------------------------------------------
progress("vpar_num_childful()");

check_equal(vpar_num_childful([]), 0, "vpar_num_childful() empty");
check_equal(vpar_num_childful([0]), 0, "vpar_num_childful() singleton");
check_equal(vpar_num_childful([1]), 1, "vpar_num_childful() self-loop");

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        got=vpar_num_childful(vpar));
     check_equal(got, #setminus(Set(vpar),[0]),
                 "vpar_num_childful() vs distinct vpar entries");
     check_equal(got + vpar_num_childless(vpar), #vpar,
                 "vpar_num_childful() plus vpar_num_childless()");
  );
}

{
  \\ Ordered forests total num childful = binomial(2*n-1,n-2)
  \\ A002054
  check_equal(vector(#A002054_samples,n,n++; binomial(2*n-1,n-2)),
              A002054_samples);  \\ up to offset
  for(n=0,10,
     my(vpar=vpar_preorder_first(n),
        total=0);
     until((vpar=vpar_preorder_next(vpar,1))=='none,  \\ forests
           total += vpar_num_childful(vpar));
     check_equal(total,
                 binomial(2*n-1,n-2),
                 "vpar_num_childful() total over ordered forests");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_childless_vertices()");

check_equal(vpar_childless_vertices([]),[]);
check_equal(vpar_childless_vertices([0]),[1], "vpar_childless_vertices() singleton");
check_equal(vpar_childless_vertices([0,1]),[2], "vpar_childless_vertices() path-2");
check_equal(vpar_childless_vertices([2,0]),[1]);

test_childless_vertices_by_gnot(vpar) =
{
  \\ Previous code:
  \\ must full Vec for select(), not vpar_INTERNAL_num_children_vecsmall()
  Vec(select(vpar_INTERNAL_gnot, vpar_num_children_vector(vpar), 1));
}

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        got=vpar_childless_vertices(vpar));
     check_equal(got, test_select_vertices(vpar,vpar_is_childless),
                 "vpar_childless_vertices() vs vpar_is_childless()");
     check_equal(got, test_childless_vertices_by_gnot(vpar),
                 "vpar_childless_vertices() vs test_childless_vertices_by_gnot()");
     check_equal(#got, vpar_num_childless(vpar),
                 "vpar_childless_vertices() count vs vpar_num_childless()");

     check_equal(#got, vpar_num_childless(vpar));
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_is_childless()");

check_equal(vpar_is_childless([0],1), 1);
check_equal(vpar_is_childless([0,1],1), 0);
check_equal(vpar_is_childless([0,1],2), 1);

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     check_equal(vpar_is_childless(vpar,0),
                 vpar_num_roots(vpar)==0,
                 "vpar_is_childless() v=0 vs vpar_num_roots() per addhelp");
     for(v=0,#vpar,
        check_equal(vpar_is_childless(vpar,v),
                    vpar_num_children_of_vertex(vpar,v)==0,
        "vpar_is_childless() vs vpar_num_children_of_vertex() per addhelp")));
}


\\-----------------------------------------------------------------------------
progress("vpar_is_childful()");

check_equal(vpar_is_childful([],1), 0);
check_equal(vpar_is_childful([0],1), 0);
check_equal(vpar_is_childful([0,1],1), 1);
check_equal(vpar_is_childful([0,1],2), 0);

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     check_equal(vpar_is_childful(vpar,0),
                 vpar_num_roots(vpar)!=0,
                 "vpar_is_childful() v=0 vs vpar_num_roots() per addhelp");
     for(v=0,#vpar,
        check_equal(vpar_is_childful(vpar,v),
                    vpar_num_children_of_vertex(vpar,v)!=0,
                    "vpar_is_childful() vs vpar_num_children_of_vertex()");
        check_equal(vpar_is_childful(vpar,v),
                    1 - vpar_is_childless(vpar,v),
        "vpar_is_childful() and vpar_is_childless() are opposites")));
}


\\-----------------------------------------------------------------------------
progress("vpar_balanced_binary_prev()");

test_balanced_binary_prev_by_search(b) =
{
  until(vpar_is_balanced_binary(b),
        b--; if(b<0, error()));
  b;
}
{
  check_equal(test_to_binary
         (test_balanced_binary_prev_by_search(test_from_binary(111110000010))),
                                                               111101100000);
  check_equal(test_to_binary
              (vpar_balanced_binary_prev(test_from_binary(111110000010))),
                                                          111101100000);

  check_equal(test_to_binary
              (vpar_balanced_binary_prev(test_from_binary(10101010))),
                                                            111000);
  check_equal(test_to_binary
              (vpar_balanced_binary_prev(test_from_binary(10))),
                                                           0);
  check_equal(test_to_binary
              (vpar_balanced_binary_prev(test_from_binary(1110001010101010))),
                                                          1101111110000000);
  check_equal(test_to_binary
              (vpar_balanced_binary_prev(test_from_binary(1110001010))),
                                                          1101110000);
  check_equal(test_to_binary
              (vpar_balanced_binary_prev(test_from_binary(1100))),
              1010);
}

{
  for(b=2,1024,
     vpar_is_balanced_binary(b) || next;
     \\ print("b="test_to_binary(b));
     \\ print(test_to_binary(vpar_balanced_binary_prev(b)));
     check_equal(vpar_balanced_binary_prev(b),
                 test_balanced_binary_prev_by_search(b),
     "vpar_balanced_binary_prev() vs test_balanced_binary_prev_by_search()"));
}


\\-----------------------------------------------------------------------------
progress("vpar_balanced_binary_next()");

test_balanced_binary_next_by_search(b) = \
  until(vpar_is_balanced_binary(b), b++); b;
{
  for(b=0,1024,
     vpar_is_balanced_binary(b) || next;
     check_equal(vpar_balanced_binary_next(b),
                 test_balanced_binary_next_by_search(b),
     "vpar_balanced_binary_next() vs test_balanced_binary_next_by_search()"));
}

{
  my(l=List([]), b=0);
  while(#l<#A085192_samples,
        my(t=vpar_balanced_binary_next(b));
        listput(l,t-b); b=t);
  check_equal(Vec(l), A085192_samples, "A085192 balanced binary increments");
}
{
  my(l=List([]), b=0);
  while(#l<#A014486_samples,
        listput(l,b);
        b=vpar_balanced_binary_next(b));
  check_equal(Vec(l), A014486_samples, "A014486 balanced binary (in decimal)");
}
{
  my(l=List([]), b=0);
  while(#l<#A063171_samples,
        listput(l, vpar_INTERNAL_fromdigits(binary(b)));
        b=vpar_balanced_binary_next(b));
  check_equal(Vec(l), A063171_samples, "A063171 balanced binary (in binary)");
}
{
  my(l=List([]), b=0);
  while(#l<#A071152_samples,
        listput(l, 2*vpar_INTERNAL_fromdigits(binary(b)));
        b=vpar_balanced_binary_next(b));
  check_equal(Vec(l),
              A071152_samples,
              "A071152 balanced binary (in binary), * 2");
}
{
  my(l=List([]), b=2);  \\ starting b=2
  while(#l<#A080237_samples,
        listput(l, valuation(b,2));
        b=vpar_balanced_binary_next(b));
  check_equal(Vec(l), A080237_samples, "A080237 num low zero bits");
}


\\-----------------------------------------------------------------------------
progress("vpar_num_claws()");

test_num_claws_by_search(vpar) =
{
  \\ FIXME: Would check against b,c,d edges to see only induced claws.
  my(ret=0);
  for(a=1,#vpar,
     for(b=1,#vpar,
        if(a==b, next);
        vpar_is_neighbour(vpar,a,b) || next;
        for(c=b+1,#vpar,
           if(a==c, next);
           vpar_is_neighbour(vpar,a,c) || next;
           for(d=c+1,#vpar,
              if(a==d, next);
              vpar_is_neighbour(vpar,a,d) || next;
              \\ print("found "a" "b" "c" "d);
              ret++))));
  ret;
}

check_equal(vpar_num_claws(vpar_make_star(4)), 1);

\\ maximum number of claws for n=7 vertices, equal maximum with some cyclics
check_equal(vpar_num_claws(vpar_make_star(7)), 20);

\\ check_equal(vpar_num_claws([1]),0, "vpar_num_claws() self-loop");
\\ {
\\   my(vpar=[1,1]);
\\   check_equal(vpar_degrees_vector(vpar), [3,1]);
\\   check_equal(vpar_num_claws(vpar),0, "vpar_num_claws() self-loop and 1 leaf");
\\ }
\\ check_equal(vpar_num_claws([1,1,1]),0, "vpar_num_claws() self-loop and 2 leaf");
\\ check_equal(vpar_num_claws([1,1,1,1]),1, "vpar_num_claws() self-loop and 3 leaf");

{
  \\   2
  \\   |
  \\   3
  \\  / \
  \\ 1   4
  my(vpar=[3, 0, 2, 3]);
  vpar_validate(vpar);
  check_equal(vpar_num_claws(vpar),1, "vpar_num_claws() claw");
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     \\ print(vpar);
     check_equal(vpar_num_claws(vpar),
                 test_num_claws_by_search(vpar),
                 "vpar_num_claws() vs test_num_claws_by_search()"));
}


\\-----------------------------------------------------------------------------
progress("vpar_is_automorphism()");

{
  \\          1
  \\        /   \
  \\      2      6
  \\     / \    / \
  \\    3   5  7   8
  \\    |          |
  \\    4          9
  my(vpar=[0,1,2,3,2,1,6,6,8]);
  vpar_validate(vpar);

  \\          2 3 4 5  6 7 8 9
  my(perm=[1, 6,8,9,7, 2,5,3,4]);
  check_equal(test_is_perm(perm), 1);
  check_equal(vpar_is_automorphism(vpar,perm), 1);
}
{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        perm=vpar_INTERNAL_random_perm(#vpar));

     check_equal(vpar_is_automorphism(vpar,perm),
                 vpar==vpar_relabel_perm(vpar,perm),
                 "vpar_is_automorphism() vs vpar_relabel_perm()");
  );
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        perm=vpar_INTERNAL_random_perm(#vpar));
     \\ by definition, vpar is isomorphic to any relabelling
     check_equal(vpar_is_isomorphic(vpar, vpar_relabel_perm(vpar,perm)),
                 1,
                 "vpar_is_isomorphic() to any relabelling");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_subdivide()");

check_equal(vpar_subdivide([]),[], "vpar_subdivide() empty tree");
check_equal(vpar_subdivide([],10),[], "vpar_subdivide() empty tree");
check_equal(vpar_subdivide([0]),[0]);
check_equal(vpar_subdivide([0,0]),[0,0]);
check_equal(vpar_subdivide([0,1]),[0,3,1]);
check_equal(vpar_subdivide([0,1,1]),[0,4,5,1,1]);
check_equal(vpar_subdivide([0,1,1],0),[0,1,1]);
check_equal(vpar_subdivide(Vecsmall([0,1,1]),3),[0,6,9, 1,4,5, 1,7,8]);

\\ addhelp() example
\\  1 <- 2
\\  1 <- 3,4,5,6 <- 2
check_equal(vpar_subdivide([0,1],4), [0,6, 1,3,4,5]);

check_equal(vpar_subdivide([1],1),[2,1]);
check_equal(vpar_subdivide([1],3),[4,1,2,3]);

\\    1 --> 2        1 -> 5,4,3 -> 2
\\      <--            <- 6,7,8 <-
check_equal(vpar_subdivide([2,1],3), [5,8, 2,3,4, 1,6,7]);

test_subdivide_size(vpar,s) = #vpar + s*vpar_num_edges(vpar);
{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        s=random(5),
        got=vpar_subdivide(vpar,s));
     check_equal(#got, test_subdivide_size(vpar,s)));
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        s=random(5),
        got=vpar_subdivide(vpar,s));
     vpar_validate(got);
     check_equal(#got, test_subdivide_size(vpar,s)));
}


\\-----------------------------------------------------------------------------
progress("vpar_relabel_swap()");

\\ 1<--2<--3<--4<--5<--6
\\ 1<--4<--3<--2<--5<--6
{
  check_equal(vpar_relabel_swap(vpar_make_path(6),2,4),
              [0, 3, 4, 1, 2, 5],
              "relabel swap");
  \\           1  2  3  4  5  6
}

check_equal(vpar_relabel_swap([1],1,1), [1]);
check_equal(vpar_relabel_swap([2,1],1,2), [2,1]);

\\  1 <-> 2     3 <-> 2
\\  |           |
\\  3           1
check_equal(vpar_relabel_swap([2,1,1],1,3), [3,3,2]);


\\-----------------------------------------------------------------------------
progress("vpar_minmax_ancestor_of_vertex()");

check_equal(vpar_minmax_ancestor_of_vertex([0],1,min), 1);
check_equal(vpar_minmax_ancestor_of_vertex([0],1,max), 1);
check_equal(vpar_minmax_ancestor_of_vertex([1],1,min), 1);
check_equal(vpar_minmax_ancestor_of_vertex([1],1,max), 1);
check_equal(vpar_minmax_ancestor_of_vertex([2,1],1,min), 1);
check_equal(vpar_minmax_ancestor_of_vertex([2,1],1,max), 2);

test_minmax_ancestor_of_vertex_by_limited(vpar,v,func) =
{
  \\ Looking upwards at most #vpar steps is enough to see all of a cycle,
  \\ but of course may traverse it many more times than actually needed.
  my(m=0);
  for(i=1,#vpar,
     m=func(if(m,m,v),v);
     if(!(v=vpar[v]), break()));
  m;
}
{
  my(funcs=[min,max]);
  for(f=1,#funcs,
     my(func=funcs[f]);
     for(i=1,#test_sample_forests_and_cyclics,
        my(vpar=test_sample_forests_and_cyclics[i]);
        for(v=1,#vpar,
           check_equal(vpar_minmax_ancestor_of_vertex(vpar,v,func),
                       test_minmax_ancestor_of_vertex_by_limited(vpar,v,func),
                       "vpar_minmax_ancestor_of_vertex() vs test_minmax_ancestor_of_vertex_by_limited()"))));
}

test_minmax_ancestor_of_vertex_by_unlimited(vpar,v,func) =
{
  \\ Previous code with an unlimited search upwards.
  if(vpar_is_cyclic(vpar), error("for acyclics only"));

  my(m=v); while((v=vpar[v]), m=func(m,v)); m;
}
{
  my(funcs=[min,max]);
  for(f=1,#funcs,
     my(func=funcs[f]);
     for(i=1,#test_sample_forests,
        my(vpar=test_sample_forests[i]);
        for(v=1,#vpar,
           check_equal(vpar_minmax_ancestor_of_vertex(vpar,v,func),
                       test_minmax_ancestor_of_vertex_by_unlimited(vpar,v,func),
                       "vpar_minmax_ancestor_of_vertex() vs test_minmax_ancestor_of_vertex_by_unlimited()"))));
}


\\-----------------------------------------------------------------------------
progress("vpar_minmax_ancestors_vector()")

check_equal(vpar_minmax_ancestors_vector([],max), []);
check_equal(vpar_minmax_ancestors_vector([0],max), [1]);
check_equal(vpar_minmax_ancestors_vector([0,0],max), [1,2]);
check_equal(vpar_minmax_ancestors_vector([0,1,0],max), [1,2,3]);
check_equal(vpar_minmax_ancestors_vector([0,0,0,0],max), [1,2,3,4]);

check_equal(vpar_minmax_ancestors_vector([],min), []);
check_equal(vpar_minmax_ancestors_vector([0],min), [1]);
check_equal(vpar_minmax_ancestors_vector([0,0],min), [1,2]);
check_equal(vpar_minmax_ancestors_vector([0,1,0],min), [1,1,3]);
check_equal(vpar_minmax_ancestors_vector([0,0,0,0],min), [1,2,3,4]);

check_equal(vpar_minmax_ancestors_vector([10,5,6,8,0,1,3,10,9,10],max), \
            [10,5,10,10,5,10,10,10,9,10], "vpar_minmax_ancestors_vector() max");
check_equal(vpar_minmax_ancestors_vector([8,8,10,5,4,7,0,5,4,2],max), \
            [8,8,10,5,5,7,7,8,9,10], "vpar_minmax_ancestors_vector() max");

\\ cyclic allowed
check_equal(vpar_is_cyclic([2,1]), 1);
check_equal(vpar_minmax_ancestors_vector([2,1],min), [1,1]);
check_equal(vpar_minmax_ancestors_vector([2,1],max), [2,2]);

\\ cycle other than 7,8
check_equal(vpar_is_cyclic([6,1,2,3,4,5,1,7]),1);
check_equal(vpar_minmax_ancestors_vector([6,1,2,3,4,5,1,7],max), \
            [6,6,6,6,6,6,7,8]);
check_equal(vpar_minmax_ancestors_vector([6,1,2,3,4,5,1,7],min), \
            [1,1,1,1,1,1,1,1]);

{
  my(funcs=[min,max]);
  for(f=1,#funcs,
     my(func=funcs[f]);
     for(i=1,#test_sample_forests_and_cyclics,
        my(vpar=test_sample_forests_and_cyclics[i]);

        check_equal(vpar_minmax_ancestors_vector(vpar,func),
                 vector(#vpar,v, vpar_minmax_ancestor_of_vertex(vpar,v,func)),
                    "vpar_minmax_ancestors_vector() vs individual vpar_minmax_ancestor_of_vertex()");
     ));
}


\\-----------------------------------------------------------------------------
progress("vpar_minmax_descendants_vector()");

\\ 1 <- 2
check_equal(vpar_minmax_descendants_vector([0,1],min), [1,2], \
            "vpar_minmax_descendants_vector() path-2 reverse, min");
check_equal(vpar_minmax_descendants_vector([0,1],max), [2,2], \
            "vpar_minmax_descendants_vector() path-2 reverse, max");

\\ 2 <- 1
check_equal(vpar_minmax_descendants_vector([2,0],min), [1,1], \
            "vpar_minmax_descendants_vector() path-2 reverse, min");
check_equal(vpar_minmax_descendants_vector([2,0],max), [1,2], \
            "vpar_minmax_descendants_vector() path-2 reverse, max");

\\ 3 <- 2 <- 1 <- 4
check_equal(vpar_minmax_descendants_vector([2,3,0,1],min), [1,1,1,4]);

test_minmax_descendants_vector_by_subtree(vpar,func) =
{
  my(vecfunc=if(func==min,vecmin,
                func==max,vecmax,
                error()));
  vector(#vpar,v, vecfunc(vpar_subtree_vertices(vpar,v)));
}
test_minmax_descendants_vector_by_seq(vpar,func) =
{
  \\ Previous code working upwards with a seq.
  if(vpar_is_cyclic(vpar), error("for acyclics only"));

  my(seq=vpar_INTERNAL_seq_upwards(vpar),
     vec=[1..#vpar],
     v,p);
  for(i=1,#seq,
     if((p=vpar[v=seq[i]]), vec[p]=func(vec[p],vec[v])));
  vec;
}

{
  my(funcs=[min,max]);
  for(f=1,#funcs,
     my(func=funcs[f]);

     for(i=1,#test_sample_forests,
        my(vpar=test_sample_forests[i]);
        check_equal(vpar_minmax_descendants_vector(vpar,func),
                    test_minmax_descendants_vector_by_seq(vpar,func),
                    "vpar_minmax_descendants_vector() vs test_minmax_descendants_vector_by_seq()"));

     for(i=1,#test_sample_forests_and_cyclics,
        my(vpar=test_sample_forests_and_cyclics[i]);

        check_equal(vpar_minmax_descendants_vector(vpar,func),
                    vector(#vpar,v, vpar_minmax_descendant_of_vertex(vpar,v,func)),
                    "vpar_minmax_descendants_vector() vs individual vpar_minmax_descendant_of_vertex()");

        check_equal(vpar_minmax_descendants_vector(vpar,func),
                    test_minmax_descendants_vector_by_subtree(vpar,func),
                    "vpar_minmax_descendants_vector() vs test_minmax_descendants_vector_by_subtree()"));
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_subtree_sizes_vector()");

check_equal(vpar_subtree_sizes_vector([]), [], \
            "vpar_subtree_sizes_vector() empty");
check_equal(vpar_subtree_sizes_vector([0]), [1], \
            "vpar_subtree_sizes_vector() singleton");
check_equal(vpar_subtree_sizes_vector([1]), [1], \
            "vpar_subtree_sizes_vector() self-loop");
check_equal(vpar_subtree_sizes_vector([2,1]), [2,2]);
check_equal(vpar_subtree_sizes_vector([2,3,1,1]), [4,4,4,1]);

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        got=vpar_subtree_sizes_vector(vpar));
     check_equal(got,
                 vector(#vpar,v, vpar_subtree_size(vpar,v)),
       "vpar_subtree_sizes_vector() vs individual vpar_subtree_size()"));
}

test_subtree_sizes_vector_by_seq(vpar) =
{
  \\ Previous code working upwards using a seq, for acyclics only.
  if(vpar_is_cyclic(vpar), error("for acyclics only"));

  my(seq=vpar_INTERNAL_seq_upwards(vpar),
     sizes=vector(#vpar,v,1));
  for(i=1,#seq,
     my(v=seq[i], p=vpar[v]);
     if(p, sizes[p]+=sizes[v]));
  sizes;
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     check_equal(vpar_subtree_sizes_vector(vpar),
                 test_subtree_sizes_vector_by_seq(vpar),
       "vpar_subtree_sizes_vector() vs test_subtree_sizes_vector_by_seq()"));
}



\\ Return 1 if vpar has same-size sibling subtrees in sets of at most 2.
\\ So a vertex can have 1 other sibling of same subtree size, but not 2 or
\\ more others.
test_is_max_two_same_size_subtrees(vpar) =
{
  my(subtree_sizes=vpar_subtree_sizes_vector(vpar),
     seq=vecsort(subtree_sizes,,1),
     prev_size=vectorsmall(#vpar+1),
     count =vectorsmall(#vpar+1));
  for(i=1,#seq,
     my(v=seq[i],
        p=vpar[v]);
     if(prev_size[p+1] == subtree_sizes[v],
        count[p+1]++;
        count[p+1]<=2 || return(0);
        ,
        prev_size[p+1] = subtree_sizes[v];
        count[p+1]=1);
     );
  1;
}

{
  \\ Abelian automorphisms group (C2), but 3 same size sibling subtrees.
  \\        1
  \\      / | \
  \\     2  5  8
  \\     |  |  |\
  \\     3  6  9 10
  \\     |  |
  \\     4  7
  my(vpar=[0,1,2,3, 1,5,6, 1,8,8]);
  vpar_validate(vpar);
  check_equal(test_is_max_two_same_size_subtrees(vpar), 0);
  check_equal(vpar_is_automorphisms_le(vpar,2,1), 1); \\ Abelian
}

{
  for(n=1,12,
     my(vpar=vpar_unlabelled_first(n),
        count=0);
     until((vpar=vpar_unlabelled_next(vpar))=='none,
           count += test_is_max_two_same_size_subtrees(vpar));
     check_equal(count, A213920_samples[n+1],
                 "A213920 same size sibling subtrees in sets of at most 2"));
}

{
  \\ A000435 normalized height, is total subtree sizes
  my(forest=1);
  for(n=0, test_limit_trees_search - forest,
     my(total=0);
     forvec(vec=vpar_forvec_from(n,forest), \\ forests
           my(vpar=vpar_from_upascend(vec));
           total += vecsum(vpar_subtree_sizes_vector(vpar)));
     check_equal(total, A000435_samples[n+1],
                "A000435 subtree sizes total, forests"));
}
{
  \\ A001865 connected functions on n nodes, is total subtree sizes
  \\   sum(k=1,n, n!/(n-k)! * n^(n-k-1))
  \\   1/n * sum(k=1,n, binomial(n,k) * (n-k)^(n-k) * k^k)  \\ Paul D. Hanna
  \\   A000435(n) + n^(n-1)
  my(forest=-1);
  for(n=1, test_limit_trees_search - forest,
     my(total=0);
     forvec(vec=vpar_forvec_from(n,forest), \\ root=1
           my(vpar=vpar_from_upascend(vec));
           total += vecsum(vpar_subtree_sizes_vector(vpar)));
     check_equal(total, A001865_samples[n],
                 "A001865 subtree sizes total, root=1"));
}
{
  \\ A063169 = n*A001865, total subtree sizes
  \\ where
  my(forest=0);
  for(n=1, test_limit_trees_search - forest,
     my(total=0);
     forvec(vec=vpar_forvec_from(n,forest), \\ trees
           my(vpar=vpar_from_upascend(vec));
           total += vecsum(vpar_subtree_sizes_vector(vpar)));
     check_equal(total, A063169_samples[n],
                "A063169 subtree sizes total, trees");
  );
}


\\ sqrt(3)-sqrt(2)
\\ x - (sqrt(2)+sqrt(3))
\\ polroots((x-sqrt(2))^2 - 3)
\\ polroots(x^2 - 2*sqrt(2)*x + 2 - 3)
\\ 2*sqrt(2)*x = x^2 - 1
\\ polroots(8*x^2 - (x^2 - 1)^2)
\\ polroots(x^4 - 10*x^2 + 1)
\\ poldisc(x^4 - 10*x^2 + 1) == 2^14*3^2
\\ poldisc(x^2 - 10*x + 1) == 3*32
\\ 5 + 2*sqrt(6)
\\ sqrt(5 + 2*sqrt(6))

{
  \\ J M Pallo, "Enumerating, Ranking and Unranking Binary Trees", The Computer
  \\ Journal, volume 29, number 2, 1986, pages 171-175.
  \\
  \\ Sample tree, labelled external nodes.
  \\
  \\                  *                        5         8
  \\             /       \                   / | \      / \
  \\          *             *               2  3  4    6   7
  \\       /    \          / \              |
  \\      *      *        *   9             1
  \\     / \    / \      / \
  \\    1   2  3   *    6   *            "natural" correspondence
  \\              / \      / \           postorder labelling
  \\             4   5    7   8
  \\
  \\ v=1  2  3  4  5  6  7  8   (9)
  \\   1, 2, 1, 1, 5, 1, 1, 3   (9)

  my(vpar=[2,5,5,5,0, 8,8,0]);
  vpar_validate(vpar);
  check_equal(vpar_is_postorder(vpar), 1);
  check_equal(vpar_subtree_sizes_vector(vpar), [1,2, 1,1,5, 1,1,3],
              "vpar_subtree_sizes_vector() postorder Pallo weights vector");

  \\ Preorder
  \\      1         6
  \\    / | \      / \
  \\   2  4  5    7   8
  \\   |
  \\   3
  \\
  my(vpar2=[0,1,2,1,1, 0,6,6]);
  vpar_validate(vpar2);
  check_equal(vpar_is_preorder(vpar2), 1);
  check_equal(vpar_relabel_postorder(vpar2), vpar);
}

{
  \\ Ro-Yu Wu, Jou-Ming Chang, Yue-Li Wang, "A Linear Time Algorithm for
  \\ Binary Tree Sequences Transformation Using Left-arm and Right-arm
  \\ Rotations"
  \\
  \\                  6                       6         9
  \\             /       \                   / \       / \
  \\          2             9               2   5     7   8
  \\       /    \          / \             /   / \
  \\      1      5        7   *           1   3   4
  \\     / \    / \      / \
  \\    *   *  3   *    *   8          "natural" correspondence
  \\          / \          / \
  \\         *   4        *   *
  \\            / \
  \\           *   *


  my(vpar=[2,6,5,5,6,0, 9,9,0]);
  vpar_validate(vpar);
  check_equal(vpar_is_postorder(vpar), 1);
  check_equal(vpar_subtree_sizes_vector(vpar), [1,2, 1,1,3,6, 1,1,3],
              "vpar_subtree_sizes_vector() postorder Wu et al");

  \\ flipped
  \\                  6                        6      2    1
  \\             /       \                    / \    / \
  \\          9              2               9   7  5   3
  \\         / \           /   \                 |      |
  \\        *   7        5       1               8      4
  \\           / \      / \     / \
  \\          8   *    *   3   *   *       natural correspondence
  \\         / \          / \
  \\        *   *        4   *
  \\

  my(vpar2=[0, 0,2,3,2, 0,6,7,6]);
  vpar_validate(vpar2);
  check_equal(vpar_is_preorder(vpar2), 1);
  check_equal(vpar_subtree_sizes_vector(vpar2),
              [1,4,2, 1,1,4,2, 1,1],
              "vpar_subtree_sizes_vector() RW Wu et al");

  check_equal(vpar2, vpar_transpose(vpar));

  \\ print(vpar2);
  \\ print(vpar_transpose(vpar));
  \\ print(vpar_relabel_preorder(vpar_transpose(vpar_relabel_preorder(vpar))));
}


\\-----------------------------------------------------------------------------
progress("vpar_subtree_size()");

check_equal(vpar_subtree_size([0],1), 1, "vpar_subtree_size() singleton");
check_equal(vpar_subtree_size([0,0],1), 1, "vpar_subtree_size() 2 singletons");
check_equal(vpar_subtree_size([0,1],2), 1, "vpar_subtree_size() path-2 v=2");
check_equal(vpar_subtree_size([0,1],1), 2, "vpar_subtree_size() path-2 v=1");

check_equal(vpar_subtree_size(vpar_make_path(10),4), 7);   \\ middle
check_equal(vpar_subtree_size(vpar_make_path(10),1), 10);  \\ its root
check_equal(vpar_subtree_size([0,1,0,3],1), 2);  \\ forest
check_equal(vpar_subtree_size([0,1,0,3],2), 1);
check_equal(vpar_subtree_size([0,1,0,3],3), 2);
check_equal(vpar_subtree_size([0,1,0,3],4), 1);

\\ cyclics
check_equal(vpar_subtree_size([1],1), 1, "vpar_subtree_size() self-loop");
check_equal(vpar_subtree_size([2,1,1,1,4],1), 5);
check_equal(vpar_subtree_size([2,1,1,1,4],3), 1);
check_equal(vpar_subtree_size([2,1,1,1,4],4), 2);

test_subtree_size_by_upwards(vpar,v) =
{
  \\ Previous implementation of vpar_subtree_sizes().
  \\ Can be done with just one temp vector of info instead of a seq.
  if(vpar_is_cyclic(vpar), error("for acyclics only"));

  my(seq=vpar_INTERNAL_seq_upwards(vpar),
     sizes=vectorsmall(#vpar,v,1));
  for(i=1,#seq,
     my(c=seq[i]);
     if(c==v,break);
     if(vpar[c], sizes[vpar[c]]+=sizes[c]));
  sizes[v];
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     check_equal(vector(#vpar,v, vpar_subtree_size(vpar,v)),
                 vector(#vpar,v, test_subtree_size_by_upwards(vpar,v)),
                 "vpar_subtree_size() vs test_subtree_size_by_upwards()"));
}


\\-----------------------------------------------------------------------------
progress("vpar_row_vertices_vector()");

check_equal(vpar_row_vertices_vector([]), [], \
            "vpar_row_vertices_vector() empty");
check_equal(vpar_row_vertices_vector([0]), [ [1] ], \
            "vpar_row_vertices_vector() singleton");
check_equal(vpar_row_vertices_vector([0,0]), [ [1,2] ]);

test_row_vertices_vector_by_select(vpar) =
{
  my(depths=vpar_depths_vector(vpar),
     height=vpar_height(vpar));
  vector(height+1,d,
         Vec(select(depth->depth==d-1, depths, 1)));
}
\\ Return the number of self-loops in vpar.
test_selfloops(vpar) = select(v->vpar[v]==v, [1..#vpar]);
check_equal(test_selfloops([]), [], "test_selfloops() empty");
check_equal(test_selfloops([0]), [], "test_selfloops() singleton");
check_equal(test_selfloops([1]), [1], "test_selfloops() self-loop");
check_equal(test_selfloops([2,1,1,4]), [4]);

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        got=vpar_row_vertices_vector(vpar));

     check_equal(sum(i=1,#got, #got[i]), #vpar,
                 "vpar_row_vertices_vector() total length n");

     check_equal(if(#got,got[1],[]),
                 setunion(vpar_roots(vpar), test_selfloops(vpar)),
                 "vpar_row_vertices_vector() first vpar_roots() per addhelp");

     check_equal(apply(length,got), vpar_row_widths_vector(vpar),
          "vpar_row_vertices_vector() lengths vs vpar_row_widths_vector()");
     check_equal(got, test_row_vertices_vector_by_select(vpar),
          "vpar_row_vertices_vector() vs test_row_vertices_vector_by_select()");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_row_width_of_vertex()");

check_equal(vpar_row_width_of_vertex([0],1), 1, \
            "vpar_row_width_of_vertex() singleton");
check_equal(vpar_row_width_of_vertex([1],1), 1, \
            "vpar_row_width_of_vertex() self-loop");

test_row_width_of_vertex_vector_by_row_vertices(vpar) =
{
  \\ Return a vector where vec[v] = row width of the row containing v.
  \\ This is not the same as vpar_row_widths_vector() (it is width at each
  \\ successive depth, not successive vertex).

  my(rows=vpar_row_vertices_vector(vpar),
     ret=vector(#vpar));
  for(r=1,#rows,
     my(width=#rows[r]);
     for(i=1,width,
        ret[rows[r][i]] = width));
  ret;
}
{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        got=vpar_row_widths_vector(Vecsmall(vpar)));
     check_equal(vector(#vpar,v, vpar_row_width_of_vertex(vpar,v)),
                 test_row_width_of_vertex_vector_by_row_vertices(vpar),
                 "vpar_row_width_of_vertex() vs test_row_width_of_vertex_vector_by_row_vertices()");
  );
}

\\-----------------------------------------------------------------------------
progress("vpar_row_widths_vector()");

check_equal(vpar_row_widths_vector([]),[], "vpar_row_widths_vector() empty");

test_row_widths_vector_by_depths(vpar) =
{
  \\ The same approach as current vpar_row_widths_vector(), but here with
  \\ row_widths[] vector built explicitly rather than via
  \\ vpar_INTERNAL_histogram().

  my(depths=vpar_depths_vector(vpar),
     height=vpar_height(vpar),
     row_widths=vector(height+1));
  for(v=1,#vpar,
     row_widths[depths[v]+1]++);
  row_widths;
}
{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        got=vpar_row_widths_vector(Vecsmall(vpar)));
     check_equal(got,
                 test_row_widths_vector_by_depths(vpar),
           "vpar_row_widths_vector() vs test_row_widths_vector_by_depths()");

     \\ Row width * depth is its contribution to sum subtree sizes,
     \\ including for cyclics.
     my(row_widths=vpar_row_widths_vector(Vecsmall(vpar)));
     check_equal(sum(i=1,#row_widths, i*row_widths[i]),
                 vecsum(vpar_subtree_sizes_vector(vpar)),
     "vpar_row_widths_vector() for vecsum of vpar_subtree_sizes_vector()" ));
}


\\-----------------------------------------------------------------------------
progress("vpar_rowpos_of_vertex()");

check_equal(vpar_rowpos_of_vertex([0,3,0],1), 0, \
            "vpar_rowpos_of_vertex() singleton and path-2 1");
check_equal(vpar_rowpos_of_vertex([0,3,0],2), 0, \
            "vpar_rowpos_of_vertex() singleton and path-2 2");
check_equal(vpar_rowpos_of_vertex([0,3,0],3), 1, \
            "vpar_rowpos_of_vertex() singleton and path-2 3");

check_equal(vpar_rowpos_of_vertex([0,3,0],1,1), 1, \
            "vpar_rowpos_of_vertex() singleton and path-2 1 down");
check_equal(vpar_rowpos_of_vertex([0,3,0],2,1), 0, \
            "vpar_rowpos_of_vertex() singleton and path-2 2 down");
check_equal(vpar_rowpos_of_vertex([0,3,0],3,1), 0, \
            "vpar_rowpos_of_vertex() singleton and path-2 3 down");

check_equal(vpar_rowpos_of_vertex([3,0,0],1), 0, \
            "vpar_rowpos_of_vertex() path-2 and singleton 1");
check_equal(vpar_rowpos_of_vertex([3,0,0],2), 0, \
            "vpar_rowpos_of_vertex() path-2 and singleton 2");
check_equal(vpar_rowpos_of_vertex([3,0,0],3), 1, \
            "vpar_rowpos_of_vertex() path-2 and singleton 3");

check_equal(vpar_rowpos_of_vertex([2,3,4,0,0,0],6), 2, \
            "vpar_rowpos_of_vertex() path-4 and singletons, v=6");
check_equal(vpar_rowpos_of_vertex([2,3,4,0,0,0],3), 0, \
            "vpar_rowpos_of_vertex() path-4 and singletons, v=3");

check_equal(vpar_rowpos_of_vertex([0],1), 0, "vpar_rowpos_of_vertex() singleton");
check_equal(vpar_rowpos_of_vertex([0,1],1), 0, "vpar_rowpos_of_vertex() path-2");
check_equal(vpar_rowpos_of_vertex([0,1],2), 0, "vpar_rowpos_of_vertex() path-2");
check_equal(vpar_rowpos_of_vertex([2,0],1), 0, "vpar_rowpos_of_vertex() path-2 rev 1");
check_equal(vpar_rowpos_of_vertex([2,0],2), 0, "vpar_rowpos_of_vertex() path-2 rev 2");

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     for(v=1,#vpar,
        check_equal(vpar_row_width_of_vertex(vpar,v),
                    1 + sum(down=0,1, vpar_rowpos_of_vertex(vpar,v,down)),
                    "vpar_rowpos_of_vertex() up and down width per addhelp")));
}

test_rowpos_of_vertex_by_search(vpar,v,down=0) =
{
  \\ Previous code with its own search up loop, for forests only.
  if(vpar_is_cyclic(vpar), error("for acyclics only"));

  \\ seen[i] = 0 when vertex i has not yet had its depth established.
  \\ seen[i] > 0 is depth of vertex i.
  \\ v_depth is depth of v, set to #vpar so all depths >0.
  \\ r_depth is depth of root, established by working upwards from v.
  \\ In the loop, if seen[i] not yet established then search upwards from i
  \\ for either a seen[] depth or the root (which is r_depth).  Re-traverse
  \\ upwards to set depths based on that top.
  \\ Having established seen[i] can count how many of i<v are seen[i]==v_depth
  \\ so in the row of v and before v.
  \\ The effect is vpar_depths_vector() but run only far enough to find
  \\ depths of vertices <v, and with an arbitrary offset to all depth values.
  \\
  \\ ENHANCE-ME: Maybe cyclics could be allowed by reckoning a row as all of
  \\ given depth, under the depth = num proper ancestors definition.
  \\ vpar_rowpos_vector() is already so because it uses vpar_depths_vector().

  \\ print("vpar_rowpos_of_vertex() "vpar"  v="v);

  my(seen=vectorsmall(#vpar),
     ret=0,
     v_depth = #vpar,
     r_depth = v_depth+1,
     p=v);
  \\ print(" v_depth "v_depth);

  until(!(p=vpar[p]), seen[p]=r_depth--);
  \\ print(" initial seen "Vec(seen));
  \\ print(" r_depth "r_depth);

  my(step=if(down,-1,1));
  forstep(i=if(down,#vpar,1), v-step, step,
     if(!seen[i],
        \\ print(" i="i);
        p=i;
        my(d = r_depth+1); \\ +1 ready for pre-decrement
        while(p=vpar[p],
              d++;
              if(seen[p],
                 \\ print("  found seen["p"]="seen[p]" at d="d);
                 d += seen[p] - r_depth;   \\ change so relative to seen[p]
                 break));
        \\ print("  d="d);
        p=i;
        until(!(p=vpar[p]) || seen[p],
              \\ print("  set seen["p"] = "d);
              seen[p] = d--));
     ret += (seen[i]==v_depth));

  \\ print(" final seen "Vec(seen));
  \\ print(" ret "ret);
  ret;
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     for(v=1,#vpar,
        for(down=0,1,
           check_equal(vpar_rowpos_of_vertex(vpar,v,down),
                       test_rowpos_of_vertex_by_search(vpar,v,down),
                       "vpar_rowpos_of_vertex() vs test_rowpos_of_vertex_by_search()"))));
}


\\-----------------------------------------------------------------------------
progress("vpar_rowpos_vector()");

check_equal(vpar_rowpos_vector([]), [], "vpar_rowpos_vector() empty");
check_equal(vpar_rowpos_vector([0]), [0], "vpar_rowpos_vector() singleton");

test_rowpos_vector_by_row_vertices(vpar,down) =
{
  my(rows=vpar_row_vertices_vector(vpar),
     ret=vector(#vpar));
  for(i=1,#rows,
     my(width=#rows[i]);
     for(j=1,width,
        ret[rows[i][j]] = if(down, width-j, j-1)));
  ret;
}
{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     \\ print(vpar);
     for(down=0,1,
        my(got=vpar_rowpos_vector(Vecsmall(vpar), down));
        check_equal(got,
                    test_rowpos_vector_by_row_vertices(vpar,down),
          "vpar_rowpos_vector() vs test_rowpos_vector_by_row_vertices()");
        check_equal(got,
                    vector(#vpar,v, vpar_rowpos_of_vertex(vpar,v,down)),
          "vpar_rowpos_vector() vs individual vpar_rowpos_of_vertex()"));
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_is_ancestor()");

check_equal(vpar_is_ancestor([0,1],1,2), 0, "vpar_is_ancestor()");
check_equal(vpar_is_ancestor([0,1],2,1), 1);
check_equal(vpar_is_ancestor([0,1],1,1), 1);

{
  \\ For forests vpar_is_ancestor(vpar,v,ancestor) is equivalent to
  \\
  \\   vpar_common_ancestor(vpar,v,ancestor) == ancestor
  \\
  \\ But if v,ancestor are in a cycle then the shortest distance rule in
  \\ vpar_common_ancestor() means it could be either one of them which is
  \\ returned by vpar_common_ancestor().

  \\ v-----------------+
  \\ 3 -> 1 -> 2 -> 7 -+
  \\
  my(vpar=[2, 7, 1, 6, 2, 6, 3, 6],
     v=1, ancestor=3);
  check_equal(vpar_common_ancestor(vpar,v,ancestor), 1);
  check_equal(vpar_is_ancestor(vpar,v,ancestor), 1);
}

\\ safe to cyclics
test_is_ancestor_by_search(vpar,v,ancestor) =
{
  for(i=1,#vpar,
     if(v==ancestor,return(1));
     (v=vpar[v]) || break);
  0;
}
{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     for(rep=1,5,
        my(u=vpar_random_vertex(vpar),
           v=vpar_random_vertex(vpar));
        u || next();

        \\ print(vpar" "u" "v);
        check_equal(vpar_is_ancestor(vpar,u,v),
                    test_is_ancestor_by_search(vpar,u,v),
                    "vpar_is_ancestor() vs test_is_ancestor_by_search()")));
}


\\-----------------------------------------------------------------------------
progress("minmax index calculations");

{
  \\ min = upwards
  \\ max = downwards
  my(ses = (vpar,func) ->
     my(start= func(1,#vpar),
        end  = #vpar+1-start,
        step = -func(1,-1));
     [start,end,step]);

  check_equal(ses([0,1,2,3],min), [1,4,1]);
  check_equal(ses([0],min),       [1,1,1]);
  check_equal(ses([],min),        [0,1,1]);  \\ not right on empty vpar

  check_equal(ses([0,1,2,3],max), [4,1,-1]);
  check_equal(ses([0],max),       [1,1,-1]);
  check_equal(ses([],max),        [1,0,-1]);  \\ not right on empty vpar
}


\\-----------------------------------------------------------------------------
progress("vpar_minmax_descendant_of_vertex()");

{
  check_equal(iferr(vpar_minmax_descendant_of_vertex([0,1,2],1,min);'ok, e, 'err),
              'ok, "vpar_minmax_descendant_of_vertex() error detection");
  check_equal(iferr(vpar_minmax_descendant_of_vertex([0,1,2],0,min);'ok, e, 'err),
              'err, "vpar_minmax_descendant_of_vertex() error detection");
  check_equal(iferr(vpar_minmax_descendant_of_vertex([0,1,2],4,max);'ok, e, 'err),
              'err, "vpar_minmax_descendant_of_vertex() error detection");
}

test_minmax_descendant_of_vertex_by_ancestor(vpar,v,func) =
{
  \\ smallest or biggest vertex i which has v as ancestor
  if(func==min,
     for(i=1,v, if(vpar_is_ancestor(vpar,i,v), return(i))),
     func==max,
     forstep(i=#vpar,v,-1, if(vpar_is_ancestor(vpar,i,v), return(i))));
  error("oops");
}

{
  my(funcs=[min,max]);
  for(f=1,#funcs,
     my(func=funcs[f]);
     for(i=1,#test_sample_forests_and_cyclics,
        my(vpar=test_sample_forests_and_cyclics[i]);
        for(v=1,#vpar,
           check_equal(vpar_minmax_descendant_of_vertex(vpar,v,func),
                       test_minmax_descendant_of_vertex_by_ancestor(vpar,v,func),
                       "vpar_minmax_descendant_of_vertex() vs test_minmax_descendant_of_vertex_by_ancestor()"))));
}


\\-----------------------------------------------------------------------------
progress("vpar_subtree_vertices()");

check_equal(vpar_subtree_vertices([0],1), [1], \
            "vpar_subtree_vertices() singleton");
check_equal(vecsort(vpar_subtree_vertices(vpar_make_path(10),4)), \
            [4,5,6,7,8,9,10]);
check_equal(vpar_subtree_size(vpar_make_path(10),4), 7);   \\ middle

\\ cyclics
check_equal(vpar_subtree_vertices([1,2],1), [1]);
check_equal(vpar_subtree_vertices([1,2],2), [2]);
check_equal(vpar_subtree_vertices([2,1],2), [2,1]);
check_equal(vpar_subtree_vertices([2,1,1,1,1],2), [2,5,4,3,1]);

{
  my(vpar=vpar_make_star(5),
     root=vpar_roots(vpar)[1]);
  check_equal(vpar_subtree_vertices(vpar,root), [1,5,4,3,2]);
  check_equal(vpar_subtree_vertices(vpar,5), [5]);
  check_equal(vpar_subtree_size(vpar,5), 1);
  check_equal(vpar_subtree_sizes_vector(vpar)[5], 1);
}
check_equal(vpar_subtree_sizes_vector([]), []);
check_equal(vpar_subtree_sizes_vector([0]), [1]);

test_subtree_vertices_by_children_1(vpar,v) =
{
  if(vpar_is_cyclic(vpar), error("for acyclics only"));
  my(children=vpar_children_vector(vpar),
     subtree_vertices=List([v]),
     pending=List([v]));
  while(#pending,
        v=pending[#pending]; listpop(pending);
        my(c=children[v]);
        for(i=1,#c,
           listput(subtree_vertices,c[i]);
           listput(pending,c[i])));
  Vec(subtree_vertices);
}
test_subtree_vertices_by_children_2(vpar,v) =
{
  if(vpar_is_cyclic(vpar), error("for acyclics only"));
  my(l=List([]),
     pending=List([v]));
  while(#pending,
        my(v=pending[#pending]);
        listpop(pending);
        listput(l,v);
        my(c=vpar_children_of_vertex(vpar,v));
        for(i=1,#c, listput(pending,c[i])));
  Vec(l);
}
test_subtree_vertices_by_tail(vpar,v) =
{
  \\ Old code:
  \\
  \\ chain[] is a linked list of vertices in the subtree below each v.
  \\ The vertices are v, then chain[v], then chain[chain[v]], etc until
  \\ reaching a chain[v_last]=0.
  \\ tail[v]=v_last is the location of that final entry for v list.
  \\ At a child c its list is appended to the list at its parent p by
  \\ changing parent chain[tail[p]] and updating tail[p] to the child's
  \\ tail.  On reaching the desired v the chain is turned into a Vec.

  my(seq   = vpar_INTERNAL_seq_upwards(vpar),
     chain = vectorsmall(#vpar),
     tail  = vpar_INTERNAL_identity_perm(#vpar));
  for(i=1,#seq,
     my(c=seq[i]);
     if(c==v,break());
     my(p=vpar[c]);
     if(p,
        chain[tail[p]]=c;
        tail[p]=tail[c]);
  );
  \\ use tail now as a counter, postion of previous store in seq
  tail=0; while(v, seq[tail++]=v;v=chain[v]);  \\ pre-increment
  Vec(seq,tail);
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     if(#vpar,
        for(vrep=1,3,
           my(v=vpar_random_vertex(vpar));
           check_equal(vecsort(vpar_subtree_vertices(vpar,v)),
                       vecsort(test_subtree_vertices_by_children_1(vpar,v)),
                       "vpar_subtree_vertices() vs test_subtree_vertices_by_children_1()");
           check_equal(vecsort(vpar_subtree_vertices(vpar,v)),
                       vecsort(test_subtree_vertices_by_children_2(vpar,v)));
           check_equal(vecsort(vpar_subtree_vertices(vpar,v)),
                       vecsort(test_subtree_vertices_by_tail(vpar,v)));
        )));
}

test_subtree_vertices_by_ancestor(vpar,v) = \
  test_select_vertices(vpar,(vpar,i)->vpar_is_ancestor(vpar,i,v));

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     if(#vpar,
        for(vrep=1,3,
           my(v=vpar_random_vertex(vpar));
           check_equal(vecsort(vpar_subtree_vertices(vpar,v)),
                       vecsort(test_subtree_vertices_by_ancestor(vpar,v)));
        )));
}


\\-----------------------------------------------------------------------------
progress("vpar_siblingpos_of_vertex()");
\\ vpar_siblingpos_vector()

check_equal(vpar_siblingpos_vector([0]), [0], \
            "vpar_siblingpos_vector() singleton");
check_equal(vpar_siblingpos_of_vertex([0],1), 0, \
            "vpar_siblingpos_of_vertex() singleton");
check_equal(vpar_siblingpos_of_vertex([1],1), 0, \
            "vpar_siblingpos_of_vertex() self-loop");
check_equal(vpar_siblingpos_vector(vpar_make_path(5)), [0,0,0,0,0]);

check_equal(vpar_siblingpos_vector(vpar_make_star(5)), [0,0,1,2,3]);
check_equal(vpar_siblingpos_vector(vpar_make_star(5),1), [0,3,2,1,0]);
check_equal(vpar_siblingpos_vector([1]), [0], \
            "vpar_siblingpos_vector() self-loop");

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     for(down=0,1,
        check_equal(vpar_siblingpos_vector(vpar,down),
                    vector(#vpar,v, vpar_siblingpos_of_vertex(vpar,v,down)),
        "vpar_siblingpos_vector() vs vpar_siblingpos_of_vertex() individual")));
}

\\ Return the total siblingpos of all vertices in all labelled rooted trees
\\ of n vertices.  "forest" can be 1 for total over forests, or forest can
\\ be -1 for total over root=1 trees.
\\
\\ There are n*vpar_count(n,forest) vertices counted so mean siblingpos ->
\\ 1/2.
\\
test_siblingpos_total(n,forest=0) =
{
  if(n==0, 0,
     if(forest==1, n++; forest=-1);  \\ forest same as root=1 at n+1
     1/2*(n-2)*(n-1)/n * vpar_count(n,forest));  \\ tree is n* root=1
}
{
  for(forest=-1,1,
     for(n=1, test_limit_trees_search - forest,
        my(total=0);
        forvec(vec=vpar_forvec_from(n,forest), \\ trees or forests
              my(vpar=vpar_from_upascend(vec),
                 vec=vpar_siblingpos_vector(vpar));
              total += vecsum(vec));
        \\ print(total); \\ / vpar_count(n,forest));
        check_equal(total, test_siblingpos_total(n,forest),
                    Str("vpar_siblingpos_vector() total, forest="forest));
     ));
}

\\ vector(8,n,n--; test_siblingpos_total(n,-1))
{
  \\ A053507 binomial(n-1,2)*n^(n-3)
  check_equal(vector(#A053507_samples,n, test_siblingpos_total(n,-1)),
              A053507_samples);    \\ OFFSET=1
  check_equal(vector(#A053507_samples,n,n--; test_siblingpos_total(n,1)),
              A053507_samples);

  \\ tree is n* root=1
  check_equal(vector(#A053507_samples,n, if(n==0,0, test_siblingpos_total(n)/n)),
              A053507_samples);    \\ OFFSET=1
  \\ vector(8,n,n--; test_siblingpos_total(n,0))
  \\ vector(8,n,n--; n*test_siblingpos_total(n,-1))
  \\ not in OEIS: 3, 48, 750, 12960, 252105
}


\\-----------------------------------------------------------------------------
progress("vpar_siblings_of_vertex()");

check_equal(vpar_siblings_of_vertex([0],1), [1]);
check_equal(vpar_siblings_of_vertex([0,1,1,1],3), [2,3,4]);

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     for(v=1,#vpar,
        my(siblings=vpar_siblings_of_vertex(vpar,v));
        check_equal(#siblings, vpar_num_siblings_of_vertex(vpar,v) + 1,
          "vpar_siblings_of_vertex() count vs vpar_num_siblings_of_vertex()");
        check_equal(siblings, test_select_vertices(vpar,
                                                   (vpar,u)->vpar[u]==vpar[v]),
          "vpar_siblings_of_vertex() vs test_select_vertices() same parent")));
}


\\-----------------------------------------------------------------------------
progress("vpar_seq_siblings()");

check_equal(vpar_seq_siblings([]), Vecsmall([]), "vpar_seq_siblings() empty");
check_equal(vpar_seq_siblings([0,1,1]), Vecsmall([1,2,3]));
check_equal(vpar_seq_siblings([2,0,2]), Vecsmall([1,3,2]));
check_equal(vpar_seq_siblings([3,3, 0, 3]), Vecsmall([1,2,4, 3]));

test_seq_siblings_by_next(vpar) =
{
  my(vec=vpar_next_sibling_vector(vpar),
     seq=vectorsmall(#vpar),
     upto=0);
  for(i=1,#vpar,
     if(vec[i]>=0,
        my(v=i, t);
        until(!(v=t), seq[upto++]=v; t=vec[v]; vec[v]=-1)));
  seq;
}
test_seq_siblings_by_sort(vpar) = \
  vpar_INTERNAL_seq_bucket_sort(vpar_minmax_sibling_vector(vpar,min));
test_seq_siblings_by_concat(vpar) = \
  if(#vpar, Vecsmall(concat(vpar_siblings_sets(vpar))), Vecsmall([]));

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        seq=vpar_seq_siblings(vpar));
     check_equal(seq, test_seq_siblings_by_sort(vpar),
                "vpar_seq_siblings() vs test_seq_siblings_by_sort()");
     check_equal(seq, test_seq_siblings_by_concat(vpar),
                "vpar_seq_siblings() vs test_seq_siblings_by_concat()");
     check_equal(seq, test_seq_siblings_by_next(vpar),
                "vpar_seq_siblings() vs test_seq_siblings_by_next()");
     check_equal(vpar_seq_is_ordered(vpar,seq), 1,
                "vpar_seq_siblings() is vpar_seq_is_ordered()");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_siblings_setnum_vector()");

check_equal(vpar_siblings_setnum_vector([]), []);
check_equal(vpar_siblings_setnum_vector([0,0,0]), [1,1,1]);
check_equal(vpar_siblings_setnum_vector([0,1,1, 8,8,8,8,0]), [1,2,2, 3,3,3,3,1]);

test_sibling_setnum_vector_by_min(vpar) =
{
  my(min_sibling=vpar_minmax_sibling_vector(vpar,min),
     min_to_pos=vector(#vpar),
     upto=0);
  for(v=1,#min_sibling,
     my(m=min_sibling[v]);
     if(!min_to_pos[m], min_to_pos[m]=upto++));
  vecextract(min_to_pos,min_sibling);
}

\\ vec is a vector of integers in the range 1..#vec.
\\ Those integers which do not occur in vec are "collapsed" by mapping those
\\ which do occur to 1..d  (where d is how many distinct integers occur).
\\ The mapping is returned in the form of a vector m[1..#vec].  m[i] = the
\\ collapsed index when i occurs in vec, or m[i]=0 if it does not occur.
test_vector_collapse_down(vec) =
{
  my(seen=vectorsmall(#vec));
  for(i=1,#vec, seen[vec[i]]=1);
  my(ret=vector(#vec),
     upto=0);
  for(i=1,#seen,
     if(seen[i], ret[i] = upto++));
  ret;
}
test_sibling_setnum_vector_by_collapse(vpar) =
{
  my(min_sibling=vpar_minmax_sibling_vector(vpar,min),
     mvec=test_vector_collapse_down(min_sibling));
  apply(m->mvec[m], min_sibling);
}

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     check_equal(vpar_siblings_setnum_vector(vpar),
                 test_sibling_setnum_vector_by_min(vpar));
     check_equal(vpar_siblings_setnum_vector(vpar),
                 test_sibling_setnum_vector_by_collapse(vpar)));
}


\\-----------------------------------------------------------------------------
progress("Baxter Permutations");

{
  \\ A046996, formula adapted from Neil Sloane there
  check_equal(vector(#A046996_samples,n,n++;
                     sum(k=1,n, binomial(n+1,k-1) * binomial(n+1,k) * binomial(n+1,k+1))
                     / ( 2 * binomial(n+1,1) * binomial(n+1,2))),
              A046996_samples);    \\ OFFSET=2

  check_equal(vector(#A046996_samples,n,n++;
                     sum(k=1,n,
                         1
                         * n!
                         / (n-k)!
                         * (n+1)!
                         * (n-1)!
                         / (n-k+1)!
                         / (n-k+2)!
                         / (k-1)!
                         / k!
                         / (k+1)!
                     )),
              A046996_samples);    \\ OFFSET=2

  check_equal(vector(#A046996_samples,n,n++;
                     sum(k=1,n,
                         1
                         * prod(i=n-k+3,n+1, i)
                         * prod(i=n-k+2,n, i)
                         * prod(i=n-k+1,n-1, i)
                         / (k-1)!
                         / k!
                         / (k+1)!
                     )),
              A046996_samples);    \\ OFFSET=2
}


\\-----------------------------------------------------------------------------
progress("vpar_domnum()");
\\ state table

\\ Return a vector [ new_mtype, increment ] which is the new mtype for
\\ p_mtype accumulating v_mtype under it, and increment = 0 or 1 to add to
\\ domnum.
\\
\\ mtype 0=undecided, 1=dominated, 2=include
\\
test_domnum_mtype_add_and_inc(p_mtype,v_mtype) =
{
  my(new_mtype = p_mtype,
     new_inc = 0);

  if(v_mtype==0,  \\ v is undominated, include parent
     new_inc = (p_mtype!=2); \\ count if p not already included
     new_mtype=2;
     ,
     v_mtype==2,  \\ v is included, raise parent from undecided to dominated
     if(p_mtype==0, new_mtype=1));

  [ new_mtype, new_inc ];
}

{
  my(l=List([]));
  for(p=0,2,
     for(v=0,2,
        my(mtype_and_inc=test_domnum_mtype_add_and_inc(p,v),
           new_mtype = mtype_and_inc[1],
           new_inc = mtype_and_inc[2]);
        listput(l, 2*new_mtype + mtype_and_inc[2])));

  if(0,
     for(p=0,2,
        for(v=0,2,
           printf("%2d,",l[3*p+v+1]));
        print()));

  check_equal(vpar_INTERNAL_domnum_table[1..#l], Vecsmall(l));
}
{
  \\ increments
  check_equal(apply(t->t%2, Vec(vpar_INTERNAL_domnum_table)[1..9]),
              [1, 0, 0,
               1, 0, 0,
               0, 0, 0],
              "vpar_INTERNAL_domnum_table increments");
}
{
  \\ new mtypes
  \\ Row mtype=2 included stays as included.
  check_equal(apply(t->t\2, Vec(vpar_INTERNAL_domnum_table)[1..9]),
              [2, 0, 1,
               2, 1, 1,
               2, 2, 2],
              "vpar_INTERNAL_domnum_table new mtypes");
}


\\ Return the increment 0 or 1 to add to domnum if reach state mtype at the
\\ root of a tree.
\\ If mtype is not already dominated then take a child to dominate it.
\\ If have_child==0 then cannot dominate, and return -1.
\\
test_domnum_mtype_top_inc(mtype) = mtype==0;

\\ This form by explicit separate tests of mtype whether a vertex present
\\ etc.
\\
test_domnum_by_separate(vpar) =
{
  my(seq=vpar_INTERNAL_seq_upwards(vpar),
     mtype=vectorsmall(#vpar),  \\ 0=undecided, 1=dominated, 2=include
     domnum=0);
  for(i=1,#seq,
     my(v=seq[i], p=vpar[v]);
     if(mtype[v]==0,  \\ v is undominated, include parent
        if(p,
           domnum+=(mtype[p]!=2); \\ count if p not already included
           mtype[p]=2,
           \\ no parent for v, include v
           domnum++)
        ,
        mtype[v]==2,  \\ v is included, raise parent from undecided to dominated
        if(p && mtype[p]==0, mtype[p]=1)));
  domnum;
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     check_equal(vpar_domnum(vpar),
                 test_domnum_by_separate(vpar),
                 "vpar_domnum() vs test_domnum_by_separate()");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_domnum()");
\\ vpar_domnum_and_count()

check_equal(vpar_domnum([0]), 1, "vpar_domnum() singleton");

check_equal(vpar_domnum([1]), 1, "vpar_domnum() self-loop");
check_equal(vpar_domnum([2,1]), 1, "vpar_domnum() 2-cycle");
check_equal(vpar_domnum([2,3,1]), 1, "vpar_domnum() 3-cycle");
check_equal(vpar_domnum([2,3,4,1]), 2, "vpar_domnum() 4-cycle");
check_equal(vpar_domnum([3,1,2, 1,4]), 2, "vpar_domnum() 3-cycle 2-tail");

{
  my(vpar=[0,1,2,3,4,5,6,3,2,1,10,11,12,13,14]);
  check_equal(vpar_domnum(vpar), 5, "vpar_domnum()");
  check_equal(vpar_indomnum(vpar), 6, "vpar_indomnum()");
  check_equal(vpar_semitotdomnum(vpar), 7, "vpar_semidomnum()");
  check_equal(vpar_totdomnum(vpar), 8, "vpar_totdomnum()");
}

test_domnum_of_acyclic(vpar) =
{
  \\ Previous code, for acyclic only.
  if(vpar_is_cyclic(vpar), error("for acyclics only"));

  my(table=Vecsmall([
                     5, 0, 2,   \\ p mtype = 0 row
                     5, 2, 2,   \\ p       = 1
                     4, 4, 4    \\ p       = 2
                    ]));

  my(seq=vpar_INTERNAL_seq_upwards(vpar),
     state=vectorsmall(#vpar),
     ret=0);
  for(i=1,#seq,
     my(v=seq[i], p=vpar[v],
        t=table[state[v] + if(p,3*state[p]) + 1]);
     if(p, state[p] = t>>1);
     ret += bitand(t,1));
  ret;
}
{
  \\ acyclics
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
       got=vpar_domnum(vpar),
       got_and_count=vpar_domnum_and_count(vpar));
     check_equal(got, got_and_count[1],
                 "vpar_domnum() vs vpar_domnum_and_count() domnum");

     if(#vpar <= test_limit_flags_search,
        check_equal(got,
                    test_domnum_and_count_by_flags(vpar)[1],
                    "vpar_domnum() vs test_domnum_and_count_by_flags()");
        check_equal(got_and_count,
                    test_domnum_and_count_by_flags(vpar),
                    "vpar_domnum_and_count() vs test_domnum_and_count_by_flags()"));

     check_equal(got,
                 test_domnum_of_acyclic(vpar),
                 "vpar_domnum() vs test_domnum_of_acyclic()"));
}

{
  \\ cyclics
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
       got=vpar_domnum(vpar));

     if(#vpar <= test_limit_flags_search,
        check_equal(got,
                    test_domnum_and_count_by_flags(vpar)[1],
                    "vpar_domnum() vs test_domnum_and_count_by_flags()"));
     );
}

{
  check_equal(vector(50,n,n--; vpar_domnum(vpar_make_binomial_tree(n))),
              vector(50,n,n--; if(n==1,1, floor(n/2))),
              "vpar_domnum() of binomial tree n vertices");
}


\\-----------------------------------------------------------------------------
progress("vpar_indnum()");

check_equal(vpar_indnum([]),   0, "vpar_indnum() empty"); \\ one empty set
check_equal(vpar_indnum([0]),  1, "vpar_indnum() singleton");
check_equal(vpar_indnum([0,1]),1, "vpar_indnum() path-2");
check_equal(vpar_indnum([2,0]),1, "vpar_indnum() path-2 down");

check_equal(vpar_indnum([1]),         0, "vpar_indnum() self-loop");
check_equal(test_indnum_by_flags([1]),0, "test_indnum_by_flags() self-loop");
check_equal(vpar_indnum([2,1]),1, "vpar_indnum() 2-cycle");

\\ Return the maximum independence number occurring for trees without
\\ duplicated leaves.  Without duplicated leaves means no vertex has 2 or
\\ more leaf (degree-1) neighbours.  This is per
\\
\\     Min-Jen Jou, Jenq-Jong Lin, "Independence Numbers in Trees", Open
\\     Journal of Discrete Mathematics, volume 5, 2015, pages 27-31.
\\     http://dx.doi.org/10.4236/ojdm.2015.53003
\\
test_Jou_and_Lin_max_indnum(n) = floor((2*n-1)/3);

{
  \\ n=11 star arms 3,3,3,1, Jou and Lin no duplicated leaves n=12 max indnum
  my(vpar=vpar_join(vpar_subdivide(vpar_make_star(4),2),1,[0]));
  check_equal(#vpar,11);
  check_equal(vpar_indnum(vpar), test_Jou_and_Lin_max_indnum(#vpar));
  check_equal(test_Jou_and_Lin_max_indnum(11), 7);
}
{
  \\ n=11, Jou and Lin no duplicated leaves n=12 max indnum
  \\
  \\     *
  \\    / \
  \\   *   *
  \\  /|   |\
  \\ * *   * *
  \\ |       |
  \\ *       *
  \\ |       |
  \\ *       *
  \\
  \\ Min-Jen Jou, Jenq-Jong Lin, "Independence Numbers in Trees", Open
  \\ Journal of Discrete Mathematics, volume 5, 2015, pages 27-31.
  \\ http://dx.doi.org/10.4236/ojdm.2015.53003

  my(vpar=vpar_make_path(9));
  vpar=vpar_join(vpar,4,[0]);
  vpar=vpar_join(vpar,6,[0]);
  check_equal(#vpar,11);
  check_equal(vpar_indnum(vpar), test_Jou_and_Lin_max_indnum(#vpar));
}

test_indnum_of_acyclic(vpar) =
{
  \\ Previous code, for acyclic only.
  if(vpar_is_cyclic(vpar), error("for acyclics only"));

  my(seq=vpar_INTERNAL_seq_upwards(vpar),
     exclude=vectorsmall(#vpar),
     indnum=0);
  for(i=1,#seq,
     my(v=seq[i]);
     if(!exclude[v],
        indnum++;
        if(vpar[v], exclude[vpar[v]]=1)));
  indnum;
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
       got=vpar_indnum(vpar));
     check_equal(got, test_indnum_of_acyclic(vpar),
                 "vpar_indnum() vs test_indnum_of_acyclic()");

     check_equal(got>=#vpar/2, 1,
                 "vpar_indnum() always >=#vpar/2 per addhelp()");
     check_equal(got==#vpar/2,
                 vpar_has_perfect_matching(vpar),
                 "vpar_indnum() addhelp describing indnum==n/2 equiv vpar_has_perfect_matching()");

     my(component_vpar=vpar_component_trees(vpar));
     check_equal(got,
                 sum(i=1,#component_vpar, vpar_indnum(component_vpar[i])),
                 "vpar_indnum() sum of components, per addhelp");
  );
}

check_equal(vpar_indnum([2,1]), 1);

\\ 1 == 2 <- 3
check_equal(test_indnum_by_flags([2,1, 2]), 2);
check_equal(vpar_indnum([2,1, 2]), 2);

check_equal(test_indnum_by_flags([2,3,4,1, 2,1]), 3);
check_equal(vpar_indnum([2,3,4,1, 2,1]), 3);
\\  +<-------------+
\\  6 -> 3 -> 4 -> 7
\\  |    |         |
\\  1    5         2

check_equal(test_indnum_by_flags([6,7, 4,7, 3,3, 6]), 4);
check_equal(vpar_indnum([6,7, 4,7, 3,3, 6]),          4);

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
       got=vpar_indnum(vpar));

     if(#vpar <= test_limit_flags_search,
        check_equal(got, test_indnum_by_flags(vpar),
                    "vpar_indnum() vs test_indnum_by_flags()"));

     my(component_vpar=vpar_component_trees(vpar));
     check_equal(got,
                 sum(i=1,#component_vpar, vpar_indnum(component_vpar[i])),
                 "vpar_indnum() sum of components, per addhelp");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_is_domset()");

check_equal(vpar_is_domset([],[]), 1, "vpar_is_domset_flags() empty");
check_equal(vpar_is_domset([0,1,2,3,4], [1,3,5]), 1);

test_is_domset_by_bits(vpar,set) =
{
  my(dominated=vectorsmall(#vpar));  \\ bits 1=in set, 2=dom by set
  for(i=1,#set,
     my(v=set[i], p=vpar[v]);
     dominated[v] = bitor(dominated[v],1);
     if(p, dominated[p] = bitor(dominated[p],2)));
  for(v=1,#vpar,
     if(!(dominated[v]
          || (vpar[v] && bitand(dominated[vpar[v]],1))),
        return(0)));
  1;
}
test_is_domset_flags_by_neighbours(vpar,flags) =
{
  my(dominated=vectorsmall(#vpar));  \\ bits 1=in set, 2=dom by set
  for(v=1,#vpar,
     if(flags[v],
        dominated[v]=1;
        ,
        my(neighbours=vpar_neighbours_of_vertex(vpar,v));
        for(i=1,#neighbours,
           if(flags[neighbours[i]], dominated[v]=1; next(2)))));
  hammingweight(dominated) == #vpar;
}

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     for(rep=1,max(1,#vpar),
        my(flags=vpar_random_flags(vpar),
           set=test_flags_to_Set(flags),
           want=test_is_domset_by_bits(vpar,set));
        check_equal(vpar_is_domset_flags(vpar,flags),
                    want,
                    "vpar_is_domset_flags() vs test_is_domset_by_bits()");
        check_equal(vpar_is_domset(vpar,set),
                    want,
                    "vpar_is_domset() vs test_is_domset_by_bits()");
        check_equal(test_is_domset_flags_by_neighbours(vpar,flags),
                    want,
                    "test_is_domset_flags_by_neighbours() vs test_is_domset_by_bits()");
        ));
}


\\-----------------------------------------------------------------------------
progress("test_domnum_and_count_by_flags()");

check_equal(test_domnum_and_count_by_flags(vpar_make_path(2)), [1,2]);
check_equal(test_domnum_and_count_by_flags(vpar_make_path(3)), [1,1]);
check_equal(test_domnum_and_count_by_flags(vpar_make_path(4)), [2,4]);


\\-----------------------------------------------------------------------------
progress("vpar_INTERNAL_domnum_of_path()");

{
  for(n=0,12,
     check_equal(vpar_INTERNAL_domnum_of_path(n),
                 test_domnum_and_count_by_flags(vpar_make_path(n))[1],
                 "vpar_INTERNAL_domnum_of_path() vs test_domnum_and_count_by_flags()"));
}
check_equal(vpar_INTERNAL_domnum_of_path(0), 0, "vpar_INTERNAL_domnum_of_path() 0");
check_equal(vpar_INTERNAL_domnum_of_path(1), 1, "vpar_INTERNAL_domnum_of_path() 1");
check_equal(vpar_INTERNAL_domnum_of_path(2), 1);
check_equal(vpar_INTERNAL_domnum_of_path(3), 1, "vpar_INTERNAL_domnum_of_path() 3");
check_equal(vpar_INTERNAL_domnum_of_path(4), 2);
check_equal(vpar_INTERNAL_domnum_of_path(5), 2);
check_equal(vpar_INTERNAL_domnum_of_path(6), 2);


\\-----------------------------------------------------------------------------
progress("vpar_centres()");

check_equal(vpar_centres([]),  [],  "vpar_centres() empty");
check_equal(vpar_centres([0]), [1], "vpar_centres() singleton");

check_equal(vpar_centres(vpar_make_path(5)),[3]);
check_equal(vpar_centres(vpar_make_path(3)),[2]);
check_equal(vpar_centres(vpar_make_path(7)),[4]);

check_equal(vpar_centres(vpar_make_path(4)),[2,3], "vpar_centres() path-4");
check_equal(vpar_centres(vpar_make_path(6)),[3,4]);
check_equal(vpar_centres(vpar_make_path(8)),[4,5]);

\\ non-adjacent 4 centres
check_equal(vpar_centres([8,1,2,3,4,5,6,7, 1,3,5,7]), [2,4,6,8]);

\\ Not sure how useful an unspecified path.
\\ Maybe something definite like smallest vertex number start, and smallest
\\ vertex number at diameter from there as end.
\\
\\ ecc_path[] could be a linked list instead of creating vectors.
\\
test_diameter_path(vpar) =
{
  if(#vpar,
     my(seq=vpar_INTERNAL_seq_upwards(vpar),
        ecc_path=vector(#vpar,v,[v]), \\ downwards, including self
        diameter_path=[1]);
     for(i=1,#seq,
        my(v=seq[i], p=vpar[v]);
        if(p,
           my(prospective_diameter=#ecc_path[p] + #ecc_path[v]);
           if(prospective_diameter>#diameter_path,
              diameter_path=concat([Vecrev(ecc_path[v]),ecc_path[p]]));
           if(#ecc_path[p]<#ecc_path[v]+1,
              ecc_path[p] = concat(p,ecc_path[v])),
           if(#ecc_path[v]>#diameter_path,
              diameter_path=ecc_path[v])));
     diameter_path;
     ,
     []);  \\ vpar==[]
}
\\ {
\\ addhelp(test_diameter_path,
\\ "vec = test_diameter_path(vpar)
\\ Return a vector of vertices which comprise a diameter of vpar.
\\ There might be multiple paths attaining the diameter.
\\ It is unspecified which one is returned, or which way around its endpoints.");
\\ }

check_equal(test_diameter_path([]), [], "empty");
check_equal(test_diameter_path([0]), [1]);

check_equal(test_diameter_path(Vecsmall(vpar_make_path(5))),[5,4,3,2,1]);
check_equal(test_diameter_path(vpar_make_path(4)),[4,3,2,1]);

check_equal(test_diameter_path([0,1,2,3,4]),[5,4,3,2,1]);

test_centres_by_diameter_path(vpar) =
{
  \\ Could find just two diameter endpoints and which one is deeper, then go
  \\ up diam/2 to the centre, like vpar_centres_of_forest(), or even just
  \\ use vpar_centres_of_forest().

  my(v=test_diameter_path(vpar));
  if(#v,v[(#v+1)>>1 .. (#v+2)>>1],[]);
}

test_centres_by_minimum_eccentricity(vpar) =
{
  my(eccs=vpar_eccentricities_vector(vpar),
     radius=vecmin(concat(#vpar, \\ no empty vecmin
                          vpar_eccentricities_vector(vpar))));
  Vec(select(e->e==radius, eccs, 1));  \\ v with eccs[v]==radius

  \\ my(centres=List([]),
  \\    minecc=#vpar);
  \\ for(v=1,#vpar,
  \\    my(ecc=vpar_eccentricity_of_vertex(vpar,v));
  \\    if(ecc<minecc, minecc=ecc; centres=List([v]),
  \\       ecc==minecc,listput(centres,v)));
  \\ Vec(centres);
}
{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i]);
     check_equal(vecsort(vpar_centres(vpar)),
                 test_centres_by_minimum_eccentricity(vpar),
                 "vpar_centres() vs test_centres_by_minimum_eccentricity()");
     check_equal(vecsort(vpar_centres(vpar)),
                 vecsort(test_centres_by_diameter_path(vpar)),
                 "vpar_centres() vs test_centres_by_diameter_path()");
  );
}
{
  for(i=1,#test_sample_trees_and_unicyclics,
     my(vpar=test_sample_trees_and_unicyclics[i]);
     check_equal(vecsort(vpar_centres(vpar)),
                 test_centres_by_minimum_eccentricity(vpar),
                 "vpar_centres() vs test_centres_by_minimum_eccentricity()");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_indnum() of cycle");

\\ Return indnum of an n-cycle.
test_indnum_of_cycle(n) = n>>1;

{
  for(n=0,12,
     my(want=test_indnum_of_cycle(n),
        vpar=test_make_cycle(n));
     if(#vpar <= test_limit_flags_search,
        check_equal(test_indnum_by_flags(vpar), want,
                    "test_indnum_by_flags() vs test_indnum_of_cycle()"));
     check_equal(vpar_indnum(vpar), want,
                 "vpar_idnnum() vs test_indnum_of_cycle()"));
}
check_equal(test_indnum_of_cycle(1), 0);
check_equal(test_indnum_of_cycle(2), 1);
check_equal(test_indnum_of_cycle(3), 1);
check_equal(test_indnum_of_cycle(4), 2);
check_equal(test_indnum_of_cycle(5), 2);
check_equal(test_indnum_of_cycle(6), 3);

check_equal(test_indnum_of_cycle(2), 1);


\\-----------------------------------------------------------------------------
progress("vpar_indnum() of path");

\\ Return indnum of an n-path.
test_indnum_of_path(n)  = (n+1)>>1;

{
  for(n=0,12,
     my(want=test_indnum_of_path(n),
        vpar=vpar_make_path(n));
     if(#vpar <= test_limit_flags_search,
        check_equal(test_indnum_by_flags(vpar), want,
                    "test_indnum_by_flags() vs test_indnum_of_path()"));
     check_equal(vpar_indnum(vpar), want,
                 "vpar_idnnum() vs test_indnum_of_path()"));
}

check_equal(test_indnum_of_path(1), 1, "test_indnum_of_path() 0");
check_equal(test_indnum_of_path(2), 1);
check_equal(test_indnum_of_path(3), 2, "test_indnum_of_path() 3");
check_equal(test_indnum_of_path(4), 2);
check_equal(test_indnum_of_path(5), 3);
check_equal(test_indnum_of_path(6), 3);


\\-----------------------------------------------------------------------------
progress("vpar_is_indset()");
\\ vpar_is_indset_flags()

check_equal(vpar_is_indset_flags([],[]), 1, "vpar_is_indset() empty");

\\ cyclics
check_equal(vpar_is_indset_flags([1],[]), 1, "vpar_is_indset() empty");
check_equal(vpar_is_indset_flags([0],[1]), 1, "vpar_is_indset() singleton");
check_equal(vpar_is_indset_flags([1],[1]), 0, "vpar_is_indset() self-loop");

check_equal(vpar_is_indset_flags([],[]), 1, "vpar_is_indset_flags() empty");
check_equal(vpar_is_indset_flags([0],[1]), 1, "vpar_is_indset_flags() singleton");
check_equal(vpar_is_indset_flags([0,1,2,3,4],[1,0,1,0,1]), 1);
check_equal(vpar_is_indset_flags([0,1,2,3,4],[1,1,0,0,1]), 0);
check_equal(vpar_is_indset_flags([0,1,2,3,4],[1,0,0,0,0]), 1);
check_equal(vpar_is_indset_flags([0,1,2,3,4],[1]), 1);

\\ cyclics
check_equal(vpar_is_indset_flags([1],[0]), 1);
check_equal(vpar_is_indset_flags([1],[1]), 0, "vpar_is_indset_flags() self-loop");
check_equal(vpar_is_indset_flags([2,1],[1,0]), 1);
check_equal(vpar_is_indset_flags([2,1],[1,1]), 0);

test_is_indset_flags_by_neighbours(vpar,flags) =
{
  for(v=1,#vpar,
     if(flags[v],
        my(neighbours=vpar_neighbours_of_vertex(vpar,v));
        for(i=1,#neighbours,
           if(flags[neighbours[i]], return(0)))));
  1;
}
{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     for(rep=1,#vpar+1,
        my(flags=vpar_random_flags(vpar),
           set=test_flags_to_Set(flags),
           want=test_is_indset_flags_by_neighbours(vpar,flags));
        check_equal(vpar_is_indset(vpar,set),
                    want,
                    "vpar_is_indset() vs test_is_indset_flags_by_neighbours()");
        check_equal(vpar_is_indset_flags(vpar,flags),
                    want,
                    "vpar_is_indset_flags() vs test_is_indset_flags_by_neighbours()")));
}


\\-----------------------------------------------------------------------------
progress("test_sample_trees_and_single_cyclics");

{
  for(i=1,#test_sample_trees_and_unicyclics,
     my(vpar=test_sample_trees_and_unicyclics[i]);
     check_equal(vpar_num_components(vpar) <= 1, 1));
}


\\-----------------------------------------------------------------------------
progress("test_cyclic_roots()");

check_equal(test_cyclic_roots([0]), [1]);
check_equal(test_cyclic_roots([0,0]), test_make_cycle(2));
for(n=3,8, check_equal(test_cyclic_roots(vector(n)), test_make_cycle(n)));


\\-----------------------------------------------------------------------------
progress("vpar_diameter()");
\\ vpar_diameter_and_count()

check_equal(vpar_diameter([]),-1, "vpar_diameter() empty");
check_equal(vpar_diameter([0]),0, "vpar_diameter() single vertex");

check_equal(vpar_diameter([1]),0, "vpar_diameter() self-loop");
check_equal(vpar_diameter([2,1]),1, "vpar_diameter() 2-cycle");

check_equal(vpar_diameter_and_count([]), [-1,1], \
            "vpar_diameter_and_count() empty");
check_equal(vpar_diameter_and_count([0]), [0,1], \
            "vpar_diameter_and_count() singleton");

{
  \\ path
  for(n=0,10,
     my(vpar=vpar_make_path(n));
     if(n>=1,
        check_equal(vpar_degree_of_vertex(vpar,1),if(n==1,0,1)));
     check_equal(vpar_diameter(Vecsmall(vpar)), n-1,
                 Str("vpar_diameter() path "n));
     check_equal(vpar_diameter_and_count(vpar),
                 [n-1, 1],
                 Str("vpar_diameter_and_count() path "n)));
}
{
  \\ star
  for(n=3,10,
     check_equal(vpar_diameter_and_count(vpar_make_star(n)),
                 [2,binomial(n-1,2)],
                 Str("vpar_diameter_and_count() star "n)));
}
{
  \\ cycle
  for(n=1,20,
     my(vpar=test_make_cycle(n),
        want_diameter=n\2);
     check_equal(vpar_diameter(vpar), want_diameter,
                 "vpar_diameter() cycle");

     \\ PENDING
     \\ check_equal(vpar_diameter_and_count_cyclic(vpar),
     \\             [want_diameter,
     \\              if(n%2==0, n/2, n)],
     \\             "vpar_diameter_and_count_cyclic() cycle");
  );
}

test_diameter_of_acyclic(vpar) =
{
  \\ Previous code, for acyclic only.
  if(vpar_is_cyclic(vpar), error("for acyclics only"));

  \\ Code roughly in the manner of
  \\
  \\   S. Mitchell Hedetniemi, E. J. Cockayne and S. T. Hedetniemi, "Linear
  \\   Algorithms for Finding the Jordan Centre and Path Centre of a Tree",
  \\   Transportation Science, volume 15, number 2, May 1981, pages 98-114.
  \\   http://www.jstor.org/stable/25768007
  \\
  \\ height[v] is the height of the subtree at and below v.
  \\ Working upwards from leaves (childless vertices) height[p] is the maximum
  \\ of its children.
  \\ Prospective diameters are between existing height[p] and new child
  \\ height[v], so up and down the children of maximum and second maximum
  \\ (the second maximum possibly being equal).

  my(seq=vpar_INTERNAL_seq_upwards(vpar),
     height   = vectorsmall(#vpar),
     diameter = min(0,#vpar-1));
  for(i=1,#seq,
     my(v=seq[i], p=vpar[v]);
     if(p,
        my(b=height[v]+1,
           prospective=height[p]+b);
        diameter=max(diameter,prospective);
        height[p]=max(height[p],b)));
  diameter;
}

test_diameter_and_count_by_max_eccentricity(vpar) =
{
  \\ Max eccs are diameter endpoints, but only some of them are pairwise
  \\ diameter distance.
  \\ When <=1 max ecc have only one diameter endpoint which is reckoned 1
  \\ empty path.
  my(eccs=vpar_eccentricities_vector(vpar),
     d=vecmax(concat(-1,eccs)),  \\ no empty vecmax
     ends=select(v->eccs[v]==d, [1..#vpar], 1),
     pairs=sum(i=1,#ends, sum(j=i+1,#ends, vpar_path_length(vpar,ends[i],ends[j])==d)));
  [d, if(#ends<=1,1,pairs)];
}

\\ [4, 0, 4, 2, 2] got [3, 2]
\\     2
\\    / \
\\   4   5
\\  / \
\\ 1   3

{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i],
        got=vpar_diameter_and_count(Vecsmall(vpar)));
     \\ print(vpar" got "got);
     check_equal(got[1], test_diameter_of_acyclic(vpar),
                 "vpar_diameter_and_count() vs test_diameter_of_acyclic()");
     check_equal(got[1], vpar_diameter(vpar),
                 "vpar_diameter_and_count() vs vpar_diameter()");
     check_equal(got,
                 test_diameter_and_count_by_max_eccentricity(vpar),
                 "vpar_diameter_and_count() vs test_diameter_and_count_by_max_eccentricity()"));
}

{
  for(i=1,#test_sample_trees_and_unicyclics,
     my(vpar=test_sample_trees_and_unicyclics[i],
        got=vpar_diameter(Vecsmall(vpar)));
     #vpar>0 || next;
     \\ print(vpar" got "got);
     check_equal(got,
                 vecmax(vpar_path_lengths_matrix(vpar)),
                 "vpar_diameter() vs max vpar_path_lengths_matrix()");
     check_equal(got,
                 vecmax(vpar_eccentricities_vector(vpar)),
                 "vpar_diameter() vs max vpar_eccentricities_vector()");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_eccentricities_vector()");

check_equal(vpar_eccentricities_vector([2,3,4,5,6,7,1]), [3,3,3,3,3,3,3], \
            "vpar_eccentricities_vector() 7-cycle");

check_equal(vpar_eccentricities_vector([2,1]), [1,1], \
            "vpar_eccentricities_vector() 2-cycle");

check_equal(vpar_eccentricities_vector([2,3,4,1]), [2,2,2,2], \
            "vpar_eccentricities_vector() 4-cycle");

{
  \\        4
  \\       v  ^
  \\      1 -> 3
  \\      |
  \\      2
  check_equal(vpar_eccentricities_vector([3,1,4,1]), [1,2,2,2],
              "vpar_eccentricities_vector() 3-cycle and leaf");
}

{
  \\      2 <- 4
  \\      v    ^
  \\      1 -> 3
  my(vpar=[3,1,4,2]);
  check_equal(vpar_eccentricities_vector([3,1,4,2]), [2,2,2,2],
              "vpar_eccentricities_vector() 4-cycle jumbled");
  check_equal(vpar_diameter([3,1,4,2]), 2,
              "vpar_diameter() 4-cycle jumbled");
  check_equal(vpar_radius([3,1,4,2]), 2,
              "vpar_radius() 4-cycle jumbled");
}


test_eccentricties_vector_of_acyclic(vpar) = \
{
  \\ Previous code, for acyclic only.
  if(vpar_is_cyclic(vpar), error("for acyclics only"));

  my(seq=vpar_INTERNAL_seq_upwards(vpar),
     ecc       =vector(#vpar),
     ecc_second=vectorsmall(#vpar),
     ecc_above =ecc_second);

  \\ Leaf upwards ecc[]=ecc_below[] and ecc_second[] as the second biggest
  \\ eccentricity below v.
  for(i=1,#seq,
     my(v=seq[i], p=vpar[v]);
     if(p,
        my(b=ecc[v]+1);  \\ accumulate ecc[v] into ecc[p]
        if(b>ecc[p],
           ecc_second[p]=ecc[p];
           ecc[p]=b;
           ,
           ecc_second[p]=max(ecc_second[p],b))));

  \\ Root downwards augment ecc[] with eccentricity above.
  \\ If ecc_below[p]==ecc[v]+1 then v is where ecc_below[p] is attained, so
  \\ want ecc_second[p] for paths other than down through v.
  \\ ecc[p] is overwritten so that it's not ecc_below[p] but is bigger if
  \\ something above is bigger than ecc_below[p].  When that happens
  \\ ecc[p]==ecc[v]+1 does not identify when v is the ecc_below[p].  But
  \\ this doesn't matter because ecc_above[p] is then >= ecc_below[p]+1
  \\ anyway, so "b" is still correct.
  \\
  forstep(i=#seq,1,-1,
         my(v=seq[i], p=vpar[v]);
         if(p,
            my(b=max(ecc_above[p], if(ecc[p]==ecc[v]+1,ecc_second,ecc)[p])
                 + 1);
            ecc_above[v]=b;
            ecc[v]=max(ecc[v],b)));
  ecc;
}

{
  my(vpar=[2, 4, 4, 0]);
  \\ print("vpar ",vpar);
  \\ print("num_children "vpar_num_children_vector(vpar));
  \\ vpar_view(vpar);
  check_equal(vpar_eccentricities_vector(vpar),
              vector(#vpar,v, vpar_eccentricity_of_vertex(vpar,v)),
              "sample 4");
}
{
  my(vpar=[0,1]);
  check_equal(vpar_eccentricities_vector(vpar),
              vector(#vpar,v, vpar_eccentricity_of_vertex(vpar,v)));
}
{
  my(vpar=[0,1,2,3,4,5,6,7]);
  check_equal(vpar_eccentricities_vector(vpar),
              vector(#vpar,v, vpar_eccentricity_of_vertex(vpar,v)),
              "vpar_eccentricities_vector() path-3");
}

{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i],
        got=vpar_eccentricities_vector(vpar));
     check_equal(got,
                 test_eccentricties_vector_of_acyclic(vpar),
                 "vpar_eccentricities_vector() vs test_eccentricties_vector_of_acyclic()"));
}

{
  for(i=1,#test_sample_trees_and_unicyclics,
     my(vpar=test_sample_trees_and_unicyclics[i],
        got=vpar_eccentricities_vector(vpar));
     check_equal(got,
                 vector(#vpar,v, vpar_eccentricity_of_vertex(vpar,v)),
                 "vpar_eccentricities_vector() vs vpar_eccentricity_of_vertex()");
     if(#vpar,
        \\ print(vpar);
        check_equal(vecmax(got), vpar_diameter(vpar),
                    "vpar_eccentricities_vector() maximum is diameter, per addhelp()");
        check_equal(vecmin(got), vpar_radius(vpar),
                    "vpar_eccentricities_vector() minimum is radius, per addhelp()");
        check_equal(got,
                    apply(vecmax,Vec(vpar_path_lengths_matrix(vpar))),
                    "vpar_eccentricities_vector() vs vpar_path_lengths_matrix()"));
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_neighbours_of_vertex()");
\\ vpar_neighbours_vector()

{
  my(vpar=vpar_make_star(5));
  check_equal(vpar_neighbours_of_vertex(vpar,1), [2,3,4,5], "neighbours");
  check_equal(vpar_children_of_vertex(vpar,1), [2,3,4,5], "children");
  check_equal(vpar_children_of_vertex(vpar,2), [], "children");
  check_equal(vpar_neighbours_of_vertex(vpar,2), [1], "neighbours");
  check_equal(vpar_children_vector(vpar), [[2,3,4,5],[],[],[],[]]);
}

\\ cyclic
check_equal(vpar_neighbours_of_vertex([1],1), [1,1], \
            "vpar_neighbours_of_vertex() self-loop");
check_equal(vpar_neighbours_of_vertex([2,1],1), [2,2]);
check_equal(vpar_neighbours_of_vertex([2,3,1],1), [2,3]);

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        vec=vpar_neighbours_vector(vpar));
     check_equal(vec,
                 vector(#vpar,v,vpar_neighbours_of_vertex(vpar,v)),
                 "vpar_neighbours_vector() vs individual vpar_neighbours_of_vertex()"));
}


\\-----------------------------------------------------------------------------
progress("vpar_common_ancestor()");

\\        10
\\       / | \
\\     3   4   9          Knuth 2.3.4.4
\\    / \     / \
\\   1   2   5   8
\\              / \
\\             6   7
{
  my(vpar=[3,3,10,10,9,8,8,9,10, 0]);
  check_equal(vpar_common_ancestor(vpar,1,2),3);
  check_equal(vpar_common_ancestor(vpar,1,3),3);
  check_equal(vpar_common_ancestor(vpar,3,3),3);
  check_equal(vpar_common_ancestor(vpar,2,5),10);
  check_equal(vpar_common_ancestor(vpar,5,7),9);
  check_equal(vpar_common_ancestor(vpar,5,4),10);
  check_equal(vpar_degree_of_vertex(vpar,3),3, "degree");
  check_equal(vpar_degree_of_vertex(vpar,4),1, "degree");
  check_equal(vpar_degree_of_vertex(vpar,10),3);

  vpar=vpar_concat_forest(vpar,vpar);
  check_equal(vpar_common_ancestor(vpar,10,20),0);
  check_equal(vpar_common_ancestor(vpar,1,20),0);
  check_equal(vpar_common_ancestor(vpar,15,14),20);
}
{
  \\ 8-> 6<-5<-4 <- 9
  \\     v     ^
  \\     1->2->3 <- 7
  my(vpar=[2,3,4,5,6,1, 3,6,4]);
  check_equal(vpar_common_ancestor(vpar,1,1), 1);
  check_equal(vpar_common_ancestor(vpar,1,2), 2);
  check_equal(vpar_common_ancestor(vpar,1,3), 3);
  check_equal(vpar_common_ancestor(vpar,1,4), 1);
  check_equal(vpar_common_ancestor(vpar,1,5), 1);
  check_equal(vpar_common_ancestor(vpar,1,6), 1);

  check_equal(vpar_common_ancestor(vpar,7,8), 3);
  check_equal(vpar_common_ancestor(vpar,8,7), 6);

  check_equal(vpar_common_ancestor(vpar,7,9), 4);
  check_equal(vpar_common_ancestor(vpar,9,7), 4);

  check_equal(vpar_common_ancestor(vpar,8,9), 6);
  check_equal(vpar_common_ancestor(vpar,9,8), 6);
}

test_common_ancestor_by_depth(vpar,u,v) =
{
  \\ Previous code, for acyclic.
  if(vpar_is_cyclic(vpar), error("for acyclics only"));

  my(u_depth=vpar_depth_of_vertex(vpar,u),
     v_depth=vpar_depth_of_vertex(vpar,v));
  while(u_depth>v_depth, u=vpar[u]; u_depth--);
  while(v_depth>u_depth, v=vpar[v]; v_depth--);
  while(u&&u!=v, u=vpar[u]; v=vpar[v]);
  u;
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     for(u=1,#vpar,
        for(v=1,#vpar,
           check_equal(vpar_common_ancestor(vpar,u,v),
                       test_common_ancestor_by_depth(vpar,u,v)))));
}

\\ A286778(n) = for complete binary tree of height n,
\\              sum over pairs u,v, the depth of common ancestor of u,v
\\ u,v can have u==v and is both ways u,v and v,u
\\ also
\\ A286778(n) = sum squares of subtree sizes at non-root vertices
\\
A286778_by_common_ancestor(n) =
{
  my(vpar=vpar_make_complete_tree(2,n));  \\ height n
  sum(u=1,#vpar,
        sum(v=1,#vpar,
              my(c=vpar_common_ancestor(vpar,u,v));
              vpar_depth_of_vertex(vpar,c)));
}
A286778_by_subtree_size(n) =
{
  my(vpar=vpar_make_complete_tree(2,n));
  sum(v=2,#vpar, vpar_subtree_size(vpar,v)^2);
}
{
  check_equal(vector(5,n,n--;A286778_by_common_ancestor(n)),
              A286778_samples[1..5]);
  check_equal(vector(5,n,n--;A286778_by_subtree_size(n)),
              A286778_samples[1..5]);
}


\\-----------------------------------------------------------------------------
progress("vpar_path_lengths_vector()");

check_equal(vpar_path_lengths_vector([0],1), [0]);
check_equal(vpar_path_lengths_vector([0,1],1), [0,1]);
check_equal(vpar_path_lengths_vector([0,1],2), [1,0]);
check_equal(vpar_path_lengths_vector([0,0],1), [0,-1]);
check_equal(vpar_path_lengths_vector([0,0,0],1), [0,-1,-1]);
check_equal(vpar_path_lengths_vector([0,1,2],1), [0,1,2]);
check_equal(vpar_path_lengths_vector([0,1,2],2), [1,0,1]);
check_equal(vpar_path_lengths_vector([0,1,2,3,4],5), [4,3,2,1,0]);
check_equal(vpar_path_lengths_vector([0, 0, 2],1), [0,-1,-1]);

check_equal(vpar_path_lengths_vector([0],1), [0]);

\\               /8--7
\\     1--2--3--4    |
\\               \5--6
check_equal(vpar_path_lengths_vector([2,3,4, 5,6,7,8,4],1), [0,1,2,3, 4,5,5,4]);

\\               /9--8\
\\     1--2--3--4      7
\\               \5--6/
check_equal(vpar_path_lengths_vector([2,3,4, 5,6,7,8,9,4],1), [0,1,2,3, 4,5,6,5,4]);

\\               /6
\\     1--2--3--4 |
\\               \5
check_equal(vpar_path_lengths_vector([2,3,4, 5,6,4],1), [0,1,2,3, 4,4]);

\\               /7\
\\     1--2--3--4   6
\\               \5/
check_equal(vpar_path_lengths_vector([2,3,4, 5,6,7,4],1), [0,1,2,3, 4,5,4]);

check_equal(vpar_path_lengths_vector([1],1), [0]);

check_equal(vpar_path_lengths_vector([2,3,0],3), [2,1,0]);

check_equal(vpar_path_lengths_vector([2,1],1), [0,1]);
check_equal(vpar_path_lengths_vector([2,1],2), [1,0]);

test_path_lengths_vector_by_seq(vpar,v) = \
{
  \\ Past code using a seq.
  if(vpar_is_cyclic(vpar), error("for acyclics only"));

  my(seq=vpar_INTERNAL_seq_upwards(vpar),
     distances=vector(#vpar,v,-1));

  \\ upwards from v, set distances[v] for ancestors of v
  my(p=-1); \\ distance above v
  while(distances[v]=p++; (v=vpar[v]), ); \\ pre-increment

  \\ downwards all, set distances[] for descendants of ancestors,
  \\ including descendants of v itself
  forstep(i=#seq,1,-1,
         if(distances[v=seq[i]]==-1,
            if((p=vpar[v]) && distances[p]>=0,
               distances[v]=distances[p]+1)));
  distances;
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     for(v=1,#vpar,
        my(got=vpar_path_lengths_vector(Vecsmall(vpar),v));
        check_equal(got,
                    vector(#vpar,u, vpar_path_length(vpar,u,v)),
                    "vpar_path_lengths_vector() vs individual vpar_path_length()");
        check_equal(got,
                    test_path_lengths_vector_by_seq(vpar,v),
                    "vpar_path_lengths_vector() vs individual test_path_lengths_vector_by_seq()")
     ));
}

test_path_lengths_vector_by_Dijkstra(vpar,v) =
{
  \\ Working outwards from v, setting new vertices encountered to distance d.
  \\ New vertices are where ret[t]==-1 initial value.
  \\ Unreachables remain that initial -1.
  \\ Cycles allowed, since any loop back around is not a new vertex (and the
  \\ already seen way to the vertex is <= this way).

  my(ret=vector(#vpar,i,-1),
     d=0,
     pending=[v]);
  while(#pending,
        for(i=1,#pending, ret[pending[i]]=d);
        d++;
        pending=vpar_neighbours_of_set(vpar,pending);
        pending=select(v->ret[v]<0, pending));  \\ not already seen
  ret;
}
{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     for(v=1,#vpar,
        my(got=vpar_path_lengths_vector(vpar,v));
        check_equal(got, test_path_lengths_vector_by_Dijkstra(vpar,v));
        check_equal(got, vector(#vpar,u,vpar_path_length(vpar,u,v)))));
}


\\-----------------------------------------------------------------------------
progress("vpar_path_length()");

check_equal(vpar_path_length([0],1,1), 0);
check_equal(vpar_path_length([0,0],1,2), -1);
check_equal(vpar_path_length([0,1,0,3],2,4), -1);
check_equal(vpar_path_length([0,1,0,3],2,1), 1);

{
  my(vpar=vpar_make_path(10));
  check_equal(vpar_path_length(vpar,1,10),9);

  \\   1  2  3  4  5
  \\ 3--4--5--6--7--8
  check_equal(vpar_path_length(vpar,3,8),5);
}
{
  \\      1
  \\     v ^
  \\    2 > 3
  \\   /     \
  \\  4       5
  my(vpar=[2,3,1,2,3]);
  check_equal(vpar_path_length(vpar,4,5), 3);
  check_equal(vpar_path_length(vpar,1,3), 1);
}


test_path_length_by_path_vertices(vpar,u,v) = \
  #vpar_path_vertices(vpar,u,v) - 1;

test_path_length_by_old_code(vpar,u,v) =
{
  my(u_depth=vpar_depth_and_root_of_vertex(vpar,u),
     v_depth=vpar_depth_and_root_of_vertex(vpar,v));
  if(u_depth[2]!=v_depth[2],return(-1));
  u_depth=u_depth[1];  \\ just depths now
  v_depth=v_depth[1];
  my(len=0);
  while(u_depth>v_depth, u=vpar[u];u_depth--;len++);
  while(v_depth>u_depth, v=vpar[v];v_depth--;len++);
  while(u&&u!=v, u=vpar[u]; v=vpar[v]; len+=2);
  len;

  \\ Remembering the ancestors seen of u,v rather than going up to the root.
  \\ So worst search up is 2*length.
  \\ if(u==v,return(0));
  \\ my(dist=vectorsmall(#vpar,v,-1),vv=[u,v],dd=[0,0]);
  \\ dist[u]=0; dist[v]=0;
  \\ while(vecsum(vv),
  \\       for(i=1,2,
  \\          if(vv[i],
  \\             \\ print("at i="i"  "vv[i]" "dist);
  \\             vv[i]=vpar[vv[i]];
  \\             \\ print(" parent "vv[i]);
  \\             if(vv[i],
  \\                dd[i]++;
  \\                if(dist[vv[i]]>=0,
  \\                   \\ print("end i="i" vv="vv" dd="dd);
  \\                   \\ print(" dist "dist"  "dist[vv[i]]);
  \\                   return(dist[vv[i]] + dd[i]));
  \\                dist[vv[i]]=dd[i]))));
  \\ vecsum(dd);
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     #vpar>=1 || next();
     for(rep=1,4,
        my(u=vpar_random_vertex(vpar),
           v=vpar_random_vertex(vpar),
           got=vpar_path_length(Vecsmall(vpar),u,v));
        check_equal(got, test_path_length_by_path_vertices(vpar,u,v));
        check_equal(got, test_path_length_by_old_code(vpar,u,v))));
}


\\-----------------------------------------------------------------------------
progress("vpar_depth_and_root_of_vertex()");

check_equal(vpar_depth_and_root_of_vertex([0],1), [0,1]);
check_equal(vpar_depth_and_root_of_vertex([1],1), [0,0]);

\\  4 --- 3
\\  |     |
\\  5 ->6 2  d, d+l
\\        |
\\        1
\\        |
\\        0
\\ u=2=d is v=4
\\ further -d mod l steps to u==v
\\ so u==v at udepth = d + (-d mod l), and vdepth = 2*udepth
\\ l=4,  -2 == 2 mod l, so udepth=4 vdepth=8
\\ d = q*l + r with 0 <= r < l
\\ d mod l = r,   -d mod l = if(r==0,0,l-r)
\\ udepth = d + if(r==0,0,l-r)
\\        = if(r==0, d, d+l-r)
\\ d = udepth   or udepth + r-l
\\
\\
\\ u at d after d steps, v at 2*d
\\ which is v ahead by 2*d-d = d mod l
\\ further
\\
\\ cycle=0 for depth up to root or cycle
\\ cycle=1 for depth up and around the cycle too, so to top non-repeat
\\
test_depth_and_root_of_vertex_by_specific(vpar,v) = \
{
  \\ Previous code.

  \\ u is a search up from v looking for a root, ie. vpar[u]==0.
  \\ h is a half-speed search up.
  \\ If find u==h then they are part of a cycle.
  \\ Cycle length is by working upwards until u==h again.
  \\ The non-cycle length up from v is then found by re-traversing upwards
  \\ with u ahead by the cycle length, so that find u==v at the first
  \\ cyclic.
  \\
  \\ Time taken is linear in the returned depth d and the cycle length c (if
  \\ cyclic).  The first loop has two stopping conditions, either u==h or u
  \\ through #vpar many vertices.  The former catches small d and c, the
  \\ latter caps at #vpar if big c (instead of 2 cycles around big c).
  \\
  \\ The depth d at which u==h found does not say anything about the desired
  \\ depth.  Different depth reach u==h with the same d.  For example
  \\
  \\    5 <-- 4
  \\      \   |       from u=1 reach u=4,h=4 at d=6, desiring depth=2
  \\       -> 3
  \\          |       from u=2 reach u=5,h=5 at d=6, desiring depth=2
  \\          2
  \\          |
  \\          1
  \\
  \\ The d with u==h is a multiple of the cycle length c though.  If the
  \\ desired depth is t, and du=2*dh steps up by u and h, then have
  \\ dh-t == du-t mod c so that u==h, and hence dh==2*dh mod c and
  \\ dh==0 mod c.  What multiple of c is du or dh depends on t, so doesn't
  \\ help to find c.

  \\ print("vpar_depth_and_root_of_vertex() v="v);
  my(u=v, h=v);
  for(d=0,#vpar,
     vpar[u] || return([d,u]);
     u=vpar[u];
     d++;
     vpar[u] || return([d,u]);
     u=vpar[u];
     if((h=vpar[h])==u, break));

  \\ print(" cyclic at u="u" h="h);
  my(c=1);
  h=u; while((u=vpar[u])!=h, c++);
  \\ print(" cycle length c="c);

  u=v; for(i=1,c, u=vpar[u]);      \\ ahead by c steps
  for(d=0,#vpar,
     if(u==v, return([d+c-1,0]));  \\ equal at first cyclic
     u=vpar[u]; v=vpar[v]);
  [0,0];
}

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     for(v=1,#vpar,
        my(got = vpar_depth_and_root_of_vertex(vpar,v));
        \\ print(vpar" "got);
        check_equal(#got, 2, "vpar_depth_and_root_of_vertex() vec length 2");
        check_equal(got[1], vpar_depth_of_vertex(vpar,v));
        check_equal(got[2], if(vpar_is_cyclic_vertex(vpar,v),
                               0, vpar_root_of_vertex(vpar,v)),
                    "vpar_depth_and_root_of_vertex() vs vpar_root_of_vertex()");

        check_equal(got, test_depth_and_root_of_vertex_by_specific(vpar,v),
                    "vpar_depth_and_root_of_vertex() vs test_depth_and_root_of_vertex_by_specific()")));
}


\\-----------------------------------------------------------------------------
progress("vpar_depth_of_vertex()");

check_equal(vpar_depth_of_vertex([0],1), 0);
check_equal(vpar_depth_of_vertex([0,1],1), 0);
check_equal(vpar_depth_of_vertex([0,1],2), 1);

\\ cyclic
check_equal(vpar_depth_of_vertex([1],1), 0, \
            "vpar_depth_of_vertex() self-loop");
check_equal(vpar_depth_of_vertex([2,1],1), 1);
check_equal(vpar_depth_of_vertex([2,1],2), 1);

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     for(v=1,#vpar,
        my(got = vpar_depth_of_vertex(vpar,v));
        check_equal(got==0, vpar[v]==0,
                   "vpar_depth_of_vertex() root has depth==0")));
}

{
  \\ same d example comments in the code
  \\
  \\    5 <-- 4
  \\      \   |       from u=1 reach u=4,h=4 at d=6
  \\       -> 3
  \\          |       from u=1 reach u=5,h=5 at d=6
  \\          2
  \\          |
  \\          1

  my(vpar=[2,3,4,5,3]);
  check_equal(vpar_is_cyclic(vpar), 1);
  check_equal(vpar_num_cycles(vpar), 1);
  check_equal(vpar_num_components(vpar), 1);

  my(u=1); for(d=1,6,   u=vpar[u]);
  my(h=1); for(i=1,6/2, h=vpar[h]);
  check_equal(u,4);
  check_equal(h,4);

  my(u=2); for(d=1,6,   u=vpar[u]);
  my(h=2); for(i=1,6/2, h=vpar[h]);
  check_equal(u,5);
  check_equal(h,5);
}

\\ previous code, cycles not allowed
test_depth_of_vertex_by_acyclic(vpar,v) = \
  for(depth=0,#vpar, (v=vpar[v]) || return(depth)); -1;

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     check_equal(vector(#vpar,v, vpar_depth_of_vertex(vpar,v)),
                 vector(#vpar,v, test_depth_of_vertex_by_acyclic(vpar,v)),
                 "vpar_depth_of_vertex() vs test_depth_of_vertex_by_acyclic()");
  );
}


{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     for(v=1,#vpar,
        check_equal(vpar_depth_of_vertex(vpar,v),
                    sum(i=1,#vpar, i!=v && vpar_is_ancestor(vpar,v,i)),
                    "vpar_depth_of_vertex() vs count vpar_is_ancestor()")));
}


\\-----------------------------------------------------------------------------
progress("vpar_INTERNAL_top_depth_length()");

check_equal(vpar_INTERNAL_top_depth_length([0],1), [1,0,0]);
check_equal(vpar_INTERNAL_top_depth_length([0,1,2,3],4), [1,3,0]);

check_equal(vpar_INTERNAL_top_depth_length([2,3,4,1, 3,5],6), [3,2,4]);
check_equal(vpar_INTERNAL_top_depth_length([2,3,4,1, 3,5],1), [1,0,4]);


\\-----------------------------------------------------------------------------
progress("vpar_INTERNAL_common_ancestor_and_distances()");

check_equal(vpar_INTERNAL_common_ancestor_and_distances([0,1,1],2,3), \
            [1,1,1]);
check_equal(vpar_INTERNAL_common_ancestor_and_distances([0,1,1,3],2,4), \
            [1,1,2]);
check_equal(vpar_INTERNAL_common_ancestor_and_distances([0,1,1,3],1,4), \
            [1,0,2]);
check_equal(vpar_INTERNAL_common_ancestor_and_distances([0,1,1,3],4,1), \
            [1,2,0]);

\\ cyclic
check_equal(vpar_INTERNAL_common_ancestor_and_distances([2,3,4,5,1],1,2), \
            [2,1,0]);
check_equal(vpar_INTERNAL_common_ancestor_and_distances([2,3,4,5,1],2,1), \
            [2,0,1]);

{
  my(vpar=[9, 1, 8, 8, 8, 11, 5, 0, 12, 4, 2, 5]);
  vpar_validate(vpar);
  check_equal(vpar_is_tree(vpar), 1);
  check_equal(vpar_root_of_vertex(vpar,2), 8);
  check_equal(vpar_root_of_vertex(vpar,10), 8);
  check_equal(vpar_depth_of_vertex(vpar,2), 5);
  check_equal(vpar_depth_of_vertex(vpar,10), 2);
  check_equal(vpar_INTERNAL_common_ancestor_and_distances(vpar,2,10),
              [8, 5,2]);
}

test_common_ancestor_and_distances_acyclic(vpar,u,v) =
{
  \\ Previous code, looking only upwards for common under root.
  my(u_depth=vpar_depth_of_vertex(vpar,u),
     v_depth=vpar_depth_of_vertex(vpar,v));
  u_depth -= v_depth;
  v_depth=max(-u_depth,0);    \\ >0 when v deeper than u
  u_depth=max(u_depth,0);     \\ >0 when u deeper than v
  for(i=1,u_depth, u=vpar[u]);
  for(i=1,v_depth, v=vpar[v]);
  while(u&&u!=v,
        u=vpar[u]; v=vpar[v]; u_depth++; v_depth++);
  if(u, [u,u_depth,v_depth],
        [0,0,0]);
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     for(u=1,#vpar,
        for(v=1,#vpar,
           \\ print(vpar" u="u" v="v);
           my(got=vpar_INTERNAL_common_ancestor_and_distances(vpar,u,v));
           if(got[1],
              check_equal(vpar_is_ancestor(vpar,u,got[1]), 1);
              check_equal(vpar_is_ancestor(vpar,v,got[1]), 1));
           check_equal(got,
                       test_common_ancestor_and_distances_acyclic(vpar,u,v),
                       "vpar_INTERNAL_common_ancestor_and_distances() vs test_common_ancestor_and_distances_acyclic()"))));
}

test_common_ancestor_and_distances_cyclic_by_ancestors(vpar,u,v) =
{
  \\ print("test_common_ancestor_and_distances_cyclic_by_ancestors() "vpar" u="u" v="v);
  my(u_ancestors=vpar_ancestors(vpar,u),
     v_ancestors=vpar_ancestors(vpar,v),
     common=setintersect(Set(u_ancestors),Set(v_ancestors)));
  if(#common==0, return([0,0,0]));

  my(u_distances=vector(#vpar));
  for(i=1,#u_ancestors, u_distances[u_ancestors[i]]=i-1);
  my(v_distances=vector(#vpar));
  for(i=1,#v_ancestors, v_distances[v_ancestors[i]]=i-1);
  \\ print("  u_distances "u_distances);
  \\ print("  v_distances "v_distances);

  my(keys=vector(#common,i, [u_distances[common[i]] + v_distances[common[i]],
                             u_distances[common[i]] ]));
  \\ print("  keys "keys);
  my(seq=vecsort(keys,,1));
  \\ print("  seq "Vec(seq));
  my(c=common[seq[1]]);
  [c, u_distances[c], v_distances[c]];
}

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     for(u=1,#vpar,
        for(v=1,#vpar,
           \\ print(vpar" "u" "v);
           my(got=vpar_INTERNAL_common_ancestor_and_distances(vpar,u,v));
           if(got[1],
              check_equal(vpar_is_ancestor(vpar,u,got[1]), 1);
              check_equal(vpar_is_ancestor(vpar,v,got[1]), 1));
           check_equal(got,
                       test_common_ancestor_and_distances_cyclic_by_ancestors(vpar,u,v),
                       "vpar_INTERNAL_common_ancestor_and_distances() vs test_common_ancestor_and_distances_cyclic_by_ancestors()"))));
}



\\-----------------------------------------------------------------------------
progress("vpar_path_lengths_matrix()");

test_path_lengths_matrix_by_vector(vpar) = \
  Mat(vector(#vpar,v, vpar_path_lengths_vector(vpar,v)~));

test_path_lengths_matrix_by_seq(vpar) =
{
  \\ Previous code.
  \\ Wrong results for cyclic vpar, but shouldn't inf-loop.
  \\ Suspect not greatly better than individual vpar_path_length_vector().

  my(seq=vpar_seq_successive_paths(vpar,1),
     distances=matrix(#vpar,#vpar,u,v, -1));

  \\ upwards set distances[u,v] for u ancestor of v
  my(p);
  for(v=1,#vpar,
     my(u=v);
     p=-1; \\ distance above v
     while(distances[u,v]<0,   \\ stop at cycle
           distances[u,v]=p++; \\ pre-increment
           (u=vpar[u]) || break));

  \\ downwards set distances[u,v] for u descendant of ancestor of v,
  \\ including descendant of v itself
  for(i=1,#seq,
     my(u=seq[i]);
     if((p=vpar[u]),
        for(v=1,#vpar,
           if(distances[u,v]==-1 && distances[p,v]>=0,
              distances[u,v] = distances[p,v]+1))));
  distances;
}

{
  my(vpar=[0,1,1]);
  \\ vpar_view(vpar);
  check_equal(vpar_path_lengths_matrix(vpar),
              [ 0,1,1; 1,0,2; 1,2,0 ],
              "vpar_path_lengths_matrix() star-3");
  check_equal(vpar_path_lengths_matrix(vpar),
              test_path_lengths_matrix_by_vector(vpar),
              "vpar_path_lengths_matrix() star-3 vs vector");
}

check_equal(vpar_path_lengths_matrix([0]), Mat([0]), \
            "vpar_path_lengths_matrix() singleton");

check_equal(vpar_path_lengths_matrix([1]), Mat([0]), \
            "vpar_path_lengths_matrix() self-loop");

check_equal(vpar_path_lengths_matrix([2,1]), [0,1; 1,0], \
            "vpar_path_lengths_matrix() 2-cycle");

\\ 3 <- 1 <- 2 <- 3
check_equal(vpar_path_lengths_matrix([3,1,2]), [0,1,1; 1,0,1; 1,1,0], \
            "vpar_path_lengths_matrix() cycle-3");

check_equal(vpar_path_lengths_matrix([0,1]), [0,1; 1,0], \
            "vpar_path_lengths_matrix() path-2");
check_equal(vpar_path_lengths_matrix([2,0]), [0,1; 1,0], \
            "vpar_path_lengths_matrix() path-2 reverse");

check_equal(vpar_path_lengths_matrix([2, 3, 0]), [0,1,2; 1,0,1; 2,1,0], \
            "vpar_path_lengths_matrix() path-3 reverse");

\\ 2 --> 1 --> 3
check_equal(vpar_path_lengths_matrix([3,1,0]), [0,1,1; 1,0,2; 1,2,0], \
            "vpar_path_lengths_matrix() path 3,1,0");

\\ 1 <-> 2
\\       |
\\       3
check_equal(vpar_path_lengths_matrix([2,1,2]), [0,1,2; 1,0,1; 2,1,0], \
            "vpar_path_lengths_matrix() 2,1,2");
check_equal(vpar_path_lengths_matrix([2,1,2,4]), [0,1,2,-1; 1,0,1,-1; 2,1,0,-1; -1,-1,-1,0], \
            "vpar_path_lengths_matrix() 2,1,2,4");

{
  \\  4 < 3
  \\   v ^
  \\    2
  \\    |
  \\    1
  my(vpar=[2, 3,4,2],
     want=[0,1,2,2;
           1,0,1,1;
           2,1,0,1;
           2,1,1,0]);
  check_equal(test_path_lengths_matrix_by_vector(vpar), want,
              "test_path_lengths_matrix_by_vector() four");
  check_equal(vpar_path_lengths_matrix(vpar), want,
              "vpar_path_lengths_matrix() four");
}
{
  \\  3 < 2
  \\  v   ^    5-cycle
  \\  4   1
  \\    5
  my(vpar=[2,3,4,5, 1],
     want=[0,1,2,2,1;
           1,0,1,2,2;
           2,1,0,1,2;
           2,2,1,0,1;
           1,2,2,1,0]);
  check_equal(test_path_lengths_matrix_by_vector(vpar), want,
              "test_path_lengths_matrix_by_vector() 5-cycle");
  check_equal(vpar_path_lengths_matrix(vpar), want,
              "vpar_path_lengths_matrix() 5-cycle");
}
{
  \\  3 < 2
  \\  v   ^    4-cycle
  \\  4 > 1
  my(vpar=[2,3,4,1],
     want=[0,1,2,1;
           1,0,1,2;
           2,1,0,1;
           1,2,1,0]);
  check_equal(test_path_lengths_matrix_by_vector(vpar), want,
              "test_path_lengths_matrix_by_vector() four B");
  check_equal(vpar_path_lengths_matrix(vpar), want,
              "vpar_path_lengths_matrix() four B");
}
{
  \\          4 < 5
  \\           v ^
  \\ 1 loop     3
  \\            |
  \\            2
  my(vpar=[1, 3, 5,3,4],
     want=[0,  -1,-1,-1,-1;
           -1, 0,1,2,2;
           -1, 1,0,1,1;
           -1, 2,1,0,1;
           -1, 2,1,1,0]);
  check_equal(test_path_lengths_matrix_by_vector(vpar), want,
              "test_path_lengths_matrix_by_vector() five");
  check_equal(vpar_path_lengths_matrix(vpar), want,
              "vpar_path_lengths_matrix() five");
}
{
  \\          4 < 1
  \\           v ^
  \\ 2 root     5
  \\            |
  \\            3
  my(vpar=[4, 0, 5, 5, 1],
     want=[0, -1, 2,1,1;
           -1, 0, -1,-1,-1;
           2, -1, 0,2,1;
           1, -1, 2,0,1;
           1, -1, 1,1,0]);
  check_equal(test_path_lengths_matrix_by_vector(vpar), want,
              "test_path_lengths_matrix_by_vector() five B");
  check_equal(vpar_path_lengths_matrix(vpar), want,
              "vpar_path_lengths_matrix() five B");
}

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        m=vpar_path_lengths_matrix(vpar));
     check_equal(m,m~, "vpar_path_lengths_matrix() symmetric");

     check_equal(vpar_path_lengths_matrix(vpar),
                 matrix(#vpar,#vpar,u,v, vpar_path_length(vpar,u,v)));
  );
}

{
  for(i=1,#test_sample_trees_and_forests,
     my(vpar=test_sample_trees_and_forests[i],
        m=vpar_path_lengths_matrix(vpar));
     check_equal(m,m~, "vpar_path_lengths_matrix() symmetric");
     check_equal(m, test_path_lengths_matrix_by_vector(vpar),
       "vpar_path_lengths_matrix() vs test_path_lengths_matrix_by_vector()");

     \\ per addhelp of vpar_path_lengths_matrix()
     check_equal(apply(d->d==1, m),  vpar_adjacency_matrix(vpar),
       "vpar_path_lengths_matrix() to vpar_adjacency_matrix()");

     \\ R.L. Graham and H.O. Pollak, "On the Addressing Problem for Loop
     \\ Switching", Bell Systems Tech J, volume 50, 1971, pages 2495-2519.
     \\ http://www.math.ucsd.edu/~ronspubs/71_05_loop_switching.pdf
     if(#vpar >= 2 && vpar_is_tree(vpar),
        my(n = #vpar);
        check_equal(matdet(m), (-1)^(n-1) * (n-1) * 2^(n-2),
                 "vpar_path_lengths_matrix() matdet of tree"));

     \\ Merris, "Laplacian Matrices of Graphs: A Survey", Linear Algebra
     \\ Applications, volume 197 and 198, 1994, pages 143-176.
     \\ FIXME: this not right ...
     \\ if(#vpar >= 2 && vpar_is_tree(vpar),
     \\    my(A=vpar_adjacency_matrix(vpar));
     \\    print(A*vpar_path_lengths_matrix(vpar)*A~);
     \\   \\ check_equal(A~*m*A, 2*matid(#vpar));
     \\   );

     \\ Graham and Lovasz, "Distance Matrix Polynomials of Trees", Bell
     \\ Systems Technical Journal, lemma formula for distance matrix inverse.
     if(vpar_is_tree(vpar) && vpar_num_edges(vpar)>=1,
        my(minv=m^-1,
           degrees=vpar_degrees_vector(vpar));
        check_equal(minv,
                    matrix(#vpar,#vpar,i,j,
                           (2-degrees[i])*(2-degrees[j]) / ( 2*(#vpar-1) )
                           + if(i==j, -1/2*degrees[i],
                                1/2*vpar_is_neighbour(vpar,i,j))),
          "vpar_path_lengths_matrix() inverse of tree");
        check_equal(minv,
                    matrix(#vpar,#vpar,i,j,
                           (2-degrees[i])*(2-degrees[j]) / ( 2*(#vpar-1) ))
                    - 1/2*vpar_Laplacian_matrix(vpar),
          "vpar_path_lengths_matrix() inverse, with Laplacian, of tree"));
  );
}

{
  \\ distinct path lengths matrices occurring = count unrooted
  for(forest=0,1,
     for(n=2, test_limit_trees_search - forest,
        my(l=List([]));
        forvec(vec=vpar_forvec_from(n,forest),
              my(vpar=vpar_from_upascend(vec),
                 m=vpar_path_lengths_matrix(vpar));
              my(pos=setsearch(l,m,1));
              if(pos, listinsert(l,m,pos)));

        check_equal(#l, vpar_count_unrooted(n,forest),
          "vpar_path_lengths_matrix() distinct count is vpar_count_unrooted()");
     ));
}

{
  for(n=0,10,
     my(vpar=vector(n));
     check_equal(matdet(vpar_path_lengths_matrix(vpar)), 1-n,
                 "vpar_path_lengths_matrix() matdet of singletons"));
}

{
  \\ A085750 matdet of star-n path lengths matrix
  check_equal(vector(#A085750_samples,n,
                     matdet(vpar_path_lengths_matrix(vpar_make_star(n)))),
              A085750_samples,    \\ OFFSET=1
              "A085750 matdet vpar_path_lengths_matrix() on path-n");
}


\\------
\\ path
{
  \\ A085750 matdet of path-n path lengths
  \\ (-1)^(n-1) * (n-1) * 2^(n-2),
  \\ 1 0 -1 4 -12 32 -80 192 -448 1024 -2304
  check_equal(vector(#A085750_samples,n,
                     (-1)^(n-1) * (n-1) * 2^(n-2)),
              A085750_samples,    \\ OFFSET=1
              "A085750 formula");
  check_equal(vector(#A085750_samples,n,
                     matdet(vpar_path_lengths_matrix(vpar_make_path(n)))),
              A085750_samples,    \\ OFFSET=1
              "A085750 matdet vpar_path_lengths_matrix() on path-n");
}
{
  \\ R. B. Bapat and S. Sivasubramanian, "Product Distance Matrix of a Graph
  \\ and Squared Distance Matrix of a Tree", Applicable Analysis and
  \\ Discrete Mathematics, volume 7, number 2, October 2013, pages 285-301.
  \\ http://www.jstor.org/stable/43660717
  \\
  \\ Theorem 17 determinant of path lengths matrix entries squared of a tree.

  for(n=1,12,
     my(vpar=vpar_free_first(n));
     until((vpar=vpar_free_next(vpar))=='none,
           my(m = apply(sqr, vpar_path_lengths_matrix(vpar)),
              by_det = matdet(m),
              degrees=vpar_degrees_vector(vpar),
              tau=apply(d->2-d, degrees),           \\ 2 - degree
              by_formula = (-1)^n
                          * 4^(n-2)
                          * ( (2*n-1) * prod(i=1,n, tau[i])
                             - 2* sum(i=1,n, prod(j=1,n, if(i==j,1, tau[j])))));
           check_equal(by_det, by_formula,
                       "vpar_path_lengths_matrix() squared entries matdet")));
}

{
  \\ R. B. Bapat and S. Sivasubramanian, "Product Distance Matrix of a Graph
  \\ and Squared Distance Matrix of a Tree", Applicable Analysis and
  \\ Discrete Mathematics, volume 7, number 2, October 2013, pages 285-301.
  \\ http://www.jstor.org/stable/43660717
  \\
  \\ Corollary 18, determinant of path length matrix entries squared of
  \\ complete binary tree.

  check_equal(vector(5,h,
                     matdet(apply(sqr,
                     vpar_path_lengths_matrix(vpar_make_complete_tree(2,h))))),
              vector(5,h,
                     my(n=#vpar_make_complete_tree(2,h));
                     2*4^(n-2)),
              "A085750 matdet vpar_path_lengths_matrix() on path-n");

  for(h=1,5,
     my(n=#vpar_make_complete_tree(2,h));
     check_equal(n, 2^(h+1)-1));

  check_equal(vector(5,h,
                     my(n=#vpar_make_complete_tree(2,h));
                     2*4^(n-2)),
              vector(5,h, 1/32 * 16^(2^h)));
  \\ vector(4,h, 1/32 * 16^(2^h))
  \\ not in OEIS: 8, 2048, 134217728, 576460752303423488
}


\\-----------------------------------------------------------------------------
progress("vpar_free_branch_reduce()");

check_equal(vpar_free_branch_reduce([]), []);
check_equal(vpar_free_branch_reduce([0]), [0], \
            "vpar_free_branch_reduce() singleton");
check_equal(vpar_free_branch_reduce([1]), []);
check_equal(vpar_free_branch_reduce([0,1]), [0,1]);
check_equal(vpar_free_branch_reduce([0,1,2]), [0,1]);
check_equal(vpar_free_branch_reduce([2,3,1]), []);

check_equal(vpar_free_branch_reduce([0,1,1]), [0,1]);

test_find_degree_2(vpar) =
{
  my(degrees=vpar_degrees_vector(vpar));
  for(v=1,#vpar,
     if(degrees[v]==2 && vpar[v],  \\ non-root first
        return(v)));
  for(v=1,#vpar,
     if(degrees[v]==2,             \\ then any
        return(v)));
  0;
}

test_free_branch_reduce_by_deletes(vpar) =
{
  my(v);
  while(v=test_find_degree_2(vpar),
        my(neighbours=vpar_neighbours_of_vertex(vpar,v));
        #neighbours==2 || error();
        if(vpar_is_ancestor(vpar,neighbours[1],neighbours[2]),
           vpar[neighbours[1]]=neighbours[2],   \\ upwards adopt
           vpar_is_ancestor(vpar,neighbours[2],neighbours[1]),
           vpar[neighbours[2]]=neighbours[1],   \\ upwards adopt
           vpar[vecmax(neighbours)]=vecmin(neighbours)); \\ bigger -> smaller
        vpar=vpar_delete_vertex(vpar,v));
  vpar;
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     check_equal(vpar_free_branch_reduce(vpar),
                 test_free_branch_reduce_by_deletes(vpar),
                 "vpar_free_branch_reduce() vs test_free_branch_reduce_by_deletes()"));
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        degrees=vpar_degrees_vector(vpar),
        rerooted=vpar);
     for(v=1,#vpar,
        if(degrees[v]!=2, rerooted=vpar_reroot(rerooted,v)));
     my(roots=vpar_roots(rerooted));
     for(i=1,#roots,
        check_equal(vpar_degree_of_vertex(rerooted,roots[i])!=2, 1,
                    "rerooted root degree"));
     rerooted=vpar_branch_reduce(rerooted);

     my(free=vpar_free_branch_reduce(vpar));
     \\ print(vpar);
     \\ print(rerooted);
     \\ print(free);
     rerooted=vpar_reroot_forest(rerooted,vpar_roots(free));
     check_equal(free,rerooted,
                 "vpar_free_branch_reduce() vs rerooted vpar_branch_reduce()");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_is_singleton()");

check_equal(vpar_is_singleton([0],1), 1);
check_equal(vpar_is_singleton([0,1],1), 0);
check_equal(vpar_is_singleton([0,1],2), 0);
check_equal(vpar_is_singleton([0,0,1],1), 0);
check_equal(vpar_is_singleton([0,0,1],2), 1);
check_equal(vpar_is_singleton([0,0,1],3), 0);

\\ cyclics
check_equal(vpar_is_singleton([1],1), 0);
check_equal(vpar_is_singleton([2,1],1), 0);
check_equal(vpar_is_singleton([2,1],2), 0);

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     for(v=1,#vpar,
        my(got=vpar_is_singleton(vpar,v));
        check_equal(got, vpar_degree_of_vertex(vpar,v)==0,
                    "vpar_is_singleton() vs vpar_degree_of_vertex()");
        check_equal(got, vpar[v]==0 && vpar_is_childless(vpar,v),
                    "vpar_is_singleton() vs no parent and vpar_is_childless()");
     ));
}


\\-----------------------------------------------------------------------------
progress("vpar_singletons()");

check_equal(vpar_singletons([]), []);
check_equal(vpar_singletons([0]), [1]);
check_equal(vpar_singletons([0,1]), []);
check_equal(vpar_singletons([0,0]), [1,2]);
check_equal(vpar_singletons([0,0,1]), [2]);

\\ cyclics
check_equal(vpar_singletons([1]), []);
check_equal(vpar_singletons([2,1]), []);

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        got=vpar_singletons(vpar));
     check_equal(got, test_select_vertices(vpar,vpar_is_singleton),
                 "vpar_singletons() vs vpar_is_singleton()");
     check_equal(got, setintersect(vpar_roots(vpar),
                                   vpar_childless_vertices(vpar)),
                 "vpar_singletons() vs vpar_roots() setintersect vpar_childless_vertices() per addhelp");

     check_equal(#setintersect(got, vpar_cyclic_vertices(vpar)), 0,
                 "vpar_singletons() disjoint from vpar_cyclic_vertices() per addhelp");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_leaves()");

check_equal(vpar_leaves([]), []);
check_equal(vpar_leaves([0]), [1]);
check_equal(vpar_leaves([0,1]), [1,2]);
check_equal(vpar_leaves([0,1,2]), [1,3]);

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        got=vpar_leaves(vpar));
     check_equal(got, test_select_vertices(vpar,vpar_is_leaf),
                 "vpar_leaves() vs vpar_is_leaf()");
     check_equal(#got, vpar_num_leaves(vpar),
                 "vpar_leaves() length vs vpar_num_leaves()");
     check_equal(#setintersect(got, vpar_cyclic_vertices(vpar)), 0,
                 "vpar_leaves() not cyclic so disjoint from vpar_cyclic_vertices()");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_num_leaves()");

check_equal(vpar_num_leaves([]), 0);
check_equal(vpar_num_leaves([0]), 1);
check_equal(vpar_num_leaves([0,1]), 2);
check_equal(vpar_num_leaves([0,1,2]), 2);
check_equal(vpar_num_leaves([0,0]), 2);

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        got=vpar_num_leaves(vpar));
     check_equal(got, sum(v=1,#vpar, vpar_is_leaf(vpar,v)),
                 "vpar_num_leaves() vs count vpar_is_leaf()");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_is_leaf()");

check_equal(vpar_is_leaf([0],1), 1, "vpar_is_leaf() singleton");
check_equal(vpar_is_leaf([0,1],1), 1);
check_equal(vpar_is_leaf([0,1],2), 1);
check_equal(vpar_is_leaf([0,1,2],1), 1);
check_equal(vpar_is_leaf([0,1,2],2), 0);
check_equal(vpar_is_leaf([0,1,2],3), 1);

\\ cyclic
check_equal(vpar_is_leaf([1],1), 0, "vpar_is_leaf() self-loop");

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     for(v=1,#vpar,
        my(got=vpar_is_leaf(vpar,v));
        check_equal(got, vpar_degree_of_vertex(vpar,v)<=1,
                    "vpar_is_leaf() vs vpar_degree_of_vertex()");
        if(vpar_is_cyclic_vertex(vpar,v),
           check_equal(got, 0,
                       "vpar_is_leaf() false at on vpar_is_cyclic_vertex()"));
     ));
}


\\-----------------------------------------------------------------------------
progress("vpar_degree_of_vertex()");

check_equal(vpar_degree_of_vertex([0,1,2],1), 1);
check_equal(vpar_degree_of_vertex([0,1,2],2), 2);
check_equal(vpar_degree_of_vertex([0,1,2],3), 1);

check_equal(vpar_degree_of_vertex([0],1), 0);

\\ cyclics
check_equal(vpar_degree_of_vertex([1],1), 2, \
            "vpar_degree_of_vertex() self-loop = 2");
check_equal(vpar_degree_of_vertex([2,1],1), 2, \
            "vpar_degree_of_vertex() 2-cycle");
check_equal(vpar_degree_of_vertex([2,1],2), 2);

check_equal(vpar_num_edges([1]), 1, \
            "vpar_num_edges() self-loop = 1");


\\-----------------------------------------------------------------------------
progress("vpar_degrees_vector()");

check_equal(vpar_degrees_vector([]), [], "vpar_degrees_vector() empty");
check_equal(vpar_degrees_vector([0]), [0]);
check_equal(vpar_degrees_vector([0,1]), [1,1]);
check_equal(vpar_degrees_vector([2,1]), [2,2]);

{
  \\ cycles
  for(n=1,10,
     my(vpar=test_make_cycle(n));
     check_equal(vpar_degrees_vector(vpar), vector(n,i, 2),
                "vpar_degrees_vector() cycle all 2s"));
}
{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        vec=vpar_degrees_vector(vpar));
     check_equal(vecsum(vec),
                 2*vpar_num_edges(vpar),
                 "vpar_degrees_vector() vecsum = 2*num edges");
     check_equal(vec,
                 vector(#vpar,v, vpar_degree_of_vertex(vpar,v)),
       "vpar_degrees_vector() vs individual vpar_degree_of_vertex()"));
}


\\-----------------------------------------------------------------------------
progress("vpar_eccentricity_of_vertex()");

{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i]);
     for(v=1,#vpar,
        check_equal(vpar_eccentricity_of_vertex(Vecsmall(vpar),v),
                    vpar_height(vpar_reroot(vpar,v)),
                    "vpar_eccentricity_of_vertex() vs vpar_height() of reroot")));
}


\\-----------------------------------------------------------------------------
progress("vpar_delete_vertex()");

check_equal(vpar_delete_vertex(vpar_make_path(10),4), \
            vpar_concat_forest(vpar_make_path(3),vpar_make_path(6)));

\\ cyclics
check_equal(vpar_delete_vertex([1],1), []);
check_equal(vpar_delete_vertex([2,3,4,1],1), [2,3,0]);
check_equal(vpar_delete_vertex([2,3,4,1],3), [2,0,1]);
check_equal(vpar_delete_vertex([1, 1,1,1],1), [0,0,0]);
check_equal(vpar_delete_vertex([4,4,4,4],1), [3,3,3]);


\\-----------------------------------------------------------------------------
progress("vpar_contract_vertex()");

check_equal(vpar_contract_vertex([0],1), []);
check_equal(vpar_contract_vertex([0,1,2],2), [0,1]);

\\ cyclics
check_equal(vpar_contract_vertex([1],1), []);
check_equal(vpar_contract_vertex([2,3,1, 1,1],1), [2,1, 1,1]);
check_equal(vpar_contract_vertex([2,3,1, 3,3],3), [2,1, 1,1]);
check_equal(vpar_contract_vertex([2,3,1, 3,3],2), [2,1, 2,2]);
check_equal(vpar_contract_vertex([1, 1,1,1],1), [0,0,0]);
check_equal(vpar_contract_vertex([4,4,4, 4],1), [3,3,3]);
check_equal(vpar_contract_vertex([4,4,4, 4],4), [0,0,0]);


\\  1 --> 2 --> 3
\\
check_equal(vpar_delete_vertex([2,3,0],2), [0,0], \
            "vpar_delete_vertex() middle");
check_equal(vpar_contract_vertex([2,3,0],1), [2,0], \
            "vpar_contract_vertex() first");
check_equal(vpar_contract_vertex([2,3,0],2), [2,0], \
            "vpar_contract_vertex() middle");
check_equal(vpar_contract_vertex([2,3,0],3), [2,0], \
            "vpar_contract_vertex() end");

test_contract_vertex(vpar,v) =
{
  for(c=1,#vpar,
     if(vpar[c]==v, vpar[c]=vpar[v]));  \\ to grandparent
  vpar_delete_vertex(vpar,v);
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     for(v=1,#vpar,
        check_equal(vpar_contract_vertex(vpar,v),
                    test_contract_vertex(vpar,v))));
}


\\-----------------------------------------------------------------------------
progress("vpar_branch_reduce()");

check_equal(vpar_branch_reduce([]), []);
check_equal(vpar_branch_reduce([0]), [0]);
check_equal(vpar_branch_reduce([0,1]), [0,1]);
check_equal(vpar_branch_reduce([0,1,1,1,1]), [0,1,1,1,1]);
check_equal(vpar_branch_reduce([0,1,2,3,4]), [0,1]);

\\ cyclics
check_equal(vpar_branch_reduce([1]), [], \
            "vpar_branch_reduce() self-loop alone delete");
check_equal(vpar_branch_reduce([2,3,1]), [], \
            "vpar_branch_reduce() cycle alone delete");
check_equal(vpar_branch_reduce([1,1,1,1]), [1,1,1,1], \
            "vpar_branch_reduce() self-loop and more keep");
check_equal(vpar_branch_reduce([2,3,1,1]), [1,1], \
            "vpar_branch_reduce() loop and one below shrink");

\\ when want to reduce
test_vertex_is_branch_reduce(vpar,v) = \
  vpar[v] && vpar_num_children_of_vertex(vpar,v) == 1;

test_branch_reduce(vpar) =
{
  my(keep=vector(#vpar,v, !test_vertex_is_branch_reduce(vpar,v)));
  for(v=1,#vpar,
     my(p=vpar[v]);
     if(p && keep[v],
        while(p && !keep[p], p=vpar[p]);   \\ up past delete parents
        vpar[v]=p));
  vpar_keep_flags(vpar,keep);
}
test_branch_reduce_by_contract(vpar) =
{
  forstep(v=#vpar,1,-1,
     if(test_vertex_is_branch_reduce(vpar,v),
        vpar=vpar_contract_vertex(vpar,v)));
  vpar;
}
test_branch_reduce_by_seq(vpar) =
{
  \\ Past code using a seq.
  if(vpar_is_cyclic(vpar), error("for acyclics only"));

  my(vec = vpar_INTERNAL_num_children_vecsmall(vpar),
     seq = vpar_INTERNAL_seq_upwards(vpar,vec));
  for(v=1,#vec, vec[v] = !(vpar[v] && vec[v]==1));

  \\ If parent p is to be deleted, then set vpar[v]=grandparent.
  \\ Vertices are considered root downwards so in a chain of deletes the
  \\ grandparent being kept becomes the parent of all the deleted.  On
  \\ reaching the kept vertex at the bottom of the chain the grandparent is
  \\ also being kept.
  \\
  forstep(i=#seq,1,-1,
         my(v=seq[i], p=vpar[v]);
         if(p&&!vec[p], vpar[v]=vpar[p]));

  vpar_keep_flags(vpar,vec);
}

\\  1 --> 2 --> 3
\\
check_equal(test_branch_reduce([2,3,0]), [2,0]);
check_equal(test_branch_reduce_by_contract([2,3,0]), [2,0]);

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got=vpar_branch_reduce(vpar));
     check_equal(got,
                 test_branch_reduce_by_seq(vpar),
                 "test_branch_reduce() vs test_branch_reduce_by_seq()");
  );
}
{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        got=vpar_branch_reduce(vpar));
     check_equal(got,
                 test_branch_reduce(vpar));
     check_equal(got,
                 test_branch_reduce_by_contract(vpar),
                 "test_branch_reduce() vs test_branch_reduce_by_contract()");
  );
}


\\-----------------------------------------------------------------------------
progress("test_ordered_Bell()");

check_equal(vector(#A000670_samples,n,n--; test_ordered_Bell(n)), \
            A000670_samples);    \\ OFFSET=0
check_equal(vector(20,n, sum(k=1,n, binomial(n,k)*test_ordered_Bell(n-k))), \
            vector(20,n, test_ordered_Bell(n)));


\\-----------------------------------------------------------------------------
progress("vpar_depths_vectors_count()");

check_equal(vpar_depths_vectors_count(0),   1, "vpar_depths_vectors_count() n=0");
check_equal(vpar_depths_vectors_count(0,1), 1);
check_equal(vpar_depths_vectors_count(1),   1, "vpar_depths_vectors_count() n=1");
check_equal(vpar_depths_vectors_count(1,1), 1);
check_equal(vpar_depths_vectors_count(2),   2, "vpar_depths_vectors_count() n=2");
check_equal(vpar_depths_vectors_count(2,1), 3);

{
  \\ negative n comes out 0, which is nice, but not a documented feature
  for(forest=-1,1,
     for(n=-5,-1,
        check_equal(vpar_depths_vectors_count(-1,forest), 0)));
}

{
  \\ number of different depths vectors occurring
  for(n=0,6,
     my(count_trees=0, count_forests=0, count_root1=0);
     forvec(vec=vector(n,i,[0,n-1]),
           count_trees   += vpar_is_depths_vector(vec);
           count_forests += vpar_is_depths_vector(vec,1);
           count_root1   += vpar_is_depths_vector(vec,-1));

     check_equal(count_forests, test_ordered_Bell(n),
                 Str("vpar_is_depths_vector() count_forests n=",n));
     check_equal(count_trees,   if(n, n*test_ordered_Bell(n-1), 1),
                 Str("vpar_is_depths_vector() count_trees n=",n));
     check_equal(count_root1,   if(n, test_ordered_Bell(n-1), 1),
                 Str("vpar_is_depths_vector() count_root1 n=",n));

     check_equal(count_trees, vpar_depths_vectors_count(n),
                 Str("vpar_depths_vectors_count() vs vpar_is_depths_vector() trees n=",n));
     check_equal(count_forests, vpar_depths_vectors_count(n,1),
                 Str("vpar_depths_vectors_count() vs vpar_is_depths_vector() forests n=",n));
     check_equal(count_root1, vpar_depths_vectors_count(n,-1),
                 Str("vpar_depths_vectors_count() vs vpar_is_depths_vector() root=1 trees n=",n));
  );
}

{
  \\ Single-component functional digraphs same as forests.
  \\
  \\ Trees and self-loop cyclics are same depths vectors, correspond to the
  \\ trees within the forests.
  \\
  \\ Cyclic of len >= 2 correspond to forests >= 2 roots.  The cyclic depths
  \\ vector has nothing < len-1, and can map to forest depths vector by
  \\ identifying the cyclic vertices with the roots and subtracting their
  \\ len-1 in the depths vector.

  for(n=0,6,
     my(l=List([]));
     forvec(vpar=vector(n,i,[0,n]),
           vpar_num_components(vpar)<=1 || next;
           my(depths=vpar_depths_vector(vpar),
              pos=setsearch(l,depths,1));
           if(pos, listinsert(l,depths,pos)));
     l=Vec(l);
     check_equal(setisset(l), 1);
     check_equal(#l, vpar_depths_vectors_count(n,1),
                 "vpar_depths_vectors_count() vs distinct vpar_depths_vector() one-component functional digraphs"));
}

\\ {
\\   \\ Any-component functional digraphs -- formula for gaps etc.
\\   for(n=0,6,
\\      my(l=List([]));
\\      forvec(vpar=vector(n,i,[0,n]),
\\            my(depths=vpar_depths_vector(vpar),
\\               pos=setsearch(l,depths,1));
\\            if(pos, listinsert(l,depths,pos)));
\\      l=Vec(l);
\\      check_equal(setisset(l), 1);
\\      check_equal(#l, ...
\\                  "any-component functional digraphs distinct vpar_depths_vector()"));
\\ }


\\-----------------------------------------------------------------------------
progress("test_is_Motzkin_forest()");

\\ Balanced binary 101 is not Motzkin, since it is a vertex with absent left
\\ but present right.
test_is_Motzkin_forest_by_balanced_binary(vpar) =
{
  my(b=binary(vpar_to_balanced_binary(vpar)));
  for(i=1,#b-2,
     if(b[i] && !b[i+1] && b[i+2], return(0)));
  1;
}

{
  \\ A001006 Motzkin numbers
  \\ 1,1,2,4,9,21,51,127,323,835,2188,
  for(n=1,10,
     my(vpar=vpar_preorder_first(n),
        count=0);
     until((vpar=vpar_preorder_next(vpar,1))=='none,
           my(got=test_is_Motzkin_forest(vpar));
           count += got;
           check_equal(got,
                       test_is_Motzkin_forest_by_balanced_binary(vpar)));
     check_equal(count,A001006_samples[n]));
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     for(v=1,#vpar,
        my(vec=vpar_ancestors(vpar,v));
        check_equal(test_is_Motzkin_forest(vpar),
                    test_is_Motzkin_forest_by_balanced_binary(vpar))));
}


\\-----------------------------------------------------------------------------
progress("vpar_ancestors()");


{
  my(vpar=[2,0,1,0]);
  check_equal(vpar_ancestors(vpar,1), [1,2]);
  check_equal(vpar_ancestors(vpar,2), [2]);
  check_equal(vpar_ancestors(vpar,3), [3,1,2]);
  check_equal(vpar_ancestors(vpar,4), [4]);
}
{
  \\ cyclic
  my(vpar=[2,3,1,3,3,2,6]);
  check_equal(vpar_ancestors(vpar,1), [1,2,3]);
  check_equal(vpar_ancestors(vpar,2), [2,3,1]);
  check_equal(vpar_ancestors(vpar,3), [3,1,2]);
  check_equal(vpar_ancestors(vpar,4), [4,3,1,2]);
  check_equal(vpar_ancestors(vpar,5), [5,3,1,2]);
  check_equal(vpar_ancestors(vpar,6), [6,2,3,1]);
  check_equal(vpar_ancestors(vpar,7), [7,6,2,3,1]);
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     for(v=1,#vpar,
        my(vec=vpar_ancestors(vpar,v));
        check_equal(#vec, vpar_depth_of_vertex(vpar,v) + 1);
        check_equal(vec, vpar_path_vertices(vpar,v,vpar_root_of_vertex(vpar,v)));
     ));
}
{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     for(v=1,#vpar,
        my(vec=vpar_ancestors(vpar,v));
        check_equal(#vec, vpar_depth_of_vertex(vpar,v) + 1);

        check_equal(vecsort(vec),
                    select(i->vpar_is_ancestor(vpar,v,i), [1..#vpar]),
                    "vpar_ancestors() vs individual vpar_is_ancestor()");

        for(i=1,#vec-1,
           check_equal(vpar[vec[i]], vec[i+1],
                       "vpar_ancestors() is successive parents"))));
}


\\-----------------------------------------------------------------------------
progress("test_concat_forest_vector()");

check_equal(test_concat_forest_vector([]), [], \
            "test_concat_forest_vector() none");
check_equal(test_concat_forest_vector([ [] ]), [], \
            "test_concat_forest_vector() 1 empty");
check_equal(test_concat_forest_vector([ [],[],[] ]), [], \
            "test_concat_forest_vector() 3 empties");
check_equal(test_concat_forest_vector([ [0],[0] ]), [0,0]);


\\-----------------------------------------------------------------------------
progress("test_mod_nearest()");

check_equal(test_mod_nearest(4,3), 1);
check_equal(test_mod_nearest(5,3), -1);
check_equal(test_mod_nearest(6,3), 0);

check_equal(test_mod_nearest(4,4), 0);
check_equal(test_mod_nearest(5,4), 1);
check_equal(test_mod_nearest(6,4), 2);
check_equal(test_mod_nearest(7,4), -1);


\\-----------------------------------------------------------------------------
progress("vpar_INTERNAL_Catalan_triangle()");

check_equal(vector(10,n,n--; vpar_INTERNAL_Catalan_triangle(n,0)), \
            vector(10,n,n--; 1), \
            "vpar_INTERNAL_Catalan_triangle() n,0");

check_equal(vector(10,n,n--; vpar_INTERNAL_Catalan_triangle(n,n)), \
            vector(10,n,n--; vpar_INTERNAL_Catalan_number(n)), \
     "vpar_INTERNAL_Catalan_triangle() n,n = vpar_INTERNAL_Catalan_number()");

check_equal(vector(10,n, vpar_INTERNAL_Catalan_triangle(n,n-1)), \
            vector(10,n, vpar_INTERNAL_Catalan_number(n)), \
     "vpar_INTERNAL_Catalan_triangle() n,n-1 = vpar_INTERNAL_Catalan_number()");

{
  \\ A009766 Catalan triangle by rows
  \\ binomial(n+m,n)*(n-m+1)/(n+1)
  my(l=List([]));
  for(n=0,#A009766_samples,
     for(k=0,n,
        if(#l==#A009766_samples, break(2));
        listput(l,vpar_INTERNAL_Catalan_triangle(n,k))));
  check_equal(Vec(l), A009766_samples,
              "A009766 vpar_INTERNAL_Catalan_triangle()");
}
{
  \\ A030237 Catalan triangle by rows, sans duplicate at last of each row
  my(l=List([]));
  for(n=0,#A030237_samples,
     for(k=0,n-1,
        if(#l==#A030237_samples, break(2));
        listput(l,vpar_INTERNAL_Catalan_triangle(n,k))));
  check_equal(Vec(l), A030237_samples,
              "A030237 vpar_INTERNAL_Catalan_triangle()");
}

{
  \\ Preorder forests with initial path down length k,
  \\ per comments in the code.
  for(n=0,10,
     my(i=1);
     my(vpar=vpar_preorder_first(n),
        counts=vector(n+1));
     until((vpar=vpar_preorder_next(vpar,1))=='none,  \\ forests
           for(i=0,n,    \\ i many preceding vertices path down
              counts[i+1]++;
              if(i+1>n || vpar[i+1]!=i,
                 next(2))));
     for(i=0,n,     \\ preceding vertices, so depth of first of remaining
        my(k=n-i);  \\ remaining vertices
        check_equal(counts[i+1],
                    vpar_INTERNAL_Catalan_triangle(n,k),
        "vpar_INTERNAL_Catalan_triangle() num preorder forests with prefix"));
     );
}
{
  \\ Preorder forests with depth k of last vertex,
  \\ reversed triangle row n-1, n-1-k.
  for(n=1,10,
     my(vpar=vpar_preorder_first(n),
        counts=vector(n));
     until((vpar=vpar_preorder_next(vpar,1))=='none,  \\ forests
           counts[vpar_depth_of_vertex(vpar,n)+1]++);
     check_equal(counts,
                 vector(n,k,k--; vpar_INTERNAL_Catalan_triangle(n-1,n-1-k)),
                 "vpar_INTERNAL_Catalan_triangle() num preorder forests with last vertex depth k"));
}


\\-----------------------------------------------------------------------------
progress("Fibonacci Lattice");

\\ Richard P. Stanley, "The Fibonacci Lattice", Fibonacci Quarterly, volume
\\ 13, number 3, October 1975, pages 215-232.
\\ https://fq.math.ca/13-3.html
\\ https://fq.math.ca/Scanned/13-3/stanley.pdf
\\
\\         1          0 segs
\\        / \
\\       2   3        1 segs  L or R
\\      / \ / \
\\     4   5   6      2 segs  LL  L+R  RR
\\       S(3)

{
  my(smaller_list=[ [], [1],[1], [2], [2,3], [3] ]);
  check_equal(#smaller_list,6);

  my(is_ideal=(flags)->
     for(i=1,#flags,
        if(flags[i],
           for(j=1,#smaller_list[i],
              if(!flags[smaller_list[i][j]],
                 return(0)))));
     1);

  my(count_ideals=0);
  forvec(vec=vector(6,i,[0,1]),
        count_ideals += is_ideal(vec));
  check_equal(count_ideals,
              vpar_INTERNAL_Catalan_number(4));
}
{
  my(n=10,
     count_by_total=vector(2*n));
  for(n=1,n,
     forvec(vec=vector(n,i,[1,2]),
           my(total=vecsum(vec));
           count_by_total[total]++));
  check_equal(count_by_total[1..n],
              vector(n,i, fibonacci(i+1)));
}


\\-----------------------------------------------------------------------------
progress("vpar_count()");

check_equal(vpar_count(0,-1), 1, "n=0 root=1 trees, 1 of");

{
  \\ trees root=1
  \\ n=3     1      1 <- 2 <- 3
  \\        / \
  \\       2   3    1 <- 3 <- 2
  \\
  \\   n = 0  1  2   3    4     5      6
  my(want=[1, 1, 1,  3,  16,  125,  1296, 16807]);
  check_equal(vector(#want,n,n--; vpar_count(n,-1)), want);

  check_equal(vector(#want,n, vpar_count(n,-1)),
              vector(#want,n, n^(n-2)));

  check_equal(vector(#want,n, vpar_count(n,-1)),
              vector(#want,n, vpar_count(n-1,1)),
              "vpar_count() root=1 is forests of n-1");
}

{
  \\   n = 0  1  2   3    4     5      6
  my(want=[1, 1, 2,  9,  64,  625,  7776]);
  check_equal(vector(#want,n,n--; vpar_count(n)), want);
}
check_equal(vector(#want,n, vpar_count(n)), \
            vector(#want,n, n^(n-1)), \
            "vpar_count() formula in vpar.gp seq comments");

{
  \\   n = 0  1  2   3    4     5      6
  my(want=[1, 1, 3, 16, 125, 1296, 16807]);
  check_equal(vector(#want,n,n--; vpar_count(n,1)), want);
}
check_equal(vector(#want,n, vpar_count(n,1)), \
            vector(#want,n, (n+1)^(n-1)), \
            "vpar_count() formula in vpar.gp seq comments");

{
  \\ A000169 trees
  check_equal(vector(#A000169_samples,n, vpar_count(n)),
              A000169_samples);    \\ OFFSET=1
}
{
  \\ A000272 trees root=1, or forests one further n
  check_equal(vector(#A000272_samples,n,n--; vpar_count(n,-1)),
              A000272_samples);    \\ OFFSET=0

  my(v=A000272_samples[^1]);
  check_equal(vector(#v,n,n--; vpar_count(n,1)), v);
}

{
  \\ A053506 n * count forests = total vertices in labelled forests
  \\ (n-1)*n^(n-2)
  \\ = count funcs with a 2-cycle a<->b and all reaching that cycle
  \\ = bi-rooted labelled trees with the roots adjacent
  check_equal(vector(#A053506_samples,n,n--; n*vpar_count(n,1)),
              A053506_samples);
}

{
  \\ A066274 endofunctions  (n-1)*n^(n-1)
  \\ functional digraphs not a single tree  n^n - n^(n-1)
  check_equal(vector(#A066274_samples,n, (n-1)*vpar_count(n)),
              A066274_samples);    \\ OFFSET=1
}

{
  \\ A060946 Vandermonde trace, sum of powers
  \\ cumulative trees count
  check_equal(vector(#A060946_samples,n, sum(i=1,n,vpar_count(i))),
              A060946_samples);    \\ OFFSET=1
}

{
  \\ A117887 unrooted labelled trees <= n
  \\ vector(8,n, sum(k=2,n, k^(k-2)))
  \\ Excluding n=1 or something.
  \\ Unrooted meaning root=1.
  \\
  check_equal(vector(#A117887_samples,n,n++; sum(i=2,n,vpar_count(i,-1))),
              A117887_samples,
              "A117887 unrooted labelled trees <= n");
}
{
  \\ A023879 number of partitions in expanding space
  A023879_samples=A023879_samples[^1];
  check_equal(vpar_INTERNAL_Euler_transform
              (vector(#A023879_samples,n, vpar_count(n))),
              A023879_samples);
}
{
  \\ A076728 = (n-1)^2 * n^(n-2)
  check_equal(vector(#A076728_samples,n, (n-1)^2 * n^(n-2)),
              A076728_samples,    \\ OFFSET=1
              "A076728 samples");
  check_equal(vector(#A076728_samples,n, (n-1)^2*vpar_count(n,-1)),
              A076728_samples,
              "A076728 samples, root=1 trees");
  check_equal(vector(#A076728_samples,n, (n-1)^2*vpar_count(n-1,1)),
              A076728_samples,
              "A076728 samples, forests");
  check_equal(vector(#A076728_samples,n, n--; n^2*vpar_count(n,1)),
              A076728_samples,
              "A076728 samples, forests n-1");
  check_equal(vector(#A076728_samples,n, (n-1)^2 * 1/n*vpar_count(n)),
              A076728_samples,
              "A076728 samples, trees");
}

{
  my(limit = 10,
     L = O(x^limit),
     gB = test_func_to_egf(n->vpar_count(n,-1), limit),   \\ root=1 n^(n-2)
     gT = test_func_to_egf(vpar_count, limit),            \\ trees  n^(n-1)
     gF = test_func_to_egf(n->vpar_count(n,1), limit),    \\ forests (n+1)^(n-1)
     gT1 = gT-1);  \\ 0 at n=0 instead of 1

  check_equal(gB, test_func_to_egf(n->if(n<2,1, n^(n-2)), limit),
             "vpar_count() gB");
  check_equal(gB', test_func_to_egf(n->if(n==0,1, (n+1)^(n-1)), limit-1),
             "vpar_count() gB'");
  check_equal(gB', gF + L/x,
             "vpar_count() gB' = gF is root=1 as shifted up forests");
  check_equal(x*gB, test_func_to_egf(n->if(n==0,0, n==1,1, n==2,2,
                                           n*(n-1)^(n-3)), limit+1),
             "vpar_count() x*gB");

  check_equal(gT, test_func_to_egf(n->if(n<2,1, n^(n-1)), limit),
             "vpar_count() gT");
  check_equal(x*gT', test_func_to_egf(n->if(n==0,0,  n*n^(n-1)), limit),
             "vpar_count() gT'");
  check_equal(gT', test_func_to_egf(n-> (n+1)^n, limit-1),
             "vpar_count() gT'");

  \\ gT' * gT'
  \\ 1,4,26,236,2760,39572,672592,13227804,295579520
  \\ A001863 Normalized total height of rooted trees with n nodes.
  \\ sum(k=0, n-2, (n-2)! * n^k/k!),
  \\
  \\ print(test_egf_terms(gT'*gT',limit));
  \\ check_equal(gT'*gT', test_func_to_egf(n-> (n+1)^n, limit),
  \\            "vpar_count() gT' * gT'");

  check_equal(gT, (gB-1)'*x+1);
  check_equal(gT, 1 + x*gB');
  check_equal(x*gB', test_func_to_egf(n->if(n,n^(n-1)), limit));

  check_equal(gF, test_func_to_egf(n->if(n<1,1, (n+1)^(n-1)), limit),
             "vpar_count() gT");
  check_equal(gF', test_func_to_egf(n-> (n+2)^n, limit-1),
             "vpar_count() gF'");

  check_equal(x*gF' + gF, test_func_to_egf(n-> (n+1)^n, limit),
             "vpar_count() x*gF' + gF");
  check_equal(gT',  x*gF' + gF + L/x,  "gT'");

  check_equal(gF, exp(gT1), "gF");
  check_equal(gF + L/x, (gT-1)/x, "gF 2");
  check_equal(gT, 1 + x*gF + L, "gT");
  check_equal(gT', (x*gF + L)', "gT'");

  \\ Gotthold Eisenstein, "Entwicklung von alpha^alpha^alpha^...", Journal
  \\ fur die Reine und Angewandte Mathematik, volume 28, 1844, pages 49-52.
  \\ (As referred to in Janson, Knuth, Luczak, Pittel, "Birth of the Giant
  \\ Component".)
  \\
  check_equal(gT1, x*exp(gT1) + L, "gT");

  check_equal(1/(2-gT),   test_func_to_egf(n-> n^n, limit), "1/(2-gT)");
  check_equal(1/(1-gT1),  test_func_to_egf(n-> n^n, limit));
  check_equal(1/(1-gT1),  1+x*gT');
  \\ A001865 sum(k=1,n, n! * n^(n-k-1) / (n-k)!)
  \\ print(test_egf_terms(log(1/(2-gT)), 10));

  check_equal(gT', exp(gT-1) / (2-gT)  + L/x, "gT'");
  check_equal(gT1', exp(gT1) / (1-gT1) + L/x);
  check_equal(gT1', gF * (1+x*gT')     + L/x);

  \\----------

  my(gW = test_LambertW(limit));

  \\ T(x) = -W(-x)
  check_equal(gT, 1-subst(gW,x,-x),  "gT in terms of Lambert W");
  check_equal(gT1, -subst(gW,x,-x),  "gT1 in terms of Lambert W");

  \\----------

  \\ print(test_egf_terms(gB^2,limit));
  \\ not in OEIS: 1, 2, 4, 12, 62, 470, 4752, 60368, 925696, 16651818

  \\ gT^2  = 2*(2*n-1) * n^(n-2)
  \\ A089945  (2*n+1) * (n+1)^(n-1)
  check_equal(gT^2, test_func_to_egf(n->if(n==0,1, 2*(2*n-1)*n^(n-2)), limit));

  \\ gT1^2
  \\ A206855  = 2*(n-1)*n^(n-2)
  \\ total degree of root over labelled rooted trees
  check_equal(gT1^2,
              test_func_to_egf(n->if(n==0,0, 2*(n-1)*n^(n-2)), limit+1),
              "gT1^2");

  \\ gT1^3
  \\ A206855  = 2*(n-1)*n^(n-2)
  \\ total degree of root over labelled rooted trees
  check_equal(gT1^2,
              test_func_to_egf(n->if(n==0,0, 2*(n-1)*n^(n-2)), limit+1),
              "gT1^2");

  \\ print(test_egf_terms(gT1^3,limit));
  \\ not in OEIS: 6, 72, 900, 12960, 216090, 4128768, 89282088
  \\ gT1^3 / 6
  \\ A053507 binomial(n-1,2)*n^(n-3)
  \\ 1, 12, 150, 2160, 36015, 688128, 14880348
  check_equal(gT1^3/6,
              test_func_to_egf(n->if(n==0,0,
                                     binomial(n-1,2)*n^(n-3)), limit+2),
              "gT1^3");

  \\ A065888 endofunctions
  \\ not in OEIS: 24, 480, 8640, 164640, 3440640, 79361856
  \\ not A065888 which is 1/4
  check_equal(gT1^4/4, test_func_to_egf(n->if(n==0,0,
                                      (n-3)*(n-2)*(n-1)*n^(n-4)), limit+3));
  \\ A053508
  \\ 1, 20, 360, 6860, 143360, 3306744
  check_equal(gT1^4/24, test_func_to_egf(n->if(n==0,0,
                                         binomial(n-1,3)*n^(n-4)), limit+3),
              "gT1^4, A053508");

  \\ print(test_egf_terms(gT1^4,limit)/24);

  \\ gF^2  = 2*(n+2)^(n-1)
  \\ A007334 spanning trees of complete-n graph contract one edge
  \\       = 2 * n^(n-3)  for n>=2
  \\   number of acyclic funcs 1..n -> 1..n+2
  \\   egf (-W(-x)/x)*exp(-W(-x))
  check_equal(gF^2, test_func_to_egf(n-> 2*(n+2)^(n-1), limit));
}


\\-----------------------------------------------------------------------------
progress("test_vector_run_lengths()");

check_equal(test_vector_run_lengths([]), []);
check_equal(test_vector_run_lengths([1,1]), [2]);
check_equal(test_vector_run_lengths([1,1,0]), [2,1]);
check_equal(test_vector_run_lengths([1,1,0,0,0,1]), [2,3,1]);
check_equal(test_vector_run_lengths([1,1,1,0,1,1,0,0]), [3,1,2,2]);


\\-----------------------------------------------------------------------------
progress("Kreweras Lattice Num Edges");

\\ num ordered forests of n vertices of which k many are roots
test_Ck(n,k) =
{
  n >= 0 || error();
  vpar_INTERNAL_Catalan_triangle(n-1,n-k);
}
{
  for(n=1,8,
     my(vpar=vpar_preorder_first(n),
        count_by_numroots=vector(n));
     until((vpar=vpar_preorder_next(vpar,1))=='none,  \\ forests
           count_by_numroots[vpar_num_roots(vpar)]++);
     check_equal(vector(n,k, test_Ck(n,k)),
                 count_by_numroots,
                 "test_Ck() vs vpar_num_roots()");
     check_equal(vector(n,k, test_Ck(n,k)),
                 vector(n,k, binomial(2*n-1-k,n-1) * k/n)));
}

\\ test_S(n) returns number of possible splits of one siblings set into two
\\ siblings set in an ordered forest of n vertices.
\\ This is the number of edges (covers) in the Kreweras lattice.
test_S(n) =
{
  n>=0 || error("oops S() n="n);
  sum(k=1,n, test_SS(n,k));
}
\\ test_SS(n,k) returns number of possible splits of one siblings set into
\\ two siblings set in an ordered forest of n vertices of which k many are
\\ root vertices.  The root vertices are the set containing 1.
test_SS(n,k) =
{
  \\ 1 ..... c ... n         first partition k terms
  \\   |---| |-----|
  \\ 1  c-2   n-c+1
  \\    any  firsts k-1
  \\  C(c-2)  Ck(n-c+1,k-1)
  \\
  \\ k-1 splits which start from v=1, times Ck(n,k) of forests
  \\ +
  \\ S(c-2) splits in the c-2 sub-forest, times n-c+1,k-1 different second part
  \\ +
  \\ SS(n-c+1,k-1) splits in the second part, being now k-1 roots only,
  \\ and times the Catalan(c-2) different first parts

  n>=0 || error("oops SS() n="n);
  (1 <= k && k <= n) || error("oops SS() n="n" k="k);
  if(k==1, test_S(n-1),
     (k-1) * test_Ck(n,k)
     + sum(c=2,n-k+2,
           test_Ck(n-c+1,k-1)                   * test_S(c-2)
           + vpar_INTERNAL_Catalan_number(c-2)  * test_SS(n-c+1,k-1)));
}
check_equal(test_S(0), 0);
check_equal(test_S(1), 0);
check_equal(test_S(2), 1);   \\ 2 forests, one edge between

test_SS_first_part(n,k) = sum(c=2,n-k+2, test_Ck(n-c+1,k-1) * test_S(c-2));
\\ for(n=1,10, print(vector(n-1,k,k++; test_SS_first_part(n,k))));
\\ for(n=1,10, print(vector(n-1,k,k++; binomial(2*n-k-1,n+1))));
{
  \\ A176992 Triangle T(n,m) = binomial(2n-k+1,n+1) read by rows 0<=k<=n
  for(n=1,12,
     for(k=2,n,
        check_equal(test_SS_first_part(n,k), binomial(2*n-k-1,n+1))));

  my(l=List([]));
  for(n=0,#A176992_samples,
     for(k=0,n,
        #l<#A176992_samples || break(2);
        listput(l, binomial(2*n-k+1,n+1))));
  check_equal(Vec(l), A176992_samples,
              "A176992 binomial(2*n-k+1,n+1)");
}

test_SS_second_part(n,k) = \
  sum(c=2,n-k+2, vpar_INTERNAL_Catalan_number(c-2) * test_SS(n-c+1,k-1));
\\ for(n=1,10, print(vector(n-1,k,k++; test_SS_second_part(n,k))));
\\ for(n=1,10, print(Vecrev(vector(n-1,k,k++; test_SS_second_part(n,k)))));
\\ not in OEIS: 715, 737, 660, 490, 275, 105, 21
\\ not in OEIS: 21, 105, 275, 490, 660, 737, 715

\\ n is the number of siblings in a set.
\\ Return how many ways to split some of it out to a new non-crossing set.
\\ Set member 1 remains in the first set.
\\ The split part is choose start x, end y, for x+1 to y inclusive.
test_setsize_to_num_split(n) = binomial(n,2);
check_equal(test_setsize_to_num_split(1), 0);
check_equal(test_setsize_to_num_split(2), 1);
check_equal(test_setsize_to_num_split(3), 3);  \\ 1,23 3,12 2,13

\\ setsizes is a vector of sibling set sizes, return total possible splits out of them
test_setsizes_to_num_split(setsizes) = \
  sum(i=1,#setsizes, test_setsize_to_num_split(setsizes[i]));

{
  for(n=1,8,
     my(vpar=vpar_preorder_first(n),
        total=0,
        by_firstsize=vector(n));
     until((vpar=vpar_preorder_next(vpar,1))=='none,
           my(sets=vpar_siblings_sets(vpar),
              setsizes=apply(length,sets),
              splits=test_setsizes_to_num_split(setsizes));
           total += splits;
           by_firstsize[setsizes[1]] += splits);

     check_equal(total, test_S(n));
     check_equal(total, binomial(2*n,n-2));
     check_equal(by_firstsize,
                 vector(n,k, test_SS(n,k)));
     );
}

\\ [0]
\\ [0, 1]
\\ [1, 2, 3]
\\ [6, 7, 9, 6]
\\ [28, 28, 30, 24, 10]
\\ [120, 114, 108, 88, 50, 15]
\\ [495, 462, 405, 324, 205, 90, 21]
\\ [2002, 1859, 1551, 1210, 800, 411, 147, 28]
\\
\\ k=n:    1,2,3,6,10 = triangular
\\ k=n-1:  2,9,24,50,90,147 = n*(n+1)^2/2
\\ k=n-2:  7,30,88,205,411
\\ 1,6,28,128,495,2002

\\ not in OEIS: 120, 114, 108, 88, 50, 15
\\ not in OEIS: 15, 50, 88, 108, 114, 120
check_equal(vecsum([120, 114, 108, 88, 50, 15]), binomial(2*6,6-2));
\\
\\ for(n=1,10, print(vector(n,k, test_SS(n,k))));
\\ for(n=1,10, print(Vecrev(vector(n,k, test_SS(n,k)))));

{
  check_equal(vector(10,n,my(k=n); test_SS(n,k)),
              vector(10,n,my(k=n); binomial(n,2)),
              "test_SS(n,n)");

  check_equal(vector(10,n,n++;my(k=n-1); test_SS(n,k)),
              vector(10,n,n++;my(k=n-1); (n-2)*(n-1)^2/2),
              "test_SS(n,n-1)");

  check_equal(vector(10,n,n+=2;my(k=n-2); test_SS(n,k)),
              vector(10,n,n+=2;my(k=n-2); 1/4*(n-2)*(n^3 - 4*n^2 + n + 10)),
              "test_SS(n,n-2)");

  check_equal(vector(10,n,n+=3;my(k=n-3); test_SS(n,k)),
              vector(10,n,n+=3;my(k=n-3);
                     1/12*(n-3)*(n + 2)*(n^3 - 6*n^2 + 5*n + 24)),
              "test_SS(n,n-3)");

  check_equal(vector(10,n,n+=4;my(k=n-4); test_SS(n,k)),
              vector(10,n,n+=4;my(k=n-4);
                     1/48*(n-4)*(n+2)*(n+3)*(n^3 - 8*n^2 + 11*n + 44)),
              "test_SS(n,n-3)");
}
\\ [0, 0, 0, 0, 0, 28, 114, 405, 1210, 3080, 6903, 14014]


test_SS_simplified(n,k) =
{
  n>=0 || error("oops SS() n="n);
  (1 <= k && k <= n) || error("oops SS() n="n" k="k);
  if(k==1, test_S(n-1),
                    (2*n-k-1)!/(n+1)!/(n-k)! * (k-1) * k * (n+1)
     + if(n-k-2>=0, (2*n-k-1)!/(n+1)!/(n-k)! * (n-k-1) * (n-k))
     + sum(c=2,n-k+2,
           vpar_INTERNAL_Catalan_number(c-2) * test_SS(n-c+1,k-1)));
}
{
  for(n=1,12,
     for(k=2,n,
        check_equal(test_SS_simplified(n,k), test_SS(n,k),
                   "test_SS_simplified()")));
}

\\-----------------------------------------------------------------------------
progress("vpar_diameter_num_endpoints()");
\\ vpar_diameter_num_vertices()

check_equal(vpar_diameter_num_endpoints([]),0);
check_equal(vpar_diameter_num_vertices([]),0);
check_equal(vpar_diameter_num_endpoints([0]),1);
check_equal(vpar_diameter_num_vertices([0]),1);

test_diameter_num_endpoints_by_eccentricities(vpar) =
{
  \\ Diameter endpoints are where eccentricity == diameter.
  my(diameter=vpar_diameter(vpar),
     eccs=vpar_eccentricities_vector(vpar));
  sum(v=1,#vpar, eccs[v]==diameter);
}
test_diameter_num_vertices_by_paths(vpar) =
{
  my(diameter=vpar_diameter(vpar),
     seen=vector(#vpar));
  for(u=1,#vpar,
     for(v=1,#vpar,
        if(vpar_path_length(vpar,u,v)==diameter,
           my(path=vpar_path_vertices(vpar,u,v));
           for(i=1,#path,
              seen[path[i]]=1))));
  vecsum(seen);
}

{
  \\       1 2 3 4 5 6 7 8
  my(vpar=[0,1,2,2,2,2,2,2,8,9,10,11,12,12]);
  \\ vpar_view(vpar);
  check_equal(vpar_diameter_num_endpoints(vpar),
              test_diameter_num_endpoints_by_eccentricities(vpar), "12,12");
}
{
  for(n=0,10,
     my(vpar=vpar_make_star(n),
        want_num_endpoints=if(n<3,n,n-1));
     check_equal(vpar_diameter_num_endpoints(vpar),
                 test_diameter_num_endpoints_by_eccentricities(vpar)));
}
{
  \\       1 2 3 4 5 6 7 8
  my(vpar=[0,1,2,1,4]);
  check_equal(vpar_diameter_num_endpoints(vpar), 2);
}
{
  \\       1 2 3 4 5 6 7 8 9
  my(vpar=[0,1,2,1,4,1,6,1,8,9]);
  check_equal(vpar_diameter_num_endpoints(vpar), 4);
}
{
  \\       1 2 3 4 5 6 7 8
  my(vpar=[8,4,4,0,2,2,9,6,6,5]);
  check_equal(test_diameter_num_endpoints_by_eccentricities(vpar), 4,
              "test_diameter_num_endpoints_by_eccentricities() 10");
  check_equal(vpar_diameter_num_endpoints(vpar), 4);
}

{
  my(vpar=[0,1,1,1,4,5]);
  check_equal(vpar_diameter_num_endpoints(vpar), 3);
}
{
  my(vpar=[0,1,2,2,4,5,6]);
  check_equal(vpar_diameter_num_endpoints(vpar), 3);
}
{
  \\       1 2 3 4 5 6 7
  my(vpar=[0,1,2,2,2,2,6,7,8]);
  check_equal(vpar_diameter_num_endpoints(vpar), 5);
}

\\ height 3 is levels 2^0 to 2^3 inclusive is 8 leaf
check_equal(vpar_diameter_num_endpoints(vpar_make_complete_tree(2,3)), 8, \
            "vpar_diameter_num_endpoints() of binary tree");

check_equal(vpar_diameter_num_endpoints([]), 0, \
            "vpar_diameter_num_endpoints() empty");
check_equal(vpar_diameter_num_endpoints([0]), 1, \
            "vpar_diameter_num_endpoints() singleton");
check_equal(vpar_diameter_num_endpoints(vpar_make_star(6)), 5, "star-6");
check_equal(vpar_diameter_num_endpoints(vpar_make_path(2)), 2, "path-2");
check_equal(vpar_diameter_num_endpoints(vpar_make_path(3)), 2, "path-3");
check_equal(vpar_diameter_num_endpoints(vpar_make_path(6)), 2, "path-6");

{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i]);

     my(got =vpar_diameter_num_endpoints(vpar),
        want=test_diameter_num_endpoints_by_eccentricities(vpar));
     \\ if(got!=want,print(vpar);vpar_view(vpar));
     check_equal(got,want, "vpar_diameter_num_endpoints() random");

     \\ per addhelp, ends are leaves
     my(num_leaves=vpar_num_leaves(vpar));
     check_equal(got <= num_leaves, 1,
            "vpar_diameter_num_endpoints() <= vpar_num_leaves()");

     \\ when leaves are all same ecc then all of them are diameter endpoints,
     \\ and conversely when they are all diameter endpoints they are all same
     \\ ecc (the diameter)
     my(eccentricities=vpar_eccentricities_vector(vpar),
        leaves=vpar_leaves(vpar),
        leaves_eccenticities=vecextract(eccentricities,leaves),
        leaves_eccsame=test_vector_is_all_same(leaves_eccenticities));
     check_equal(leaves_eccsame, got == num_leaves,
                 "vpar_diameter_num_endpoints() == vpar_num_leaves() equiv to leaves all same ecc");

     my(got =vpar_diameter_num_vertices(vpar),
        want=test_diameter_num_vertices_by_paths(vpar));
     \\ if(got!=want,print(vpar);vpar_view(vpar));
     check_equal(got,want, "vpar_diameter_num_vertices() random"));
}


\\-----------------------------------------------------------------------------
progress("test_untriangular_floor()");

{
  for(n=0,100,
     my(t=test_untriangular_floor(n));
     check_equal(test_triangular(t) <= n, 1);
     check_equal(test_triangular(t+1) > n, 1));

  check_equal(vector(#A003056_samples,n,n--; test_untriangular_floor(n)),
              A003056_samples,  \\ OFFSET=0
              "A003056 n appears n+1 times = test_untriangular_floor()");
}


\\-----------------------------------------------------------------------------
progress("vpar_Laplacian_matrix()");

check_equal(vpar_Laplacian_matrix([0,1]), [1,-1; -1,1], \
            "vpar_Laplacian_matrix()");

{
  my(vpar=[3,3,10,10,9,8,8,9,10, 0]);     \\ Knuth 2.3.4.4
  check_equal(vpar_Laplacian_matrix(vpar),
              matdiagonal(vpar_degrees_vector(vpar))
                - vpar_adjacency_matrix(vpar));
}

check_equal(Vec(Mat([123])), [ [123]~ ]);
check_equal(vecsum([ [123]~ ]), [123]~);
check_equal(vecsum([123]~), 123);

check_equal(Vec(Mat([])), [ ]);

test_matsum(m) = m=Vec(m); if(#m,m=vecsum(m)); vecsum(m);
check_equal(test_matsum(Mat([])), 0);
check_equal(test_matsum(Mat([123])), 123);
check_equal(test_matsum(Mat([1,2;3,4])), 10);

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        m = vpar_Laplacian_matrix(vpar));

     check_equal(m, matdiagonal(vpar_degrees_vector(vpar))
                    - vpar_adjacency_matrix(vpar),
     "vpar_Laplacian_matrix() from vpar_adjacency_matrix() per addhelp");

     check_equal(test_matsum(apply(s->s<0,m)),
                 2*vpar_num_edges(vpar),
                 "vpar_Laplacian_matrix() num negatives");

     if(#vpar,
        my(v=vpar_random_vertex(vpar),
           r=m[^v,^v]);
        \\ print(matdet(r)" "vpar_num_roots(vpar)==1);
        check_equal(matdet(r), vpar_is_tree(vpar),
                    "vpar_Laplacian_matrix() matdet per addhelp");
        );

     \\ print(matrank(m)" "#vpar-vpar_num_roots(vpar)"  "vpar);
     check_equal(matrank(m), #vpar-vpar_num_roots(vpar),
                 "vpar_Laplacian_matrix() matrank vs vpar_num_roots()");
     check_equal(matrank(m), vpar_num_edges(vpar),
                 "vpar_Laplacian_matrix() matrank vs vpar_num_edges()");

     \\ Signed and signless Laplacian matrix have same spectrum because they
     \\ are similar, by matrix D = diagonal +/-1 sign according as which
     \\ bipartite side each vertex.
     \\
     my(depths=vpar_depths_vector(vpar),
        D=matdiagonal(apply(d->if(d%2,1,-1), depths)));
     check_equal(abs(m), D*m*D^-1);
  );
}

\\ m is a matrix.  Return m with rows i and j and columns i and j deleted.
test_matrix_delete_two_rowcolumns(m,i,j) =
{
  if(i>i, m=m[^i,^i]; m[^j,^j],
          m=m[^j,^j]; m[^i,^i]);
}
{
  \\ Damir Vukicevic and Ivan Gutman, "Laplacian Matrix and Distance In
  \\ Trees", Kragujevac J. Math. 26 (2004), pages 19-24.
  \\ Distinct vertices i,j, Laplacian delete both those rows and columns,
  \\ matdet = path length from i to j.
  \\ Bapat and Kulkarni: This is also the number of 2-component spanning
  \\ forests where i,j in different components of that forest.

  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i],
        m = vpar_Laplacian_matrix(vpar));

     my(path_lengths=vpar_path_lengths_matrix(vpar));
     for(i=1,#vpar,
        for(j=i+1,#vpar,
           check_equal(matdet(test_matrix_delete_two_rowcolumns(m,i,j)),
                       path_lengths[i,j],
                       "vpar_Laplacian_matrix() matdet for path length")));

     \\ Gutman and Xiao, "Distance In Trees and Laplacian Matrix"
     check_equal(m*path_lengths*m,  -2*m,
                 "vpar_Laplacian_matrix() L*D*L = -2L");
  );
}

\\----

test_Wiener_index_by_Laplacian(vpar) =
{
  my(m = vpar_Laplacian_matrix(vpar),
     p = charpoly(m),
     n = poldegree(p));
  n==#vpar || error();
  polcoeff(p,0)==0 || error("oops, expected low coeff ==0");
  p /= variable(p);
  polcoeff(p,0)!=0 || error("oops, expected second low coeff !=0");
  n * vecsum(apply(x->1/x, polroots(p)));
}
test_Wiener_index_by_Laplacian_addhelp(vpar) =
{
 \\ #vpar*vecsum(polroots(Polrev(Vec(charpoly(vpar_Laplacian_matrix(vpar))/'x))))
 \\  ;

  \\ my(v=polroots(charpoly(vpar_Laplacian_matrix(vpar))));
  \\ #vpar*sum(i=1,#v, if(v[i],1/v[i]));

  my(v=mateigen(vpar_Laplacian_matrix(vpar),1)[1]);
  #vpar*sum(i=1,#v, if(v[i],1/v[i]));
}
{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i]);
     if(#vpar==0,next());

     my(W =vpar_Wiener_index(vpar),
        WL=test_Wiener_index_by_Laplacian(vpar),
        WLH=test_Wiener_index_by_Laplacian_addhelp(vpar));
     check_equal(abs(W-WL) < 1e-12, 1,
                 "test_Wiener_index_by_Laplacian()");

     \\ print(W);
     \\ print(WLH);
     check_equal(abs(W-WLH) < 1e-10, 1,
                 "test_Wiener_index_by_Laplacian_addhelp()");
  );
}

{
  \\ count of how many vpar_Laplacian_matrix() occur, per the addhelp
  for(forest=0,1,
     for(n=0, test_limit_trees_search - forest,
        my(l=List([])); \\ distinct matrices
        forvec(vec=vpar_forvec_from(n,forest), \\ forests
              my(vpar=vpar_from_upascend(vec),
                 m=vpar_Laplacian_matrix(vpar));
              my(pos=setsearch(l,m,1));
              if(pos, listinsert(l,m,pos)));

        check_equal(#l, vpar_count_unrooted(n,forest),
                    "vpar_Laplacian_matrix() count distinct")));
}


\\-----------------------------------------------------------------------------
progress("vpar_to_balanced_binary()");

check_equal(vpar_to_balanced_binary([]), 0, \
            "vpar_to_balanced_binary() empty");
check_equal(vpar_to_balanced_binary([0]), test_from_binary(10), \
            "vpar_to_balanced_binary() singleton");
check_equal(vpar_to_balanced_binary([0,1]), test_from_binary(1100), \
            "vpar_to_balanced_binary() path-2");
check_equal(vpar_to_balanced_binary([0,1,2,3]), test_from_binary(11110000), \
            "vpar_to_balanced_binary() path-4");
check_equal(vpar_to_balanced_binary([0,1,2,2]), test_from_binary(11101000), \
            "vpar_to_balanced_binary() claw");

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got=vpar_to_balanced_binary(vpar));
     check_equal(vpar_is_balanced_binary(got), 1,
                "vpar_to_balanced_binary() vs test_is_balanced_binary()");
     check_equal(#binary(got), 2*#vpar,
                "vpar_to_balanced_binary() binary length");
     check_equal(if(got,1+logint(got,2),0), 2*#vpar,
                "vpar_to_balanced_binary() logint length");
  );
}

\\ like cmp() actually
test_lex_by_length_first(vec1,vec2) = \
  if(#vec1==#vec2, lex(vec1,vec2), lex(#vec1,#vec2));

{
  \\ addhelp() describing b increases with increasing preorder lex()
  for(forest=0,1,
     my(prev_b=-1, prev_vpar);
     for(n=0,8-forest,
        my(vpar=vpar_preorder_last(n,forest));
        until((vpar=vpar_preorder_prev(vpar,forest))=='none,  \\ increasing lex()
              my(b=vpar_to_balanced_binary(vpar));
              \\ print("n="n"  "prev_b" "b"  "test_to_binary(prev_b)" "test_to_binary(b));
              \\ print("  "prev_vpar"  "vpar);
              check_equal(prev_b < b, 1,
                          "vpar_to_balanced_binary() is increasing b with increasing lex()");
              if(n,
                 check_equal(test_lex_by_length_first(prev_vpar,vpar),  -1,
                             "vpar_to_balanced_binary() increasing test_lex_by_length_first()"));
              prev_vpar=vpar;
              prev_b=b;
)));
}

{
  \\ complete binary tree as balanced binary
  check_equal(vector(4,h,h--; vpar_to_balanced_binary(vpar_make_complete_tree(2,h))),
              [2, 52, 14952, 1026847952]);
  \\ not in OEIS: 2, 52, 14952, 1026847952
  \\ not in OEIS: 1, 26, 7476, 513423976

  check_equal(vector(4,h,h--; test_to_binary(vpar_to_balanced_binary(vpar_make_complete_tree(2,h)))),
              [10, 110100, 11101001101000, 111101001101000111010011010000]);
  \\ not in OEIS: 10, 110100, 11101001101000, 111101001101000111010011010000
  \\ not in OEIS: 1, 11010, 1110100110100, 11110100110100011101001101000

  for(h=1,4,
     my(vpar=vpar_make_complete_tree(2,h),
        want=vpar_to_balanced_binary(vpar),
        vpar_prev=vpar_make_complete_tree(2,h-1),
        prev=vpar_to_balanced_binary(vpar_prev));

     \\ binary:  1 prev  prev 0
     check_equal(concat([1, binary(prev), binary(prev), 0]),
                 binary(want));
     check_equal(#binary(want), 2^(h+2)-2);

     check_equal(test_to_binary((1 << (2^(h+2)-3)) + (prev << (2^(h+1)-1)) + (prev<<1)),
                 test_to_binary(want)));
}

test_highest_0s_run_extents(n) =
{
  my(pos=0, hi=-1, lo=0);
  while(1,
        while(bitand(n,1), n>>=1; pos++);   \\ strip low 1s
        n || break;
        lo=pos;   \\ lowest 0
        while(!bitand(n,1), hi=pos; n>>=1; pos++));  \\ count and strip low 0s
  [hi,lo];
}
check_equal(test_highest_0s_run_extents(test_from_binary(0)), [-1,0]);
check_equal(test_highest_0s_run_extents(test_from_binary(1)), [-1,0]);
check_equal(test_highest_0s_run_extents(test_from_binary(10)), [0,0]);
check_equal(test_highest_0s_run_extents(test_from_binary(111000111)), [5,3]);
check_equal(test_highest_0s_run_extents(test_from_binary(1100101011)), [7,6]);

test_highest_0s_run_length(n) =
{
  my(ret=0);
  while(1,
        while(bitand(n,1), n>>=1);   \\ strip low 1s
        n || break;
        ret=0;
        while(!bitand(n,1), n>>=1; ret++));  \\ count and strip low 0s
  ret;
}
check_equal(test_highest_0s_run_length(test_from_binary(0)), 0);
check_equal(test_highest_0s_run_length(test_from_binary(1)), 0);
check_equal(test_highest_0s_run_length(test_from_binary(10)), 1);
check_equal(test_highest_0s_run_length(test_from_binary(100)), 2);
check_equal(test_highest_0s_run_length(test_from_binary(10001)), 3);
check_equal(test_highest_0s_run_length(test_from_binary(111000111)), 3);
check_equal(test_highest_0s_run_length(test_from_binary(1100101011)), 2);

{
  \\ balanced binary total num 0s in highest bit run 0001
  \\ A014138 sum Catalans
  my(forest=1);
  for(n=2,min(8,#A014138_samples),
     my(vpar=vpar_preorder_first(n),
        total=0);
     until((vpar=vpar_preorder_next(vpar,forest))=='none,  \\ increasing lex()
           my(b=vpar_to_balanced_binary(vpar));
           total += test_highest_0s_run_length(b));
     check_equal(total,
                 A014138_samples[n+1],  \\ OFFSET=0
                 "A014138 total 0s highest run 0s of balanced binary"));

  check_equal(vector(#A014138_samples,n,n--;
                     sum(k=1,n, vpar_INTERNAL_Catalan_number(k))),
              A014138_samples,  \\ OFFSET=0
              "A014138 sum Catalans");
}
{
  check_equal(vector(#A284968_samples,n,
                     sum(k=1,n, vpar_INTERNAL_Catalan_number(k)) - n),
              A284968_samples,  \\ OFFSET=1
              "A284968 sum Catalans - n");
}

{
  \\ Knuth fasc4a.ps exercise 11.
  \\ Reversing the order of all siblings gives Dyck word with left/right
  \\ parens reversed.
  for(n=1,5,
     my(vpar=vpar_preorder_first(n));
     until((vpar=vpar_preorder_next(vpar,1))=='none,  \\ increasing lex()
           my(b=vpar_to_balanced_binary(vpar),
              vparrev=vpar_relabel_reverse(vpar));
           check_equal(vpar_to_balanced_binary(vparrev),
                       test_bit_complement_and_reverse(b))))
}


\\-----------------------------------------------------------------------------
progress("vpar_is_neighbour()");

check_equal(vpar_is_neighbour([0],1,1), 0);
check_equal(vpar_is_neighbour([1],1,1), 1, "vpar_is_neighbour() self-loop");
check_equal(vpar_is_neighbour([1,1,1],1,2), 1);
check_equal(vpar_is_neighbour([1,1,1],2,3), 0);

check_equal(vpar_is_neighbour([0,0,2,3,4,5],2,3),1);
check_equal(vpar_is_neighbour([0,0,2,3,4,5],3,2),1);
check_equal(vpar_is_neighbour([0,0,2,3,4,5],2,4),0);
check_equal(vpar_is_neighbour([0,0,2,3,4,5],1,2),0);


\\-----------------------------------------------------------------------------
progress("vpar_root_of_vertex_vector()");

check_equal(vpar_root_of_vertex_vector([]), []);
check_equal(vpar_root_of_vertex_vector([0]), [1]);
check_equal(vpar_root_of_vertex_vector([0,1]), [1,1]);
check_equal(vpar_root_of_vertex_vector([2,0]), [2,2]);
check_equal(vpar_root_of_vertex_vector([0,3,0]), [1,3,3]);

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        got=vpar_root_of_vertex_vector(vpar));
     check_equal(got, vector(#vpar,v, vpar_root_of_vertex(vpar,v)),
       "vpar_root_of_vertex_vector() vs individual vpar_root_of_vertex()"));
}

\\ old code for acyclics only
test_root_of_vertex_vector_of_tree(vpar) =
{
  \\ roots[v] = the root of v, or 0 if not yet determined.
  \\ At each vertex, follow up its ancestors to find the root or an ancestor
  \\ already determined in roots[].
  \\ Then re-traverse upwards to fill in.

  my(roots=vector(#vpar));
  for(i=1,#vpar,
     if(!roots[i],
        my(v=i,r);
        until(!(v=vpar[r=v]), if(r=roots[v],break));
        v=i;
        until(!(v=vpar[v]) || roots[v], roots[v]=r)));
  roots;
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     check_equal(vpar_root_of_vertex_vector(vpar),
                 test_root_of_vertex_vector_of_tree(vpar),
       "vpar_root_of_vertex_vector() vs test_root_of_vertex_vector_of_tree()"));
}

{
  \\ count distinct vpar_root_of_vertex_vector() over forests, per the addhelp
  \\ A000248
  my(forest=1);
  for(n=0, test_limit_trees_search - forest,
     my(l=List([])); \\ distinct vecs seen
     forvec(vec=vpar_forvec_from(n,forest), \\ forests
           my(vpar=vpar_from_upascend(vec),
              roots=vpar_root_of_vertex_vector(vpar));
           my(pos=setsearch(l,roots,1));
           if(pos, listinsert(l,roots,pos)));

     check_equal(#l,
                 A000248_samples[n +1],  \\ OFFSET=0
                 "A000248 vpar_root_of_vertex_vector() occurring forests");
     check_equal(#l, vpar_count_height_le(n,1,1),
     "vpar_root_of_vertex_vector() occurring forests, vpar_count_height_le()");

     \\ k roots, rest under them
     check_equal(#l, sum(k=0,n, binomial(n,k)*k^(n-k)),
                 "vpar_root_of_vertex_vector() occurring forests");

     \\ or k non-roots, swapping k <-> n-k
     check_equal(#l, sum(k=0,n, binomial(n,k)*(n-k)^k),
                 "vpar_root_of_vertex_vector() occurring forests");
  );
}

test_A080108(n) = sum(k=1,n, binomial(n-1,k-1)*k^(n-k));
{
  check_equal(vector(#A080108_samples,n, test_A080108(n)), \\ OFFSET=1
              A080108_samples,  \\ OFFSET=1
              "A080108 test_A080108(n)");
}
{
  \\ count distinct vpar_root_of_vertex_vector() over functional digraphs,
  \\ per the addhelp
  \\ A080108

  for(n=0,6,
     my(l=List([])); \\ distinct vecs seen
     forvec(vpar=vector(n,i,[0,n]),  \\ functional digraphs
           my(roots=vpar_root_of_vertex_vector(vpar));
           my(pos=setsearch(l,roots,1));
           if(pos, listinsert(l,roots,pos)));

     check_equal(#l, test_A080108(n+1),
       "A080108 vpar_root_of_vertex_vector() occurring functional digraphs");

     \\ choose k roots, rest are under one of those k or are in a cycle (so
     \\ no root), so (k+1)^(n-k)
     check_equal(#l, sum(k=0,n, binomial(n,k)*(k+1)^(n-k)),
       "vpar_root_of_vertex_vector() occurring functional digraphs");

     \\ or k non-roots, swapping k <-> n-k
     check_equal(#l, sum(k=0,n, binomial(n,k)*(n-k+1)^k),
       "vpar_root_of_vertex_vector() occurring functional digraphs");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_root_of_vertex()");

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     for(v=1,#vpar,
        my(got=vpar_root_of_vertex(vpar,v));
        check_equal(got==v, vpar[v]==0,
                   "vpar_root_of_vertex() root=v at root")));
}

{
  for(n=1,5,
     my(vpar=vpar_make_path(n));
     for(v=1,n,
        check_equal(vpar_root_of_vertex(vpar,v), 1,
                    "vpar_root_of_vertex() on path")));
  for(n=1,5,
     my(vpar=test_make_cycle(n));
     for(v=1,n,
        check_equal(vpar_root_of_vertex(vpar,v), 0,
                    "vpar_root_of_vertex() on cycle")));
}


\\-----------------------------------------------------------------------------

{
  \\ R. B. Bapat and S. Sivasubramanian, "Product Distance Matrix of a Graph
  \\ and Squared Distance Matrix of a Tree", Applicable Analysis and
  \\ Discrete Mathematics, volume 7, number 2, October 2013, pages 285-301.
  \\ http://www.jstor.org/stable/43660717

  \\        1
  \\        |
  \\   3 -- 2 -- 5 -- 6
  \\        |    |
  \\        4    7
  my(vpar=[0,1,2,2,2, 5,5]);
  vpar_validate(vpar);
  check_equal(apply(sqr, vpar_path_lengths_matrix(vpar)),
              [0,1,4,4,4,9,9;
               1,0,1,1,1,4,4;
               4,1,0,4,4,9,9;
               4,1,4,0,4,9,9;
               4,1,4,4,0,1,1;
               9,4,9,9,1,0,4;
               9,4,9,9,1,4,0]);
}


\\-----------------------------------------------------------------------------
progress("logint()");

{
  for(b=2,10,
     for(i=1,b^4,
        check_equal(#digits(i,b), logint(i,b)+1,
                    "logint() digits length")));

  for(i=1,256,
     my(l=logint(i,2));
     check_equal(#binary(i), l+1,
                 "logint() bit length");
     check_equal(bittest(i,l), 1);
     for(a=l+1,l+5,
        check_equal(bittest(i,a), 0)));
}


\\-----------------------------------------------------------------------------
progress("vpar_indpoly() and vpar_perfect_dompoly() both all real roots");

{
  for(n=0,8,
     my(vpar=vpar_free_first(n),
        l=List([]));
     until((vpar=vpar_free_next(vpar))=='none,
           if(test_poly_is_all_real_roots(vpar_indpoly(vpar))
              && test_poly_is_all_real_roots(vpar_perfect_dompoly(vpar)),
              listput(l,vpar)));
     l=Vec(l);
     \\ print(l);
     if(n==0, check_equal(l,[ [] ]),        \\ empty
        n==1, check_equal(l,[ [0] ]),       \\ singleton
        n==2, check_equal(l,[ [0,1] ]),     \\ path-2
        n==3, check_equal(l,[ [0,1,1] ]),   \\ path-3
        n==4, check_equal(#l, 0),           \\ none 4
        n==5, check_equal(l,[ [0,1,2,1,1] ]),   \\ bistar 2,3
        n==6, check_equal(l,[ [0,1,2,2, 1,1] ]),   \\ bistar 3,3

        \\         7
        \\         |               n=8 only
        \\ 3---2---1---5---6
        \\     |   |
        \\     4   8
        n==8, check_equal(l,[ [0,1,2,2, 1,5,1,1] ]),
        );
  );
}


\\-----------------------------------------------------------------------------
progress("test_is_path()");

check_equal(test_is_path([]), 1);
check_equal(test_is_path([0]), 1);
check_equal(test_is_path([2,0]), 1);
check_equal(test_is_path([0,1,2,3]), 1, "test_is_path()");
check_equal(test_is_path([2,3,0]), 1);
check_equal(test_is_path([0,1,1]), 0);
\\ cyclics not recognised
\\ check_equal(test_is_path([1]), 0);
\\ check_equal(test_is_path([2,1]), 0);

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     check_equal(test_is_path(vpar),
                 vpar_is_isomorphic(vpar, vpar_make_path(#vpar)),
                 "test_is_path() vs vpar_is_isomorphic()"));
}


\\-----------------------------------------------------------------------------
progress("test_is_free_path()");

check_equal(test_is_free_path([]), 1);
check_equal(test_is_free_path([0]), 1);
check_equal(test_is_free_path([2,0]), 1);

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     check_equal(test_is_free_path(vpar),
                 vpar_is_free_isomorphic(vpar, vpar_make_path(#vpar)),
                 "test_is_free_path() vs vpar_is_isomorphic()"));
}


\\-----------------------------------------------------------------------------
progress("test_is_free_star()");

check_equal(test_is_free_star([]), 1);
check_equal(test_is_free_star([0]), 1);
check_equal(test_is_free_star([2,0]), 1);
check_equal(test_is_free_star([0,1,1]), 1);
check_equal(test_is_free_star([4,4,4,0]), 1);
check_equal(test_is_free_star([4,4,4,5,6,0]), 0);

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     check_equal(test_is_free_star(vpar),
                 vpar_is_free_isomorphic(vpar, vpar_make_star(#vpar)),
                 "test_is_free_star() vs vpar_is_isomorphic()"));
}


\\-----------------------------------------------------------------------------
progress("vpar_count_updown()");

{
  check_equal(vpar_count_updown(5,1), 120);
  check_equal(iferr(vpar_count_updown(5,'forest);0, e, 1),
              1,
  "vpar_count_updown() error on forest=symbol (whereas it should be 0,1)");
}


\\-----------------------------------------------------------------------------
progress("vpar_make_random_downwards()");

{
  vpar_make_random_downwards(5,1);
  check_equal(iferr(vpar_make_random_downwards(5,'forest);0, e, 1),
              1,
              "vpar_make_random_downwards() error on forest=symbol (whereas it should be 0,1)");
}

\\ vpar_make_random_downwards() but calling to test_random()
test_vpar_make_random_downwards = \
  eval(test_random_to_test_random(vpar_make_random_downwards));
{
  for(forest=0,1,
     for(n=1,10-forest,
        my(count=vpar_count_updown(n));
        for(rep=1,min(n,3),
           my(i=random(count) + 1,  \\ 1 to count, inclusive
              vpar=vpar_downwards_ith(n,i,forest));
           \\ print("n="n" count="count"  i="i" "vpar);

           test_random_list      =List([]);
           test_random_limit_list=List([]);
           if(!forest, check_equal(vpar_is_tree(vpar), 1));
           if(#vpar, check_equal(vpar[1], 0));
           for(v=2,#vpar,
              \\ tree 1 to v-1, forest 0 to v-1
              listput(test_random_list,       vpar[v] - !forest);
              listput(test_random_limit_list, v - !forest));

           my(got=test_vpar_make_random_downwards(n,forest));
           check_equal(#test_random_list, 0, "test_random_list consumed");
           check_equal(#test_random_limit_list, 0, "test_random_limit_list consumed");
           check_equal(got, vpar, "vpar_make_random_downwards() from test_random()");
        )));
}


\\-----------------------------------------------------------------------------
progress("vpar_downwards_to_i()");

check_equal(vpar_downwards_to_i([]), 1);
check_equal(vpar_downwards_to_i([],1), 1);

check_equal(vpar_downwards_to_i([0]), 1, "vpar_downwards_to_i() singleton");
check_equal(vpar_downwards_to_i([0],1), 1);

check_equal(vpar_downwards_to_i([0,1]), 1);
check_equal(vpar_downwards_to_i([0,1],1), 1);
check_equal(vpar_downwards_to_i([0,0],1), 2, "vpar_downwards_to_i() 2 singletons");

check_equal(vpar_downwards_to_i([0,1,2]), 1, "vpar_downwards_to_i() path-3");
check_equal(vpar_downwards_to_i([0,1,2],1), 1);
check_equal(vpar_downwards_to_i([0,1,1]), 2);
check_equal(vpar_downwards_to_i([0,1,1],1), 2);

{
  check_equal(vpar_downwards_to_i([0,1,2,3,4],1), 1);
  check_equal(iferr(vpar_downwards_to_i([0,1,2,3,4],'forest);0, e, 1),
              1,
              "vpar_downwards_to_i() error on forest=symbol (whereas it should be 0,1)");
}


\\-----------------------------------------------------------------------------
progress("vpar_downwards_ith()");

check_equal(vpar_downwards_ith(0,1), []);
check_equal(vpar_downwards_ith(0,2), 'none);

check_equal(vpar_downwards_ith(1,1), [0]);
check_equal(vpar_downwards_ith(1,2), 'none);
check_equal(vpar_downwards_ith(1,1,1), [0]);
check_equal(vpar_downwards_ith(1,2,1), 'none);

check_equal(vpar_downwards_ith(2,1), [0,1]);
check_equal(vpar_downwards_ith(2,2), 'none);
check_equal(vpar_downwards_ith(2,1,1), [0,1]);
check_equal(vpar_downwards_ith(2,2,1), [0,0]);
check_equal(vpar_downwards_ith(2,3,1), 'none);

{
  check_equal(vpar_downwards_ith(5,1,1), [0,1,2,3,4]);
  check_equal(iferr(vpar_downwards_ith(5,1,'forest);0, e, 1),
              1,
              "vpar_downwards_ith() error on forest=symbol (whereas it should be 0,1)");
}

{
  for(forest=0,1,
     for(n=0, test_limit_trees_search - forest,
        my(l=List([]),
           vpar);
        for(i=1,(n+3)!,
           vpar=vpar_downwards_ith(n,i,forest);
           if(vpar=='none,break());
           vpar_validate(vpar);
           listput(l,vpar);
           check_equal(vpar_downwards_to_i(vpar,forest), i,
                       "vpar_downwards_ith() vs vpar_downwards_to_i()");
        );
        check_equal(vpar, 'none,
                    "vpar_downwards_ith() stop at 'none");
        check_equal(#l, vpar_count_updown(n,forest),
                    "vpar_downwards_ith() count vs vpar_count_updown()");

        l=Set(l);
        forvec(vec=vpar_forvec_from(n,forest), \\ forests
              my(vpar=vpar_from_upascend(vec));
              check_equal(setsearch(l,vpar)!=0,
                          vpar_is_downwards(vpar),
                          "vpar_downwards_ith() all and only vpar_is_downwards()"));
     ));
}

test_relabel_func(vpar,func) =
{
  my(new_vpar=vpar);
  for(v=1,#vpar,
     new_vpar[func(v)] = if(vpar[v],func(vpar[v])));
  new_vpar;
}
test_relabel_rotate(vpar,offset=1) = \
  test_relabel_func(vpar, v->if((v+=offset)>#vpar,v-#vpar,v));
\\   1      2
\\   |  =>  |
\\   2      3
\\   |      |
\\   3      1
check_equal(test_relabel_rotate([0,1,2],1), [3,0,2]);

{
  \\ vpar_downwards_ith() tree n is forest n-1 with root above
  for(n=1,5,
     for(i=1,vpar_count_updown(n),
        my(vpar_tree=vpar_downwards_ith(n,i),
           vpar_forest=vpar_downwards_ith(n-1,i,1));
        check_equal(vpar_tree,
                    test_relabel_rotate(vpar_add_root_above_forest(vpar_forest),1),
                    "vpar_downwards_ith() tree n as forest n-1 with root above");
     ));
}

{
  \\ A225824 upwards labelled trees which are asymmetric as unlabelled
  my(limit=9,
     got=vector(limit,n,
                sum(i=1,vpar_count_updown(n),
                    my(vpar=vpar_downwards_ith(n,i));
                    vpar_is_asymmetric(vpar))));
  check_equal(got, A225824_samples[1..limit],
              "A225824 upwards labelled which are asymmetric as unlabelled");

  \\ A032301 upwards labelled trees where all siblings are different size
  \\ subtrees.
  \\ Per Geoffrey Critzer in A225824, this is the same as asymmetric for n=1..8,
  \\ but differs at n=9 where there are two asymmetric trees of same size 4,
  \\
  \\       *
  \\     /   \
  \\    *     *
  \\    |    / \
  \\    *   *   *
  \\    |       |
  \\    *       *
  \\    |
  \\    *
  \\
  my(eight=min(8,limit));
  check_equal(got[1..eight], A032301_samples[1..eight]);
  check_equal(A225824_samples[9] != A032301_samples[9], 1);
  check_equal(vector(4,n,vpar_count_unlabelled_asymmetric(n)), [1,1,1,2])
}

\\ Return 1 if all siblings have different size subtrees below.
\\ Roots of a forest are reckoned as siblings.
test_sibling_subtrees_different_sizes(vpar) =
{
  my(sizes=vpar_subtree_sizes_vector(vpar),
     children=vpar_children_vector(vpar));
  for(p=1,#vpar,
     test_vector_is_all_different(vecextract(sizes, children[p]))
       || return(0));
  test_vector_is_all_different(vecextract(sizes, vpar_roots(vpar)));
}
{
  \\ A032301 upwards labelled trees where all siblings have different size
  \\ subtrees
  \\
  my(limit=9,
     got=vector(limit,n,
                sum(i=1,vpar_count_updown(n),
                    my(vpar=vpar_downwards_ith(n,i));
                    test_sibling_subtrees_different_sizes(vpar))));
  check_equal(got, A032301_samples[1..limit],
              "A032301 upwards labelled all siblings different size subtrees");
}{
  \\ egf relation per Geoffrey Critzer in A032301
  my(limit=#A032301_samples,
     L = O(x^limit),
     gA032301 = test_func_to_egf(n->if(n,A032301_samples[n]), limit));
  check_equal(gA032301',
              prod(n=1,limit, 1 + A032301_samples[n]*x^n/n! + L/x));
}
\\ cf.
\\ upwards labelled forests where all siblings have different size subtrees,
\\ but the roots not reckoned as siblings
\\ not in OEIS: 1, 1, 2, 5, 18, 74, 372, 2160, 14476

check_equal(permtonum([1,2,3]), 0);
check_equal(permtonum([1,3,2]), 1);
check_equal(permtonum([2,1,3]), 2);
check_equal(permtonum([2,3,1]), 3);
check_equal(permtonum([3,1,2]), 4);
check_equal(permtonum([3,2,1]), 5);

\\ gp 2.11 incompatible change numtoperm() returns Vecsmall
check_equal(Vec(numtoperm(3,0)), [1,2,3]);
check_equal(Vec(numtoperm(3,1)), [1,3,2]);
check_equal(Vec(numtoperm(3,5)), [3,2,1]);
{
  \\ mapping perm -> downwards forest in vpar.gp "Upwards and Downwards"
  \\ intro comments
  for(n=0,4,
     for(i=0,n!-1,
        my(perm=numtoperm(n,i),
           vpar = vpar_downwards_ith(#perm, permtonum(perm)+1, 1));
        vpar_validate(vpar);
        check_equal(vpar_is_downwards(vpar), 1);

        my(perm_again=numtoperm(#vpar, vpar_downwards_to_i(vpar,1)-1));
        check_equal(perm_again, perm, "perm_again");
     ));
}
{
  \\ vpar.gp initial comments "Upwards and Downwards Trees" on perm to vpar
  \\ numtoperm() is Vec in 2.7.x and Vecsmall in 2.11.x, but in either case
  \\ get back etc the same
  my(perm=numtoperm(5,1));
  check_equal(Vec(perm), [1,2,3,5,4]);
  my(vpar = vpar_downwards_ith(#perm, permtonum(perm)+1, 1));
  check_equal(vpar, [0, 1, 2, 3, 3]);
  my(perm_again = numtoperm(#vpar, vpar_downwards_to_i(vpar,1)-1));
  check_equal(perm_again, perm, "perm_again 1");

  perm=numtoperm(5,51);
  check_equal(Vec(perm), [3,1,4,5,2]);
  check_equal(test_is_perm(perm), 1);
  vpar = vpar_downwards_ith(#perm, permtonum(perm)+1, 1);
  check_equal(vpar, [0, 1, 0, 1, 3]);
  perm_again = numtoperm(#vpar, vpar_downwards_to_i(vpar,1)-1);
  check_equal(perm_again, perm, "perm_again 2");
}


\\-----------------------------------------------------------------------------
progress("vpar_preorder_to_i()");

{
  check_equal(vpar_preorder_to_i([0,1,2,3,4],1), 1);
  check_equal(iferr(vpar_preorder_to_i([0,1,2,3,4],'forest);0, e, 1),
              1,
              "vpar_preorder_to_i() error on forest=symbol (whereas it should be 0,1)");
}

check_equal(vpar_preorder_to_i([0,0,0,0,0],1), vpar_count_ordered(5,1), \
            "vpar_preorder_to_i() 5 singletons");

check_equal(vpar_preorder_to_i([]), 1, "vpar_preorder_to_i() empty");
check_equal(vpar_preorder_to_i([],1), 1);

check_equal(vpar_preorder_to_i([0]), 1, "vpar_preorder_to_i() singleton");
check_equal(vpar_preorder_to_i([0],1), 1);

check_equal(vpar_preorder_to_i([0,1]), 1, "vpar_preorder_to_i() path-2");
check_equal(vpar_preorder_to_i([0,1],1), 1);
check_equal(vpar_preorder_to_i([0,0],1), 2, "vpar_preorder_to_i() 2 singletons");

check_equal(vpar_preorder_to_i([0,1,2]), 1, "vpar_preorder_to_i() path-3");
check_equal(vpar_preorder_to_i([0,1,2],1), 1);
check_equal(vpar_preorder_to_i([0,1,1]), 2);
check_equal(vpar_preorder_to_i([0,1,1],1), 2);


\\-----------------------------------------------------------------------------
progress("vpar_preorder_ith()");

check_equal(vpar_preorder_ith(0,1), []);
check_equal(vpar_preorder_ith(0,2), 'none);

check_equal(vpar_preorder_ith(1,1), [0]);
check_equal(vpar_preorder_ith(1,2), 'none);
check_equal(vpar_preorder_ith(1,1,1), [0]);
check_equal(vpar_preorder_ith(1,2,1), 'none);

check_equal(vpar_preorder_ith(2,1), [0,1]);
check_equal(vpar_preorder_ith(2,2), 'none);
check_equal(vpar_preorder_ith(2,1,1), [0,1]);
check_equal(vpar_preorder_ith(2,2,1), [0,0]);
check_equal(vpar_preorder_ith(2,3,1), 'none);

{
  check_equal(vpar_preorder_ith(5,1,1), [0,1,2,3,4]);
  check_equal(iferr(vpar_preorder_ith(5,1,'forest);0, e, 1),
              1,
              "vpar_preorder_ith() error on forest=symbol (whereas it should be 0,1)");
}

{
  for(forest=0,1,
     for(n=0,8-forest,
        my(i=1);
        my(vpar=vpar_preorder_first(n));
        until((vpar=vpar_preorder_next(vpar,forest))=='none,
              my(got_vpar=vpar_preorder_ith(n,i,forest));
              check_equal(got_vpar,vpar,
                          "vpar_preorder_ith() vs vpar_preorder_next()");
              check_equal(vpar_preorder_to_i(vpar,forest), i,
                          "vpar_preorder_ith() vs vpar_preorder_to_i()");
              i++);
        check_equal(vpar, 'none,
                    "vpar_preorder_ith() stop at 'none");
        check_equal(i, vpar_count_ordered(n,forest)+1,
                    "vpar_preorder_next() count");

        check_equal(vpar_preorder_ith(n,0,forest), 'none,
                    "vpar_preorder_ith() i=0 none");
        check_equal(vpar_preorder_ith(n,-123,forest), 'none,
                    "vpar_preorder_ith() i=-123 none");

        check_equal(vpar_preorder_ith(n,i,forest), 'none,
                    "vpar_preorder_ith() i=max+1 none");

        check_equal(vpar_preorder_ith(n,i+123,forest), 'none,
                    "vpar_preorder_ith() i=max+123 none");

     ));
}


\\-----------------------------------------------------------------------------
progress("vpar_unlabelled_prev()");

{
  my(vpar=vpar_unlabelled_last(5));
  check_equal(iferr(vpar_unlabelled_prev(vpar,'forest);0, e, 1),
              1,
              "vpar_unlabelled_prev() error on forest=symbol (whereas it should be 0,1)");
}

check_equal(vpar_unlabelled_next([0,1,2,2,2,2,1,1]), [0,1,2,2,2, 1,6,6]);

{
  \\     1              1
  \\    / \ \          / \
  \\   2  4  5    =>  2   4
  \\   |              |   |
  \\   3              3   5
  my(vpar=[0, 1, 2, 1, 1],
     prev_vpar=[0,1,2, 1,4]);
  vpar_validate(vpar);
  vpar_validate(prev_vpar);
  check_equal(prev_vpar, vpar_to_unlabelled(prev_vpar));
  check_equal(vpar_unlabelled_prev(vpar), prev_vpar);
}
{
  \\     1              1
  \\   // \\           / \ \
  \\  2 3 4 5    =>   2  4  5
  \\                  |
  \\                  3
  my(vpar=[0,1,1,1,1],
     prev_vpar=[0,1,2, 1,1]);
  check_equal(vpar, vpar_unlabelled_last(5));
  check_equal(prev_vpar, vpar_to_unlabelled(prev_vpar));
  check_equal(vpar_unlabelled_prev(vpar), prev_vpar);
}
{
  \\      1              1
  \\     / \            /
  \\    2   4    =>    2
  \\    |              | \
  \\    3              3  4
  my(vpar=[0,1,2, 1],
     prev_vpar=[0,1,2, 2]);
  vpar_validate(vpar);
  vpar_validate(prev_vpar);
  check_equal(vpar,      vpar_to_unlabelled(vpar));
  check_equal(prev_vpar, vpar_to_unlabelled(prev_vpar));
  check_equal(vpar_unlabelled_next(prev_vpar), vpar);
  check_equal(vpar_unlabelled_prev(vpar), prev_vpar);
}
{
  \\     1            1
  \\     |            | \
  \\     2      =>    2  5
  \\   / | \          |
  \\  3  4  5         3        increase depth of 4
  \\                   \
  \\                    4
  my(vpar=[0, 1, 2, 2, 2],
     prev_vpar=[0, 1, 2, 3, 1]);
  vpar_validate(vpar);
  vpar_validate(prev_vpar);
  check_equal(vpar,      vpar_to_unlabelled(vpar));
  check_equal(prev_vpar, vpar_to_unlabelled(prev_vpar));
  check_equal(vpar_unlabelled_next(prev_vpar), vpar);
  check_equal(vpar_unlabelled_prev(vpar), prev_vpar);
}
{
  \\   1              1
  \\   | \            | \
  \\   2  4    =>     2  5
  \\   |   \          |\       increase depth of 4
  \\   3    5         3 4
  my(vpar=[0, 1, 2, 1, 4],
     prev_vpar=[0, 1, 2, 2, 1]);
  vpar_validate(vpar);
  vpar_validate(prev_vpar);
  check_equal(vpar,      vpar_to_unlabelled(vpar));
  check_equal(prev_vpar, vpar_to_unlabelled(prev_vpar));
  check_equal(vpar_unlabelled_next(prev_vpar), vpar);
  check_equal(vpar_unlabelled_prev(vpar), prev_vpar);
}
{
  \\      1            1
  \\    / | \         /  \ \
  \\   2  4  6   =>  2    5 6
  \\   |  |          | \
  \\   3  5          3  4
  \\
  \\   not to     1     since that is lex subtree(4) > subtree(2)
  \\            / |
  \\           2  4
  \\           |  | \
  \\           3  5  6
  my(vpar=[0, 1, 2, 1, 4, 1],
     prev_vpar=[0, 1, 2, 2, 1, 1]);
  vpar_validate(vpar);
  vpar_validate(prev_vpar);
  check_equal(vpar,      vpar_to_unlabelled(vpar));
  check_equal(prev_vpar, vpar_to_unlabelled(prev_vpar));
  check_equal(vpar_unlabelled_next(prev_vpar), vpar, "vpar_unlabelled_next() 6");
  check_equal(vpar_unlabelled_prev(vpar), prev_vpar, "vpar_unlabelled_prev() 6");
}
{
  \\      1                1
  \\    /   \             / \ \          equality of top trees,
  \\   2     5    =>    2    6 7         cannot increase the last
  \\  / \   / \        /|\
  \\ 3  4   6  7      3 4 5
  \\
  my(vpar=[0, 1, 2, 2, 1, 5, 5],
     prev_vpar=[0, 1, 2, 2, 2, 1, 1]);
  vpar_validate(vpar);
  vpar_validate(prev_vpar);
  check_equal(vpar,      vpar_to_unlabelled(vpar));
  check_equal(prev_vpar, vpar_to_unlabelled(prev_vpar));
  check_equal(vpar_unlabelled_next(prev_vpar), vpar, "vpar_unlabelled_next() 7");
  check_equal(vpar_unlabelled_prev(vpar), prev_vpar, "vpar_unlabelled_prev() 7");
}

{
  for(forest=0,1,
     for(n=0,12-forest,
        my(vpar=vpar_unlabelled_first(n),
           prev_vpar='none);
        until(vpar=='none,
              check_equal(vpar_unlabelled_prev(vpar,forest), prev_vpar,
                          "vpar_unlabelled_prev() inverse of vpar_unlabelled_next()");
              prev_vpar=vpar;
              vpar=vpar_unlabelled_next(vpar,forest))));
}


\\-----------------------------------------------------------------------------
progress("vpar_count_free_asymmetric()");

{
  check_equal(vpar_count_free_asymmetric(10,1), 9);
  check_equal(iferr(vpar_count_free_asymmetric(10,'forest);0, e, 1),
              1,
  "vpar_count_free_asymmetric() error on forest=symbol (whereas it should be 0,1)");
}

{
  \\ free trees
  for(n=0,11,
     my(vpar=vpar_free_first(n),
       count=0);
     until((vpar=vpar_free_next(vpar))=='none,
           count += vpar_is_free_asymmetric(vpar));
     check_equal(count, vpar_count_free_asymmetric(n),
       Str("vpar_count_free_asymmetric() vs vpar_is_free_asymmetric() trees n="n));
  );
}

{
  \\ free forests, but selecting from unlabelled forests
  for(n=1,11,
     my(vpar=vpar_unlabelled_first(n),
        l=List([]));
     until((vpar=vpar_unlabelled_next(vpar,1))=='none,
           vpar_is_free_asymmetric(vpar) || next();
           for(i=1,#l,
              if(vpar_is_free_isomorphic(vpar,l[i]), next(2)));
           listput(l,vpar));
     check_equal(#l, vpar_count_free_asymmetric(n,1),
       Str("vpar_count_free_asymmetric() vs vpar_is_free_asymmetric() forests n="n));
     );
}

{
  \\ A000220 number of free asymmetric trees, n>=1
  check_equal(vector(#A000220_samples,n, vpar_count_free_asymmetric(n)),
              A000220_samples,    \\ OFFSET=1
              "A000220 number of free asymmetric trees");
}
{
  \\ A000220 gf A(x)-A^2(x)/2-A(x^2)/2 where A is A004111
  my(limit=#A000220_samples+1,
     gTL = test_func_to_gf(n->if(n==0,0,vpar_count_unlabelled_asymmetric(n)), limit),
     gTF = gTL - gTL^2/2 - subst(gTL,'x,'x^2)/2);
  check_equal(test_gf_terms(gTF,limit)[^1],
              A000220_samples);  \\ n=1 upwards
}

{
  \\ A035056 number of free asymmetric forests, n>=0
  check_equal(vector(#A035056_samples,n,n--; vpar_count_free_asymmetric(n,1)),
              A035056_samples,    \\ OFFSET=0
              "A035056 number of free asymmetric forests");
}


\\-----------------------------------------------------------------------------
progress("vpar_count_hereditaryleastsingle()");

{
  check_equal(vpar_count_hereditaryleastsingle(5,1), 326);
  check_equal(iferr(vpar_count_hereditaryleastsingle(5,'forest);0, e, 1),
              1,
  "vpar_count_hereditaryleastsingle() error on forest=symbol (whereas it should be 0,1)");
}

{
  for(forest=0,1,
     for(n=0, test_limit_trees_search - forest,
        my(count=0);
        forvec(vec=vpar_forvec_from(n,forest), \\ forests or trees
              my(vpar=vpar_from_upascend(vec));
              count+=vpar_is_hereditaryleastsingle(vpar));
        check_equal(count, vpar_count_hereditaryleastsingle(n,forest),
                    "vpar_count_hereditaryleastsingle() vs count vpar_is_hereditaryleastsingle()");
     ));
}


\\-----------------------------------------------------------------------------
progress("vpar_indomsets_count_max()");

check_equal(vector(20,n,n--; vpar_indomsets_count_max(n,-1)), \
            vector(20,n,n--; vpar_indomsets_count_max(n)), \
            "vpar_indomsets_count_max() root=1 same as all trees");

{
  check_equal(vpar_indomsets_count_max(5,1), 4);
  check_equal(iferr(vpar_indomsets_count_max(5,'forest);0, e, 1),
              1,
  "vpar_indomsets_count_max() error on forest=symbol (whereas it should be 0,1)");
}

\\ Formula shown in vpar_indomsets_count_max() addhelp.
test_indomsets_count_max_by_formula(n) =
{
  if(n==0,1,
     n%2==0, 2^(n/2-1)+1,     \\ even
             2^((n-1)/2));    \\ odd
}
check_equal(vector(20,n,n--; vpar_indomsets_count_max(n)), \
            vector(20,n,n--; test_indomsets_count_max_by_formula(n)), \
        "vpar_indomsets_count_max() vs test_indomsets_count_max_by_formula()");

test_indomsets_count_max_by_search(n,forest=0) =
{
  \\ PENDING: Search over free forests instead of unlabelled, when free
  \\ forests code is ready.

  \\ root=1 same as all trees, indomsets don't depend on root
  if(forest<0,forest=0);

  my(m=0,
     vpar=vpar_unlabelled_first(n));
  until((vpar=vpar_unlabelled_next(vpar,forest))=='none,
        m=max(m,vpar_indomsets_count(vpar)));
  m;
}
{
  for(forest=-1,1,
     my(limit=10-forest);
     check_equal(vector(limit,n,n--; vpar_indomsets_count_max(n,forest)),
                 vector(limit,n,n--; test_indomsets_count_max_by_search(n,forest)),
                 "vpar_indomsets_count_max() vs test_indomsets_count_max_by_search()"));
}

{
  \\ per vpar_indomsets_count_max() addhelp, all counts except n=2 are <2^(n/2)
  check_equal(vector(20,n, vpar_indomsets_count_max(n) < 2^(n/2)),
              vector(20,n, n!=2));
  check_equal(vector(20,n,n=2*n-1; vpar_indomsets_count_max(n)),
              vector(20,n,n=2*n-1; 2^((n-1)/2)));
}

{
  \\ Wilf's example of largest number of indomsets, n=9
  \\
  \\ Herbert S. Wilf, "The Number of Maximal Independent Sets in a Tree",
  \\ SIAM Journal on Algebraic and Discrete Methods, volume 7, number 1,
  \\ January 1986, pages 125-130.
  \\ https://www.math.upenn.edu/~wilf/website/Maximal%20independent%20sets%20in%20a%20tree.pdf

  my(vpar=vpar_subdivide(vpar_make_star(6)));
  vpar_validate(vpar);
  check_equal(#vpar, 11);
  check_equal(vpar_indomsets_count(vpar), 32);
  check_equal(vpar_indomsets_count_max(#vpar), 32);
}
{
  \\ Wilf's example of largest number of indomsets, n=10
  \\ one arm length 1
  \\
  \\ Herbert S. Wilf, "The Number of Maximal Independent Sets in a Tree",
  \\ SIAM Journal on Algebraic and Discrete Methods, volume 7, number 1,
  \\ January 1986, pages 125-130.
  \\ https://www.math.upenn.edu/~wilf/website/Maximal%20independent%20sets%20in%20a%20tree.pdf

  my(vpar=vpar_join(vpar_subdivide(vpar_make_star(5)),1,[0]));
  vpar_validate(vpar);
  check_equal(#vpar, 10);
  check_equal(vpar_indomsets_count(vpar), 17);
  check_equal(vpar_indomsets_count_max(#vpar), 17);
}


\\-----------------------------------------------------------------------------
progress("vpar_is_preorder_depths_vector()");

{
  check_equal(vpar_is_preorder_depths_vector([0,0],1), 1);
  check_equal(iferr(vpar_is_preorder_depths_vector([0,0],'forest);0, e, 1),
              1,
  "vpar_is_preorder_depths_vector() error on forest=symbol (whereas it should be 0,1)");
}
{
  for(forest=0,1,
     for(n=0, test_limit_trees_search - forest,
        my(seen=List([]));  \\ Set() of depths vectors seen
        my(vpar=vpar_preorder_first(n));
        until((vpar=vpar_preorder_next(vpar,forest))=='none,
              my(depths=vpar_depths_vector(vpar));
              check_equal(vpar_is_preorder_depths_vector(depths,forest), 1,
                          "vpar_is_preorder_depths_vector() of iterate");
              my(pos=setsearch(seen,depths,1));
              if(pos, listinsert(seen,depths,pos)));
        check_equal(#seen, vpar_count_ordered(n,forest),
                    "vpar_depths_vector() of preorder all distinct");

        forvec(depths=vector(n,i,[-1,n+1]),
              my(is_depths = vpar_is_depths_vector(depths,forest),
                 is_preorder_depths = vpar_is_preorder_depths_vector(depths,forest));
              check_equal(is_depths >= is_preorder_depths, 1,
                          "vpar_is_depths_vector() >= vpar_is_preorder_depths_vector()");
              check_equal(is_preorder_depths,
                          setsearch(seen,depths)!=0,
                          "vpar_is_preorder_depths_vector() vs seen from vpars"));
     ));
}

\\ {
\\   for(i=1,#test_sample_forests,
\\      my(vpar=test_sample_forests[i],
\\         depths=vpar_depths_vector(vpar));
\\         check_equal(vpar_is_preorder_depths_vector(depths),
\\                     vpar_is_preorder(vpar),
\\                     "vpar_is_preorder_depths_vector() vs vpar_is_preorder()");
\\      );
\\ }

\\ n=4
\\ [0, 1, 2, 3]  vpar [0, 1, 2, 3]
\\ [0, 1, 2, 2]  vpar [0, 1, 2, 2]
\\ [0, 1, 2, 1]  vpar [0, 1, 2, 1]
\\ [0, 1, 1, 1]  vpar [0, 1, 1, 1]

\\ [0, 1, 2, 3]
\\ [0, 1, 2, 2]
\\ [0, 1, 2, 1]
\\ [0, 1, 1, 2]
\\ [0, 1, 1, 1]

\\ vpar_make_preorder_depths([0, 1, 2, 1]) == [0,1,2,1]
\\ vpar_make_preorder_depths([0, 1, 1, 2]) == [0,1,1,3]
\\ vpar_view([0,1,2,1])
\\ vpar_view([0,1,1,3])

\\ Catalan numbers
\\ 1 1 1 2 5 14 42 132


\\-----------------------------------------------------------------------------
progress("vpar_is_depths_vector()");

check_equal(vpar_is_depths_vector([]),1, "vpar_is_depths_vector() empty");
check_equal(vpar_is_depths_vector([-1]),0, "vpar_is_depths_vector() negative");

check_equal(vpar_is_depths_vector([0]),1, "vpar_is_depths_vector() singleton");
check_equal(vpar_is_depths_vector([0],1),1, "vpar_is_depths_vector() singleton");
check_equal(vpar_is_depths_vector([0],-1),1, "vpar_is_depths_vector() singleton as root=1");

check_equal(vpar_is_depths_vector([0,1]),1);
check_equal(vpar_is_depths_vector([0,1],1),1);
check_equal(vpar_is_depths_vector([0,1],-1),1);
check_equal(vpar_is_depths_vector([1,0]),1, "vpar_is_depths_vector() path-2 reverse");
check_equal(vpar_is_depths_vector([1,0],1),1);
check_equal(vpar_is_depths_vector([1,0],-1),0);

check_equal(vpar_is_depths_vector([0,0,0,0]),0, "vpar_is_depths_vector() 4 singletons");
check_equal(vpar_is_depths_vector([0,0,0,0],-1),0);
check_equal(vpar_is_depths_vector([0,0,0,0],1),1);
check_equal(vpar_is_depths_vector([1]),0);
check_equal(vpar_is_depths_vector([0,1,2]),1);
check_equal(vpar_is_depths_vector([0,1,2,3,4,5]),1);
check_equal(vpar_is_depths_vector([2,1,0]),1);
check_equal(vpar_is_depths_vector([2,2,2,1,1,1,0,0,0]),0);
check_equal(vpar_is_depths_vector([2,2,2,1,1,1,0,0,0],1),1);
check_equal(vpar_is_depths_vector([5,4,3,2,1,0]),1);
check_equal(vpar_is_depths_vector([5,4,3,2,1,1]),0);
check_equal(vpar_is_depths_vector([0,5,5,5,5,5]),0);
check_equal(vpar_is_depths_vector([-1]),0);
check_equal(vpar_is_depths_vector([-100]),0, "vpar_is_depths_vector() -100");
check_equal(vpar_is_depths_vector([1,-1]),0);
check_equal(vpar_is_depths_vector([100,-100]),0);
check_equal(vpar_is_depths_vector([0,1,-100]),0, \
            "vpar_is_depths_vector() 0,1,-100");
check_equal(vpar_is_depths_vector([0,1,2^1000]),0, \
            "vpar_is_depths_vector() wild positive");
check_equal(vpar_is_depths_vector([0,1,-2^1000]),0, \
            "vpar_is_depths_vector() wild negative");

{
  check_equal(vpar_is_depths_vector([0,0],1), 1);
  check_equal(iferr(vpar_is_depths_vector([0,0],'forest);0, e, 1),
              1,
  "vpar_is_depths_vector() error on forest=symbol (whereas it should be 0,1)");
}

{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i],
        depths=vpar_depths_vector(vpar));
     check_equal(vpar_is_depths_vector(depths),1,
                "vpar_is_depths_vector() of tree");
     check_equal(vpar_is_depths_vector(depths,1),1,
                "vpar_is_depths_vector() of tree, as forest");

     if(#vpar, vpar=vpar_reroot(vpar,1));
     depths=vpar_depths_vector(vpar);
     check_equal(vpar_is_depths_vector(depths,-1),1,
                "vpar_is_depths_vector() of tree root=1");
     check_equal(vpar_is_depths_vector(depths),1,
                "vpar_is_depths_vector() of tree root=1, as tree");
     check_equal(vpar_is_depths_vector(depths,1),1,
                "vpar_is_depths_vector() of tree root=1, as forest");
  );
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        depths=vpar_depths_vector(vpar));
     check_equal(vpar_is_depths_vector(depths,1),1,
                 "vpar_is_depths_vector() of forest"));
}


\\-----------------------------------------------------------------------------
progress("vpar_degrees_vectors_count()");

{
  check_equal(vpar_degrees_vectors_count(5,1), 136);
  check_equal(iferr(vpar_degrees_vectors_count(5,'forest);0, e, 1),
              1,
  "vpar_degrees_vectors_count() error on forest=symbol (whereas it should be 0,1)");
}

\\ formula shown in the addhelp() of vpar_degrees_vectors_count()
test_degrees_vectors_count_F(n) = \
  1 + sum(m=2,n, binomial(n,m) * test_degrees_vectors_count_C(m));
test_degrees_vectors_count_C(m) = \
  sum(c=0,floor(m/2)-1, binomial(2*m-3-2*c, m-1));

check_equal(vector(30,n,n--; vpar_degrees_vectors_count(n,1)), \
            vector(30,n,n--; test_degrees_vectors_count_F(n)));

{
  \\ A014301 = degrees vectors of forests of no singletons
  \\         = number of "protected points"
  \\ a(n) = (1/2)*Sum_{j=0..floor(n/2)} binomial(2n-2j-2, n-2)
  \\ a(n) = Sum_{k=0..n} (-1)^(n-k)*C(n+k,k-1)
  \\
  check_equal(vector(#A014301_samples,m, test_degrees_vectors_count_C(m)),
              A014301_samples,     \\ OFFSET=1
              "test_degrees_vectors_count_C() vs A014301_samples");
}

{
  my(forest=0);
  for(n=1, test_limit_trees_search - forest,
     my(l=List([]));   \\ distinct degrees vectors
     forvec(vec=vpar_forvec_from(n,forest), \\ trees
           my(vpar=vpar_from_upascend(vec),
              degrees=vpar_degrees_vector(vpar));
           my(pos=setsearch(l,degrees,1));
           if(pos, listinsert(l,degrees,pos)));

     check_equal(#l, vpar_degrees_vectors_count(n),
                 "vpar_degrees_vectors_count() trees vs degrees search");
     check_equal(#l, vpar_degrees_vectors_count(n,-1),
                 "vpar_degrees_vectors_count() trees root=1 vs degrees search");
  );
}
{
  my(forest=1);
  for(n=1, test_limit_trees_search - forest,
     my(l=List([]));   \\ distinct degrees vectors
     forvec(vec=vpar_forvec_from(n,forest), \\ forests
           my(vpar=vpar_from_upascend(vec),
              degrees=vpar_degrees_vector(vpar));
           my(pos=setsearch(l,degrees,1));
           if(pos, listinsert(l,degrees,pos)));

     check_equal(#l, test_degrees_vectors_count_F(n),
                 "test_degrees_vectors_count_F() vs degrees search");
     check_equal(#select(degrees->#degrees && vecmin(degrees)>=1, l),
                 test_degrees_vectors_count_C(n),
                 "test_degrees_vectors_count_C() vs degrees search");
     check_equal(#l, vpar_degrees_vectors_count(n,1),
                 "vpar_degrees_vectors_count() forests, vs degrees search");
  );
}

{
  my(limit = 10,
     L = O(x^limit),
     gT = test_func_to_gf(vpar_degrees_vectors_count, limit));

  check_equal(gT, 1+x+x^2*(1-sqrt(1-4*x+L))/(2*x)/sqrt(1-4*x+L) + L,
             "vpar_degrees_vectors_count() gT");

  check_equal(gT, 1 + x + 1/2*x*(1/sqrt(1-4*x+L) - 1) + L);
  check_equal(gT, 1 + 1/2*x*(1 + 1/sqrt(1-4*x+L)) + L);
  check_equal(gT, 1 + 1/2*x + 1/2*x*sqrt(1-4*x+L)/(1-4*x) + L);
  check_equal(gT, ( 1 - 7/2*x - 2*x^2 + 1/2*x*sqrt(1-4*x+L) )/(1-4*x) + L);

  check_equal(1 - 2/x - 2 + 2/x*gT, 1/sqrt(1-4*x+L) + L/x);
  check_equal( 1/x^2* (  -x^2 + (1-4*x)*(x - 2 - 2*x + 2*gT)^2 ), 0 + L/x);
  check_equal( 1/x^2* (  -x^2 + (1-4*x)*(x - 2 - 2*x + 2*gT)^2 ), 0 + L/x);

  \\---------------
  my(gC = test_func_to_gf(test_degrees_vectors_count_C, limit));

  \\ per Emeric Deutsch in A014301
  check_equal(gC, (1-2*x-sqrt(1-4*x))/(3*sqrt(1-4*x)-1+4*x) + L,
             "vpar_degrees_vectors_count() gC");

  check_equal(gC, (1-2*x - sqrt(1-4*x))/(3*sqrt(1-4*x) - (1-4*x)) + L);
  check_equal(gC, (1-2*x - sqrt(1-4*x)) * (3*sqrt(1-4*x) + (1-4*x)) / (9*(1-4*x) - (1-4*x)^2) + L);
  check_equal(gC, ( - 2 + 6*x + 8*x^2  + 2*(1-x) * sqrt(1-4*x)
                  ) / 4 / (2+x)/ (1-4*x) + L);

  check_equal(gC, 1/2 * ( -(1+x)*(1-4*x) + (1-x)*sqrt(1-4*x) )
                      / (2+x) / (1-4*x) + L);
  check_equal(gC, 1/2 *(-1-x + (1-x)/sqrt(1-4*x)) / (2+x) + L);

  \\---------------
  my(gF = test_func_to_gf(n->vpar_degrees_vectors_count(n,1), limit));

  \\ Superseeker GUESSGF
  \\        2                   4       3       2            1/2
  \\    10 x  - 17 x + 3 + (20 x  - 44 x  + 33 x  - 10 x + 1)
  \\ [- --------------------------------------------------------, ogf]
  \\                        2
  \\                  2 (5 x  - 11 x + 2) (x - 1)

  check_equal(gF, ( (3-2*x)*(1-5*x)  + (1-2*x)*sqrt( (1-x)*(1-5*x) + L) )
                  / ( 2* (2-x) * (1-x) * (1-5*x) ) );

  check_equal(gF, 1/2 * 1/(2-x) * ( (3-2*x)*(1-5*x)  + (1-2*x)*sqrt( (1-x)*(1-5*x) + L) )
                  / ( (1-x) * (1-5*x) ) );
  check_equal(gF, 1/2 * 1/(2-x) * ( (3-2*x)*(1-5*x) / ( (1-x) * (1-5*x) )
                                   + (1-2*x)/sqrt( (1-x)*(1-5*x) + L) ));
  check_equal(gF, 1/2 * 1/(2-x)
            * ( 2 + 1/(1-x) + (1-2*x) / sqrt( (1-x)*(1-5*x) + L) ));

  check_equal(test_gf_terms(1/2 * 1/(2-x), 10),
              vector(10,n, 1/2^(n+1)));

  \\ print(test_gf_terms((1-2*x) / sqrt( (1-x)*(1-5*x) + L), limit));
  \\        1, 1, 5, 23, 105, 483, 2243, 10505, 49549, 235075
  \\ not in OEIS: 5, 23, 105, 483, 2243, 10505, 49549, 235075

  \\ print(test_gf_terms(1 / sqrt( (1-x)*(1-5*x) + L), limit));
  \\ 1, 3, 11, 45, 195, 873, 3989, 18483, 86515, 408105
  \\ A026375 sum(k=0,n, binomial(n,k)*binomial(2*k,k))
  \\ binomial transform of A000984 central binomial(2*n,n)
  check_equal(1 / sqrt( (1-x)*(1-5*x) + L),
              test_func_to_gf(n->sum(k=0,n, binomial(n,k)*binomial(2*k,k)), limit));
}

{
  \\ A128732  DL's in skew Dyck paths semilength n
  \\ origin to X axis, set U up, D down, L left with U and L steps not overlapping
  my(A128732_samples = [
   0, 0, 1, 5, 23, 106, 493, 2312, 10917, 51840, 247319, 1184557, 5692517,
   27434578, 132547877, 641789941, 3113487683, 15130119784, 73637665027,
   358883327591, 1751237017413, 8555108199294, 41836182269267,
   204779733440086 ]);
  check_equal(vector(#A128732_samples,m,m-=2;
                  if(m==-1,0, test_degrees_vectors_count_F(m+1)
                              - test_degrees_vectors_count_F(m))),
              A128732_samples,
              "test_degrees_vectors_count_F() increments vs A128732_samples");

  my(A128732 = n->A128732_samples[n+1],  \\ n=0 upwards
     limit = #A128732_samples,
     L = O(x^limit),
     gA128732 = test_func_to_gf(A128732, limit));
  check_equal(gA128732,
              x* (1 - x - sqrt(1 - 6*x + 5*x^2 + L))
              / (1 - 6*x + 5*x^2 + (1+x)*sqrt(1 - 6*x + 5*x^2 + L))  + L,
              "A128732 gf");
  check_equal((1 - 6*x + 5*x^2),  (1-5*x)*(1-x));
}


\\-----------------------------------------------------------------------------
progress("vpar_num_siblings_vectors_count()");

{
  check_equal(vpar_num_siblings_vectors_count(5,1), 42);
  check_equal(iferr(vpar_num_siblings_vectors_count(5,'forest);0, e, 1),
              1,
  "vpar_num_siblings_vectors_count() error on forest=symbol (whereas it should be 0,1)");
}
{
  check_equal(vector(#A178682_samples,n,n--; vpar_num_siblings_vectors_count(n,1)),
              A178682_samples,    \\ OFFSET=0
              "vpar_num_siblings_vectors_count() forests = A178682");
}


\\-----------------------------------------------------------------------------
progress("vpar_is_num_siblings_vector()");

check_equal(vpar_is_num_siblings_vector([]), 1, \
            "vpar_is_num_siblings_vector() empty");
check_equal(vpar_is_num_siblings_vector([0]), 1, \
            "vpar_is_num_siblings_vector() singleton");
check_equal(vpar_is_num_siblings_vector([0,0]), 1);
\\ check_equal(vpar_is_num_siblings_vector([1,1]), 0);
check_equal(vpar_is_num_siblings_vector([1,1],1), 1);
check_equal(vpar_is_num_siblings_vector([1,-1]), 0);

check_equal(vpar_num_siblings_vector([0,1,1]), [0,1,1]);
check_equal(vpar_is_num_siblings_vector([0,1]), 0);
check_equal(vpar_is_num_siblings_vector([0,1,1]), 1);
check_equal(vpar_is_num_siblings_vector([0,1,1,1]), 0);

{
  check_equal(vpar_is_num_siblings_vector([0,0],1), 1);
  check_equal(iferr(vpar_is_num_siblings_vector([0,0],'forest);0, e, 1),
              1,
  "vpar_is_num_siblings_vector() error on forest=symbol (whereas it should be 0,1)");
}

{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i],
        num_siblings=vpar_num_siblings_vector(vpar));
     \\ print(vpar"  "num_siblings);
     check_equal(vpar_is_num_siblings_vector(num_siblings),1,
                 "vpar_is_num_siblings_vector() of tree");
     check_equal(vpar_is_num_siblings_vector(num_siblings,1),1,
                 "vpar_is_num_siblings_vector() of tree as forest");

     if(#vpar, vpar=vpar_reroot(vpar,1));
     num_siblings=vpar_num_siblings_vector(vpar);
     \\ print(vpar"  "num_siblings);
     check_equal(vpar_is_num_siblings_vector(num_siblings,-1), 1,
                 "vpar_is_num_siblings_vector() of tree root=1");
     check_equal(vpar_is_num_siblings_vector(num_siblings), 1,
                 "vpar_is_num_siblings_vector() of tree root=1 as tree");
     check_equal(vpar_is_num_siblings_vector(num_siblings,1), 1,
                 "vpar_is_num_siblings_vector() of tree root=1 as forest"));
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        num_siblings=vpar_num_siblings_vector(vpar));
     check_equal(vpar_is_num_siblings_vector(num_siblings,1),1,
                 "vpar_is_num_siblings_vector() of forest"));
}

{
  for(forest=-1,1,
     for(n=0, test_limit_trees_search - forest,
        my(seen=List([]));  \\ Set() of num_siblings vectors seen
        forvec(vec=vpar_forvec_from(n,forest),
              my(vpar=vpar_from_upascend(vec),
                 num_siblings=vpar_num_siblings_vector(vpar));
              check_equal(vpar_is_num_siblings_vector(num_siblings,forest),1,
                          "vpar_is_num_siblings_vector() of all");
              my(pos=setsearch(seen,num_siblings,1));
              if(pos, listinsert(seen,num_siblings,pos)));
        \\ print(seen);
        \\ print(#seen);

        check_equal(#seen, vpar_num_siblings_vectors_count(n,forest),
                    "vpar_num_siblings_vectors_count() vs seen");
     ));
}


\\-----------------------------------------------------------------------------
progress("vpar_is_num_children_vector()");

check_equal(vpar_is_num_children_vector([]),1, \
            "vpar_is_num_children_vector() empty tree");
check_equal(vpar_is_num_children_vector([],1),1, \
            "vpar_is_num_children_vector() empty forest");
check_equal(vpar_is_num_children_vector([],-1),1, \
            "vpar_is_num_children_vector() empty root=1");

check_equal(vpar_is_num_children_vector([0]),1, \
            "vpar_is_num_children_vector() singleton tree");
check_equal(vpar_is_num_children_vector([0],1),1, \
            "vpar_is_num_children_vector() singleton forest");
check_equal(vpar_is_num_children_vector([0],-1),1, \
            "vpar_is_num_children_vector() singleton root=1");

check_equal(vpar_is_num_children_vector([1,0],-1),1, \
            "vpar_is_num_children_vector() root=1 path-2");
check_equal(vpar_is_num_children_vector([0,1],-1),0, \
            "vpar_is_num_children_vector() root=1 path-2, not");

check_equal(vpar_is_num_children_vector([0,0,0,0]),0, \
            "vpar_is_num_children_vector() on forest");
check_equal(vpar_is_num_children_vector([0,0,0,0],1),1);
check_equal(vpar_is_num_children_vector([1]),0);
check_equal(vpar_is_num_children_vector([0,1]),1);
check_equal(vpar_is_num_children_vector([1,0]),1);
check_equal(vpar_is_num_children_vector([-1]),0);
check_equal(vpar_is_num_children_vector([-100]),0);
check_equal(vpar_is_num_children_vector([1,-1]),0);
check_equal(vpar_is_num_children_vector([100,-100]),0);
check_equal(vpar_is_num_children_vector([0,1,-100]),0, \
            "vpar_is_num_children_vector() 0,1,-100");
check_equal(vpar_is_num_children_vector([0,1,2^1000]),0, \
            "vpar_is_num_children_vector() wild positive");
check_equal(vpar_is_num_children_vector([0,1,-2^1000]),0, \
            "vpar_is_num_children_vector() wild negative");

{
  check_equal(vpar_is_num_children_vector([0,0],1), 1);
  check_equal(iferr(vpar_is_num_children_vector([0,0],'forest);0, e, 1),
              1,
  "vpar_is_num_children_vector() error on forest=symbol (whereas it should be 0,1)");
}

{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i],
        num_children=vpar_num_children_vector(vpar));
     check_equal(vpar_is_num_children_vector(num_children),1,
                "vpar_is_num_children_vector() of tree");

     if(#vpar, vpar=vpar_reroot(vpar,1));
     num_children=vpar_num_children_vector(vpar);
     check_equal(vpar_is_num_children_vector(num_children,-1), 1,
                 "vpar_is_num_children_vector() of tree root=1"));
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        num_children=vpar_num_children_vector(vpar));
     check_equal(vpar_is_num_children_vector(num_children,1),1,
                 "vpar_is_num_children_vector() of forest"));
}


\\-----------------------------------------------------------------------------
progress("vpar_num_children_vectors_count()");

{
  check_equal(vpar_num_children_vectors_count(5,1), 126);
  check_equal(iferr(vpar_num_children_vectors_count(5,'forest);0, e, 1),
              1,
  "vpar_num_children_vectors_count() error on forest=symbol (whereas it should be 0,1)");
}

\\  central binomial coefficient
\\ vector(10,n,distinct_num_children_tree(n))
test_distinct_num_children_tree(n) = if(n==0,1, binomial(2*n-2,n-1));

test_distinct_num_children_tree_root1(n) = if(n==0,1, binomial(2*n-3,n-1));

\\ A001700 a(n) = binomial(2n+1, n+1)
\\ vector(10,n,distinct_num_children_forest(n))
test_distinct_num_children_forest(n) = if(n==0,1, binomial(2*n-1,n));

{
  \\ counts vs vpar_is_num_children_vector() tests
  for(n=0,5,
     my(count_trees_root1=0, count_trees=0, count_forests=0);
     forvec(vec=vector(n,i,[-1,n]),
           count_trees_root1 += vpar_is_num_children_vector(vec,-1);
           count_trees       += vpar_is_num_children_vector(vec);
           count_forests     += vpar_is_num_children_vector(vec,1));

     check_equal(count_trees_root1, vpar_num_children_vectors_count(n,-1),
                 Str("vpar_num_children_vectors_count() "
                     "vs vpar_is_num_children_vector() n="n" root=1"));
     check_equal(count_trees, vpar_num_children_vectors_count(n),
                 Str("vpar_num_children_vectors_count() "
                     "vs vpar_is_num_children_vector() n="n" trees"));
     check_equal(count_forests, vpar_num_children_vectors_count(n,1),
                 Str("vpar_num_children_vectors_count() "
                     "vs vpar_is_num_children_vector() n="n" forests"));

     check_equal(count_trees_root1, test_distinct_num_children_tree_root1(n));
     check_equal(count_trees, test_distinct_num_children_tree(n));
     check_equal(count_forests, test_distinct_num_children_forest(n));
  );
}


\\-----------------------------------------------------------------------------
progress("test_any_singletons()");

check_equal(test_any_singletons([]), 0);
check_equal(test_any_singletons([0]), 1);
check_equal(test_any_singletons([0,1]), 0);
check_equal(test_any_singletons([2,0]), 0);
check_equal(test_any_singletons([2,0,0]), 1);
check_equal(test_any_singletons([2,0,0,0]), 1);


\\-----------------------------------------------------------------------------
progress("test_num_free_automorphism_generators_max()");

{
  \\ trees
  my(want=[0,  0, 1, 1, 2, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8]);
  check_equal(vector(#want,n,n--;
                     test_num_free_automorphism_generators_max(n)),
              want);
}
{
  \\ forests
  my(want=[0,  0, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 7, 8, 8, 9]);
  check_equal(vector(#want,n,n--;
                     test_num_free_automorphism_generators_max(n,1)),
              want);
}


\\-----------------------------------------------------------------------------
progress("test_perms_is_involutions()");

{
  check_equal(test_perms_is_involutions([ Vecsmall([1,3,2]) ]), 1);
  check_equal(test_perms_is_involutions([ Vecsmall([3,1,2]) ]), 0);

  check_equal(test_perms_is_involutions
              ([ Vecsmall([1,2]), Vecsmall([2,1]) ]), 1,
              "test_perms_is_involutions()");
}


\\-----------------------------------------------------------------------------
progress("vpar_centroids_and_weight()");

check_equal(vpar_centroids_and_weight([]),[[],0], "vpar_centroids_and_weight() empty");
check_equal(vpar_centroids_and_weight([0]),[[1],0], "vpar_centroids_and_weight() singleton");

test_children_subtree_sizes_vector(vpar) =
{
  my(size_lists=vector(#vpar,i,List([])),
     seq=vpar_INTERNAL_seq_upwards(vpar));
  for(i=1,#seq,
     my(v=seq[i], p=vpar[v],
        l=(size_lists[v]=Vec(size_lists[v])));
     if(p, listput(size_lists[p],1+vecsum(l))));
  size_lists;
}
test_neighbour_subtree_sizes_vector(vpar) =
{
  apply(v->concat(v,[#vpar-vecsum(v)-1]),
       test_children_subtree_sizes_vector(vpar));
}
\\ return vector where vec[v] is weight of v
test_weights_vector_by_neighbour_max(vpar) = \
  apply(vecmax,test_neighbour_subtree_sizes_vector(vpar));

\\ return [[c1,c2], weight]
test_centroids_and_weight_by_weights_min(vpar) =
{
  my(weights=vpar_weights_vector(vpar),
     centroid_weight=vecmin(concat(#vpar, \\ no empty vecmin
                                   weights)));
  [Vec(select(w->w==centroid_weight,weights,1)),
   centroid_weight];
}

\\ Per addhelp() string, centroid is those vertices which have all
\\ neighbouring subtree sizes <= 1/2 total vertices.
test_centroids_by_half_vertices(vpar) =
{
  my(l=List([]));
  for(v=1,#vpar,
     my(vpar=vpar_reroot(vpar,v),
        children=vpar_children_of_vertex(vpar,v));
     for(i=1,#children,
        my(c=children[i]);
        if(!(vpar_subtree_size(vpar,c) <= 1/2*#vpar),
           next(2)));
     listput(l,v));
  Vec(l);
}

{
  check_equal(test_centroids_and_weight_by_weights_min(vpar_make_bistar(5,5)),
              [[1,6],5],
              "test_centroids_and_weight_by_weights_min");
}
check_equal(vpar_centroids_and_weight(vpar_make_bistar(5,5)),\
            [[1,6],5]);

{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i]);
     check_equal(vpar_weights_vector(vpar),
                 test_weights_vector_by_neighbour_max(vpar));

     my(got_centroids_and_weight=vpar_centroids_and_weight(vpar));
     my(by_test=test_centroids_and_weight_by_weights_min(vpar));
     check_equal(#got_centroids_and_weight[1] <= 2, 1);
     check_equal(got_centroids_and_weight, by_test);

     my(got_centroids=got_centroids_and_weight[1]);
     check_equal(got_centroids,
                 test_centroids_by_half_vertices(vpar),
                 "vpar_centroids_and_weight() vs half vertices");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_to_sparse6()");

check_equal(vpar_to_sparse6([]), ":?");
check_equal(vpar_to_sparse6(vector(8)), ":G");
check_equal(vpar_to_sparse6([5, 1, 4, 5, 0]), ":Daq_~");
{
  \\ n=31 with self-loop
  my(vpar=vector(31));
  vpar[1]=1;  check_equal(vpar_to_sparse6(vpar),
              concat([":",Strchr([31+63,63]),""]));
}
check_equal(vpar_to_sparse6([1]), concat([":",Strchr([63+1,63+31]),""]), \
            "vpar_to_sparse6() n=1 with self-loop");
check_equal(vpar_to_sparse6([1,0]), concat([":",Strchr([63+2,63+7]),""]), \
            "vpar_to_sparse6() n=2 with self-loop 1<->1");
check_equal(vpar_to_sparse6([1,2]), concat([":",Strchr([63+2,63+15]),""]), \
            "vpar_to_sparse6() n=2 both self-loops");

check_equal(vpar_to_sparse6([0,1,0,0]), \
            concat([":",Strchr([63+4,63+32+7]),""]), \
            "vpar_to_sparse6() n=4 edge 0-1 (0-based)");
{
  \\ : N(4) 1-10    0-00
  \\        set v=2 edge from=0    (0-based)
  check_equal(Vecsmall("O"), Vecsmall([63+ 16 ]));
  check_equal(vpar_to_sparse6([3,0,0,0]), ":Co",
              "vpar_to_sparse6() n=4 edge 0-2 (0-based)");

  \\ : N(4) 1-10    0-00 0-01   0-11
  \\        set v=2 edge edge   pad   0-based
  check_equal(Vecsmall("oJ"), Vecsmall([111,74]));
  check_equal(Vecsmall("oJ"), Vecsmall([63+ 48, 63+ 11 ]));
  check_equal(vpar_to_sparse6([3,3,0,0]), ":CoJ",
              "vpar_to_sparse6() n=4 edges 0-2 1-2 (0-based), pad 0-11");
}


\\-----------------------------------------------------------------------------
progress("vpar_is_balanced_binary()");

check_equal(vpar_is_balanced_binary(0), 1, "vpar_is_balanced_binary() 0");
check_equal(vpar_is_balanced_binary(1), 0);
check_equal(vpar_is_balanced_binary(2), 1);
check_equal(vpar_is_balanced_binary(test_from_binary(1010)), 1);
check_equal(vpar_is_balanced_binary(test_from_binary(1100)), 1);
check_equal(vpar_is_balanced_binary(test_from_binary(1001)), 0);

check_equal(vpar_is_balanced_binary(test_from_binary(110)), 0);
check_equal(vpar_is_balanced_binary(test_from_binary(11001010)), 1);
check_equal(vpar_is_balanced_binary(test_from_binary(11001001)), 0);
check_equal(vpar_is_balanced_binary(test_from_binary(1000)), 0);

{
  for(n=1,8,
     my(lo=4^(n-1),
        hi=4^n-1);
     check_equal(#digits(lo-1,4), n-1, "lo-1");
     check_equal(#digits(lo,4), n, "lo");
     check_equal(#digits(hi,4), n, "hi");
     check_equal(#digits(hi+1,4), n+1, "hi+1");
     check_equal(sum(b=lo,hi, vpar_is_balanced_binary(b)),
                 vpar_INTERNAL_Catalan_number(n),
                 "vpar_is_balanced_binary() Catalan number"));
}

test_is_balanced_binary_by_binary(b) =
{
  my(vec=binary(b),
     ones=0);
  for(i=1,#vec,
     if(vec[i], ones++,
        ones-- < 0, return(0)));
  ones==0;
}
{
  for(b=0,1024,
     check_equal(vpar_is_balanced_binary(b),
                 test_is_balanced_binary_by_binary(b),
     "vpar_is_balanced_binary() vs test_is_balanced_binary_by_binary()"));
}

{
  check_equal(vector(#A080116_samples,b,b--; vpar_is_balanced_binary(b)),
              A080116_samples,    \\ OFFSET=0
              "A080116 is balanced binary predicate 0,1");
}


\\-----------------------------------------------------------------------------
progress("vpar_minmax_child_vector()");

check_equal(vpar_minmax_child_vector([],min), []);
check_equal(vpar_minmax_child_vector([],max), []);

test_minmax_child_vector_by_children(vpar,func) =
{
  my(vecfunc=if(func==min,vecmin,
                func==max,vecmax,
                error()));
  apply(c->if(#c==0,0,vecfunc(c)), vpar_children_vector(vpar));
}
{
  my(funcs=[min,max]);
  for(f=1,#funcs,
     my(func=funcs[f]);
     for(i=1,#test_sample_forests,
        my(vpar=test_sample_forests[i]);
        check_equal(vpar_minmax_child_vector(vpar,func),
                    test_minmax_child_vector_by_children(vpar,func),
                    "vpar_minmax_child_vector() vs test_minmax_child_vector_by_children()");
        check_equal(vpar_minmax_child_vector(vpar,func),
                    vector(#vpar,v, vpar_minmax_child_of_vertex(vpar,v,func)),
                    "vpar_minmax_child_vector() vs individual vpar_minmax_child_of_vertex()");
     ));
}

{
  \\ vpar_minmax_child_vector() addhelp, forests distinct vectors occurring
  \\ is num sets of ordered lists (A000262)
  my(funcs=[min,max],
     forest=1);
  for(f=1,#funcs,
     my(func=funcs[f]);
     for(n=0, test_limit_trees_search - forest,
        my(l=List([]));
        forvec(vec=vpar_forvec_from(n,forest),  \\ forests
              my(vpar=vpar_from_upascend(vec),
                 vec=vpar_minmax_child_vector(vpar,func),
                 pos=setsearch(l,vec,1));
              if(pos, listinsert(l,vec,pos)));
        \\ print1(#l",");
        check_equal(#l, A000262_samples[n+1],
                    "vpar_minmax_child_vector() distinct vectors = A000262");
     ));
}

{
  \\ vpar_minmax_child_vector() addhelp, count forests with increasing
  \\ min/max child is Bell numbers
  my(forest=1);
  for(n=0, test_limit_trees_search - forest,
     my(count_min_child_increasing=0,
        count_max_child_increasing=0,
        count_both_increasing=0);
     forvec(vec=vpar_forvec_from(n,forest),  \\ forests
           my(vpar=vpar_from_upascend(vec),
              min_child=vpar_minmax_child_vector(vpar,min),
              max_child=vpar_minmax_child_vector(vpar,max),
              is_min_child_increasing = test_vector_is_increasing(min_child),
              is_max_child_increasing = test_vector_is_increasing(max_child));
           count_min_child_increasing += is_min_child_increasing;
           count_max_child_increasing += is_max_child_increasing;
           count_both_increasing += (is_min_child_increasing
                                  && is_max_child_increasing);
     );
     check_equal(count_min_child_increasing, vpar_INTERNAL_Bell_number(n));
     check_equal(count_max_child_increasing, vpar_INTERNAL_Bell_number(n));
     check_equal(count_both_increasing, vpar_INTERNAL_Catalan_number(n));
  );
}
{
  \\ and trees Bell numbers 1 smaller
  my(forest=0);
  for(n=0, test_limit_trees_search - forest,
     my(count_min_child_increasing=0,
        count_max_child_increasing=0);
     forvec(vec=vpar_forvec_from(n,forest),  \\ trees
           my(vpar=vpar_from_upascend(vec),
              min_child=vpar_minmax_child_vector(vpar,min),
                 max_child=vpar_minmax_child_vector(vpar,max));
           count_min_child_increasing += test_vector_is_increasing(min_child);
           count_max_child_increasing += test_vector_is_increasing(max_child);
     );
     check_equal(count_min_child_increasing,
                 if(n==0,1,vpar_INTERNAL_Bell_number(n-1)));
     check_equal(count_max_child_increasing,
                 if(n==0,1,vpar_INTERNAL_Bell_number(n-1)));
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_minmax_child_of_vertex()");

check_equal(vpar_minmax_child_of_vertex([],0,min), 0);
check_equal(vpar_minmax_child_of_vertex([],0,max), 0);
check_equal(vpar_minmax_child_of_vertex([2,0],0,min), 2);
check_equal(vpar_minmax_child_of_vertex([2,0],0,max), 2);
check_equal(vpar_minmax_child_of_vertex([2,0],2,min), 1);
check_equal(vpar_minmax_child_of_vertex([2,0],2,max), 1);
check_equal(vpar_minmax_child_of_vertex([0,1,0,3],0,min), 1);
check_equal(vpar_minmax_child_of_vertex([0,1,0,3],0,max), 3);

{
  my(funcs=[min,max]);
  for(f=1,#funcs,
     my(func=funcs[f]);
     for(i=1,#test_sample_forests,
        my(vpar=test_sample_forests[i]);
));
}


\\-----------------------------------------------------------------------------
progress("test_poly_num_real_roots()");

check_equal(test_poly_num_real_roots(1), 0, \
            "test_poly_num_real_roots() constant");
check_equal(test_poly_num_real_roots(x), 1, \
            "test_poly_num_real_roots() x");
check_equal(test_poly_num_real_roots(x^2), 2);
check_equal(test_poly_num_real_roots((x-2)^4), 4);
check_equal(test_poly_num_real_roots(x^2+1), 0, \
            "test_poly_num_real_roots() I");
check_equal(test_poly_num_real_roots((x-2)^2*(x-4)^3), 5);
check_equal(test_poly_num_real_roots((x^2+2)*(x-4)^3), 3, \
            "test_poly_num_real_roots() imags");


\\-----------------------------------------------------------------------------
progress("vpar_make_preorder_depths()");

check_equal(vpar_make_preorder_depths([]), [], \
            "vpar_make_preorder_depths() empty");
check_equal(vpar_make_preorder_depths([0]), [0]);
check_equal(vpar_make_preorder_depths([0,0]), [0,0], \
            "vpar_make_preorder_depths() 2 singletons");
check_equal(vpar_make_preorder_depths([0,1,1,1,1]), [0,1,1,1,1]);
check_equal(vpar_make_preorder_depths([0,1]), [0,1]);
check_equal(vpar_make_preorder_depths([0,1,2,3,4]), [0,1,2,3,4]);

\\ previous code of vpar_make_preorder_depths() using extra space last[]
test_make_preorder_depths_by_last(depths) =
{
  \\ last[d+1] is the last vertex seen at depth d (depth d=0 upwards).
  \\ last[1] is the last root vertex seen.
  \\
  \\ Vertex v with depths[v]==0 is a root, so vpar[v]=0 and last[1]=v.
  \\ Vertex v with depths[v]==d >=1 gets parent last[d] which is the last at
  \\ depth above it, and becomes the new last[d+1]=v at this depth d.

  my(last=vectorsmall(#depths),
     vpar=vector(#depths));
  for(v=1,#vpar,
     my(d=depths[v]);
     if(d, vpar[v]=last[d]);
     last[d+1]=v);
  vpar;
}

\\ preceding p as per vpar_make_preorder_depths() addhelp description
test_make_preorder_depths_by_preceding(depths) =
{
  vector(#depths,v,
         if(depths[v]==0, 0,
            my(p=v-1);
            while(depths[p] != depths[v]-1, p--);
            p));
}

\\ old code going high to low with a linked list of those at depth d waiting
\\ to have their parent set
test_make_preorder_depths_by_chain(depths) =
{
  \\ Working from n to 1, for each depth d>=1 keep a linked list of vertex
  \\ numbers v with depths[v]==d.
  \\ At a depths[p]==d-1, set p as the parent of the vertices in the d
  \\ linked list and zap that list to empty again.
  \\ In this way a depths[v]==d gets parent which is the closest preceding
  \\ depths[p]==d-1.
  \\ The linked lists are kept in the vpar[] return being constructed.
  \\ vpar[d]=u = first vertex in the chain of depth d, then vpar[u]=next,
  \\ and so on until 0 for no more (including vpar[d]=0 for empty list).
  \\ Once the parents of a chain have been set they are not touched again.
  \\ The decreasing depths due to pre-order mean the linked lists do not
  \\ overlap the parents written.

  \\ print("vpar_make_preorder_depths() "depths);
  my(vpar=vector(#depths));
  forstep(v=#vpar,1,-1,
     my(d=depths[v]);
     \\ print(" v="v" d="d"  vpar="vpar);

     my(t=vpar[d+1], u);  \\ traverse linked list setting parent v
     while(u=t,
           \\ print("  chain u="u" parent v="v);
           t=vpar[u]; vpar[u]=v);
     vpar[d+1]=0;  \\ zap to empty list at d+1

     if(d,
        vpar[v]=vpar[d]; vpar[d]=v,  \\ prepend v to chain d
        vpar[v]=0));   \\ d==0 is root
  vpar;
}

\\ Total number of up steps in the vpar_make_preorder_depths() code.
\\ This is net n - depths[#depths] - 1.
test_make_preorder_depths_num_up_steps(depths) =
{
  my(total=0);
  for(v=2,#depths,
     for(i=depths[v],depths[v-1], total++));
  total;
}



{
  for(i=1,#test_sample_preorder_forests,
     my(vpar=test_sample_preorder_forests[i],
        depths=vpar_depths_vector(vpar),
        got_vpar=vpar_make_preorder_depths(depths));
     check_equal(vpar_depths_vector(got_vpar), depths,
                 "vpar_make_preorder_depths() depths again");

     check_equal(got_vpar,
                 test_make_preorder_depths_by_last(depths),
     "vpar_make_preorder_depths() vs test_make_preorder_depths_by_last()");

     check_equal(got_vpar,
                 test_make_preorder_depths_by_preceding(depths),
     "vpar_make_preorder_depths() vs test_make_preorder_depths_by_preceding()");

     check_equal(got_vpar,
                 test_make_preorder_depths_by_chain(depths),
     "vpar_make_preorder_depths() vs test_make_preorder_depths_by_chain()");

     check_equal(if(#depths, #depths - depths[#depths] - 1,  0),
                 test_make_preorder_depths_num_up_steps(depths),
                 "test_make_preorder_depths_num_up_steps() formula");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_count_min_ancestor_increasing()");

{
  \\ not in OEIS: 2, 8, 47, 365, 3570, 42796, 616610
  my(want=[1, 1, 2, 8, 47, 365, 3570, 42796, 616610, 10492857, 207316715,
           4678768094]);
  check_equal(vector(#want,n,n--; vpar_count_min_ancestor_increasing(n)),
              want,
              "vpar_count_min_ancestor_increasing() vs samples trees");
}
{
  \\ not in OEIS: 3, 14, 87, 675, 6343, 71176, 947341
  my(want=[1, 1, 3, 14, 87, 675, 6343, 71176, 947341]);
  check_equal(vector(#want,n,n--; vpar_count_min_ancestor_increasing(n,1)),
              want,
              "vpar_count_min_ancestor_increasing() vs samples forests");
}

{
  for(forest=-1,1,
     for(n=0, test_limit_trees_search - forest,        \\ bit slow for n>=6
        my(count_min_increasing=0,
           count_max_increasing=0,
           count_min_decreasing=0,
           count_max_decreasing=0,
           count_min_both=0,
           count_max_both=0);
        forvec(vec=vpar_forvec_from(n,forest),
              my(vpar=vpar_from_upascend(vec),
                 min_ancestors=vpar_minmax_ancestors_vector(vpar,min),
                 max_ancestors=vpar_minmax_ancestors_vector(vpar,max),
                 min_is_increasing =test_vector_is_increasing(min_ancestors),
                 max_is_increasing =test_vector_is_increasing(max_ancestors),
                 min_is_decreasing =test_vector_is_decreasing(min_ancestors),
                 max_is_decreasing =test_vector_is_decreasing(max_ancestors));
              count_min_increasing += min_is_increasing;
              count_max_increasing += max_is_increasing;
              count_min_decreasing += min_is_decreasing;
              count_max_decreasing += max_is_decreasing;
              count_min_both       += (min_is_increasing && min_is_decreasing);
              count_max_both       += (max_is_increasing && max_is_decreasing);
        );
        my(got=vpar_count_min_ancestor_increasing(n,forest));

        check_equal(got,
                    count_min_increasing,
                    Str("vpar_count_min_ancestor_increasing() as min incr"
                        " n="n" forest="forest));

        \\ max increasing - same for trees and forests, but root=1 no longer
        \\ all of vpar_count but only forests n-1
        \\
        check_equal(if(forest==-1, if(n==0,1,vpar_count_min_ancestor_increasing(n-1,1)),
                       got),
                    count_max_increasing,
                    "ancestors count_max_increasing");

        my(got_all_root1=vpar_count(n,-1));
        check_equal(got_all_root1, count_min_decreasing,
                    "ancestors count_min_decreasing");

        check_equal(if(forest==-1, (n<=1), got_all_root1),
                    count_max_decreasing,
                    "ancestors count_max_decreasing");

        check_equal(got_all_root1,
                    count_min_both,
                    "ancestors count_min_both increasing&decreasing");
        check_equal(if(forest==-1, (n<=1), got_all_root1),
                    count_max_both,
                    "ancestors count_max_both increasing&decreasing");
     ));
}


\\-----------------------------------------------------------------------------
progress("vpar_roots_ith()");

check_equal(vpar_roots_ith([],1), []);
check_equal(vpar_roots_ith([0],1), [1]);
check_equal(vpar_roots_ith([0,1],1), [1]);
check_equal(vpar_roots_ith([0,1],2), [2]);

check_equal(vpar_roots_ith([0,4,1,0],1),     [1,2]);
check_equal(vpar_roots_ith([0,4,1,0],2),     [1,4]);
check_equal(vpar_roots_ith([0,4,1,0],3), Set([3,2]));
check_equal(vpar_roots_ith([0,4,1,0],4),     [3,4]);

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     check_equal(vpar_roots_ith(vpar,1),
                 vpar_component_minmax(vpar,min),
                 "vpar_roots_ith() 1 vs vpar_component_minmax() min");
     check_equal(vpar_roots_ith(vpar, vpar_num_rootings(vpar)),
                 vpar_component_minmax(vpar,max),
                 "vpar_roots_ith() max vs vpar_component_minmax() max");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_reroot_ith()");

check_equal(vpar_reroot_ith([0,1],1), [0,1]);
check_equal(vpar_reroot_ith([0,1],2), [2,0]);


\\-----------------------------------------------------------------------------
progress("vpar_subtree_num_childless()");

check_equal(vpar_subtree_num_childless([0],1),1, \
            "vpar_subtree_num_childless() singleton");

check_equal(vpar_subtree_num_childless([0,1,2,3],1), 1, \
            "vpar_subtree_num_childless() path-4 v=1");
check_equal(vpar_subtree_num_childless([0,1,2,3],2), 1, \
            "vpar_subtree_num_childless() path-4 v=2");

check_equal(vpar_subtree_num_childless([0,1,1],1),2, \
            "vpar_subtree_num_childless() star-3");

check_equal(vpar_subtree_num_childless([1],1), 0, \
            "vpar_subtree_num_childless() self-loop singleton");
{
  for(v=1,2,
     check_equal(vpar_subtree_num_childless([2,1],v), 0,
                 "vpar_subtree_num_childless() 2-cycle"));
}
{
  for(v=1,3,
     check_equal(vpar_subtree_num_childless([2,1,1],v), 1,
                 "vpar_subtree_num_childless() cyclic 3");
     check_equal(vpar_subtree_num_childless([3,3,2],v), 1,
                 "vpar_subtree_num_childless() cyclic 3"));
}

test_cyclics_to_roots(vpar) =
{
  my(cyclics=vpar_cyclic_vertices(vpar));
  for(i=1,#cyclics, vpar[cyclics[i]]=0);
  vpar;
}
test_subtree_num_childless_by_subtree_break(vpar,v) =
{
  \\ if v is cyclic, then breaking its link to parent may make that parent
  \\ look like a childless below v.  This happens if vpar[v] doesn't have
  \\ any other children, only the 1 child v.

  my(adjust = vpar_is_cyclic_vertex(vpar,v)
             && vpar_num_children_of_vertex(vpar,vpar[v])==1);
  vpar[v]=0;
  vpar_num_childless(vpar_subtree(vpar,v)) - adjust;
}
{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     for(v=1,#vpar,
        check_equal(vpar_subtree_num_childless(vpar,v),
                    test_subtree_num_childless_by_subtree_break(vpar,v),
                    "vpar_subtree_num_childless() vs test_subtree_num_childless_by_subtree_break()");
        check_equal(vpar_subtree_num_childless(vpar,v),
                    vpar_num_childless(vpar_subtree(vpar,v)),
                    "vpar_subtree_num_childless() vs vpar_num_childless() of vpar_subtree()")));
}


\\-----------------------------------------------------------------------------
progress("vpar_subtree_childless_vertices()");

check_equal(vpar_subtree_childless_vertices([0],1), [1], \
            "vpar_subtree_childless_vertices() singleton");

check_equal(vpar_subtree_childless_vertices([2,3,4,1],1), [], \
            "vpar_subtree_childless_vertices() 4-cycle");

test_subtree_childless_vertices_by_intersect(vpar,v) =
{
  setintersect(vpar_childless_vertices(vpar),
               Set(vpar_subtree_vertices(vpar,v)));
}
{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     for(v=1,#vpar,
        my(got=vpar_subtree_childless_vertices(vpar,v));
        check_equal(got,
                    test_subtree_childless_vertices_by_intersect(vpar,v),
                    "vpar_subtree_childless_vertices() vs test_subtree_childless_vertices_by_intersect()");
        check_equal(#got, vpar_subtree_num_childless(vpar,v),
                    "vpar_subtree_childless_vertices() length vs vpar_subtree_num_childless()")));
}


\\-----------------------------------------------------------------------------
progress("vpar_subtree_num_childless_vector()");

check_equal(vpar_subtree_num_childless_vector([2,1]), [0,0]);
check_equal(vpar_subtree_num_childless_vector([2,1,1]), [1,1,1]);
check_equal(vpar_subtree_num_childless_vector([2,1,2,1]), [2,2,1,1]);
check_equal(vpar_subtree_num_childless_vector([2,1,2, 1, 4,4]), [3,3,1,2,1,1]);

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     check_equal(vpar_subtree_num_childless_vector(vpar),
                 vector(#vpar,v, vpar_subtree_num_childless(vpar,v)),
                 "vpar_subtree_num_childless_vector() vs individual vpar_subtree_num_childless()"));
}

test_subtree_num_childless_vector_of_acyclic(vpar) =
{
  my(seq=vpar_INTERNAL_seq_upwards(vpar),
     ret=vector(#vpar));
  for(i=1,#seq,
     my(v=seq[i]);
     if(!ret[v], ret[v]=1);
     if(vpar[v], ret[vpar[v]] += ret[v]));
  ret;
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     check_equal(vpar_subtree_num_childless_vector(vpar),
                 test_subtree_num_childless_vector_of_acyclic(vpar),
                 "vpar_subtree_num_childless_vector() vs test_subtree_num_childless_vector_of_acyclic()"));
}


\\------------------------------------------------------------------------------
progress("test_num_automorphism_generators_max()");

check_equal(test_num_automorphism_generators_max(0), 0);
check_equal(test_num_automorphism_generators_max(1), 0);
check_equal(test_num_automorphism_generators_max(2), 0);
check_equal(test_num_automorphism_generators_max(0,1), 0);
check_equal(test_num_automorphism_generators_max(1,1), 0);
check_equal(test_num_automorphism_generators_max(2,1), 1);
{
  check_equal(vector(#A004524_samples,n,n--;
                     test_num_automorphism_generators_max(n)), \\ trees
              A004524_samples,    \\ OFFSET=0
              "A004524 three even, one odd = max generators for tree, in current generators style");
}


\\------------------------------------------------------------------------------
progress("vpar_num_free_automorphisms()");

check_equal(vpar_num_free_automorphisms([]), 1, \
           "vpar_num_free_automorphisms() empty");

check_equal(vpar_num_free_automorphisms([0,1]), 2, \
           "vpar_num_free_automorphisms() path-2");
check_equal(vpar_num_automorphisms([0,1]), 1, \
           "vpar_num_automorphisms() path-2");

test_num_free_automorphisms_by_search(vpar) = \
  sum(i=1,(#vpar)!, vpar_is_free_automorphism(vpar,numtoperm(#vpar,i)));

test_num_free_automorphisms_of_forest_by_components(vpar) =
{
  my(component_vpar=vpar_component_trees(vpar),
     ret=prod(i=1,#component_vpar, vpar_num_free_automorphisms(component_vpar[i])),
     same=vector(#component_vpar,i,1));
  for(i=1,#component_vpar,
     for(j=1,i-1,
        if(vpar_is_free_isomorphic(component_vpar[i],component_vpar[j]),
           ret *= same[j]++;   \\ accumulating in the first of isomorphics
           break())));
  ret;
}
{
  \\  2 3 5 6 7
  \\    | |
  \\    1 4
  my(vpar=[3, 0, 0, 5, 0, 0, 0]);
  check_equal(2*2 * 2 * 3!, 48);
  check_equal(vpar_num_free_automorphisms(vpar), 48,
              "vpar_num_free_automorphisms() 7");
  check_equal(test_num_free_automorphisms_of_forest_by_components(vpar), 48,
              "test_num_free_automorphisms_of_forest_by_components() 7");
}


{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got=vpar_num_free_automorphisms(vpar));

     check_equal(1 <= got && got <= (#vpar)!, 1,
                 "vpar_num_free_automorphisms() range");

     if(#vpar <= test_limit_perms_search,
        check_equal(got, test_num_free_automorphisms_by_search(vpar),
                    "vpar_num_free_automorphisms() vs search");
       );

     check_equal(got,
                 test_num_free_automorphisms_of_forest_by_components(vpar),
                 "vpar_num_free_automorphisms() vs test_num_free_automorphisms_of_forest_by_components()");
  );
}
{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i],
        got=vpar_num_free_automorphisms(vpar));

     if(#vpar_centres(vpar)==1,
        check_equal(got <= (#vpar-1)!, 1,
          "vpar_num_free_automorphisms() maximum of unicentral tree"));

     \\ Jaroslav Nesetril, "A Congruence Theorem For Asymmetric Trees",
     \\ Pacific Journal of Mathematics, volume 37, number 3, March 1971,
     \\ pages 771-778.  https://projecteuclid.org/euclid.pjm/1102970478
     \\
     \\ Corollary 1.2, removing a leaf node from a free asymmetric tree
     \\ gives num automorphisms <= 2 in the remainder.
     \\
     if(got==1, \\ asymmetric
        my(degrees=vpar_degrees_vector(vpar));
        for(v=1,#vpar,
           if(degrees[v]==1,
              my(sub=vpar_delete_vertex(vpar,v));
              \\ print(vpar" "sub);
              check_equal(vpar_num_free_automorphisms(sub) <= 2,
                          1,
        "vpar_num_free_automorphisms() of asymmetric with leaf-deleted"))));
     );
}

\\ Return 1 if vpar is a path-2, of any labelling.
test_is_path2(vpar) = vpar==[0,1] || vpar==[2,0];
{
  \\ only trees of all perms automorphisms are path 0,1,2
  for(n=0,8,
     my(vpar=vpar_free_first(n),
        target=n!);
     until((vpar=vpar_free_next(vpar))=='none,
           check_equal(vpar_num_free_automorphisms(vpar) == target,
                       vpar==[] || vpar==[0] || test_is_path2(vpar),
                       "vpar_num_free_automorphisms() trees n! max singletons or path-2"));
  );
}
{
  \\ Per vpar_num_free_automorphisms() addhelp, all n! perms as
  \\ automorphisms occurs when vpar is path-2 or forest of singletons.
  for(n=0,6,
     my(vpar=vpar_unlabelled_first(n),
        target=n!);
     until((vpar=vpar_unlabelled_next(vpar,1))=='none,  \\ forests
           check_equal(vpar_num_free_automorphisms(vpar) == target,
                       test_is_all_singletons(vpar) || test_is_path2(vpar),
                       "vpar_num_free_automorphisms() trees n! max singletons or path-2");
           \\ if(vpar_num_free_automorphisms(vpar) == target,
           \\    print(vpar));
          );
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_free_automorphisms()");

check_equal(vpar_free_automorphisms([]), [ Vecsmall([]) ], \
            "vpar_free_automorphisms() empty");
check_equal(vpar_free_automorphisms([0]), [ Vecsmall([1]) ], \
            "vpar_free_automorphisms() singleton");
check_equal(vpar_free_automorphisms([0,0]), \
            [ Vecsmall([1,2]), Vecsmall([2,1]) ], \
            "vpar_free_automorphisms() two singletons");
check_equal(vpar_free_automorphisms([0,1]), \
            [ Vecsmall([1,2]), Vecsmall([2,1]) ], \
            "vpar_free_automorphisms() path-2");
check_equal(vpar_free_automorphisms([0,1,1]), \
            [ Vecsmall([1,2,3]), Vecsmall([1,3,2]) ], \
            "vpar_free_automorphisms() star-3");

test_free_automorphisms_by_generators(vpar) = \
  test_perms_generators_to_group(vpar_free_automorphism_generators(vpar),#vpar);

\\ Return a vector of all automorphism permutations of vpar (including the
\\ identify permutation).
test_free_automorphisms_by_search(vpar) =
{
  my(l=List([]));
  for(i=0,(#vpar)!-1,
     my(perm=numtoperm(#vpar,i));
     if(vpar_is_free_automorphism(vpar,perm),
        listput(l,Vecsmall(perm))));
  Vec(l);
}
check_equal(test_free_automorphisms_by_search([0]), [Vecsmall([1])]);

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        num_free_automorphisms=vpar_num_free_automorphisms(vpar));
     num_free_automorphisms <= test_limit_sets || next;

     my(got=vpar_free_automorphisms(vpar));
     check_equal(#got, num_free_automorphisms,
                 "vpar_free_automorphisms() count vs vpar_num_free_automorphisms()");
     check_equal(#got, #Set(got),
                 "vpar_free_automorphisms() no duplicates");
     for(i=1,#got,
        check_equal(#got[i], #vpar);
        check_equal(test_is_perm(got[i]), 1));
     check_equal(test_perms_is_group(got), 1,
                 "vpar_free_automorphisms() vs test_perms_is_group()");

     got = vecsort(got);
     check_equal(got, test_free_automorphisms_by_generators(vpar),
       "vpar_free_automorphisms() vs test_free_automorphisms_by_generators()");
     if(#vpar <= test_limit_perms_search,
        check_equal(got, test_free_automorphisms_by_search(vpar),
          "vpar_free_automorphisms() vs test_free_automorphisms_by_search()"));

     \\ FIXME: No, generators not currently minimum.
     \\ check_equal(vpar_num_free_automorphism_generators(vpar),
     \\             test_perms_minimum_num_generators(got),
     \\             "vpar_num_free_automorphism_generators() vs test_perms_minimum_num_generators()");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_num_cycles()");

check_equal(vpar_num_cycles([]), 0, "vpar_num_cycles() empty");
check_equal(vpar_num_cycles([0]), 0);
check_equal(vpar_num_cycles([0,1]), 0);
check_equal(vpar_num_cycles([2,1]), 1);
check_equal(vpar_num_cycles([2,1,3]), 2);

\\ Previous implementation.
test_num_cycles_by_old_code(vpar) =
{
  \\ Search upwards from each vertex i, similar to vpar_is_cyclic().
  \\ seen[v]=i for the vertices traversed.
  \\ If reach seen[v]!=0 and !=i then it is a previous search up and any
  \\ cycle above it has been counted.
  \\ If reach seen[v]==i then it is a re-visit so a cycle.

  my(seen=vectorsmall(#vpar),
     ret=0);
  for(i=1,#vpar,
     my(v=i);
     until(!(v=vpar[v]),
           if(seen[v], ret += (seen[v]==i); break());
           seen[v]=i));
  ret;
}

\\ WRONG
\\ test_num_cycles_by_cyclic_vertex(vpar) =
\\ {
\\   my(count=0);
\\   for(v=1,#vpar,
\\      if(vpar_is_cyclic_vertex(vpar,v),
\\         count += (v==vpar_minmax_ancestor_of_vertex(vpar,v,min))));
\\   count;
\\ }
     \\ check_equal(got, test_num_cycles_by_cyclic_vertex(vpar));
{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        got=vpar_num_cycles(vpar));
     check_equal(got, test_num_cycles_by_old_code(vpar)));
}

{
  \\ total num cycles in all functional digraphs
  \\ A216534
  check_equal(vector(6,n,n--;  \\ n=0 upwards
                     my(total=0);
                     forvec(vpar=vector(n,i,[0,n]),
                           total += vpar_num_cycles(vpar));
                     total),
              A216534_samples[1..6],
              "A216534 total vpar_num_cycles()");

  my(limit = min(20,#A216534_samples),
     gA216534 = test_vector_to_egf(A216534_samples,0,limit),
     gT = test_func_to_egf(vpar_count, limit),            \\ trees  n^(n-1)
     gF = test_func_to_egf(n->vpar_count(n,1), limit));   \\ forests (n+1)^(n-1)
  check_equal(test_egf_terms(gA216534,limit), A216534_samples[1..limit],
              "gA216534 terms");

  my(gT1 = gT-1);
  check_equal(gT', gT1');
  \\ print("gT1   ",test_egf_terms(gT1, 10));
  \\ print("recip ",test_egf_terms(1/(1-gT1) - 1, 10));

  \\ per egf in A216534 by Geoffrey Critzer
  check_equal(gA216534, exp(gT1)*log(1/(1-gT1))/(1-gT1));

  check_equal(gA216534, gT' * log(1/(2-gT)));
  check_equal(gA216534, gT1' * log(1/(1-gT1)));
  check_equal(gA216534, gT' * log(1+x*gT'));

  \\ print(test_egf_terms(intformal(log(1+x*gT')), 10));
}

{
  \\ num cyclic 1-component functional digraphs
  \\ A001865 connected functions on n nodes
  \\ being "cyclic" excludes trees, since they have no function value at the
  \\ root
  check_equal(vector(5,n,  \\ n=1 upwards
                     my(count=0);
                     forvec(vpar=vector(n,i,[0,n]),
                           count += (vpar_is_cyclic(vpar)
                                     && vpar_num_components(vpar)==1));
                     count),
              A001865_samples[1..5],
              "A001865 num 1-component, 1-cycle functional digraphs");
  check_equal(vector(#A001865_samples,n, sum(k=1,n, n!/(n-k)! * n^(n-k-1))),
              A001865_samples);    \\ OFFSET=1
}
{
  \\ all components cyclic, is n^n
  \\ no roots, so all parents all possible 1..n,
  \\ which is n edges as all n vertices have parents
  check_equal(vector(5,n,n--;  \\ n=0 upwards
                     my(count=0);
                     forvec(vpar=vector(n,i,[0,n]),
                           my(want = (vpar_num_components(vpar)
                                      == vpar_num_cycles(vpar)));
                           \\ check_equal(want, vpar_num_edges(vpar)==n);
                           count += want);
                     count),
              vector(5,n,n--; n^n));
}

{
  \\ Svante Janson, Donald E. Knuth, Tomasz Luczak, Boris Pittel, "Birth of
  \\ the Giant Component", Random Structures and Algorithms, volume 4, 1993,
  \\ pages 233-358.
  \\ https://arxiv.org/abs/math/9310236
  \\ https://onlinelibrary.wiley.com/doi/abs/10.1002/rsa.3240040303
  \\ Section 3. Trees, Unicycles and Bicycles.

  \\ A057500 num connected labelled graphs n nodes, n edges
  my(Vhat = 1/6*x^3 + 5/8*x^4 + 37/20*x^5 + 61/12*x^6);
  check_equal(test_egf_terms(Vhat,7), [0, 0, 0, 1, 15, 222, 3660]);
}
{
  \\ num unicyclic functional digraphs (incl self-loops)
  \\          n = 0  1  2   3    4     5
  \\ not in OEIS: 0, 1, 5, 35, 328, 3894
  \\ num <=1 cycles functional digraphs
  \\ not in OEIS: 1, 2, 8, 51, 453, 5190

  \\ check_equal(vector(6,n,n--;  \\ n=0 upwards
  \\                    my(count=0);
  \\                    forvec(vpar=vector(n,i,[0,n]),
  \\                          count += (vpar_num_cycles(vpar)==1));
  \\                    count),
  \\             A_samples[1..6],
  \\             "A num vpar_num_cycles()==1");
}


\\-----------------------------------------------------------------------------
progress("test_vector_to_egf()");

check_equal(test_vector_to_egf([100,200,300]), \
            100*x/1 + 200*x^2/2! + 300*x^3/3! + O(x^4));
check_equal(test_vector_to_egf([100,200,300],0), \
            100 + 200*x/1 + 300*x^2/2! + O(x^3));
check_equal(test_vector_to_egf([100,200,300],0,2), \
            100 + 200*x/1 + O(x^2));


\\------------------------------------------------------------------------------
progress("Various EGFs f(n) Variations");

{
  my(limit=10,
     L = O(x^limit),
     v = vector(limit,i, prime(i+20)),   \\ vector of random values

     \\ func of random values, v_func(n) for 1<=n<=limit
     v_func = n->n++; if(n<1||n>#v,0, v[n]),

     \\ egf of v values
     g = test_func_to_egf(v_func,limit));

  check_equal(test_egf_terms(g,limit), v, "test_egf_terms()");

  \\ f(n+1) = g'
  check_equal(test_func_to_egf(n-> v_func(n+1), limit-2),
              g'                                         + L/x^2,
              "g' shift down, f(n+1)");

  \\ f(n+2) = g''
  check_equal(test_func_to_egf(n-> v_func(n+2), limit-3),
              g''                                        + L/x^3);

  \\ f(n-1) = intformal(g)
  check_equal(test_func_to_egf(n-> v_func(n-1), limit),
              intformal(g)                               + L,
              "intformal(g) shift up");

  \\ n*f(n-1) = x*g
  check_equal(test_func_to_egf(n-> n*v_func(n-1), limit),
              x*g  + L,
              "n*f(n-1)");
  \\ n*f(n) = x*g'
  check_equal(test_func_to_egf(n-> n*v_func(n), limit),
              x*g',
              "n*f(n)");
  \\ (n+1)*f(n) = x*g' + g
  check_equal(test_func_to_egf(n-> (n+1)*v_func(n), limit),
              x*g' + g);

  \\ 1/(n+1)*f(n+1)
  check_equal(test_func_to_egf(n-> 1/(n+1)*v_func(n+1), limit-1),
              (g-v[1])/x,
              "1/(n+1)*f(n+1)");
  \\ 1/n*f(n)
  check_equal(test_func_to_egf(n->if(n, 1/n*v_func(n) ), limit),
              intformal((g-v[1])/x));

  \\ 1/(n+1)*f(n)
  check_equal(test_func_to_egf(n-> 1/(n+1)*v_func(n), limit),
              intformal(g)/x);

  \\ (n-1)*n*f(n)
  check_equal(test_func_to_egf(n-> (n-1)*n*v_func(n), limit),
              x^2*g'');

  \\ n^2*f(n)
  check_equal(test_func_to_egf(n-> n^2*v_func(n), limit),
              x*(x*g')');
  check_equal(test_func_to_egf(n-> n^2*v_func(n), limit),
              x*(x'*g' + x*g''));
  check_equal(test_func_to_egf(n-> n^2*v_func(n), limit),
              x*g' + x^2*g'');

  \\ exponential transform with 1s  = g*exp(x)
  check_equal(test_func_to_egf(n-> sum(t=0,n, binomial(n,t)*v_func(t)), limit),
              test_egf1s(limit) * g);
  check_equal(test_func_to_egf(n-> sum(t=0,n, binomial(n,t)*v_func(t)), limit),
              g*exp(x));


  \\---------------------------------------------------------------------------
  \\ product of egfs

  my(u = vector(limit,i, prime(i+40)),
     u_func = n->n++; if(n<1||n>#u,0, u[n]),
     h = test_func_to_egf(u_func,limit));
  check_equal(test_egf_terms(h,limit), u, "test_egf_terms h");

  \\ binomial cross product sum
  check_equal(test_func_to_egf(n-> sum(t=0,n,
                    binomial(n,t) * v_func(t) * u_func(n-t)), limit),
              g*h);

  check_equal(matrix(6,6,n,t,n--;t--; binomial(n,t)),
              matrix(6,6,n,t,n--;t--; if(t==0,1, binomial(n-1,t-1)*n/t)));

  check_equal(test_func_to_egf(n->
                    v_func(0) * u_func(n) +
                    sum(t=1,n,
                    binomial(n-1,t-1)*n/t * v_func(t) * u_func(n-t)), limit),
              g*h);

  \\ integration by parts
  check_equal(g*h - v[1]*u[1], intformal(g'*h) + intformal(g*h'));


  \\---------------------------------------------------------------------------
  \\ total something in all trees n, to total in all forests n
  \\ gTotalF = gNumF * gTotalT,   expecting that total(0) = 0

  my(NumT_func   = n->if(n==0,1, v_func(n)),
     TotalT_func = n->if(n==0,0, u_func(n)),
     gNumT   = test_func_to_egf(NumT_func,limit),
     gTotalT = test_func_to_egf(TotalT_func,limit));
  check_equal(TotalT_func(0), 0);

  my(NumF_vec = vector(limit));
  for(n=1,#NumF_vec,
     NumF_vec[n] = NumT_func(n)
          + sum(t=1,n-1,
                \\ print("n="n" t="t" binomial "binomial(n,t)
                \\       " product ",NumT_func(t)," * ",NumF_vec[n-t]);
                binomial(n-1,t-1) * NumT_func(t) * NumF_vec[n-t]));
  my(NumF_func = n->if(n<1||n>#NumF_vec,0, NumF_vec[n]));

  \\ vpar_count(2) == 2
  \\ vpar_count(2,1) == 3
  \\ 1--2 or 2--1  or 1 2
  check_equal(NumF_func(2), NumT_func(2) + NumT_func(1)^2);
  check_equal(NumF_func(1), NumT_func(1));

  my(NumF_func = n->if(n<0,0, n==0,1, n>#NumF_vec,0, NumF_vec[n]),
     gNumF     = test_func_to_egf(NumF_func,limit));
  check_equal(gNumF, exp(gNumT-1));  \\ exponential transform

  my(TotalF_vec = vector(limit-1));
  for(n=1,#TotalF_vec,
     TotalF_vec[n] = TotalT_func(n)
          + sum(t=1,n-1,
              binomial(n-1,t-1)
              * (  TotalT_func(t) * NumF_func(n-t)
                 + TotalF_vec[n-t] * NumT_func(t))));
  my(TotalF_func = n->if(n<1||n>#TotalF_vec,0, TotalF_vec[n]),
     gTotalF = test_func_to_egf(TotalF_func,limit));
  check_equal(TotalF_func(0), 0);
  check_equal(TotalF_func(1), TotalT_func(1));

  check_equal(TotalF_func(2), TotalT_func(2) + 2*TotalT_func(1)*NumT_func(1));

  check_equal(vector(limit-1,n, TotalF_func(n)),
              vector(limit-1,n,
                     TotalT_func(n)
                     + sum(t=1,n-1,
                           binomial(n-1,t-1)
                           * (  TotalT_func(t) * NumF_func(n-t)
                              + TotalF_func(n-t) * NumT_func(t)))));

  check_equal(vector(limit-1,n, TotalF_func(n)),
              vector(limit-1,n,
                     TotalT_func(n)
                     + sum(t=1,n-1,   \\ k=0 to n-2=m-1
                           my(m=n-1, k=t-1);  \\ t=k+1
                           binomial(m,k)
                           * (  TotalT_func(k+1) * NumF_func(m-k)
                              + TotalF_func(m-k) * NumT_func(k+1)))));

  check_equal(vector(limit-1,n, TotalF_func(n)),
              vector(limit-1,n,
                     my(m=n-1);
                     TotalT_func(n)
                     + sum(k=0,m-1,
                           binomial(m,k)
                           * (  TotalT_func(k+1) * NumF_func(m-k)
                              + TotalF_func(m-k) * NumT_func(k+1)))));

  check_equal(vector(limit-1,n, TotalF_func(n)),
              vector(limit-1,n,
                     my(m=n-1);
                       sum(k=0,m,
                           binomial(m,k)
                           * (  TotalT_func(k+1) * NumF_func(m-k)
                              + TotalF_func(m-k) * NumT_func(k+1)))));

  check_equal(gTotalF, intformal(gTotalT'*gNumF + gTotalF * gNumT'));

  check_equal(gTotalF', gTotalT'*gNumF + gTotalF * gNumT');
  check_equal(gTotalF' + (-gNumT')*gTotalF, gTotalT'*gNumF);

  \\ first order ordinary differential equation
  \\ y' + P(x)*y = Q(x)
  \\ y = 1/u(x) int Q(x)*u(x)  where u(x) = exp(int P(x))

  my(P = -gNumT',
     Q = gTotalT'*gNumF,
     u = exp(intformal(P)));

  check_equal(u, exp(intformal(-gNumT')), "ODE u");
  check_equal(u, exp(-gNumT+1),           "ODE u");
  check_equal(1/u, exp(gNumT-1),          "ODE 1/u");
  check_equal(1/u, gNumF);

  check_equal(gTotalF,  1/u * intformal( Q*u ));
  check_equal(gTotalF,  gNumF * intformal( gNumF * gTotalT' * 1/gNumF ));
  check_equal(gTotalF,  gNumF * intformal(gTotalT'));
  check_equal(gTotalF,  gNumF * gTotalT);
}


\\-----------------------------------------------------------------------------
progress("vpar_count_unlabelled_asymmetric()");
\\ vpar_count_unlabelled_asymmetric_vector()

{
  \\ A004111 num asymmetric rooted unlabelled trees
  my(samples=A004111_samples);
  check_equal(samples[1], 0, "A004111 count 0 at n=0");
  samples[1] = 1;
  check_equal(vector(#samples,n,n--; vpar_count_unlabelled_asymmetric(n)),
              samples,
              "A004111 number of asymmetric rooted unlabelled trees");

  check_equal(vector(#samples,n,n-=2;
                     if(n<0,1, vpar_count_unlabelled_asymmetric(n,1))),
              samples,
    "A004111 number of asymmetric rooted unlabelled trees, as forests");
}
{
  for(forest=0,1,
     for(len=0,5,
        check_equal(vpar_count_unlabelled_asymmetric_vector(len,forest),
                    vector(len,n, vpar_count_unlabelled_asymmetric(n,forest)),
     "vpar_count_unlabelled_asymmetric_vector() vs individual vpar_count_unlabelled_asymmetric()")));
}
{
  for(forest=0,1,
     for(n=0,12-forest,
        my(vpar=vpar_unlabelled_first(n),
           count=0);
        until((vpar=vpar_unlabelled_next(vpar,forest))=='none,
              count += vpar_is_asymmetric(vpar));
        check_equal(count, vpar_count_unlabelled_asymmetric(n,forest),
                    Str("vpar_count_unlabelled_asymmetric() vs vpar_is_asymmetric(), forest="forest));
     ));
}
{
  for(forest=0,1,
     for(n=-5,-1,
        check_equal(iferr(vpar_count_unlabelled_asymmetric(n,forest);0, e, 1),
                    1,
                    "vpar_count_unlabelled_asymmetric() error on negative n")));
}

{
  my(T=vector(10,n, vpar_count_unlabelled_asymmetric(n)),
     F=vector(10,n, vpar_count_unlabelled_asymmetric(n,1)));
  check_equal(test_weigh_transform(T), F,
  "vpar_count_unlabelled_asymmetric() unlabelled weigh transform trees -> forests");
}
{
  my(limit=15,
     L=O(x^limit),
     g=x+L);      \\ c*x^n = how many n vertex asymmetric unlabelled rooted trees
  for(n=1,limit-1,
     g *= (1 + x^n +L)^vpar_count_unlabelled_asymmetric(n));
  check_equal(Vec(g), vector(limit-1,n, vpar_count_unlabelled_asymmetric(n)),
              "vpar_count_unlabelled_asymmetric() gf product");
}


\\-----------------------------------------------------------------------------
progress("vpar_is_free_asymmetric()");

\\ any relabel perm other than the identity is not asymmetric
test_is_free_asymmetric_by_search(vpar) =
{
  for(i=1,(#vpar)!-1,  \\ non-identity perms
     my(perm=numtoperm(#vpar,i));
     if(vpar_is_free_automorphism(vpar,perm),
        return(0)));
  1;
}
check_equal(test_is_free_asymmetric_by_search(vpar_make_path(4)), 0, \
           "test_is_free_asymmetric_by_search() path-4");
check_equal(test_is_free_asymmetric_by_search(vpar_make_path(6)), 0);
check_equal(test_is_free_asymmetric_by_search(vpar_make_path(8)), 0);
check_equal(test_is_free_asymmetric_by_search(vpar_free_first(8)), 0);
check_equal(vpar_is_free_asymmetric(vpar_free_first(8)), 0);

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     if(#vpar <= test_limit_perms_search,
        \\ print(vpar);
        check_equal(vpar_is_free_asymmetric(vpar),
                    test_is_free_asymmetric_by_search(vpar),
                    "vpar_is_free_asymmetric() vs test_is_free_asymmetric_by_search()")));
}


\\-----------------------------------------------------------------------------
progress("vpar_count_height_le()");

check_equal(vpar_count(2,1), 3, "vpar_count() n=2 forests");
check_equal(vpar_count_height_le(0,-1,1), 1, \
            "vpar_count_height_le() n=0 h=-1 forests");
check_equal(vpar_count_height_le(0,0,1), 1);
check_equal(vpar_count_height_le(0,1,1), 1);
check_equal(vpar_count_height_le(0,99,1), 1);

check_equal(vpar_count_height_le(1,-1,1), 0, \
            "vpar_count_height_le() n=1 h=-1 forest=1");
check_equal(vpar_count_height_le(1,0,1), 1, \
            "vpar_count_height_le() n=1 h=0 forest=1");
check_equal(vpar_count_height_le(1,1,1), 1);
check_equal(vpar_count_height_le(1,2,1), 1);
check_equal(vpar_count_height_le(1,99,1), 1);

check_equal(vpar_count_height_le(2,0,1), 1, \
            "vpar_count_height_le() n=2 h=0 forests");
check_equal(vpar_count_height_le(2,1,1), 3);
check_equal(vpar_count_height_le(2,2,1), 3);

{
  \\ h<=n-1 is all trees or forests
  for(forest=-1,1,
     for(n=1,10,
        check_equal(vpar_count_height_le(n,n-1,forest),
                    vpar_count(n,forest),
                    "vpar_count_height_le() h=n-1 vs full vpar_count()")));
}

{
  for(forest=-1,1,
     for(n=0, test_limit_trees_search - forest,
        my(offset=3,
           count=vector(n+offset+1));  \\ count[h+offset] for count of h
        forvec(vec=vpar_forvec_from(n,forest),
              my(vpar=vpar_from_upascend(vec),
                 height=vpar_height(vpar));
              for(i=height+offset,#count, count[i]++));
        \\ print(count);

        for(i=1,#count,
           my(h=i-offset);
           check_equal(vpar_count_height_le(n,h,forest), count[i],
                       Str("vpar_count_height_le()"
                           " n="n" h="h" forest="forest)));
        ));
}

{
  \\ A236396 count trees of height <= h
  my(l=List([]));
  for(n=1,#A236396_samples,
     for(h=0,n-1,
        if(#l==#A236396_samples,break(2));
        listput(l,vpar_count_height_le(n,h))));
  check_equal(Vec(l),A236396_samples,
              "vpar_count_height_le() vs A236396_samples");

}
{
  \\ A210725 count forests of height <= h
  my(l=List([]));
  for(n=1,#A210725_samples,
     for(h=0,n-1,
        if(#l==#A210725_samples,break(2));
        listput(l,vpar_count_height_le(n,h,1))));
  check_equal(Vec(l),A210725_samples);

}

{
  \\ A000248 count forests height <=1
  \\ exponential transform of 1,2,3,...
  my(got=vector(30,n,vpar_count_height_le(n,1,1)));
  check_equal(got, vpar_INTERNAL_exponential_transform(vector(30,n, n)));

  \\ Paul D. Hanna in A000248
  \\ Choose k roots, others (n-k)^k under those.
  check_equal(got, vector(30,n, sum(k=0,n, binomial(n,k)*(n-k)^k)));

  check_equal(vector(#A000248_samples,n,n--; vpar_count_height_le(n,1,1)),
              A000248_samples,  \\ OFFSET=0
              "A000248 vpar_count_height_le(n,1,1)");
}

\\ forests <= h, by recurrence
\\ {
\\   if(n<0 || h<-1, return(0));
\\   if(h==-1,return(n==0));
\\   if(n==0,return(1));
\\
\\   \\ print("vpar_count_height_le() n="n" h="h);
\\   h=min(h,n-1);
\\   sum(t=1,n,
\\       \\ print(" t="t" "t * binomial(n-1,t-1)" "vpar_count_height_le(t-1,h-1)" "vpar_count_height_le(n-t,h));
\\       t * binomial(n-1,t-1)
\\       * vpar_count_height_le(t-1,h-1)
\\       * if(t==n,1,vpar_count_height_le(n-t,h)));
\\ }


\\-----------------------------------------------------------------------------
progress("vpar_INTERNAL_preorder_prev_sibling_of_vertex()");

{
  for(i=1,#test_sample_preorder_forests,
     my(vpar=test_sample_preorder_forests[i]);
     check_equal(vector(#vpar,v,
                        \\ allowing gnil here, for the internals
                        my(p=vpar_INTERNAL_preorder_prev_sibling_of_vertex(vpar,v));
                        if(p,p,0)),
                 vpar_next_sibling_vector(vpar,1),
       "vpar_INTERNAL_preorder_prev_sibling_of_vertex() vs vpar_next_sibling_vector()"));
}


\\-----------------------------------------------------------------------------
progress("vpar_INTERNAL_factorials_vector()");

check_equal(vpar_INTERNAL_factorials_vector(10), \
            vector(10,n, n!), \
            "vpar_INTERNAL_factorials_vector()");
check_equal(vpar_INTERNAL_factorials_vector(10,1), \
            vector(10,n, (n-1)!), \
            "vpar_INTERNAL_factorials_vector() offset=1");


\\-----------------------------------------------------------------------------
progress("test_to_dreadnaut()");

check_equal(test_to_dreadnaut([]), "n=0 $=1 g.");
check_equal(test_to_dreadnaut([0]), "n=1 $=1 g.");
check_equal(test_to_dreadnaut([0,1]), "n=2 $=1 g; 1.");
check_equal(test_to_dreadnaut([0,1,2]), "n=3 $=1 g; 1; 2.");
check_equal(test_to_dreadnaut([2,3,0]), "n=3 $=1 g2; 3.");

\\ down=1
check_equal(test_to_dreadnaut([],1), "n=0 $=1 g.");
check_equal(test_to_dreadnaut([0],1), "n=1 $=1 g.");
check_equal(test_to_dreadnaut([0,1],1), "n=2 $=1 g2.", \
            "test_to_dreadnaut() path-2, down");
check_equal(test_to_dreadnaut([0,1,2]), "n=3 $=1 g; 1; 2.", \
            "test_to_dreadnaut() path-3, down");
check_equal(test_to_dreadnaut([0,1,0,2]), "n=4 $=1 g; 1 4:2.");
check_equal(test_to_dreadnaut([3,3,0,1]), "n=4 $=1 g3; 3 4:1.");

\\ down=2, both ways
check_equal(test_to_dreadnaut([0,1],2), "n=2 $=1 g2; 1.");

check_equal(test_to_dreadnaut([0,1,1,1]), "n=4 $=1 g; 1; 1; 1.", \
            "test_to_dreadnaut() star-4");
check_equal(test_to_dreadnaut([0,1,1,1],1), "n=4 $=1 g2,3,4.", \
            "test_to_dreadnaut() star-4, down");

check_equal(test_to_dreadnaut([1]), "n=1 $=1 g1.", \
            "test_to_dreadnaut() self-loop 1");
check_equal(test_to_dreadnaut([1,2]), "n=2 $=1 g1; 2.", \
            "test_to_dreadnaut() self-loop 2");
check_equal(test_to_dreadnaut([3,1,2]), "n=3 $=1 g3; 1; 2.", \
            "test_to_dreadnaut() cycle");

\\   2    3
\\   |
\\   5
\\   |
\\   1
\\   |
\\   4
check_equal(test_to_dreadnaut([5,0,0,1,2]),    "n=5 $=1 g5 4:1; 2.");
check_equal(test_to_dreadnaut([5,0,0,1,2], 1), "n=5 $=1 g4; 5 5:1.");

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got=test_to_dreadnaut(vpar));
  );
}


\\-----------------------------------------------------------------------------
progress("test_dreadnaut_roots()");

check_equal(test_dreadnaut_roots([]),        "f=[]");
check_equal(test_dreadnaut_roots([0]),       "f=[1]");
check_equal(test_dreadnaut_roots([0,1]),     "f=[1]");
check_equal(test_dreadnaut_roots([2,0]),     "f=[2]");
check_equal(test_dreadnaut_roots([0,0]),     "f=[1:2]");
check_equal(test_dreadnaut_roots([1,2]),     "f=[]");
check_equal(test_dreadnaut_roots([0,1,0,0]), "f=[1 3:4]");
check_equal(test_dreadnaut_roots([0,1,0,0,0,0,1,1,0]), "f=[1 3:6 9]");
check_equal(test_dreadnaut_roots([0,1,0,1,0]), "f=[1 3 5]");
check_equal(test_dreadnaut_roots([0,0,1,1,0]), "f=[1:2 5]");


\\-----------------------------------------------------------------------------
progress("test_perms_direct_product()");

{
  my(S2=test_perms_Sn(2));
  check_equal(test_perms_direct_product(S2,S2),
              [Vecsmall([1, 2, 3, 4]), Vecsmall([1, 2, 4, 3]),
               Vecsmall([2, 1, 3, 4]), Vecsmall([2, 1, 4, 3])],
              "test_perms_direct_product() S2 x S2");
}


\\-----------------------------------------------------------------------------
progress("vpar_count_asymmetric()");

check_equal(vpar_count_asymmetric(0), 1);
{
  for(forest=0,1,
     check_equal(vector(10,n,n--; vpar_count_asymmetric(n,forest)),
                 vector(10,n,n--; n!*vpar_count_unlabelled_asymmetric(n,forest)),
                "vpar_count_asymmetric() labelled = n! * unlabelled"));

  check_equal(vector(10,n,n--; vpar_count_asymmetric(n,-1)),
              vector(10,n,n--; vpar_count_asymmetric(n, 0) / if(n,n,1)),
  "vpar_count_asymmetric() labelled root=1 = all trees / n");

  check_equal(vector(10,n,n--; vpar_count_asymmetric(n,-1)),
              vector(10,n,n--; if(n==0,1,(n-1)!)*vpar_count_unlabelled_asymmetric(n)),
  "vpar_count_asymmetric() labelled root=1 = (n-1)! * unlabelled trees");
}

{
  for(forest=-1,1,
     for(n=0, test_limit_trees_search - forest,
        my(count=0);
        forvec(vec=vpar_forvec_from(n,forest), \\ forests
              my(vpar=vpar_from_upascend(vec));
              count += vpar_is_asymmetric(vpar));
        check_equal(count,
                    vpar_count_asymmetric(n,forest),
                    Str("vpar_count_asymmetric()"
                        " n="n" forest="forest));
     ));
}

{
  check_equal(vector(#A228159_samples,n,n--;
                     if(n==0,0, vpar_count_asymmetric(n))),
              A228159_samples,  \\ OFFSET=0
              "A228159 count labelled rooted asymmetric trees");
}


\\-----------------------------------------------------------------------------
progress("vpar_INTERNAL_S3_generators");
\\ vpar_INTERNAL_S4_generators

{
  my(zero_to_one = (vec)->apply(s->for(i=1,#s,s[i]++);s, vec),
     gens = zero_to_one(vpar_INTERNAL_S3_generators),
     perms= test_perms_generators_to_group(gens));
  check_equal(apply(test_permorder,gens), [3,2]);
  check_equal(#perms, 6);
  check_equal(test_perms_is_isomorphic(perms, test_perms_Sn(3)), 1,
              "vpar_INTERNAL_S3_generators vs test_perms_Sn(3)");

  gens = zero_to_one(vpar_INTERNAL_S4_generators);
  perms= test_perms_generators_to_group(gens);
  check_equal(#perms, 24);
  check_equal(apply(test_permorder,gens), [2,2,3,2]);
  check_equal(test_perms_is_isomorphic(perms, test_perms_Sn(4)), 1,
              "vpar_INTERNAL_S3_generators vs test_perms_Sn(4)");
}


\\-----------------------------------------------------------------------------
progress("test_perms_isomorphic_perm()");

{
  check_equal(test_perms_isomorphic_perm([Vecsmall([1,2]), Vecsmall([2,1])],
                                         [Vecsmall([2,1]), Vecsmall([1,2])]),
              Vecsmall([2,1]));
}
{
  my(perms1=test_perms_generators_to_group([Vecsmall([2,1, 3,4]), Vecsmall([1,2, 4,3])]),
     perms2=test_perms_generators_to_group([Vecsmall([2,3,4,1])]));
  check_equal(#perms1, 4);
  check_equal(#perms2, 4);
  check_equal(test_perms_isomorphic_perm(perms1,perms2), 'none);
  check_equal(test_perms_isomorphic_perm(perms1,perms1), Vecsmall([1,2,3,4]));
}

{
  \\ my(perms1=vpar_automorphisms([0, 1, 2, 3, 3, 3]),
  \\    perms2=vpar_automorphisms([0, 1, 2, 2, 2, 1]));
  my(perms1=test_perms_generators_to_group([Vecsmall([2,3,1, 4,5,6])]),
     perms2=test_perms_generators_to_group([Vecsmall([1,2,3, 5,6,4])]));
  check_equal(#perms1, 3, "perms1");
  check_equal(#perms2, 3, "perms2");
  check_equal(test_perms_is_isomorphic(perms1,perms2), 1);

  my(got=test_perms_isomorphic_perm(perms1,perms2));
  check_equal(got, Vecsmall([1,2,3]),
              "test_perms_isomorphic_perm() C3 to C3 in 6");
  \\ check_equal(test_perms_is_map(perms1,perms2,got), 1);
}

{
  for(n=1,6,
     my(perms1=test_vector_ranperm(test_perms_An(n)),
        perms2=test_vector_ranperm(perms1));
     check_equal(#perms1, #perms2, "perms1");
     check_equal(test_perms_is_isomorphic(perms1,perms2), 1);

     my(got=test_perms_isomorphic_perm(perms1,perms2));
     \\ check_equal(test_perms_is_map(perms1,perms2,got), 1);
  );
}


\\------------------------------------------------------------------------------
progress("vpar_num_automorphisms()");

check_equal(vpar_num_automorphisms([]), 1, \
           "vpar_num_automorphisms() empty");
check_equal(vpar_num_automorphisms([0]), 1);
check_equal(vpar_num_automorphisms([0,1]), 1, \
           "vpar_num_automorphisms() path-2");
check_equal(vpar_num_automorphisms([0,0]), 2!, \
            "vpar_num_automorphisms() 2 singletons");
check_equal(vpar_num_automorphisms([0,0,0]), 3!, \
            "vpar_num_automorphisms() 3 singletons");
check_equal(vpar_num_automorphisms([3,3,4,0]), 2, \
            "vpar_num_automorphisms() claw");
{
  for(n=1,10,
     check_equal(vpar_num_automorphisms(vector(n)), n!,
                 "vpar_num_automorphisms() of singletons"));
}

{
  \\        1
  \\   / / / \ \ \
  \\  2 3 4  5 6 7
  \\         | | |
  \\         8 9 10
  my(vpar=[0, 1,1,1, 1,1,1, 5,6,7]);
  vpar_validate(vpar);
  check_equal(vpar_num_automorphisms(vpar), 3!*3!,
                 "vpar_num_automorphisms() of 3 and 3");
}

test_num_automorphisms_by_search(vpar) = \
  sum(i=1,(#vpar)!, vpar_is_automorphism(vpar,numtoperm(#vpar,i)));

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got=vpar_num_automorphisms(vpar));

     check_equal(1 <= got && got <= (#vpar)!, 1,
                 "vpar_num_automorphisms() range");

     \\ even or 1
     check_equal(got==1 || got%2==0, 1,
                 "vpar_num_automorphisms() == 1 or even");

     my(n=#vpar);
     check_equal((n!) % got, 0,
                 "vpar_num_automorphisms() divisor of n!");
     if(vpar_is_tree(vpar) && #vpar,
        check_equal(((n-1)!) % got, 0,
                    "vpar_num_automorphisms() tree, divisor of (n-1)!"));

     if(#vpar <= test_limit_perms_search,
        check_equal(got, test_num_automorphisms_by_search(vpar),
                    "vpar_num_free_automorphisms() vs search");
       );
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_component_cycle_lengths()");

check_equal(vpar_component_cycle_lengths([]), []);
check_equal(vpar_component_cycle_lengths([1]), [1]);
check_equal(vpar_component_cycle_lengths([2,1]), [2]);
check_equal(vpar_component_cycle_lengths([2,1, 4,5,3,4,4,4]), [2,3]);

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     check_equal(vpar_component_cycle_lengths(vpar),
                 vector(vpar_num_components(vpar)),   \\ [0,0,...]
                 "vpar_component_cycle_lengths() of forest"));
}
{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     check_equal(#vpar_component_cycle_lengths(vpar),
                 vpar_num_components(vpar),
       "vpar_component_cycle_lengths() length vpar_num_components()"));
}

{
  for(n=1,5,
     my(total=0,
        vpar=vpar_unlabelled_first(n),
        l=List([]));  \\ Set() of cycle lengths vectors seen
     forvec(vpar=vector(n,i,[0,n]),  \\ all digraphs
           my(vec=vpar_component_cycle_lengths(vpar),
              pos=setsearch(l,vec,1));
           if(pos, listinsert(l,vec,pos)));
     check_equal(#l, fibonacci(2*n+2) - 1));

  check_equal(vector(#A035508_samples,n,n--; fibonacci(2*n+2) - 1),
              A035508_samples,    \\ OFFSET=0
              "A035508 fibonacci(2*n+2) - 1");
  check_equal(vector(#A001906_samples,n,n--; fibonacci(2*n)),
              A001906_samples,    \\ OFFSET=0
              "A001906 fibonacci(2*n)");
}


\\-----------------------------------------------------------------------------
progress("vpar_to_digraph6()");

my(down=0); check_equal(bitnegimply(1,down), 1, "bitnegimply");
my(down=1); check_equal(bitnegimply(1,down), 0);
my(down=2); check_equal(bitnegimply(1,down), 1);

my(down=0); check_equal(down!=0, 0);
my(down=1); check_equal(down!=0, 1);
my(down=2); check_equal(down!=0, 1);

check_equal(vpar_to_digraph6([]), "&?", "vpar_to_digraph6() empty");
check_equal(vpar_to_digraph6([],1), "&?");
check_equal(vpar_to_digraph6([],2), "&?");
{
  my(want="&B?_",
     edges = [ [2,0] ],
     vpar=vector(3));
  for(i=1,#edges, vpar[edges[i][1]+1] = edges[i][2]+1);
  check_equal(2*3+0, 6);
  check_equal(vpar_to_digraph6(vpar), want);
}

{
  \\ formats.txt example edge 0->2 in n=5 vertices
  \\     2  3  4  5
  \\        |
  \\        1
  \\ x = 00100 00000 00000 00000 00000
  \\     |-----||-----||-----||-----||--|   6-bit blocks
  check_equal(vpar_to_digraph6([3,0,0,0,0]),
              Strchr([38, 63+5, 63+8, 63,63,63,63]),
              "vpar_to_digraph6() formats.txt example");
  \\ or downwards 2->0
  \\ x = 00000 00000 10000 00000 00000
  \\     |-----||-----||-----||-----||--|   6-bit blocks
  check_equal(vpar_to_digraph6([3,0,0,0,0],1),
              Strchr([38, 63+5, 63, 63+2, 63,63,63]),
              "vpar_to_digraph6() formats.txt example, down=1");
  \\ or both 0->2 and 2->0
  check_equal(vpar_to_digraph6([3,0,0,0,0],2),
              Strchr([38, 63+5, 63+8, 63+2, 63,63,63]),
              "vpar_to_digraph6() formats.txt example, down=2");
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        str=vpar_to_digraph6(vpar),
        vec=Vecsmall(str));
     check_equal(#vec>=2, 1, "vpar_to_digraph6() length");
     \\ check_equal(vec[#vec], 10, "vpar_to_digraph6() final newline");
  );
}


\\------------------------------------------------------------------------------
progress("vpar_free_orbit_sets()");

check_equal(vpar_free_orbit_sets([]), [], \
           "vpar_free_orbit_sets() empty");
check_equal(vpar_free_orbit_sets([0,1]), [ [1,2] ], \
           "vpar_free_orbit_sets() path-2");

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got = vpar_free_orbit_sets(vpar));
     check_equal(vecsum(apply(length,got)), #vpar,
                 "vpar_free_orbit_sets() total length");
     check_equal(sum(i=1,#got, #got[i]), #vpar);
     check_equal(#got, vpar_num_free_orbits(vpar),
                "vpar_free_orbit_sets() length vs vpar_num_free_orbits()");
     for(i=1,#got, check_equal(setisset(got[i]), 1));

     \\ asymmetric if and only if all in their own orbit
     check_equal(vpar_is_free_asymmetric(vpar),
                 sum(i=1,#got, #got[i]==1) == #vpar,
                 "vpar_free_orbit_sets() singles when asymmetric"));
}


\\------------------------------------------------------------------------------
progress("vpar_free_orbitnum_vector()");

check_equal(vpar_free_orbitnum_vector([]), [], \
           "vpar_free_orbitnum_vector() empty");
check_equal(vpar_free_orbitnum_vector([0]), [1]);
check_equal(vpar_free_orbitnum_vector([0,1]), [1,1]);
check_equal(vpar_free_orbitnum_vector([0,1,1]), [1,2,2]);

test_free_automorphism_orbitnum_vector_by_generators(vpar) =
{
  my(vec=vpar_free_automorphism_generators(vpar));
  if(#vec==0, return([1..#vpar]));
  test_perms_orbitnum_vector(vec);
}

{
  \\     6          10
  \\     |           |
  \\     4           5
  \\     |           |
  \\     1          11
  \\    / \         / \
  \\   2   3       7   8
  \\       |           |
  \\      12           9
  my(vpar=[4, 1, 1, 6, 10, 0, 11, 11, 8, 0, 5, 3]);
  vpar_validate(vpar);
  check_equal(vpar_free_orbit_sets(vpar),
              [[1,11], [2,7], [3,4,5,8], [6,9,10,12]]);

  my(got_orbitnum_vector=vpar_free_orbitnum_vector(vpar));
  check_equal(got_orbitnum_vector,
              [1, 2, 3, 3, 3, 4, 2, 3, 4, 4, 1, 4]);
  \\        n=    2              7

  check_equal(vpar_num_free_automorphisms(vpar), 8);
  my(generators=vpar_free_automorphism_generators(vpar));
  check_equal(generators,
              [Vecsmall([11, 7, 5, 8, 3, 9, 2, 4, 6, 12,  1, 10]),
  \\                   n= 1  2  3  4  5  6  7  8  9  10  11  12
  \\                   (6 9)(4 8)(1 11)(2 7)(3 5)(12 10)    components
               Vecsmall([1, 2, 3, 4, 8, 6, 7, 5, 10,  9, 11, 12])]);
  \\                   n=1  2  3  4  5  6  7  8   9  10  11  12
  \\                   (8 5)(9 10)      arms in one component
  check_equal(apply(test_permorder,generators), [2,2]);

  check_equal(got_orbitnum_vector,
              test_free_automorphism_orbitnum_vector_by_generators(vpar));
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);

     my(got = vpar_free_orbitnum_vector(vpar));
     check_equal(#got, #vpar);
     check_equal(test_vector_is_increasing(test_compnum_to_smallest(got)), 1,
     "vpar_free_orbitnum_vector() is by increasing smallest member");

     check_equal(got,
                 test_free_automorphism_orbitnum_vector_by_generators(vpar),
                 "vpar_free_orbitnum_vector() vs test_free_automorphism_orbitnum_vector_by_generators()");

     check_equal(vpar_is_free_asymmetric(vpar),
                 got == [1..#vpar],
        "vpar_free_orbitnum_vector() on asymmetric is all 1..n"));
}


\\------------------------------------------------------------------------------
progress("vpar_orbit_sets()");

check_equal(vpar_orbit_sets([]), [], \
            "vpar_orbit_sets() empty");

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got = vpar_orbit_sets(vpar));
     check_equal(vecsum(apply(length,got)), #vpar);
     check_equal(sum(i=1,#got, #got[i]), #vpar);
     check_equal(#got, vpar_num_orbits(vpar));
     for(i=1,#got, check_equal(setisset(got[i]), 1));

     \\ asymmetric if and only if all in their own orbit
     check_equal(vpar_is_asymmetric(vpar),
                 sum(i=1,#got, #got[i]==1) == #vpar,
                 "vpar_orbit_sets() singles when asymmetric");
  );
}
{
  \\ example in vpar.gp initial comments
  my(vpar=[0, 1,1,1, 2,2,2, 3,3,3, 4,4]);
  vpar_validate(vpar);
  my(want=[[1], [2,3], [4],
           [5,6,7,8,9,10], [11,12]]);
  check_equal(vpar_orbit_sets(vpar), want);
  check_equal(vpar_free_orbit_sets(vpar), want); \\ same
}


\\------------------------------------------------------------------------------
progress("vpar_orbitnum_vector()");

check_equal(vpar_orbitnum_vector([]), [], \
           "vpar_orbitnum_vector() empty");
check_equal(vpar_orbitnum_vector([0]), [1]);
check_equal(vpar_orbitnum_vector([0,1]), [1,2]);
check_equal(vpar_orbitnum_vector([0,0]), [1,1]);
check_equal(vpar_orbitnum_vector([0,1,1]), [1,2,2]);

test_automorphism_orbitnum_vector_by_generators(vpar) =
{
  my(vec=vpar_automorphism_generators(vpar));
  if(#vec==0, return([1..#vpar]));
  test_perms_orbitnum_vector(vec);
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     my(got = vpar_orbitnum_vector(vpar));
     check_equal(#got, #vpar);
     check_equal(test_vector_is_increasing(test_compnum_to_smallest(got)), 1,
                "vpar_orbitnum_vector() is by increasing smallest member");

     check_equal(got,
                 test_automorphism_orbitnum_vector_by_generators(vpar),
                 "vpar_orbitnum_vector() vs test_automorphism_orbitnum_vector_by_generators()");

     check_equal(vpar_is_asymmetric(vpar),
                 got == [1..#vpar],
                 "vpar_orbitnum_vector() asymmetric is all 1..n"));
}


\\------------------------------------------------------------------------------
progress("vpar_num_orbits()");

check_equal(vpar_num_orbits([]), 0, \
           "vpar_num_orbits() empty");

{
  \\ A000107 = total number of orbits in unlabelled
  \\           of n vertices
  for(n=0,min(10,#A000107_samples),
     my(total=0,
        vpar=vpar_unlabelled_first(n));
     until((vpar=vpar_unlabelled_next(vpar))=='none,   \\ trees
           total += vpar_num_orbits(vpar));
     check_equal(total,
                 A000107_samples[n+1],
                 "A000107 total vpar_num_orbits() of trees"));
}
{
  \\ A000243 = 1,3,9,26, starting n=2.
  \\ Total num orbits over forests is this starting n=1.
  \\
  for(n=1,min(10,#A000243_samples),
     my(total=0,
        vpar=vpar_unlabelled_first(n));
     until((vpar=vpar_unlabelled_next(vpar,1))=='none,   \\ forests
           total += vpar_num_orbits(vpar));
     check_equal(total,
                 A000243_samples[n],
                 "A000243 total vpar_num_orbits() of forests"));
}
{
  \\ A000243 = A000107 - A000081  (respecting offsets)
  my(A000107_samples = A000107_samples[3..#A000107_samples],
     A000081_samples = A000081_samples[3..#A000081_samples],
     len=vecmin([#A000243_samples, #A000107_samples, #A000081_samples]));
  check_equal(A000243_samples[1..len],
              A000107_samples[1..len] - A000081_samples[1..len]);
}


\\-----------------------------------------------------------------------------
progress("test_quaternion_mul()");

check_equal(test_quaternion_mul_table[1,], test_quaternion_mul_elements);

{
  for(s=0,1,
     my(s=if(s,-1,1));
     for(t=0,1,
        my(t=if(s,-1,1));
        check_equal(test_quaternion_mul(s*i,t*i), -1 *s*t);
        check_equal(test_quaternion_mul(s*j,t*j), -1 *s*t);
        check_equal(test_quaternion_mul(s*k,t*k), -1 *s*t);
        check_equal(test_quaternion_mul(s*i,t*j), k *s*t);
        check_equal(test_quaternion_mul(s*j,t*k), i *s*t);
        check_equal(test_quaternion_mul(s*k,t*i), j *s*t)));
}


\\-----------------------------------------------------------------------------
progress("test_perms_Cn()");

{
  my(C4 = test_perms_Cn(4));
  check_equal(#C4[1], 4, "test_perms_Cn(4) elem length");
  check_equal(#C4,    4, "test_perms_Cn(4) num perms");
  check_equal(C4, [Vecsmall([1,2,3,4]), Vecsmall([2,3,4,1]),
                   Vecsmall([3,4,1,2]), Vecsmall([4,1,2,3])],
              "test_perms_Cn(4)");
}


\\-----------------------------------------------------------------------------
progress("test_perms_Dn()");

{
  for(n=1,6,
     my(perms = test_perms_Dn(n));
     check_equal(#perms, if(n<3,n,2*n), "test_perms_Dn() num");
     check_equal(#Set(perms), #perms, "test_perms_Dn() distinct"));
}


\\-----------------------------------------------------------------------------
progress("test_perms_minimum_num_generators()");

{
  my(perms=[Vecsmall([1,2,3])]);
  check_equal(test_perms_is_group(perms), 1);
  check_equal(test_perms_minimum_num_generators(perms), 0);
}
{
  my(perms=test_perms_Cn(2));
  check_equal(test_perms_minimum_num_generators(perms), 1);
}
{
  my(perms=test_perms_Cn(3));
  check_equal(test_perms_minimum_num_generators(perms), 1);
}
{
  my(perms=test_perms_Sn(3));
  check_equal(test_perms_minimum_num_generators(perms), 2);
}
{
  my(perms=test_perms_Cn(4));
  check_equal(test_perms_minimum_num_generators(perms), 1);
}
{
  my(perms=test_perms_An(4));
  check_equal(test_perms_minimum_num_generators(perms), 2);
}
{
  my(perms=test_perms_Sn(4));
  check_equal(test_perms_minimum_num_generators(perms), 2);
}

\\-----------------------------------------------------------------------------
progress("vpar_seq_successive_paths()");

check_equal(vpar_seq_successive_paths([]), Vecsmall([]), \
            "vpar_seq_successive_paths() empty");
check_equal(vpar_seq_successive_paths([0]), Vecsmall([1]));
check_equal(vpar_seq_successive_paths([0,1]), Vecsmall([1,2]));

check_equal(vpar_seq_successive_paths([1]), Vecsmall([1]));
check_equal(vpar_seq_successive_paths([2,1]), Vecsmall([1,2]));
check_equal(vpar_seq_successive_paths([3,0,1]), Vecsmall([1,3,2]));

\\ 1 --> 2  path root 1 up to root 2
check_equal(vpar_seq_successive_paths([2,0]), Vecsmall([1,2]));

{
  \\        2
  \\      / | \
  \\    1   4   3
  \\    |
  \\    5
  \\    |
  \\    6
  my(vpar = [2, 0, 2, 2, 1, 5]);
  check_equal(vpar_seq_successive_paths(vpar), Vecsmall([1,2, 3,4,5,6]));
}

test_seq_successive_paths_by_paths(vpar) =
{
  my(seq=List([]));
  for(v=1,#vpar,
     my(path=vpar_ancestors(vpar,v),
        set=Set(seq));
     for(i=1,#path,
        if(!setsearch(set,path[i]), listput(seq,path[i]))));
  Vecsmall(seq);
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     check_equal(vpar_seq_successive_paths(vpar),
                 test_seq_successive_paths_by_paths(vpar),
                 "vpar_seq_successive_paths() vs by_paths"));
}

test_seq_successive_paths_by_sort_depths(vpar) =
{
  my(min_descendant = vpar_minmax_descendants_vector(vpar,min),
     depths         = vpar_depths_vector(vpar),
     keys = vector(#vpar,v, [min_descendant[v],-depths[v]]));
  vecsort(keys,,1);
}
test_seq_successive_paths_by_sort_heights(vpar) =
{
  my(min_descendant  = vpar_minmax_descendants_vector(vpar,min),
     subtree_heights = vpar_subtree_heights_vector(vpar),
     keys = vector(#vpar,v, [min_descendant[v],subtree_heights[v]]));
  vecsort(keys,,1);
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     check_equal(vpar_seq_successive_paths(vpar),
                 test_seq_successive_paths_by_sort_depths(vpar),
                 "vpar_seq_successive_paths() vs by_sort depths");
     check_equal(vpar_seq_successive_paths(vpar),
                 test_seq_successive_paths_by_sort_heights(vpar),
                 "vpar_seq_successive_paths() vs by_sort heights"));
}

{
  \\ vpar_seq_successive_paths() round trips
  my(forest=0);
  for(n=0, test_limit_trees_search - forest,
     for(down=0,1,
        my(count=0);
        forvec(vec=vpar_forvec_from(n,forest), \\ trees
              my(vpar=vpar_from_upascend(vec));
              my(seq =vpar_seq_successive_paths(vpar,down),
                 vec =vecextract(vpar,seq));

              my(vpar_again = vpar_from_successive_paths(vec,down));
              check_equal(vpar_again,vpar,
                          "vpar_seq_successive_paths() round trip");

              \\ my(seq_again = vpar_INTERNAL_seqfrom_successive_paths(vec,down));
              \\ check_equal(seq_again,seq,
              \\             "vpar_INTERNAL_seqfrom_successive_paths() round trip");

              vpar_again = vecextract(vec,seq^-1);
              check_equal(vpar_again,vpar,
                          "vpar_seq_successive_paths() vpar by vecextract");

           \\ seq =vpar_seq_successive_paths(vpar,1);
           \\ vec =vecextract(vpar,seq);
           \\ seq_again = vpar_INTERNAL_seqfrom_successive_paths(vec);
           \\ check_equal(seq_again,seq,
           \\             "vpar_INTERNAL_seqfrom_successive_paths() down=1 round trip");
        )));
}


\\-----------------------------------------------------------------------------
progress("vpar_seq_upascend()");
\\ vpar_INTERNAL_seq_upwards()

\\ check_equal(vpar_seq_upascend([1]), Vecsmall([1]));

{
  \\         1       max   10
  \\         6       desc  10
  \\        10             10
  \\         4             9
  \\       3   9         8   9
  \\       8   7         8   7
  \\       2   5         2   5
  \\
  my(vpar=[0,8,4,10,7,1,9,3,4,6]);
  \\       1 2 3  4 5 6 7 8 9 10
  vpar_validate(vpar);
  my(seq=vpar_seq_upascend(vpar));
  check_equal(seq, Vecsmall([2,5,7,8,3,9,4,10,6,1]),
              "vpar_seq_upascend()");
  my(vec=vecextract(vpar,seq));
  \\
  check_equal(vec, [8,7,9,3,4,4,10,6,1,0]);
  check_equal(vpar_from_upascend(vec),vpar,
              "vpar_from_upascend()");

  check_equal(vpar_minmax_descendants_vector(vpar,max),
              [10, 2, 8, 9, 5, 10, 7, 8, 9, 10],
              "vpar_minmax_descendants_vector() max");

  \\ at 5 unseen 2,3,8 seen 1,4,6,7,9,10
  seq=vpar_seq_upascend(vpar,1);
  check_equal(seq, Vecsmall([2,8,3,5,7,9,4,10,6,1]),
              "vpar_seq_upascend() ahead=1");
  vec=vecextract(vpar,seq);
  check_equal(vec, [8,3,4,7,9,4,10, 6,1,0]);
  \\                2,8,3,5,7,9, 4,10,6,1
  \\                      *

  my(vpar_again=vpar_from_upascend(vec,1));
  vpar_validate(vpar_again);
  check_equal(vpar_again,vpar, "vpar_from_upascend() ahead=1");
}

{
  \\        10
  \\       / | \
  \\     7   1   9
  \\    / \     / \
  \\   5   4   6   8
  \\   |          / \
  \\   3        2   11
  \\
  \\ upascend        1,2,3, 4,5, 6,7, 11,8,9,10
  \\ upascend ahead  1,2,3, 5, 4,7,6, 11,8,9,10      takes 5 ahead
  \\
  my(vpar=[10,8,5,7,7,9,10,9,10,0,8]);
  my(seq=vpar_seq_upascend(vpar));
  check_equal(seq, Vecsmall([1,2,3, 4,5, 6,7, 11,8,9,10]),
              "vpar_seq_upascend()");
  my(vec=vecextract(vpar,seq));
  \\                                     7        10
  check_equal(vec, [10, 8, 5, 7, 7,  9, 10, 8, 9, 10, 0]);
  check_equal(vpar_from_upascend(vec),vpar,
              "vpar_from_upascend()");

  seq=vpar_seq_upascend(vpar,1);
  check_equal(seq, Vecsmall([1,2,3, 5, 4,7,6, 11,8,9,10]),
              "vpar_seq_upascend() ahead=1");
  vec=vecextract(vpar,seq);
  \\                 1               6   7        10
  check_equal(vec, [10, 8, 5, 7, 7, 10,  9, 8, 9, 10, 0]);
  \\                 1  2  3  5  4   7   6 11  8   0 10

  my(vpar_again=vpar_from_upascend(vec,1));
  vpar_validate(vpar_again);
  check_equal(vpar_again,vpar, "vpar_from_upascend() ahead=1");
}

check_equal(vpar_seq_upascend([]), Vecsmall([]));
check_equal(vpar_seq_upascend([3,3,5,5,0]), Vecsmall([1,2,3,4,5]));
check_equal(vpar_seq_upascend([5,4,4,5,0]), Vecsmall([1,2,3,4,5]));

\\ Algorithm D of
\\
\\     Tim Paulden and David K. Smith, "Developing New Locality Results for
\\     the Prufer Code Using a Remarkable Linear-Time Decoding Algorithm",
\\     The Electronic Journal of Combinatorics, volume 14, 2007, #R55.
\\
\\ which is by Cho et al.
\\
test_from_upascend_algorithm_D(code) =
{
  my(n=#code+2,
     tight=vector(n),
     vpar=vector(n),
     largest=n-1,
     j=2, v);
  tight[n]=1;
  code=Vec(code,n-1);
  code[n-1]=n;
  \\ print("code "code" length "#code);
  while(1,
        \\ print("at j="j" n-j="n-j" tight "Vec(select(t->t,tight,1)));
        if(tight[code[n-j]],
           while(tight[largest], largest--);
           v=largest;
           \\ print(" found largest loose "v);
           ,
           v=code[n-j];
           \\ print(" code[n-j]="code[n-j]" is loose");
          );
        my(prev=code[n-j+1]);
        \\ print(" edge v="v" to prev "prev);
        if(vpar[v], error("oops overwrite at v=",v," vpar[v]=",vpar[v]));
        vpar[v] = prev;
        tight[v]=1;
        j++;
        if(!(j<n), break()));

  vecsum(tight)==n-1 || error();
  while(tight[largest], largest--);  \\ last loose
  v=largest;
  if(vpar[v], error("oops overwrite at v=",v," vpar[v]=",vpar[v]));
  vpar[v] = code[1];
  vpar;
}

{
  \\ Tim Paulden and David K. Smith, "Developing New Locality Results for
  \\ the Prufer Code Using a Remarkable Linear-Time Decoding Algorithm",
  \\ The Electronic Journal of Combinatorics, volume 14, 2007, #R55.
  \\ Section 1.2.1 tree with 15 reckoned root.
  my(vpar=[3,12,15,6,15,3,15,6,6,11,1,1,11,15,0]);

  my(code=[12,6,15,15,6,6,3,11,1,11,1,3,15]);
  check_equal(#code,13);
  check_equal(test_from_upascend_algorithm_D(code),vpar);

  my(vec=concat(code,[15,0]));
  check_equal(vpar_from_upascend(vec),vpar);
}

{
  \\ Evan Thompson, Tim Paulden, and David K. Smith, "The Dandelion Code: A
  \\ New Coding of Spanning Trees for Genetic Algorithms", IEEE Transactions
  \\ ON Evolutionary Computation, volume 11, number 1, February 2007.
  \\ Section II Prufer code example:
  \\  4 --> 2 --> 6 --> 5 --> 1 --> 8 --> 7 root
  \\                          |
  \\                          3
  my(vpar=[8,6,1,2,1,5,0,7]);
  check_equal(#vpar,8);
  vpar_validate(vpar);
  check_equal(vpar_degrees_vector(vpar), [3,2,1,1,2,2,1,2], "vector D");

  my(seq=vpar_seq_upascend(vpar,1));

  my(vec=vecextract(vpar,seq));
  check_equal(vec,[1,2,6,5,1,8, 7,0]);
  check_equal(vpar,vpar_from_upascend([1,2,6,5,1,8, 7,0]),"vpar_from_upascend");
}
{
  \\ Caminiti thesis section 2.2 figure 2.4 example, Third Neville code
  \\        7
  \\        |
  \\        3
  \\      / |  \
  \\    6   8    4
  \\        |   / \
  \\        1  2   5
  \\
  \\       1   3   5   7
  my(vpar=[8,4,7,3,4,3,0,3]);
  vpar_validate(vpar);
  my(seq=vpar_seq_upascend(vpar,1));
  check_equal(seq,Vecsmall([1,8,2,5,4,6,3, 7]));
  my(vec=vecextract(vpar,seq));
  check_equal(vec,[8,3,4,4,3,3,7, 0]);
}
{
  \\        3              3
  \\      / | \          / | \
  \\    6   4   5      4   5   6
  \\   / \                    / \
  \\  1   2                  1   2
  my(vpar=[6,6,0,3,3,3]);
  vpar_validate(vpar);
  check_equal(vpar_INTERNAL_seq_upwards(vpar), Vecsmall([1, 2, 6, 4, 5, 3]));
  check_equal(vpar_seq_upascend(vpar), Vecsmall([1, 2, 4, 5, 6, 3]));
}


{
  \\         3                    7             1
  \\       / | \                / | \         / | \
  \\     4   5   6            2   3   6     6   5   2
  \\     |      / \           |      / \    |      / \
  \\     7     1   2          1     4   5   7     4   3
  my(vpar=[6,6,0,3,3,3,4]);
  vpar_validate(vpar);
  check_equal(vpar_INTERNAL_seq_upwards(vpar), Vecsmall([1, 2, 6, 5, 7, 4, 3]));
  my(upascend=vpar_seq_upascend(vpar));
  check_equal(upascend, Vecsmall([1, 2, 5, 6, 7, 4, 3]));
  check_equal(vector(#vpar,i,vpar[upascend[i]]), [6, 6, 3, 3, 4, 3, 0]);
}

{
  \\        10
  \\       / | \
  \\     3   4   9
  \\    / \     / \
  \\   1   2   5   8
  \\              / \
  \\             6   7
  \\
  \\       1 2  3  4 5 6 7 8  9 10
  my(vpar=[3,3,10,10,9,8,8,9,10, 0]);     \\ Knuth 2.3.4.4
  check_equal(#vpar,10, "Knuth 2.3.4.4");
  vpar_validate(vpar);
  check_equal(vpar_INTERNAL_seq_upwards(vpar), Vecsmall([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]));
  check_equal(vpar_seq_upascend(vpar), Vecsmall([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]));

  \\         1                       1
  \\       / | \                   / | \
  \\     8   7   2               2   7   8
  \\    / \     / \             / \     / \
  \\  10   9   6   3           3   6   9   10
  \\              / \         / \
  \\             5   4       4   5
  my(rvpar=vpar_relabel_reverse(vpar));
  check_equal(rvpar, [0,1,2,3,3,2,1,1,8,8]);
  check_equal(vpar_seq_upascend(rvpar), Vecsmall([4, 5, 3, 6, 2, 7, 9, 10, 8, 1]));
}

\\ upascend equivalent definition in terms of sort, as described in the
\\ addhelp
test_seq_upascend_by_sort(vpar) =
{
  my(max_descendant = vpar_minmax_descendants_vector(vpar,max),
     depths         = vpar_depths_vector(vpar),
     keys = vector(#vpar,v, [max_descendant[v],-depths[v]]));
  vecsort(keys,,1);
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     check_equal(vpar_seq_upascend(vpar),
                 test_seq_upascend_by_sort(vpar),
                 "vpar_seq_upascend() vs test_seq_upascend_by_sort()"));
}

\\ PENDING vpar_minmax_childless_descendants_vector maybe
test_minmax_childless_vector(vpar,func) =
{
  my(seq=vpar_seq_upascend(vpar),
     ret=vector(#vpar));
  for(i=1,#seq,
     my(v=seq[i],p);
     if(ret[v]==0,ret[v]=v);
     if((p=vpar[v]),
        ret[p] = if(ret[p], func(ret[p],ret[v]), ret[v])));
  ret;
}
\\ upascend ahead=1 equivalent definition in terms of sort, as described in
\\ the addhelp
test_seq_upascend_ahead_by_sort(vpar) =
{
  my(max_descendant = test_minmax_childless_vector(vpar,max),
     depths         = vpar_depths_vector(vpar),
     keys = vector(#vpar,v, [max_descendant[v],-depths[v]]));
  vecsort(keys,,1);
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     check_equal(vpar_seq_upascend(vpar,1),
                 test_seq_upascend_ahead_by_sort(vpar),
                 "vpar_seq_upascend() vs test_seq_upascend_ahead_by_sort()"));
}


\\-----------------------------------------------------------------------------
progress("vpar_make_complete_tree()");

check_equal(vpar_make_complete_tree(2,1), [0,1,1]);
{
  for(c=2,3,
     for(h=0,4,
        my(vpar=vpar_make_complete_tree(c,h));
        vpar_validate(vpar);
        check_equal(vpar_height(vpar),h, "complete_tree height");
        check_equal(#vpar, (c^(h+1)-1)/(c-1));
        if(c==2,
           check_equal(#vpar, 2^(h+1)-1));
     ));
}


\\-----------------------------------------------------------------------------
progress("vpar_INTERNAL_automorphisms() 30 = vpar_orbitnum_vector()");

\\ Return 1 if vpar is a forest of paths per test_is_path().
test_is_paths(vpar) =
{
  my(components=vpar_component_trees(vpar));
  for(i=1,#components,
     test_is_path(components[i]) || return(0));
  1;
}
check_equal(test_is_paths([0]), 1, "test_is_paths() singleton");
check_equal(test_is_paths([0,0]), 1, "test_is_paths() 2 singletons");

\\ Structure of return is slightly arbitrary.
\\ {
\\   for(i=1,#test_sample_forests,
\\      my(vpar=test_sample_forests[i],
\\         got=vpar_INTERNAL_automorphisms(vpar,30));
\\      print(vpar" "got);
\\      check_equal(test_is_paths(vpar), 1));
\\ }



\\-----------------------------------------------------------------------------
progress("vpar_subtree()");

check_equal(vpar_subtree([0,1,2,3,4],1), [0,1,2,3,4]);
check_equal(vpar_subtree([0,1,2,3,4],3), [0,1,2]);

check_equal(vpar_subtree([2,1,1],1), [2,1,1]);
check_equal(vpar_subtree([2,1,1],2), [2,1,1]);
check_equal(vpar_subtree([2,1,1],3), [0]);


test_subtree_by_ancestor(vpar,v) = \
  vpar_keep_flags(vpar,vector(#vpar,i, vpar_is_ancestor(vpar,i,v)));
check_equal(test_subtree_by_ancestor([0],1), [0], \
            "test_subtree_by_ancestor() singleton");

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     #vpar || next();
     my(v=vpar_random_vertex(vpar));
     \\ print(vpar" v="v);
     check_equal(vpar_subtree(vpar,v),
                 test_subtree_by_ancestor(vpar,v),
                 "vpar_subtree() vs test_subtree_by_ancestor()"));
}


\\------------------------------------------------------------------------------
progress("test_perm_is_involution()");

{
  for(n=0,10,
     my(perm=vpar_INTERNAL_random_perm(n),
        got=test_perm_is_involution(perm));
     check_equal(got, perm^2 == test_identity_perm(#perm),
                 "test_perm_is_involution() vs perm^2");
     check_equal(got, test_permorder(perm)<=2,
                 "test_perm_is_involution() vs test_permorder()"));
}

test_perm_involution_count(n) =
{
  \\ formula per Muir given at Mathworld,  A000085
  1 + sum(k=0,floor((n-1)/2), prod(i=0,k, binomial(n-2*i, 2)) / (k+1)!);
}
{
  for(n=1,7,
     my(got=sum(i=1,n!, test_perm_is_involution(numtoperm(n,i))));
     check_equal(got, test_perm_involution_count(n),
       "test_perm_is_involution() count vs test_perm_involution_count()");

     \\ formula per Skiena given at Mathworld
     check_equal(got, n! * sum(k=0,floor(n/2), 1/(2^k*k!*(n-2*k)!)),
                "test_perm_is_involution() count vs second formula"));
}
{
  \\ recurrence Muir and Skiena given at Mathworld
  for(n=2,15,
     check_equal(test_perm_involution_count(n),
       test_perm_involution_count(n-1) + (n-1)*test_perm_involution_count(n-2),
                 "test_perm_involution_count() recurrence"));

  check_equal(vector(#A000085_samples,n,n--; test_perm_involution_count(n)),
              A000085_samples,     \\ OFFSET=0
              "A000085 test_perm_involution_count()");

  check_equal(vector(20,n,n--; test_perm_involution_count(n)),
              vector(20,n,n--; vpar_siblingpos_vectors_count(n,1)),
              "test_perm_involution_count() vs vpar_siblingpos_vectors_count()");
}

\\-----------------------------------------------------------------------------
progress("vpar_is_equaldepths()");

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     check_equal(vpar_is_equaldepths(vpar),
                 vpar==vpar_depths_vector(vpar),
                 "vpar_is_equaldepths() vs vpar==vpar_depths_vector(vpar)"));
}


\\-----------------------------------------------------------------------------
progress("test_triangular()");

check_equal(test_triangular(0), 0, "test_triangular() 0");
check_equal(test_triangular(1), 1);
check_equal(test_triangular(2), 3);
check_equal(test_triangular(3), 6);
{
  check_equal(vector(#A000217_samples,n,n--; test_triangular(n)),
              A000217_samples,   \\ OFFSET=0
              "A000217 test_triangular()");

  my(want=vector(10,t,t--; test_triangular(t)),
     got=select(test_is_triangular,[0..vecmax(want)]));
  check_equal(got,want, "test_is_triangular() vs test_triangular()");
}


\\------------------------------------------------------------------------------
progress("vpar_cyclic_vertices()");

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        got=vpar_cyclic_vertices(vpar));
     check_equal(got,
                 test_select_vertices(vpar,vpar_is_cyclic_vertex),
                 "vpar_cyclic_vertices()() vs select vpar_is_cyclic_vertex()");
     check_equal(#got,
                 vpar_num_cyclic_vertices(vpar),
                 "vpar_cyclic_vertices()() length vs vpar_num_cyclic_vertices()"));
}


\\------------------------------------------------------------------------------
progress("vpar_num_cyclic_vertices()");

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        got=vpar_siblings_sets(vpar));
     check_equal(vpar_num_cyclic_vertices(vpar),
                 sum(v=1,#vpar, vpar_is_cyclic_vertex(vpar,v)),
                 "vpar_num_cyclic_vertices()() vs count vpar_is_cyclic_vertex()"));
}


\\-----------------------------------------------------------------------------
progress("vpar_INTERNAL_gnot()");

check_equal(vpar_INTERNAL_gnot(0), 1, "vpar_INTERNAL_gnot() 0");
check_equal(vpar_INTERNAL_gnot(0.0), 1, "vpar_INTERNAL_gnot() 0.0");
check_equal(vpar_INTERNAL_gnot(-1), 0);
check_equal(vpar_INTERNAL_gnot('x), 0);
check_equal(vpar_INTERNAL_gnot([123]), 0);


\\-----------------------------------------------------------------------------
progress("vpar_is_cyclic_vertex()");

check_equal(vpar_is_cyclic_vertex([0],1), 0, \
            "vpar_is_cyclic_vertex() singleton");
check_equal(vpar_is_cyclic_vertex([0,2],1), 0);
check_equal(vpar_is_cyclic_vertex([0,2],2), 1);
check_equal(vpar_is_cyclic_vertex([3,1,2],1), 1);
check_equal(vpar_is_cyclic_vertex([2,3,4,2,1], 1), 0);

test_is_cyclic_vertex_by_seen(vpar,v) =
{
  my(seen=vectorsmall(#vpar),
     p=v);
  while(p=vpar[p],
        if(p==v, return(1));   \\ cycle back to v
        if(seen[p], break);    \\ in a cycle without v
        seen[p]=1);
  0;
}
{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     for(v=1,#vpar,
        check_equal(vpar_is_cyclic_vertex(vpar,v),
                    test_is_cyclic_vertex_by_seen(vpar,v))));
}


\\-----------------------------------------------------------------------------
progress("test_perm_powers()");

check_equal(test_perm_powers(Vecsmall([])), [ Vecsmall([]) ]);
check_equal(test_perm_powers(Vecsmall([1])), [ Vecsmall([1]) ]);
check_equal(test_perm_powers(Vecsmall([2,1])), \
            [ Vecsmall([1,2]), Vecsmall([2,1]) ]);
check_equal(test_perm_powers(Vecsmall([3,1,2])), \
            [ Vecsmall([1,2,3]), Vecsmall([3,1,2]), Vecsmall([2,3,1]) ]);

{
  for(n=0,10,
     my(perm=vpar_INTERNAL_random_perm(n),
        got=test_perm_powers(perm));
     check_equal(#got, test_permorder(perm)));
}


\\-----------------------------------------------------------------------------
progress("test_perms_orbitnum_vector()");

check_equal(test_perms_orbitnum_vector([ Vecsmall([1,2]) ]), [1,2]);
check_equal(test_perms_orbitnum_vector([ Vecsmall([2,1]) ]), [1,1]);
{
  my(perms=[ Vecsmall([2,1,3]), Vecsmall([2,3,1]) ]);
  check_equal(apply(test_is_perm,perms), [1,1]);
  check_equal(apply(test_permorder,perms), [2,3]);
  check_equal(#test_perms_generators_to_group(perms), 3!);
  check_equal(test_perms_orbitnum_vector(perms), [1,1,1]);
}


\\-----------------------------------------------------------------------------
progress("vpar_count_subtree_heights_increasing()");

check_equal(vpar_count_subtree_heights_increasing(0),1);
check_equal(vpar_count_subtree_heights_increasing(1),1);
check_equal(vpar_count_subtree_heights_increasing(2),1);   \\ path-2 only
check_equal(vpar_count_subtree_heights_increasing(3),2);
check_equal(vpar_count_subtree_heights_increasing(3,1),1+3+1);

test_W(n,m,e) =
{
  if(n<0||m<0||e<0, error());
  if(n==0,(m==0),  if(m,m*test_W(n-1,m-1,e+1)) + e*test_W(n-1,m,e));
}

test_C(n,l) =
{
  \\ print("test_C "n" "l);
  if(n<0 || l<0 || l>n,error());
  if(l==0, return(n==0));
  sum(f=0,n-l,
        my(c = test_C(n-l,f),
           w = test_W(l,f,n-l-f+1));
        \\ print(" f="f" recurse C("n-l","f") = "c","
        \\       " term W("l","f","n-l-f+1") = "w);
        c*w );
}
test_CC(n) = sum(l=0,n, test_C(n,l));
check_equal(test_C(0,0), 1, "test_C 0,0");
check_equal(test_C(1,1), 1, "test_C 1,1");

check_equal(test_C(2,2), 1, "test_C 2,2");
check_equal(test_C(2,1), 1);
check_equal(test_CC(2), 2);  \\ path-2 increasing, or 2 isolated

check_equal(test_C(3,1), 1, "test_C 1,1");

check_equal(test_C(3,3), 1, "test_C 3,3");
\\ 1->3 2    1 2->3   1->3<-2
check_equal(test_C(3,2), 3, "test_C 3,2");  \\ which has 3 above
check_equal(test_CC(3), 5);

check_equal(vector(12,n,n--; vpar_count_subtree_heights_increasing(n,1)), \
            vector(12,n,n--; test_CC(n)), "test_CC");

{
  \\ increasing
  \\ not in OEIS: 2, 5, 16, 61, 268, 1329, 7342, 44703, 297314, 2143479
  my(want=[1, 1, 2, 5, 16, 61, 268, 1329, 7342, 44703, 297314, 2143479,
           16643124, 138413449, 1227204644]);
  check_equal(vector(#want,n,vpar_count_subtree_heights_increasing(n)),
              want,
              "vpar_count_subtree_heights_increasing() samples");

  \\ shown in per addhelp
  want=[1,1,1,2,5,16,61,268];
  check_equal(vector(#want,n,n--; vpar_count_subtree_heights_increasing(n)),
              want);
}
check_equal(vector(8,n,n--; vpar_count_subtree_heights_increasing(n)), \
            [1,1,1,2,5,16,61,268], \
            "vpar_count_subtree_heights_increasing() samples in addhelp");

{
test_W(n,f,e) = if(n==0,(f==0),
                   f*test_W(n-1,f-1,e+1) + e*test_W(n-1,f,e));
}
{
test_C(n,l) = if(l==0, (n==0),
              sum(f=0,n-l, test_C(n-l,f) * test_W(l,f,n-l-f+1)));
}
{
test_count_subtree_heights_increasing(n) = sum(l=0,n, test_C(n,l));
}
\\ check_equal(test_count_subtree_heights_increasing(-1),1);
check_equal(test_count_subtree_heights_increasing(0),1);
check_equal(test_count_subtree_heights_increasing(1),1);
{
  check_equal(vector(12,n,n--; vpar_count_subtree_heights_increasing(n,1)),
              vector(12,n,n--; test_count_subtree_heights_increasing(n)),
              "vpar_count_subtree_heights_increasing() vs addhelp formulas, forest");
}

{
test_C(n,l) = if(l==0,(n==0),  l==1,1,
              sum(f=0,n-l, test_C(n-l,f) * test_W(l,f,n-l-f)));
}
check_equal(test_C(1,1),1);
{
  check_equal(vector(12,n,n--; vpar_count_subtree_heights_increasing(n)),
              vector(12,n,n--; test_count_subtree_heights_increasing(n)),
              "vpar_count_subtree_heights_increasing() vs addhelp formulas, tree");

  my(want=[1,1,1,2,5,16,61,268]);
  check_equal(vector(#want,n,n--; test_count_subtree_heights_increasing(n)),
              want);

  check_equal(matrix(10,10,n,f,n--;f--; test_W(n,f,0)),
              matrix(10,10,n,f,n--;f--; f!*stirling(n,f,2)),
              "vpar_count_subtree_heights_increasing() W vs stirling");
}

check_equal(test_W(5,0,0), 0);

test_count_subtree_heights_increasing_of_partition(p,forest) =
{
  \\ print(p);
  prod(i=2,#p,
       my(w=test_W(p[i], p[i-1],
                   forest + if(i>2,vecsum(Vec(p,i-2)))));
     \\ print("W "p[i]" "p[i-1]" "forest+if(i>2,vecsum(Vec(p,i-2)))" = "w);
       w);
}

check_equal(Set(partitions(3)), \
            [Vecsmall([3]),Vecsmall([1,2]),Vecsmall([1,1,1])], \
            "partitions() 3");

check_equal(Set(partitions(4)), \
            [Vecsmall([4]),Vecsmall([1,3]),Vecsmall([2,2]),Vecsmall([1,1,2]), \
             Vecsmall([1,1,1,1])], \
            "partitions() 4");
check_equal(test_count_subtree_heights_increasing_of_partition([4],1), 1);
check_equal(test_count_subtree_heights_increasing_of_partition([1,3],1),7, \
            "test_count_subtree_heights_increasing_of_partition() [1,3]");
\\   1 2->3->4   2 1->3->4   1->3->4   1->3->4   2->3->4
\\                             /            /         /
\\                            2            2         1
check_equal(test_count_subtree_heights_increasing_of_partition([1,1,2],1),5);
check_equal(test_count_subtree_heights_increasing_of_partition([1,1,1,1],1),1);

test_count_subtree_heights_increasing_by_partition(n,forest) =
{
  if(forest,
  \\ print(partitions(n));
     vecsum(apply(p->prod(i=2,#p, test_W(p[i],p[i-1],
                                         1+if(i>2,vecsum(Vec(p,i-2))))),
                 partitions(n)));
     ,
     vecsum(apply(p->
                 (#p==0 || p[1]==1) *
                 prod(i=2,#p, test_W(p[i],p[i-1],
                                     if(i>2,vecsum(Vec(p,i-2))))),
                 partitions(n)));
    );
}
check_equal(vpar_count_subtree_heights_increasing(3,1), \
            test_count_subtree_heights_increasing_by_partition(3,1), \
            "vpar_count_subtree_heights_increasing() n=3 vs partition");
check_equal(vpar_count_subtree_heights_increasing(4,1), \
            test_count_subtree_heights_increasing_by_partition(4,1));

test_rStirling2(n,m,r) =
{
  if(n<r, 0,
     n==r, (m==r),
     m*test_rStirling2(n-1,m,r) + test_rStirling2(n-1,m-1,r));
}
{
  \\ table 1 in Broder
  my(r=1);
  check_equal(matrix(6,6,n,m, test_rStirling2(n,m,r)),
              [1, 0, 0, 0, 0,0;
               1, 1, 0, 0, 0,0;
               1, 3, 1, 0, 0,0;
               1, 7, 6, 1, 0,0;
               1,15,25,10, 1,0;
               1,31,90,65,15,1]);
}
{
  \\ table 2 in Broder
  my(r=2);
  check_equal(matrix(6,6,n,m,n++;m++; test_rStirling2(n,m,r)),
              [ 1,  0,  0,  0, 0,0;
                2,  1,  0,  0, 0,0;
                4,  5,  1,  0, 0,0;
                8, 19,  9,  1, 0,0;
               16, 65, 55, 14, 1,0;
               32,211,285,125,20,1]);
}
{
  \\ table 3 in Broder
  my(r=3);
  check_equal(matrix(6,6,n,m,n+=2;m+=2; test_rStirling2(n,m,r)),
              [  1,  0,  0,  0, 0,0;
                 3,  1,  0,  0, 0,0;
                 9,  7,  1,  0, 0,0;
                27, 37, 12,  1, 0,0;
                81,175, 97, 18, 1,0;
               243,781,660,205,25,1]);
}
\\ default(output,2);
\\ my(r=2); matrix(5,10,n,f,n--;f--; test_W(n,f,r))
\\ my(r=2); matrix(5,10,n,f,n--;f--; f!*test_rStirling2(n+r,f+r,r))
\\ [l,f,n-l-f]
/*
  GP-Test  ['l + 'n-'l-'f, 'f + 'n-'l-'f, 'n-'l-'f] == \
  GP-Test  ['n - 'f, 'n - 'l, 'n - 'l - 'f]
*/
check_equal(matrix(8,8,n,f,n--;f--;vector(8,e,e--; test_W(n,f,e))), \
            matrix(8,8,n,f,n--;f--;vector(8,e,e--; f!*test_rStirling2(n+e,f+e,e))));
{
  for(forest=0,1,
     check_equal(vector(10,n,n--; vpar_count_subtree_heights_increasing(n,forest)),
                 vector(10,n,n--; test_count_subtree_heights_increasing_by_partition(n,forest)),
                 Str("vpar_count_subtree_heights_increasing() vs partition"
                     " forest="forest));
  );
}

test_count_subtree_heights_increasing_by_search(n,forest) =
{
  my(count=0);
  forvec(vec=vpar_forvec_from(n,forest),
        my(vpar=vpar_from_upascend(vec),
           subtree_heights=vpar_subtree_heights_vector(vpar));
        count += test_vector_is_increasing(subtree_heights));
  count;
}
{
  for(forest=0,1,
     for(n=0, test_limit_trees_search - forest,
        check_equal(vpar_count_subtree_heights_increasing(n,forest),
                    test_count_subtree_heights_increasing_by_search(n,forest),
                    Str("vpar_count_subtree_heights_increasing() "
                        "vs test_count_subtree_heights_increasing_by_search()"
                        " n=",n," forest="forest))));
}


\\-----------------------------------------------------------------------------
progress("vpar_minmax_sibling_vector()");

check_equal(vpar_minmax_sibling_vector([],min), [], \
            "vpar_minmax_sibling_vector() empty");
check_equal(vpar_minmax_sibling_vector([],max), []);

{
  \\ Example forest where siblings are consecutive vertex numbers so that
  \\ vpar_minmax_sibling_vector(vpar,min) is ascending.
  \\
  \\           3       4        vpar = [3,3, 0, 0,4, 1,1,1, 5,2]
  \\          / \      |
  \\         1   2     5        vpar_minmax_sibling_vector(vpar,min)
  \\        /|\   \    |           == [1,1,3,3,5,6,6,6,9,10]
  \\       6 7 8  10   9
  my(vpar=[3,3, 0, 0,4, 1,1,1, 5,2]);
  check_equal(vpar_minmax_sibling_vector(vpar,min),
              [1,1,3,3,5,6,6,6,9,10]);
}

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     for(m=0,1,
        my(func=if(m,max,min),
           vec=vpar_minmax_sibling_vector(vpar,func),
           nsib=vpar_next_sibling_vector(vpar,!m));
        \\ print(vpar" "vec" nsib "nsib);
        for(v=1,#vpar,
           my(mv=vec[v]);
           check_equal(vpar[v],vpar[mv],
                      "vpar_minmax_sibling_vector() is silbling");
           check_equal(nsib[mv], 0,
                      "vpar_minmax_sibling_vector() sibling is min/max"))));
}


\\-----------------------------------------------------------------------------
progress("vpar_indsets()");

{
  my(vpar=[0,1]);
  check_equal(vpar_indsets(vpar),[[], [2], [1]]);
}

test_indsets_by_flags(vpar) =
{
  my(l=List([]));
  forvec(flags=vector(#vpar,i,[0,1]),
        if(vpar_is_indset_flags(vpar,flags),
           listput(l,test_flags_to_Set(flags))));
  Vec(l);
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        indsets_count=vpar_indsets_count(vpar));
     if(indsets_count > test_limit_sets, next());

     my(got=vpar_indsets(vpar));
     check_equal(#got, indsets_count);
     for(i=1,#got,
        check_equal(setisset(got[i]), 1);
        check_equal(vpar_is_indset(vpar,got[i]), 1));

     check_equal(vecsort(got), vecsort(test_indsets_by_flags(vpar)));

     check_equal(Pol(0,'y) + sum(i=1,#got, 'y^#got[i]),
                 vpar_indpoly(vpar,'y));
  );
}

\\ Return a Vecsmall of the positions i in vector v where v[i]==target.
test_vector_positions_equal(v,target) = select(value->value==target, v, 1);


\\-----------------------------------------------------------------------------
progress("bitnegimply()");

{
  for(a=1,10,
     for(b=1,10,
        check_equal(bitnegimply(a,b), bitand(a,bitneg(b)))));
}


\\-----------------------------------------------------------------------------
progress("test_fromdigits()");

check_equal(test_fromdigits([1,2,3]), 123);
check_equal(test_fromdigits([1,1,0,1],2), 13);


\\-----------------------------------------------------------------------------
progress("vpar_component_minmax()");

check_equal(vpar_component_minmax([],min), [], \
            "vpar_component_minmax() min, empty");
check_equal(vpar_component_minmax([],max), []);

check_equal(vpar_component_minmax([0,1,0,3],min), [1,3]);
check_equal(vpar_component_minmax([0,1,0,3],max), [2,4]);

test_component_minmax_by_vertices(vpar,func) =
{
  func=if(func==min,vecmin,vecmax);
  vecsort(apply(func,vpar_component_vertices(vpar)));
}
test_component_minmax_by_roots(vpar,func) =
{
  my(roots=vpar_roots(vpar),
     descendants=vpar_minmax_descendants_vector(vpar,func));
  vecsort(apply(r->descendants[r], roots));
}
test_component_minmax_by_old_forest(vpar,func) =
{
  if(vpar_is_cyclic(vpar), error("for acyclics only"));

  \\ For min, take vertices low to high.  At each follow it up to the root
  \\ or to an already-seen vertex.  If reach an unseen root then i is the
  \\ minimum vertex in it.
  \\ Conversely for max, take vertices high to low and unseen root is
  \\ maximum vertex in that component.
  \\
  \\ For max, when vpar=[] empty have func(1,#vpar) = 0 and #vpar-start+1 =
  \\ 1 so forstep goes i=0,1 whereas want no loops when empty.  Could fiddle
  \\ with the start calculation but just as easy to skip entirely #vpar==0.

  my(ret=vector(vpar_num_roots(vpar)));
  if(#vpar,
     my(seen =vectorsmall(#vpar),
        start=func(1,#vpar),     \\ 1 for min, #vpar for max
        step =-func(1,-1),       \\ 1 for min, -1 for max
        upto =func(0,#ret+1));   \\ position in ret[] for pre-inc by step
     forstep(i=start, #vpar-start+1, step,
            my(v=i);
            while(!seen[v],
                  seen[v]=1;
                  if(!(v=vpar[v]), ret[upto+=step]=i; break()))));
  ret;
}
{
  my(funcs=[min,max]);
  for(f=1,#funcs,
     my(func=funcs[f]);
     for(i=1,#test_sample_forests_and_cyclics,
        my(vpar=test_sample_forests_and_cyclics[i]);
        check_equal(vpar_component_minmax(vpar,func),
                    test_component_minmax_by_vertices(vpar,func),
                    "vpar_component_minmax() vs test_component_minmax_by_vertices()")));
}
{
  my(funcs=[min,max]);
  for(f=1,#funcs,
     my(func=funcs[f]);
     for(i=1,#test_sample_forests,
        my(vpar=test_sample_forests[i]);
        check_equal(vpar_component_minmax(vpar,func),
                    test_component_minmax_by_roots(vpar,func),
                    "vpar_component_minmax() vs test_component_minmax_by_roots()");
        check_equal(vpar_component_minmax(vpar,func),
                    test_component_minmax_by_old_forest(vpar,func),
                    "vpar_component_minmax() vs test_component_minmax_by_roots()");
     ));
}


\\-----------------------------------------------------------------------------
progress("vpar_component_random_vertices()");

test_vpar_component_random_vertices = \
  eval(test_random_to_test_random(vpar_component_random_vertices));
{
  my(vpar=[0,1,0,3],
     want=[ [1,3], [1,4], [2,3], [2,4] ],
     upto=1);
  forvec(v=[[0,1], [0,1]],
        test_random_list      =List(v);
        test_random_limit_list=List([2,2]);
        my(got=test_vpar_component_random_vertices(vpar));
        check_equal(#test_random_list, 0, "test_random_list consumed");
        check_equal(#test_random_limit_list, 0, "test_random_limit_list empty");
        check_equal(got, want[upto]);
        upto++);
}

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        num_components=vpar_num_components(vpar),
        component_sizes=vpar_component_sizes(vpar),
        component_lists=vpar_component_vertices(vpar),
        want_indexes=vector(num_components,c,random(component_sizes[c])),
        want_vertices=vector(#component_lists, c,
                             component_lists[c][want_indexes[c]+1]));

     test_random_limit_list=List(component_sizes);
     test_random_list=List(want_indexes);

     my(vertices=test_vpar_component_random_vertices(vpar));
     check_equal(#test_random_list, 0, "test_random_list consumed");
     check_equal(#test_random_limit_list, 0, "test_random_limit_list consumed");

     check_equal(vertices,want_vertices,
                 "vpar_component_random_vertices() vertices");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_component_trees()");

check_equal(vpar_num_roots([]), 0);
check_equal(vpar_component_sizes([]), []);
check_equal(vpar_component_trees([]), [ ], "vpar_component_trees() empty");
check_equal(vpar_component_trees([0]), [ [0] ]);
check_equal(vpar_component_trees([0,0]), [ [0], [0] ]);
check_equal(vpar_component_trees([0,0,2]), [ [0], [0,1] ]);
check_equal(vpar_component_trees([1]), [ [1] ]);
check_equal(vpar_component_trees([1,2]), [ [1], [1] ]);

test_component_trees_by_vertices(vpar) = \
  apply(set->vpar_keep_vertices(vpar,set), vpar_component_vertices(vpar));

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        got=vpar_component_trees(vpar));
     check_equal(got, test_component_trees_by_vertices(vpar)));
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got=vpar_component_trees(vpar));

     \\ Re-concatenating is a relabelling due to making contiguous vertex
     \\ ranges.  Within the components sibling order is preserved, but root
     \\ order is not (components going by min vertex not by root number).
     my(conc=test_concat_forest_vector(got));

     check_equal(vpar_is_isomorphic(conc,vpar), 1,
                 "vpar_component_trees() re-concatenated is isomorphic"));
}


\\-----------------------------------------------------------------------------
progress("test_num_compositions()");

{
  check_equal(vector(#A011782_samples,n,n--; test_num_compositions(n)),
              A011782_samples,    \\ OFFSET=0
              "A011782 num compositions");
}


\\-----------------------------------------------------------------------------
progress("vpar_component_sizes()");

check_equal(vpar_component_sizes([]), []);
check_equal(vpar_component_sizes([0]), [1]);
check_equal(vpar_component_sizes([0,0,0]), [1,1,1]);

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i]);
     check_equal(vpar_component_sizes(vpar),
                 apply(length, vpar_component_vertices(vpar))));
}

{
  \\ number of different component sizes vectors occurring, as described in
  \\ vpar_component_sizes() addhelp
  my(forest=1);
  for(n=0, test_limit_trees_search - forest,
     my(l=List([]));
     forvec(vec=vpar_forvec_from(n,forest), \\ forests
           my(vpar=vpar_from_upascend(vec),
              component_sizes=vpar_component_sizes(vpar),
              pos = setsearch(l,component_sizes,1));
           if(pos, listinsert(l,component_sizes,pos)));
     check_equal(#l, 2^max(0,n-1)));
}


\\-----------------------------------------------------------------------------
progress("vpar_num_components()");

check_equal(vpar_num_components([]), 0, "vpar_num_components() empty");
check_equal(vpar_num_components([0]), 1, "vpar_num_components() singleton");
check_equal(vpar_num_components([1]), 1, "vpar_num_components() singleton self-loop");
check_equal(vpar_num_components([0,1]), 1);
check_equal(vpar_num_components([0,0]), 2);
check_equal(vpar_num_components([2,1]), 1);
check_equal(vpar_num_components([1,2]), 2);

test_num_components_by_decrease(vpar) =
{
  \\ ret starts as if every vertex is a separate component, then decreases
  \\ if a vertex is in a component already seen.
  \\ Search upwards from each vertex i looking for a vertex already seen,
  \\ or otherwise set seen[v]=i to show seen.
  \\ If reach seen[v]!=0 then seen[v]==i is a cycle not already seen so ret
  \\ unchanged, or seen[v]!=i is a component already seen so ret decrease.
  \\ If reach vpar[v]==0 without anything seen then this is a root not
  \\ already seen so ret unchanged.
  \\ While working upwards each vertex is in the same component, not
  \\ separate, but ret is not adjusted.  That happens later when i = those
  \\ other vertices and they are found to have their seen[v] != i.

  my(seen=vectorsmall(#vpar),
     ret=#vpar);
  for(i=1,#vpar,
     my(v=i);
     until(!(v=vpar[v]),
           if(seen[v], ret-=(seen[v]!=i); break());
           seen[v]=i));
  ret;
}

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        got=vpar_num_components(vpar));
     check_equal(got,  vpar_num_roots(vpar) + vpar_num_cycles(vpar));
     check_equal(got,  test_num_components_by_decrease(vpar));
     if(vpar_is_cyclic(vpar),
        check_equal(got >= vpar_num_roots(vpar), 1);
        ,
        check_equal(got, vpar_num_roots(vpar)));
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_compnum_vector()");

check_equal(vpar_compnum_vector([]), [], "vpar_compnum_vector() empty");
check_equal(vpar_compnum_vector([0]), [1], "vpar_compnum_vector() singleton");
check_equal(vpar_compnum_vector([1]), [1], "vpar_compnum_vector() cyclic");
check_equal(vpar_compnum_vector([0,0]), [1,2]);
check_equal(vpar_compnum_vector([0,1]), [1,1]);
check_equal(vpar_compnum_vector([2,1]), [1,1], "vpar_compnum_vector() cyclic2");
check_equal(vpar_compnum_vector(Vecsmall([0,1])), [1,1]);

test_compnum_vector_by_roots(vpar) =
{
  if(vpar_is_cyclic(vpar), error("for acyclics only"));

  my(roots=vpar_root_of_vertex_vector(vpar),
     root_to_component=vector(#vpar),
     upto=0);
  for(v=1,#vpar,
     my(r=roots[v]);
     if(!root_to_component[r], root_to_component[r]=upto++));
  vector(#vpar,v, root_to_component[roots[v]]);
}

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        got = vpar_compnum_vector(vpar));
     check_equal(test_vector_is_increasing(test_compnum_to_smallest(got)), 1,
                "vpar_compnum_vector() is by increasing smallest member"));
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got = vpar_compnum_vector(vpar));
     check_equal(got, test_compnum_vector_by_roots(vpar),
                 "vpar_compnum_vector() vs test_compnum_vector_by_roots()"));
}

{
  \\ count of how many vpar_compnum_vector() occur, per the addhelp
  my(forest=1);
  for(n=0, test_limit_trees_search - forest,
     my(l=List([])); \\ distinct vecs seen
     forvec(vec=vpar_forvec_from(n,forest), \\ forests
           my(vpar=vpar_from_upascend(vec),
              compnum=vpar_compnum_vector(vpar));
           my(pos=setsearch(l,compnum,1));
           if(pos, listinsert(l,compnum,pos)));
     \\ print(#l);
     check_equal(#l, vpar_INTERNAL_Bell_number(n),
                 "vpar_compnum_vector() count occurring = Bell");
  );
}


\\-----------------------------------------------------------------------------
progress("test_perms_is_normal_subgroup()");

{
  for(n=1,5,
     my(T=[ test_identity_perm(n) ], \\ trivial group
        C=test_perms_Cn(n),
        A=test_perms_An(n),
        S=test_perms_Sn(n));
     check_equal(test_perms_is_normal_subgroup(C,T), 1);
     check_equal(test_perms_is_normal_subgroup(A,T), 1);
     check_equal(test_perms_is_normal_subgroup(S,T), 1);
     \\ check_equal(test_perms_is_normal_subgroup(A,C), 1,
     \\             Str("n="n" C in A"));
     \\ check_equal(test_perms_is_normal_subgroup(S,C), 1,
     \\             Str("n="n" C in S"));
     check_equal(test_perms_is_normal_subgroup(S,A), 1,
                 Str("n="n" A in S"));
  );
}


\\-----------------------------------------------------------------------------
progress("test_perms_is_nilpotent()");

check_equal(test_perms_is_nilpotent([ Vecsmall([]) ]), 1, \
            "test_perms_is_nilpotent() trivial empty");
check_equal(test_perms_is_nilpotent([ Vecsmall([1,2]) ]), 1);
check_equal(test_perms_is_nilpotent([ Vecsmall([1,2]), Vecsmall([2,1]) ]), 1);
{
  my(perms=test_perms_Sn(3),
     reduced=test_perms_commutator_group(perms,perms));
  check_equal(reduced,
              [ Vecsmall([1, 2, 3]), Vecsmall([2, 3, 1]), Vecsmall([3, 1, 2]) ],
              "test_perms_commutator_group(S3,S3) = C3");
  reduced=test_perms_commutator_group(perms,reduced);
  check_equal(reduced,
              [ Vecsmall([1, 2, 3]), Vecsmall([2, 3, 1]), Vecsmall([3, 1, 2]) ],
              "test_perms_commutator_group(S3,C3) = C3 again");
  check_equal(test_perms_is_nilpotent(perms), 0,
              "test_perms_is_nilpotent() C3 not");
}
{
  for(n=1,5,
     my(perms=test_perms_Sn(n));
     check_equal(test_perms_is_nilpotent(perms),  n<=2,
                 Str("test_perms_is_nilpotent() of S"n)));
}


\\-----------------------------------------------------------------------------
progress("vpar_children_of_vertex()");

check_equal(vpar_children_of_vertex([0],1), []);

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     check_equal(vpar_children_of_vertex(vpar,0), vpar_roots(vpar)));
}


\\------------------------------------------------------------------------------
progress("test_perm_insert_n()");

check_equal(test_perm_insert_n(Vecsmall([]),1), Vecsmall([1]));
check_equal(test_perm_insert_n(Vecsmall([1]),1), Vecsmall([1,2]));
check_equal(test_perm_insert_n(Vecsmall([1]),2), Vecsmall([1,2]));
check_equal(test_perm_insert_n(Vecsmall([2,3,1]),1), Vecsmall([1,3,4,2]));
check_equal(test_perm_insert_n(Vecsmall([2,3,1]),4), Vecsmall([2,3,1,4]));
check_equal(test_perm_insert_n(Vecsmall([2,3,4,5,6,1]),3), \
            Vecsmall([2,4, 3, 5,6,7,1]));


\\-----------------------------------------------------------------------------
progress("vpar_isomorphic_perm()");

check_equal(vpar_isomorphic_perm([],[]), Vecsmall([]));

{
  \\ example in vpar.gp initial comments
  my(vpar1=[3,1,0,1,3],
     vpar2=[3,3,5,5,0],
     perm=[3,1,5,2,4]);
  check_equal(vpar_isomorphic_perm(vpar1,vpar2),
              Vecsmall(perm),
              "vpar_isomorphic_perm() example");
  check_equal(vpar_relabel_perm(vpar1,perm), vpar2,
              "vpar_relabel_perm() isomorphism example");
}
{
  my(vpar1=[7,6,1,1,6,0,0],   \\ with 2,5,6
     vpar2=[6,6,5,5,7,0,0],   \\ with 1,2,6
     perm=Vecsmall([5,1,3,4,2,6,7]));
  check_equal(test_is_perm(perm), 1);
  check_equal(perm^-1, Vecsmall([2,5,3,4,1,6,7]));
  check_equal(vpar_relabel_perm(vpar1,perm), vpar2, "perm for forest 7");

  my(got_perm=vpar_isomorphic_perm(vpar1,vpar2));
  check_equal(test_is_perm(got_perm), 1);

  my(vpar2_again=vpar_relabel_perm(vpar1,got_perm));
  check_equal(vpar2_again, vpar2, "vpar_isomorphic_perm() forest 7");
}
{
  my(vpar1=[0,1,2],
     vpar2=[2,0,1],
     perm=Vecsmall([2,1,3]));
  check_equal(test_is_perm(perm), 1);
  check_equal(vpar_relabel_perm(vpar1,perm), vpar2, "perm for path-2");
  check_equal(vpar_isomorphic_perm(vpar1,vpar2), perm,
              "vpar_isomorphic_perm() path-2");
}
{
  \\      2        1     relabel 2->1  = [3,1,2]
  \\     /        /              1->3
  \\    1        3               3->2
  \\   /        /
  \\  3        2        seq 3,1,2  other 2,3,1
  \\                    at 3 store 2, etc
  my(vpar1=[2,0,1],
     vpar2=[0,3,1],
     perm=Vecsmall([3,1,2]));
  check_equal(test_is_perm(perm), 1);

  my(seq1=Vecsmall([3,1,2]),  \\ rotate up
     seq2=Vecsmall([2,3,1]),  \\ rotate down
     id  =Vecsmall([1,2,3]));
  check_equal(seq1*seq2, id);
  check_equal(seq2*seq1, id);
  check_equal(vecextract(seq1,seq2), id);
  check_equal(vecextract(seq2,seq1), id);
  check_equal(vecextract(seq1,seq2^-1), seq2);
  check_equal(vecextract(seq2,seq1^-1), perm);

  check_equal(vpar_relabel_perm(vpar1,perm), vpar2, "perm for path-2 second");
  check_equal(vpar_isomorphic_perm(vpar1,vpar2), perm,
              "vpar_isomorphic_perm() path-2 second");
}
{
  for(i=1,#test_sample_forests,
     my(vpar1=test_sample_forests[i],
        rand_perm=vpar_INTERNAL_random_perm(#vpar1));
     check_equal(test_is_perm(rand_perm), 1);

     my(vpar2=vpar_relabel_perm(vpar1,rand_perm));
     \\ print("vpar1  "vpar1"  rand_perm "Vec(rand_perm));
     \\ print("vpar2 "vpar2);

     my(got_perm=vpar_isomorphic_perm(vpar1,vpar2));
     check_equal(got_perm!='none, 1,
                 "vpar_isomorphic_perm() random relabel, found");
     check_equal(test_is_perm(got_perm), 1);
     check_equal(vpar_relabel_perm(vpar1,got_perm), vpar2,
                 "vpar_isomorphic_perm() random relabel, equal after perm");

     my(num_automorphisms=vpar_num_automorphisms(vpar1));
     my(automorphisms1=if(num_automorphisms<=test_limit_sets,
                          vpar_automorphisms(vpar1),
                          vpar_automorphism_generators(vpar1)));
     for(i=1,#automorphisms1,
        my(a_perm = got_perm * automorphisms1[i]);
        check_equal(vpar_relabel_perm(vpar1,a_perm), vpar2,
                    "vpar_isomorphic_perm() times automorphism1"));

     check_equal(num_automorphisms,vpar_num_automorphisms(vpar2));
     my(automorphisms2=if(num_automorphisms<=test_limit_sets,
                          vpar_automorphisms(vpar2),
                          vpar_automorphism_generators(vpar2)));
     for(i=1,#automorphisms2,
        my(a_perm = automorphisms2[i] * got_perm);
        check_equal(vpar_relabel_perm(vpar1,a_perm), vpar2,
                    "vpar_isomorphic_perm() times automorphism2"));
  );
}


\\------------------------------------------------------------------------------
progress("vpar_free_automorphism_generators()");

check_equal(vpar_free_automorphism_generators([]), [], \
            "vpar_free_automorphism_generators() empty");

check_equal(vpar_free_automorphism_generators([0,1,1]), \
            [Vecsmall([1, 3,2])], \
            "vpar_free_automorphism_generators() swap star-3");
check_equal(vpar_free_automorphism_generators([0,1,2]), \
            [Vecsmall([3,2,1])], \
            "vpar_free_automorphism_generators() reverse path-3");
{
  \\  1  3  4
  \\     |
  \\     2
  check_equal(vpar_free_automorphism_generators([0,3,0,0]),
              [Vecsmall([4, 2,3, 1]),
               Vecsmall([1, 3,2, 4])],
              "vpar_free_automorphism_generators() two swaps");
}
check_equal(vpar_free_automorphism_generators([0,0]), \
            [ Vecsmall([2,1]) ], \
            "vpar_free_automorphism_generators() 2 singletons");
check_equal(vpar_free_automorphism_generators([0,0,0]), \
            [ Vecsmall([1, 3,2]), Vecsmall([3,1,2]) ], \
            "vpar_free_automorphism_generators() 3 singletons");
check_equal(vpar_free_automorphism_generators([0,0,0,0]), \
            [ Vecsmall([1,2, 4,3]), Vecsmall([4,1,2,3]) ], \
            "vpar_free_automorphism_generators() 4 singletons");
{
  \\       1
  \\     /   \
  \\   2      3
  \\  / \    / \
  \\ 4   5  6   7
  my(vpar=[0,1,1,2,2,3,3]);
  vpar_validate(vpar);
  my(got=vpar_free_automorphism_generators(vpar));
  check_equal(got,
              [ Vecsmall([1, 3,2, 6,7, 4,5]),   \\ subtrees 2,3 swap
                Vecsmall([1,2,3, 5,4, 6,7]) ],   \\ leaf 4,5 swap
              "vpar_free_automorphism_generators() complete binary h=2");
  check_equal(test_perm_generators_is_minimal(got), 1);

  \\ PENDING vpar_num_free_automorphism_generators
  \\ check_equal(#got, vpar_num_free_automorphism_generators(vpar),
  \\             Str("vpar_free_automorphism_generators() complete binary h=2,"
  \\                 " vs vpar_num_free_automorphism_generators()"));
}
{
  \\       1     n=6
  \\       |
  \\       6
  \\     /| |\
  \\    2 3 4 5
  my(vpar=[0, 6, 6, 6, 6, 1]);
  vpar_validate(vpar);
  check_equal(vpar_free_automorphism_generators(vpar),
              [ Vecsmall([1,2,3, 5,4, 6]),     \\ swap leaves
                Vecsmall([5,1,2,3,4,  6]) ],   \\ cycle leaves
              "vpar_free_automorphism_generators() tree 6");
  \\ PENDING vpar_num_free_automorphism_generators
  \\ check_equal(vpar_num_free_automorphism_generators(vpar), 2);
}

{
  my(vec=[Vecsmall([1, 5, 6, 4, 2, 3]),    \\ (2,5) (3,6)
          Vecsmall([1, 6, 3, 5, 4, 2]),    \\ (2,6) (4,5)
          Vecsmall([2, 3, 1, 5, 6, 4]),    \\ (1,2,3) (4,5,6)
          Vecsmall([6, 4, 2, 1, 5, 3])]);  \\ (1,2,3,4,6)
  check_equal(test_perm_powers(vec[1]),
              [Vecsmall([1, 2, 3, 4, 5, 6]), Vecsmall([1, 5, 6, 4, 2, 3])]);
  check_equal(test_perm_powers(vec[2]),
              [Vecsmall([1, 2, 3, 4, 5, 6]), Vecsmall([1, 6, 3, 5, 4, 2])]);
  check_equal(test_perm_powers(vec[3]),
              [Vecsmall([1, 2, 3, 4, 5, 6]), Vecsmall([2, 3, 1, 5, 6, 4]),
               Vecsmall([3, 1, 2, 6, 4, 5])]);
  check_equal(test_perm_powers(vec[4]),
              [Vecsmall([1, 2, 3, 4, 5, 6]), Vecsmall([6, 4, 2, 1, 5, 3]),
               Vecsmall([3, 1, 4, 6, 5, 2]), Vecsmall([2, 6, 1, 3, 5, 4]),
               Vecsmall([4, 3, 6, 2, 5, 1])]);

  check_equal(#test_perms_generators_to_group(vec), 6!/2);
  check_equal(vector(#vec,i, #test_perms_generators_to_group(vec[^i])),
              [360, 360, 360, 60]);
  check_equal(test_perm_generators_is_minimal(vec), 0,
              "test_perm_generators_is_minimal() 6");
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     \\ print("vpar "vpar);

     my(got_generators = vpar_free_automorphism_generators(vpar));

     \\ PENDING
     \\ check_equal(#got_generators,
     \\             vpar_num_free_automorphism_generators(vpar),
     \\             Str("vpar_free_automorphism_generators()"
     \\                 " vs vpar_num_free_automorphism_generators()"));

     for(i=1,#got_generators,
        check_equal(#got_generators[i], #vpar,
                    "vpar_free_automorphism_generators() perm size");
        check_equal(test_is_perm(got_generators[i]), 1,
                    "vpar_free_automorphism_generators() is perm"));
     check_equal(test_perm_generators_is_minimal(got_generators), 1);

     if(#vpar <= test_limit_perms_search,
        my(got_perms = test_perms_generators_to_group(got_generators,#vpar));
        got_perms  = vecsort(got_perms);
        check_equal(#got_perms, vpar_num_free_automorphisms(vpar),
          "vpar_free_automorphism_generators() size vs vpar_num_free_automorphisms()");
       );
  );
}


\\ Return a vector of all free_automorphism permutations of vpar, not including
\\ the identify permutation.
test_free_automorphism_perms_by_search(vpar) =
{
  my(l=List([]));
  for(i=1,(#vpar)!-1,
     my(perm=numtoperm(#vpar,i));
     if(vpar_is_free_automorphism(vpar,perm),
        listput(l,Vecsmall(perm))));
  Vec(l);
}


\\-----------------------------------------------------------------------------
progress("test_perm_to_cycles()");

check_equal(test_perm_to_cycles(Vecsmall([])), [], \
            "test_perm_to_cycles() empty");
check_equal(test_perm_to_cycles(Vecsmall([1,2])), \
            [Vecsmall([1]), Vecsmall([2])])
check_equal(test_perm_to_cycles(Vecsmall([2,1])), \
            [Vecsmall([1,2])]);
check_equal(test_perm_to_cycles(Vecsmall([3,4,1,2])), \
            [Vecsmall([1,3]), Vecsmall([2,4])]);


\\-----------------------------------------------------------------------------
progress("test_perm_is_even()");

check_equal(test_perm_is_even(Vecsmall([])), 1, "test_perm_is_even() empty");
check_equal(test_perm_is_even(Vecsmall([1,2])), 1, "test_perm_is_even() id-2");
check_equal(test_perm_is_even(Vecsmall([1,2,3,4,5])), 1, \
            "test_perm_is_even() id-5");
check_equal(test_perm_is_even(Vecsmall([2,1])), 0, \
            "test_perm_is_even() swap");
check_equal(test_perm_is_even(Vecsmall([2,3,4,1])), 0, \
            "test_perm_is_even() 4-cycle");
check_equal(test_perm_is_even(Vecsmall([2,3,1])), 1, \
            "test_perm_is_even() 3-cycle is even");
check_equal(test_perm_is_even(Vecsmall([2,1,4,3])), 1);

test_perm_is_even_by_orders(perm) =
{
  my(vec=test_perm_elem_orders_vector(perm));
  sum(i=1,#vec,
      if(vec[i]%2==0,  \\ even length, is odd set of transpositions
         1/vec[i]))
  % 2 == 0;
}

\\ permsign() new in gp 2.11
test_have_permsign = iferr(permsign(Vecsmall([2,1,4,3])),e,0);
{
  for(n=0,10,
     for(rep=1,n+1,
        my(perm=vpar_INTERNAL_random_perm(n));

        check_equal(test_perm_is_even_by_cycles(perm),
                    test_perm_is_even_by_orders(perm),
           "test_perm_is_even_by_cycles() vs test_perm_is_even_by_orders()");

        check_equal(test_perm_is_even(perm),
                    test_perm_is_even_by_orders(perm),
                    "test_perm_is_even() vs test_perm_is_even_by_orders()");
     ));
}


\\-----------------------------------------------------------------------------
progress("test_free_r_m_e()");

check_equal(test_free_r_m_e([]), [0,0,0], "test_free_r_m_e() empty");
check_equal(test_free_L1_top([]), 0);
check_equal(test_free_L1_height([]), -1);
check_equal(test_free_L2_height([]), -1);

check_equal(test_free_r_m_e([0]), [1,0,1], "test_free_r_m_e() singleton");
check_equal(test_free_L1_top([0]), 0);
check_equal(test_free_L1_height([0]), -1);
check_equal(test_free_L2_height([0]), 0, "test_free_L2_height() singleton");

check_equal(test_free_r_m_e([0,1]), [1,0,2], "test_free_r_m_e() path-2");
check_equal(test_free_L1_top([0,1]), 2);
check_equal(test_free_L1_height([0,1]), 0);
check_equal(test_free_L2_height([0,1]), 0);


\\-----------------------------------------------------------------------------
progress("test_Schultz_index()");
\\ test_modified_Schultz_index()

test_Schultz_index_by_sum(vpar) =
{
  my(m=vpar_path_lengths_matrix(vpar),
     degrees=vpar_degrees_vector(vpar));
  sum(u=1,#vpar,
      sum(v=u+1,#vpar,
          (degrees[u]+degrees[v]) * m[u,v]));
}
test_modified_Schultz_index_by_sum(vpar) =
{
  my(m=vpar_path_lengths_matrix(vpar),
     degrees=vpar_degrees_vector(vpar));
  sum(u=1,#vpar,
      sum(v=u+1,#vpar,
          degrees[u]*degrees[v] * m[u,v]));
}

{
  my(vpar=[0,1]);
  check_equal(test_modified_Schultz_index_by_sum(vpar), 1);
  check_equal(vpar_Wiener_index(vpar), 1);
  check_equal(4*1 - 2*(2*1-1), 2);
  check_equal(test_modified_Schultz_index(vpar), 1);
}
{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i]);
     \\ print(vpar"  n="#vpar"  "test_modified_Schultz_index_by_sum(vpar) - 4*vpar_Wiener_index(vpar));
     check_equal(test_Schultz_index(vpar),
                 test_Schultz_index_by_sum(vpar),
                 "test_Schultz_index() vs test_Schultz_index_by_sum()");
     check_equal(test_modified_Schultz_index(vpar),
                 test_modified_Schultz_index_by_sum(vpar),
     "test_modified_Schultz_index() vs test_modified_Schultz_index_by_sum()");
  );
}

\\ Gutman: "Some Realations Between Distance-Based Polynomials of
\\ Trees". Bulle Classe des Sciences Mathematiques et Naturelles, Sciences
\\ Mathematiques, volume 30, 2005, pages 1-7.
\\ With Gutman setting constant term to n, maybe.
test_Schultz_poly_by_sum(vpar) =
{
  my(m=vpar_path_lengths_matrix(vpar),
     degrees=vpar_degrees_vector(vpar));
  sum(u=1,#vpar,
      sum(v=u+1,#vpar,
          (degrees[u]+degrees[v]) * x^m[u,v]))
  + #vpar;
}
test_modified_Schultz_poly_by_sum(vpar) =
{
  my(m=vpar_path_lengths_matrix(vpar),
     degrees=vpar_degrees_vector(vpar));
  sum(u=1,#vpar,
      sum(v=u+1,#vpar,
          degrees[u]*degrees[v] * x^m[u,v]))
  + #vpar;
}

{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i],
        poly=test_Schultz_poly_by_sum(vpar));
     check_equal(subst(poly', 'x,1),
                 test_Schultz_index(vpar),
                 "test_Schultz_poly_by_sum() derivative test_Schultz_index()");
     check_equal(polcoeff(poly,0), #vpar,
                 "test_Schultz_index() constant term");
  );
}

{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i],
        poly=test_modified_Schultz_poly_by_sum(vpar));
     check_equal(subst(poly', 'x,1),
                 test_modified_Schultz_index(vpar),
                 "test_modified_Schultz_poly_by_sum() derivative test_modified_Schultz_index()");
     check_equal(polcoeff(poly,0), #vpar,
                 "test_modified_Schultz_index() constant term");
  );
}


\\-----------------------------------------------------------------------------
progress("test_seq_is_rooted_stablizing()");

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     if(#vpar <= test_limit_perms_search,
        my(seq=Vecrev(vpar_seq_premax(vpar)));
        check_equal(test_seq_is_rooted_stablizing(vpar,seq), 1,
                    "test_seq_is_rooted_stablizing() of vpar_seq_premax()"));
  );
}


\\-----------------------------------------------------------------------------
progress("test_Narayana()");

{
  \\ A001263 = Narayana
  my(l=List([]));
  for(n=0,#A001263_samples,
     for(k=1,n,
        if(#l==#A001263_samples, break(2));
        listput(l,test_Narayana(n,k))));
  check_equal(Vec(l),
              A001263_samples,
              "A001263 Narayana triangle");
}
{
  for(n=0,10,
     check_equal(sum(k=0,n, test_Narayana(n,k)),
                 vpar_INTERNAL_Catalan_number(n),
                 "test_Narayana() row sum = vpar_INTERNAL_Catalan_number()");
     for(k=0,n,
        my(got=test_Narayana(n,k));
        if(n>=1,
           check_equal(got, test_Narayana(n,n-k+1),
                       "test_Narayana() reverse"));
        if(k>=1,
           check_equal(got,  binomial(n,k)*binomial(n-1,k-1)/(n-k+1),
                       "test_Narayana() n,k n-1,k-1 formula"));

        \\ this form in for example in Hackl, Heuberger, Kropf, Prodinger,
        \\ "Fringe Analysis of Plane Trees Related to Cutting and Pruning",
        \\ arxiv 1704.01095
        if(n>=1,
           check_equal(got,  binomial(n,k)*binomial(n,k-1)/n,
                       "test_Narayana() /n formula"));

     ));
}

\\ for(n=1,8, for(k=0,n, printf("%6d",test_Narayana(n,k))); print());
\\ for(n=1,8, for(k=0,n, printf("%6d",binomial(n,k)*binomial(n,k-1)/n)); print());


\\------------------------------------------------------------------------------
progress("test_make_ordered_partitions()");

check_equal(test_make_ordered_partitions(0), [ [] ], \
            "test_make_ordered_partitions() 0");
check_equal(test_make_ordered_partitions(1), [ [[1]] ]);

{
  my(want=[1, 1, 3, 13, 73, 501, 4051]);  \\ A000262 sets of lists
  for(n=0,#want-1,
     my(vec=test_make_ordered_partitions(n),
        count_by_parts=vector(n));
     for(i=1,#vec,
        test_ordered_partition_validate(vec[i]);
        if(n, count_by_parts[#vec[i]]++));
     check_equal(#vec, want[n+1]);
     check_equal(count_by_parts, vector(n,k, test_A105278(n,k)));
  );
}

\\ check_exact(#make_all_sets(2),  2);
\\ check_exact(#make_all_sets(3),  5);
\\ check_exact(#make_all_sets(4), 15);


\\-----------------------------------------------------------------------------
progress("test_permorder()");

check_equal(test_permorder(Vecsmall([1,2,3])), 1);
check_equal(test_permorder(Vecsmall([2,1,3])), 2);
check_equal(test_permorder(Vecsmall([5,1,2,3,4])), 5);
check_equal(test_permorder(Vecsmall([5,1,2,3,4, 6,7,8])), 5);
check_equal(test_permorder(Vecsmall([2,3,1, 5,4])), 6);
check_equal(test_permorder(Vecsmall([2,1, 4,3])), 2);


\\------------------------------------------------------------------------------
progress("test_perm_is_swap1()");

check_equal(test_perm_is_swap1([]), 0, "test_perm_is_swap1() empty");
check_equal(test_perm_is_swap1([1,2]), 0);
check_equal(test_perm_is_swap1([2,1]), 1, "test_perm_is_swap1() swap");
check_equal(test_perm_is_swap1([2,3,1]), 0, "test_perm_is_swap1() 3-cycle");
check_equal(test_perm_is_swap1([2,1,3,4]), 1);
check_equal(test_perm_is_swap1([2,1,4,3]), 0, "test_perm_is_swap1() 2 swaps");


\\------------------------------------------------------------------------------
progress("test_seq_free_stabilizing_by_search()");

check_equal(test_seq_free_stabilizing_by_search([]), Vecsmall([]));
check_equal(test_seq_free_stabilizing_by_search([0]), Vecsmall([1]));
check_equal(test_seq_free_stabilizing_by_search([0,1]), Vecsmall([1,2]));
check_equal(test_seq_free_stabilizing_by_search([0,1,2]), Vecsmall([2,1,3]));
check_equal(test_seq_free_stabilizing_by_search([0,1,2,3]), 'none);
check_equal(test_seq_free_stabilizing_by_search([0,1,2,3,4]), 'none);
check_equal(test_seq_free_stabilizing_by_search([0,1,2,3,4,5]), 'none);
check_equal(test_seq_free_stabilizing_by_search([0,1,1,1,1]), Vecsmall([1,2,3,4,5]));
check_equal(test_seq_free_stabilizing_by_search([5,5,5,5,0]), Vecsmall([1,2,5,3,4]));


\\------------------------------------------------------------------------------
progress("test_perm_delete_n()");

check_equal(test_perm_delete_n([1,2,3,4],4), [1,2,3]);
check_equal(test_perm_delete_n([1],1), []);
check_equal(test_perm_delete_n(Vecsmall([2,3,1]),1), Vecsmall([2,1]), \
            "test_perm_delete_n() cycle-3 del 1");
check_equal(test_perm_delete_n(Vecsmall([2,3,1]),2), Vecsmall([2,1]), \
            "test_perm_delete_n() cycle-3 del 2");
check_equal(test_perm_delete_n(Vecsmall([2,3,1]),3), Vecsmall([2,1]), \
            "test_perm_delete_n() cycle-3 del 3");


\\-----------------------------------------------------------------------------
progress("test_perm_reversal()");

check_equal(test_perm_reversal(0), Vecsmall([]));
check_equal(test_perm_reversal(4), Vecsmall([4,3,2,1]));
check_equal(test_perm_reversal(5), Vecsmall([5,4,3,2,1]));


\\-----------------------------------------------------------------------------
progress("test_perm_resize()");

check_equal(test_perm_resize(Vecsmall([]), 0), Vecsmall([]));
check_equal(test_perm_resize(Vecsmall([1,2,3]), 0), Vecsmall([]));
check_equal(test_perm_resize(Vecsmall([2,1]), 2), Vecsmall([2,1]));
check_equal(test_perm_resize(Vecsmall([2,1]), 4), Vecsmall([2,1, 3,4]));


\\-----------------------------------------------------------------------------
progress("test_perms_An()");

{
  my(A4 = test_perms_An(4));
  check_equal(#A4[1], 4,  "test_perms_An(4) elem length");
  check_equal(#A4,    12, "test_perms_An(4) num perms");
}


\\-----------------------------------------------------------------------------
progress("test_perms_generators_to_group()");

{
  check_equal(test_perms_generators_to_group([ Vecsmall([2,1]) ]),
              [ Vecsmall([1,2]), Vecsmall([2,1])  ],
              "test_perms_generators_to_group() of [2,1]");
}
{
  my(vec=[ Vecsmall([2,3,4,5,6,1]), Vecsmall([2,1, 3,4,5,6]) ]);
  check_equal(#test_perms_generators_to_group(vec), 6!,
              "test_perms_generators_to_group() 6");
}


\\-----------------------------------------------------------------------------
progress("test_perm_generators_is_minimal()");

{
  check_equal(test_perm_generators_is_minimal([ Vecsmall([2,1]) ]), 1,
              "test_perm_generators_is_minimal() of [2,1]");
  check_equal(test_perm_generators_is_minimal([ Vecsmall([2,1,3]),
                                               Vecsmall([3,2,1]) ]), 1);
  check_equal(test_perm_generators_is_minimal([ Vecsmall([1,2, 4,3]),
                                               Vecsmall([2,1, 3,4]) ]), 1);
  check_equal(test_perm_generators_is_minimal([ Vecsmall([1,2, 4,3]),
                                               Vecsmall([2,1, 3,4]),
                                               Vecsmall([2,1, 4,3])  \\ product
                                              ]), 0);
}


\\-----------------------------------------------------------------------------
progress("test_perms_is_solvable()");

check_equal(test_perms_is_solvable([ Vecsmall([]) ]), 1, \
            "test_perms_is_solvable() trivial empty");
check_equal(test_perms_is_solvable([ Vecsmall([1,2]) ]), 1);
check_equal(test_perms_is_solvable([ Vecsmall([1,2]), Vecsmall([2,1]) ]), 1);
{
  for(n=1,5,
     my(perms=test_perms_Sn(n));
     check_equal(test_perms_is_solvable(perms),  n<=4,
                 Str("test_perms_is_solvable() S"n)));
}


\\-----------------------------------------------------------------------------
progress("test_perms_nilpotency_class()");

check_equal(test_perms_nilpotency_class([ Vecsmall([]) ]), 1, \
            "test_perms_nilpotency_class() trivial empty");
check_equal(test_perms_nilpotency_class([ Vecsmall([1,2]),Vecsmall([2,1]) ]), \
            2,  "test_perms_nilpotency_class() of C2");


\\-----------------------------------------------------------------------------
progress("test_Padovan()");

test_Padovan_by_recurrence(n) =
{
  if(n==0, return(1));
  if(n==1, return(0));
  if(n==2, return(0));
  my(a=vector(n));
  a[3]=1;
  for(n=4,n, a[n]=a[n-2]+a[n-3]);
  a[n];
}
{
  \\ A000931 Padovan numbers
  check_equal(vector(50,n,n--; test_Padovan(n)),
              vector(50,n,n--; test_Padovan_by_recurrence(n)));
}
{
  \\ A000931 Padovan numbers
  check_equal(vector(#A000931_samples,n,n--; test_Padovan(n)),
              A000931_samples,    \\ OFFSET=0
              "test_Padovan() = A000931");
}


\\-----------------------------------------------------------------------------
progress("test_vector_product()");

check_equal(test_vector_product([]),1);
check_equal(test_vector_product([3]),3);
check_equal(test_vector_product([3,4]),12);
check_equal(test_vector_product([ Vecsmall([3,1,2]), Vecsmall([2,3,1]) ]), \
                                Vecsmall([1,2,3]));


\\-----------------------------------------------------------------------------
progress("test_func_to_gf()");
\\ test_func_to_egf()

check_equal(test_func_to_gf(n->n^2+1, 5), \
            1 + 2*x + 5*x^2 + 10*x^3 + 17*x^4 + O(x^5));

check_equal(test_func_to_egf(n->n^2+1, 5), \
            1 + 2*x + 5/2!*x^2 + 10/3!*x^3 + 17/4!*x^4 + O(x^5));


\\------------------------------------------------------------------------------
progress("test_egf1s()");

{
  \\ egf exp(x) is all 1s
  my(limit=10);
  check_equal(test_egf1s(limit), exp(x) + O(x^limit),
              "test_egf1s() is exp(x)");
}


\\------------------------------------------------------------------------------
progress("test_egfN()");

{
  my(limit=10,
     L=O(x^limit));
  check_equal(test_egfN(limit),  x*test_egf1s(limit) + L,
              "test_egf1s() from test_egf1s()");
  check_equal(test_egfN(limit),  x*exp(x) + L,
              "test_egfN() is x*exp(x)");
}


\\------------------------------------------------------------------------------
progress("test_LambertW()");

\\ 0, 1, -2, 9, -64, 625, -7776, 117649, -2097152, 43046721
{
  \\ A213236 values whole egf is Lambert W
  my(samples=concat(0, A213236_samples));
  check_equal(test_egf_terms(test_LambertW(#samples),#samples), samples,
             "A213236 Lambert W");
}
{
  my(limit=10,
     L = O(x^limit),
     W = test_LambertW(limit));

  \\ W = functional inverse of x*exp(x)
  check_equal(subst(W,x,x*exp(x+L)), x + L);
  check_equal(subst(x*exp(x+L),x,W), x + L);

  check_equal(W, x/exp(W) + L, "x/exp(W)");
  check_equal(exp(W) + L/x, x/W, "x/W");
  check_equal(1/W, exp(W)/x + L/x^2, "1/W");

  \\ exp(exp(x)-1) = Bell numbers
  check_equal(test_func_to_egf(vpar_INTERNAL_Bell_number, limit),
              exp(exp(x+L)-1));
}


\\------------------------------------------------------------------------------
progress("test_weigh_transform()");

{
  \\ A035079 = weigh transform of A007561
  my(input_samples=A007561_samples[^1],
     output_samples=A035079_samples[^1],
     len=min(#input_samples,#output_samples));
  input_samples =input_samples[1..len];
  output_samples=output_samples[1..len];
  check_equal(test_weigh_transform(input_samples), output_samples);
}

{
  \\ A000009 weigh transform of 1s = partitions of n
  my(samples=A000009_samples[^1]);
  check_equal(test_weigh_transform(vector(#samples,i,1)),
              samples);
}
{
  \\ A022567 weigh transform of 2s
  my(samples=A022567_samples[^1]);
  check_equal(test_weigh_transform(vector(#samples,i,2)),
              samples);
}
{
  \\ A026007 weigh transform of f(n)=n is partitions with n types of size n
  my(samples=A026007_samples[^1]);
  check_equal(test_weigh_transform([1..#samples]),
              samples);
}
{
  \\ A294836 weigh transform of hexagonals A000384 = n*(2*n-1)
  my(samples=A294836_samples[^1]);
  check_equal(test_weigh_transform(vector(#samples,n, n*(2*n-1))),
              samples,
              "A294836 weigh transform of hexagonals");

  my(hexagonal = (n)->n*(2*n-1),
     a=vector(#samples));
  for(n=1,#a,
     \\ a[n] = 1/n * sum(k=1,n, if(n-k==0,1,a[n-k]) * sumdiv(k,d, d*hexagonal(d)*(-1)^(1+k/d)));
     a[n] = -1/n * ( sumdiv(n,d, d*hexagonal(d)*(-1)^(n/d))
                   + sum(k=1,n-1, if(n-k==0,1,a[n-k]) * sumdiv(k,d, d*hexagonal(d)*(-1)^(k/d))));
  );
  check_equal(a, samples,
             );
}


\\------------------------------------------------------------------------------
progress("test_vector_is_increasing()");
\\ test_vector_is_decreasing()

check_equal(test_vector_is_increasing([]), 1, \
            "test_vector_is_increasing() empty");
check_equal(test_vector_is_increasing([9]), 1);
check_equal(test_vector_is_increasing([8,9]), 1);
check_equal(test_vector_is_increasing([9,8]), 0);
check_equal(test_vector_is_increasing([ [1,3], [2] ]), 1);

check_equal(test_vector_is_decreasing([]), 1, \
            "test_vector_is_decreasing() empty");
check_equal(test_vector_is_decreasing([9]), 1);
check_equal(test_vector_is_decreasing([8,9]), 0);
check_equal(test_vector_is_decreasing([9,8]), 1);
check_equal(test_vector_is_decreasing(Vecsmall([1,1,0])), 1);
check_equal(test_vector_is_decreasing([ [2], [1,3] ]), 1);


\\------------------------------------------------------------------------------
progress("test_vector_is_all_different()");

check_equal(test_vector_is_all_different([]), 1);
check_equal(test_vector_is_all_different([1,2,3]), 1);
check_equal(test_vector_is_all_different([1,2,2]), 0);


\\------------------------------------------------------------------------------
progress("test_vector_search()");

check_equal(test_vector_search([1,9,1,3],9), 2);
check_equal(test_vector_search([1,9,1,3],1), 1);
check_equal(test_vector_search([1,9,1,3],6), 0);


\\------------------------------------------------------------------------------
progress("test_vecsort_vecmin()");

check_equal(test_vecsort_vecmin([]), []);
check_equal(test_vecsort_vecmin([ [9,2], [3,1] ]), [ [3,1], [9,2] ]);


\\------------------------------------------------------------------------------
progress("test_is_pow2()");

check_equal(test_is_pow2(0), 0);
check_equal(test_is_pow2(1), 1);
check_equal(test_is_pow2(2), 1);
check_equal(test_is_pow2(3), 0);
check_equal(test_is_pow2(4), 1);
check_equal(test_is_pow2(5), 0);
check_equal(test_is_pow2(7), 0);
check_equal(test_is_pow2(8), 1);


\\------------------------------------------------------------------------------
progress("test_compositions()");

check_equal(test_compositions(1,[0,1],2), \
            [Vecsmall([0,1]), Vecsmall([1,0])]);
check_equal(test_compositions(2,[0,2],3), \
            vecsort([Vecsmall([0,0,2]), Vecsmall([0,2,0]), Vecsmall([2,0,0]), \
                     Vecsmall([0,1,1]), Vecsmall([1,0,1]), Vecsmall([1,1,0])]));
check_equal(test_compositions(0), [Vecsmall([])]);
check_equal(test_compositions(1), [Vecsmall([1])], "test_compositions() n=1");
check_equal(test_compositions(2), [Vecsmall([2]), Vecsmall([1,1])]);
{
  \\ number of compositions into n pieces
  my(want=[1, 1,2,4,8,16,32,64]);
  check_equal(vector(#want,n,n--; #test_compositions(n)), want);
}
{
  \\ number of compositions into n pieces of length l, binomial coeff
  \\ 1
  \\ 1, 1
  \\ 1, 2, 1
  \\ 1, 3, 3, 1
  \\ 1, 4, 6, 4, 1
  check_equal(matrix(6,6,n,l,n--; #test_compositions(n,,[l,l])),
              matrix(6,6,n,l,n--; if(n==0,1, binomial(n-1,l-1))));
}


\\------------------------------------------------------------------------------
progress("A000262 = partitions of 1..n into ordered lists");

\\ 1, 3, 13, 73, 501, 4051, 37633, 394353,
\\ n=1  [1]
\\ n=2  [[1],[2]]  [[1,2]]  [[2,1]]

A000262_vector(n) =
{
  \\ recurrence per OEIS A000262
  my(ret=vector(n));
  if(n>=1,
     ret[1]=1;
     if(n>=2,
        ret[2]=3;
        for(n=3,n, ret[n] = (2*n-1)*ret[n-1] - (n-1)*(n-2)*ret[n-2])));
  ret;
}
{
  my(want=A000262_samples[^1]);  \\ starting from n=1
  check_equal(A000262_vector(#want), want,
              "A000262_vector() vs A000262_samples");
}
A000262(n) = if(n==0,1, A000262_vector(n)[n]);

\\ list containing 1 size k, choose its k-1 others, ordered, times rest F
A000262_by_recurrence(n) =
{
  my(vec=A000262_vector(n),
     F=(n)->if(n==0,1, vec[n]));
  if(n<=1,1, sum(k=1,n, binomial(n-1,k-1) * k! * F(n-k)));
}
check_equal(A000262_vector(100), vector(100,n,A000262_by_recurrence(n)), \
            "A000262 by binomial recurrence")

check_equal(vpar_INTERNAL_exponential_transform(vector(20,n, n!)), \
            A000262_vector(20), \
            "A000262 = exponential transform of factorials");

{
  my(l=List([]));
  for(n=1,#A105278_samples,
     for(k=1,n,
        if(#l>=#A105278_samples,break(2));
        listput(l, test_A105278(n,k))));
  check_equal(Vec(l),
              A105278_samples,
              "A105278 function");
}

\\ T(n) = num trees consecutive siblings so vpar_minmax_sibling_vector()
\\ is increasing
\\
A000262_T(n) = if(n<=1,1, sum(k=0,n-1, (n-k+1)*test_A105278(n-1,k)));

{
  my(limit = 20,
     L = O('x^limit),
     gFactorials_sans_0 = test_func_to_egf(n->if(n==0,0,n!), limit));
  check_equal(gFactorials_sans_0, 1/(1-x)-1 + L);
  check_equal(gFactorials_sans_0, x/(1-x) + L);
  check_equal(gFactorials_sans_0', 1/(1-x)^2 + L/x);

  \\ F(n) = A000262 = num forests consecutive siblings so
  \\ vpar_minmax_sibling_vector() is increasing
  \\
  my(gF = test_func_to_egf(A000262, limit));
  check_equal(gF, exp(gFactorials_sans_0),
              "A000262 egf exp of factorials");
  check_equal(gF,  exp(x/(1-x) + L));

  check_equal(vector(20,n,n--; A000262(n)),
              vector(20,n,n--; if(n==0,1, 1/n * sum(l=1,n, binomial(n,l) * l*l! * A000262(n-l)))));

  check_equal(x*gFactorials_sans_0',
              test_func_to_egf(n->if(n==0,0,n*n!), limit));
  check_equal(gF,  1 + intformal( x*gFactorials_sans_0' * gF / x));
  check_equal(gF,  1 + intformal( gFactorials_sans_0' * gF));

  \\ T(n) = num trees consecutive siblings so vpar_minmax_sibling_vector()
  \\ is increasing
  \\
  my(gT = test_func_to_egf(A000262_T, limit));

  \\ from OEIS superseeker starting at n=1
  check_equal(gT + L/x^2,
              1 - x + intformal( (2-4*x+3*x^2) / (1-x)^2 * exp(x/(1-x)) ),
              "A000262_T intformal");
  check_equal(poldisc(2-4*x+3*x^2), -8);

  check_equal(gT,
              1 - x + intformal( (2-4*x+3*x^2) / (1-x)^2 * gF ) + L,
              "A000262_T vs gF");

  \\ recurrence from OEIS superseeker starting n=0
  \\     4      3      2              3      2                   2
  \\ [{(n  + 2 n  - 3 n ) a(n) + (-2 n  - 5 n  - n) a(n + 1) + (n  + n) a(n + 2),
  \\     a(0) = 1, a(1) = 1, a(2) = 2}, ogf]

  check_equal(vector(20,n,n+=2; (2*n^2 - 3*n - 1)/(n-1)*A000262_T(n-1)
                              - (n+1)*(n-2)*(n-3)/(n-1)*A000262_T(n-2)),
              vector(20,n,n+=2; A000262_T(n)));
  check_equal(poldisc(2*n^2-3*n-1), 17);


  \\ print(test_egf_terms(1 - x + intformal((2-4*x+3*x^2)/(x-1)^2  *  exp(x/(1-x))), 20));
  \\ print(test_egf_terms(1 - x + intformal((2-4*x+3*x^2)/(x-1)^2  *  exp(x/(1-x))), 20));
  \\ print(test_egf_terms(gT, 20));

  \\ \\ print(test_egf_terms(gT / exp(x/(1-x))  *(x-1)^2, 10));
  \\ quit;
  \\ check_equal(gT, (2-4*x+3*x^2) * exp(gFactorials_sans_0) / gFactorials_sans_0');
  \\ check_equal(gT/gFactorials_sans_0', 0);
}

\\ Superseeker T(n) starting n=0 gives, with Ei() the exponential integral.
\\
\\          1              1                1
\\  exp(- -----) - Ei(1, -----) + 3 exp(- -----) (x - 1)
\\        x - 1          x - 1            x - 1
\\ [---------------------------------------------------- - x
\\                         exp(1)
\\        3 exp(1) + Ei(1, -1)
\\      + --------------------, egf]
\\               exp(1)
\\
\\     4      3      2              3      2                   2
\\ [{(n  + 2 n  - 3 n ) a(n) + (-2 n  - 5 n  - n) a(n + 1) + (n  + n) a(n + 2),
\\     a(0) = 1, a(1) = 1, a(2) = 2}, ogf]
\\     4      3      2              4      3      2
\\ [{(n  + 2 n  - 3 n ) a(n) + (-2 n  - 7 n  - 6 n  - n) a(n + 1)
\\          4      3      2
\\      + (n  + 4 n  + 5 n  + 2 n) a(n + 2), a(0) = 1, a(1) = 1, a(2) = 1}, egf]




\\-----------------------------------------------------------------------------
progress("vpar_seq_downqueue()");

test_seq_downqueue_by_sort(vpar) =
{
  my(depths=vpar_depths_vector(vpar),
     keys = vector(#vpar,v,
                   concat([depths[v], Vecrev(vpar_ancestors(vpar,v))])));
  vecsort(keys,,1);
}
test_seq_downqueue_by_preorder(vpar) =
{
  \\ vecsort() is "stable", so sort by depths and among equal depths by
  \\ preorder seq
  my(depths=vpar_depths_vector(vpar),
     seq=vpar_seq_preorder(vpar));
  Vecsmall(vecsort(Vec(seq), v->depths[v]));
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);

     my(seq=vpar_seq_downqueue(vpar));
     check_equal(test_is_perm(seq), 1,
                 "vpar_seq_downqueue() is perm");
     check_equal(vpar_seq_is_ordered(vpar,seq), 1,
                 "vpar_seq_downqueue() is vpar_seq_is_ordered()");
        check_equal(test_seq_is_siblings_consecutive(vpar,seq), 1,
                    "vpar_seq_downqueue() is test_seq_is_siblings_consecutive() per intro comments");

     check_equal(seq, test_seq_downqueue_by_sort(vpar),
                 "vpar_seq_downqueue() vs test_seq_downqueue_by_sort()");
     check_equal(seq, test_seq_downqueue_by_preorder(vpar),
                 "vpar_seq_downqueue() vs test_seq_downqueue_by_preorder()");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_INTERNAL_preorder_subtree_cmp()");

{
  \\      1              1
  \\    /   \
  \\   2      6
  \\  /|\    /|\\
  \\ 3 4 5  7 8 9 10
  my(vpar=[0,1,2,2,2, 1,6,6,6,6]);
  vpar_validate(vpar);
  check_equal(vpar_is_preorder(vpar), 1);
  check_equal(vpar_INTERNAL_preorder_subtree_cmp(vpar,3,4), 0,
              "vpar_INTERNAL_preorder_subtree_cmp() 3,4");
  check_equal(vpar_INTERNAL_preorder_subtree_cmp(vpar,2,6), -1);
}

\\ UNTESTED
\\ test_random_siblings(vpar) =
\\ {
\\   my(children=vpar_children_vector(vpar),
\\      r=random(sum(i=1,#children, binomial(#children[i],2))));
\\   for(i=1,#children,
\\      my(t=binomial(#children[i],2));
\\      if(r>=t, r-=t; next);
\\      forstep(j=#children[i],1,-1,
\\         if(r>=j, r-=j; next);
\\         return([r,j]));
\\      error("oops"));
\\   error("oops"));
\\ }

{
  for(i=1,#test_sample_preorder_forests,
     my(vpar=test_sample_preorder_forests[i]);
     #vpar>=1 || next;
     for(rep=1,10,
        my(u=vpar_random_vertex(vpar),
           s=vpar_children_of_vertex(vpar,vpar[u]),
           v=s[random(#s)+1]);
        check_equal(vpar[u],vpar[v], "u,v are siblings");
        check_equal(vpar_INTERNAL_preorder_subtree_cmp(vpar,u,v),
                    lex(vpar_subtree(vpar,u), vpar_subtree(vpar,v)),
                    "vpar_INTERNAL_preorder_subtree_cmp()")));
}


\\-----------------------------------------------------------------------------
progress("vpar_preorder_prev()");

{
  \\      1               1
  \\     / \             / \
  \\    2   3     ->    2   4       drop 3
  \\         \         /            set 4 under root
  \\          4       3
  my(vpar=[0, 1, 1, 3],
     want=[0, 1, 2, 1]);
  check_equal(vpar_preorder_next(want), vpar);
  check_equal(vpar_preorder_prev(vpar), want,
              "vpar_preorder_prev() 4");
}

{
  for(forest=0,1,
     for(n=0,10-forest,
        my(vpar=vpar_preorder_first(n),
           prev_vpar='none);
        until(vpar=='none,
              check_equal(vpar_preorder_prev(vpar,forest), prev_vpar,
                          "vpar_preorder_prev() inverse of vpar_preorder_next()");
              prev_vpar=vpar;
              vpar=vpar_preorder_next(vpar,forest))));
}


\\-----------------------------------------------------------------------------
progress("vpar_count_unrooted()");

{
  check_equal(vector(#A001858_samples,n,n--; vpar_count_unrooted(n,1)),
              A001858_samples,  \\ OFFSET=0
              "A001858 = vpar_count_unrooted() forests");
}

\\ No, not currently an error.
\\ {
\\   for(forest=0,1,
\\      for(n=-5,-1,
\\         check_equal(iferr(vpar_count_unrooted(n,forest);0, e, 1),
\\                     1,
\\                     "vpar_count_unrooted() error on negative n")));
\\ }

\\ Variant on Michael Somos' code in A001858
test_count_unrooted_forests_by_double_sum(n) =
{
  \\ n-m-j>=0 is j<=n-m
  \\ n-1-(n-m-j) = m+j-1 >= 0 is j>=1-m
  \\ j>=0
  \\ m-j>=0 is j<=m

  if(n<=1, 1,
     sum(m=0,n,
           sum(j=max(0,1-m),min(m,n-m),
                 (n-1)! * n^(n-m-j) * (m+j)
                 / (n-m-j)! / (m-j)! / j! / (-2)^j)));
}
{
  check_equal(vector(20,n,n--; vpar_count_unrooted(n,1)),
              vector(20,n,n--; test_count_unrooted_forests_by_double_sum(n)),
  "vpar_count_unrooted() vs test_count_unrooted_forests_by_double_sum()");
}


\\-----------------------------------------------------------------------------
progress("vpar_num_children_vector()");

check_equal(vpar_num_children_vector(vpar_make_path(5)),[1,1,1,1,0]);
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        vec=vpar_num_children_vector(vpar));
     check_equal(vec, vector(#vpar,v, vpar_num_children_of_vertex(vpar,v)),
                 "vpar_num_children_vector() vs individual vpar_num_children_of_vertex()");
     check_equal(vecsum(vec), vpar_num_edges(vpar),
                 "vpar_num_children_vector() vecsum = num edges"));
}

\\ vpar_num_children_vector() uniques vs partitions(), per addhelp
{
  my(forest=0);
  for(n=2, test_limit_trees_search - forest,
     my(parts=List([]),
        comps=List([]));
     forvec(vec=vpar_forvec_from(n,forest), \\ trees
           my(vpar=vpar_from_upascend(vec),
              num_children=Vecsmall(vpar_num_children_vector(vpar)),
              pos=setsearch(comps,num_children,1));
           if(pos, listinsert(comps,num_children,pos));

           num_children=vecsort(num_children);
           pos=setsearch(parts,num_children,1);
           if(pos, listinsert(parts,num_children,pos)));

     check_equal(Vec(parts),      partitions  (n-1,[0,n-1],n));
     check_equal(Vec(comps), test_compositions(n-1,[0,n-1],n));
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_num_siblings_of_vertex()");
\\ vpar_num_siblings_vector()

check_equal(vpar_num_siblings_vector(vpar_make_path(5)),[0,0,0,0,0]);
check_equal(vpar_num_siblings_vector(vpar_make_star(5)),[0,3,3,3,3]);
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     check_equal(vpar_num_siblings_vector(vpar),
                 vector(#vpar,v, vpar_num_siblings_of_vertex(vpar,v))));
}

{
  \\ A071720 total mean num siblings, forests
  \\   num spanning trees in K(n)-e, complete graph less one edge
  \\   (n-2)*n^(n-3)
  my(forest=1);
  for(n=1, test_limit_trees_search - forest,
     my(total=0);
     forvec(vec=vpar_forvec_from(n,forest), \\ forests
           my(vpar=vpar_from_upascend(vec));
           total += vecsum(vpar_num_siblings_vector(vpar)) / n);  \\ mean
     check_equal(total, A071720_samples[n],
                 "A071720 mean siblings total, forests"));
}
{
  \\ A065513 total num siblings, root=1 trees
  \\   endofunctions with a 3-cycle a->b->c->a and some iterate f^k(x)=a
  \\   (n-1)*(n-2)*n^(n-3)
  my(forest=-1);
  for(n=3, test_limit_trees_search - forest,
     my(total=0);
     forvec(vec=vpar_forvec_from(n,forest), \\ root=1
           my(vpar=vpar_from_upascend(vec));
           total += vecsum(vpar_num_siblings_vector(vpar)));
     check_equal(total, A065513_samples[n-2],
                 "A065513 num siblings total, root=1"));

  \\ A065513 total mean num siblings, trees
  my(forest=0);
  for(n=3, test_limit_trees_search - forest,
     my(total=0);
     forvec(vec=vpar_forvec_from(n,forest), \\ trees
           my(vpar=vpar_from_upascend(vec));
           total += vecsum(vpar_num_siblings_vector(vpar)) / n);  \\ mean
     check_equal(total, A065513_samples[n-2],
                 "A065513 mean siblings total, trees"));
}


\\-----------------------------------------------------------------------------
progress("vpar_skew_adjacency_poly()");

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        poly=vpar_skew_adjacency_poly(vpar));
     check_equal(type(poly), "t_POL");

     \\ print();
     \\ print(vpar);
     \\ print(vpar_skew_adjacency_poly(vpar));
     \\ print(charpoly(vpar_skew_adjacency_matrix(vpar)));

     check_equal(poly, charpoly(vpar_skew_adjacency_matrix(vpar)),
                 "vpar_skew_adjacency_poly() vs matrix charpoly");

     check_equal(poly, subst(vpar_matchpoly(vpar), 'x, 'x^-2) * x^#vpar,
                 "vpar_skew_adjacency_poly() vs matchpoly, per addhelp");

     \\ print(poly," vs ",subst(Polrev(vpar_matchpoly(vpar)), 'x, 'x^2)"  from matchpoly "vpar_matchpoly(vpar)" vpar="vpar);
     \\ check_equal(poly, Pol(subst(Polrev(vpar_matchpoly(vpar)), 'x, 'x^2)),
     \\             "vpar_skew_adjacency_poly() vs matchpoly Polrev");


     \\ L. Collatz and U. Sinogowitz, Spektren Endlicher Grafen,
     \\ Abh. Math. Sem. Univ. Hamburg, volume 21, 1957, pages 63-77.
     check_equal(polcoeff(poly,#vpar), 1,
       "vpar_skew_adjacency_poly() high term 1 empty matching, per addhelp");

     \\ Collatz and Sinogowitz for second term corresonding to trace(m)==0
     forstep(i=1,#vpar,2,
        check_equal(polcoeff(poly,#vpar-i), 0,
                 "vpar_skew_adjacency_poly() odd term == 0, per addhelp"));

     check_equal(polcoeff(poly,#vpar-2), vpar_num_edges(vpar),
       "vpar_skew_adjacency_poly() second high term num edges, per addhelp");

     check_equal(polcoeff(poly,0), vpar_has_perfect_matching(vpar),
       "vpar_skew_adjacency_poly() constant term vpar_has_perfect_matching(), per addhelp");

     check_equal(valuation(poly,'x), #vpar-2*vpar_matchnum(vpar),
     "vpar_skew_adjacency_poly() valuation is 2*vpar_matchnum(), per addhelp");

     my(matchings_count=vpar_matchings_count(vpar));
     check_equal(subst(poly,'x,1), matchings_count,
     "vpar_skew_adjacency_poly() sum vpar_matchings_count(), per addhelp");
     check_equal(vecsum(Vec(poly)), matchings_count,
     "vpar_skew_adjacency_poly() sum vpar_matchings_count(), per addhelp");

     my(component_vpar=vpar_component_trees(vpar));
     check_equal(poly,
                 Pol(prod(i=1,#component_vpar, vpar_skew_adjacency_poly(component_vpar[i]))),
                 "vpar_skew_adjacency_poly() product of components, per addhelp");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_INTERNAL_Bell_number()");

check_equal(vector(7,n,n--; vpar_INTERNAL_Bell_number(n)), [1,1,2,5,15,52,203]);
check_equal(vector(#A000110_samples,n,n--; vpar_INTERNAL_Bell_number(n)), \
                    A000110_samples);  \\ OFFSET=0


\\-----------------------------------------------------------------------------
progress("vpar_reachability_matrix()");

check_equal(vpar_reachability_matrix([]), Mat([]));
check_equal(vpar_reachability_matrix([0]), Mat([0]));
check_equal(vpar_reachability_matrix([0,1,2,3]), \
            [0,1,1,1; 1,0,1,1; 1,1,0,1; 1,1,1,0]);


\\ check_equal(vpar_reachability_matrix([0,1],1), [0,0; 1,0], \
\\             "vpar_reachability_matrix() path-2 directed=1");
\\ check_equal(vpar_reachability_matrix([0,1],2), [0,1; 0,0], \
\\             "vpar_reachability_matrix() path-2 directed=2");
check_equal(vpar_reachability_matrix([0,1]), [0,1; 1,0], \
            "vpar_reachability_matrix() path-2 directed=0");
\\
\\ check_equal(vpar_reachability_matrix([0,1,2],1), [0,0,0; 1,0,0; 1,1,0], \
\\             "vpar_reachability_matrix() path-3 directed=1");
\\ check_equal(vpar_reachability_matrix([0,1,2],2), [0,1,1; 0,0,1; 0,0,0], \
\\             "vpar_reachability_matrix() path-3 directed=2");
check_equal(vpar_reachability_matrix([0,1,2]), [0,1,1; 1,0,1; 1,1,0], \
            "vpar_reachability_matrix() path-3 directed=0");

check_equal(vpar_reachability_matrix([0,1,1]), [0,1,1; 1,0,1; 1,1,0], \
            "vpar_reachability_matrix() star-3 directed=0");
check_equal(vpar_reachability_matrix([0,1,2,2]), \
            [0,1,1,1; 1,0,1,1; 1,1,0,1; 1,1,1,0]);

test_reachability_matrix_by_adjacency_power(vpar) =
{
  my(m=vpar_adjacency_matrix(vpar),
     p=m,
     ret=m);
  for(n=2,#vpar,
     p*=m;  \\ so m^n
     ret+=p);
  for(i=1,#vpar, ret[i,i]=0);
  apply(c->c!=0, ret);
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);

     check_equal(vpar_reachability_matrix(vpar),
                 test_reachability_matrix_by_adjacency_power(vpar),
                 Str("vpar_reachability_matrix() "
                     "vs test_reachability_matrix_by_adjacency_power()"));

     check_equal(vpar_reachability_matrix(vpar),
                 apply(d->d>=1, vpar_path_lengths_matrix(vpar)),
                 Str("vpar_reachability_matrix() "
                     "vs test_reachability_matrix_by_path_length_matrix()"));
  );
}

{
  my(forest=1);
  for(n=0, test_limit_trees_search - forest,
     my(l=List([]));   \\ distinct reachability matrices, of forests
     forvec(vec=vpar_forvec_from(n,forest),
           my(vpar=vpar_from_upascend(vec),
              m=vpar_reachability_matrix(vpar));
           my(pos=setsearch(l,m,1));
           if(pos, listinsert(l,m,pos)));

     check_equal(#l, vpar_INTERNAL_Bell_number(n),
                 "vpar_reachability_matrix() distinct count is Bell numbers");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_INTERNAL_for_free_automorphisms()");

check_equal(vpar_INTERNAL_for_free_automorphisms([]), []);
check_equal(vpar_INTERNAL_for_free_automorphisms([0]), [0]);

\\ Return 1 if vpar has any bicentral component tree.
test_any_bicentral(vpar) =
{
  my(centres=vpar_centres_of_forest(vpar));
  for(i=1,#centres, if(#centres[i]==2, return(1)));
  0;
}
check_equal(test_any_bicentral([2, 3, 0]), 0);
check_equal(test_any_bicentral([0,1,2,3]), 1);

test_for_free_automorphism_by_types(vpar) =
{
  if(test_any_bicentral(vpar),
     vpar_add_roots_above_centres(vpar);
     ,
     my(centres=vpar_centres_of_forest(vpar));
     vpar_reroot_forest(vpar, if(#centres, concat(centres), [])));
}
check_equal(vpar_add_roots_above_centres([0,3,0]), [4,5,5,0,0]);

test_all_centres(vpar) =
{
  my(vec=vpar_centres_of_forest(vpar));
  if(#vec, concat(vec), vec);
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
       got=vpar_INTERNAL_for_free_automorphisms(vpar));

     check_equal(got, test_for_free_automorphism_by_types(vpar),
     "vpar_INTERNAL_for_free_automorphisms() vs test_for_free_automorphism_by_types()");

     check_equal(got, if(test_any_bicentral(vpar),
                         vpar_add_roots_above_centres(vpar),
                         vpar_reroot_forest(vpar, test_all_centres(vpar))));
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_is_strongly_monotone_forest()");

\\ "outer" and "inner" are 2-element vectors [min,max].
\\ Return 1 if inner is nested within the range of outer.
\\    outer_min <=  inner_min <= inner_max  <= outer_max
test_range_is_nested(outer,inner) =
{
  outer[1] <= outer[2] || error();
  inner[1] <= inner[2] || error();
  outer[1] <= inner[1] && inner[2] <= outer[2];
}

\\ per vpar_is_strongly_monotone_forest() addhelp, its condition is
\\ equivalent to vertex ranges not nested (only part overlap or disjoint)
test_is_strongly_monotone_forest_by_no_nesting(vpar) =
{
  my(v=vpar_component_vertices(vpar));
  v=apply(elem->[vecmin(elem),vecmax(elem)], v);
  for(i=1,#v,
     for(j=1,#v,
        if(i!=j && test_range_is_nested(v[i],v[j]), return(0))));
  1;
}

test_is_strongly_monotone_forest_by_minmax(vpar) =
{
  my(mins=vpar_component_minmax(vpar,min),
     maxs=vpar_component_minmax(vpar,max),
     root_of_vertex=vpar_root_of_vertex_vector(vpar));
  apply(v->root_of_vertex[v], mins) == apply(v->root_of_vertex[v], maxs);
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     \\ print(vpar"  compnums "vpar_compnum_vector(vpar));
     check_equal(vpar_is_strongly_monotone_forest(vpar),
                 test_is_strongly_monotone_forest_by_no_nesting(vpar),
                 "vpar_is_strongly_monotone_forest() vs no nesting");
     check_equal(vpar_is_strongly_monotone_forest(vpar),
                 test_is_strongly_monotone_forest_by_minmax(vpar),
                 "vpar_is_strongly_monotone_forest() vs min,max");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_strongly_monotone_partitions_count()");

{
  check_equal(vpar_INTERNAL_strongly_monotone_transform([1,1,1,1]),
              [[1],
               [2, 1],
               [4, 2, 1],
               [9, 5, 3, 1]]);
}
{
  \\ A092920 = strongly monotone partitions,
  \\           or non-overlapping partitions
  check_equal(vector(#A092920_samples,n,n--;
                     vpar_strongly_monotone_partitions_count(n)),
              A092920_samples,  \\ OFFSET=0
              "vpar_strongly_monotone_partitions_count()");
}


\\-----------------------------------------------------------------------------
progress("vpar_count_strongly_monotone_forests()");

{
  \\ not in OEIS: 3, 14, 97, 917, 11103, 164411, 2880568, 58291264, 1337891607
  my(want=[
       1, 1, 3, 14, 97, 917, 11103, 164411, 2880568, 58291264, 1337891607,
       34340500987, 974704320607, 30312971790320, 1025066867236797
     ]);
  check_equal(vector(#want,n,n--; vpar_count_strongly_monotone_forests(n)),
              want,
              "vpar_count_strongly_monotone_forests()");
}

{
  my(forest=1);
  for(n=0, test_limit_trees_search - forest,
     my(count=0);
     forvec(vec=vpar_forvec_from(n,forest), \\ forests
           my(vpar=vpar_from_upascend(vec));
           count += vpar_is_strongly_monotone_forest(vpar));
     check_equal(count, vpar_count_strongly_monotone_forests(n),
     "vpar_count_strongly_monotone_forests() vs vpar_is_strongly_monotone_forest()"));
}


\\-----------------------------------------------------------------------------
progress("vpar_seq_lexmin()");
\\ vpar_to_lexmin()

check_equal(vpar_seq_lexmin([]), Vecsmall([]));
check_equal(vpar_seq_lexmin([0]), Vecsmall([1]));
check_equal(vpar_seq_lexmin([0,0,0,0]), Vecsmall([1,2,3,4]));
check_equal(vpar_seq_lexmin([0,1,1,1,1]), Vecsmall([1,2,3,4,5]));
check_equal(vpar_seq_lexmin([0,0,1,1,1,2,2]), Vecsmall([1,2,3,4,5,6,7]));
check_equal(vpar_seq_lexmin([0,0,2,2,1,1,1]), Vecsmall([1,2, 5,6,7, 3,4]));
check_equal(vpar_seq_lexmin([0,0,2,2,2,1,1,1]), Vecsmall([1,2, 6,7,8, 3,4,5]));

test_relabel_premax_rows(vpar) =
{
  vpar=vpar_to_unlabelled(vpar);
  vpar_relabel_seq(vpar,vpar_seq_rows(vpar));
}

{
  \\          5
  \\       /     \
  \\      1        4
  \\     / \      / \
  \\    2  10    7   8
  \\        |    |   |
  \\        6    9   3

  \\ 3 = [0, -1, [[1, 0, []]],  3]
  \\ 1 = [0, 0, [],             1]
  \\ 2 = [1, 0, [],             2]
  \\
  \\       1  2  3  4  5   6  7  8  9  10
  my(vpar=[5, 1, 8, 5, 0, 10, 4, 4, 7,  1],
     want_seq=Vecsmall([5, 4,1, 7,8,10,2, 9,3,6]),
     want_vpar=[0, 1,1, 2,2, 3,3, 4,5,6]);
  vpar_validate(vpar);
  check_equal(test_is_perm(want_seq), 1);
  check_equal(vpar_relabel_seq(vpar,want_seq), want_vpar);

  check_equal(vpar_seq_lexmin(vpar), want_seq,
             "vpar_seq_lexmin() 10");
}
{
  \\   1    3
  \\        |
  \\        2

  \\ 3 = [0, -1, [[1, 0, []]],  3]
  \\ 1 = [0, 0, [],             1]
  \\ 2 = [1, 0, [],             2]
  \\
  my(vpar=[0, 3, 0],
     want_seq=Vecsmall([3,1,2]),
     want_perm=want_seq^-1,
     want_vpar=[0,0,1]);
  vpar_validate(vpar);
  check_equal(vpar_relabel_perm(vpar,want_perm), want_vpar);
  check_equal(want_perm, Vecsmall([2, 3, 1]));

  check_equal(vpar_seq_lexmin(vpar), want_seq,
             "vpar_seq_lexmin() forest 3");
}
{
  \\            12
  \\         / |  | \         premax takes subtree 2 first
  \\        2  4  8  9        lexmin takes subtree 4 first
  \\       /   |\     \
  \\      7    3 13    1
  \\     /|    |  |
  \\    5 6   10 11
  my(vpar=[9, 12, 4, 12, 7, 7, 2, 12, 12, 3, 13, 0, 4]);
  check_equal(vpar_seq_premax(vpar),
              Vecsmall([12,2,7,5,6, 4,3,10,13,11, 9,1, 8]));
  check_equal(vpar_to_unlabelled(vpar),
              [0,1,2,3,3, 1,6,7,6,9, 1,11, 1],
              "vpar_to_unlabelled() 12");
  check_equal(vpar_seq_lexmin(vpar),
              Vecsmall([12, 4,2,9,8, 3,13,7,1, 10,11, 5,6]),
              "vpar_seq_lexmin() 12");
  check_equal(vpar_to_lexmin(vpar),
              [0,1,1,1,1, 2,2,3,4, 6,7,8,8]);
  check_equal(test_relabel_premax_rows(vpar),
              [0,1,1,1,1, 2,3,3,4, 6,6,7,8]);  \\ is bigger at 3
}
{
  \\    4     6    3
  \\   / \    |
  \\  1   5   2

  \\ 4 = [0, -2, [[1, 0, []], [1, 0, []]], 4]
  \\ 6 = [0, -1, [[1, 0, []]], 6]
  \\ 3 = [0, 0, [], 3]
  \\ 1 = [1, 0, [], 1]
  \\ 5 = [1, 0, [], 5]
  \\ 2 = [1, 0, [], 2]
  \\
  my(vpar=[4, 6, 0, 0, 4, 0],
     want_seq=Vecsmall([4,6,3, 1,5,2]),
     want_perm=want_seq^-1,
     want_vpar=[0,0,0,1,1,2]);
  vpar_validate(vpar);
  check_equal(vpar_relabel_perm(vpar,want_perm), want_vpar);
  check_equal(want_perm, Vecsmall([4, 6, 3, 1, 5, 2]));

  check_equal(vpar_seq_lexmin(vpar), want_seq,
             "vpar_seq_lexmin() forest 6");
}

{
  \\      1           2     3
  \\     / \         / \    |
  \\   4     5      6   7   8
  \\  / \   / \    / \  |   |
  \\ 9 10  11 12  13 14 15  16
  \\
  my(vpar=[0,0,0, 1,1,2,2,3, 4,4,5,5, 6,6,7, 8]);
  vpar_validate(vpar);

  my(got_seq=vpar_seq_lexmin(vpar));
  check_equal(got_seq,
              vpar_INTERNAL_identity_perm(#vpar),
              "vpar_seq_lexmin() 16 is identity");

  my(got_seq=vpar_seq_lexmin(vpar));
  check_equal(got_seq,
              vpar_INTERNAL_identity_perm(#vpar),
             "vpar_seq_lexmin() 16");
}

test_minmax_perm_by_search(vpar,func) =
{
  my(best_perm=numtoperm(#vpar,0),
     best_vpar=vpar,
     want_lex=func(1,-1));
  for(i=1,(#vpar)!-1,
     my(this_perm=numtoperm(#vpar,i),
        this_vpar = vpar_relabel_perm(vpar,this_perm));
     if(lex(this_vpar,best_vpar) == want_lex,
        best_perm = this_perm;
        best_vpar = this_vpar));
  Vecsmall(best_perm);
}
test_to_lexminmax_by_search(vpar,func) = \
  vpar_relabel_perm(vpar, test_minmax_perm_by_search(vpar,func));

check_equal(vecsort([0,1,2],,1), Vecsmall([1,2,3]));
check_equal(vpar_relabel_perm([0,1,2], [1,2,3]), [0,1,2]);
{
  my(vpar=[2,3,0],
     perm=vecsort(vpar,,1));  \\ for min
  check_equal(perm, Vecsmall([3,1,2]));
  check_equal(vpar_relabel_perm(vpar, perm),    [2,0,1]);
  check_equal(vpar_relabel_perm(vpar, perm^-1), [0,3,1]);
}


{
  \\      6
  \\     / \
  \\    2   5
  \\   / \   \
  \\  1   4   3
  \\
  \\
  my(vpar=[2, 6, 5, 2, 6, 0],
     want_seq =Vecsmall([6, 2,5, 1,4,3]));
  check_equal(test_is_perm(want_seq), 1);

  my(want_perm=want_seq^-1,
     want_vpar=[0, 1, 1, 2, 2, 3]);
  vpar_validate(vpar);
  check_equal(test_minmax_perm_by_search(vpar,min), want_perm,
             "test_minmax_perm_by_search() 6");
  check_equal(vpar_relabel_perm(vpar,want_perm), want_vpar,
             "want_perm 6");

  my(got_seq = vpar_seq_lexmin(vpar));
  check_equal(got_seq, want_seq,
             "vpar_seq_lexmin() 6");
  check_equal(test_to_lexminmax_by_search(vpar,min), want_vpar);

  my(got_vpar = vpar_to_lexmin(vpar));
  check_equal(got_vpar, want_vpar, "vpar_to_lexmin() 5");
}
{
  \\  1   3
  \\  |   |
  \\  2   5
  \\      |
  \\      4
  \\
  \\
  my(vpar=[0, 1, 0, 5, 3],
     want_perm=Vecsmall([2,4,1,5,3]),
     want_seq=Vecsmall([3,1, 5,2, 4]),
     want_vpar=[0, 0, 1, 2, 3]);
  vpar_validate(vpar);
  check_equal(test_minmax_perm_by_search(vpar,min), want_perm,
             "test_minmax_perm_by_search() 5");
  check_equal(vpar_relabel_perm(vpar,want_seq^-1), want_vpar,
             "want_seq 5");

  my(got_seq = vpar_seq_lexmin(vpar));
  check_equal(got_seq, want_seq, "vpar_seq_lexmin() 5");
  check_equal(test_to_lexminmax_by_search(vpar,min), want_vpar);

  my(got_vpar = vpar_to_lexmin(vpar));
  check_equal(got_vpar, want_vpar, "vpar_to_lexmin() 5");
}
{
  \\      1       2
  \\     / \      |
  \\    3   4     5
  \\   / \       / \
  \\  6   7     8   9
  \\
  \\
  my(vpar=[0,0, 1,1,2, 3,3,5,5]);
  vpar_validate(vpar);
  check_equal(test_minmax_perm_by_search(vpar,min),
              test_identity_perm(#vpar),
              "test_minmax_perm_by_search() 9");
  check_equal(vpar_seq_lexmin(vpar),
              test_identity_perm(#vpar),
              "vpar_seq_lexmin() 9");
  check_equal(vpar_to_lexmin(vpar_relabel_random(vpar)),
              vpar,
              "vpar_seq_lexmin() random 9");
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got_seq = vpar_seq_lexmin(vpar),
        got_vpar = vpar_to_lexmin(vpar));
     \\ print(vpar);

     check_equal(vpar_seq_is_downwards(vpar,got_seq), 1,
                 "vpar_seq_lexmin() is vpar_seq_is_downwards()");
     check_equal(got_vpar, vpar_relabel_seq(vpar,got_seq),
                 "vpar_to_lexmin() vs vpar_seq_lexmin()");

     if(#vpar <= test_limit_perms_search,
        check_equal(vpar_to_lexmin(vpar),
                    test_to_lexminmax_by_search(vpar,min),
                    Str("vpar_to_lexmin() vs test_to_lexminmax_by_search()"))));
}


\\-----------------------------------------------------------------------------
progress("vpar_free_isomorphic_perm()");

{
  for(i=1,#test_sample_forests,
     my(vpar1=test_sample_forests[i],
        rand_perm=vpar_INTERNAL_random_perm(#vpar1));
     check_equal(test_is_perm(rand_perm), 1);

     my(vpar2=vpar_reroot_random(vpar_relabel_perm(vpar1,rand_perm)));

     \\ print("vpar1 "vpar1"  rand_perm "Vec(rand_perm));
     \\ print("vpar2 "vpar2);

     my(got_perm=vpar_free_isomorphic_perm(vpar1,vpar2));
     check_equal(got_perm!='none, 1,
                 "vpar_free_isomorphic_perm() random relabel, found");
     check_equal(test_is_perm(got_perm), 1);

     my(vpar1_relabelled=vpar_relabel_perm(vpar1,got_perm));
     my(vpar1_relabelled_and_rerooted=vpar_reroot_forest(vpar1_relabelled,
                                                         vpar_roots(vpar2)));
     check_equal(vpar1_relabelled_and_rerooted, vpar2,
                 Str("vpar_free_isomorphic_perm() random relabel,"
                     " equal after perm+relabel vpar2"));

     my(funcs=[min,max]);
     for(f=1,#funcs,
        my(func=funcs[f]);
        check_equal(vpar_reroot_minmax(vpar1_relabelled,func),
                    vpar_reroot_minmax(vpar2,           func),
                    Str("vpar_free_isomorphic_perm() random relabel,"
                        " equal after perm+relabel minmax")));


     my(num_automorphisms=vpar_num_free_automorphisms(vpar1));
     my(automorphisms1=if(num_automorphisms<=test_limit_sets,
                          vpar_free_automorphisms(vpar1),
                          vpar_free_automorphism_generators(vpar1)));

     my(vpar2_rerooted_min=vpar_reroot_minmax(vpar2,min));
     for(i=1,#automorphisms1,
        my(a_perm = got_perm * automorphisms1[i]);
        check_equal(vpar_reroot_minmax(vpar_relabel_perm(vpar1,a_perm),min),
                    vpar2_rerooted_min,
                    "vpar_free_isomorphic_perm() times automorphism1"));

     check_equal(num_automorphisms,vpar_num_free_automorphisms(vpar2));
     my(automorphisms2=if(num_automorphisms<=test_limit_sets,
                          vpar_free_automorphisms(vpar2),
                          vpar_free_automorphism_generators(vpar2)));
     for(i=1,#automorphisms2,
        my(a_perm = automorphisms2[i] * got_perm);
        check_equal(vpar_reroot_minmax(vpar_relabel_perm(vpar1,a_perm),min),
                    vpar2_rerooted_min,
                    "vpar_free_isomorphic_perm() times automorphism2"));

  );
}


\\-----------------------------------------------------------------------------
progress("vpar_automorphisms()");

check_equal(vpar_automorphisms([]), [ Vecsmall([]) ], \
            "vpar_automorphisms() empty");
check_equal(vpar_automorphisms([0]), [ Vecsmall([1]) ], \
            "vpar_automorphisms() singleton");
check_equal(vpar_automorphisms([0,0]), [ Vecsmall([1,2]), Vecsmall([2,1]) ], \
            "vpar_automorphisms() two singletons");
check_equal(vpar_automorphisms([0,1,1]), [ Vecsmall([1,2,3]), Vecsmall([1,3,2]) ], \
            "vpar_automorphisms() star-3");

test_automorphisms_by_generators(vpar) = \
  test_perms_generators_to_group(vpar_automorphism_generators(vpar), #vpar);

{
  \\ gens order 2, 3
  \\ total num 6
  my(vpar=[0,1,1,1]);
  check_equal(vpar_num_automorphisms(vpar), 6);
  check_equal(vecsort(vpar_automorphisms(vpar)),
              vecsort([Vecsmall([1, 2,3,4]), Vecsmall([1, 2,4,3]),
                       Vecsmall([1, 3,2,4]), Vecsmall([1, 3,4,2]),
                       Vecsmall([1, 4,2,3]), Vecsmall([1, 4,3,2]) ]),
              "vpar_automorphisms() star-4");
}
{
  \\ gens order 2, 4
  \\ total num 4! = 24
  my(vpar=[0,1,1,1,1]);
  check_equal(vpar_num_automorphisms(vpar), 24);
  my(got=vpar_automorphisms(vpar));
  check_equal(#got, 4!);
  check_equal(vecsort(got),
              test_automorphisms_by_generators(vpar),
              "vpar_automorphisms() vs test_automorphisms_by_generators()");
}
{
  \\ Complete Trees
  \\       1               1
  \\      / \             /|\
  \\     2   3           2 3 4
  \\    / \ / \
  \\   4  5 6  7
  \\
  \\ GP-Test  2^2-1 == 3
  \\ GP-Test  my(h=4); 2^(2^h-1) == 32768
  \\ #vpar_make_complete_tree(5,4) == 781
  \\
  for(c=2,5,
     for(h=1,4,
        my(vpar=vpar_make_complete_tree(c,h),
           want_num_automorphisms = (c!) ^ (#vpar-vpar_num_childless(vpar)),
           want_num_generators    = if(c<=2,1,2) * h);

        check_equal(vpar_num_automorphisms(vpar), want_num_automorphisms,
                    Str("vpar_num_automorphisms() complete tree c="c" h="h));

        if(want_num_automorphisms <= test_limit_sets,
           my(got=vpar_automorphisms(vpar));
           check_equal(#got, want_num_automorphisms,
                       "vpar_automorphisms() complete binary tree h=2, count");
           check_equal(vecsort(got),
                       test_automorphisms_by_generators(vpar),
             "vpar_automorphisms() vs test_automorphisms_by_generators()"))));
}


{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        num_automorphisms=vpar_num_automorphisms(vpar));
     if(num_automorphisms > test_limit_sets, next);

     my(got=vpar_automorphisms(vpar));
     check_equal(#got, num_automorphisms);
     check_equal(#got, #Set(got),
                 "vpar_automorphisms() no duplicates");
     for(i=1,#got,
        check_equal(#got[i], #vpar);
        check_equal(test_is_perm(got[i]), 1));
     check_equal(test_perms_is_group(got), 1,
                 "vpar_automorphisms() vs test_perms_is_group()");

     got=vecsort(got);
     check_equal(got, test_automorphisms_by_generators(vpar),
                 "vpar_automorphisms() vs test_automorphisms_by_generators()");
     if(#vpar <= test_limit_perms_search,
        check_equal(got, test_automorphisms_by_search(vpar),
                    "vpar_automorphisms() vs test_automorphisms_by_search()"));

    \\ FIXME: No, generators not currently minimum.
    \\ check_equal(vpar_num_automorphism_generators(vpar),
    \\             test_perms_minimum_num_generators(got),
    \\             "vpar_num_automorphism_generators() vs test_perms_minimum_num_generators()");

  );
}


\\-----------------------------------------------------------------------------
progress("vpar_seq_childless_paths()");
\\ vpar_from_childless_paths()

check_equal(vpar_seq_childless_paths([]), Vecsmall([]));
check_equal(vpar_seq_childless_paths([],1), Vecsmall([]));

test_seq_childless_paths_by_lists(vpar,down=0) =
{
  my(seen=vectorsmall(#vpar),
     childless=vpar_childless_vertices(vpar),
     retlist=List([]));
  for(i=1,#childless,
     my(v=childless[i],l=List([]));
     while(1,
           listput(l,v); seen[v]=1;
           v=vpar[v];
           if(!v || seen[v], break()));
     l=if(down,Vecrev(l),Vec(l));
     listput(retlist,l));

  if(#retlist, Vecsmall(concat(Vec(retlist))), Vecsmall([]));
}
{
  \\        10
  \\       / | \
  \\     3   4   9
  \\    / \     / \
  \\   1   2   5   8
  \\              / \
  \\             6   7
  \\
  \\       1 2  3  4 5 6 7 8  9 10
  my(vpar=[3,3,10,10,9,8,8,9,10, 0]);     \\ Knuth 2.3.4.4
  check_equal(#vpar,10);
  vpar_validate(vpar);
  my(vec=vecextract(vpar,vpar_seq_childless_paths(vpar,1)));
  \\ print("vec "vec);
  check_equal(vec, [0,10,3,3,10, 10,9,9,8,8]);
  check_equal(vpar_from_childless_paths(vec,1), vpar);

  check_equal(vpar_seq_childless_paths(vpar,1),
              Vecsmall([10,3,1,2,4,9,5,8,6,7]));
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     for(down=0,1,
        check_equal(vpar_seq_childless_paths(vpar,down),
                    test_seq_childless_paths_by_lists(vpar,down),
                    "vpar_seq_childless_paths() vs lists")));
}


\\-----------------------------------------------------------------------------
progress("vpar_from_upascend()");

check_equal(vpar_from_upascend([]),[]);
check_equal(vpar_from_upascend([0]),[0]);
check_equal(vpar_from_upascend([1,0]),[0,1]);
check_equal(vpar_from_upascend([2,0]),[2,0]);
check_equal(vpar_from_upascend([1,1,0]),[0,1,1]);
check_equal(vpar_from_upascend([2,1,0]),[0,1,2]);
check_equal(vpar_from_upascend([3,1,0]),[0,3,1]);
check_equal(vpar_from_upascend([1,2,0]),[2,0,1]);
check_equal(vpar_from_upascend([1,3,0]),[3,1,0], "upascend");
{
  my(vpar=[]);
  check_equal(vecextract(vpar,vpar_seq_upascend(vpar)),[]);
}
{ my(vpar=[0]);
  check_equal(vecextract(vpar,vpar_seq_upascend(vpar)),[0]);
}


{
  \\ Cho, Kim, Seo, Shin, "Coloured Pruufer Codes For K-Edge Coloured
  \\ Trees", Electronic Journal of Combinatorics, volume 11, #N10, 2004.
  \\ http://mathsci.kaist.ac.kr/~dskim/
  \\ http://mathsci.kaist.ac.kr/~dskim/papers/Tree.pdf
  \\ Example
  \\         1
  \\      /  |  \
  \\     3   6    7
  \\    / \  |
  \\   2  4  5
  \\ Think they work from high to low taking and removing childless.
  \\
  my(vpar=[0,3,1,3,6,1,1],
     rev=(n)->if(n,8-n));
  vpar_validate(vpar);
  my(vpar_rev = vpar_relabel_reverse(vpar));

  my(got_seq_rev=vpar_seq_upascend(vpar_rev),
     got_seq = apply(rev, Vec(got_seq_rev)));
  check_equal(got_seq, [7,5,6,4,2,3,1],
              "vpar_seq_upascend() 7");

  my(got_vec=vecextract(vpar,got_seq));
  check_equal(got_vec, [1,6,1,3,3,1,0],
              "vpar_seq_upascend() 7 vec");
}
{
  \\  3 <- 4 <- 2 <- 1
  \\
  my(vpar=[2, 4, 0, 3]);
  vpar_validate(vpar);
  for(ahead=0,1,
     my(got_seq=vpar_seq_upascend(vpar,ahead));
     check_equal(got_seq, Vecsmall([1,2,4,3]),
                 "vpar_seq_upascend() 4 path");
     my(got_vec=vecextract(vpar,got_seq));
     check_equal(got_vec, [2,4,3,0],
                 "vpar_seq_upascend() 4 path extract");
     check_equal(vpar_from_upascend(got_vec,ahead), vpar,
                 "vpar_from_upascend() 4"));
}
{
  \\         2
  \\         |
  \\         3
  \\        / \
  \\       1   4
  \\
  my(vpar=[3, 0, 2, 3]);
  vpar_validate(vpar);
  for(ahead=0,1,
     my(got_seq=vpar_seq_upascend(vpar,ahead));
     check_equal(got_seq, Vecsmall([1,4,3,2]),
                 "vpar_seq_upascend() 4");
     my(got_vec=vecextract(vpar,got_seq));
     check_equal(got_vec, [3,3,2,0],
                 "vpar_seq_upascend() 4 extract");
     check_equal(vpar_from_upascend(got_vec), vpar,
                 "vpar_from_upascend() 4"));
}
{
  \\         1
  \\         |
  \\         6
  \\         |
  \\        10
  \\         |
  \\         4
  \\        / \
  \\       3   9
  \\      /     \
  \\     8       7
  \\    /         \
  \\   2           5
  \\
  my(vpar=[0,8,4,10,7,1,9,3,4,6]);
  vpar_validate(vpar);
  my(got_seq=vpar_seq_upascend(vpar,1));
  check_equal(got_seq, Vecsmall([2,8,3,5,7,9,4,10,6,1]),
             "vpar_seq_upascend() ahead=1, 10");
  my(got_vec=vecextract(vpar,got_seq));
  check_equal(got_vec, [8,3,4,7,9,4,10,6,1,0],
              "vpar_seq_upascend() ahead=1, 10 extract");

  check_equal(vpar_from_upascend(got_vec,1), vpar,
              "vpar_from_upascend() ahead=1, 10");
}

{
  \\ Caminiti slides, Prufer encode rooted
  \\       4
  \\      / \
  \\     1   3
  \\    /   / \
  \\   2   5   6
  \\
  my(vpar=[4,1,4,0,3,3]);
  vpar_validate(vpar);
  check_equal(vecextract(vpar,vpar_seq_upascend(vpar)),
              [1,4,3,3,4, 0]);
}

test_smallest_not_in_list(v) =
{
  my(s=Set(Vec(v)));
  for(i=1,vecmax(s)+1, if(!setsearch(s,i),return(i)));
  error("oops");
}
test_from_canonical_by_not_in_lists(can) =
{
  my(vpar=vector(#can),
     V=vector(#can));
  for(i=1,#can,
     my(v=test_smallest_not_in_list(concat(can[i..#can],V[1..i-1])));
     V[i]=v;
     vpar[v]=can[i]);
  vpar;
}

check_equal(vpar_forvec_from(3), [[1,3], [1,3], [0,0]]); \\ tree

test_from_upascend_by_num_children(vpar) =
{
  \\ Knuth gives procedure: find V1 which is smallest not in vec[1..#vec],
  \\ and it is edge V1 up to vec[1] to store in vpar, ie. vpar[V1]=vec[1].
  \\ Then V2 is smallest not in V1,vec[2..#vec] stored to vpar[V2]=vec[2],
  \\ etc.
  \\
  \\ The approach here is as follows.  vec[] is parent vertex numbers and
  \\ num_children[] is constructed as count of how many children each parent
  \\ has.  Loop i = 1 to n looks for num_children[V1]==0.  This is V1
  \\ smallest not in vec[].  Its parent is p=vec[1] and store vpar[V1]=p.
  \\ So p=vec[1] has been handled and num_children[p] is decreased, leaving
  \\ num_children[] as for vec[2..#vec].  If p<V1 then V2=p is the next
  \\ smallest not in vec[2..#vec], since only num_children[p] has changed.
  \\ This chain up by p continues while p<i, where i is the start of this
  \\ chain of V.  The chain stops when p>i because it's possible there's
  \\ somewhere num_children[i+1..p-1] == 0.  Resuming the i loop searches
  \\ for them.
  \\
  \\ For ahead=1 the variation is to continue to chain up while
  \\ num_children[p]==0, so don't stop at p<i.
  \\
  \\ The net result is 1 pass to make num_children[], 1 pass for i chain
  \\ starts, and up to 1 further pass for the chains.  If all p>i then no
  \\ chaining back.  The most chaining is when the last i=n is the first
  \\ num_children[i]==0 and it chains back through all preceding n-1 vertices.
  \\
  \\ This is much the same approach as
  \\
  \\     Xiaodong Wang, Lei Wang, Yingjie Wu, "An Optimal Algorithm for Prufer
  \\     Codes", Journal of Software Engineering & Applications, 2009, 2:
  \\     pages 111-115.  doi:10.4236/jsea.2009.22016
  \\     https://www.scirp.org/journal/PaperDownload.aspx?paperID=532
  \\
  \\ They loop over vec[] and run the vertex forward as necessary to find a
  \\ childless leaf, whereas here loop over the vertex with chains up and
  \\ take from vec[].

  my(num_children=vpar_INTERNAL_num_children_vecsmall(vec),
     vpar=vector(#vec),
     upto=0);  \\ position in vec[]
  for(i=1,#vec,
     \\ print("i="i" num_children "num_children[i]);
     my(v=i);
     while((ahead || v<=i) && num_children[v]==0,
           my(p=vec[upto++]);  \\ pre-increment
           \\ print("  vpar store v="v" p="p);
           num_children[v]=-1; \\ no revisit "ahead"
           (vpar[v]=p) || break();
           num_children[p]--;
           v=p);
  );
  vpar;
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     check_equal(vpar_seq_upascend(vpar),
                 test_seq_upascend_by_sort(vpar),
                 "vpar_seq_upascend() vs by_sort");
     check_equal(vpar_seq_upascend(vpar,1),
                 test_seq_upascend_ahead_by_sort(vpar),
                 "vpar_seq_upascend() ahead=1 vs by_sort");

     for(ahead=0,1,
        my(seq=vpar_seq_upascend(vpar,ahead),
           vec=vecextract(vpar,seq));
        \\ print("vpar="vpar" ahead="ahead);
        \\ print("  seq "Vec(seq));
        \\ print("  vec "vec);
        check_equal(vpar_from_upascend(vec,ahead), vpar,
                    Str("vpar_from_upascend() back to vpar, ahead="ahead)));
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_indpoly()");

check_equal(vpar_indpoly([]),    Pol(1), "empty");
check_equal(type(vpar_indpoly([])), "t_POL", "empty");
check_equal(vpar_indpoly([0]),   x+1,    "single");
check_equal(vpar_indpoly([0,1]), 2*x+1,  "path-2");
check_equal(vpar_indpoly([0,1,1,1,1]), x^4 + 4*x^3 + 6*x^2 + 5*x + 1, "star-5");
check_equal(vpar_indpoly([0,1,2,3,4]), x^3         + 6*x^2 + 5*x + 1, "path-5");

test_indpoly_by_flags(vpar) =
{
  my(poly=Pol(0));
  forvec(flags=vector(#vpar,i,[0,1]),
        if(vpar_is_indset_flags(vpar,flags),
           poly += x^vecsum(flags)));
  poly;
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        poly=vpar_indpoly(vpar));
     check_equal(type(poly), "t_POL");

     \\ print(vpar);
     \\ print(poly);

     check_equal(polcoeff(poly,0), 1);
     check_equal(polcoeff(poly,1), #vpar);
     check_equal(polcoeff(poly,2), binomial(#vpar,2) - vpar_num_edges(vpar));
     check_equal(polcoeff(poly,#vpar), vpar_num_edges(vpar)==0);
     check_equal(poldegree(poly), vpar_indnum(vpar));
     check_equal(pollead(poly), vpar_indnum_and_count(vpar)[2]);

     check_equal(simplify(subst(poly,'x,1)), vpar_indsets_count(vpar),
                 "vpar_indpoly() subst to vpar_indsets_count() per addhelp");
     check_equal(vpar_indsets_count(vpar), vecsum(Vec(poly)));

     if(#vpar <= test_limit_flags_search,
        check_equal(poly, test_indpoly_by_flags(vpar)));

     my(component_vpar=vpar_component_trees(vpar));
     check_equal(poly,
                 Pol(prod(i=1,#component_vpar, vpar_indpoly(component_vpar[i]))),
                 "vpar_indpoly() product of components");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_perfect_dompoly()");

check_equal(vpar_perfect_dompoly([]),    Pol(1), \
            "vpar_perfect_dompoly() empty");
check_equal(type(vpar_perfect_dompoly([])), "t_POL", "empty");

check_equal(vpar_perfect_dompoly([0]),          x, \
            "vpar_perfect_dompoly() singleton");
check_equal(vpar_perfect_dompoly([0,1]),        x^2 + 2*x, \
            "vpar_perfect_dompoly() path-2");
check_equal(vpar_perfect_dompoly([0,1,2]),      x^3 + 2*x^2 + x, \
            "vpar_perfect_dompoly() path-3");
check_equal(vpar_perfect_dompoly([0,1,1,1]),  \
      x^4 + binomial(3,2)*x^3 + binomial(3,1)*x^2 + x, \
      "vpar_perfect_dompoly() star-4");
check_equal(vpar_perfect_dompoly([0,1,1,1,1]),  \
      x^5 + binomial(4,3)*x^4 + binomial(4,2)*x^3 + binomial(4,1)*x^2 + x, \
      "vpar_perfect_dompoly() star-5");

test_perfect_dompoly_by_flags(vpar) =
{
  my(poly=Pol(0));
  forvec(flags=vector(#vpar,i,[0,1]),
        if(vpar_is_perfect_domset_flags(vpar,flags),
           poly += x^vecsum(flags)));
  poly;
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        poly=vpar_perfect_dompoly(vpar));
     check_equal(type(poly), "t_POL");

     if(#vpar <= test_limit_flags_search,
        check_equal(poly,
                    test_perfect_dompoly_by_flags(vpar),
                    Str("vpar_perfect_dompoly() "
                        "vs test_perfect_dompoly_by_flags()")));

     my(component_vpar=vpar_component_trees(vpar));
     check_equal(poly,
                 Pol(prod(i=1,#component_vpar, vpar_perfect_dompoly(component_vpar[i]))),
                 "vpar_perfect_dompoly() product of components");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_perfect_domnum()");
\\ vpar_perfect_domnum_and_count()

check_equal(vpar_perfect_domnum_and_count([]), [0,1]);
check_equal(vpar_perfect_domnum_and_count([0]), [1,1]);
check_equal(vpar_perfect_domnum_and_count(vpar_make_star(3)), [1,1]);

test_perfect_domnum_and_count_by_flags(vpar) =
{
  if(#vpar==0,return([0,1]));  \\ empty set only
  my(perfect_domnum=#vpar,count=0);
  forvec(flags=vector(#vpar,i,[0,1]),
        if(vpar_is_perfect_domset_flags(vpar,flags),
           my(size=vecsum(flags));
           if(size<perfect_domnum,  perfect_domnum=size;count=1, \\ new low
              size==perfect_domnum, count++)));                  \\ equal low
  [perfect_domnum,count];
}
check_equal(test_perfect_domnum_and_count_by_flags(vpar_make_path(2)), [1,2]);
check_equal(test_perfect_domnum_and_count_by_flags(vpar_make_path(3)), [1,1]);
check_equal(test_perfect_domnum_and_count_by_flags(vpar_make_path(4)), [2,2]);
check_equal(test_perfect_domnum_and_count_by_flags(vpar_make_star(3)), [1,1]);

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        by_flags=test_perfect_domnum_and_count_by_flags(vpar));
     \\ print(vpar);
     check_equal(vpar_perfect_domnum_and_count(vpar), by_flags,
                 Str("vpar_perfect_domnum_and_count() "
                     "vs test_perfect_domnum_and_count_by_flags()"));

     check_equal(vpar_perfect_domnum(vpar), by_flags[1],
                 Str("vpar_perfect_domnum() "
                     "vs test_perfect_domnum_and_count_by_flags()"));
  );
}

{
  check_equal(vector(8,h,h--;
                     vpar_perfect_domnum(vpar_make_complete_tree(2,h))),
              vector(8,h,h--; if(h==0,1, 2^h-1)),
              "vpar_perfect_domnum() of complete binary tree");
}

{
  check_equal(vector(10,n,n--; vpar_perfect_domnum(vpar_make_star(n))),
              vector(10,n,n--; if(n==0,0, 1)),
              "vpar_perfect_domnum() of star");
  check_equal(vector(10,n,n--; vpar_perfect_domnum
                                (vpar_subdivide(vpar_make_star(n)))),
              vector(10,n,n--; if(n==0,0, max(1,n-1))),
              "vpar_perfect_domnum() of subdivided star");
}
{
  check_equal(vector(20,n,n--; vpar_perfect_domnum(vpar_make_binomial_tree(n))),
              vector(20,n,n--; if(n==1,1, floor(n/2))),
              "vpar_perfect_domnum() of binomial tree n");
}
{
  check_equal(vector(7,k,k--;
                   vpar_perfect_domnum(vpar_make_binomial_tree(2^k))),
              vector(7,k,k--; if(k==0,1, 2^(k-1))),
              "vpar_perfect_domnum() of binomial tree 2^k");
}


\\-----------------------------------------------------------------------------
progress("vpar_is_root_minmax()");

check_equal(vpar_is_root_minmax([],min), 1, "vpar_is_root_minmax() empty");
check_equal(vpar_is_root_minmax([],max), 1);
check_equal(vpar_is_root_minmax([0],min), 1, "vpar_is_root_minmax() singleton");
check_equal(vpar_is_root_minmax([0],max), 1);

\\ 1  3
\\    |
\\    2
check_equal(vpar_is_root_minmax([0,3,0],max), 1);
check_equal(vpar_is_root_minmax([0,3,0],min), 0, "vpar_is_root_minmax() 3 min");

test_is_root_minmax_by_descendants(vpar,func) =
{
  my(minmax_descendants=vpar_minmax_descendants_vector(vpar,func));
  for(v=1,#vpar,
     if(vpar[v]==0 && minmax_descendants[v]!=v, return(0)));
  1;
}
test_is_root_minmax_by_root_of_vertex_vector(vpar,func) =
{
  my(roots=vpar_root_of_vertex_vector(vpar));
  for(v=1,#vpar,
     if(func(v,roots[v])!=roots[v], return(0)));
  1;
}
{
  my(funcs=[min,max]);
  for(f=1,#funcs,
     my(func=funcs[f]);
     for(i=1,#test_sample_forests,
        my(vpar=test_sample_forests[i]);

        check_equal(vpar_is_root_minmax(vpar,func),
                    test_is_root_minmax_by_descendants(vpar,func),
        "vpar_is_root_minmax() vs test_is_root_minmax_by_descendants()");

        check_equal(vpar_is_root_minmax(vpar,func),
                    test_is_root_minmax_by_root_of_vertex_vector(vpar,func),
  "vpar_is_root_minmax() vs test_is_root_minmax_by_root_of_vertex_vector()")));
}


\\-----------------------------------------------------------------------------
progress("vpar_height_num_components()");

check_equal(vpar_height_num_components([]),0);
check_equal(vpar_num_roots([]),0);

check_equal(vpar_height_num_components([0,0,0]),3, \
            "vpar_height_num_components() singletons");
check_equal(vpar_height_num_components([0,0,0,1]),1, \
            "vpar_height_num_components() forest");
check_equal(vpar_height_num_components([0,0,0,1,1,2]),2, \
            "vpar_height_num_components() forest");

test_height_num_components_by_roots(vpar) =
{
  my(roots=vpar_roots(vpar));
  roots=apply(v->vpar_subtree_height(vpar,v), roots);
  my(height=if(#roots,vecmax(roots)));
  sum(i=1,#roots, roots[i]==height);
}
test_height_num_components_by_depths(vpar) =
{
  my(components=vpar_compnum_vector(vpar),
     depths=vpar_depths_vector(vpar),
     height=if(#vpar,vecmax(depths)),
     seen=vectorsmall(if(#vpar,vecmax(components))));
  sum(v=1,#vpar,
      if(depths[v]==height,
         seen[components[v]] = 1));
  hammingweight(seen);
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got=vpar_height_num_components(vpar));
     check_equal(got, test_height_num_components_by_roots(vpar));
     check_equal(got, test_height_num_components_by_depths(vpar));
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_is_component_heights()");

\\ no components is n=0 empty only
check_equal(vpar_is_component_heights([],0), 1, \
            "vpar_is_component_heights() empty");
check_equal(vpar_is_component_heights([],1), 0);
check_equal(vpar_is_component_heights([],-1), 0);
check_equal(vpar_is_component_heights([],-100), 0);
check_equal(vpar_is_component_heights([],100), 0);

\\ single component height=0 is n=1 singleton only
check_equal(vpar_is_component_heights([0],1), 1, \
            "vpar_is_component_heights() singleton");
check_equal(vpar_is_component_heights([0],0), 0);
check_equal(vpar_is_component_heights([0],2), 0);
check_equal(vpar_is_component_heights([0],-1), 0);

\\ size < 0 no such
check_equal(vpar_is_component_heights([-1],0), 0, \
            "vpar_is_component_heights() size<0");
check_equal(vpar_is_component_heights([1,-1],1), 0);

check_equal(vpar_is_component_heights([0,0],-1), 0);
check_equal(vpar_is_component_heights([0,0],1), 0);
check_equal(vpar_is_component_heights([0,0],2), 1);
check_equal(vpar_is_component_heights([0,0],3), 0);

check_equal(vpar_is_component_heights([1],2), 1);
check_equal(vpar_is_component_heights([1,0],2), 0);

{
  my(forest=1);
  for(n=0, test_limit_trees_search - forest,
     \\ print("n="n);
     my(count=0,
        l=List([]));

     \\ all vpar forests heights should pass
     forvec(vec=vpar_forvec_from(n,1), \\ forests
           my(vpar=vpar_from_upascend(vec),
              component_heights=vpar_component_heights(vpar));
           \\ print(" "vpar" heights "component_heights);
           check_equal(vpar_is_component_heights(component_heights,n),1,
                       "vpar_is_component_heights() of heights");
           my(pos=setsearch(l,component_heights,1));
           if(pos, listinsert(l,component_heights,pos)));
     check_equal(#l, 2^n-n);

     \\ among arbitrary vecs, only those from forests should pass
     \\ print(" l = "Vec(l));
     for(len=0,n,
        forvec(vec=vector(len,i,[-1,n]),
           \\ print(" vec "vec" "n);
              check_equal(vpar_is_component_heights(vec,n),
                          setsearch(l,vec)!=0,
                          "vpar_is_component_heights() of vec")));
     );
}


\\-----------------------------------------------------------------------------
progress("vpar_height_num_vertices()");

check_equal(vpar_height_num_vertices([]),0);
check_equal(vpar_height_num_vertices([0]),1);
check_equal(vpar_height_num_vertices([0,1]),1);
check_equal(vpar_height_num_vertices([0,1,1]),2);
check_equal(vpar_height_num_vertices([0,1,1,0,4,4,4]),5);
check_equal(vpar_height_num_vertices([0,0,2,3,4,5]),1, \
            "vpar_height_num_vertices() forest");

test_height_num_vertices_by_depths(vpar) =
{
  my(depths=vpar_depths_vector(vpar),
     height=vpar_height(vpar));
  sum(v=1,#vpar, depths[v]==height);
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     check_equal(vpar_height_num_vertices(vpar),
                 test_height_num_vertices_by_depths(vpar));
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_height_vertices()");

check_equal(vpar_height_vertices([]), []);
check_equal(vpar_height_vertices([0]), [1]);
check_equal(vpar_height_vertices([0,1]), [2]);

test_height_vertices_by_depth(vpar) =
{
  my(l=List([]),
     height=vpar_height(vpar));
  for(v=1,#vpar,
     if(vpar_depth_of_vertex(vpar,v)==height, listput(l,v)));
  Vec(l);
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got=vpar_height_vertices(Vecsmall(vpar)));
     check_equal(#got, vpar_height_num_vertices(vpar));
     check_equal(got, test_height_vertices_by_depth(vpar)));
}


\\-----------------------------------------------------------------------------
progress("vpar_subtree_height()");

test_subtree_height_by_vecmax(vpar,v) =
{
  my(depths=vpar_depths_vector(vpar));
  vecmax(apply(v->depths[v],vpar_subtree_vertices(vpar,v))) - depths[v];
}

check_equal(vpar_subtree_height([0,1],1), 1);
check_equal(vpar_subtree_height([0,1],2), 0, \
            "vpar_subtree_height() at child");
{
  my(vpar=vpar_make_complete_tree(2,3));
  check_equal(vector(#vpar,v,vpar_subtree_height(vpar,v)),
              [3, 2,2, 1,1,1,1, 0,0,0,0, 0,0,0,0]);
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     for(v=1,#vpar,
        check_equal(vpar_subtree_height(Vecsmall(vpar),v),
                    test_subtree_height_by_vecmax(vpar,v),
                    "vpar_subtree_height() vs test_subtree_height_by_vecmax()")));
}


\\-----------------------------------------------------------------------------
progress("vpar_subtree_heights_vector()");

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     check_equal(vpar_subtree_heights_vector(Vecsmall(vpar)),
                 vector(#vpar,v, vpar_subtree_height(vpar,v)),
                 "vpar_subtree_heights_vector() vs individual vpar_subtree_height()"));
}


\\-----------------------------------------------------------------------------
progress("vpar_indnum_sets()");

check_equal(vpar_indnum_sets([]), [ [] ], "vpar_indnum_sets() empty");
check_equal(vpar_indnum_sets([0]), [ [1] ]);

\\ order in returned vec is arbitrary, but this is how they come out presently
check_equal(vpar_indnum_sets([0,1]), [ [1],[2] ], "vpar_indnum_sets() path-2");
check_equal(vpar_indnum_sets([2,0]), [ [2],[1] ]);

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        want_indnum_and_count = vpar_indnum_and_count(vpar),
        want_indnum = want_indnum_and_count[1],
        want_count = want_indnum_and_count[2]);
     if(want_count > test_limit_sets, next());

     my(got = vpar_indnum_sets(vpar));
     check_equal(#got, want_count);
     for(i=1,#got,
        check_equal(setisset(got[i]), 1, "vpar_indnum_sets() is set");
        check_equal(#got[i], want_indnum, "vpar_indnum_sets() set size");
        check_equal(vpar_is_indset(vpar,got[i]), 1,
                    "vpar_indnum_sets() is indset")));
}


\\-----------------------------------------------------------------------------
progress("vpar_radius()");

check_equal(vpar_radius([]), 0, "vpar_radius() empty");
check_equal(vpar_radius([0]), 0, "vpar_radius() singleton");
check_equal(vpar_radius([0,1]), 1);
check_equal(vpar_radius([0,1,2]), 1);
check_equal(vpar_radius([0,1,2,3]), 2);
for(n=0,6, check_equal(vpar_radius(vpar_make_path(n)), ceil((n-1)/2)));

check_equal(vpar_radius([2,3,4,1]), 2);
check_equal(vpar_diameter([2,3,4,1]), 2);

test_radius_by_eccentricities_vector(vpar) = \
  if(#vpar, vecmin(vpar_eccentricities_vector(vpar)), 0);

{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i],
        got=vpar_radius(vpar));
     check_equal(got, test_radius_by_eccentricities_vector(vpar));
     check_equal(got, ceil(vpar_diameter(vpar)/2)));

     \\ (vpar_diameter(vpar)+1)>>1; \\ for acyclics
}
{
  for(i=1,#test_sample_trees_and_unicyclics,
     my(vpar=test_sample_trees_and_unicyclics[i],
        got=vpar_radius(vpar));
     check_equal(got, test_radius_by_eccentricities_vector(vpar)));
}


\\-----------------------------------------------------------------------------
progress("vpar_domnum_sets()");
\\ no disjoint minimum domset pair though no vertex in common

\\ flags1,flags2 are vectors of booleans 0,1, of the same length.
\\ Return 1 if they are disjoint, so they don't have 1 in the same places.
test_flags_is_disjoint(flags1,flags2) = \
  for(i=1,min(#flags1,#flags2), if(flags1[i]&&flags2[i], return(0))); 1;

\\ v is a vector of Set()s of vertex numbers.
\\ Return 1 if any pair of them is disjoint, or 0 if not.
test_sets_any_disjoint_pair(v) =
{
  for(i=1,#v,
     for(j=i+1,#v,
        if(setintersect(v[i],v[j])==[], return(1))));  \\ a disjoint pair
  0;
}

{
  \\ Lowenstein example no disjoint minimum domsets, though no single vertex
  \\ in common.
  \\
  \\           1
  \\      /    |    \
  \\   4      10      16
  \\  / \     / \     / \
  \\ 3   5   9  11  15  17
  \\ |   |   |   |   |   |
  \\ 2   6   8  12  14  18
  \\     |       |       |
  \\     7      13      19
  \\
  \\ Christian Lowenstein, "In the Complement of a Dominating Set",
  \\ dissertation, April 2010.
  \\ https://www.db-thueringen.de/servlets/MCRFileNodeServlet/dbt_derivate_00021011/ilm1-2010000233.pdf
  \\
  \\ https://hog.grinvin.org/ViewGraphInfo.action?id=30696

  my(vpar=vpar_reroot(vpar_make_path(6),3));
  vpar=vpar_join(vpar_join(vpar_join([0],1,vpar),1,vpar),1,vpar);

  check_equal(vpar_minimal_domsets_count(vpar), 445);
  check_equal(vpar_domnum_and_count(vpar), [7, 20]);

  my(flags1 = [1, 0,1,0,0,1,0, 0,1,0,0,1,0, 0,1,0,0,1,0],
     flags2 = [0, 1,0,1,0,0,1, 0,1,0,0,1,0, 0,1,0,0,1,0]);
  check_equal(vpar_is_minimal_domset_flags(vpar,flags1), 1);
  check_equal(vpar_is_minimal_domset_flags(vpar,flags2), 1);
  check_equal(vecsum(flags1), 7);
  check_equal(vecsum(flags2), 7);
  check_equal(test_flags_is_disjoint(flags1,flags2), 0,
              "flags 1,2 disjoint");

  \\ my(flags3 = [0, 1,0,1,0,0,1,  0,0,1,0,0,1, 1,0,1,0,0,1]);
  \\ check_equal(vecsum(flags3), 7, "flags3 size");
  \\ check_equal(vpar_is_minimal_domset_flags(vpar,flags3), 1);
  \\ check_equal(test_flags_is_disjoint(flags1,flags3), 0);

  my(sets=vpar_domnum_sets(vpar));
  check_equal(#sets, 20);
  check_equal(#sets[1], 7);
  check_equal(test_setintersect_sets(sets), []);
  check_equal(test_sets_any_disjoint_pair(sets), 0);
}

{
  \\ n=16 no disjoint minimum domset pair though no vertex in common
  \\             *
  \\          /  |  \
  \\        *    *    *
  \\      / |    |    | \
  \\     *  *    *    *  *
  \\     |  |    |    |  |
  \\     *  *    *    *  *
  \\     |               |
  \\     *               *
  \\ https://hog.grinvin.org/ViewGraphInfo.action?id=30698

  my(vpar=[2, 3, 4, 5, 0, 5, 6, 7, 8, 6, 10, 5, 12, 13, 4, 15]);
  check_equal(vpar_minimal_domsets_count(vpar), 162);
  check_equal(vpar_domnum_and_count(vpar), [6, 18]);

  my(sets=vpar_domnum_sets(vpar));
  check_equal(#sets, 18);
  check_equal(#sets[1], 6);
  check_equal(test_setintersect_sets(sets), []);
  check_equal(test_sets_any_disjoint_pair(sets), 0);
}


\\-----------------------------------------------------------------------------
progress("vpar_disjoint_domnum()");
\\ vpar_disjoint_domnum_and_count()

check_equal(vpar_disjoint_domnum([]), 0, "vpar_disjoint_domnum() empty");
check_equal(vpar_disjoint_domnum_and_count([]), [0,1], \
            "vpar_disjoint_domnum_and_count() empty");

check_equal(vpar_disjoint_domnum([0]), -1);
check_equal(vpar_disjoint_domnum_and_count([0]), [-1,0]);

check_equal(vpar_disjoint_domnum([0,1]), 2, "vpar_disjoint_domnum() path-2");
check_equal(vpar_disjoint_domnum_and_count([0,1]), [2,1], \
            "vpar_disjoint_domnum_and_count() path-2");

check_equal(test_is_disjoint_domsets_flags(vpar_make_path(5),[1,2,0,2,1]), 0, \
            "test_is_disjoint_domsets_flags()");
check_equal(test_is_disjoint_domsets_flags(vpar_make_path(5),[1,2,0,1,2]), 1);

check_equal(test_disjoint_domnum_and_count_by_flags(vpar_make_path(2)), [2,1], \
           "test_disjoint_domnum_and_count_by_flags()");
check_equal(test_disjoint_domnum_and_count_by_flags(vpar_make_path(4)), [4,2]);
check_equal(test_disjoint_domnum_and_count_by_flags(vpar_make_star(4)), [4,1]);
check_equal(test_disjoint_domnum_and_count_by_flags(vpar_make_path(5)), [4,1], \
           "test_disjoint_domnum_and_count_by_flags()");

{
  \\ paths and rerooted
  \\
  \\ 5    with 4[2] 4[2]  without 10[0] 4[4] 9[0]
  \\            ^^
  \\ 4    with 4[2] 3[2]  without 9[0] 3[2] 8[0]
  \\
  \\ 3    with 3[1] 3[1]  without 6[0] 2[1] 8[0]
  \\
  \\ 2    with 2[1] 5[0]  without 8[0] 4[0] 4[0]
  \\
  \\ 1    with 4[0] 1[1]  without 4[0] 4[0] 0[1]
  \\
  \\ n=3
  \\   1--2--3
  \\   2  1       without_dom_undom size=2 in 1 way
  \\
  \\ n=4
  \\   1--2--3--4
  \\   1  2  1  2       size=4 in 2 ways
  \\   1  2  2  1
  \\   1  2     1  with_undom 3 in 1 way
  \\ n=5
  \\   1--2--3--4--5
  \\   1  2     1  2    size=4 in 1 way
  \\
  \\ print(test_disjoint_domnum_sets_by_flags(vpar));

  for(n=1,10,
     my(vpar=vpar_relabel_reverse(vpar_make_path(n)));
     for(rep=0,1,
        my(by_flags=test_disjoint_domnum_and_count_by_flags(vpar),
           name=Str("path n="n,if(rep," rerooted","")));

        \\ check_equal(vpar_disjoint_domnum_and_count_quantities(vpar),
        \\             test_disjoint_domnum_and_count_quantities_by_flags(vpar),
        \\             Str("vpar_disjoint_domnum_and_count_quantities() "name));

        check_equal(vpar_disjoint_domnum(vpar), by_flags[1]);
        check_equal(vpar_disjoint_domnum_and_count(vpar), by_flags,
                    Str("vpar_disjoint_domnum_and_count() "
                        "vs test_disjoint_domnum_and_count_by_flags() "name));

        \\ for next loop
        vpar=vpar_reroot_random(vpar);
        ));
}

{
  \\ singleton
  check_equal(test_disjoint_domnum_and_count_quantities_by_flags([0]),
              [-1, 0;   \\ with_dom
               1, 1;    \\ with_undom
               -1, 0;
               -1, 0;
               0, 1;
               -1, 0],
              "test_disjoint_domnum_and_count_quantities_by_flags() singleton");
}
{
  \\ path-2
  check_equal(test_disjoint_domnum_and_count_quantities_by_flags([0,1]),
              [2, 1;   \\ with_dom
               -1, 0;
               -1, 0;  \\ without_dom_dom
               -1, 0;
               -1, 0;
               2, 1],   \\ domsets
              "test_disjoint_domnum_and_count_quantities_by_flags() path-2");
}

{
  my(vpar=vpar_make_path(3));
  check_equal(vpar_disjoint_domnum(vpar), 3, "vpar_disjoint_domnum() path-3");
  check_equal(test_disjoint_domnum_and_count_by_flags(vpar), [3,1]);
  check_equal(vpar_disjoint_domnum_and_count(vpar), [3,1],
              "vpar_disjoint_domnum_and_count() path-3");

  check_equal(test_disjoint_domnum_and_count_quantities_by_flags(vpar),
              [3, 1;  \\ with_dom
               3, 1;
               -1, 0;  \\ without_dom_dom
               2, 1;
               -1, 0;
               3,1],
              "test_disjoint_domnum_and_count_quantities_by_flags() path-3");
}
{
  \\   1
  \\  / \
  \\ 2   3
  my(vpar=vpar_make_star(3));
  check_equal(test_disjoint_domnum_and_count_by_flags(vpar), [3,1]);
  check_equal(vpar_disjoint_domnum(vpar), 3, "vpar_disjoint_domnum() star-3");
  check_equal(vpar_disjoint_domnum_and_count(vpar), [3,1]);

  \\ check_equal(vpar_disjoint_domnum_and_count_quantities(vpar),
  \\             test_disjoint_domnum_and_count_quantities_by_flags(vpar),
  \\             "vpar_disjoint_domnum_and_count_quantities() vs test_disjoint_domnum_and_count_quantities_by_flags() star-3");
}

check_equal(vpar_disjoint_domnum([0,1,2,3]), 4, \
            "vpar_disjoint_domnum() path-4");

{
  \\ path-4 rerooted
  \\     2
  \\    / \
  \\   1   3
  \\        \
  \\         4
  my(vpar=[2,0,2,3]);
  check_equal(vpar_is_domset_flags(vpar,[1,0,1,0]), 1);
  check_equal(vpar_is_domset_flags(vpar,[0,1,0,1]), 1);
  check_equal(vpar_is_domset_flags(vpar,[1,0,0,1]), 1);
  check_equal(vpar_is_domset_flags(vpar,[0,1,1,0]), 1);
  check_equal(test_is_disjoint_domsets_flags(vpar,[1,2,1,2]), 1);
  check_equal(test_is_disjoint_domsets_flags(vpar,[1,2,2,1]), 1);

  check_equal(test_disjoint_domnum_and_count_by_flags(vpar), [4,2],
              "test_disjoint_domnum_and_count_by_flags() path-4 rerooted");
  check_equal(vpar_disjoint_domnum(vpar), 4,
              "vpar_disjoint_domnum() path-4 rerooted");
  check_equal(vpar_disjoint_domnum_and_count(vpar), [4,2],
              "vpar_disjoint_domnum_and_count() path-4 rerooted");

  \\ check_equal(vpar_disjoint_domnum_and_count_quantities(vpar),
  \\             test_disjoint_domnum_and_count_quantities_by_flags(vpar),
  \\             "vpar_disjoint_domnum_and_count_quantities() vs test_disjoint_domnum_and_count_quantities_by_flags() path-4 rerooted");
}

check_equal(vpar_disjoint_domnum([0,1,1,1]), 4, \
            "vpar_disjoint_domnum() star-4");
check_equal(vpar_disjoint_domnum([0,1,2,3,4]), 4, \
            "vpar_disjoint_domnum() path-5");

check_equal(vpar_disjoint_domnum([2,6,4,6,6,0]), 6, \
            "vpar_disjoint_domnum() 6 leaf last");
check_equal(vpar_disjoint_domnum([2,6,6,5,6,0]), 6, \
            "vpar_disjoint_domnum() 6 leaf middle");
check_equal(vpar_disjoint_domnum([6,3,6,5,6,0]), 6, \
            "vpar_disjoint_domnum() 6 leaf first");

{
  \\     4     with 5[3] 5[1]  without 4[1] 4[1] 24[0]
  \\    / \
  \\   5   1   with 2[1] 7[0]  without 12[0] 6[0] 6[0]
  \\   |   |
  \\   3   2
  my(vpar=[4, 1, 5, 0, 4]);

  check_equal(test_disjoint_domnum_and_count_quantities_by_flags(vpar),
              [5, 3;  \\ with_dom
               5, 1;
               4, 1;  \\ without_dom_dom
               4, 1;
               -1, 0;
               4,1],
              "test_disjoint_domnum_and_count_quantities_by_flags() 5b");
  \\ check_equal(vpar_disjoint_domnum_and_count_quantities(vpar),
  \\             test_disjoint_domnum_and_count_quantities_by_flags(vpar),
  \\             "vpar_disjoint_domnum_and_count_quantities() vs test_disjoint_domnum_and_count_quantities_by_flags() 5b");

  check_equal(test_disjoint_domnum_and_count_by_flags(vpar), [4,1],
              "vpar_disjoint_domnum() 5b");
  check_equal(vpar_disjoint_domnum(vpar), 4,
              "vpar_disjoint_domnum() 5b");
  check_equal(vpar_disjoint_domnum_and_count(vpar), [4,1],
              "vpar_disjoint_domnum() 5b");
}

{
  \\     5     with 6[4] 5[3]  without 11[0] 5[3] 4[1]
  \\     |
  \\     4     with 5[3] 5[1]  without 4[2] 4[1] 28[0]
  \\    / \
  \\   6   1   with 2[1] 8[0]  without 14[0] 7[0] 7[0]
  \\   |   |
  \\   3   2
  my(vpar=[4, 1, 6, 5, 0, 4]);
  check_equal(test_disjoint_domnum_and_count_by_flags(vpar), [6,4],
              "vpar_disjoint_domnum() 6b");
  check_equal(vpar_disjoint_domnum(vpar), 6,
              "vpar_disjoint_domnum() 6b");
  check_equal(vpar_disjoint_domnum_and_count(vpar), [6,4],
              "vpar_disjoint_domnum() 6b");

  check_equal(test_disjoint_domnum_and_count_quantities_by_flags(vpar),
              [6, 4;  \\ with_dom
               5, 3;
               -1, 0;  \\ without_dom_dom
               5, 3;
               4, 1;
               6,4],
              "test_disjoint_domnum_and_count_quantities_by_flags() 6b");
  \\ check_equal(vpar_disjoint_domnum_and_count_quantities(vpar),
  \\             test_disjoint_domnum_and_count_quantities_by_flags(vpar),
  \\             "vpar_disjoint_domnum_and_count_quantities() vs test_disjoint_domnum_and_count_quantities_by_flags() 6b");
}
{
  \\         6       with 6[3] 13[0]  without 13[0] 12[0] 12[0]
  \\        / \
  \\       4   5     at 4   = 5[3] 5[1]  without 4[1] 4[1] 32[0]
  \\      / \        4 to 6 = 6[4] 5[2]  without 12[0] 5[3] 4[1]
  \\     1   7
  \\     |   |
  \\     2   3
  my(vpar=[4, 1, 7, 6, 6, 0, 4]);

  check_equal(test_disjoint_domnum_and_count_quantities_by_flags(vpar),
              [6, 3;  \\ with_dom
               -1, 0;
               -1, 0;  \\ without_dom_dom
               -1, 0;
               -1, 0;
               6,3],
              "test_disjoint_domnum_and_count_quantities_by_flags() 7b");
  \\ check_equal(vpar_disjoint_domnum_and_count_quantities(vpar),
  \\             test_disjoint_domnum_and_count_quantities_by_flags(vpar),
  \\             "vpar_disjoint_domnum_and_count_quantities() vs test_disjoint_domnum_and_count_quantities_by_flags() 7b");

  check_equal(test_disjoint_domnum_and_count_by_flags(vpar), [6,3],
              "vpar_disjoint_domnum() 7b");
  check_equal(vpar_disjoint_domnum(vpar), 6,
              "vpar_disjoint_domnum() 7b");
  check_equal(vpar_disjoint_domnum_and_count(vpar), [6,3],
              "vpar_disjoint_domnum() 7b");
}
{
  \\       7          with 7[3] 7[3]  without 7[6] 6[2] 22[0]
  \\     /   \
  \\    6     4            5[3] 5[1]  without 4[2] 4[1] 36[0]
  \\    |    / \
  \\    5   8   1
  \\        |   |
  \\        3   2
  my(vpar=[4, 1, 8, 7, 6, 7, 0, 4]);

  \\ 1,3,6 = dom and 2,4,5,8 = dom
  check_equal(vpar_is_domset(vpar,[1,3,6]), 1);
  check_equal(vpar_is_domset(vpar,[2,4,5,8]), 1);
  check_equal(test_flags_to_Set([1,2,1,2,2,1,0,2],1), [1, 3, 6]);
  check_equal(test_flags_to_Set([1,2,1,2,2,1,0,2],2), [2, 4, 5, 8]);
  check_equal(test_is_disjoint_domnum_quantities_by_flags(vpar,[1,2,1,2,2,1,0,2]),
              [0, 0, 1, 0, 0]);

  \\ 1,6,8 and 2,3,4,5
  check_equal(test_flags_to_Set([1,2,2,2,2,1,0,1],1), [1, 6, 8]);
  check_equal(vpar_is_domset(vpar,[1,6,8]), 1);
  check_equal(vpar_is_domset(vpar,[2,3,4,5]), 1);
  check_equal(test_is_disjoint_domnum_quantities_by_flags(vpar,[1,2,2,2,2,1,0,1]),
              [0, 0, 1, 0, 0]);

  check_equal(test_disjoint_domnum_and_count_quantities_by_flags(vpar),
              [7, 3;  \\ with_dom
               7, 3;
               7, 3;  \\ without_dom_dom
               6, 2;
               -1, 0;
               7,6],
              "test_disjoint_domnum_and_count_quantities_by_flags() 8b");
  \\ check_equal(vpar_disjoint_domnum_and_count_quantities(vpar),
  \\             test_disjoint_domnum_and_count_quantities_by_flags(vpar),
  \\             "vpar_disjoint_domnum_and_count_quantities() vs test_disjoint_domnum_and_count_quantities_by_flags() 8b");

  check_equal(test_disjoint_domnum_and_count_by_flags(vpar), [7,6],
              "test_disjoint_domnum_and_count_by_flags() 8b");
  check_equal(vpar_disjoint_domnum(vpar), 7,
              "vpar_disjoint_domnum() 8b");
  check_equal(vpar_disjoint_domnum_and_count(vpar), [7,6],
              "vpar_disjoint_domnum_and_count() 8b");
}
{
  \\       8
  \\     /   \
  \\    7     4
  \\    |    / \
  \\    6   9   1
  \\    |   |   |
  \\    5   3   2
  my(vpar=[4,1,9,8,6,7,8,0,4]);
  check_equal(test_disjoint_domnum_and_count_by_flags(vpar), [8,13],
              "vpar_disjoint_domnum() 9b");
  check_equal(vpar_disjoint_domnum(vpar), 8,
              "vpar_disjoint_domnum() 9b");
  check_equal(vpar_disjoint_domnum_and_count(vpar), [8,13],
              "vpar_disjoint_domnum() 9b");
}

{
  \\       6
  \\     / | \
  \\    1  3  2
  \\    |  |  |
  \\    9  8  5
  \\    |     |
  \\    7     4
  my(vpar=[6,6,6,5,2,0,9,3,1]);

  \\ check_equal(vpar_disjoint_domnum_and_count_quantities(vpar),
  \\             test_disjoint_domnum_and_count_quantities_by_flags(vpar),
  \\             "vpar_disjoint_domnum_and_count_quantities() vs test_disjoint_domnum_and_count_quantities_by_flags() 9");

  check_equal(test_disjoint_domnum_and_count_by_flags(vpar), [7,1],
              "vpar_disjoint_domnum() 9");
  check_equal(vpar_disjoint_domnum([6,6,6,5,2,0,9,3,1]), 7,
              "vpar_disjoint_domnum() 9");
}

{
  \\         6
  \\       / |
  \\     8   9
  \\   / |   |
  \\  2  3   4
  \\  |  |   |
  \\  7  5   1
  my(vpar=[4, 8, 8, 9, 3, 0, 2, 6, 6]);

  \\ check_equal(vpar_disjoint_domnum_and_count_quantities(vpar),
  \\             test_disjoint_domnum_and_count_quantities_by_flags(vpar),
  \\             "vpar_disjoint_domnum_and_count_quantities() vs test_disjoint_domnum_and_count_quantities_by_flags() 10-2");
}
{
  \\         7
  \\       / |  \          [1, 2, 6, 10, 11]  [3, 4, 5, 8]
  \\     9  10   5         [1, 2, 6, 7, 11]   [3, 4, 5, 8]
  \\   / |   |   |         [1, 3, 8, 10, 11]  [2, 4, 5, 6]
  \\  2  3   4  11         [1, 3, 7, 8, 11]   [2, 4, 5, 6]
  \\  |  |   |             [1, 6, 7, 8, 11]   [2, 3, 4, 5]
  \\  8  6   1                                  2,6 or 2,3 or 3,8 with 4,5
  \\
  \\ at 10  with 3[1] 3[1]  without 14[0] 2[1] 24[0]
  \\        to p 4[2] 3[1]  without 15[0] 3[1] 14[0]
  \\
  \\ at 9   with 5[3] 5[1]  without 4[1] 4[1] 24[0]
  \\        to p 8[10] 7[3]  without 8[3] 7[1] 18[0]
  \\
  \\ at 5   with 2[1] 13[0]  without 24[0] 12[0] 12[0]
  \\
  my(vpar=[4, 9, 9, 10, 7, 3, 0, 2, 7, 7, 5]);
  \\ print(test_disjoint_domnum_sets_by_flags(vpar));

  my(want_domnum_and_count = test_disjoint_domnum_and_count_by_flags(vpar),
     got_domnum_and_count = vpar_disjoint_domnum_and_count(vpar),
     got_domnum = vpar_disjoint_domnum(vpar));

  check_equal(test_disjoint_domnum_and_count_quantities_by_flags(vpar),
              [9, 3;  \\ with_dom
               9, 3;
               9, 2;  \\ without_dom_dom
               9, 2;
               -1, 0;
               9,5],
              "test_disjoint_domnum_and_count_quantities_by_flags() 10");
  \\ check_equal(vpar_disjoint_domnum_and_count_quantities(vpar),
  \\             test_disjoint_domnum_and_count_quantities_by_flags(vpar),
  \\             "vpar_disjoint_domnum_and_count_quantities() vs test_disjoint_domnum_and_count_quantities_by_flags() 9");

  check_equal(got_domnum, want_domnum_and_count[1]);
  check_equal(got_domnum_and_count, want_domnum_and_count,
              "vpar_disjoint_domnum_and_count() 10");
}
{
  \\    6
  \\    |
  \\    2
  \\   / \
  \\  3   1
  \\  |   |
  \\  5   4
  \\
  \\ [1, 2, 3]  [4, 5, 6]   size=6  with v=6 have 4 ways
  \\ [1, 2, 5]  [3, 4, 6]
  \\ [1, 3, 6]  [2, 4, 5]
  \\ [1, 5, 6]  [2, 3, 4]
  \\
  my(vpar=[2, 6, 2, 1, 3, 0]);
  \\ print(test_disjoint_domnum_sets_by_flags(vpar));

  my(want_domnum_and_count=test_disjoint_domnum_and_count_by_flags(vpar),
     got_domnum_and_count=vpar_disjoint_domnum_and_count(vpar),
     got_domnum=vpar_disjoint_domnum(vpar));
  check_equal(got_domnum,want_domnum_and_count[1]);
  check_equal(got_domnum_and_count,want_domnum_and_count);

  check_equal(test_disjoint_domnum_and_count_quantities_by_flags(vpar),
              [6, 4;  \\ with_dom
               5, 3;
               -1,0;  \\ without_dom_dom
               5, 3;
               4, 1;
               6, 4],
              "test_disjoint_domnum_and_count_quantities_by_flags() 6c");
  \\ check_equal(vpar_disjoint_domnum_and_count_quantities(vpar),
  \\             test_disjoint_domnum_and_count_quantities_by_flags(vpar),
  \\             Str("vpar_disjoint_domnum_and_count_quantities() vs "
  \\                 "test_disjoint_domnum_and_count_quantities_by_flags() 6c"));
}
{
  \\      1
  \\     / \
  \\    9   4
  \\    |   |
  \\    3   7
  \\   / \
  \\  5   2
  \\  |   |
  \\  8   6
  \\
  \\ i=3 new low 8
  \\ flags=[0, 1, 1, 1, 1, 2, 2, 2, 2] size 8  [0, 0, 1, 0, 0]
  \\ flags=[0, 2, 2, 2, 2, 1, 1, 1, 1] size 8  [0, 0, 1, 0, 0]
  \\
  \\ flags=[0, 1, 1, 1, 2, 2, 2, 1, 2] size 8  [0, 0, 1, 0, 0]
  \\ flags=[0, 2, 2, 2, 1, 1, 1, 2, 1] size 8  [0, 0, 1, 0, 0]
  \\
  \\ flags=[0, 1, 2, 2, 1, 2, 1, 2, 1] size 8  [0, 0, 1, 0, 0]
  \\ flags=[0, 2, 1, 1, 2, 1, 2, 1, 2] size 8  [0, 0, 1, 0, 0]
  \\
  \\ flags=[0, 1, 2, 2, 2, 2, 1, 1, 1] size 8  [0, 0, 1, 0, 0]
  \\ flags=[0, 2, 1, 1, 1, 1, 2, 2, 2] size 8  [0, 0, 1, 0, 0]

  my(vpar=[0, 3, 9, 1, 3, 2, 4, 5, 1]);
  \\ print(test_disjoint_domnum_sets_by_flags(vpar));

  my(want_domnum_and_count=test_disjoint_domnum_and_count_by_flags(vpar),
     got_domnum_and_count=vpar_disjoint_domnum_and_count(vpar),
     got_domnum=vpar_disjoint_domnum(vpar));
  check_equal(got_domnum,want_domnum_and_count[1]);
  check_equal(got_domnum_and_count,want_domnum_and_count);

  check_equal(test_disjoint_domnum_and_count_quantities_by_flags(vpar),
              [8, 9;  \\ with_dom
               8, 3;
               8, 4;  \\ without_dom_dom
               8, 4;
               -1, 0;
               8, 13],
              "test_disjoint_domnum_and_count_quantities_by_flags() 9c");
  \\ check_equal(vpar_disjoint_domnum_and_count_quantities(vpar),
  \\             test_disjoint_domnum_and_count_quantities_by_flags(vpar),
  \\             "vpar_disjoint_domnum_and_count_quantities() vs test_disjoint_domnum_and_count_quantities_by_flags() 9c");
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     if(#vpar <= test_limit_flags_search,
        my(by_flags=test_disjoint_domnum_and_count_by_flags(vpar));
        \\ print(vpar);
        check_equal(vpar_disjoint_domnum_and_count(vpar), by_flags,
                    Str("vpar_disjoint_domnum_and_count() "
                        "vs test_disjoint_domnum_and_count_by_flags()"));

        \\ print(vpar);
        check_equal(vpar_disjoint_domnum(vpar), by_flags[1],
                    Str("vpar_disjoint_domnum() "
                        "vs test_disjoint_domnum_and_count_by_flags()"));
       ));
}

test_disjoint_domnum_of_complete_binary_tree(h) = \
  1/7*(12*2^h + if(h%3==0,2, h%3==1,-3, -6));
{
  \\ complete binary tree
  check_equal(vector(14,h, vpar_disjoint_domnum(vpar_make_complete_tree(2,h))),
              vector(14,h, test_disjoint_domnum_of_complete_binary_tree(h)),
              "vpar_disjoint_domnum() of complete binary tree");

  \\ ratio exactly 6/7 when h%3==1
  my(vpar=vpar_make_complete_tree(2,6));
  check_equal(vpar_disjoint_domnum(vpar)/#vpar, 6/7 + 8/889);
  vpar=vpar_make_complete_tree(2,8);
  check_equal(vpar_disjoint_domnum(vpar)/#vpar, 6/7);
  vpar=vpar_make_complete_tree(2,10);
  check_equal(vpar_disjoint_domnum(vpar)/#vpar, 6/7 + 3/14329);

  \\ not in OEIS: 3, 6, 14, 27, 54, 110, 219, 438, 878, 1755, 3510, 7022, 14043, 28086, 56174, 112347
  \\ size         3  7  15  31  63  127  255
  \\ ways
  \\ not in OEIS: 33, 297, 88209, 682934601537
  \\ factor(88209)
  \\ factor(682934601537)

  check_equal(vector(6,h, vpar_disjoint_domnum_and_count(vpar_make_complete_tree(2,h))[2]),
              [1, 1, 33, 297, 88209, 682934601537],
              "vpar_disjoint_domnum_and_count() of complete binary tree");
}

{
  check_equal(vector(10,n,n++; vpar_disjoint_domnum(vpar_make_star(n))),
              vector(10,n,n++; n),
              "vpar_disjoint_domnum() of star");
  check_equal(vector(10,n,n++; vpar_disjoint_domnum_and_count(vpar_make_star(n))[2]),
              vector(10,n,n++; 1),
              "vpar_disjoint_domnum_and_count() of star");

  \\ 2 vertex arms
  check_equal(vector(10,n,n++; vpar_disjoint_domnum
                                 (vpar_subdivide(vpar_make_star(n)))),
              vector(10,n,n++; if(n==2,3, 2*n-2)),
              "vpar_disjoint_domnum() of subdivided star");
  check_equal(vector(10,n,n++; vpar_disjoint_domnum_and_count
                                 (vpar_subdivide(vpar_make_star(n)))[2]),
              vector(10,n,n++; if(n<=2,1, 2^(n-2)-1)),
              "vpar_disjoint_domnum_and_count() of subdivided star");

  \\ arms 3 vertices, 3*n-2 vertices
  \\ difference without_dom_dom to with_dom is n-3 arbitrarily large
  check_equal(vector(10,n,n++; vpar_disjoint_domnum
                                 (vpar_subdivide(vpar_make_star(n),2))),
              vector(10,n,n++; 2*n),
              "vpar_disjoint_domnum() of 2-subdivided star");
  check_equal(vector(10,n,n++; vpar_disjoint_domnum_and_count
                                 (vpar_subdivide(vpar_make_star(n),2))[2]),
              vector(10,n,n++; if(n==3,5, 2*n-2)),
              "vpar_disjoint_domnum_and_count() of 2-subdivided star");
  \\ check_equal(vector(10,n,n++; vpar_disjoint_domnum_and_count_quantities
  \\                                (vpar_subdivide(vpar_make_star(n),2))),
  \\             vector(10,n,n++; [2*n,                2*n-2;         \\ with_dom
  \\                               2*n-1,              1;
  \\                               if(n==2,-1, 3*n-3), 2^(n-2)-1;     \\ without_dom_dom
  \\                               3*n-3,              1;             \\ without_dom_undom = #vpar-1
  \\                               -1,                 0;
  \\                               2*n,                if(n==3,5, 2*n-2) ]),
  \\             "vpar_disjoint_domnum_and_count() of 2-subdivided star");
}
{
  check_equal(vector(20,n,n++; vpar_disjoint_domnum
                                (vpar_make_binomial_tree(n))),
              vector(20,n,n++; n),
              "vpar_disjoint_domnum() of binomial n");
  check_equal(vector(20,n,n++; vpar_disjoint_domnum_and_count
                                (vpar_make_binomial_tree(n))[2]),
              vector(20,n,n++; 2^floor(n/2-1)),
              "vpar_disjoint_domnum_and_count() of binomial n");
}


\\-----------------------------------------------------------------------------
progress("vpar_centroids()");

check_equal(vpar_centroids([]), [], "vpar_centroids() empty tree");

{
  \\ smallest centre and centroid disjoint
  my(vpar=[0,1,2,3,4,4,4,4]);
  \\           8   7
  \\            \ /
  \\ 1---2---3---4---6
  \\         C  G|
  \\             5
  check_equal(#vpar,8);
  check_equal(vpar_centres(vpar),  [3]);
  check_equal(vpar_centroids(vpar),[4]);
}
{
  \\ smallest bi-centre and bi-centroid disjoint
  \\ 3  2
  \\   \|
  \\ 4--1--7--8--9--10--11--12
  \\   /|
  \\ 5  6
  my(vpar=[0,1,1,1,1,1,1,7,8,9,10,11]);
  check_equal(#vpar,12);
  check_equal(vecsort(vpar_centres(vpar)), [8,9]);
  check_equal(vpar_centroids(vpar),[1,7]);
}


\\-----------------------------------------------------------------------------
progress("vpar_totdomsets()");

check_equal(vpar_totdomsets([]), [ [] ]);
check_equal(vpar_totdomsets([0]), [  ]);
check_equal(vpar_totdomsets([0,1]), [ [1,2] ]);

\\ order unspecified in the return, but this is how it comes out now
check_equal(vpar_totdomsets([0,1,2]), [ [1,2,3], [1,2], [2,3] ]);

test_totdomsets_from_domsets(vpar) = \
  select(s->vpar_is_totdomset(vpar,s), vpar_domsets(vpar));

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        totdomsets_count = vpar_totdomsets_count(vpar));
     if(totdomsets_count > test_limit_sets, next());

     my(sets=vpar_totdomsets(vpar));
     check_equal(#sets, totdomsets_count, "vpar_totdomsets() count");

     for(i=1,#sets,
        check_equal(setisset(sets[i]), 1, "vpar_totdomsets() is set");
        check_equal(vpar_is_totdomset(vpar,sets[i]), 1,
                    "vpar_totdomsets() is totdomset"));

     check_equal(Pol(0,'y) + sum(i=1,#sets, 'y^#sets[i]),
                 vpar_totdompoly(vpar,'y),
                 "vpar_totdomsets() size counts vs vpar_totdompoly()");

     if(vpar_domsets_count(vpar) <= test_limit_sets,
        check_equal(vecsort(sets),
                    vecsort(test_totdomsets_from_domsets(vpar)),
                    "vpar_totdomsets() vs test_totdomsets_from_domsets()"));
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_domnum_sets()");

check_equal(vpar_domnum_sets([]), [ [] ]);
check_equal(vpar_domnum_sets([0]), [ [1] ]);

\\ order in returned vec is arbitrary, but this is how they come out presently
check_equal(vpar_domnum_sets([0,1]), [ [1],[2] ]);
check_equal(vpar_domnum_sets([2,0]), [ [2],[1] ]);

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        want_domnum_and_count = vpar_domnum_and_count(vpar),
        want_domnum = want_domnum_and_count[1],
        want_count = want_domnum_and_count[2]);
     if(want_count > test_limit_sets, next());

     my(got = vpar_domnum_sets(vpar));
     check_equal(#got, want_count);
     for(i=1,#got,
        check_equal(setisset(got[i]), 1, "vpar_domnum_sets() is set");
        check_equal(#got[i], want_domnum, "vpar_domnum_sets() set size");
        check_equal(vpar_is_minimal_domset(vpar,got[i]), 1,
                    "vpar_domnum_sets() is minimal domset"));

     \\ When no isolated vertices, the set complement of a minimal
     \\ dominating set is dominating, because every vertex in a minimal
     \\ dominating set has at least neighbour absent from the set (or it
     \\ would not be minimal, all the neighbours would dominate the vertex
     \\ so no need for it in the set).
     \\
     for(i=1,#got,
        my(comp=vpar_set_complement(vpar,got[i]));
        \\ print(vpar" set "got[i]" comp "comp);
        check_equal(setisset(comp), 1, "vpar_domnum_sets() complement is set");
        check_equal(vpar_is_domset(vpar,comp),
                    ! test_any_singletons(vpar),
                    "vpar_domnum_sets() complement is dominating"));
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_indomnum_sets()");

check_equal(vpar_indomnum_sets([]), [ [] ]);
check_equal(vpar_indomnum_sets([0]), [ [1] ]);

\\ order in returned vec is arbitrary, but this is how they come out presently
check_equal(vpar_indomnum_sets([0,1]), [ [1],[2] ]);
check_equal(vpar_indomnum_sets([2,0]), [ [2],[1] ]);

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        want_indomnum_and_count = vpar_indomnum_and_count(vpar),
        want_indomnum = want_indomnum_and_count[1],
        want_count = want_indomnum_and_count[2]);
     if(want_count > test_limit_sets, next());

     my(got = vpar_indomnum_sets(vpar));
     check_equal(#got, want_count);
     for(i=1,#got,
        check_equal(setisset(got[i]), 1, "vpar_indomnum_sets() is set");
        check_equal(#got[i], want_indomnum, "vpar_indomnum_sets() set size");
        check_equal(vpar_is_indomset(vpar,got[i]), 1,
                    "vpar_indomnum_sets() is indomset")));
}


\\-----------------------------------------------------------------------------
progress("vpar_indomnum()");

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     if(#vpar <= test_limit_flags_search,
        check_equal(vpar_indomnum(vpar),
                    vpar_indomnum_and_count(vpar)[1])));
}


\\-----------------------------------------------------------------------------
progress("vpar_indomnum_and_count()");

check_equal(vpar_indomnum_and_count([]), [0,1], \
            "vpar_indomnum_and_count() empty");

test_indomnum_and_count_by_flags(vpar) =
{
  if(#vpar==0,return([0,1]));  \\ empty set only
  my(indomnum=#vpar+1,count=0);
  forvec(flags=vector(#vpar,i,[0,1]),
        if(vpar_is_indomset_flags(vpar,flags),
           my(size=vecsum(flags));
           if(size<indomnum,  indomnum=size;count=1,  \\ new low
              size==indomnum, count++)));           \\ equal low
  if(count==0,indomnum=-1);
  [indomnum,count];
}
check_equal(test_indomnum_and_count_by_flags(vpar_make_path(0)), [0,1], \
            "test_indomnum_and_count_by_flags of path 0");
check_equal(test_indomnum_and_count_by_flags(vpar_make_path(1)), [1,1], \
            "test_indomnum_and_count_by_flags of path 1");
check_equal(test_indomnum_and_count_by_flags(vpar_make_path(2)), [1,2]);
check_equal(test_indomnum_and_count_by_flags(vpar_make_path(3)), [1,1]);
check_equal(test_indomnum_and_count_by_flags(vpar_make_path(4)), [2,3]);

{
  for(n=0,10,
     my(vpar=vpar_make_path(n),
        by_flags=test_indomnum_and_count_by_flags(vpar),
        got=vpar_indomnum_and_count(vpar));
     check_equal(got,by_flags,
                 Str("vpar_indomnum_and_count() path n="n));
     check_equal(got[1], (n+2)\3);
  );
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got_indomnum_and_count=vpar_indomnum_and_count(vpar),
        got_indomnum = got_indomnum_and_count[1],
        got_count    = got_indomnum_and_count[2]);

     if(#vpar,
        check_equal(got_indomnum>=1, 1,
                    "vpar_indomnum_and_count() indomnum>=1"));

     check_equal(got_count <= vpar_domsets_count(vpar), 1);
     check_equal(got_count <= vpar_indsets_count(vpar), 1);
     if(#vpar <= test_limit_flags_search,
        check_equal(got_indomnum_and_count,
                    test_indomnum_and_count_by_flags(vpar),
          "vpar_indomnum_and_count() vs test_indomnum_and_count_by_flags()");
       ));
}


\\-----------------------------------------------------------------------------
progress("vpar_indomsets()");

test_indomsets_by_flags(vpar) =
{
  my(l=List([]));
  forvec(flags=vector(#vpar,i,[0,1]),
        if(vpar_is_indomset_flags(vpar,flags),
           listput(l,test_flags_to_Set(flags))));
  Vec(l);
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        want_count=vpar_indomsets_count(vpar));
     if(want_count > test_limit_sets, next());

     my(got = vpar_indomsets(vpar));
     check_equal(#got, want_count);
     for(i=1,#got,
        check_equal(setisset(got[i]), 1);
        check_equal(vpar_is_indomset(vpar,got[i]), 1));

     check_equal(vecsort(got), vecsort(test_indomsets_by_flags(vpar)));
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_is_domset_flags()");

check_equal(vpar_is_domset_flags([],[]), 1, "vpar_is_domset_flags() empty");
check_equal(vpar_is_domset_flags([0,1,2,3,4],Vecsmall([1,0,1,0,1])), 1);
check_equal(vpar_is_domset_flags([0,1,2,3,4],[1,1,0,0,1]), 1);
check_equal(vpar_is_domset_flags([0,1,2,3,4],[1,0,0,0,0]), 0);
check_equal(vpar_is_domset_flags([0,1,2,3,4],[1,1,1,1,0]), 1);
check_equal(vpar_is_domset_flags(vpar_make_path(2), [1,0]), 1);

test_is_domset_flags_by_passes(vpar,flags) =
{
  my(dominated=vectorsmall(#vpar));
  for(v=1,#vpar,
     my(p=vpar[v]);
     if(flags[v],
        dominated[v]=1;           \\ v dominates itself
        if(p, dominated[p]=1));   \\ v dominates p
     if(p && flags[p],
        dominated[v]=1));         \\ p dominates v

  for(v=1,#vpar, if(!dominated[v],return(0)));
  1;
}
test_is_domset_flags_by_hammingweight(vpar,flags) =
{
  my(dominated=Vecsmall(flags,#vpar));  \\ all of flags itself dominated
  for(v=1,#flags,
     my(p=vpar[v]);
     if(p,
        dominated[p]+=flags[v];     \\ p dominated by v
        dominated[v]+=flags[p]));   \\ v dominated by p
  hammingweight(dominated)==#vpar;
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     for(rep=1,max(1,#vpar),
        my(flags=vpar_random_flags(vpar));
        check_equal(vpar_is_domset_flags(vpar,flags),
                    test_is_domset_flags_by_passes(vpar,flags));
        check_equal(vpar_is_domset_flags(vpar,flags),
                    test_is_domset_flags_by_hammingweight(vpar,flags))));
}


\\-----------------------------------------------------------------------------
progress("vpar_minimal_dompoly()");

check_equal(vpar_minimal_dompoly([]),    Pol(1), \
            "vpar_minimal_dompoly() empty");
check_equal(type(vpar_minimal_dompoly([])), "t_POL", "empty");

check_equal(vpar_minimal_dompoly([0]),          x, \
            "vpar_minimal_dompoly() singleton");
check_equal(vpar_minimal_dompoly([0,1]),        2*x, \
            "vpar_minimal_dompoly() path-2");
check_equal(vpar_minimal_dompoly([0,1,1,1,1]),  x^4 + x, \
            "vpar_minimal_dompoly() star-5");
check_equal(vpar_minimal_dompoly([0,1,2,3,4]),  x^3 + 3*x^2, \
            "vpar_minimal_dompoly() path-5");
check_equal(vpar_minimal_dompoly([0,1,2,3,4],'y),  'y^3 + 3*'y^2, \
            "vpar_minimal_dompoly() path-5 y");

test_minimal_dompoly_by_flags(vpar) =
{
  my(poly=Pol(0));
  forvec(flags=vector(#vpar,i,[0,1]),
        if(vpar_is_minimal_domset_flags(vpar,flags),
           poly += x^vecsum(flags)));
  poly;
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        poly=vpar_minimal_dompoly(vpar));
     check_equal(type(poly), "t_POL");

     check_equal(valuation(poly,'x), vpar_domnum(vpar));
     check_equal(polcoeff(poly,valuation(poly,'x)),vpar_domnum_and_count(vpar)[2]);
     check_equal(simplify(subst(poly,'x,1)), vpar_minimal_domsets_count(vpar));

     if(#vpar <= test_limit_flags_search,
        check_equal(poly,
                    test_minimal_dompoly_by_flags(vpar),
                    Str("vpar_minimal_dompoly() "
                        "vs test_minimal_dompoly_by_flags()")));

     my(component_vpar=vpar_component_trees(vpar));
     check_equal(poly,
                 Pol(prod(i=1,#component_vpar, vpar_minimal_dompoly(component_vpar[i]))),
                 "vpar_minimal_dompoly() product of components");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_num_edges() total shown in addhelp");

{
  my(forest=0);
  for(n=0, test_limit_trees_search - forest,
     my(total=0);
     forvec(vec=vpar_forvec_from(n,forest), \\ trees
           my(vpar=vpar_from_upascend(vec));
           total+=vpar_num_edges(vpar));
     check_equal(total, if(n==0,0, (n-1)*vpar_count(n)),
                 "vpar_num_edges() trees total"));
}

{
  my(forest=1);
  for(n=0, test_limit_trees_search - forest,
     my(total=0);
     forvec(vec=vpar_forvec_from(n,forest), \\ forests
           my(vpar=vpar_from_upascend(vec));
           total+=vpar_num_edges(vpar));
     check_equal(total, (n-1)*n*(n+1)^(n-2),
                 "vpar_num_edges() forests total"));
}
check_equal(vector(#A065513_samples,n,n++; (n-1)*n*(n+1)^(n-2)), \
            A065513_samples);


\\-----------------------------------------------------------------------------
progress("vpar_is_free_isomorphic()");

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        vpar2=vpar_relabel_random(vpar));
     vpar2=vpar_reroot_random(vpar2);

     check_equal(vpar_is_free_isomorphic(vpar,vpar2),1,
                 "vpar_is_free_isomorphic() random relabel and reroot"));
}


\\-----------------------------------------------------------------------------
progress("vpar_is_hereditaryleastsingle()");

{
  \\ Samples from Chunwei Song's thesis end ofsection 3.2 page 64.
  \\ https://www.math.upenn.edu/grad/dissertations/SongThesis.pdf
  \\ not in OEIS: 4,15,96,665,6028,60907,725560
  my(want=[1,1,1,4,15,96,665,6028,60907,725560]);
  check_equal(vector(#want,n,vpar_count_hereditaryleastsingle(n,-1)), want,
             "vpar_count_hereditaryleastsingle() root=1 samples");
}

test_is_hereditaryleastsingle_by_mins(vpar) =
{
  my(min_child=vpar_minmax_child_vector(vpar,min));
  for(v=1,#vpar,
     my(c=min_child[v]);
     if(c && min_child[c], return(0)));  \\ least child has a child
  1;
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        roots=vpar_roots(vpar));
     check_equal(vpar_is_hereditaryleastsingle(vpar),
                 test_is_hereditaryleastsingle_by_mins(vpar)));
}


\\-----------------------------------------------------------------------------
progress("vpar_is_leastchildmonk()");

check_equal(vpar_is_leastchildmonk([]), 1, "vpar_is_leastchildmonk() empty");
check_equal(vpar_is_leastchildmonk([0]),1,"vpar_is_leastchildmonk() singleton");
check_equal(vpar_is_leastchildmonk([0,0]), 1);

\\ n=2 forests
\\    1    2     1 2
\\    |    |
\\    2    1
check_equal(vpar_is_leastchildmonk([0,1]),1);
check_equal(vpar_is_leastchildmonk([2,0]),1);
check_equal(vpar_is_leastchildmonk([0,0]),1);
check_equal(vpar_count(2,1), 3);  \\ n=3 all forests

\\ Example n=4 trees root=0 in Chunwei Song's thesis figure 3.2 page 47.
\\ https://www.math.upenn.edu/grad/dissertations/SongThesis.pdf
\\ Numbered here +1 so root=1.
\\
\\      1         1        1         1
\\    / | \      / \      / \       / \
\\   2  3  4    2   3    2   4     3   4
\\                  |        |         |
\\                  4        3         2
check_equal(vpar_is_leastchildmonk([0,1,1,1]), 1, \
            "vpar_is_leastchildmonk() 4 star");
check_equal(vpar_is_leastchildmonk([0,1,1,3]), 1, \
            "vpar_is_leastchildmonk() 4 second");
check_equal(vpar_is_leastchildmonk([0,1,4,3]), 1);
check_equal(vpar_is_leastchildmonk([0,4,1,1]), 1);

\\ 2 <- 1 <- 3
check_equal(vpar_is_leastchildmonk([2,0,1]), 0);
\\ 1 <- 3 <- 2
check_equal(vpar_is_leastchildmonk([0,3,1]), 0);

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        roots=vpar_roots(vpar));
     check_equal(vpar_is_leastchildmonk(vpar),
                 test_is_leastchildmonk_by_min_child(vpar)));
}


\\-----------------------------------------------------------------------------
progress("vpar_count_leastchildmonk()");

check_equal(vpar_count_leastchildmonk(0,-1), 1, \
            "vpar_count_leastchildmonk() n=0 trees root=1");
check_equal(vpar_count_leastchildmonk(0,0), 1, \
            "vpar_count_leastchildmonk() n=0 trees");
check_equal(vpar_count_leastchildmonk(0,1), 1);

check_equal(vpar_count_leastchildmonk(1,-1), 1);
check_equal(vpar_count_leastchildmonk(1,0),  1);
check_equal(vpar_count_leastchildmonk(1,1),  1);

check_equal(vpar_count_leastchildmonk(2,-1), 1);
check_equal(vpar_count_leastchildmonk(2,0),  2);
check_equal(vpar_count_leastchildmonk(2,1),  3);

\\ Example n=4 trees root=0 in Chunwei Song's thesis figure 3.2 page 47.
\\ https://www.math.upenn.edu/grad/dissertations/SongThesis.pdf
check_equal(vpar_count_leastchildmonk(4,-1), 4);

{
  for(forest=-1,1,
     for(n=0, test_limit_trees_search - forest,
        my(count=0);
        forvec(vec=vpar_forvec_from(n,forest), \\ forests or trees
              my(vpar=vpar_from_upascend(vec));
              count+=vpar_is_leastchildmonk(vpar));
        check_equal(count, vpar_count_leastchildmonk(n,forest),
        "vpar_count_leastchildmonk() vs count vpar_is_leastchildmonk()");
     ));
}


\\-----------------------------------------------------------------------------
progress("vpar_count_leastchildmonk_and_rootmin()");

\\ n=2 trees
check_equal(vpar_is_leastchildmonk([0,1]), 1);
check_equal(vpar_is_leastchildmonk([2,0]), 1);
check_equal(vpar_is_root_minmax([0,1],min), 1);
check_equal(vpar_is_root_minmax([2,0],min), 0);

{
  \\ Sample forest values in Chunwei Song's thesis figure 3.2 page 50.
  \\ https://www.math.upenn.edu/grad/dissertations/SongThesis.pdf
  \\ Shown as 1,2,5,18,93,104 but believe the last is a misprint should be 704.
  \\ not in OEIS: 2,5,18,93,704
  my(want=[1,  1,2,5,18,93,704]);
  check_equal(vector(#want,n,n--; vpar_count_leastchildmonk_and_rootmin(n,1)),
              want,
              "vpar_count_leastchildmonk_and_rootmin() vs samples");
}

\\ FIXME: This not right.
\\ {
\\   my(n=5,
\\      t=test_func_to_egf(vpar_count_leastchildmonk_and_rootmin, n),
\\      f=test_func_to_egf(i->vpar_count_leastchildmonk_and_rootmin(i,1), n));
\\   print(t);
\\   print(f);
\\   check_equal(exp(t), f);
\\ }

test_count_leastchildmonk_and_rootmin_forests_by_compositions(n) =
{
  \\ Formula in Chunwei Song thesis corollary 3.2.5 page 50.
  \\ https://www.math.upenn.edu/grad/dissertations/SongThesis.pdf
  \\ The sum shown is over compositions.
  \\ The abs ensures when v[i]==1 that get (v[i]-2)^(v[i]-2) = 1.

  my(total=0, comps=test_compositions(n));
  for(i=1,#comps,
     my(v=comps[i]);
     total += test_multinomial(v)
             * prod(i=1,#v, abs((v[i]-2)^(v[i]-2)))
             / (#v)!);
  total;
}
{
  check_equal(vector(10,n,n--; vpar_count_leastchildmonk_and_rootmin(n,1)),
              vector(10,n,n--; test_count_leastchildmonk_and_rootmin_forests_by_compositions(n)),
  Str("vpar_count_leastchildmonk_and_rootmin() vs "
      "test_count_leastchildmonk_and_rootmin_forests_by_compositions()"));
}

{
  for(forest=0,1,
     for(n=0, test_limit_trees_search - forest,
        my(count=0);
        forvec(vec=vpar_forvec_from(n,forest),
              my(vpar=vpar_from_upascend(vec));
              count += vpar_is_leastchildmonk(vpar)
                      && vpar_is_root_minmax(vpar,min));
        check_equal(count, vpar_count_leastchildmonk_and_rootmin(n,forest),
                    Str("vpar_count_leastchildmonk_and_rootmin() "
                        "vs count predicates, n="n" forest="forest))));
}


\\-----------------------------------------------------------------------------
progress("vpar_root()");

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        roots=vpar_roots(vpar));
     check_equal(vpar_root(vpar),
                 if(#roots,roots[1],0)));
}


\\-----------------------------------------------------------------------------
progress("vpar_add_root_above_forest()");

check_equal(vpar_add_root_above_forest([]),[0]);
check_equal(vpar_add_root_above_forest([0]),[2,0]);
check_equal(vpar_add_root_above_forest([0,0,0]),[4,4,4,0]);

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        added=vpar_add_root_above_forest(vpar));
     check_equal(vpar_roots(added), [#vpar+1]));
}


\\-----------------------------------------------------------------------------
progress("vpar_reverse_cycles()");

check_equal(vpar_reverse_cycles([]), []);
check_equal(vpar_reverse_cycles([0]), [0]);
check_equal(vpar_reverse_cycles([0,1,2,3,4]), [0,1,2,3,4]);
check_equal(vpar_reverse_cycles([1]), [1]);
check_equal(vpar_reverse_cycles([1,2]), [1,2]);
check_equal(vpar_reverse_cycles([2,1]), [2,1]);
check_equal(vpar_reverse_cycles([3,1,2]), [2,3,1]);

\\   1 <- 2 <- 3 <- 4 <- 5 <- 6      1 -> 2 -> 3 -> 4 -> 5 -> 6
\\   |                        ^      ^                        |
\\   +-----------------------/        \-----------------------+
check_equal(vpar_reverse_cycles([6,1,2,3,4,5]), [2,3,4,5,6,1]);

\\   +--------\           /--------+
\\   |         v         v         |
\\   4 <- 6 <- 9         4 -> 6 -> 9
\\   |    |    |         |    |    |
\\   3    2    1         3    2    1
\\        |   / \             |   / \
\\        5  7   8            5  7   8
check_equal(vpar_reverse_cycles([9,6,4,9,2,4,1,1,6]), \
                                [9,6,4,6,2,9,1,1,4]);

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     check_equal(vpar_reverse_cycles(vpar), vpar));
}
{
  for(n=0,15,
     my(vpar=vector(n,i, random(n+1)));  \\ 0..n
     check_equal(vpar_reverse_cycles(vpar_reverse_cycles(vpar)),
                 vpar));
}


\\-----------------------------------------------------------------------------
progress("vpar_forvec_from()");

{
  my(l=List([]));
  forvec(v=[], listput(l,v));     \\ empty forvec gives single empty v
  check_equal(l, List([ [] ]));
}
{
  \\ n=0 root=1 giving single empty tree
  my(l=List([]));
  forvec(v=vpar_forvec_from(0,-1), listput(l,v));
  check_equal(l, List([ [] ]));
}

check_equal(vpar_forvec_from(0,-1), []);  \\ one empty
check_equal(vpar_forvec_from(0, 0), []);
check_equal(vpar_forvec_from(0, 1), []);

check_equal(vpar_forvec_from(1,-1), [[0,0]]);
check_equal(vpar_forvec_from(1, 0), [[0,0]]);
check_equal(vpar_forvec_from(1, 1), [[0,0]]);

check_equal(vpar_forvec_from(3,-1), [[1,3], [1,1], [0,0]], \
            "vpar_forvec_from() n=3 trees, root=1");
check_equal(vpar_forvec_from(3, 0), [[1,3], [1,3], [0,0]], \
            "vpar_forvec_from() n=3 trees");
check_equal(vpar_forvec_from(3, 1), [[0,3], [0,3], [0,0]], \
            "vpar_forvec_from() n=3 forests");


\\-----------------------------------------------------------------------------
progress("vpar_reroot_minmax()");

check_equal(vpar_reroot_minmax([],min), []);
check_equal(vpar_reroot_minmax([],max), []);
check_equal(vpar_reroot_minmax([0],min), [0]);
check_equal(vpar_reroot_minmax([0],max), [0]);
check_equal(vpar_reroot_minmax([0,1,0,3],min), [0,1,0,3]);
check_equal(vpar_reroot_minmax([0,1,0,3],max), [2,0,4,0]);


\\-----------------------------------------------------------------------------
progress("vpar_reroot_random()");

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     vpar_validate(vpar_reroot_random(vpar)));
}

\\-----------------------------------------------------------------------------
progress("vpar_count_contiguous_forests()");

\\ vector(10,n,vpar_count_contiguous_forests(n))

check_equal(vpar_count_contiguous_forests(0),1);
check_equal(vpar_count_contiguous_forests(1),1);
check_equal(vpar_count_contiguous_forests(2),3, \
            "vpar_count_contiguous_forests() n=2");

test_count_contiguous_forests_by_compositions(n) = \
  vecsum(apply(c->prod(i=1,#c, vpar_count(c[i])), test_compositions(n)));
check_equal(vector(7,n, vpar_count_contiguous_forests(n)), \
            vector(7,n, test_count_contiguous_forests_by_compositions(n)), \
            "vpar_count_contiguous_forests() vs compositions");

test_count_contiguous_forests_by_explicit_transform(n) =
{
  \\ Component tree of t vertices is t^(t-1) many, and preceding it n-t
  \\ many vertices of increasing component.
  \\ a[1] is count for n=0, etc, so i index is 1 bigger.

  my(a=vector(n,i, vpar_count(i)),   \\ each trees count
     b=vector(n+1));                 \\ output
  b[1]=1;  \\ n=0
  for(i=2,n+1, b[i]=sum(t=1,i-1, b[i-t]*a[t]));
  b[n+1];
}
{
  check_equal
  (vector(20,n, vpar_count_contiguous_forests(n)),
   vector(20,n, test_count_contiguous_forests_by_explicit_transform(n)),
   Str("vpar_count_contiguous_forests() "
       "vs test_count_contiguous_forests_by_explicit_transform()"));
}

\\-----------------------------------------------------------------------------
progress("vpar_reroot()");

check_equal(vpar_reroot(vpar_make_path(2),1),[0,1], "var_reroot()");
check_equal(vpar_reroot(vpar_make_path(2),2),[2,0], "var_reroot()");
check_equal(vpar_reroot(vpar_make_path(5),4),[2,3,4,0,4]);


\\-----------------------------------------------------------------------------
progress("vpar_reroot_forest()");

test_reroot_forest_by_individual(vpar,roots) = \
  for(i=1,#roots, vpar=vpar_reroot(vpar,roots[i])); vpar;
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        roots=vpar_random_set(vpar));
     check_equal(vpar_reroot_forest(vpar,roots),
                 test_reroot_forest_by_individual(vpar,roots)));
}


\\-----------------------------------------------------------------------------
progress("vpar_centres_of_forest()");

\\ "centres" is a vector of vectors of vertices [ [c1,c2], [c1,c2], [c1] ]
\\ Return it with each element [c1,c2] sorted in ascending order.
test_sort_centres_of_forest(centres) =
{
  \\ print("test_sort_centres_of_forest() "centres);
  vecsort(apply(c->vecsort(c),centres));
}

check_equal(vpar_centres_of_forest([]), []);
check_equal(vpar_centres_of_forest([0]), [[1]]);
check_equal(test_sort_centres_of_forest(vpar_centres_of_forest([0,1])), \
            [[1,2]]);
check_equal(vpar_centres_of_forest([0,1,2]), [[2]]);

test_centres_of_forest_by_eccentricities(vpar) =
{
  my(eccs=vpar_eccentricities_vector(vpar),
     root=vector(#vpar,v,vpar_root_of_vertex(vpar,v)),
     root_to_min_ecc=vector(#vpar,v,#vpar),
     root_to_centres=vector(#vpar,v,List([])));
  for(v=1,#vpar,
     my(r=root[v]);
     root_to_min_ecc[r]=min(root_to_min_ecc[r],eccs[v]));
  \\ print("root_to_min_ecc "root_to_min_ecc);
  for(v=1,#vpar,
     \\ print("v="v" ecc "eccs[v]);
     my(r=root[v]);
     if(eccs[v]==root_to_min_ecc[r],
        listput(root_to_centres[r],v)));
  test_sort_centres_of_forest(apply(Vec,select(c->#c,root_to_centres)));
}
check_equal(test_centres_of_forest_by_eccentricities([0,1,2]), [[2]]);

{
  my(vpar=[7, 1, 0, 7, 6, 7, 3]);
  check_equal(vpar_centres_of_forest(vpar), [[7]],
              "vpar_centres_of_forest() of 7");
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     check_equal
     (test_sort_centres_of_forest(vpar_centres_of_forest(vpar)),
      test_centres_of_forest_by_eccentricities(vpar),
      "vpar_centres_of_forest() vs test_centres_of_forest_by_eccentricities()"));
}


\\-----------------------------------------------------------------------------
progress("vpar_INTERNAL_seq_bucket_sort()");

check_equal(vpar_INTERNAL_seq_bucket_sort([0,0,0,0]), Vecsmall([1,2,3,4]), \
            "vpar_INTERNAL_seq_bucket_sort() zeros");

check_equal(vpar_INTERNAL_seq_bucket_sort([0,3,3,0]), Vecsmall([1,4,2,3]), \
            "vpar_INTERNAL_seq_bucket_sort() 2,2");

check_equal(vpar_INTERNAL_seq_bucket_sort([0,0,3,3,3,0]), \
            Vecsmall([1,2,6, 3,4,5]), \
            "vpar_INTERNAL_seq_bucket_sort() 3,3");
check_equal(vpar_INTERNAL_seq_bucket_sort([0,0,3,3,3,0],1), \
            Vecsmall([3,4,5, 1,2,6]), \
            "vpar_INTERNAL_seq_bucket_sort(),down=1 3,3");


\\-----------------------------------------------------------------------------
progress("vpar_INTERNAL_random_perm()");

\\ perm is a permutation.
\\ Return a vector of the random numbers which vpar_INTERNAL_random_perm()
\\ would need to generate this permutation.  These numbers are successive
\\ swaps of position i with somewhere i through n, as an offset r=0 to n-i
\\ inclusive.
test_perm_to_swaps(perm) =
{
  if(#perm<2, return([]));
  my(undone=test_identity_perm(#perm),
     ret=vector(#perm-1));
  for(i=1,#perm-1,
     my(t=perm[i], pos=i);
     while(undone[pos]!=t, pos++);
     \\ print("t="t" at pos="pos" in undone="Vec(undone));
     my(t=undone[i]); undone[i]=undone[pos]; undone[pos]=t;
     ret[i]=pos-i);
  ret;
}
check_equal(test_perm_to_swaps([1,2]), [0], "test_perm_to_swaps()");
check_equal(test_perm_to_swaps([2,1]), [1]);
check_equal(test_perm_to_swaps([1,2,3]), [0,0], "test_perm_to_swaps()");
check_equal(test_perm_to_swaps([2,1,3]), [1,0], "test_perm_to_swaps()");
check_equal(test_perm_to_swaps([1,3,2]), [0,1], "test_perm_to_swaps()");
{
  \\ all swaps vectors arise for some permutation
  for(n=0,5,
     my(l=List([]));
     for(p=0,n!-1,
        my(perm=Vecsmall(numtoperm(n,p)),
           swaps=test_perm_to_swaps(perm),
           pos=setsearch(l,swaps,1));
        check_equal(pos!=0, 1, "test_perm_to_swaps() no duplication");
        listinsert(l,swaps,pos));

     my(want=List([]));
     forvec(swaps=vector(max(0,n-1),i, [0,n-i]),
           listput(want,swaps));
     check_equal(#l,#want);
     check_equal(l,want);
  );
}

\\ vpar_INTERNAL_random_perm() calling test_random()
test_vpar_INTERNAL_random_perm = \
  eval(test_random_to_test_random(vpar_INTERNAL_random_perm));

{
  \\ vpar_INTERNAL_random_perm() generates all permutations for suitable
  \\ random numbers
  for(n=0,5,
     if(option_verbose,print("n="n));
     for(p=0,n!-1,
        my(want=Vecsmall(numtoperm(n,p)));
        check_equal(test_is_perm(want),1);

        test_random_list      =List(test_perm_to_swaps(want));
        test_random_limit_list=List(if(n<2,[],vector(n-1,i,n-i+1)));
        if(option_verbose,
           print(" want="want);
           print(" list  ="test_random_list);
           print(" limits="test_random_limit_list));

        my(got=test_vpar_INTERNAL_random_perm(n));
        check_equal(#test_random_list, 0, "test_random_list consumed");
        check_equal(#test_random_limit_list, 0, "test_random_limit_list consumed");

        check_equal(test_is_perm(got),1);
        check_equal(got,want, "test_vpar_INTERNAL_random_perm()")));
}


\\-----------------------------------------------------------------------------
progress("vpar_seq_upqueue()");

{
  \\ Caminiti thesis section 2.2 figure 2.5 example, Third Neville code
  \\             2
  \\          /     \
  \\         7       1
  \\       / |  \    |
  \\      6 10   9   8
  \\    / |  |
  \\   5  3  4
  \\
  my(vpar=[2,0,6,10,6,7,2,1,7,7]);
  vpar_validate(vpar);
  my(seq=vpar_seq_upqueue(vpar));
  check_equal(seq,Vecsmall([3,4,5,8,9,10,6,1,7,2]));
  my(vec=vecextract(vpar,seq));
  check_equal(vec,[6,10,6,1,7,7,7,2,2, 0]);
}

\\ vpar_seq_upqueue() is sort by ascending subtree height, and among equal
\\ such by maximum vertex at that height.
\\
\\ Descendants which are not deepest do not cause their parent to go into
\\ the queue at that point so can be ignored.  The maximum at height was
\\ last in the initial queue, which determines where the vertex falls among
\\ queue phases.
\\
test_seq_upqueue_by_sort(vpar) =
{
  my(subtree_heights=vpar_subtree_heights_vector(vpar),
     max_at_heights=test_minmax_at_height_vector(vpar,max),
     keys=vector(#vpar,v, [subtree_heights[v], max_at_heights[v], v]));
  vecsort(keys,lex,1);
}

\\ PENDING vpar_minmax_height_descendants_vector()
\\ Will drop the "proper" parameter.
test_minmax_at_height_vector(vpar,func,proper=0) =
{
  my(seq=vpar_INTERNAL_seq_upwards(vpar),
     subtree_heights = vector(#vpar),
     descendants     = vector(#vpar,v,if(proper,0,v)));
  for(i=1,#seq,
     my(v=seq[i], p=vpar[v]);
     if(p,
        my(h=subtree_heights[v]+1);
        \\ print("at v="v" p="p"  vheight="h" pheight="subtree_heights[p]);
        if(h > subtree_heights[p],
           descendants[p]=descendants[v];
           if(proper && !descendants[p], descendants[p]=v);
           subtree_heights[p]=h;
           \\ print("  new biggest height at p");
           ,
           h == subtree_heights[p],
           descendants[p]=func(descendants[p],descendants[v]);
           \\ print("  equal biggest height at p, descendant to ",descendants[p]);
          )));
  descendants;
}

check_equal(vecsort([[0,1],[1,1]]), [[0,1],[1,1]]);
check_equal(vecsort([[0,1],[1,1]],lex,1), Vecsmall([1,2]));

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     my(by_upqueue = vpar_seq_upqueue(vpar),
        by_sort    = test_seq_upqueue_by_sort(vpar));

     \\ if(by_sort != by_upqueue,
     \\    my(subtree_heights=vpar_subtree_heights_vector(vpar),
     \\       min_at_heights=vpar_minmax_height_descendants_vector(vpar,min),
     \\       max_at_heights=vpar_minmax_height_descendants_vector(vpar,max),
     \\       keys=vector(#vpar,v, [subtree_heights[v], min_at_heights[v]]));
     \\    print("subtree_heights ",subtree_heights);
     \\    print("min_at_heights ",min_at_heights);
     \\    print("max_at_heights ",max_at_heights);
     \\    print("keys ",keys);
     \\    print("by_upqueue ",by_upqueue);
     \\    vpar_view(vpar));

     check_equal(by_upqueue,by_sort,
                 "vpar_seq_upqueue() vs test_seq_upqueue_by_sort()"));
}

\\-----------------------------------------------------------------------------
progress("vpar_seq_subtree_size()");
\\ vpar_INTERNAL_exto_subtree_sizes_vector()

check_equal(vpar_subtree_sizes_vector(vpar_make_path(5)),[5,4,3,2,1]);

check_equal(vpar_subtree_sizes_vector([0,1]), [2,1]);
check_equal(vecsort([2,1],,1), Vecsmall([2,1]));
check_equal(vecsort([2,1],,1+4), Vecsmall([1,2]));  \\ down
check_equal(vecsort(-[2,1],,1), Vecsmall([1,2]));   \\ down

check_equal(vpar_seq_subtree_size([0,1]),   Vecsmall([2,1]));
check_equal(vpar_seq_subtree_size([0,1],1), Vecsmall([1,2]));

check_equal(vpar_INTERNAL_exto_subtree_sizes_vector([0]),  [1]);
check_equal(vpar_INTERNAL_exto_subtree_sizes_vector([0],1),[1]);

{
  my(vpar=[2,0]);    \\ 1 --> 2, down
  check_equal(vpar_subtree_sizes_vector(vpar),[1,2]);
  my(seq=vpar_seq_subtree_size(vpar,1));
  check_equal(seq,Vecsmall([2,1]));
  my(vec=vecextract(vpar,seq));
  check_equal(vec, [0,2]);
  my(subtree_sizes_again=vpar_INTERNAL_exto_subtree_sizes_vector(vec,1));
  check_equal(subtree_sizes_again, [1,2], "subtree_sizes_again");
}
{
  my(vpar=[0,1],     \\ 2 --> 1, down
     seq=vpar_seq_subtree_size(vpar,1));
  check_equal(seq,Vecsmall([1,2]));
  my(vec=vecextract(vpar,seq));
  check_equal(vec, [0,1]);
  my(subtree_sizes_again=vpar_INTERNAL_exto_subtree_sizes_vector(vec,1));
  check_equal(subtree_sizes_again, [2,1]);
}

{
  for(i=1,#test_sample_forests,
     for(down=0,1,
        my(vpar=test_sample_forests[i],
           subtree_sizes=vpar_subtree_sizes_vector(vpar),
           seq=vpar_seq_subtree_size(vpar,down),
           vec=vecextract(vpar,seq));

        my(subtree_sizes_again=vpar_INTERNAL_exto_subtree_sizes_vector(vec,down));
        check_equal(subtree_sizes_again, subtree_sizes);

        my(seq_again=vpar_INTERNAL_seqfrom_subtree_size(vec,down));
        check_equal(seq_again, seq);

        my(vpar_again=vpar_from_subtree_size(vec,down));
        check_equal(vpar_again, vpar);
     ));
}


\\-----------------------------------------------------------------------------
progress("vpar_is_cyclic()");

check_equal(vpar_is_cyclic([1,1,1]), 1);
check_equal(vpar_is_cyclic([0,1,1]), 0);

{
  \\ count how many vectors are cyclic and not
  for(n=1,5,
     my(count_all=0, count_cyclic=0, count_acyclic=0, count_acyclic_trees=0);
     forvec(vpar=vector(n,i,[0,n]),
           my(is_cyclic = vpar_is_cyclic(vpar),
              num_roots = vpar_num_roots(vpar));
           count_all++;
           count_cyclic        +=  is_cyclic;
           count_acyclic       += !is_cyclic;
           count_acyclic_trees += !is_cyclic && num_roots==1);
     check_equal(count_all,           (n+1)^n,     Str("count_all n="n));
     check_equal(count_cyclic,      n*(n+1)^(n-1), "count_cyclic");
     check_equal(count_acyclic,       vpar_count(n,1), "count_acyclic");
     check_equal(count_acyclic_trees, vpar_count(n),   "count_acyclic_trees");
     );
}


\\-----------------------------------------------------------------------------
progress("vpar_preorder_last()");

{
  for(forest=0,1,
     for(n=0,10-forest,
        my(vpar=vpar_preorder_last(n,forest));
        vpar_validate(vpar);
        if(!forest,
           check_equal(vpar_is_tree(vpar), 1));
        check_equal(vpar_is_preorder(vpar), 1);
        check_equal(vpar_preorder_next(vpar,forest), 'none);
        check_equal(vpar_depths_vector(vpar), vpar,
                    "vpar_preorder_last() vpar==depths");

        \\ PENDING vpar_num_ordered_isomorphic()
        \\ check_equal(vpar_num_ordered_isomorphic(vpar),
        \\             if(forest||n==0,1, n));
     ));
}


\\-----------------------------------------------------------------------------
progress("vpar_preorder_first()");

{
  for(n=0,10,
     my(vpar=vpar_preorder_first(n));
     vpar_validate(vpar);
     check_equal(vpar_is_preorder(vpar), 1);
     for(forest=0,1,
        check_equal(vpar_preorder_prev(vpar,forest), 'none));
     check_equal(vpar_depths_vector(vpar), vpar,
                 "vpar_preorder_first() vpar==depths");

     \\ PENDING vpar_num_ordered_isomorphic()
     \\ check_equal(vpar_num_ordered_isomorphic(vpar),
     \\             if(forest||n==0,1, n));
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_is_equimatchable()");

{
  \\   5
  \\  / \
  \\ 4   3
  \\ |
  \\ 2
  \\ |
  \\ 1
  my(vpar=[2, 4, 5, 5, 0]);
  check_equal(vpar_is_equimatchable(vpar), 1,
              "vpar_is_equimatchable() path-5 rerooted");
}
check_equal(vpar_is_equimatchable(vpar_make_path(3)), 1);
check_equal(vpar_is_equimatchable(vpar_make_path(4)), 0);
check_equal(vpar_is_equimatchable(vpar_make_path(5)), 1, \
              "vpar_is_equimatchable() path-5");
check_equal(vpar_is_equimatchable(vpar_make_path(6)), 0);
check_equal(vpar_is_equimatchable(vpar_make_path(7)), 0);

test_is_equimatchable_by_lower_matchnum(vpar) =
{
  \\ print("vpar_is_equimatchable_by_lower_matchnum() "vpar);

  my(seq=vpar_INTERNAL_seq_upwards(vpar),
     with               = vectorsmall(#vpar),
     with_exists        = with,
     without            = with,
     without_max        = with,
     without_max_exists = vectorsmall(#vpar,v,1),  \\ empty is maximal
     matched            = with,
     matchnum           = with);
  for(i=1,#seq,
     my(v=seq[i],
        p=vpar[v],
        v_max = min(with[v],
                          if(without_max_exists[v], without_max[v],
                             with[v])));

     \\ print(" v="v" with "with[v]" without "without[v]" max "without_max[v]"["without_max_exists[v]"]");
     \\ print("  v_max "v_max" matchnum "matchnum[v]);

     if(with[v] && with[v] < matchnum[v],
        \\ print("  early exit");
        return(0));

     if(p,
        matchnum[p] += matchnum[v];
        if(!matched[v] && !matched[p],
           matchnum[p]++;
           matched[p]=1);

        with[p] = min(if(with[p], with[p] + v_max,
                               #vpar),
                            without[p] + without[v] + 1);  \\ new

        without[p] += v_max;

        without_max[p] += with[v];
        if(with[v]==0,
           without_max_exists[p] = 0);

        \\ print("  set p="p" with "with[p]" without "without[p]" max "without_max[p]"["without_max_exists[p]"]");
        ,
        \\ print("  at root");
        if(v_max != matchnum[v],
           \\ print("  different");
           return(0));    \\ at root
       ));
  1;
}
test_is_equimatchable_by_matchpoly(vpar) =
{
  my(poly=vpar_maximal_matchpoly(vpar));
  poldegree(poly) == valuation(poly,'x);
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        matchnum_and_count = vpar_matchnum_and_count(vpar),
        matchnum       = matchnum_and_count[1],
        matchnum_count = matchnum_and_count[2],
        maximals_count = vpar_maximal_matchings_count(vpar),
        got            = vpar_is_equimatchable(vpar));
     \\ print(vpar);
     check_equal(matchnum_count <= maximals_count, 1);

     check_equal(got, matchnum_count==maximals_count,
                 "vpar_is_equimatchable() vs matchnum_count==maximals_count");
     check_equal(got, test_is_equimatchable_by_lower_matchnum(vpar),
                 "vpar_is_equimatchable() vs test_is_equimatchable_by_lower_matchnum()");
     check_equal(got, test_is_equimatchable_by_matchpoly(vpar),
                 "vpar_is_equimatchable() vs test_is_equimatchable_by_matchpoly()");

     \\ PENDING vpar_lower_matchnum() or whatever name
     \\ check_equal(got, matchnum==vpar_lower_matchnum(vpar),
     \\             "vpar_is_equimatchable() vs matchnum==lower_matchnum");
  );
}

{
  \\ infinite size equimatchable of degree <= 3
  \\ p=2 path-2
  \\      * 2
  \\      |
  \\      *
  \\ p=1 v=2 is p=3
  \\      * 3
  \\      |
  \\      *                          * 2
  \\      |                        / |
  \\      *                       *  * 3
  \\ p=1 v=3,v=1 is p=2              |
  \\      * 2                        * 2
  \\    / |                        / |
  \\   *  *                       *  *
  \\      |                          |
  \\      *                          *
  \\      |                          |
  \\      *                          *
  my(vpar=vpar_make_path(2));
  for(i=1,30,
     vpar_validate(vpar);
     check_equal(vpar_is_equimatchable(vpar), 1);
     vpar = vpar_join([0,1,1],2,vpar));
}

{
  \\ vpar_is_equimatchable() count

  \\          n = 1 2 3  4   5    6
  my(want       =[1,2,9,16,325,2196],
     want_state2=[0,2,3, 4,125, 726],
     forest=0);
  for(n=1, min(#want, test_limit_trees_search - forest),
     my(count=0,
        count_state2=0,
        count_ordered=0);
     forvec(vec=vpar_forvec_from(n,forest), \\ trees
           my(vpar=vpar_from_upascend(vec));
           if(vpar_is_equimatchable(vpar),
              count++;
              count_state2 += (test_equimatchable_get_state(vpar)==2);
              count_ordered += vpar_is_preorder(vpar)));
     \\ print(count_state2);
     check_equal(count, want[n],
                Str("vpar_is_equimatchable() count n="n));
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_INTERNAL_histogram()");

check_equal(vpar_INTERNAL_histogram([]), Vecsmall([]));
check_equal(vpar_INTERNAL_histogram([1,1,1,3]), Vecsmall([3,0,1]));
check_equal(vpar_INTERNAL_histogram([1,1,1,3],0), Vecsmall([0,3,0,1]));
check_equal(vpar_INTERNAL_histogram([4,4,4,5],4), Vecsmall([3,1]));


\\-----------------------------------------------------------------------------
progress("vpar_weights_vector()");
\\ vpar_weight_of_vertex()

test_weights_vector_by_individual(vpar) = \
  vector(#vpar,v,vpar_weight_of_vertex(vpar,v));
{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i],
        got =vpar_weights_vector(vpar),
        want=test_weights_vector_by_individual(vpar));
     if(got!=want,
        \\ vpar_view(vpar);
        print("vpar="vpar" children ",vpar_children_of_vertex(vpar,v)));
     check_equal(got,want,
                "vpar_weights_vector() vs individual vpar_weight_of_vertex()"));
}
  \\    print("at u="u"  "Vec(num_below),"  weight="weight);
  \\ print(vpar," v="v"  final "Vec(num_below),"  weight="weight);


\\-----------------------------------------------------------------------------
progress("vpar_indnum_and_count()");

check_equal(vpar_indnum_and_count([]),[0,1]); \\ one empty set
check_equal(vpar_indnum_and_count([0]),[1,1]);
check_equal(vpar_indnum_and_count([0,1]),[1,2]);  \\ path-2
check_equal(vpar_indnum_and_count([2,0]),[1,2]);  \\ path-2

test_indnum_and_count_by_flags(vpar) =
{
  my(indnum=0,count=1);
  if(#vpar,
     forvec(flags=vector(#vpar,i,[0,1]),
           if(vpar_is_indset_flags(vpar,flags),
              my(size=vecsum(flags));
              if(size>indnum,  indnum=size;count=1,  \\ new high
                 size==indnum, count++))));           \\ equal high
  [indnum,count];
}
check_equal(test_indnum_and_count_by_flags([]),[0,1], \
            "test_indnum_and_count_by_flags() empty");
check_equal(test_indnum_and_count_by_flags([0]),[1,1]);

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got=vpar_indnum_and_count(vpar));

     my(by_test=test_indnum_and_count_by_flags(vpar));
     my(indnum_by_func=vpar_indnum(vpar));
     check_equal(indnum_by_func,by_test[1], "vpar_indnum()");

     check_equal(got,by_test, "vpar_indnum_and_count()");

     my(component_vpar=vpar_component_trees(vpar),
        component_indnum_and_count=apply(vpar_indnum_and_count,component_vpar));
     check_equal(got,
                 [sum(i=1,#component_indnum_and_count, component_indnum_and_count[i][1]),
                  prod(i=1,#component_indnum_and_count, component_indnum_and_count[i][2])],
                 "vpar_indnum_and_count() components sum and product, per addhelp");
  );
}

\\ Jou and Lin example no duplicated leaves maximum indnum
{
  my(vpar=[0,1,2,3, 1,5,6, 1,8,9, 1,11,12, 1],
     n=#vpar);
  check_equal(vpar_indnum(vpar), floor((2*n-1)/3));
}


\\-----------------------------------------------------------------------------
progress("vpar_totdomsets_count()");

check_equal(vpar_totdomsets_count([]),1, "vpar_totdomsets_count() empty");
check_equal(vpar_totdomsets_count([0]),0, "vpar_totdomsets_count() singleton");
check_equal(vpar_totdomsets_count([0,0]),0, "vpar_totdomsets_count() two isolated");
check_equal(vpar_totdomsets_count([0,1,0,3]),1, "vpar_totdomsets_count() two pairs");

test_totdomsets_count_by_flags(vpar) =
{
  my(count=0);
  forvec(flags=vector(#vpar,i,[0,1]),
        count+=vpar_is_totdomset_flags(vpar,flags));
  count;
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     if(#vpar <= test_limit_flags_search,
        check_equal(vpar_totdomsets_count(vpar),
                    test_totdomsets_count_by_flags(vpar))));
}

{
  \\ number of total dominating sets in a complete binary tree
  check_equal(vector(6,h,h--;
                     vpar_totdomsets_count(vpar_make_complete_tree(2,h))),
              [0, 3, 25, 1088, 2059683, 9594697265625]);
  \\ not in OEIS: 3, 25, 1088, 2059683, 9594697265625
}
\\
\\ with(h)          = if(h==0,1, (with(h-1) + without(h-1))^2);
\\ with_undom(h)    = if(h==0,1, without(h-1)^2);
\\ with_dom(h)      = with(h) - with_undom(h);
\\ without(h)       = if(h==0,1, totdomsets(h-1)^2);
\\ without_undom(h) = if(h==0,1, without_dom(h-1)^2);
\\ without_dom(h)   = without(h) - without_undom(h);
\\ totdomsets(h)    = with_dom(h) + without_dom(h);
\\
\\ with(h)          = if(h==0,1,h==1,4, \
\\         ( totdomsets(h-2)^2 - with(h-1) )^2 );
\\ without_undom(h) = if(h==0,1,h==1,0, \
\\         ( totdomsets(h-2)^2 - without_undom(h-1) )^2);
\\ totdomsets(h)    = if(h==0,0,h==1,3, \
\\         with(h) - without_undom(h) + totdomsets(h-1)^2 - totdomsets(h-2)^4);
\\ vector(6,h,h--; totdomsets(h))
\\
\\ vector(6,h,h--; with(h))
\\ not in OEIS: 1, 4, 16, 625, 1562500, 7541856107536
\\ vector(6,h,h--; with_undom(h))
\\ not in OEIS: 81, 390625, 1401249857536
\\ vector(6,h,h--; without(h))
\\ not in OEIS: 9, 625, 1183744, 4242294060489
\\ vector(6,h,h--; without_undom(h))
\\ not in OEIS: 81, 295936, 788203044864


\\-----------------------------------------------------------------------------
progress("vpar_delete_vertices()");

check_equal(vpar_delete_vertices([],[]), [], "vpar_delete_vertices() empty");
check_equal(vpar_delete_vertices([0],[]), [0]);
check_equal(vpar_delete_vertices([0],[1]), []);
check_equal(vpar_delete_vertices([0,1,1,1,4],[1,2]), [0,0,2]);
check_equal(vpar_delete_vertices([0,1,1,1,4],[1]), [0,0,0,3]);
check_equal(vpar_delete_vertices([0,1,1,1,4],[5]), [0,1,1,1]);

check_equal(iferr(vpar_delete_vertices([0,1,1,1,4],[99]); 0, \
                  e, 1), 1, "vpar_delete_vertices() out of range is error");
check_equal(iferr(vpar_delete_vertices([0,1,1,1,4],[5,5]); 0, \
                  e, 1), 1, "vpar_delete_vertices() duplicate is error");
check_equal(iferr(vpar_delete_vertices([0,1,1,1,4],[1,3,1]); 0, \
                  e, 1), 1, "vpar_delete_vertices() duplicate is error");

\\ same as vpar_delete_vertices(), done by individual deletes
test_delete_vertices_by_individual(vpar,vec) =
{
  vec=vecsort(vec,,4);  \\ sort in descending order
  for(i=1,#vec, vpar=vpar_delete_vertex(vpar,vec[i]));
  vpar;
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        del=vpar_random_set(vpar));
     del = test_vector_ranperm(del);
     check_equal(vpar_delete_vertices(vpar,del),
                 test_delete_vertices_by_individual(vpar,del));
     );
}

\\-----------------------------------------------------------------------------
progress("vpar_keep_vertices()");

check_equal(vpar_keep_vertices([0,1,1,1,4],[]), [], \
            "vpar_keep_vertices() keep empty");
check_equal(vpar_keep_vertices([0,1,1,1,4],[1,2,3,5]), [0,1,1,0], \
            "vpar_keep_vertices()");
check_equal(vpar_keep_vertices([0,1,1,1,4],[5,4,3,2]), [0,0,0,3]);
check_equal(vpar_keep_vertices([0,1,1,1,4],[1,2,3,4]), [0,1,1,1]);

check_equal(iferr(vpar_keep_vertices([0,1,1,1,4],[99]); 0, \
                  e, 1), 1, "vpar_keep_vertices() out of range is error");
check_equal(iferr(vpar_keep_vertices([0,1,1,1,4],[5,5]); 0, \
                  e, 1), 1, "vpar_keep_vertices() duplicate is error");
check_equal(iferr(vpar_keep_vertices([0,1,1,1,4],[1,3,1]); 0, \
                  e, 1), 1, "vpar_keep_vertices() duplicate is error");

{
  \\ vpar_keep_vertices() vs vpar_delete_vertices()
  for(rep=1,10,
     my(vpar=vpar_make_random(10,1), \\ forest
        keep=vpar_random_set(vpar),
        del =vpar_set_complement(vpar,keep));
     keep = test_vector_ranperm(keep);
     del  = test_vector_ranperm(del);
     check_equal(vpar_delete_vertices(vpar,del),
                 vpar_keep_vertices(vpar,keep));
     );
}


\\-----------------------------------------------------------------------------
progress("vpar_keep_flags()");

check_equal(vpar_keep_flags([],[]), []);
check_equal(vpar_keep_flags([0],[]), [], "vpar_keep_flags() keep none");
check_equal(vpar_keep_flags([0],[0]), []);
check_equal(vpar_keep_flags([0],[1]), [0]);
check_equal(vpar_keep_flags([0,1,1,1,4],[0,0,0,0]), [], \
            "vpar_keep_flags() keep none");
check_equal(vpar_keep_flags([0,1,1,1,4],[0,0,0,0,0]), []);
check_equal(vpar_keep_flags([0,1,1,1,4],[1,0,1,1,1]), [0,1,1,3]);

{
  \\ vpar_keep_vertices() vs vpar_keep_flags()
  for(rep=1,10,
     my(vpar=vpar_make_random(10,1), \\ forest
        flags=vpar_random_flags(9),
        vertices=test_flags_to_Set(flags));
     vertices = test_vector_ranperm(vertices);
     check_equal(vpar_keep_flags(vpar,flags),
                 vpar_keep_vertices(vpar,vertices),
                "vpar_keep_flags() vs vpar_keep_vertices()");
     );
}


\\-----------------------------------------------------------------------------
progress("vpar_make_star()");

{
  for(n=0,5,
     my(vpar=vpar_make_star(n));
     vpar_validate(vpar);
     check_equal(#vpar,n));
}


\\-----------------------------------------------------------------------------
progress("vpar_make_bistar()");

{
  for(n=0,5,
     for(m=0,5,
        my(vpar=vpar_make_bistar(n,m));
        vpar_validate(vpar);
        check_equal(#vpar,n+m)));
}

check_equal(vpar_make_bistar(0,4), \
            vpar_make_star(4));
check_equal(vpar_make_bistar(4,0), \
            vpar_make_star(4));
check_equal(vpar_make_bistar(4,1), \
            vpar_make_star(5), "bistar 4,1");
{
  my(vpar=vpar_make_bistar(1,4));
  vpar=vpar_relabel_swap(vpar, 1,2);
  vpar=vpar_reroot(vpar,1);
  check_equal(vpar, vpar_make_star(5), "bistar 1,4");
}


\\-----------------------------------------------------------------------------
progress("vpar_from_upqueue()");

check_equal(vpar_from_upqueue([]),[]);
check_equal(vpar_from_upqueue([0]),[0]);
check_equal(vpar_seq_upqueue([0,1]),Vecsmall([2,1]));
check_equal(vpar_from_upqueue([1,0]),[0,1],"vpar_from_upqueue()");
check_equal(vpar_from_upqueue([2,0]),[2,0]);


\\-----------------------------------------------------------------------------
progress("vpar_seq_subtree_height()");

check_equal(vpar_seq_subtree_height([]), Vecsmall([]), \
            "vpar_seq_subtree_height() empty");
check_equal(vpar_seq_subtree_height([0]), Vecsmall([1]));
check_equal(vpar_seq_subtree_height([0,1]), Vecsmall([2,1]), \
            "vpar_seq_subtree_height() path-2");
check_equal(vpar_seq_subtree_height([0,1],1), Vecsmall([1,2]), \
            "vpar_seq_subtree_height(),down=1 path-2");

{
  \\    1     6
  \\         /  \
  \\       2     4
  \\             |
  \\             5
  \\             |
  \\             3
  my(vpar=[0, 6, 5, 6, 4, 0],
     subtree_heights = [0,0,0,2,1,3],
     seq = Vecsmall([1, 2, 3, 5, 4, 6]),
     vec = [0, 6, 5, 4, 6, 0]);
  check_equal(vpar_subtree_heights_vector(vpar), subtree_heights,
              "vpar_seq_subtree_height() six vpar_subtree_heights_vector");
  check_equal(vpar_seq_subtree_height(vpar), seq,
              "vpar_seq_subtree_height() six");
  check_equal(vecextract(vpar,seq),vec,
              "vpar_seq_subtree_height() six extract");
  check_equal(vpar_INTERNAL_seqfrom_subtree_height(vec), seq,
              "vpar_INTERNAL_seqfrom_subtree_height()");
  check_equal(vpar_from_subtree_height(vec), vpar,
              "vpar_from_subtree_height()");

  seq=vpar_seq_subtree_height(vpar,1);
  check_equal(seq,Vecsmall([6, 4, 5, 1,2,3]));
  vec=vecextract(vpar,seq);
  check_equal(vec, [0, 6, 4, 0, 6, 5]);
  check_equal(vpar_INTERNAL_exto_subtree_heights_vector(vec,1),
              subtree_heights);
}

{
  \\ Caminiti slides Encoding: Second Neville code
  \\ same in thesis
  \\             7 (3,7)
  \\       /     |      \        \
  \\  2 (2,2)    |         \       \
  \\  |          |          |         \
  \\  1 (1,1)   10 (1,10)   6 (1,6)      |
  \\  |          |         / \           |
  \\  8 (0,8)    4(0,4)  3(0,3) 5(0,5)  9(0,9)
  \\
  my(vpar=[2,7,6,10,6,7,0,1,7,7]);
  vpar_validate(vpar);
  check_equal(vpar_subtree_heights_vector(vpar),[1,2,0,0,0,1,3,0,0,1]);

  my(seq=vpar_seq_subtree_height(vpar));
  check_equal(seq, Vecsmall([3,4,5,8,9,1,6,10,2, 7]));

  my(vec=vecextract(vpar,seq));
  check_equal(vec, [6,10,6,1,7,2,7,7,7, 0]);
}
{
  \\    2     3
  \\    |     |
  \\    4     1
  \\          |
  \\          5
  my(vpar=[3, 0, 0, 2, 1],
     subtree_heights = [1,1,2,0,0],
     seq=Vecsmall([4,5, 1,2, 3]),
     vec=[2, 1, 3, 0, 0]);
  check_equal(vpar_subtree_heights_vector(vpar), subtree_heights,
              "vpar_seq_subtree_height() five vpar_subtree_heights_vector");
  check_equal(vpar_seq_subtree_height(vpar), seq,
              "vpar_seq_subtree_height() five");
  check_equal(vecextract(vpar,seq),vec,
              "vpar_seq_subtree_height() five extract");
  check_equal(vpar_INTERNAL_seqfrom_subtree_height(vec), seq,
              "vpar_INTERNAL_seqfrom_subtree_height()");
  check_equal(vpar_from_subtree_height(vec), vpar,
              "vpar_from_subtree_height()");

  seq=vpar_seq_subtree_height(vpar,1);
  check_equal(seq,Vecsmall([3, 1,2, 4,5]));
  vec=vecextract(vpar,seq);
  check_equal(vec,[0, 3,0, 2,1]);
  check_equal(vpar_INTERNAL_exto_subtree_heights_vector(vec,1),
              subtree_heights);
}

\\ Deo and Micikevicius sample tree, with 11 as the root
{
  my(vpar=[11,9,10,7,7,7,1,3,4,4,0]);
  \\ vpar_view(vpar);

  my(seq,vec);
  my(seq=vpar_seq_upascend(vpar));         \\ Neville's 1st
  my(vec=vecextract(vpar,seq));
  check_equal(vec,[9,7,7,3,10,4,4,7,1, 11,0]);

  seq=vpar_INTERNAL_seq_upwards(vpar);  \\ Neville's 3rd
  \\ print(seq);
  vec=vecextract(vpar,seq);
  check_equal(vec,[9,4,7,7,3,10,4,7,1, 11,0]);

  seq=vpar_seq_upqueue(vpar);  \\ Deo and Micikevicius
  check_equal(seq,Vecsmall([2,5,6,8, 9,3, 10,4,7,1,11]));

  \\ with centre 4 as root
  my(vpar4=[7,9,10,0,7,7,4,3,4,4,1]);
  check_equal(vpar4,vpar_reroot(vpar,4), "reroot");

  seq=vpar_seq_subtree_height(vpar4);
  vec=vecextract(vpar4,seq);
  check_equal(vec,[9,7,7,3,1,7,10,4,4, 4,0], "leafphases");

  seq=vpar_seq_upqueue(vpar4);
  \\ print(seq);
  vec=vecextract(vpar4,seq);
  \\ check_equal(vec,[9,7,7,3,1,7,10,4,4, 4,0], "upqueue");
}


test_seq_subtree_height_by_phases(vpar) =
{
  my(num_children=vpar_INTERNAL_num_children_vecsmall(vpar),
     seq=vectorsmall(#vpar),
     upto=0,                 \\ starting after pre-increment
     phase=1);

  for(v=1,#vpar,
     if(num_children[v]==0,  \\ leaf vertices
        num_children[v]=-1;
        seq[upto++]=v));     \\ pre-increment

  while(upto<#vpar,
        my(new_upto=upto);
        for(i=phase,upto,
           my(v=seq[i], p=vpar[v]);
           if(vpar[v] && num_children[p]--==0,
              seq[new_upto++]=p));  \\ pre-increment
        phase = upto+1; upto=new_upto;
        my(s=vecsort(seq[phase..upto]));
        for(i=1,#s, seq[phase+i-1]=s[i]));
  seq;
}
test_seq_subtree_height_by_sort(vpar) =
{
  my(dir=1, \\ if(up,-1,1)
     heights=vpar_subtree_heights_vector(vpar),
     \\ sort first by height then by vertex number
     keys=vector(#vpar,v, [dir*heights[v], v]));
  vecsort(keys,,1);
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        seq=vpar_seq_subtree_height(vpar));

     check_equal(seq, test_seq_subtree_height_by_phases(vpar),
                 "vpar_seq_subtree_height() vs by phases");
     check_equal(seq, test_seq_subtree_height_by_sort(vpar));
     );
}


\\-----------------------------------------------------------------------------
progress("vpar_INTERNAL_exponential_transform()");

\\ per addhelp() of vpar_count(),
\\ all labelled rooted trees -> all labelled rooted forests
check_equal(vpar_INTERNAL_exponential_transform(vector(20,n, vpar_count(n))), \
                                               vector(20,n, vpar_count(n,1)));
{
  my(limit = 20,
     T = test_func_to_egf(vpar_count, limit),
     F = test_func_to_egf(n->vpar_count(n,1), limit));
  \\ print("T      = "T);
  \\ print("exp(T) = "exp(T-1));
  \\ print("F      = "F);
  check_equal(exp(T-1), F);
}
{
  my(limit = 20,
     T = 1 + x + 456/2!*x^2 + 789/3!*x^3 + 1011/4!*x^4 + O(x^5));
  \\ print("T        = "T);
  \\ print("exp(T-1) = "exp(T-1));
  \\ print("exp(T-1) = "exp(T-1 - x));

  \\ TS = T - (intformal(exp(x))-intformal(intformal(exp(x))))
  \\ F = exp(T-1)
  \\ intformal(exp(x))-intformal(intformal(exp(x)))
  \\ S = exp(TS-1)
  \\ for(n=0,limit-1, print(polcoeff(TS,n)*n!))

  \\             inf      e^-t
  \\ eint1(x) = integral  ----  dt   = incgam(0,x)
  \\              x         t
}


\\ constant 1 -> Bell numbers
check_equal(vpar_INTERNAL_exponential_transform(vector(10,n, 1)), \
            [1, 2, 5, 15, 52, 203, 877, 4140, 21147, 115975]);

\\ n -> A000248 count forests height <=1
check_equal(vpar_INTERNAL_exponential_transform(vector(10,n, n)), \
            [1, 3, 10, 41, 196, 1057, 6322, 41393, 293608, 2237921]);

\\ n twice-> A007550 exponentiated twice
{
  my(v=vector(10,n, n));
  for(i=1,2, v=vpar_INTERNAL_exponential_transform(v));
  check_equal(v, [1,4,20,127,967,8549,85829,962308,11895252,160475855]);
}

\\ n^2 = A033462 exponential transform of squares
check_equal(vpar_INTERNAL_exponential_transform(vector(9,n, n^2)), \
            [1, 5, 22, 125, 836, 6277, 52396, 479593, 4757680]);

\\ vpar_INTERNAL_exponential_transform() recurrence becomes compositions of
\\ n for sizes of component trees, multinomial to choose the ways vertices
\\ can be assigned to those compositions, but divide out num components
\\ factorial since want to order components by smallest vertex they contain.
\\
test_exponential_transform_by_compositions(a) =
{
  vector(#a,n,
         my(comps=test_compositions(n), total=0);
         for(i=1,#comps,
            my(v=comps[i]);
            total += test_multinomial(v) * prod(i=1,#v, a[v[i]]) / (#v)!);
         total);
}
{
  for(reps=1,5,
     my(a=vector(8,i, random(20)));
     check_equal(vpar_INTERNAL_exponential_transform(a),
                 test_exponential_transform_by_compositions(a)));
}


\\-----------------------------------------------------------------------------
progress("vpar_num_singletons()");

check_equal(vpar_num_singletons([]), 0);
check_equal(vpar_num_singletons([0]), 1);
check_equal(vpar_num_singletons([0,1]), 0);
check_equal(vpar_num_singletons([0,0]), 2);

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        got=vpar_num_singletons(vpar));
     check_equal(got, sum(v=1,#vpar, vpar_is_singleton(vpar,v)),
                 "vpar_num_singletons() vs count vpar_is_singleton()");
     check_equal(got!=0, test_any_singletons(vpar),
                 "vpar_num_singletons() vs test_any_singletons()");
  );
}

{
  \\ vpar_num_singletons() total shown in its addhelp
  my(TotS = (n)-> if(n==0,0, vpar_count(n)));
  my(forest=1);
  for(n=0, test_limit_trees_search - forest,
     my(total=0,
       count_no_singletons=0);
     forvec(vec=vpar_forvec_from(n,forest), \\ forests
           my(vpar=vpar_from_upascend(vec),
              singletons=vpar_num_singletons(vpar));
           total += singletons;
           count_no_singletons += (singletons==0));

     check_equal(total, TotS(n));
     check_equal(count_no_singletons, vpar_count_no_singletons(n));
  );
}

check_equal(vector(30,n, vpar_count_no_singletons(n)), \
            vpar_INTERNAL_exponential_transform(vector(30,n, if(n<2,0, vpar_count(n)))));

check_equal(vector(30,n, vpar_count_no_singletons(n)), \
            vector(30,n, sum(k=0,n, (-1)^(n-k)*binomial(n,k)*vpar_count(k,1))), \
            "vpar_count_no_singletons() formula shown in its addhelp()");

\\ Sum_{j=1..n-1} binomial(n-1,j) (j+1)^j a(n-1-j) by Alois P. Heinz in A105785
test_count_no_singletons_by_recurrence(n) =
{
  if(n==0, 1,
      sum(t=2,n, binomial(n-1,t-1) * vpar_count(t) * vpar_count_no_singletons(n-t)));
}
{
  check_equal(vector(10,n,n--; test_count_no_singletons_by_recurrence(n)),
              vector(10,n,n--; vpar_count_no_singletons(n)),
              "vpar_count_no_singletons() vs recurrence");
}


{
  my(limit = 8,
     L = O(x^limit),
     gT = test_func_to_egf(vpar_count, limit),
     gS = test_func_to_egf(vpar_count_no_singletons, limit));
  \\ print("gT        = "gT);
  \\ print("exp(gT-1) = "exp(gT-1));
  \\ print("exp(gT-1) = "exp(gT-1 - x));
  \\ print("gS        = "gS);

  check_equal(intformal(exp(x+L)) + L,  exp(x+L) - 1,
              "intformal(exp(x))");
  check_equal(exp(gT-1-x), gS, "vpar_count_no_singletons() gS");

  my(gtotal = test_func_to_egf(n->if(n==0,0,vpar_count(n)), limit),
     gF = test_func_to_egf(n->vpar_count(n,1), limit),
     gF_times_n = test_func_to_egf(n->n*vpar_count(n-1,1), limit));
  check_equal(gtotal, gS*x*exp(x) + L,
              "vpar_count_no_singletons() gtotal");
  check_equal(gtotal, exp(gT-1-x)*x*exp(x) + L);
  check_equal(gtotal, x*exp(gT-1) + L);
  check_equal(gtotal, x*gF + L);        \\ x*g = shift up, times n
  check_equal(gF_times_n, x*gF + L);
  \\ check_equal(gF_times_n, gtotal);

  \\ TS = T - (intformal(exp(x))-intformal(intformal(exp(x))))
  \\ F = exp(T-1)
  \\ intformal(exp(x))-intformal(intformal(exp(x)))
  \\ S = exp(TS-1)
  \\ for(n=0,limit-1, print(polcoeff(TS,n)*n!))
}
{
  \\ P(x) = x
  my(limit = 8,
     L = O(x^limit),
     gOnes = test_func_to_egf(n->1, limit),
     gN    = test_func_to_egf(n->n, limit));
  check_equal(gOnes, exp(x) + L);
  check_equal(gN,    x*gOnes + L,  "gN");
  check_equal(gN,    x*exp(x) + L, "gN");
}


\\-----------------------------------------------------------------------------
progress("vpar_count_eccentricities_vectors()");

\\ samples from devel/try.gp
check_equal(vector(8,n,n--; vpar_eccentricities_vectors_count(n)), \
            [1, 1, 1, 3, 10, 45, 231, 1428]);

{
  \\ A032032 n labelled things partitioned into labelled buckets >=2 in each
  my(A032032 = n->A032032_samples[n+1],  \\ n=0 upwards
     limit = #A032032_samples,
     L = O(x^limit),
     gA032032 = test_func_to_egf(A032032, limit));

  check_equal(gA032032, 1/(2+x-exp(x+L)));
  check_equal(vpar_INTERNAL_ge2_vector(limit-1),
              A032032_samples[^1],
              "A032032");

  check_equal(vector(limit-1,n,n++; vpar_eccentricities_vectors_count(n)),
              vector(limit-1,n,n++; n*A032032(n-1) + n*(n-1)/2*A032032(n-2)));

  my(gE = test_func_to_egf(vpar_eccentricities_vectors_count, limit));
  check_equal(gE, 1 + x*(1+x/2)/(2+x-exp(x+L)) + L, "gE");
  check_equal(gE, 1 + x*(1+x/2)*gA032032       + L, "gE 2");
}


\\-----------------------------------------------------------------------------
progress("vpar_2height()");

check_equal(vpar_2height([]), -2, "vpar_2height() of empty");
check_equal(vpar_2height([0]), -1, "vpar_2height() of singleton");
check_equal(vpar_2height([0,1]), 0, "vpar_2height() of path-2");
check_equal(vpar_2height([0,0]), 0, "vpar_2height() of two singletons");
{
  for(n=0,10,
     check_equal(vpar_2height(vpar_make_path(n)), n-2,
                 "vpar_2height() of path is n-2"));
}

test_2height_by_vecmax(vpar) =
{
  my(depths=vecsort(vpar_depths_vector(vpar)));
  if(#depths>=2, depths[#depths-1], #depths-2);
}
test_2height_by_endpoints(vpar) =
{
  if(#vpar<2, #vpar-2,
     vpar_height(vpar) - (vpar_height_num_vertices(vpar) <= 1));
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     check_equal(vpar_2height(vpar),
                 test_2height_by_vecmax(vpar),
                 "vpar_2height() vs test_2height_by_vecmax()");

     check_equal(vpar_2height(vpar),
                 test_2height_by_endpoints(vpar),
                 "vpar_2height() vs test_2height_by_endpoints()");
  );
}

\\ vector(10,n, test_count_plantedplane(n))
\\ A000108 Catalan numbers
\\ Ordered rooted trees, n vertices not including the root.
test_count_plantedplane(n) = if(n<=2,1, 1/n*binomial(2*n-2,n-1));

{
  my(limit = 10,
     L = O(x^limit),
     gP = test_func_to_gf(n->test_count_plantedplane(n), limit));
  check_equal(gP, 1 + (1 - sqrt(1-4*x + L) )/2,   \\ ordinary gf
              "test_count_plantedplane() gf");
  check_equal(gP, (3 - sqrt(1-4*x + L))/2,
              "test_count_plantedplane() gf");

  \\ for(n=0,limit-1, print(polcoeff(gP,n)));
  \\ for(n=0,limit-1, print(polcoeff(gP,n)*n!));
  \\ for(n=0,limit-1, print(polcoeff(gPt,n)*n!));
}

\\ 0 1 2 12 120 1680 30240 665280 17297280 518918400
\\ A001813 Quadruple factorial numbers: a(n) = (2n)!/n!.


\\-----------------------------------------------------------------------------
progress("vpar_maximal_matchpoly()");

check_equal(vpar_maximal_matchpoly([]),  Pol(1));
check_equal(vpar_maximal_matchpoly([0]), Pol(1));
check_equal(vpar_maximal_matchpoly([0,1],'y), 'y);
check_equal(vpar_is_maximal_matching_flags([],[]), 1);

test_maximal_matchpoly_by_flags(vpar) =
{
  my(poly=Pol(0));
  forvec(flags=vector(#vpar,i,[0,1]),
        if(vpar_is_maximal_matching_flags(vpar,flags),
           poly += x^vecsum(flags)));
  poly;
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        poly=vpar_maximal_matchpoly(vpar));
     check_equal(type(poly),"t_POL");

     if(#vpar <= test_limit_flags_search,
        check_equal(poly, test_maximal_matchpoly_by_flags(vpar),
        "vpar_maximal_matchings_count() vs test_maximal_matchpoly_by_flags()"));

     \\ in the addhelp
     check_equal(poly!=0, 1);
     check_equal(poldegree(poly), vpar_matchnum(vpar));
     check_equal(pollead(poly), vpar_matchnum_and_count(vpar)[2]);
     check_equal(simplify(subst(poly,'x,1)), vpar_maximal_matchings_count(vpar));

     \\ As noted by for example
     \\     Tomislav Doslic, Ivana Zubac, "Counting Maximal Matchings In
     \\     Linear Polymers", Ars Mathematica Contemporanea, volume 11,
     \\     2016, pages 255-276, http://amc-journal.eu
     \\ there are maximal matchings of all sizes from smallest matching up
     \\ to matchnum, no gaps.
     \\
     for(n=valuation(poly,'x), poldegree(poly),
        check_equal(polcoeff(poly,n)>0, 1));

     my(component_vpar=vpar_component_trees(vpar));
     check_equal(poly,
                 Pol(prod(i=1,#component_vpar, vpar_maximal_matchpoly(component_vpar[i]))),
                 "vpar_maximal_matchpoly() product of components");
  );
}

{
  \\ Number of maximal matchings in path of n edges, per
  \\     Tomislav Doslic, Ivana Zubac, "Counting Maximal Matchings In Linear
  \\     Polymers", Ars Mathematica Contemporanea, volume 11, 2016, pages
  \\     255-276, http://amc-journal.eu
  \\ A misprint there shows x^n for what believe should be x^k.
  for(n=3,12,
     my(vpar=vpar_make_path(n+1));
     check_equal(vpar_maximal_matchpoly(vpar),
                 sum(k=ceil(n/3),ceil(n/2), binomial(k+1,3*k-n)*x^k),
                 "vpar_maximal_matchpoly() of path"));

  check_equal(vector(20,n,n--; test_Padovan(n+6)),
            vector(20,n,n--; sum(k=ceil(n/3),ceil(n/2), binomial(k+1,3*k-n))));
}
\\ my(n=12); for(k=0,n, print(binomial(k+1,3*k-n)));


\\-----------------------------------------------------------------------------
progress("vpar_INTERNAL_invert_transform()");

check_equal(vpar_INTERNAL_invert_transform(vector(10,n,1)), \
            vector(10,n, 2^(n-1)));
{
  \\ A001906 every second Fibonacci
  my(samples=A001906_samples[^1]);  \\ not initial 0 in A001906
  check_equal(vpar_INTERNAL_invert_transform(vector(#samples,n, n)),
              samples,
  "vpar_INTERNAL_invert_transform() of n = every second Fibonacci A001906");
}
{
  \\ A051296 invert transform of factorials
  my(samples=A051296_samples[^1]);  \\ not initial extra 1 in A051296
  check_equal(vpar_INTERNAL_invert_transform(vector(#samples,n, n!)),
              samples, "A051296");

  my(A051296 = n->samples[n+1],  \\ n=0 upwards
     gA051296 = test_func_to_gf(A051296, #samples),
     limit = #samples,
     L = O(x^limit),
     gF = test_func_to_gf(n->n!, limit),
     gF1 = test_func_to_gf(n->if(n==0,0,n!), limit));

  check_equal(1+x*gA051296 + L, 1/(1-(gF-1)));
  check_equal(1+x*gA051296 + L, 1/(1-gF1));
  check_equal(gA051296 + L/x, 1/(1-gF1)/x - 1/x, "gA051296");
}

{
  \\ invert transform is -1 + 1/(1-g))
  \\
  my(limit=10,
     v = vector(limit,i, prime(i+20)),
     v_func = n-> if(n<1||n>#v,0, v[n]),
     g = test_func_to_gf(v_func,limit),
     inv = vpar_INTERNAL_invert_transform(v),
     inv_func = n-> if(n<1||n>#v,0, inv[n]),
     ginv = test_func_to_gf(inv_func,limit));

  check_equal(ginv, -1 + 1/(1-g));
}



\\-----------------------------------------------------------------------------
progress("vpar_count_height()");

{
  for(forest=-1,1,
     for(n=0, test_limit_trees_search - forest,
        my(offset=4,
           count=vector(n+offset+1));  \\ count[h+offset] for count of h
        forvec(vec=vpar_forvec_from(n,forest),
              my(vpar=vpar_from_upascend(vec),
                 height=vpar_height(vpar));
              count[height+offset]++);
        \\ print(count);

        for(i=1,#count,
           my(h=i-offset);
           check_equal(vpar_count_height(n,h,forest), count[i],
                       Str("vpar_count_height() vs forvec"
                           " n="n" h="h" forest="forest)));
        ));
}

{
  for(forest=-1,1,
     for(n=1,10,
        check_equal(sum(h=-2,n+1, vpar_count_height(n,h,forest)),
                    vpar_count(n,forest),
                    "vpar_count_height() total vpar_count()")));
}

\\ trees
for(n=-1,10, check_equal(vpar_count_height(n,-1), n==0));
for(n=-1,10, check_equal(vpar_count_height(n,0), n==1));
{
  for(n=-1,10,
     check_equal(vpar_count_height(n,1), if(n>=2,n,0),
                "vpar_count_height() trees height=1 is n"));
}

\\ forests
for(n=-1,10, check_equal(vpar_count_height(n,-1,1), n==0));
for(n=-1,10, check_equal(vpar_count_height(n,0,1), n>=1, \
                         "vpar_count_height() forest h=0"));
check_equal(vpar_count_height(-1,1,1), 0);
check_equal(vpar_count_height(0,1,1), 0);
check_equal(vpar_count_height(1,1,1), 0);
check_equal(vpar_count_height(2,1,1), 2);
\\ check_equal(vpar_count_height(3,1,1), vpar_count(3)-1);

{
  \\ h==n-1 is path only, with permutation of labels
  check_equal(vector(20,n,n--; vpar_count_height(n,n-1)),
              vector(20,n,n--; n!),
              "vpar_count_height() trees h=n-1 path permutations");
  check_equal(vector(20,n,n--; vpar_count_height(n,n-1,1)),
              vector(20,n,n--; n!),
              "vpar_count_height() forests h=n-1 path permutations");
  check_equal(vector(20,n,n--; vpar_count_height(n,n-1,-1)),
              vector(20,n,n--; if(n==0,1,(n-1)!)),
              "vpar_count_height() root=1 h=n-1 path permutations");
}

{
  \\ A034855 table count trees of height == h
  my(l=List([]));
  for(n=2,#A034855_samples,
     for(h=min(1,n),n-1,
        if(#l==#A034855_samples,break(2));
        listput(l,vpar_count_height(n,h))));
  check_equal(Vec(l),A034855_samples, "A034855");
}
{
  \\ A235595 table count forests of height == h
  my(l=List([]));
  for(n=1,#A235595_samples,
     for(h=0,n-1,
        if(#l==#A235595_samples,break(2));
        \\ print1(" "vpar_count_height(n,h,1));
        listput(l,vpar_count_height(n,h,1)));
     \\ print();
  );
  check_equal(Vec(l),A235595_samples, "A235595 table");
}
{
  \\ A235596 count forests height == 1
  check_equal(vector(#A235596_samples,n,n--; vpar_count_height(n,1,1)),
              A235596_samples, "A235596");
}

{
  my(limit = 10,
     L = O(x^limit),
     gN = test_func_to_egf(n->n, limit));
  check_equal(gN, x*exp(x+L) + L);

  my(gTeq1 = test_func_to_egf(n->vpar_count_height(n,1), limit));
  check_equal(gTeq1, gN-x, "gTeq1");

  my(gTle1 = test_func_to_egf(n->vpar_count_height_le(n,1), limit));
  check_equal(gTle1, gN+1);

  my(gFle1 = test_func_to_egf(n->vpar_count_height_le(n,1,1), limit));
  check_equal(polcoeff(gTle1-1, 0), 0);
  check_equal(polcoeff(gTle1-1, 1), 1);
  check_equal(gFle1, exp(gTle1-1),        "gFle1");
  check_equal(gFle1, exp(x*exp(x+L) + L), "gFle1");

  my(gOne = exp(x+L));
  my(gFeq1 = sum(n=0,limit-1, vpar_count_height(n,1,1) /n! * x^n) + L);
  check_equal(gFeq1, gFle1 - gOne);

  check_equal(gFeq1, exp(x*exp(x+L)) - exp(x+L));
}

\\ F1(m) = vpar_INTERNAL_exponential_transform(vector(10,m, if(m>=2,m,0)))[m];
\\ vector(10,m, F1(m))
\\ F(n) = sum(m=2,n, binomial(n,m)*F1(m));
\\ vector(10,m, F(m))


\\-----------------------------------------------------------------------------
progress("vpar_totdompoly()");

check_equal(vpar_totdompoly([]),    Pol(1), "vpar_totdompoly() empty");
check_equal(vpar_totdompoly([0]),   Pol(0), "vpar_totdompoly() single");
check_equal(vpar_totdompoly([0,1]), x^2, "vpar_totdompoly() path-2");
check_equal(vpar_totdompoly([0,1,1,1,1]),x^5 + 4*x^4 + 6*x^3 + 4*x^2,"star-5");
check_equal(vpar_totdompoly([0,1,2,3,4]),x^5 + 3*x^4 +   x^3,        "path-5");

test_totdompoly_by_flags(vpar) =
{
  my(poly=Pol(0));
  forvec(flags=vector(#vpar,i,[0,1]),
        if(vpar_is_totdomset_flags(vpar,flags),
           poly += x^vecsum(flags)));
  poly;
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        poly=vpar_totdompoly(vpar));
     check_equal(type(poly), "t_POL");

     check_equal((poly!=0), (vpar_num_singletons(vpar)==0));
     check_equal(simplify(subst(poly,x,1)),
                 vpar_totdomsets_count(vpar),
                 "vpar_totdompoly() for vpar_totdomsets_count()");

     if(vpar_num_singletons(vpar)==0,
        check_equal(polcoeff(poly,#vpar), 1);
        check_equal(valuation(poly,'x), vpar_totdomnum(vpar));
        check_equal(polcoeff(poly,valuation(poly,'x)),
                    vpar_totdomnum_and_count(vpar)[2]);
        );

     my(totdomnum_and_count=vpar_totdomnum_and_count(vpar));
     check_equal(totdomnum_and_count[1], if(poly,valuation(poly,x),-1));
     check_equal(totdomnum_and_count[2], test_poly_low_coefficient(poly));

     if(#vpar <= test_limit_flags_search,
        check_equal(poly, test_totdompoly_by_flags(vpar)));

     my(component_vpar=vpar_component_trees(vpar));
     check_equal(poly,
                 Pol(prod(i=1,#component_vpar, vpar_totdompoly(component_vpar[i]))),
                 "vpar_totdompoly() product of components");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_is_isomorphic()");
\\
\\ See also try_isomorphism_to_gp() in nautyextra devel/misc.c

check_equal(vpar_is_isomorphic([],[0]), 0, \
            "vpar_is_isomorphic() empty and single");

\\      1            1
\\     / \          /       directed not isomorphic
\\    2   3        2        free isomorphic
\\                /
\\               3
check_equal(vpar_is_isomorphic([2,0,1],[0,1,1]), 0, "vpar_is_isomorphic()");
check_equal(vpar_is_free_isomorphic([2,0,1],[0,1,1]), 1, \
            "vpar_is_free_isomorphic()");

check_equal(vpar_is_isomorphic([7,6,1,1,6,0,0], [6,6,5,5,7,0,0]), 1,\
            "vpar_is_isomorphic()");

check_equal(vpar_is_isomorphic([0,0],[0,1]),0, "forest != tree");
check_equal(vpar_is_isomorphic([0,1],[0,0]),0, "tree != forest");

check_equal(vpar_is_isomorphic([],[]),1);
check_equal(vpar_is_isomorphic([0,3,1],[0,1,2]),1);
check_equal(vpar_is_isomorphic([0,1,1,3],[0,4,1,1]),1);
check_equal(vpar_is_isomorphic([0,1,4,2],[0,4,1,3]),1);
check_equal(vpar_is_isomorphic([0,1,2,2],[0,3,1,2]),0);
check_equal(vpar_is_isomorphic([0,4,4,1,4],[2,4,2,0,2]),1, \
           "5-star, leaf root, relabelled");
check_equal(vpar_is_isomorphic([0,3,0],[0,0,1]),1, \
           "isolated and 2-path, relabelled");

\\   1               1
\\   |              / \
\\   2             2   3
\\  / \            |
\\ 3   4           4
check_equal(vpar_is_isomorphic([0,1,2,2],[0,1,1,2]),0);

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        vpar2=vpar_relabel_random(vpar));
     check_equal(vpar_is_isomorphic(vpar,vpar2),1,
                 "vpar_is_isomorphic() random relabel"));
}


\\-----------------------------------------------------------------------------
progress("vpar_ordered_isomorphic_perm()");
\\ vpar_is_ordered_isomorphic()

\\ [2, 3, 0]   3 <- 2 <- 1
\\ [3, 0, 2]   2 <- 3 <- 1
\\
{
  my(vpar1=[0, 0, 1, 2],  \\  1<-3 2<-4
     vpar2=[0, 4, 1, 0]); \\  1<-3 4<-2
  check_equal(vpar_relabel_perm(vpar1,vpar_seq_rows(vpar1)^-1), [0, 0, 1, 2]);
  check_equal(vpar_relabel_perm(vpar2,vpar_seq_rows(vpar2)^-1), [0, 0, 2, 1]);
}
\\
{
  \\      4          1                1           1
  \\      |          |                |           |
  \\      3    ->    2                2     ->    2
  \\     / \        / \              / \         / \
  \\    1   6      3   4            3   6       3   4
  \\    |   |      |   |            |   |       |   |
  \\    5   2      6   5            4   5       5   6
  my(vpar1=[3, 6, 4, 0, 1, 3],
     vpar2=[0, 1, 2, 3, 6, 2]);
  check_equal(vpar_relabel_perm(vpar1,vpar_seq_rows(vpar1)^-1), [0, 1, 2, 2, 4, 3]);
  check_equal(vpar_relabel_perm(vpar2,vpar_seq_rows(vpar2)^-1), [0, 1, 2, 2, 3, 4]);
}

\\ vpar1,vpar2 ordered isomorphic = exists an ordered permutation
\\ relabelling one to the other
test_is_ordered_isomorphic_by_search(vpar1,vpar2) =
{
  my(n=#vpar1);
  for(i=1,n!,
     my(perm=Vecsmall(numtoperm(n,i)));
     if(vpar_seq_is_ordered(vpar1,perm^-1) && vpar_relabel_perm(vpar1,perm) == vpar2,
        return(1)));
  0;
}

{
  for(i=1,#test_sample_forests,
     my(vpar1=test_sample_forests[i],
        vpar2=vpar_relabel_random(vpar1),
        got_perm=vpar_ordered_isomorphic_perm(vpar1,vpar2),
        got_iso=vpar_is_ordered_isomorphic(vpar1,vpar2));

     check_equal(got_iso, got_perm!='none,
                 "vpar_is_ordered_isomorphic() vs vpar_ordered_isomorphic_perm()");
     if(got_perm!='none,
        check_equal(vpar_relabel_perm(vpar2,got_perm), vpar1,
                   "vpar_ordered_isomorphic_perm() relabelling");
       );

     check_equal(got_iso,
                 vpar_relabel_perm(vpar1,vpar_seq_postorder(vpar1)^-1)
              == vpar_relabel_perm(vpar2,vpar_seq_postorder(vpar2)^-1),
                 "vpar_is_ordered_isomorphic() vs postorder relabel");
     if(got_perm!='none,
        check_equal(got_perm,
                    vpar_seq_postorder(vpar1) * vpar_seq_postorder(vpar2)^-1,
                    "vpar_ordered_isomorphic_perm() vs postorder perms"));

     if(#vpar1 <= test_limit_perms_search,
        check_equal(got_iso,
                    test_is_ordered_isomorphic_by_search(vpar1,vpar2));
        );

     \\ print(vpar1);
     \\ print(vpar2);
     \\ if(vpar_is_tree(vpar1),
     \\ check_equal(got_iso,
     \\             vpar_relabel_perm(vpar1,vpar_seq_rows(vpar1)^-1)
     \\          == vpar_relabel_perm(vpar2,vpar_seq_rows(vpar2)^-1),
     \\             "vpar_is_ordered_isomorphic() vs rows relabel"));
  );
}

{
  \\ vpar_is_ordered_isomorphic()
  \\ vs test_ordered_to_unlabelled() and plain vpar_is_isomorphic()

  my(funcs=[vpar_relabel_preorder,
            vpar_relabel_postorder,
            vpar_relabel_random]);
  for(f=1,#funcs,
     my(func=funcs[f]);
     for(i=1,#test_sample_forests,
        my(vpar  = test_sample_forests[i],
           vpar2 = func(vpar),
           vpar_unlabelled  = test_ordered_to_unlabelled(vpar),
           vpar2_unlabelled = test_ordered_to_unlabelled(vpar2));
        \\ print(vpar" "vpar2);

        check_equal(#vpar_unlabelled,
                    #vpar + (3*#vpar + vecsum(vpar_siblingpos_vector(vpar))));

        check_equal(vpar_is_isomorphic(vpar_unlabelled,vpar2_unlabelled),
                    vpar_is_ordered_isomorphic(vpar,vpar2),
                    Str("f="f));
  ));
}


\\------------------------------------------------------------------------------
progress("vpar_num_rootings_max()");

check_equal(vpar_num_rootings_max(0), 1, "vpar_num_rootings_max() 0");
check_equal(vpar_num_rootings_max(1), 1, "vpar_num_rootings_max() 1");
check_equal(vpar_num_rootings_max(2), 2);
check_equal(vpar_num_rootings_max(3), 3);

{
  for(n=0,20,
     my(m=0); \\ maximum product
     forpart(p=n,
             m=max(m,test_vector_product(p)));
     check_equal(vpar_num_rootings_max(n), m,
                 "vpar_num_rootings_max() vs partition product search"));
}


\\------------------------------------------------------------------------------
progress("vpar_num_free_isomorphic()");

check_equal(vpar_num_free_isomorphic([]), 1, \
           "vpar_num_free_isomorphic() empty");
check_equal(vpar_num_free_isomorphic([0]), 1, \
           "vpar_num_free_isomorphic() singleton");

check_equal(vpar_is_isomorphic([0,1],[0,1]), 1, \
           "vpar_is_isomorphic() path-2 to itself");
check_equal(vpar_is_isomorphic([0,1],[2,0]), 1, \
           "vpar_is_isomorphic() path-2 relabelled");
check_equal(vpar_num_isomorphic([0,1]), 2, \
           "vpar_num_isomorphic() path-2");

check_equal(vpar_num_isomorphic([3,3,0]), 3, \
           "vpar_num_isomorphic() path-2 root above");
check_equal(vpar_num_rootings([3,3,0]), 3, \
           "vpar_num_isomorphic() path-2 root above");

check_equal(vpar_is_free_isomorphic([0,1],[0,1]), 1);
check_equal(vpar_is_free_isomorphic([0,1],[2,0]), 1);
check_equal(vpar_num_free_isomorphic([0,1]), 1, \
           "vpar_num_free_isomorphic() path-2");

check_equal(vpar_num_isomorphic([0,0]), 1, \
           "vpar_num_isomorphic() 2 singletons");
check_equal(vpar_num_rootings([0,0]), 1, \
           "vpar_num_rootings() 2 singletons");
check_equal(vpar_num_isomorphic([3,4,0,0]), 12, \
           "vpar_num_isomorphic() 2 singletons, roots above");
check_equal(vpar_num_rootings([3,4,0,0]), 4, \
           "vpar_num_rootings() 2 singletons, roots above");
check_equal(vpar_num_free_isomorphic([0,0]), 1, \
           "vpar_num_free_isomorphic() 2 singletons");

\\ Return a vector [ [r1,r2], [r1,r3], ... ].
\\ Each element is a vector of possible root vertex numbers, being one
\\ vertex from each component of vpar, in all combinations.
\\
test_rerooting_roots(vpar) =
{
  \\ my(component_vertices=vpar_component_vertices(vpar),
  \\    ret = [ [] ]);
  \\ for(i=1,#component_vertices,
  \\    ret = vpar_INTERNAL_setproduct(ret, apply(v->[v],component_vertices[i])));
  \\ ret;

  my(component_vertices=vpar_component_vertices(vpar),
     component_ranges=vector(#component_vertices,i, [1,#component_vertices[i]]),
     roots_list=List([]));
  forvec(r=component_ranges,
        listput(roots_list, vector(#r,i,component_vertices[i][r[i]])));
  Vec(roots_list);
}
check_equal(test_rerooting_roots([0,1,1,1]), \
            [ [1], [2], [3], [4] ]);
check_equal(test_rerooting_roots([0,1,0,3]), \
            [ [1,3], [1,4], [2,3], [2,4] ]);

\\ Return [num_unrooted, num_rooted].
\\ num_unrooted = how many root=min
test_num_free_isomorphic_by_search(vpar) =
{
  my(num_unrooted=0,
     num_rooted=0);
  forvec(vec=vpar_forvec_from(#vpar, !vpar_is_tree(vpar)),
        my(vpar2=vpar_from_upascend(vec));
        if(vpar_is_free_isomorphic(vpar,vpar2),
           num_rooted++;
           num_unrooted += vpar_is_root_minmax(vpar2,min)));
  [num_unrooted, num_rooted];
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got=vpar_num_free_isomorphic(vpar));
     \\ print(vpar" got "got);
     check_equal(1 <= got && got <= (#vpar)!, 1,
                 "vpar_num_free_isomorphic() range");

     if(#vpar <= test_limit_trees_search,
        my(by_search = test_num_free_isomorphic_by_search(vpar),
           num_unrooted = by_search[1],
           num_rooted    = by_search[2]);
        check_equal(got, num_unrooted,
                    "vpar_num_free_isomorphic() vs search, unrooted");
        check_equal(got * vpar_num_rootings(vpar),
                    num_rooted,
                    "vpar_num_free_isomorphic() vs search, rooted");
       );
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_totdomnum_sets()");

check_equal(vpar_totdomnum_sets([]), [ [] ], "vpar_totdomnum_sets() empty");
check_equal(vpar_totdomnum_sets([0]), [ ], "vpar_totdomnum_sets() singleton");
check_equal(vpar_totdomnum_sets([0,1]), [ [1,2] ], \
            "vpar_totdomnum_sets() path-2");

\\    1
\\   / \    sole set 1,3
\\  2   3
{
  my(vpar=[0,1,1]);
  check_equal(vpar_totdomnum_and_count(vpar), [2,2]);
  check_equal(vpar_totdomnum_sets(vpar), [ [1,2], [1,3] ],
            "vpar_totdomnum_sets() star-3");
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        want_totdomnum_and_count = vpar_totdomnum_and_count(vpar),
        want_totdomnum = want_totdomnum_and_count[1],
        want_count = want_totdomnum_and_count[2]);
     if(want_count > test_limit_sets, next());

     my(got = vpar_totdomnum_sets(vpar));
     check_equal(#got, want_count, "vpar_totdomnum_sets() count");
     for(i=1,#got,
        check_equal(setisset(got[i]), 1, "vpar_totdomnum_sets() is set");
        check_equal(#got[i], want_totdomnum, "vpar_totdomnum_sets() set size");
        check_equal(vpar_is_totdomset(vpar,got[i]), 1,
                    "vpar_totdomnum_sets() is domset"));
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_add_roots_above_centres()");

check_equal(vpar_add_roots_above_centres([]), []);
check_equal(vpar_add_roots_above_centres([0]), [2,0]);

\\                           4
\\                           |
\\ 1 -> 2 -> 3  becomes 1 -> 2 <- 3
check_equal(vpar_add_roots_above_centres([0,1,2]), [2,4,2,0], \
            "vpar_add_roots_above_centres()");

\\ 1 <- 2  becomes 1 -> 3 <- 2
check_equal(vpar_add_roots_above_centres([0,1]), [3,3,0]);
check_equal(vpar_add_roots_above_centres([2,0]), [3,3,0]);

\\                                      5
\\                                    /  \
\\ 1 -> 2 -> 3 -> 4  becomes   1 -> 2      3 <- 4
check_equal(vpar_add_roots_above_centres([0,1,2,3]), [2,5,5,3,0], \
            "vpar_add_roots_above_centres()");


\\ 1 -> 2 -> 3 -> 4  becomes   1 -> 2 -> 9 <- 3 <- 4
\\ 5 <- 6 <- 7 <- 8  becomes   5 -> 6 -> 10  <- 7 <- 8
check_equal(vpar_add_roots_above_centres([0,1,2,3, 0,5,6,7]), \
            [2,9,9,3, 6,10,10,7, 0,0], \
            "vpar_add_roots_above_centres()");


\\------------------------------------------------------------------------------
progress("vpar_num_isomorphic()");

check_equal(vpar_num_isomorphic([]), 1, \
           "vpar_num_isomorphic() empty");

test_num_isomorphic_by_search(vpar) =
{
  \\ l = set of distinct vpar vectors from relabellings of vpar
  my(l=List([]));
  for(i=1,(#vpar)!,
     my(new_vpar = vpar_relabel_perm(vpar, numtoperm(#vpar,i)));
     my(pos=setsearch(l,new_vpar,1));
     if(pos, listinsert(l,new_vpar,pos)));
  #l;
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got=vpar_num_isomorphic(vpar));

     check_equal(1 <= got && got <= (#vpar)!, 1,
                 "vpar_num_isomorphic() range");

     if(#vpar <= test_limit_perms_search,
        check_equal(got, test_num_isomorphic_by_search(vpar),
                    "vpar_num_isomorphic() vs search");
       );
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_is_asymmetric()");

check_equal(vpar_is_asymmetric([]), 1, "vpar_is_asymmetric() empty");
check_equal(vpar_is_asymmetric([0]), 1);
check_equal(vpar_is_asymmetric([0,1]), 1);
check_equal(vpar_is_asymmetric([0,1,2]), 1);
check_equal(vpar_is_asymmetric([0,1,1]), 0, "vpar_is_asymmetric() star-3");

{
  \\ n=7 free asymmetric
  \\     3
  \\    / \
  \\   2   4
  \\      / \
  \\     6   7
  \\        / \
  \\       5   1
  my(vpar=[7, 3, 0, 3, 7, 4, 4]);
  vpar_validate(vpar);
  my(perm=[5, 2,3,4, 1, 6,7]);
  check_equal(vpar_relabel_perm(vpar,perm), vpar, "vpar_relabel_perm() same");

  \\ PENDING
  \\ check_equal(vpar_automorphism_ids_vector(vpar), [1, 1,5,4, 1, 1,3]);
  check_equal(vpar_is_asymmetric(vpar), 0);
}

{
  for(n=0,6,
     my(vpar=vpar_make_star(n));
     check_equal(vpar_is_asymmetric(vpar), n<3,
                 "vpar_make_star() asymmetric"));
}
{
  \\ n=7 free asymmetric
  \\       1
  \\     / | \
  \\   2   3   5
  \\       |   |
  \\       4   6
  \\           |
  \\           7
  my(vpar=[0,1,1,3,1,5,6]);
  vpar_validate(vpar);
  check_equal(vpar_is_asymmetric(vpar), 1, "vpar_is_asymmetric() 7");
}
{
  \\ n=8 free asymmetric
  \\         1             1
  \\       /   \         / | \
  \\     2       6     2   3   5
  \\    / \      |         |   |
  \\   3   5     7         4   6
  \\   |         |             |
  \\   4         8             7
  \\                           |
  \\                           8
  my(vpar=[0,1,2,3,2,1,6,7]);
  vpar_validate(vpar);
  check_equal(vpar_is_asymmetric(vpar), 1);

  my(vpar2=[0,1,1,3,1,5,6,7]);
  check_equal(vpar_is_free_isomorphic(vpar,vpar2), 1);
}
{
  \\ n=9 free asymmetric
  \\       1              1             1
  \\     / | \          / | \         / | \
  \\   2   3   5      2   3   5     2   3   5
  \\       |   |          |   |         |   | \
  \\       4   6          4   6         4   6  9
  \\           |          |   |             |
  \\           7          9   7             7
  \\           |              |             |
  \\           8              8             8
  \\           |
  \\           9
  my(vpar=[0,1,2,3,2,1,6,7]);
  vpar_validate(vpar);
  check_equal(vpar_is_asymmetric(vpar), 1,
              "vpar_is_asymmetric() 9");

  my(vpar2=[0,1,1,3,1,5,6,7]);
  check_equal(vpar_is_free_isomorphic(vpar,vpar2), 1);
}

{
  \\   2   4
  \\   |   |
  \\   1   3
  my(vpar=[2, 0, 4, 0]);
  vpar_validate(vpar);
  check_equal(vpar_is_asymmetric(vpar), 0,
              "vpar_is_asymmetric() two path-2");

  check_equal(vpar_relabel_perm(vpar,[3,4, 1,2]), vpar);
}

\\ Return 1 if vpar is asymmetric, by searching for a automorphism permutation
\\ other than the identity permutation.
test_is_asymmetric_by_perm_search(vpar) =
{
  for(i=1,(#vpar)!-1,
     my(perm=numtoperm(#vpar,i));
     if(vpar_relabel_perm(vpar,perm)==vpar, return(0)));
  1;
}

\\ Return 1 if vpar is asymmetric, by ensuring no parent has two children of
\\ the same subtree id, including roots reckoned as children of "p=0".
\\ vecsort flag=8 delete duplicates, so all distinct is unchanged length.
test_is_asymmetric_by_child_ids(vpar) =
{
  my(ids=vpar_automorphism_ids_vector(vpar));
  #vecsort(vector(#vpar,v,[ids[v],vpar[v]]),,9) == #vpar;

\\   print("v "vpar);
\\   print("i "ids);
\\   print();
\\   print("s "vecsort(vector(#vpar,x,[ids[x],vpar[x]]),
\\     (x,y)->my(s=lex(ids[x],ids[y])); if(s,s,lex(vpar[x],vpar[y]));
\\                     print("cmp "x" "y" is "lex([ids[x],vpar[x]],[ids[y],vpar[y]]));
\\ lex([ids[x],vpar[x]],[ids[y],vpar[y]]),
\\     1));
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     \\ PENDING
     \\ check_equal(vpar_is_asymmetric(vpar),
     \\             test_is_asymmetric_by_child_ids(vpar),
     \\             "vpar_is_asymmetric() vs child ids");
     if(#vpar <= test_limit_perms_search,
        check_equal(vpar_is_asymmetric(vpar),
                    test_is_asymmetric_by_perm_search(vpar),
                    "vpar_is_asymmetric() vs perm search")));
}



\\-----------------------------------------------------------------------------
progress("vpar_indompoly()");

check_equal(vpar_indompoly([]), Pol(1), \
            "vpar_indompoly() empty tree");
check_equal(vpar_indompoly(vpar_make_path(1)), x, \
            "vpar_indompoly() single vertex");
check_equal(vpar_indompoly(vpar_make_path(2)), 2*x, \
            "vpar_indompoly() path-2");
check_equal(vpar_indompoly(vpar_make_path(3)), x^2 + x, \
            "vpar_indompoly() path-3");
check_equal(vpar_indompoly(vpar_make_path(4)), 3*x^2, \
            "vpar_indompoly() path-4");

test_indompoly_by_flags(vpar) =
{
  my(poly=Pol(0));
  forvec(flags=vector(#vpar,i,[0,1]),
        if(vpar_is_indomset_flags(vpar,flags),
           poly += x^vecsum(flags)));
  poly;
}
check_equal(test_indompoly_by_flags(vpar_make_path(3)), x^2 + x, \
            "test_indompoly_by_flags() path-3");

\\ This variation keeping without_dom[] instead of keeping without[] and
\\ subtracting the undom to get dom.
\\
test_indompoly_by_dom(vpar,var='x) =
{
  \\ print("test_indompoly_by_dom() "vpar);
  my(seq=vpar_INTERNAL_seq_upwards(vpar),
     with          = vector(#vpar,v,var),  \\ 1 set size 1
     without_dom   = vector(#vpar),        \\ 0 sets
     without_undom = vector(#vpar,v,1),    \\ 1 set size 0 undominated
     ret           = var/var);             \\ 1 as poly
  for(i=1,#seq,
     my(v=seq[i],
        p=vpar[v],
        v_domsets = with[v] + without_dom[v]);
     \\ print(" v="v" with="with[v]" without dom "without_dom[v]" undom "without_undom[v]);
     \\ print("  domsets "v_domsets);
     if(p,
        with[p]          *= without_dom[v] + without_undom[v];

        \\ print("  p without_dom extend "without_dom[p] * v_domsets" new "without_undom[p] * with[v]);
        without_dom[p]    = without_dom[p] * v_domsets
                           + without_undom[p] * with[v];

        without_undom[p] *= without_dom[v];

        \\ print("  gives p with="with[p]" without dom "without_dom[p]" undom "without_undom[p]);
        ,
        ret *= v_domsets);

     with[v]          =  \\ kill bignum intermediate values
     without_dom[v]   =
     without_undom[v] = 0);

  ret;
}
\\ vpar_indompoly([0,1,1])
\\ test_indompoly_by_dom([0,1,1])

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        poly=vpar_indompoly(vpar));
     check_equal(type(poly), "t_POL");
     check_equal(poly, test_indompoly_by_dom(vpar),
                 "vpar_indompoly() vs test_indompoly_by_dom()");

     check_equal(valuation(poly,'x), vpar_indomnum(vpar));
     check_equal(polcoeff(poly,valuation(poly,'x)), vpar_indomnum_and_count(vpar)[2]);

     my(indomnum_and_count=vpar_indomnum_and_count(vpar));
     check_equal(indomnum_and_count[1], if(poly,valuation(poly,x),-1));
     check_equal(indomnum_and_count[2], test_poly_low_coefficient(poly));

     check_equal(simplify(subst(poly,x,1)),
                 vpar_indomsets_count(vpar),
                 "vpar_indompoly() subst for vpar_indomsets_count()");

     if(#vpar <= test_limit_flags_search,
        check_equal(poly, test_indompoly_by_flags(vpar),
                    "vpar_indompoly() vs test_indompoly_by_flags()"));

     my(component_vpar=vpar_component_trees(vpar));
     check_equal(poly,
                 Pol(prod(i=1,#component_vpar, vpar_indompoly(component_vpar[i]))),
                 "vpar_indompoly() product of components");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_dompoly()");

check_equal(vpar_dompoly([]),    Pol(1),         "vpar_dompoly() empty");
check_equal(type(vpar_dompoly([])), "t_POL", "empty");
check_equal(vpar_dompoly([0]),   x,         "vpar_dompoly() single");
check_equal(vpar_dompoly([0,1]), x^2 + 2*x, "vpar_dompoly() path-2");
check_equal(vpar_dompoly([0,1],'y), 'y^2 + 2*'y, "vpar_dompoly() path-2 y");
check_equal(vpar_dompoly([0,1,1,1,1]), x^5 + 5*x^4 + 6*x^3 + 4*x^2 + x, \
            "star-5");
check_equal(vpar_dompoly([0,1,2,3,4]), x^5 + 5*x^4 + 8*x^3 + 3*x^2, \
            "path-5");

test_dompoly_by_flags(vpar) =
{
  my(poly=Pol(0));
  forvec(flags=vector(#vpar,i,[0,1]),
        if(vpar_is_domset_flags(vpar,flags),
           poly += x^vecsum(flags)));
  poly;
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        poly=vpar_dompoly(vpar));
     check_equal(type(poly), "t_POL");

     check_equal(polcoeff(poly,0), #vpar==0);
     check_equal(polcoeff(poly,1)!=0,
                 #vpar && vpar_is_tree(vpar) && vpar_diameter(vpar)<=2);
     check_equal(polcoeff(poly,#vpar), 1);
     check_equal(polcoeff(poly,#vpar-1), #vpar - vpar_num_singletons(vpar));

     check_equal(valuation(poly,x), vpar_domnum(vpar));
     check_equal(polcoeff(poly,valuation(poly,'x)), vpar_domnum_and_count(vpar)[2]);

     my(domnum_and_count=vpar_domnum_and_count(vpar));
     check_equal(domnum_and_count[1], valuation(poly,x));
     check_equal(domnum_and_count[2], test_poly_low_coefficient(poly));

     check_equal(simplify(subst(poly,x,1)), vpar_domsets_count(vpar));

     if(#vpar <= test_limit_flags_search,
        check_equal(poly, test_dompoly_by_flags(vpar)));

     my(component_vpar=vpar_component_trees(vpar));
     check_equal(poly,
                 Pol(prod(i=1,#component_vpar, vpar_dompoly(component_vpar[i]))),
                 "vpar_dompoly() product of components");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_matchnum_and_count()");

check_equal(vpar_matchnum_and_count([]), [0,1], \
            "vpar_matchnum_and_count() empty tree");
check_equal(vpar_matchnum_and_count([0]), [0,1], \
            "vpar_matchnum_and_count() singleton");
check_equal(vpar_matchnum_and_count([0,1]), [1,1], \
            "vpar_matchnum_and_count() path-2");

test_matchnum_and_count_by_search(vpar) =
{
  my(matchnum=0,count=0);
  forvec(flags=vector(#vpar,i,[0,1]),
        if(vpar_is_matching_flags(vpar,flags),
           my(s=vecsum(flags));
           if(s>matchnum, matchnum=s;count=1,  \\ new high
              s==matchnum, count++)));         \\ equal high
  [matchnum,count];
}
{
  for(n=4,8,
     my(vpar=vpar_make_path(n));
     check_equal(vpar_matchnum_and_count(vpar),
                 test_matchnum_and_count_by_search(vpar),
                 Str("vpar_matchnum_and_count() path "n)));
}
{
  for(n=4,8,
     my(vpar=vpar_make_star(n));
     check_equal(vpar_matchnum_and_count(vpar),
                 test_matchnum_and_count_by_search(vpar),
                 Str("vpar_matchnum_and_count() star "n)));
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got = vpar_matchnum_and_count(vpar));
     check_equal(got,
                 test_matchnum_and_count_by_search(vpar),
                 Str("vpar_matchnum_and_count() "
                     "vs test_matchnum_and_count_by_search()"));
     check_equal(got[1], vpar_matchnum(vpar));
     check_equal(got[1]==#vpar/2, vpar_has_perfect_matching(vpar));
  );
}

test_make_stars_chain(v) =
{
  my(vpar=vpar_make_star(v[1]));
  for(i=2,#v,
     vpar = vpar_join(vpar, #vpar, vpar_make_star(v[i]-1)));
  vpar;
}

check_equal(vpar_matchnum_and_count(test_make_stars_chain([4,4])), \
            [2,8]);  \\ n=7
check_equal(vpar_matchnum_and_count(test_make_stars_chain([4,5])), \
            [2,11]); \\ n=8
check_equal(vpar_matchnum_and_count(test_make_stars_chain([5,5])), \
            [2,15]);  \\ n=9
{
  \\ n=10
  my(vpar=test_make_stars_chain([4,4,4]));
  check_equal(#vpar, 10);
  check_equal(vpar_matchnum_and_count(vpar), [3,21]);
}
check_equal(vpar_matchnum_and_count(test_make_stars_chain([5,6])), \
            [2,19]); \\ n=10 by stars
check_equal(4*5-1, 19);

check_equal(vpar_matchnum_and_count(test_make_stars_chain([4,5,4])), \
            [3,30]);  \\ n=11
check_equal(vpar_matchnum_and_count(test_make_stars_chain([5,5,4])), \
            [3,41]);  \\ n=12
check_equal(vpar_matchnum_and_count(test_make_stars_chain([5,5,5])), \
            [3,56]);  \\ n=13
{
  \\ n=14 different   5,4,4,4 around the 5 instead of chain
  my(vpar=[0,1]);
  for(i=1,3, vpar=vpar_join(vpar,1,vpar_reroot(vpar_make_star(4),2)));
  check_equal(#vpar, 14);
  check_equal(vecsum([5,4,4,4])-3, 14);
  check_equal(vpar_matchnum_and_count(vpar), [4,81]);

  vpar = test_make_stars_chain([5,4,4,4]);
  check_equal(#vpar, 14);
  check_equal(vpar_matchnum_and_count(vpar), [4,76]);
}
{
  \\ n=15
  my(vpar=test_make_stars_chain([4,5,5,4]));
  check_equal(#vpar, 15);
  check_equal(vpar_matchnum_and_count(vpar), [4,112]);
}
{
  \\ n=16
  my(vpar=test_make_stars_chain([5,5,5,4]));
  check_equal(#vpar, 16);
  check_equal(vpar_matchnum_and_count(vpar), [4,153]);
}
{
  \\ n=17 different    5,4,4,4,4 around the 5 instead of chain
  my(vpar=[0]);
  for(i=1,4, vpar=vpar_join(vpar,1,vpar_reroot(vpar_make_star(4),2)));
  check_equal(#vpar, 17);
  check_equal(vecsum([5,4,4,4,4])-4, 17);
  check_equal(vpar_matchnum_and_count(vpar), [5,216]);

  vpar = test_make_stars_chain([5,4,4,4,4]);
  check_equal(#vpar, 17);
  check_equal(vpar_matchnum_and_count(vpar), [5,199]);
}

\\ Biedl, Demaine, Duncan, Fleischer, Kobourov, "Tight Bounds on Maximal
\\ and Maximum Matchings", Proceedings ISAAC 2001, Lecture Notes in
\\ Computer Science volume 2223, 2001, pages 308-319.
\\ https://www2.cs.arizona.edu/~kobourov/art.ps
\\ http://erikdemaine.org/papers/MatchingBounds_ISAAC2001/
\\
\\ Show every degree <= 3 graph has a matching of size >= (n-1)/3.
\\
\\ Make Biedl et al's example tree, with t top row vertices, t>=1.
\\ Their example is t=4.
test_make_matchnum_nsub1over3(t) =
{
  my(w=2*t+3,
     vpar=vpar_make_path(w),
     star=vpar_make_star(3));
  forstep(v=2,w-1,2, vpar=vpar_join(vpar,v,[0]));
  forstep(v=3,w-2,2, vpar=vpar_join(vpar,v,star));
  vpar;
}
\\ vpar_view(test_make_matchnum_nsub1over3(1));
{
  for(t=1,8,
     my(vpar=test_make_matchnum_nsub1over3(t));
     check_equal(#vpar, 6*t+4,
                "test_make_matchnum_nsub1over3() num vertices");
     check_equal(vecmax(vpar_degrees_vector(vpar)) <= 3, 1);
     my(got=vpar_matchnum(vpar));
     check_equal(got, (#vpar-1)/3));
}
{
  my(vpar=test_make_matchnum_nsub1over3(4));
  check_equal(vpar_matchnum(vpar), 9);
  check_equal(vpar_matchnum_and_count(vpar), [9, 5200]);
}


\\-----------------------------------------------------------------------------
progress("vpar_matchnum_and_count() form using prospective");

\\ This form makes the extend or prospective size calculations clearer

test_matchnum_and_count_by_prospective(vpar) =
{
  \\ with[v] = size of largest matching with v
  \\ with_count[v] = number of such matchings
  \\ without[v] = size of largest matching without v
  \\ without_count[v] = number of such matchings
  \\
  \\ "extend" extends with[p] by taking the bigger of with[v] and/or
  \\ without[v] below it and no new pairings.
  \\ "prospective" is a new pairing between without[p] and without[v].
  \\
  \\ Always have with[v] = without[v] 0 or 1.  Supposing that holds at v and
  \\ p, the choice between extend and prospective for new with[p] maintains.
  \\ So with[v] is the bigger of with and without.  v_max_count is the
  \\ number of matchings of that with[v] max size, by including
  \\ without_count[v] when with==without.
  \\
  my(seq=vpar_INTERNAL_seq_upwards(vpar),
     with          = vectorsmall(#vpar),
     without       = vectorsmall(#vpar),
     with_count    = vector(#vpar),
     without_count = vector(#vpar,v,1),
     ret       = 0,
     ret_count = 1);
  for(i=1,#seq,
     my(v=seq[i],
        p=vpar[v],
        v_max_count = with_count[v]
                     + if(without[v]==with[v], without_count[v]));
     if(p,
        my(extend      = with[p] + with[v],
           prospective = without[p] + without[v] + 1);
        with[p] = max(extend, prospective);
        with_count[p] = if(with[p]==extend, with_count[p] * v_max_count)
              + if(with[p]==prospective, without_count[p] * without_count[v]);

        without[p]       += with[v];  \\ v_max
        without_count[p] *= v_max_count;
        ,
        ret       += with[v];     \\ v_max
        ret_count *= v_max_count);

     with_count[v]    =     \\ kill bignum intermediate values
     without_count[v] = 0);

  [ret,ret_count];
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     check_equal(vpar_matchnum_and_count(vpar),
                 test_matchnum_and_count_by_prospective(vpar)));
}


\\-----------------------------------------------------------------------------
progress("vpar_is_stable_equimatchable() state table");

\\ rel is vector [with_list, without, non]
\\   with_list = Set() of equimatchable states occurring from each possible
\\               cut below v.  Empty [] when no edges below to cut.
\\   without   = equimatchable state when no cut below v (the same as in
\\               vpar_is_equimatchable())
\\   ok        = boolean 1=ok, 0=not edge stable equimatchable.
\\
test_stable_equimatchable_rel_initial = [[], 1, 1];
test_stable_equimatchable_rel_non = [[], 0, 0];

test_stable_equimatchable_rel_validate(rel) =
{
  \\ print("test_stable_equimatchable_rel_validate() "rel);
  my(with_list,without,ok);
  [with_list,without,ok] = rel;
  if(!ok,
     without==0 || error();
     with_list==[] || error();
     ,
     (1 <= without && without <= 5) || error();
     for(i=1,#with_list,
        (1 <= with_list[i] && with_list[i] <= 5) || error());
     );
}
test_stable_equimatchable_rel_validate \
  (test_stable_equimatchable_rel_initial);
test_stable_equimatchable_rel_validate \
  (test_stable_equimatchable_rel_non);

test_stable_equimatchable_rel_add(p_rel,v_rel) =
{
  \\ print("test_stable_equimatchable_rel_add() p="p_rel"  v="v_rel);
  test_stable_equimatchable_rel_validate(p_rel);
  test_stable_equimatchable_rel_validate(v_rel);

  my(p_with_list,p_without,p_ok);
  [p_with_list,p_without,p_ok] = p_rel;
  my(v_with_list,v_without,v_ok);
  [v_with_list,v_without,v_ok] = v_rel;

  if(!p_ok || !v_ok,
     return(test_stable_equimatchable_rel_non));

  \\ p_without edge to v_without
  \\ If new_without==0 then that is the whole never equimatchable, no matter
  \\ what else above or below.
  my(new_without = test_is_equimatchable_table_lookup(p_without,v_without));
  if(new_without==0,
     \\ print(" without "p_without" and "v_without" to non");
     return(test_stable_equimatchable_rel_non));

  my(new_with_list=List([]));

  \\ extend existing with[p] by without[v]
  for(i=1,#p_with_list,
     my(new_state = test_is_equimatchable_table_lookup
                      (p_with_list[i], v_without));
     listput(new_with_list,new_state));

  \\ build with[p] by p_without cut to v_without
  \\ must have v_without!=4 or it is a non-equimatchable piece below
  if(v_without==4, return(test_stable_equimatchable_rel_non));
  listput(new_with_list, p_without);

  \\ build with[p] by p_without edge to each v_with
  for(i=1,#v_with_list,
     my(new_state = test_is_equimatchable_table_lookup
                      (p_without, v_with_list[i]));
     listput(new_with_list,new_state));

  new_with_list = Set(new_with_list);
  \\ any new_state==0 is a non-equimatchable combination
  if(new_with_list[1]==0,
     return(test_stable_equimatchable_rel_non));

  [new_with_list, new_without, 1];
}

test_stable_equimatchable_rel_by_delete(vpar) =
{
  my(seen=vector(test_is_equimatchable_num_states));
  for(v=1,#vpar,
     if(vpar[v],
        if(! vpar_is_equimatchable(vpar_subtree(vpar,v)),
           return(test_stable_equimatchable_rel_non));
        my(state=test_equimatchable_get_state(vpar_delete_subtree(vpar,v)));
        if(!state,
           return(test_stable_equimatchable_rel_non));
        seen[state]=1));
  [ Vec(select(f->f!=0, seen, 1)),
   test_equimatchable_get_state(vpar),
   1 ];
}

test_stable_equimatchable_rel_list = List([]);
{
  \\ 1 = singleton
  listput(test_stable_equimatchable_rel_list,
          test_stable_equimatchable_rel_initial);
  check_equal(test_stable_equimatchable_rel_list[1], [ [], 1, 1 ]);
  check_equal(test_stable_equimatchable_rel_list[1],
              test_stable_equimatchable_rel_by_delete([0]));
}
{
  \\ 2 = path-2
  \\    *
  \\    |
  \\    *
  listput(test_stable_equimatchable_rel_list,
          test_stable_equimatchable_rel_add
          (test_stable_equimatchable_rel_list[1],
           test_stable_equimatchable_rel_list[1]));
  check_equal(test_stable_equimatchable_rel_list[2],
              [ [1], 2, 1 ]);
  check_equal(test_stable_equimatchable_rel_list[2],
              test_stable_equimatchable_rel_by_delete([0,1]));
}
{
  \\ 3 = path-3
  \\    *
  \\    |
  \\    *
  \\    |
  \\    *
  listput(test_stable_equimatchable_rel_list,
          test_stable_equimatchable_rel_add
          (test_stable_equimatchable_rel_list[1],
           test_stable_equimatchable_rel_list[2]));
  check_equal(test_stable_equimatchable_rel_list[3],
              [ [1,2],  3, 1 ]);
  check_equal(test_stable_equimatchable_rel_list[3],
              test_stable_equimatchable_rel_by_delete([0,1,2]));
}
{
  \\ 4 = path-4
  \\    *
  \\    |
  \\    *
  \\    |
  \\    *
  \\    |
  \\    *
  listput(test_stable_equimatchable_rel_list,
          test_stable_equimatchable_rel_add
          (test_stable_equimatchable_rel_list[1],
           test_stable_equimatchable_rel_list[3]));
  check_equal(test_stable_equimatchable_rel_list[4],
              [ [1,2,3], 4, 1 ]);
  check_equal(test_stable_equimatchable_rel_list[4],
              test_stable_equimatchable_rel_by_delete([0,1,2,3]));
}
{
  \\ 5 = star-3
  \\    *
  \\   / \
  \\  *   *
  listput(test_stable_equimatchable_rel_list,
          test_stable_equimatchable_rel_add
          (test_stable_equimatchable_rel_list[2],
           test_stable_equimatchable_rel_list[1]));
  check_equal(test_stable_equimatchable_rel_list[5],
              [ [2], 2, 1 ]);
  check_equal(test_stable_equimatchable_rel_list[5],
              test_stable_equimatchable_rel_by_delete([0,1,1]));
}
{
  \\ 6 = claw root at arm
  \\    *
  \\    |
  \\    *
  \\   / \
  \\  *   *
  listput(test_stable_equimatchable_rel_list,
          test_stable_equimatchable_rel_add
          (test_stable_equimatchable_rel_list[1],
           test_stable_equimatchable_rel_list[5]));
  check_equal(test_stable_equimatchable_rel_list[6],
              [ [1,3], 3, 1 ]);
  check_equal(test_stable_equimatchable_rel_list[6],
              test_stable_equimatchable_rel_by_delete([0,1,2,2]));
}
{
  \\ 7 claw + 1 above
  \\    *
  \\    |
  \\    *
  \\    |
  \\    *
  \\   / \
  \\  *   *
  listput(test_stable_equimatchable_rel_list,
          test_stable_equimatchable_rel_add
          (test_stable_equimatchable_rel_list[1],
           test_stable_equimatchable_rel_list[6]));
  check_equal(test_stable_equimatchable_rel_list[7],
              [ [1,2,4], 4, 1 ]);
  check_equal(test_stable_equimatchable_rel_list[7],
              test_stable_equimatchable_rel_by_delete([0,1,2,3,3]));
}
{
  \\ 8
  \\      *
  \\     / \
  \\    *   *
  \\    |
  \\    *
  \\   / \
  \\  *   *
  listput(test_stable_equimatchable_rel_list,
          test_stable_equimatchable_rel_add
          (test_stable_equimatchable_rel_list[2],
           test_stable_equimatchable_rel_list[6]));
  check_equal(test_stable_equimatchable_rel_list[8],
              [ [2,4], 2, 1 ]);
  check_equal(test_stable_equimatchable_rel_list[8],
              test_stable_equimatchable_rel_by_delete([0,1,2,3,3,1]));
}
{
  \\ 9
  \\      1
  \\      |
  \\      2
  \\     / \
  \\    3   4
  \\    |
  \\    5
  \\   / \
  \\  6   7
  listput(test_stable_equimatchable_rel_list,
          test_stable_equimatchable_rel_add
          (test_stable_equimatchable_rel_list[1],
           test_stable_equimatchable_rel_list[8]));
  check_equal(test_stable_equimatchable_rel_list[9],
              [ [1,3,5], 3, 1 ]);
  check_equal(test_stable_equimatchable_rel_list[9],
              test_stable_equimatchable_rel_by_delete([0,1,2,2,3,5,5]));
}
{
  \\ 10
  \\        *
  \\      /   \
  \\    *       *
  \\   / \     / \
  \\  *   *   *   *
  listput(test_stable_equimatchable_rel_list,
          test_stable_equimatchable_rel_add
          (test_stable_equimatchable_rel_list[6],
           test_stable_equimatchable_rel_list[5]));
  check_equal(test_stable_equimatchable_rel_list[10],
              [ [3], 3, 1 ]);
  check_equal(test_stable_equimatchable_rel_list[10],
              test_stable_equimatchable_rel_by_delete([0,1,1,2,2,3,3]));
}
{
  \\ 11
  \\        *
  \\        |
  \\        *
  \\      /   \
  \\    *       *
  \\   / \     / \
  \\  *   *   *   *
  listput(test_stable_equimatchable_rel_list,
          test_stable_equimatchable_rel_add
          (test_stable_equimatchable_rel_list[1],
           test_stable_equimatchable_rel_list[10]));
  check_equal(test_stable_equimatchable_rel_list[11],
              [ [1,4], 4, 1 ]);
  check_equal(test_stable_equimatchable_rel_list[11],
              test_stable_equimatchable_rel_by_delete([0,1,2,2,3,3,4,4]));
}
{
  \\ 12
  \\        1
  \\     /     \
  \\    2       3
  \\    |       |
  \\    4       5
  \\   / \     / \
  \\  6   7   8   9
  listput(test_stable_equimatchable_rel_list,
          test_stable_equimatchable_rel_add
          (test_stable_equimatchable_rel_list[7],
           test_stable_equimatchable_rel_list[6]));
  check_equal(test_stable_equimatchable_rel_list[12],
              [ [2,4], 4, 1 ]);
  check_equal(test_stable_equimatchable_rel_list[12],
              test_stable_equimatchable_rel_by_delete([0,1,1,2,3,4,4,5,5]));
}
{
  \\ 13
  \\                 *
  \\            /         \
  \\        *                *
  \\      /   \            /   \
  \\    *       *        *       *
  \\   / \     / \      / \     / \
  \\  *   *   *   *    *   *   *   *
  listput(test_stable_equimatchable_rel_list,
          test_stable_equimatchable_rel_add
          (test_stable_equimatchable_rel_list[11],
           test_stable_equimatchable_rel_list[10]));
  check_equal(test_stable_equimatchable_rel_list[13],
              [ [4], 4, 1 ]);
  check_equal(test_stable_equimatchable_rel_list[13],
              test_stable_equimatchable_rel_by_delete([0,1,1,2,2,3,3, 4,4,5,5,6,6,7,7]));
}
test_stable_equimatchable_rel_list \
  = Vec(test_stable_equimatchable_rel_list);

{
  if(0,
     for(i=1,#test_stable_equimatchable_rel_list,
        my(rel=test_stable_equimatchable_rel_list[i]);
        print("state="i"  ",rel));
     );
}

\\ Return the state number 1 to 9 for the given rel, or return 0 if not one
\\ of the 9 states in test_equimatchable_rel_list.
test_stable_equimatchable_rel_find(rel) =
{
  for(i=1,#test_stable_equimatchable_rel_list,
     if(test_stable_equimatchable_rel_list[i] == rel, return(i)));
  0;
}
{
  test_stable_equimatchable_rel_table
  = matrix(#test_stable_equimatchable_rel_list,
           #test_stable_equimatchable_rel_list,
           p,v,
           test_stable_equimatchable_rel_find
           (test_stable_equimatchable_rel_add
            (test_stable_equimatchable_rel_list[p],
             test_stable_equimatchable_rel_list[v])));
}

{
  \\ each parent/child state combination must transition to a new known state,
  \\ or be non
  for(p=1,#test_stable_equimatchable_rel_list,
     for(v=1,#test_stable_equimatchable_rel_list,
        my(rel=test_stable_equimatchable_rel_add
               (test_stable_equimatchable_rel_list[p],
                test_stable_equimatchable_rel_list[v]));
        check_equal(test_stable_equimatchable_rel_find(rel)
                    || rel==test_stable_equimatchable_rel_non, 1,
                    Str("test_stable_equimatchable_rel_list transition"
                        " p="p" v="v" to "rel))));
}

{
  check_equal(test_stable_equimatchable_rel_list,
              [[ [],      1, 1 ],   \\ 1 = singleton
               [ [1],     2, 1 ],   \\ 2 = path=2
               [ [1,2],   3, 1 ],   \\ 3 = path-3
               [ [1,2,3], 4, 1 ],   \\ 4 = path-4
               [ [2],     2, 1 ],   \\ 5 = star-3
               [ [1,3],   3, 1 ],   \\ 6 = claw
               [ [1,2,4], 4, 1 ],   \\ 7 = claw + 1
               [ [2,4],   2, 1 ],   \\ 8 = claw + 2 up, root at 1
               [ [1,3,5], 3, 1 ],   \\ 9 = 1 above 8
               [ [3],     3, 1 ],   \\ 10 = complete binary n=7 vertices
               [ [1,4],   4, 1 ],   \\ 11
               [ [2,4],   4, 1 ],   \\ 12
               [ [4],     4, 1 ]    \\ 13
              ],
              "test_stable_equimatchable_rel_list");
}
check_equal(#test_stable_equimatchable_rel_list, 13);

\\ Return 1 if rel is not an edge stable equimatchable state as a root.
\\ Per vpar_is_equimatchable() states, cannot have 4 as that is a
\\ non-equimatchable as a root.
\\
test_stable_equimatchable_rel_is_non_at_root(rel) =
{
  my(with_list,without,ok);
  [with_list,without,ok] = rel;

  !ok || without==4 || #setintersect(with_list,[4]);
}
{
  test_stable_equimatchable_rel_ok_at_root
  = apply(rel-> ! test_stable_equimatchable_rel_is_non_at_root(rel),
         test_stable_equimatchable_rel_list);
}

test_stable_equimatchable_rel_is_equivalent(x,y) =
{
  \\ same final ok-ness
  test_stable_equimatchable_rel_ok_at_root[x] ==
  test_stable_equimatchable_rel_ok_at_root[y] || return(0);

  \\ same transitions when add new child under
  test_stable_equimatchable_rel_table[x,] ==
  test_stable_equimatchable_rel_table[y,] || return(0);

  \\ same transitions when are a child under some parent
  test_stable_equimatchable_rel_table[,x] ==
  test_stable_equimatchable_rel_table[,y] || return(0);

  1;
}

{
  \\ 4 is not at root, and goes to non for anything under or above
  \\ mangle the table to reckon it non

  check_equal(test_stable_equimatchable_rel_table[4,],
              [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]);
  check_equal(test_stable_equimatchable_rel_table[,4],
              [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]~);
  check_equal(test_stable_equimatchable_rel_ok_at_root[4], 0);

  for(p=1,#test_stable_equimatchable_rel_list,
     for(v=1,#test_stable_equimatchable_rel_list,
        if(test_stable_equimatchable_rel_table[p,v]==4,
           test_stable_equimatchable_rel_table[p,v] = 0)));
}

{
  \\ 3,9 same, 7,12 same, 11,13 same
  for(x=1,#test_stable_equimatchable_rel_list,
     for(y=x+1,#test_stable_equimatchable_rel_list,
        check_equal(test_stable_equimatchable_rel_is_equivalent(x,y),
                       (x==3 && y==9)
                    || (x==7 && y==12)
                    || (x==11 && y==13),
                    Str("test_stable_equimatchable_rel_is_equivalent()"
                        " x="x" y="y))));
}

\\ 7,12 same - both ok, but both doomed to be non when anything below, or
\\ when they are below anything else
check_equal(test_stable_equimatchable_rel_is_equivalent(7,12), 1, \
            "test_stable_equimatchable_rel_is_equivalent() 7,12");
check_equal(test_stable_equimatchable_rel_table[7,], \
            test_stable_equimatchable_rel_table[12,]);
check_equal(test_stable_equimatchable_rel_table[,7], \
            test_stable_equimatchable_rel_table[,12]);
check_equal(test_stable_equimatchable_rel_ok_at_root[7], 0, \
            "test_stable_equimatchable_rel_ok_at_root 7");
check_equal(test_stable_equimatchable_rel_ok_at_root[12], 0);
check_equal(test_stable_equimatchable_rel_table[,7], \
            vector(#test_stable_equimatchable_rel_list)~);
check_equal(test_stable_equimatchable_rel_table[,12], \
            vector(#test_stable_equimatchable_rel_list)~);
check_equal(test_stable_equimatchable_rel_table[7,], \
            [8, 0, 0, 0, 0, 12, 0, 0, 0, 12, 0, 0, 0]);
check_equal(test_stable_equimatchable_rel_table[12,], \
            [8, 0, 0, 0, 0, 12, 0, 0, 0, 12, 0, 0, 0]);

\\ drop 4
\\ 9 -> 3
\\ 12 -> 7 is 6 shunk
\\ 13 -> 11 is 9 shrunk
\\                                      1,2,3,4,5,6,7,8,9,10,11,12,13
test_stable_equimatchable_rel_shrink = [1,2,3,0,4,5,6,7,3, 8, 9, 6, 9];
test_stable_equimatchable_rel_shrink_size = \
  #test_stable_equimatchable_rel_list - 4;
check_equal(test_stable_equimatchable_rel_shrink_size, 9, \
           "test_stable_equimatchable_rel_shrink_size");

{
  \\ vpar_is_stable_equimatchable() ok constant, inline in code
  \\ generated
  my(want=0);
  for(i=1,#test_stable_equimatchable_rel_list,
     if(test_stable_equimatchable_rel_ok_at_root[i],
        my(state=test_stable_equimatchable_rel_shrink[i]);
        if(!state, next());
        want=bitor(want, 1 << state)));
  check_equal(318,
              want,
              "vpar_is_stable_equimatchable() at root bits constant");
  check_equal(test_is_stable_equimatchable_at_root_bits,
              want);
}
{
  \\ test_is_stable_equimatchable_at_root
  for(i=1,#test_stable_equimatchable_rel_list,
        my(state=test_stable_equimatchable_rel_shrink[i]);
        if(!state, next());
     check_equal(test_is_stable_equimatchable_at_root(state),
                 test_stable_equimatchable_rel_ok_at_root[i],
                 "test_is_stable_equimatchable_at_root()"));
}

{
  \\ vpar_INTERNAL_is_stable_equimatchable_table[]

  check_equal(#vpar_INTERNAL_is_stable_equimatchable_table,
              test_is_stable_equimatchable_num_states^2,
              "vpar_INTERNAL_is_stable_equimatchable_table[] length N^2");

  \\ generated
  my(m=matrix(test_stable_equimatchable_rel_shrink_size,
              test_stable_equimatchable_rel_shrink_size));
  for(p=1,#test_stable_equimatchable_rel_list,
     my(ps=test_stable_equimatchable_rel_shrink[p]);
     if(!ps, next());
     for(v=1,#test_stable_equimatchable_rel_list,
        my(vs=test_stable_equimatchable_rel_shrink[v]);
        if(!vs, next());

        my(t=test_stable_equimatchable_rel_find
              (test_stable_equimatchable_rel_add
               (test_stable_equimatchable_rel_list[p],
                test_stable_equimatchable_rel_list[v])));
        if(t, t=test_stable_equimatchable_rel_shrink[t]);
        m[ps,vs] = t));

  if(0,
     for(p=1,test_stable_equimatchable_rel_shrink_size,
        for(v=1,test_stable_equimatchable_rel_shrink_size,
           printf(" %2d%s",
                  m[p,v],
                  if(v==test_stable_equimatchable_rel_shrink_size
                     && p==test_stable_equimatchable_rel_shrink_size,"",",")));
     print()));

  my(l=List([]));
  for(p=1,test_stable_equimatchable_rel_shrink_size,
     for(v=1,test_stable_equimatchable_rel_shrink_size,
        listput(l,m[p,v])));
  check_equal(vpar_INTERNAL_is_stable_equimatchable_table,
              Vecsmall(l),
              "vpar_INTERNAL_is_stable_equimatchable_table generated");
}


{
  \\ vpar_INTERNAL_is_stable_equimatchable_table[]
  \\ all states reached

  my(seen=vector(test_stable_equimatchable_rel_shrink_size),
    more=1);
  seen[1]=1;  \\ singleton
  while(more,
        more=0;
        for(p=1,test_stable_equimatchable_rel_shrink_size,
           for(v=1,test_stable_equimatchable_rel_shrink_size,
              if(seen[p] && seen[v],
                 my(t=test_is_stable_equimatchable_table_lookup(p,v));
                 if(t && !seen[t],
                    seen[t]=1;
                    more=1)))));
  check_equal(vecsum(seen), test_stable_equimatchable_rel_shrink_size,
             "vpar_INTERNAL_is_stable_equimatchable_table all reached");
}

test_stable_equimatchable_state_is_equivalent(x,y) =
{
  \\ same final ok-ness
  test_is_stable_equimatchable_at_root(x) ==
  test_is_stable_equimatchable_at_root(y) || return(0);

  \\ same transitions when add new child under
  for(v=1,test_is_stable_equimatchable_num_states,
     if(test_is_stable_equimatchable_table_lookup(x,v) !=
        test_is_stable_equimatchable_table_lookup(y,v),
        return(0)));

  \\ same transitions when under some parent
  for(p=1,test_is_stable_equimatchable_num_states,
     if(test_is_stable_equimatchable_table_lookup(p,x) !=
        test_is_stable_equimatchable_table_lookup(p,y),
        return(0)));

  1;
}
{
  for(x=1,test_is_stable_equimatchable_num_states,
     for(y=x+1,test_is_stable_equimatchable_num_states,
        check_equal(test_stable_equimatchable_state_is_equivalent(x,y),
                    0,
                    Str("test_stable_equimatchable_state_is_equivalent()"
                        " x="x" y="y))));
}


\\-----------------------------------------------------------------------------
progress("vpar_is_stable_equimatchable()");

check_equal(vpar_is_stable_equimatchable([0]), 1, \
            "vpar_is_stable_equimatchable() singleton");
check_equal(vpar_is_stable_equimatchable([0,1]), 1, \
            "vpar_is_stable_equimatchable() path-2");
check_equal(vpar_is_stable_equimatchable([0,1,2]), 1, \
            "vpar_is_stable_equimatchable() path-3");
check_equal(vpar_is_stable_equimatchable([0,1,2,3]), 0, \
            "vpar_is_stable_equimatchable() path-4");
check_equal(vpar_is_stable_equimatchable([0,1,2,3,4]), 0, \
            "vpar_is_stable_equimatchable() path-5");

{
  \\     5         whole not equimatchable
  \\     |         cut 2->1 (or 3->1) is path-4 not equimatchable
  \\     4
  \\     |
  \\     1
  \\    / \
  \\   2   3
  my(vpar=[4, 1, 1, 5, 0]);
  check_equal(vpar_is_equimatchable(vpar), 0,
              "vpar_is_equimatchable() for edge stable 5");
  check_equal(vpar_is_stable_equimatchable(vpar), 0);
  check_equal(test_stable_equimatchable_get_state(vpar), 6,
              "test_stable_equimatchable_get_state() 5");
}
{
  \\     2
  \\    / \        path-4 re-rooted, not equimatchable
  \\   1   3          p=3 and under v=1
  \\       |
  \\       4
  my(vpar=[2,0,4,2]);
  check_equal(vpar_is_equimatchable(vpar), 0,
              "vpar_is_equimatchable() path-4 rerooted");
  check_equal(vpar_is_stable_equimatchable(vpar), 0,
              "vpar_is_stable_equimatchable() path-4 rerooted");
}
{
  \\    1
  \\    |\ \      not equimatchable due to pair 1-2 smaller
  \\    2 3 4
  \\    |
  \\    5
  my(vpar=[0,1,1,1,2]);
  check_equal(vpar_is_equimatchable(vpar), 0,
              "vpar_is_equimatchable() 5");
  check_equal(vpar_is_stable_equimatchable(vpar), 0,
              "vpar_is_stable_equimatchable() 5");
}
{
  \\     3
  \\    / \        path-5 re-rooted
  \\   4   5       not edge stable equimatchable as cut to path-4
  \\       |
  \\       2
  \\       |
  \\       1
  my(vpar=[2, 5, 0, 3, 3]);
  check_equal(vpar_is_equimatchable(vpar), 1,
              "vpar_is_equimatchable() path-5 rerooted");
  check_equal(vpar_is_stable_equimatchable(vpar), 0,
              "vpar_is_stable_equimatchable() path-5 rerooted");
}
{
  \\      3       state=3 (shrunk)
  \\      |
  \\      6
  \\     / \
  \\    1   7
  \\    |
  \\    5
  \\   / \
  \\  2   4
  my(vpar=[6, 5, 0, 5, 1, 3, 6]);
  check_equal(vpar_is_equimatchable(vpar), 1,
              "vpar_is_equimatchable() 9 trim");
  check_equal(test_is_edge_stable_equimatchable_by_cut(vpar), 1,
              "test_is_edge_stable_equimatchable_by_cut() 9 trim");
  check_equal(test_stable_equimatchable_get_state(vpar), 3,
              "test_stable_equimatchable_get_state() 9 trim");
  check_equal(vpar_is_stable_equimatchable(vpar), 1,
              "vpar_is_stable_equimatchable() 9 trim");
}
{
  \\      3
  \\      |     8    6
  \\      7
  \\     / \
  \\    1   9
  \\    |
  \\    5
  \\   / \
  \\  2   4
  my(vpar=[7, 5, 0, 5, 1, 0, 3, 0, 7]);
  check_equal(vpar_is_equimatchable(vpar), 1,
              "vpar_is_equimatchable() 9");
  check_equal(vpar_is_stable_equimatchable(vpar), 1,
              "vpar_is_stable_equimatchable() 9");
}

\\ 1 = singleton
check_equal(test_stable_equimatchable_get_state([0]), 1, \
            "test_stable_equimatchable_get_state() 1");

\\ 2 = path-2
check_equal(test_is_stable_equimatchable_table_lookup(1,1), 2);
check_equal(vpar_INTERNAL_is_stable_equimatchable_table[1],2);
check_equal(test_stable_equimatchable_get_state([0,1]), 2, \
            "test_stable_equimatchable_get_state() 2");

\\ 3 = path-3
check_equal(test_stable_equimatchable_get_state([0,1,2]), 3, \
            "test_stable_equimatchable_get_state() 3");
\\ 4 = path-4, shrunk to 0 = non
check_equal(test_stable_equimatchable_get_state([0,1,2,3]), 0, \
            "test_stable_equimatchable_get_state() 4");
\\ 5 = star-3
check_equal(test_stable_equimatchable_get_state([0,1,1]), 4, \
            "test_stable_equimatchable_get_state() star-3");
\\ 6 = claw
check_equal(test_stable_equimatchable_get_state([0,1,2,2]), 5);
{
  \\ 7 = shrunk 6
  \\      1
  \\     /
  \\    2
  \\    |
  \\    3
  \\   / \
  \\  4   5
  my(vpar=[0,1,2,3,3]);
  check_equal(test_stable_equimatchable_get_state(vpar), 6,
            "test_stable_equimatchable_get_state() 7");

  \\ not equimatchable due to pair 2-3 smaller
  check_equal(vpar_is_equimatchable(vpar), 0,
            "vpar_is_equimatchable() of edge stable state 7");
  check_equal(vpar_maximal_matchpoly(vpar), 'x + 2*'x^2);
}
{
  \\ 8 -> shrunk 7
  \\      1
  \\     / \
  \\    3   2
  \\    |
  \\    4
  \\   / \
  \\  5   6
  my(vpar=[0,1,1,3,4,4]);
  check_equal(test_stable_equimatchable_get_state(vpar), 7,
              "test_stable_equimatchable_get_state() 8");

  \\ is equimatchable without cuts
  check_equal(vpar_is_equimatchable(vpar), 1,
              "vpar_is_equimatchable() of edge stable state 8");

  \\ but cut 2 off gives state 6 which is not equimatchable due to 3-4
  \\ smaller in the rest
  vpar[2]=0;
  check_equal(vpar_is_equimatchable(vpar), 0,
              "vpar_is_equimatchable() of edge stable state 8 cut 2");
}
{
  \\ 9 = 3 shrunk
  \\      1
  \\      |
  \\      2
  \\     / \
  \\    4   3
  \\    |
  \\    5
  \\   / \
  \\  6   7
  my(vpar=[0,1,2,2,4,5,5]);
  check_equal(test_stable_equimatchable_get_state(vpar), 3);
}
{
  \\ 10 -> 8 shrunk
  \\        *
  \\      /   \
  \\    *       *
  \\   / \     / \
  \\  *   *   *   *
  my(vpar=vpar_make_complete_tree(2,2));
  check_equal(test_stable_equimatchable_get_state(vpar), 8);
}
{
  \\ 11 -> 9 shrunk
  \\        1
  \\        |
  \\        2    s=8
  \\       / \
  \\      3   4
  \\    / |   | \
  \\   5  6   7  8
  my(vpar=vpar_join([0],1,vpar_make_complete_tree(2,2)));
  check_equal(vpar, [0, 1, 2, 2, 3, 3, 4, 4]);
  vpar_validate(vpar);
  check_equal(vpar_is_equimatchable(vpar), 0);

  check_equal(test_is_edge_stable_equimatchable_by_cut(vpar), 0);
  check_equal(test_is_vertex_stable_equimatchable_by_removals(vpar), 0);
  check_equal(test_stable_equimatchable_get_state(vpar), 9);
  check_equal(vpar_is_stable_equimatchable(vpar), 0,
              "vpar_is_stable_equimatchable() 11");

  check_equal(vpar_maximal_matchpoly(vpar), 4*'x^2 + 4*'x^3);
  check_equal(vpar_matchnum_and_count(vpar), [3, 4]);
  check_equal(vpar_is_matching_set(vpar,Set([5,7,2])), 1);
  check_equal(vpar_is_matching_set(vpar,Set([3,8])), 1);
  check_equal(test_is_edge_stable_equimatchable_no_cut_list(vpar),
              [0, 3, 4, 5, 6, 7, 8]);
}

{
  \\            6
  \\            |
  \\           10
  \\          /   \
  \\  s=8   9      5
  \\       / \
  \\      7   8
  \\    / |   | \
  \\   1  3   2  4
  my(vpar=[7, 8, 7, 8, 10, 0, 9, 9, 10, 6]);
  vpar_validate(vpar);
  check_equal(vpar_matchnum_and_count(vpar), [3, 20]);
  check_equal(vpar_maximal_matchpoly(vpar), 20*'x^3);
  check_equal(vpar_is_equimatchable(vpar), 1);
  check_equal(vpar_is_stable_equimatchable(vpar), 1);
  check_equal(test_is_edge_stable_equimatchable_by_cut(vpar), 1,
              "test_is_edge_stable_equimatchable_by_cut() 10");
  check_equal(test_is_vertex_stable_equimatchable_by_removals(vpar), 1,
              "test_is_vertex_stable_equimatchable_by_removals() 10");

  check_equal(test_stable_equimatchable_get_state(vpar), 3);
  check_equal(test_is_edge_stable_equimatchable_no_cut_list(vpar),
              [],
              "test_is_edge_stable_equimatchable_no_cut_list() 10");

  my(sans_6 = vpar_delete_vertices(vpar,[6]));
  check_equal(vpar_maximal_matchpoly(sans_6), 12*'x^3,
              "vpar_maximal_matchpoly() sans_6");
  check_equal(vpar_is_equimatchable(sans_6), 1);
  check_equal(test_is_edge_stable_equimatchable_by_cut(sans_6), 0,
              "test_is_edge_stable_equimatchable_by_cut() sans_6");
  check_equal(test_is_vertex_stable_equimatchable_by_removals(sans_6), 0,
              "test_is_vertex_stable_equimatchable_by_removals() sans_6");
  check_equal(vpar_is_stable_equimatchable(sans_6), 0,
              "vpar_is_stable_equimatchable() sans_6");
  check_equal(test_stable_equimatchable_get_state(sans_6), 7);
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got=vpar_is_stable_equimatchable(vpar));
     \\ print(vpar);
     check_equal(got, test_is_edge_stable_equimatchable_by_cut(vpar),
                 Str("vpar_is_stable_equimatchable()"
                     " vs test_is_edge_stable_equimatchable_by_cut()"));
     check_equal(got, test_is_vertex_stable_equimatchable_by_removals(vpar),
                 Str("vpar_is_stable_equimatchable()"
                     " vs test_is_vertex_stable_equimatchable_by_removals()"));
     check_equal(got <= vpar_is_equimatchable(vpar), 1);
  );
}

{
  \\ infinite size edge stable equimatchable of degree <= 4
  \\
  \\         *  4
  \\      // |           repeat 4,5 with successive star-4 above
  \\     **  *  5
  \\         |
  \\         *  4
  \\       / |
  \\      *  *  1
  \\
  my(vpar=[0,1,1]);
  for(i=1,30,
     vpar_validate(vpar);
     check_equal(vpar_is_stable_equimatchable(vpar), 1);
     vpar = vpar_join([0,1,1,1],2,vpar));
}

\\--------------------
progress("vertex removal or edge cut gives same set of components left");

test_undup_free_isomorphic(vec) =
{
  my(ret=List([]));
  for(i=1,#vec,
     for(j=1,#ret,
        if(vpar_is_free_isomorphic(vec[i],ret[j]), next(2)));
     listput(ret,vec[i]));
  Vec(ret);
}
test_forest_components(vpar) =
{
  apply(vec->vpar_keep_vertices(vpar,vec),
       vpar_component_vertices(vpar));
}
test_vertex_removal_components(vpar) =
{
  my(vec=[vpar]);
  for(v=1,#vpar,
     my(vpar_del=vpar_delete_vertex(vpar,v));
     vec=concat(vec,test_forest_components(vpar_del)));
  vec=select(vpar->#vpar, vec);  \\ no empties
  vec=test_undup_free_isomorphic(vec);
  Set(vec);
}
check_equal(test_vertex_removal_components([0,1]), [ [0], [0,1] ]);

test_edge_cut_components(vpar) =
{
  my(vec=[vpar]);
  for(v=1,#vpar,
     my(vpar_cut=vpar);
     vpar_cut[v]=0;
     vec=concat(vec,test_forest_components(vpar_cut)));
  vec=select(vpar->#vpar, vec);  \\ no empties
  vec=test_undup_free_isomorphic(vec);
  Set(vec);
}
check_equal(test_vertex_removal_components([0,1]), [ [0], [0,1] ]);

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     if(#vpar==0, next());

     my(e = test_edge_cut_components(vpar),
        v = test_vertex_removal_components(vpar));
     \\ print("vpar="vpar"  e="e"  v="v);
     check_equal(#e, #v,
                 "edge cut and vertex remove same number of components");
     check_equal(test_undup_free_isomorphic(concat(e,v)), e,
                 "edge cut and vertex remove same components");
  );
}

{
  \\ equimatchable state of path-n is smallest having its state
  for(n=2,5,
     my(path=vpar_make_path(n),
        vpar=vpar_unlabelled_first(n));
     until((vpar=vpar_unlabelled_next(vpar)) == 'none,
           my(state=test_equimatchable_get_state(vpar));
           \\ print("n="n" state="state" in "vpar);
           check_equal(state <= n, 1,
                      "vpar_is_equimatchable() expect no vpar has n < state");
           if(state==n,
              check_equal(vpar_is_isomorphic(vpar,path), 1,
                      "vpar_is_equimatchable() only vpar n == state is path"));
          ));
}


\\-----------------------------------------------------------------------------
progress("vpar_INTERNAL_graph6_size_str()");

check_equal(vpar_INTERNAL_graph6_size_str(30),    Strchr([93]), \
            "vpar_INTERNAL_graph6_size_str() 30");
check_equal(vpar_INTERNAL_graph6_size_str(12345), Strchr([126, 66,63,120]), \
            "vpar_INTERNAL_graph6_size_str() 12345");

\\ 460175067 = octal 3333333333  < 2^29
check_equal(digits(460175067,8), [3,3,3,3,3,3,3,3,3,3]);
check_equal(vpar_INTERNAL_graph6_size_str(460175067), \
            Strchr([126,126,  63, 90, 90, 90, 90, 90]));

check_equal(vpar_INTERNAL_graph6_size_str(2^31 - 1), \
            Strchr([126,126,  63+1, 126, 126, 126, 126, 126]))
check_equal(vpar_INTERNAL_graph6_size_str(2^36 - 1), \
            Strchr([126,126,  126, 126, 126, 126, 126, 126]))


\\-----------------------------------------------------------------------------
progress("vpar_to_graph6()");

check_equal(vpar_to_graph6([]), "?");
check_equal(vpar_to_graph6([0]), "@");
check_equal(vpar_to_graph6([0,0]), "A?");
check_equal(vpar_to_graph6([2,0]), Strchr([63+2, 63+32]));
check_equal(vpar_to_graph6([0,1]), Strchr([63+2, 63+32]));
check_equal(Strchr([63+2, 63+32]), "A_");

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        str=vpar_to_graph6(vpar),
        vec=Vecsmall(str));
     check_equal(#vec >= 1, 1, "vpar_to_graph6() length");
     \\ check_equal(vec[#vec], 10, "vpar_to_graph6() final newline");
  );
}


\\------------------------------------------------------------------------------
progress("vpar_free_next()");

check_equal(vpar_free_first(0), []);
check_equal(vpar_free_next([]), 'none);

{
  \\ Michael Dineen, "Constant Time Generation of Free Trees"
  \\ example of condition (b) on heights fails

  \\         1
  \\       / | \  \  \
  \\     2   6  8  9  10
  \\     |   |
  \\     3   7
  \\    / \
  \\   4   5
  \\
  my(vpar=[0,1,2,3,3, 1,6, 1,1,1]);
  check_equal(#vpar, 10);
  vpar_validate(vpar);
  check_equal(vpar_is_preorder(vpar), 1);
  check_equal(vpar_centres(vpar), [1,2]);
  my(depths=vpar_depths_vector(vpar));
  check_equal(depths,    [0,1,2,3,3, 1,2, 1,1,1]);

  \\          1
  \\        /   \
  \\      2      8
  \\    / | \    |
  \\   3  5  7   9
  \\   |  |      |
  \\   4  6     10
  \\
  my(next_vpar=vpar_free_next(vpar),
     next_depths=vpar_depths_vector(next_vpar));
  \\ print("next_vpar   "next_vpar);
  \\ print("next_depths "next_depths);
  check_equal(vpar_is_preorder(next_vpar), 1,
              "vpar_free_next() 10(b) preorder");
  check_equal(next_depths, [0,1,2,3,2,3,2, 1,2,3],
              "vpar_free_next() 10(b) depths");
  \\ check_equal(vpar_preorder_depths_is_free_primary(next_depths), 1,
  \\   "vpar_free_next() 10(b) vpar_preorder_depths_is_free_primary()");
  check_equal(next_vpar, [0,1,2,3,2,5, 2, 1,8,9],
              "vpar_free_next() 10(b) vpar");
}
{
  \\ Michael Dineen, "Constant Time Generation of Free Trees"
  \\ example condition (c) fails

  \\         1
  \\       /    \  \         l=7 level=2
  \\     2       8  10
  \\   /|||\     |
  \\ 3 4 5 6 7   9
  \\
  my(vpar=[0,1,2,2,2,2,2, 1,8, 1]);
  vpar_validate(vpar);
  check_equal(vpar_is_preorder(vpar), 1);
  check_equal(vpar_centres(vpar), [1]);

  my(depths=vpar_depths_vector(vpar));
  check_equal(depths, [0,1,2,2,2,2,2, 1,2,1]);
  \\ check_equal(vpar_preorder_depths_is_free_primary(depths), 1,
  \\             "vpar_preorder_depths_is_free_primary() 10(c)");

  \\         1
  \\       /   \
  \\     2      7
  \\   /||\    /|\
  \\ 3 4 5 6  8 9 10
  \\
  my(next_vpar=vpar_free_next(vpar),
     next_depths=vpar_depths_vector(next_vpar));
  \\ print("next_vpar   "next_vpar);
  \\ print("next_depths "next_depths);
  check_equal(vpar_is_preorder(next_vpar), 1,
              "vpar_free_next() 10(c) preorder");
  check_equal(next_depths, [0,1,2,2,2,2, 1,2,2,2],
              "vpar_free_next() 10(c) depths");
  \\ check_equal(vpar_preorder_depths_is_free_primary(next_depths), 1,
  \\   "vpar_free_next() 10(c) vpar_preorder_depths_is_free_primary()");
  check_equal(next_vpar, [0,1,2,2,2,2, 1,7,7,7],
              "vpar_free_next() 10(c) vpar");
}
{
  \\ Michael Dineen, "Constant Time Generation of Free Trees"
  \\ example condition (d) fails

  \\          1
  \\        /  \  \
  \\      2     7   9                 l=6 level=2
  \\    / |     |   |
  \\   3  5     8   10
  \\   |  |
  \\   4  6
  \\
  my(vpar=[0,1,2,3,2,5, 1,7, 1,9]);
  vpar_validate(vpar);
  check_equal(vpar_is_preorder(vpar), 1);
  check_equal(vpar_centres(vpar), [1,2]);

  my(depths=vpar_depths_vector(vpar));
  check_equal(depths, [0,1,2,3,2,3, 1,2, 1,2]);
  \\ check_equal(vpar_preorder_depths_is_free_primary(depths), 1,
  \\             "vpar_preorder_depths_is_free_primary() 10(d)");

  \\         1
  \\       /   \
  \\     2      8
  \\   /||\     |
  \\ 3 5 6 7    9
  \\ |          |
  \\ 4         10
  \\
  my(next_vpar=vpar_free_next(vpar),
     next_depths=vpar_depths_vector(next_vpar));
  \\ print("next_vpar   "next_vpar);
  \\ print("next_depths "next_depths);
  check_equal(vpar_is_preorder(next_vpar), 1,
              "vpar_free_next() 10(d) preorder");
  \\ check_equal(vpar_preorder_depths_is_free_primary(next_depths), 1,
  \\   "vpar_free_next() 10(d) vpar_preorder_depths_is_free_primary()");

  check_equal(next_vpar, [0,1,2,3,2,2,2, 1,8,9],
              "vpar_free_next() 10(d) vpar");
  check_equal(next_depths, [0,1,2,3,2,2,2, 1,2,3],
              "vpar_free_next() 10(d) depths");

}

{
  \\ n=10 first
  \\         1
  \\       /  \
  \\     2     7
  \\     |     |
  \\     3     8
  \\     |     |
  \\     4     9
  \\     |     |
  \\     5    10
  \\     |
  \\     6
  \\
  my(vpar=vpar_free_first(10));
  check_equal(vpar, [0, 1, 2, 3, 4, 5, 1, 7, 8, 9]);

  my(next_vpar=vpar_free_next(vpar),
     next_depths=vpar_depths_vector(next_vpar));
  \\ print(next_vpar);
  \\ print(next_depths);
  check_equal(vpar_is_preorder(next_vpar), 1);
  \\ check_equal(vpar_preorder_depths_is_free_primary(next_depths), 1);
}
{
  \\ n=5
  \\         1                 1                      1
  \\       /  \               /| \                /  / \  \
  \\     2     4            2  4  5              2  3   4  5
  \\     |     |           /
  \\     3     5          3
  \\   h=1   h=2         h=1   h=1             h=0   h=1
  \\                    size=2 size=3
  \\
  my(vpar=vpar_free_first(5));
  check_equal(vpar, [0, 1, 2, 1, 4]);

  \\ check_equal(vpar_preorder_depths_is_free_primary([0,1,2,1,2]), 1);
  \\ check_equal(vpar_preorder_depths_is_free_primary([0,1,2,1,1]), 1);
  \\ check_equal(vpar_preorder_depths_is_free_primary([0,1,1,1,1]), 1);

  my(next_vpar=vpar_free_next(vpar),
     next_depths=vpar_depths_vector(next_vpar));
  \\ print("next_vpar   "next_vpar);
  \\ print("next_depths "next_depths);
  check_equal(vpar_is_preorder(next_vpar), 1);
  \\ check_equal(vpar_preorder_depths_is_free_primary(next_depths), 1);
  check_equal(test_depths_is_free_primary_by_vectors(next_depths), 1);

  next_vpar=vpar_free_next(next_vpar);
  next_depths=vpar_depths_vector(next_vpar);
  \\ print("next_next_vpar   "next_vpar);
  \\ print("next_next_depths "next_depths);
  check_equal(vpar_is_preorder(next_vpar), 1);
  \\ check_equal(vpar_preorder_depths_is_free_primary(next_depths), 1);
  check_equal(test_depths_is_free_primary_by_vectors(next_depths), 1);
  check_equal(next_vpar, vpar_free_last(5));

  check_equal(vpar_count_free(5), 3);
}
{
  \\ n=7
  \\         1               1
  \\       /  \            /  \  \
  \\     2     5         2     5   7
  \\     |    / \        |     |
  \\     3   6   7       3     6
  \\     |               |
  \\     4               4
  \\    h=2   h=2       h=2   h=2
  \\   size=3 size=4   size=3 size=4
  \\
}
{
  \\ n=10
  \\           1                   1
  \\         /  \  \             /  \
  \\       2     7  10         2     7
  \\      /      |            /     /|\
  \\    3        8          3      8 9 10
  \\   /|\       |         /|\
  \\  4 5 6      9        4 5 6
  \\   h=2     h=3         h=2     h=2
  \\ size=5  size=5      size=5   size=5
  \\
  my(vpar=[0, 1, 2, 3, 3, 3, 1, 7, 8, 1]);
  vpar_validate(vpar);

  my(next_vpar=vpar_free_next(vpar),
     next_depths=vpar_depths_vector(next_vpar));
  \\ print("next_vpar   "next_vpar);
  \\ print("next_depths "next_depths);
  check_equal(vpar_is_preorder(next_vpar), 1);
  \\ check_equal(vpar_preorder_depths_is_free_primary(next_depths), 1);

  check_equal(test_depths_is_free_primary_by_vectors(next_depths), 1);
  \\ check_equal(next_vpar, test_free_next_by_skip(vpar));
}

{
  for(n=0,10,
     \\ print("n="n);
     my(vpar=vpar_free_first(n),
        l=List([]));
     until((vpar=vpar_free_next(vpar))=='none,
           vpar_validate(vpar);
           listput(l,vpar);

           check_equal(vpar_is_preorder(vpar), 1,
                       "vpar_free_next() is preorder");
           my(depths=vpar_depths_vector(vpar));
           \\ check_equal(vpar_preorder_depths_is_free_primary(depths), 1,
           \\ "vpar_free_next() vs vpar_preorder_depths_is_free_primary()");
          );
     \\ print(l);
     check_equal(#l, vpar_count_free(n),
                 Str("vpar_free_next() vs vpar_count_free() n="n));

     for(i=1,#l,
        for(j=1,i-1,
           check_equal(vpar_is_free_isomorphic(l[i],l[j]), 0,
           "vpar_free_next() not vpar_is_free_isomorphic() to each other")));
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_is_almost_equimatchable() states");

\\                                          /-- smallest maximal --\
\\                                         with   without without_max   matchnum
test_almost_equimatchable_rel_initial = ['no_with,   0,       0,       'unmatched];

test_almost_equimatchable_rel_validate(rel) =
{
  my(with,without,without_max,matched);
  [with,without,without_max,matched] = rel;

  if(without_max!='no_without_max,
     without_max >= without || error());
}

test_almost_equimatchable_rel_add(p_rel,v_rel) =
{
  \\ print("test_almost_equimatchable_rel_add() "p_rel"  "v_rel);
  test_almost_equimatchable_rel_validate(p_rel);
  test_almost_equimatchable_rel_validate(v_rel);

  my(v_with,v_without,v_without_max,v_matched);
  [v_with,v_without,v_without_max,v_matched] = v_rel;
  my(p_with,p_without,p_without_max,p_matched);
  [p_with,p_without,p_without_max,p_matched] = p_rel;

  my(v_max = min(if(v_with=='no_with,               99, v_with),
                 if(v_without_max=='no_without_max, 99, v_without_max)));
  \\ print(" v_max  "v_max);
  if(v_max==99, error());

  \\ with[p] = min(if(with_exists[p], with[p] + v_max,  #vpar),
  \\               without[p] + without[v] + 1);
  my(ret_with = min(if(p_with=='no_with, 99,  p_with + v_max),
                    p_without + v_without  + 1));
  \\ print(" ret_with  "ret_with);
  if(ret_with==99, error());

  ret_without = p_without + v_max;

  \\ without_max[p] += with[v];
  \\ if(!with_exists[v], without_max_exists[p] = 0);
  my(ret_without_max = if(p_without_max=='no_without_max || v_with=='no_with,
                          'no_without_max,
                          p_without_max + v_with));
  if(ret_without_max==99, error());
  \\ print(" ret_without_max  "ret_without_max);

  my(ret_matched=p_matched,
     matchnum_offset=0);
  if(p_matched=='unmatched && v_matched=='unmatched,
     ret_matched='matched; matchnum_offset=1);

  \\ without[p] += v_max;
  [if(ret_with=='no_with, ret_with,
                          ret_with        - matchnum_offset),
   ret_without                            - matchnum_offset,
   if(ret_without_max=='no_without_max, ret_without_max,
                          ret_without_max - matchnum_offset),
   ret_matched];
}

\\ rel of states 1 to 5
test_almost_equimatchable_rel_list = List([]);

listput(test_almost_equimatchable_rel_list, \
        test_almost_equimatchable_rel_initial);
{
  \\ 2 = path-2
  \\    *
  \\    |
  \\    *
  listput(test_almost_equimatchable_rel_list,
          test_almost_equimatchable_rel_add
          (test_almost_equimatchable_rel_list[1],
           test_almost_equimatchable_rel_list[1]));
}
{
  \\ 3 = path-3
  \\    *
  \\    |
  \\    *
  \\    |
  \\    *
  listput(test_almost_equimatchable_rel_list,
          test_almost_equimatchable_rel_add
          (test_almost_equimatchable_rel_list[1],
           test_almost_equimatchable_rel_list[2]));
}
{
  \\ 4 = path-4
  \\    *
  \\    |
  \\    *
  \\    |
  \\    *
  \\    |
  \\    *
  listput(test_almost_equimatchable_rel_list,
          test_almost_equimatchable_rel_add
          (test_almost_equimatchable_rel_list[1],
           test_almost_equimatchable_rel_list[3]));
}
{
  \\ 5 = path-5
  listput(test_almost_equimatchable_rel_list,
          test_almost_equimatchable_rel_add
          (test_almost_equimatchable_rel_list[1],
           test_almost_equimatchable_rel_list[4]));
}
{
  \\ 6 = path-6
  listput(test_almost_equimatchable_rel_list,
          test_almost_equimatchable_rel_add
          (test_almost_equimatchable_rel_list[1],
           test_almost_equimatchable_rel_list[5]));
}
{
  \\ 7 = path-7
  listput(test_almost_equimatchable_rel_list,
          test_almost_equimatchable_rel_add
          (test_almost_equimatchable_rel_list[1],
           test_almost_equimatchable_rel_list[6]));
}
{
  \\ 8 = path-8
  listput(test_almost_equimatchable_rel_list,
          test_almost_equimatchable_rel_add
          (test_almost_equimatchable_rel_list[1],
           test_almost_equimatchable_rel_list[7]));
}
{
  \\ 9 = path-9
  listput(test_almost_equimatchable_rel_list,
          test_almost_equimatchable_rel_add
          (test_almost_equimatchable_rel_list[1],
           test_almost_equimatchable_rel_list[8]));
}
{
  \\ 10 = path-10
  listput(test_almost_equimatchable_rel_list,
          test_almost_equimatchable_rel_add
          (test_almost_equimatchable_rel_list[1],
           test_almost_equimatchable_rel_list[9]));
}
{
  \\ 11 = path-11
  listput(test_almost_equimatchable_rel_list,
          test_almost_equimatchable_rel_add
          (test_almost_equimatchable_rel_list[1],
           test_almost_equimatchable_rel_list[10]));
}
{
  \\ 12 = path-4 root at second
  \\     *
  \\   / |
  \\  *  *
  \\     |
  \\     *
  listput(test_almost_equimatchable_rel_list,
          test_almost_equimatchable_rel_add
          (test_almost_equimatchable_rel_list[2],
           test_almost_equimatchable_rel_list[2]));
}
{
  \\ 13 = path-6 root at second
  \\     *
  \\   / |
  \\  *  *
  \\     |
  \\     *
  \\     |
  \\     *
  \\     |
  \\     *
  listput(test_almost_equimatchable_rel_list,
          test_almost_equimatchable_rel_add
          (test_almost_equimatchable_rel_list[2],
           test_almost_equimatchable_rel_list[4]));
}
{
  \\ 14 = path-7 root at third
  \\     *
  \\   / |
  \\  *  *
  \\  |  |
  \\  *  *
  \\     |
  \\     *
  \\     |
  \\     *
  listput(test_almost_equimatchable_rel_list,
          test_almost_equimatchable_rel_add
          (test_almost_equimatchable_rel_list[3],
           test_almost_equimatchable_rel_list[4]));
}
{
  \\ 15 = path-9 root at middle
  \\     *
  \\   / |
  \\  *  *
  \\  |  |
  \\  *  *
  \\  |  |
  \\  *  *
  \\  |  |
  \\  *  *
  listput(test_almost_equimatchable_rel_list,
          test_almost_equimatchable_rel_add
          (test_almost_equimatchable_rel_list[5],
           test_almost_equimatchable_rel_list[4]));
}
{
  \\ 16 = path-13 root at middle
  \\     *
  \\   / |
  \\  *  *
  \\  |  |
  \\  *  *
  \\  |  |
  \\  *  *
  \\  |  |
  \\  *  *
  \\  |  |
  \\  *  *
  \\  |  |
  \\  *  *
  listput(test_almost_equimatchable_rel_list,
          test_almost_equimatchable_rel_add
          (test_almost_equimatchable_rel_list[7],
           test_almost_equimatchable_rel_list[6]));
}
test_almost_equimatchable_rel_list = Vec(test_almost_equimatchable_rel_list);

{
  if(0,
     for(i=1,#test_almost_equimatchable_rel_list,
        my(rel=test_almost_equimatchable_rel_list[i]);
        print("state="i"  ",rel));
     );
}

\\ Return the smallest size of a maximal matching at rel, in the form of a
\\ difference matchnum - size.  This is always <= 0.
test_almost_equimatchable_rel_v_max(rel) =
{
  my(with,without,without_max,matched);
  [with,without,without_max,matched] = rel;

  my(ret=min(if(with=='no_with,               99999, with),
             if(without_max=='no_without_max, 99999, without_max)));
  ret <= 0 || error();
  ret;
}
{
  if(0,
     for(i=1,#test_almost_equimatchable_rel_list,
        my(rel=test_almost_equimatchable_rel_list[i]);
        print("n="i"  "test_almost_equimatchable_rel_v_max(rel),"   ",rel)));
}

\\ Return the state number 1 to 5 for the given rel, or return 0 if not one
\\ of the 5 states in test_almost_equimatchable_rel_list.
test_almost_equimatchable_rel_find(rel) =
{
  for(i=1,#test_almost_equimatchable_rel_list,
     if(test_almost_equimatchable_rel_list[i] == rel, return(i)));
  0;
}

\\ Return 1 if rel is a non almost-equimatchable state, meaning that no
\\ matter what parent or further children, this vertex has a maximal <
\\ maximum - 1 and so makes its whole tree likewise not almost
\\ equimatchable.
\\
test_almost_equimatchable_rel_is_non(rel) =
{
  test_almost_equimatchable_rel_validate(rel);

  my(with,without,without_max,matched);
  [with,without,without_max,matched] = rel;

  with !='no_with && with <= -2;
}
{
  for(i=1,#test_almost_equimatchable_rel_list,
     my(rel=test_almost_equimatchable_rel_list[i]);
     check_equal(test_almost_equimatchable_rel_is_non(rel), 0,
                 "test_almost_equimatchable_rel_is_non() on states"));
}
check_equal(test_almost_equimatchable_rel_is_non([-2,0,0,'matched]),   1);
check_equal(test_almost_equimatchable_rel_is_non([-2,0,0,'unmatched]), 1);
check_equal(test_almost_equimatchable_rel_is_non(['no_with,0,0,'matched]),   0);
check_equal(test_almost_equimatchable_rel_is_non([0,-2,-2,'matched]),   0);
check_equal(test_almost_equimatchable_rel_is_non([0,-2,-2,'unmatched]), 0);

\\ {
\\   \\ extend test_almost_equimatchable_rel_list with necessary further states
\\   my(more=1);
\\   while(more,
\\         more=0;
\\         for(p=1,#test_almost_equimatchable_rel_list,
\\            for(v=1,#test_almost_equimatchable_rel_list,
\\               my(rel=test_almost_equimatchable_rel_add
\\                      (test_almost_equimatchable_rel_list[p],
\\                       test_almost_equimatchable_rel_list[v]));
\\               if(test_almost_equimatchable_rel_is_non(rel), next());
\\               if(!test_almost_equimatchable_rel_find(rel),
\\                  print("add "rel);
\\                  more=1;
\\                  listput(test_almost_equimatchable_rel_list,rel)))));
\\ }
{
  for(p=1,#test_almost_equimatchable_rel_list,
     for(v=1,#test_almost_equimatchable_rel_list,
        my(rel=test_almost_equimatchable_rel_add
               (test_almost_equimatchable_rel_list[p],
                test_almost_equimatchable_rel_list[v]));

        check_equal(test_almost_equimatchable_rel_find(rel)
                    || test_almost_equimatchable_rel_is_non(rel), 1,
                    Str("test_almost_equimatchable_rel transition"
                        " p="p" v="v" to "rel))));
}
\\ print("test_almost_equimatchable_rel_list  ",#test_almost_equimatchable_rel_list);

{
  test_almost_equimatchable_rel_table
  = matrix(#test_almost_equimatchable_rel_list,
           #test_almost_equimatchable_rel_list,
           p,v,
           test_almost_equimatchable_rel_find
           (test_almost_equimatchable_rel_add
            (test_almost_equimatchable_rel_list[p],
             test_almost_equimatchable_rel_list[v])));
}

{
  check_equal(test_almost_equimatchable_rel_list,
              [['no_with, 0,        0,        'unmatched],   \\ 1
               [    0,   -1, 'no_without_max, 'matched],     \\ 2
               [    0,    0,        0,        'unmatched],   \\ 3
               [    0,   -1,       -1,        'matched],     \\ 4
               [    0,   -1,        0,        'unmatched],   \\ 5
               [   -1,   -1,       -1,        'matched],
               [    0,   -1,       -1,        'unmatched],
               [   -1,   -2,       -1,        'matched],
               [   -1,   -1,       -1,        'unmatched],
               [   -1,   -2,       -2,        'matched],     \\ 10
               [   -1,   -2,       -1,        'unmatched],   \\ 11

               [   -1,   -1, 'no_without_max, 'matched],     \\ 12
               [   -1,   -2, 'no_without_max, 'matched],
               [   -1,   -1,        0,        'unmatched],
               [   -1,   -2,        0,        'unmatched],
               [   -1,   -2,       -2,        'unmatched]    \\ 16
              ]);
}

\\ Return 1 if rel is not an edge almost equimatchable state as a root.
\\ Per vpar_is_equimatchable() states, cannot have 4 as that is a
\\ non-equimatchable as a root.
\\
test_almost_equimatchable_rel_is_ok_at_root(rel) =
{
  test_almost_equimatchable_rel_v_max(rel) == -1;
}
{
  test_almost_equimatchable_rel_ok_at_root
  = apply(rel-> test_almost_equimatchable_rel_is_ok_at_root(rel),
         test_almost_equimatchable_rel_list);
}

\\ 11,15 are doomed above or below, but are ok as a root
check_equal(vecsum(test_almost_equimatchable_rel_table[11,]), 0);
check_equal(vecsum(test_almost_equimatchable_rel_table[,11]), 0);
check_equal(test_almost_equimatchable_rel_ok_at_root[11], 1);
\\
check_equal(vecsum(test_almost_equimatchable_rel_table[15,]), 0);
check_equal(vecsum(test_almost_equimatchable_rel_table[,15]), 0);
check_equal(test_almost_equimatchable_rel_ok_at_root[15], 1);

{
  \\ 16 is not at root, and goes to non for anything under or above
  check_equal(vecsum(test_almost_equimatchable_rel_table[16,]), 0);
  check_equal(vecsum(test_almost_equimatchable_rel_table[,16]), 0);
  check_equal(test_almost_equimatchable_rel_ok_at_root[16], 0);

  \\ drop 16 as a target
  check_equal(test_almost_equimatchable_rel_table[7,6], 16);
  check_equal(test_almost_equimatchable_rel_table[7,12], 16);
  my(count=0);
  for(p=1,#test_almost_equimatchable_rel_list,
     for(v=1,#test_almost_equimatchable_rel_list,
        my(t=test_almost_equimatchable_rel_table[p,v]);
        if(t==16,
           test_almost_equimatchable_rel_table[p,v] = 0;
           count++)));
  check_equal(count, 2);   \\ 16 occurs in 2 places
}

test_almost_equimatchable_rel_is_equivalent(x,y) =
{
  \\ same final ok-ness
  test_almost_equimatchable_rel_ok_at_root[x] ==
  test_almost_equimatchable_rel_ok_at_root[y] || return(0);

  my(m=apply(s->if(s==y,x,s), test_almost_equimatchable_rel_table));

  \\ same transitions when add new child under
  m[x,] == m[y,] || return(0);

  \\ same transitions when are a child under some parent
  m[,x] == m[,y] || return(0);

  1;
}

{
  for(x=1,#test_almost_equimatchable_rel_list,
     for(y=x+1,#test_almost_equimatchable_rel_list,
        check_equal(test_almost_equimatchable_rel_is_equivalent(x,y),
                       (x==6 && y==12)
                    || (x==8 && y==13)
                    || (x==9 && y==14)
                    || (x==11 && y==15)
                    ,
                    Str("test_almost_equimatchable_rel_is_equivalent()"
                        " x="x" y="y))));
}

\\ Might be able to make an argument why these variations in without_max
\\ don't make a difference, not with the conditions of the other fields at
\\ least.  For now easy enough to notice mechanically the equivalence in
\\ test_almost_equimatchable_rel_table etc.
\\
check_equal(test_almost_equimatchable_rel_list[6],  [-1, -1, -1, 'matched]);
check_equal(test_almost_equimatchable_rel_list[12], [-1, -1, 'no_without_max, 'matched]);

check_equal(test_almost_equimatchable_rel_list[8],  [-1, -2, -1, 'matched]);
check_equal(test_almost_equimatchable_rel_list[13], [-1, -2, 'no_without_max, 'matched]);

check_equal(test_almost_equimatchable_rel_list[9],  [-1, -1, -1, 'unmatched]);
check_equal(test_almost_equimatchable_rel_list[14], [-1, -1,  0, 'unmatched]);

check_equal(test_almost_equimatchable_rel_list[11], [-1, -2, -1, 'unmatched]);
check_equal(test_almost_equimatchable_rel_list[15], [-1, -2,  0, 'unmatched]);

\\ drop 16
\\ 15 -> 11
\\ 14 -> 9
\\ 13 -> 8
\\ 12 -> 6
\\                                      1,2,3,4,5,6,7,8,9,10,11,12,13 14 15 16
test_almost_equimatchable_rel_shrink = [1,2,3,4,5,6,7,8,9,10,11, 6, 8, 9,11, 0];
test_almost_equimatchable_rel_shrink_size = \
  #test_almost_equimatchable_rel_list - 5;
{
  check_equal(test_almost_equimatchable_rel_shrink_size, 11,
              "test_almost_equimatchable_rel_shrink_size");
  check_equal(vecmax(test_almost_equimatchable_rel_shrink),
              test_almost_equimatchable_rel_shrink_size,
              "vecmax test_almost_equimatchable_rel_shrink");
}
{
  \\ vpar_INTERNAL_is_almost_equimatchable_table[]
  \\ generated
  my(m=matrix(test_almost_equimatchable_rel_shrink_size,
              test_almost_equimatchable_rel_shrink_size));
  for(p=1,#test_almost_equimatchable_rel_list,
     my(ps=test_almost_equimatchable_rel_shrink[p]);
     if(!ps, next());
     for(v=1,#test_almost_equimatchable_rel_list,
        my(vs=test_almost_equimatchable_rel_shrink[v]);
        if(!vs, next());

        my(t=test_almost_equimatchable_rel_find
              (test_almost_equimatchable_rel_add
               (test_almost_equimatchable_rel_list[p],
                test_almost_equimatchable_rel_list[v])));
        if(t, t=test_almost_equimatchable_rel_shrink[t]);
        m[ps,vs] = t));

  if(0,
     for(p=1,test_almost_equimatchable_rel_shrink_size,
        for(v=1,test_almost_equimatchable_rel_shrink_size,
           printf(" %2d%s",
                  m[p,v],
                  if(v==test_almost_equimatchable_rel_shrink_size
                     && p==test_almost_equimatchable_rel_shrink_size,"",",")));
     print()));

  my(l=List([]));
  for(p=1,test_almost_equimatchable_rel_shrink_size,
     for(v=1,test_almost_equimatchable_rel_shrink_size,
        listput(l,m[p,v])));
  check_equal(vpar_INTERNAL_is_almost_equimatchable_table,
              Vecsmall(l),
              "vpar_INTERNAL_is_almost_equimatchable_table generated");
}

check_equal(test_is_almost_equimatchable_num_states, \
            test_almost_equimatchable_rel_shrink_size);

{
  \\ vpar_is_almost_equimatchable() ok constant, inline in code
  \\ generated
  my(want=0);
  for(i=1,#test_almost_equimatchable_rel_list,
     if(test_almost_equimatchable_rel_ok_at_root[i],
        my(state=test_almost_equimatchable_rel_shrink[i]);
        if(!state, next());
        want=bitor(want, 1 << state)));
  check_equal(3024,
              want,
              "vpar_is_almost_equimatchable() at root bits constant");
  check_equal(test_is_almost_equimatchable_at_root_bits,
              want);
}

test_almost_equimatchable_state_is_equivalent(x,y) =
{
  \\ same final ok-ness
  test_is_almost_equimatchable_at_root(x) ==
  test_is_almost_equimatchable_at_root(y) || return(0);

  \\ same transitions when add new child under
  for(v=1,test_is_almost_equimatchable_num_states,
     if(test_is_almost_equimatchable_table_lookup(x,v) !=
        test_is_almost_equimatchable_table_lookup(y,v),
        return(0)));

  \\ same transitions when under some parent
  for(p=1,test_is_almost_equimatchable_num_states,
     if(test_is_almost_equimatchable_table_lookup(p,x) !=
        test_is_almost_equimatchable_table_lookup(p,y),
        return(0)));

  1;
}
{
  for(x=1,test_is_almost_equimatchable_num_states,
     for(y=x+1,test_is_almost_equimatchable_num_states,
        check_equal(test_almost_equimatchable_state_is_equivalent(x,y),
                    0,
                    Str("test_almost_equimatchable_state_is_equivalent()"
                        " x="x" y="y))));
}

{
  \\ almost equimatchable states are as at the end of respective path-n
  for(n=1,11,
     check_equal(test_almost_equimatchable_get_state(vpar_make_path(n)), n));
}
\\ but 6 also second of path-4
check_equal(test_almost_equimatchable_get_state([0,1,1,2]), 6);
{
  \\ and 6 also
  \\    *
  \\   / \
  \\  *   *
  \\     / \
  \\    *   *
  my(vpar=[0,1,1,2,2]);
  vpar_validate(vpar);
  check_equal(test_almost_equimatchable_get_state(vpar), 6);
}
{
  \\ and 6 also
  \\    *
  \\   /|\
  \\  * * *
  \\      |
  \\      *
  my(vpar=[0,1,1,1,2]);
  vpar_validate(vpar);
  check_equal(test_almost_equimatchable_get_state(vpar), 6);
}
{
  \\ and 7 also
  \\     *
  \\     |
  \\     *
  \\   / |
  \\  *  *
  \\     |
  \\     *
  my(vpar=[0,1,2,2,3]);
  vpar_validate(vpar);
  check_equal(test_almost_equimatchable_get_state(vpar), 7);
}
{
  \\ and 8 also
  \\     *
  \\     |
  \\     *
  \\     |
  \\     *
  \\   / |
  \\  *  *
  \\     |
  \\     *
  my(vpar=[0, 1, 2, 3, 4, 3]);
  vpar_validate(vpar);
  check_equal(test_almost_equimatchable_get_state(vpar), 8);
}
{
  \\ almost equimatchable state of path-n is not the smallest having its state,
  \\ or perhaps only state<=5
  for(n=2,6,
     my(path=vpar_make_path(n),
        vpar=vpar_unlabelled_first(n));
     until((vpar=vpar_unlabelled_next(vpar)) == 'none,
           my(state=test_almost_equimatchable_get_state(vpar));
           if(state==6 && vpar==[0, 1, 2, 1], next());
           if(state==6 && vpar==[0, 1, 2, 2, 1], next());
           if(state==6 && vpar==[0, 1, 2, 1, 1], next());
           if(state==7, next());
           if(state==8, next());
           \\ print("n="n" state="state" in "vpar);
           check_equal(state <= n, 1,
                       Str("vpar_is_almost_equimatchable()"
                           " expect no vpar has n < state\n"
                           " state="state" n="n" vpar="vpar));
          ));
}


\\-----------------------------------------------------------------------------
progress("vpar_is_almost_equimatchable()");

check_equal(test_is_almost_equimatchable_by_matchpoly([]), 0);
check_equal(test_is_almost_equimatchable_by_matchpoly([0]), 0);
check_equal(test_is_almost_equimatchable_by_matchpoly([0,1]), 0);
check_equal(test_is_almost_equimatchable_by_matchpoly([0,1,2]), 0);
check_equal(test_is_almost_equimatchable_by_matchpoly([0,1,2,3]), 1);
check_equal(test_is_almost_equimatchable_by_matchpoly([0,1,2,3,0]), 0);

check_equal(vpar_is_almost_equimatchable([0]), 0, \
            "vpar_is_almost_equimatchable() singleton");
check_equal(vpar_is_almost_equimatchable([0,1]), 0, \
            "vpar_is_almost_equimatchable() path-2");
check_equal(vpar_is_almost_equimatchable([0,1,2]), 0, \
            "vpar_is_almost_equimatchable() path-3");
check_equal(vpar_is_almost_equimatchable([0,1,2,3]), 1, \
            "vpar_is_almost_equimatchable() path-4");
check_equal(vpar_is_almost_equimatchable([0,1,2,3,4]), 0, \
            "vpar_is_almost_equimatchable() path-5");

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        got=vpar_is_almost_equimatchable(vpar));
     \\ print(vpar);
     check_equal(got, test_is_almost_equimatchable_by_matchpoly(vpar),
                 Str("vpar_is_almost_equimatchable()"
                     " vs test_is_almost_equimatchable_by_matchpoly()"));
     check_equal(got && vpar_is_equimatchable(vpar), 0,
                "vpar_is_almost_equimatchable() never also vpar_is_equimatchable()");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_is_equimatchable() states");

\\ See comments in vpar.gp after vpar_is_equimatchable() on the with,without
\\ quantities here.

\\                                   /-- smallest maximal --\
\\                                  with   without without_max   matchnum
test_equimatchable_rel_initial = ['no_with,   0,       0,       'unmatched];

test_equimatchable_rel_validate(rel) =
{
  my(with,without,without_max,matched);
  [with,without,without_max,matched] = rel;

  if(without_max!='no_without_max,
     without_max >= without || error());
}

test_equimatchable_rel_add(p_rel,v_rel) =
{
  \\ print("test_equimatchable_rel_add() "p_rel"  "v_rel);
  test_equimatchable_rel_validate(p_rel);
  test_equimatchable_rel_validate(v_rel);

  my(v_with,v_without,v_without_max,v_matched);
  [v_with,v_without,v_without_max,v_matched] = v_rel;
  my(p_with,p_without,p_without_max,p_matched);
  [p_with,p_without,p_without_max,p_matched] = p_rel;

  my(v_max = min(if(v_with=='no_with,               99, v_with),
                 if(v_without_max=='no_without_max, 99, v_without_max)));
  \\ print(" v_max  "v_max);
  if(v_max==99, error());

  \\ with[p] = min(if(with_exists[p], with[p] + v_max,  #vpar),
  \\               without[p] + without[v] + 1);
  my(ret_with = min(if(p_with=='no_with, 99,  p_with + v_max),
                    p_without + v_without  + 1));
  \\ print(" ret_with  "ret_with);
  if(ret_with==99, error());

  ret_without = p_without + v_max;

  \\ without_max[p] += with[v];
  \\ if(!with_exists[v], without_max_exists[p] = 0);
  my(ret_without_max = if(p_without_max=='no_without_max || v_with=='no_with,
                          'no_without_max,
                          p_without_max + v_with));
  if(ret_without_max==99, error());
  \\ print(" ret_without_max  "ret_without_max);

  my(ret_matched=p_matched,
     matchnum_offset=0);
  if(p_matched=='unmatched && v_matched=='unmatched,
     ret_matched='matched; matchnum_offset=1);

  \\ without[p] += v_max;
  [if(ret_with=='no_with, ret_with,
                          ret_with        - matchnum_offset),
   ret_without                            - matchnum_offset,
   if(ret_without_max=='no_without_max, ret_without_max,
                          ret_without_max - matchnum_offset),
   ret_matched];
}

\\ rel of states 1 to 5
test_equimatchable_rel_list \
  = vector(5,i, test_equimatchable_rel_initial);
{
  for(i=2,#test_equimatchable_rel_list,
     test_equimatchable_rel_list[i]
     = test_equimatchable_rel_add(test_equimatchable_rel_initial,
                                  test_equimatchable_rel_list[i-1]));
}
\\ {
\\   forstep(i=#test_equimatchable_rel_list,1,-1,
\\      my(rel=test_equimatchable_rel_list[i]);
\\      print("n="i"  "test_equimatchable_rel_v_max(rel),"   ",rel));
\\ }
{
  check_equal(test_equimatchable_rel_list,
              [['no_with, 0,        0,        'unmatched],   \\ 1
               [    0,   -1, 'no_without_max, 'matched],     \\ 2
               [    0,    0,        0,        'unmatched],   \\ 3
               [    0,   -1,       -1,        'matched],     \\ 4
               [    0,   -1,        0,        'unmatched]]); \\ 5
  check_equal(#test_equimatchable_rel_list,
              test_is_equimatchable_num_states); \\ in test-helpers.gp
}

\\ Return the state number 1 to 5 for the given rel, or return 0 if not one
\\ of the 5 states in test_equimatchable_rel_list.
test_equimatchable_rel_find(rel) =
{
  for(i=1,#test_equimatchable_rel_list,
     if(test_equimatchable_rel_list[i] == rel, return(i)));
  0;
}

\\ Return the smallest size of a maximal matching at rel, in the form of a
\\ difference matchnum-size.  This is always <= 0.
test_equimatchable_rel_v_max(rel) =
{
  my(with,without,without_max,matched);
  [with,without,without_max,matched] = rel;

  my(ret=min(if(with=='no_with,               99999, with),
             if(without_max=='no_without_max, 99999, without_max)));
  ret <= 0 || error();
  ret;
}
check_equal(apply(test_equimatchable_rel_v_max,test_equimatchable_rel_list), \
            [0,0,0, -1, 0]);

{
  my(l=List([]));
  for(p=1,#test_equimatchable_rel_list,
     for(v=1,#test_equimatchable_rel_list,
        my(new_state=test_equimatchable_rel_find
                       (test_equimatchable_rel_add
                        (test_equimatchable_rel_list[p],
                         test_equimatchable_rel_list[v])));
        listput(l, new_state);
        check_equal(new_state,
                    test_is_equimatchable_table_lookup(p,v),
      "test_equimatchable_rel_add() vs test_is_equimatchable_table_lookup()");
     ));
  check_equal(vpar_INTERNAL_is_equimatchable_table, Vecsmall(l));
}

\\ Return 1 if rel is a non-equimatchable state, meaning that no matter what
\\ parent or further children, this vertex has a maximal < maximum and so
\\ makes its whole tree likewise non-equimatchable.
\\ Condition here as described in the vpar_is_equimatchable() comments.
\\
test_equimatchable_rel_is_non(rel) =
{
  test_equimatchable_rel_validate(rel);

  my(with,without,without_max,matched);
  [with,without,without_max,matched] = rel;

  with !='no_with && with < 0;

  \\ These are non-equimatchable combinations, but the condition on
  \\ "matched" is not needed.
  \\ , allow_without_max=0
  \\ if(with !='no_with && with < 0 && matched=='matched,
  \\   return(1));   \\ not equimatchable
  \\
  \\ if(without_max != 'no_without_max && without_max < 0 && matched=='unmatched,
  \\    if(!allow_without_max,
  \\       error("oops, expected without_max case to be unnecessary"));
  \\    return(1));   \\ not equimatchable
  \\ 0;
}
check_equal(apply(test_equimatchable_rel_is_non,test_equimatchable_rel_list), \
            [0,0,0,0,0], \
            "test_equimatchable_rel_is_non() on five");
check_equal(test_equimatchable_rel_is_non([-1,0,0,'matched]),   1);
check_equal(test_equimatchable_rel_is_non([-1,0,0,'unmatched]), 1);
check_equal(test_equimatchable_rel_is_non(['no_with,0,0,'matched]),   0);
check_equal(test_equimatchable_rel_is_non([0,-1,-1,'matched]),   0);
check_equal(test_equimatchable_rel_is_non([0,-1,-1,'unmatched]), 0);

{
  for(p=1,#test_equimatchable_rel_list,
     for(v=1,#test_equimatchable_rel_list,
        my(rel=test_equimatchable_rel_add
               (test_equimatchable_rel_list[p],
                test_equimatchable_rel_list[v]));

        check_equal(test_equimatchable_rel_find(rel)
                    || test_equimatchable_rel_is_non(rel), 1,
          "test_equimatchable_rel five transitions to five or non")));
}

{
  \\ path-n is the minimum having equimatchable state n
  for(n=2,5,
     my(path=vpar_make_path(n),
        vpar=vpar_unlabelled_first(n));
     until((vpar=vpar_unlabelled_next(vpar)) == 'none,
           my(state=test_equimatchable_get_state(vpar));
           \\ print("n="n" state="state" in "vpar);
           check_equal(state <= n, 1,
                      "vpar_is_equimatchable() no vpar has n < state");
           if(state==n,
              check_equal(vpar_is_isomorphic(vpar,path), 1,
                      "vpar_is_equimatchable() only vpar n == state is path"));
          ));
}


\\-----------------------------------------------------------------------------
progress("vpar_delete_subtree()");

check_equal(vpar_delete_subtree([0,1,2],2), [0]);


\\-----------------------------------------------------------------------------
progress("vpar_relabel_perm()");

check_equal(vpar_relabel_perm([],[]), []);
check_equal(vpar_relabel_perm([0],[1]), [0]);
check_equal(vpar_relabel_perm(vpar_make_path(5),[1,2,3,4,5]), \
            [0,1,2,3,4]);

check_equal(vpar_relabel_perm([0,1,2], [1,2,3]), \
            [0,1,2]);
check_equal(vpar_relabel_perm([0,1,2], [2,1,3]), \
            [2,0,1]);

{
  my(vpar_orig=[0,1,2,3,4,5]);
  vpar_validate(vpar_orig);

  my(a=Vecsmall([2,3,4,5,6,1]),
     b=Vecsmall([1,5,3,4,2,6]));
  check_equal(test_is_perm(a), 1, "perm a");
  check_equal(test_is_perm(b), 1, "perm b");
  check_equal(a*b, Vecsmall([2, 6, 4, 5, 3, 1]));
  check_equal(b*a, Vecsmall([5, 3, 4, 2, 6, 1]));
  check_equal(test_is_perm(a*b), 1);
  check_equal(test_is_perm(b*a), 1);

  my(vpar_perm=vpar_relabel_perm(vpar_orig, a));
  vpar_perm = vpar_relabel_perm(vpar_perm, b);

  check_equal(vpar_relabel_perm(vpar_orig, b*a), vpar_perm,
              "vpar_relabel_perm() b*a");
  check_equal(vpar_relabel_perm(vpar_orig, a*b)
              != vpar_relabel_perm(vpar_orig, b*a), 1);
}

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        perm=vpar_INTERNAL_random_perm(#vpar),
        got=vpar_relabel_perm(vpar,perm));
     check_equal(vpar, vpar_relabel_perm(got,perm^-1));
  );
}

\\ WRONG
\\ test_relabel_perm_by_vecextract(vpar,perm) =
\\ {
\\   vpar=vecextract(vpar,perm);
\\   for(i=1,#vpar, if(vpar[i],vpar[i]=perm[vpar[i]]));
\\   vpar;
\\ }
     \\ check_equal(vpar_relabel_perm(vpar,perm),
     \\             test_relabel_perm_by_vecextract(vpar,perm));


\\-----------------------------------------------------------------------------
progress("vpar_children_of_set()");
\\ vpar_num_children_of_set()

check_equal(vpar_children_of_set([],[]), [], "vpar_children_of_set() empty");
check_equal(vpar_children_of_set([0],[]), []);
check_equal(vpar_children_of_set([0],[1]), []);
check_equal(vpar_children_of_set([0,1],[1]), [2]);
check_equal(vpar_children_of_set([0,1],[2]), []);
check_equal(vpar_children_of_set([0,1],[1,2]), []);

test_children_of_set_by_union(vpar,set) =
{
  my(ret=[]);
  for(i=1,#set,
     ret=setunion(ret,Set(vpar_children_of_vertex(vpar,set[i]))));
  setminus(ret,set);
}
{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        set=vpar_random_set(vpar),
        got=vpar_children_of_set(vpar,set));
     \\ print(vpar"  set "set" got "got);
     check_equal(got, test_children_of_set_by_union(vpar,set));
     check_equal(#got, vpar_num_children_of_set(vpar,set),
                 "vpar_children_of_set() size vs vpar_num_children_of_set()");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_parents_of_set()");

check_equal(vpar_parents_of_set([],[]), [], "vpar_parents_of_set() empty");
check_equal(vpar_parents_of_set([0],[]), []);
check_equal(vpar_parents_of_set([0],[1]), []);
check_equal(vpar_parents_of_set([0,1],[1]), []);
check_equal(vpar_parents_of_set([0,1],[2]), [1]);
check_equal(vpar_parents_of_set([0,1],[1,2]), []);

test_parents_of_set_by_select(vpar,set) =
{
  my(ret=apply(v->vpar[v],set)); \\ the parents
  ret=select(p->p, ret);     \\ the non-zeros
  setminus(Set(ret), set);
}
test_parents_of_set_by_union(vpar,set) =
{
  my(ret=[]);
  for(i=1,#set,
     my(v=set[i], p=vpar[v]);
     if(p, ret=setunion(ret,[p])));
  setminus(ret,set);
}

{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        set=vpar_random_set(vpar),
        got=vpar_parents_of_set(vpar,set));
     \\ print(vpar"  set "set" got "got);
     check_equal(got, test_parents_of_set_by_select(vpar,set));
     check_equal(got, test_parents_of_set_by_union(vpar,set));
     check_equal(#got, vpar_num_parents_of_set(vpar,set),
                 "vpar_parents_of_set() size vs vpar_num_parents_of_set()");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_num_neighbours_of_set()");

check_equal(vpar_num_neighbours_of_set([],[]), 0);
check_equal(vpar_num_neighbours_of_set([0],[]), 0);
check_equal(vpar_num_neighbours_of_set([0],[1]), 0);
check_equal(vpar_num_neighbours_of_set([0,1],[1]), 1);
check_equal(vpar_num_neighbours_of_set([0,1],[1,2]), 0);
check_equal(vpar_num_neighbours_of_set([0,1,2],[2]), 2);
check_equal(vpar_num_neighbours_of_set([0,1,2,3],[2,3]), 2);
check_equal(vpar_num_neighbours_of_set([1],[1]), 0, \
            "vpar_num_neighbours_of_set() self-loop");
check_equal(vpar_num_neighbours_of_set([2,1],[1]), 1, \
            "vpar_num_neighbours_of_set() 2-cycle");


\\-----------------------------------------------------------------------------
progress("vpar_neighbours_of_set()");

test_neighbours_of_set_by_union(vpar,set) =
{
  my(ret=[]);
  for(i=1,#set,
     ret=setunion(ret,Set(vpar_neighbours_of_vertex(vpar,set[i]))));
  setminus(ret,set);
}
{
  for(i=1,#test_sample_forests_and_cyclics,
     my(vpar=test_sample_forests_and_cyclics[i],
        set=vpar_random_set(vpar),
        got=vpar_neighbours_of_set(vpar,set));
     \\ print(vpar"  set "set" got "got);
     check_equal(got, test_neighbours_of_set_by_union(vpar,set));
     check_equal(#got, vpar_num_neighbours_of_set(vpar,set),
                 "vpar_neighbours_of_set() size vs vpar_num_neighbours_of_set()");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_path_vertices()");

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     #vpar>=1 || next();
     for(rep=1,4,
        my(u=vpar_random_vertex(vpar),
           v=vpar_random_vertex(vpar),
           c=vpar_common_ancestor(vpar,u,v),
           got=vpar_INTERNAL_common_ancestor_and_distances(vpar,u,v));
        \\ print(vpar"  "u" "v" roots "vpar_root_of_vertex(vpar,u)" "vpar_root_of_vertex(vpar,u)" common "c);
        if(c,
           check_equal(got,
                       [c, vpar_path_length(vpar,u,c),
                           vpar_path_length(vpar,v,c)],
                       "vpar_INTERNAL_common_ancestor_and_distances()"),
           check_equal(got[1], c)));
  );
}



\\-----------------------------------------------------------------------------
progress("vpar_path_vertices()");

check_equal(vpar_path_vertices([0,1,2,3,4],4,2), [4,3,2]);

\\        10
\\       / | \
\\     3   4   9          Knuth 2.3.4.4
\\    / \     / \
\\   1   2   5   8
\\              / \
\\             6   7
{
  my(vpar=[3,3,10,10,9,8,8,9,10, 0]);
  check_equal(vpar_path_vertices(vpar,2,9),[2,3,10,9]);
  check_equal(vpar_path_vertices(vpar,8,4),[8,9,10,4]);
  check_equal(vpar_path_vertices(vpar,10,10),[10]);
  check_equal(vpar_path_vertices(vpar,1,1),[1]);
  check_equal(vpar_path_vertices(vpar,3,10),[3,10]);
}

test_path_vertices_by_common_ancestor(vpar,u,v) =
{
  my(c=vpar_common_ancestor(vpar,u,v));
  if(!c, return([]));
  my(u_list=List([]),
     v_list=List([]));
  while(u!=c, listput(u_list,u); u=vpar[u]);
  while(v!=c, listput(v_list,v); v=vpar[v]);
  concat([Vec(u_list), c, Vecrev(v_list)]);
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     #vpar>=1 || next();
     my(u=vpar_random_vertex(vpar),
        v=vpar_random_vertex(vpar));
     check_equal(vpar_path_vertices(vpar,u,v),
                 test_path_vertices_by_common_ancestor(vpar,u,v),
       "vpar_path_vertices() vs test_path_vertices_by_common_ancestor()");
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_concat_forest()");

check_equal(vpar_concat_forest([],[]), []);
check_equal(vpar_concat_forest([0],[]), [0]);
check_equal(vpar_concat_forest([],[0]), [0]);
check_equal(vpar_concat_forest([0,1,2],[0,1,2,3]), [0,1,2, 0,4,5,6]);
check_equal(vpar_concat_forest(Vecsmall([1]), Vecsmall([1])), \
            Vecsmall([1,2]), \
            "vpar_concat_forest() Vecsmall and cycles");


\\-----------------------------------------------------------------------------
progress("vpar_is_siblingpos_vector()");

check_equal(vpar_is_siblingpos_vector([],0,-1), 1);
check_equal(vpar_is_siblingpos_vector([],0,0), 1);
check_equal(vpar_is_siblingpos_vector([],0,1), 1);

check_equal(vpar_is_siblingpos_vector([0],0,-1), 1, \
            "vpar_is_siblingpos_vector() singleton root=1");
check_equal(vpar_is_siblingpos_vector([0],0,0), 1);
check_equal(vpar_is_siblingpos_vector([0],0,1), 1);

{
  \\      1
  \\      |
  \\      3
  \\     / \
  \\    2   4
  my(vpar=[0, 3, 1, 3]);
  vpar_validate(vpar);
  check_equal(vpar_roots(vpar), [1]);
  my(siblingpos=[0, 0, 0, 1]);
  check_equal(vpar_siblingpos_vector(vpar), siblingpos);
  check_equal(vpar_is_siblingpos_vector(siblingpos), 1);
  check_equal(vpar_is_siblingpos_vector(siblingpos,,-1), 1);
}
{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i]);
     for(down=0,1,
        my(siblingpos=vpar_siblingpos_vector(vpar,down));
        \\ print(vpar"  "siblingpos);
        check_equal(vpar_is_siblingpos_vector(siblingpos,down),1,
                    Str("vpar_is_siblingpos_vector() of tree, down="down));
        check_equal(vpar_is_siblingpos_vector(siblingpos,down,1),1,
                    Str("vpar_is_siblingpos_vector() of tree as forest, down="down)));

     if(#vpar, vpar=vpar_reroot(vpar,1));
     for(down=0,1,
        my(siblingpos=vpar_siblingpos_vector(vpar,down));
        \\ print(vpar"  "siblingpos);
        check_equal(vpar_is_siblingpos_vector(siblingpos,down,-1), 1,
                    Str("vpar_is_siblingpos_vector() of tree root=1, down="down));
        check_equal(vpar_is_siblingpos_vector(siblingpos,down), 1,
                    Str("vpar_is_siblingpos_vector() of tree root=1 as tree, down="down));
        check_equal(vpar_is_siblingpos_vector(siblingpos,down,1), 1,
                    Str("vpar_is_siblingpos_vector() of tree root=1 as forest, down="down))));
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     for(down=0,1,
        my(siblingpos=vpar_siblingpos_vector(vpar,down));
        check_equal(vpar_is_siblingpos_vector(siblingpos,down,1),1,
                    "vpar_is_siblingpos_vector() of forest")));
}

{
  \\ l = list of all siblingpos vectors which occur from iterating.
  \\ All must pass vpar_is_siblingpos_vector().

  for(down=0,1,
     for(forest=0,1,
        for(n=0, test_limit_trees_search - forest,
           my(l=List([]));
           forvec(vec=vpar_forvec_from(n,forest), \\ trees or forests
                 my(vpar=vpar_from_upascend(vec),
                    siblingpos=vpar_siblingpos_vector(vpar,down),
                    pos=setsearch(l,siblingpos,1));
                 \\ print(vpar" degs "siblingpos" pos "pos);
                 if(pos,listinsert(l,siblingpos,pos));
                 check_equal(vpar_is_siblingpos_vector(siblingpos,down,forest), 1,
                 "vpar_is_siblingpos_vector() of vpar_forvec_from()"));


           check_equal(vpar_siblingpos_vectors_count(n,forest),
                       #l,
                       Str("vpar_siblingpos_vectors_count() forest="forest));

           forvec(siblingpos=vector(n,i,[-1,n+1]),
                 \\ print(siblingpos"  "setsearch(l,siblingpos));
                 check_equal(vpar_is_siblingpos_vector(siblingpos,down,forest),
                             setsearch(l,siblingpos)!=0,
                             "vpar_is_siblingpos_vector() vs setsearch"));
        )));
}


\\-----------------------------------------------------------------------------
progress("vpar_siblingpos_vectors_count()");

\\ tableau is a vector of vectors like
\\    [ [1,2,3],
\\      [4,5],
\\      [6]
\\    ]
\\ Return 1 if it is a Young tableau, meaning vector lengths decreasing, and
\\ each row entries increasing and each column entries increasing.
\\
test_is_tableau(tableau) =
{
  test_vector_is_decreasing(apply(length,tableau)) || return(0);
  for(i=1,#tableau, test_vector_is_increasing(tableau[i]) || return(0));
  for(j=1,#tableau[1],
     for(i=1,#tableau-1,
        if(j>#tableau[i+1], break());
        if(tableau[i][j] > tableau[i+1][j], return(0))));
  1;
}
check_equal(test_is_tableau([[1,2,3],[4],[5]]), 1);

\\ shape is a tableau shape.
\\ perm is a vector of entries to put into the rows of that shape.
\\ The number of entries is n=vecsum(shape).
\\ Must have #perm == n.
test_tableau_shape_fill(shape,perm) =
{
  my(upto=1);
  for(i=1,#shape,
     my(t=upto+shape[i]);
     shape[i] = perm[upto..t-1];
     upto=t);
  shape;
}

\\ shape is a vector of integers.
\\ Return 1 if it is a Young tableau shape, meaning all >=1 and non-increasing.
test_is_tableau_shape(shape) = \
  vecmin(shape)>=1 && test_vector_is_decreasing(shape);

\\ Return the number of Young tableaux of "shape".
\\ Same as vpar_INTERNAL_tableau_shape_count() but by counting how many
\\ valid tableau are made by permuting 1..n=vecsum(shape) in the shape.
test_tableau_shape_count_by_perms(shape) =
{
  my(n=vecsum(shape),
     ret=0);
  for(i=1,n!,
     ret += test_is_tableau(test_tableau_shape_fill(shape,numtoperm(n,i))));
  ret;
}
{
  my(shape=[4,1,1,1],
     want = 7! /7 /3/2 /3/2);
  check_equal(test_is_tableau_shape(shape), 1);
  check_equal(test_tableau_shape_count_by_perms(shape), want);
  check_equal(vpar_INTERNAL_tableau_shape_count(shape), want);
}
{
  for(n=1,6,
     forpart(p=n,
             my(shape=Vecrev(p));
             check_equal(test_is_tableau_shape(shape), 1,
                        "test_is_tableau_shape() of partition");
             check_equal(vpar_INTERNAL_tableau_shape_count(shape),
                         test_tableau_shape_count_by_perms(shape));
            ));
}


\\-----------------------------------------------------------------------------
progress("vpar_is_subtree_heights_vector()");

check_equal(vpar_is_subtree_heights_vector([]),1);
check_equal(vpar_is_subtree_heights_vector([0]),1);

check_equal(vpar_is_subtree_heights_vector([1,0]), 1, \
            "vpar_is_subtree_heights_vector() path-2");
check_equal(vpar_is_subtree_heights_vector([1,0],1), 1, \
            "vpar_is_subtree_heights_vector() path-2, forest");
check_equal(vpar_is_subtree_heights_vector([1,0],-1), 1, \
            "vpar_is_subtree_heights_vector() path-2, root=1");

check_equal(vpar_is_subtree_heights_vector([0,1]), 1, \
            "vpar_is_subtree_heights_vector() path-2 reverse");
check_equal(vpar_is_subtree_heights_vector([0,1],1), 1, \
            "vpar_is_subtree_heights_vector() path-2 reverse, forest");
check_equal(vpar_is_subtree_heights_vector([0,1],-1), 0, \
            "vpar_is_subtree_heights_vector() path-2 reverse, root=1");

{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i],
        subtree_heights=vpar_subtree_heights_vector(vpar));
     check_equal(vpar_is_subtree_heights_vector(subtree_heights), 1,
                 "vpar_is_subtree_heights_vector() of tree");
     check_equal(vpar_is_subtree_heights_vector(subtree_heights,1), 1,
                 "vpar_is_subtree_heights_vector() of tree, as forest");

     if(#vpar, vpar=vpar_reroot(vpar,1));
     subtree_heights=vpar_subtree_heights_vector(vpar);
     check_equal(vpar_is_subtree_heights_vector(subtree_heights,-1), 1,
                 "vpar_is_subtree_heights_vector() of root=1");
     check_equal(vpar_is_subtree_heights_vector(subtree_heights), 1,
                 "vpar_is_subtree_heights_vector() of root=1, as tree");
     check_equal(vpar_is_subtree_heights_vector(subtree_heights,1), 1,
                 "vpar_is_subtree_heights_vector() of root=1, as forest");
);
}
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        subtree_heights=vpar_subtree_heights_vector(vpar));
     check_equal(vpar_is_subtree_heights_vector(subtree_heights,1), 1,
                 "vpar_is_subtree_heights_vector() of forest"));
}


\\-----------------------------------------------------------------------------
progress("vpar_make_random()");

{
  \\ trees root=1
  for(n=0,8,
     for(rep=1,n+1,
        my(vpar=vpar_make_random(n,-1));
        vpar_validate(vpar);
        check_equal(vpar_is_tree(vpar), 1, "vpar_make_random() is tree");
        check_equal(vpar_root(vpar), if(#vpar,1,0))));
}
{
  \\ trees
  for(n=0,8,
     for(rep=1,n+1,
        my(vpar=vpar_make_random(n));
        vpar_validate(vpar);
        check_equal(vpar_is_tree(vpar), 1, "vpar_make_random() is tree");
     ));
}
{
  \\ forests
  for(n=0,8,
     for(rep=1,n+1,
        my(vpar=vpar_make_random(n,1));
        vpar_validate(vpar)));
}

\\ vpar_make_random() with fake random() generates all trees

\\ Run vpar=func(forest) to see it generates all possible valid trees from
\\ test_random() sequences.
check_make_random_all(name,func) =
{
  for(n=0,5,
     for(forest=-1,1,
        my(l=List([]),  \\ list of all vpar created
           vec_params=if(forest==-1,vector(max(0,n-2),i,[0,n-1]),
                         forest==0, vector(max(0,n-1),i,[0,n-1]),
                         forest==1, vector(max(0,n-1),i,[0,n]),
                         error()));
        if(option_verbose, print(name" n="n" forest="forest
                                 " vec_params "vec_params));
        forvec(vec=vec_params,
              test_random_list=List(vec);
              test_random_limit_list=List(vector(#vec,i,n+(forest>0)));
              my(vpar=func(n,forest));
              check_equal(#test_random_list, 0, "test_random_list consumed");
              check_equal(#test_random_limit_list, 0,
                          "test_random_limit_list consumed");
              vpar_validate(vpar);
              if(forest<=0,
                 check_equal(vpar_is_tree(vpar), 1,
                             "vpar_make_random() is tree"));
              listput(l,vpar));
        check_equal(#Set(l),#l)));
}

\\ same as vpar_make_random() but calling test_random() instead of random()
test_vpar_make_random=eval(test_random_to_test_random(vpar_make_random));
if(option_verbose, print(test_vpar_make_random));

check_make_random_all("vpar_make_random",test_vpar_make_random);

\\ vpar_from_upascend() in the manner of Cho et al, used with successive
\\ random numbers.  v is random for i=1 to n-1 inclusive.  prev is the
\\ previous random number.
\\ Same as vpar_make_random() but upto runs downwards.
test_make_random_by_upascend(n,forest=0) =
{
  my(vpar = vector(n,v,-1),
     upto = n+1,  \\ ready for pre-decrement
     prev = 0,
     v    = 1);
  for(i=1,n,
     \\ random() 1 to n for tree, 0 to n for forest
     vpar[if(i==-forest, 1,
             i==n || !(v=test_random(n+(forest>0))+(forest<=0)) || vpar[v]>=0,
             while(vpar[upto--]>=0, ); upto,
             v)] = prev;
     prev=v);
  vpar;
}
check_make_random_all("test_make_random_by_upascend", \
                      test_make_random_by_upascend);

\\ per comments in vpar.gp
test_make_random_by_upqueue(n,forest=0) =
{
  vpar_from_upqueue
    (vector(n,i, if(i==n-1 && forest==-1, 1,
                    i==n, 0,
                    test_random(n+(forest>0))+(forest<=0))));
}
check_make_random_all("test_make_random_by_upqueue", \
                      test_make_random_by_upqueue);


\\-----------------------------------------------------------------------------
progress("vpar_minimal_domsets()");

check_equal(vpar_minimal_domsets([]), [[]]);

\\ order unspecified in the return, but this is how they come out now
check_equal(vpar_minimal_domsets([0,1]), [[1], [2]]);
check_equal(vpar_minimal_domsets([0,1,2]), [[1,3], [2]]);
check_equal(vpar_minimal_domsets([0,1,2,3]), [[1,4], [1,3], [2,3], [2,4]]);
check_equal(vpar_minimal_domsets([0,1,1]), [[1], [2,3]], \
            "vpar_minimal_domsets() star-3");
check_equal(vpar_minimal_domsets([0,1,1,1]), [[1], [2,3,4]], \
            "vpar_minimal_domsets() star-4");

\\ 1 <-- 3 <-- 2 <-- 4
\\        \--5
\\ at v=1 [2, 3] is without_dom_notsole
\\ at v=3 [2,3] is with_min_req  needing 3 for 5
check_equal(vpar_minimal_domsets([0,3,1,2,3]), \
            [[1,2,5], [1,4,5], [3,4], [2,3]], \
            "vpar_minimal_domsets() 5");

\\ 3 <-- 4 <-- 2 <-- 1     path-7 middle
\\  \--5 <-- 6
check_equal(vpar_minimal_domsets([2,4,0,3,3,5]), \
            [[1,3,5], [1,3,6], [2,3,6], [2,4,6], [1,4,5], [1,4,6], [2,5]], \
            "vpar_minimal_domsets() path-7 middle");

{
  \\       /--3 <-- 4 <-- 5 <-- 6
  \\ 1 <- 2
  \\       \--7 <-- 8 <-- 9 <-- 10
  \\
  \\ [2, 3, 6, 8, 9]  has 8 not minimal
  \\ into p=2   8,9 is without_dom_sole, 7 dom again is not minimal
  \\ at v=2  3,6 is with_min_req
  \\         2,3,6 is with_notmin_notreq
  \\         with_notmin_notreq extends by without_dom_notsole
  \\
  my(vpar=[0,1,2,3,4,5,2,7,8,9]);
  vpar_validate(vpar);
  my(got=vecsort(vpar_minimal_domsets(vpar)),
     want=vecsort(test_minimal_domsets_by_flags(vpar)),
     want_count=vpar_minimal_domsets_count(vpar));
  \\ print("got  has extra "setminus(Set(got),Set(want)));
  \\ print("want has extra "setminus(Set(want),Set(got)));
  check_equal(#want, want_count);
  check_equal(#got, want_count, "vpar_minimal_domsets() 10");
  check_equal(got,want, "vpar_minimal_domsets() 10");
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        want_count=vpar_minimal_domsets_count(vpar));
     if(want_count > test_limit_sets, next());

     my(got=vpar_minimal_domsets(vpar));
     check_equal(#got, want_count,
              "vpar_minimal_domsets() count vs vpar_minimal_domsets_count()");
     for(i=1,#got,
        check_equal(setisset(got[i]), 1);
        check_equal(vpar_is_minimal_domset(vpar,got[i]), 1));

     check_equal(vecsort(got), vecsort(test_minimal_domsets_by_flags(vpar)));

     check_equal(Pol(0,'y) + sum(i=1,#got, 'y^#got[i]),
                 vpar_minimal_dompoly(vpar,'y));
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_totdomnum() state table");

\\ present, dominated, have_child
test_totdomnum_rel_initial = [0, 0, 0];

test_totdomnum_rel_validate(rel) =
{
  \\ print("test_totdomnum_rel_validate() "rel);
  my(present,dominated,have_child);
  [present,dominated,have_child] = rel;
  present==0 || present==1 || error();
  dominated==0 || dominated==1 || error();
  have_child==0 || have_child==1 || error();
}
test_totdomnum_rel_validate(test_totdomnum_rel_initial);

\\ Return a vector [ new_rel, increment ] which is the new rel for p_rel
\\ accumulating v_rel under it, and increment = 0 or 1 to add to totdomnum.
\\
test_totdomnum_rel_add_and_inc(p_rel,v_rel) =
{
  \\ print("test_totdomnum_rel_add() p="p_rel"  v="v_rel);
  test_totdomnum_rel_validate(p_rel);
  test_totdomnum_rel_validate(v_rel);

  my(p_present,p_dominated,p_have_child);
  [p_present,p_dominated,p_have_child] = p_rel;
  my(v_present,v_dominated,v_have_child);
  [v_present,v_dominated,v_have_child] = v_rel;


  my(new_inc=0,
     new_present    = p_present,
     new_dominated  = p_dominated,
     new_have_child = p_have_child);

  if(!v_dominated,
     new_inc = !p_present;  \\ take parent to dominate child
     new_present = 1);

  new_have_child = 1;

  if(v_present,
     new_dominated = 1);  \\ child dominates parent

  [ [new_present,new_dominated,new_have_child], new_inc ];
}
test_totdomnum_rel_add(p_rel,v_rel) = \
  test_totdomnum_rel_add_and_inc(p_rel,v_rel)[1];

\\ rel of states 1 to 5
test_totdomnum_rel_list = {[
  test_totdomnum_rel_initial,   \\ 1 = singleton
  [1, 0, 1],                    \\ 2 = path-2
  [1, 1, 1],                    \\ 3 = path-3
  [0, 1, 1],                    \\ 4 = path-4
  [0, 0, 1]                     \\ 5 = path-5
]};
{
  if(0,
     forstep(i=#test_totdomnum_rel_list,1,-1,
            my(rel=test_totdomnum_rel_list[i]);
            print("state="i"  ",rel));
     );
}

\\ Return the state number 1 to 9 for the given rel, or return 0 if not one
\\ of the 9 states in test_equimatchable_rel_list.
test_totdomnum_rel_find(rel) =
{
  for(i=1,#test_totdomnum_rel_list,
     if(test_totdomnum_rel_list[i] == rel, return(i)));
  0;
}

{
  for(p=1,#test_totdomnum_rel_list,
     for(v=1,#test_totdomnum_rel_list,
        my(rel=test_totdomnum_rel_add
               (test_totdomnum_rel_list[p],
                test_totdomnum_rel_list[v]));
        \\ print("p="p" v="v" to "rel);
        check_equal(test_totdomnum_rel_find(rel) != 0, 1,
        "test_totdomnum_rel_list transitions")));
}

{
  my(l=List([]));
  for(p=1,#test_totdomnum_rel_list,
     for(v=1,#test_totdomnum_rel_list,
        my(rel_and_inc=test_totdomnum_rel_add_and_inc
                       (test_totdomnum_rel_list[p],
                        test_totdomnum_rel_list[v]),
           new_rel = rel_and_inc[1],
           new_inc = rel_and_inc[2],
           new_state = test_totdomnum_rel_find(new_rel));
        if(! new_state,
           error("oops, no state for p=",p," v=",v));
        listput(l, 2*(new_state-1) + rel_and_inc[2])));

  if(0,
     for(p=1,#test_totdomnum_rel_list,
        for(v=1,#test_totdomnum_rel_list,
           printf("%2d,",l[5*p+v-5]));
        print()));

  check_equal(vpar_INTERNAL_totdomnum_table, Vecsmall(l));
}


\\ Return the increment 0 or 1 to add to totdomnum if reach state rel at the
\\ root of a tree.
\\ If rel is not already dominated then take a child to dominate it.
\\ If have_child==0 then cannot dominate, and return -1.
\\
test_totdomnum_rel_top_inc(rel) =
{
  my(present,dominated,have_child);
  [present,dominated,have_child] = rel;

  if(!dominated,      \\ no parent and not already dominated
     if(have_child,
        return(1),      \\ take child to dominate parent
        return(-1)));   \\ no child, isolated vertex, cannot totdom
  0;
}

{
  \\ The increment coded into the first 5 entries of
  \\ vpar_INTERNAL_totdomnum_table corresponds to the increment of
  \\ test_totdomnum_rel_top_inc(), and state 1 (a singleton) is the only
  \\ state where -1 for no total dominating sets at all.

  for(v=1,#test_totdomnum_rel_list,
     my(top=test_totdomnum_rel_top_inc(test_totdomnum_rel_list[v]));
     check_equal(v==1, top<0);
     if(top>=0,
        check_equal(bitand(vpar_INTERNAL_totdomnum_table[v],1),
                    top>0));
  );
}
{
  check_equal(apply(t->t%2, Vec(vpar_INTERNAL_totdomnum_table)),
              [1, 1, 0, 0, 1,
               0, 0, 0, 0, 0,
               0, 0, 0, 0, 0,
               1, 1, 0, 0, 1,
               1, 1, 0, 0, 1]);
}
{
  check_equal(apply(t->t\2, Vec(vpar_INTERNAL_totdomnum_table)),
              [1, 2, 3, 4, 1,
               1, 2, 2, 1, 1,
               2, 2, 2, 2, 2,
               2, 2, 3, 3, 2,
               1, 2, 3, 4, 1]);
}

\\ Return the totdomnum state 1 to 5 at the root of vpar.
\\ vpar must be a tree, not a forest.
\\
test_totdomnum_get_state(vpar) =
{
  my(seq=vpar_INTERNAL_seq_upwards(vpar),
     state = vectorsmall(#vpar));
  for(i=1,#seq,
     my(v=seq[i], p=vpar[v]);
     if(p,
        state[p] = vpar_INTERNAL_totdomnum_table[5*state[p]+state[v]+1] >> 1,
        return(state[v]+1)));
  0;
}
{
  \\ state n is as at the end of a path-n
  for(n=1,#test_totdomnum_rel_list,
     check_equal(test_totdomnum_get_state(vpar_make_path(n)), n,
                 Str("test_totdomnum_get_state() of path "n)));
}


\\-----------------------------------------------------------------------------
progress("vpar_maximal_matchings_count()");

test_maximal_matchings_count_by_flags(vpar) =
{
  my(count=0);
  forvec(flags=vector(#vpar,i,[0,1]),
        count += vpar_is_maximal_matching_flags(vpar,flags));
  count;
}

test_maximal_matchsets_by_flags(vpar) =
{
  my(l=List([]));
  forvec(flags=vector(#vpar,i,[0,1]),
        if(vpar_is_maximal_matching_flags(vpar,flags),
           listput(l,test_flags_to_Set(flags))));
  Vec(l);
}
check_equal(test_maximal_matchsets_by_flags([]), [ [] ]);


check_equal(vpar_maximal_matchings_count([]), 1, \
            "vpar_maximal_matchings_count() empty");
check_equal(vpar_maximal_matchings_count([0]), 1, \
            "vpar_maximal_matchings_count() singleton");
check_equal(vpar_maximal_matchings_count([0,1]), 1, \
            "vpar_maximal_matchings_count() path-2");
check_equal(vpar_maximal_matchings_count([0,1,2]), 2, \
            "vpar_maximal_matchings_count() path-3");

{
  \\      1        24      <-
  \\      |        34
  \\      4        34 52   <-
  \\     / \       41
  \\    2   3      41 52   <-
  \\    |          52
  \\    5
  my(vpar=[0, 4, 4, 1, 2]);
  check_equal(vecsort(vpar_matchsets(vpar)),
              [[], [2], [3], [3,5], [4], [4,5], [5]]);
  check_equal(test_maximal_matchings_count_by_flags(vpar), 3);
  check_equal(vpar_maximal_matchings_count(vpar), 3,
              "vpar_maximal_matchings_count() 5");
}

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     if(#vpar <= test_limit_flags_search,
        check_equal(vpar_maximal_matchings_count(vpar),
                    test_maximal_matchings_count_by_flags(vpar),
                    "vpar_maximal_matchings_count() vs by flags"));

     if(#vpar <= test_limit_sets,
        check_equal(vpar_maximal_matchings_count(vpar),
                    #test_maximal_matchsets_by_flags(vpar),
                    "vpar_maximal_matchings_count() vs sets"));
  );
}

{
  \\ Number of maximal matchings in path, per
  \\     Tomislav Doslic, Ivana Zubac, "Counting Maximal Matchings In Linear
  \\     Polymers", Ars Mathematica Contemporanea, volume 11, 2016, pages
  \\     255-276, http://amc-journal.eu
  for(n=0,12,
     my(vpar=vpar_make_path(n));
     check_equal(vpar_maximal_matchings_count(vpar),
                 test_Padovan(n+5),
                 "vpar_maximal_matchings_count() path is Padovan"));
}


\\-----------------------------------------------------------------------------
progress("vpar_relabel_random()");

vpar_validate(vpar_relabel_random(vpar_make_path(0)));
vpar_validate(vpar_relabel_random(vpar_make_path(1)));
vpar_validate(vpar_relabel_random(vpar_make_path(2)));
vpar_validate(vpar_relabel_random(vpar_make_path(5)));

\\ Return 1 if all values in vector v are equal.
vector_is_all_same(v) = #v==0 || #vecsort(v,,8)==1;

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     my(rel=vpar_relabel_random(vpar));
     vpar_validate(rel));
}

{
  \\ As described in the addhelp of vpar_relabel_random(), various different
  \\ vpar arise from relabelling.  Each different vpar arises from the same
  \\ number of perms, that number being the number of perms in the
  \\ automorphism group.

  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        n=#vpar);
     n <= test_limit_perms_search-1 || next();

     my(l=List([]));    \\ relabelled vpars
     my(c=List([]));    \\ count how many perms make each
     my(automorphisms_count=0);
     for(i=1,n!,
        my(perm=numtoperm(#vpar,i));
        my(vpar_relabelled=vpar_relabel_perm(vpar,perm));
        automorphisms_count += (vpar==vpar_relabelled);

        my(pos=setsearch(l,vpar_relabelled));
        if(pos,
           c[pos]++;  \\ existing entry
           ,
           pos=setsearch(l,vpar_relabelled,1); \\ new entry
           listinsert(l,vpar_relabelled,pos);
           listinsert(c,1,pos)));

     check_equal(vecsum(Vec(c)), n!, "total perm relabellings");
     check_equal(#l >= 1, 1);
     check_equal(#c==#l, 1);
     for(i=1,#c,
        check_equal(c[i], automorphisms_count, "each c = automorphisms_count"));
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_is_eccentricities_vector()");

check_equal(vpar_is_eccentricities_vector([]),1, \
            "vpar_is_eccentricities_vector() empty tree");
check_equal(vpar_is_eccentricities_vector([0]),1, \
            "vpar_is_eccentricities_vector() singleton");

check_equal(vpar_eccentricities_vector([0,1]), [1,1], \
            "vpar_eccentricities_vector() path-2");
check_equal(vpar_is_eccentricities_vector([1,1]),1, \
            "vpar_is_eccentricities_vector() path-2");
{
  my(vpar=vpar_make_path(4),
     eccentricities=vpar_eccentricities_vector(vpar));
  check_equal(vpar_eccentricities_vector(vpar), [3,2,2,3],
            "vpar_is_eccentricities_vector() 3,2,2,3");
}
check_equal(vpar_is_eccentricities_vector([2, 1,1, 2]), 0);

\\ GP-Test  my(diameter=3); ceil(diameter/2) == 2
check_equal(vpar_is_eccentricities_vector([3,3, 2, 3,3]), 0);

\\ GP-Test  my(diameter=3); ceil(diameter/2) == 2
check_equal(vpar_is_eccentricities_vector([3, 3, 2, 3, 4]), 0, \
            "vpar_is_eccentricities_vector() 3,3,2,3,4");

{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i],
        eccentricities=vpar_eccentricities_vector(vpar));
     check_equal(vpar_is_eccentricities_vector(eccentricities),1,
                 Str("vpar_is_eccentricities_vector() of tree "
                     eccentricities));
  );
}
{
  for(forest=0,0,
     for(n=1, test_limit_trees_search - forest,
        my(l=List([]));
        forvec(vec=vpar_forvec_from(n,forest), \\ trees or forests
              my(vpar=vpar_from_upascend(vec),
                 eccentricities=vpar_eccentricities_vector(vpar),
                 pos=setsearch(l,eccentricities,1));
              if(pos,
                 listinsert(l,eccentricities,pos);
                 check_equal(vpar_is_eccentricities_vector(eccentricities), 1,
                             "vpar_is_eccentricities_vector() of ecc vector")));
        l=Vec(l);
        check_equal(setisset(l),1,
                    "accumulated eccentricities is a set");

        \\ among arbitrary vecs, only those from trees should pass
        forvec(vec=vector(n,i,[-1,n]),
              \\ print(" vec "vec" "n);
              check_equal(vpar_is_eccentricities_vector(vec),
                          setsearch(l,vec)!=0,
                          Str("vpar_is_eccentricities_vector() of "vec)));
     ));
}

\\-----------------------------------------------------------------------------
progress("vpar_seq_blob()");
\\ vpar_from_blob()

check_equal(vpar_seq_blob([]),Vecsmall([]), "vpar_seq_blob() empty");
check_equal(vpar_seq_blob([0,0]),Vecsmall([2,1]));

\\ list is a vector of integers.
\\ Return a permutation (Vecsmall) of 1..n in which each element of list is
\\ rotated down each element of list.
test_make_perm_cycle_down(n,list) =
{
  my(perm=test_identity_perm(n));
  for(i=1,#list, perm[list[i]]=list[if(i==#list,1,i+1)]);
  perm;
}
check_equal(test_make_perm_cycle_down(8,[2,5,6]), \
            Vecsmall([1, 5, 3,4, 6, 2, 7,8]));

\\ Return 1 if v > ancestors in vpar, so v the maximum of it and its ancestors.
test_vertex_gt_ancestors(vpar,v) = \
  v == vpar_minmax_ancestor_of_vertex(vpar,v,max);

\\ Return a vector of those vertex numbers which have v>ancestors,
\\ so all those satisfying test_vertex_gt_ancestors(vpar,v).
test_vertices_gt_ancestors_list(vpar) = \
  test_select_vertices(vpar, test_vertex_gt_ancestors);

\\ Return a vector of the vertices from v to its root, inclusive of v and
\\ the root.
test_path_vertices_to_root(vpar,v) =
{
  my(l=List([]));
  while(v && #l<=#vpar, listput(l,v);v=vpar[v]);
  Vec(l);
}

\\ Return blob code for vpar.
\\ vpar must be a tree with root=1.
\\ Per "BLOB ENCODING" algorithm of Micikevicius, Caminiti, Deo.
\\ test_vec_blob_by_MCD() returns the full modified vpar array.
\\ test_code_blob_by_MCD() returns their code form which skips the fixed
\\ new_vpar[1]=root=1 and new_vpar[#vpar]=0 fixed elements.
\\
test_vec_blob_by_MCD(vpar) =
{
  vpar_roots(vpar)==[1] || error();
  my(last=vpar[#vpar],
     new_vpar=vpar);
  forstep(v=#vpar-1,2,-1,
         vecmax(test_path_vertices_to_root(vpar,v))
         == vecmax(test_path_vertices_to_root(new_vpar,v)) || error();

         if(vecmax(test_path_vertices_to_root(vpar,v))==v,
            my(t=new_vpar[v]); new_vpar[v]=last; last=t));
  new_vpar;
}
test_code_blob_by_MCD(vpar) =
{
  test_vec_blob_by_MCD(vpar)[2..#vpar-1];
}

\\ Picciotto thesis section 4.1
test_code_blob_by_surgery(vpar) =
{
  my(n=#vpar,
     blob=[n],
     blob_parent=vpar[n],
     code=[],
     i=1);
  while(1,
        my(path=test_path_vertices_to_root(vpar,n-i));
        path=Set(path);
        if(#setintersect(path,blob),
           code=concat([vpar[n-i]],code);
           vpar[n-i]=0;
           blob=setunion(blob,[n-i]);
           ,
           code=concat([blob_parent],code);
           blob_parent=vpar[n-i];
           vpar[n-i]=0;
           blob=setunion(blob,[n-i]);
           );
        i++;
        if(i==n,break()));
  code;
}

{
  \\ Kreweras and Moszkowski, "Tree Codes That Preserve Increases and Degree
  \\ Sequences", Discrete Mathematics, volume 87, 1991, pages 291-296.
  \\ Figure 2 example.
  \\
  \\             7
  \\          /      \
  \\       14            4
  \\        |         /  |  \
  \\        5        3   6    11
  \\     / / \ \        / \    |
  \\    2 9  12 13     1   8  10
  \\
  my(vpar=[6,5,4,7,14,4,0,6,5,11,4,5,5,7]);
  vpar_validate(vpar);

  my(max_ancestors=vpar_minmax_ancestors_vector(vpar,max));
  check_equal(max_ancestors, [7,14,7,7,14,7,7,8,14,11,11,14,14,14]);

  my(L = [1,2,3,4,5,6,9,10,12,13],
     Ldash = [7,8,11,14]);
  check_equal(select(v->max_ancestors[v]>v, [1..#vpar]),
              L, "Kreweras and Moszkowski example L");
  check_equal(select(v->max_ancestors[v]==v, [1..#vpar]),
              Ldash, "Kreweras and Moszkowski example Ldash");

  \\                n = 1         6  7   8  9 10  11  12,13  14
  my(want_seq=Vecsmall([1,2,3,4,5,6, 8, 11, 9,10, 14, 12,13,  7]));
  check_equal(test_is_perm(want_seq),1,
              "Kreweras and Moszkowski example seq is perm");
  check_equal(vpar_seq_blob(vpar),want_seq,
              "Kreweras and Moszkowski seq is blob");

  \\                          *            *
  \\        n = 1          6  7  8  9 10  11  12,13   omit 14
  my(want_code=[6,5,4,7,14,4, 6, 4, 5,11,  7,  5, 5]);
  check_equal(vecextract(vpar,want_seq)[^#vpar], want_code,
              "Kreweras and Moszkowski example vecextract");
}

{
  \\ Caminiti thesis section 5.2 figure 5.2 blob
  \\
  \\              0
  \\            /   \
  \\          5       9
  \\          |        \
  \\          2         11
  \\        /   \
  \\       7    10
  \\              \
  \\               4 -- 6 -- 3 -- 1 -- 8
  \\
  \\ path 1 to root
  \\ v > max(pathset(1 to v))

  \\                      0 1 2 3  4 5 6 7 8 9 10 11        v0
  my(vpar=apply(p->p+1, [-1,3,5,6,10,0,4,2,1,0, 2, 9])); \\ parent
  check_equal(#vpar,12);
  check_equal(test_vertex_gt_ancestors(vpar,12), 1,
              "12 (11 0-based) is stable");

  my(want_stable0=vecsort([7,10,5,0,9,11]),
     want_stable=apply(n->n+1, want_stable0));
  check_equal(test_vertices_gt_ancestors_list(vpar), want_stable,
              "Caminiti example stables");

  \\ Correction to sample numbers shown in Caminiti's thesis.
  \\ Believe second "5" of code should be 2, as per the digraph shown in
  \\ section 5.4.1 "Blob Transformation" page 71.  Position 5 parent of
  \\ vertex 5 is 2.       |
  \\                      v
  my(want_code0=[3,5,6,10,2,4,0,1,2, 9]);
  \\             1 2 3  4 5 6 7 8 9 10
  check_equal(#want_code0, 10);

  \\            v          v     v      v  v  v   vertices to rotate down
  \\            0,1,2,3,4, 5, 6, 7, 8,  9,10,11
  my(want_seq0=[5,1,2,3,4, 7, 6, 9, 8, 10,11,0 ],
     want_seq=Vecsmall(apply(n->n+1, want_seq0)));

  my(seq=vpar_seq_blob(vpar));
  check_equal(test_make_perm_cycle_down(#vpar,want_stable),want_seq,
              "Caminiti example seq vs stable cycle");
  check_equal(seq,want_seq, "Caminiti example seq");

  my(want_code=apply(n->n+1, want_code0));
  check_equal(test_code_blob_by_MCD(vpar),want_code,
              "Caminiti example, code by MCD");

  my(want_vec=concat([1,want_code,0]));
  my(vec=vecextract(vpar,seq));
  check_equal(vec,want_vec,"vec Caminiti");
}

{
  \\ Paulius Micikevicius, Saverio Caminiti, Narsingh Deo, "Linear-Time
  \\ Algorithms for Encoding Trees as Sequences of Node Labels",
  \\ Congressus Numerantium, volume 183, 2006, pages 65-75.
  \\ (URLs with vpar_seq_subtree_height() code.)
  \\
  \\         1 *
  \\         |                              n       n
  \\         6 *        =>      after       6 *     1 *   self-loops
  \\         |                  blob            n
  \\         8 *                                8 *
  \\       /   \                               /     root
  \\      7     9 *                           7       9 *
  \\      |     |                             |       |
  \\      2     5     1,6,8,9                 2       5
  \\      |     |                             |       |
  \\      3     4                             3       4

  \\       1 2 3 4 5 6 7 8 9    \\ vertex
  my(vpar=[0,7,2,5,9,1,8,6,8]); \\ parent

  my(seq=vpar_seq_blob(vpar));

  \\ shown in section 3 codes as "Happy" but is Blob, per part (d) of
  \\ figure 1
  \\           1  2 3 4 5 6 7 8  9    \\ vertex
  my(want_vec=[0, 7,2,5,9,6,8,8, 8]); \\ digraph parent after blob encoding
  check_equal(test_vec_blob_by_MCD(vpar),want_vec);

  \\ Micikevicius, Deo, Caminiti show digraph 9->8 and 1->nothing which is
  \\ per the changes made by their "BLOB ENCODING" algorithm and which don't
  \\ matter since last 9 and root 1 are not part of the n-1 blob code.
  \\ For the vecextract here complete the rotations so 1 parent self and
  \\ 9 is root.
  want_vec[1]=1;
  want_vec[9]=0;

  my(vec=vecextract(vpar,seq));
  check_equal(vec,want_vec);
}

{
  \\ Picciotto thesis example 3.4.1
  \\ 2 -> 1 -> 3 -> 0
  my(vpar = [3,1,0],
     seq = vpar_seq_blob(vpar),
     vec=vecextract(vpar,seq));
  check_equal(vec,[3,1,0]);

  check_equal(test_code_blob_by_surgery(vpar),[3,1]);
}
{
  \\ Picciotto thesis example in section 4.1
  \\      5
  \\     / \     with 5 in place of root 0
  \\    3   4
  \\   / \
  \\  1   2
  my(vpar = [3,3,5,5,0],
     seq = vpar_seq_blob(vpar),
     vec=vecextract(vpar,seq));
  check_equal(seq,Vecsmall([1,2,3,4,5]));
  check_equal(vec,[3,3,5,5, 0]);

  check_equal(test_code_blob_by_surgery(vpar),[3,3,5,5]);
}
{
  \\ example in Picciotto thesis section 9.1 page 115 (first)
  \\        0
  \\        |
  \\        4
  \\      / | \
  \\    5   2  7
  \\       / \
  \\      6   3
  \\      |
  \\      1
  my(vpar=[6,4,2,0,4,2,4]);  \\ root=0
  vpar_validate(vpar);
  check_equal(test_code_blob_by_surgery(vpar),
              [6,4,2,4,2,4],
              "test_code_blob_by_surgery");
}

check_equal(vpar_from_blob([]),[], "vpar_from_blob() empty");
check_equal(vpar_from_blob([0,0,0,0]),[0,0,0,0]);

{
  for(i=1,#test_sample_trees,
     my(vpar=test_sample_trees[i]);
     #vpar>=2 || next();
     check_equal(vecextract(vpar,vpar_seq_blob(vpar))[^#vpar],
                 test_code_blob_by_surgery(vpar)));
}

\\ vpar_seq_blob() children in order per addhelp and comments
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i],
        seq=vpar_seq_blob(vpar));
     check_equal(test_seq_is_ordered_except_roots(vpar,seq), 1,
                 "vpar_seq_blob() is test_seq_is_ordered_except_roots()");
     if(vpar_is_tree(vpar),
        check_equal(vpar_seq_is_ordered(vpar,seq), 1,
                 "vpar_seq_blob() on tree is vpar_seq_is_ordered()"));
  );
}


\\-----------------------------------------------------------------------------
progress("vpar_set_complement()");

check_equal(vpar_set_complement([],[]), []);
check_equal(vpar_set_complement([0,1,2],[1,3]), [2]);

{
  for(n=0,10,
     my(vpar=vpar_make_path(n));
     for(rep=1,min(n,4),
        my(set=vpar_random_set(vpar),
           comp=vpar_set_complement(vpar,set));
        \\ print("n="n" set "set" comp "comp);
        check_equal(setisset(set),1);
        check_equal(setisset(comp),1);
        check_equal(setintersect(set,comp),[],
                    "vpar_set_complement() setintersect");
        check_equal(setunion(set,comp), [1..n],
                    "vpar_set_complement() and setunion is everything");
     ));
}


\\-----------------------------------------------------------------------------
progress("vpar_random_set()");

{
  my(vpar=[0,1,2],
     set=vpar_random_set(vpar));
  check_equal(0 <= #set && #set <= #vpar, 1);
  for(i=1,#set,
     check_equal(1 <= set[i] && set[i] <= #vpar, 1));
}


\\-----------------------------------------------------------------------------
progress("vpar_random_flags()");

{
  my(vpar=[0,1,2],
     flags=vpar_random_flags(vpar));
  check_equal(#flags,#vpar);
  for(i=1,#flags,
     check_equal(flags[i]==0 || flags[i]==1, 1));
}


\\-----------------------------------------------------------------------------
progress("vpar_children_vector()");

check_equal(vpar_children_vector([]), []);
check_equal(vpar_children_vector([0]), [ [] ]);
check_equal(vpar_children_vector([0,1]), [ [2], [] ]);

test_children_vector_by_select(vpar) = \
  vector(#vpar,p, Vec(select(sp->sp==p, vpar, 1)));
{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     check_equal(vpar_children_vector(vpar),
                 test_children_vector_by_select(vpar)));
}


\\-----------------------------------------------------------------------------
progress("vpar_seq_is_ordered()");

check_equal(vpar_seq_is_ordered([],[]), 1, "vpar_seq_is_ordered() empty");
check_equal(vpar_seq_is_ordered([0,1,2],[1,2,3]), 1);
check_equal(vpar_seq_is_ordered([0,1,1],[1,2,3]), 1);
check_equal(vpar_seq_is_ordered([0,1,1],[1,3,2]), 0);
check_equal(vpar_seq_is_ordered([0,1,1],[3,1,2]), 0);
check_equal(vpar_seq_is_ordered([0,1,1],[2,1,3]), 1);

check_equal(vpar_seq_is_ordered([0,1,2],[1,2,3],1), 1, "vpar_seq_is_ordered() path-3 down");
check_equal(vpar_seq_is_ordered([0,1,1],[1,2,3],1), 0);
check_equal(vpar_seq_is_ordered([0,1,1],[1,3,2],1), 1);
check_equal(vpar_seq_is_ordered([0,1,1],[3,1,2],1), 1);
check_equal(vpar_seq_is_ordered([0,1,1],[2,1,3],1), 0);

{
  for(i=1,#test_sample_forests,
     my(vpar=test_sample_forests[i]);
     check_equal(vpar_seq_is_ordered(vpar, vpar_seq_preorder(vpar)), 1,
                "vpar_seq_is_ordered() of preorder");
     check_equal(vpar_seq_is_ordered(vpar, vpar_seq_postorder(vpar)), 1);
     );
}


\\-----------------------------------------------------------------------------
progress("Sequences Reversibility and Uniqueness");

{
  my(forest=1,
     n_max= test_limit_trees_search - forest,
     debug=0,
     elems=
     [
      ["vpar_seq_upascend",  vpar_seq_upascend,  vpar_from_upascend,
       'no_seqfrom, 'upwards, 'not_ordered, 0
      ],
      ["upascend,ahead=1",
       (vpar)->vpar_seq_upascend(vpar,1),
       (vec)->vpar_from_upascend(vec,1),
       'no_seqfrom, 'upwards, 'not_ordered, 0
      ],

      ["vpar_seq_successive_paths",
       vpar_seq_successive_paths,
       vpar_from_successive_paths,
       'no_seqfrom,
       'bothwards, 'not_ordered, 0
      ],
      ["vpar_seq_successive_paths,down=1",
       vpar->vpar_seq_successive_paths(vpar,1),
       vec->vpar_from_successive_paths(vec,1),
       'no_seqfrom,
       'downwards, 'not_ordered, 0
      ],

      ["vpar_seq_childless_paths",
       vpar_seq_childless_paths,
       vpar_from_childless_paths,
       'no_seqfrom,
       'bothwards, 'not_ordered, 0
      ],
      ["vpar_seq_childless_paths,down=1",
       vpar->vpar_seq_childless_paths(vpar,1),
       vec->vpar_from_childless_paths(vec,1),
       'no_seqfrom,
       'downwards, 'not_ordered, 0
      ],

      ["vpar_seq_blob",
       vpar_seq_blob,
       vpar_from_blob,
       vec->vpar_INTERNAL_seq_rotate_down(vpar_minmax_ancestors_vector(vec,max)),
       'bothwards, 'ordered_except_roots, 0
      ],

      ["vpar_seq_num_children",
       vpar_seq_num_children,
       vpar_from_num_children,
       vpar_seq_num_children,  \\ is its own seqfrom
       'bothwards, 'not_ordered, 0
      ],
      ["vpar_seq_num_children,down=1",
       vpar->vpar_seq_num_children(vpar,1),
       vec->vpar_from_num_children(vec,1),
       vec->vpar_seq_num_children(vec,1),  \\ is its own seqfrom
       'bothwards, 'not_ordered, 0
      ],

      ["vpar_seq_upqueue",
       vpar_seq_upqueue,
       vpar_from_upqueue,
       vpar_INTERNAL_seqfrom_upqueue,
       'upwards, 'not_ordered, 0
      ],

      ["vpar_seq_subtree_height",
       vpar_seq_subtree_height,
       vpar_from_subtree_height,
       vpar_INTERNAL_seqfrom_subtree_height,
       'upwards, 'not_ordered, 0
      ],
      ["vpar_seq_subtree_height,down=1",
       vpar->vpar_seq_subtree_height(vpar,1),
       vec->vpar_from_subtree_height(vec,1),
       vec->vpar_INTERNAL_seqfrom_subtree_height(vec,1),
       'downwards, 'not_ordered, 0
      ],

      ["vpar_seq_subtree_size",
       vpar_seq_subtree_size,
       vpar_from_subtree_size,
       vpar_INTERNAL_seqfrom_subtree_size,
       'upwards, 'not_ordered, 0
      ],
      ["vpar_seq_subtree_size,down=1",
       vpar->vpar_seq_subtree_size(vpar,1),
       vpar->vpar_from_subtree_size(vpar,1),
       vpar->vpar_INTERNAL_seqfrom_subtree_size(vpar,1),
       'downwards, 'not_ordered, 0
      ]

     ]);
  for(i=1,#elems,
     length(elems[i])==7 || error();
     my(func_name   =elems[i][1],
        seq_func    =elems[i][2],
        from_func   =elems[i][3],
        seqfrom_func=elems[i][4],
        direction   =elems[i][5],
        ordered     =elems[i][6],
        min_size    =elems[i][7]);
     if(debug, print(func_name));

     setsearch(Set(['upwards,'downwards,'bothwards]), direction)
       || error();
     setsearch(Set(['ordered,'ordered_except_roots,'not_ordered]), ordered)
       || error();

     for(i=1,#test_sample_forests,
        my(vpar=test_sample_forests[i]);
        if(#vpar<min_size,next());
        if(debug,print(func_name," vpar "vpar));

        my(seq=seq_func(vpar));
        if(debug,print("  seq "seq));
        check_equal(test_is_perm(seq),1, Str(func_name" seq is perm"));

        my(vec=vecextract(vpar,seq),
           vpar_again = from_func(vec));
        check_equal(vpar_again,vpar, Str(func_name" vpar again"));
        );

     for(n=min_size,n_max,
        if(debug, print("n="n" "func_name));

        my(count=0,
           count_downwards=0,
           count_upwards=0,
           count_not_downwards=0,
           count_not_upwards=0,
           count_ordered=0,
           count_not_ordered=0);
        forvec(vec=vpar_forvec_from(n,forest), \\ forests
              my(vpar=vpar_from_upascend(vec));
              count++;
              if(debug,print(func_name," vpar "vpar));

              my(seq=seq_func(vpar));
              if(debug,print("  seq "seq));

              my(is_upwards=vpar_seq_is_upwards(vpar,seq),
                 is_downwards=vpar_seq_is_downwards(vpar,seq));
              count_upwards   += is_upwards;
              count_downwards += is_downwards;
              count_not_upwards   += ! is_upwards;
              count_not_downwards += ! is_downwards;

              if(direction=='upwards,
                 check_equal(is_upwards,1,
                             Str(func_name" should be upwards")));
              if(direction=='downwards,
                 check_equal(is_downwards,1,
                             Str(func_name" should be downwards")));

              my(is_ordered = vpar_seq_is_ordered(vpar,seq));
              count_ordered     += is_ordered;
              count_not_ordered += ! is_ordered;
              if(ordered=='ordered,
                 check_equal(is_ordered, 1,
                             Str(func_name" should be sibling ordered")));

              my(is_ordered_except_roots
                 = test_seq_is_ordered_except_roots(vpar,seq));
              count_ordered     += is_ordered;
              count_not_ordered += ! is_ordered;
              if(ordered=='ordered_except_roots,
                 check_equal(is_ordered_except_roots, 1,
                    Str(func_name" should be sibling ordered except roots")));

              my(vec=vecextract(vpar,seq));

              my(vpar_again=from_func(vec));
              vpar_validate(vpar_again);
              check_equal(vpar_again,vpar,
                          Str(func_name" vpar_again by from_func"));

              if(seqfrom_func!='no_seqfrom,
                 my(seq_again=seqfrom_func(vec));
                 check_equal(seq_again,seq,
                             Str(func_name" seq_again by seqfrom_func")));

              if(n>=2 && func_name=="vpar_seq_blob",
                 \\ for tree vec[1]==1 iff vertex 1 is root
                 \\ for forest vec[1]==0or1 iff vertex 1 is root
                 my(roots=vpar_roots(vpar));
                 if(#roots==1,
                    check_equal(vec[1]==1, roots==[1],
                                "blob tree vec[1]==1 is root 1"),
                    check_equal(vec[1]<=1, vpar[1]==0,
                                "blob forest vec[1]==0,1 is root 1")));
              );

        check_equal(count,(n+1)^(n-1));
        if(direction=='bothwards,
           check_equal(count_upwards && count_downwards, 1,
                          Str(func_name" should be bothwards"));
           if(n>=3,
              check_equal(count_not_upwards && count_not_downwards, 1,
                          Str(func_name" n="n" bothwards should have not up and not down")));
          );

        if(ordered=='not_ordered && n>=4,
           check_equal(count_not_ordered >= 1, 1,
                       Str(func_name" n="n" not_ordered should have some unordered")));
     ));
}


\\-----------------------------------------------------------------------------
progress("vpar_make_num_children()");

{
  \\   2
  \\   |
  \\   1
  \\   |
  \\   3
  \\ must v=3 be childless, then v=1,2 both num_children=1 could be either
  \\ way around but the specified order is by ascending v, so 1 above 3

  my(num_children=[1,1,0],
     vpar=vpar_make_num_children(num_children));
  check_equal(vpar, [2,0,1]);
}

{
  for(n=2,5,
     my(comps=test_compositions(n-1,[0,n-1],n),
        count_upwards=0,
        count_downwards=0,
        count_nonewards=0,
        count_num_children_increasing=0,
        count_num_children_decreasing=0);
     for(i=1,#comps,
        my(num_children=comps[i]);
        check_equal(vpar_is_num_children_vector(num_children), 1);
        \\ print("num_children "num_children);

        my(vpar=vpar_make_num_children(num_children));
        \\ print(" vpar "vpar);
        vpar_validate(vpar);
        check_equal(vpar_num_children_vector(vpar), Vec(num_children),
                    "vpar_make_num_children() gives num_children");

        my(is_upwards  =vpar_is_upwards(vpar),
           is_downwards=vpar_is_downwards(vpar));
        count_upwards   += is_upwards;
        count_downwards += is_downwards;
        count_nonewards += (!is_upwards && !is_downwards);
        \\ print(" is_upwards "is_upwards" is_downwards "is_downwards);

        my(num_children_is_increasing=test_vector_is_increasing(num_children),
           num_children_is_decreasing=test_vector_is_decreasing(num_children));
        count_num_children_increasing += num_children_is_increasing;
        count_num_children_decreasing += num_children_is_decreasing;

        \\ some non-monotonic num_children can also give vpar_is_upwards()
        if(num_children_is_increasing,
           check_equal(is_upwards, 1, "vpar_make_num_children() is_upwards"));

        \\ if(num_children_is_decreasing,
        \\    check_equal(is_downwards, 1, "vpar_make_num_children() is_downwards"));
        );

     check_equal(count_upwards  >=1, 1, "vpar_make_num_children() count_upwards");
     check_equal(count_downwards>=1, 1, "vpar_make_num_children() count_downwards");
     check_equal(count_nonewards>=1, if(n>=3,1,0), "count_nonewards");

     check_equal(count_num_children_increasing >=1, 1);
     check_equal(count_num_children_decreasing >=1, 1);
  );
}


\\==============================================================================
progress("No globals left over");
\\ test-symbols.txt is all symbols appearing in vpar.gp.

check_equal(vpar,'vpar);
{
  my(test_symbols=readstr("test-symbols.txt"));
  check_equal(#test_symbols>0,1);
  for(test_i=1,#test_symbols,
     my(test_symbol=test_symbols[test_i]);
     check_equal(eval(test_symbol),
                 eval(Str("'"test_symbol)),
                 Str("leftover global variable: "test_symbol)));
}
\\ {
\\   my(test_symbols=readstr("test-funcs.txt"));
\\   for(test_i=1,#test_symbols,
\\      my(test_symbol=test_symbols[test_i]);
\\      my(value=eval(test_symbol));
\\      check_equal(value == eval(Str("'"test_symbol))
\\                  || type(value) == "t_CLOSURE", 1,
\\                  Str("leftover non-function: "test_symbol)));
\\ }

\\-----------------------------------------------------------------------------
{
  printf("total %d tests in %.1f seconds\n",
         check_count,
         if(option_progress,progress_total_time,gettime()) /1000);
}

\\-----------------------------------------------------------------------------
